//Menu points
var PgMenuPointColumnSize = function(placement, menu, force_size, auto_col, gh) {

    PgElementPageViewMenuPoint.call(this, placement, menu);

    this.options = gh.options;
    this.gh = gh;

    this.value = 4;
    this.prevClass = null;
    this.unitSize = 30;
    this.placement = placement;

    this.menu = menu;

    this.editSize = force_size || this.options.size_for_all;
    this.colSize = 0;
    this.forceSize = force_size;

    this.autoColPoint = auto_col;
    this.hidePointInDrag = null;

    this.factor = placement == 'right' ? 1 : -1;
    this.tooltip = true;

    this.update();
}

extendObject(PgMenuPointColumnSize, PgElementPageViewMenuPoint)

PgMenuPointColumnSize.prototype.getUnitSize = function () {
    var parent_w = this.menu.$el.parent().innerWidth();
    return this.snapX = Math.max(10, (parent_w / this.options.num_columns) * this.menu.pageView.zoom);
}

PgMenuPointColumnSize.prototype.getColSize = function () {
    return Math.max(1, Math.min(this.options.num_columns, Math.round(this.menu.$el.width() / this.unitSize)));
}

PgMenuPointColumnSize.prototype.update = function() {
    var sizes = this.options.sizes;

    this.unitSize = this.getUnitSize();
    this.colSize = this.getColSize();

    this.editSize = this.forceSize || this.options.size_for_all;

    var map = this.gh.getColumnSizes(this.menu.pgel);

    for(var i = 0; i < sizes.length; i++) {
        if(!this.gh.isSizeVisible(this.menu.pageView, sizes[i])) break;
        if(this.forceSize && sizes[i] != this.forceSize) continue;

        if(sizes[i] in map) {
            this.editSize = sizes[i];
            this.colSize = map[sizes[i]];
            this.prevClass = this.gh.getClassForSize(this.editSize, this.options.col_prefix_class, this.colSize);

            if(this.colSize == '' || this.colSize == 'auto') {
                this.colSize = Math.round(this.menu.$el.width() / this.unitSize);
            }
            this.colSize = parseInt(this.colSize);
        }
    }

    var sizeUp = this.editSize.toUpperCase();

    if(!(this.editSize in map)) {
        this.setIsNotSet(true);
        this.setText(this.editSize);
        this.setTooltipText('Column size for ' + sizeUp + ' is not set. Drag the handle to set it. Click to open the column size control.');
    } else {
        this.setIsNotSet(false);

        this.setText(this.editSize + this.gh.getColumnSizeAsShortString(map[this.editSize], ' '));
        this.setTooltipText('Drag the handle to resize the column for ' + sizeUp + ' and larger devices. Click to open the column size control.');
    }

    if(this.autoColPoint) {
        if(this.autoColPoint.editSize == this.editSize) {
            this.autoColPoint.hide();
        } else {
            //Handle situation where size for smaller is not set and size for current is also not set.
            //Only the current size point will handle this for both points
            if(this.autoColPoint.editSize == this.options.size_for_all && !(this.options.size_for_all in map)) {
                this.autoColPoint.hide();
            } else {
                this.autoColPoint.show();
            }
        }
    }

}


PgMenuPointColumnSize.prototype.add = function(menu, $container) {
    PgElementPageViewMenuPoint.prototype.add.call( this, menu, $container );
    this.update();
}

PgMenuPointColumnSize.prototype.onReposition = function() {
    //update text in box
    if(!this.in_drag) {
        this.update();
    }
}

PgMenuPointColumnSize.prototype.onDragStart = function(x, y) {
    this.setLabel(this.colSize);
    this.setText(this.editSize);
    if(this.hidePointInDrag) this.hidePointInDrag.hide();
}

PgMenuPointColumnSize.prototype.onDrag = function(x, y) {
    var _this = this;
    var span = Math.min(this.options.num_columns, Math.max(1, Math.round(_this.colSize + _this.factor * x / _this.unitSize)));
    var cls = this.gh.getClassForSize(_this.editSize, this.options.col_prefix_class, span);

    pinegrow.getCollectionForElement(this.menu.pgel).forEachElement(function(pgel, i) {
        if(cls != _this.prevClass) {
            pgel.removeClass(_this.prevClass);
            pgel.addClass(cls);
            _this.prevClass = cls;
            _this.menu.reposition();

            if(_this.autoColPoint) {
                //check if auto point became the same as me
                if(_this.autoColPoint.editSize == _this.editSize) {
                    //we don't need it any more
                    _this.autoColPoint.hide();
                }
            }
            if(i == 0) {
                _this.setLabel(span);
            }
        }
    })
    return {x: (span - this.colSize) * this.unitSize, y:y};
}

PgMenuPointColumnSize.prototype.onClick = function() {
    this.gh.options.onClickOnColumnPoint.call(this);
}

PgMenuPointColumnSize.prototype.onDragStop = function(x, y) {
    this.setLabel();

    pinegrow.changeManager.didChange(pinegrow.getCollectionForElement(this.menu.pgel), 'Change column size');
    pinegrow.selectedElements.reselect();
}