class PgDmDesignSkillPlainHTMLOptions extends PgDmDesignSkill {
    constructor() {
        super('plainhtml', 'Design options');

        this.addControl('reset', {
            name: 'Normalize styles',
            type: 'checkbox',
            value: 'true',
            dm_no_lock: true,
            helptext: `<p>Add normalize.css to make the default styling the same in all major browsers.</p>`
        })
    }

    populateDefaultParams() {
        this.setParam('reset', 'true');
    }

    onSurpriseMe(done) {
        done && done();
        /*
        if(!this.isLocked('gradients')) {
            this.setParam('gradients', Math.random() < 0.3 ? '1' : '');
        }
        this.surpriseMeForRangeControls(done);
         */
    }

    onGetCSS(part) {
        var _this = this;
        var code = '';

        switch(part) {

            case 'init':
                if(this.getParam('reset') === 'true') {
                    code += `
/*! normalize.css v8.0.1 | MIT License | github.com/necolas/normalize.css */html{line-height:1.15;-webkit-text-size-adjust:100%}body{margin:0}main{display:block}h1{font-size:2em;margin:.67em 0}hr{box-sizing:content-box;height:0;overflow:visible}pre{font-family:monospace,monospace;font-size:1em}a{background-color:transparent}abbr[title]{border-bottom:none;text-decoration:underline;text-decoration:underline dotted}b,strong{font-weight:bolder}code,kbd,samp{font-family:monospace,monospace;font-size:1em}small{font-size:80%}sub,sup{font-size:75%;line-height:0;position:relative;vertical-align:baseline}sub{bottom:-.25em}sup{top:-.5em}img{border-style:none}button,input,optgroup,select,textarea{font-family:inherit;font-size:100%;line-height:1.15;margin:0}button,input{overflow:visible}button,select{text-transform:none}[type=button],[type=reset],[type=submit],button{-webkit-appearance:button}[type=button]::-moz-focus-inner,[type=reset]::-moz-focus-inner,[type=submit]::-moz-focus-inner,button::-moz-focus-inner{border-style:none;padding:0}[type=button]:-moz-focusring,[type=reset]:-moz-focusring,[type=submit]:-moz-focusring,button:-moz-focusring{outline:1px dotted ButtonText}fieldset{padding:.35em .75em .625em}legend{box-sizing:border-box;color:inherit;display:table;max-width:100%;padding:0;white-space:normal}progress{vertical-align:baseline}textarea{overflow:auto}[type=checkbox],[type=radio]{box-sizing:border-box;padding:0}[type=number]::-webkit-inner-spin-button,[type=number]::-webkit-outer-spin-button{height:auto}[type=search]{-webkit-appearance:textfield;outline-offset:-2px}[type=search]::-webkit-search-decoration{-webkit-appearance:none}::-webkit-file-upload-button{-webkit-appearance:button;font:inherit}details{display:block}summary{display:list-item}template{display:none}[hidden]{display:none}
                    
`;
                }

                code += `*, *::before, *::after {
    box-sizing: border-box;
}`
                break;

            case 'rules':
                break;
        }
        return code;
    }
}

class PgDmDesignSkillColorsForPlainHTML extends PgDmDesignSkillColors {
    constructor(colors) {
        super('plainhtml.colors', colors, function() {
            var css = `\n:root {
`;
            colors.forEachColor(function(key, val) {
                css += `    --${key}: ${val};
`;
            })
            css += '}';
            return css;
        }, 'rules');
    }

    populateDefaultParams(keep_existing) {

        if(!keep_existing) this.colors.clear();

        this.colors.add('primary', '#0d6efd', false, null, true);
        this.colors.add('secondary', '#6c757d', false, null, true);
        this.colors.add('color3', '#0dcaf0', false, null, true);
        this.colors.add('color4', '#198754', false, null, true);
        this.colors.add('white', 'white', true, null, true);
        this.colors.add('light', '#f8f9fa', true, null, true);
        this.colors.add('dark', '#212529', true, null, true);

        super.populateDefaultParams();
    }
}


class PgDmDesignProviderForPlainHTML extends PgDmDesignProvider {

    constructor(f) {
        super('plainhtml', 'Plain HTML Design');
        this.framework = f;
        this.prefix = 'plainhtml';

        var _this = this;

        this.on_engine_ready_funcs = []

        this.colors = new PgDmDesignColors();
        this.colors.can_edit = true;
        this.colors.mode = 'css';

        this.addSkills();
    }

    addSkills() {
        var _this = this;

        var colors_skill = new PgDmDesignSkillColorsForPlainHTML(this.colors);
        this.addSkill( colors_skill );

        this.fonts_skill_main = new PgDmDesignSkillFonts(false, colors_skill, 'css');
        this.addSkill( this.fonts_skill_main );
        this.fonts_skill_headings = new PgDmDesignSkillFonts(true, colors_skill, 'css');
        this.addSkill( this.fonts_skill_headings );

        var bck_skill = new PgDmDesignSkillBackground(colors_skill);
        bck_skill.onGetBackgroundColorCSS = function(col) {
            return !col ? '' : `
body {
    background-color: var(--${col});            
}
`
        }
        this.addSkill( bck_skill );
        this.addSkill( new PgDmDesignSkillPlainHTMLOptions() );
    }

    canEditColors() {
        return true;
    }

    getActiveFonts() {
        var r = []
        this.fonts_skill_main.getActiveFonts(r);
        this.fonts_skill_headings.getActiveFonts(r);
        return r;
    }

    getEngine(done) {
        var _this = this;
        if(!this.engine) {
            this.engine = new PgDmEngineForPlainHTML(function(engine) {
                _this.on_engine_ready_funcs.forEach(function(func) {
                    func(engine);
                })
                _this.on_engine_ready_funcs = [];
                done && done(engine);
            }, this.framework);
        } else if(this.engine.isReady()) {
            done(this.engine);
        } else {
            _this.on_engine_ready_funcs.push(done);
        }
    }

    destroyEngine() {
        this.engine && this.engine.destroy();
        this.engine = null;
    }
}


class PgDmEngineForPlainHTML extends PgDmEngine {

    constructor(on_engine_ready, provider) {
        super(on_engine_ready);

        var _this = this;

        this.delayedTasks = new PgDelayedTasks();
        this.design_provider = provider;
        this.is_ready = false;

        this.done_list = [];

        this.busy = false;
        this.update_on_done = false;

        var p = pinegrow.getCurrentProject();
        this.custom_cs_url = null;

        if(isApp()) {
            var custom_file = p.getAbsolutePath('css/theme.css');
            this.custom_cs_url = crsaMakeUrlFromFile(custom_file);
        } else {
            this.custom_cs_url = p.getAbsoluteUrl('css/theme.css');
        }

        var outcs = this.outcs = pinegrow.getStylesheetByUrl(this.custom_cs_url);

        if(!outcs) {
            if(!isApp()) {
                var project = pinegrow.getCurrentProject();
                if(!project.hasContentForUrl('css/theme.css')) {
                    project.setContentOfUrl('css/theme.css', '');
                }
                pinegrow.refreshCurrentProjectDelayed(true);
            }
            outcs = this.outcs = PgCSSStylesheet();
            outcs.setUrl(this.custom_cs_url);
            pinegrow.stylesheets.addCrsaStylesheet(outcs);

            outcs.load(this.custom_cs_url, function () {
                _this.is_ready = true;
                outcs.setIgnore(true);
                on_engine_ready && on_engine_ready(_this);
            }, false, true, true);
        } else {
            _this.is_ready = true;
            on_engine_ready && on_engine_ready(_this);
        }

        if(!outcs.isLocked()) {
            outcs.setLocked(true, 'The stylesheet is generated by the Design panel.');
        }
    }

    activateOnPage(page, done) {
        var _this = this;

        if(page.sourceNode) {
            var replacePgel = null;
            if (!page.stylesheets.has(this.outcs)) {
                page.sourceNode.withoutEvents(function () {
                    page.sourceNode.find('link').forEach(function (pgel) {
                        if (pgel.getAttribute('rel') === 'stylesheet') {
                            var href = pgel.getAttribute('href');
                            if (href && href.match(/normalize\.css/i)) {
                                replacePgel = pgel;
                            }
                        }
                    })
                    page.stylesheets.attachTo(_this.outcs, null, function () {
                        pinegrow.changeManager.didChange(null, 'Add theme stylesheet', null, page);
                        $('body').trigger('crsa-stylesheets-changed');
                        page.refresh(function() {
                            //trigger insight
                            pinegrow.insight.inspectPageOrPgel(page, null);
                        });
                        done && done();
                    }, replacePgel, true /* first css if no replace pgel */);
                })
                page.setPageChanged(true);
            } else {
                done && done();
            }
        } else {
            done && done();
        }
    }

    getOutputCS(page) {
        return this.outcs;
    }

    update(done) {
        done && this.done_list.push(done);
        this.delayedTasks.run('update', this.updateTask.bind(this), 10);
    }

    updateTask() {
        var _this = this;

        var page = pinegrow.getSelectedPage();
        if(!page) {
            _this.done_list.forEach(function(f) {
                f();
            })
            _this.done_list = [];
            return;
        }

        if(this.busy) {
            this.update_on_done = true;
            return;
        }

        var o = {
            css: '',
            cs_url: this.custom_cs_url
        }

        pinegrow.dispatchEvent('on_dm_get_css', null, 'import', o);
        pinegrow.dispatchEvent('on_dm_get_css', null, 'init', o);
        pinegrow.dispatchEvent('on_dm_get_css', null, 'rules', o);
        var source_css = o.css;

        var p = new CrsaProfile();

        this.busy = true;

        var cs = _this.getOutputCS(page);
        if(cs) {
            cs.ignore_imports_in_apply = true;
            cs.setSource(o.css, function() {
                p.show('Applied to page');

                _this.busy = false;
                if(_this.update_on_done) {
                    _this.update();
                } else {
                    _this.done_list.forEach(function(f) {
                        f();
                    })
                    _this.done_list = [];
                }
                _this.update_on_done = false;

                pinegrow.dispatchEvent('on_css_stylesheet_changed', null, cs);
                pinegrow.selectedElements.repositionMenusDelayed();
                pinegrow.refreshPageChangedStatusForAllPages(true);
            });

            var fonts = pgDmStore.design.fonts.getFontUrls();
            _this.addFontsToPageViews(cs, fonts);
        }
    }

    saveCSS() {
        var page = pinegrow.getSelectedPage();
        var cs = this.getOutputCS(page);
        if(cs && (cs.localFile || !isApp())) {
            cs.save(cs.localFile);
            cs.setChanged(false);
        }
    }

    destroy() {
        this.design_provider = null;
    }
}