$(function () {

    var init = function() {
        var live_version = 2;

        console.log(`PinegrowLive ${live_version} loaded.`);

        pinegrow.addEventHandler('on_key_pressed', function(page, e, status) {
            if (status.key == '.' && status.ctrl) {
                status.done = true;
                pinegrow.showAlert('Resize to 1920x1080?', 'Sure?', 'No', 'Yes', null, function() {
                    var win = require('nw.gui').Window.get();
                    var w = 1440;
                    win.width = w;//1080 * 2250/1500;//1920;
                    win.height = w * (1080 / 1920) + (2160 - 2112)/2;
                })
                //2880 x 1620
                return;
            }
        })

        var current_version = parseFloat(crsaGetVersion());

        if(parseFloat(crsaGetVersion()) >= 5.98 && parseFloat(crsaGetVersion()) < 6) {

            //BrowserSync
            var PgBrowserSync = function() {
                change_count = 0;

                var bs_url = pinegrow.getSetting('browsersync-url', 'http://localhost:3000');

                var reload = [];

                var tasks = new PgDelayedTasks();

                var enabled = pinegrow.getSetting('browsersync-enabled', '0') === '1';

                var last_error_msg = 0;

                var reloadDelayed = function() {
                    tasks.run('reload', async function() {
                        if(reload.length) {
                            var url = bs_url + '/__browser_sync__?method=reload';
                            reload.forEach(function(u) {
                                url += `&args=${u}`;
                            })
                            //console.log('BrowserSync: ' + url)
                            reload = [];

                            try {
                                await pgFetchRemoteContent(url);
                            } catch(err) {
                                console.log(`BrowserSync is not running on ${bs_url}`);

                                var t = new Date().getTime();
                                if(t - last_error_msg > 60000) {
                                    pinegrow.showQuickError(`BrowserSync is not running on ${bs_url}. Start it or switch off the sync in the top toolbar with <i class="icon icon-27"></i>.`);
                                    last_error_msg = t;
                                }
                            }
                        }
                    }, 250);
                }

                var changedUrl = function(url) {
                    if(!enabled) return;
                    var p = pinegrow.getCurrentProject();
                    if (p) {
                        //url = pinegrow.getProxyUrl(url);

                        url = p.getRelativeUrl(url);

                        if(reload.indexOf(url) < 0) {
                            reload.push(url);
                            reloadDelayed();
                        }
                    }

                }

                pinegrow.addEventHandler('on_page_changes_done', function (page) {
                    changedUrl(page.url);
                })

                pinegrow.addEventHandler('on_css_stylesheet_changed', function (pagenull, cs) {
                    cs.url && changedUrl(cs.url);
                })

                var $icon = $('.commands-toolbar');

                $icon.get(0).addEventListener('click', function(e) {
                    if(e.target.matches('.icon-27') || e.target.querySelector('.icon-27')) {
                        e.preventDefault();
                        e.stopPropagation();

                        var page = pinegrow.getSelectedPage();

                        if(!page) {
                            pinegrow.showQuickMessage('Nothing to preview in browser. Open a page first.');
                            return;
                        }

                        var actions = [
                            {
                                label: 'Open the page in browser',
                                //kbd: 'CMD O',
                                action: function() {
                                    if(!enabled) {
                                        page.openInBrowser();
                                    } else {
                                        var p = pinegrow.getCurrentProject();
                                        if(!p) {
                                            pinegrow.showQuickError('This feature requires projects. Open the page as Project first.');
                                        } else {
                                            if(!p.isUrlInProject(page.url)) {
                                                pinegrow.showQuickError('The page is not part of the current project. Opening in the browser instead...');
                                                page.openInBrowser();
                                            } else {
                                                var url = p.getRelativeUrl(page.url);

                                                var gui = require('nw.gui');
                                                url = bs_url + '/' + url;

                                                gui.Shell.openExternal(url);
                                            }
                                        }
                                    }
                                }
                            },
                            {
                                type: 'divider'
                            },
                            {
                                type: 'header',
                                label: 'BrowserSync'
                            },
                            {
                                label: 'Use BrowserSync',
                                check: enabled,
                                action: function() {
                                    enabled = !enabled;
                                    pinegrow.setSetting('browsersync-enabled', enabled ? '1' : '0');
                                    pinegrow.showMode('BrowserSync', enabled)
                                }
                            },
                            {
                                label: 'How to start BrowserSync',
                                action: function() {
                                    pinegrow.showAlert(`<p>BrowserSync must be running for this feature to work.</p><p>To start BrowserSync:</p><ul>
<li>Open the terminal / command prompt</li>
<li>Run <code>browser-sync start --proxy "${pinegrow.getApiUrl()}"</code></li>
</ul>
<p><a href="https://pinegrow.com/docs/master-pinegrow/using-external-code-editors/live-preview-on-any-device-or-browser-with-browsersync/" class="external">Learn more about using BrowserSync in Pinegrow.</a></p>`, 'How to start BrowserSync')
                                }
                            },
                            {
                                label: 'Settings...',
                                action: function() {
                                    var fields = {
                                        url: {
                                            name: 'BrowserSync url',
                                            type: 'text',
                                            value: bs_url
                                        }
                                    };

                                    pinegrow.showQuickDialog(`BrowserSync Settings`, 'Settings', `<p>BrowserSync is an external and free tool that let's you preview pages on multiple browsers / devices while you edit them in Pinegrow. All changes will be shown in the browser, without having to save them first.</p><p><a href="https://pinegrow.com/docs/master-pinegrow/using-external-code-editors/live-preview-on-any-device-or-browser-with-browsersync/" class="external">Learn more about using BrowserSync in Pinegrow.</a></p>`, fields, 'Cancel', 'Save', null, function(values) {
                                        bs_url = values.url || 'http://localhost:3000';
                                        pinegrow.setSetting('browsersync-url', bs_url);
                                    })
                                }
                            },
                        ]

                        var menu = new CrsaContextMenu(actions, $(e.target));
                        menu.showForElement($(e.target));
                    }
                }, true);

            }

            var pgBrowserSync = new PgBrowserSync();
            //End BrowserSync
        }

        if(parseFloat(crsaGetVersion()) >= 5.99 && parseFloat(crsaGetVersion()) < 6) {
            //Fix overwritting external links
            pinegrow.copyMissingExternalFiles = function (page, sourceUrl, destUrl, projectBaseUrl) {

                page.sourceNode.walk(function (node) {
                    if (node.isElement) {
                        node.mapLinksToNewLocation(projectBaseUrl, destUrl, true /* only if dest file exists */);
                    }
                    return true;
                })

                var remote_template = crsaIsRemoteUrl(sourceUrl);

                this.goThroughAllLinkedResources(page, function (el, attr, url) {
                    if (el.tagName == "a") return;

                    if (url && !crsaIsAbsoluteUrl(url)) {
                        var newUrl = crsaGetUrlToSmartCombine(destUrl, url);
                        var new_file_url = crsaCombineUrlWith(destUrl, newUrl);

                        var fs = require('fs');
                        var source_file_url = crsaCombineUrlWith(sourceUrl, url);
                        var new_file_path = crsaMakeFileFromUrl(new_file_url);

                        if (crsaIsRemoteUrl(source_file_url)) {
                            //console.log(`REMOTE FILE ${source_file_url}`);
                            if (!crsaIsFileExist(new_file_path)) {
                                pgCopyRemoteUrlToLocalPath(source_file_url, new_file_path);

                                var cs = pinegrow.stylesheets.getByUrl(new_file_url);
                                if (cs) {
                                    cs.reload();
                                }
                            }
                        } else {

                            var source_file_path = crsaMakeFileFromUrl(source_file_url);

                            if (crsaIsFileExist(source_file_path) && !crsaIsFileExist(new_file_path)) {
                                crsaCreateDirs(crsaGetFileDir(new_file_path), fs);
                                crsaCopyFileSync(fs, source_file_path, new_file_path);

                                //Check if stylesheet is already loaded, but has empty content because the file didn't exist before
                                var cs = pinegrow.stylesheets.getByUrl(new_file_url);
                                if (cs) {
                                    cs.reload();
                                }
                            }
                        }
                    }
                });
            }
        }

        //Hotpatch CSS helper tool
        if(current_version === 6) {
            if (typeof PgCSSFieldExtension !== 'undefined') {
                console.log('Patching CSS Field helper...')

                var patch_map = {
                    'line-height': true,
                    'z-index': true,
                    'opacity': true,
                    'background': true,
                    'background-image': true,
                    'flex-grow': true,
                    'flex-shrink': true,
                    'order': true,
                    'grid-row': true,
                    'grid-column': true,
                    'grid-row-start': true,
                    'grid-row-end': true,
                    'grid-column-start': true,
                    'grid-column-end': true,
                    //               'grid-': true,
                }

                var oldonSetValue = PgCSSFieldExtension.prototype.onSetValue;
                var oldonGetValue = PgCSSFieldExtension.prototype.onGetValue;

                PgCSSFieldExtension.prototype.onSetValue = function (val) {
                    if (patch_map[this.css_property]) {
                        this.add_default_unit = '';
                    }
                    return oldonSetValue.call(this, val);
                }

                PgCSSFieldExtension.prototype.onGetValue = function (val, event) {
                    if (patch_map[this.css_property]) {
                        this.add_default_unit = '';
                    }
                    return oldonGetValue.call(this, val, event);
                }

                setTimeout(function () {
                    //patch this.findCrsaStylesheetForUniqueName
                    pinegrow.stylesheets.old_findCrsaStylesheetForUniqueName = pinegrow.stylesheets.findCrsaStylesheetForUniqueName;

                    pinegrow.stylesheets.findCrsaStylesheetForUniqueName = function (uniqueName) {

                        var cs = this.old_findCrsaStylesheetForUniqueName(uniqueName);
                        if (!cs) {
                            cs = this.getCrsaStylesheetByUniqueName(uniqueName);
                        }
                        return cs;
                    }
                }, 100)

                var bsf = pinegrow.findFrameworkByKey('bs5');

                pinegrow.addEventHandler('on_get_floating_tools_sections', function (page, sections) {
                    if (page && bsf && page.hasFramework(bsf)) {
                        //console.log(sections);
                        sections.forEach(function (s) {
                            s.controls.forEach(function (c) {
                                if (c.def.key === 'raduisborder') {
                                    c.def.options = [
                                        {key: 'rounded', name: 'All sides'},
                                        {key: 'rounded-0', name: 'All sides none'},
                                        {key: 'rounded-1', name: 'All sides small'},
                                        {key: 'rounded-2', name: 'All sides medium'},
                                        {key: 'rounded-3', name: 'All sides large'},
                                        {key: 'rounded-top', name: 'Top'},
                                        {key: 'rounded-bottom', name: 'Bottom'},
                                        {key: 'rounded-start', name: 'Left'},
                                        {key: 'rounded-end', name: 'Right'},
                                        {key: 'rounded-circle', name: 'Circle'},
                                        {key: 'rounded-pill', name: 'Pill'}
                                    ]
                                    c.options.submenu = null;
                                }
                            })
                        })
                    }
                })

                console.log('Live patching PgDmEngineForBootstrap.destroy...');
                PgDmEngineForBootstrap.prototype.destroy = function () {
                    console.log('Using patched PgDmEngineForBootstrap.destroy');

                    try {
                        if (this.css_mode === 'generate') {
                            this.outcs.remove();
                            pinegrow.stylesheets.removeStylesheetByKey(this.outcs.url);

                            this.cs.remove();
                            pinegrow.stylesheets.removeStylesheetByKey(this.cs.url);

                            this.dummy_outcs.remove();
                            pinegrow.stylesheets.removeStylesheetByKey(this.dummy_outcs.url);
                        }
                    } catch (err) {
                        console.error(err)
                    }
                }

                //Patch JS file colorization, doesn't work
                /*
                try {
                    if(!pgIsDev()) {
                        PgMonacoEditor.languages.forEach(function (lang) {
                            if (lang.id === 'javascript') {
                                lang.mimetypes.push('application/javascript');
                            }
                        })
                    }
                } catch(err) {
                    console.error(err)
                }

                 */

                //Patch ids in BS files for DP
                console.log('Live patching PgDmEngineForBootstrap.prototype.update');
                PgDmEngineForBootstrap.prototype.oldUpdate = PgDmEngineForBootstrap.prototype.update;


                PgDmEngineForBootstrap.prototype.update = function (done) {
                    if (this.css_mode === 'generate') {

                        var forAll = function (importedFiles, func) {
                            for (var i = 0; i < importedFiles.length; i++) {
                                var importedFile = importedFiles[i];
                                func(importedFile);
                                if (importedFile.importedFiles.length > 0) forAll(importedFile.importedFiles, func);
                            }
                        }

                        this.customCs.importedFiles && forAll(this.customCs.importedFiles, function (ics) {
                            if (!ics.original_getSourceForRules) {
                                ics.original_getSourceForRules = ics.getSourceForRules;
                                ics.getSourceForRules = function (node, withParams, forceFormatting) {
                                    return this.original_getSourceForRules(node, false, forceFormatting);
                                }
                            }
                        });
                    }
                    //this.outcs && (this.outcs.changed = true);
                    return this.oldUpdate(done);
                }

                PgDmEngineForBootstrap.prototype.saveCSS = function () {
                    var outfile = crsaMakeFileFromUrl(this.out_url);
                    if (this.css_mode === 'generate' && outfile) {
                        this.outcs.save(outfile);
                        this.outcs.setChanged(false);
                    }
                }
            }
        }

        if(current_version <= 6.1) {
            console.log('LIVE: Patch shouldHaveId on PgNodeEvent.');
            PgNodeEvent.prototype.shouldHaveId = function() {
                return false;
            }
        }

        if(current_version >= 6) {
            console.log('LIVE: Disable WP warnAboutActionThatIsNotOnMasterOrInContent');
            var wpf = pinegrow.findFrameworkByKey('wordpress.pinegrow');
            if(wpf) {
                wpf.warnAboutActionThatIsNotOnMasterOrInContent = function () {
                }

                console.log('LIVE: Fix WP wpIsPGTCProject');
                wpf.wpIsPGTCProject = function(project) {
                    if(!project.db) {
                        project.db = new PgProjectDB(project);
                        project.db.init(true /* skip create */);
                    }
                    if(project.db.is_loaded) {
                        if(project.db.findOne('elements')) return true;
                    }
                    return false;
                }
            }
        }

        var dumpObserver = function() {
            //Looking for crashdumps
            console.log('LIVE: Setting up the dump observer...');
            try {
                var gui = require('nw.gui');
                var data_path = gui.App.dataPath;
                var path = require('path');
                var dir;
                if (is_mac) {
                    dir = path.join(data_path, '../Crashpad/completed')
                } else if (is_windows) {
                    dir = path.join(data_path, 'User Data/CrashPad/completed');
                } else {
                    dir = path.join(data_path, 'Crash Reports/completed');
                }
                var list = []
                var dialog_shown = false;
                var fs = require('fs');
                fs.watch(dir, {persistent: true}, function (event, name) {
                    console.log('LIVE Dump watcher: a crash report was generated ' + name + ', ' + event);
                    var t = Date.now();
                    list.push({time: t, name: name});
                    var cut = t - 20 * 1000; //how many in 10 seconds
                    var cut_idx = -1;
                    for (var i = 0; i < list.length; i++) {
                        if (list[i].time < cut) {
                            cut_idx = i;
                        } else {
                            break;
                        }
                    }
                    if (cut_idx >= 0) {
                        list.splice(0, cut_idx + 1);
                    }
                    if (list.length > 5 && !dialog_shown) {
                        dialog_shown = true;

                        pinegrow.showAlert(`<p>A large number of crash dumps (${list.length} in the last 20s) are being created in <code>${dir}</code> folder. This usually happens in connection with embeded browser rendered crashes.</p><p>Sorry about this! We are working on a better solution. For now, please save your work and restart Pinegrow. You might also want to delete all files in the above mentioned folder.</p><p>Please contact our support on support@pinegrow.com if this happens often.</p>`, 'Too many crash dumps generated');
                    }

                    if(pgIsDev()) {
                        pinegrow.showQuickError(`DEV: Crash dump created!`);
                    }
                })
                console.log(`LIVE: Dump observer is watching ` + dir);
            } catch(err) {
                console.warn(`LIVE: Unable to launch the dump observer.`, err);
            }
        }


        if(current_version < 6.31) {

            dumpObserver();

            console.log('LIVE: Path CSS for Menu points');
            $('head').append(`<style>
div.pg-sel-menu-point {
    left: 0;
    top: 0;
    }
</style>`);
        }

        if(current_version < 6.61) {
            //patch function crsaIsAbsoluteUrl
            window.crsaIsAbsoluteUrl = function(url) {
                return !!url && (url.indexOf('://') >= 0 || url.indexOf('//') === 0);
            }
        }

        //Add action icons styles for breadcrumbs
        $('head').append(`<style>
.status-bar-crumbs li.item action.icon-Scene_Scroll, .status-bar-crumbs li.item action.icon-hide, .status-bar-crumbs li.item action.icon-see, .status-bar-crumbs li.item action.icon-113, .status-bar-crumbs li.item action.icon-Applay_to_many {
            display: inline-block !important;
            color: inherit;
        }
</style>`);

        if(current_version === 6.7) {
            //patch wp map
            setTimeout(function() {
                var wpf = pinegrow.findFrameworkByKey('wordpress.pinegrow');
                if(wpf && document.body.isWPEnabledForProject_map && !wpf.isWPEnabledForProject_map) {
                    console.log('LIVE: Fixing WP isWPEnabledForProject_map map')
                    wpf.isWPEnabledForProject_map = document.body.isWPEnabledForProject_map;
                }
            }, 1000)
        }

    }

    //This is loaded with async. Perhaps pinegrow is already initialized?
    if(typeof pinegrow !== 'undefined' && pinegrow.ready) {
        init();
    } else {
        $('body').one('pinegrow-ready', function (e, pinegrow) {
            init()
        })
    }
})