
var PgDmCSSCodeMaker = function(data) {
    this.indents = [
        '','    ','        '
    ];
    this.data = data;
    this.css = '';
}

PgDmCSSCodeMaker.prototype.getParam = function(key) {
    if(key in this.data) {
        return this.data[key];
    }
    return null;
}

PgDmCSSCodeMaker.prototype.p = function(key, prop, important, val) {
    if(val === undefined) val = this.getParam(key);
    if(val !== null && val !== '') {
        return `${prop}: ${val}${important ? ' !important' : ''};`;
    }
    return null;
}

PgDmCSSCodeMaker.prototype.cssvar = function(key, prop, important, val) {
    if(val === undefined) val = this.getParam(key);
    if(val !== null && val !== '') {
        return `${prop}: var(--${val})${important ? ' !important' : ''};`;
    }
    return null;
}

PgDmCSSCodeMaker.prototype.v = function(key, varname, val) {
    if(val === undefined) val = this.getParam(key);
    if(val !== null && val !== '') {
        this.css += `$${varname}: ${val};
`;
    }
}


PgDmCSSCodeMaker.prototype.rule = function(sel, props, level) {
    if(level === undefined) level = 1;
    var body = '';

    for(var i = 0; i < props.length; i++) {
        if(props[i]) {
            body += this.indents[level] + props[i] + "\n";
        }
    }

    if(body.length) {
        this.css += `
${sel} {
${body}}`;
    }
}

class PgDmDesignColor {
    constructor(name, value, shades, can_rename) {
        this.name = name;
        this.value = value;
        this.shades = shades;
        this.can_rename = can_rename;
        this.locked = false;
        //this.generate_shades = true;
    }

    setValue(val, shade) {
        if(shade) {
            this.shades[shade] = val;
        } else {
            this.value = val;
        }
    }

    getValue(shade) {
        if(shade) {
            return this.shades[shade];
        } else {
            return this.value;
        }
    }

    copyFromColor(color, leave_shades) {
        this.value = color.value;
        if(!leave_shades) {
            this.shades = {};

            var _this = this;
            color.forEachShade(function (key, col) {
                _this.shades[key] = col;
            })
        } else {
            this.calculateShades();
        }
    }

    isLocked() {
        return this.locked;
    }

    setLocked(v) {
        this.locked = v;
    }

    clone() {
        return new PgDmDesignColor(this.name, this.value, cloneObject(this.shades));
    }

    hasShades() {
        return this.shades;
    }

    forEachShade(func) {
        if(this.shades) {
            for (var key in this.shades) {
                func(key, this.shades[key]);
            }
        }
    }

    setShades(map, calc, only_calc_for_these_list) {
        if(map) {
            this.shades = {}
            for (var key in map) {
                this.shades[key] = map[key];
            }
        } else {
            this.shades = null;
        }
        calc && this.calculateShades(only_calc_for_these_list);
    }

    setShadeKeys(list) {
        if(!list || list.length === 0) {
            this.shades = null;
        } else {
            this.shades = {};
            for(var i = 0; i < list.length; i++) {
                this.shades[list[i]] = null;
            }
            this.calculateShades();
        }
    }

    calculateShades(only_calc_for_these_list) {

        if(!this.shades) return;

        var shade_keys = Object.keys(this.shades);

        var numColors = shade_keys.length;

        if(numColors === 0) return;

        var base_idx = Math.floor(numColors / 2);

        //b=1

        var base = this.value;

        var dist_light = 0.15;
        var dist_dark = 0.35;

        switch(numColors) {
            case 2:
            case 3:
                dist_dark = dist_light = 0.75;
                break;
            case 4:
            case 5:
                dist_dark = dist_light = 0.5;
                break;
            case 6:
            case 7:
                dist_light = 0.25;
                break;
        }

        if(base_idx > 0) {
            var max_tint = 90;
            var tint_step = max_tint / base_idx;
            var tint = tint_step;

            const colors = chroma.bezier([chroma.mix('white', base, dist_light), base])
                .scale()
                .correctLightness()
                .colors(base_idx + 1);

            for (var i = 0; i < base_idx; i++) {
                //this.shades[shade_keys[i]] = chroma.mix('white', base, tint / 100).css();
                if(!only_calc_for_these_list || only_calc_for_these_list.indexOf(shade_keys[i]) >= 0) {
                    this.shades[shade_keys[i]] = colors[i];
                }
                tint += tint_step;
            }
        }

        if(!only_calc_for_these_list || only_calc_for_these_list.indexOf(shade_keys[base_idx]) >= 0) {
            this.shades[shade_keys[base_idx]] = base;
        }

        if(base_idx < shade_keys.length - 1) {

            var max_tint = 90;
            var tint_step = max_tint / base_idx;
            var tint = max_tint;

            const colors = chroma.bezier([base, chroma.mix('black', base, dist_dark)])
                .scale()
                .correctLightness()
                .colors(numColors - base_idx);


            for(var i = base_idx + 1; i < shade_keys.length; i++) {
                if(!only_calc_for_these_list || only_calc_for_these_list.indexOf(shade_keys[i]) >= 0) {
                    this.shades[shade_keys[i]] = colors[i - base_idx];
                }
                //this.shades[shade_keys[i]] = chroma.mix('black', base, tint / 100).css();
                tint -= tint_step;
            }
        }
/*
        var colors = Object.values(this.shades);

        const stepsLeft = chroma.scale(colors)
            .correctLightness()
            .colors(colors.length);

        for(var i = 0; i < stepsLeft.length; i++) {
            this.shades[shade_keys[i]] = stepsLeft[i];
        }

 */


    }

    calculateShades2() {

        if(!this.shades) return;

        var shade_keys = Object.keys(this.shades);

        var numColors = shade_keys.length;

        if(numColors === 0) return;

        var _range = function(start, end, step) {
            var r = [];
            while(start < end) {
                r.push(start);
                start += step;
            }
            return r;
        }

        var autoGradient = function(color, numColors) {
            const lab = chroma(color).lab();
            const lRange = 100 * (0.95 - 1/numColors);
            const lStep = lRange / (numColors-1);
            let lStart = (100-lRange)*0.5;
            const range = _range(lStart, lStart+numColors*lStep, lStep);
            let offset = 0;
            if (true) {
                offset = 9999;
                for (let i=0; i < numColors; i++) {
                    let diff = lab[0] - range[i];
                    if (Math.abs(diff) < Math.abs(offset)) {
                        offset = diff;
                    }
                }
            }
            return range.map(l => chroma.lab([l + offset, lab[1], lab[2]]));
        }

        var autoColors = function(color, numColors) {
            return autoGradient(color, numColors);
        }

        const color = this.value;
        const colors2 = []

        const even = numColors % 2 === 0;
        const numColorsLeft = numColors;
        const numColorsRight = 0;
        const genColors = autoColors(color, numColorsLeft);
        //const genColors2 = autoColors(colors2[0], numColorsRight, true);
        const stepsLeft = chroma.scale(genColors)
            .correctLightness()
            .colors(numColorsLeft);
        const stepsRight = [];
        const steps = stepsLeft;//(stepsLeft).concat(stepsRight.slice(1));

        var key_i = 0;
        for(var i = steps.length - 1; i >= 0; i--) {
            this.shades[shade_keys[key_i++]] = steps[i];
        }
    }
}

var PgDmDesignColors = function() {

    this.can_edit = false;
    this.mode = 'sass_var';

    this.default_shades = null;

    var map = {}
    var locked = {}

    var _this = this;

    this.onReplaceWith = null;

    this.showContextMenu = function(key, $el) {
        var a = [];

        if(this.onReplaceWith) {
            var submenu = [];
            this.forEachColor(function(k, val) {
                if(k !== key) {
                    submenu.push({
                        label: k,
                        action: function() {
                            _this.onReplaceWith(key, k);
                        }
                    })
                }
            })
            a.push({
                label: 'Replace this color with',
                submenu: submenu
            })
        }

        if(a.length) {
            var m = new CrsaContextMenu(a, $el);
            m.showForElement($el);
        }
    }

    this.add = function(key, value, is_locked, shades, can_rename) {
        var c = new PgDmDesignColor(key, value, shades, can_rename);
        map[key] = c;
        if(is_locked !== undefined) {
            c.setLocked(is_locked);
            //locked[key] = is_locked;
        }
        return c;
    }

    this.deleteColor = function(color_object) {
        delete map[color_object.name];
    }

    this.renameColor = function(color_object, name) {
        //shold keep the order the same
        var keys = Object.keys(map);
        var new_map = {};
        var idx = keys.indexOf(color_object.name);
        for(var i = 0; i < keys.length; i++) {
            if(idx === i) {
                new_map[name] = color_object;
            } else {
                new_map[keys[i]] = map[keys[i]];
            }
        }
        if(idx < 0) {
            new_map[name] = color_object;
        }
        map = new_map;
        color_object.name = name;
    }

    this.clear = function() {
        map = {};
    }

    this.setColor = function(key, value, shade) {
        map[key].setValue(value, shade);
    }

    this.getColor = function(key, shade) {
        return map[key].getValue(shade);
    }

    this.getColorObject = function(key) {
        return map[key] || null;
    }

    this.getShades = function(key) {
        return map[key].shades || null;
    }

    this.copyFromColor = function(key, color, leave_shades) {
        map[key].copyFromColor(color, leave_shades)
    }

    this.getColorKeyByIndex = function(i, value) {
        var keys = Object.keys(map);
        if(keys.length > i) {
            return keys[i];
        }
    }

    this.setColorByIndex = function(i, value) {
        var keys = Object.keys(map);
        if(keys.length > i) {
            this.setColor(keys[i], value);
        }
    }

    this.getNumberOfColors = function() {
        return Object.keys(map).length;
    }

    this.forEachColor = function(f) {
        $.each(map, function(key, val) {
            f(key, val.getValue(), val);
        })
    }

    this.isLocked = function(key) {
        return map[key].isLocked();
        //return locked[key] || false;
    }

    this.setLocked = function(key, v) {
        map[key].setLocked(v);
        //locked[key] = v;
    }

    this.getNumberOfUnlockedColors = function() {
        var c = 0;
        this.forEachColor(function(key, val, co) {
            if(!co.isLocked()) c++;
        })
        return c;
    }

    this.setUnlockedColorByIndex = function(i, value) {
        var c = 0;
        this.forEachColor(function(key, val, co) {
            if(!co.isLocked()) {
                if(i === c) {
                    map[key].setValue(value);
                }
                c++;
            }
        })
    }

    this.getUnlockedColorKeyByIndex = function(i) {
        var c = 0;
        var k = null;
        this.forEachColor(function(key, val, co) {
            if(!co.isLocked()) {
                if(i === c) {
                    k = key;
                }
                c++;
            }
        })
        return k;
    }

    this.getColorsForPicker = function(prefix, rows) {
        prefix = prefix || (this.mode === 'sass_var' ? '$' : '');
        rows = rows || [];
        this.forEachColor(function(key, col, color_object) {
            var row = [];
            if(!color_object.hasShades()) {
                row.push(
                    {
                        key: prefix + key,
                        name: key,
                        color: col
                    }
                )
            } else {
                color_object.forEachShade(function(subkey, subcol) {
                    const k = prefix + key + '-' + subkey;
                    row.push(
                        {
                            key: k,
                            name: k,
                            color: subcol
                        }
                    )
                })
            }
            rows.push(row)
        })
        return rows;
    }

    this.debug = function() {
        console.warn(map)
    }
}

class PgDmDesignSkillColors extends PgDmDesignSkill {

    constructor(key, colors, get_code, get_code_stage) {
        super(key, 'Colors');
        this.colors = colors;
        this.get_code = get_code;
        this.get_code_stage = get_code_stage || 'init';

        this.colors_control = null;

        var _this = this;

        //create controls
        this.addControl('colors',
            {
                name: 'Colors',
                type: 'custom',
                dm_no_lock: true,
                control: function() {
                    //will keep helper functions and vars private
                    //Create control, needs a unique id
                    var c = _this.colors_control = new PgCustomPropertyControl('pg.dm.colors');

                    var swatches = new PgDmColorSwatches(_this.colors, _this);

                    swatches.onColorChanged = function(key, color, moving_in_picker) {

                        if(moving_in_picker) {
                            //_this.colors.setColor(key, color); already set by swatch
                            pgDmStore.update();
                        } else {
                            //finished color selection
                            pgDmStore.db.makeChanges(null, 'Set color', function () {
                                //_this.colors.setColor(key, color); already set by swatch
                                _this.controlSetValue(c.obj, color, {}, null, 'change', {
                                    subkey: key
                                })
                            })
                        }
                        swatches.debug();
                    }

                    swatches.onColorsReordered = function(mode) {
                        pgDmStore.db.makeChanges(null, mode === 'copy' ? 'Copy color' : 'Swap colors', function () {
                            _this.colors.forEachColor(function(key, color) {
                                _this.updateColorNode(key, color);
                            })
                        })
                        pgDmStore.update();
                    }

                    swatches.onColorDeleted = function(color_object) {
                        pgDmStore.db.makeChanges(null, 'Delete color', function () {
                            _this.deleteColorNode(color_object.name);
                        })
                        _this.colors.deleteColor(color_object);
                        pgDmStore.getActiveDesignProvider().onColorListChanged();
                        swatches.updateDisplay(true);
                        pgDmStore.update();
                    }

                    swatches.onColorRenamed = function(old_name, name, color_object) {
                        pgDmStore.db.makeChanges(null, 'Rename color', function () {
                            _this.colors.renameColor(color_object, name);
                            _this.renameColorNode(old_name, name);
                        })
                        pgDmStore.getActiveDesignProvider().onColorListChanged();
                        swatches.updateDisplay(true);
                        pgDmStore.update();
                    }

                    //Register subfields.
                    c.onDefine = function() {
                        //no subfields
                    }

                    //Show control. Return the control $el.
                    c.onShow = function() {
                        return swatches.$element;
                    }

                    //Called when control is recycled and new values are set
                    c.onSetValues = function() {
                        //console.warn(`onSetValues`);
                        //colors.debug();
                        swatches.updateDisplay(true);
                        /*
                        _this.colors.forEachColor(function(key, val, color_object) {
                            swatches.setColor(key, val, color_object);
                            swatches.setLocked(key, _this.colors.isLocked(key));
                        });*/
                    }

                    c.onHasData = function() {
                        return true;
                    }

                    //Return the control
                    return c;
                }
            }
        )

        /*
this.addControl('interval',
    {
        name: 'Color interval',
        type: 'range',
        range_min: 5,
        range_max: 50,
        range_marks: true
    }
)

 */
        this.addControl(`button.from_image`,
            {
                name: 'Get from the selected image',
                type: 'button',
                add_class: 'pg-dm-colors-buttons',
                dm_no_lock: false, //force lock for button
                func: function() {
                    var dp = pgDmStore.getActiveDesignProvider();

                    var doWithBeforeSurprise = function(func) {
                        var f = function() {
                            pinegrow.showQuickMessage(ok_msg);
                            setTimeout(function() {
                                func();
                            }, 50);
                        }

                        if(dp.onBeforeSurpriseMe) {
                            dp.onBeforeSurpriseMe(f)
                        } else {
                            f();
                        }
                    }

                    var err_msg = 'The selected element is not an image and it doesn\'t have a background image.';
                    var ok_msg = 'Getting colors from the image...';

                    var doTask = function() {
                        var pgel = pinegrow.selectedElements.getLastSelected();
                        if(!pgel) {
                            pinegrow.showQuickError('Please select an element first.');
                        } else if(pgel.tagName != 'img') {
                            //look for background
                            var $el = pgel.get$DOMElement();
                            var bck_img = $el.css('background-image');
                            if(bck_img) {
                                var m = bck_img.match(/url\(["']?([^"'\)]+)["']?\)/);
                                if(m) {
                                    bck_img = m[1];
                                    doWithBeforeSurprise(function() {
                                        _this.getPaletteFromImage(bck_img, function() {
                                            pgDmStore.update();
                                            pgDmStore.designPanel.updatePanelDisplay();
                                        }, 'Get colors from image')
                                    })
                                } else {
                                    pinegrow.showQuickError(err_msg)
                                }
                            } else {
                                pinegrow.showQuickError(err_msg)
                            }
                        } else {
                            var $img = pgel.get$DOMElement();
                            if($img) {
                                doWithBeforeSurprise(function() {
                                    if(isApp()) {
                                        _this.setColorsFromImage($img.get(0), function () {
                                            pgDmStore.update();
                                            pgDmStore.designPanel.updatePanelDisplay();
                                        }, 'Get colors from image')
                                    } else {
                                        //reload image due to cors in browser
                                        _this.getPaletteFromImage($img.get(0).src, function() {
                                            pgDmStore.update();
                                            pgDmStore.designPanel.updatePanelDisplay();
                                        }, 'Get colors from image')
                                    }
                                })
                            } else {
                                pinegrow.showQuickError(err_msg)
                            }
                        }
                    }

                    doTask();

                    //we call this to get the chance to dp to add extra unlocked colors
                    //moved to doTasks to check before calling onbefore
                    /*
                    if(dp.onBeforeSurpriseMe) {
                        dp.onBeforeSurpriseMe(doTask)
                    } else {
                        doTask();
                    }*/
                }
            }
        )

        this.addControl('button.more', {
            type: 'button',
            name: '<i class="icon icon-down" style="position: relative;top: 2px;"></i>',//pgSVGIcons.three_dots,
            append_to: '.pg-dm-colors-buttons',
            button_class: 'pg-button-svg-icon',
            helptext: '<p>Select the strategy for picking colors and restore defaults.</p>',
            func: function(obj, $button, values, $fc, event) {
                var harmony = _this.getParam('harmony');

                var setStrategy = function() {
                    _this.setParam('harmony', this.data);

                    var bck_skill = pgDmStore.getDesignSkill('background');
                    var url = bck_skill.getParam('image');
                    if(url) {
                        bck_skill.getColorsFromImage(url, function () {
                            pgDmStore.update();
                        })
                    }
                }

                var a = [
                    {
                        label: 'Strategy for getting colors from images',
                        type: 'header'
                    },
                    {
                        label: 'Get all colors from the image',
                        data: '',
                        action: setStrategy,
                        check: !harmony
                    },
                    {
                        label: '2 from image + 1 complimentary',
                        data: 'complimentary',
                        action: setStrategy,
                        check: harmony === 'complimentary'
                    },
                    {
                        label: '2 from image + 2 split',
                        data: 'split',
                        action: setStrategy,
                        check: harmony === 'split'
                    },
                    {
                        label: '2 from image + 2 triad',
                        data: 'triad',
                        action: setStrategy,
                        check: harmony === 'triad'
                    },
                    {
                        label: '2 from image + 2 similar',
                        data: 'analog',
                        action: setStrategy,
                        check: harmony === 'analog'
                    },
                    {
                        type: 'divider'
                    },
                    {
                        label: 'Restore default colors',
                        action: function() {
                            _this.restoreDefaultColors();
                        }
                    }
                ]

                event.stopPropagation();
                event.stopImmediatePropagation();

                var m = new CrsaContextMenu(a, $button);
                m.showForElement($button);
            }
        })
    }

    restoreDefaultColors() {
        var _this = this;
        this.populateDefaultParams(true);
        pgDmStore.db.makeChanges(null, 'Restore colors', function () {
            _this.colors.forEachColor(function(key, color, color_obj) {
                _this.updateColorNode(key, color);
                pgDmStore.setParamLocked(_this, key, color_obj.isLocked())
            })
        })
        pgDmStore.update();
        pgDmStore.designPanel.updatePanelDisplay();
    }

    updateColorNode(idx, color, name) {
        var _this = this;

        var node = _this.node.findOne(`[key="${idx}"]`);
        if(!node) {
            node = pgCreate(`<dmcolor key="${idx}" />`);
            node.appendTo(_this.node);
        }
        node.setAttr('color', color);
        var co = this.colors.getColorObject(idx);
        node.removeAttrIfStartsWithStringWithEvents('shade-');
        co.forEachShade(function(skey, scol) {
            node.setAttribute('shade-' + skey, scol);
        })
        //this.colors.isLocked(idx) ? node.setAttribute('locked') : node.removeAttribute('locked');
    }

    deleteColorNode(idx) {
        var node = this.node.findOne(`[key="${idx}"]`);
        node && node.remove();
    }

    renameColorNode(old_idx, idx) {
        var node = this.node.findOne(`[key="${old_idx}"]`);
        node && node.setAttribute('key', idx);
    }

    onLoad(node) {
        var _this = this;
        var nodes = node.find('dmcolor');
        if(nodes.length) {
            _this.colors.clear();
        }
        nodes.forEach(function(c) {
            var key = c.getAttribute('key');
            var css_color = c.getAttribute('color');

            var shades = {};
            var has_shades = false;
            c.getAttrList().forEach(function(a) {
                if(a.name.startsWith('shade-')) {
                    shades[a.name.replace('shade-', '')] = a.value;
                    has_shades = true;
                }
            })
            var co = _this.colors.getColorObject(key);
            if(!co) {
                co = _this.colors.add(key, css_color, false, null, true);
            } else {
                co.setValue(css_color);
            }
            co.shades = has_shades ? shades : null;

            //_this.colors.setLocked(key, c.hasAttribute('locked'));
        })
    }

    onSetParam(name, value) {
        if(name === 'harmony') {
            this.node.setAttribute('harmony', value);
        } else {
            this.updateColorNode(name, value);
        }
    }

    onGetParam(name, value) {
        if(name === 'harmony') {
            return this.node.getAttribute('harmony');
        }
        return value;
    }

    setColorsFromImage(src, done, undo_action) {
        var _this = this;
        var pal = new PgDmPalette([], src, _this.colors.getNumberOfUnlockedColors());
        pal.harmony = this.getParam('harmony');
        this.setFromPalette(pal, function() {
            _this.colors_control && _this.colors_control.onSetValues();
            done();
        }, undo_action);
    }

    setFromPalette(pal, done, undo_action) {
        var _this = this;
        var i = 0;

        var task = function(list) {
            list.forEach(function(c) {

                var css_color = c.color.css();

                var color_key = _this.colors.getUnlockedColorKeyByIndex(i);

                if(color_key) {
                    _this.colors.setColor(color_key, css_color);
                    var co = _this.colors.getColorObject(color_key);
                    co.calculateShades();
                    _this.updateColorNode(color_key, css_color);
                }
                i++;
            })
            done && done();
        }
        pal.getColorsInfo(function(list) {
            if(undo_action) {
                pgDmStore.db.makeChanges(null, undo_action, function() {
                    task(list);
                }, 'design_panel');
            } else {
                task(list)
            }
        })
    }

    getPaletteFromImage(val, done, undo_action, harmony) {
        var _this = this;
        var img = new Image();
        img.onload = function () {

            var pal = new PgDmPalette([], img);
            pal.harmony = harmony;
            _this.setFromPalette(pal, done, undo_action);
        }
        img.onerror = function() {
            done && done();
        }
        if(!isApp()) {
            img.crossOrigin = "anonymous";
        }
        img.src = val;
    }

    areAllColorsLocked() {
        return this.isLocked(`button.from_image`);
    }

    onSurpriseMe(done) {
        if(this.areAllColorsLocked) {
            done();
            return;
        }
        var bck_skill = pgDmStore.getDesignSkill('background');
        if(bck_skill && !bck_skill.isLocked('image')) {
            done();
            //random image will also set the palette
            return;
        }

        var p = pgDmStore.design.palettes.getRandomPalette();
        this.setFromPalette(p);
        done();
    }

    onGetCSS(part) {
        if(part === this.get_code_stage) {
            return this.get_code();
        } else {
            return '';
        }
    }

    populateDefaultParams(keep_existing) {
        super.populateDefaultParams();

        this.node.find('dmcolor').forEach(function(cn) {
            cn.remove();
        })

        var _this = this;
        this.colors.forEachColor(function(key, col, color_object) {
            _this.updateColorNode(key, col);
            pgDmStore.setParamLocked(_this, key, color_object.isLocked());
        })
    }
}

class PgDmDesignSkillFonts extends PgDmDesignSkill {
    constructor(finetune, colors_skill, mode) {
        super(finetune ? 'headings' : 'fonts', finetune ? 'Finetune headings' : 'Typography');
        var _this = this;

        this.mode = mode || 'sass';
        this.get_code_stage = mode === 'css' ? 'rules' : 'init';

        this.finetune = finetune;

        this.effects = {
            shadow: {
                name: 'Shadow',
                css: `text-shadow: 0.03em 0.05em 0.05em rgba(0, 0, 0, 0.3);`
            },
            shadow_hard: {
                name: 'Hard shadow',
                css: `text-shadow: 0.08em 0.08em 0px rgba(0,0,0,0.2);`
            },
            shadow_distant: {
                name: 'Distant shadow',
                css: `text-shadow: 0em 0.05em 0em rgba(0,0,0,0.2),0em 0.1em 0.1em rgba(0,0,0,0.15),0em 0.2em 0.03em rgba(0,0,0,0.1),0em 0.45em 0.34em rgba(0,0,0,0.1);`
            },
            shadow_heavy: {
                name: 'Heavy shadow',
                css: `text-shadow: 0em 0.05em 0.04em rgba(0,0,0,0.25),0em 0.1em 0.14em rgba(0,0,0,0.1),0em 0.18em 0.23em rgba(0,0,0,0.1);`
            }
        }

        this.fonts = [
            {
                key: 'text',
                name: 'Regular text',
                usage: 'text',
                toCSS: function(part) {
                    var css = new PgDmCSSCodeMaker(_this.getFontParams('text'));

                    var props = []

                    if(part === _this.get_code_stage) {
                        var family = _this.getFontParam('text', 'family');
                        if(family) {
                            var f = pinegrow.fontLibrary.getFontByFamily(family);
                            if(f) {
                                family = f.getFamilyWithFallback();
                            }
                            if(_this.mode === 'sass') {
                                css.v('family', 'font-family-sans-serif', family);
                            } else {
                                props.push(css.p(null, 'font-family', false, family));
                            }
                        }
                        if(_this.mode === 'sass') {
                            css.v('weight', 'font-weight-base');
                            css.v('height', 'line-height-base');
                            css.v('color', 'body-color');
                        } else {
                            props.push(css.p('weight', 'font-weight'));
                            props.push(css.p('height', 'line-height'));
                            props.push(css.cssvar('color', 'color'));
                        }

                        if(_this.mode === 'css') {
                            css.rule('body', props);
                        }

                    }
                    if(part === 'rules') {
                        css.rule('p, li', _this.addEffect('text', _this.getCSSDetailsProps(css)))
                    }
                    return css.css;
                },
                surpriseMe: function(f) {
                    var key = f.key;
                    var f = pinegrow.fontLibrary.getRandomFont('text');
                    if(!_this.isLocked(key, 'family')) {
                        _this.setFontParam(key, 'family', f.family);
                    }
                    _this.resetDetails('text');
                }
            },
            {
                key: 'headings',
                name: 'Headings',
                usage: 'headings',
                toCSS: function(part) {
                    var css = new PgDmCSSCodeMaker(_this.getFontParams('headings'));

                    var props = []

                    if(part === _this.get_code_stage) {
                        var family = _this.getFontParam('headings', 'family');
                        if(family) {
                            var f = pinegrow.fontLibrary.getFontByFamily(family);
                            if(f) {
                                family = f.getFamilyWithFallback();
                            }
                            if(_this.mode === 'sass') {
                                css.v('family', 'headings-font-family', family);
                            } else {
                                props.push(css.p(null, 'font-family', false, family));
                            }
                        }
                        if(_this.mode === 'sass') {
                            css.v('weight', 'headings-font-weight');
                            css.v('height', 'headings-line-height');
                            css.v('color', 'headings-color');
                        } else {
                            props.push(css.p('weight', 'font-weight'));
                            props.push(css.p('height', 'line-height'));
                            props.push(css.cssvar('color', 'color'));
                            props = props.concat(_this.addEffect('headings', _this.getCSSDetailsProps(css)));
                        }

                        if(_this.mode === 'css') {
                            css.rule('h1, h2, h3, h4, h5, h6', props);
                        }

                    }

                    if(part === 'rules' && _this.mode === 'sass') {
                        css.rule('h1, h2, h3, h4, h5, h6', _this.addEffect('headings', _this.getCSSDetailsProps(css)));
                    }
                    return css.css;
                },
                surpriseMe: function(f) {
                    var key = f.key;
                    var f = pinegrow.fontLibrary.getRandomFont('headings');
                    if(!_this.isLocked(key, 'family')) {
                        _this.setFontParam(key, 'family', f.family);
                    }
                    _this.resetDetails('headings');
                }
            },
            {
                key: 'lead',
                name: 'Lead',
                usage: 'text',
                toCSS: function(part) {
                    var css = new PgDmCSSCodeMaker(_this.getFontParams('lead'));

                    if(part === _this.get_code_stage) {
                        css.v('weight', 'lead-font-weight');

                    }

                    if(part === 'rules') {
                        var family = _this.getFontParam('lead', 'family');
                        if(family) {
                            var f = pinegrow.fontLibrary.getFontByFamily(family);
                            if(f) {
                                family = f.getFamilyWithFallback();
                            }
                        }

                        css.rule('.lead', _this.addEffect('lead', _this.getCSSDetailsProps(css).concat([
                            css.p('family', 'font-family', false, family),
                            css.p('height', 'line-height'),
                            css.p('color', 'color'),
                        ])))
                    }
                    return css.css;
                }
            },
            {
                key: 'h1',
                name: 'H1',
                usage: 'h1',
                //label_before: 'Finetune headings',
                toCSS: function(part) {
                    return _this.toCSSForHeadings(part, 'h1')
                },
                surpriseMe: function(f) {
                    var key = f.key;
                    if(!_this.isLocked(key, 'family')) {
                        //in most cases we just let the main headings setting be
                        //in other we can set a different h1
                        if(Math.random() < 0.4) {
                            var f = pinegrow.fontLibrary.getRandomFont('h1');
                            _this.setFontParam(key, 'family', f.family);
                            _this.resetDetails('h1');

                            //how about h2
                            if(!_this.isLocked('h2', 'family')) {
                                if (Math.random() < 0.5) {
                                    _this.setFontParam('h2', 'family', f.family);
                                } else {
                                    _this.setFontParam('h2', 'family', '');
                                }
                                _this.resetDetails('h2');
                            }
                        } else {
                            _this.setFontParam(key, 'family', '');
                            if(!_this.isLocked('h2', 'family')) {
                                _this.setFontParam('h2', 'family', '');
                            }
                            _this.resetDetails('h1');
                            _this.resetDetails('h2');
                        }
                    }
                }
            },
            {
                key: 'h2',
                name: 'H2',
                usage: 'h2',
                toCSS: function(part) {
                    return _this.toCSSForHeadings(part, 'h2')
                }
            },
            {
                key: 'h3',
                name: 'H3',
                usage: 'headings',
                toCSS: function(part) {
                    return _this.toCSSForHeadings(part, 'h3')
                }
            },
            {
                key: 'h4',
                name: 'H4',
                usage: 'headings',
                toCSS: function(part) {
                    return _this.toCSSForHeadings(part, 'h4')
                }
            },
            {
                key: 'h5',
                name: 'H5',
                usage: 'headings',
                toCSS: function(part) {
                    return _this.toCSSForHeadings(part, 'h5')
                }
            }
        ]

        this.fonts = this.fonts.filter(function(f) {
            if(_this.mode === 'css' && f.key === 'lead') return false;

            var is_h = f.key.length === 2;
            return _this.finetune ? is_h : !is_h;
        })

        var bm = new PgToggleButtonMaker();

        var defineFont = function(font) {

            var name, key;

            var i = font.key;
            var name = font.name;

            _this.addControl(`${i}_name`,
                {
                    name: 'Name',
                    type: 'hidden',
                    html: key,
                    end_section: true,
                    aastart_repeat_item: name,
                    aarepeat_item_open: true,
                    aaadd_repeat_item_label: 'Add heading level'
                }
            )

            var d = {
                name: name,
                type: 'select',
                show_empty: true,
                fill_list_on_open: true,
                can_add_items: true,
                options: function() {
                    return pinegrow.fontLibrary.getSelectOptions(font.usage);
                },
                dm_no_lock: font.surpriseMe ? false : 'placeholder'
            }

            if(font.label_before) {
                d.start_repeat_item = font.label_before;
                d.repeat_item_open = true;
                d.add_repeat_item_label = 'none';
            }

            _this.addControl(`${i}_family`,d);

            _this.addControl(`${i}_weight`,
                {
                    name: 'Weight',
                    type: 'select',
                    start_section: name + ' details',
                    show_empty: true,
                    fill_list_on_open: true,
                    can_add_items: true,
                    options: function() {
                        return pinegrow.fontLibrary.getSelectOptionsForWeight(_this.getFontParam(i, 'family'));
                    },
                    dm_no_lock: font.surpriseMe ? false : 'placeholder'
                }
            )

            /*
            _this.addControl(`font_${i}_height`,
                {
                    name: 'Line height',
                    type: 'slider',
                    slider_def_unit: ''
                }
            )

             */

            _this.addControl(`${i}_color`,
                {
                    name: 'Color',
                    type: 'color',
                    colors: function() {
                        return colors_skill.colors.getColorsForPicker();
                    },
                    dm_no_lock: font.surpriseMe ? false : 'placeholder'
                }
            )

            _this.addControl(`${i}_spacing`,
                {
                    name: 'Spacing',
                    type: 'slider',
                    slider_def_unit: 'px',
                    dm_no_lock: font.surpriseMe ? false : 'placeholder'
                }
            )

            if(i === 'h1' || true) {
                _this.addControl(`${i}_effect`,
                    {
                        name: 'Effect',
                        type: 'select',
                        show_empty: true,
                        options: function() {
                            var r = [];
                            $.each(_this.effects, function(key, d) {
                                r.push({key: key, name: d.name});
                            })
                            return r;
                        },
                        dm_no_lock: font.surpriseMe ? false : 'placeholder'
                    })
            }

            _this.addControl(`font_${i}_options_container`,
                {
                    name: 'Options',
                    type: 'displayhtmlwithlabel',
                    html: '',
                    class: `crsa-field-text-font-options-container crsa-field-text-font-options-container-${i}`,
                    dm_no_lock: font.surpriseMe ? false : 'placeholder'
                }
            )

            _this.addControl(`${i}_italic`,
                {
                    name: 'Italic',
                    type: 'select',
                    toggle_buttons: true,
                    append_to: `.crsa-field-text-font-options-container-${i}`,
                    without_text: true,
                    show_empty: true,
                    dm_no_lock: true,
                    options : [
                        {
                            key : 'italic',
                            name : 'Italic',
                            html : bm.makeIcon('style_italic', { attributes: { 'title': 'Italic' } })
                        }
                    ]
                }
            )

            _this.addControl(`${i}_smallcaps`,
                {
                    name: 'Small caps',
                    type: 'select',
                    toggle_buttons: true,
                    append_to: `.crsa-field-text-font-options-container-${i}`,
                    without_text: true,
                    show_empty: true,
                    dm_no_lock: true,
                    options : [
                        {
                            key : 'small-caps',
                            name : 'Small caps',
                            html : bm.makeText('Abc', {
                                styles: {'font-family': 'serif', 'font-size': '15px', 'font-variant': 'small-caps'},
                                attributes: {'title': 'Small caps'}
                            })
                        }
                    ]
                }
            )

            _this.addControl(`${i}_uppercase`,
                {
                    name: 'Uppercase',
                    type: 'select',
                    toggle_buttons: true,
                    append_to: `.crsa-field-text-font-options-container-${i}`,
                    without_text: true,
                    show_empty: true,
                    dm_no_lock: true,
                    options: [
                        {
                            key : 'uppercase',
                            name : 'Uppercase',
                            html : bm.makeIcon('Transform-uppercase', { attributes: { 'title': 'Uppercase' } })
                        }
                    ]
                }
            )



        }

        this.fonts.forEach(function(font) {
            defineFont(font);
        })

        if(!finetune) {
            this.addControl(`manage`,
                {
                    name: 'Manage fonts...',
                    type: 'button',
                    end_section: true,
                    func: function() {
                        pinegrow.fontLibrary.showLibrary();
                    }
                }
            )
        }

    }

    onSurpriseMe(done) {
        for(var i = 0; i < this.fonts.length; i++) {
            this.fonts[i].surpriseMe && this.fonts[i].surpriseMe(this.fonts[i]);
        }
        done();
    }

    onGetCSS(part, cs_url) {

        var _this = this;
        var code = '';

        switch(part) {

            case 'import':
                var fonts_helper = pgDmStore.design.fonts;
                if(!_this.finetune) fonts_helper.clear();

                for(var i = 0; i < _this.fonts.length; i++) {
                    var family = _this.getFontParam(_this.fonts[i].key, 'family');
                    if(family) {
                        var info = {
                            key: _this.fonts[i].key,
                            family: family,
                            weight: _this.getFontParam(_this.fonts[i].key, 'weight'),
                            italic: _this.getFontParam(_this.fonts[i].key, 'italic')
                        }
                        fonts_helper.setFont(info);
                    }
                }

                var css = '';

                if(_this.finetune) {
                    var imports = fonts_helper.getImportStatements(cs_url);
                    css += imports.length ? (imports.join("\n") + "\n\n") : '';

                    var faces = fonts_helper.getFontFontFaceStatements();
                    css += faces.length ? (faces.join("\n\n") + "\n\n") : '';
                    return css;
                } else {
                    return ''; //two steps, return all fonts in finetune stage
                }

            case 'init':
            case 'rules':
                for(var i = 0; i < _this.fonts.length; i++) {
                    code += _this.fonts[i].toCSS(part)
                }
                break;

        }
        return code;
    }

    getActiveFonts(list) {
        var r = list || []
        for(var i = 0; i < this.fonts.length; i++) {
            var family = this.getFontParam(this.fonts[i].key, 'family');
            if(family && r.indexOf(family) < 0) {
                r.push(family);
            }
        }
        return r;
    }

    getFontParam(i, p) {
        return this.getParam(`${i}_${p}`);
    }

    isLocked(key, p) {
        return super.isLocked(`${key}_${p}`);
    }

    setFontParam(key, p, v) {
        return this.setParam(`${key}_${p}`, v);
    }

    getFontParams(key) {
        var r = {}
        var fk = `${key}_`;
        $.each(this.params, function(f,v) {
            if(f.startsWith(fk)) {
                r[f.replace(fk,'')] = v;
            }
        })
        return r;
    }

    addEffect(key, list) {
        var effect = this.getFontParam(key, 'effect');
        if(effect && this.effects[effect]) {
            list.push(this.effects[effect].css)
        }
        return list;
    }

    getCSSDetailsProps(css) {
        return [
            css.p('italic', 'font-style'),
            css.p('smallcaps', 'font-variant'),
            css.p('uppercase', 'text-transform'),
            css.p('spacing', 'letter-spacing')
        ]
    }

    toCSSForHeadings(part, key) {
        var css = new PgDmCSSCodeMaker(this.getFontParams(key));

        if(part === 'init') {

        } else if(part === 'rules') {
            var imp = true;

            var family = this.getFontParam(key, 'family');
            if(family) {
                var f = pinegrow.fontLibrary.getFontByFamily(family);
                if(f) {
                    family = f.getFamilyWithFallback();
                }
            }

            var rules = [
                css.p('family', 'font-family', imp, family),
                css.p('weight', 'font-weight', imp),
                css.p('height', 'line-height', imp),
                this.mode === 'sass' ? css.p('color', 'color', imp) : css.cssvar('color', 'color', imp)
            ]

            this.addEffect(key, rules);

            css.rule(key, rules.concat(this.getCSSDetailsProps(css)))
        }
        return css.css;
    }

    resetDetails(key) {
        var _this = this;
        var list = ['weight', 'color', 'spacing', 'effect', 'italic', 'smallcaps', 'uppercase'];
        list.forEach(function(prop) {
            if(!_this.isLocked(key, prop)) {
                _this.setFontParam(key, prop, '');
            }
        })
    }
}

class PgDmDesignSkillBackground extends PgDmDesignSkill {
    constructor(colors_skill, debug) {
        super('background', 'Background');

        this.colors_skill = colors_skill;

        var _this = this;

        this.addControl('color',
            {
                name: 'Color',
                type: 'color',
                colors: function() {
                    return colors_skill.colors.getColorsForPicker();
                }
            }
        )

        var picked_url = null;

        this.addControl('image',
            {
                name: 'Image',
                type: 'image',
                live_update: false,
                file_picker_use_project_as_parent: true,
                file_picker_use_image_menu: true,
                /*
                file_picker_parent_url: function() {
                    return pinegrow.getCurrentProject().getUrl() + '/dummy.html'
                },

                 */

                /*
                file_picker_use_image_menu: function(actions, on_set_value) {
                    actions.push({
                        label: 'Surprise me!',
                        action: function() {
                            pgDmStore.unsplash.getRandomPhoto({
                                //featured: true,
                                query: pinegrow.getPhotoBrowserSearchTerm() || 'wallpaper'
                            }, function(unp) {
                                var src = pgDmStore.unsplash.getImageUrl(unp);
                                on_set_value(src);
                            })
                        }
                    })
                },

                 */
                /*
                file_picker_on_picked: function(url, obj, unp) {
                    picked_url = url;

                    _this.setParam('image_sm', unp ? pgDmStore.unsplash.getImageUrl(unp, 'custom', {w: 640}) : null);
                    _this.setParam('image_lg', unp ? pgDmStore.unsplash.getImageUrl(unp, 'custom', {w: 2400}) : null);
                    _this.getField('image_sm').setValue( _this.getParam('image_sm'));
                    _this.getField('image_lg').setValue( _this.getParam('image_lg'));

                    return url;
                },*/
                on_changed: function(obj, prop, url, oldValue, fieldDef, $field, eventType) {
                    if(eventType === 'change') {
                        url && _this.getColorsFromImage(url, function () {
                            picked_url = url;
                            pgDmStore.update();
                            pgDmStore.designPanel.updatePanelDisplay();
                        });
                    }
                }
            }
        )

        this.addControl('image_for_colors', {
            name: 'Only use the image for colors',
            type: 'checkbox',
            value: 'true',
            dm_no_lock: true,
            start_section: 'Image options',
            helptext: `<p>Check to let Surprise Me use the random image for getting colors without setting the image as the background of the page.</p>`
        })

        this.addControl('selector', {
            name: 'Selector',
            type: 'text',
            default_value: 'body',
            placeholder: 'body',
            dm_no_lock: true,
            live_update: false,
            helptext: `<p>Selector for the HTML element to which to apply the background image. Default is <code>body</code>.</p><p>To set the src attribute of an image element on the current page, use the selector that starts with <code>img</code>, for example <code>img.poster</code>.</p><p>You can list multiple selectors, separated with commas.</p>`,
            show_if: function() {
                return _this.getParam('image_for_colors') !== 'true'
            }
        })

        this.addControl('advanced', {
            name: 'Support fixed bck. on mobile &amp; filters',
            type: 'checkbox',
            value: 'true',
            negvalue: 'false',
            default_value: 'true',
            dm_no_lock: true,
            helptext: `<p>Set the background image with the <code>:before</code> pseudo element, so that the fixed image on the body works in mobile Safari, and filters can be used. The <code>z-index</code> of the target element is set to <code>0</code> and if the target element is not Body, it must have position relative, absolute or fixed.</p><p>Note, these settings might cause conflicts with the existing CSS styling. If the conflicts can not be resolved, disable this option.</p>`,
            show_if: function() {
                //var sel = _this.getParam('selector') || 'body';
                return _this.getParam('image_for_colors') !== 'true';
            }
        })

        this.addControl('source', {
            name: 'Source',
            type: 'select',
            default_value: 'unsplash',
            options: function() {
                var r = [
                    {key: 'unsplash', name: 'Free online images'}
                ];
                if(isApp()) {
                    r.push({key: 'folder', name: 'Folder'})
                }
                return r;
                },
            dm_no_lock: true,
            helptext: `<p>Select the source of the random image.</p>`
        })

        this.addControl('keyword', {
            name: 'Query',
            type: 'text',
            placeholder: 'wallpaper',
            dm_no_lock: true,
            live_update: false,
            show_if: function() {
                return _this.getParam('source') !== 'folder'
            },
            helptext: `<p>Search query for random online images, for example vivid or house. Default is wallpaper.</p>`
        })

        this.addControl('folder', {
            name: 'Folder',
            type: 'text',
            file_picker: true,
            file_picker_no_url: true,
            file_picker_folder: true,
            dm_no_lock: true,
            live_update: false,
            show_if: function() {
                return _this.getParam('source') === 'folder'
            },
            helptext: `<p>Select the local folder on your computer that contains the images.</p>`,
            on_changed : function(pgel, prop, value, oldValue, fieldDef, $field, eventType, values, crsaPage) {
                if(value) {
                    if(!pinegrow.getCurrentProject().isFileInProject(value)) {
                        pinegrow.showAlert(`<p>The selected folder is located outside of the project.</p><p>Selected images will show up when editing the project in Pinegrow, but will not work in the browser. To make the images show up, copy the selected images into a subfolder within the project and include them in the design from there.</p>`, 'A note about external images', null, 'Got it!');
                    }
                }
            }
        })

        this.addControl('image_sm',
            {
                name: 'Small Image',
                type: 'image',
                live_update: false,
                file_picker_use_project_as_parent: true,
                file_picker_use_image_menu: true,
                dm_no_lock: true,
                helptext: `<p>URL of the image to use on small screens (less than 768px). Use Settings to customize the media query.</p>`,
                show_if: function() {
                    return _this.getParam('image_for_colors') !== 'true'
                }
            }
        )


        this.addControl('image_lg',
            {
                name: 'Large Image',
                type: 'image',
                live_update: false,
                file_picker_use_project_as_parent: true,
                file_picker_use_image_menu: true,
                dm_no_lock: true,
                helptext: `<p>URL of the image to use on large screens (larger than 1440px). Use Settings to customize the media query.</p>`,
                show_if: function() {
                    return _this.getParam('image_for_colors') !== 'true'
                }
            }
        )



        this.default_mq_small = '@media (max-width: 767px)';
        this.default_mq_large = '@media (min-width: 1440px), (min-width: 992px) and (-webkit-min-device-pixel-ratio: 2), (min-width: 992px) and (min-resolution: 192dpi)';

        this.addControl('button.options', {
            type: 'button',
            name: 'Settings...',
            button_class: 'pg-dm-bck-settings-button',
            func: function () {
                var provider = pgDmStore.getActiveDesignProvider();

                var settings = provider.settings;

                var fields = {
                    'background.media_small': {
                        name: 'Small media query',
                        type: 'text',
                        value: settings['background.media_small'] || _this.default_mq_small,
                        placeholder: _this.default_mq_small,
                        helptext: `<p>Media query for displaying the small background image.</p>`
                    },
                    'background.media_large': {
                        name: 'Large media query',
                        type: 'text',
                        value: settings['background.media_large'] || _this.default_mq_large,
                        placeholder: _this.default_mq_large,
                        helptext: `<p>Media query for displaying the large background image. Default media query for the large size detects the retina screen and uses the large image if pixel size is 992px or larger.</p>`
                    }
                }

                pinegrow.showQuickDialog('Background settings', `Media queries`, `<p>Set the media queries for displaying small and large background image.</p><p>Note, these settings are stored in the project settings and therefore apply to all designs in the project.</p>`, fields, 'Cancel', 'Save', null, function (values) {
                    provider.setSettings(values);
                })
            }
        })

        this.filters = [
            {
                filter: 'blur',
                name: 'Blur',
                min: 0,
                max: 20,
                unit: 'px',
                start_section: 'Filters',
                end_section: true
            },
            {
                filter: 'brightness',
                name: 'Brightness',
                min: 0,
                max: 200,
                unit: '%',
                step: 20,
                def: 100
            },
            {
                filter: 'contrast',
                name: 'Contrast',
                min: 0,
                max: 200,
                unit: '%',
                step: 20,
                def: 100
            },
       /*     {
                filter: 'drop-shadow',
                type: 'slider',
                unit: ''
            },*/
            {
                filter: 'grayscale',
                name: 'Grayscale',
                min: 0,
                max: 100,
                unit: '%',
                step: 1,
                def: 0
            },
            {
                filter: 'hue-rotate',
                name: 'Hue rotate',
                min: 0,
                max: 360,
                unit: 'deg',
                step: 1,
                def: 0
            },
            {
                filter: 'invert',
                name: 'Invert',
                min: 0,
                max: 100,
                unit: '%',
                step: 1,
                def: 0
            },
            {
                filter: 'opacity',
                name: 'Opacity',
                min: 0,
                max: 100,
                unit: '%',
                step: 1,
                def: 100
            },
            {
                filter: 'saturate',
                name: 'Saturate',
                min: 0,
                max: 200,
                unit: '%',
                step: 20,
                def: 100
            },
            {
                filter: 'sepia',
                name: 'Sepia',
                min: 0,
                max: 100,
                unit: '%',
                step: 1,
                def: 0
            }
        ]

        this.addControl('filter_notice', {
            type: 'displayhtml',
            dm_no_lock: true,
            start_section: 'Filters',
            end_section: true,
            name: '<p style="text-align: left;">Disable "Use image only for colors" checkbox and enable "Mobile &amp; filters" checkbox in Image options to use the filters.</p>',
            show_if: function() {
                return _this.getParam('advanced') !== 'true' || _this.getParam('image_for_colors') === 'true';
            }
        })

        this.filters.forEach(function(f) {
            _this.addControl('filter_' + f.filter, {
                name: f.name,
                type: f.type || 'range',
                range_min: f.min,
                range_max: f.max,
                range_step: f.step || 1,
                slider_def_unit: f.unit || '',
                dm_no_lock: true,
                default_value: f.def || 0,
                //start_section: f.start_section || null,
                //end_section: f.end_section || null,
                helptext: `<p>Apply the filter ${f.filter} to the background image.</p>`,
                show_if: function() {
                    //var sel = _this.getParam('selector') || 'body';
                    return _this.getParam('advanced') === 'true' && _this.getParam('image_for_colors') !== 'true';
                }
            })
        })

        this.addControl('button.filter_defaults', {
            type: 'button',
            name: 'Reset defaults',
            button_class: 'pg-dm-bck-settings-button',
            func: function () {
                pgDmStore.db.makeChanges(null, 'Reset image filters', function() {
                    _this.filters.forEach(function(f) {
                        _this.setParam('filter_' + f.filter, ('def' in f) ? f.def : 0);
                    })
                }, 'design_panel');
                pinegrow.showQuickMessage('Image filters were reset.');
                pgDmStore.update();
                pgDmStore.designPanel.updatePanelDisplay();
            },
            show_if: function() {
                //var sel = _this.getParam('selector') || 'body';
                return _this.getParam('advanced') === 'true' && _this.getParam('image_for_colors') !== 'true';
            }
        })
    }

    getSmallMediaQuery() {
        var provider = pgDmStore.getActiveDesignProvider();
        var m = provider.settings['background.media_small'] || this.default_mq_small;
        if(!m.startsWith('@media')) {
            m = '@media' + m;
        }
        return m;
    }

    getLargeMediaQuery() {
        var provider = pgDmStore.getActiveDesignProvider();
        var m = provider.settings['background.media_large'] || this.default_mq_large;
        if(!m.startsWith('@media')) {
            m = '@media' + m;
        }
        return m;
    }

    populateDefaultParams() {
        var _this = this;
        this.setParam('image_for_colors', 'true');
        this.setParam('advanced', 'true');
        this.filters.forEach(function(f) {
            _this.setParam('filter_' + f.filter, f.def || 0);
        })
    }

    onSurpriseMe(done) {
        var _this = this;
        var scope = 'body';//this.getParam('scope');
        var source = this.getParam('source') || 'unsplash';

        if(!this.isLocked('image')) {
            if(scope === 'element' && !pinegrow.selectedElements.getLastSelected()) {
                pinegrow.showQuickError(`Can't set background image because no element is selected.`)
                done();
            }
            var query = this.getParam('keyword') || 'wallpaper';
            var image_size = this.getParam('image_for_colors') === 'true' ? 'small' : null;

            image_size = null;

            if(source === 'unsplash') {
                pgDmStore.unsplash.getRandomPhoto({
                    query: query //pinegrow.getPhotoBrowserSearchTerm() || 'wallpaper'
                }, function (unp) {
                    var src = pgDmStore.unsplash.getImageUrl(unp, 'custom', {w: 1200});
                    _this.setParam('image', src);
                    _this.getColorsFromImage(src, done);

                }, function () {
                    pinegrow.showQuickError('Unable to get a random photo!');
                    done && done();
                })
            } else {
                var folder = this.getParam('folder');
                if(folder) {
                    pgDmStore.localImages.getRandomImageFromFolder(folder, function(src) {
                        if(src) {
                            _this.setParam('image', src);
                            _this.getColorsFromImage(src, done);
                        } else {
                            done && done();
                        }
                    })
                }
            }
        } else {
            done();
        }
    }

    getColorsFromImage(url, done) {
        var _this = this;
        if(this.colors_skill.areAllColorsLocked()) {
            done();
        } else {
            var img = new Image();
            img.onload = function () {
                _this.colors_skill.setColorsFromImage(img, function () {
                    done();
                }, null)
            }
            img.onerror = function() {
                done && done();
            }
            if(!isApp()) {
                img.crossOrigin = "anonymous";
            }
            img.src = url;
        }
    }

    onGetCSS(part, cs_url) {
        var code = '';

        switch(part) {

            case 'init':
                if (this.getParam('color')) {
                    if(this.onGetBackgroundColorCSS) {
                        //code = this.onGetBackgroundColorCSS(this.getParam('color'));
                    } else {
                        code += this.setVar('body-bg', this.getParam('color'));
                    }
                }
                break;

            case 'rules':
                if (this.getParam('color')) {
                    if(this.onGetBackgroundColorCSS) {
                        code = this.onGetBackgroundColorCSS(this.getParam('color'));
                    } else {
                        //code += this.setVar('body-bg', this.getParam('color'));
                    }
                }

                var project = pinegrow.getCurrentProject();

                var makeUrl = function(url) {
                    if(crsaIsRemoteUrl(url)) return url;
                    url = project.getAbsoluteUrl(url);
                    if(project.isUrlInProject(url)) {
                        return crsaMakeLinkRelativeTo(url, cs_url);
                    } else {
                        return url;
                    }
                }

                var selector = this.getParam('selector') || 'body';

                var adv_mode = this.getParam('advanced') === 'true';

                var _this = this;

                var doImageForSelector = function(selector) {
                    var on_body = selector === 'body';
                    var set_relative = false;

                    var imp = '!important';

                    var filters = [];

                    _this.filters.forEach(function (f) {
                        var val = _this.getParam('filter_' + f.filter);
                        if (val && val != f.def) {
                            filters.push(`${f.filter}(${val}${f.unit})`)
                        }
                    })

                    //console.log(filters)

                    if (_this.getParam('image') && !_this.getParam('image_for_colors')) {
                        var image = _this.getParam('image');

                        var image_src_mode = false;

                        if (selector.match(/^img($|[^a-z])/)) {
                            //set src on image
                            var mode_desc = 'Image element mode: ';
                            var page = pinegrow.getSelectedPage();
                            if (!page) {
                                pinegrow.showQuickError(mode_desc + 'Page is not open.');
                            } else {
                                var images = page.sourceNode.find(selector);
                                if (images.length === 0) {
                                    pinegrow.showQuickError(mode_desc + `No images with the selector ${selector} were found.`);
                                } else {
                                    var col = new PgCollection();
                                    col.setList(images);
                                    var image_rel_url = page.makeRelativeUrl(image);
                                    image_src_mode = true;

                                    var do_set = false;
                                    col.forEachElement(function (pgel) {
                                        if (pgel.getAttribute('src') !== image_rel_url) {
                                            do_set = true;
                                        }
                                    })
                                    if (do_set) {
                                        var api = new PgApi();
                                        api.setAttribute(col, 'src', image_rel_url);
                                    }
                                }
                            }
                        }

                        if (adv_mode) {
                            var main_selector = selector;
                            if (!image_src_mode) {
                                selector = selector + ':before';
                            }

                            if (!on_body && !image_src_mode) {
                                //warn if element is position static
                                var page = pinegrow.getSelectedPage();
                                if (page) {
                                    var $el = page.get$Html()?.find(main_selector);
                                    if (!$el || $el.length === 0) {

                                    } else {
                                        if ($el.css('position') === 'static' && !$el.is('body,html')) {
                                            //warn instead
                                            pinegrow.showQuickError(`The background element <code>${main_selector}</code> should have position set to <code>relative</code>, <code>absolute</code> or <code>fixed</code>.`)
                                            //set_relative = true;
                                        }
                                    }
                                }
                            }

                            //we need to establish a new stacking context on the main element, so that we can push bck to -1
                            if (!image_src_mode) {
                                code += `
${main_selector} {
    z-index: 0;
`;
                                if (set_relative) {
                                    code += `    position: relative;
`;
                                }
                                code += `}
`
                            }
                        }

                        if(!image_src_mode || filters.length) {
                            code += `
${selector} {`;

                            if (!image_src_mode) {
                                code += `
    background-image: url(${makeUrl(image)})${imp};
    background-size: cover;
    background-position: center center;
    background-repeat: no-repeat;
`

                                if (adv_mode) {
                                    code += `    content: ' ';
    top:0;
    left:0;
    width:100%;
    height:100%;
    z-index:-1;
`;
                                    if (on_body) {
                                        code += `    position: fixed;
`;
                                    } else {
                                        code += `    position: absolute;
`;
                                    }

                                } else {

                                    if (on_body) {
                                        code += `    background-attachment: fixed;
`;
                                    }
                                }
                            }

                            if (filters.length && (adv_mode || image_src_mode)) {
                                code += `    filter: ${filters.join(' ')};
`;
                            }
                            code += `}
`;
                        }

                        if (!image_src_mode && !on_body && adv_mode && !selector.match(/^[a-z]/i) && selector.indexOf(',') < 0) {
                            //if not body and selector is not a tag, add an additional rule with fixed
                            code += `
body${selector} {
position: fixed;
}                        
`;
                        }

                        var image_sm = _this.getParam('image_sm');
                        var image_lg = _this.getParam('image_lg');
                        var image_xl = null;

                        if (image.indexOf('//images.unsplash.com') >= 0) {
                            if (!image_sm) {
                                //image_sm = image.replace(/\&h=[0-9]+/, '');
                                //image_sm = image_sm.replace(/\&w=[0-9]+/, '&w=640');
                            }
                            if (!image_lg) {
                                image_lg = image.replace(/\&h=[0-9]+/, '');
                                image_lg = image_lg.replace(/\&w=[0-9]+/, '&w=2000');
                            }
                        }

                        if(!image_src_mode) {
                            if (image_sm) {
                                code += `${_this.getSmallMediaQuery()} {
    ${selector} {                            
        background-image: url(${makeUrl(image_sm)})${imp};   
    }                             
}
`
                            }


                            if (image_lg) {
                                code += `${_this.getLargeMediaQuery()} {
    ${selector} {                        
        background-image: url(${makeUrl(image_lg)})${imp};         
    }                           
}
`
                            }
                        }

                    }
                }

                selector.split(',').forEach(function(sel_item) {
                    doImageForSelector(sel_item.trim())
                })

                if(true || code) {
                    code += `body, html {
    min-height: 100vh;
}\n`
                }

                break;

        }
        return code;
    }
}

