class PgFontType {

    constructor(node, family) {
        this.type = 'abstract';
        this.family = family;
        this.weights = [];
        this.styles = [];
        this.category = null;
        this.error = null;
        this.url = null;
        this.can_refresh = false;
        this.usage = 'all';
        this.fallback = null;

        if(node) {
            this.fromNode(node);
        }
    }

    getTypeLabel() {
        return this.type;
    }

    isUsedFor(usage) {
        if(usage === 'h2') usage = 'h1';
        if(!usage || !this.usage || this.usage === 'all' || (this.usage === 'headings' && usage === 'h1') || this.usage === usage) return true;
        return false;
    }

    addStyle(style) {
        if(this.styles.indexOf(style) < 0) {
            this.styles.push(style);
        }
    }

    addWeight(w) {
        if(this.weights.indexOf(w) < 0) {
            this.weights.push(w);
        }
    }

    describeStyles() {
        return this.weights.join(' ');
    }

    getRandomWeight(use) {
        if(this.weights.length) {
            return pgGetRandomItem(this.weights)
        }
    }

    getPreview(text) {
        return `<span style="font-family:${this.family}">${text || this.family}</span>`;
    }

    getCode() {
        return {
            import: null,
            import_url: null,
            css: null,
            js: null
        }
    }

    getFamilyWithFallback() {
        var family = this.family;
        if(family.indexOf('"') < 0 && family.indexOf("'") < 0 && family.indexOf(',') < 0) {
            family = `'${family}'`;
        }

        if(!this.fallback) {
            return family;
        } else {
            return family + ', ' + this.fallback;
        }
    }

    addToPageView($html) {
        var $head = $html.find('head');

        var d = this.getCode();
        if(d.url) {
            var key = this.family;
            var $link = $head.find(`link[data-pg-temp-font="${key}"]`);
            if($link.length === 0) {
                var url = crsaIsRemoteUrl(d.url) ? d.url : pinegrow.getProxyUrl(d.url);
                if($head.find(`link[href="${url}"]`).length === 0) {
                    var html = `<link href="${url}" rel="stylesheet" data-pg-temp-font="${key}"/>`;
                    $head.append(html);
                }
            }
        }
        if(d.css) {
            var key = this.family;
            $head.find(`style[data-pg-temp-font="${key}"]`).remove();
            var html = `<style data-pg-temp-font="${key}">${d.css}</style>`;
            $head.append($(html));
        }
    }

    activateInProject(p) {
        this.addToPageView($('html'));
    }

    deactivateInProject(p) {

    }

    toNode(node) {
        node.type = this.type;
        node.family = this.family;
        node.weights = this.weights;
        node.styles = this.styles;
        node.category = this.category;
        node.url = this.url;
        node.usage = this.usage;
        node.fallback = this.fallback;

        /*
        node.setAttribute('type', this.type);
        node.setAttribute('family', this.family);
        node.setAttribute('weights', this.weights.join(' '));
        node.setAttribute('styles', this.styles.join(' '));

         */
    }

    fromNode(node) {
        this.family = node.family;
        this.weights = node.weights || [];
        this.styles = node.styles || [];
        this.category = node.category;
        this.url = node.url;
        this.usage = node.usage || 'all';
        this.fallback = node.fallback;

        /*
        this.family = node.getAttribute('family');
        this.weights = (node.getAttribute('weights') || '').split(' ');
        this.styles = (node.getAttribute('styles') || '').split(' ');

         */
    }
}


/*
Local is now derived from remote

class PgFontTypeLocal extends PgFontType {
    constructor(node, folder, family) {
        super(node);

        this.type = 'local';

        if(!node) {
            this.folder = folder;
            this.url = crsaMakeUrlFromFile(folder);
        }

        this.stylesheet_url = new URL('stylesheet.css', this.url + '/').toString();
        var file = require('path').join(this.folder, 'stylesheet.css');

        if (!crsaIsFileExist(file)) {
            this.error = `The file ${file} is not found.`;
        } else {
            if (family) {
                this.family = family;
            } else {
                //detect family name
                try {
                    var css = require('fs').readFileSync(file, {encoding: 'utf8'});
                    var m = css.match(/font\-family:\s*['"](.*)['"]/m);

                    if (m) {
                        this.family = m[1];
                    } else {
                        this.error = 'Unknown font family';
                    }
                } catch (err) {
                    this.error = err.toString();
                }
            }
        }
    }

    getCode() {
        return {
            import: `@import url('${this.stylesheet_url}');`,
            url: this.stylesheet_url,
            css: null,
            js: null
        }
    }

    toNode(node) {
        super.toNode(node);
    }

    fromNode(node) {
        super.fromNode(node);
        this.folder = crsaMakeFileFromUrl(this.url);
    }
}


 */

class PgFontTypeSystem extends PgFontType {
    constructor(node, family) {
        super(node, family);
        this.type = 'system';
    }
}

class PgFontTypeUploadedWebfont extends PgFontType {
    constructor(node, family) {
        super(node, family);
        this.type = 'uploaded';
        this.max_variants = 10;

        if(!node) {
            this.variants = [];
            this.is_variable = false;
        }
    }

    getTypeLabel() {
        return 'self-hosted';
    }

    fromNode(node) {
        var _this = this;
        this.family = node.family;
        this.node = node;
        this.weights = [];
        this.styles = [];
        this.is_variable = node.is_variable || false;

        node.variants.forEach(function(v) {
            if(_this.weights.indexOf(v.weight) < 0) {
                _this.weights.push(v.weight);
            }
            if(_this.styles.indexOf(v.style) < 0) {
                _this.styles.push(v.style);
            }
        })

        this.variants = node.variants || [];

        this.category = null;//node.category;
        this.url = null;
        this.usage = node.usage || 'all';
        this.fallback = node.fallback;
    }

    toNode(node) {
        node.type = this.type;
        node.family = this.family;
        node.fallback = this.fallback;
        node.usage = this.usage;
        node.variants = this.variants;
        node.is_variable = this.is_variable;
    }

    edit(on_save) {
        var _this = this;
        var fields = {}
        var values = {
            family: this.family,
            fallback: this.fallback
        }

        var max_variants = this.max_variants;

        var vi = 0;

        var d = new PgDialog('Add or edit a self-hosted web font');

        d.sections = {
            name: {
                name: 'Font family name',
                is_closable: false,
                fields: {
                    family: {
                        name: 'Font family name',
                        type: 'text',
                        placeholder: 'for example, My Cool Font',
                        helptext: 'The font family name is used to reference the font in CSS rules.',
                        validate: function(obj, prop, value, fieldDef, $field, values) {
                            if(!value) {
                                return pinegrow.getValidationError(fieldDef.name, 'req');
                            }
                            return null;
                        }
                    },
                    fallback: {
                        name: 'Fallback font family',
                        type: 'text',
                        helptext: 'Set the font that will be used if for some reason the web font is not available',
                        placeholder: 'for example, sans-serif'
                    },
                    /*
                    is_variable: {
                        name: 'This is a variable font',
                        type: 'checkbox',
                        value: '1',
                        helptext: ''
                    }*/
                }
            },
        }

        for(var i = 1; i <= max_variants; i++) {
            var variant = (vi < this.variants.length) ? this.variants[vi] : null;

            var sec_fields = {}

            values['weight_' + i] = variant ? variant.weight : null;
            values['style_' + i] = variant ? variant.style : null;
            values['file_woff2_' + i] = variant ? variant.woff2 : null;
            values['file_woff_' + i] = variant ? variant.woff : null;
            values['file_ttf_' + i] = variant ? variant.ttf : null;
            values['variable_' + i] = variant ? variant.variable : null;
            values['extras_' + i] = variant ? variant.extras : null;

            sec_fields['weight_' + i] = {
                name: 'Weight',
                type: 'select',
                options: [
                    {key: 'normal', name: 'normal'},
                    {key: '100', name: '100 (Thin)'},
                    {key: '200', name: '200 (Extra Light)'},
                    {key: '300', name: '300 (Light)'},
                    {key: '400', name: '400 (Normal)'},
                    {key: '500', name: '500 (Medium)'},
                    {key: '600', name: '600 (Semi Bold)'},
                    {key: '700', name: '700 (Bold)'},
                    {key: '800', name: '800 (Extra Bold)'},
                    {key: '900', name: '900 (Black)'},
                ],
                helptext: 'The font weight for which this font variant should be used.'
            };
            sec_fields['style_' + i] = {
                name: 'Style',
                type: 'select',
                options: [
                    {key: 'normal', name: 'normal'},
                    {key: 'italic', name: 'italic'},
                ],
                helptext: 'The font style for which this font variant should be used.'
            }
            sec_fields['file_woff2_' + i] = {
                name: 'WOFF2 file url',
                type: 'text',
                helptext: `<p>Enter the URL of the .woff2 file that is uploaded to the Media library. It is best to use relative URLs.</p>`,
                placeholder: 'WOFF2 is supported by all modern browsers and offers better compression than woff.'
            }
            sec_fields['file_woff_' + i] = {
                name: 'WOFF file url',
                type: 'text',
                helptext: `<p>Enter the URL of the .woff file that is uploaded to the Media library. It is best to use relative URLs.</p>`,
                placeholder: 'WOFF is supported by all modern browsers.'
            }
            sec_fields['file_ttf_' + i] = {
                name: 'TTF file url',
                type: 'text',
                helptext: `<p>Enter the URL of the .ttf file that is uploaded to the Media library. It is best to use relative URLs.</p>`,
                placeholder: 'Optionally, add TTF as a fallback for older browsers.'
            }

            d.sections['variant_' + i] = {
                name: 'Font variant ' + i,
                is_closable: true,
                closed: !variant,
                fields: sec_fields
            }
            vi++;
        }

        d.values = values;
        d.okButtonLabel = 'Save';
        d.cancelButtonLabel = 'Cancel';
        d.validate_on_submit = true;

        d.before = `<p>Add a web font that is hosted on your site. Enter the font family name and then add up to ten font variants for different weights and styles.</p><p>Each font variant has a font file in one or more font formats, that <b>needs to be uploaded into the Media library beforehand</b>.</p><p><a href="https://css-tricks.com/understanding-web-fonts-getting/" class="external">Learn more about different font formats.</a></p>`;

        d.onOK = function() {
            var values = d.values;
            var node = {
                family: values.family,
                fallback: values.fallback,
                variants: []
            }

            var formats = ['woff2', 'woff', 'ttf'];

            for(var i = 1; i <= _this.max_variants; i++) {
                var has = false;
                formats.forEach(function(f) {
                    if(values['file_' + f + '_' + i]) {
                        has = true;
                    }
                })
                if(has) {
                    node.variants.push({
                        weight: values['weight_' + i] || 'normal',
                        style: values['style_' + i] || 'normal',
                        ttf: values['file_ttf_' + i] || null,
                        woff: values['file_woff_' + i] || null,
                        woff2: values['file_woff2_' + i] || null,
                    })
                }
            }

            _this.fromNode(node);
            on_save && on_save();
        }

        d.show();
    }

    getCode() {
        var css = '';
        var values = this.node;
        this.variants.forEach(function(v) {
            var srcs = [];

            if(v.woff2) {
                srcs.push(`url('${v.woff2}') format('woff2')`);
            }
            if(v.woff) {
                srcs.push(`url('${v.woff}') format('woff')`);
            }
            if(v.ttf) {
                srcs.push(`url('${v.ttf}') format('truetype')`);
            }

            css += `
@font-face {
    font-family: '${values.family}';
    src: ${srcs.join(', ')};
    font-weight: ${v.weight};
    font-style: ${v.style};
}`;
        })

        return {
            import: null,
            url: null,
            css: css,
            js: null
        }
    }
}

class PgFontTypeRemote extends PgFontType {
    constructor(node, url, family) {
        super(node);

        this.type = 'remote';
        this.can_refresh = true;

        if(!node) {
            this.url = url;
            this.family = family;
        }
    }

    getCode() {
        return {
            import: `@import url('${this.url}');`,
            url: this.url,
            css: null,
            js: null
        }
    }
}

class PgFontTypeLocal extends PgFontTypeRemote {
    constructor(node, url, family) {
        super(node, url, family);

        this.type = 'local';
    }

    copyToProjectIfNeeded() {
        if(!isApp()) return;
        var p = pinegrow.getCurrentProject();
        if(p.isUrlInProject(this.url)) return;

        var path = require('path');
        var fs = require('fs');

        var folder = p.getAbsolutePath('fonts/' + this.family);
        var css_name = crsaGetNameFromUrl(this.url);
        var new_url = p.getAbsoluteUrl('fonts/' + encodeURIComponent(this.family) + '/' + encodeURIComponent(css_name));
        var _this = this;

        try {
            crsaCreateDirs(folder, fs);
            crsaCopyFolderSync(path.dirname(crsaMakeFileFromUrl(this.url)), folder);
            _this.url = new_url;
            pinegrow.fontLibrary.save();
            pinegrow.refreshCurrentProjectDelayed(true);
            return true;
        } catch(error) {
            pinegrow.showQuickError('Could not copy local fonts to the project folder.');
            console.error(error);
        }
    }

    getUrl() {
        this.copyToProjectIfNeeded();
        return this.url;
        /*
        var p = pinegrow.getCurrentProject();
        var rel_url = p.getRelativeUrl(this.url);
        return rel_url;

         */
    }
}

class PgFontTypeAdobe extends PgFontTypeRemote {
    constructor(node, url, family) {
        super(node, url, family);

        this.type = 'adobe';
    }
}


class PgFontTypeGoogle extends PgFontType {
    constructor(node, gf_obj) {
        super(node);

        this.type = 'google';

        var variants_map = {
            regular: '400',
            thin: '100',
            bold: '700'
        }

        if(!node) {
            this.family = gf_obj.name;
            this.category = gf_obj.category;

            var _this = this;

            this.styles = [];
            this.weights = [];

            (gf_obj.wanted_styles || []).forEach(function(v) {
                v = variants_map[v] || v;
                var m = v.match(/^([0-9]+)/);
                var w;
                if(!m) {
                    v = '400' + v;
                    w = '400';
                } else {
                    w = m[1];
                }
                _this.styles.push({
                    weight: w,
                    italic: v.indexOf('italic') >= 0
                })

                if(_this.weights.indexOf(w) < 0) {
                    _this.weights.push(w);
                }
            });

            _this.styles.sort(function(a, b) {
                return a.weight.localeCompare(b.weight);
            })
        }

        switch (this.category) {
            case 'sans':
                this.fallback = 'sans-serif';
                break;
            case 'serif':
                this.fallback = 'serif';
                break;
            default:
                this.fallback = 'sans-serif';
        }

        this.url = this.url || `https://fonts.googleapis.com/css?family=${this.getGoogleFontUrlToken()}`;
    }

    getGoogleFontUrlToken() {
        var t = this.family.replace(/ /g, '+');
        if(this.styles.length) {
            t += ':';
            t += this.styles.map(function(s) {
                return s.weight + (s.italic ? 'italic' : '')
            }).join(',');
        }
        return t;
    }

    describeStyles() {
        return this.styles.map(function(s) {
            return s.weight + (s.italic ? 'i' : '')
        }).join(' ');
    }

    getCode() {
        return {
            import: `@import url('${this.url}');`,
            import_url: this.url,
            url: this.url,
            css: null,
            js: null
        }
    }
}

//Collection of fonts activated for the project
var PgFontTypeLibrary = function() {

    var _this = this;

    var all_fonts = [];
    var activated_fonts = [];
    var activated_map = {}

    this.init = function() {


    }

    this.addFont = function(font, insert_at_idx) {
        if(activated_map[font.family]) {
            this.removeFont(font.family)     ;
        }
        activated_map[font.family] = font;
        activated_fonts.push(font);
        font.addToPageView($('html'));
    }

    this.getFontByFamily = function(family) {
        return activated_map[family] || null;
    }

    this.removeFont = function(family) {
        var font = activated_map[family] || null;
        if(font) {
            var idx = activated_fonts.indexOf(font);
            idx >= 0 && activated_fonts.splice(idx, 1);
            delete activated_map[family];
        }
    }

    this.forEachFont = function(func) {
        activated_fonts.forEach(func);
    }


    this.getSelectOptions = function(usage) {
        var r = [];
        this.forEachFont(function(font) {
            if(font.isUsedFor(usage)) {
                r.push({
                    key: font.family,
                    name: font.family,
                    html: font.getPreview(font.family)
                })
            }
        })

        r.push({
            key: 'font-manager',
            name: 'Manage fonts...',
            action: function () {
                _this.showLibrary();
            }
        })
        return r;
    }

    this.getSelectOptionsForWeight = function(family) {
        var default_list = [
            {key:'normal', name:'Normal'},
            {key:'bold', name:'Bold'},
            {key:'lighter', name:'Lighter'},
            {key:'bolder', name:'Bolder'},
            {key:'100', name:'100'},
            {key:'200', name:'200'},
            {key:'300', name:'300'},
            {key:'400', name:'400'},
            {key:'500', name:'500'},
            {key:'600', name:'600'},
            {key:'700', name:'700'},
            {key:'800', name:'800'},
            {key:'900', name:'900'},
            {key:'inherit', name:'Inherit'}
        ]

        var map = {
            normal: '400',
            bold: '700'
        }

        var font = this.getFontByFamily(family);
        if(font) {
            return default_list.filter(function(i) {
                if(i.key === 'lighter' || i.key === 'bolder') return true;
                var val = map[i.key] || i.key;
                return font.weights.indexOf(val) >= 0;
            })
        } else {
            return default_list;
        }
    }

    this.showLibrary = function() {
        var fv = new PgFontTypeLibraryView();
        pinegrow.showViewInDialog('Font library', '', fv.view);
        fv.show();
    }

    this.clear = function() {
        activated_fonts = [];
        activated_map = {}
    }

    this.save = function() {
        var pi = pinegrow.getCurrentProjectInfo();
        var data = {
            activated: []
        };
        this.forEachFont(function(font) {
            var i = {};
            font.toNode(i);
            data.activated.push(i);
        })
        pi.saveFile('fonts.json', JSON.stringify(data));
    }
    
    this.getFontClassForType = function(type) {
        var cls = PgFontType;
        switch(type) {
            case 'local':
                cls = PgFontTypeLocal;
                break;
            case 'remote':
                cls = PgFontTypeRemote;
                break;
            case 'adobe':
                cls = PgFontTypeAdobe;
                break;
            case 'google':
                cls = PgFontTypeGoogle;
                break;
            case 'system':
                cls = PgFontTypeSystem;
                break;
            case 'uploaded':
                cls = PgFontTypeUploadedWebfont;
                break;
        }
        return cls;
    }

    this.load = function() {
        this.clear();

        var pi = pinegrow.getCurrentProjectInfo();
        var json = pi.loadFile('fonts.json');
        if(json) {
            var data = JSON.parse(json);
            if(data.activated) {
                data.activated.forEach(function(node) {
                    var cls = _this.getFontClassForType(node.type);
                    
                    if(cls) {
                        var font = new cls(node);
                        if(font.family) {
                            _this.addFont(font);
                            font.activateInProject();
                        }
                    }
                })
            }
        } else {
            this.restoreDefaultFonts();
        }
    }
    
    this.restoreDefaultFonts = function() {

        /*
        Usage values can be:
        - all: the font can be used for text and headings
        - h1: the font can only be used for H1 and H2. This is for display fonts usually
        - headings: the font is suitable for headings, not for regular text
         */

        var default_fonts = [
            {type: 'system', family: 'Arial, Helvetica, sans-serif', usage: 'all', category: 'sans'},
            {type: 'system', family: '"Arial Black", Gadget, sans-serif', usage: 'headings', category: 'sans'},
            {type: 'system', family: 'Impact, Charcoal, sans-serif', usage: 'headings', category: 'sans'},
            {type: 'system', family: '"Lucida Sans Unicode", "Lucida Grande", sans-serif', usage: 'all', category: 'sans'},
            {type: 'system', family: 'Tahoma, Geneva, sans-serif', usage: 'all', category: 'sans'},
            {type: 'system', family: 'Verdana, Geneva, sans-serif', usage: 'all', category: 'sans'},
            {type: 'system', family: '"Times New Roman", Times, serif', usage: 'all', category: 'serif'},
            {type: 'system', family: '"Palatino Linotype", "Book Antiqua", Palatino, serif', usage: 'all', category: 'serif'},
            {type: 'system', family: 'Georgia, serif', usage: 'all', category: 'serif'},
            {type: 'system', family: '"Courier New", Courier, monospace', usage: 'all', category: 'mono'},
            {type: 'system', family: '"Lucida Console", Monaco, monospace', usage: 'all', category: 'mono'},

            {type: 'google', family: 'Abril Fatface', usage: 'h1', category: 'display'},
            {type: 'google', family: 'Alegreya', usage: 'all', category: 'serif', w: '400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Arima Madurai', usage: 'headings', category: 'display', w: '100, 200, 300, 400, 500, 700, 800, 900'},
            {type: 'google', family: 'Arvo', usage: 'h1', category: 'serif', w: '400, 700'},
            {type: 'google', family: 'Asap', usage: 'all', category: 'sans', w: '400, 500, 600, 700'},

            {type: 'google', family: 'BioRhyme', usage: 'all', category: 'serif', w: '200, 300, 400, 700, 800'},

            {type: 'google', family: 'Cabin', usage: 'all', category: 'sans', w: '400, 500, 600, 700'},
            {type: 'google', family: 'Catamaran', usage: 'headings', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Caveat', usage: 'all', category: 'display', w: '400, 500, 600, 700'},
            {type: 'google', family: 'Caveat Brush', usage: 'all', category: 'display'},
            {type: 'google', family: 'Cormorant Garamond', usage: 'h1', category: 'serif', w: '300, 400, 500, 600, 700'},

            {type: 'google', family: 'Dancing Script', usage: 'headings', category: 'display', w: '400, 500, 600, 700'},
            {type: 'google', family: 'DM Sans', usage: 'all', category: 'sans', w: '400, 500, 700'},
            {type: 'google', family: 'Dosis', usage: 'all', category: 'sans', w: '200, 300, 400, 500, 600, 700, 800'},

            {type: 'google', family: 'Expletus Sans', usage: 'headings', category: 'display', w: '400, 500, 600, 700'},

            {type: 'google', family: 'Fira Sans', usage: 'all', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Fraunces', usage: 'all', category: 'serif', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'IBM Plex Serif', usage: 'all', category: 'serif', w: '100, 200, 300, 400, 500, 600, 700'},
            {type: 'google', family: 'Inter', usage: 'all', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Josefin Sans', usage: 'headings', category: 'sans', w: '200, 300, 400, 500, 600, 700'},
            {type: 'google', family: 'Jost', usage: 'headings', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Lato', usage: 'all', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Libre Baskerville', usage: 'headings', category: 'serif', w: '400, 700'},
            {type: 'google', family: 'Lora', usage: 'h1', category: 'serif', w: '400, 500, 600, 700'},

            {type: 'google', family: 'Kalam', usage: 'all', category: 'display', w: '300, 400, 700'},
            {type: 'google', family: 'Kufam', usage: 'headings', category: 'display', w: '400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Mali', usage: 'all', category: 'display', w: '200, 300, 400, 500, 600, 700'},
            {type: 'google', family: 'Martel', usage: 'headings', category: 'serif', w: '200, 300, 400, 600, 700, 800, 900'},
            {type: 'google', family: 'Maven Pro', usage: 'all', category: 'sans', w: '400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Montserrat', usage: 'headings', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Mulish', usage: 'all', category: 'sans', w: '200, 300, 400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Nunito Sans', usage: 'text', category: 'sans', w: '200, 300, 400, 600, 700, 800, 900'},

            {type: 'google', family: 'Open Sans', usage: 'all', category: 'sans', w: '300, 400, 600, 700, 800'},
            {type: 'google', family: 'Oswald', usage: 'headings', category: 'sans', w: '200, 300, 400, 500, 600, 700'},
            {type: 'google', family: 'Overlock', usage: 'all', category: 'display', w: '400, 700, 900'},
            {type: 'google', family: 'Oxanium', usage: 'headings', category: 'display', w: '200, 300, 400, 500, 600, 700, 800'},

            {type: 'google', family: 'Permanent Marker', usage: 'h1', category: 'display', w: '400'},
            {type: 'google', family: 'Playfair Display', usage: 'headings', category: 'sans', w: '400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Poppins', usage: 'headings', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Prompt', usage: 'headings', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Proza Libre', usage: 'all', category: 'sans', w: '400, 500, 600, 700, 800'},

            {type: 'google', family: 'Quicksand', usage: 'all', category: 'sans', w: '300, 400, 500, 600, 700'},

            {type: 'google', family: 'Rakkas', usage: 'h1', category: 'display', w: '400'},
            {type: 'google', family: 'Raleway', usage: 'all', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Rokkitt', usage: 'h1', category: 'serif', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Ruda', usage: 'all', category: 'sans', w: '400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Sansita', usage: 'headings', category: 'serif', w: '400, 700, 800, 900'},
            {type: 'google', family: 'Sansita Swashed', usage: 'h1', category: 'display', w: '300, 400, 500, 600, 700, 800, 900'},
            {type: 'google', family: 'Sora', usage: 'all', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800'},
            {type: 'google', family: 'Space Grotesk', usage: 'all', category: 'sans', w: '300, 400, 500, 600, 700'},
            {type: 'google', family: 'Spectral', usage: 'headings', category: 'serif', w: '200, 300, 400, 500, 600, 700, 800'},

            {type: 'google', family: 'Tillana', usage: 'headings', category: 'display', w: '400, 500, 600, 700, 800'},
            {type: 'google', family: 'Titillium Web', usage: 'all', category: 'sans', w: '200, 300, 400, 600, 700, 900'},
            {type: 'google', family: 'Trirong', usage: 'h1', category: 'serif', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Ubuntu', usage: 'all', category: 'sans', w: '300, 400, 500, 700'},
            
            {type: 'google', family: 'Work Sans', usage: 'headings', category: 'sans', w: '100, 200, 300, 400, 500, 600, 700, 800, 900'},

            {type: 'google', family: 'Zilla Slab', usage: 'all', category: 'serif', w: '300, 400, 500, 600, 700'},
        ]

        default_fonts.forEach(function(node) {
            if(!_this.getFontByFamily(node.family)) {
                var cls = _this.getFontClassForType(node.type);

                if (cls) {
                    if(node.w) {
                        node.styles = node.w.split(/,\s?/).map(function(w) {
                            return { weight: w}
                        });
                        node.weights = node.w.split(/,\s?/);
                    }
                    var font = new cls(node);
                    _this.addFont(font);
                    font.activateInProject();
                }
            }
        })
        this.save();
    }

    this.removeAll = function() {
        this.clear();
        this.save();
    }

    this.refreshFontsFromUrl = function(url, done) {
        var font_class = null;
        var idx = -1;

        var list = activated_fonts.slice(0);
        list.forEach(function(font) {
            if(font.url === url) {
                if(idx < 0) {
                    idx = activated_fonts.indexOf(font);
                }
                _this.removeFont(font.family);
                if(!font_class) {
                    font_class = font.constructor;
                }
            }
        })

        var gui = require('nw.gui');
        gui.App.clearCache();

        this.extractRemoteFontsFromUrl(url, function() {
            done && done();
        }, font_class, idx)
    }

    this.extractRemoteFontsFromUrl = function(url, done, font_class, insert_at_idx) {

        if(insert_at_idx === undefined) insert_at_idx = -1;

        font_class = font_class || PgFontTypeRemote;

        var parseCSS = function(css) {
            var r = [];
            var map = {}

            var a = css.split('@font-face').forEach(function(chunk) {
                var m = chunk.match(/font\-family:\s*['"](.*)['"]/m);
                if(m) {
                    var family = m[1];
                    var m = chunk.match(/font\-style:\s*([^;]*);/m);
                    if(m) {
                        var style = m[1];
                    }
                    var m = chunk.match(/font\-weight:\s*([^;]*);/m);
                    if(m) {
                        var weight = m[1];
                    }
                    if(!map[family]) {
                        map[family] = new font_class(null, url, family);
                        r.push(map[family]);

                        //try looking for fallback font
                        var re = new RegExp('font\\-family\\:\\s*[\'"]?' + pgParserEscapeRegExp(family) + '[\'"]?,\\s*(.*)\\;');
                        var m = css.match(re);
                        if(m) {
                            map[family].fallback = m[1];
                        }

                        pinegrow.fontLibrary.addFont(map[family], insert_at_idx >= 0 ? insert_at_idx++ : undefined);
                    }
                    style && map[family].addStyle(style);
                    weight && map[family].addWeight(weight);
                }
            })

            r.forEach(function(font) {
                font.activateInProject();
            })
        }

        if(crsaIsFileUrl(url)) {
            try {
                var file = crsaMakeFileFromUrl(url);
                var css = require('fs').readFileSync(file, {encoding: 'utf8'});
                parseCSS(css);
                done();
            } catch(err) {
                pinegrow.showQuickError('Could not read file ' + file);
                done();
            }
        } else {
            fetch(url)
                .then(function (response) {
                    response.text().then(function (css) {
                        parseCSS(css)
                        done();
                    });
                });
        }
    }

    this.getRandomFont = function(usage) {
        var r = activated_fonts.filter(function(font) {
            return font.isUsedFor(usage);// !font.usage || font.usage === usage;
        })
        if(!r.length) return null;
        var font = r[Math.round(Math.random() * (r.length - 1))];
        return font;
    }



    this.init();

    pinegrow.addEventHandler('on_project_loaded', function() {
        _this.load();
    })

    pinegrow.addEventHandler('on_project_closed', function() {
        _this.clear();
    })
}

var PgFontTypeLibraryView = function() {
    var fontLibrary = pinegrow.fontLibrary;

    var _this = this;

    var $element = $(`<div class="pg-font-lib-view">
<p>The fonts on this list are used in the Design panel. Only the fonts that are actually used in the active design are imported in the generated CSS stylesheet.</p>

<table class="table table-striped table-hover activated-fonts"></table>
<p>
<button class="add-font pg-button on-white-bck">Add font...</button>
</p>
</div>`);

    var $list = $element.find('.activated-fonts');

    $element.on('click', '.remove-font', function(e) {
        var $tr = $(this.closest('tr'));
        var family = pgDecodeAttribute($tr.attr('data-family'));
        fontLibrary.removeFont(family);
        fontLibrary.save();
        $tr.remove();

        pinegrow.showQuickMessage(`${family} was removed.`);
    })

    $element.on('click', '.edit-font', function(e) {
        e.preventDefault();

        var $tr = $(this.closest('tr'));
        var family = pgDecodeAttribute($tr.attr('data-family'));
        var font = fontLibrary.getFontByFamily(family);
        var orig_family = font.family;
        font.edit(function() {
            if(orig_family !== font.family) {
                fontLibrary.removeFont(orig_family);
                fontLibrary.addFont(font);
                font.activateInProject();
            }
            fontLibrary.save();
            _this.show();
        })
    })

    $element.on('click', '.refresh-font', function(e) {
        var $tr = $(this.closest('tr'));
        var family = pgDecodeAttribute($tr.attr('data-family'));
        var font = fontLibrary.getFontByFamily(family);

        pinegrow.showQuickMessage(`Refreshing fonts from ${font.url}...`);

        $tr.find('td').html('&nbsp;').first().html('Refreshing... Please wait.');

        fontLibrary.refreshFontsFromUrl(font.url, function() {
            fontLibrary.save();
            _this.show();
            pinegrow.showQuickMessage('Done.');
        })

    })

    $element.on('click', '.set-usage', function(e) {
        var $tr = $(this.closest('tr'));
        var family = pgDecodeAttribute($tr.attr('data-family'));
        var font = fontLibrary.getFontByFamily(family);
        var $el = $(this);

        var setUsage = function(u) {
            font.usage = u;
            fontLibrary.save();
            $el.html(u === 'h1' ? 'h1, h2' : u);
        }

        var actions = [
            {
                label: 'all',
                check: font.usage === 'all',
                action: function() {
                    setUsage('all')
                }
            },
            {
                label: 'text',
                check: font.usage === 'text',
                action: function() {
                    setUsage('text')
                }
            },
            {
                label: 'headings',
                check: font.usage === 'headings',
                action: function() {
                    setUsage('headings')
                }
            },
            {
                label: pusage('h1'),
                check: font.usage === 'h1',
                action: function() {
                    setUsage('h1')
                }
            },
        ]

        var m = new CrsaContextMenu(actions, $el);
        m.in_dialog = true;
        m.showForElement($el);

        return false;
    })

    


    $element.on('click', '.add-font', function(e) {
        var actions = [
            {
                label: 'System font...',
                action: function() {
                    pinegrow.showPrompt(`<p>Enter the name(s) of the system font.</p>`, 'Add a system font', '', '"Times New Roman", Times, serif', null, function(family) {
                        if(family) {
                            var font = new PgFontTypeSystem(null, family);
                            fontLibrary.addFont(font);
                            font.activateInProject();
                            fontLibrary.save();
                            _this.show();
                        }
                    })
                }
            },
            {
                label: 'Google Fonts...',
                action: function() {
                    pinegrow.pgGoogleFontManager.show(function (fonts) {
                        console.log(fonts)
                        fonts.forEach(function(gf_obj) {
                            var font = new PgFontTypeGoogle(null, gf_obj);
                            fontLibrary.addFont(font);
                            font.activateInProject();
                        })
                        fontLibrary.save();
                        _this.show();
                    });
                }
            },
            {
                label: isApp() ? 'Local web font...' : 'Self-hosted web font...',
                action: function() {

                    if(isApp()) {
                        var fields = {
                            url: {
                                name: 'Local font CSS file',
                                type: 'text',
                                file_picker_use_project_as_parent: true,
                                file_picker: true,
                                file_picker_no_proxy: true,
                                file_picker_no_copy_to_folder: true,
                                helptext: `<p>Select the CSS file located in the local font bundle.</p>`,
                                validate: function (obj, fn, value) {
                                    if (!value) {
                                        return 'Please select the CSS file.'
                                    }
                                }
                            }
                        }

                        pinegrow.showQuickDialog('Add a local web font', `Stylesheet`, `<p>Select the local CSS file that contains <code>@font-face</code> declaration of the web font bundle.</p><p>Note, you can use a service such as <a href="https://www.fontsquirrel.com/tools/webfont-generator" class="external">Font Squirrel</a> to upload font files and generate the web font bundle.</p>`, fields, 'Cancel', 'Add', null, function (values) {
                            fontLibrary.extractRemoteFontsFromUrl(pinegrow.getCurrentProject().getAbsoluteUrl(values.url), function () {
                                fontLibrary.save();
                                _this.show();
                            }, PgFontTypeLocal)
                        })
                    } else {
                        var font = new PgFontTypeUploadedWebfont(null, '');
                        font.edit(function() {
                            fontLibrary.addFont(font);
                            font.activateInProject();
                            fontLibrary.save();
                            _this.show();
                        });
                    }
                }
            },
            {
                label: 'Adobe fonts (TypeKit)...',
                action: function() {
                    pinegrow.showPrompt(`<p>Enter the <b>Project ID</b>. Projects ID are in the form <code>abc8efg</code>.</p>`, 'Add Adobe Fonts project', '', 'abc8efg', null, function(pid) {
                        if(pid) {
                            var url = `https://use.typekit.net/${pid}.css`
                            fontLibrary.extractRemoteFontsFromUrl(url, function() {
                                fontLibrary.save();
                                _this.show();
                            }, PgFontTypeAdobe)
                        }
                    })
                }
            },
            {
                label: 'Import from URL...',
                action: function() {
                    pinegrow.showPrompt(`<p>Enter the URL of the CSS stylesheet that contains @font-face declarations.</p>` + (isApp() ? '' : `<p>If the URL is located on a remote domain, the URL needs to be cross-origin accessible.</p>`), 'Import from URL', '', 'https://example.com/fonts/style.css', null, function(url) {
                        fontLibrary.extractRemoteFontsFromUrl(url, function() {
                            fontLibrary.save();
                            _this.show();
                        })
                    })
                }
            },
            {
                type: 'divider'
            },
            {
                label: 'Add default fonts',
                action: function() {
                    fontLibrary.restoreDefaultFonts();
                    _this.show();
                }
            },
            {
                type: 'divider'
            },
            {
                label: 'Remove all...',
                action: function() {
                    pinegrow.showAlert(`<p>Are you sure you want to remove all fonts from the list?</p>`, 'Are you sure?', 'Cancel', 'Yes, remove all fonts', null, function() {
                        fontLibrary.removeAll();
                        _this.show();
                    })
                }
            },
        ]

        var $el = $(this);
        var m = new CrsaContextMenu(actions, $el);
        m.in_dialog = true;
        m.showForElement($el);

        return false;
    })

    var view = this.view = new PgUIView($element, 'font_library');

    var usage_map = {
        h1: 'h1, h2'
    }

    var pusage = function(u) {
        return usage_map[u] || u || 'all';
    }

    this.show = function() {
        view.whenVisible(function() {
            var html = `<thead>
<tr>
<td>Family name</td>
<td>Use for</td>
<td>Source</td>
<td>Styles</td>
<td style="text-align:center;">Refresh</td>
<td style="text-align:center;">Remove</td>
</tr>
</thead><tbody>`;
            fontLibrary.forEachFont(function(font) {
                html += `<tr data-family="${pgEncodeAttribute(font.family)}">
    <td>${font.getPreview()}${font.edit ? ' <a href="#" class="edit-font">edit</a>' : ''}</td>
    <td><a href="#" class="set-usage">${pusage(font.usage)}</a></td>
    <td>${font.getTypeLabel()}</td>
    <td>${font.describeStyles()}</td>
    <td style="text-align:center;">${font.can_refresh ? '<i class="icon icon-refresh refresh-font" title="Reload ' + font.url + ' and refresh fonts."></i>' : '&nbsp;'}</td>
    <td style="text-align:center;"><i class="icon icon-close remove-font"></i></td>
</tr>`;
            })
            html += '</tbody>';

            $list.html(html);

            addTooltip($list.find('.refresh-font'), 'title', {use_title_attr: true});
            addTooltip($list.find('.remove-font'), 'Remove the font from the list.');
            addTooltip($list.find('.set-usage'), 'Click to select where this font should be used.');

        }, 'show');
    }
}


var PgDmFontLibrary = function() {

    var text_fonts = [
        {name: 'Lato'},
        {name: 'Montserrat'},
        {name: 'Playfair Display'},
        {name: 'Open Sans'},
        {name: 'Merriweather'},
        {name: 'Georgia', system: true}
    ]

    var heading_fonts = [
        {name: 'Faster One', w: 400, s: '2px'},
        {name: 'Chonburi'},
        {name: 'Tulpen One', w: 400},
        {name: 'Oswald'},
        {name: 'Open Sans', w: 700},
        {name: 'Montserrat', w: 400, s: '2px', t: 'u'},
        {name: 'Playfair Display'},
        {name: 'Indie Flower'},
        {name: 'Georgia',system: true, style: 'italic'}
    ]

    var all_fonts = {}

    var collectAllFonts = function(list) {
        list.forEach(function(f) {
            all_fonts[f.name] = f;
        })
    }

    collectAllFonts(text_fonts);
    collectAllFonts(heading_fonts);

    var getFontStyles = function(f, getstr, only) {
        var s = {};
        s['font-family'] = '\'' + f.name + '\'';
        if(f.style) {
            s['font-style'] = f.style;
        }
        if(f.w) {
            s['font-weight'] = f.w;
        }
        if(f.s) {
            s['letter-spacing'] = f.s;
        }
        if(f.t) {
            s['text-transform'] = f.t == 'u' ? 'uppercase' : 'lowercase';
        }
        if(getstr) {
            var str = '';
            $.each(s, function(key, val) {
                if(only && only.indexOf(key) < 0) return true;
                str += key + ':' + val + ';';
            })
            return str;
        }
        return s;
    }

    var getFontUrl = function(f) {
        var ws = [f.w];
        if(f.leadw) ws.push(f.leadw);
        return `https://fonts.googleapis.com/css?family=${f.name.replace(/ /g, '+')}:${ws.join(',')}${f.style ? f.style[0] : ''}`;
    }

    this.getRandomFont = function(fonts) {
        return fonts[Math.round(Math.random() * (fonts.length - 1))];
    }

    this.getRandomTextFont = function() {
        return this.getRandomFont(text_fonts)
    }

    this.getRandomHeadingFont = function() {
        return this.getRandomFont(heading_fonts)
    }

    font_sample_text = 'AaBbCcDdEeFfGgHhIiJjKk';

    this.getFontOptions = function(fonts) {
        var idx = 1;
        return fonts.map(function(f) {
            return {key: f.name, name: f.name, html: '<span style="' + getFontStyles(f, true) + '">' + f.name + '</span>'}
        })
    }

    this.getTextFontOptions = function() {
        return this.getFontOptions(text_fonts);
    }

    this.getHeadingFontOptions = function() {
        return this.getFontOptions(heading_fonts);
    }

    this.getTextFontByKey = function(idx) {
        return text_fonts[idx-1];
    }

    this.getHeadingFontByKey = function(idx) {
        return heading_fonts[idx-1];
    }

    this.getTextFontKey = function(f) {
        return text_fonts.indexOf(f) + 1;
    }

    this.getHeadingFontKey = function(f) {
        return heading_fonts.indexOf(f) + 1;
    }

}
