$(function() {

    //Wait for Pinegrow to wake-up
    $('body').one('pinegrow-ready', function(e, pinegrow) {

        //if(!pgIsMagic()) return;

        //Create new Pinegrow framework object
        var f = new PgFramework('pg.dm.lib.components', 'PG DM Lib');

        f.type = "pg.dm.lib";
        f.allow_single_type = true;

        f.description = '';
        f.author = 'Pinegrow';
        f.author_link = 'http://pinegrow.com';
        f.default = pgIsMagic();

        var prefix = 'pg.dm.lib.';

        //Tell Pinegrow about the framework
        pinegrow.addFramework(f);

        var list_containers = [];
        var list_content = [];
        var list_components = [];

        var addComponent = function(type, name, sel, code, on_can_insert_into, options) {
            var container = new PgComponentType(prefix + type, name);
            container.selector = sel;
            container.code = code;
            container.empty_placeholder = true;
            container.tags = 'major';

            container.on_can_insert_into = on_can_insert_into;

            if(options) {
                $.each(options, function(key, val) {
                    container[key] = val;
                })
            }
            //Add it to our framework
            f.addComponentType(container);
            return container;
        }

        //Containers

        var parentsAre = function(pgel, sel) {
            var p = pgel;
            while(p && !p.rootNode) {
                if(!p.isSelector(sel)) return false;
                p = p.parent;
            }
            return true;
        }

        var sections_selector = 'header,section,footer';
        var blocks_selector = sections_selector + ',div';

        var insertInBlocksOnly = function(pgel) {
            if(parentsAre(pgel, 'div,body,html,' + sections_selector)) return true;
            return false;
        }

        var insertInBlocksAndLinksOnly = function(pgel) {
            if(parentsAre(pgel, 'a,div,body,html,' + sections_selector)) return true;
            return false;
        }

        var isSection = function(pgel) {
            return pgel.isSelector(sections_selector);
        }

        var isBlock = function(pgel) {
            return pgel.isSelector(blocks_selector);
        }

        var canInsertSection = function(pgel, def) {
            //can insert into pgel?

            if(pgel.closest(sections_selector)) {
                return false;
            }
            if(pgel.tagName === 'body') return true;

            if(parentsAre(pgel, 'div,body,html')) return true;

            return false;
        }

        var canInsertContent = function(pgel, def) {
            //must be in block
        }

        var insertInSectionsOnly = function(pgel, def) {
            return pgel.parent.isSelector(sections_selector);
        }

        list_containers.push(addComponent('header', 'Header', 'header', `<header></header>`, canInsertSection));

        list_containers.push(addComponent('section', 'Section', 'section', `<section></section>`, canInsertSection));

        list_containers.push(addComponent('footer', 'Footer', 'footer', `<footer></footer>`, canInsertSection));

        list_containers.push(addComponent('div', 'Div', 'div', `<div></div>`, insertInBlocksOnly));

        list_containers.push(addComponent('columns.centered', 'Columns - Centered', '.container', `<div class="container"><div class="row"><div class="col-md-4 pg-empty-placeholder"></div><div class="col-md-4 pg-empty-placeholder"></div><div class="col-md-4 pg-empty-placeholder"></div></div></div>`, function(pgel, def) {
            return isSection(pgel);
        }));

        list_containers.push(addComponent('columns', 'Columns', '.container-fluid', `<div class="container-fluid"><div class="row"><div class="col-md-4"></div><div class="col-md-4"></div><div class="col-md-4"></div></div></div>`));

        // Content

        var h_sel = 'h1,h2,h3,h4,h5,h6';
        var text_sel = 'p,span,button,.btn,b,em,i'

        for(var h = 1; h <= 6; h++) {
            list_content.push(addComponent('h' + h, 'H' + h, 'h' + h, `<h${h}>Heading ${h}</h${h}>`, function(pgel) {
                //can insert into
                return isBlock(pgel);

                //if(pgel.closest(h_sel)) return false;
                //if(insertInBlocksAndLinksOnly(pgel)) return true;
            }));
        }

        list_content.push(addComponent('p', 'Paragraph', 'p', `<p>This is some nice text.</p>`, function(pgel) {
            return isBlock(pgel);
        }));

        list_content.push(addComponent('img', 'Image', 'img', `<img src="${pinegrow.getPlaceholderImage()}">`, function(pgel, def) {
            return isBlock(pgel) || pgel.isSelector('p');
        }));

        /*
        list_content.push(addComponent('link', 'Link', 'a', `<a href="">Link</a>`, function(pgel, def) {
            return isBlock(pgel) || pgel.isSelector('p');
        }));
*/

        list_content.push(addComponent('button', 'Button', '.btn', `<button class="btn btn-primary">Click me</button>`, function(pgel, def) {
            return isBlock(pgel) || pgel.isSelector('p');
        }));

        list_content.push(addComponent('icon', 'Icon', '.far,.fas,.fab', `<i class="far fa-star"></i>`, function(pgel, def) {
            return isBlock(pgel) || pgel.isSelector(h_sel + ',' + text_sel.replace(',i', ''));
        }, {
            empty_placeholder: false
        }));

        // Components

        var c_navbar = addComponent('navbar', 'Navbar', '.navbar',
            //code
            function() {
                var id = getUniqueId('navbarToggler');
                var ddId = getUniqueId('navbarDropdownMenuLink');

                return `<nav class="navbar navbar-expand-lg navbar-light bg-light">
                        <a class="navbar-brand" href="#">Brand</a>
                        <button class="navbar-toggler" type="button" data-toggle="collapse" data-target="#${id}" aria-controls="" aria-expanded="false" aria-label="Toggle navigation">
                            <span class="navbar-toggler-icon"></span>
                        </button>
                        <div class="collapse navbar-collapse" id="${id}">
                            <ul class="navbar-nav mr-auto mt-2 mt-lg-0">
                                <li class="nav-item active">
                                    <a class="nav-link" href="#">Home <span class="sr-only">(current)</span></a>
                                </li>
                                <li class="nav-item">
                                    <a class="nav-link" href="#">Link</a>
                                </li>\
                                <li class="nav-item">
                                    <a class="nav-link disabled" href="#">Disabled</a>
                                </li>
                                <li class="nav-item dropdown">
                                    <a class="nav-link dropdown-toggle" href="#" id="${ddId}" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">
                                        Dropdown link
                                    </a>
                                    <div class="dropdown-menu" aria-labelledby="${ddId}">
                                        <a class="dropdown-item" href="#">Action</a>
                                        <a class="dropdown-item" href="#">Another action</a>
                                        <a class="dropdown-item" href="#">Something else here</a>
                                    </div>
                                </li>
                            </ul>
                        </div>
                    </nav>`;
            },
            function(pgel) {
                //can insert into pgel?
                if(parentsAre(pgel, 'div,body,html,header,section,footer')) return true;

                return false;
            }
        )

        c_navbar.on_filter_tree_children = function (pgel) {
            var r = [];
            r = r.concat(pgel.find('.navbar-brand'));
            r = r.concat(pgel.find('.navbar-nav'));
            return r;
        }

        c_navbar.on_get_minitree_structure = function(pgel) {
            var d = {
                brand: {
                    def: c_navbar_brand
                },
                sections: {
                    def: c_navbar_sections,
                    children: {
                        link: {
                            def: c_navbar_link
                        },
                        dropdown: {
                            def: c_dropdown
                        },
                        form: {
                            def: c_navbar_form
                        }
                    }
                }
            }
        }

        var list = [];
        var createMiniTreeStructure = function(pgel, ms) {

            var createLevel = function(node, d, level) {

                var defs = []

                //Find any elements
                node.children.forEach(function(chel) {
                    if(chel.isElement) {

                        $.each(d, function(key, info) {

                            if(chel.isSelector(info.def.selector)) {

                                var a = {}

                                a.name = chel.getName();
                                a.pgel = chel;

                                if(info.children) {
                                    createLevel(info.children, level + 1);
                                }

                            }

                            defs.push(info.def);
                        })

                    }
                })
                //add addnew
                var defs = currentPage.getAllTypes(null, node);
                var add = [];
                $.each(defs, function(i, def) {
                    if(def.action_menu && def.action_menu.add) {
                        add.push(def.action_menu);
                    }
                });
            }
        }





        f.componentFiltersTree(c_navbar);

        list_components.push(c_navbar);

        var c_navbar_nav_item = addComponent('navbar.nav.item', 'Navbar Nav Item', '.nav-item',
            `<li class="nav-item">
                <a class="nav-link" href="#">Link</a>
            </li>`)

        c_navbar_nav_item.on_filter_tree_children = function (pgel) {
            return [];
        }

        f.componentFiltersTree(c_navbar_nav_item);

        list_components.push(addComponent('div', 'Slider', 'div', `<div></div>`));
        list_components.push(addComponent('div', 'Tabs', 'div', `<div></div>`));
        list_components.push(addComponent('div', 'Accordion', 'div', `<div></div>`));
        list_components.push(addComponent('div', 'Modal', 'div', `<div></div>`));
        list_components.push(addComponent('div', 'Dropdown', 'div', `<div></div>`));

        var section = new PgFrameworkLibSection('pg.dm.lib.containers', 'Layout');
        //Pass components in array
        section.setComponentTypes(list_containers);
        f.addLibSection(section);

        section = new PgFrameworkLibSection('pg.dm.lib.content', 'Content');
        //Pass components in array
        section.setComponentTypes(list_content);
        f.addLibSection(section);

        section = new PgFrameworkLibSection('pg.dm.lib.components', 'Components');
        //Pass components in array
        section.setComponentTypes(list_components);
        f.addLibSection(section);

        var body_comp = addComponent('body', 'Body', 'body', `<div></div>`)

        body_comp.action_menu = {
            add : [prefix + 'header', prefix + 'section', prefix + 'footer', prefix + 'navbar'],
            on_add : function(pgel, pgnew, newdef, prepend) {
                if (newdef.type == prefix + 'header' || newdef.type == prefix + 'navbar') {
                    pgel.prepend(pgnew);
                }
                else if (newdef.type == prefix + 'footer') {
                    pgel.append(pgnew);
                } else {
                    if (prepend) {
                        pgel.prepend(pgnew);
                    } else {
                        pgel.append(pgnew);
                    }
                }
            }
        }
    });
});