class PgTutorialsManager {
    constructor() {
        this.list = []

        var _this = this;

        pinegrow.addEventHandler('on_recent_list_populate', function(pagenull, list) {
            _this.updateStartScreen();
        });

        pinegrow.addEventHandler('on_before_load_project_from_server', function(pagenull, project, done) {
            if(project.url.startsWith('https://tutorial_')) {

                var key = project.url.replace('https://tutorial_', '');
                var tutorial = _this.getTutorialForLessonKey(key);
                var lesson = tutorial.findLesson(key);

                lesson.addFilesToProject(project);

                if(lesson.main_lesson) {
                    //project.projectSaveKey = 'https://tutorial_' + lesson.main_lesson.key;
                    var prev_lesson = tutorial.getPreviousLesson(lesson);
                    if(prev_lesson) {
                        project.projectLoadKeyIfNotFound = 'tutorial_' + prev_lesson.key;
                    }
                }

                project.lesson = lesson;
                project.name = lesson.name;
                project.isCreatedWithNewProject = lesson.tutorial.startProjects || false;

                pgf.open_all_files_in_code = lesson.open_all_files_in_code || false;
                done();

            } else {
                //normal project
                var lesson = new PgTutorialLesson('dummy');

                lesson.resources.forEach(function (r) {
                    project.addSourceUrlToDOMUrlMap(r.source_url, r.dom_url);
                })

                pgf.open_all_files_in_code = pgf.kids || false;
                done();
            }
        })

        pinegrow.addEventHandler('on_project_loaded', function(pagenull, project) {

            if(true) {
                var files = project.getEditableFiles();
                var first = true;

                var open_pages = project.getListOfOpenPages();
                var open_all = pgf.kids || false;

                if(open_pages.length === 0) {
                    open_pages.push('index.html');
                    project.pageWasOpenedOrClosed({url: 'index.html'}, true);
                }

                files = files.filter(function(f) {
                    var name = crsaGetNameFromUrl(f.url);
                    return !name.startsWith('_') || pinegrow.currentUser.isAdmin();
                })

                var pi = pinegrow.getCurrentProjectInfo();

                //open page with a slight delay so that other on_project_loaded handlers might do their stuff, for example registering mapped resources
                setTimeout(function() {
                    for(var i = files.length - 1; i >= 0; i--) {
                        var rel_url = project.getRelativeUrl(files[i].url);
                        if(open_all || open_pages.indexOf(rel_url) >= 0) {
                            (function(first) {
                                /*
                                pinegrow.openPage(files[i].url, function(page) {
                                    first && page.show(false);
                                }, first, null, {
                                    tab_index: 0,
                                    do_not_remember_page_open: true,
                                    open_with_wrapper: pgf.use_page_wrappers && pi.getSettingForUrl(files[i].url, 'open-with-wrapper')
                                });
                                */

                                pinegrow.openOrShowPage(files[i].url, function(page) {
                                    first && page.show(false);
                                }, first, null, {
                                    tab_index: 0,
                                    do_not_remember_page_open: true,
                                    open_with_wrapper: pgf.use_page_wrappers && pi.getSettingForUrl(files[i].url, 'open-with-wrapper')
                                })
                            })(first);
                            first = false;
                        }
                    }
                }, 10)

                if(project.lesson && project.lesson.description) {
                    if( pinegrow.infoPanel) {
                        pinegrow.infoPanel.setContent(null, {
                            title: (project.lesson.display_title || project.lesson.name) + (project.lesson ? '<a href="#" class="pg-tutorial-restart-button"><i class="icon icon-refresh"></i></a>' : ''),
                            content: project.lesson.description
                        })

                        var $video_button = pinegrow.infoPanel.get$Content().find('.pg-tutorial-video-button');

                        $video_button.on('click', function (e) {
                            e.preventDefault();
                            showVideo($(this).attr('data-vid'), project.lesson.subtitles);
                        })

                        addTooltip($video_button, 'Watch the video to learn about the adventure.');

                        var $restart_button = pinegrow.infoPanel.get$Content().find('.pg-tutorial-restart-button');

                        $restart_button.on('click', function (e) {
                            e.preventDefault();
                            var h = new PgProjectMenuHelper(project);
                            h.restartLesson();
                        })

                        addTooltip($restart_button, `Restart the current ${pgf.lesson_name}.`);

                        pinegrow.showUIPanel('info', false, false, false, 'pages');
                        //$('.menu-current-project').hide()
                        $('.menu-current-project').show().find('span').html(project.lesson.name);
                        $('.pg-lesson-check-button').show();
                        $('body').addClass('in-lesson');

                        if (project.isLoadedFromServer) {
                            pinegrow.backend.setIndicatorState('saved-lesson');

                            //new PgUIElementExplainer($('.menu-current-project'), `<p>You are continuing a saved adventure.</p><p>Use Restart to restart the adventure.</p>`, 1000, 8000);
                        }

                        //disabled because cause lesson to be saved immediatelly after restart
                        if (!project.isLoadedFromServer || project.isLoadedFromBackupId) {
                            //Save the project so that previews work
                            //pinegrow.backend.saveCurrentProject();
                        }

                        //var vid = $('.pg-tutorial-video-button[data-vid]').attr('data-vid');

                        var data = {
                            "l": [{
                                "name": "Show",
                                "trg": "no",
                                "a": "jello",
                                "po": "true",
                                "rpt": 3,
                                "dly": 1
                            }]
                        };

                        pgia.play($('.pg-tutorial-video-button[data-vid]').get(0), 'Show', data);

                        //vid && showVideo(vid);
                    } else {
                        pinegrow.showAlert(`<p>The Info panel can not be displayed. Please restore the workspace layout to one of the presets in the Window menu.</p>`, 'Info panel is missing');
                    }

                } else {
                    if(project.hasContentForUrl('_info.html')) {
                        project.getContentOfUrl('_info.html', function(info) {
                            try {
                                var p = new pgParser();
                                p.assignIds = false;
                                p.parse(info);
                                var body = p.rootNode.findOne('body');
                                var h1 = body.findOne('h1');

                                var info_data = {
                                    title: h1.html()
                                }

                                h1.remove();

                                info_data.content = _this.addInteractiveElementsToHTML(body.html());

                                pinegrow.infoPanel.setContent(null, info_data);
                            } catch(err) {
                                console.error(err);
                            }
                            pinegrow.showUIPanel('info', false, false, false, 'pages');
                        })
                    } else {
                        pinegrow.infoPanel.setContent(null, {
                            title: 'Nothing to see here',
                            content: `<p>Create the page _info.html to set the content of the info panel.</p>`
                        })
                        pinegrow.hideUIPanel('info');
                    }

                    $('.menu-current-project').show().find('span').html(project.name);
                    $('.pg-lesson-check-button').hide();
                    $('body').removeClass('in-lesson');
                }
            }

        })

        pinegrow.addEventHandler('on_project_closed', function() {
            pinegrow.infoPanel.setContent(null, {
                title: 'Nothing to see here',
                content: `<p>Create the page _info.html to set the content of the info panel.</p>`
            })
            pinegrow.hideUIPanel('info');
        })

        var video_tip_shown = false;

        pinegrow.addEventHandler('on_user_projects_received_from_backend', function(pagenull, data) {

            if(data.lessons) {
                $.each(data.lessons, function(lid, status) {
                    var tutorial = _this.getTutorialForLessonKey(lid);
                    if(tutorial) {
                        var lesson = tutorial.findLesson(lid);
                        lesson && lesson.setStatus(status);
                    }
                })
            }
            //check if video was done
            if(pgf.kids && (!data.lessons['intro.video.kids'] || data.lessons['intro.video.kids'] !== 'done') && !video_tip_shown) {
                //intro video not yet seen
                video_tip_shown = true;

                setTimeout(function() {
                    if(!pinegrow.getCurrentProject()) {
                        var explainer = new PgUIElementExplainer($('.pg-lessons-intro-video'), 'Start by watching the introduction video.', {
                            show_delay: 0,
                            close_delay: 50000,
                            focus: false,
                            position: 'up',
                            scroll_container: $('.pg-start-screen'),
                            button: 'Got it!',
                            on_button_clicked: function() {
                                pinegrow.backend.setLessonStatus({
                                    key: 'intro.video.kids'
                                }, 'done');
                            }
                        });
                    }
                }, 1000);
            }
            //look for active

            _this.getTutorials().forEach(function(tutorial) {

                var has_active = false;
                var lessons = tutorial.getLessons(true);

                for(var i = 0; i < lessons.length; i++) {
                    if(lessons[i].status === 'active') {
                        has_active = lessons[i];
                        break;
                    }
                }
                if(!has_active) {
                    //find first active (first locked)
                    for(var i = 0; i < lessons.length; i++) {
                        if(lessons[i].status === 'locked' || lessons[i].status === 'none') {
                            has_active = lessons[i];
                            break;
                        }
                    }
                    if(!has_active) {
                        for (var i = lessons.length - 2; i >= 0; i--) {
                            if (lessons[i].status === 'done') {
                                has_active = lessons[i + 1];
                                break;
                            }
                        }
                    }
                }

                has_active && has_active.setStatus('active');
            })

            _this.updateStartScreen(data.projects);
        });



        pinegrow.addEventHandler('on_public_projects_received_from_backend', function(pagenull, list) {
            var html = '';
            var global = pgf.featured_projects === 'global';
            if(list.length) {
                html += `<div class="pg-lessons-chapter"><h3>${global ? 'Projects belonging to other users' : 'Featured projects'}</h3>`;
                html += `<div class="lessons-grid">`;

                list.forEach(function (pi) {
                    html += `<li class="lesson-item lesson-item-with-image lesson-status-done" data-project="${pi.project_id}">`;

                    if (pi.thumb) {
                        html += _this.getHtmlForProjectThumb(pi);
                    } else {
                        html += _this.getHtmlForProjectPlaceholder(pi);
                    }

                    html += _this.getHTMLForProjectInfo(pi);
                    html += `</li>`;
                })
                html += '</div></div>';
            }

            $('.projects .public-projects').html(html);
        });

        pinegrow.addEventHandler('on_toolbar_created', function(pagenull, toolbar) {
            toolbar.addSection('tutorial');

            var checkButton = new PgButtonToolbarItem('<span class="pg-button pg-button-primary">Check <i class="icon icon-check"></i></span>', 'Check if the answer is correct.');
            checkButton.$element.addClass('pg-lesson-check-button')
            toolbar.addItem('tutorial', checkButton);
            pg$Hide(checkButton.$element);

            var ignore_check_click = false;

            checkButton.onClick = function() {
                if(ignore_check_click) return;

                var p = pinegrow.getCurrentProject();
                if(p.lesson) {
                    var tutorial = p.lesson.tutorial;
                    var status = p.lesson.check();

                    if(status.result !== 'ok') {
                        status.message += `<br><br>Note, you can re-watch the video to see the instructions again.`;
                    }

                    var html = `<div class="pg-tutorial-check-box-overlay">
<div class="pg-tutorial-check-box pg-tutorial-check-box-${status.result}">
<h1>${status.title}</h1>
<i class="close-check icon icon-close"></i>
<div class="pg-tutorial-check-box-msg">${status.message}</div>`;

                    if(status.decoration) {
                        html += status.decoration;
                    }

                    if(status.result === 'ok') {
                        html += `<a class="pg-tutorial-check-box-button pg-tutorial-check-box-button-back"><span style="transform:scaleX(-1);display: inline-block;"><i class="icon icon-right-arrow"></i></span> Back to the ${pgf.lesson_name}</a>&nbsp;<a class="pg-tutorial-check-box-button pg-tutorial-check-box-button-ok">Next ${pgf.lesson_name} <i class="icon icon-right-arrow"></i></a><p><a href="#" class="aapg-tutorial-check-box-button pg-tutorial-check-box-button-launchpad">Leave the ${pgf.lesson_name} and go to the Start screen</a></p>`;
                    } else {
                        html += `<a class="pg-tutorial-check-box-button pg-tutorial-check-box-button-error"><span style="transform:scaleX(-1);display: inline-block;"><i class="icon icon-right-arrow"></i></span> Try again</a><p><a href="#" class="aapg-tutorial-check-box-button pg-tutorial-check-box-button-launchpad">Leave the ${pgf.lesson_name} and go to the Start screen</a></p>`;
                    }

                    html += '</div></div>';
                    var $check_box = $(html);

                    addTooltip($check_box.find('[title]').on('click', function(e) {
                        pinegrow.dispatchEvent('on_info_panel_tooltip_clicked', null, this, e);
                    }), 'title', {
                    });

                    ignore_check_click = true;

                    setTimeout(function() {
                        $('body').append($check_box);
                        ignore_check_click = false;
                    }, status.delay || 0)


                    $check_box.find('.pg-tutorial-check-box-button-back, .pg-tutorial-check-box-button-error').on('click', function(e) {
                        e.preventDefault();
                        $check_box.remove();
                    })

                    $check_box.find('.pg-tutorial-check-box-button-ok').on('click', function(e) {
                        e.preventDefault();
                        $check_box.remove();
                        tutorial.openNextLesson();
                    })

                    $check_box.find('.pg-tutorial-check-box-button-launchpad').on('click', function(e) {
                        e.preventDefault();
                        $check_box.remove();
                        pinegrow.backend.saveAndCloseCurrentProject();
                        /*
                        pinegrow.closeCurrentProject(false, true, function() {

                        });

                         */
                    })

                    $check_box.find('i.close-check').on('click', function(e) {
                        e.preventDefault();
                        $check_box.remove();
                    })

                    $check_box.find('.pg-tutorial-check-box').on('click', function(e) {
                        return false;
                    })

                    $check_box.on('click', function(e) {
                        e.preventDefault();
                        $check_box.remove();
                    })
                }
            }

            var homeButton = new PgButtonToolbarItem('<span class="pg-button">Menu</span>', 'Go back to home screen and other actions.');
            //homeButton.$element.addClass('pg-lesson-check-button')
            //toolbar.addItem('tutorial', homeButton);

            homeButton.$element.addClass('pg-lesson-home-button')

            /*
            $('.menu-project-menu').on('click', function(e) {
                homeButton.onClick();
                e.preventDefault();
            })
            */

            homeButton.onClick = function() {

            }

        });

        var video_player = null;

        var showVideo = function(vid, srt) {
            video_player = new PgTutorialVideo(vid, srt);
        }

        var wireStartScreenEvents = function() {
            $('.pg-start-screen .lessons ul').on('click', 'li', function(e) {
                e.preventDefault();
                var $li = $(this);

                var lesson_id = $li.attr('data-key');
                var tutorial = _this.getTutorialForLessonKey(lesson_id);
                var lesson = tutorial.findLesson(lesson_id);

                if(!pinegrow.currentUser.isLoggedInAndActive()) {
                    if(!lesson.demo) {
                        pinegrow.currentUser.onlyForMembersNotice('go on this adventure')

                        return;
                    }
                }

                var only_locked = true;

                if(!pinegrow.currentUser.isLoggedInAndActive()) {
                    only_locked = false;
                }
                if(pinegrow.currentUser.isAdmin() && false) {
                    only_locked = false;
                }

                if(tutorial.allowRandomAccess) {
                    only_locked = false;
                }

                if(lesson && lesson.status === 'locked') {
                    if(only_locked) {
                        pinegrow.showQuickError(`This ${pgf.lesson_name} is locked. Unlock it by completing the ${pgf.lesson_name}s before it.`);
                    } else {
                        if(tutorial.allowRandomAccess) {
                            if(tutorial.warnOnRandomAccess) {
                                pinegrow.showNotice(`<p>Lessons in this tutorial build upon each other and it is recommended to do them one by one. The added benefit is that whatever you build in one lesson will serve as the starting point for the next lesson.</p><p>That said, it is up to you how you learn. You can go straight to lessons that interest you.</p>`, 'A quick note before you jump into this lesson', 'tutorial-random-access', function () {
                                    tutorial.openLesson(lesson_id);
                                }, false, false, false, true, false, null, {
                                    ok_label: 'Open the lesson',
                                    cancel_label: 'Cancel'
                                });
                            } else {
                                tutorial.openLesson(lesson_id);
                            }
                        }
                    }
                } else {
                    tutorial.openLesson(lesson_id);
                }
            })

            $('.pg-start-screen .lessons ul,.pg-start-screen .projects ul').on('mouseenter', 'li', function(e) {
                e.preventDefault();
                var img = this.querySelectorAll('.lesson-img');
                if(img.length) {
                    gsap.killTweensOf(img[0]);
                    gsap.to(img[0], {
                        bottom: this.classList.contains('lesson-status-done') ? 0 : -20,
                        duration: 0.35
                    })

                    if(img.length === 2) {
                        gsap.killTweensOf(img[1]);
                        gsap.to(img[1], {
                            left: '100px',
                            bottom: '-40px',
                            rotateZ: '0deg',
                            duration: 6
                        })
                    }
                }
            })

            $('.pg-start-screen .lessons ul,.pg-start-screen .projects ul').on('mouseleave', 'li', function(e) {
                e.preventDefault();
                var img = this.querySelectorAll('.lesson-img');
                if(img.length) {
                    gsap.killTweensOf(img[0]);
                    gsap.to(img[0], {
                        bottom: this.classList.contains('lesson-status-done') ? -20 : -44,
                        duration: 1,
                        delay: 0.5
                    })

                    if(img.length === 2) {
                        gsap.killTweensOf(img[1]);
                        gsap.to(img[1], {
                            left: '120%',
                            bottom: '120%',
                            rotateZ: '45deg',
                            ease: 'power2.easeIn',
                            duration: 4
                        })

                    }
                }
            })

            $('.pg-start-screen .projects ul')
                .on('click', 'li', function(e) {
                    if($(e.target).is('a.author')) return;
                    if($(e.target).closest('.project-dropdown').length) return;

                    e.preventDefault();
                    var $li = $(this);
                    var pid = $li.attr('data-project');
                    var a = pid.split('/');
                    var project_user_slug = null;
                    if(a.length === 2) {
                        project_user_slug = a[0];
                        pid = a[1];
                    }
                    if(pid === 'new') {
                        //moved to start screen
                    } if(pid === 'open') {
                        //moved to start screen

                    } else {
                        if(isApp()) {
                            var is_file = $li.attr('data-file');
                            if(pid.startsWith('internal:')) {
                                var url = pid.replace('internal:', '');
                                pinegrow.openExample(url);
                            } else {
                                if(is_file) {
                                    pinegrow.openOrShowPage(pid);
                                } else {
                                    pinegrow.openProject(pid);
                                }
                            }
                        } else {
                            _this.openProject('https://' + pid, null, project_user_slug);
                        }
                    }
                })
                .on('contextmenu', 'li', function(e) {
                    showProjectContextMenu.call(this, e);
                    return false;
                })
                .on('click', '.project-dropdown', function(e) {
                    showProjectContextMenu.call(this, e);
                    return false;
                })
        }

        var showProjectContextMenu = function(e) {
            var $li = $(this).closest('li');
            var pid = $li.attr('data-project');

            var a = pid.split('/');
            var project_user_slug = null;
            if(a.length === 2) {
                project_user_slug = a[0];
                pid = a[1];
            }

            var actions = [];

            if(pid !== 'new') {
                actions.push({
                    label: 'Edit',
                    action: function() {
                        if(isApp()) {
                            pinegrow.openProject(pid);
                        } else {
                            _this.openProject('https://' + pid, null, project_user_slug);
                        }
                    }
                })

                if(isApp()) {
                    actions.push({
                        label: 'Edit in new window',
                        action: function() {
                            pinegrow.openProjectInNewWindow(pid);
                        }
                    })

                    actions.push({
                        type: 'divider'
                    })

                    actions.push({
                        label: 'Remove from the start screen',
                        action: function() {
                            pinegrow.recentFiles.remove(pid, false);
                            $li.remove();
                        }
                    })
                } else {
                    actions.push({
                        label: 'Edit in new window',
                        dont_handle_click: true,
                        link_url: pinegrow.backend.getProjectEditUrl(pid)
                    })

                    actions.push({
                        type: 'divider'
                    })

                    actions.push({
                        label: 'Set image...',
                        action: function () {
                            var p = pinegrow.backend.makePlaceholderProjectFromProjectId(pid, project_user_slug, decodeURIComponent($li.attr('data-project-name')));

                            //hacks
                            var orig_getCurrentProject = pinegrow.getCurrentProject;
                            pinegrow.getCurrentProject = function () {
                                return p;
                            };

                            var done = function () {
                                pinegrow.getCurrentProject = orig_getCurrentProject;
                            }

                            pinegrow.focusRect.showFor$El($li);

                            var images_view = new PgQuickImageBrowser(null, 'Select the project image');
                            images_view.imageBrowser.show_sizes = false;
                            images_view.imageBrowser.enable_unsplash = false;

                            images_view.show($li);

                            images_view.onHide = function () {
                                done();
                            }

                            images_view.setOnSelected(function (url, img) {
                                $li.find('.placeholder, .project-thumb').remove();
                                p.thumb = url;
                                $li.prepend($(pinegrow.tutorialsManager.getHtmlForProjectThumb(p)));

                                pinegrow.backend.setProjectFeaturedImage(p, img.id, function () {
                                    pinegrow.showQuickMessage('Project image set.');
                                })
                            });
                            images_view.with_focus_rect = true;

                        }
                    })

                    if ($li.find('.project-thumb').length) {
                        actions.push({
                            label: 'Remove image',
                            action: function () {
                                var p = pinegrow.backend.makePlaceholderProjectFromProjectId(pid, project_user_slug, decodeURIComponent($li.attr('data-project-name')));

                                $li.find('.placeholder, .project-thumb').remove();
                                $li.prepend($(pinegrow.tutorialsManager.getHtmlForProjectPlaceholder(p)));

                                pinegrow.backend.setProjectFeaturedImage(p, 0, function () {
                                    pinegrow.showQuickMessage('Project image removed.');
                                })
                            }
                        })
                    }

                    actions.push({
                        type: 'divider'
                    })

                    actions.push({
                        label: 'Rename project...',
                        action: function () {
                            var p = pinegrow.backend.makePlaceholderProjectFromProjectId(pid, project_user_slug, decodeURIComponent($li.attr('data-project-name')));
                            pinegrow.backend.askToRenameProject(p, function () {
                                $li.attr('data-project', p.root.url.replace('https://', ''));
                                $li.attr('data-project-name', encodeURIComponent(p.name));
                                $li.find('.project-name').html(p.name);
                                pinegrow.showQuickMessage('Project renamed.');
                            })
                        }
                    })

                    actions.push({
                        type: 'divider'
                    })

                    actions.push({
                        type: 'header',
                        label: 'Danger zone'
                    })

                    actions.push({
                        label: 'Delete project...',
                        action: function () {
                            var p = pinegrow.backend.makePlaceholderProjectFromProjectId(pid, project_user_slug, decodeURIComponent($li.attr('data-project-name')));
                            var h = new PgProjectMenuHelper(p);
                            h.delete(function () {
                                $li.remove();
                                pinegrow.backend.getProjects();
                                pinegrow.showQuickMessage('Project deleted.');
                            });
                        }
                    })
                }

                var m = new CrsaContextMenu(actions, $li);
                m.showAt(e.pageX, e.pageY, $li.closest('.pg-start-screen'));
            }
        }

        var selectNewProjectTemplateWizard = function() {

            var tree = {
                choice: {
                    question: 'What would you like to build?',
                    back_name: 'the project type selection',
                    options: [
                        {
                            key: 'blocks',
                            name: 'Blocks for an existing site',
                            desc: 'Build one or more custom Gutenberg blocks and export them as a plugin that can be used on any WordPress site.',
                            thumb: 'blocks.png',
                            choice: {
                                question: 'Which CSS framework is the site using?',
                                options: [
                                    {
                                        key: 'custom',
                                        name: 'Custom CSS',
                                        project_id: 'new_html',
                                        thumb: 'blank_html.png',
                                        desc: 'The right choice for most cases. Select this if the site theme is not using Bootstrap or Tailwind CSS.'
                                    },
                                    {
                                        key: 'bootstrap',
                                        name: 'Bootstrap',
                                        project_id: 'new_bs',
                                        thumb: 'blank_bootstrap.jpg',
                                        desc: 'Select if the site theme is using Bootstrap.'
                                    },
                                    {
                                        key: 'tailwind',
                                        name: 'Tailwind CSS',
                                        project_id: 'new_tw',
                                        thumb: 'blank_tailwind.png',
                                        desc: 'Select if the site theme is using Tailwind CSS.'
                                    }
                                ]
                            }
                        },
                        {
                            key: 'theme',
                            name: 'A complete theme',
                            desc: 'Build a complete WordPress theme with one or more templates.',
                            thumb: 'theme.png',
                            pro: true,
                            pro_name: 'Theme project template',
                            choice: {
                                question: 'Which CSS framework do you want to use?',
                                options: [
                                    {
                                        key: 'bootstrap',
                                        name: 'Bootstrap',
                                        project_id: 'new_bs',
                                        thumb: 'blank_bootstrap.jpg',
                                        desc: 'Select to build the theme using Bootstrap CSS and JavaScript components.'
                                    },
                                    {
                                        key: 'tailwind',
                                        name: 'Tailwind CSS',
                                        project_id: 'new_tw',
                                        thumb: 'blank_tailwind.png',
                                        desc: 'Select to style the theme with Tailwind CSS.'
                                    },
                                    {
                                        key: 'custom',
                                        name: 'Plain CSS',
                                        project_id: 'new_html',
                                        thumb: 'blank_html.png',
                                        desc: 'Select to build the theme with custom CSS styling.'
                                    },
                                ]
                            }
                        },
                        {
                            key: 'blank',
                            name: 'Just give me a blank project',
                            desc: 'Start with a blank project to create a static HTML project, theme or a plugin.',
                            thumb: 'blank.png',
                            pro: true,
                            pro_name: 'Blank project template',
                            choice: {
                                question: 'Which CSS framework do you want to use?',
                                options: [
                                    {
                                        key: 'bootstrap',
                                        name: 'Bootstrap',
                                        project_id: 'new_bs',
                                        thumb: 'blank_bootstrap.jpg',
                                    },
                                    {
                                        key: 'tailwind',
                                        name: 'Tailwind CSS',
                                        project_id: 'new_tw',
                                        thumb: 'blank_tailwind.png',
                                    },
                                    {
                                        key: 'custom',
                                        name: 'Plain CSS',
                                        project_id: 'new_html',
                                        thumb: 'blank_html.png',
                                    },
                                ]
                            }
                        }
                    ]
                }
            }

            var current_path = '';

            var pathToChoice = function(path) {
                if(path === '') return tree;
                var a = path.split('-');
                var c = tree;
                var j = 0;
                while(c.choice && j < a.length) {
                    var has = false;
                    for(var i = 0; i < c.choice.options.length; i++) {
                        if (a[j] === c.choice.options[i].key) {
                            c = c.choice.options[i];
                            j++;
                            has = true;
                            break;
                        }
                    }
                    if(!has) return null;
                }
                return c;
            }

            var askQuestion = function(current) {
                var parent = getParent();
                var msg = '';

                if(parent) {
                    msg += `<p style="margin-bottom: 20px;"><a href="#" class="back">← Back to ${parent.choice.back_name}</a></p>`
                }

                $d.find('.project-type').html(current.name ? (' - ' + current.name) : '');

                msg += `<h2 class="pg-project-templates-title">${current.choice.question}</h2>`;

                msg += `<div class="pg-project-templates">`;
                var pp = current_path.length ? (current_path + '-') : '';
                current.choice.options.forEach(function(pt) {
                    msg += `<div data-path="${pp + pt.key}" style="overflow:visible;"><img src="images/project_templates/${pt.thumb || ''}">`;

                    if(pt.pro && !pinegrow.currentUser.isPro()) {
                        msg += `<div class="pg-pro-feature-badge" style="position: absolute;top:-4px;right:-4px;">Pro</div>`;
                    }

                    msg += `<div class="caption" style="padding-right:8px;"><h3>${pt.name}</h3><p>${pt.desc || ''}</p>`;

                    msg += `</div></div>`;
                })
                msg += `</div>`;

                if(!parent) {
                    msg += `<h2 class="pg-project-templates-title" style="margin-top:20px;margin-bottom: 10px;">Import a project`;

                    if(!pinegrow.currentUser.isPro()) {
                        msg += `<div class="pg-pro-feature-badge" style="margin-left:8px;">Pro</div>`;
                    }

                    msg += `</h2><p style="color:#777;">Upload a HTML project. Note: do not upload WordPress themes and plugins.</p><p><a class="upload" href="#">Import a project from a ZIP file</a></p>`;
                }

                $d.find('.wizard').html(msg);
            }

            var getParent = function() {
                if(current_path === '') return null;
                var a = current_path.split('-');
                a.pop();
                if(a.length === 0) return tree;
                return pathToChoice(a.join('-'));
            }

            var back = function() {
                var a = current_path.split('-');
                a.pop();
                current_path = a.join('-');
                var q = pathToChoice(current_path);
                askQuestion(q);
            }

            var list = [];

            pinegrow.dispatchEvent('on_get_new_project_templates', null, list);

            var $d = pinegrow.showAlert(`<div class="wizard"></div>`, 'Start a new project<span class="project-type"></span>', 'Cancel');

            askQuestion(tree);

            $d.on('click', '.pg-project-templates [data-path]', function(e) {
                e.preventDefault();
                var path = $(this).attr('data-path');
                var q = pathToChoice(path);
                current_path = path;

                if(q.choice) {
                    if(q.pro && !pinegrow.currentUser.isPro()) {
                        pinegrow.currentUser.onlyForProNotice(q.pro_name || q.name);
                    } else {
                        askQuestion(q);
                    }
                } else if(q.project_id) {
                    var pid = 'tutorial_' + q.project_id;
                    var a = pid.split('/');
                    var project_user_slug = null;

                    if(!pid.startsWith('https://')) {
                        if (a.length === 2) {
                            project_user_slug = a[0];
                            pid = a[1];
                        }
                        pid = 'https://' + pid;
                    }
                    $d.modal('hide');
                    _this.openProject(pid, function(project) {

                    }, project_user_slug, {
                        isCreatedWithNewProject: true,
                        source_project_id: pid
                    });
                }
            })

            $d.on('click', '.back', function(e) {
                e.preventDefault();
                back();
            })

            $d.on('click', '.upload', function(e) {
                e.preventDefault();

                if(!pinegrow.currentUser.isPro()) {
                    pinegrow.currentUser.onlyForProNotice('Project import')
                } else {
                    var dl = new PgProjectDownloader();
                    dl.uploadProject(null, function () {
                        $d.modal('hide');
                    });
                }
            })

            $d.find('.modal-body').css({
                backgroundColor: '#eeeeee',
                borderTop: '1px solid #dddddd'
            })

            $d.find('.modal-footer').css({
                marginTop: '0',
                borderTop: '1px solid #cfcfcf'
                //borderTop: '1px solid #dddddd'
            })

        }

        var selectNewProjectTemplate = this.selectNewProjectTemplate = function() {

            if(pgf.product === "WPPLUGIN") {
                selectNewProjectTemplateWizard();
                return;
            }

            var list = [];

            pinegrow.dispatchEvent('on_get_new_project_templates', null, list);

            var msg = '';
            //msg += `<div class="pg-tutorial-video-button"><i class="icon icon-play-round"></i> Learn how to create projects</div>`;

            msg += `<h2 class="pg-project-templates-title">Choose a project template to get started:</h2>`;

            msg += `<div class="pg-project-templates">`;

            list.forEach(function(pt) {
                msg += `<div data-pid="${pt.pid}"><img src="${pt.thumb}"><div class="caption" style="${pt.vid ? '' : 'padding-right:8px;'}"><h3>${pt.name}</h3><p>${pt.desc}</p>`;
                if(pt.vid) {
                    msg += `<div class="pg-tutorial-video-button" data-vid="${pt.vid}" data-subid="${pt.subid}"><i class="icon icon-play-round"></i> Guide</div>`;
                }
                msg += `</div></div>`;
            })
            msg += `</div>`;

            if(pinegrow.currentUser.isAdmin()) {
                msg += `<h2 class="pg-project-templates-title">Import a project</h2><p><a class="upload" href="#">Import a project from a ZIP file</a>`;
            }

            var $d = pinegrow.showAlert(msg, 'Start a new project', 'Cancel');

            $d.find('.pg-project-templates [data-pid]').on('click', function(e) {
                e.preventDefault();
                var pid = $(this).attr('data-pid');
                var a = pid.split('/');
                var project_user_slug = null;

                if(!pid.startsWith('https://')) {
                    if (a.length === 2) {
                        project_user_slug = a[0];
                        pid = a[1];
                    }
                    pid = 'https://' + pid;
                }
                $d.modal('hide');
                _this.openProject(pid, function(project) {

                }, project_user_slug, {
                    isCreatedWithNewProject: true,
                    source_project_id: $(this).attr('data-pid')
                });
            })

            $d.find('.upload').on('click', function(e) {
                e.preventDefault();
                var dl = new PgProjectDownloader();
                dl.uploadProject(null,function() {
                    $d.modal('hide');
                });
            })

            addTooltip($d.find('.pg-tutorial-video-button').on('click', function(e) {
                e.preventDefault();
                var vid = this.getAttribute('data-vid');
                var subid = this.getAttribute('data-subid');
                var video_player = new PgTutorialVideo(vid, pgTutorialSubtitles.get(subid));
                return false;
            }), 'Watch a video about how to use this template.');

            $d.find('.modal-body').css({
                backgroundColor: '#eeeeee',
                borderTop: '1px solid #dddddd'
            })

            $d.find('.modal-footer').css({
                marginTop: '0',
                borderTop: '1px solid #cfcfcf'
                //borderTop: '1px solid #dddddd'
            })

        }

        wireStartScreenEvents();
    }

    addTutorial(t) {
        this.list.push(t);
    }

    getTutorials() {
        return this.list;
    }

    getLessons(for_start) {
        var r = []
        this.list.forEach(function(t) {
            if(!for_start || t.showOnStartScreen) {
                r = r.concat(t.getLessons())
            }
        })
        return r;
    }

    getTutorialForLessonKey(key) {
        for(var i = 0; i < this.list.length; i++) {
            if(this.list[i].prefix && key.startsWith(this.list[i].prefix)) {
                return this.list[i];
            }
        }
        return null;
    }

    getLessonByKey(key) {
        var tutorial = this.getTutorialForLessonKey(key);
        return tutorial ? tutorial.findLesson(key) : null;
    }

    registerTutorials() {
        pinegrow.dispatchEvent('on_register_tutorials', null, this);
        pinegrow.recentFiles.populate();
        if(pgf.featured_projects) {
            pinegrow.backend.getPublicProjects();
        }
    }

    getHtmlForProjectThumb(pi, no_menu, with_name, with_desc) {
        var dmenu = no_menu ? '' : `<div class="project-dropdown"><i class="icon icon-down"></i></div>`;
        var html = `<div class="image-wrapper-4x3 project-thumb ${with_name ? 'placeholder' : ''}"><img src="${pi.thumb}" class="">`;
        with_name && (html += `<span><spanname>${pi.name}</spanname></span>`);

        if(with_desc && pi.short_description) {
            html += `<div class="description">${pi.short_description}</div>`;
        }
        html += this.getHTMLForProjectTags(pi)
        html += `${dmenu}</div>`;
        return html;
    }

    getHTMLForProjectTags(pi) {
        var _this = this;
        var html = '';
        if(pi && pi.tags && pi.tags.length) {
            html += `<tags class="tags">${pi.tags.map(function(i) {
                return _this.showTag(i.toLowerCase());
            }).join('')}</tags>`;
        }
        return html;
    }

    showTag(tag) {
        var map = {
            'bootstrap 5': '<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-bootstrap-fill" viewBox="0 0 16 16">\n' +
                '  <path d="M6.375 7.125V4.658h1.78c.973 0 1.542.457 1.542 1.237 0 .802-.604 1.23-1.764 1.23H6.375zm0 3.762h1.898c1.184 0 1.81-.48 1.81-1.377 0-.885-.65-1.348-1.886-1.348H6.375v2.725z"/>\n' +
                '  <path d="M4.002 0a4 4 0 0 0-4 4v8a4 4 0 0 0 4 4h8a4 4 0 0 0 4-4V4a4 4 0 0 0-4-4h-8zm1.06 12V3.545h3.399c1.587 0 2.543.809 2.543 2.11 0 .884-.65 1.675-1.483 1.816v.1c1.143.117 1.904.931 1.904 2.033 0 1.488-1.084 2.396-2.888 2.396H5.062z"/>\n' +
                '</svg>',
            'bootstrap': '<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-bootstrap-fill" viewBox="0 0 16 16">\n' +
                '  <path d="M6.375 7.125V4.658h1.78c.973 0 1.542.457 1.542 1.237 0 .802-.604 1.23-1.764 1.23H6.375zm0 3.762h1.898c1.184 0 1.81-.48 1.81-1.377 0-.885-.65-1.348-1.886-1.348H6.375v2.725z"/>\n' +
                '  <path d="M4.002 0a4 4 0 0 0-4 4v8a4 4 0 0 0 4 4h8a4 4 0 0 0 4-4V4a4 4 0 0 0-4-4h-8zm1.06 12V3.545h3.399c1.587 0 2.543.809 2.543 2.11 0 .884-.65 1.675-1.483 1.816v.1c1.143.117 1.904.931 1.904 2.033 0 1.488-1.084 2.396-2.888 2.396H5.062z"/>\n' +
                '</svg>',
            'tailwind css': '<svg style="width: 24px;" xmlns="http://www.w3.org/2000/svg" fill="none" viewBox="0 0 54 33"><g clip-path="url(#prefix__clip0)"><path fill="currentColor" fill-rule="evenodd" d="M27 0c-7.2 0-11.7 3.6-13.5 10.8 2.7-3.6 5.85-4.95 9.45-4.05 2.054.513 3.522 2.004 5.147 3.653C30.744 13.09 33.808 16.2 40.5 16.2c7.2 0 11.7-3.6 13.5-10.8-2.7 3.6-5.85 4.95-9.45 4.05-2.054-.513-3.522-2.004-5.147-3.653C36.756 3.11 33.692 0 27 0zM13.5 16.2C6.3 16.2 1.8 19.8 0 27c2.7-3.6 5.85-4.95 9.45-4.05 2.054.514 3.522 2.004 5.147 3.653C17.244 29.29 20.308 32.4 27 32.4c7.2 0 11.7-3.6 13.5-10.8-2.7 3.6-5.85 4.95-9.45 4.05-2.054-.513-3.522-2.004-5.147-3.653C23.256 19.31 20.192 16.2 13.5 16.2z" clip-rule="evenodd"/></g><defs><clipPath id="prefix__clip0"><path fill="#fff" d="M0 0h54v32.4H0z"/></clipPath></defs></svg>',
            'plain html': `<svg style="width: 16px;" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 512 512">
  <path fill="currentColor" d="M71,460 L30,0 481,0 440,460 255,512"/>
  <path fill="#FFF" d="M256,208 L181,208 176,150 256,150 256,94 255,94 114,94 115,109 129,265 256,265zM256,355 L255,355 192,338 188,293 158,293 132,293 139,382 255,414 256,414z"/>
  <path fill="#FFF" d="M255,208 L255,265 325,265 318,338 255,355 255,414 371,382 372,372 385,223 387,208 371,208zM255,94 L255,129 255,150 255,150 392,150 392,150 392,150 393,138 396,109 397,94z"/>
</svg>`,
            'interactions': `<i class="icon icon-113" style=""></i>`,
            'wordpress': `<i class="icon icon-WP" style=""></i>`,
            'tutorial': `<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-lightbulb" viewBox="0 0 16 16">
  <path d="M2 6a6 6 0 1 1 10.174 4.31c-.203.196-.359.4-.453.619l-.762 1.769A.5.5 0 0 1 10.5 13a.5.5 0 0 1 0 1 .5.5 0 0 1 0 1l-.224.447a1 1 0 0 1-.894.553H6.618a1 1 0 0 1-.894-.553L5.5 15a.5.5 0 0 1 0-1 .5.5 0 0 1 0-1 .5.5 0 0 1-.46-.302l-.761-1.77a1.964 1.964 0 0 0-.453-.618A5.984 5.984 0 0 1 2 6zm6-5a5 5 0 0 0-3.479 8.592c.263.254.514.564.676.941L5.83 12h4.342l.632-1.467c.162-.377.413-.687.676-.941A5 5 0 0 0 8 1z"/>
</svg>`,
            css: `<i class="icon icon-Pin15" style=""></i>`,
            html: `<i class="icon icon-KODA" style=""></i>`
        }


        if(map[tag]) {
            return `<tag title="${tag}">${map[tag]}</tag>`;
        } else {
            return tag;
        }
    }

    getHTMLForProjectInfo(pi) {
        var _this = this;
        var html = `<div class="project-info">`

        html += `<div class="name-author"><span><b>${pi.name}</b></span>`;

        if(pi.author) {
            html += `<a title="View profile" class="author" href="/@${pi.author}/" target="_blank">@${pi.author}</a>`;
        }

        html += `</div>`;

        if(pi.tags && pi.tags.length) {
            html += `<div class="tags">`;
            pi.tags.forEach(function(tag) {
                html += `${_this.showTag(tag)}`;
            })
            html += `</div>`;
        }
        html += `</div>`;
        return html;
    }

    getHtmlForProjectPlaceholder(pi, no_menu, with_desc) {
        var dmenu = no_menu ? '' : `<div class="project-dropdown"><i class="icon icon-down"></i></div>`;
        var c_sum = 0;
        for(var ci = 0; ci < pi.name.length; ci++) {
            c_sum += pi.name.charCodeAt(ci);
        }

        var dark = false;//pinegrow.getWorkspaceTheme() === 'gray';

        var c_hue = 200 + (c_sum % 180);
        var bck_color = chroma.hsl(c_hue, dark ? 0.2 : 0.4, dark ? 0.9 : 0.90);
        var txt_color = chroma.hsl(c_hue, dark ? 0.5 : 0.9, dark ? 0.75 : 0.75);

        if(dark) {
            var a = bck_color;
            bck_color = txt_color;
            txt_color = a;
        }

        var html = `<div class="placeholder" style="background-color:${bck_color.hex()};color:${txt_color.hex()};">${pgSVGIcons.pg_logo}<span style="">${shortenString(pi.name, 50, '', false)}</span>`;

        if(with_desc && pi.short_description) {
            html += `<div class="description">${pi.short_description}</div>`;
        }

        html += this.getHTMLForProjectTags(pi)

        html += `${dmenu}</div>`;

        return html;

        //color:${txt_color.hex()};
        return`<div class="placeholder" style="background-color:${bck_color.hex()};"><span style="color:${txt_color.hex()};">${shortenString(pi.name, 2, '', false)}</span>${dmenu}</div>`;
    }


    updateStartScreen(project_list, force) {
        var _this = this;
        //show info
        var hi_name = ' ' + pinegrow.currentUser.getName();
        if(!pinegrow.currentUser.isLoggedInAndActive()) {
            hi_name = '';
        }
        var $info = $('.pg-start-screen .info-inner').html(`<h3>${pgf.welcome_message.replace('#NAME', hi_name)}</h3>`);

        var $menu = $('<div class="info-inner-menu"></div>').appendTo($info);


        var actions = [];
        pinegrow.dispatchEvent('on_get_actions_for_start_menu', null, actions);

        pinegrow.showIllustratedMenu($menu, actions);


        var $ul = $('.pg-start-screen .lessons ul').html('');

        var html = '';

        var chapter = null;

        var demo_guest = !pinegrow.currentUser.isLoggedInAndActive();

        var tutorials = this.getTutorials();

        if(tutorials.length /*pgf.kids*/) {
            tutorials.forEach(function(tutorial) {
                if(!tutorial.showOnStartScreen) return;

                var lessons = tutorial.getLessons();
                html += `<div class="pg-lessons-chapter"><h3>${tutorial.name}</h3>`;

                if(tutorial.description) {
                    html += `<div class="pg-tutorial-description">${tutorial.description}</div>`;
                }

                html += `<div class="lessons-grid">`;

                for (var i = 0; i < lessons.length; i++) {

                    if (pgf.kids && lessons[i].chapter !== chapter) {
                        if (chapter) {
                            html += '</div></div>';
                        }
                        html += `<div class="pg-lessons-chapter"><h3>${lessons[i].chapter}</h3><div class="lessons-grid">`;
                        chapter = lessons[i].chapter;
                    }
                    if (i === 0 && lessons[i].status === 'locked') {
                        lessons[i].setStatus('active');
                    }

                    var a = lessons[i].name.split(' - ');
                    var name = `<b>${a[0]}</b>`;
                    a.shift();
                    a.length && (name += '<br>' + a.join(' - '));

                    var lesson_status = lessons[i].status;

                    if (lessons[i].demo && demo_guest && lesson_status === 'locked') {
                        lesson_status = 'demo';
                    }

                    if(lesson_status === 'locked' && lessons[i].tutorial.allowRandomAccess) {
                        lesson_status = 'none';
                    }

                    if (lessons[i].image && false) {
                        html += `<li class="lesson-item lesson-item-with-image lesson-status-${lesson_status}" data-key="${lessons[i].key}"><img src="images/kids/lessons/${lessons[i].image}.jpg"><div class="lesson-item-desc"><span>${name}</span>
<i class="lock icon icon-Lock_open"></i>
<i class="done icon icon-check"></i></div>`;
                    } else {
                        html += `<li class="lesson-item lesson-status-${lesson_status} ${lessons[i].icon ? '' : 'lesson-item-with-image'} ${lessons[i].short_description ? 'with-description' : ''}" data-key="${lessons[i].key}">`;

                        if (lessons[i].icon) {
                            html += `<span>${name}</span>`;
                        } else if(lessons[i].thumb) {
                            html += _this.getHtmlForProjectThumb(lessons[i], true, true, true);
                            //html += `<span><b>${name}</b></span>`;
                        } else {
                            html += _this.getHtmlForProjectPlaceholder(lessons[i], true, true);
                        }

                        html += `<i class="lock icon icon-Lock_open"></i>
<i class="done icon icon-check"></i>
<i class="preview icon icon-Lock_closed"></i>`;


                        if (lessons[i].icon) {
                            html += `<img src="images/kids/${lessons[i].icon}.png" class="lesson-img">`;

                            if (i === 0) {
                                html += `<img src="images/kids/rocket4.png" class="lesson-img" style="left:120%;bottom:120%;transform:rotateZ(45deg);">`;
                            }
                        }

                        //html += _this.getHTMLForProjectInfo(lessons[i]);
                    }

                    html += `</li>`;

                    //lessons[i].tooltip && addTooltip($li, lessons[i].tooltip);
                }
                if (chapter) {
                    html += '</div></div>';
                }
                html += '</div></div>';
            })
        }



        var images = [];
        for (var i = 1; i <= 7; i++) images.push(`rocket${i}.png`);
        for (var i = 1; i <= 15; i++) images.push(`food${i}.png`);

        $ul.html(html);

        var $ul = $('.pg-start-screen .projects ul.my-projects');//.html('');

        /*
        var $li = $(`<li class="lesson-item" data-project="new"><span><i class="icon icon-plus"></i>New project</span></li>`).appendTo($ul);

         */


        var showProjects = function (list) {
            var html = ``;

            //html += `<p>`

            if(!pinegrow.currentUser.isLoggedInAndActive()) {
                html += `<div class="local-storage-notice">You are not signed in. Your projects will be stored in the browser storage.</div>`;
            }

            if(list.length) {
                html += `<div class="lessons-grid">`;
                list.forEach(function (pi) {
                    //var rocket = pgStringToNumberRange(pi.project_id, 1, images.length);
                    html += `<li class="lesson-item lesson-item-with-image lesson-status-done" data-project="${pi.project_id}" data-project-name="${encodeURIComponent(pi.name)}" ${pi.is_file ? 'data-file="true"' : ''}>`;
                    if (pi.thumb) {
                        html += _this.getHtmlForProjectThumb(pi);
                    } else {
                        html += _this.getHtmlForProjectPlaceholder(pi);
                    }

                    html += `<div class="name-author"><span><b class="project-name">${pi.name}</b></span></div></li>`;
                    //$(html).appendTo($ul);
                })
                html += '</div>';
            } else {
                //html += `<div class="no-projects-message"><div><h4>You don't have any recent projects.</h4><p>Use the blue buttons above ↑ to create a new project, open an existing one or get up to speed with short &amp; fun interactive tutorials.</p></div></div>`;

                html += `<div class="no-projects-message"><h4>${pgf.no_projects_message}</h4><p>Use the blue buttons above <span>↑↑↑</span> to create a new project, open an existing project<br>or get up to speed with short &amp; fun interactive tutorials.</p></div>`;

            }
            html += '</div>';
            $ul.html(html);
        }

        if (project_list) {
            showProjects(project_list);
        } else {
            //should get projects?

            if (pinegrow.backend.shouldRefreshProjects() || force) {
                pinegrow.backend.getProjects(function (list) {
                    //showProjects(list);
                }, 'project')
            }
        }
    }

    addInteractiveElementsToHTML(adventure) {
        var _this = this;
        var tutorial = new PgTutorial();

        var map = {
            'Watch the video': '.pg-tutorial-video-button',
            'Text section': 'style:niga,[data-key=sec_open_Text_null]',
            'Font family': true,
            'Font size': true,
            'Text transform': true,
            'Text color': 'style:color',
            'Text align': true,
            'Text decoration': true,
            'Background section': 'style:niga,[data-key=sec_open_Background_null]',
            'Background image': true,
            'Background color': true,
            'Background position': true,
            'Background size': true,
            'Background attachment': true,
            'Background repeat': true,
            'Margin': 'style:niga,.margin-and-padding-container|.crsa-field-margin-top,.crsa-field-margin-left,.crsa-field-margin-bottom,.crsa-field-margin-right',
            'Margin left / right': 'style:niga,.margin-and-padding-container|.crsa-field-margin-left,.crsa-field-margin-right',
            'Margin left': 'style:niga,.margin-and-padding-container|.crsa-field-margin-left',
            'Margin top': 'style:niga,.margin-and-padding-container|.crsa-field-margin-top',
            'Margin bottom': 'style:niga,.margin-and-padding-container|.crsa-field-margin-bottom',
            'Margin right': 'style:niga,.margin-and-padding-container|.crsa-field-margin-right',
            'Padding': 'style:niga,.padding-container|.crsa-field-padding-left,.crsa-field-padding-right,.crsa-field-padding-top,.crsa-field-padding-bottom',
            'Padding left': 'style:niga,.padding-container|.crsa-field-padding-left',
            'Padding top': 'style:niga,.padding-container|.crsa-field-padding-top',
            'Padding bottom': 'style:niga,.padding-container|.crsa-field-padding-bottom',
            'Padding right': 'style:niga,.padding-container|.crsa-field-padding-right',
            'Flex Container': 'style:niga,[data-key=\'sec_open_Flex container_\']',
            'Dimensions section': 'style:niga,[data-key=sec_open_Dimension_null]',
            'Width': true,
            'Height': true,
            'Border': 'style:niga,.border-radius-container',
            'Border width': 'style:niga,.border-radius-container|.crsa-field-border-top input.crsa-input,.crsa-field-border-left input.crsa-input,.crsa-field-border-bottom input.crsa-input,.crsa-field-border-right input.crsa-input',
            'Border color': 'style:niga,.border-radius-container|.crsa-field-border-top .sp-replacer,.crsa-field-border-left .sp-replacer,.crsa-field-border-bottom .sp-replacer,.crsa-field-border-right .sp-replacer',
            'Border style': 'style:niga,.border-radius-container|.crsa-field-border-top .border-style,.crsa-field-border-left .border-style,.crsa-field-border-bottom .border-style,.crsa-field-border-right .border-style',
            'Border radius': 'style:niga,.border-radius-container|.crsa-field-border-top-left-radius,.crsa-field-border-bottom-left-radius,.crsa-field-border-top-right-radius,.crsa-field-border-bottom-right-radius',
            'right corners': 'style:niga,.border-radius-container|.crsa-field-border-top-right-radius,.crsa-field-border-bottom-right-radius',
            'Position': true,
            'Left': true,
            'Top': true,
            'Right': true,
            'Bottom': true,
            'Z-Index': 'style:z-index',
            'Create new CSS rule': 'style:niga,.pg-active-add-rule .create-dialog',
            'Style attribute': 'style:niga,.style-attribute-style-pane',
            'Classes': 'prop:rules',
            'Id': 'prop:element_id',
            'Save the style': 'style:niga,.to-rule-icon',
            'Save the inline style': 'style:niga,.to-rule-icon',
            'Float': true,
            'Clear': true,
            'Src': 'prop:src',
            'Href': 'prop:html-link-href',
            'Padding top / bottom': 'style:niga,.padding-container|.crsa-field-padding-top,.crsa-field-padding-bottom',
            'Padding left / right': 'style:niga,.padding-container|.crsa-field-padding-left,.crsa-field-padding-right',
            'Display': true,
            'Display section': 'style:niga,[data-key=sec_open_Display_null]',
            'Title': 'prop:page-title',
            'Description': 'prop:page-description',
            'Keywords': 'prop:page-keywords',
            'Author': 'prop:page-author',
            'Max width': true,
            'Min width': true,
            'Max height': true,
            'Min height': true,
            'Object fit': true,
            'Box shadow': 'style:niga,[data-key=\'sec_open_Box shadow_null\']',
            'Outline': true,
            'Position section': 'style:niga,[data-key=sec_open_Display_null]',
            'Position fields': 'style:niga,.position-container',
            'Equal borders': 'style:niga,.border-radius-container .change-all-icon',
            'Equal radius': 'style:niga,.border-radius-container .change-all-border-raduis',
            'add a shadow': 'style:niga,.box-shadow-container .add-btn',
            'Distance': 'style:niga,.crsa-field-box-shadow-distance-0',
            'Angle': 'style:niga,#collapse-box-shadow-0 .angle-selector',
            'Blur': 'style:niga,.crsa-field-box-shadow-blur-0',
            'Shadow color': 'style:niga,.crsa-field-box-shadow-color-0',
            'Tree panel': 'tooltip:tree_panel',
            'Style panel': 'tooltip:style',
            'Interactions panel': 'tooltip:interactions_panel',
            'Edit animation': 'animations:niga,.pg-ae-control-buttons .edit',
            'CSS grid': 'tooltip:css_grid',
            'WordPress menu': 'tooltip:wp_menu',
            'WordPress panel': 'tooltip:wp_panel',
            'Export the plugin': 'tooltip:wp_menu',
            'Project panel': 'tooltip:project_panel',
            'Insert button': 'tooltip:insert_menu',
            'Insert panel': 'tooltip:insert_menu',
            'Library panel': 'tooltip:insert_menu',
            'Library panel in the fixed UI': 'tooltip:library_panel',
            'Activate the text editor': 'tooltip:edit_text',
            'Inner HTML': 'prop:html-link-label',
            'Properties panel': 'tooltip:attributes_panel',
            'Refresh the page': 'tooltip:refresh_page',
            'Visual CSS editor settings': 'style:niga,.icon-tools',
            'Delete the element': 'tooltip:delete',
            'Delete it': 'tooltip:delete',
            'Preview the page': 'tooltip:preview'
        }
        var re_txt = '';
        $.each(map, function(field, sel) {
            if(re_txt.length) {
                re_txt += '|';
            }
            re_txt += field;
        })

        adventure = adventure.replace(/ICON_([a-z0-9\-_]+)/g, function(a, b, c) {
            return `<i class="icon icon-${b}"></i>`;
        })

        var re = new RegExp('<b>(' + re_txt + ')<\/b>', 'g');
        return adventure.replace(re, function(a, b, c) {
            var f = b.toLowerCase().replace(/\s/g, '-');
            var val = map[b];
            if(typeof val === 'string' && val.startsWith('tooltip:')) {
                val = val.replace('tooltip:', '');
                return `<b>${tutorial.tooltip(b, val)}</b>`;
            } else {
                var sel = val === true ? ('style:' + f + '|.crsa-field-' + f) : val;
                return `<b>${tutorial.field(b, sel)}</b>`;
            }
        })
    }

    openProject(url, done, project_user_slug, options, revision_id) {

        window.parent && window.parent.postMessage({
            action: 'pgStartedLoadingProject'
        }, '*');

        var open = function() {
            var p = new CrsaProject();
            if(options) {
                $.each(options, function(key, val) {
                    p[key] = val;
                })
            }
            p.fromServer(url, function(p) {
                pinegrow.setCurrentProject(p);
                done && done(p);
            }, project_user_slug, revision_id)
        }

        var cp = pinegrow.getCurrentProject();
        if(cp) {
            pinegrow.closeCurrentProject(false, true, function() {
                open();
            });
        } else {
            open();
        }
    }

    reopenProject(done) {
        var _this = this;
        var p = pinegrow.getCurrentProject();
        pinegrow.backend.saveAndCloseCurrentProject(function() {
            _this.openProject(p.url, done, p.isLoadedFromAnotherUser);
        });
    }
}