$(function() {
    var init = function() {

        if(!pgAddOnManager.isAddonActive('TW')) {
            console.log('TailwindUI is not active.');
            return;
        }

        var f = new PgFramework('tw.ui', 'Tailwind UI');

        f.description = 'Adds support for working with <a href="https://tailwindui.com/">TailwindUI</a> to Tailwind Visual Editor. Please note that <b>Tailwind UI library itself is not included</b> or distributed with Tailwind Visual Editor. You need to purchase the appropriate license directly from the <a href="https://tailwindui.com/">TailwindUI website</a>.';
        f.author = 'Pinegrow';
        f.author_link = 'http://pinegrow.com';

        f.setScriptFileByScriptTagId('plugin-tailwind-ui'); //get url if script is included directly into edit.html

        f.ignore_css_files = [/(^|\/)tailwind\-ui\.(css|less)/i, /(^|\/)tailwind\-ui\.min\.(css|less)/i];

        //Tell Pinegrow about the framework
        pinegrow.addFramework(f, 3);

        var def_all = new PgComponentType('tailwind.ui.all', 'Component', {
            selector: function (pgel) {
                return (pgel.getAttr('id') || '').startsWith('component-');
            },
            tags: 'major',
            priority : -1
        })

        f.addComponentType(def_all);

        var login = function(view) {
            var $d = pinegrow.showAlert(`<p>Please use this form to login to your TailwindUI account. Logging in directly in the Library panel doesn't work because of cookie issues.</p><p><b>After you login - or, if you are already logged in - click on the "Done" button at the bottom of this dialog box.</b></p><div class="vw" style="height:50vh;"></div>`, 'Login to TailwindUI', 'Cancel', `Done, I'm logged in!`, null, function() {
                chrome.cookies.getAll({
                    domain: '.tailwindui.com',
                    storeId: webview.getCookieStoreId()
                }, function(cookies) {
                    //console.log('Found cookies');
                    cookies.forEach(function(co) {

                        chrome.cookies.remove({
                            name: co.name,
                            url: 'https://tailwindui.com'
                        }, function() {
                            var c = {
                                domain: 'tailwindui.com',
                                expires: co.expirationDate,
                                httpOnly: co.httpOnly,
                                name: co.name,
                                path: co.path,
                                sameSite: 'None',
                                secure: co.secure,
                                value: co.value,
                                url: 'https://tailwindui.com'
                            }
                            //console.log('Cookie set', c);

                            pinegrow.devTools.setCookies([c]);
                            //chrome.cookies.set(c)
                        })

                    })
                })

                setTimeout(function() {
                    $d.modal('hide');
                    view.reload(view.item.url);
                }, 500);

                return false;
            } )

            var $vw = $(`<webview src="https://tailwindui.com/login" style="width:100%;height:650px;outline: 1px solid #cccccc;"></webview>`).appendTo($d.find('.vw'));

            //var $vw = $d.find('webview');
            var webview = $vw.get(0);
            //$vw.attr('src', 'https://tailwindui.com/login');
        }

        f.on_get_page_libs = function (page, list) {

            var component_name = null;
            var webview = null;

            list.push({
                name: 'TailwindUI',
                tab_name: 'UI',
                url: 'https://tailwindui.com/components',
                selector: "[id^='component-']",
                framework: f,
                removeClasses: [],
                direct: true,
                no_spreading: true,
                get_code: function($el, pick_all) {
                    component_name = '';
                    if(pick_all) {
                        return $el.get(0).outerHTML;
                    }
                    var srcdoc = $el.find('iframe').attr('srcdoc');
                    var p = new pgParser();
                    p.assignIds = false;
                    p.parse(srcdoc);
                    var body = p.rootNode.findOne('body');
                    if(body) {
                        component_name = $el.find('> header h2').text().trim();
                        return body.html();
                    }

                    return '<div>Sorry, could not get the component code.</div>';
                },
                on_pgel_created: function(pgel) {
                    pgel.findIncludingSelf('[src]').forEach(function(img) {
                        var url = img.getAttr('src');
                        try {
                            url = (new URL(url, 'https://tailwindui.com/')).toString();
                            img.setAttr('src', url);
                        } catch(err) {}
                    })
                    if(component_name) {
                        pgel.setAttribute('data-pg-name', component_name);
                    }
                },
                on_page_lib_loaded: function(doc, view) {
                    if(doc.body.querySelector('form[action="https://tailwindui.com/login"]')) {
                        login(view);
                    }

                    $(doc.body).on('click', 'a', function(e) {
                        var url = this.href;
                        if(crsaIsAbsoluteUrl(url)) {
                            e.preventDefault();
                            view.reload(url);
                        }
                    })
                },
                on_page_lib_menu: function(actions, view) {
                    actions.push({
                        label: 'Login to TailwindUI',
                        action: function() {
                            login(view)
                        }
                    })

                }
            })
        }




        f.addResource('https://cdn.jsdelivr.net/npm/@tailwindcss/ui@latest/dist/tailwind-ui.min.css', /tailwind\-ui(\.min|)\.css/i);
        f.addResource('https://rsms.me/inter/inter.css', /inter\.css/i);
        //f.addResource('https://cdn.jsdelivr.net/gh/alpinejs/alpine@v2.x.x/dist/alpine.min.js', /alpine(\.min|)\.js/i);

        var url = 'https://library.pinegrow.com/tailwindui/';
        var templates_url = url + 'template/';
        var thumbs_url = templates_url + 'screenshots/';

        var pages = [];

        pages.push({
            name: 'index.html',
            desc_name: 'Empty Tailwind UI page',
            url: templates_url + 'index.html',
            image_direct: thumbs_url + 'index.jpg'
        })

        f.addTemplateProjectFromDefinition({
            url: null,
            pages: pages,
            templateBrowserIndex: 1
        })


        //Register starting page template
        f.addTemplateProjectFromResourceFolder('template', null, 0.1);
    }

    //This is loaded with async. Perhaps pinegrow is already initialized?
    if(typeof pinegrow !== 'undefined' && pinegrow.ready) {
        init();
    } else {
        $('body').one('pinegrow-ready', function (e, pinegrow) {
            init()
        })
    }
});