(function( $ ) {
    $.fn.crsacss = function( method ) {
        if ( methods[method] ) {
            return methods[method].apply( this, Array.prototype.slice.call( arguments, 1 ));
        } else if ( typeof method === 'object' || ! method ) {
            return methods.init.apply( this, arguments );
        } else {
            $.error( 'Method ' +  method + ' does not exist on jQuery.crsacss' );
        }
    };

    var $iframes = [];
    var opts;

    var methods = {
        init : function( options ) {
            opts = $.extend( {}, $.fn.crsacss.defaults, options );

            $('body').on('crsa-page-added', function(event, iframe) {
                $iframes = $('iframe.content-iframe');

            });

            $('div.canvas').on('crsa-page-closed-removed', function(e, crsaPage) {
                $iframes = $('iframe.content-iframe');
            });

            return this.each(function(i,e){

            });
        },
        addStylesheetFromUrl : function(url, done) {
            var cs = pinegrow.stylesheets.getStylesheetByKey(url);
            if (cs) {
                if(done) done(cs);
            }
            else {
                cs = PgCSSStylesheet();
                cs.load(url, function() {
                    pinegrow.stylesheets.addCrsaStylesheet(cs);
                    if(done) done(cs);
                });
            }

            pinegrow.stylesheets.stylsheetUsed({
                url: url,
                pgStylesheet: cs
            });
        },
        loadLessStyles : function(iframe, done, include_dynamic) {
            var $iframe = $(iframe);
            var oldps = getCrsaPageStylesForPage($iframe, true /* dont try to reload */);
            if (oldps) {
                oldps.removeAllInlineSheets();
                oldps.destroy();
            }
            var cp = $iframe.data('crsa-page');

            if(cp) {
                var ps = new CrsaPageStyles(cp);

                cp.setStylesheets(ps);
                //$iframe.data('crsa-page-styles', ps);

                ps.loadAllStylesheets($iframe, function () {
                    console.log('all stylesheets loaded');
                    if (done) done();
                }, include_dynamic);
            } else {
                //page closed before loading done
                //if (done) done();
            }
        },
        renameRuleSelector : function(rule, new_selector, done, edited_by) {
            var cs = pgStylesheetHelper.getStylesheetFromNode(rule);
            cs.renameRuleSelector(rule, new_selector, done, edited_by);
        },
        addRule : function(rule) {
            var $rules = $();
            $iframes.each(function(i,e){
                var $frame = $(e);
                var ss = $frame.data('crsa-css-ss');
                ss.insertRule( rule, ss.cssRules ? ss.cssRules.length : 0 );
                $rules = $rules.add(ss.cssRules[ ss.cssRules.length - 1 ]);
            });
            return $rules;
        },
        css : function($rules, prop, value) {
            $rules.each(function(i,e) {
                $(e).css(prop, value);
            });
            return this;
        },
        find : function(sel) {
            var rules = [];
            $iframes.each(function(i,e){
                var $frame = $(e);
                var ss = $frame.data('crsa-css-ss');
                if(ss && ss.cssRules) {
                    var a = filterRules($(ss.cssRules), sel);
                    $.each(a, function (i, e) {
                        rules.push(e);
                    });
                }
            });
            var r = $(rules);
            return r;
        },
        getCss : function($rules, prop) {
            if($rules.length == 0) return null;
            var r = $rules[0];
            var s = r.style;
            if(s[prop]) return s[prop];
            return null;
        },
        getClassesForElement : function($el) {
            var r = [];
            if(!el) return r;

            var el = $el.get(0);
            if(typeof el.className !== 'string') return [];

            var list = el.className.split(/\s+/);

            for(var n = 0; n < list.length; n++) {
                if(list[n].indexOf('crsa-') != 0) {
                    r.push('.' + list[n]);
                }
            }
            return r;
        },
        getRulesForElement : function($el, add_classes, return_selectors) {
            var list = [];
            var selectors = [];
            var node = $el.get(0);
            if(node.ownerDocument.defaultView && node.ownerDocument.defaultView.getMatchedCSSRules) {
                var rules = node.ownerDocument.defaultView.getMatchedCSSRules(node, '');
                var dir = {};
                if(rules) {
                    var i = rules.length;
                    while (i--) {
                        dir[normalizeSelector(rules[i].selectorText)] = true;
                    }
                    for(var i = 0; i < pinegrow.stylesheets.crsaStylesheetsLength(); i++) {
                        var cs = pinegrow.stylesheets.getCrsaStylesheetById(i);
                        if(cs.ignore) continue;
                        dir = cs.filterRuleObjects(dir);
                        $.each(cs.getAllRules(), function(i,r) {
                            var sel;
                            if(r.selector && (sel = pinegrow.pgPostCSSHelper.getSelector(r)) in dir) {
                                list.push(r);
                                if(selectors.indexOf(sel) < 0) {
                                    sel = cs.filterSource(sel);
                                    if (selectors.indexOf(sel) < 0) {
                                        selectors.push(sel);
                                    }
                                }
                            }
                        });
                    }
                }
                if(add_classes) {
                    var cls = methods.getClassesForElement( $el);
                    for(var i = 0; i < cls.length; i++) {
                        var c = normalizeSelector(cls[i]);
                        if(selectors.indexOf(c) < 0) selectors.push(c);
                    }
                }
                return return_selectors ? selectors : list;
            } else {
                for(var i = 0; i < pinegrow.stylesheets.crsaStylesheetsLength().length; i++) {
                    var cs = pinegrow.stylesheets.getCrsaStylesheetById(i);
                    $.each(cs.getAllRules(), function(i,r) {
                        try {
                            if(r.selector && $el.is(r.selector)) {
                                var sel = normalizeSelector(r.selector);
                                list.push(r);
                                if(selectors.indexOf(sel) < 0) selectors.push(sel);
                            }
                        }
                        catch(err) {}
                    });
                    if(add_classes) {
                        var cls = methods.getClassesForElement( $el);
                        for(var i = 0; i < cls.length; i++) {
                            var c = normalizeSelector('.' + cls[i]);
                            var found = false;
                            for(j = 0; j < list.length; j++) {
                                var r = list[j];
                                if(typeof r == 'object' && r.selector) {
                                    if(c == normalizeSelector(r.selector)) {
                                        found = true;
                                        break;
                                    }
                                }
                            }
                            if(!found) {
                                selectors.push(c);
                            }
                        }
                    }
                }
            }
            return return_selectors ? selectors : list;
        },
        updateVarList : function() {
            $('body').trigger('crsa-var-updated');
        },
        showVariables : function() {
            $('body').trigger('crsa-var-updated');
        },
        showStylesheetsManager : function() {
            if($('.crsa-sm').length > 0) return;

            var $b = $('<ul/>', {class: 'crsa-man crsa-sm'});

            var notLoadedMessage = function() {
                showAlert("The stylesheet is not loaded. Load it (press on X in front of the stylesheet name in CSS panel) before renaming.", "Notice");
            }

            var showEditForm = function($li, save_as) {

                $li.append('<form class="form-inline" role="form" style="position:relative;">\
                    <div class="form-group" style="width:65%;">\
                    <label class="sr-only">Url</label>\
                    <div class="input-group" style="width:98%;">\
                    <input type="url" class="form-control url" placeholder="Url of .css file">\
                    <span class="input-group-addon"><i class="icon-folder" style="color:#757575;"></i></span>\
                    </div>\
                    <p class="help-block"></p>\
                    </div>\
                    <button type="submit" class="btn btn-primary add" style="position:absolute;top:2px;">Add</button>\
                    <button class="btn btn-link cancel">Cancel</button>\
                    </form>');
                $li.find('>a,>button').hide();
                var $form = $li.find('form');

                var $url = $form.find('input.url');
                var $select_file = $form.find('span.input-group-addon');
                var $url_group = $url.closest('.form-group');
                var $url_help = $url_group.find('.help-block').hide();

                var cs = $li.data('crsa-cs');
                if(cs) {
                    $url.val(cs.url);
                    $form.find('button.add').html('Rename');
                } else {
                    $form.find('button.cancel').hide();
                }

                if(isApp()) {
                    $select_file.on('click', function(e) {
                        e.preventDefault();
                        crsaChooseFile(function(url, file) {
                            $url.val(url);
                        }, (cs != null ? crsaGetNameFromUrl(cs.url) : null) || (save_as ? 'style.css' : false));
                    });
                } else {
                    $select_file.remove();
                }

                $form.find('button.add').on('click', function(e) {
                    e.preventDefault();
                    var $url = $(e.target).closest('form').find('input.url');
                    var url = $.trim($url.val());

                    if(url.length == 0) {
                        $url_group.addClass('has-error');
                        $url_help.html("Enter the url of existing or new stylesheet.").show();
                        return;
                    }
                    if(cs) {
                        var a = pinegrow.stylesheets.findCrsaStylesheetForUrl(url, true);
                        if(a.length > 0 && a[0] != cs) {
                            $url_group.addClass('has-error');
                            $url_help.html("File with this url is already loaded.").show();
                        } else if(!cs.loaded) {
                            notLoadedMessage();
                        } else {
                            var sheets = pinegrow.stylesheets.getAllSheetsForPages(cs);
                            cs.setUrl(url);

                            cs.reattachToAll(sheets, function() {
                                $('body').trigger('crsa-stylesheets-changed');
                                updateList();
                            });
                            if(isApp()) {
                                showAlert("Save the stylesheet or the page to which this stylesheet is attached to actually save it under this new name.", "Notice");
                            }
                        }
                    } else {
                        if(!crsaIsAbsoluteUrl(url)) {
                            var cp = pinegrow.getSelectedPage();
                            if (cp) {
                                url = crsaCombineUrlWith(cp.url, url);
                                $url.val(url);
                            }
                        }
                        if (pinegrow.stylesheets.getStylesheetByKey(url)) {
                            $url_group.addClass('has-error');
                            $url_help.html("File with this url is already loaded.").show();
                        } else {
                            $url_group.removeClass('has-error');
                            $url_help.hide();

                            var newcs = PgCSSStylesheet();

                            pinegrow.stylesheets.stylsheetUsed({
                                url: url,
                                pgStylesheet: newcs
                            });

                            newcs.load(url, function() {
                                pinegrow.stylesheets.addCrsaStylesheet(newcs);
                                if(newcs.changed && newcs.localFile) {
                                    newcs.save(newcs.localFile, function() {
                                        updateList();
                                    })
                                } else {
                                    updateList();
                                }
                                showAlert('Attach the new stylesheet to the page if you want to use it on that page. Btw, one stylesheet can be attached to multiple pages.', 'A Tip');
                            }, false, false, true, true);
                        }
                    }

                });

                $form.find('button.cancel').on('click', function(e) {
                    var cs = $li.data('crsa-cs');
                    if(cs) {
                        $form.remove();
                        $li.find('>a,>button').show();
                        e.preventDefault();
                    }
                });
            }

            var updateList = function() {
                $b.html('');
                var list = pinegrow.stylesheets.getAllCrsaStylesheets();
                // var selectedPage = $.fn.crsa('getSelectedPage');

                var showSourceFileUrl = function ($li, url) {
                    var $sourceFileUrl = $li.find('.source-file-url');
                    var $useScssLess = $li.find('.use-scss-and-less');

                    $sourceFileUrl.html('<b>Source file:</b> ' + url).show();
                    $useScssLess.data('already-used-file', true);
                    // $useScssLess.hide();

                    $('<a href="#">Cancel</>').appendTo($sourceFileUrl).on('click', function (e) {
                        e.preventDefault();

                        var $sourceFileUrl = $li.find('.source-file-url');
                        var $useScssLess = $li.find('.use-scss-and-less');

                        var cs = $sourceFileUrl.closest('li').data('crsa-cs');
                        var sourceStylesheet = cs.sourceStylesheet;

                        cs.resetSourceStylesheet();

                        var pages = pinegrow.stylesheets.getPagesForStylesheet(cs);
                        for (var i = 0; i < pages.length; i++) {
                            pinegrow.stylesheets.removeUsedStylesheet(sourceStylesheet, pages[i]);
                        }
                        sourceStylesheet && sourceStylesheet.remove(true);

                        $sourceFileUrl.html('').hide();
                        // $useScssLess.show();
                        $useScssLess.data('already-used-file', false);
                        cs.setSource(cs.source, function() {
                            $('body').trigger('crsa-stylesheets-changed');
                        });
                    });
                }

                for (var i = 0; i < list.length; i++) {
                    if (list[i].inline) continue;
                    var cs = list[i];
                    (function() {
                        var sourceStylesheet = null;
                        if (cs.sourceStylesheet) sourceStylesheet = cs.sourceStylesheet;

                        var $li = $('<li/>').appendTo($b).data('crsa-cs', cs);

                        var $a = $('<span/>', {}).html(cs.name + (cs.changed ? ' *' : '')).appendTo($li);

                        var $dup = $('<a/>', {href: '#'}).html("Duplicate").appendTo($li);
                        var $edit = $('<a/>', {href: '#'}).html("Rename").appendTo($li);
                        var $useScssLess = $('<a/>', {href: '#', class: 'use-scss-and-less'}).html('Use SCSS/LESS').appendTo($li);
                        var $close = $('<a/>', {href: '#'}).html("Unload").appendTo($li);
                        var $save = null;
                        if (cs.changed && isApp() && cs.loaded) {
                            $save = $('<a/>', {href: '#'}).html("<b>Save</b>").appendTo($li);
                        }

                        $('<div/>', {}).html('<b>CSS file:</b> ' + cs.url).appendTo($li);
                        var $sourceFileUrl = $('<div/>', {'class': "source-file-url" , 'style': "display:none;" }).html('').appendTo($li);

                        if (sourceStylesheet) {
                            showSourceFileUrl($li, sourceStylesheet.url);
                        }
                        if ($save != null) {
                            $useScssLess.data('should-save-file', true);
                        }
                        if (cs.url.startsWith('http') /* && !cs.mappedRemote */) {
                            $useScssLess.data('has-absolute-link', true);
                        }
                        // else {
                        //     if (crsaIsFileUrl(cs.url) && $save == null) {
                        //         $useScssLess.data('should-save-file', false);
                        //     }
                        // }

                        var $scssLessContainer = $('<div/>', {}).appendTo($li);

                        $dup.on('click', function(e) {
                            e.preventDefault();
                            var $li = $(e.delegateTarget).closest('li');
                            var cs = $li.data('crsa-cs');
                            if(!cs.loaded) {
                                notLoadedMessage();
                            } else {
                                cs.duplicate(function(new_cs) {
                                    pinegrow.stylesheets.addCrsaStylesheet(new_cs);
                                    updateList();
                                });
                            }
                        });

                        $edit.on('click', function(e) {
                            e.preventDefault();
                            var selectedPage = pinegrow.getSelectedPage();
                            if (!selectedPage.localFile) {
                                pinegrow.showAlert('<p>Please save the page before renaming the stylesheet.</p><p>After saving the page go to <b>Page -&gt; Manage stylesheets...</b> to rename the stylesheet.</p>', 'The page is not saved yet');
                                return;
                            }

                            var $li = $(e.delegateTarget).closest('li');
                            var cs = $li.data('crsa-cs');
                            if(!cs.loaded) {
                                notLoadedMessage();
                            } else {
                                showEditForm($li);
                            }
                        });

                        $close.on('click', function(e) {
                            e.preventDefault();
                            var $li = $(e.delegateTarget).closest('li');
                            var cs = $li.data('crsa-cs');

                            if (pinegrow.stylesheets.getPagesForStylesheet(cs).length > 0) {
                                showAlert("This stylesheet is attached to at least one open document. Close the document or detach the stylesheet before closing it.", "Can't close this stylesheet");
                            } else {
                                /*
                                if (cs.sourceStylesheet) {
                                    cs.sourceStylesheet.remove();
                                    cs.sourceStylesheet = null;
                                }
                                pinegrow.stylesheets.removeCrsaStylesheet(cs);

                                 */
                                //cs.unload();
                                cs.remove();
                                $('body').trigger('crsa-stylesheets-changed');
                                updateList();
                            }
                        });

                        if($save) {
                            $save.on('click', function(e) {
                                e.preventDefault();
                                var $li = $(e.delegateTarget).closest('li');
                                var cs = $li.data('crsa-cs');

                                var fs = require('fs');

                                if(!cs.localFile/*!crsaIsFileUrl(cs.url)*/) {
                                    showAlert("This is a remote stylesheet. Rename it first, then save it.", "Can't save this stylesheet");
                                } else {
                                    cs.save(cs.localFile/* crsaMakeFileFromUrl(cs.url) */, function(err) {
                                        if(err) {
                                            showAlert(err, "Can't save this stylesheet");
                                        } else {
                                            crsaQuickMessage(crsaGetNameFromUrl(cs.url) + ' saved!');
                                        }
                                        updateList();
                                    }, fs);
                                }
                            });
                        }

                        var $inlineForm = $(`<form class="form-inline" role="form" style="display:none;">
                            <div class="form-group" style="width:65%;">
                                <label class="sr-only">Url</label>
                                <div class="input-group" style="width:98%;">
                                    <input type="url" class="form-control url less-scss-input" placeholder="Url of scss/less file"> <span class="input-group-addon"><i class="icon-folder"></i></span></div>
                                <p class="help-block" style="display: none;"></p>
                            </div>
                            <button type="submit" class="btn btn-default use">Use</button>
                            <button class="btn btn-link cancel" style="display: none;">Cancel</button>
                        </form>`).appendTo($scssLessContainer);

                        var $urlInput = $inlineForm.find('input.url');
                        var $select_file = $inlineForm.find('.input-group-addon');
                        var $useBtn = $inlineForm.find('button.use');

                        $select_file.on('click', function(e) {
                            e.preventDefault();
                            var $urlInput = $(this).closest('li').find('input.url');

                            crsaChooseFile(function(url, file) {
                                $urlInput.val(url);
                            });
                        });

                        $useScssLess.on('click', function (e) {
                            e.preventDefault();
                            if (!selectedPage.localFile) {
                                pinegrow.showAlert('<p>Please save the page before using the SCSS/LESS stylesheet.</p><p>After saving the page go to <b>Page -&gt; Manage stylesheets...</b> to use SCSS/LESS stylesheet.</p>', 'The page is not saved yet');
                                return;
                            }

                            var $target = $(this);
                            if ($target.data('already-used-file')) {
                                pinegrow.showQuickError('This file already has source file.');
                                return;
                            }
                            if ($target.data('should-save-file')) {
                                pinegrow.showQuickError('You should save the file first.');
                                return;
                            }
                            if ($useScssLess.data('has-absolute-link')) {
                                pinegrow.showQuickError('This is a remote stylesheet.');
                                return;
                            }

                            $target.closest('li').find('form').show();
                        });

                        $useBtn.on('click', function (e) {
                            e.preventDefault();
                            var $li = $(this).closest('li');
                            var $inlineForm = $li.find('form');
                            var url = $li.find('input.url').val();
                            var cs = $li.data('crsa-cs');

                            var ext = crsaGetExtFromUrl(url);
                            var newStylesheet = null;

                            if (!crsaIsAbsoluteUrl(url)) {
                                url = crsaCombineUrlWith(cs.url, url);
                            }
                            showSourceFileUrl($(this).closest('li'), url);
                            $inlineForm.hide();

                            newStylesheet = pinegrow.stylesheets.getStylesheetByKey(url);

                            if (!newStylesheet) {
                                if (ext == "scss") {
                                    newStylesheet = PgSCSSStylesheet();
                                    cs.ignore = false;
                                }
                                else if (ext == "less") {
                                    newStylesheet = PgLESSStylesheet();
                                    cs.ignore = false;
                                }
                                else {
                                    showAlert('Please select file with <code>.scss</code> or <code>.less</code> extention', 'Can\'t use file');
                                    return;
                                }

                                newStylesheet.index = cs.getIndexForSourceStylesheet();
                                var baseFilePath = crsaMakeFileFromUrl(crsaGetBaseForUrl(url));
                                crsaCreateDirs(baseFilePath);

                                var filePath = crsaMakeFileFromUrl(url);
                                crsaCreateFile(filePath);

                                newStylesheet.load(url, function () {
                                    pinegrow.dispatchEvent('on_stylesheet_closed', null, cs);
                                    cs.setSourceStylesheet(newStylesheet);
                                    $('body').trigger('crsa-stylesheets-changed');
                                }, false, null, cs);
                            }
                            else {
                                cs.sourceStylesheet = newStylesheet;
                                newStylesheet.addCompiledStylesheet(cs);
                                $('body').trigger('crsa-stylesheets-changed');
                            }

                            pinegrow.stylesheets.stylsheetUsed({
                                url: url,
                                pgStylesheet: newStylesheet
                            }, 'source', cs);
                        });

                        var selectedPage = pinegrow.getSelectedPage();
                        if (selectedPage) {
                            var $attachDetachContainer = $('<div/>', {'class': 'attach-detach'}).appendTo($li);

                            var $attachLink = $('<a class="crsa-sm-attach" href="#">Attach to ' + selectedPage.name + '</a>').appendTo($attachDetachContainer);
                            var $detachLink = $('<a class="crsa-sm-attach" href="#">Detach From ' + selectedPage.name + '</a>').appendTo($attachDetachContainer);

                            var attachedDone = function () {
                                $attachLink.hide();
                                $detachLink.show();
                                $li.addClass('crsa-active');
                            }

                            var detachedDone = function () {
                                $detachLink.hide();
                                $attachLink.show();
                                $li.removeClass('crsa-active');
                            }

                            if (pinegrow.stylesheets.isPageHasStylesheet(selectedPage, cs)) {
                                attachedDone();
                            }
                            else {
                                detachedDone();
                            }

                            $attachLink.on('click', function (e) {
                                e.preventDefault();
                                var $li = $(e.delegateTarget).closest('li');
                                var cs = $li.data('crsa-cs');
                                selectedPage.stylesheets.attachTo(cs, null, function (e) {
                                    attachedDone();
                                    pinegrow.changeManager.didChange(null, 'Attach stylesheet', null, selectedPage);
                                    //delay a bit, give cs a time to load. otherwise status is disabled
                                    setTimeout(function() {
                                        $('body').trigger('crsa-stylesheets-changed');
                                    }, 500)
                                });
                                /*
                                cs.attachTo(selectedPage, null, function (e) {
                                    attachedDone();
                                });*/
                            });

                            $detachLink.on('click', function (e) {
                                e.preventDefault();
                                var $li = $(e.delegateTarget).closest('li');
                                var cs = $li.data('crsa-cs');
                                selectedPage.stylesheets.detachFrom(cs);
                                detachedDone();
                                pinegrow.changeManager.didChange(null, 'Detach stylesheet', null, selectedPage);
                            });
                        }
                    })()
                }
                var $li = $('<li/>').appendTo($b);

                var tabs = '<ul class="nav nav-tabs" style="margin-top:15px;">\
                    <li class="active"><a href="#ss-add-new" data-toggle="tab">New...</a></li>\
                    <li><a href="#ss-add-open" data-toggle="tab">Open file...</a></li>\
                    <li><a href="#ss-add-url" data-toggle="tab">Open url...</a></li>\
                </ul>\
                    <div class="tab-content">\
                        <div class="tab-pane active" id="ss-add-new"><p class="">Create a new stylesheet (if you select an existing stylesheet it will be loaded instead).</p></div>\
                        <div class="tab-pane" id="ss-add-open"><p class="">Open an existing stylesheet.</p></div>\
                        <div class="tab-pane" id="ss-add-url"><p class="">Enter an absolute or relative url of the stylesheet that you want to create or load.</p></div>\
                    </div>';

                if($tabs) $tabs.remove();
                if($tabsh) $tabsh.remove();
                $tabsh = $('<h4>Add or load a stylesheet:</h4>').appendTo($main);
                $tabs = $(tabs).appendTo($main);
                showEditForm($tabs.find('#ss-add-new'), true);
                showEditForm($tabs.find('#ss-add-open'));
                showEditForm($tabs.find('#ss-add-url'));
            }
            var $tabs;
            var $tabsh;
            var $main = $('<div class="manager"/>').append($b);
            var $d = makeModalDialog("Stylesheets Manager", "Close", null, $main, function () {
                pgSetDataOfObjsTo($main.find('> .crsa-sm > li'), 'crsa-cs', null);
                $('body').off('.cssman');
            });
            updateList();

            $('body').on('crsa-page-selected.cssman', function(e, selectedPage) {
                updateList();
            });
        },
        showFrameworkManager : function(crsaPage) {
            var text;
            if(crsaPage) {
                text = 'Select which libraries and plugins you want to use with this page in <b>Lib</b> and <b>Prop</b> panels.';
            } else {
                text = 'The list of libraries and plugins. Select Libraries &amp; Plugins Manager for a specific page to manage which plugins are active for that page.';
            }
            var $div = $('<div><p class="text-muted">' + text + '</p></div>');
            var $b = $('<ul/>', {class: 'crsa-man crsa-fm'}).appendTo($div);

            var loadComponentLibrary = function() {
                pinegrow.selectAndLoadLibrary(function(project, f) {
                    updateList();
                });
            }

            var showEditForm = function($li) {
                $li.append('<form class="form-inline load-plugin" role="form">\
                    <h4>Load your own plugins</h4>\
                    <p><b>PLEASE LOAD ONLY PINEGROW PLUGINS!</b> Loading anything else <b>WILL</b> break the app.</p>\
                    <div class="form-group" style="width:65%;">\
                    <label class="sr-only" for="exampleInputEmail2">Url</label>\
                    <div class="input-group" style="width:98%;">\
                    <input type="url" class="form-control url" id="exampleInputEmail2" placeholder=".js file">\
                    <span class="input-group-addon"><i class="icon-folder"></i></span>\
                    </div>\
                    <p class="help-block"></p>\
                    </div>\
                    <button type="submit" class="btn btn-default add">Add</button>\
                    <button class="btn btn-link cancel">Cancel</button>\
                    </form>');
                $li.find('>a,>button').hide();
                var $form = $li.find('form').hide();

                var $url = $form.find('input.url');
                var $select_file = $form.find('span.input-group-addon');
                var $url_group = $url.closest('.form-group');
                var $url_help = $url_group.find('.help-block').hide();

                $form.find('button.cancel').hide();

                if(isApp()) {
                    $select_file.on('click', function(e) {
                        e.preventDefault();
                        crsaChooseFile(function(url, file) {
                            $url.val(url);
                        });
                    });
                } else {
                    $select_file.remove();
                }

                $form.find('button.add').on('click', function(e) {
                    e.preventDefault();
                    var url = $.trim($url.val());

                    if(url.length == 0) {
                        $url_group.addClass('has-error');
                        $url_help.html("Select the file first.").show();
                        return;
                    }
                    var file = crsaMakeFileFromUrl(url);
                    if(pinegrow.findFrameworkWithUrl(url)) {
                        $url_group.addClass('has-error');
                        $url_help.html("This file is already loaded.").show();
                    } else {
                        $url_group.removeClass('has-error');
                        $url_help.hide();

                        if(confirm("Are you sure? Only Pinegrow PLUGINS and COMPONENTS can be loaded. Anything else (framework JavaScript files, etc...) WILL break the app.")) {

                            pinegrow.loadFrameworkFromUrl(url, true, function(err) {
                                if(err) {
                                    showAlert('Unable to load ' + file + '. ' + err, "Unable to load plugin.");
                                }
                                updateList();
                            });
                        }

                    }
                });
            }

            var updateList = function() {
                $b.html('');
                var list = pinegrow.getFrameworks();

                $.each(list, function(i, fm) {
                    if(!fm.show_in_manager) return true;
                    var $li = $('<li/>').appendTo($b).data('crsa-fm', fm);

                    var $a = $('<span/>', {}).html(fm.name).appendTo($li);
                    $('<i class="icon-check"></i>').appendTo($a);

                    var $add;
                    var $unload;
                    var activated = false;
                    if(crsaPage) {
                        if(crsaPage.frameworks.indexOf(fm) < 0) {
                            $add = $('<a/>', {class: 'crsa-sm-attach', href: '#'}).html('<i class="icon-plus"></i> Activate').appendTo($li);
                        } else {
                            $li.addClass('crsa-active');
                            $add = $('<a/>', {class: 'crsa-sm-attach', href: '#'}).html('<i class="icon-minus"></i> Deactivate').appendTo($li);
                            activated = true;
                        }

                        $add.on('click', function(e) {

                            var $li = $(e.delegateTarget).closest('li');
                            var fm = $li.data('crsa-fm');

                            var onInitDone = function(f) {
                                //added, check for resources
                                if(f.resources.has()) {
                                    pinegrow.showUpdateResourcesDialog(crsaPage,  pinegrow.getCurrentProject(), [f]);
                                }
                                updateList();
                            }

                            var idx = crsaPage.frameworks.indexOf(fm);
                            if(idx >= 0) {
                                crsaPage.removeFramework(fm);
                                $li.removeClass('crsa-active');
                                updateList();
                                crsaPage.setPageChanged(true);
                            } else {
                                if(!crsaPage.addFramework(fm, false, onInitDone)) {
                                    showAlert('Only one ' + fm.type + ' library or plugin can be added to a page at the same time. Remove the other one first.', "Could not add framework");
                                } else {
                                    $li.addClass('crsa-active');
                                    //if project, add to all open pages from project
                                    var project = pinegrow.getCurrentProject();
                                    if(project) {
                                        var first = true;
                                        project.forEachOpenPage(function(page) {
                                            page.addFramework(fm, false, first ? onInitDone : null);
                                            first = false;
                                        }, crsaPage);
                                    }

                                }

                                crsaPage.setPageChanged(true);
                            }

                            e.preventDefault();
                        });
                    }

                    if(activated && fm.resources.has()) {
                        $('<a/>', {class: 'crsa-sm-attach', href: '#'}).html('<i class="icon-281"></i> Resources').appendTo($li)
                            .on('click', function(e) {
                                e.preventDefault();
                                var $li = $(e.delegateTarget).closest('li');
                                var fm = $li.data('crsa-fm');
                                pinegrow.showUpdateResourcesDialog(crsaPage,  pinegrow.getCurrentProject(), [fm]);
                            });
                    }


                    if(fm.changed) {
                        // TODO: @add-icon save icon (fa-save)
                        $('<a/>', {class: 'crsa-sm-attach', href: '#'}).html('<i class="fa fa-save"></i> <b>Save</b>').appendTo($li)
                            .on('click', function(e) {
                                e.preventDefault();
                                var $li = $(e.delegateTarget).closest('li');
                                var fm = $li.data('crsa-fm');

                                if(fm.pluginUrl) {
                                    fm.save(fm.pluginUrl, function() {
                                        crsaQuickMessage("Library saved.");
                                        updateList();
                                        pinegrow.frameworksChanged();
                                    });
                                } else {
                                    crsaChooseFile(function(url, file) {
                                        fm.save(file, function() {
                                            crsaQuickMessage("Library saved.");
                                            updateList();
                                            pinegrow.frameworksChanged();
                                        });
                                    }, fm.getFileName());
                                }

                            });
                    }

                    if(fm.pluginUrl) {
                        // TODO: @add-icon times icon (fa-times)
                        $('<a/>', {class: 'crsa-sm-attach', href: '#'}).html('<i class="fa fa-times"></i> Unload').appendTo($li)
                            .on('click', function(e) {
                                e.preventDefault();
                                var $li = $(e.delegateTarget).closest('li');
                                var fm = $li.data('crsa-fm');
                                if(fm.changed) {
                                    showAlert("Plugin has unsaved changes. Sure you want to close it?", "Unsaved changes", "Cancel", "Close it", null, function() {
                                        pinegrow.unloadFramework(fm);
                                        updateList();
                                    });
                                } else {
                                    pinegrow.unloadFramework(fm);
                                    updateList();
                                }
                            });
                        $('<div/>', {}).html(fm.pluginUrl).appendTo($li);
                    }


                });
                var $li = $('<li/>').appendTo($b);
                showEditForm($li);
            }

            updateList();

            $div.append('<p><a href="#" class="load-lib">Load component library</a> - <a href="#" class="show-load-plugins">Load plugin</a> - <a href="https://pinegrow.com/docs/developers/" target="_blank" class="link">Learn how to create Pinegrow plugins.</a></p>')

            $div.find('a.load-lib').on('click', function(e) {
                e.preventDefault();
                loadComponentLibrary();
            });

            $div.find('a.show-load-plugins').on('click', function(e) {
                e.preventDefault();
                $d.find('.load-plugin').show();
            });
            var $d = makeModalDialog(crsaPage ? "Libraries &amp; Plugins Manager for " + crsaPage.name : "Libraries &amp; Plugins Manager", "Close", null, $div);

            if(isApp()) {
                $d.find('a.link').on('click', function(e) {
                    var url = $(e.delegateTarget).attr('href');
                    if(url) {
                        e.preventDefault();
                        var gui = require('nw.gui');
                        var url = $(e.delegateTarget).attr('href');
                        gui.Shell.openExternal(url);
                    }
                })
            }

            $('body').on('crsa-framework-loaded', function(e, fm) {
                updateList();
            });
        }
    };

    var filterRules = function($rules, s ){
        var o;
        if( !s ) s = /./;//just keep them all.
        if( s.split ){
            o = $.trim(s).toLowerCase().split(/\s*,\s*/);
            s = function(){
                if(!this.selectorText) return false; //MTCH
                return !!$.grep( this.selectorText.toLowerCase().split(/\s*,\s*/), function( sel ){
                    return $.inArray( sel, o ) != -1;
                }).length;
            };
        }else if( s.exec ){//string regex, or actual regex
            o = s;
            s = function(){ return o.test(this.selectorText); };
        }
        var r = $([]).pushStack($.grep($rules, function( e, i ){
            return s.call( e, i );
        }));
        return r;
    };

    // Plugin defaults – added as a property on our plugin function.
    $.fn.crsacss.defaults = {
        title : 'crsa'
    };
})( jQuery );



function getIframeDocument(iframe) {
    return iframe.contentDocument || iframe.contentWindow.document;
}

function getCrsaPageStylesForPage($iframe, dont_reload) {
    if(!$iframe) return null;
    var cp = $iframe.data('crsa-page');
    if(!cp) return null;
    return cp.getStylesheets(dont_reload);
    //return $iframe.data('crsa-page-styles');
}

function normalizeSelector(sel) {
    var a = sel.split(',');
    for(var i = 0; i < a.length; i++) {
        a[i] = $.trim(a[i].replace("\n","").replace(/\s{2,}/ig," ").replace(/\s*>\s*/, '>'));
    }
    return a.join(',');
}

function getRuleName(r) {
    if(r.selector) return r.selector;
    if(r.raw) shortenString(r.raw, 18);
    return 'Rule';
}

var _crsaCss_re_class = /^\.(-?[_a-zA-Z]+[_a-zA-Z0-9-]*)$/;

function getClassFromSelector(sel) {
    var m = sel.match(_crsaCss_re_class);
    return m ? m[1] : null;
}