
var crsaPageUid = 1;

var CrsaPageTemplate = function() {

    this.type = null;
    this.header = null;
    this.html = null;

    this.detect = function(source) {
        //check liquid
        if(source.indexOf('---') != 0) return false;

        var a = source.split('---');
        if(a.length < 3) return false;

        this.type = 'liquid';
        this.header = '---' + a[1] + '---';
        a.splice(0,2);

        this.html = a.length == 1 ? a[0] : a.join('---');
        return true;
    }

    this.getTemplateSource = function(full_html) {
        if(!this.type) return full_html;
        return this.header + "\n" + full_html;
    }

    this.getHtml = function() {
        return this.html;
    }

}

var pgDefaultClassSetter = {

    hasClass: function(pgel, cls) {
        return pgel.hasClass(cls);
    },

    addClass: function(pgel, cls, sorter) {
        pgel.addClass(cls, sorter);
    },

    removeClass: function(pgel, cls, sorter) {
        pgel.removeClass(cls, sorter);
    },

    canRemoveClass: function(pgel, cls) {
        return pgel.canRemoveClass(cls);
    },

    getClasses: function(pgel) {
        return pgel.getClasses();
    }
}

var CrsaPage = function(not_interactive) {
    this.url = null;
    this.wrapper_url = null;

    this.wrapper_selector = null;
    this.name = null;
    this.localFile = null;
    this.localFileWatcher = null;
    this.lastSavedAt = 0;
    this.$iframe = null;
    this.$page = null;
    this.activeView = null;
    this.visible = true;
    this.loading = false;
    this.changed = true;
    this.force_close = false;
    this.live_update = null;
    this.save_parent = false;
    this.uid = crsaPageUid++;

    this.undoHistory = new PgUndoHistory(this);

    this.webapp = false;

    this.exists = false;
    this.crsaProjectTemplate = null;
    this.allowReload = false;
    this.readOnly = false;
    this.treeTop = null;
    this.treeRepaintOnShow = null;
    this.treeCurrentRoot = null;
    this.load_source = null;
    this.frameworks_added = 0;
    this.breakpoints = [];
    this.openedInEditor = true;
    this.onlyCode = false;
    this.code = null;
    this.isBackgroundPage = false;

    this.data = {};

    this.firstLoad = true;

    this.tree = new PgPageTree(this);

    this.currentFrameworkHandlerReturnValues = {};

    this.sourceNode = null;
    this.selected_element_pgid = null;

    this.scrollMode = true; //crsaGetCustomLogic().scrollMode;
    this.deviceWidth = 1024;
    this.zoom = 1.0;

    this.template = new CrsaPageTemplate();
    this.frameworks = [];
    this.animationsStopped = false;
    this.javascriptEnabled = pinegrow.getSetting('javascript-enabled', '1') == '1';

    this.assetsCache = {};

    this.delayedTasks = new PgDelayedTasks();

    this.show_head_in_tree = false;
    this.externalWindow = null;

    this.mappedRemote = false;
    this.mappedRemoteFirstSave = false;

    this.classSetter = pgDefaultClassSetter;

    this.default_wrapper_selector = 'article,main,body';

    this.handleWrapperOnPageLoad = function(statusFunc) {
        var page = this;

        var project = pinegrow.getCurrentProject();
        if(!project) {
            statusFunc('done');
            return;
        }

        var onWrapperUrlError = function(error) {
            pinegrow.showPageWrapperDialog(page.url, `<div class="alert alert-info"><p>Unable to load the wrapper page <code>${page.wrapper_url}</code>. The error is "${error.toString()}".</p><p>Select another wrapper page or disable the wrapper.</p></div>`, true, page)

            page.wrapper_node = null;
        }

        var default_wrapper_selector = this.default_wrapper_selector;

        var pi = pinegrow.getCurrentProjectInfo();

        if(page.wrapper_url) {
            //load wrapper

            var url_test = page.wrapper_url;

            !isApp() && pgSetCookie('pinegrow-no-admin-bar', '1', 10000);

            pgFetchRemoteContent(url_test).then(function (data) {
                var html;
                try {
                    if (data === undefined) throw 'Error getting remote file ' + url_test;
                    html = new TextDecoder().decode(data);
                } catch (err) {
                    html = 'Error getting remote file ' + url_test + err;
                    onWrapperUrlError(err);
                    pgRemoveCookie('pinegrow-no-admin-bar');
                    statusFunc('done');
                    return;
                }
                !isApp() && pgRemoveCookie('pinegrow-no-admin-bar');

                var p = new pgParser();
                p.assignIds = false;
                p.parse(html);
                page.wrapper_node = p.rootNode;

                if(!page.wrapper_node.findOneKeepFindOrder(page.wrapper_selector || default_wrapper_selector)) {
                    onWrapperUrlError(`The element with the selector <code>${page.wrapper_selector}</code> can not be found on the page.`)
                }
                pi.setSettingForUrl(page.url, 'open-with-wrapper', !!page.wrapper_node);
                pi.save();
                statusFunc('done');
            }).catch(function (error) {
                console.log('Error getting remote file ' + url_test, error);
                onWrapperUrlError(error)
                statusFunc('done');
            })
        } else {
            if(pgf.use_page_wrappers) {
                pi.setSettingForUrl(page.url, 'open-with-wrapper', false);
                pi.save();
            }
            statusFunc('done');
        }
    }

    this.getHTMLNodeForBrowserWithSupportForWrapper = function(html_full_node) {

        var page = this;
        var node = html_full_node;

        if(page.wrapper_node) {
            var wrapper_base_url = page.wrapper_url;

            var wrapper_node = page.wrapper_node.clone(true);

            var wrapper_pgel = wrapper_node.findOneKeepFindOrder(page.wrapper_selector || page.default_wrapper_selector);
            if(wrapper_pgel) {
                var body_scripts = []
                var body_els = []

                wrapper_pgel.setAttribute('data-pg-wrapper', 'true');

                var body = node.findOne('body');
                body.children.forEach(function(child) {
                    if(child.isElement) {
                        if(child.tagName === 'script') {
                            var script_url = child.getAttribute('src') || child.getAttribute('data-pg-script-src');
                            if(!script_url || wrapper_node.findOne(`script[src="${script_url}"]`) === null) {
                                body_scripts.push(child)
                            }
                        } else {
                            body_els.unshift(child)
                        }
                    }
                })

                var append_to_head = [];

                var head = node.findOne('head');
                if(head) {
                    var wrapper_head = wrapper_node.findOne('head')
                    if(head && wrapper_head) {
                        head.find('style,link[rel="stylesheet"]').forEach(function(s) {
                            var style_url = s.getAttribute('href') || s.getAttribute('data-pg-script-src');
                            var style_in_wrapper = style_url ? wrapper_node.findOne(`link[href="${style_url}"]`) : null;

                            append_to_head.push(s);

                            if(style_in_wrapper) {
                                style_in_wrapper.remove();
                            }
                        })

                        var basehref = wrapper_head.findOne('base');
                        if(basehref) {
                            wrapper_base_url = basehref.getAttribute('href') || wrapper_base_url;
                        }
                    }
                }

                var wrapper_body = wrapper_node.findOne('body');

                wrapper_body.setAttribute('data-pg-id', body.getId());

                wrapper_node.goThroughAllUrls(function(url, attr, pgel) {
                    try {
                        if(!crsaIsAbsoluteUrl(url)) {
                            url = new URL(url, wrapper_base_url).toString();
                        }
                    } catch(err) {

                    }
                    return url;
                })

                append_to_head.forEach(function(el) {
                    wrapper_head.append(el);
                })

                body_els.forEach(function(s) {
                    wrapper_pgel.prepend(s);
                })

                body_scripts.forEach(function(s) {
                    //if(!wrapper_body.findOne(`script`))
                    wrapper_body.append(s);
                })

                html_full_node = wrapper_node;
            } else {

            }
        }

        return html_full_node;
    }


    this.stylesheets = null;

    this.setStylesheets = function(ps) {
        this.stylesheets = ps;
    }

    this.getStylesheets = function(dont_reload) {
        if(!this.stylesheets && this.openedInEditor && !dont_reload) {
            //strange???
            this.updateStylesheetsList();
            return new CrsaPageStyles(this); //dummy
        }
        return this.stylesheets;
    }

    this.getAttachedStylesheets = function(options) {
        /*
         options = {
         loaded: true,
         ignored: true,
         imported: true,
         inline: true
         }
         */
        if(!this.stylesheets) return [];
        return this.stylesheets.getList(options);
    }

    this.hasAttachedStylesheetWithUrl = function(url) {
        var list = this.getAttachedStylesheets();
        for(var i = 0; i < list.length; i++) {
            if(list[i].url && list[i].url === url) return list[i];
        }
        return null;
    }

    this.getPageLibs = function() {
        var list = [];
        this.frameworks.forEach(function(f) {
            var libs = f.getPageLibs();
            libs.forEach(function(lib) {
                lib.framework = f;
                list.push(lib);
            })
        })
        this.callFrameworkHandler('on_get_page_libs', list);

        list.forEach(function(i) {
            i.dynamic = true;
        })
        return list;
    }

    this.setUrl = function(url) {
        this.url = pgDecodeUrl(url);
    }

    this.isRemote = function() {
        return this.url && !crsaIsFileUrl(this.url);
    }

    this.isNew = function() {
        return !this.localFile && (!this.isRemote() || this.pageIsRemoteTemplate);
    }

    this.getName = function() {
        return this.tab_name || this.name;
    }

    this.scrollToHash = function(hash) {
        pageViews.forEach(function(pv) {
            pv.scrollToHash(hash);
        })
    }

    //Dynamic elements
    var dynamicPgels = [];

    this.registerDynamicElement = function(el) {

        if(this.wrapper_node && !el.closest('[data-pg-id]')) {
            return null;
        }

        var $el = $(el);
        var tag = el.tagName.toLowerCase();

        var html = '<' + tag;

        var attrs = [];

        for (var i = 0; i < el.attributes.length; i++) {
            var a = el.attributes[i];
            if (a.specified) {
                var astr = a.name;
                if(a.value !== null && a.value !== undefined) {
                    var quote = a.value.indexOf('"') >= 0 ? `'` : `"`;
                    astr += '=' + quote + a.value + quote;
                }
                attrs.push(astr);
            }
        }
        if(attrs.length) {
            html += ' ' + attrs.join(' ');
        }

        if(pgSingleTags.indexOf(tag) >= 0) {
            html += '>';
        } else {
            html += `></${tag}>`;
        }
        var p = new pgParser();
        p.assignIds = true;
        p.parse(html);

        var pgel = p.rootNode.children[0];

        pgel.dynamic = true;

        pgel.page = this;
        pgel.document = this.sourceNode;
        pgel.$el = $el;
        $el.attr('data-pg-id', pgel.getId());

        dynamicPgels.push(pgel);

        return pgel;
    }
    //End dyn el

    var pageViews = [];

    this.addPageView = function(pv) {
        pageViews.push(pv);
    }

    this.removePageView = function(pv) {
        var idx = pageViews.indexOf(pv);
        if(idx >= 0) {
            pageViews.splice(idx, 1);
        }
    }

    this.getPageViews = function() {
        return pageViews;
    }

    this.getPageViewOf$DOMElement = function($el, $iframe) {
        $iframe = $iframe || getIframeOfElement($el);
        if(!$iframe) return null;
        for(var i = 0; i < pageViews.length; i++) {
            if($iframe.get(0) == pageViews[i].get$Iframe().get(0)) return pageViews[i];
        }
        return null;
    }

    this.getPageViewById = function(pvid) {
        for(var i = 0; i < pageViews.length; i++) {
            if(pageViews[i].getId() == pvid) return pageViews[i];
        }
        return null;
    }

    this.closePageView = function(pv) {
        var idx = pageViews.indexOf(pv);
        if(idx >= 0) {
            pageViews.splice(idx, 1);
        }
    }

    this.get$Iframes = this.getPageView$Iframes = function() {
        return pageViews.map(function(pv) {
            return pv.get$Iframe();
        })
    }

    this.showInExternalWindow = function() {
        var gui = require('nw.gui');

        this.externalWindow = gui.Window.open('empty.html', {
            focus: true
        });

        this.externalWindow.on('close', function () {
            _this.externalWindow.close(true);
            _this.externalWindow = null;
        });
        this.externalWindow.setAlwaysOnTop(true);

        var loaded = false;
        this.externalWindow.on('loaded', function() {
            if(!loaded) {
                //debugger;
                _this.allowReload = true;
                var $win_body = $(_this.externalWindow.window.document.body);
                _this.$page.appendTo($win_body.find('#win-content'));
                loaded = true;
            }
        });
    }

    this.fromFile = function(file, only_code, watch_changes) {

        if(typeof only_code == 'undefined') only_code = false;

        var file_url = isApp() ? crsaMakeUrlFromFile( file ) : file;
        this.setUrl(file_url);
        this.name = crsaGetNameFromUrl( this.url );
        this.localFile = file;
        this.visible = false;
        this.changed = false;
        this.exists = true;
        this.openedInEditor = false;
        this.onlyCode = only_code;

        var profile = new CrsaProfile(true);

        var html;
        var project = pinegrow.getCurrentProject();

        if(!project || !project.loaded_from_backend) {
            var fs = require('fs');

            try {
                var data = fs.readFileSync(file);
                html = data.toString('utf8');
            } catch (err) {
                html = '';
            }
        } else {
            //browser
            html = project.getContentOfUrl(this.url);
        }

        if(only_code) {
            this.code = html;
        } else {
            var p = new pgParser();
            p.assignIds = false;
            p.nodeCatalogue = null;
            p.parse(html);
            this.sourceNode = p.rootNode;

            this.callGlobalFrameworkHandler('on_page_parsed_in_proxy', this.sourceNode, this.url);

            profile.show('FROM FILE PARSE ' + this.localFile);
        }
        if(watch_changes) {
            this.setLocalFile(this.localFile);
        }
    }

    this.getCodeForSave = function() {
        return this.getCode(pinegrow.getSetting('keep-collapsed', '0') == '1', true /* for save */);
    }

    this.getCode = function(keep_collapsed, for_save) {

        if(this.onlyCode) {
            return this.code;
        } else {
            return this.sourceNode.toStringForCodeEditorOrSave(this, for_save, keep_collapsed);
        }
    }

    this.setCode = function(code, undoInfo, skipDidChange) {
        var project = pinegrow.getCurrentProject();
        if(this.onlyCode) {
            this.code = code;
            if(project && project.loaded_from_backend) {
                project.setContentOfUrl(this.url, code);
            }
        } else {
            var oldValue = (undoInfo && this.sourceNode) ? this.sourceNode.toStringWithIds(true) : null;
            var new_node;
            if(code instanceof pgParserNode) {
                new_node = code;
            } else {
                var p = new pgParser();
                p.assignIds = true;
                p.parse(code);
                new_node = p.rootNode;
            }
            if(this.sourceNode) {
                this.sourceNode.withoutEvents(function() {
                    _this.sourceNode.remove();
                })
            }
            this.sourceNode = new_node;

            if(undoInfo) {
                var pgEvent = new PgNodeEvent(this.sourceNode, 'setCode', null, {
                    oldValue: oldValue
                }, this);
                var undoEvent = new PgUndoEvent(pgEvent);
                this.undoHistory.addEvent(undoEvent);
                if(!skipDidChange) {
                    pinegrow.changeManager.didChange(this.sourceNode, undoInfo.action || 'Set code', undoInfo, this)
                }
            }
        }
    }

    this.addPageStateToUndoHistory = function(oldValue) {
        var pgEvent = new PgNodeEvent(this.sourceNode, 'setCode', null, {
            oldValue: oldValue
        }, this);
        var undoEvent = new PgUndoEvent(pgEvent);
        this.undoHistory.addEvent(undoEvent);
    }

    this.setLocalFile = function(file) {
        var project = pinegrow.getCurrentProject();
        var watch = !project || !project.loaded_from_backend;

        if(watch) {
            this.stopWatchingFileForChanges();
        }

        this.localFile = file;

        if(watch) {
            this.watchFileForChanges();
        }
    }

    this.setJavascriptEnabled = function(value) {
        if(this.javascriptEnabled != value) {
            this.javascriptEnabled = value;
            //pinegrow.setSetting('javascript-enabled', value ? '1' : '0');
            this.refresh();
        }
    }

    this.hasScripts = function(at_least) {
        if(typeof at_least == 'undefined') at_least = 4;
        var list = this.sourceNode.find('script');
        return list.length >= at_least;
    }

    this.makeRelativeUrl = function(url) {
        return crsaMakeLinkRelativeTo(url, crsaMakeUrlAbsolute(_this.url));
    }

    this.makeAbsoluteUrl = function(relative_url) {
        if(crsaIsAbsoluteUrl(relative_url)) return relative_url;
        return require('url').resolve(this.url, relative_url);
        //return crsaGetBaseForUrl(this.url) + '/' + relative_url;
    }


    this.watchFileOnFileChanged = function(curr, prev) {
        //console.log('ccccccccc');
        //return;
        var fs = require('fs');

        var updatePage = function(v) {
            var ret = _this.applyChangesToSource(v, false, false, true);

            if(ret.updated) {
                crsaQuickMessage("Auto refreshed!");
                if(ret.changes && ret.changes.length == 1) {
                    ret.replace_selected.forEach(function(d) {
                        pinegrow.selectedElements.replace(d.old_pgel, d.new_pgel);
                    })
                }
            } else if(ret.changes && !ret.update) {
                //whole document is changed
                pinegrow.selectedElements.cleanDeleted();
                _this.refreshWithUpdateStylesheets();
            }
            _this.setPageChanged(false);

            _this.callFrameworkHandler('on_page_loaded_from_external');
        }

        if(curr.mtime > prev.mtime && !_this.live_update && pinegrow.getSetting('auto-reloading', '1') == '1') {
            var ct = (new Date()).getTime();
            if(ct - _this.lastSavedAt > 5000) {
                console.log(_this.localFile + ' changed');

                var code = fs.readFileSync(_this.localFile, {encoding: "utf8"});

                pinegrow.codeEditors.isCodeForUrlSameInExternalEditor(_this.url, code, function(same) {

                    if(_this.watchReloadDialog) {
                        try {
                            _this.watchReloadDialog.modal('hide');
                            _this.watchReloadDialog = null;
                        } catch(err) {}
                    }

                    if(same) return;

                    if(_this.changed) {

                        _this.watchReloadDialog = showAlert('<p><b>' + _this.name + '</b> was modified outside of Pinegrow. The file has unsaved changes. Do you want to reload it?</p><p><em>You can disable auto-reloading in Support -&gt; Settings.</em></p>',  "Unsaved file modified outside of Pinegrow", "Don't reload", "Reload", function() {
                            _this.watchReloadDialog = null;
                        }, function() {
                            _this.watchReloadDialog = null;
                            updatePage(code);
                        });
                    } else {
                        updatePage(code);
                    }
                })

            } else {
               // console.log(_this.localFile + ' changed in PG');
            }
        }
    }

    this.isWatchingFileForChanges = false;
    this.watchReloadDialog = null;

    this.watchFileForChanges = function() {
        if(this.localFile && pinegrow.getSetting('auto-reloading', '1') == '1') {
            var fs = require('fs');
            //this.localFileWatcher = fs.watchFile(this.localFile, {persistent: true, interval: 1102}, this.watchFileOnFileChanged);
            if(this.nodeFileWatcher) {
                this.nodeFileWatcher.close();
                this.nodeFileWatcher = null;
            }
            try {
                this.nodeFileWatcher = fs.watch(this.localFile, {persistent: true}, function (event, filename) {
                    //console.log(event, filename);
                    _this.delayedTasks.run('reloadFile', function () {
                        var curr = {
                            mtime: 1
                        }
                        var prev = {
                            mtime: 0
                        }
                        _this.watchFileOnFileChanged(curr, prev);
                    }, 10);
                })
                this.isWatchingFileForChanges = true;
            } catch(err) {}
        }
    }

    this.stopWatchingFileForChanges = function() {
        if(this.localFile && this.isWatchingFileForChanges) {
            //var fs = require('fs');
            //fs.unwatchFile(this.localFile, this.watchFileOnFileChanged);
            if(this.nodeFileWatcher) {
                this.nodeFileWatcher.close();
                this.nodeFileWatcher = null;
            }
            this.isWatchingFileForChanges = false;
        }
    }

    this.getAsset = function(url) {
        if(url in this.assetsCache) return this.assetsCache[url];
        return null;
    }

    this.addAsset = function(url, data, mime) {
        this.assetsCache[url] = {data: data, mime: mime};
    }

    this.clearAssetsCache = function() {
        this.assetsCache = {};
    }

    var _this = this;
    var pageChangedTimer = null;
    var cssid = 'crsa-inline-styles';

    var componentTypes = null;
    var componentTypesDict = {};
    var libSections = null;
    var actionsSections = null;

    this.parsedStatus = null;
    this.parsedError = null;

    var whenParsedQueue = [];

    this.whenParsed = function(loadFunc, doFunc, errorFunc) {
        if(this.parsedStatus == 'done') {
            doFunc(this);

        } else if(!this.parsedStatus == 'error') {
            errorFunc(this, _this.parsedError);

        } else if(this.parsedStatus === null) {
            _this.parsedStatus = 'progress';

            whenParsedQueue.push({doFunc: doFunc, errorFunc: errorFunc});

            loadFunc(this, function(status, error) {
                _this.parsedStatus = status;
                _this.parsedError = error;

                whenParsedQueue.forEach(function(funcs) {
                    (status == 'done' ? funcs.doFunc : funcs.errorFunc)(_this, _this.parsedError);
                })
                whenParsedQueue = [];
            })

        } else if(this.parsedStatus = 'progress') {
            whenParsedQueue.push({doFunc: doFunc, errorFunc: errorFunc});
        }
    }


/*
    if(pinegrow.sourceParser && !not_interactive) {
        $('body').on('crsa-server-page-loaded', onServerPageLoaded);
    }
*/

    this.onClose = function() {
        $('body')
            //.off('crsa-server-page-loaded', onServerPageLoaded)
            .off('.page' + this.uid);

        this.setPageChanged(false, true);
        this.setLocalFile(null); //will remove file watchers

        if(this.tree) {
            this.tree.destroy();
            this.tree = null;
        }

        pinegrow.selectedElements.removeForPage(this);

        if(pinegrow.pageHasIsolatedElement(this)) {
            pinegrow.isolateElement(null);
        }

        if(this.codeEditor) {
            this.codeEditor.closeTab();
            this.codeEditor = null;
        }

        this.undoHistory.destroy();
        this.undoHistory = null;

        pinegrow.undoManager.purgePage(this);

        var ps = this.getStylesheets();
        if(ps) {
            ps.destroy();
            this.setStylesheets(null);
        }
        //this.$iframe.data('crsa-page-styles', null);

        if(this.$iframe) {
            this.$iframe.off('.crsa');
        }

        pageViews = null;
        this.activeView = null;

        //pgParserNodeCatalogueInstance.logStat();
        if(this.sourceNode) {
            this.sourceNode.withoutEvents(function () {
                _this.sourceNode.remove();
            })
            this.sourceNode = null;
        }
        //pgParserNodeCatalogueInstance.logStat();

        var project = pinegrow.getCurrentProject();
        if(project) {
            project.removeBackgroundPageForUrl(this.url);
        }

        dynamicPgels.forEach(function(pgel) {
            pgel.removeWithoutEvents();
        })
        dynamicPgels = [];

        this.data = {};
    }

    this.askClose = function(on_closed) {
        var ask_close_html = "This page has unsaved changes. Are you sure you want to close it?";
        var ask_close_css = "Stylesheets attached to this page have unsaved changes. Are you sure you want to proceed?";

        var askOnCloseDialog = function(ask, leave_css) {
            var $alert = showAlert(ask,  "Page has unsaved changes", "Cancel", "Save &amp; Close", null, function() {
                if(canSavePage(_this)) {
                    if(!isAppForProject()) {
                        pinegrow.showQuickMessage('Please wait, saving...');
                        pinegrow.backend.saveCurrentProject(function(error) {
                            if(!error) {
                                _this.view.closeTab();
                                on_closed && on_closed();
                            } else {
                                pinegrow.showQuickError('Error saving the project.');
                            }
                        })
                    } else {
                        _this.save(function () {
                            _this.view.closeTab();
                            on_closed && on_closed();
                        }, true, true);
                    }
                }
            })
            $('<a href="#" class="btn pull-left">Don\'t save</a>').appendTo($alert.find('.modal-footer')).on('click', function(e) {
                e.preventDefault();
                _this.view.closeTab();
                //$.fn.crsapages('closePage', crsaPage, /*leave_css*/);
                $alert.modal('hide');
            });
        }

        var ask = null;
        if(_this.hasCssChanges()) {
            ask = ask_close_css;
        }
        if(_this.changed && !_this.force_close) {
            ask = ask_close_html;
        }

        var project = pinegrow.getCurrentProject();

        if(project && project.loaded_from_backend) ask = false; //saving changes is handled on the project level in the browser

        if(ask) {
            askOnCloseDialog(ask);
        } else {
            _this.view.closeTab();
            on_closed && on_closed();
        }
    }


    this.close = function () {
        $.fn.crsapages('closePage', this);
    }

    this.show = function(flash) {
        this.view.show(flash);
    }

    this.refreshDisplay = function() {
        //disabled30
        return;
/*
        var t = this.$iframe.css('transform');
        this.$iframe.css('transform', 'none');

        var h = this.$iframe.get(0).offsetHeight;
        this.$iframe.css('transform', t);
        */

        var $contents = this.$iframe.contents();
        var o = $contents.scrollTop();
        $(this.getBody()).redraw();
        $contents.scrollTop(o);
/*
        var element = this.getBody();
        var disp = element.style.display;
        element.style.display = 'none';
        var trick = element.offsetHeight;
        element.style.display = disp;*/
    }

    this.getPageUrlFromLoadedUrl = function(url) {
        url = pinegrow.getOriginalUrl(url);
        url = url.replace(/&*pgid=[0-9]+/, '');
        url = url.replace(/&*pgpvid=[0-9]+/, '');
        url = url.replace('?pgedit=1', '').replace('&pgedit=1', '').replace('pgedit=1', '');
        if(url.length && url.charAt(url.length-1) == '?') {
            url = url.substr(0, url.length-1);
        }
        return url;
    }

    this.getPreviewUrl = function(noids) {
        var url = crsaMakeUrlAbsolute(this.url);

       // var a = ['pglive=1'];
        //if(noids) a.push('pgnoids=1');

        //url += 'pgid=' + this.uid + '&pglive=1';
        //url = crsaAppendQueryToUrl(url, a);
        url = pinegrow.getProxyUrl(url);

        return url;
    }

    var resetFrameworksCache = function() {
        componentTypes = null;
        libSections = null;
        actionsSections = null;
        componentTypesDict = null;
    }

    this.getComponentTypes = function() {
        if(!componentTypes) {
            componentTypes = [];

            for(var i = 0; i < this.frameworks.length; i++) {
                var f = this.frameworks[i];
                if(f.not_main_types) continue;
                $.each(f.getComponentTypes(), function(key, def) {
                    componentTypes.push(def);
                });
            }
            componentTypes.sort(function(a,b) {
                return a.priority - b.priority;
            });
            componentTypesDict = {};
            $.each(componentTypes, function(i, def) {
                if(!componentTypesDict[def.type]) {
                    componentTypesDict[def.type] = def;
                }
            });
        }
        return componentTypes;
    }

    this.setComponentTypesInformation = function(info) {
        componentTypes = info.list;
        componentTypesDict = info.dict;
        this.frameworks = info.frameworks;
    }

    this.getComponentTypesInformation = function() {
        this.getComponentTypes();
        return {list: componentTypes, dict: componentTypesDict, frameworks: this.frameworks};
    }

    this.getPageStylesheets = function() {
        return this.stylesheets;
    }

    this.getPageStylesheet = function(re) {
        var list = this.getPageStylesheets().stylesheets;
        for(var i = 0; i < list.length; i++) {
            if(typeof re == 'string') {
                if(list[i].url == re) return list[i];
            } else {
                if (list[i].url.match(re)) return list[i];
            }
        }
        return null;
    }

    this.hasLinkedStylesheet = function(url) {
        var links = this.sourceNode.find('link');
        for(var i = 0; i < links.length; i++) {
            if(links[i].getAttr('href') == url) return links[i];
        }
        return null;
    }

    this.getViewHTMLForElement = function(pgel, disable_js) {
        disable_js = true;
        var origjs = this.javascriptEnabled;
        if(disable_js) this.javascriptEnabled = false;
        pgCurrentRequestContext.setCurrentRequestContext(this.url, this.sourceNode);

        var html = pgel.toStringWithIds(true, pinegrow.getFormatHtmlOptions(), pgCurrentRequestContext.createProxyUrlNodeOutputFilter);
        //console.log(html);

        if(false && pgel.hasAttr('react')) {
            var React = require('react');
            var ReactDOMServer = require('react-dom/server');

            var js = Babel.transform(html, {
                plugins: ["transform-react-jsx"]
            });

            html = ReactDOMServer.renderToString(eval(js.code));
/*
            var p = new pgParser();
            p.assignIds = false;
            p.parse(html);
            p.rootNode.children[0].setAttr('data-pg-id', pgel.getId());

            html = p.rootNode.children[0].toStringWithIds(true);*/
            console.log(html);
        }
        this.javascriptEnabled = origjs;
        return html;
    }

    this.getViewInnerHTMLForElement = function(pgel, disable_js) {
        if(pgel.tagName == 'script') {
            return pgel.content;
        }
        if(pgel.tagName == 'style') {
            var cs = pinegrow.stylesheets.findCrsaInlineStylesheet(pgel);
            var str = pgel.content;
            if(cs) {
                str = cs.getSourceForRules(null, true);
            }
            str = pgAddProxyLinksToStyleAttribute(str);
            return str;
        }
        disable_js = true;
        var origjs = this.javascriptEnabled;
        if(disable_js) this.javascriptEnabled = false;
        pgCurrentRequestContext.setCurrentRequestContext(this.url, this.sourceNode);
        var html = pgel.toStringWithIds(true, pinegrow.getFormatHtmlOptions(), pgCurrentRequestContext.createProxyUrlNodeOutputFilter, true);
        this.javascriptEnabled = origjs;
        return html;
    }

    this.addStylesheet = function(url, update_list, done, replace_pgel, add_before) {
        var loaded = this.openedInEditor;

        if(loaded && !replace_pgel) {
            var cs = this.getPageStylesheet(url);
            if (cs) {
                if (done) done(cs);
                return false;
            }
        }
        url = this.makeRelativeUrl(url);
        if(this.hasLinkedStylesheet(url)) {
            if(done) done();
            return false;
        }
        var code = '<link rel="stylesheet" href="' + url + '">';
        var head_pgel = this.sourceNode.findOne('head');

        if(!head_pgel) {
            if(done) done();
            return false;
        }

        var css_el = pgCreateNodeFromHtml(code);

        if(replace_pgel) {
            //replace_pgel.replaceWith(css_el);
            replace_pgel.setAttribute('href', url);
        } else {
            var added = false;
            if(add_before) {
                var before = head_pgel.findOne(`link[href="${add_before}"]`);
                if(before) {
                    css_el.insertBefore(before);
                    added = true;
                }
            }
            if(!added) {
                head_pgel.append(css_el);
            }
        }

        if(update_list && loaded) {
            var update = function() {
                _this.updateStylesheetsList(function() {
                    cs = _this.getPageStylesheet(url);
                    if(done) done(cs);
                });
            }
            if(pgf.load_srcdoc) {
                _this.refresh(update);
            } else {
                update();
            }

        }
        return true;
    }

    this.updateStylesheetsList = function(done) {
        //do a bit of delay
        this.delayedTasks.run('updateStylesheetsList', function() {
            $.fn.crsacss('loadLessStyles', _this.$iframe.get(0), function() {
                $('body').trigger('crsa-stylesheets-changed');
                _this.addCrsaStyles();
                if(done) done();
            });
        }, 500);
    }

    this.hasScriptInHeaderOrFooter = function(url, footer) {
        var node = this.sourceNode.findOne(footer ? 'body' : 'head');
        if(!node) return null;
        var list = node.find('script');
        for(var i = 0; i < list.length; i++) {
            if(list[i].getAttr('src') == url) return list[i];
        }
        return null;
    }

    this.addScript = function(url, footer, replace_pgel, code) {
        url = this.makeRelativeUrl(url);

        if(footer === undefined) footer = false;

        var has_in_wrong_place = this.hasScriptInHeaderOrFooter(url, !footer);
        if(has_in_wrong_place && !replace_pgel) {
            has_in_wrong_place.remove();
        }

        if(this.hasScriptInHeaderOrFooter(url, footer) && !replace_pgel) return false;

        var html = code ? `<script>${code}</script>` : '<script src="' + url + '"></script>';
        var body = this.sourceNode.findOne('body');

        if(!body) return false;

        var head = this.sourceNode.findOne('head');

        if(!head) return false;

        var script = pgCreateNodeFromHtml(html);

        if(replace_pgel) {
            //replace_pgel.replaceWith(script);
            if(!code) {
                replace_pgel.setAttribute('src', url);
            } else {
                replace_pgel.html(code);
            }
        } else {
            if (footer) {
                body.append(script);
            } else {
                head.append(script);
            }
        }
        return true;
    }

    this.getTypeDefinition = function(type) {
        if(componentTypesDict == null) {
            this.getComponentTypes();
        }
        return componentTypesDict[type] ? componentTypesDict[type] : null;
    }

    this.hasStylesheet = function(re) {
        var doc = this.$iframe.get(0).contentDocument || this.$iframe.get(0).contentWindow.document;

        for(var i=0; i<doc.styleSheets.length; i++) {
            var sheet = doc.styleSheets[i];

            if(sheet.href) {
                if(sheet.href.match(re)) return sheet;
            } else if(sheet.ownerNode) {
                var href = sheet.ownerNode.getAttribute('data-pg-style-url');
                if(href && href.match(re)) return sheet;
            }
        }
        return false;
    }

    this.getAllDOMStylesheets = function() {
        var r = [];
        var doc = this.$iframe.get(0).contentDocument || this.$iframe.get(0).contentWindow.document;

        for(var i=0; i<doc.styleSheets.length; i++) {
            var sheet = doc.styleSheets[i];

            if(sheet.href) {
                r.push(sheet);
            }
        }
        return r;
    }

    this.stylesheetsThatContain = function(str_re, re, done, force_reload) {
        var list = [];
        var matches = [];
        var async = false;

        var ps = this.getStylesheets();
        if (!ps) {
            done(list);
            return;
        }

        var all = ps.getAllCrsaStylesheets();
        var idx = -1;

        var doOne = function() {
            idx++;
            if(idx < all.length) {
                var cs = all[idx];
                var m = null;

                if((cs.fileType == 'css') && (!re || (cs.url && cs.url.match(re)))) {
                    if (!force_reload && (cs.loaded || cs.getSource())) {
                        if(m = (cs.getSource() || '').match(str_re)) {
                            list.push(cs);
                            matches.push(m);
                        }
                        doOne();
                    } else {
                        if(cs.localFile) {
                            try {
                                var fs = require('fs');
                                var code = fs.readFileSync(cs.localFile, {encoding: 'utf8'});
                                if(code && (m = code.match(str_re))) {
                                    list.push(cs);
                                    matches.push(m);
                                }
                                cs.source = code;
                            } catch(err) {

                            }
                            doOne();
                        } else if(cs.url) {
                            try {
                                async = true;
                                $.ajax({
                                    url: cs.url,
                                    data: null,
                                    dataType: 'text'
                                }).done(function (code) {
                                    if(code && (m = code.match(str_re))) {
                                        list.push(cs);
                                        matches.push(m);
                                    }
                                    cs.source = code;
                                    doOne();
                                }).fail(function () {
                                    doOne();
                                });
                            } catch(err) {
                                doOne();
                            }
                        } else {
                            doOne();
                        }
                    }
                } else {
                    doOne();
                }
            } else {
                done(list, matches, async);
            }
        }

        doOne();
    }

    this.hasScript = function(re) {
        var doc = this.$iframe.get(0).contentDocument || this.$iframe.get(0).contentWindow.document;

        var found = false;
        $(doc).find('script[src]').each(function(i, s) {
            var src = $(s).attr('src');
            if(src && src.match(re)) {
                found = true;
                return false;
            }
        })
        return found;
    }

    this.getLibSections = function() {
        if(!libSections) {
            libSections = [];
            for(var i = 0; i < this.frameworks.length; i++) {
                var f = this.frameworks[i];
                $.each(f.getLibSections(), function(i, section) {
                    libSections.push(section);
                });
            }
        }
        return libSections;
    }

    this.getActionsSections = function() {
        if(!actionsSections) {
            actionsSections = [];
            for(var i = 0; i < this.frameworks.length; i++) {
                var f = this.frameworks[i];
                $.each(f.getActionsSections(), function(i, section) {
                    actionsSections.push(section);
                });
            }
        }
        return actionsSections;
    }

    this.getAllTypes = function($el, pgel, skip_actions, only_auto_updatable) {
        var s = new Date().getTime();
        var r = [];
        for(var i = 0; i < this.frameworks.length; i++) {
            var f = this.frameworks[i];
            if(only_auto_updatable && !f.auto_updatable) continue;
            var a = f.getTypes($el, pgel, skip_actions);
            if(a.length) {
                r = r.concat(a);
            }
        }
        total_in_get_type += new Date().getTime() - s;
        return r;
    }

    var total_in_get_type = 0;

   /*
    setInterval(function() {
        if(total_in_get_type > 0) console.log('total_in_get_type = ' + total_in_get_type + 'ms');
    }, 10000);
*/

    this.getMainType = function($el, pgel, skip_actions, only_auto_updatable) {
        //var s = new Date().getTime();

        if($el instanceof pgParserNode) {
            pgel = $el;
            $el = null;
        }

        for(var i = 0; i < this.frameworks.length; i++) {
            var f = this.frameworks[i];
            if(only_auto_updatable && !f.auto_updatable) continue;
            var def = f.getType($el, pgel, skip_actions);
            //total_in_get_type += new Date().getTime() - s;
            if(def) return def;
        }
        //total_in_get_type += new Date().getTime() - s;
        return null;
    }

    this.getTypeThatFiltersTree = function(pgel) {
        for(var i = 0; i < this.frameworks.length; i++) {
            var f = this.frameworks[i];
            if(f.has_types_that_filter_tree) {
                var def = f.getTypeThatFiltersTree(pgel);
                if (def) {
                    return def;
                }
            }
        }
        return null;
    }

    this.getAllActionTypes = function(pgel) {
        var r = [];
        for(var i = 0; i < this.frameworks.length; i++) {
            var f = this.frameworks[i];
            if(f.has_actions) {
                var a = f.getActionTypes(pgel);
                if(a.length) {
                    r = r.concat(a);
                }
            }
        }
        return r;
    }

    this.editCode = function() {
        $('body').crsa('editCode', this.$iframe);
    }

    this.exitEditCode = function() {
        if(pinegrow.codeEditor.isInEdit(this.$iframe)) {
            pinegrow.codeEditor.exitEdit(false);
        }
    }

    this.toggleEditCode = function() {
        if(pinegrow.codeEditor.isInEdit(this.$iframe)) {
            pinegrow.codeEditor.exitEdit(false);
        } else {
            this.editCode();
        }
    }

    this.getActionTag = function(pgel, for_search, short) {
        var a = [];
        for(var i = 0; i < this.frameworks.length; i++) {
            var f = this.frameworks[i];
            if(f.has_actions) {
                var s = f.getActionTag(pgel, for_search, short);
                if(s) a.push(s);
            }
        }
        return a.length ? a.join(', ') : null;
    }

    this.detectAndAddFrameworks = function(on_done, stage) {

        if(pgf.use_webview) {
            if(on_done) on_done();
            return;
        }

        var _this = this;

        var project_info = this.getProjectInfo();

        var info = project_info.getSettingsForFile( this.localFile );

        var added_types = [];

        var template_frameworks = [];

        if(this.crsaProjectTemplate) {
            template_frameworks = this.crsaProjectTemplate.frameworks;
        }

        if(stage <= 1) {
            this.frameworks = [];
        }

        var addFramework = function(f, index) {
            if (_this.canAddFramework(f)) {
                if(_this.frameworks.indexOf(f) < 0) {
                    if (index) {
                        _this.frameworks.splice(index, 0, f);
                    }
                    else {
                        _this.frameworks.unshift(f);
                    }
                    if(f.on_plugin_activated) {
                        f.on_plugin_activated(_this);
                    }
                    if(f.type) added_types.push(f.type);

                    if(!f.default) {
                        pinegrow.stats.using(f.key);
                    }
                    pinegrow.dispatchEventDelayed('on_page_frameworks_changed', _this);
                }
            }
        }

        if(template_frameworks.length > 0) {
            $.each(pinegrow.getFrameworks(), function(key, f) {
                if(template_frameworks.indexOf(key) >= 0) {
                    addFramework(f);
                }
            });
        }

        if((!info || !info.frameworks || info.frameworks.length == 0) && this.crsaProjectTemplate && this.crsaProjectTemplate.framework) {
            addFramework(this.crsaProjectTemplate.framework);
        }

        //add project framework
        var project = pinegrow.getCurrentProject();
        if(project && project.framework) {
            addFramework(project.framework);
        }

        var all_frameworks = Object.values(pinegrow.getFrameworks());
        var idx = -1;

        var doFramework = function() {
            idx++;
            if(idx < all_frameworks.length) {
                var f = all_frameworks[idx];
                var auto_detect = true;
                if(info && info.frameworks) {
                    var infoFrameworkIndex = info.frameworks.indexOf(f.key);
                    if(infoFrameworkIndex >= 0) {
                        addFramework(f, infoFrameworkIndex);
                    }
                    auto_detect = false;
                } else if(f.default) {
                    addFramework(f);
                    auto_detect = false;

                }
                if(!f.show_in_manager) auto_detect = true;

                if(auto_detect && f.detect && _this.sourceNode && stage === 1) {
                    if(f.detect(_this)) {
                        if(f.type && added_types.indexOf(f.type) >= 0) {
                            doFramework();
                            return true;
                        }
                        addFramework(f);
                        if(!f.silent_detection) {
                            crsaQuickMessage(f.name + " detected.", 2000);
                        }
                    }
                }
                if(auto_detect && f.detect_async && _this.sourceNode && stage === 2) {
                    f.detect_async(_this, function(add) {
                        if(add) {
                            addFramework(f);
                            if(!f.silent_detection) {
                                crsaQuickMessage(f.name + " detected.", 2000);
                            }
                        } else {
                            if (!f.show_in_manager && f.default) {
                                addFramework(f);
                            }
                        }
                        doFramework();
                    })
                } else {
                    if (!f.show_in_manager && f.default) {
                        addFramework(f);
                    }
                    doFramework();
                }
            } else {
                doneFirstPass();
            }
        }

        var doneFirstPass = function() {
            if(info && info.frameworks) {
                var missing = [];
                for(var i = 0; i < info.frameworks.length; i++) {
                    var found = false;
                    for(var j = 0; j < _this.frameworks.length; j++) {
                        if(_this.frameworks[j].key == info.frameworks[i]) {
                            found = true;
                            break;
                        }
                    }
                    if(!found) {
                        missing.push(info.frameworks[i]);
                    }
                }

                if(missing.length) {
                    //showAlert('<p>The following Pinegrow plugins that are used on this page were not found: <b>' + missing.join(', ') + '.</b></p><p>The most likely cause is that the plugin was upgraded to the new version in this release of Pinegrow.</p><p>Please use <b>Page -&gt; Manage frameworks</b> to activate appropriate plugins.</p>', 'Some plugins are missing');

                    var all = Object.values(pinegrow.getFrameworks());
                    var idx = -1;

                    var doMissing = function() {
                        idx++;
                        if(idx < all.length) {
                            var f = all[idx];
                            if(f.default) {
                                addFramework(f);
                                doMissing();
                            } else if(f.detect && _this.sourceNode) {
                                if(f.detect(_this)) {
                                    if(f.type && added_types.indexOf(f.type) >= 0) {
                                        doMissing();
                                        return true;
                                    }
                                    addFramework(f);
                                    if(!f.silent_detection) {
                                        crsaQuickMessage(f.name + " detected.", 2000);
                                    }
                                }
                                doMissing();
                            } else if(f.detect_async && _this.sourceNode) {
                                f.detect_async(_this, function(add) {
                                    if(add) {
                                        if(f.type && added_types.indexOf(f.type) >= 0) {
                                            doMissing();
                                            return true;
                                        }
                                        addFramework(f);
                                        if(!f.silent_detection) {
                                            crsaQuickMessage(f.name + " detected.", 2000);
                                        }
                                    }
                                    doMissing();
                                })
                            } else {
                                doMissing();
                            }
                        } else {
                            doneAll();
                        }
                    }

                    doMissing();

                } else {
                    doneAll();
                }
            } else {
                doneAll();
            }
        }

        var doneAll = function() {

            sortFrameworksByOrder();

            resetFrameworksCache();
            _this.frameworks_added = stage || 2;
            //$('body').trigger('crsa-frameworks-changed');

            pinegrow.dispatchEventDelayed('on_page_frameworks_changed', _this);

            //read breakpoints
            if (project_info) {

                var widthChanged = false;

                var breakpoints = project_info.getSetting('breakpoints');
                if (breakpoints && breakpoints.length) {
                    breakpoints = $.unique(breakpoints);
                    for (var i = 0; i < breakpoints.length; i++) {
                        if (String(breakpoints[i]).match(/^[0-9\.]*$/)) breakpoints[i] = String(breakpoints[i]) + 'px';
                    }
                    _this.breakpoints = breakpoints;
                    widthChanged = true;
                }

                if (widthChanged) $('body').trigger('crsa-breakpoints-changed');
            }

            if(on_done) on_done();
        }

        doFramework();
    }

    var sortFrameworksByOrder = function() {
        _this.frameworks = _this.frameworks.sort(function (a, b) {
            if (a.order > b.order) return 1;
            if (a.order < b.order) return -1;
            return 0;
        });
    }

    this.setResourceNamespaceForFramework = function(f, res_namespace) {
        //hardcode for now, not used
        var project_info = this.getProjectInfo();
        var resources = project_info.getSetting('resources') || {};
        resources[f.type] = res_namespace;
        project_info.setSetting('resources', resources);
        project_info.save();
    }

    this.getResourceNamespaceForFramework = function(f, detect) {
        if(!this.localFile) return '';

        //detect
        if(detect) {
            var folder = this.getProjectOrFileDir();
            var namespace = 'components/' + f.type + '/';

            if(folder) {
                var path = require('path').join(folder, namespace);
                try {
                    var stat = require('fs').statSync(path);
                    //exists
                    return namespace;
                }
                catch(err) {
                    //doesn't exist
                    return '';
                }
            }
        }

        var project_info = this.getProjectInfo();
        var template_id = project_info.getSetting('template_framework_id');

        if(template_id && template_id == f.type) return ''; //same project

        return 'components/' + f.type + '/';

        //hardcode for now
        var project_info = this.getProjectInfo();
        var resources = project_info.getSetting('resources') || {};
        return (f.type) in resources ? resources[f.type] : '';
    }

    this.findFrameworkOfType = function(type) {
        for(var i = 0; i < this.frameworks.length; i++) {
            if(this.frameworks[i].type === type) return this.frameworks[i];
        }
        return null;
    }

    this.hasFramework = function(f) {
        return this.frameworks.indexOf(f) >= 0;
    }

    this.canAddFramework = function(fm) {
        if(fm.allow_single_type && fm.type) {
            var efm = this.findFrameworkOfType(fm.type);
            if(efm) return false;
        }
        return true;
    }

    this.getFrameworks = function() {
        return this.frameworks;
    }

    this.addFramework = function(fm, skip_changed, init_done_func) {
        var changed = false;
        var idx = this.frameworks.indexOf(fm);
        if(idx < 0) {
            if(!this.canAddFramework(fm)) return false;
            this.frameworks.unshift(fm);
            sortFrameworksByOrder();
            changed = true;
            //this.callFrameworkHandler('on_plugin_activated');
            if(fm.on_plugin_activated) {
                fm.on_plugin_activated(this, init_done_func);
            } else {
                if(init_done_func) init_done_func(fm);
            }
            this.addCrsaStyles();
        }
        if(changed && !skip_changed) {
            pinegrow.frameworksChanged();
        }
        return true;
    }

    this.frameworksChanged = function() {
        resetFrameworksCache();
    }

    this.removeFramework = function(fm) {
        var idx = this.frameworks.indexOf(fm);
        if(idx >= 0) {
            this.frameworks.splice(idx, 1);
            resetFrameworksCache();

            if(fm.on_plugin_deactivated) {
                fm.on_plugin_deactivated(this);
            }
            $('body').trigger('crsa-frameworks-changed');
            this.addCrsaStyles();
        }
    }

    var arraysEqual = function(a, b) {
        var i = a.length;
        if (i != b.length) return false;
        while (i--) {
            if (a[i] !== b[i]) return false;
        }
        return true;
    }

    this.copyContentOfPage = function(crsaPage, skip_refresh) {

        var ret = this.applyChangesToSource(crsaPage.sourceNode, false, true, pinegrow.getSelectedPage() == this);

        if(ret.changes && !ret.updated && !skip_refresh) {
            this.refresh();
        }
    }

    var copyContentFromPage = function(crsaPage) {
        if(pageChangedTimer) {
            clearTimeout(pageChangedTimer);
        }
        pageChangedTimer = setTimeout(function() {
            _this.copyContentOfPage(crsaPage);

            if(!_this.readOnly) {
                $.fn.crsa('buildTree', _this.$iframe);
                _this.treeRepaintOnShow = true;
            }
            pageChangedTimer = null;

            if(!arraysEqual(_this.breakpoints, crsaPage.breakpoints)) {
                _this.setAllBreakpoints(crsaPage.breakpoints);
                $('body').trigger('crsa-breakpoints-changed');
            }

            _this.autoSize();
            _this.setPageChanged(true);
        }, 500);
    }

    this.onPageChanged = function(crsaPage) {
        //if(_this.live_update == crsaPage) {
            copyContentFromPage(crsaPage);
        //}
    }

    this.setSyntaxErrorIndicator = function(has_errors, fileNode, project) {
        if(has_errors) {
            //this.setFileNodeTag(new CrsaFileTag('syntax', 'danger', 'The page has HTML syntax errors. Right-click -&gt; Validate HTML for info.', 'code'), fileNode, project);
        } else {
            //this.removeFileNodeTag('syntax', fileNode, project);
        }
    }

    this.refreshPageChangedStatus = function(force) {
        var val = this.changed || this.hasCssChanges();
        this.setPageChanged(val, force, true);
    }

    this.setPageChanged = function(val, force, just_update_display) {
        var project = pinegrow.getCurrentProject();
        var is_app = !project || !project.loaded_from_backend;

        if(this.changed !== val || force) {

            if(!just_update_display) {
                this.changed = val;
            }

            if(val && !is_app) {
                if(this.onlyCode) {
                    if(project) {
                        project.settingsFileIsChanged(project.getRelativeUrl(this.url));
                    }
                } else {
                    pinegrow.backend.saveDelayed();
                }
            }

            if(this.view) {
                this.view.setModified(val);
            }

            if(this.codeEditor) {
                this.codeEditor.setModified(val);
            }

            if(is_app) {
                if (val) {
                    // TODO: @add-icon asterisk icon (asterisk)
                    this.setFileNodeTag(new CrsaFileTag('&#x2731;', 'info', 'Save changes.', null))
                } else {
                    this.removeFileNodeTag('&#x2731;');
                }
            }
        }
    }


    this.openInBrowser = function() {
        var project = pinegrow.getCurrentProject();

        if(project && project.loaded_from_backend) {

        } else {
            if (false && !this.localFile && this.url.startsWith('file://')) {
                pinegrow.showAlert('<p>After creating a new page, you have to <b>save the page</b> first, before using Preview in browser. After the page is saved on disk, you can use Preview in browser without saving the page.</p>', 'Page needs a home first');
            } else {
                var gui = require('nw.gui');
                var url = this.getPreviewUrl(true);

                gui.Shell.openExternal(url);

                pinegrow.showNotice('<p>The page was opened in your default browser. Click on <b>the link icon</b> next to Page -&gt; Preview in Browser to <b>copy the preview link to clipboard</b> from where you can use it to open the preview in any browser.</p><p>The preview link will work as long as the page is opened in Pinegrow.</p>', 'About the preview link', 'page-preview-link');
            }
        }
    }

    this.getProject = function() {
        var p = pinegrow.getCurrentProject();
        if(p && (!p.localFile || p.isPageInProject(this))) return p;
        return null;
    }

    this.getProjectOrFileDir = function() {
        var p = pinegrow.getCurrentProject();
        if(p && (!p.localFile || p.isPageInProject(this))) return p.getDir();
        if(!this.localFile) return null;
        return require('path').dirname(this.localFile);
    }

    this.getFileNode = function() {
        var project = pinegrow.getCurrentProject();
        if(project) {
            return project.getFileForUrl(this.url);
        }
    }

    this.setFileNodeTag = function(tag_obj, file, project, index) {
        if(!project) project = pinegrow.getCurrentProject();
        if(project) {
            if(!file) file = project.getFileForUrl(this.url);
            if(file) {
                file.addTag(tag_obj, index);
                project.updateTagsForFile(file);
            }
        }
    }

    this.removeFileNodeTag = function(tag, file, project) {
        if(!project) project = pinegrow.getCurrentProject();
        if(project) {
            if(!file) file = project.getFileForUrl(this.url);
            if(file) {
                file.removeTag(tag);
                project.updateTagsForFile(file);
            }
        }
    }

    this.getMirroredPages = function() {
        var r = [];
        for(var i = 0; i < crsaPages.length; i++) {
            if(crsaPages[i] != this && crsaPages[i].live_update == this) {
                r.push(crsaPages[i]);
            }
        }
        return r;
    }

    this.elementWasSelected = function($el) {
        //removed30
        debugger;
    }

    this.pageLoaded = function() {
        var pages = this.getMirroredPages();
        for(var i = 0; i < pages.length; i++) {
            var cp = pages[i];
            if(cp.url != this.url) {
                cp.rename(this.url);
                cp.reload();
            }
        }
    }

    this.hasCssChanges = function() {
        if(this.force_close) return false;
        if(!this.openedInEditor) return false;
        var ps = this.stylesheets;
        var has = false;
        if(ps) {
            $.each(ps.stylesheets, function(i, cs) {
                // if(cs.inline) return true;
                if(!cs.loaded) return true;
                var csChanged = has || cs.changed;
                if (!csChanged && cs.sourceStylesheet) {
                    cs.sourceStylesheet.forEachStylesheet(function (stylesheet) {
                        csChanged = stylesheet.changed;
                        if (csChanged) return false;
                    })
                }
                has = csChanged;
            });
        }
        return has;
    }

    this.hasChanges = function() {
        return this.changed || (!this.onlyCode && this.hasCssChanges());
    }

    this.getBreakpointsFromCss = function(done) {
        var ps = getCrsaPageStylesForPage(this.$iframe);
        var has = false;
        var r = [];

        var doDone = function() {
            var list = [];
            for(var i = 0; i < r.length; i++) {
                var cv = crsaSplitCSSValue(r[i]);
                var cvup = crsaOneUpCSSValue(cv);
                if(cv.value == 0 || r.indexOf(cvup.value + cvup.unit) >= 0) {
                    r[i] = -1;
                } else {
                    list.push(r[i]);
                }
            }
            /*
            for(var i = 0; i < r.length; i++) {
                if(list.indexOf(r[i]) < 0) {
                    list.push(r[i]);
                }
            }
            */
            list.sort(function (a, b) {
                var cva = crsaSplitCSSValue(a);
                var cvb = crsaSplitCSSValue(b);
                return cva.value - cvb.value;
            });

            list = pgMakeStringArrayUnique(list)

            done(list);
        }

        if(ps) {
            var c = 0;
            $.each(ps.stylesheets, function(i, cs) {
                //if(cs.inline) return true;
                c++;
                if(!cs.loaded) {
                    cs.setIgnore(false);
                    cs.load(cs.url, function() {
                        cs.getBreakpoints(function(list) {
                            r = $.unique( r.concat(list) );
                            cs.setIgnore(true);
                            c--;
                            if(c == 0) doDone();
                        })
                    });
                } else {
                    setTimeout(function() {
                        if (!cs.ignore || true) { //get from ignored css as well. for example, bootstrap.css is ignored by default
                            cs.getBreakpoints(function(list) {
                                r = $.unique( r.concat(list) );
                                c--;
                                if(c == 0) doDone();
                            })
                        } else {
                            c--;
                            if(c == 0) doDone();
                        }

                    }, 100);
                }
            });
            if(c == 0) doDone();
        } else {
            doDone();
        }
    }

    this.changeBreakpoint = function(old, b, done) {
        if(old == b) return;
        for(var i = 0; i < this.breakpoints.length; i++) {
            if(b == this.breakpoints[i]) {
                throw "Breakpoint " + b + " is already set.";
            }
        }
        for(var i = 0; i < this.breakpoints.length; i++) {
            if(old == this.breakpoints[i]) {
                this.breakpoints[i] = b;
            }
        }
        sortBreakpoints();

        this.setPageChanged(true);
        describeMediaQueryCache = {};

        var ps = getCrsaPageStylesForPage(this.$iframe);
        var has = false;
        var r = [];
        var changed = false;

        var doDone = function() {
            done(changed);
        }

        if(ps) {
            var c = 0;
            $.each(ps.stylesheets, function(i, cs) {
                //if(cs.inline) return true;
                c++;
                if (cs.compiledFile) cs = cs.sourceStylesheet;

                if(!cs.loaded) {
                    cs.ignore = false;
                    cs.load(cs.url, function() {
                        cs.changeBreakpoint(old, b, function(ch) {
                            c--;
                            changed = changed || ch;
                            _this.setPageChanged(true);
                            cs.ignore = !ch;
                            if(c == 0) doDone();
                        });
                    });
                } else {
                    setTimeout(function() {
                        cs.changeBreakpoint(old, b, function(ch) {
                            c--;
                            changed = changed || ch;
                            _this.setPageChanged(true);
                            if(c == 0) doDone();
                        });
                    }, 40);
                }

            });
            if(c == 0) doDone();
        } else {
            doDone();
        }
    }

    this.addBreakpoint = function(b) {
        //b = parseInt(b);
        for(var i = 0; i < this.breakpoints.length; i++) {
            if(b == this.breakpoints[i]) {
                throw "Breakpoint " + b + " is already set.";
            }
        }
        this.breakpoints.push(b);
        sortBreakpoints();
        this.setPageChanged(true);
        describeMediaQueryCache = {};
    }

    var sortBreakpoints = function() {
        _this.breakpoints.sort(function(a, b){
            var apx = parseFloat(crsaConvertCSSValueToPx(a, _this.$iframe));
            var bpx = parseFloat(crsaConvertCSSValueToPx(b, _this.$iframe));
            return apx-bpx;
        });
    }

    this.removeBreakpoint = function(b) {
        var r = [];
        for(var i = 0; i < this.breakpoints.length; i++) {
            if(b != this.breakpoints[i]) {
                r.push(this.breakpoints[i]);
            }
        }
        this.breakpoints = r;
        this.setPageChanged(true);
        describeMediaQueryCache = {};
    }


    this.setAllBreakpoints = function(list) {
        this.breakpoints = pgMakeStringArrayUnique(list);
        this.breakpoints.sort(function(a, b){return a-b});
        //this.setPageChanged(true);
        describeMediaQueryCache = {};
    }

    this.showManageBreakpoints = function() {
        var project = pinegrow.getCurrentProject();

        var save = function() {
            if(project && project.loaded_from_backend) {
                var pi = pinegrow.getCurrentProjectInfo();
                pi.setSetting('breakpoints', _this.breakpoints);
                pi.save();
            }
            pinegrow.selectedElements.reselect(); //trigger update in style panel
        }

        var showEditBox = function($el, old) {
            var eid = getUniqueIdFast();
            if($el.data('popover-active')) {
                $el.popover('destroy');
                $el.data('popover-active', null);
                return;
            }

            var ensureRemove = function($e) {
                setTimeout(function() {
                    $e.closest('.popover').remove();
                }, 750);
            }

            var pop = $el.popover({
                html: true,
                placement: 'right',
                trigger: 'manual',
                title: old ? 'Edit breakpoint' : 'Add new breakpoint',
                container: 'body',
                content: '<form id="' + eid + '"><form><input class="form-control" placeholder="width with unit" style="margin-bottom:8px;"/><button class="ok btn">' + (old ? "Set" : "Add") + '</button><button class="useit btn btn-link">Use current width</button><button class="closeit btn btn-link">Cancel</button></form>'
            })
                .on('shown.bs.popover', function() {
                    var $d = $('#' + eid);
                    var $i = $d.find('input').focus();
                    var $b = $d.find('button.ok');
                    var $bc = $d.find('button.closeit');
                    var $bu = $d.find('button.useit');

                    if(old) {
                        $i.val(old);
                    }

                    var doAdd = function() {
                        var val = $.trim($i.val());
                        if(val.length > 0) {
                            //val = val.replace(/[^0-9]/g, '');

                            var cv = crsaSplitCSSValue(val);
                            val = cv.value + cv.unit;

                            try {
                                if(!old) {
                                    _this.addBreakpoint(val);
                                    save();
                                } else {
                                    if($el) {
                                        $el.html('Please wait...').addClass('disabled');
                                    }
                                    //_this.undoStack.add("Change breakpoint");
                                    _this.changeBreakpoint(old, val, function(changed) {
                                        $el.html('Edit').removeClass('disabled');
                                        if(changed) {
                                            setTimeout(function() {
                                                crsaQuickMessage("Stylesheets were updated!", 3000);
                                            }, 250);
                                            setTimeout(function() {
                                                $('body').trigger('crsa-stylesheets-changed');
                                            }, 50);
                                        }
                                        save();
                                    });

                                }
                                //didMakeChange(_this.$iframe);todo30
                                $el.popover('hide');
                                ensureRemove($d);
                                updateList();
                                $('body').trigger('crsa-breakpoints-changed');
                            }
                            catch(err) {
                                pinegrow.alert(err);
                            }
                        }
                    }

                    $b.on('click', function(e) {
                        e.preventDefault();
                        doAdd();
                    });

                    $d.on('submit', function(e) {
                        e.preventDefault();
                        doAdd();
                    });

                    $bu.on('click', function (e) {
                        e.preventDefault();
                        $i.val(_this.activeView.deviceWidth + 'px');
                        //doAdd();
                    })

                    $bc.on('click', function(e) {
                        $el.popover('hide');
                        e.preventDefault();
                        ensureRemove($d);
                    });
                })
                .on('hidden.bs.popover', function() {
                    setTimeout(function() {
                        $el.popover('destroy').data('popover-active', null);
                    },10);
                });
            $el.popover('show').data('popover-active', true);
        }

        var $container_div = $('<div/>');

        var updateList = function() {
            $container_div.html('');
            var $ul = $('<ul/>', {style: 'margin-bottom: 0px'}).appendTo($container_div);
            var $custom_ul = $('<ul/>', {style: 'margin-bottom: 0px'}).appendTo($container_div);

            var addNewWidth = function (width, $container) {
                var $button = $('<a href="#" class="btn btn-sm btn-link">Edit</a>');
                var $del = $('<a href="#" class="btn btn-sm btn-link">Delete</a>');
                $('<li></li>').html(width).appendTo($container).append($button).append($del).data('breakpoint', width);
                $button.on('click', function(e) {
                    var $li = $(e.delegateTarget).closest('li');
                    e.preventDefault();
                    showEditBox($(e.delegateTarget), $li.data('breakpoint'));
                });
                $del.on('click', function(e) {
                    var $li = $(e.delegateTarget).closest('li');
                    e.preventDefault();
                    if(confirm('Are you sure? Existing CSS rules will not be affected.')) {
                        _this.removeBreakpoint($li.data('breakpoint'));
                        updateList();
                        save();
                        $('body').trigger('crsa-breakpoints-changed');
                        //didMakeChange(_this.$iframe);//todo30
                    }
                });
            }

            for(var i = 0; i < _this.breakpoints.length; i++) {
                addNewWidth(_this.breakpoints[i], $ul);
            }

        }

        updateList();

        var $b = $container_div;

        var $main = $('<div class="manager"><p>Breakpoints:</p></div>').append($b);

        var msg = '';

        $('<a href="#" class="btn btn-link add">Add breakpoint...</a>').appendTo($main)
            .on('click', function(e) {
                e.preventDefault();
                showEditBox($(this));
            });

        $('<a href="#" class="btn btn-link add">Get from stylesheets</a>').appendTo($main)
            .on('click', function(e) {
                e.preventDefault();
                var $b = $(e.delegateTarget);
                $b.html('Please wait...').addClass('disabled');

                _this.getBreakpointsFromCss(function(list) {
                    if(list.length) {

                        var dofunc = function() {
                            _this.setAllBreakpoints(list);
                            //didMakeChange(_this.$iframe);todo30
                            crsaQuickMessage("Done!", 2000);
                            save();
                        }

                        if(_this.breakpoints.length) {
                            var msg = 'Breakpoints [' + list.join(', ') + '] were found in CSS. Do you want to use them instead of current breakpoints? This can\'t be undone.';
                            if(confirm(msg)) {
                                dofunc();
                            }
                        } else {
                            dofunc();
                        }

                    } else {
                        crsaQuickMessage("No breakpoints found!", 2000);
                    }
                    updateList();
                    $('body').trigger('crsa-breakpoints-changed');
                    $b.html('Get from stylesheets').removeClass('disabled');
                });
            });

        //$('<a href="http://docs.pinegrow.com/styling/creating-responsive-layouts" target="_blank" class="btn btn-link add external">Help</a>').appendTo($main);

        crsaHandleExternalLinks($main);

        //todo30 - updateMetaAdd
        var updateMetaAdd = function() {
            var $ma = $main.find('.meta-add');
            if(!$ma.length) {
                $ma = $('<div class="meta-add well" style="margin-top:12px;"></div>').appendTo($main);
            }
            var pghead = _this.sourceNode.findOne('head');
            if(pghead) {
                var pgm = pghead.findOne('meta[name="viewport"]');
                if (pgm) {
                    $ma.hide();
                } else {
                    $ma.html('<p>Document needs <code>&lt;meta name=&quot;viewport&quot; ... &gt;</code> directive to enable responsive behaviour on mobile devices. <a href="#" class="">Add it!</a></p>');
                    var $add = $ma.find('a');

                    $ma.show();
                    $add.on('click', function (e) {
                        e.preventDefault();
                        //_this.undoStack.add("Meta viewport directive added");
                        var pgm = pgCreate('<meta name="viewport" content="width=device-width, initial-scale=1.0">');
                        pghead.append(pgm)
                        _this.setPageChanged(true);
                        pinegrow.changeManager.didChange(pgm, 'Add viewport declaration');
                        updateMetaAdd();
                        crsaQuickMessage("Viewport declaration added!", 2000);
                    });
                }
            }
        }

        updateMetaAdd();


        var $d = makeModalDialog("Manage breakpoints", "Close", null, $main);
        $d.removeAttr('tabindex');
        $d.find('.modal-dialog').css('width', '500px');
    }

    var describeMediaQueryCache = {};

    this.describeMediaQuery = function(q) {
        var rule = null;

        if(q in describeMediaQueryCache) {
            return describeMediaQueryCache[q];
        }

        var idx = q.indexOf('(');
        if(idx >= 0) {
            rule = q.substr(idx);
            if(q.indexOf('only') >= 0) {
                rule = rule.replace('only ','');
            }
            $.each(['screen', 'print'], function(i,t) {
                if(q.indexOf(t) >= 0) {
                    type = t;
                    rule = rule.replace(t,'');
                }
            });
        } else {
            rule = q;
        }
        rule = rule.replace(/\s/g,'');

        var res = null;
        if(_this.breakpoints.length) {
            for(var i = 0; i < _this.breakpoints.length; i++) {
                var cv = crsaSplitCSSValue(_this.breakpoints[i]);
                var x = cv.value;
                var cvdown = crsaOneUpCSSValue(cv, -1); //one down

                var mq = '(max-width:' + (cvdown.value) + cv.unit + ')';
                if(rule == mq) {
                    res = '<span class="line"></span><span class="excluding">' + x + '</span>';
                    break;
                }
                mq = '(min-width:' + (x) + cv.unit + ')';
                if(rule == mq) {
                    res = '<span class="including">' + x + '</span><span class="line"></span>';
                    break;
                }
                if(i < _this.breakpoints.length -1) {

                    var ncv = crsaSplitCSSValue(_this.breakpoints[i+1]);
                    var nx = ncv.value;
                    var ncvdown = crsaOneUpCSSValue(ncv, -1); //one down

                    mq = '(min-width:' + (x) + 'px)and(max-width:' + (ncvdown.value) + cv.unit + ')';
                    if(rule == mq) {
                        res = '<span class="including">' + x + '</span><span class="line"></span><span class="excluding">' + (nx) + '</span>';
                        break;
                    }
                }
            }
        }
        describeMediaQueryCache[q] = res ? res : q;
        return res ? res : q;
    }

    this.showMediaQueryHelper = function($el) {
        var offset = $el.offset();
        var place = offset.left > $(window).width()/2 ? 'left' : 'right';
        //place = 'auto';
        var eid = getUniqueIdFast();
        if($el.data('tool-active')) {
            $el.data('tool-active', null);

            return;
        }
        var ensureRemove = function($e) {
            $dialog.hide();
            $el.data('tool-active', null);
            tool.closed = true;
            setTimeout(function() {
                $dialog.remove();
            }, 500);
        }

        var original = $.trim($el.val());
        var original_just_rule = '';
        var has_only = localStorage.mediaToolOnly ? true : false;
        var type = localStorage.mediaToolType ? localStorage.mediaToolType : null;
        var strange = true;

        if(original) {
            var idx = original.indexOf('(');
            if(idx >= 0) {
                original_just_rule = original.substr(idx);
                if(original.indexOf('only') >= 0) {
                    has_only = true;
                    original_just_rule = original_just_rule.replace('only ','');
                }
                $.each(['screen', 'print'], function(i,t) {
                    if(original.indexOf(t) >= 0) {
                        type = t;
                        original_just_rule = original_just_rule.replace(t,'');
                    }
                });
                original_just_rule = original_just_rule.replace(/\s\s+/g,' ');
                strange = false;
            } else {
                original_just_rule = original;
            }
        } else {
            strange = false;
        }

        var done = false;

        var s = '';
        if(_this.breakpoints.length) {
            for(var i = 0; i < _this.breakpoints.length; i++) {

                var cv = crsaSplitCSSValue(_this.breakpoints[i]);
                var x = cv.value;
                var cvdown = crsaOneUpCSSValue(cv, -1); //one down

                var mq = '(max-width:' + (cvdown.value) + cv.unit + ')';
                s += '<li class="media-query" data-media="' + mq + '"><i class="fa fa-check"></i><span class="line"></span><span class="excluding">' + x + '</span><small>smaller than ' + x + ' ' + cv.unit + '</small></li>';
                mq = '(min-width:' + (x) + cv.unit + ')';
                s += '<li class="media-query" data-media="' + mq + '"><i class="fa fa-check"></i><span class="including">' + x + '</span><span class="line"></span><small>' + x + ' ' + cv.unit + ' and up</small></li>';
                if(i < _this.breakpoints.length -1) {

                    var ncv = crsaSplitCSSValue(_this.breakpoints[i+1]);
                    var nx = ncv.value;
                    var ncvdown = crsaOneUpCSSValue(ncv, -1); //one down

                    mq = '(min-width:' + (x) + cv.unit + ') and (max-width:' + (ncvdown.value) + cv.unit + ')';
                    s += '<li class="media-query submedia-last" data-media="' + mq + '"><i class="fa fa-check"></i><span class="including">' + x + '</span><span class="line"></span><span class="excluding">' + (nx) + '</span><small>from ' + x + ' ' + cv.unit + ' to ' + (ncv.value) + ' ' + cv.unit + '</small></li>';
                }
            }
        } else {
            s += '<li>No breakpoints are defined. Use &quot;Manage breakpoints&quot; to define them.</li>';
        }

        var triggerTimeout = null;

        var triggerChange = function($el) {
            if(triggerTimeout) {
                clearTimeout(triggerTimeout);
            }
            triggerTimeout = setTimeout(function() {
                $el.focus();
                $el.trigger('change');
                triggerTimeout = null;
            }, 50);
        }

        var form = '<form role="form" class="form-inline">\
            <div class="form-group" style="margin-right:10px;">\
            <label class="control-label sr-only" for="formInput1">Field label</label>\
            <select id="formInput1" class="form-control input-sm">\
            <option></option>\
            <option>screen</option>\
            <option>print</option>\
        </select>\
        </div>\
            <div class="checkbox">\
                <label class="control-label">\
                    <input type="checkbox"> Ignore in old browsers</label>\
                </div>\
            </form>';

        var $c = $('<div id="' + eid + '" class="media-tool"><ul class="media-list">' + s + '</ul>' + form + '</div>');

        var $body = $('body');
        var $dialog = makeDialog("Select media query", "Cancel", "Ok", $c).css('width','400px');
        $body.append($dialog);

        $dialog.find('.modal-footer').prepend('<a href="#" class="pull-left manage btn btn-link">Manage breakpoints...</a>').css('margin-top', '0px');

        $dialog.find('.modal-body').css('padding-bottom', '0px');

        $dialog.on('keydown', function(e) {
            if(e.which == 27) {
                $dialog.find('button.cancel').trigger('click');
                e.preventDefault();
            }
        });

        var elo = $el.offset();
        var bw = $body.width();
        var bh = $body.height();

        var x,y;

        if(elo.left < bw / 2) {
            x = elo.left + $el.width() + 35;
        } else {
            x = elo.left - $dialog.width() - 10;
        }
        y = elo.top - $dialog.height()/2;

        x = localStorage.mediaToolX ? parseInt(localStorage.mediaToolX) : x;
        y = localStorage.mediaToolY ? parseInt(localStorage.mediaToolY) : y;

        if(y < 10) y = 10;
        if(y + $dialog.height() > bh) {
            y = bh - 10 - $dialog.height();
        }
        if(x < 10) x = 10;
        if(x + $dialog.width() > bw) {
            x = bw - 10 - $dialog.width();
        }


        $dialog.css('top', y + 'px').css('left', x + 'px');
        $dialog.draggable({handle: '.modal-header'})
            .on('dragstart', function() {
                $.fn.crsapages('showOverlays');
            })
            .on('dragstop', function() {
                $.fn.crsapages('showOverlays', true);
                var ofs = $dialog.offset();
                localStorage.mediaToolX = ofs.left;
                localStorage.mediaToolY = ofs.top;
            });

        var $d = $('#' + eid);
        var $ul = $d.find('ul');
        var $b = $dialog.find('button.ok');
        var $bc = $dialog.find('button.close,button.cancel');

        var $select = $d.find('select');
        var $only = $d.find('input[type="checkbox"]');

        if(strange) {
            $select.attr('disabled', 'disabled');
            $only.attr('disabled', 'disabled');
        } else {
            $select.val(type);
            if(has_only) {
                $only.attr('checked', 'checked');
            }
            $select.on('change', function(e) {
                apply(original_just_rule);
                localStorage.mediaToolType = $select.val();
            });
            $only.on('change', function(e) {
                apply(original_just_rule);
                localStorage.mediaToolOnly = $only.is(':checked') ? '1' : '0';
            });
        }

        var apply = function(q) {
            var s = $select.val();
            var has_and = false;
            if(s) {
                if(q.length) {
                    q = s + ' and ' + q;
                    has_and = true;
                } else {
                    q = s;
                }
            }
            if($only.is(':checked')) {
                q = 'only ' + (!s ? 'screen and ' : '') + (!has_and ? '' : '') + q;
            }
            $el.val(q);
            triggerChange($el);
        }

        var current = original_just_rule.replace(/\s/g, '');

        $ul.find('li').each(function(i,e) {
            var $li = $(e);
            var q = $li.attr('data-media');
            if(q) {
                if(q.replace(/\s/g,'') == current) {
                    $li.addClass('media-current');
                }
                var win = _this.getWindow();
                if(win && win.matchMedia(q).matches) {
                    $li.addClass('media-match');
                }
            }
        });

        $ul.find('i.fa-check:visible').tooltip({container: 'body', title: 'Matches current window size (' + _this.deviceWidth + 'px).'});

        $ul.find('li')
            .on('mouseover', function(e) {
                if(!done) {
                    var $e = $(e.delegateTarget);
                    var q = $e.attr('data-media');
                    if(q) {
                        apply(q);
                    }
                }
            })
            .on('mouseout', function(e) {
                if(!done) {
                    var q = $(e.delegateTarget).attr('data-media');
                    if(q) {
                        if(strange) {
                            $el.val(original);
                            triggerChange($el);
                        } else {
                            apply(original_just_rule);
                        }
                    }
                }
            })
            .on('click', function(e) {
                done = true;
                e.preventDefault();
                var $e = $(e.delegateTarget);
                var q = $e.attr('data-media');
                if(q) {
                    $ul.find('li.media-current').removeClass('media-current');
                    $e.addClass('media-current');
                    apply(q);
                    ensureRemove($d);
                }
            });

        $b.on('click', function(e) {
            e.preventDefault();
            done = true;
            ensureRemove($d);
        });

        $bc.on('click', function(e) {
            done = true;
            $el.val(original);
            triggerChange($el);
            e.preventDefault();
            ensureRemove($d);
        });

        $dialog.find('.manage').on('click', function(e) {
            done = true;
            $el.val(original);
            triggerChange($el);
            e.preventDefault();
            ensureRemove($d);
            _this.showManageBreakpoints();
        });

        $el.data('tool-active', true);

        var tool = {
            closed: false,
            close: function() {
                ensureRemove();
            }
        }
        return tool;
    }

    this.getProjectInfo = function() {
        return new pgProjectInfo(_this.localFile);
    }

    this.getDocument = function() {
        return this.$iframe.get(0).contentDocument;
    }

    this.getWindow = function() {
        return this.$iframe.get(0).contentWindow;
    }

    this.getBody = function() {
        return getIframeBody(this.$iframe.get(0));
    }

    this.get$Html = function() {
        var doc = this.getDocument();
        return doc.documentElement ? $(doc.documentElement) : null;
    }

    this.getElementWithPgId = function(pgid) {
        var $el = this.get$Html().find('[data-pg-id="' + pgid + '"]');
        return $el.length ? $el : null;
    }

    this.getCSSVariableValue = function(v) {
        var $body = this.activeView.get$Body();
        var val = this.activeView.getWindow().getComputedStyle($body.get(0)).getPropertyValue(v);
        if(typeof val === 'string') val = val.trim();
        return val;
    }

    this.getElementFromPgParserNode = function(pgel) {
        return this.getElementWithPgId(pgel.getId());
    }

    this.getData = function($el, key) {
        try {
            var id = getUniqueIdFast('crsa');
            $el.attr('data-tmp', id);
            var windowjQuery = this.$iframe[0].contentWindow.$;
            var f = this.$iframe.contents().find('[data-tmp="' + id + '"]');
            $el.removeAttr('data-tmp');
            return f.length > 0 ? windowjQuery.data(f.get(0), key) : null;
        } catch(err) {}
        return null;
    }

    this.saveHTML = function() {
        this.save(function() {}, true, false, false, false);
    }

    this.saveAs = function(done) {
        if(pinegrow.codeEditor) pinegrow.codeEditor.refreshBeforeSaveIfNeeded();
        this.save(function(err) {
            if(!err) {
                pinegrow.recentFiles.add(_this.url);
                pinegrow.pageTabs.updateDisplay();
                if(done) done(_this);
            }
        }, true, true, true);
    }


    this.save = function(done, save_html, save_css, save_as, save_project, first_save, ask_css_overwrite) {
        var project = pinegrow.getCurrentProject();

        if(!project || !project.loaded_from_backend) {

            if(!pgf.save_css) {
                save_css = false;
            }

            if(!this.openedInEditor || this.onlyCode) {
                save_css = false;
                save_project = false;
                save_html = true;

                //check if it is css, then save cs instead
                var cs = pinegrow.stylesheets.getByUrl(this.url, true);
                if(cs) {
                    //check if editing a stylesheet of a page that is not saved yet, in order to prevent changing the template file
                    var pages = pinegrow.stylesheets.getPagesForStylesheet(cs);
                    for (var i = 0; i < pages.length; i++) {
                       if(pages[i].isNew()) {
                           //the page is not saved yet
                           pinegrow.showQuickError(`Please save the page ${pages[i].getName()} before saving the stylesheet.`);
                           //done && done(false);
                           return;
                       }
                    }

                    cs.save(cs.localFile, function() {
                        pinegrow.refreshPageChangedStatusForAllPages(true);

                        //Will check active frameworks for all open pages, needed for wp
                        pinegrow.dispatchEvent('on_css_file_edited_in_code_editor_saved', pinegrow.getAllPages(), cs);

                        if(done) done();
                    });
                    return;
                }
            }

            if(typeof save_project == 'undefined') save_project = true;

            var html;

            if(this.onlyCode) {
                html = this.code;
            } else {
                html = this.getCodeForSave();
            }

            var fs = require('fs');
            var pm = require('path');
            var project = pinegrow.getCurrentProject();

            pinegrow.dispatchEvent('on_before_save', this);

            if(this.localFile && !save_as) {
                try {
                    if(this.live_update && this.save_parent) {
                        if(done) done();
                        showAlert("The page was not saved because it is a mirrored page. You should save the parent page " + this.live_update.name + " instead.", "Page was not saved");
                        return;
                    }

                    var dir = crsaGetFileDir(this.localFile);

                    //create only project dirs
                    if (!project || (project && !project.isPageInProject(this))) {
                        if(this.crsaProjectTemplate && save_project) {
                            this.crsaProjectTemplate.copyRequiredFilesTo(dir, null, null, true);
                        }
                    }

                    var notice = [];

                    var saveDone = function(err, saved_csss) {
                        if(pgf.save_html) {
                            if (!project || (project && !project.isPageInProject(_this))) {
                                if (_this.crsaProjectTemplate && save_project) {
                                    _this.crsaProjectTemplate.copyRequiredFilesTo(dir);
                                }
                            }

                            if(_this.pageIsRemoteTemplate) {
                                var reslist = new PgResourcesList();
                                var page_changed = reslist.detectRequiredResources(_this, _this.source_framework);
                                if(page_changed) {
                                    //pg-require tags were removed from source, lets save again
                                    html = _this.getCode(pinegrow.getSetting('keep-collapsed', '0') == '1', true /* for save */);
                                    crsaWriteFileWithBackup(fs, _this.localFile, html, "utf8");
                                }
                            }

                            var url = crsaMakeUrlFromFile(_this.localFile);
                            if (url != _this.url && (crsaIsFileUrl(_this.url) || _this.pageIsRemoteTemplate)) _this.rename(url);

                            if(_this.pageIsRemoteTemplate) {
                                pinegrow.addFrameworkResourcesToPageOrProject(_this.source_framework, _this, {
                                    resources: reslist,
                                    only_add_to_page: _this,
                                    overwrite: function(file) {
                                        if(file === 'pinegrow.json' && project_info && project_info.was_just_created) {
                                            //overwrite pinegrow.json if auto-generated one was just saved right now
                                            return true;
                                        }
                                        return false;
                                    },
                                    done: function() {
                                        var updateCS = function() {
                                            pinegrow.stylesheets.removeAllUnusedStylesheets();
                                            pinegrow.removeEventHandler('on_page_stylesheets_loaded', updateCS);
                                        }

                                        pinegrow.addEventHandler('on_page_stylesheets_loaded', updateCS)

                                        _this.refresh(function() {
                                            //remove unused stylesheets

                                        });
                                    }
                                })

                                isFile = true;
                                _this.pageIsRemoteTemplate = false;
                            }

                        }

                        if(!err) {
                            if(!_this.onlyCode) {
                                _this.callFrameworkHandler('on_page_saved', saved_csss, first_save);

                                _this.view && _this.view.rememberPageViews();
                            } else {
                                _this.callFrameworkHandler('on_code_page_saved');
                            }
                        }

                        if(done) done(err);

                        if(save_html) {
                            _this.changed = false;
                            _this.refreshPageChangedStatus(true);
                            //_this.setPageChanged(false, true);
                        }

                        //_this.refreshPageChangedStatus(true);
                        pinegrow.refreshPageChangedStatusForAllPages(true);

                        var getError = function(info) {
                            var e = 'Unknown error';
                            switch(info.error) {
                                case 'syntax':
                                    e = 'Syntax error (check CSS panel for more)';
                                    break;
                                case 'url':
                                    e = 'Invalid or remote url';
                                    break;
                                case 'file':
                                    e = 'Error writting to file: ' + info.msg;
                                    break;
                                default:
                                    e = info.error;
                                    break;
                            }
                            return e;
                        }

                        var b = '';
                        if((!isFile && first_save)) {

                            b += isFile ? '</ul>' : '</ul><p>Images, scripts and other resources were not saved.</p><p>If you want to save all changes it is recommended to first save the whole webpage to your computer (Save As Complete page in your browser) and then open it in Pinegrow.</p>';
                        }
                        if(notice.length > 0) {
                            b += isFile ? "<p>The following stylesheets were not saved:</p><ul>" : "<p>The following stylesheets were not saved because they have syntax errors or their paths would place them outside of the folder selected for saving the HTML file:</p><ul>";
                            for(var i = 0; i < notice.length; i++) {
                                b += '<li>' + notice[i].url + ' - ' + getError(notice[i]) + '</li>';
                            }
                        }

                        if(first_save || _this.mappedRemoteFirstSave) {
                            pinegrow.refreshCurrentProjectDelayed(true);
                            _this.mappedRemoteFirstSave = !html_was_saved;
                        }

                        if(b.length > 0) {
                            showAlert(b, 'Some files were not saved');
                        } else if(html_was_saved) {
                            crsaQuickMessage((save_css || _this.onlyCode || !pgf.edit_css) ? _this.name + " was saved." : _this.name + " (HTML only) was saved.")
                        }
                    }

                    var isFile = crsaIsFileUrl(this.url);

                    var html_was_saved = false;

                    if(save_html) {

                        if(pgf.save_html) {
                            if(!this.mappedRemote || this.changed || this.pageIsRemoteTemplate) {
                                this.lastSavedAt = (new Date()).getTime();
                                crsaWriteFileWithBackup(fs, this.localFile, html, "utf8");
                                html_was_saved = true;
                            }

                            var project = pinegrow.getCurrentProject();
                            if (project) {
                                if (!this.openedInEditor && !this.codeEditor) {
                                    project.removeBackgroundPageForUrl(this.url);
                                }
                            }

                            if (save_project) {

                                var file_or_project_dir = _this.localFile;
                                var project = pinegrow.getCurrentProject();
                                if (project && project.isFileInProject(_this.localFile)) {
                                    file_or_project_dir = project.getDir();
                                }

                                var project_info = new pgProjectInfo(file_or_project_dir, first_save /* don't search for parent project info if this is the new file */);

                                var pageinfo = {frameworks: []};
                                for (var i = 0; i < _this.frameworks.length; i++) {
                                    pageinfo.frameworks.push(_this.frameworks[i].key);
                                }
                                project_info.setSettingForFile(_this.localFile, 'frameworks', pageinfo.frameworks);
                                project_info.setSetting('breakpoints', _this.breakpoints);
                                //project_info.setSetting('custom_width', _this.custom_width);
                                //project_info.setSettingForFile(_this.localFile, 'last_page_width', _this.$iframe.width());

                                project_info.setSetting('frameworks', pageinfo.frameworks);

                                //moved this here, we're saving page info already
                                if (_this.source_framework) {
                                    project_info.setSetting('template_framework_id', _this.source_framework.type)
                                }

                                project_info.save();
                            }
                        }
                    }


                    //debugger;
                    if(save_css) {
                        var saved_csss = [];
                        var cp = getCrsaPageStylesForPage(this.$iframe);
                        var count = 0;
                        var fileWriter = null;

                        if(ask_css_overwrite) {
                            fileWriter = new CrsaOverwriteProtectionFileWriter('dummy_source', null);
                            fileWriter.use_export_cache = false;
                            if(first_save) {
                                fileWriter.skip_existing_files = true; //don't ask to overwrite existing files. just skip them.
                            }
                        }

                        var getRelativePathForCs = function (cs) {
                            var nakedUrl = cs.url;
                            var rel_url = crsaMakeLinkRelativeTo(nakedUrl, crsaMakeUrlAbsolute(_this.url));
                            return crsaRemoveUrlParameters(rel_url);
                        }

                        var getPathForCs = function (cs, rel_url) {
                            var path;
                            rel_url = rel_url || getRelativePathForCs(cs);

                            if (rel_url.indexOf('://') >= 0) {
                                path = crsaMakeFileFromUrl(rel_url);
                            } else {
                                //hmm... why not use localFile instead?
                                path = require('path').resolve(dir, rel_url);
                            }

                            return path;
                        }

                        $.each(cp.getAllCrsaStylesheets(), function(i, cs) {
                            var mainCs = null;

                            if (cs.sourceStylesheet) {
                                mainCs = cs;
                                cs = cs.sourceStylesheet;
                            }

                            if(cs.inline) {
                                cs.fileSaved();
                                return true;
                            }
                            if(!cs.loaded) return true;
                            if(cs.ignoreInSave) return true;

                            if(cs.loadingFailed) return true;

                            var rel_url = getRelativePathForCs(cs);
                            var path = getPathForCs(cs, rel_url);

                            var cs_error = null;

                            if(cs_error = cs.getError()) {
                                notice.push({url: cs.url, error: cs_error || 'syntax'});
                            } else {
                                var skip = false;

                                if((rel_url.indexOf('://') >= 0 && !crsaIsFileUrl(rel_url)) || (!isFile && rel_url.indexOf('..') >= 0)) {
                                    //Is this mapped remote url?
                                    if(cs.localFile) {
                                        path = cs.localFile;
                                    } else {
                                        if (cs.changed) {
                                            notice.push({url: cs.url, error: 'url'});
                                        }
                                        skip = true;
                                    }
                                }

                                if(!skip) {
                                    if (!cs.needsSave(path)) {
                                        if (!mainCs || !mainCs.needsSave()) return true;

                                        cs = mainCs;
                                        path = getPathForCs(cs);
                                    }

                                    var fileSaved = function () {
                                        if (count == 0) {
                                            if(fileWriter) {
                                                fileWriter.askIfNeeded(function() {
                                                    saveDone(null, saved_csss);
                                                })
                                            } else {
                                                saveDone(null, saved_csss);
                                            }
                                        }
                                    }

                                    count++;
                                    cs.save(path, function(err) {
                                        if(err) {
                                            //debugger;
                                            //throw(err);
                                            notice.push({url: cs.url, error: 'file', msg: err});
                                        } else {
                                            if(mainCs) {
                                            }
                                        }
                                        count--;
                                        saved_csss.push({cs: cs, path: path, error: err});
                                        fileSaved();
                                    }, fs, null, fileWriter);
                                }
                            }
                        });

                        if (count == 0) {
                            if(fileWriter) {
                                fileWriter.askIfNeeded(function() {
                                    saveDone();
                                })
                            } else {
                                saveDone();
                            }
                        }
                    } else {
                        saveDone();
                    }
                }
                catch(err) {
                    if(done) done(err);
                    var hint = '<p>A HINT: By default, Save dialog will select the root folder of your computer. Files can\'t be written there of course. Select the folder where you have write access and it will be ok.</p>';
                    showAlert("<p>Could not save file: " + err.message + '</p>' + hint, "Error");
                }
            } else {
                crsaChooseFile(function(url, file) {

                    var dofunc = function() {
                        var origLocalFile = _this.localFile;
                        var origSaveParent = _this.save_parent;
                        _this.setLocalFile(file);
                        _this.save_parent = false;
                        _this.save(function(err) {

                            if(!err) {
                                if(crsaIsFileUrl(_this.url)) {
                                    _this.rename(url);
                                    $.fn.crsapages('updateFilesName');
                                    _this.reload(function() {
                                        //workaround for situation when new page is saved in existing project and csss still point to the resource versions of csss
                                        _this.updateStylesheetsList();
                                    });
                                }
                            } else {
                                _this.setLocalFile(_this.localFile);
                                _this.save_parent = origSaveParent;
                            }
                            $body.trigger('crsa-page-saved-as', _this);
                            done();
                        }, save_html, save_css, false, true, true, true /* ask css overwrite */);
                    }

                    if(url.indexOf('.htm') < 0) {

                        showAlert("Do you want to use the file without the <b>.html</b> extension? Some browsers will not recognize it.", "No .html extension", "Use as is", "Add .html extension", function() {
                            dofunc();
                        }, function() {
                            file = file + '.html';
                            url = url + '.html';
                            dofunc();
                        });
                    } else {
                        dofunc();
                    }

                }, this.name);
            }
        } else {
            // Not used?
            //browser save
            var store = new CrsaBrowserStorage();
            if(save_html) {
                store.save(this.url, html);
                this.setPageChanged(false);
            }

            if(save_css) {
                var cp = getCrsaPageStylesForPage(this.$iframe);
                var count = 0;
                $.each(cp.getAllCrsaStylesheets(), function(i, cs) {
                    if(cs.inline) return true;
                    var less = cs.getLessSource();
                    store.save(cs.url, less);
                    cs.setChanged(false);
                });
            }
            if(done) done();
        }
    }


    this.saveChangedFrameworks = function(done) {
        pinegrow.saveChangedFrameworks(this, done);
    }


    this.rename = function(new_url) {
        this.setUrl(new_url);
        this.name = this.tab_name = getPageName(new_url);
        this.updateNameDisplay();
  }

    this.updateNameDisplay = function() {
        this.view.updateTitle();
        if(this.codeEditor) {
            //this.codeEditor.updateTitle();
            this.codeEditor.pageWasRenamed();
        }
        //this.$page.find('.name').html(this.name);
        //this.$page.find('a.crsa-preview').attr('href', this.url);
    }


    this.duplicate = function() {
        var cp = new CrsaPage();
        cp.wrapper_url = this.wrapper_url;
        cp.wrapper_selector = this.wrapper_selector;
        cp.scrollMode = this.scrollMode;
        cp.javascriptEnabled = this.javascriptEnabled;
        cp.parsedStatus = this.parsedStatus;

        cp.loaded = this.loaded;
        if(this.sourceNode) {
            cp.sourceNode = this.sourceNode.clone();
        }

        cp.frameworks = this.frameworks.slice(0);
        cp.frameworks_added = this.frameworks.length > 0 ? 2 : 0;
        cp.frameworksChanged();

        cp.breakpoints = this.breakpoints.slice(0);
        var skip = [];
        for(var j = 0; j < crsaPages.length; j++) {
            skip.push(crsaPages[j].name);
        }

        /*
        cp.name = crsaDuplicateName(this.name, this.localFile, skip);
        if(this.localFile) {
            cp.setLocalFile(crsaGetFileDir(this.localFile) + cp.name);
        }*/

        cp.name = this.name;
        cp.localFile = this.localFile;

        cp.setUrl(this.url);
        return cp;
    }



    this.reload = function(done) {
        this.view.reload(done);
    }

    var refresh_done_callbacks = [];

    this.refresh = function(done, silent_multi_refresh) {
        done && refresh_done_callbacks.push(done);

        this.delayedTasks.run('refresh', function() {
            _this.view.refresh(function(a,b,c,d) {
                var list = refresh_done_callbacks.slice(0);
                refresh_done_callbacks = [];
                list.forEach(function(d) {
                    d(a, b, c, d);
                })
            }, null, silent_multi_refresh);
            pgf.srcset_mode && !silent_multi_refresh && pinegrow.showQuickMessage('Page view refreshed.');
        }, 50);
    }

    this.refreshWithUpdateStylesheets = function() {
        this.refresh(function($iframe, page, refresh, pv) {
            if(pv == _this.activeView) {
                _this.updateStylesheetsList();
            }
        });
    }

    this.autoSize = function() {
        this.view.updateSize();

        //$.fn.crsapages('autoSizePage', this.$iframe, this.scrollMode);
    }


    this.setLiveUpdate = function(crsaPage) {
        var $body = $('body');
        if(!crsaPage) {
            if(this.live_update) {
                //$body.off('crsa-page-changed', onPageChanged);
            }
            this.$page.removeClass('mirrored');
            this.live_update = null;
        } else {
            //$body.on('crsa-page-changed', onPageChanged);
            this.$page.addClass('mirrored');
            var source = crsaPage.live_update ? crsaPage.live_update : crsaPage;
            this.live_update = source;
        }
    }

    this.getCurrentFrameworkHandlerReturnValue = function(handler) {
        return this.currentFrameworkHandlerReturnValues[handler] || null;
    }

    this.callFrameworkHandler = function(name, a, b, c, d, e, f) {
        /*
        this.currentFrameworkHandlerReturnValues[name] = null;
        for(var i = 0; i < _this.frameworks.length; i++) {
            var f = _this.frameworks[i];
            if(name in f && f[name]) {
                this.currentFrameworkHandlerReturnValues[name] = f[name](_this, a, b, c, d, e, f);
            }
        }
        */
        //will also dispatch to page framework handlers
        return pinegrow.dispatchEvent(name, this, a, b, c, d, e, f);
/*
        var ret = this.currentFrameworkHandlerReturnValues[name];
        this.currentFrameworkHandlerReturnValues[name] = null;
        return ret;*/
    }

    this.callGlobalFrameworkHandler = function(name, a, b, c, d, e, f) {
        return pinegrow.callGlobalFrameworkHandler(name, a, b, c, d, e, f, this);
    }

    this.getSource = function(full_source, skip_beautify) {

        return this.getCode(true);
    }

    this.applyChangesToSource = function(html_or_node, abort_on_errors, skip_was_changed, update_stylesheets, event_type) {//todo30 - ?????
        /* returns
            true - updated
            array - errors
            false - whole doc changed, not updated
            previewUpdated
         */
        var selected = pinegrow.selectedElements.getSelectedOnPage(this);
        var ret = {changes: null, errors: null, updated: false, stylesheets_changed: false, scripts_changed: false, whole_changed: false, replace_selected:[]};

        var rootNode;
        var clone = false;

        var parser = null;

        var profile = new CrsaProfile();

        if(this.onlyCode) {
            var code = typeof html_or_node == 'string' ? html_or_node : html_or_node.toStringOriginal(true);
            if(code != this.code) {
                this.code = code;
                var project = pinegrow.getCurrentProject();

                if(project && project.loaded_from_backend) {
                    project.setContentOfUrl(this.url, code);
                }
                this.setPageChanged(true);
            }
            ret.updated = true;
            return ret;
        }

        if(typeof html_or_node == 'string') {
            parser = new pgParser();
            parser.assignIds = false;
            parser.parse(html_or_node);
            rootNode = parser.rootNode;
        } else {
            rootNode = html_or_node;
            clone = true;
        }

        ret.errors = rootNode.validateTree(true /* doing strict validation */);
        if(ret.errors.length && abort_on_errors) {
            //debugger;
            //console.log('ERRORS');
            //console.log(ret.errors);
            return ret;
        }

        profile.show('applyChangesToSource PARSED');

        var changes = pgFindChangedNodesInPage(this.sourceNode, rootNode);
        ret.changes = changes;

        //console.log(changes);

        profile.show('applyChangesToSource FOUND CHANGES');

        //return ret;

        var $update_els = $([]);

        var done = false;
        var stylesheets_changed = false;

        var css_tags = ['style', 'html', 'head'];
        var whole_tags = ['html', 'head'];

        var isCSS = function(node) {
            return node.tagName == 'style' || (node.tagName == 'link' && node.getAttr('rel') == 'stylesheet');
        }

        var containsCSS = function(node) {
            var has = false;
            node.walk(function(n) {
                if(n.isElement && isCSS(n)) {
                    has = true;
                    return false;
                }
                return true;
            })
            return has;
        }

        if(changes.length && this.openedInEditor) {
            pgCurrentRequestContext.setCurrentRequestContext(this.url, rootNode);

            var whole = false;
            for(var i = 0; i < changes.length; i++) {
                var ch = changes[i];
                if(ch.original.parent == null || whole_tags.indexOf(ch.original.tagName) >= 0 || whole_tags.indexOf(ch.changed.tagName) >= 0) {
                    whole = true;
                    ret.whole_changed = true;
                    break;
                }
            }
            if(!whole) {
                for(var i = 0; i < changes.length; i++) {
                    var ch = changes[i];

                    var pgid = ch.original.getId();

                    var pgel = ch.original;

                    if(clone) {
                        ch.changed = ch.changed.clone(true);
                    }
                    ch.changed.assignIdAndAddToCatalog(true);

                    if(selected.indexOf(ch.original) >= 0) {
                        ret.replace_selected.push({
                            old_pgel: ch.original,
                            new_pgel: ch.changed
                        })
                    }

                    //check if big errors when not abort_on_errors
                    /*
                    if(!abort_on_errors && ret.errors.length) {
                        if(ch.changed.text().indexOf('data-pg-id') >= 0) {
                            console.error('Big error', ch.changed);
                        }
                    }
                    */
                    ch.original.replaceWith(ch.changed);

                    if(!stylesheets_changed) {
                        if(isCSS(ch.original) || containsCSS(ch.original) || isCSS(ch.changed) || containsCSS(ch.changed)) {
                            stylesheets_changed = true;
                        }
                    }

                    if(ch.original.tagName == 'script' || ch.changed.tagName == 'script') {
                        ret.scripts_changed = true;
                    }

                    done = true;
                }
            }
            ret.stylesheets_changed = stylesheets_changed;

            if(done) {
                if(stylesheets_changed && update_stylesheets) {
                    _this.updateStylesheetsList();
                }
                this.setPageChanged(true);
                if(!skip_was_changed && this.openedInEditor) {
//                    didMakeChange(this.$iframe, null, null, null, event_type);
                }
                ret.updated = true;

                profile.show('applyChangesToSource DONE');

                return ret;
            }
        } else {
            //console.log('no changes');
            ret.updated = true;
            profile.show('applyChangesToSource DONE');

            return ret;
        }
        var old_html = this.sourceNode.toStringWithIds(true);

        this.sourceNode.removeWithoutEvents();

        this.sourceNode = clone ? rootNode.clone(true) : rootNode;
        this.sourceNode.assignIdAndAddToCatalog(true);


        //manually add event
        var pgEvent = new PgNodeEvent(this.sourceNode, 'setCode', null, {
            oldValue: old_html
        }, this);
        var undoEvent = new PgUndoEvent(pgEvent);
        pinegrow.undoManager.getHistoryForPage(this).addEvent(undoEvent);
        //end add event

        //console.log('whole document is changed.');
        //this.refresh();
        profile.show('applyChangesToSource DONE');

        return ret;
    }


    this.setSource = function(html, done, no_init, skip_dom_update) {

        var parser = new pgParser();
        parser.parse(html, function() {
            _this.sourceNode = parser.rootNode;
            var htmlNodes = parser.rootNode.find('>html');
            if(htmlNodes && htmlNodes.length) {
                var htmlCode = htmlNodes[0].html(null, true);

                var doc = getIframeDocument(_this.$iframe.get(0));
                var $html = $(doc).find('> html');

                $html[0].innerHTML = htmlCode;

                var attributes = $.map($html[0].attributes, function(item) {
                    return item.name;
                });
                $.each(attributes, function(i, item) {
                    $html.removeAttr(item);
                });

                var attrs = htmlNodes[0].getAttrList();
                for(var i = 0; i < attrs.length; i++) {
                    $html.attr(attrs[i].name, attrs[i].value);
                }

                if(!no_init) {
                    setTimeout(function() {
                        $.fn.crsacss('loadLessStyles', _this.$iframe.get(0), function() {
                            $.fn.crsa('updateStructureAndWireAllElemets', _this.$iframe);
                            $('body').trigger('crsa-stylesheets-changed');
                            _this.addCrsaStyles();
                            if(done) done();
                        });
                    }, 500);
                } else {
                    if(done) done();
                }
            } else {
                if(done) done();
            }
        });
    }

    this.runScripts = function(done) {
        if(done) done();
        return; //disable for now

        var crsaPage = this;
        setTimeout(function() {

            try {
                /*
                var highestTimeoutId = setTimeout(function() {}, 1000);
                for (var i = 0 ; i <= highestTimeoutId ; i++) {
                    clearTimeout(i);
                }
                var highestIntervalId = setInterval(function() {}, 1000);
                for (var i = 0 ; i <= highestIntervalId ; i++) {
                    clearInterval(i);
                }
                */
            }
            catch(err) {}

            var evalInlineScripts = function() {
                var scripts = crsaPage.getDocument().getElementsByTagName('script');
                for (var ix = 0; ix < scripts.length; ix++) {
                    var s = scripts[ix];
                    if(s.text) {
                        var scr = document.createElement('script');
                        scr.async = false;
                        scr.text = s.text;
                        $.each(s.attributes, function(idx, attr) {
                            scr.setAttribute(attr.nodeName, attr.nodeValue);
                        });
                        s.parentNode.replaceChild(scr, s);
                        //eval.call(crsaPage.getWindow(), s.text);
                    }
                }
                if(done) done();
            }

            var count = 0;
            var scripts = crsaPage.getDocument().getElementsByTagName('script');
            for (var ix = 0; ix < scripts.length; ix++) {
                var s = scripts[ix];
                if(s.src) {
                    var scr = document.createElement('script');
                    scr.async = false;
                    $(scr).on('load', function(e) {
                        setTimeout(function() {
                            count--;
                            if(count == 0) {
                                evalInlineScripts();
                            }
                        },1);
                    });
                    count++;
                    $.each(s.attributes, function(idx, attr) {
                        scr.setAttribute(attr.nodeName, attr.nodeValue);
                    });

                    s.parentNode.replaceChild(scr, s);
                }
            }
            if(count == 0) {
                evalInlineScripts();
            }
        }, 500);
    }

    this.addCrsaStyles = function($iframe) {
        if(pgf.use_webview) return;

        if(!$iframe) $iframe = this.$iframe;
        var o = {css: '.crsa-highlighted {\
            box-shadow: 0 0 10px rgba(0, 255, 0, 0.9) !important;\
        }\
        .aaaacrsa-selected {\
            box-shadow: 0 0 0px 2px rgba(0, 180, 255, 1) !important;\
        }\
        .crsa-edit {\
            cursor: default;\
        }\
        .pg-partial-container {\
            box-shadow: 0 0 0px 1px rgba(0, 236, 255, 1) !important;\
        }\
        [data-pg-hidden] {display: none !important;}\
        script-disabled {display: none !important;}\
        *[contenteditable="true"] {\
            outline: rgb(255, 215, 0) dotted 2px !important;\
            outline-offset: 0px !important;\
        }\
        .crsa-disable-hover {\
            pointer-events: none;\
        }\
        .crsa-in-drag body {\
            padding-bottom:200px !important;\
        }\
        [data-pg-preview-ref] {\
            box-shadow: 0 0 0px 1px rgba(0, 236, 255, 1) !important;\
        }\
        * {\
            user-select:none;\
        }\
        [contenteditable="true"], [contenteditable="true"] * {\
            user-select:auto;\
        }\
        '};

        ////rgb(255, 215, 0) dotted 2px !important;

        //keep for back compatibility
        if(/*!pgf.empty_ph_attribute &&*/ pinegrow.getSetting('show-placeholders', '1') == '1') {
            o.css += '.pg-empty-placeholder {\
            min-height: 100px;\
            min-width: 100px;\
            box-shadow: 0 0 0 1px rgba(0,0,0,0.15);\
            background-image:url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAYAAACp8Z5+AAAADklEQVQIW2NgQAXGZHAAGioAza6+Hk0AAAAASUVORK5CYII=);\
        }\
        ';
            o.css += '*.pg-no-placeholder:empty {\
            min-height: inherit;\
        }\
        ';
        }
        //, div:empty, section:empty, nav:empty, article:empty, aside:empty, header:empty, footer:empty, iframe:empty, embed:empty, object:empty, video:empty, source:empty, table:empty, form:empty, fieldset:empty, menu:empty, ul:empty, ol:empty {\

        o.css += '\
        .crsa-placed {\
            opacity: 0.5 !important;\
        }\
        html.crsa-stop-animations * {\
            -webkit-transition-property: none !important;\
            transition-property: none !important;\
            -webkit-transform: none !important;\
            transform: none !important;\
            -webkit-animation: none !important;\
            animation: none !important;\
        }';

        o.css += `
:root {
    scroll-behavior: auto !important;
}

html.pg-dm-surprise {
    transition: opacity 0.3s;
    opacity: 0 !important;
}

html.pg-dm-surprise-ended {
    transition: opacity 0.5s;
    opacity: 1 !important;
}

`;


        if(pgf.empty_ph_attribute && pinegrow.getSetting('show-placeholders', '1') == '1') {
            o.css += `
[${pgf.empty_ph_attribute}] {
            min-height: 100px;
            padding-top:1px;
            padding-bottom:1px;
            outline: rgba(200,200,200,0.75) dashed 1px;
    outline-offset: -1px;
            background-image:url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAYAAACp8Z5+AAAADklEQVQIW2NgQAXGZHAAGioAza6+Hk0AAAAASUVORK5CYII=);
}
`
        }
        if(pgf.kids) {
            o.css += `
h1:empty,p:empty {
    min-height:1em;
}
`
        }

        /*
        o.css += `
.pinegrow-ui-helper {
    pointer-events:none;
}
`
*/

        /*
        o.css += `
*::-webkit-scrollbar {
    height: 10px;
    width: 10px;
    background: #373738;
}
*::-webkit-scrollbar-corner {
    background: #373738;
}
*::-webkit-scrollbar-thumb {
    background: #4f4f4f;
}
`;
*/


        this.callFrameworkHandler('on_set_inline_style', o);

        var css = '<style id="' + cssid + '">' + o.css + '</style>';

        var $head = $(getIframeHead($iframe.get(0)));
        var $cssel = $head.find('#' + cssid);
        if($cssel.length == 0) {
            $head.prepend(css);
        } else {
            $cssel.replaceWith($(css));
        }
    }

    this.stopAnimations = function() {
        var $html = $(getIframeHtmlElement(this.$iframe.get(0)));
        $html.addClass('crsa-stop-animations');
        this.animationsStopped = true;
    }

    this.startAnimations = function() {
        var $html = $(getIframeHtmlElement(this.$iframe.get(0)));
        $html.removeClass('crsa-stop-animations');
        this.animationsStopped = false;
    }

    this.hasStoppedAnimations = function() {
        var $html = $(getIframeHtmlElement(this.$iframe.get(0)));
        return $html.hasClass('crsa-stop-animations');
    }

    this.doWithoutCrsaStyles = function(func) {
        var $head = $(getIframeHead(this.$iframe.get(0)));
        var $c = $head.find('#' + cssid);
        $c.detach();
        func();
        $head.append($c);
    }

    this.removeCrsaStyles = function() {
        if(pgf.use_webview) return;
        var $head = $(getIframeHead(this.$iframe.get(0)));
        var $c = $head.find('#' + cssid);
        $c.detach();
    }

    this.getCompatibleUrl = function(url) {
        if(crsaIsFileUrl(url)) {
            if(pinegrow.sourceParser) {
                url = pinegrow.httpServer.makeUrl(url);
            }
        } else {
            if(!this.webapp && pinegrow.httpServer) {
                url = pinegrow.httpServer.makeProxyUrl(url);
            }
        }
        return url;
    }

    var show_name_classes_filter = null;

    this.getShowNameClassesFilter = function() {
        return show_name_classes_filter;
    }

    this.setShowNameClassesFilter = function(v) {
        show_name_classes_filter = v;
    }

    this.warmupDevTools = function(done) {
        //maybe not needed anymore
        //still needed
        done && done();
        return;

        var page = this;

        pinegrow.devTools.updateIframes(function() {

            var $html = page.activeView.get$Html();

            var onDone = function() {
                $html.find('link[data-pg-href]').each(function(i, n) {
                    n.setAttribute('href', n.getAttribute('data-pg-href'));
                    n.removeAttribute('data-pg-href');
                })
                done && done();
            }

            $html.find('link[rel="stylesheet"]').each(function(i, n) {
                n.setAttribute('data-pg-href', n.getAttribute('href'));
                n.removeAttribute('href');
            })

            var selectedIframe = pinegrow.devTools.getSelectedChromeIframe({
                $iframe: page.$iframe
            });

            if (selectedIframe) {
                pinegrow.devTools.runDummyGetMatchedCSSRules(selectedIframe, function(a) {
                    console.log('GOT DUMMY CSS RULES', a);

                    onDone();
                })
            } else {
                onDone();
            }
        });
    }
}


var PgCurrentRequestContext = function() {

    var _this = this;

    var ctx = {documentNode: null, crsaPage: null, isRemote: false, remotePath: null, remotePathParts: null, remoteHost: null, base: "", body_element: null, html_element: null, pageView: null};

    this.makeProxyUrl = function(url) {
        return pinegrow.getProxyUrl(url);
    }

    this.releaseRequestContextForPage = function(page) {
        if(ctx.crsaPage == page) {
            this.setCurrentRequestContext(null, null, false);
        }
    }

    this.setCurrentRequestContext = function(fullPath, documentNode, preview_mode, page, pageView) {
        if(page) {
            ctx.crsaPage = page;
        } else {
            ctx.crsaPage = documentNode ? pinegrow.getCrsaPageOfPgParserNode(documentNode) : null;
        }
        ctx.documentNode = documentNode;
        ctx.previewMode = preview_mode || pgf.srcset_mode; //pgf.kids
        ctx.html_element = null;
        ctx.body_element = null;
        ctx.pageView = pageView;

        if(documentNode) {
            ctx.html_element = documentNode.findOne('html');
            ctx.body_element = documentNode.findOne('body');
        }

        if(!fullPath || crsaIsFileUrl(fullPath)) {
            ctx.isRemote = false;
            ctx.remotePath = null;
            ctx.remotePathParts = null;
            ctx.remoteHost = null;
            ctx.base = ""
        } else {
            //get base
            ctx.base = "";
            if(documentNode) {
                var baseList = documentNode.find('base', true); //single
                if(baseList.length) {
                    ctx.base = baseList[0].getAttr('href');
                }
            }
            ctx.isRemote = true;

            ctx.remotePath = crsaRemoveUrlParameters(fullPath).replace(/^[a-z]+:\/\//, '');
            ctx.remotePathParts = ctx.remotePath.split('/');

            if(ctx.remotePathParts.length) {
                ctx.remoteHost = ctx.remotePathParts.shift();

                if(ctx.remotePathParts.length) {
                    ctx.remotePathParts.pop();
                    ctx.remotePath = ctx.remotePathParts.join('/');
                }
            }
        }

    }


    this.createProxyUrlNodeOutputFilter = function(linkNode, str, type, attr_name, attr_value, attr_quote) {
        if(type === 'attrs' && !linkNode.script_info) {
            str = crsaRemoveScripts(str);

            if(ctx.crsaPage && !ctx.crsaPage.javascriptEnabled) {
                str = str.replace(/\s(on[a-z]*\="[^"]*")/g, ' pg-disabled-$1'); //remove onX JS attributes
            }
        }
        if(type === 'attrs' && (linkNode.tagName === 'script' || linkNode.tagName === 'link' || linkNode.tagName === 'img' || linkNode.tagName === 'base')) {

            var is_remote = ctx.isRemote;
            var remotePath = ctx.remotePath;
            var remoteHost = ctx.remoteHost;
            var remotePathParts = ctx.remotePathParts;

            var attrs = linkNode.getAttributesString(function(linkNode, attrName, attrValue, attrQuote) {
                var attrLower = attrName.toLowerCase();

                if(attrLower === 'href' || attrLower === 'src') {
                    var href = attrValue;

                    if(attrValue) {
                        //strip extra ../
                        if(crsaIsAbsoluteUrl(attrValue)) {

                            if(attrValue.indexOf('//') == 0) {
                                attrValue = 'http:' + attrValue;
                            }

                            var change = true;
                            if(attrValue.match(/^https?\:\/\//) && !attrValue.match(/^https?\:\/\/pinegrow\.com\/placeholders/)) {
                                change = false;
                            }

                            if(linkNode.tagName === 'link' && linkNode.getAttribute('rel') === 'stylesheet') {
                                change = true;
                                //pg node ids are not in Active CSS otherwise, have to use proxy

                                if(ctx.crsaPage.wrapper_url && !linkNode.getId()) {
                                    //this is a remote stylesheet of the wrapper element. load it as it is
                                    change = false;
                                }
                            }
                            //include without proxy... mages should be also ok. Make proxy for PG images so that server can map them to local
                            if(change) {
                                attrValue = _this.makeProxyUrl(attrValue);
                            }

                        } else if(is_remote) {
                            if(linkNode.tagName !== 'base') {
                                if(remotePath !== null) {
                                    var a = attrValue.split('/');
                                    var last = remotePathParts.length;
                                    for(var j = 0; j < a.length; j++) {
                                        if(a[j] != '..') break;
                                        last--;
                                    }
                                    if(last < 0) {
                                        while(last < 0) {
                                            a.shift();
                                            last++;
                                        }
                                        attrValue = a.join('/');
                                        //console.log('CSS rel link fixed: ' + href);
                                    }
                                }
                                if(attrValue.length) {
                                    if(attrValue.charAt(0) !== '/') {
                                        attrValue = ctx.base + attrValue;
                                    }
                                    if(attrValue.charAt(0) === '/') {
                                        attrValue = 'http://' + remoteHost + attrValue;
                                        attrValue = _this.makeProxyUrl(attrValue);
                                    }

                                    //Handle project level mapped urls
                                    var project = pinegrow.getCurrentProject();
                                    if(project) {
                                        attrValue = project.mapSourceUrlToDOMUrl(attrValue);
                                    }
                                }
                            }
                        } else {
                            //local
                            if(attrValue.charAt(0) === '/') {
                                var project = pinegrow.getCurrentProject();
                                if(project && !pgf.root_urls) {
                                    attrValue = project.getUrl() + attrValue;
                                    attrValue = _this.makeProxyUrl(attrValue);
                                }
                            }
                        }
                    }
                } else if(attrLower === 'integrity') {
                    return '';
                }
                var str = attrValue === null ? attrName : attrName + '=' + attrQuote + attrValue + attrQuote;
                str = crsaRemoveScripts(str);
                return str;
            });

            if(linkNode.tagName === 'script' && ctx.crsaPage && !ctx.crsaPage.javascriptEnabled) {
                attrs = attrs.replace(/type\=[\"\'][^\"\']*[\"\']/g, '') + ' type="pinegrow/disabled"';
            }
            return attrs;
        } else if(type === 'attrs' && (linkNode.tagName === 'a' || linkNode.tagName === 'form') && !ctx.previewMode) {
            //force remove target
            str = str.replace(/starget\=["'][^"']*["']/g, '');
            //if(!str.match(/\starget\=/i)) {
            str += ' target="_pg_blank"';
            //}
        } else if(type === 'node' && linkNode.script_info) {
            if(linkNode.script_info) {
                if(ctx.html_element && !linkNode.closest('html') && linkNode.document && linkNode.document == ctx.documentNode) {
                    //script is outside of html element, dont show it
                    str = '';
                } else {
                    str = '<php data-pg-id="' + linkNode.getId() + '" data-pg-script-type="' + linkNode.script_info.type + '">' + escapeHtmlCode( linkNode.content ) + '</php>';
                }
            }
        } else if(type === 'attribute') {
            if(attr_name == 'style') {
                if(str) {
                    if(ctx.pageView) {
                        str = ctx.pageView.emulateCSSFeatures(str);
                    }
                    return pgAddProxyLinksToStyleAttribute(str);
                }
            }
            return str;
        } else if(type === 'content' && linkNode.tagName == 'style') {
            //get cs with ids
            var cs = pinegrow.stylesheets.findCrsaInlineStylesheet(linkNode);
            if(cs) {
                str = cs.getSourceForRules(null, true) || str;
            }
            str = pgAddProxyLinksToStyleAttribute(str);
            if(ctx.pageView) {
                str = ctx.pageView.emulateCSSFeatures(str);
            }
        } else if(type === 'node' && linkNode.tagName === 'html' && ctx.pageView) {
            if(ctx.pageView.isMobileDevice) {
                str = str.replace(/\<html/i, '<html data-pg-mobile');
            }
            if(!ctx.pageView.isPgiaEnabled) {
                str = str.replace(/\<html/i, '<html data-pg-ia-disabled');
            }
        } /*else if(type === 'node' && linkNode.tagName === 'iframe' && pgf.kids) {
            return `<pgiframe>${str}</pgiframe>`;
        }*/
        return str;
    }
}

var pgCurrentRequestContext = new PgCurrentRequestContext();

var getCrsaPageForIframe = function($iframe) {
    if(!$iframe) return null;
    var cp = $iframe.data('crsa-page');
    if(!cp) {
        var pv = getPageViewOf$Iframe($iframe);
        if(pv) {
            cp = pv.getPage();
            $iframe.data('crsa-page', cp);
        }
    }
    return cp;
}


var getPageName = function(url) {
    var n = url.split(/[\\\/]/).pop();
    if(n.length == 0) n = 'index.html';
    var a = n.split('?');
    n = a.length > 1 ? a[0] : n;
    if(!n.length) n = 'index.html';
    return pgUrlToDisplayString(n);
}

//(function( $ ) {

    $.fn.crsapages = function( method ) {

        //var opts = $.extend( {}, $.fn.hilight.defaults, options );
        if ( methods[method] ) {
            return methods[method].apply( this, Array.prototype.slice.call( arguments, 1 ));
        } else if ( typeof method === 'object' || ! method ) {
            return methods.init.apply( this, arguments );
        } else {
            $.error( 'Method ' +  method + ' does not exist on jQuery.crsacss' );
        }
    };

    var options;
    var pages;
    var crsaPages = [];
    var currentZoom = 1;
    var centerLeft = 300;
    var centerRight = 300;
    var fitZoom = false;

    var $body = $('body');

    var $empty;
    var customPageActions = [];

    var s_ajaxListener = new Object();

    var handleXMLHTTPOpen = function(a,b,async,user,pass) {
        console.log('Sync AJAX call to ' + b + ' changed to async. PG would block otherwise.');
        s_ajaxListener.tempOpen.call(this, a, b, true, user, pass);
    }

var methods = {
        init : function( opts ) {
            options = $.extend( {}, $.fn.crsapages.defaults, opts );
            $empty = $('.empty-canvas');
            updatePagesList(this);

            if(isApp()) {
                var gui = require('nw.gui');
                gui.Window.get().on('document-start', function(win) {
                    if(win) {
                        var pv = getPageViewOfWindow(win);
                        if(pv) {
                            ///
                            //Disable ServiceWorkers
                            if(win.ServiceWorkerContainer) {
                                win.ServiceWorkerContainer.prototype.register = function(url) {
                                    console.log(`Service workers are disabled in Pinegrow: ${url}`);
                                    return new Promise(function(resolve, reject) {
                                        reject('Service workers are disabled in Pinegrow.');
                                    });
                                };
                            }
                            s_ajaxListener.tempOpen = win.XMLHttpRequest.prototype.open;
                            win.XMLHttpRequest.prototype.open = handleXMLHTTPOpen;
                            pv.get$Iframe().trigger('frame-document-start');
                        }

                    }
                });
                gui.Window.get().on('document-end', function(frame) {
                    //console.log('document-end');
                });
                gui.Window.get().on('new-win-policy', function(frame, url, policy) {
                    console.log('Win open: ' + url);
                    if(url != window.location.href) {
                        //policy.forceCurrent();

                        policy.ignore();
                        /*if(url.indexOf('autopilothq') > 0) {
                            require('nw.gui').Shell.openExternal(url);
                        }*/
                    }
                });
            }

        },
        getAllPages : function() {
            return crsaPages;
        },
        clearUndoSetFlag : function() {
            $.each(crsaPages, function(i,p) {
                p.undoSetFlag = false;
            })
        },
        saveAll : function(close_on_save) {

            pinegrow.dispatchEvent('on_before_save');

            var savedProjects = [];
            $.each(crsaPages, function(i,cp) {
                //cp.saveChangedFrameworks();
                if(!cp) return;

                if(cp.live_update && cp.save_parent) return true;

                if(cp.changed || cp.hasCssChanges()) {
                    var do_proj = false;
                    if(cp.crsaProjectTemplate && savedProjects.indexOf(cp.crsaProjectTemplate) < 0) {
                        do_proj = true;
                        savedProjects.push(cp.crsaProjectTemplate);
                    }
                    cp.save(function() {
                        if(close_on_save) {
                            cp.force_close = true;
                            if(cp.view) cp.view.closeTab();
                        }
                    }, true, true, false, do_proj);
                }
            })
            //background pages
            var project = pinegrow.getCurrentProject();
            if(project) {
                project.getAllBackgroundPages().forEach(function(cp) {
                    if(cp.changed) {
                        cp.save(function() {
                        });
                    }
                })
            }
            pinegrow.saveChangedFrameworks();
        },
        downloadAllPages : function() {
            crsaLoadScript('lib/jszip/jszip.js', function() {
                var $b = $('<p><i class="fa fa-refresh fa-spin"></i> Preparing the Zip file. Please wait...</p>');
                showAlert($b, "Download project");

                var zip = new JSZip();

                var doneAll = function() {
                    $b.html('<a download="PinegrowProject.zip">Download the project Zip file</a></p><p>The zip file contains .html sources of all open pages, stylesheets in .css and .less format as well as Bootstrap files.');
                    $a = $b.find('a');
                    $a.attr('href', window.URL.createObjectURL(zip.generate({type:"blob"})));
                }

                var doneCss = function() {
                    if(project) {
                        project.copyRequiredFilesToZip(zip, doneAll);
                    } else {
                        doneAll();
                    }
                }

                var project = null;

                $.each(crsaPages, function(i, cp) {
                    zip.file(cp.name, cp.getSource(true));
                    if(cp.crsaProjectTemplate) project = cp.crsaProjectTemplate;
                    cp.setPageChanged(false);
                });

                var cslist = pinegrow.stylesheets.getAllCrsaStylesheets();

                var count = 0;
                $.each(cslist, function(i, cs) {
                    count++;
                    var less = cs.getLessSource();
                    cs.getCssSource(function(css) {
                        setTimeout(function() {
                            zip.file(cs.name, css);
                            var less_name = cs.name.replace(/\.css$/i, '.less');
                            if(less_name == cs.name) less_name += '.less';
                            zip.file(less_name, less);
                            count--;
                            if(count == 0) doneCss();
                        }, 10);
                    });
                });
                if(count == 0) {
                    doneCss();
                }
            });
        },
        addCustomPageAction : function(name, func) {
            //depreceated
            customPageActions.push({label: name, func: func});
        },
        addPage : function(src, loaded, providedCrsaPage, onLoadCanceled, options) {
            var name = getPageName(src);
            var _this = this;

            var crsaPage;
            if(!providedCrsaPage) {
                crsaPage = new CrsaPage();
                crsaPage.name = name;

                crsaPage.webapp = pgf.edit_apps && !crsaIsFileUrl(src);

                if(crsaPage.webapp) {
                    var p = new pgParser();
                    p.parse('<html></html>');
                    crsaPage.sourceNode = p.rootNode;
                }
                crsaPage.setUrl(src);
            } else {
                crsaPage = providedCrsaPage;
            }
            crsaPage.openOptions = options || {};
            crsaPage.loading = true;

            if(!crsaPage.view) {
                crsaPage.view = new PgPageViewTab(crsaPage);

            } else {
                crsaPage.view.addView(crsaPage);
            }
            crsaPage.setPageChanged(false, true);



            if(options) {
                if(options.wrapper_url) crsaPage.wrapper_url = options.wrapper_url;
                if(options.wrapper_selector) crsaPage.wrapper_selector = options.wrapper_selector;
                if(options.original_new_url) {
                    crsaPage.original_new_url = options.original_new_url;
                }
                if(options.on_source_parsed) {
                    crsaPage.onSourceParsed = options.on_source_parsed;
                }
            }

            crsaPages.push(crsaPage);

            crsaPage.view.load(function(a, b) {
                if(loaded) loaded(a, b);
            }, onLoadCanceled);

            methods.updateFilesName();

            updatePagesList(this);

            crsaPage.autoSize();
            return this;
        },
        closePage: function(crsaPage, leave_css) {
            var project = pinegrow.getCurrentProject();

            if(project && project.loaded_from_backend && crsaPage.changed) {
                //save the code to the project
                var p = pinegrow.getCurrentProject();
                if(p) {
                    p.setContentOfUrl(crsaPage.url, crsaPage.getCodeForSave());
                }
            }
            if(!crsaPage.onlyCode) {
                var $page = crsaPage.$page;
                var $iframe = $page.find('iframe,webview');
                var $canvas = $('div.canvas');
                methods.updateFilesName(crsaPage);

                pinegrow.clearSelectedElements();
                $('body').trigger('crsa-page-closed', crsaPage);

                var ps = getCrsaPageStylesForPage($iframe);
                if (ps && !leave_css) {
                    ps.removeAllExclusiveSheets();
                }
            }
            var idx = crsaPages.indexOf(crsaPage);
            if(idx >= 0) crsaPages.splice(idx, 1);

            crsaPage.closed_without_saving = crsaPage.changed;

            crsaPage.onClose();

            if(crsaPage.view) {
                crsaPage.view.closePage(crsaPage);
            }

            pinegrow.lastActivePage.closePage(crsaPage);

            var project = pinegrow.getCurrentProject();
            if(project && !project.is_closing_all_pages) {
                project.pageWasOpenedOrClosed(crsaPage, false);
            }

            updatePagesList($canvas);
            methods.refresh();
            pgCurrentRequestContext.releaseRequestContextForPage(crsaPage);
            $('body').trigger('crsa-stylesheets-changed');
            $body.trigger('crsa-page-closed-removed', crsaPage);
            pinegrow.dispatchEvent('on_page_closed', crsaPage);
        },
        updateFilesName : function (excludePage) {
            var pagesObj = {};
            for (var i=0; i<crsaPages.length; i++) {
                if (crsaPages[i] == excludePage) continue;
                crsaPages[i].tab_name = crsaPages[i].name;
              if (!pagesObj[crsaPages[i].name]) pagesObj[crsaPages[i].name] = [];
              pagesObj[crsaPages[i].name].push(crsaPages[i]);
            }

            for (var name in pagesObj) {
              if (pagesObj[name].length > 1) {
                var pagesWithSameName = pagesObj[name];
                for (var i=0; i < pagesWithSameName.length; i++) {
                  var fileName = pagesWithSameName[i].name;
                  var dirUrl = crsaGetBaseForUrl(pagesWithSameName[i].url);
                  var dirName = crsaGetNameFromUrl(dirUrl);
                  if(path) {
                      pagesWithSameName[i].tab_name = path.join(dirName, fileName);
                  } else {
                      pagesWithSameName[i].tab_name = dirName + '/' + fileName;
                  }
                  //pagesWithSameName[i].$page.find('span.name').html(pagesWithSameName[i].tab_name);
                    pagesWithSameName[i].view.updateTitle(pagesWithSameName[i].tab_name);
                }
              }
              else {
                //pagesObj[name][0].$page.find('span.name').html(pagesObj[name][0].tab_name);
                  pagesObj[name][0].view.updateTitle(pagesObj[name][0].tab_name);
              }
            }
        },
        checkIfPageExists : function(url, done) {
            var project = pinegrow.getCurrentProject();

            if((!project || !project.loaded_from_backend) && crsaIsFileUrl(url))
            {
                var file = crsaMakeFileFromUrl(url, true);

                try {
                    var fs = require('fs');
                    var stat = fs.statSync(file);
                    if(stat.isFile()) {
                        done(true);
                    } else {
                        done(false, 'NOT A FILE - Path is a folder');
                    }
                } catch(err) {
                    done(false, err);
                }
                return;

                //console.log('Checking file ' + file);
                if(crsaIsFileOrDir(file) == 'file') {
                    //console.log('File ' + file + ' found');
                    done(true);
                } else {
                    console.log('File ' + file + ' not found.');
                    done(false);
                }
            } else {
                $.ajax({
                    url: url,
                    data: null,
                    dataType: 'html'
                })
                    .done(function(data) {
                        done(true);
                    }).fail(function(a, b, c) {
                        done(false);
                    });
            }
        },
        showOverlays : function(hide, cls, animate) {
            var add = function($p, $if) {
                var $c = $p.find('.content');
                var $o = $p.find('.iframe-overlay');

                if (!hide) {
                    var pos = $if.position();

                    if(pgf.use_webview) {
                        pos.top = 0;
                    }
                    if ($o.length == 0) {
                        $o = $('<div/>', {'class': 'iframe-overlay'}).appendTo($c).data('iframe_element', $if).css('top', pos.top + 'px');
                    }
                    gsap.killTweensOf($o.get(0));
                    cls && $o.addClass(cls);

                    if(animate) {
                        gsap.to($o.get(0), {
                            opacity: 1,
                            duration: 0.25,
                            onComplete: function() {
                            }
                        })
                    }

                } else {
                    if(animate) {
                        gsap.to($o.get(0), {
                            opacity: 0,
                            duration: 1,
                            onComplete: function() {
                                $o.remove();
                                cls && $o.removeClass(cls);
                            }
                        })
                    } else {
                        $o.remove();
                        cls && $o.removeClass(cls);
                    }
                }
            }

            pinegrow.getAllPages().forEach(function(crsaPage) {
                if(crsaPage.view) {
                    crsaPage.getPageViews().forEach(function(pv) {
                        var $p = pv.get$Page();
                        var $if = pv.get$Iframe();
                        add($p, $if);
                    })
                }
            });
            $('.page-lib').each(function(i, pl) {
                var $p = $(pl);
                var $if = $p.find('iframe');
                add($p, $if);
            })
        },
        showLoadingOverlay : function(pv, hide, onCancel) {

            //if(pgf.use_webview) return null;

            var $if = pv.get$Iframe();
            var $p = pv.get$Page();
            var $c = $p.find('.content');
            var $o = $p.find('.iframe-loading-overlay');
            var pos = $if.position();

            if(pgf.use_webview) {
                pos.top = 0;
            }

            if(!hide) {
                if($o.length == 0) {
                    var msg = pgf.use_webview ? '<a href="#">Cancel</a>' : 'If the page doesn\'t load check your firewall settings or / and try changing the internal web server port in Support -&gt; Settings.';

                    $o = $('<div/>', {'class' : 'iframe-loading-overlay'}).html('<div><i class="icon-REFRESH icon-spin"></i>Loading, please wait...<p style="padding-left:40px;padding-right:40px;"><small>' + msg + '</small></p></div>').appendTo($c).css('top',pos.top + 'px');
                    $o.find('a').on('click', function(e) {
                        e.preventDefault();
                        methods.showLoadingOverlay(pv, true);
                        if(onCancel) onCancel();
                    })
                }
            } else {
                $o.remove();
            }
        },
        showDeviceValue: function (pageView) {
            var $device = pageView.get$Page().find('.device-menu');
            if (!$device) return;

            var done = false;
            $.each(options.sizes, function(key, size) {
                if(pageView.deviceWidth == size) {
                    $device.find('>button').html(size + 'px <span class="caret"></span>');
                    done = true;
                    return false;
                }
            });
            if(!done) {
                $device.find('>button').html(pageView.deviceWidth + 'px <span class="caret"></span>');
            }
        },
        getZoom : function(page) {
            return page ? page.zoom : 1.0;//currentZoom;
        },
        refresh : function() {
            //debugger;
            var cw = 0;
            var $code = $("#textedit");
            if($code.is(':visible')) {
                cw = $code.width() + 5;
            }
            //methods.centerPages($('#crsa-left-plane').is(':visible') ? centerLeft : 10, ($('#crsa-tree').is(':visible') ? $('#crsa-tree').width() : 10) + cw);

            methods.centerPages(centerLeft, centerRight);

            if(fitZoom) {
                methods.zoom(getFitZoomScale());
            } else {
                methods.zoom(currentZoom);
            }
        },
        autoSizePage : function($if, scroll_mode) {
            sizePage($if, currentZoom, scroll_mode);
        },
        setFit : function(fit) {
            fitZoom = fit;
            methods.refresh();
        },
        getFit : function(fit) {
            return fitZoom;
        },
        canvasResized : function() {

            methods.refresh();
            return;
            if(fitZoom) {
                methods.zoom(getFitZoomScale());
            }
        },
        zoom : function(zoom) {
            return;


            var $pages = pages.find('> div.content');
            var $iframes = pages.find('iframe,webview');
            var w = 800;

            currentZoom = zoom;
           // methods.centerPages(centerLeft, $('#crsa-tree').width());

            $pages.each(function(i,e) {
                var $e = $(e);

                if($e.closest('.page-tab').length) return true;

                var $if = $e.find('iframe,webview');
                var cp = getCrsaPageForIframe($if);

                w = cp.deviceWidth;
                var cw = Math.ceil(w * zoom);

                //.css('height', '') leave it out
                $if.css('width',w + 'px').css('transform','scale(' + zoom + ', ' + zoom + ')').css('transform-origin', '0 0');

                sizePage($if, zoom, cp.scrollMode, w);

                $if.data('zoom', zoom);
            });
            repositionSelectedElementMenu();
            //$.fn.crsa('refreshSelectElement'); maybe dont need this
            return this;
        },
        centerPages : function(leftp, rightp) {
            return;

            //debugger;
            centerLeft = leftp;
            centerRight = rightp;
            var $canvas = $('div.canvas');
            var w = $(window).width();// $canvas.outerWidth();
            var space = w - leftp - rightp;
            var margin = 20;
            var $pages = $canvas.find('div.page');
            var pagew = $pages.width() + margin;
            var pagesPerLine = Math.floor(space / pagew);
            if(pagesPerLine < 1) pagesPerLine = 1;
            if(pagesPerLine > $pages.length) pagesPerLine = $pages.length;
            var pagesWidth = pagew * pagesPerLine;
            var paddingLeft = leftp + (space - pagesWidth)/2 - 20;
            var paddingRight = rightp + (space - pagesWidth)/2 - 20;

            paddingLeft = centerLeft;
            paddingRight = centerRight;
            //$canvas.css('padding-left', paddingLeft + 'px').css('padding-right', paddingRight + 'px');
            if(typeof testGoldenLayout != 'function') {
                $canvas.css('overflow-y', 'auto').css('overflow-x', 'auto').css('margin-left', paddingLeft + 'px').css('margin-right', (paddingRight + 0) + 'px');
            }
          /*  if(pagesWidth < space) {
                $canvas.css('overflow-x', 'hidden');
            } */

        }
    };

//w * z + space = cw
//z = (cw - space)/w

    var getFitZoomScale = function() {
        var spc = 24;
        var w = 0,
            sw = spc;
        $.each(crsaPages, function(i,cp) {
            if(cp.visible) {
                w += cp.deviceWidth;
                sw += spc;
            }
        });
        var space = $('div.canvas').width();// - 12;// - centerLeft - centerRight;
        //console.log('fit ' + w + ' ' + space);
        var z = (space - sw) / w;
        var min = 200;//230;//214;


        w = 0;
        sw = spc;
        do {
            var all_min = true;
            var len = sw;
            $.each(crsaPages, function(i,cp) {
                if(cp.visible) {
                    if(cp.deviceWidth * z < min) {
                        len += min;
                    } else {
                        len += cp.deviceWidth * z;
                        all_min = false;
                    }
                    len += spc;
                }
            });
            if(len > space) z = z * 0.95;
        } while(!all_min && len > space);

        if(z > 1.0) z = 1.0;
        return z;
    }

    var sizePage = function($if, zoom, scroll_mode, w) {
        var $e = $if.closest('.content');
        if(!w) w = $if.width();
        var cw = Math.ceil(w * zoom);
        //$if.css('height', ''); disable scroll_mode
        var h;
        var eh;
        var $canvas = $('div.canvas');
        if(scroll_mode) {
            var bh = /*$body.height() - 90 */ $canvas.height() - 50 - $.fn.crsa('getPageTabs').getHeight() - pinegrow.statusBar.getHeight() /* status bar */;
            h = bh / zoom;
            eh = bh;
        } else {
            h = Math.ceil($if[0].contentWindow.document.body.scrollHeight * 1.05);
            eh = Math.ceil(h * zoom);
        }
        $if.css('height', h + 'px').css('width',w + 'px');
        $e.css('width', cw + 'px').css('height', eh + 'px' );

        var $p = $e.parent();
        if(cw < 225) {
            $p.addClass('narrow');
        } else {
            $p.removeClass('narrow');
        }

    }

    var panels_hidden = true;

    var updatePagesList = function($el) {
        $('#crsa-tree').show();
        pages = $('div.page');
        return;

        pages = $('div.page');
        if(pages.length == 0 && !pinegrow.getCurrentProject()) {
            $empty.show();
            pinegrow.statusBar.$element.hide();
            alignEmptyCanvas();
            if(!pinegrow.getCurrentProject()) {
                $('#crsa-left-plane').hide();
                $('#crsa-tree').hide();
                $(window).trigger('resize');
                panels_hidden = true;
            }
            //pinegrow.hideLeftPanel();
        } else {
            $empty.hide();

           // if(panels_hidden) {
                $('#crsa-left-plane').show();
                $('#crsa-tree').show();
                panels_hidden = false;
                $(window).trigger('resize');
            //pinegrow.statusBar.$element.show();
           // }
        }
    }
    // Plugin defaults – added as a property on our plugin function.
    $.fn.crsapages.defaults = {
        title : 'crsa',
        sizes : {
            'desktop' : 1600,
            'laptop' : 1280,
            'tablet-landscape' : 1024,
            'tablet-portrait' : 768,
            'phone' : 320
        }
    };


    function getIframeBody(iframe) {
        var doc = iframe.contentDocument || iframe.contentWindow.document;
        return doc.body;
    }

    function getIframeHtmlElement(iframe) {
        var doc = iframe.contentDocument || iframe.contentWindow.document;
        var head  = doc.getElementsByTagName('html')[0];
        return head;
    }

    function getIframeHead(iframe) {
        var doc = iframe.contentDocument || iframe.contentWindow.document;
        var head  = doc.getElementsByTagName('head')[0];
        return head;
    }

function getIframeTag(iframe, tag) {
    var doc = iframe.contentDocument || iframe.contentWindow.document;
    var head  = doc.getElementsByTagName(tag)[0];
    return head;
}

//})( jQuery );

var pgGetActionsForPageMenu = function(menu, crsaPage) {

    crsaPage = crsaPage || pinegrow.getSelectedPage();
    var $body = $('body');

    if(!crsaPage) {
        menu.add({
            label: 'Open a page first!',
            action: function() {

            }
        })

        return;
    }

    if(pgf.edit_html) {

        menu.add({
            label: 'Page properties...',
            action: function() {
                var html = crsaPage.sourceNode.findOne('html');
                if(!html) {
                    pinegrow.showQuickError('The page does not have the top HTML element.')
                } else {
                    pinegrow.showQuickMessage('Select the top HTML element and go to the Properties panel.');

                    pinegrow.selectElement(html);
                    var view = pinegrow.showUIPanel('prop');

                    setTimeout(function() {
                        pgRevealElement( view.get$Element().find('.crsa-field-page-title').parent());
                    }, 200)
                }
            }
        });
/*
        menu.add({
            label: 'Manage fonts...',
            action: function () {
                pinegrow.fontLibrary.showLibrary();
            }
        })


 */
        menu.add({
            type: 'divider'
        })

        menu.add({
            label: 'Duplicate',
            action: function () {
                pinegrow.duplicatePage(crsaPage);
            }
        })
    }

    /* Not sure this is useful
    menu.add({
        label: 'Revert to saved',
        action: function() {
            pinegrow.reloadPage(crsaPage);
        }
    })*/

    menu.add({
        label: 'Refresh views',
        kbd: 'CMD R',
        action: function() {
            crsaPage.refresh();
        }
    })

    /*
    menu.add({
        label: 'Rename',
        action: function() {
            $body.trigger('crsa-page-rename', crsaPage);
        }
    })*/

    if(pgf.edit_html) {
        menu.add({
            label: 'Manage stylesheets...',
            action: function () {
                $.fn.crsacss('showStylesheetsManager');
            }
        })


        menu.add({
            label: 'Manage libraries &amp; plugins...',
            action: function () {
                $.fn.crsacss('showFrameworkManager', crsaPage);
            }
        })

        menu.add({
            label: 'Manage breakpoints...',
            action: function () {
                crsaPage.showManageBreakpoints();
            }
        })

        if (!pinegrow.isContributorMode()) {

            menu.add({
                label: 'Edit code',
                kbd: 'CMD E',
                action: function () {
                    pinegrow.editCode(crsaPage);
                }
            })
        }

        menu.add({
            label: 'Preview in browser',
            kbd: 'CMD B',
            action: function () {
                crsaPage.openInBrowser();
            },
            check: '<i class="icon icon-Pin2"></i>',
            check_action: function (e) {
                if (isApp()) {
                    var gui = require('nw.gui');
                    var clipboard = gui.Clipboard.get();
                    var $link = $(e.delegateTarget).closest('a');
                    clipboard.set(crsaPage.getPreviewUrl(), 'text');

                    pinegrow.showQuickMessage('Link to the page was copied to the clipboard.')

                    e.stopPropagation();
                    e.preventDefault();
                }
            }
        })

        menu.add({
            label: 'Get free stock photos from Unsplash...',
            kbd: null,
            action: function () {
                pgShowQuickImageBrowser();
            }
        })
    }

    var customPageActions = [];

    if(crsaPage.sourceNode) {
        crsaPage.callFrameworkHandler('on_page_menu', customPageActions);
    }

    menu.addMany(customPageActions);

    menu.add({
        type: 'divider',
        action: function() {

        }
    })

    menu.add({
        label: 'CSS animations',
        check: function() {
            return !crsaPage.animationsStopped;
        },
        action: function() {
            var val = !this.check_value;
            if(val) {
                crsaPage.startAnimations();
            } else {
                crsaPage.stopAnimations();
            }
            this.setCheckValue(val);
            pinegrow.showQuickMessage('CSS animations on ' + crsaPage.getName() + ' are ' + (val ? 'ON' : 'OFF'));
        }
    })

    menu.add({
        label: 'JavaScript enabled',
        check: function() {
            return crsaPage.javascriptEnabled;
        },
        action: function() {
            var val = !this.check_value;
            crsaPage.setJavascriptEnabled(val);
            this.setCheckValue(val);
            pinegrow.showQuickMessage('Javascript on ' + crsaPage.getName() + ' is ' + (val ? 'ON' : 'OFF'));
        }
    })

    menu.add({
        type: 'divider',
        action: function() {

        }
    })

    menu.add({
        label: 'Close',
        action: function() {
            crsaPage.askClose();
        }
    })

    return;
}
