


var CrsaHttpServer = function(status_callback) {

    var config = {
        directory: '',
        port: parseInt(process.env[pgf.product + '_WEBSERVER_PORT'] || pinegrow.getSetting('webserver-port', '40000')),
        host : process.env[pgf.product + '_WEBSERVER_HOST'] || pinegrow.getSetting('webserver-host', 'localhost')
    };

    var libs  = {
        http: require('http'),
        url: require('url'),
        fs: require('fs'),
        path: require('path')
    };

    /*

    ON by default now

    if(pinegrow.getSetting('root-urls', '0') === '1') {
        pgf.root_urls = true;
        pinegrow.showMode('Serve local projects from server root', true);
    }
    */

    var _this = this;
    this.serverStarted = false;

    this.url = 'http://' + $.trim(config.host) + ':' + config.port;

    this.encodeUrl = function(url) {
        return url;
        //return url.replace(/#/g, '%23').replace(/\?/g, '%3F').replace(/;/g, '%3B').replace(/\&/g, '%26');
    }

    this.decodeUrl = function(url) {
        return url;
        //return decodeURIComponent(url);
    }

    this.makeUrl = function(url) {
        if(!pgf.use_server) return url;

        if(pgf.root_urls) {
            var p = pinegrow.getCurrentProject();
            if(p) {
                var rel = p.getRelativeUrl(url);
                if(!crsaIsAbsoluteUrl(rel)) {
                    return this.url + '/' + rel;
                }
            }
        }

        if(url.indexOf('file://') === 0) {
            url = this.encodeUrl(url);
        }
        return url.replace('file://', this.url + '/file://');
    }

    this.getOriginalUrl = function(url) {
        if(url && url.indexOf(this.url + '/') == 0) {
            url = url.replace(this.url + '/', '');

            if(pgf.root_urls) {
                if(url.indexOf('://') < 0) {
                    var p = pinegrow.getCurrentProject();
                    if(p) {
                        return p.getAbsoluteUrl(url);
                    }
                }
            }
            url = this.decodeUrl(url);
            if(url.indexOf('://') < 0) url = 'file://' + url;
        }
        return url;
    }

    this.makeProxyUrl = function(url, live) {
        if (url == null) return "";

        if(url.indexOf('file://') === 0) {
            url = this.makeUrl(url);
        } else if(!crsaIsAbsoluteUrl(url)) {
            //console.log('MakeProxyUrl: url is relative: ' + url);
        } else if(url.indexOf(this.url) != 0) {
            url = this.url + '/' + this.encodeUrl(url);
        }
        if(live) {
            //url = crsaAppendQueryToUrl(url, ['pglive=1']);
        }
        return url;
    }

    this.getUrlWithoutPinegrowParameters = function(url) {
        if(!url) return url;
        url = url.replace(/[\&]*(pgedit|pglive|pgid|pgids|pgnoids|pgpvid)=[a-z0-9]+/g,'');
        url = url.replace(/\?$/, '');
        return url;
    }

    var getMimeType = function(file){
        file = crsaRemoveUrlParameters(file);

        //We need this because editable file types can be edited in Settings. For example, .whatever can be listed as editable. I moved the if down, to only use it in case of unrecognized types
        // if(pinegrow.isFileEditable(file)) {
        //     // return "text/html";
        // }

        //PGCore Handle file supported in PG
        //todo PEP find a way to handle support for csharp, vue and more
        var i = file.lastIndexOf("."),
            ext = (i === -1) ? "default" : file.substr(i),
            mimeTypes = {
                ".bmp": "image/bmp",
                ".css": "text/css",
                ".gif": "image/gif",
                ".ico": "image/x-icon",
                ".htm": "text/html",
                ".html": "text/html",
                ".php": "application/x-php",
                ".php5": "text/html",
                ".asp": "text/html",
                ".aspx": "text/html",
                ".cfm": "text/html",
                ".cfml": "text/html",
                ".cfc": "text/html",
                ".shtml": "text/html",
                ".jpg": "image/jpeg",
                ".jpeg": "image/jpeg",
                ".webp": "image/webp",
                ".js": "application/javascript",
                ".ts": "application/typescript",
                ".cshtml": "text/x-cshtml",
                ".razor": "text/x-cshtml",
                ".xml": "application/xml",
                ".pgml": "application/xml",
                ".json": "application/json",
                ".otf": "font/opentype",
                ".woff": "application/font-woff",
                ".woff2": "application/font-woff",
                ".ttf": "application/x-font-ttf",
                ".svg": "image/svg+xml",
                ".png": "image/png",
                ".text": "text/plain",
                ".txt": "text/plain",
                ".mp3": "audio/mpeg",
                ".mp4": "video/mp4",
                ".webm": "video/webm",
                ".ogv": "video/ogg",
                ".ogg": "audio/ogg",
                ".less": "text/x-less",
                ".scss": "text/x-scss",
                "default": null
                //"default": "application/octet-stream"
            };
        var mt = mimeTypes[ext.toLowerCase()] || mimeTypes['default'];

        //handle any unrecognized editable types as html
        if(!mt && pinegrow.isFileEditable(file)) {
            mt = "text/html";
        }
        return mt;
    }

    this.getMimeType = function(file) {
        return getMimeType(file);
    }

    this.testServer = function(done) {

        function handleErrors(response) {
            if (!response.ok) {
                throw Error(response.statusText);
            }
            return response;
        }

        function timeout(ms, promise) {
            return new Promise(function(resolve, reject) {
                setTimeout(function() {
                    reject(new Error("The request is taking too long."))
                }, ms)
                promise.then(resolve, reject)
            })
        }

        timeout(5000, fetch(this.url + '/pginternal/detect'))
            .then(handleErrors)
            .then(function(response) {
                return response.text();
            })
            .then(function(data) {
                if(data.indexOf('<h1>OK</h1>') >= 0) {
                    done('OK');
                } else {
                    done('ANOTHER_SERVER', '<p><b>Another server is already running</b> on network ports selected for Pinegrow (<b>' + config.port + ' to ' + (config.port + 10) + '</b>).</p><p>Stop the app that is using these ports or go to <b>Support -&gt; Settings</b> and change the value of the <b>Internal webserver port</b> setting.</p><p>Note: Pinegrow needs its own internal webserver during editing. Other webservers like Apache, Nginx etc will not do.</p>');
                }
            })
            .catch(function(error) {
                console.error(error)
                done('NOT_ACCESSIBLE', '<p><b>The internal webserver is not accessible at ' + this.url + '.</b> Error received: <code>' + (error || '') + '</code></p><p>The most likely reason is that your firewall or anti-virus software is blocking Pinegrow from accessing the network. In that case add a firewall rule that will allow Pinegrow to listen to the incoming network connections on port range <b>' + config.port + ' to ' + (config.port + 10) + '</b>.</p>');
            })
    }

    this.releaseRequestContextForPage = pgCurrentRequestContext.releaseRequestContextForPage;

    this.setCurrentRequestContext = pgCurrentRequestContext.setCurrentRequestContext;

    this.createProxyUrlNodeOutputFilter = pgCurrentRequestContext.createProxyUrlNodeOutputFilter;

    this.readRemoteUrl = function(fullPath, done) {
        var oReq = new XMLHttpRequest();
        oReq.open("GET", fullPath, true);
        oReq.responseType = "arraybuffer";

        oReq.onreadystatechange = function (oEvent) {
            if(oReq.readyState == 4) {
                if(oReq.status == 200) {
                    var arrayBuffer = oReq.response; // Note: not oReq.responseText
                    if (arrayBuffer) {
                        var buffer = new Buffer( new Uint8Array(arrayBuffer) );
                        var mime = oReq.getResponseHeader('content-type');
                        if(mime) {
                            mime = mime.split(';')[0];
                        }
                        done(null, buffer, mime);
                    }
                } else {
                    done("Error " + oReq.status + ": Not found or can't get remote file: " + fullPath);
                }
            }
        };

        oReq.onerror = function (oEvent) {
            done("Not found or can't get remote file: " + fullPath + "Error: " + oEvent.toString());
        }

        try {
            oReq.send(null);
        } catch(err) {
            done("Not found or can't get remote file: " + fullPath + "Error: " + err.toString());
        }
    }

    var lastSocketKey = 0;
    var socketMap = {};
    var server = null;

    this.destroy = function(done) {
        if(server) {
            console.log('Stopping internal server...');
            var p = new CrsaProfile('stopInternalServer');

            Object.keys(socketMap).forEach(function(socketKey){
                socketMap[socketKey].destroy();
            });

            server.close(function() {
                console.log('Internal server stopped.');
                p.show('Stopped');
                if(done) done();
            });
            server = null;
        } else {
            if(done) done();
        }
    }

    try {
        if(pgf.use_server) {
            server = libs.http.createServer(function (request, response) {

                var sendError = function (code, msg) {
                    response.writeHead(code, {
                        'content-type': "text/html"
                    });
                    response.write("<h1>" + code + " - " + escapeHtmlCode(msg) + "</h1>");
                    response.end();
                }


                var doRequest = function () {
                    var url = libs.url.parse(request.url, true),
                        path = null,
                        result = null,
                        fileCount = 0;

                    try {
                        path = decodeURIComponent(url.pathname);
                        result = request.method + " " + path.bold
                    } catch (err) {
                        sendError(500, 'Malformed url');
                        return;
                    }

                    //console.log('SERVER: ' + path);



                    // Default file to index.html
                    if (path === "/") path += "index.html";

                    var pathHandlers = {};

                    pathHandlers['/pginternal/detect'] = function (path, request, response) {
                        response.writeHead(200, {
                            'content-type': "text/html",
                            'Cache-Control': 'no-cache, no-store, must-revalidate',
                            'Pragma': 'no-cache',
                            'Expires': '0'
                        });
                        response.write("<h1>OK</h1>");
                        response.end();
                    }

                    if (path in pathHandlers) {
                        pathHandlers[path](path, request, response);
                        return;
                    }


                    var referer = null;
                    var referer_original = null;

                    if ('referer' in request.headers) {
                        referer = _this.getOriginalUrl(request.headers.referer);
                        referer_original = referer;
                        referer = referer.replace(/[\?\#].*/g, '');
                    }

                    if (path.match(/https?\:\/\/pinegrow\.com\/placeholders/)) {
                        var m = path.match(/img.*$/);
                        if (m) {
                            path = decodeURIComponent(crsaMakeUrlFromFile(crsaGetAppDir()) + 'placeholders/' + m[0]);
                        }
                    } else if (path.indexOf(pinegrow.thumbnailMaker.host) >= 0) {
                        path = decodeURIComponent(pinegrow.thumbnailMaker.getStoredThumbnailFilePath(path));

                    } else if(pgf.root_urls) {
                        var p = pinegrow.getCurrentProject();
                        if(p) {
                            if(path.startsWith('/') && path.indexOf('://') < 0) {
                                path = decodeURIComponent(p.getAbsoluteUrl(path));
                            }
                        }
                    }

                    //console.log('SERVER3: ' + path);

                    var is_remote = ((path.indexOf('://') >= 0 || path.indexOf('///') == 0) || (referer && referer.indexOf('://') >= 0 && referer.indexOf('file://') < 0)) && (path.indexOf('file://') < 0);

                    var fullPath;
                    var fullPathUrl;

                    if (is_remote) {
                        try {
                            fullPath = decodeURIComponent(url.path.substr(1));
                            //Sometimes this returns URI malformed exception
                        } catch (err) {
                            fullPath = url.path.substr(1);
                        }
                        fullPathUrl = _this.getUrlWithoutPinegrowParameters(fullPath);
                    } else {
                        //console.log('HTTP Server path: ' + path);
                        fullPath = libs.path.join(config.directory, path.replace('file://', ''));

                        while (fullPath.length > 1 && fullPath.charAt(0) == '\\') {
                            fullPath = fullPath.substr(1);
                        }
                        if (libs.path.sep == '\\') {
                            //win
                            if (!fullPath.match(/^[a-z]\:/i)) {
                                fullPath = '\\\\' + fullPath;
                            }
                        }
                        fullPathUrl = crsaIsFileUrl(fullPath) ? fullPath : crsaMakeUrlFromFile(fullPath);
                    }

                    //console.log('HTTP Server: ' + fullPath);


                    var handleWrapper = function (html_for_browser, rootNode, done) {

                        if (!crsaPage) {
                            done(html_for_browser);
                            //debugger;
                            return;
                        }

                        /*
                        var onHtmlReady = function (html) {
                            var wp = new pgParser();
                            wp.assignIds = false;
                            wp.replaceExistingIds = false;
                            wp.parse(html);
                            var htmlel = wp.rootNode.findOne(crsaPage.wrapper_selector || 'body');
                            if (htmlel) {
                                var remove = [];
                                for (var i = 0; i < htmlel.children.length; i++) {
                                    if (htmlel.children[i].tagName != 'script') remove.push(htmlel.children[i]);
                                }
                                for (var i = 0; i < remove.length; i++) {
                                    remove[i].remove();
                                }
                                htmlel.prepend(rootNode.clone(true, true));

                                htmlel.pgId = rootNode.getId();
                                htmlel.addClass('pg-partial-container');
                            }
                            var opts = pinegrow.getFormatHtmlOptions();
                            opts.assign_missing_ids = false;
                            html_for_browser = wp.rootNode.toStringWithIds(true, opts, pgCurrentRequestContext.createProxyUrlNodeOutputFilter);
                            done(html_for_browser);
                        }

                         */

                        if (crsaPage.wrapper_url) {

                            var node = rootNode.clone(true, true);
                            var node = crsaPage.getHTMLNodeForBrowserWithSupportForWrapper(node);

                            var html_options = {
                                indent: '    ',
                                php_ids: false,
                                assign_missing_ids: false,
                                remove_collapsed: false,
                                remove_hidden: false
                            };

                            html_for_browser = node.toStringWithIds(true, html_options, pgCurrentRequestContext.createProxyUrlNodeOutputFilter);

                            done(html_for_browser);
                            /*
                            var wcp = pinegrow.getCrsaPageByUrl(crsaPage.wrapper_url);
                            if (wcp) {
                                var html = wcp.sourceNode.toStringOriginal(true, pinegrow.getFormatHtmlOptions());
                                onHtmlReady(html);

                            } else if (crsaIsFileUrl(crsaPage.wrapper_url)) {
                                var f = crsaMakeFileFromUrl(crsaPage.wrapper_url);
                                var html = libs.fs.readFileSync(f, {encoding: "utf8"});
                                onHtmlReady(html);
                            } else {
                                $.ajax({
                                    url: crsaPage.wrapper_url,
                                    data: null,
                                    dataType: 'text'
                                }).done(function (data) {
                                    onHtmlReady(data);
                                }).fail(function () {
                                    onHtmlReady('File ' + crsaPage.wrapper_url + ' not found.');
                                });
                            }

                             */

                        } else if (!rootNode.findIncludingSelf('body', true)) {
                            html_for_browser = '<body data-pg-id="' + rootNode.getId() + '">' + html_for_browser + '</body>';
                            done(html_for_browser);
                        } else {
                            done(html_for_browser);
                        }
                    }

                    var onContentRead = function (data, serverMime) {
                        var mime = serverMime ? serverMime : getMimeType(fullPath);

                        //force text/html for all editable files, otherwise strange things happen with php in iframe, for example
                        if(pinegrow.isFileEditable(fullPath)) {
                            mime = "text/html";
                        }

                        if (!mime) {
                            var project = pinegrow.getCurrentProject();
                            if (project && project.isFileInProject(fullPath)) {
                                //allow all files in project dir
                                mime = 'application/octet-stream';
                            }
                        }
                        //debugger;
                        //response.setTimeout(5000);

                        if (mime) {
                            if ((['image/gif', 'image/jpeg', 'image/png', 'application/octet-stream'].indexOf(mime) >= 0 || mime.startsWith('image')) && (is_remote)) {

                                response.writeHead(200, {
                                    'content-type': mime,
                                    'Cache-Control': 'max-age=60000, public',
                                    'Content-Length': Buffer.byteLength(data)
                                });
                            } else {
                                response.writeHead(200, {
                                    'content-type': mime,
                                    'Cache-Control': 'no-cache, no-store, must-revalidate',
                                    'Pragma': 'no-cache',
                                    'Expires': '0',
                                    'Content-Length': Buffer.byteLength(data)
                                });
                            }
                            /*
                             var resp_socket = response.socket;

                             response.socket.on('drain', function() {
                             resp_socket.destroy();
                             })*/
                            response.write(data);
                            response.end();
                        } else {
                            sendError(403, "Type not allowed");
                        }
                    }

                    var pgid = null;
                    var pglive = false;
                    var pglib = url.query.pgpagelib && url.query.pgpagelib == '1';
                    var pgpvid = null;

                    var crsaPage = pinegrow.getCrsaPageByUrl(fullPathUrl);

                    if (!crsaPage || pglib) {
                        crsaPage = pinegrow.pageLibManager.get(crsaRemoveUrlParameters(fullPathUrl));
                    }

                    var pgnoids = false;

                    if (url.query.pglive) {
                        pglive = true;
                    }
                    if (url.query.pgnoids && url.query.pgnoids == '1') {
                        pgnoids = true;
                    }

                    if (url.query.pgpvid) {
                        pgpvid = parseInt(url.query.pgpvid);
                    }

                    if(request.headers['user-agent'] && request.headers['user-agent'].indexOf('PinegrowInternal') < 0) {
                        //not internal iframe
                        pglive = true;
                        pgnoids = true;
                    }

                    if(request.headers['user-agent'] && request.headers['user-agent'].indexOf('PinegrowPageLib') >= 0) {
                        pglib = true;
                    }

                    var pageView = null;

                    if (crsaPage) {

                        if (pgpvid) {
                            pageView = crsaPage.getPageViewById(pgpvid);
                        }

                        crsaPage.whenParsed(
                            function (page, statusFunc) {

                                try {

                                    var onData = function (data) {
                                        var p = new pgParser();
                                        p.replaceExistingIds = true;

                                        var profile = new CrsaProfile(true);

                                        var original_html = data.toString('utf8');

                                        var project = pinegrow.getCurrentProject();
                                        if (project) {
                                            var bck_page = project.getBackgroundPageForUrl(fullPathUrl);
                                            if (bck_page) {
                                                original_html = bck_page.getCode();
                                            }
                                        }
                                        //console.log(original_html);
                                        p.parse(original_html);
                                        page.sourceNode = p.rootNode;

                                        //we need this handler to map urls in add new page in projects
                                        page.onSourceParsed && page.onSourceParsed(page);

                                        page.handleWrapperOnPageLoad(statusFunc);
                                        //statusFunc('done');

                                        profile.show('HTML parse');
                                    }

                                    if (!is_remote) {
                                        if (crsaPage.original_new_url) {
                                            fullPath = crsaMakeFileFromUrl(crsaPage.original_new_url)
                                        }
                                        var data = libs.fs.readFileSync(fullPath);
                                        onData(data);
                                    } else {

                                        var mappedFile = pinegrow.getMappedRemoteFileForUrl(fullPath, 'html');

                                        if (mappedFile && crsaIsFileExist(mappedFile)) {
                                            var data = libs.fs.readFileSync(mappedFile);
                                            onData(data);
                                        } else {

                                            if(crsaPage.forceReload) {
                                                fullPath = crsaAppendQueryToUrl(fullPath, [`pgz=${(new Date()).getTime()}`]);
                                            }

                                            _this.readRemoteUrl(fullPath, function (error, buffer, mime) {
                                                if (error) {
                                                    statusFunc('error', error);
                                                } else {
                                                    onData(buffer);
                                                }
                                            })
                                        }
                                    }

                                } catch (error) {
                                    console.log('File ' + fullPath + ' can not be read by Pinegrow: ' + error);
                                    statusFunc('error', error);
                                }
                            },
                            function (page) {

                                page.callGlobalFrameworkHandler('on_page_parsed_in_proxy', page.sourceNode, page.url, pglive);
                                var original_js = page.javascriptEnabled;
                                if (pglive) {
                                    page.javascriptEnabled = true;
                                }
                                pgCurrentRequestContext.setCurrentRequestContext(fullPathUrl, page.sourceNode, pglive /* false */, page, pageView);

                                var html_for_browser;

                                if (!pgnoids || page.forcePgIds) {
                                    html_for_browser = page.sourceNode.toStringWithIds(true, null, pgCurrentRequestContext.createProxyUrlNodeOutputFilter);
                                } else {
                                    html_for_browser = page.sourceNode.toStringOriginal(true, null, pgCurrentRequestContext.createProxyUrlNodeOutputFilter);
                                }

                                page.javascriptEnabled = original_js;

                                handleWrapper(html_for_browser, page.sourceNode, function (html_for_browser) {

                                    var data = new Buffer(html_for_browser);

                                    response.writeHead(200, {
                                        'content-type': 'text/html',
                                        'Cache-Control': 'no-cache, no-store, must-revalidate',
                                        'Pragma': 'no-cache',
                                        'Expires': '0',
                                        "Content-Length": Buffer.byteLength(data)
                                    });

                                    response.write(data);
                                    response.end();
                                });
                            },
                            function (page, error) {
                                sendError(404, "File " + fullPath + " can not be read by Pinegrow: " + error);

                            }
                        )
                        return;
                    }

                    if (referer) {
                        var refurl = libs.url.parse(referer, true);

                        //refurl.href;

                        var refererPage = null;

                        var m = referer_original.match(/[&\?]pgpvid=([0-9]+)/);
                        if (m) {
                            pgpvid = parseInt(m[1]);
                            pageView = pinegrow.getPageViewById(pgpvid);
                            if (pageView) {
                                referedPage = pageView.getPage();
                            }
                        }

                        if (referer.match(/[&\?]pglive=1/)) pglive = true;

                        //hmmm... can we have stylesheets without css in the url?
                        if (fullPathUrl.match(/\.css($|[^a-z])/g)) {
                            //var path = is_remote ? fullPath : 'file://' + fullPath;

                            //var hrefPath = encodeURI(path);
                            var cslist = pinegrow.stylesheets.findCrsaStylesheetForUrl(fullPathUrl, false, true);
                            //console.log(cslist.length + ' stylesheets found');

                            var cs = cslist.length > 0 ? cslist[0] : null;

                            if (cs && cs.loaded) {
                                cs.getCssSource(function (css) {
                                    css = css || '';
                                    css = pgAddProxyLinksToStyleAttribute(css);
                                    if (pageView) {
                                        css = pageView.emulateCSSFeatures(css);
                                    }
                                    var data = new Buffer(css);
                                    onContentRead(data, 'text/css');
                                    /*
                                     response.writeHead(200, {
                                     'content-type': 'text/css'
                                     })
                                     response.write(data);
                                     response.end();*/
                                }, cs.ignore !== true);
                                return;
                            }
                        }
                    }

/*
                    if (crsaPage && pgid && crsaPage.wrapper_url) {
                        fullPath = crsaMakeFileFromUrl(crsaPage.url);
                        is_remote = false;
                    }*/

                    if (is_remote) {

                        if ('referer' in request.headers) {
                            var a = _this.getOriginalUrl(request.headers.referer).replace(/^[a-z]+:\/\//i, '').split(/[//\?\#]/);
                            if (a.length) {
                                var host = a[0];
                                //console.log('Referer host: ' + host);
                                if (fullPath.indexOf('//') === 0) {
                                    var a = document.createElement('a');
                                    a.href = request.headers.referer;
                                    fullPath = a.protocol + fullPath;
                                } else if (fullPath.indexOf('://') < 0) {
                                    //fullPath = fullPath.replace('://', '://' + host + '/');
                                    fullPath = 'http://' + host + '/' + fullPath;
                                    //console.log('Path fixed: ' + fullPath);
                                }
                            }
                        } else {
                            if (fullPath.indexOf('//') === 0) {
                                fullPath = 'http:' + fullPath;
                            }
                        }

                        var mapped = pinegrow.getMappedRemoteFileForUrl(fullPath, null);

                        if (mapped && crsaIsFileExist(mapped)) {
                            libs.fs.readFile(mapped, function (error, data) {
                                if (!error) {
                                    onContentRead(data);
                                } else {
                                    sendError(404, "Not found - File " + mapped + " can not be read by Pinegrow: " + error);
                                    console.log('File ' + mapped + ' can not be read by Pinegrow: ' + error);
                                }
                            });
                            return;
                        }

                        if (crsaPage) {
                            var data = crsaPage.getAsset(fullPath);
                            if (data !== null) {
                                if (data.data === false) {
                                    sendError(404, "Not found or can't get remote file");
                                } else {
                                    onContentRead(data.data, data.mime);
                                }
                                //console.log('Asset found in cache: ' + fullPath);
                                return;
                            }
                        }

                        _this.readRemoteUrl(fullPath, function (error, buffer, mime) {
                            if (error) {
                                sendError(404, error);
                                if (crsaPage) {
                                    crsaPage.addAsset(fullPath, false, mime);
                                }
                            } else {
                                if (crsaPage) {
                                    crsaPage.addAsset(fullPath, buffer, mime);
                                }
                                onContentRead(buffer, mime);
                            }
                        })
                        return;
                    } else {
                        libs.fs.readFile(fullPath, function (error, data) {
                            //console.log('READ ' + fullPath);
                            if (!error) {
                                onContentRead(data);
                                //console.log('SENT ' + fullPath);
                            } else {
                                sendError(404, "Not found - File " + fullPath + " can not be read by Pinegrow: " + error);
                                console.log('File ' + fullPath + ' can not be read by Pinegrow: ' + error);
                            }
                        });
                    }
                }

                if (request.client.localAddress == request.client.remoteAddress) {
                    //local connection, ok
                    doRequest();
                } else {
                    //remote
                    var addr = request.client.remoteAddress;

                    if (block_remote[addr]) {
                        sendError(403, "Connection not allowed");
                        pinegrow.showQuickMessage('Connection from <b>' + addr + '</b> was blocked. Restart Pinegrow to flush blocked list.', 4000, true);
                        return;
                    }
                    if (allow_remote[addr]) {
                        //allowed, ok
                        doRequest();
                    } else {
                        //ask
                        pinegrow.showAlert(`<p>Should Pinegrow accept network connections from <b>${addr}</b>?</p><p>Local connections are always allowed. Restart Pinegrow to flush the allowed and the blocked lists.</p>`, 'Allow network connection?', 'Block', 'Allow', function () {
                            //block
                            block_remote[addr] = true;
                            sendError(403, "Connection not allowed");
                        }, function () {
                            //allow
                            allow_remote[addr] = true;
                            doRequest();
                        });
                    }
                }
            });

            var allow_remote = {};
            var block_remote = {};

            var port_idx = 0;

            server.keepAliveTimeout = 0;
            server.maxConnections = 100;

            server.on('listening', function (e) {
                _this.testServer(function (status, details) {
                    if (status_callback) {
                        status_callback(status, details, _this);
                    }
                })
                pinegrow.dispatchEvent('on_server_created');
            });

            server.on('error', function (e) {
                console.log('Could not start internal http server.');

                if (port_idx < 20) {
                    port_idx += 2;

                    _this.url = 'http://' + $.trim(config.host) + ':' + (config.port + port_idx);
                    _this.port = config.port + port_idx;
                    server.listen(config.port + port_idx);
                } else {
                    //pinegrow.alert('Could not start the internal webserver on port ' + config.port + '. Please make sure that the port is available and that connections are not blocked by a firewall. Use Support -> Settings to set the port.');

                    if (status_callback) {
                        status_callback('NOT_STARTED', '<p><b>Could not start the internal webserver</b> on any of the ports in range <b>' + config.port + ' to ' + (config.port + 10) + '</b>.</p><p>The most likely reasons are:</p><ul><li> Another app is using these ports. Go to <b>Support -&gt; Settings</b> and change the value of the <b>Internal webserver port</b> setting.</li><li>Your firewall or anti-virus software is blocking Pinegrow from using these ports. In that case add a firewall rule that will allow Pinegrow to use the network port.</li></ul>', _this);
                    }
                }
            });


            server.on('connection', function (socket) {
                var socketKey = ++lastSocketKey;
                socketMap[socketKey] = socket;
                socket.on('close', function () {
                    delete socketMap[socketKey];
                });
            });

            _this.port = config.port + port_idx;
            server.listen(config.port + port_idx);
        }
    }
    catch(err) {

    }
}





var PgApiServer = function(port) {

    if(!pgf.editor_sync) return;

    var io = require('socket.io')(port, {
        allowEIO3: true,
        //pingInterval: 3000
       // transports: ["websocket"]
    });

    this.addEndPoint = function(endpoint, on_connection) {
        return io
            .of('/' + endpoint)
            .on('connection', function (socket) {
                on_connection(socket);
            })
            .on('error', function(error) {
                console.log('Socket error ' + error);
            })
            .on("connect_error", function(err) {
                console.log(`connect_error due to ${err.message}`);
            })
            .on('disconnect', function(reason, details) {
                console.log(details.context.status);
                console.log(details.context.responseText);
            })
    }

    this.addEndPoint('core', function(socket) {
        var name = "Pinegrow";
        var project = pinegrow.getCurrentProject();
        if(project) name = project.getName();
        socket.emit('introduceInstance', {name: name, port: port});
    })
}
