(function( $ ) {

    $.fn.crsastorage = function( method ) {

        if ( methods[method] ) {
            return methods[method].apply( this, Array.prototype.slice.call( arguments, 1 ));
        } else if ( typeof method === 'object' || ! method ) {
            return methods.init.apply( this, arguments );
        } else {
            $.error( 'Method ' +  method + ' does not exist on jQuery.crsastorage' );
        }
    };


    var options;
    var pages;
    var client;
    var projects = [];
    var load_count = 0;

    var methods = {
        init : function( opts ) {
            options = $.extend( {}, $.fn.crsastorage.defaults, opts );
        },
        auth : function() {
            client = new Dropbox.Client({ key: "iz3y3vk51f3xtza" });
            client.authDriver(new Dropbox.AuthDriver.Popup({receiverUrl: "http://localhost/dropbox.html"}));

            if (client.isAuthenticated()) {
                methods.authDone();
            } else {

                client.authenticate(function(error, client) {
                    if (error) {
                        showError(error);
                        return;
                    }
                    methods.authDone();
                });
            }
        },
        authDone : function() {
            loadProjects();
        },
        createProject : function(name) {
            client.mkdir("/" + name, function(error, stat) {
                if (error) {
                    showError(error);
                    return;
                }
                projects.push(stat);
                client.makeUrl(stat.path, {long : true}, function(error, share) {
                    if (error) {
                        showError(error);
                        return;
                    }
                    console.log(share);
                });
            });
        },
        openProject : function(name) {
            load_count = 0;
            var project = null;

            var onDone = function() {
                methods.openProjectBrowser(project);
            }

            client.readFile("/" + name + "/pinegrow.json", {}, function(error, str, stat, range) {
                if(!error && str) {
                    project = JSON.parse(str);
                }
                if(!project) {
                    project = {'crsa_project_name' : name, 'crsa_files' : []};
                    loadProjectFolder("/" + name, project, function() {

                        saveProjectCache(project);
                        alert('load finished');
                        console.log(project);

                        onDone();
                    });
                } else {
                    console.log('Project loaded from cache');
                    console.log(project);

                    onDone();
                }

            });

        },
        openProjectBrowser : function(p) {
            if(typeof p == 'undefined' || !p) p = project;
            var $b = $('<div/>', {'class' : 'crsa-project-browser'}).appendTo($('body'));
            var $title = $('<h2/>').html(p.crsa_project_name).appendTo($b);
            var $tree = $('<div/>', {'class' : 'crsa-project-tree'}).appendTo($b);
            var $grid = $('<div/>', {'class' : 'crsa-project-grid'}).appendTo($b);

            var makeTreeLevel = function(node, $dest) {
                var $li = $('<li/>').html('<div>' + (node.crsa_project_name ? node.crsa_project_name : node.name) + '</div>').appendTo($dest);
                $li.data('crsa-node', node);
                if(node.crsa_files) {
                    var $ul = $('<ul/>').appendTo($li);
                    $.each(node.crsa_files, function(i,sn) {
                        if(sn.isFolder) {
                            makeTreeLevel(sn, $ul);
                        }
                    });
                }
            }

            var showGrid = function(node) {
                $grid.html('');
                if(!node.crsa_files) return;
                $.each(node.crsa_files, function(i,s) {
                    var $e = $('<div/>').html(s.name);
                });
            }

            var selectedNode = null;

            var $root = $('<ul/>').appendTo($tree);
            makeTreeLevel(p, $root);

            $root.find('li').on('click', function(event) {
                var $li = $(event.delegateTarget);
                var node = $li.data('crsa-node');
                showGrid(node);
                if(selectedNode) selectedNode.removeClass('active');
                $li.addClass('active');
                selectedNode = $li;
                return false;
            });

            return $b;
        }
    };

    var saveProjectCache = function(project) {
        var pf = null;
        $.each(project.crsa_files, function(i,s) {
            if(s.name == 'pinegrow.json') {
                pf = s;
                return true;
            }
        });

        var name = project.crsa_project_name;

        client.writeFile("/" + name + "/pinegrow.json", JSON.stringify(project), {lastVersionTag : (pf ? pf.versionTag : null)}, function(error, stat) {
            if(error) {
                showError(error);
            } else {
                console.log("Project cache saved");
            }
        });
    }

    var getPublicUrl = function(path, func, onDone) {
        load_count++;
        client.makeUrl(path, {downloadHack : true}, function(error, share) {
            if (error) {
                //showError(error);
                load_count--;
                if(onDone && load_count <= 0) onDone();
                return;
            }
            func(share);
            load_count--;
            if(onDone && load_count <= 0) onDone();
        });
    }

    var loadProjectFolder = function(path, node, onDone) {
        load_count++;
        client.readdir(path, {}, function(error, names, currentStat, stats) {
            if(error) {

            } else {
                $.each(stats, function(i,s) {
                    node.crsa_files.push(s);

                    if(s.isFolder) {
                        s.crsa_files = [];
                        loadProjectFolder(s.path, s, onDone);
                    }
                    if(s.isFile && !s.crsa_share) {
                        getPublicUrl(s.path, function(share) {
                            s.crsa_share = share;
                        }, onDone);
                    }
                });
            }
            load_count--;

            if(load_count <= 0) {
                onDone();
            }
        });
    }

    var loadProjects = function() {
        projects = [];
        client.readdir("/", {}, function(error, names, currentStat, stats) {
            if (error) {
                showError(error);
                return;
            }
            console.log(stats);
            $.each(stats, function(i,s) {
                if(s.isFolder) {
                    var p = s;
                    projects.push(p);
                    console.log('Project ' + p.name);
                }
            });
        });
    }


    var showError = function(error) {
        alert(error);
        switch (error.status) {
            case Dropbox.ApiError.INVALID_TOKEN:
                // If you're using dropbox.js, the only cause behind this error is that
                // the user token expired.
                // Get the user through the authentication flow again.
                break;

            case Dropbox.ApiError.NOT_FOUND:
                // The file or folder you tried to access is not in the user's Dropbox.
                // Handling this error is specific to your application.
                break;

            case Dropbox.ApiError.OVER_QUOTA:
                // The user is over their Dropbox quota.
                // Tell them their Dropbox is full. Refreshing the page won't help.
                break;

            case Dropbox.ApiError.RATE_LIMITED:
                // Too many API requests. Tell the user to try again later.
                // Long-term, optimize your code to use fewer API calls.
                break;

            case Dropbox.ApiError.NETWORK_ERROR:
                // An error occurred at the XMLHttpRequest layer.
                // Most likely, the user's network connection is down.
                // API calls will not succeed until the user gets back online.
                break;

            case Dropbox.ApiError.INVALID_PARAM:
            case Dropbox.ApiError.OAUTH_ERROR:
            case Dropbox.ApiError.INVALID_METHOD:
            default:
            // Caused by a bug in dropbox.js, in your application, or in Dropbox.
            // Tell the user an error occurred, ask them to refresh the page.
        }
    }


    var updatePagesList = function($el) {
        return $el.each(function(i,e){
            pages = $(e).find('div.page');
        });
    }

    // Plugin defaults – added as a property on our plugin function.
    $.fn.crsastorage.defaults = {
    };



})( jQuery );
