/**
 * Created by Matjaz on 11/9/13.
 */

var CrsaBrowserStorage = function() {

    var getKey = function(url) {
        return 'localSave_' + url;
    }

    this.save = function(url, data) {
        var key = getKey(url);
        localStorage[key] = data;
    }

    this.exists = function(url) {
        return getKey(url) in localStorage;
    }

    this.load = function(url) {
        var key = getKey(url);
        return (key in localStorage) ? localStorage[key] : null;
    }

    this.saveProject = function() {
        var d = {pages: [], css: []};
        $.each($.fn.crsapages('getAllPages'), function(i, cp) {
            d.pages.push({url: cp.url, html: cp.getSource(false), name: cp.name});
            cp.setPageChanged(false);
        });
        if(d.pages.length == 0) return null;
        $.each(pinegrow.stylesheets.getAllCrsaStylesheets(), function(i, cs) {
            if(!cs.inline && cs.loaded) {
                d.css.push({url: cs.url, less: cs.getSource()});
                cs.setChanged(false);
            }
        });
        var str = JSON.stringify(d);
        if(str.length > 2000000) {
            pinegrow.alert('Sorry, the project is too big to save. Use File -> Save.');
        } else {
            localStorage.savedProject = str;
        }
        return d;
    }

    this.hasSavedProject = function() {
        return 'savedProject' in localStorage;
    }

    this.restoreProject = function(done) {
        if(!this.hasSavedProject()) return null;

        try {
            var json = localStorage.savedProject;
            var d = JSON.parse(json);

            var idx = 0;

            var cssidx = 0;

            var loadCss = function() {
                if(cssidx >= d.css.length) {
                    //done
                    $('body').trigger('crsa-stylesheets-changed');
                    if(done) done();
                } else {
                    var css = d.css[cssidx];
                    var cslist = pinegrow.stylesheets.findCrsaStylesheetForUrl(css.url);
                    var cs;
                    if(cslist && cslist.length > 0) {
                        cs = cslist[0];
                        cs.genSetSource(css.less, function() {
                            cssidx++;
                            loadCss();
                        });
                    } else {
                        $.fn.crsacss('addStylesheetFromUrl', css.url, function(cs) {
                            cs.genSetSource(css.less, function() {
                                cssidx++;
                                loadCss();
                            });
                        });
                    }
                }
            }

            var loadPage = function() {
                var page = d.pages[idx];
                $.fn.crsa("openPage", page.url, function() {

                }, function(crsaPage) {

                    setPageSource(crsaPage.$iframe, page.html, null, true);

                    crsaPage.runScripts(function() {
                        crsaPage.callFrameworkHandler('on_page_loaded');
                        $.fn.crsa('updateIfNeeded');
                        crsaPage.addCrsaStyles();
                    });

                    if(page.name) {
                        crsaPage.name = page.name;
                        crsaPage.updateNameDisplay();
                    }
                }, null, function(crsaPage) {
                    //done
                    idx++;
                    if(idx < d.pages.length) {
                        loadPage();
                    } else {
                        //load csss
                        loadCss();
                    }
                });
            }

            if(d.pages.length > 0) {
                loadPage();
            }

            return d;
        }
        catch(err) {
            return null;
        }
    }
}





var CrsaFileTag = function(text, type, desc, icon, cls) {
    this.text = text;
    this.type = type;
    this.desc = desc;
    this.icon = icon;
    this.class = cls;
    this.show = true;
}

var CrsaFile = function (entityMapObject) {
    this.isFile = true;
    this.name = null;
    this.path = null;
    this.url = null;
    this.children = [];
    this.entity = null;
    this.image = null;
    this.required = false;
    this.tag = null;
    this.tags = [];
    this.parent = null;

    this.entityMap = entityMapObject;
}

CrsaFile.prototype.addChild = function(node) {
    this.children.push(node);
    node.parent = this;
}

CrsaFile.prototype.withParents = function(func) {
    var p = this.parent;
    while(p) {
        func(p);
        p = p.parent;
    }
}

CrsaFile.prototype.fromEntity = function (entity) {
    this.isFile = entity.isFile;
    this.name = entity.name;
    this.path = entity.fullPath;
    if (this.entityMap) this.entityMap[this.path] = entity;
    return this;
}

CrsaFile.prototype.fromFile = function (filepath) {
    var fs = require('fs');
    var path = require('path');
    var stat = fs.statSync(filepath);

    this.isFile = stat.isFile();
    this.name = path.basename(filepath);
    this.url = crsaMakeUrlFromFile(filepath);
    this.path = filepath;
}

CrsaFile.prototype.fromURL = function (url) {
    this.isFile = true;
    this.name = crsaGetNameFromUrl(url);
    this.url = url;
}

CrsaFile.prototype.getEntity = function () {
    if (this.entityMap && this.path && this.path in this.entityMap) return this.entityMap[this.path];
    return null;
}

CrsaFile.prototype.walkThroughFiles = function (func) {
    for (var i = 0; i < this.children.length; i++) {
        var n = this.children[i];
        func(n);
        n.walkThroughFiles(func);
    }
}

CrsaFile.prototype.getMappedUrl = function() {
    if(this.content && this.content.startsWith('@mapurl:')) {
        return this.content.replace('@mapurl:', '').trim();
    }
    return null;
}

CrsaFile.prototype.getMappedUrlOrUrl = function() {
    return this.getMappedUrl() || this.url;
}

CrsaFile.prototype.addTag = function (tagObj, index) {
    for (var i = 0; i < this.tags.length; i++) {
        if (this.tags[i].text == tagObj.text) return;
    }
    if(index === undefined) {
        this.tags.push(tagObj);
    } else {
        if(index > this.tags.length) {
            index = this.tags.length;
        }
        this.tags.splice(index, 0, tagObj)
    }
}

CrsaFile.prototype.hasTag = function (tag) {
    for (var i = 0; i < this.tags.length; i++) {
        if (this.tags[i].text == tag) return true;
    }
    return false;
}

CrsaFile.prototype.getTagByType = function (type) {
    for (var i = 0; i < this.tags.length; i++) {
        if (this.tags[i].type === type) return this.tags[i];
    }
    return null;
}

CrsaFile.prototype.removeTag = function (tag) {
    for (var i = 0; i < this.tags.length; i++) {
        if (this.tags[i].text == tag) {
            this.tags.splice(i, 1);
            return;
        }
    }
}

CrsaFile.prototype.removeTagByType = function (type) {
    for (var i = 0; i < this.tags.length; i++) {
        if (this.tags[i].type === type) {
            this.tags.splice(i, 1);
            return;
        }
    }
}

CrsaFile.prototype.removeTags = function () {
    this.tags = [];
}

CrsaFile.prototype.updateTags = function() {
    var p = pinegrow.getCurrentProject();
    if(p) {
        p.updateTagsForFile(this);
    }
}

CrsaFile.prototype.copyFile = function (currentFilePath, newFilePath, done) {
    var path = require('path');
    var fs = require('fs');

    var _this = this;

    if (crsaIsRemoteUrl(this.url)) {
        //copy remote file
        pgCopyRemoteUrlToLocalPath(this.url, newFilePath, function() {
            var new_file = new CrsaFile();
            new_file.fromFile(newFilePath);
            new_file.parent = _this.parent;
            done && done(new_file);
        });
    } else {
        crsaCopyFileSync(fs, currentFilePath, newFilePath);
        var new_file = new CrsaFile();
        new_file.fromFile(newFilePath);
        new_file.parent = _this.parent;
        done && done(new_file);
        return new_file;
    }
}

CrsaFile.prototype.duplicate = function (new_name, overwrite) {
    var path = require('path');
    var fs = require('fs');
    var new_path = path.join(path.dirname(this.path), new_name);
    var exists = false;
    try {
        fs.statSync(new_path);
        exists = true;
    } catch (err) {
    }
    if (exists && !overwrite) {
        throw "File " + new_name + " already exists.";
    }
    return this.copyFile(this.path, new_path);
}

CrsaFile.prototype.copyFileToNewPath = function (new_path, done) {
    var fs = require('fs');

    var exists = false;
    try {
        fs.statSync(new_path);
        exists = true;
    } catch (err) {
    }
    if (exists) {
        throw "File " + new_path + " already exists.";
    }
    return this.copyFile(this.path, new_path, done);
}

CrsaFile.prototype.deleteFileAndFolder = function (currentPath) {
    var path = require('path');
    var fs = require('fs');

    var _this = this;

    if (fs.lstatSync(currentPath).isDirectory()) {
        fs.readdirSync(currentPath).forEach(function (currnetFile, index) {
            var curPath = path.join(currentPath, currnetFile);
            if (fs.lstatSync(curPath).isDirectory()) { // recurse
                _this.deleteFileAndFolder(curPath);
            } else { // delete file
                fs.unlinkSync(curPath);
            }
        });
        fs.rmdirSync(currentPath);
    }
    else {
        fs.unlinkSync(currentPath);
    }
}

CrsaFile.prototype.delete = function () {
    try {
        this.deleteFileAndFolder(this.path);
    }
    catch (err) {
        return false;
    }
    return true;
}

CrsaFile.prototype.rename = function (new_name) {
    var path = require('path');
    var fs = require('fs');

    var new_path = path.join(path.dirname(this.path), new_name);

    fs.renameSync(this.path, new_path);

    if (fs.lstatSync(new_path).isDirectory()) return null;

    var new_file = new CrsaFile();
    new_file.fromFile(new_path);
    new_file.parent = this.parent;
    return new_file;
}

CrsaFile.prototype.doesFileExistInSameDir = function (file_name) {
    if(isApp()) {
        var path = require('path');
        var fs = require('fs');

        var file_path = path.join(path.dirname(this.path), file_name);
        return fs.existsSync(file_path);
    } else {
        var url = crsaGetBaseForUrl(this.path) + '/' + file_name;
        var p = pinegrow.getCurrentProject();
        return p.getFileForUrl(url);
    }
}

CrsaFile.prototype.isDirectoryInTheSameDir = function (file_path) {
    if(!isApp()) return false;

    var path = require('path');
    var fs = require('fs');

    var new_path = path.join(path.dirname(this.path), file_path);
    return fs.lstatSync(new_path).isDirectory();
}

CrsaFile.prototype.isImage = function() {
    var mimeType = pinegrow.getMimeType(this.name);
    return mimeType && mimeType.startsWith('image/');
}

CrsaFile.prototype.getSimpleType = function() {
    return pinegrow.getSimpleMimeType(this.name);
}

CrsaFile.prototype.addScriptToPage = function(page, footer) {
    var url = page.makeRelativeUrl(this.url);
    page.addScript(url, footer);
}

CrsaFile.prototype.hasScriptOnPage = function(page) {
    var url = page.makeRelativeUrl(this.url);
    if(page.hasScriptInHeaderOrFooter(url)) return 'head';
    if(page.hasScriptInHeaderOrFooter(url, true)) return 'body';
    return false;
}

CrsaFile.prototype.addStylesheetToPage = function(page) {
    var url = page.makeRelativeUrl(this.url);
    page.addStylesheet(url, true);
}

CrsaFile.prototype.hasStylesheetOnPage = function(page) {
    var url = page.makeRelativeUrl(this.url);
    return page.hasLinkedStylesheet(url);
}

var CrsaStore = function() {
    this.type = "AbstractStore"

    this.listProjects = function(search) {

    }

    this.getProject = function(pid, done) {

    }

    this.createProjectFromTemplate = function(template) {

    }

    this.saveProject = function(project) {

    }

    this.deleteProject = function(pid) {

    }

    this.createProjectFile = function(project, path, data) {

    }

    this.saveProjectFile = function(project, path, data) {

    }

    this.deleteProjectFile = function(project, path) {

    }

    this.copyProjectFile = function(project, path_source, path_dest) {

    }

    this.getProjectFileUrl = function(project, path) {

    }
};

var CrsaFileStore = function() {
    this.name = "FileStore";
    this.app = location.protocol == 'chrome-extension:';

    var _this = this;

    this.listProjects = function(search) {

    }

    this.getProject = function(pid, done) {

        chrome.fileSystem.chooseEntry({type: 'openDirectory'}, function(dir) {
            console.log(dir);
            _this.loadProject(dir, done);
        });
    }

    this.loadProject = function(dir, done) {
        var project = new CrsaProject();
        project.root = (new CrsaFile(project.entityMap)).fromEntity(dir);

        chrome.fileSystem.getDisplayPath(dir, function(path) {

            project.name = project.root.name;
            project.localPath = path;

            var dirReader = dir.createReader();

            var readingLevel = 0;
            // Call the reader.readEntries() until no more results are returned.
            var readEntries = function(dirReader, node, done) {
                readingLevel++;
                dirReader.readEntries (function(results) {
                    readingLevel--;
                    if (!results.length) {
                        if(readingLevel <= 0) done();
                    } else {
                        for(var n = 0; n < results.length; n++) {
                            var subNode = new CrsaFile(project.entityMap);
                            subNode.fromEntity(results[n]);

                            node.children.push(subNode);

                            if(!subNode.isFile) {
                                var subdirReader = results[n].createReader();
                                readEntries(subdirReader, subNode, done);
                            }
                        }
                        readEntries(dirReader, node, done);
                    }
                }, function(err) {
                    readingLevel--;
                    console.log(err);
                    if(readingLevel <= 0) done();
                });
            };

            readEntries(dirReader, project.root, function() {
                done(project);
            });

        });
        /*   chrome.fileSystem.getWritableEntry(dir, function(writable_dir) {
         console.log("writable dir " + writable_dir);
         });*/
    }

    this.saveProject = function(project) {

    }

    this.createProjectFromTemplate = function(template, done) {
        //read template json
        var url = chrome.runtime.getURL("/templates/" + template + "/pinegrow.json");
        console.log(url);

        $.ajax({
            url: url,
            data: null,
            dataType: 'json'
        })
            .done(function(data) {
                console.log(data);

                chrome.fileSystem.chooseEntry({type: 'openDirectory'}, function(dir) {
                    chrome.fileSystem.isWritableEntry(dir, function(writable) {
                        if(writable) {

                            var level = 0;

                            var decreaseLevel = function() {
                                level--;
                                if(level <= 0) {
                                    _this.loadProject(dir, function(p) {
                                        done(p);
                                    });
                                }
                            }

                            var copyTemplateNodeToDest = function(data, dir, done) {
                                console.log('copying dir ' + data.name);

                                for(var i = 0; i < data.children.length; i++) {
                                    var node = data.children[i];
                                    console.log(node.path);

                                    if(node.isFile) {
                                        level++;
                                        var url = chrome.runtime.getURL("/templates/" + template + node.path.replace("/" + template, ""));
                                        var xhr = new XMLHttpRequest();
                                        xhr.open('GET', url, true);
                                        xhr.responseType = 'blob';
                                        xhr.crsaNode = node;
                                        xhr.onload = function(e) {

                                            var node = this.crsaNode;
                                            var blob = this.response;

                                            dir.getFile(node.name, {create: true}, function(fileEntry) {

                                                fileEntry.createWriter(function(fileWriter) {

                                                    fileWriter.onwriteend = function(e) {
                                                        console.log('Write completed.');
                                                        decreaseLevel();
                                                    };

                                                    fileWriter.onerror = function(e) {
                                                        console.log('Write failed: ' + e.toString());
                                                        decreaseLevel();
                                                    };
                                                    fileWriter.write(blob);

                                                }, function(err) {
                                                    decreaseLevel();
                                                    console.log(err);
                                                });

                                            }, function(err) {
                                                decreaseLevel();
                                                console.log(err);
                                            })
                                        };

                                        xhr.send();

                                    } else {
                                        level++;

                                        var func = function(node) {
                                            dir.getDirectory(node.name, {create: true}, function(dirEntry) {

                                                copyTemplateNodeToDest(node, dirEntry, done);
                                                decreaseLevel();

                                            }, function(err) {
                                                decreaseLevel();
                                                console(err);
                                            });
                                        }
                                        func(node);
                                    }
                                }
                            }

                            copyTemplateNodeToDest(data.root, dir, function() {
                                console.log("template copied to project.")
                            });

                        } else {
                            console.log(dir + ' is not writable.');
                        }
                    });
                    console.log(dir);
                });
            })
            .fail(function(err) {
                console.log(err);
            });


        //get writable dir
        //check if dir is empty
        //copy files


    }

    this.deleteProject = function(pid) {

    }

    this.createProjectFile = function(project, path, data) {

    }

    this.saveProjectFile = function(project, path, data) {

    }

    this.deleteProjectFile = function(project, path) {

    }

    this.copyProjectFile = function(project, path_source, path_dest) {

    }

    this.getProjectFileUrl = function(project, path) {

    }
}

CrsaFileStore.prototype = new CrsaStore();
