var PgEmulatePageForUndo = function() {
    var _this = this;

    this.url = null;
    this.sourceNode = null;
    this.emulate_undo = true;
    this.name = 'Emulated page';

    this.getName = function() {
        return this.name;
    }

    this.destroy = function() {
        if(this.sourceNode) {
            this.sourceNode.withoutEvents(function () {
                _this.sourceNode.remove();
            })
        }
    }

}



var PgUndoCSSHelper = function() {

}
PgUndoCSSHelper.prototype.cssNodeToString = function(node) {
    return node.toString();
}

PgUndoCSSHelper.prototype.cssGetId = function(node) {
    return node['pg-node-id'] || null;
}

PgUndoCSSHelper.prototype.cssGetPosition = function(node) {
    var p = {};
    if(node.parent) {
        p.parentId = this.cssGetId(node.parent);
    }
    p.index = node.parent.nodes.indexOf(node);
    p.indices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
    return p;
}

var PgUndoManager = function() {

    var use_combined = pinegrow.getSetting('combined-undo', '1') == '1';

    if(!use_combined) {
        pinegrow.setSetting('combined-undo', '1');
        pinegrow.showAlert(`<p>Please note that <b>Separate undo histories for pages and CSS</b> are no longer supported. We had to remove the feature in order to be able to improve the Undo/Redo experience.</p><p>The combined Undo history will be used from now on.</p>`, 'A note about the Undo History', null, 'OK');
    }

    use_combined = true;

    var _this = this;

    var event_handler = new PgEventHandler();

    var css_active = false;

    var combinedHistory = new PgUndoHistory();

    this.setUseCombined = function(val) {
        combinedHistory.destroy();
        combinedHistory = new PgUndoHistory();

        cssHistory.destroy();
        cssHistory = new PgUndoHistory();
        cssHistory.css = true;

        pinegrow.getAllPages().forEach(function(page) {
            page.undoHistory.destroy();
            page.undoHistory = new PgUndoHistory(page);
        })
        use_combined = val;
        pinegrow.setSetting('combined-undo', val ? '1' : '0');
        pinegrow.dispatchEvent('on_undo_state_added', null);
    }

    this.getUseCombined = function() {
        return use_combined;
    }

    pinegrow.addEventHandler('on_undo_state_added', function(page) {
        css_active = page == null;
    })

    this.setCssActive = function(a) {
        css_active = a;
    }

    this.getCssActive = function() {
        return css_active;
    }

    this.getActiveHistory = function() {
        if(use_combined) return combinedHistory;

        var page = pinegrow.getSelectedPage();
        if(css_active || !page) {
            return cssHistory;
        }
        return page.undoHistory;
    }

    this.getHistoryForEvent = function(event) {
        if(use_combined) return combinedHistory;
        return event.page.undoHistory;
    }

    this.getHistoryForPage = function(page) {
        return use_combined ? combinedHistory : page.undoHistory;
    }

    this.undo = function() {
        return this.getActiveHistory().undo();
    }

    this.redo = function() {
        return this.getActiveHistory().redo();
    }

    event_handler.handle_dom_events = false;//not interested in dom-only events like executeScript

    event_handler.on_event = function(event) {
//        console.log('Undo event', event);

        var h = _this.getHistoryForEvent(event);

        if(!h.isUndoOperationInProgress() && !h.isIgnore()) {
            var undoEvent = new PgUndoEvent(event)
            if(undoEvent.operation) {
                h.addEvent(undoEvent);
            }
        }
    }

    var onPageChanged = function(page, info) {
        var h = _this.getHistoryForPage(page);

        if(!h.isUndoOperationInProgress() && !h.isIgnore()) {
            h.changeDone(info.action || 'change', info.undoCombineKey || null, info.undoSourceId || null, info.undoData || null);
        }
    }

    pinegrow.addEventHandler('on_page_changes_done', onPageChanged)

    pinegrow.addEventHandler('on_emulated_page_changed', onPageChanged)

    //CSS Events

    var cssHistory = new PgUndoHistory();
    cssHistory.css = true;

    this.getCssHistory = function() {
        return use_combined ? combinedHistory : cssHistory;
    }

    var helper = new PgUndoCSSHelper();

    var cs_sources = new WeakMap();
    var cs_timers = new WeakMap();

    var purge_done_timer = null;

    this.purgeCs = function(cs) {
        cs_sources.delete(cs);
        if(cs_timers.has(cs)) {
            clearTimeout(cs_timers.get(cs));
            cs_timers.delete(cs);
        }
        this.getCssHistory().purge(function(state) {
            if(state.events.length && state.events[0].event.cs && state.events[0].event.cs === cs) {
                return true;
            }
            return false;
        })
        if(purge_done_timer) {
            clearTimeout(purge_done_timer);
        }
        purge_done_timer = setTimeout(function() {
            pinegrow.dispatchEvent('on_undo_state_added', null)
            purge_done_timer = null;
        }, 500)
    }

    this.purgePage = function(page) {
        if(!use_combined) return;

        this.getHistoryForPage(page).purge(function(state) {
            if(state.events.length && state.events[0].event.page && state.events[0].event.page === page) {
                return true;
            }
            return false;
        })
        if(purge_done_timer) {
            clearTimeout(purge_done_timer);
        }
        purge_done_timer = setTimeout(function() {
            pinegrow.dispatchEvent('on_undo_state_added', null)
            purge_done_timer = null;
        }, 500)
    }

    var rememberCSSSource = function(cs) {
        if(cs.getRootNode() && !cs.elementStyle) {
            var p = new CrsaProfile();
            var new_source = cs.getSource();
            cs_sources.set(cs, new_source);
            p.show('rememberCSSSource');
        }
    }

    var handleCSSCodeChange = function(cs, data) {
        var p = new CrsaProfile();

        var new_source = cs.getSource();

        if(data && data.editor instanceof PgInternalCodeEditor) {
            new_source = data.editor.getCode();
        }

        if(cs_sources.has(cs)) {
            var old_source = cs_sources.get(cs);

            if(old_source != new_source) {

                if(cs.inline && cs.inlineSheetPgel) {
                    //inline cs changes should go to page history
                    var page = cs.inlineSheetPgel.getPage();
                    if(page) {
                        var h = _this.getHistoryForPage(page);

                        var pgEvent = new PgNodeEvent(cs.inlineSheetPgel, 'updateInner', null, {
                            domValue: new_source,
                            oldValue: old_source
                        }, page);
                        h.addEvent(new PgUndoEvent(pgEvent));
                        h.changeDone(data.action || 'Style update', data.undoCombineKey || null, data.undoSourceId || null, data.undoData || null, data.undoRuleName || null)
                    }
                } else {
                    var h = _this.getCssHistory();

                    var event = new PgCSSUndoEvent({
                        cs: cs,
                        operation: 'codeUpdated',
                        oldValue: old_source,
                        newValue: new_source
                    })
                    h.addEvent(event);
                    h.changeDone(data.action || 'Style update', data.undoCombineKey || null, data.undoSourceId || null, data.undoData || null, data.undoRuleName || null)
                }
                dispatchGeneralCSSChangedEvent(cs);
            } else {
                //pinegrow.showQuickMessage('No change.');
            }
        }
        cs_sources.set(cs, new_source);

        cs_timers.delete(cs);

        p.show('handleCSSCodeChange');
    }

    pinegrow.addEventHandler('on_undo_redo', function(page, info) {
        if(info.cs_col) {
            //was a css op, have to remember changed sources
            info.cs_col.forEachElement(function(cs) {
                rememberCSSSource(cs);
            })
        }
    })

    var dispatchGeneralCSSChangedEvent = function(cs) {
        pinegrow.dispatchEvent('on_css_stylesheet_changed', null, cs);
    }

    var onCSSChanged = function(cs, data) {
        var h = _this.getCssHistory();

        if(h.isUndoOperationInProgress() || h.isIgnore() || (data && data.undoIgnore)) return;

        if(cs.elementStyle) return;

        if(false && data && data.node && data.node.type == 'decl' && data.oldValue && !cs.inline) {

            //prop is changed
            var oldValue = data.oldValue

            var same = true;

            $.each(oldValue, function(key, val) {
                if(data.node[key] != val) {
                    same = false;
                    return false;
                }
            })

            if(!same) {
                var cs = data.stylesheet;
                var event = new PgCSSUndoEvent({
                    cs: cs,
                    operation: 'nodeUpdated',
                    oldValue: oldValue,
                    position: helper.cssGetPosition(data.node)
                })
                h.addEvent(event);
                h.changeDone(data.action || 'Style update', data.undoCombineKey || null, data.undoSourceId || null, data.undoData || null, data.undoRuleName || null)

                rememberCSSSource(cs);

                dispatchGeneralCSSChangedEvent(cs);
            }

        } else {
            //remember the whole cs
            handleCSSCodeChange(cs, data)
        }

    }

    pinegrow.addEventHandler('on_css_source_changed', function (p, info) {
        if(info && info.list) {
            if (info.edited_by !== 'undo' && !info.ignore_in_undo) {
                info.list.forEach(function(cs) {
                    if(!info.action) info.action = 'Edit';
                    onCSSChanged(cs, info);
                })
            }
        }
    });

    pinegrow.addEventHandler('on_css_loaded', function(pagenull, cs) {
        rememberCSSSource(cs);
    });

    pinegrow.addEventHandler('on_css_node_added', function(pagenull, data) {
        if(!data.action) data.action = 'Added';
        if(!data.undoRuleName) data.undoRuleName = pgcssh.getNodeDisplayName(data.node);
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_css_node_deleted', function(pagenull, data) {
        if(!data.action) data.action = 'Deleted';
        if(!data.undoRuleName) data.undoRuleName = pgcssh.getNodeDisplayName(data.node);
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_css_node_deleted_finished', function(pagenull, data) {
        if(!data.action) data.action = 'Deleted';
        if(!data.undoRuleName) data.undoRuleName = pgcssh.getNodeDisplayName(data.node);
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_css_node_updated', function(pagenull, data) {
        /*if(!data.action)*/ data.action = 'Set';
        if(!data.undoRuleName) data.undoRuleName = pgcssh.getNodeDisplayName(data.node);
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_css_updated', function(pagenull, data) {
        /*if(!data.action)*/ data.action = 'Set';
        if(!data.undoRuleName) data.undoRuleName = "Style";
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_stylesheet_closed', function (pagenull, cs) {
        if (cs_sources.has(cs)) cs_sources.delete(cs);
        if (cs_timers.has(cs)) cs_timers.delete(cs);

        _this.getCssHistory().purge(function (state) {
            if (state.events.length && state.events[0].event.cs && state.events[0].event.cs == cs) {
                return true;
            }
            return false;
        });
    })

    //End css events

    event_handler.start();
}

var PgUndoHistory = function(page) {

    var states = [];

    var idx = -1;
    var _this = this;

    var maxStates = 40;

    this.startState = null;

    var col = null;

    var selectIds = null;
    var executedState = null;

    var currentState = null;
    var previousState = null;

    this.css = false;

    this.getCurrentStateId = function(undo) {
        if(idx < 0) return -1;
        if(undo) {
            return states[idx].stateId;
        } else {
            if(idx + 1 < states.length) {
                return states[idx + 1].stateId;
            }
        }
        return -1;
    }

    var getCurrentState = function(event) {
        if(!currentState) {
            if(previousState) {
                //should combine states?
            }
            currentState = new PgUndoState(page);
        }
        return currentState;
    }

    this.purge = function(func) {
        var cidx = idx;
        var r = [];
        for(var i = 0; i < states.length; i++) {
            if(func(states[i])) {
                if(idx >= i) {
                    cidx--;
                }
            } else {
                r.push(states[i]);
            }
        }
        if(cidx < 0) {
            cidx = -1;
            currentState = null;
        } else {
            idx = cidx;
            if(currentState) {
                currentState = states[idx];
            }
        }
        states = r;
    }

    this.getInfo = function(cssHistory) {
        var combined = [];
        /*
        if(cssHistory) {
            //combine css states with page states
            var css = cssHistory.getInfo();

            var css_i = 0;
            var page_i = 0;

            while(true) {
                var end_css = css_i >= css.states.length;
                var end_page = page_i >= states.length;

                if(end_css && end_page) break;

                if(!end_css && (end_page || css.states[css_i].stateId < states[page_i].stateId)) {
                    combined.push(css.states[css_i]);
                    if(css.current_idx < css_i) {
                        css.states[css_i].alreadyRestored = true;
                    } else {
                        css.states[css_i].alreadyRestored = false;
                    }
                    css_i++;
                } else {
                    combined.push(states[page_i]);
                    page_i++;
                }
            }
        }*/
        return {
            combined: combined,
            states: states,
            current_idx: idx,
            current_state: (idx >= 0 && idx < states.length) ? states[idx] : null,
            start_state: this.startState
        }
    }

    this.addEvent = function(event) {
        var state = getCurrentState(event);
        if(event instanceof PgUndoEvent) {
            if(!state.element_names) {
                var pgel = event.getElement();
                if(pgel) {
                    state.element_names = pgel.getName({short: true});
                }
            }
            if(!state.page_name) {
                state.page_name = event.event.page.getName();
            }
        } else {
            if(!state.cs_name) {
                state.cs_name = event.event.cs.name;
            }
        }
        state.add(event);
    }

    this.changeDone = function(name, combineKey, sourceId, undoData, elementName) {
        getCurrentState().name = name;

        //ignore changes without events - LEAVING IT IN because it's a good way to detect wrong page change events errors
        if(currentState.events.length || true) {

            if(pgIsDev() && currentState.events.length > 1000) {
                console.error('Undo stack is huge.', currentState);
                debugger;
            }

            currentState.combineKey = combineKey;
            currentState.sourceId = sourceId;
            currentState.undoData = undoData;

            if (elementName) {
                currentState.element_names = elementName;
            }

            if (currentState.selectedElementsAtEnd === null) {
                currentState.selectedElementsAtEnd = pinegrow.selectedElements.getIds();
            }

            var combined = false;

            if (idx < states.length - 1) {
                if(idx < 0) {
                    states = [];
                } else {
                    states.splice(idx, states.length - idx);
                }
            } else if (idx >= 0 && idx < states.length) {
                var last = states[idx];
                if ((last.combineKey && combineKey && last.combineKey == combineKey) && last.sourceId == sourceId) {
                    if (new Date().getTime() - last.lastChanged < 2000) {
                        currentState.events.forEach(function (e) {
                            last.add(e);
                        })
                        last.selectedElementsAtEnd = currentState.selectedElementsAtEnd;
                        combined = true;
                    }
                }
            }

            if (!combined) {
                states.push(currentState);

                if (states.length > maxStates) {
                    this.startState = states[states.length - maxStates - 1];
                    this.startState.events = [];

                    states.splice(0, states.length - maxStates);
                    changed = true;
                }

                idx = states.length - 1;
                pinegrow.dispatchEvent('on_undo_state_added', page, currentState);

            }
        }
        currentState = null;
    }

    this.setSelectedElementsAtEnd = function(ids) {
        if(currentState) {
            currentState.selectedElementsAtEnd = ids;
        }
    }

    this.setSelectedElementsAtStart = function(ids) {
        if(currentState) {
            currentState.selectedElementsAtStart = ids;
        }
    }

    var undoOp = function(col, done) {
        if(idx >= 0 && states.length > idx) {
            executedState = states[idx];
            var ci = idx;
            selectIds = states[ci].selectedElementsAtStart;
            idx--;
            states[ci].undo(col, done);
        } else {
            pinegrow.showQuickMessage('Nothing to undo.');
            if(done) done(true);
        }
    }

    var redoOp = function(col, done) {
        if(idx < states.length - 1) {
            idx++;
            executedState = states[idx];
            selectIds = states[idx].selectedElementsAtEnd;
            states[idx].redo(col, done);
        } else {
            pinegrow.showQuickMessage('Nothing to redo.');
            if(done) done(true);
        }
    }

    this.undo = function() {
        col = new PgCollection();
        this.undoOrRedo = 'undo';
        this.start();

        try {
            undoOp(col, this.done);
        } catch(err) {
            this.done(true);
            pinegrow.showQuickMessage('Oops... Error during undo.');
            console.error('Error during undo', err);

            if(pgIsDev() && pgTestMode) throw err;
        }
    }

    this.redo = function() {
        col = new PgCollection();
        this.undoOrRedo = 'redo';
        this.start();

        try {
            redoOp(col, this.done);
        } catch(err) {
            this.done(true);
            pinegrow.showQuickMessage('Oops... Error during redo.');
            console.error('Error during redo', err);
        }
    }

    this.restoreState = function(state, done) {
        var state_idx = -1;
        if(state) {
            state_idx = states.indexOf(state);
            if (state_idx < 0) {
                pinegrow.showQuickMessage('Invalid undo state');
                return;
            }
        }
        if (state_idx == idx) return;

        col = new PgCollection();
        this.start();

        var op = state_idx < idx ? undoOp : redoOp;

        var doNext = function() {
            if(state_idx != idx) {
                op(col, doNext);
            } else {
                _this.done();
                if(done) done();
            }
        }

        doNext();
    }

    var undoOperationInProgress = false;

    this.start = function() {
        selectIds = null;
        undoOperationInProgress = true;
    }

    this.done = function(nothing) {
        if(!nothing) {
            var info = {source: this, sourceType: 'undo', undoOrRedo: _this.undoOrRedo};

            var pages = []
            var changed_page = null;

            if (executedState) {
                info.undoSourceId = executedState.sourceId;
                info.undoData = executedState.undoData;

                executedState.events.forEach(function(undoEvent) {
                    if(undoEvent.event && undoEvent.event.page) {
                        if(pages.indexOf(undoEvent.event.page) < 0) {
                            pages.push(undoEvent.event.page);

                            changed_page = undoEvent.event.page;
                        }
                    }
                })
            }

            var pgel_col = new PgCollection();
            var css_col = new PgCollection();

            col.forEachElement(function(e) {
                if(e instanceof pgParserNode) {
                    pgel_col.add(e);
                } else {
                    css_col.add(e.cs);
                }
            })

            var affected_pages = pages.slice(0);

            pgel_col.forEachPage(function(page, col) {
                pinegrow.changeManager.didChange(col, 'Undo / redo op', info, page);
                var idx = pages.indexOf(page);
                if(idx >= 0) {
                    pages.splice(idx, 1);
                }
            })

            pages.forEach(function(page) {
                pinegrow.changeManager.didChangePage(page, 'Undo / redo op', info);
            })

            if(css_col.getLength()) {
                info.cs_col = css_col;
            }

            info.pages = affected_pages;

            info.pgel_col = pgel_col;

            pinegrow.dispatchEvent('on_undo_redo', page ? page : changed_page, info);

            css_col.forEachElement(function (cs) {
                pinegrow.stylesheets.applyToPages(cs);
            })

            if(selectIds) {
                pinegrow.selectedElements.selectIds(selectIds, info);
                selectIds = null;
            } else {
                pinegrow.selectedElements.repositionMenus();
            }

            /*
            if (page) {
                pinegrow.changeManager.didChange(col, 'Undo / redo op', info, page);
                pinegrow.dispatchEvent('on_undo_redo', executedState.page, info);
            } else {
                //CSS
                var cs_col = new PgCollection();

                col.forEachElement(function (event) {
                    cs_col.add(event.cs);
                })

                info.cs_col = cs_col;

                pinegrow.dispatchEvent('on_undo_redo', null, info);

                cs_col.forEachElement(function (cs) {
                    pinegrow.stylesheets.applyToPages(cs);
                })
                pinegrow.selectedElements.repositionMenus();
            }
            if(page && selectIds) {
                pinegrow.selectedElements.selectIds(selectIds, info);
                selectIds = null;
            }*/
        }
        executedState = null;
        col = null;
        undoOperationInProgress = false;
    }

    this.isUndoOperationInProgress = function() {
        return undoOperationInProgress;
    }

    var ignore = false;

    this.isIgnore = function() {
        return ignore;
    }

    this.ignore = function(func) {
        var o = ignore;
        ignore = true;
        func();
        ignore = o;
    }

    this.destroy = function() {
        states = [];
        idx = -1;
        col = null;
        selectIds = null;
        executedState = null;
        currentState = null;
        previousState = null;
    }
}

var pgUndoStateId = 0;

var PgUndoState = function(page) {
    this.element_names = null;
    this.name = null;
    this.page = page;
    this.events = [];
    this.selectedElementsAtStart = pinegrow.selectedElements.getIds();
    this.selectedElementsAtEnd = null;
    this.lastChanged = null;
    this.combineKey = null;
    this.sourceId = null;
    this.undoData = null;
    this.stateId = pgUndoStateId++;
}

PgUndoState.prototype.add = function(event) {
    if(this.events.length && this.events[this.events.length-1].combine && this.events[this.events.length-1].combine(event)) {
        //combined
    } else {
        this.events.push(event);
    }
    this.lastChanged = new Date().getTime();
}

PgUndoState.prototype.undo = function(col, done) {
    var i = this.events.length;
    var _this = this;

    /*
    console.log('State info');
    _this.events.forEach(function(e) {
        console.log(e);
    })*/

    var doNext = function() {
        i--;
        if(i >= 0) {
            _this.events[i].undo(col, doNext);
        } else {
            done();
        }
    }

    doNext();
}

PgUndoState.prototype.redo = function(col, done) {
    var i = -1;
    var _this = this;

    var doNext = function() {
        i++;
        if(i < _this.events.length) {
            _this.events[i].redo(col, doNext);
        } else {
            done();
        }
    }

    doNext();
}



var PgCSSUndoEvent = function(event) {

    this.event = event;
    this.data = {};
    this.operation = this.operations[event.operation] || null;

    if(this.operation) {

    } else {
        console.log('UNSUPPORTED UNDO OP: ' + event.operation);
    }
}

PgCSSUndoEvent.prototype.getElement = function(indices) {
    indices = indices || (this.event.position ? this.event.position.indices : null);
    return indices ? pinegrow.pgPostCSSHelper.getNodeByIndices(this.event.cs.getRootNode(), indices) : null;
}

PgCSSUndoEvent.prototype.isSameRule = function(event) {
    if(!this.event.position || !event.event.position) return false;
    var a = this.event.position.indices;
    var b = event.event.position.indices;
    if(a.length != b.length) return false;
    for(var i = 0; i < a.length; i++) {
        if(a[i] != b[i]) return false;
    }
    return true;
}


PgCSSUndoEvent.prototype.combine = function(event) {
    if(this.operation == event.operation && this.event.cs === event.event.cs /* && this.isSameRule(event) */) {
        this.event.newValue = event.event.newValue;
        //this.oldValue = event.oldValue;
        return true;
    } else {
        return false;
    }
}

PgCSSUndoEvent.prototype.undo = function(col, done) {
    try {
        var node = this.getElement();
        if (col) col.add(this.event);
        this.operation.undo.call(this, node, done);
    } catch(err) {
        console.log('CSS Undo error ' + err);
        done();
    }
}

PgCSSUndoEvent.prototype.redo = function(col, done) {
    try {
        var node = this.getElement();
        if (col) col.add(this.event);
        this.operation.redo.call(this, node, done);
    } catch(err) {
        console.log('CSS Redo error ' + err);
        done();
    }
}

PgCSSUndoEvent.prototype.operations = {

    nodeUpdated : {
        undo : function(node, done) {
            var _this = this;

            this.data.originalValue = {
                value: node.value,
                prop: node.prop
            }
            $.each(this.event.oldValue, function(key, val) {
                node[key] = val;
            })
            pinegrow.dispatchEvent('on_css_node_updated', null, {
                node: node,
                stylesheet: _this.event.cs,
                edited_by: 'undo',
                source_type: 'undo'
            })
            done();
        },
        redo : function(node, done) {
            var _this = this;

            $.each(this.data.originalValue, function(key, val) {
                node[key] = val;
            })
            pinegrow.dispatchEvent('on_css_node_updated', null, {
                node: node,
                stylesheet: _this.event.cs,
                edited_by: 'undo',
                source_type: 'undo'
            })
            done();
        }
    },
    codeUpdated : {
        undo : function(node, done) {
            var _this = this;
            this.event.cs.setSource(this.event.oldValue, function() {
                pinegrow.dispatchEvent('on_css_source_changed', null, {list: [_this.event.cs]});
                done();
            })
        },
        redo : function(node, done) {
            var _this = this;
            this.event.cs.setSource(this.event.newValue, function() {
                pinegrow.dispatchEvent('on_css_source_changed', null, {list: [_this.event.cs]});
                done();
            })
        }
    },
    ruleUpdated : {
        undo : function(node, done) {
            var _this = this;
            this.data.code = pinegrow.pgPostCSSHelper.getSource(node);
            this.event.cs.parseSource(this.event.oldValue, function (error, postcssSource, newRootNode) {
                if(!error) {
                    pinegrow.pgPostCSSHelper.replaceNodeWithAndUpdatePgId(node, newRootNode.nodes[0]);
                    pinegrow.dispatchEvent('on_css_node_updated', null, {
                        node: newRootNode.nodes[0],
                        stylesheet: _this.event.cs,
                        edited_by: 'undo',
                        source_type: 'undo'
                    })
                } else {
                    console.log('CSS Undo - ' + error);
                }
                done();
            });
        },
        redo : function(node, done) {
            var _this = this;
            this.event.cs.parseSource(this.data.code, function (error, postcssSource, newRootNode) {
                if(!error) {
                    pinegrow.pgPostCSSHelper.replaceNodeWithAndUpdatePgId(node, newRootNode.nodes[0]);
                    pinegrow.dispatchEvent('on_css_node_updated', null, {
                        node: newRootNode.nodes[0],
                        stylesheet: _this.event.cs,
                        edited_by: 'undo'
                    })
                } else {
                    console.log('CSS Undo - ' + error);
                }
                done();
            });
        }
    }
}



var PgUndoEvent = function(event) {

    this.event = event;
    this.data = {};
    this.operation = this.operations[event.operation] || null;

    if(this.operation) {
        //this.operation.init.call(this, this.getElement(), this.event.objectId ? this.getElement(this.event.objectId) : null);
    } else {
        //console.log('UNSUPPORTED UNDO OP: ' + event.operation);
    }
}

PgUndoEvent.prototype.getElement = function(id) {
    if(id === undefined) id = this.event.pgId;
    if(id) {
        return pgParserNodeCatalogueInstance.get(id);
    } else {
        if(this.event.text) {
            return pgCreateNodeFromHtml(this.event.text);
        }
    }
    return null;
}

PgUndoEvent.prototype.combine = function(event) {
    if(this.operation == event.operation && this.event.pgId === event.event.pgId && this.operation.combine && this.operation.combine.call(this, event)) {

        return true;
    } else {
        return false;
    }
}

PgUndoEvent.prototype.undo = function(col, done) {
    try {
        var pgel = this.getElement();
        if (pgel && col) col.add(pgel);
//        console.log('UNDO event', this.event);
        this.operation.undo.call(this, pgel, this.event.objectId ? this.getElement(this.event.objectId) : null, col);
    } catch(err) {
        console.log('Undo error ' + err);
        if(pgIsDev() && pgTestMode) throw err;
    }
    done();
}

PgUndoEvent.prototype.redo = function(col, done) {
    try {
        var pgel = this.getElement();
        if (pgel && col) col.add(pgel);
        this.operation.redo.call(this, pgel, this.event.objectId ? this.getElement(this.event.objectId) : null, col);
    } catch(err) {
        console.log('Redo error ' + err);
    }
    done();
}

PgUndoEvent.prototype.operations = {

    addClass : {
        undo : function(pgel) {
            //console.log(`UNDO removeClass ${this.event.data.class}`)
            pgel.removeClass(this.event.data.class);
        },
        redo : function(pgel) {
            pgel.addClass(this.event.data.class);
        }
    },
    removeClass : {
        undo : function(pgel) {
            //console.log(`UNDO addClass ${this.event.data.class}`)
            if(this.event.data.hadClass) {
                pgel.addClass(this.event.data.class);
            }
        },
        redo : function(pgel) {
            pgel.removeClass(this.event.data.class);
        }
    },
    html : {
        undo : function(pgel) {
            pgel.html(this.event.data.oldValue);
        },
        redo : function(pgel) {
            pgel.html(this.event.data.html);
        },
        combine : function(b) {
            this.event.data.html = b.event.data.html;
            return true;
        }
    },
    setAttr : {
        undo : function(pgel) {
            if(this.event.data.oldValue) {
                pgel.setAttr(this.event.data.oldValue.name, this.event.data.oldValue.value, null, this.event.data.oldValue.quote || null);
            } else {
                pgel.removeAttr(this.event.data.attr);
            }
        },
        redo : function(pgel) {
            pgel.setAttr(this.event.data.attr, this.event.data.value, null, this.event.data.quote);
        },
        combine : function(b) {
            if(this.event.data.attr == b.event.data.attr) {
                this.event.data.value = b.event.data.value;
                this.event.data.quote = b.event.data.quote;
                return true;
            }
            return false;
        }
    },
    removeAttr : {
        undo : function(pgel) {
            if(this.event.data.oldValue) {
                pgel.setAttr(this.event.data.oldValue.name, this.event.data.oldValue.value);
            } else {
                //didnt have it before
            }
        },
        redo : function(pgel) {
            pgel.removeAttr(this.event.data.attr);
        }
    },
    remove: {
        undo : function(pgel, obj, col) {
            var parent = getPgNodeByPgId(this.event.data.oldValue.parentId);
            var el = pgCreateNodeFromHtml(this.event.data.oldValue.html);
            el.insertAtIndex(parent, this.event.data.oldValue.position);
            if(col) col.add(el);
        },
        redo : function(pgel) {
            pgel.remove();
        }
    },
    /*
    insertBefore: {
        undo : function(pgel) {
            this.data.html = pgel.toStringWithIds();
            pgel.remove();
        },
        redo : function(pgel) {
            var el = pgCreateNodeFromHtml(this.data.html);
            el.insertBefore(this.getElement(this.event.objectId));
        }
    },
    insertAfter: {
        undo : function(pgel) {
            this.data.html = pgel.toStringWithIds();
            pgel.remove();
        },
        redo : function(pgel) {
            var el = pgCreateNodeFromHtml(this.data.html);
            el.insertAfter(this.getElement(this.event.objectId));
        }
    },
    */
    appendPrepend: {
        undo : function(pgel) {
            this.data.html = pgel.toStringWithIds();
            this.data.positionBeforeUndo = pgel.getPositionInfo();

            if(!pgel.isElement) {
                //pgel is text node, can't place it, so we just restore dest and parent html
                if(this.event.data.oldValue.parent) {
                    var parent = this.getElement(this.event.data.oldValue.parent);
                    this.data.parentHtmlBeforeUndo = parent.html(null, true);
                    parent.html(this.event.data.oldValue.parentHtml);
                }
                var dest = this.getElement(this.event.objectId);
                this.data.destHtmlBeforeUndo = dest.html(null, true);
                dest.html(this.event.data.oldValue.destHtml);
                pgel.remove();

            } else if(this.event.data.oldValue.position) {
                pgel.placeAtPositionInfo(this.event.data.oldValue.position);
            } else {
                pgel.remove();
            }
        },
        redo : function(pgel, object_pgel, col) {
            if(!pgel && this.data.html) {
                pgel = pgCreateNodeFromHtml(this.data.html);
                if(col) col.add(pgel);
            }
            if(!pgel.isElement) {
                //pgel is text node, can't place it
                if(this.event.data.oldValue.parent) {
                    var parent = this.getElement(this.event.data.oldValue.parent);
                    parent.html(this.data.parentHtmlBeforeUndo);
                }
                var dest = this.getElement(this.event.objectId);
                dest.html(this.data.destHtmlBeforeUndo);
            } else {
                if (this.event.data.oldValue.position) {

                } else {
                    //was removed on undo, create new
                    pgel = pgCreateNodeFromHtml(this.data.html);
                    if(col) col.add(pgel);
                }
                pgel.placeAtPositionInfo(this.data.positionBeforeUndo);
            }
            //pgel.appendPrepend(this.getElement(this.event.objectId), this.event.data.prepend);
        }
    },
    updateInner: {
        undo : function(pgel) {
            this.data.html = pgel.html(null, true);
            pgel.html(this.event.data.oldValue);

            if (pgel.tagName === "style") {
                var cs = pinegrow.stylesheets.findCrsaInlineStylesheet(pgel);
                if (cs) {
                    cs.setSource(this.event.data.oldValue, function () {
                        pinegrow.dispatchEvent('on_css_source_changed', null, {
                            list: [cs],
                            edited_by: 'undo'
                        });
                    })
                }
            }
        },
        redo : function(pgel) {
            pgel.html(this.data.html);

            if (pgel.tagName === "style") {
                var cs = pinegrow.stylesheets.findCrsaInlineStylesheet(pgel);
                if (cs) {
                    cs.setSource(this.data.html, function () {
                        pinegrow.dispatchEvent('on_css_source_changed', null, {
                            list: [cs],
                            edited_by: 'undo'
                        });
                    })
                }
            }
        }
    },
    update: {
        undo : function(pgel, object_pgel, col) {
            this.data.html = pgel.toStringWithIds();
            var el = pgCreateNodeFromHtml(this.event.data.oldValue);
            pgel.replaceWith(el);
            this.event.pgId = el.getId();
            if(col) col.add(el);
        },
        redo : function(pgel, object_pgel, col) {
            var el = pgCreateNodeFromHtml(this.data.html);
            pgel.replaceWith(el);
            this.event.pgId = el.getId();
            if(col) col.add(el);
        }
    },
    replaceWith: {
        undo : function(pgel, obj, col) {
            pgel = obj;
            this.data.html = pgel.toStringWithIds();
            var el = pgCreateNodeFromHtml(this.event.data.oldValue);
            pgel.replaceWith(el);
            this.event.pgId = el.getId();
            if(col) col.add(obj);
        },
        redo : function(pgel, obj, col) {
            //pgel = this.getElement(this.event.objectId);
            var el = pgCreateNodeFromHtml(this.data.html);
            pgel.replaceWith(el);
            this.event.pgId = el.getId();
            if(col) col.add(el);
        }
    },
    detag: {
        undo : function(pgel) {
            var parent = getPgNodeByPgId(this.event.data.oldValue.parentId);
            this.data.parentHtml = parent.html(null, true);
            parent.html(this.event.data.oldValue.parentHtml);
        },
        redo : function(pgel) {
            var parent = getPgNodeByPgId(this.event.data.oldValue.parentId);
            parent.html(this.data.parentHtml);
        }
    },
    setCode: {
        undo : function(pgel) {
            var _this = this;
            this.data.html = this.event.page.sourceNode.toStringWithIds(true);
            var p = new pgParser();
            p.assignIds = true;
            p.parse(this.event.data.oldValue);
            if(this.event.page.sourceNode) {
                this.event.page.sourceNode.withoutEvents(function() {
                    _this.event.page.sourceNode.remove();
                })
            }
            this.event.page.sourceNode = p.rootNode;
            this.event.page.refresh(null, true /* silent_multi_refresh */);
        },
        redo : function(pgel) {
            var _this = this;
            var p = new pgParser();
            p.assignIds = true;
            p.parse(this.data.html);
            if(this.event.page.sourceNode) {
                this.event.page.sourceNode.withoutEvents(function() {
                    _this.event.page.sourceNode.remove();
                })
            }
            this.event.page.sourceNode = p.rootNode;
            this.event.page.refresh(null, true /* silent_multi_refresh */);
        },
        combine : function(b) {
            //this.event.data.oldValue = b.event.data.oldValue;
            return true;
        }
    },
}

PgUndoEvent.prototype.operations.insertBefore = PgUndoEvent.prototype.operations.insertAfter = PgUndoEvent.prototype.operations.appendPrepend;


var CrsaUndoState = function(page, name, force_cs) {

    // if (force_cs.compiledStylesheet) force_cs = force_cs.compiledStylesheet;
    var csUrl = (force_cs ? force_cs.url : "");

    var data = [];

    data.push({obj: page, data: page.sourceNode.toStringWithIds(true, pinegrow.getFormatHtmlOptions())});

    var ps = getCrsaPageStylesForPage(page.$iframe);

    if (force_cs) data.push({obj: force_cs, data: force_cs.getSource()});

    // for(var i = 0; i < ps.stylesheets.length; i++) {
    //     var cs = ps.stylesheets[i];
    //     if(cs.inline) continue;
    //     // if(!cs.loaded || !cs.changed) continue;
    //     if(!cs.loaded) continue;

    //     var found = false;
    //     if (!pinegrow.stylesheets.isStylesheetUrlEqual(csUrl, cs.url)) {
    //         for (var j = 0; j < cs.importedFiles.length; j++) {
    //             if (pinegrow.stylesheets.isStylesheetUrlEqual(csUrl, cs.importedFiles[j].url)) {
    //                 cs = cs.importedFiles[j];
    //                 found = true;
    //                 break;
    //             }
    //         }
    //     }
    //     else {
    //         found = true;
    //     }

    //     if (found) {
    //         if (cs.sourceStylesheet) cs = cs.sourceStylesheet;
    //         data.push({obj: cs, data: cs.getSource()});
    //         break;
    //     }
    // }

    data.push({obj: 'breakpoints', data: page.breakpoints.slice(0)});

    this.page = page;
    this.data = data;
    this.name = name;
    this.active = true;

    //console.log('UNDO RECORDED ' + name);

    var _this = this;

    this.apply = function(done) {
        var c = 0;
        var page_s = null;
        for(var i = 0; i < this.data.length; i++) {
            var s = this.data[i];

            if(s.obj === 'breakpoints') {
                page_s.obj.breakpoints = s.data;

            } else if(s.obj instanceof CrsaPage) {
                page_s = s;
            } else if(s.obj instanceof PgStylesheet) {
                c++;
                s.obj.setSource(s.data, function() {
                    c--;
                    if(c == 0) {
                        var parser = new pgParser();
                        parser.parse(page_s.data, function() {
                            page_s.obj.sourceNode = parser.rootNode;
                            if(done) done(_this);
                        });
                       /*
                        page_s.obj.setSource(page_s.data, function() {
                            if(done) done(_this);
                        });*/
                    }
                });
            }
        }
        if(c == 0 && page_s) {
            var parser = new pgParser();
            parser.parse(page_s.data, function() {
                page_s.obj.sourceNode = parser.rootNode;
                if(done) done(_this);
            });
            /*
            page_s.obj.setSource(page_s.data, function() {
                if(done) done(_this);
            });*/
        }
    }
}

var CrsaPageUndoStack = function(page) {

    var stack = [];
    var pointer = -1;
    var last = 'undo';
    this.page = page;


    var getLastPgStylesheet = function () {
        if (stack.length == 0) return null;

        var data = stack[stack.length - 1].data;

        for (var i = 0; i < data.length; i++) {
            if (data[i].obj instanceof PgStylesheet) return data[i].obj;
        }

        return null;
    }

    this.add = function(name, no_pointer, force_cs) {

        if (no_pointer && !force_cs) {
            force_cs = getLastPgStylesheet();
        }

        var d = new CrsaUndoState(this.page, name, force_cs);

        if(stack.length > pointer + 1) {
            stack.splice(pointer + 1, stack.length - (pointer + 1));
        }
        stack.push(d);
        if(!no_pointer) {
            pointer++;

            var max = 20;
            if(pointer > max) {
                var diff = pointer - max;
                stack.splice(0, diff);
                pointer -= diff;
                //console.log('Cleared ' + diff + ' from undo stack');
            }
        }
        return d;
    }

    this.isAtTheTip = function() {
        return pointer >= 0 && stack.length == pointer + 1;
    }

    this.remove = function() {
        if(pointer < 0) {
            return null;
        }
        pointer--;
        stack.splice(pointer + 1, 1);
    }

    this.undo = function(done) {
        var x = last == 'undo' ? 0 : 1;
        if(pointer < 0 + x) {
            if(done) done(null);
            return null;
        }
        if(last != 'undo') pointer -= 1;
        var d = stack[pointer];
        pointer -= 1;
        d.apply(done);
        last = 'undo';
        return d;
    }

    this.redo = function(done) {
        var x = last == 'undo' ? 2 : 1;
        if(pointer + x >= stack.length) {
            if(done) done(null);
            return null;
        }
        pointer += x;
        var d = stack[pointer];
        d.apply(done);
        last = 'redo';
        return d;
    }
}

var CrsaUndo = function() {

}
