var PgActiveRules = function (for_pgel) {
    var selectedElement = for_pgel || null;

    // Active nodes
    var active_nodes = null;

    var cssStates = ['active', 'hover', 'focus', 'visited'];

    var $body = $('body');

    var _this = this;

    this.matched_rules = null;
    this.show_all_rules = false;//pgf.kids || false;

    this.setShowAllRules = function(v) {
        this.show_all_rules = v;
        selectedElement = null;//will trigger rematch
    }

    var convertStatesToClass = function (selector) {
        var states = ['hover', 'active', 'focus', 'visited'];

        var newSelector = selector;
        for (var i = 0; i < states.length; i++) {
            newSelector = newSelector.replace(new RegExp(':' + states[i], 'g'), '.pg-state-' + states[i]);
        }

        return newSelector;
    }

    var canShowNodeBySelector = function (selector, $el) {

        var found = true;
        var selArr = selector.split(',');
        for (var i = 0; i < selArr.length; i++) {
            var sel = selArr[i].trim();

            sel = sel.replace(/&/g, '*');
            if (sel[0] == '>') {
                sel = sel.substr(1, sel.length).trim();
            }

            var stateFound = false;
            for (var j = 0; j < cssStates.length; j++) {
                var state = ':' + cssStates[j];
                if (sel.indexOf(state) > -1) {
                    stateFound = true;
                    if (sel.length == sel.indexOf(state) + state.length) {
                        if (state == ":active") {
                            if ($el.hasClass('pg-state-active')) return true;
                            else found = false;
                        }
                        if (state == ":hover") {
                            if ($el.hasClass('pg-state-hover')) return true;
                            else found = false;
                        }
                        if (state == ":focus") {
                            if ($el.hasClass('pg-state-focus')) return true;
                            else found = false;
                        }
                        if (state == ":visited") {
                            if ($el.hasClass('pg-state-visited')) return true;
                            else found = false;
                        }
                    } else {
                        var cleanSel = sel.replace(/>\s*/g, '>').replace(/ +(?= )/g, '').trim();
                        var cleanSel = convertStatesToClass(cleanSel);
                        var newSelArr = cleanSel.split(' ');

                        var $currEl = $el;
                        for (var k = newSelArr.length - 2; k >= 0; k--) {
                            if (newSelArr[k + 1][0] == '>') {
                                if (!$currEl.parent().length || !$currEl.parent().is(newSelArr[k])) {
                                    found = false;
                                    break;
                                }
                            } else {
                                $currEl = $currEl.closest(newSelArr[k]);
                                if ($currEl.length == 0) {
                                    found = false;
                                    break;
                                }
                            }
                        }
                    }
                }
            }

            try {

                if (!stateFound && $el[0].matches(sel)) {
                    return true;
                }
            } catch (err) {}
        }

        if (found) return true;
        else return false;
    }

    var canShowNode = function (node, $el) {
        if (!node.selector) return true;

        return canShowNodeBySelector(node.selector, $el)
    }

    var getNodeFromStylesheet = function (item, pgStylesheet) {
        var findNodeForStylesheet = function (stylesheet) {
            var node = stylesheet.getNodeWithPgId(item.id);
            return {
                node: node,
                pgStylesheet: stylesheet
            };
        }

        if (item.file) {
            if (item.file == pgStylesheet.getUniqueName()) {
                return findNodeForStylesheet(pgStylesheet);
            }

            var forAllImportedFiles = function (pgStylesheetList) {
                for (var i = 0; i < pgStylesheetList.length; i++) {
                    var importedFile = pgStylesheetList[i];
                    var importedFileId = importedFile.getUniqueName();

                    if (importedFileId == item.file) {
                        return findNodeForStylesheet(importedFile);
                    }

                    if (importedFile.importedFiles.length > 0) {
                        var stylesheet = forAllImportedFiles(importedFile.importedFiles);
                        if (stylesheet) return stylesheet;
                    }
                }
            }

            return forAllImportedFiles(pgStylesheet.importedFiles) || null;
        } else {
            return findNodeForStylesheet(pgStylesheet);
        }

        return null;
    }

    var activeStateChanged = function (curr_state) {
        if (curr_state.length != active_nodes.active_states.length) return true;

        for (var i = 0; i < curr_state.length; i++) {
            if (active_nodes.active_states.indexOf(curr_state[i]) == -1) return true;
        }

        return false;
    }

    var getPgSelector = function (selector) {
        for (var i = 0; i < cssStates.length; i++) {
            selector = selector.replace(':' + cssStates[i], '.pg-state-' + cssStates[i]);
        }

        return selector;
    }

    var clearAndUpdateActiveNodes = function (currSelectedElement, varTrigger, stylesheetChanged) {
        var changed = true;
        if (active_nodes == null) changed = false;

        _this.clear();
        selectedElement = for_pgel || currSelectedElement || pinegrow.getSelectedElement();
        _this.updateMatchedRules(true, null, {
            triggerOnNull: changed,
            varTrigger: varTrigger,
            stylesheetChanged: stylesheetChanged
        });
    }

    var onStylesheetsLoaded = function () {
        var currSelectedElement = for_pgel || pinegrow.getSelectedElement();
        if (selectedElement != currSelectedElement) {
            clearAndUpdateActiveNodes(currSelectedElement, false, true);
        }
    }

    var onStylesheetsClosed = function (p, cs) {
        if (!cs.elementStyle) {
            clearAndUpdateActiveNodes();
        }
    }

    var onCSSPseudoUpdated = function () {
        clearAndUpdateActiveNodes();
    }

    var onSelectedElementsChanged = function () {
        //update active rules every time the element is selected
        //if (selectedElement != pinegrow.getSelectedElement()) {
            clearAndUpdateActiveNodes();
        //}
    }

    var onSourceChanged = function () {
        clearAndUpdateActiveNodes();
    }

    var onRulesChanged = function () {
        clearAndUpdateActiveNodes();
    }

    var onElementClassesChanged = function () {
        clearAndUpdateActiveNodes();
    }

    var onStylesheetChanged = function () {
        clearAndUpdateActiveNodes(false, false, true);
    }

    var onActivePageViewChanged = function(page, pv) {
        if (selectedElement && selectedElement.getPage() == page && page.activeView == pv) {
            clearAndUpdateActiveNodes();
        }
    }

    var onNodeChanged = function (node, stylesheet, oldValue) {
        if (!selectedElement) return;

        var checkRuleSelector = function (rule, varTrigger) {
            if (rule) {
                var el = selectedElement.getDOMElement();
                var selector = getPgSelector(rule.selector);
                try {
                    if (el.matches(selector)) {
                        clearAndUpdateActiveNodes(null, varTrigger);
                        return true;
                    }
                } catch(err) {}
            }
            return false;
        }

        var ruleNode = null;
        var pgcssh = pinegrow.pgPostCSSHelper;
        if (pgcssh.getNodeType(node) == 'rule') {
            checkRuleSelector(node);
        }
        else if (pgcssh.getNodeType(node) == 'atrule') {
            if (node.name == 'media' ) {
                var nodes = node.nodes;
                for (var i = 0; i < nodes.length; i++) {
                    var currNode = nodes[i];
                    if (pgcssh.getNodeType(currNode) == 'rule') {
                        var updated = checkRuleSelector(currNode);
                        if (updated) break;
                    }
                }
            }
        }
        else if (pgcssh.getNodeType(node) == 'import') {
            clearAndUpdateActiveNodes();
        }
        else if (pgcssh.getNodeType(node) == 'decl') {
            if (stylesheet.isVarValue(node.value) || (oldValue && stylesheet.isVarValue(oldValue.value))) {
                var parent = node.parent;
                if (pgcssh.getNodeType(parent) == 'rule') {
                    checkRuleSelector(parent, true);
                }
            }
        }
    }

    var onNodeAdded = function (page, obj) {
        if (!obj) return;

        onNodeChanged(obj.node, obj.stylesheet);
    }

    var onNodeUpdated = function (page, obj) {
        if (!obj) return;

        onNodeChanged(obj.node, obj.stylesheet, obj.oldValue);
    }

    var onNodeDeleted = function (page, obj) {
        if (!obj) return;

        onNodeChanged(obj.node, obj.stylesheet);
    }

    // TODO enhance this function
    //      to calculate the active nodes
    //      if one of active views opened
    var getting_in_progress_pgel = null;
    var callback_list = []

    this.clearCallbackList = function() {
        callback_list = []
    }

    this.updateMatchedRules = function (updateNodes, done, options) {
        options = options || {};

        if(getting_in_progress_pgel) {
            if(getting_in_progress_pgel !== selectedElement) {
                getting_in_progress_pgel = null;
                callback_list = []
            } else {
                if(done) {
                    callback_list.push(done);
                }
            }
        }

        if(!getting_in_progress_pgel) {

            getting_in_progress_pgel = selectedElement;
            callback_list = []

            crsaGetFreshMatchedRules(selectedElement, function (curr_matched_rules, inherited_rules, inline_rules) {
                _this.matched_rules = curr_matched_rules;

                if (updateNodes) {
                    _this.updateActiveNodes(
                        options.triggerOnNull,
                        options.varTrigger,
                        options.stylesheetChanged,
                        options.rule_list, inherited_rules, inline_rules);
                }

                if (done) done(true);

                callback_list.forEach(function(func) {
                    func(true);
                })

                getting_in_progress_pgel = null;
                callback_list = []
            });
        }
    }

    //changed to return list of rules directly
    this.getMatchedRules = function(done) {
        if(this.matched_rules) {
            done(this.getShownNodes());
        }
        else {
            this.updateMatchedRules(false, function () {
                done(_this.getShownNodes());
            });
        }
    }

    function getRulesFromSelectors(rule_list, matched, more_sheets, itemsMap, nodes_to_show, pgel, inherited) {
        pgel = pgel || selectedElement;
        var page = pgel.getPage();
        if (rule_list) {
            //reorder rule list so that matched are on top
            var r = [];
            for (var im = 0; im < matched.length; im++) {
                var idx = rule_list.indexOf(matched[im]);
                if (idx >= 0) {
                    r.push(matched[im]);
                    rule_list.splice(idx, 1);
                }
            }
            for (var i = rule_list.length - 1; i >= 0; i--) {
                r.unshift(rule_list[i]);
            }
            matched = r;
        }

        var ignored_matched_rules = [];
        for (var i = matched.length - 1; i >= 0; i--) {
            var foundStylesheet = pgStylesheetHelper.getParentStylesheetBySelector(matched[i], page);
            if (!foundStylesheet) {
                if(matched[i].csurl) {
                    foundStylesheet = pinegrow.getStylesheetByUrl(matched[i].csurl);
                    if(foundStylesheet && foundStylesheet.sourceStylesheet) {
                        foundStylesheet = foundStylesheet.sourceStylesheet;
                    }
                }
                if(!foundStylesheet) continue;
            }

            if (foundStylesheet.ignore && !inherited) {
                if (!canShowNodeBySelector(matched[i].text, pgel.get$DOMElement())) continue;

                var uniqueName = foundStylesheet.getUniqueName();
                if (!more_sheets[uniqueName]) {
                    more_sheets[uniqueName] = {
                        name: foundStylesheet.name,
                        stylesheet: foundStylesheet,
                        nodesNo: 0
                    }
                }

                ignored_matched_rules.push({
                    matched_rule: matched[i],
                    pgStylesheet: foundStylesheet
                })
                more_sheets[uniqueName].nodesNo++;
                continue;
            }

            var item = {};
            var pgIdFound = false;
            var selectors = matched[i].selectors;
            for (var j = 0; j < selectors.length; j++) {
                var selector = selectors[j].text;

                if (!pgIdFound) {
                    var matchedPgId = selector.match(/\.pg-node-id-([0-9]+)/);
                    if (matchedPgId && matchedPgId.length > 1) {
                        item['id'] = matchedPgId[1];
                        pgIdFound = true;
                        continue;
                    }
                }

                var matchedPgFile = selector.match(/\.pg-file-(.*)/);
                if (matchedPgFile && matchedPgFile.length > 1) {
                    item['file'] = matchedPgFile[1];
                }
            }
            if (!pgIdFound) continue;

            if (!item['file']) item['file'] = foundStylesheet.getUniqueName();
            var itemsMapKey = item['file'] + '_' + item['id'];
            if (itemsMap[itemsMapKey]) continue;

            itemsMap[itemsMapKey] = item['id'];

            var sNode = getNodeFromStylesheet(item, foundStylesheet);
            var node = sNode.node;
            var pgStylesheet = sNode.pgStylesheet;

            if (node) {
                if (!canShowNode(node, pgel.get$DOMElement())) continue;

                nodes_to_show.push({
                    node: node,
                    pgStylesheet: pgStylesheet
                });
            }
        }

        return {
            nodes_to_show: nodes_to_show,
            ignored_matched_rules: ignored_matched_rules,
            active_states: [],
            more_nodes: more_sheets,
            matched_rules: matched
        };
    }

    this.updateActiveNodes = function (triggerOnNull, varTrigger, stylesheetChanged, rule_list, inherited_data, inline_rules) {
        var triggerEvent = false;

        var matchedFinished = function() {
            if (triggerEvent && !for_pgel) {
                var trigger = 'on_css_active_rule_changed';
                if (varTrigger) trigger = 'on_css_active_var_changed';

                var page = null;
                if (selectedElement) page = selectedElement.getPage();
                pinegrow.dispatchEvent(trigger, page, {active_nodes: active_nodes, stylesheet_changed: stylesheetChanged});
            }
        }

        var show_all = this.show_all_rules;

        if (selectedElement || show_all) {
            //var active_states = [];
            var nodes_to_show = [];
            var more_sheets = {};
            var itemsMap = {};

            var matched = rule_list || this.matched_rules || [];

            var $el = selectedElement ? selectedElement.get$DOMElement() : null;


            /*
            for (var i = 0; i < cssStates.length; i++) {
                var state = cssStates[i];
                if ($el.hasClass('pg-state-' + state)) active_states.push(state);

            var nodeid = selectedElement.getData('selectedCssRule');
            if(nodeid) {
                if(nodeid.csurl) {
                    pgel_selected_css_rule = pgcssh.getNodeFromGlobalId(nodeid);
                } else {
                    selected_rule = 'inline';
                    return;
                }
            }

            if (active_nodes && active_nodes.hasOwnProperty('nodes_to_show')) {
                if (!activeStateChanged(active_states)) return active_nodes;
            }
*/

            if(!isApp()) {

                var list = [];

                if($el && !show_all) {
                    var el = $el.get(0);
                    var idx = 0;

                    var doNode = function (pn, cs) {
                        if (el && pn.nodes) {
                            pn.nodes.forEach(function (node) {
                                if (node.type === 'rule') {

                                    try {
                                        var sel = node.selector.replace(/\:[^\s,$]+/gi, '');

                                        if (el.matches(sel)) {
                                            list.push({
                                                node: node,
                                                pgStylesheet: cs,
                                                idx: idx++,
                                                spec_array: null
                                            });
                                        }
                                    } catch (err) {
                                    }

                                } else if (node.type === 'decl') {

                                } else {
                                    doNode(node, cs);
                                }
                            })
                        }
                    }

                    pinegrow.stylesheets.getAllCrsaStylesheets().forEach(function (cs) {
                        var skip = cs.inlineSheetPgel && cs.inlineSheetPgel.hasAttribute('data-pg-injected');
                        if(cs.ignore) skip = true;
                        if(!skip) {
                            cs.rootNode && doNode(cs.rootNode, cs);
                        }
                    })

                    list.sort(function(a, b) {
                        if(!a.spec_array) {
                            a.spec_array = SPECIFICITY.calculate(a.node.selector)[0].specificityArray;
                        }
                        if(!b.spec_array) {
                            b.spec_array = SPECIFICITY.calculate(b.node.selector)[0].specificityArray;
                        }
                        var r = SPECIFICITY.compare(a.spec_array, b.spec_array);
                        if(r === 0) {
                            r = a.idx > b.idx ? 1 : -1;
                        }
                        return -1 * r;
                    })

                    //list.reverse();

                } else if(show_all) {
                    var doNodeAll = function (pn, cs) {
                        if (pn.nodes) {
                            pn.nodes.forEach(function (node) {
                                if (node.type === 'rule') {
                                    list.push({
                                        node: node,
                                        pgStylesheet: cs
                                    });
                                } else if (node.type === 'decl') {
                                } else {
                                    doNodeAll(node, cs);
                                }
                            })
                        }
                    }

                    pinegrow.stylesheets.getAllCrsaStylesheets().forEach(function (cs) {
                        if(!cs.inlineSheetPgel || !cs.inlineSheetPgel.hasAttribute('data-pg-injected')) {
                            doNodeAll(cs.rootNode, cs);
                        }
                    })
                }

                active_nodes = {
                    nodes_to_show: list,
                    ignored_matched_rules: [],
                    active_states: [],
                    more_nodes: [],
                    matched_rules: []
                };
                triggerEvent = true;
                matchedFinished();
            } else {
                active_nodes = getRulesFromSelectors(null, matched, more_sheets, itemsMap, nodes_to_show);

                active_nodes.inline_rules = inline_rules;

                active_nodes.inherited = [];

                if (inherited_data) {
                    inherited_data.forEach(function (id) {
                        //var active_states = [];
                        var nodes_to_show = [];

                        /*
                         var $el = id.pgel.get$DOMElement();
                         for (var i = 0; i < cssStates.length; i++) {
                         var state = cssStates[i];
                         if ($el.hasClass('pg-state-' + state)) active_states.push(state);
                         }
                         */
                        var an = getRulesFromSelectors(id.selectors, matched, more_sheets, {}, nodes_to_show, id.pgel, true);
                        //console.log('INHERITED', an)

                        loadInlineAsStylesheetAndGetRules(id.pgel, function (rules, cs) {
                            for (var i = rules.length - 1; i >= 0; i--) {
                                if (rules[i].nodes.length) {
                                    an.nodes_to_show.unshift({
                                        node: rules[i],
                                        pgStylesheet: cs
                                    })
                                }
                            }
                        })

                        if (an.nodes_to_show.length) {
                            active_nodes.inherited.push({
                                pgel: id.pgel,
                                data: an
                            })
                        }
                    })
                }
                triggerEvent = true;
                matchedFinished();
            }

        }
        else if (triggerOnNull) {
            triggerEvent = true;
            matchedFinished();
        }
    }

    this.getActiveNodes = function () {
        return active_nodes;
    }

    this.clear = function () {
        active_nodes = null;
        this.clearCallbackList();
    }

    this.getPgel = function() {
        return selectedElement;
    }

    this.setPgel = function(pgel) {
        selectedElement = pgel;
    }


    this.getShownNodes = function() {
        if(active_nodes) {
            return active_nodes.nodes_to_show.map(function(d) { return d.node; });
        } else {
            return [];
        }
    }

    this.isRuleShown = function(rule) {
        if(active_nodes) {
            for(var i = 0; i < active_nodes.nodes_to_show.length; i++) {
                if(active_nodes.nodes_to_show[i].node === rule) {
                    return true;
                }
            }
        }
        return false;
        //return this.getShownNodes().indexOf(rule) >= 0;
    }

    if(!for_pgel) {
        if(pgf.edit_css) {
            pinegrow.addEventHandler('on_page_stylesheets_loaded', onStylesheetsLoaded);
            pinegrow.addEventHandler('on_stylesheet_closed', onStylesheetsClosed);

            pinegrow.addEventHandler('on_css_pseudo_updated', onCSSPseudoUpdated);
            pinegrow.addEventHandler('on_selected_elements_changed', onSelectedElementsChanged);

            pinegrow.addEventHandler('on_active_page_view_changed', onActivePageViewChanged);
            pinegrow.addEventHandler('on_page_view_resized', onActivePageViewChanged);
            pinegrow.addEventHandler('on_page_closed', onActivePageViewChanged);

            pinegrow.addEventHandler('on_css_node_added', onNodeAdded);
            pinegrow.addEventHandler('on_css_node_updated', onNodeUpdated);
            pinegrow.addEventHandler('on_css_node_deleted', onNodeDeleted);

            pinegrow.addEventHandler('on_element_classes_changed', onElementClassesChanged);

            pinegrow.addEventHandler('on_css_source_changed', onSourceChanged);
            pinegrow.addEventHandler('on_element_classes_changed', onSourceChanged);

            $body.on('crsa-stylesheets-changed', onStylesheetChanged);
        }
    } else {
        clearAndUpdateActiveNodes();
    }

    this.destroy = function() {
        this.clear();

        if(!for_pgel) {
            if(pgf.edit_css) {
                pinegrow.removeEventHandler('on_page_stylesheets_loaded', onStylesheetsLoaded);
                pinegrow.removeEventHandler('on_stylesheet_closed', onStylesheetsClosed);

                pinegrow.removeEventHandler('on_css_pseudo_updated', onCSSPseudoUpdated);
                pinegrow.removeEventHandler('on_selected_elements_changed', onSelectedElementsChanged);

                pinegrow.removeEventHandler('on_active_page_view_changed', onActivePageViewChanged);
                pinegrow.removeEventHandler('on_page_view_resized', onActivePageViewChanged);
                pinegrow.removeEventHandler('on_page_closed', onActivePageViewChanged);

                pinegrow.removeEventHandler('on_css_node_added', onNodeAdded);
                pinegrow.removeEventHandler('on_css_node_updated', onNodeUpdated);
                pinegrow.removeEventHandler('on_css_node_deleted', onNodeDeleted);

                pinegrow.removeEventHandler('on_element_classes_changed', onElementClassesChanged);

                pinegrow.removeEventHandler('on_css_source_changed', onSourceChanged);
                pinegrow.removeEventHandler('on_element_classes_changed', onSourceChanged);

                $body.off('crsa-stylesheets-changed', onStylesheetChanged);
            }
        }
    }
}
