/*

(old PgCSSEditor is below)

Changes:
- uses pg grid to put sections in columns
- list stuff for shadows, transforms etc is removed. Better to create a helper class and implement list in custom CSS control for shadow, transform...
- fields & controls are drawn only once
- .setEditedRule is called to update GUI for a selected rule
- everything that uses .show now needs to be implemented with PgCustomPropertiesControl

 */

var PgCSSEditorView = function() {
    var $element = $('<div/>', {class: 'pg-css-visual crsa-css-gui'});
    var view = this.view = new PgUIView($element, 'css_visual', 'CSS Visual Editor');

    var _this = this;
    var delayed = new PgDelayedTasks();



    view.tabIcon = 'icon-window';

    var selectedRule = null;

    var editor = new PgCSSEditorNEW($element);

    view.onResize = function() {
        editor.updateLayout($element.width(), $element.height());
    }



    view.on('crsa-rule-selected', function(event, obj) {
        view.whenVisible(function() {
            if (obj) {
                showCssGui(obj.rule, obj.options);
                selectedRule = obj.rule;
            }
            else {
                selectSameRule();
            }
        })
    });

    var showCssGui = function (rule, options) {
        var obj = {
            indices: pinegrow.getSelectedRuleIndices(),
            pgStylesheet: pinegrow.getSelectedRuleStylesheet()
        }
        editor.editRule(obj, options);
    }

    var onNodeAddedOrUpdated = function (obj, deleted) {
        view.whenVisible(function () {
            if (obj.edited_by != 'gui') {

                var node = obj.node;
                var selectedRule = pinegrow.getSelectedRule();

                delayed.cancel('update');

                if (selectedRule && pinegrow.pgPostCSSHelper.isNodeEqualOrChildOf(selectedRule, node)) {

                    var show = function () {
                        var node = obj.node;
                        if (node.type == "rule" && !deleted) {
                            var rule = pinegrow.pgPostCSSHelper.getRuleForNode(node);
                            showCssGui(rule);
                        }
                        else {
                            if (pinegrow.isSelectedRule(node.parent)) {
                                showCssGui(rule);//selectSameRule();
                            }
                        }
                    }

                    if (obj.edited_by == 'undo') {
                        //lets delay a bit, maybe another undo event happens soon
                        delayed.run('update', function() {
                            show();
                        }, 100)
                    } else {
                        show();
                    }
                }
            }
        })
    }

    var onDeleted = function(page, obj) {
        onNodeAddedOrUpdated(obj, true);
    }

    var onChanged = function(page, obj) {
        onNodeAddedOrUpdated(obj);
    }

    pinegrow.addEventHandler('on_css_node_deleted', onDeleted);
    pinegrow.addEventHandler('on_css_node_updated', onChanged);
    pinegrow.addEventHandler('on_css_node_added', onChanged);

    view.onDestroy = function() {
        editor.remove();
        editor = null;
        delayed.destroy();

        pinegrow.removeEventHandler('on_css_node_deleted', onDeleted);
        pinegrow.removeEventHandler('on_css_node_updated', onChanged);
        pinegrow.removeEventHandler('on_css_node_added', onChanged);
        pinegrow.removeEventHandler('on_current_device_size_changed', onDeviceSizeChanged)
    }
}

var PgCSSEditorNEW = function ($container) {

    $container = $container || $('<div class="crsa-stylesheets crsa-css-gui"></div>');

    var grid = this.grid = new PgUISectionGrid();
    grid.minWidth = 250;
    grid.maxWidth = 500;

    var notice = new PgPanelNotice();
    notice.$element.appendTo($container);

    var css_api = new PgCSSApi();

    var device_size_control = new PgDeviceSizesControl(false, undefined, true);

    device_size_control.setStatesOptions([]);

    device_size_control.onDeviceSizeChanged = function(page, def) {
        var pgel = pinegrow.selectedElements.getLastSelected();
        pgel && (pgel.getData('selectedCssRule') !== 'none') && pgel.setData('selectedCssRule', null);
    }

    $container.prepend(device_size_control.$element);
    device_size_control.show();

    var $no_rule_overlay = $(`<div class="pg-css-editor-no-rule-overlay"><div class="alert alert-panel-notice"><p></p><button class="pg-button">Create the rule</button></div></div>`).appendTo($container);

    var no_rule_overlay_p= $no_rule_overlay.find('p').get(0);
    var no_rule_overlay_button = $no_rule_overlay.find('button').get(0);
    var no_rule_overlay_on_click = null;

    $no_rule_overlay.find('button').on('click', function(e) {
        e.preventDefault();
        no_rule_overlay_on_click && no_rule_overlay_on_click();
    })

    this.showNoRuleNotice = function(selector, size, on_click, what) {
        what = what || 'no_rule';
        var p;
        switch (what) {
            case 'no_rule':
                if (selector !== 'Style attribute') {
                    p = `The rule <span class="rule">${selector}</span> for the screen size <span class="size">${size}</span> is not defined.`
                    pgShow(no_rule_overlay_button);
                } else {
                    p = `Save the <span class="rule">${selector}</span> into a CSS rule to make it responsive.`;
                    pgHide(no_rule_overlay_button);
                }
                no_rule_overlay_button.innerHTML = 'Create the rule';
                break;
            case 'rule_not_visible':
                p = `The rule <span class="rule">${selector}</span> exists, but it is not visible on the <span class="size">${size}</span> display.`
                pgShow(no_rule_overlay_button);
                no_rule_overlay_button.innerHTML = 'Resize view to ' + size;
                break;
        }

        no_rule_overlay_p.innerHTML = p;

        no_rule_overlay_on_click = on_click;
        pg$Show($no_rule_overlay, 'flex');
    }

    this.hideNoRuleNotice = function() {
        pg$Hide($no_rule_overlay);
    }

    this.hideNoRuleNotice();

    var $filter = $('<div class="pg-css-gui-filter"></div>').appendTo($container);

    var filter_sections = [
        //{name: 'Media query', sections: ['media_query'], icon: 'MQ'},
        {name: 'Text', sections: ['text'], icon: 'icon-Pin18'},
        {name: 'Margin, padding and size', sections: [ 'margin_and_padding', 'dimension'], icon: 'icon-Pin16'},
        {name: 'Display and position', sections: [ 'display_prop', 'position_prop'], icon: 'icon-Pin17'},
        {name: 'Background', sections: ['background'], icon: 'icon-Pin21'},
        {name: 'Border and border radius', sections: ['border_prop'], icon: 'icon-Pin20'},
        {name: 'Flexbox', sections: ['flex', 'flex-child'], icon: 'icon-Pin22'},
        {name: 'Shadows', sections: ['box-shadow', 'text-shadow'], icon: 'icon-Pin23'},
        {name: 'Transforms', sections: ['transforms'], icon: 'icon-Pin24'},
        {name: 'Transitions', sections: ['transitions'], icon: 'icon-graph'},
        {name: 'List', sections: ['list'], icon: 'icon-Pin25'},
        {name: 'Grid', sections: ['grid-container', 'grid-item'], icon: 'icon-grid'}
    ]

    if(pgf.kids) {
        filter_sections = [
            {name: 'Text', sections: ['text'], icon: 'icon-Pin18'},
            {name: 'Background', sections: ['background'], icon: 'icon-Pin21'},
            {name: 'Margin, padding and dimension', sections: ['margin_and_padding', 'dimension'], icon: 'icon-Pin16'},
            {name: 'Border', sections: ['border_prop'], icon: 'icon-Pin20'},
            {name: 'Position and display', sections: ['position_prop', 'display_prop'], icon: 'icon-Pin17'},
           // {name: 'Grid', sections: ['grid-simple'], icon: 'icon-grid'},
           // {name: 'Flexbox', sections: ['flex', 'flex-child'], icon: 'icon-Pin22'},
            {name: 'Shadows', sections: ['box-shadow', 'text-shadow'], icon: 'icon-Pin23'}
           // {name: 'List', sections: ['list'], icon: 'icon-Pin25'},
        ]

        if(pgf.kids_with_transitions) {
            filter_sections.push({name: 'Transforms', sections: ['transforms'], icon: 'icon-Pin24'});
            filter_sections.push({name: 'Transitions', sections: ['transitions'], icon: 'icon-graph'})
        }
    }

    var show_sections = [];
    var auto_show_sections = false;//pinegrow.getSetting('css-auto-sections', "0") === "1";

    var toolbar = new PgButtonToolbar(true /* small */);
    toolbar.$element.appendTo($filter);

    /*
    toolbar.addSection('auto');
    var item = new PgButtonToolbarItem('icon-duplicate', 'Automatic - Show sections with used properties first and show unused sections at the end');
    item.setActive(auto_show_sections);
    toolbar.addItem('auto', item);

    item.onClick = function(item) {
        auto_show_sections = !this.getActive();
        this.setActive(auto_show_sections);
        pinegrow.setSetting('css-auto-sections', auto_show_sections ? "1" : "0")
        if(auto_show_sections) {
            toolbar.hideSection('sections');
        } else {
            toolbar.showSection('sections');
        }
        updateSections();
    }
*/

    toolbar.addSection('sections');

    filter_sections.forEach(function(sf) {
        var item = new PgButtonToolbarItem(sf.icon, sf.name, sf);
        sf.active = false;
        item.setActive(sf.active);
        toolbar.addItem('sections', item);

        sf.toolbarItem = item;

        item.onClick = function(item) {
            var key = sf.sections[0];
            sections_map[key].scrollTo();
            //this.data.active = !this.getActive();
            //this.setActive(this.data.active);

            //updateSections();
        }
    })

    if(auto_show_sections) {
        toolbar.hideSection('sections');
    } else {
        toolbar.showSection('sections');
    }

    $container.append(grid.$element);

    var on_fields_created;

    var fields_shown = false;

    this.ready = true;

    var fields = {}; //dictionary of all fields (including custom controls)

    var current_rule_obj;

    var current_element_obj;

    var stylesheet;

    var getObj = function() {
        if (current_rule_obj) {
            return current_rule_obj;
        }
        return current_element_obj
    }

    this.editRule = function(obj, options) {

        options = options || {};

        // current_element_obj = pinegrow.getSelectedElement();
        var st = this.getScrollTop();
        current_rule_obj = obj;

        var currObj = current_rule_obj || current_element_obj;

        if(pinegrow.selectedElements.getLastSelected() === null) {
            currObj = null;
        }

        var rule = null;
        // if (currObj) {
        if (currObj && currObj.indices && currObj.pgStylesheet) {
            hideMessage();
            if (!grid.$element.is(':visible')) grid.$element.show();

            if (currObj && currObj.hasOwnProperty('indices')) {
                stylesheet = currObj.pgStylesheet;
                if (!stylesheet) return;

                try {
                    rule = pinegrow.pgPostCSSHelper.getNodeByIndices(stylesheet.getRootNode(), currObj.indices);
                    if(rule.type !== 'rule') {
                        grid.$element.hide();
                        showMessage('The selected item is not a CSS rule.');
                        return;
                    }
                    var values = getValuesForObjectUsingRule(rule);
                }
                catch(err) {
                    grid.$element.hide();
                    showMessage(err.message);
                    return;
                }
            }
            else if (currObj && currObj.type === "rule") {
                stylesheet = pgStylesheetHelper.getStylesheetFromNode(currObj.data);
                rule = currObj.data;
                var values = getValuesForObjectUsingRule();
            }
            else if (currObj) {
                stylesheet = null;
                var values = getValuesForObjectUsingStyle();
            }

            if(!fields_shown) {
                showControls(currObj, values);
                fields_shown = true;
            } else {
                if (!options.skip_updating_sections) updateSections();
                //update controls for current values
                $.each(fields, function(fn, field) {
                    if (field instanceof PgCustomPropertyControl) {
                        field.setObject(currObj);
                        field.setValues(values, null, options.skip_updating_sections);
                    } else {
                        field.setValue(values[fn] || null);
                    }
                })
            }
            var scrolled = false;
            if(options.clicked_prop) {
                //find section
                for(var i = 0; i < sections.length; i++) {
                    var pgsec = sections[i];
                    if(pgsec.sec_def && pgsec.sec_def.get_has_data) {
                        if(('_' + options.clicked_prop + '_').match(pgsec.sec_def.get_has_data)) {
                            pgsec.scrollTo();
                            scrolled = true;
                            break;
                        }
                    }
                }
            }

            var media_size = css_api.getMediaSizeForRule(rule);
            var size_def = pinegrow.getSelectedPage().activeView.getDeviceSizeDefForWidth(media_size);

            device_size_control.updateOrShow(size_def);

            if(!scrolled) {
                this.setScrollTop(st);
            }
        } else {
            showMessage('CSS Rule is not selected.');

            $.each(fields, function(fn, field) {
                if (field instanceof PgCustomPropertyControl) {
                    field.setObject(null);
                }
            });
            stylesheet = null;

            grid.$element.hide();
        }

    }

    var $scrollParent = null;
    $scrollParent = $container;

    var rulesDefs = [];

    var def = $.fn.crsa.defaults.rulesDefinition;

    $.each(def.sections, function(key, ruleDef) {
        rulesDefs.push(ruleDef);
        ruleDef.section_key = key;
        ruleDef.component_definition = def;

        //Pre-create custom cotrols so that definition is done before we get values
        $.each(ruleDef.fields, function (fn, fdef) {
            if (fdef.type == 'custom') {
                if (fdef.control && !fdef.control_defined) {
                    var control = fdef.control();
                    var section = control.define(def);
                    fdef.control_defined = true;
                }
            }
        })
    });

    var showMessage = function(msg) {
        notice.show(msg);
        $filter.hide();
    }

    var hideMessage = function () {
        notice.show();
        $filter.show();
    }

    this.updateLayout = function(w, h) {
        grid.updateLayout(w, h);
    }

    var section_has_data = {};

    var getValuesForObject = function (styles, obj) {
        var values = {};
        section_has_data = {};

        var values_str = '';

        $.each(def.sections, function(sn, sv) {
            $.each(sv.fields, function(fn, fv) {
                var action = fv.action ? fv.action : 'style';

                if (fv.get_value) {
                    values[fn] = fv.get_value(obj, fn, styles);
                } else if (action == 'style' && styles) {
                    var css_prop = fv.css_property || fn;
                    var vobj = styles[css_prop] ? styles[css_prop] : null;
                    var v;
                    if (vobj) {
                        var value;
                        if(typeof vobj == 'object') {
                            value = vobj.value;
                        } else {
                            value = vobj;
                        }
                        v = value;
                        if (fv.type == 'image') {
                            v = v.replace(/url\(/g, '').replace(/\"/g,'').replace(/\'/g, "").replace(/\)/g,'');
                        }
                        values[fn] = value;
                    }

                } else if (action == 'rule_name') {
                    if (current_rule_obj) {
                        var pgStylesheet = current_rule_obj.pgStylesheet;
                        var rule = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), current_rule_obj.indices);

                        values[fn] = stylesheet.filterSource(rule.selector);
                    }
                } else if (action == 'rule_media') {
                    if (current_rule_obj) {
                        var pgStylesheet = current_rule_obj.pgStylesheet;
                        var rule = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), current_rule_obj.indices);

                        var mediaNode = pinegrow.pgPostCSSHelper.getMediaNode(rule);
                        var atRuleValue = pinegrow.pgPostCSSHelper.getAtRuleValue(mediaNode);
                        values[fn] = atRuleValue ? atRuleValue : null;
                    }
                } else if (action == 'rule_supports') {
                    if (current_rule_obj) {
                        var pgStylesheet = current_rule_obj.pgStylesheet;
                        var rule = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), current_rule_obj.indices);

                        var mediaNode = pinegrow.pgPostCSSHelper.getSupportsAtruleNode(rule);
                        var atRuleValue = pinegrow.pgPostCSSHelper.getAtRuleValue(mediaNode);
                        values[fn] = atRuleValue ? atRuleValue : null;
                    }
                }
                if(fn in values && values[fn] !== null && values[fn] !== undefined && values[fn] !== '') {
                    values_str += '_' + fn + '_';
                }
            });
        });
        //Not sure why, but values is not correctly filled for many fields, so redoing this here
        $.each(styles, function(key, val) {
            values_str += '_' + key + '_';
            if(!(key in values) && 'value' in val) {
                values[key] = val.value;
            }
        })
        $.each(def.sections, function(sn, sv) {
            if(sv.get_has_data) {
                if (values_str.match(sv.get_has_data)) {
                    section_has_data[sn] = true;
                }
            }
        });

        return values;
    }

    var getValuesForObjectUsingRule = function (rule) {
        var styles = stylesheet.getPropValuePairFor(rule);
        return getValuesForObject(styles, current_rule_obj);
    }

    var getValuesForObjectUsingStyle = function () {
        var styleStr = current_element_obj.getAttr('style');
        var styles = styleToObject(styleStr);

        return getValuesForObject(styles, current_element_obj);
    }

    //addPanelChild removed from here

    var sections = [];


    var updateSections = function() {
        grid.clear();

        var open = [];
        var not_shown = [];

        show_sections = [];
        filter_sections.forEach(function(sf) {
            var has = false;
            sf.sections.forEach(function(sn) {
                if(section_has_data[sn]) {
                    has = true;
                }
            })
            sf.toolbarItem.setActive(has);
        })

        for(var i = 0; i < sections.length; i++) {
            var s = sections[i];

            var key = s.sec_def.section_key;

            if(auto_show_sections) {

                if(!s.sec_def.get_has_data || section_has_data[key]) {
                    grid.addSection(s);
                    open.push(s);
                } else {
                    not_shown.push(s);
                }
            } else {

                if (!show_sections || show_sections.length == 0 || show_sections.indexOf(key) >= 0 || !s.sec_def.get_has_data) {
                    grid.addSection(s);

                    if (show_sections.length) {
                        open.push(s);
                    }
                }
            }
            if(section_has_data[key]) {
                s.setHasData(true);
            } else {
                s.setHasData(false);
            }
        }
        not_shown.forEach(function(s) {
            grid.addSection(s);
        })
        grid.updateLayout();

        open.forEach(function(s) {
            s.forceOpen();
        })
        not_shown.forEach(function(s) {
            //s.forceClose();
        })
    }

    var sections_map = {};

    //create controls. called only once
    var showControls = function (obj, values) {
        //grid.clear();
        on_fields_created = [];

        $.each(rulesDefs, function(i, s) {

            var key = s.section_key;

            if((s.hasOwnProperty("show") && !s.show) || !s.name) return true;

            if(pgf.kids) s.is_closable = false;

            var pgsec = new PgUISection(s.name, s.framework ? s.framework.name : null, s);

            sections.push(pgsec);
            sections_map[key] = pgsec;

            if(key == 'aaaname') {
                $container.prepend(pgsec.$element.addClass('pg-rule-name'));
            } else {
                //grid.addSection(pgsec);
            }

            var $c = pgsec.$content;

            $.each(s.fields, function(fn, fdef) {

                var $field;

                if(fdef.type == 'custom') {
                    if(fdef.control) {

                        var control = fdef.control();
                        control.setObject(obj);
                        control.setValues(values);//todo
                        control.show($scrollParent);
                        $c.append(control.$element);

                        fields[fn] = control;

                        $field = control.$element;
                    } else {
                        //not really supported anymore here
                        $field = fdef.show($c, obj, fn, fdef, values, $scrollParent);
                        console.log('fdef.show is not supported anymore. ' + fn);
                    }
                } else {
                    var field;

                    field = new PgPropertyField($c, getObj, fn, fdef, values, $scrollParent);

                    fields[fn] = field;

                    $field = field.$field;
                }

                // if (s.icons) {
                //     $field.addClass('with-icons');
                //     var icon = fdef.icon;
                //     if (!icon) icon = 'icon-' + fn;
                //
                //     $('<i class="field-icon ' + icon + '"></i>').insertAfter($field.find('>label'));
                // }
                // if(fdef.on_fields_created) {
                //     on_fields_created.push({func: fdef.on_fields_created, obj: obj, field: $field, def: fdef, name: fn, default_value: fdef.default_value || null});
                // }
                if(fdef.validate) {
                    pinegrow.validateField(obj, fn, values.hasOwnProperty(fn) ? values[fn] : null, fdef, $field, values);
                }

                if(fdef.show_if) {
                    var show = typeof fdef.show_if == 'function' ? fdef.show_if( values) : crsaEvaluateShowIfString(fdef.show_if, values, fdef, obj);

                    if($field.is(':visible') !== show) {
                        if(show) {
                            $field.show();
                        } else {
                            $field.hide();
                        }
                    }
                }

            });
        });
        updateSections();

    }

    this.getScrollTop = function () {
        return grid.$element.scrollTop();
    }

    this.setScrollTop = function (scrollTop) {
        grid.$element.scrollTop(scrollTop);
    }


        /*

         for (var i = 0; i < on_fields_created.length; i++) {
         var val = values.hasOwnProperty(on_fields_created[i].name) ? values[on_fields_created[i].name] : null;
         if (val === null && on_fields_created[i].default_value) val = on_fields_created[i].default_value;
         on_fields_created[i].func(
         on_fields_created[i].obj,
         on_fields_created[i].name,
         val,
         on_fields_created[i].def,
         on_fields_created[i].field,
         values);
         }
         on_fields_created = [];
         */


    this.remove = function () {
        current_rule_obj = null;
        sections = null;
        sections_map = null;

        pgRemoveSpectrum($container);
        pgRemoveFilePicker($container);
    }
}



