var PgFastEditor = function ($container, options) {
    if (!$container) throw "No container found";
    $container.data('pg-fast-editor', this);

    // public var

    // private
    var _this = this;
    var list = [];
    var $ul = $('<ul class="fast-editor-list"></ul>').appendTo($container);
    var $list_icon = $('<i class="icon-bin crsa-delete-item"></i>');

    this.itemSelector = 'li.crsa-node-item';

    var initOptions = function () {
        options['add-new-placeholder-text'] = options['add-new-placeholder-text'] || 'Add new...';
    }
    initOptions();

    this.itemLengthToAddPlaceholder = options['item-length-to-add-placeholder'] || 0;

    var convertToTree = function () {
        var listLength = _this.list.length;

        for (var i = 0; i < listLength; i++) {
            var item = _this.list[i];

            if (options.isAllowedToShow) {
                if (!options.isAllowedToShow(item.prop, item.value)) continue;
            }

            var $li = $('<li class="crsa-node-item"><div><span class="item-prop">' + item.prop + '</span><span class="item-value">' + item.value + '</span></div></li>').appendTo($ul);
            $li.data('list-item', item);
        }
    }

    var onClick = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('li');

        if ($target.hasClass('crsa-delete-item')) {
            var node = _this.getNodeFrom($li);
            if (options.onItemRemoved(node.prop, node.value)) {
                $li.remove();
                stylePaneHelper.onNodeRemoved();
            }
        }
    }

    var onMouseMove = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('.crsa-node-item');

        if (!$li.hasClass('add-new-node') && !$li.hasClass('placeholder-node')) {
            var $div = $li.find('> div:not(.pg-tree-quick-insert-line)');
            $list_icon.appendTo($div);
        }
    }

    var onMouseLeave = function (e) {
        $list_icon.detach();
    }

    var addTriggers = function () {
        $ul
            .off('click', 'li', onClick)
            .on ('click', 'li', onClick)

            .off('mousemove', 'li', onMouseMove)
            .on ('mousemove', 'li', onMouseMove)

            .off('mouseleave', 'li', onMouseLeave)
            .on ('mouseleave', 'li', onMouseLeave);
    }

    this.isAllowedToAddNewNodeAt = function ($li, $target) {
        var $li = $target.closest('li');

        if ($target.is('span') || $target.is('i') || $li.data('placeholder-item') || $li.hasClass('add-new-node')) return false;
        return true;
    }

    this.getNodeFrom = function ($li) {
        return $li.data('list-item');
    }

    this.getLengthWithoutPlaceholders = function (node) {
        return null;
    }

    this.pasteTriggeredForNewField = function ($el) {
        var clipboardItem = pinegrow.clipboard.getCurrentItem();
        if (clipboardItem.type == 'attributes') {
            var attrs = options.parseCodeAndGetArray(clipboardItem.code);
            for (var i = 0; i < attrs.length; i++) {
                var found = false;
                for (var j = 0; j < this.list.length; j++) {
                    if (attrs[i].prop == this.list[j].prop) {
                        this.list[i].value = attrs[i].value;
                        found = true;
                        break;
                    }
                }

                if (!found) {
                    this.list.push(attrs[i]);
                }
            }
            this.show();
        }
        else {
            //Is already inserted by system
            //document.execCommand('insertText', false, clipboardItem.code);
        }
    }

    this.addingNewNode = function ($el) {
        $el.addClass('item-prop');
    }

    this.addNewNodeUI = function ($el, helperNode, type, insertAfter) {
        var $li = $el.closest('li');
        $li.addClass('crsa-node-item');

        var text = $el.text();
        if (text != '') {
            var newNode = {
                prop: text,
                value: ''
            };

            $el.off('keydown').off('blur');

            this.list.push(newNode);
            // this will refresh the whole view
            // options.onItemAdded(newNode.prop, newNode.value);

            $li.data('node', newNode)
               .data('list-item', newNode)
               .data('placeholder-item', false)
               .data('new-added', true);

            var $nodeValue = $('<span class="item-value"></span>').appendTo($el.parent());
            $el.removeAttr('contenteditable');
            $nodeValue.attr('contenteditable', 'true');
            crsaSelectText($nodeValue[0]);

            // wait so the blur executed
            setTimeout(() => {
                $li.data('new-added', false);
            }, 0);
        }
    }

    this.nodeEditDone = function (elm) {
        var $el = $(elm);
        var $li = $el.closest('li');

        var text = $el.text();
        var node = this.getNodeFrom($li);

        if (text == "" && false /* allow empty attr value */) {
            if (options.onItemRemoved(node.prop, node.value)) {
                $li.remove();
                stylePaneHelper.onNodeRemoved();
            }
        }
        else {
            var oldText = $el.data('oldText');
            var changeType = '';

            if ($el.hasClass('item-prop')) {
                changeType = 'prop';
                node.prop  = text;
            }
            else if ($el.hasClass('item-value')) {
                changeType = 'value';
                node.value = text;
            }
            options.onItemChanged(node.prop, node.value, changeType, oldText);
        }
    }

    this.getStringOfSelectedItems = function (before, between, after) {
        before  = before  ||  '';
        after   = after   ||  '';
        between = between || ' ';
        var nodesSource = '';

        $ul.find('> .selected-item').each(function (i, el) {
            var $el = $(el);
            var node = _this.getNodeFrom($el);

            nodesSource += node.prop + between + node.value + after;
        });

        return nodesSource.trim();
    }

    this.setList = function (propList) {
        this.list = propList;
    }

    this.show = function () {
        $container.html('');
        $ul = $('<ul class="fast-editor-list"></ul>').appendTo($container);

        convertToTree();
        addTriggers();

        var helperOptions = {
            'add-new-placeholder-text': options['add-new-placeholder-text']
        }

        stylePaneHelper = new PgStylePaneHelper(this, $ul, helperOptions);
        stylePaneHelper.addTriggers($ul);
        stylePaneHelper.supportInsertBorders();
    }
}
