window.PgSCSSStylesheet = function () {
    var pgStylesheet = window.PgCompilableStylesheet();
    pgStylesheet.fileType = "scss";

    var online_sass = false;

    if(online_sass) {
        Function.prototype.call$1 = function (a) {
            return this(a);
        };
        Function.prototype.call$0 = function () {
            return this();
        };
        Function.prototype.call$2 = function (a, b) {
            return this(a, b);
        };
        Function.prototype.call$3 = function (a, b, c) {
            return this(a, b, c);
        };
        Function.prototype.call$4 = function (a, b, c, d) {
            return this(a, b, c, d);
        };
        Function.prototype.call$5 = function (a, b, c, d, e) {
            return this(a, b, c, d, e);
        };
        Function.prototype.call$1$1 = function (a) {
            return this(a);
        };
        Function.prototype.call$1$3 = function (a, b, c) {
            return this(a, b, c);
        };
        Function.prototype.call$2$2 = function (a, b) {
            return this(a, b);
        };
        Function.prototype.call$6 = function (a, b, c, d, e, f) {
            return this(a, b, c, d, e, f);
        };
        Function.prototype.call$1$0 = function () {
            return this();
        };
    }
    //var sass = require('node-sass');
    //var sass = require(isApp() ? 'node-sass' : 'dart-sass');
    //var sass = require('sass-embedded');
    //var sass = window.dartSass;

    //var sass = online_sass ? require('dart-sass') : (require(isApp() ? ('sass-embedded') : 'dart-sass'));



    var sass = require(isApp() ? 'sass-embedded' : 'dart-sass');

    var allPostcssLib = require('postcss');
    var postcss = allPostcssLib.postcss;
    var scssSyntax = allPostcssLib.postcssSCSS;
    var scssStringify = allPostcssLib.postcssSCSS.stringify;
    var scssStringifyWithParams = allPostcssLib.postcssSCSS.stringifyWithParams;


    const SASS_FUNCTIONS = ['rgb', 'rgba', 'red', 'green', 'blue', 'mix', 'hue', 'saturation', 'lightness', 'adjust-hue', 'lighten', 'darken', 'saturate', 'desaturate', 'grayscale', 'complement', 'invert', 'alpha', 'opacity', 'fade-out', 'fade-in', 'transparentize', 'adjust-color', 'scale-color', 'change-color', 'ie-hex-str', 'unquote', 'quote', 'str-length', 'str-insert', 'str-index', 'str-slice', 'to-upper-case', 'to-lower-case', 'percentage', 'round', 'ceil', 'floor', 'abs', 'min', 'max', 'random', 'length', 'nth', 'set-nth', 'join', 'append', 'zip', 'index', 'list-separator', 'is-bracketed', 'map-get', 'map-merge', 'map-remove', 'map-keys', 'map-values', 'map-has-key', 'keywords', 'selector-nest', 'selector-append', 'selector-extend', 'selector-replace', 'selector-unify', 'is-superselector', 'simple-selectors', 'selector-parse', 'feature-exists', 'variable-exists', 'global-variable-exists', 'function-exists', 'mixin-exists', 'content-exists', 'inspect', 'type-of', 'unit', 'unitless', 'comparable', 'call', 'get-function', 'if', 'unique-id'];

    pgStylesheet.getAutocompleteArr = function () {
        return SASS_FUNCTIONS
    }

    pgStylesheet.isBuitInFunction = function (functionName) {
        return SASS_FUNCTIONS.indexOf(functionName) > -1;
    }

    pgStylesheet.skipImportFile = function (path) {
        return crsaIsAbsoluteUrl(path) ||
        path.indexOf(this.getDeclCharachter()) > -1 ||
        path.indexOf("#{") > -1;
    }

    pgStylesheet.getRelativeUrlFromImported = function (url, node, withUnderScore) {
        if (url) url = url.trim();
        var filePath = crsaGetUrlAndParamsFromImport(url)[0];

        if (this.skipImportFile(filePath)) return filePath;

        var ext = '';
        if (!filePath.endsWith('.scss')) ext = '.scss';

        var importRelativeUrl = crsaGetBaseForUrl(filePath) + '/' + (withUnderScore ? '_' : '') + crsaGetNameFromUrl(filePath) + ext;

        if (importRelativeUrl[0] == '/') {
            importRelativeUrl = importRelativeUrl.substr(1, importRelativeUrl.length);
        }


        return importRelativeUrl;
    }

    var parseId = 0;

    pgStylesheet.parseSource = function (source, done, addNodeId) {
        this.resetError();

        var parsingPlugin = [];
        var _this = this;
        if (pinegrow.getSetting('disable-formating', "0") == "0") {
            // parsingPlugin.push(stylefmt);
        }

        var pid = parseId++;
        var p = new CrsaProfile('PARSE ' + pid + ' ' + this.url);
        p.show('START');

        var syntaxRes = postcss(parsingPlugin).process(source, { syntax: scssSyntax, parser: scssSyntax, stringifier: scssStringify });
        syntaxRes.then(function (postcssRes) {
            p.show('DONE');

            var rootNode = postcssRes.root;
            _this.resetTreeNodeId();

            if (addNodeId) {
                _this.processNodes(rootNode.nodes);
                var postcssSource = pinegrow.pgPostCSSHelper.getSource(rootNode, scssStringify);
                postcssSource = _this.addSemicolonToSource(postcssSource);
            }

            _this.sourceFormated = false;
            if (done) done(null, postcssSource, rootNode);
            p.show('AFTER DONE');
        }).catch(function(error) {
            //syntax error most likely
            console.log('DONE PARSE with ERROR ' + pid);
            var parsingError = error.reason + ":" + error.line;
            _this.codeError = parsingError;
            if (done) done(parsingError);
        })
/*
        if (syntaxRes.error) {
            console.log('DONE PARSE with ERROR ' + pid);
            var parsingError = syntaxRes.error.reason + ":" + syntaxRes.error.line;
            this.codeError = parsingError;
            if (done) done(parsingError);
        }*/
    }

    pgStylesheet.getSourceForRules = function (node, withParams, forceFormatting) {
        node = node || this.getRootNode();

        if(!node) return '';

        if (forceFormatting || !withParams) {
            this.updateNodeFormatting(node, forceFormatting);
        }

        var sourceWithParams = null;
        if (withParams) sourceWithParams = pinegrow.pgPostCSSHelper.getSource(node, scssStringifyWithParams);
        // Call the normal Stringifier to remove params from nodes
        var source = pinegrow.pgPostCSSHelper.getSource(node, scssStringify);

        if (withParams) return sourceWithParams;
        else return source;
    }

    pgStylesheet.regenerateCssFromScssSource = function (done) {
        var p = new CrsaProfile();
        var _this = this;

        this.sourceToSave = null;

        var source = this.getSourceForRules(null, pgStylesheet.dontAddIds ? false : true);
        if (source == "") {
            if (done) done(null, { css: "" });
            return;
        }

        var options = {
            //data: source,
            //contents: source,
            aaimporter: function (request, prev, done) {
                return _this.importer(request, prev, done);
            },
            importers: [{
                canonicalize(url) {
                    //console.warn(`IMPORTER Can ${url}`)
                    return new URL(url, _this.url);
                },
                load: function(url) {
                    url = url.toString();
                    //console.warn(`IMPORTER Load ${url}`);
                    url = url.replace(/\-pg[0-9]+/, '');

                    var url_no_ = _this.getFullUrlFor(url, false);

                    var cs_list = pinegrow.stylesheets.findCrsaStylesheetForUrl(url_no_);

                    if(cs_list.length === 0) {
                        var url_with_ = _this.getFullUrlFor(url, true);
                        cs_list = pinegrow.stylesheets.findCrsaStylesheetForUrl(url_with_);
                    }

                    if(cs_list.length) {
                        var cs = cs_list[0];
                        var buff = cs.getSourceForRules(null, _this.dontAddIds ? false : true);
                        return  {
                            contents: buff,
                            syntax: 'scss'
                        }
                    } else {
                        return  {
                            contents: '',
                            syntax: 'scss'
                        }
                    }
                }
            }]
            //data: source,
            //contents: source
        }

        if (this.includeSourceMap) {
            if (this.outFile) {
                options['sourceMap'] = true;
                options['outFile'] = this.outFile;
            }
            else {
                throw "You should enter the out file in order to generate source map";
            }
        }

        if(this.render_busy) {
            this.update_on_render_done = function() {
                _this.regenerateCssFromScssSource(done);
            }
        } else {
            this.render_busy = true;
            this.update_on_render_done = null;

            this.recursiveImports = null;

            var onFinished = function() {
                _this.render_busy = false;

                if (_this.update_on_render_done) {
                    var func = _this.update_on_render_done;
                    _this.update_on_render_done = null;
                    func();
                }
            }

            if(!isApp() || online_sass) {
                sass.render(options, function (error, result) {
                    p.show('SASS Compile')
                    _this.renderingFinished(error, result, done);
                    onFinished()
                });
            } else {
                if(this.render_promise) {
                    this.render_promise.cancel();
                    this.render_promise = null;
                }
                this.render_promise = sass.compileStringAsync(source, options)
                    .then(function(result) {
                        p.show('SASS Compile')
                        _this.render_promise = null;
                        _this.renderingFinished(null, result, done);
                        onFinished()
                    })
                    .catch(function(error) {
                        _this.render_promise = null;
                        _this.renderingFinished(error, null, done);
                        onFinished()
                    })
            }
        }
    }

    pgStylesheet.regenerateCssFromCompilableSource = function (done) {
        this.regenerateCssFromScssSource(done);
    }

    pgStylesheet.getNewStylesheet = function () {
        return PgSCSSStylesheet();
    }

    pgStylesheet.getDeclCharachter = function () {
        return "$";
    }

    pgStylesheet.stringifyNode = function (node) {
        return this.addSemicolonToSource(this.filterSource(pinegrow.pgPostCSSHelper.getSource(node, scssStringify)));
    }

    pgStylesheet.superFilter = function (node) {
        return null;
    }

    return pgStylesheet;
}