'use strict';

window.PgStylesPane = function (pgStylesheet, $container, $scrollParent, listOfNodes, options) {
    this.pgStylesheet = pgStylesheet;
    this.filter = "";
    this.itemSelector = 'li[data-node-indices]';
    this.selectingNextElm = false;

    var destroyed = false;

    var colorNamesArr = ["AliceBlue", "AntiqueWhite", "Aqua", "Aquamarine", "Azure", "Beige", "Bisque", "Black", "BlanchedAlmond", "Blue", "BlueViolet", "Brown", "BurlyWood", "CadetBlue", "Chartreuse", "Chocolate", "Coral", "CornflowerBlue", "Cornsilk", "Crimson", "Cyan", "DarkBlue", "DarkCyan", "DarkGoldenRod", "DarkGray", "DarkGreen", "DarkGrey", "DarkKhaki", "DarkMagenta", "DarkOliveGreen", "DarkOrange", "DarkOrchid", "DarkRed", "DarkSalmon", "DarkSeaGreen", "DarkSlateBlue", "DarkSlateGray", "DarkSlateGrey", "DarkTurquoise", "DarkViolet", "DeepPink", "DeepSkyBlue", "DimGray", "DimGrey", "DodgerBlue", "FireBrick", "FloralWhite", "ForestGreen", "Fuchsia", "Gainsboro", "GhostWhite", "Gold", "Goldenrod", "Gray", "Green", "GreenYellow", "Grey", "HoneyDew", "HotPink", "IndianRed", "Indigo", "Ivory", "Khaki", "Lavender", "LavenderBlush", "LawnGreen", "LemonChiffon", "LightBlue", "LightCoral", "LightCyan", "LightGoldenRodYellow", "LightGray", "LightGreen", "LightGrey", "LightPink", "LightSalmon", "LightSeaGreen", "LightSkyBlue", "LightSlateGray", "LightSlateGrey", "LightSteelBlue", "LightYellow", "Lime", "LimeGreen", "Linen", "Magenta", "Maroon", "MediumAquaMarine", "MediumBlue", "MediumOrchid", "MediumPurple", "MediumSeaGreen", "MediumSlateBlue", "MediumSpringGreen", "MediumTurquoise", "MediumVioletRed", "MidnightBlue", "MintCream", "MistyRose", "Moccasin", "NavajoWhite", "Navy", "OldLace", "Olive", "OliveDrab", "Orange", "OrangeRed", "Orchid", "PaleGoldenRod", "PaleGreen", "PaleTurquoise", "PaleVioletRed", "PapayaWhip", "PeachPuff", "Peru", "Pink", "Plum", "PowderBlue", "Purple", "Red", "RosyBrown", "RoyalBlue", "SaddleBrown", "Salmon", "SandyBrown", "SeaGreen", "SeaShell", "Sienna", "Silver", "SkyBlue", "SlateBlue", "SlateGray", "SlateGrey", "Snow", "SpringGreen", "SteelBlue", "Tan", "Teal", "Thistle", "Tomato", "Turquoise", "Violet", "Wheat", "White", "WhiteSmoke", "Yellow", "YellowGreen"]

    var $ruleList = $container.find('.rule-list').data('pg-sp', this);

    var $OldNextli = null;

    var autoCompSCSS = pgStylesheetHelper.getSCSSAutocomleteProps();
    var _this = this;

    var liClicksTime = 0;
    var spectrumActived = false;
    var $activatedColorPicker = null;
    var oldValue = '';

    var propNameArray = [];
    var fileName = pgStylesheet.getUniqueName();

    var setDefaultOptions = function () {
        if (!options) options = {};

        options['uncollapse'] = (options['uncollapse'] == undefined ? false : options['uncollapse']);
        options['uncollapse-on-click'] = options['uncollapse-on-click'] === undefined ? false : options['uncollapse-on-click'];

        options['extra-rule-el-id'] = options['extra-rule-el-id'] || '';

        options['on-node-indices-changed'] = options['on-node-indices-changed'] || null;

        options['on-node-added'] = options['on-node-added'] || null;
        options['on-node-changed'] = options['on-node-changed'] || null;
        options['on-before-node-deleted'] = options['on-before-node-deleted'] || null;
        options['on-edit-finish'] = options['on-edit-finish'] || null;
        options['on-node-replaced'] = options['on-node-replaced'] || null;
        options['on-rule-id-changed'] = options['on-rule-id-changed'] || null;
        options['on-node-uncollapsed'] = options['on-node-uncollapsed'] || null;
        options['on-rule-selected'] = options['on-rule-selected'] || null;

        options['inversed-insert'] = options['inversed-insert'] ? options['inversed-insert'] : false;

        options['is-allowed-insert-node'] = options['is-allowed-insert-node'] || null;

        options['created-by'] = options['created-by'] || '';

        options['skip-showing-comment'] = options['skip-showing-comment'] || false;
        options['skip-showing-decl']    = options['skip-showing-decl']    || false;
        options['skip-showing-atrule']  = options['skip-showing-atrule']  || false;
        options['skip-showing-rule']    = options['skip-showing-rule']    || false;

        options['item-length-to-add-placeholder'] = options['item-length-to-add-placeholder'] || 0;

        options['auth'] = options['auth'] || {};
        options['auth']['can_add_rule']  = options['auth']['can_add_rule']  == false ? false : true;
        options['auth']['can_edit_rule'] = options['auth']['can_edit_rule'] == false ? false : true;

        if(!('handle-element-selected' in options)) options['handle-element-selected'] = true;

        options['show-all'] = options['show-all'] || false;

        options['uncollapse-all'] = options['uncollapse-all'] || false;
        options['read-only'] = options['read-only'] || false;
    }
    setDefaultOptions();

    this.options = options;

    this.itemLengthToAddPlaceholder = options['item-length-to-add-placeholder'];

    this.getIndex = function () {
        return options['extra-rule-el-id'];
    }

    var $ul = $('<ul class="sass-root-node ' + pgStylesheet.getUniqueName() + '-container' + options['extra-rule-el-id'] + '"></ul>').appendTo($ruleList);

    var stylePaneHelper = new PgStylePaneHelper(this, $ul);

    var selectedRule = null;
    var selectedRule$Li = null;

    var markRuleAsSelected = function(node, $li) {
        if(selectedRule != node) {
            if(selectedRule$Li) selectedRule$Li.removeClass('pg-selected-rule');

            if(node) {
                $li.addClass('pg-selected-rule');
            }
            selectedRule = node;
            selectedRule$Li = $li;
        }
    }

    var selectRule = this.selectRule = function (node, $li) {
        var selectedRule = null;

        if (node == null) {
            markRuleAsSelected(node);
            $.fn.crsa('selectRule', null, _this);
        }
        else if (node) {
            if (node.type == "rule") {
                // var rule_obj = getObjectFromRule(node);
                markRuleAsSelected(node, $li);
                $.fn.crsa('selectRule', node, _this);
            }
            else if (node.type == "atrule" && node.name == "media") {
                markRuleAsSelected(null);
                $.fn.crsa('selectRule', null, _this);
            } else {
                selectedRule = pgcssh.getRuleForNode(node);
                if(selectedRule) {
                    var prop = null;
                    if(node.type == 'decl' && node.prop) {
                        var prop = node.prop;
                    }
                    markRuleAsSelected(selectedRule, $li.closest('li.crsa-rule'));
                    $.fn.crsa('selectRule', selectedRule, _this, null, prop);
                }
            }
        }
        if(options['on-rule-selected']) {
            options['on-rule-selected'](selectedRule || node, $li);
        }
    }

    var find$LiForNode = function(node) {
        var cs = pgStylesheetHelper.getStylesheetFromNode(node);
        if(_this.pgStylesheet !== cs) return null;

        var nodeSelector = `[data-node-indices="${pinegrow.pgPostCSSHelper.getNodeIndices(node)}"]`;
        //var nodeSelector =  "#pg-" + cs.getUniqueName() + "-rule-id-" + node['pg-node-id'] + options['extra-rule-el-id'];

        var $li = $ruleList.find('li' + nodeSelector);
        return $li.length ? $li : null;
    }

    //called on crsa-rule-selected
    var onRuleSelected = function(event, obj, skip_scroll) {
        if (!obj) return;

        if (obj.source && obj.source === _this) {
            return; //nothing to do, this panel triggered the event
        }
        if (!obj.rule) {
            //deselect
            markRuleAsSelected(null);
        } else {
            var node = obj.rule;
            var node$LiInThisPane = find$LiForNode(node);

            if (node$LiInThisPane) {
                //mark this rule as selected
                markRuleAsSelected(node, node$LiInThisPane);
                //scroll to node
                if(!skip_scroll) {
                    var offset = pgcssh.getMediaNode(node) ? -30 : 0;
                    //first triggers redraw, offset might change
                    pgScrollToItemInContainer(node$LiInThisPane, $scrollParent, false, false, null, offset);
                    //get closer
                    pgScrollToItemInContainer(node$LiInThisPane, $scrollParent, false, false, null, offset);
                }
            } else {
                //deselect
                markRuleAsSelected(null);
            }
        }
    }

    var sheetBelongsToPgStylesheet = function (currentPgStylesheet, sheet) {
        var parentPgStylesheet = pgStylesheetHelper.getParentStylesheetBySelector(sheet, pinegrow.getSelectedPage());
        var importedStylesheet = pgStylesheetHelper.getImportedStylesheetBySelector(parentPgStylesheet, sheet);

        if (importedStylesheet) parentPgStylesheet = importedStylesheet;

        if (parentPgStylesheet) {
            if (currentPgStylesheet.inline && parentPgStylesheet.inline) {
                var currentInlineSheetPgel = currentPgStylesheet.inlineSheetPgel;
                var parentInlineSheetPgel = parentPgStylesheet.inlineSheetPgel;

                if (currentInlineSheetPgel.getId() == parentInlineSheetPgel.getId()) {
                    return true;
                }
            }
            else if (currentPgStylesheet.getUniqueName() == parentPgStylesheet.getUniqueName()) {
                return true;
            }
        }
        return false;
    }

    var getSelectedElementPgIds = function (done) {
        var selectedElement = pinegrow.getSelectedElement();
        var selectedPage = pinegrow.getSelectedPage();
        if (!selectedElement || !selectedPage) {
            if (done) done([]);
            return;
        }

        var $el = selectedElement.get$DOMElement();

        if(!$el) {
            if (done) done([]);
            return;
        }

        var el = $el.get(0);

        var pgIdsAndFiles = [];

        var highlightedRules = function (nodes_to_show, ignored_matched_rules) {
            var emptyNodes = pgStylesheet.emptyNodes;
            for (var i = 0; i < emptyNodes.length; i++) {
                var emptyNode = emptyNodes[i];
                var pfIdAndFile = emptyNode.nodes;
                try {
                    if (el.matches(emptyNode.selector)) {

                        var selector = {
                            'id': emptyNode['pg-node-id']
                        };

                        if (pgStylesheet.type != "css" && pgStylesheet.imported) {
                            if (pfIdAndFile.length > 1) {
                                selector['file'] = pfIdAndFile[0].value.replace(/[\'\"]/g, '');
                            }
                        }

                        pgIdsAndFiles.push(selector);
                    }
                }
                catch(err) {}
            }

            if (pgStylesheet.ignore) {
                for (var i = 0; i < ignored_matched_rules.length; i++) {
                    var iMatchedRules = ignored_matched_rules[i];
                    if (iMatchedRules.pgStylesheet.getUniqueName() == pgStylesheet.getUniqueName()) {
                        var matched = iMatchedRules.matched_rule;

                        var item = {};
                        var pgIdFound = false;
                        var selectors = matched.selectors;
                        for (var j = 0; j < selectors.length; j++) {
                            var selector = selectors[j].text;

                            if (!pgIdFound) {
                                var matchedPgId = selector.match(/\.pg-node-id-([0-9]+)/);
                                if (matchedPgId && matchedPgId.length > 1) {
                                    item['id'] = matchedPgId[1];
                                    pgIdFound = true;
                                    continue;
                                }
                            }

                            var matchedPgFile = selector.match(/\.pg-file-(.*)/);
                            if (matchedPgFile && matchedPgFile.length > 1) {
                                item['file'] = matchedPgFile[1];
                            }
                        }

                        if (pgIdFound) pgIdsAndFiles.push(item);
                    }
                }
            }
            else {
                for (var i = 0; i < nodes_to_show.length; i++) {
                    var nToShow = nodes_to_show[i];
                    if (nToShow.pgStylesheet.getUniqueName() == pgStylesheet.getUniqueName()) {
                        var node = nToShow.node;

                        pgIdsAndFiles.push({
                            file: node['pg-file-name'],
                            id: node['pg-node-id']
                        });
                    }
                }
            }

            if (done) done(pgIdsAndFiles);
        }

        var activeNodesSet = function (active_nodes) {
            if (active_nodes) {
                var nodes_to_show = active_nodes.nodes_to_show;
                var ignored_matched_rules = active_nodes.ignored_matched_rules;
                highlightedRules(nodes_to_show, ignored_matched_rules);
                return true;
            }
            return false;
        }

        if (!activeNodesSet(pinegrow.pgActiveRules.getActiveNodes())) {
            pinegrow.pgActiveRules.updateMatchedRules(false, function () {
                /**
                 * This is a async code.
                 * The callback could be called after stylepane get destroyed.
                 */
                if (!destroyed) {
                    activeNodesSet(pinegrow.pgActiveRules.getActiveNodes());
                }
            });
        }
    }

    var updateHighlightedRulesWithSelectedElm = function () {
        $container.find('li.selected').removeClass('selected');

        getSelectedElementPgIds(function (ids) {
            var domId = ids.map(function (idAndFile) {
                if (pgStylesheet.imported && !idAndFile['file']) return false;

                return "#pg-" + (idAndFile['file'] ? idAndFile['file'] : fileName) + "-rule-id-" + idAndFile['id'] + options['extra-rule-el-id'];
            }).join(',');
            $ul.find(domId).addClass('selected');
        });
    }

    var selectNextElm = function (elm) {
        var checkAllNext = function ($elm, lookIntoNested) {
            var found = false;
            if (lookIntoNested) {
                var $nested = $(elm).parent().next();
                if ($nested.is('ul')) {
                    var $foundElement = $nested.find('> li:first-child > div > .crsa-scss-prop').first();
                    if ($foundElement.length > 0) {
                        stylePaneHelper.editField($foundElement);
                        found = true;
                    }
                }
            }
            if (!found) {
                if ($elm.is('li')) {
                    var $next = $elm.next();
                    if ($next.length == 0) {
                        stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($elm);
                    }
                    else {
                        stylePaneHelper.editField($next.find(' > div > .crsa-scss-prop').first());
                    }
                }
                else {
                    checkAllNext($elm.closest('li'), false);
                }
            }
        }

        _this.selectingNextElm = true;
        var $next = $(elm).parent().find('> span.crsa-scss-prop').last();
        if ($next.length > 0 && elm != $next[0]) {
            stylePaneHelper.editField($next);
        }
        else {
            checkAllNext($(elm), true);
        }
    }

    var selectPrevElm = function (elm) {
        var checkAllPrev = function ($elm) {
            if ($elm.is('li')) {
                var found = false;
                var $prev = $elm.prev();
                if ($prev.length == 0) {
                    var $nested = $elm.parent('ul');
                    if ($nested.length > 0) {
                        var $foundElement = $nested.siblings('div').find('> .crsa-scss-prop').last();
                        if ($foundElement.length > 0) {
                            stylePaneHelper.editField($foundElement);
                            found = true;
                        }
                        else if ($elm.parent().is('ul.sass-root-node')) {
                            $prev = $elm.parent().find('> li:last-child');
                            stylePaneHelper.editField($prev.find(' > div > .crsa-scss-prop').last());
                        }
                    }
                }

                if (!found) {
                    var $nested = $prev.find('> ul');
                    if ($nested.length > 0) {
                        stylePaneHelper.editField($nested.find('> li:last > div > .crsa-scss-prop').last());
                    }
                    else {
                        stylePaneHelper.editField($prev.find(' > div > .crsa-scss-prop').last());
                    }
                }
            }
            else {
                checkAllPrev($elm.closest('li'));
            }
        }

        _this.selectingNextElm = true;
        var $prev = $(elm).parent().find('> span.crsa-scss-prop').first();
        if ($prev.length > 0 && $prev[0] != elm) {
            stylePaneHelper.editField($prev);
        }
        else {
            checkAllPrev($(elm));
        }
    }

    var addSpectrumSpanToText = function (str) {
        var replaceWith = '$1<span style="margin-top:0;"><input class="color-input" value="$2"><span class="more-text">$2</span></span>';

        var str = str.replace(/(\#[0-9a-fA-F]{3,6})/g, '<span style="margin-top:0;"><input class="color-input" value="$1"><span class="more-text">$1</span></span>');
        str = str.replace(/(rgb\(\d{1,3}\s*,\s*\d{1,3}\s*,\s*\d{1,3}\s*\))/g, '<span style="margin-top:0;"><input class="color-input" value="$1"><span class="more-text">$1</span></span>');
        str = str.replace(/(rgba\(\d{1,3}\s*,\s*\d{1,3}\s*,\s*\d{1,3}\s*,\s*[0-1]?(?:\.[0-9]+)*\s*\))/g, '<span style="margin-top:0;"><input class="color-input" value="$1"><span class="more-text">$1</span></span>');

        var colorNamesReg = new RegExp("(^|\\s|\\(|\\,)(" + colorNamesArr.join("|") + ")\\b", 'gi');
        str = str.replace(colorNamesReg, replaceWith);

        return str;
    }

    var addSpectrumSpan = function ($el) {
        var text = getTextInsideElm($el);
        var str = addSpectrumSpanToText(text);

        if (str.length != text.length) $el[0].innerHTML = str;
        return str;
    }

    var updateContectedNode = function ($mainVar, value) {
        var connectedNodes = $mainVar.data('connected-node');
        if (connectedNodes) {
            for (var i = 0; i < connectedNodes.length; i++) {
                var varName = connectedNodes[i].varName;
                var varDeclName = pgStylesheet.getDeclCharachter() + varName;
                var $spectrumInput = connectedNodes[i].el.find('.input-color-for-' + varName);

                var $moreText = $spectrumInput.siblings('.more-text');

                if ($moreText.html() == varDeclName) {
                    var $placeholderColor = $spectrumInput.siblings('.color-value-placeholder');
                    $placeholderColor.css('background-color', value);
                    // $spectrumInput.spectrum("set", value);
                }
            }
        }
    }

    var activateSpectrum = function ($declValue) {
        var $colorPicker = null, canvasData = null, width = null, oldColor = null;
        var ctx = null;

        var updateChanges = function ($el, changeSaved, skipUpdatingSource) {
            var hexColor = $el.val();
            var declData = $el.data('decl');
            var $mainVar = null;
            if (declData) {
                var node = declData.node;
                //node.value = hexColor;
                pgcssh.setNodeProperty(node, 'value', hexColor)

                var $li = $el.data('decl').el;
                var $declValue = $li.find('.decl-value');
                $mainVar = $declValue;
                var $currentColorValue = $declValue.find('.more-text');
                $currentColorValue.text(hexColor);

                var $spectrumInput = $li.find('.decl-value .color-input');
                $spectrumInput.spectrum("set", $el.val());

                if (!skipUpdatingSource) {
                    setChangesForNodeProp($declValue[0], false, false, function () {
                        var $selectedRule = $ul.find('.pg-selected-rule');
                    });
                }

                updateContectedNode($mainVar, hexColor);
            }
            else {
                var spectrum = $el.spectrum('get');
                if (!spectrum.isValid()) return;

                var $currentColorValue = $el.siblings('.more-text');
                var color = getNewColor(spectrum);
                $currentColorValue.text(color);

                var $decVal = $el.closest('.decl-value,.atrule-params');
                $mainVar = $decVal;

                var $placeholderColor = $el.siblings('.color-value-placeholder');
                $placeholderColor.css('background-color', color);

                if (!skipUpdatingSource) {
                    setChangesForNodeProp($decVal[0], false, false, function () {
                        var $selectedRule = $ul.find('.pg-selected-rule');
                    }, !changeSaved);
                }

                updateContectedNode($mainVar, color);
            }
        }

        var colorPicked = function ($el) {
            var $ruleList = $el.closest('.rule-list');
            var stylePane = $ruleList.data('pgSp');
            var pgStylesheet = stylePane.pgStylesheet;

            var $li = $el.closest('li')
            var node = stylePane.getNodeFrom($li);
            if (node) {
                var oldValue = $li.data('node-last-value');
                pinegrow.dispatchEvent('on_css_node_updated', null, {
                    node: node,
                    stylesheet: pgStylesheet,
                    oldValue: oldValue
                });
                // $('body').trigger('crsa-variables-changed', { list: [pgStylesheet] });
            }
        }

        function rgbToHex(r, g, b) {
            if (r > 255 || g > 255 || b > 255)
                throw "Invalid color component";
            return ((r << 16) | (g << 8) | b).toString(16);
        }

        var getNewColor = function (color) {
            if (color.toRgb()['a'] == 1) return color.toHexString();
            else return color.toRgbString();
        }

        var colorPickerTimeout = null;
        var addSpectrumTo = function ($pickerInputs) {
            var spectrumHidden = function ($colorPicker) {
                setTimeout(function () {
                    try {
                        var $placeholderColorValue = $colorPicker.siblings('.color-value-placeholder');
                        $placeholderColorValue.show();

                        $colorPicker.spectrum('destroy');
                        $colorPicker.hide();
                    } catch (err) {}
                }, 10)
            }

            var addSpectrum = function ($colorPicker) {
                $colorPicker.spectrum({
                    clickoutFiresChange: true,
                    showInitial: true,
                    allowEmpty: true,
                    preferredFormat: "hex",
                    appendTo: $colorPicker.closest('body'),
                    showAlpha: true,
                    color: $colorPicker.val(),
                    showInput: true,
                    change: function (color) {
                        var $li = $(this).closest('li');
                        var node = _this.getNodeFrom($li);

                        var oldValue = $li.data('node-last-value');

                        // var oldValue = {
                        //     value: this.valueBeforeColorPickerChange || node.value,
                        //     prop: node.prop
                        // }

                        $(this).val(getNewColor(color));
                        updateChanges($(this), true);
                        if (options['on-edit-finish']) options['on-edit-finish'](node, pgStylesheet, _this);

                        pinegrow.dispatchEvent('on_css_node_updated', null, {
                            node: node,
                            stylesheet: pgStylesheet,
                            oldValue: oldValue,
                            undoCombineKey: 'colorPicked_' + node.prop,
                            action: 'Set ' + node.prop,
                            undoRuleName: null
                        });
                    },
                    move: function (color) {
                        var _curr_this = this;

                        if (colorPickerTimeout !== null) {
                            updateChanges($(_curr_this), false, true);
                            clearTimeout(colorPickerTimeout);
                        }
                        colorPickerTimeout = setTimeout(function () {

                            $(_curr_this).val(getNewColor(color));
                            updateChanges($(_curr_this));
                        }, 400);
                    },
                    show: function (color) {
                        spectrumActived = true;
                        $colorPicker = $(this);
                        $activatedColorPicker = $colorPicker
                        oldColor = getNewColor($colorPicker.spectrum('get'));

                        var page = pinegrow.getSelectedPage();
                        if (!page) return;

                        $ul.data('properties-in-edit-mode', true);

                        var $li = $(this).closest('li');
                        var node = _this.getNodeFrom($li);
                        if (node) {
                            this.valueBeforeColorPickerChange = node.value;
                            var value = pgcssh.getValuesFromNode(node);
                            $li.data('node-last-value', value);
                        }
                    },
                    hide: function (color) {
                        if ($(this).data('just-opened')) {
                            $(this).spectrum('show');
                        } else {

                            var isMasterColor = $(this).hasClass('master-color');
                            if (isMasterColor) $(this).next().hide();

                            spectrumActived = false;
                            $activatedColorPicker = null;

                            if (getNewColor(color) == oldColor) {
                                updateChanges($(this));
                            }
                        }
                        $(this).data('just-opened', false);
                        $ul.data('properties-in-edit-mode', false);

                        spectrumHidden($colorPicker);
                    }
                });

                // $colorPicker.on('input change', function (e) {
                //     $(this).spectrum("set", $(this).val());
                // });

                var pgSpectrumHelper = new PgSpectrumHelper($colorPicker, {
                    'mouse-moved': updateChanges,
                    'color-picked': colorPicked
                });
            }

            $.each($pickerInputs, function (index, elm) {
                $colorPicker = $(elm);
                $colorPicker.hide();

                var $placeholderColor = $('<div class="color-value-placeholder"></div>');
                $placeholderColor.css('background-color', $colorPicker.val());
                $placeholderColor.insertAfter($colorPicker);

                $placeholderColor.click(function (e) {
                    e.preventDefault();

                    if(options['read-only']) {
                        var $li = $(this).closest('li');
                        var node = _this.getNodeFrom($li);
                        selectRule(node, $li);

                    } else if(pgStylesheet.canEdit(false, true)) {
                        $placeholderColor.hide();
                        var $colorPicker = $placeholderColor.siblings('.color-input');
                        $colorPicker.val($placeholderColor.css('background-color'));

                        addSpectrum($colorPicker);
                        $colorPicker.spectrum("show");
                    }
                    return false;
                });
            });
        }

        addSpectrumTo($declValue.find('.color-input'));
    }

    var addSpectrumToDecl = function ($declVals) {
        $declVals.each(function (i, el) {
            var $el = $(el);
            var textValue = getTextInsideElm($el);
            var str = addSpectrumSpanToText(textValue);

            var declProps = {};
            if (pgStylesheet.fileType == "scss" || pgStylesheet.fileType == "less") {
                declProps = searchForDeclProp($el);
            }

            str = str.replace(/(\$[A-Za-z0-9\_\-]*)/g, function(match) {
                if (declProps[match]) {
                    var $li = declProps[match].el;
                    if ($li.find('.color-input').length == 1 && getTextInsideElm(declProps[match].el.find('.decl-value')).trim() == declProps[match].el.find('.more-text').text().trim()) {
                        var connectedNodeEl = declProps[match].el.find('.decl-value').data('connected-node');

                        var varName = match.substr(1, match.length)
                        if (!connectedNodeEl) connectedNodeEl = [];
                        connectedNodeEl.push({
                            el: $el,
                            varName: varName
                        });

                        declProps[match].el.find('.decl-value').data('connected-node', connectedNodeEl);
                        return '<span class="color-from-decl-prop"><input class="color-input input-color-for-' + varName + '" value="' + declProps[match].node.value + '"><span class="more-text">' + match + '</span></span>';
                    }
                    else {
                        return match;
                    }
                }
                else {
                    return match;
                }
            });

            if (str.length != textValue.length) $el[0].innerHTML = str; // Change it if color pickers inserted

            var $colorFromDeclProp = $el.find('.color-from-decl-prop');
            if ($colorFromDeclProp.length > 0) {
                var $input = $colorFromDeclProp.find('> .color-input');
                var $moreText = $colorFromDeclProp.find('> .more-text');
                var declPropName = $moreText.text();
                $moreText.data('master-color', declProps[declPropName]);
            }
        })
    }

    var updateColorPicker = function ($el) {
        addSpectrumToDecl($el);
        activateSpectrum($el);
    }

    var selectDeclParentRule = function (node, $li) {
        if (node.type == "decl") {
            var parent = node.parent;
            if (parent && parent.type == "rule") {
                selectRule(parent, $li.closest('ul').parent());
            }
        }
    }

    this.blurCallback = function (e) {
        var target = e.target;
        var $target = $(target);
        liClicksTime = 0;

        var node = this.getNodeFrom($target.closest('li'));
        if (node.type == 'decl' && $target.data('prop-name') == 'value' ||
            node.type == 'atrule' && $target.data('prop-name') == 'params') {
            updateColorPicker($target);
        }

        if ($target.hasClass('simple-autocomplete-input')) {
            // $target.autocomplete( "close" );
            var simpleAutocomplete = $target.data('simple-autocomplete');
            simpleAutocomplete.close();
        }

        if (isImportedNode(node)) {
            if (pgStylesheet.fileType == 'css') updateAllStylesheetPage(pgStylesheet);
        }
        // var rule = pinegrow.pgPostCSSHelper.getRuleForNode(node);
        // selectRule(rule);
    }

    this.editFinishedCallback = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('li');

        if (!$li.data('add-new-node')) {
            var node = this.getNodeFrom($li);
            if (!node) return; // node deleted

            if (options['on-edit-finish']) options['on-edit-finish'](node, pgStylesheet, _this);

            // selectDeclParentRule(node, $li);
            var oldValue = $li.data('node-last-value');

            pinegrow.dispatchEvent('on_css_node_updated', null, {
                node: node,
                stylesheet: pgStylesheet,
                // oldValue: pgcssh.getNodeOldValues(node),
                oldValue: oldValue,
                undoCombineKey: 'editFinishedCallback_' + (node.prop || ''),
                action: 'Set ' + (node.prop || 'CSS property')
            });

            if ($target.hasClass('decl-value')) {
                updateContectedNode($target, node.value);
            }
        }

        if(pgStylesheet) {
            var error = pgStylesheet.getCascadeError();
            if (error) pinegrow.showQuickError(error.message ? error.message : error);
        }
    }

    this.keyupCallback = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('li');
        var node = this.getNodeFrom($li);
        if (!node) return true;

        // if (!isImportedNode(node)) setChangesForNodeProp(e.target, true, false, null, true);
        setChangesForNodeProp(e.target, true, false, null, true);
        return false;
    }

    var showAddingImportNotice = function () {
        pinegrow.showQuickMessage('You can\'t add <code>@import</code> here', 4000);
    }

    var isImportedNode = function (node) {
        return (node && (node.type == "atrule" || node.type == "import") && node.name == "import");
    }

    var isMediaNode = function (node) {
        return (node && node.type == "atrule" && node.name == "media");
    }

    var isSupportsNode = function (node) {
        return (node && node.type == "atrule" && node.name == "supports");
    }

    var updatePgScssCodeUsingRules = function (done, skipVarsUpdating, singleChangedRule, action, affectedNode) {
        if (!pgStylesheet) return;
        if(singleChangedRule) {
            if(pgStylesheet.updateSingleRule) {
                var node_param = null;
                var value_param = null;
                if(action === 'delete' && affectedNode && affectedNode.type === 'decl') {
                    node_param = affectedNode.prop;
                }
                pgStylesheet.updateSingleRule(singleChangedRule, function() {
                    if (done) done(pgStylesheet.getRootNode());
                }, node_param, value_param);

                return;
            }
        }
        pgStylesheet.setSourceForRules(pgStylesheet.rootNode, function () {
            if (done) done(pgStylesheet.getRootNode());

            // if (!skipVarsUpdating) {
            //     $('body').trigger('crsa-variables-changed', { list: [pgStylesheet] });
            // }
            //didMakeChange(selectedPage);todo30 ???
        }, false, true /* fastDelayed */);
    }

    var getTextInsideElm = function ($el, notSelector) {
        var notSelector = notSelector || '.sp-replacer';
        var text = $el.clone().find(notSelector).remove().end().text() || '';
        return text.trim();
    }

    var setChangesForNodeProp = function (elm, skipElmUpdating, updateNodesSelectorIds, done, skipUpdatingVars, action, affectedNode) {
        var $el = $(elm).closest('.crsa-scss-prop');
        if ($el.length == 0) $el = $(elm);

        var $li = $el.closest('li');
        var currentNode = _this.getNodeFrom($li);

        var callback = function (action, affectedNode) {
            oldValue = oldValue || $(elm).data('old-text');
            var propName = $(elm).data('prop-name');

            if ($li.attr('data-removed') != "true") {
                if (options['on-node-changed']) options['on-node-changed'](currentNode, pgStylesheet, oldValue, propName, _this);
            }

            var singleRule = null;
            if(currentNode && currentNode.type == 'decl') {
                singleRule = pgcssh.getRuleForNode(currentNode);
            }

            updatePgScssCodeUsingRules(function (resRootNode) {
                if (updateNodesSelectorIds) {
                    updateNodeSelector(pgStylesheet.rootNode.nodes);
                    if ($li.attr('data-removed') != "true") {
                        updateNodesDomIds($li);
                    }

                    updateHighlightedRulesWithSelectedElm();
                }

                // if (isImportedNode(currentNode) && !$el.data('old-text')) {
                //     pgStylesheet.addImportedFile(currentNode, function () {
                //        if (done) done();
                //     });
                // }
                // else {
                if (done) done();
                // }

                // if (pgStylesheet.fileType == 'css') pinegrow.refreshAllPages();

                // if (done) done();
            }, skipUpdatingVars, singleRule, action, affectedNode);
        }

        var willBeUpdated = false;
        if (currentNode && $li.attr('data-removed') != "true") {
            var $moreText = $(elm).find('more-text');
            var text = getTextInsideElm($el);
            var ifAtrule = (currentNode.type == "atrule" || currentNode.type == "import") && text[0] == "@";
            var ifMixin =  pgStylesheet.fileType == 'less' && (currentNode.type == "atrule" && currentNode.mixin);

            if (ifAtrule || (ifMixin && $el.data('prop-name') == "name")) {
                text = text.substr(1, text.length);
            }

            if (currentNode.type == "decl" || (currentNode.type == "atrule" && $el.data('prop-name') == "value")) {
                var textChanged = false;
                if (text.indexOf(';') > -1) {
                    text = text.replace(/\;/g, '');
                    textChanged = true;
                }
                if (text[text.length-1] == ":") {
                    text = text.substr(0, text.length-1);
                    textChanged = true;
                }

                if (textChanged && !skipElmUpdating) {
                    $el.text(text);
                    updateColorPicker($el);
                }
            }
            if ($el.data('prop-name') == "value" && currentNode.important) currentNode.important = false;

            if (currentNode[$el.data('prop-name')] == text) return;
            currentNode[$el.data('prop-name')] = text;

            // if (isImportedNode(currentNode)) {
            //     if (!$el.data('old-text')) {
            //         willBeUpdated = true;
            //         pgStylesheet.addImportedFile(currentNode, function () {
            //             callback();
            //             // $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
            //             // pinegrow.refreshAllPages();
            //         });
            //     }
            //     // else {
            //     //     $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
            //     //     pinegrow.refreshAllPages();
            //     // }
            // }
        }
        if (!willBeUpdated) callback(action, affectedNode);

        oldValue = getTextInsideElm($el);
    }

    var moveKeyClicked = function ($el, increaseBy) {
        var startIndex = 0, endIndex = 0;
        var spanText = $el.text();
        var index = crsaGetCaretPosition($el[0])[0];
        var currentIndex = index;
        var text = "";

        currentIndex--;
        while (currentIndex >= 0 && currentIndex < spanText.length && spanText[currentIndex] != " ") {
            text = spanText[currentIndex] + text;
            currentIndex--;
        }
        startIndex = currentIndex + 1;

        currentIndex = index;
        while (currentIndex >= 0 && currentIndex < spanText.length && spanText[currentIndex] != " ") {
            text = text + spanText[currentIndex];
            currentIndex++;
        }
        endIndex = currentIndex;

        var currentNo = text.match(/^\-?[0-9]+/);
        if (currentNo && currentNo.length > 0) {
            var no = parseInt(currentNo[0]);
            var newNo = no + increaseBy;
            var unitAdded = false;

            var noRegex = new RegExp('^' + no);
            var newText = text.replace(noRegex, newNo);

            if (currentNo[0].length == text.length) {
                newText += ('px');
                unitAdded = true;
            }

            $el.text(spanText.substring(0, startIndex) + newText + spanText.substring(endIndex));

            if (unitAdded) endIndex += 2;

            if (currentNo[0].length > (newNo+'').length) endIndex--;
            else if (currentNo[0].length < (newNo+'').length) endIndex++;

            crsaSetCaretPosition($el[0], startIndex, endIndex);
            return true;
        }
        return false;
    }

    var getSecondVal = function (node) {
        if (node.type == "decl")    return node.value;
        else if (node.type == "atrule")  return node.param;
        else if (node.type == "rule")    return null;
        else if (node.type == "comment") return null;
    }

    this.keydownCallback = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('li');

        var code = e.keyCode || e.which;
        if (code == 13 || code == 9) {
            var propName = $target.data('prop-name');
            if (propName == 'value') {
                if (pgStylesheet.isVarValue($target.html())) {
                    // pinegrow.showQuickMessage('Can\'t use var for css file');
                    return false;
                }
            }

            if ($target.is('[contenteditable]')) {
                var node = this.getNodeFrom($li);
                if ($li.data('add-new-node')) {
                    $li.data('add-new-node', false);
                    stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
                }
                else {
                    if (e.shiftKey) selectPrevElm(e.target);
                    else            selectNextElm(e.target);
                }
            }

            return true;
        }
        if (code == 38) {
            e.preventDefault();
            var increaseBy = 1;
            if (e.shiftKey) increaseBy = 10;
            return moveKeyClicked($target, increaseBy);
        }
        if (code == 40) {
            e.preventDefault();
            var increaseBy = -1;
            if (e.shiftKey) increaseBy = -10;
            return moveKeyClicked($target, increaseBy);
        }
    }

    var updateAllStylesheetPage = function (currPgStylesheet) {
        var pages = pinegrow.stylesheets.getPagesForStylesheet(currPgStylesheet);
        for (var i = 0; i < pages.length; i++) {
            pages[i].refresh();
        }
    }

    var removeNodeFromParent = function (node) {
        if (!node) return;
        var parentNodes = node.parent.nodes;
        var index = parentNodes.indexOf(node);
        if (index > -1) {
            parentNodes.splice(index, 1);
            if (isImportedNode(node)) {
                pgStylesheet.removeImportedFile(node);
                // $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
                // pinegrow.refreshAllPages();

                if (pgStylesheet.fileType == 'css') updateAllStylesheetPage(pgStylesheet);
            }
        }
        return true;
    }

    this.focusCallback = function (e) {
        if(options['read-only']) return;

        oldValue = '';
        var $el = $(e.target);
        var node = this.getNodeFrom($el.closest('li'));

        if (!$el.hasClass('simple-autocomplete-input')) {
            if ($el.data('prop-name') == "prop") {
                addAutoPropCompeleteTo($el);
            }
            else if ($el.data('prop-name') == "value") {
                addAutoValueCompeleteTo($el, true)
            }
        }

        var parentNode = null;

        if (node) {
            if (node.type == "rule") parentNode = node;
            else                     parentNode = node.parent;
        }
        else {
            var $nodeItem = $el.closest('li').parent().closest('li');
            parentNode = this.getNodeFrom($nodeItem);
        }

        // selectRule(parentNode);
    }

    this.dblClickCallback = function (e) {
        var $el = $(e.target);
        var node = this.getNodeFrom($el.closest('li'));

        // selectRule(node);
    }

    var removeNode$li = function ($li) {
        var $parent = $li.parent();
        pgRemoveSpectrum($li, '.color-input');
        $li.remove();

        if (!$parent.hasClass('sass-root-node')) {
            tryToAddPlaceholderNode($parent);
        }
    }

    this.removeNode = function ($li, skipSetChanges, deleteNodeLi, done) {
        var updateNodesSelectorIds = false;
        var node = _this.getNodeFrom($li);
        // if (node.type != 'decl') {
        //     selectRule(null);
        // }

        $ul.data('properties-in-edit-mode', true);
        $li.attr('data-removed', "true");

        pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: pgStylesheet});
        if (options['on-before-node-deleted']) options['on-before-node-deleted'](node, pgStylesheet, _this);
        updateNodesDomIds($li);

        if (removeNodeFromParent(node)) {
            // selectDeclParentRule(node, $li);
            var backward = $li.parent().hasClass('sass-root-node') && options['inversed-insert'];

            updateNodeIndicesAfter($li, node, -1, backward);
            updateNodesSelectorIds = true;
        }

        $ul.data('properties-in-edit-mode', false);
        if (!skipSetChanges) {
            setChangesForNodeProp($li, false, updateNodesSelectorIds, function () {
                if (updateNodesSelectorIds) {
                    removeNode$li($li);
                    stylePaneHelper.onNodeRemoved();
                }

                pinegrow.dispatchEvent('on_css_node_deleted', null, {node: node, stylesheet: pgStylesheet});
                if (done) done();
            }, null, 'delete', node);
        }
        if (deleteNodeLi) {
            removeNode$li($li);
        }

        return updateNodesSelectorIds;
    }

    this.nodeEditDone = function (elm, liRemoved) {
        var updateNodesSelectorIds = false;
        var node = null;

        if ($(elm).text() == "") {
            var $li = $(elm).closest('li');
            node = _this.getNodeFrom($li);
            updateNodesSelectorIds = this.removeNode($li, true);
        }
        setChangesForNodeProp(elm, false, updateNodesSelectorIds, function () {
            if (updateNodesSelectorIds) {
                // $li.remove();
                pinegrow.dispatchEvent('on_css_node_deleted', null, {node: node, stylesheet: pgStylesheet});

                removeNode$li($li);
                stylePaneHelper.onListUpdated();
            }
            else {
                var $selectedRule = $ul.find('.pg-selected-rule');
                // selectRule(_this.getNodeFrom($selectedRule), $selectedRule);
            }
        });
    }

    this.recompileCode = function (done) {
        updatePgScssCodeUsingRules(done);
    }

    var addPlaceholderNode = function ($ul) {
        return $('<li class="placeholder-node crsa-node" draggable="true"></li>').appendTo($ul);
    }

    var addNestedUl = function ($li) {
        $li.addClass('has-children')
        return $('<ul class="nested"></ul>').appendTo($li);
    }

    var tryToRemovePlaceholderNode = function ($li) {
        var $ul = $li.closest('ul');
        var $parentLi = $ul.parent();

        if ($parentLi.hasClass('crsa-node')) {
            var $placeholderLi = $ul.find('> .placeholder-node');
            if ($placeholderLi.length > 0) $placeholderLi.remove();
        }
    }

    this.addNewNodeUI = function ($el, helperNode, type) {
        var $li = $el.closest('li');
        if (!type && $li.parent().hasClass('sass-root-node')) {
            helperNode = this.pgStylesheet.rootNode;
        }

        var newNode = pinegrow.pgPostCSSHelper.getNodeFromText($el.text(), helperNode, type, this.pgStylesheet);
        if (!newNode) return;
        var secondEditCallback = null;

        tryToRemovePlaceholderNode($li);
        $el.addClass('crsa-scss-prop');
        $li.attr('draggable', "true");
        $li.data('placeholder-item', false);
        $li.data('add-new-node', true);
        $li.data('node', newNode);
        $li.addClass('crsa-node crsa-' + newNode.type);
        if (newNode.type != 'rule' || (newNode.type == 'rule' && options['auth']['can_edit_rule'])) {
            $li.addClass('has-options');
        }

        var addCollapseIcon = function ($div) {
            $div.prepend('<i class="collapse-icon icon-down"></i>');
            //$div.prepend('<i class="fa fa-caret-down"></i>');
        }

        var addDeclValue = function ($declProp) {
            var $declValue = $('<span class="crsa-scss-prop decl-value"></span>').insertAfter($declProp);
            $('<span>:&nbsp;</span>').insertBefore($declValue);
            // addAutoValueCompeleteTo($declValue);

            return $declValue;
        }

        var addAtRuleParams = function ($atRuleName) {
            var $atRuleParams = $('<span class="crsa-scss-prop atrule-params"></span>').insertAfter($atRuleName);
            $('<span>&nbsp;</span>').insertBefore($atRuleParams);
            return $atRuleParams;
        }

        var addImportPath = function ($importName) {
            var $importPath = $('<span class="crsa-scss-prop import-importPath"></span>').insertAfter($importName);
            $('<span>&nbsp;</span>').insertBefore($importPath);
            return $importPath;
        }

        var addTheNewNode = function ($li, $elm, done) {
            var updateStylePane = insertNodeIntoParent($li);

            var newNode = _this.getNodeFrom($li);
            var nodeIndices = $li.attr('data-node-indices');
            setUncollapsedNodes(newNode, true, true);

            var callback = function () {
                var updateSource = function () {
                    updatePgScssCodeUsingRules(function (resRootNode) {
                        // if (updateStylePane) pinegrow.dispatchEvent('on_css_source_changed', null, {list: [pgStylesheet]});

                        updateNodeSelector(pgStylesheet.rootNode.nodes);
                        if ($li.attr('data-removed') != "true") {
                            updateNodesDomIds($li);
                        }

                        updateHighlightedRulesWithSelectedElm();

                        newNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, nodeIndices.split(','));

                        if (options['on-node-added']) options['on-node-added'](newNode, pgStylesheet, _this);
                        pinegrow.dispatchEvent('on_css_node_added', null, {node: newNode, stylesheet: pgStylesheet});

                        if (done) done();
                    });
                }

                if (isImportedNode(newNode)) {
                    pgStylesheet.addImportedFile(newNode, function (recursiveImports) {
                        if (!recursiveImports) updateSource();
                    });
                } else {
                    updateSource();
                }
            }
            callback();

            // if (isImportedNode(newNode)) {
            //     pgStylesheet.addImportedFile(newNode, function () {
            //         callback();
            //     });
            // }
            // else {
            //     callback();
            // }
        }

        var editSecondAttr = function (e, elm, callback) {
            if (!$(elm).attr('contenteditable')) return;

            var $li = $(elm).closest('li');
            var node = $li.data('node');

            var code = e.keyCode || e.which;
            if (code == '13' || code == 9) {
                e.preventDefault();
                $(elm).off('keydown', secondKeydownTrigger);
                node[$(elm).data('prop-name')] = $(elm).text();

                if (callback) callback($li);
                addTheNewNode($li, $(elm), function () {
                    stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
                });
            }
            if (code == 27) {
                e.preventDefault();
                $(elm).closest('li').remove();
            }
        }

        var secondKeydownTrigger = function (e) {
            editSecondAttr(e, this, secondEditCallback);
        }

        var firstValueWritten = function ($first, $second, callback) {
            $first.removeAttr('contenteditable');
            $second.attr('contenteditable', 'true');
            crsaSelectText($second[0]);

            secondEditCallback = callback;
            $second.on('keydown', secondKeydownTrigger);
        }

        if (newNode.type == "decl") {
            $el.addClass('decl-prop');
            $el.data('prop-name', 'prop');
            // var propvalArr = newNode.prop.split(':');
            var styleObj = styleToObject(newNode.prop);
            var propvalArr = Object.keys(styleObj);

            $el.off('keydown').off('blur');

            var $declValue = addDeclValue($el);
            $declValue.data('prop-name', 'value');

            $el.removeAttr('contenteditable');

            // addAutoPropCompeleteTo($el);

            addAutoValueCompeleteTo($declValue);
            var fullNodeAdded = false;

            var nodeSet = false;
            if (propvalArr.length > 0) {
                var propKey = propvalArr[0];
                var propVal = styleObj[propKey];
                if (propVal) {
                    nodeSet = true;
                    // getCrsaPageForIframe($.fn.crsa('getSelectedPage')).undoStack.add("Edit CSS rule", false, _this.pgStylesheet);

                    //newNode.prop = propvalArr[0];
                    pgcssh.setNodeProperty(newNode, 'prop', propvalArr[0])
                    //newNode.value = propvalArr[1].split(';')[0];

                    fullNodeAdded = true;

                    // var nodeValue = propvalArr[1].split(';')[0];
                    var nodeValue = propVal;
                    if (nodeValue) nodeValue = nodeValue.trim();
                    if (pgStylesheet.fileType == 'css') {
                        if (pgStylesheet.isVarValue(nodeValue)) {
                            if (pgStylesheet.fileType == 'css') {
                                pinegrow.showQuickMessage('You can\'t use variables on css file');
                                fullNodeAdded = false;
                            }
                        }
                    }


                    pgcssh.setNodeProperty(newNode, 'value', nodeValue);

                    $declValue[0].innerHTML = newNode.value;
                    updateColorPicker($declValue);
                }
            }
            if (!nodeSet) {
                $declValue.data('just-added', true);
            }

            $declValue.attr('contenteditable', 'true');
            $declValue.focus();
            crsaSelectText($declValue[0]);

            $el[0].innerHTML = newNode.prop;

            addTheNewNode($li, $el, function () {
                if (fullNodeAdded) {
                    stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
                }
            });
        } else if (newNode.type == "rule") {
            var $li = $el.closest('li');
            $el.addClass('rule-selector');
            $el.data('prop-name', 'selector');
            $el[0].innerHTML = newNode.selector;

            // $el.autocomplete( "destroy" );
            var simpleAutocomplete = $el.data('simple-autocomplete');
            if (simpleAutocomplete) simpleAutocomplete.destroy();

            $el.off('keydown').off('blur');

            addCollapseIcon($li.find('> div'));
            var $ul = addNestedUl($li);
            addPlaceholderNode($ul);

            addTheNewNode($li, $el, function () {
                // setUncollapsedNodes(newNode, true);
            });
            stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
        } else if (newNode.type == "atrule") {
            var $li = $el.closest('li');
            $el.addClass('atrule-name');
            $el.data('prop-name', 'name');

            var $atRuleParams = addAtRuleParams($el);
            $atRuleParams.data('prop-name', 'params');

            var nameAndParams = newNode.name.split(' ');
            $el[0].innerHTML = '@' + nameAndParams[0];
            if (nameAndParams.length > 1) {
                newNode.name = nameAndParams[0];
                // get params
                nameAndParams.splice(0, 1);
                var lastIndex = nameAndParams.length - 1;

                if (nameAndParams[lastIndex] == '{') {
                    newNode.nodes = [];
                    nameAndParams.splice(lastIndex, 1);

                    var $ul = addNestedUl($li);
                    addPlaceholderNode($ul);
                }

                newNode.params = nameAndParams.join(' ');
                $atRuleParams[0].innerHTML = newNode.params;

                updateColorPicker($atRuleParams);
                // $atRuleParams[0].innerHTML = nameAndParams[1];
                addTheNewNode($li, $atRuleParams, function () {
                    $el.removeAttr('contenteditable');
                    stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);

                    if (pgStylesheet.fileType == "css") {
                        if (isImportedNode(newNode)) updateAllStylesheetPage(pgStylesheet);
                    }
                    // $el.blur();
                });
            }
            else {
                if (pgStylesheet.fileType == "css") {
                    if (newNode.name == 'import' && pgStylesheet.isImported()) {
                        stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
                        $li.remove();
                        showAddingImportNotice();
                        return;
                    }
                }

                firstValueWritten($el, $atRuleParams, function ($li) {
                    var node = $li.data('node');
                    node.params = node.params.trim();
                    var $atRuleParams = $li.find('.atrule-params');
                    if (node.params[node.params.length - 1] == "{") {
                        node.params = node.params.substr(0, node.params.length - 1).trim();
                        $li.find('.atrule-params')[0].innerHTML = node.params;
                        node.raws.between = " ";

                        node.nodes = [];
                        addCollapseIcon($li.find('> div'));

                        var $ul = addNestedUl($li);
                        addPlaceholderNode($ul);
                    }
                    else {
                        node.raws.between = "";
                    }
                });
            }

            // $el.autocomplete( "destroy" );
            var simpleAutocomplete = $el.data('simple-autocomplete');
            if (simpleAutocomplete) simpleAutocomplete.destroy();

            $el.off('keydown').off('blur');
            if (newNode.nodes) {
                addCollapseIcon($li.find('> div'));
            }
        } else if (newNode.type == "import") {
            var $li = $el.closest('li');
            $el.addClass('import-name');
            $el.data('prop-name', 'name');

            var $importPath = addImportPath($el);
            $importPath.data('prop-name', 'importPath');

            var nameAndParams = newNode.name.split(' ');
            $el[0].innerHTML = '@' + nameAndParams[0];
            if (nameAndParams.length > 1) {
                newNode.name = nameAndParams[0];
                // get params
                nameAndParams.splice(0, 1);
                var lastIndex = nameAndParams.length - 1;

                newNode.importPath = nameAndParams.join(' ');
                $importPath[0].innerHTML = newNode.importPath;
                // $importPath[0].innerHTML = nameAndParams[1];
                addTheNewNode($li, $importPath, function () {
                    $el.removeAttr('contenteditable');
                    stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
                    // $el.blur();
                });
            }
            else {
                firstValueWritten($el, $importPath, function ($li) {
                    var node = $li.data('node');
                    node.importPath = node.importPath.trim();
                    var $importPath = $li.find('.import-importPath');
                    node.raws.between = "";
                });
            }

            // $el.autocomplete( "destroy" );
            var simpleAutocomplete = $el.data('simple-autocomplete');
            if (simpleAutocomplete) simpleAutocomplete.destroy();

            $el.off('keydown').off('blur');
        } else if (newNode.type == "comment") {
            var $li = $el.closest('li');
            $li.find('div > input').prop('checked', false);
            $el[0].innerHTML = newNode.text;
            $el.addClass('comment-text');

            $el.off('keydown').off('blur');
            addTheNewNode($li, $el);

            stylePaneHelper.addNewNodeInsideOrBesideCurrentNode($li);
            $li.data('add-new-node', false);
        }
    }

    this.addingNewNode = function ($el) {
        addAutoPropCompeleteTo($el);
    }

    this.getStringOfSelectedItems = function (deleteNodes) {
        var $selectedItems = $ul.find('.selected-item');
        var selectedNodeString = '';

        for (var i = 0; i < $selectedItems.length; i++) {
            if (i != 0) selectedNodeString += '\n';

            var node = this.getNodeFrom($($selectedItems[i]));
            selectedNodeString += pgStylesheet.stringifyNode(node);

            if (deleteNodes) {
                var $li = $($selectedItems[i]);
                var skipNodeUpdating = i + 1 < $selectedItems.length;
                _this.removeNode($li, skipNodeUpdating);
                $li.remove();
            }
        }

        if (selectedNodeString.length > 0) {
            selectedNodeString += '\n';
        }

        return selectedNodeString;
    }

    var dummyFieldKeydown = function (e) {
        if (dummyFieldKeydown != null) {
            var code = e.keyCode || e.which;
            if (code == 8 || code == 46) {
                var $selectedItems = $ul.find('.selected-item');

                if ($selectedItems.length > 0) {
                    // getCrsaPageForIframe($.fn.crsa('getSelectedPage')).undoStack.add("Edit CSS rule", false, pgStylesheet);

                    for (var i = 0; i < $selectedItems.length; i++) {
                        var $li = $($selectedItems[i]);
                        var skipNodeUpdating = i + 1 < $selectedItems.length;
                        _this.removeNode($li, skipNodeUpdating);
                        $li.remove();
                    }
                    $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });

                    return false;
                }
            }
        }
    }

    var dummyFieldKeyup = function (e) {
        if (dummyFieldKeydown != null) {
            if (spectrumActived && $activatedColorPicker) {
                var code = e.keyCode || e.which;
                if (code == 27) {
                    $activatedColorPicker.spectrum("hide");
                }
            }
        }
    }

    $('#crsa-dummy-field')
        .on('keydown', dummyFieldKeydown)
        .on('keyup', dummyFieldKeyup);

    this.hasSelectedRules = function () {
        return $ul.find('.selected-item').length > 0;
    }

    var insertNodesAfter = function ($li, nodes) {
        var sourceArray = [];
        var arrayIndex = 0;

        var forAll = function (nodes, addLi) {
            var length = nodes.length;
            for (var i = 0; i < length; i++) {
                var node = nodes[i];
                if (addLi) sourceArray[arrayIndex++] = '<li data-node-indices="' + pinegrow.pgPostCSSHelper.getNodeIndices(node) + '" class="' + getLiClassesForNode(node) + '">';

                sourceArray[arrayIndex++] = '<div>';
                sourceArray[arrayIndex++] = getStylePaneHtmlForNode(node, true);
                sourceArray[arrayIndex++] = '</div>';

                if (node.nodes && node.nodes.length > 0) {
                    sourceArray[arrayIndex++] = '<ul class="nested">';
                    forAll(node.nodes, true);
                    sourceArray[arrayIndex++] = '</ul>';
                }

                if (addLi) sourceArray[arrayIndex++] = '</li>';
            }
        }

        forAll(nodes, true);
        var $source = $(sourceArray.join(''));
        $source.insertAfter($li);

        var $declValue = $source.find('span.decl-value,span.atrule-params');
        updateColorPicker($declValue);
    }

    this.pasteTriggeredForNewField = function ($el) {
        var clipboardItem = pinegrow.clipboard.getCurrentItem();
        // if (crsaIsAbsoluteUrl(clipboardItem.code)) return;

        pgStylesheet.parseSource(clipboardItem.code, function (error, postcssSource, newRootNode) {
            if (error) {
                if (clipboardItem.type == 'css') {
                    pinegrow.showAlert('There is an error in the copied text, can\'t paste it as css code', 'Can\'t paste code');
                }
                else {
                    //Is already inserted into editable area
                    //document.execCommand('insertText', false, clipboardItem.code);
                }
                return;
            }

            if (newRootNode.nodes == 0) {
                // document.execCommand('insertText', false, clipboardItem.code);
                // pinegrow.showAlert('There are no rules to paste', 'Nothing to pase');
                return;
            }

            var $li = $el.closest('li');
            var nodePosAndParent = getIndexAndParentOfNewItem($li);
            pinegrow.pgPostCSSHelper.insertNodesAt(newRootNode.nodes, nodePosAndParent.parent, nodePosAndParent.index);

            updatePgScssCodeUsingRules(function (resRootNode) {
                // $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
                pinegrow.dispatchEvent('on_css_source_changed', null, {list: [pgStylesheet]});
            });
        }, false);
    }

    var updateRulesAndStylePane = function () {
        updatePgScssCodeUsingRules(function () {
            var scrollTop = $container.scrollTop();
            _this.show();
            $container.scrollTop(scrollTop);
        });
    }

    // Show all nodes inside list item
    var showAllNodesInside = function ($li, nodes, addLi) {
        var sourceArray = [];
        var arrayIndex = 0;

        var forAll = function (nodes, addLi) {
            var length = nodes.length;
            for (var i = 0; i < length; i++) {
                var node = nodes[i];
                if (addLi) sourceArray[arrayIndex++] = '<li data-node-indices="' + pinegrow.pgPostCSSHelper.getNodeIndices(node) + '" class="' + getLiClassesForNode(node) + '">';

                sourceArray[arrayIndex++] = '<div>';
                sourceArray[arrayIndex++] = getStylePaneHtmlForNode(node, true);
                sourceArray[arrayIndex++] = '</div>';

                if (node.nodes && node.nodes.length > 0) {
                    sourceArray[arrayIndex++] = '<ul class="nested">';
                    forAll(node.nodes, true);
                    sourceArray[arrayIndex++] = '</ul>';
                }

                if (addLi) sourceArray[arrayIndex++] = '</li>';
            }
        }

        forAll(nodes, addLi);
        $li.attr('class', getLiClassesForNode(nodes[0]));
        $li[0].innerHTML = sourceArray.join('');

        var $declValue = $li.find('span.decl-value,span.atrule-params');
        updateColorPicker($declValue);
    }

    var canShowNode = function (node) {
        var parent = node.parent;

        var parentIsAtRule = node.type == "rule" && parent && (isMediaNode(parent) || isSupportsNode(parent));
        if (!parentIsAtRule && options['skip-showing-' + node.type]) return false;

        return true;
    }

    // Place nodes inside ul
    var placeNodesInside = function ($ul, nodes, childNode) {
        var sourceArray = [];
        var arrayIndex = 0;

        var forAll = function (nodes, showOnly) {
            var length = nodes.length;
            for (var i = 0; i < length; i++) {
                var node = nodes[i];
                if (showOnly && node['pg-node-id'] != showOnly['pg-node-id']) continue;

                if (!canShowNode(node)) continue;
                if (filter_re && pgcssh.getNodeType(node) != 'decl') {
                    if (!pgcssh.searchInNode(node, filter_re, pgStylesheet)) continue;
                }

                var liIndex = arrayIndex;
                var classes = getLiClassesForNode(node);
                var hasUncollapse = classes.indexOf('uncollapse') > -1;
                sourceArray[arrayIndex++] = '<li draggable="true" data-node-indices="' + pinegrow.pgPostCSSHelper.getNodeIndices(node) + '" class="' + classes + '"' + getMoreAttrForNode(node) + '>';

                sourceArray[arrayIndex++] = '<div>';
                var nodeHtmlText = getStylePaneHtmlForNode(node, hasUncollapse);
                if (nodeHtmlText == "") {
                    sourceArray.splice(liIndex, sourceArray.length - liIndex);
                    continue;
                }
                else {
                    sourceArray[arrayIndex++] = nodeHtmlText;
                }

                sourceArray[arrayIndex++] = '</div>';

                if (hasUncollapse && node.nodes && node.nodes.length > 0) {
                    sourceArray[arrayIndex++] = '<ul class="nested">';
                    forAll(node.nodes, null);
                    sourceArray[arrayIndex++] = '</ul>';
                }

                sourceArray[arrayIndex++] = '</li>';
            }
        }

        forAll(nodes, childNode);
        $ul[0].innerHTML = sourceArray.join('');

        var $declValue = $ul.find('span.decl-value,span.atrule-params');
        updateColorPicker($declValue);
    }

    var setUncollapsedNodes = function (node, value, skipUpdateCode) {
        value = value || false;
        if (node && node.nodes) {

            if(_this.pgStylesheet && !options['uncollapse']) {
                //don't set anything if SP is uncollapsed with options
                _this.pgStylesheet.setCollapsedNode(node['pg-node-id'] || null, value);
            }
            return;


            var uncollapseNode = pinegrow.pgPostCSSHelper.getNodesWithRegex("comment", "text", /(pg-uncollapsed:(true|false))/, node)[0];
            if (!uncollapseNode) {
                uncollapseNode = pinegrow.pgPostCSSHelper.createNode('comment', '');
                pinegrow.pgPostCSSHelper.insertNodeInside(uncollapseNode, node);
            }

            var changed = false;
            if (uncollapseNode.text == '') {
                uncollapseNode.text = 'pg-uncollapsed:' + value;
                changed = true;
            }
            else {
                var uncollapseArr = uncollapseNode.text.split(':');
                var uncollapsed  = (uncollapseArr[1] == "true");
                if (uncollapsed != value) {
                    if (value == false) {
                        pinegrow.pgPostCSSHelper.removeNodeFrom(uncollapseNode);
                    }
                    else {
                        uncollapseNode.text = 'pg-uncollapsed:' + value;
                    }
                    changed = true;
                }
            }

            if (changed && !skipUpdateCode) {
                updatePgScssCodeUsingRules(null, true);
            }
        }
    }

    var getUncollapsedNodes = function (node) {
        if (node && node.nodes) {
            if(_this.pgStylesheet) {
                return _this.pgStylesheet.getCollapsedNode(node['pg-node-id'] || null);
            }

            var uncollapseNode = pinegrow.pgPostCSSHelper.getNodesWithRegex("comment", "text", /(pg-uncollapsed:(true|false))/, node);

            if (uncollapseNode.length > 0) {
                uncollapseNode = uncollapseNode[0];
                var pgUncollapse = uncollapseNode.text.split(':');
                var pgUncollapseText  = pgUncollapse[0].trim();
                var pgUncollapseValue = pgUncollapse[1].trim();

                if (pgUncollapseText == "pg-uncollapsed") {
                    if (pgUncollapseValue == "true")  return true;
                    else if (pgUncollapseValue == "false") return false;
                }
            }
        }
        return null;
    }

    var toggleNodesApperanceFor = function ($li, action, selectNode, childNode) {
        var $el = $li.find('> div > .collapse-icon');

        if ($el.hasClass('icon-down')) {
            // if (!forceUncollapse) {
            if (action != "uncollapse") {
                var $li = $el.closest('li');
                var $ul = $li.find('> ul');
                $ul.addClass('close');
                $el.removeClass('icon-down').addClass('icon-right');
                var node = _this.getNodeFrom($li);
                setUncollapsedNodes(node, false);
            }
        }
        else if ($el.hasClass('icon-right')) {
            if (action != "collapse") {
                var wasClosed = false;

                var $li = $el.closest('li');
                var $ul = $li.find('> ul');
                if (!$ul || $ul.length == 0) {
                    $ul = addNestedUl($li);
                }
                else if ($ul.hasClass('close')) {
                    $ul.removeClass('close');
                    wasClosed = true;
                }

                var node = _this.getNodeFrom($li);
                if (!wasClosed) {
                    var nodes = node.nodes;
                    if (nodes && nodes.length > 0) {
                        // convertToTree(nodes, $ul, true);
                        placeNodesInside($ul, nodes, childNode);

                        if (action == 'uncollapse-all') {
                            $ul.find('> li').each(function (i, li) {
                                toggleNodesApperanceFor($(li), action);
                            });
                        }
                        onRuleSelected(null, {rule: pinegrow.getSelectedRule()}, true /* skip scroll */);
                        for (var i = 0; i < nodes.length; i++) {
                            if (options['on-node-added']) options['on-node-added'](nodes[i], pgStylesheet, _this);
                        }
                    }

                    if ($ul.children().length == 0) addPlaceholderNode($ul);
                }

                $el.removeClass('icon-right').addClass('icon-down');
                setUncollapsedNodes(node, true);
                updateHighlightedRulesWithSelectedElm();

                if (options['on-node-uncollapsed']) options['on-node-uncollapsed']();
                if (selectNode) selectRule(node, $li);
            }
        }
    }

    var getLiCollapseState = function ($li) {
        var $el = $li.find('> div > .collapse-icon');

        if ($el.hasClass('icon-down'))  return 'uncollapsed';
        else if ($el.hasClass('icon-right')) return 'collapsed';
    }

    this.toggleCodeIgnoring = function ($el, ignore, done) {
        var $li = $el.closest('li');
        // getCrsaPageForIframe($.fn.crsa('getSelectedPage')).undoStack.add("Edit CSS rule", false, pgStylesheet);
        var currentNode = _this.getNodeFrom($li);

        var source = pgStylesheet.filterSource(pinegrow.pgPostCSSHelper.getSource(currentNode));
        if (!ignore) {

            $ul.data('properties-in-edit-mode', true);
            source = source.trim();
            if (source.substr(0, 2) == '/*' && source.substr(source.length-2, source.length) == '*/') {
                source = source.substr(2, source.length - 4).trim();
            }
            var commentSource = source.replace(/\|\|\*\|/, '/*').replace(/\|\*\|\|/, '*/');
            pinegrow.pgPostCSSHelper.parsedCSSSourceFromText(commentSource, function (err, css, root) {
                if (err || root.nodes.length == 0) {
                    pinegrow.showQuickMessage("Can't un-comment this line", 4000);
                    if (done) done(false);
                }
                else {
                    var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(currentNode);
                    var callback = function (done) {
                        updatePgScssCodeUsingRules(function (resRootNode) {
                            updateNodeSelector(pgStylesheet.rootNode.nodes);

                            if ($li.attr('data-removed') != "true") {
                                updateNodesDomIds($li);
                            }

                            updateHighlightedRulesWithSelectedElm();

                            var addedNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), nodeIndices);
                            setUncollapsedNodes(addedNode, true);
                            // if (isImportedNode(addedNode)) {
                            //     $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
                            // }

                            pinegrow.dispatchEvent('on_css_node_added', null, {node: addedNode, stylesheet: pgStylesheet, operation: 'ignore'});
                            if (options['on-node-added']) options['on-node-added'](addedNode, pgStylesheet, _this);

                            $ul.data('properties-in-edit-mode', false);
                            if (done) done(true);
                        });
                    }

                    var newNode = root.nodes[0];
                    if (pgStylesheet.fileType == "css") {
                        if (isImportedNode(newNode) && pgStylesheet.isImported()) {
                            showAddingImportNotice();
                            return;
                        }
                    }

                    currentNode['ignore'] = false;
                    $li.removeClass('crsa-comment');
                    pinegrow.pgPostCSSHelper.copyNodeRawsTo(newNode, currentNode);
                    pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: currentNode, stylesheet: pgStylesheet, operation: 'ignore'});

                    pinegrow.pgPostCSSHelper.replaceNodeWith(currentNode, newNode);
                    showAllNodesInside($li, [newNode]);
                    callback(done);
                }
            });
        }
        else {
            $ul.data('properties-in-edit-mode', true);
            $li.removeAttr('checked');
            $li.attr('data-removed', "true");
            $li.addClass('crsa-comment');
            var commentSource = source.replace(/\/\*/, '||*|').replace(/\*\//, '|*||');
            var commentNode = pinegrow.pgPostCSSHelper.createNode('comment', commentSource, null, currentNode, pgStylesheet);

            var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(currentNode);
            pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: currentNode, stylesheet: pgStylesheet});

            if (options['on-before-node-deleted']) {
                if (currentNode.nodes && currentNode.nodes.length > 0) {
                    var children = currentNode.nodes;
                    for (var i = 0; i < children.length; i++) {
                        options['on-before-node-deleted'](children[i], pgStylesheet, _this);
                    }
                }

                options['on-before-node-deleted'](currentNode, pgStylesheet, _this);
            }

            pinegrow.pgPostCSSHelper.replaceNodeWith(currentNode, commentNode);
            commentNode['pg-node-id'] = currentNode['pg-node-id']

            showAllNodesInside($li, [commentNode]);

            if (isImportedNode(currentNode)) {
                pgStylesheet.ignoreImportedFile(currentNode, true);
            }

            updatePgScssCodeUsingRules(function (resRootNode) {
                updateNodeSelector(pgStylesheet.rootNode.nodes);

                if ($li.attr('data-removed') != "true") {
                    updateNodesDomIds($li);
                }

                delete commentNode['pg-node-id'];
                $li.removeAttr('id');
                $li.attr('data-removed', "false");

                updateHighlightedRulesWithSelectedElm();

                // if (isImportedNode(currentNode)) {
                //     $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
                //     pinegrow.refreshAllPages();
                // }
                $li.attr('data-removed', "false");

                var addedNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), nodeIndices);
                pinegrow.dispatchEvent('on_css_node_added', null, {node: addedNode, stylesheet: pgStylesheet});
                $ul.data('properties-in-edit-mode', false);
                if (done) done(true);
            });
        }
    }

    var getLiToHighlightBorder = function ($target) {
        return $target.closest('li');
        var $li = $target.closest('li');
        var $ul = $li.find('> ul');
        if ($ul.length > 0) {
            if ($ul.children().length == 0) {
                return $('<li class="new-node-placeholder"></li>').appendTo($ul);
            }
        }
        return $li;
    }

    var collapseUncollapseLi = function (e) {
        e.preventDefault();
        var $target = $(e.target);
        var $li = $target.closest('li');

        var insertBorderHeightlighted = ($li.hasClass('hoverd-bottom') || $li.hasClass('hoverd-top'));
        if(!options['uncollapse-on-click'] && !insertBorderHeightlighted) {
            var node = _this.getNodeFrom($li);
            selectRule(node, $li);
            return;
        }
        if (!e.metaKey && !e.shiftKey) {

            if ($target.is('div') || $target.parent().is('div')) {
                liClicksTime++;

                var canChangeCollapse = ((!$li.hasClass('hoverd-bottom') && !$li.hasClass('hoverd-top')) || $li.hasClass('hoverd-middle')) && !$target.is('i');
                setTimeout(function() {
                    if (liClicksTime > 1) {
                        liClicksTime--;
                        return;
                    }
                    liClicksTime = 0;
                    if (canChangeCollapse) {
                        toggleNodesApperanceFor($li, null, true);
                    }
                }, 170);
            }
        }
    }

    var editLiNode = function (e) { liClicksTime++; }

    var addNewNodeBeforeCurrentNode = function ($li) {
        var $newLi = $('<li><div><span class="crsa-scss-prop"></span></div></li>');

        if ($li.length == 0) {
            $ul.append($newLi);
        }
        else {
            $newLi.insertBefore($li);
        }

        var $span = $newLi.find('span');
        stylePaneHelper.addNewNode($span, null);
    }

    var isPlaceholderLi = function ($li) {
        if ($li.data('placeholder-item') ||
            $li.hasClass('crsa-cm-placeholder') ||
            $li.hasClass('placeholder') ||
            $li.hasClass('placeholder-node') ||
            $li.hasClass('new-node-placeholder')) {
            return true;
        }
        return false;
    }

    this.isAllowedToAddNewNodeAt = function ($li, $target, mousePosition) {
        if (options['is-allowed-insert-node']) {
            if (!options['is-allowed-insert-node']($li, _this, mousePosition, this)) return false;
        }

        if ($li.data('placeholder-item') || $li.hasClass('placeholder') ||
            $li.hasClass('new-node-placeholder') || $li.hasClass('selected-item') ||
            ($li.hasClass('placeholder-node') && !$li.hasClass('crsa-node'))) return false;

        if ($target.is('span') || $target.is('i')) return false;

        var $prev = $li.prev();
        if ($prev.is(':visible') && $prev.is(_this.itemSelector) && mousePosition == this.mousePositionVals.TOP) return false;

        if ($li.hasClass('crsa-rule') && !options['auth']['can_add_rule']) return false;

        return true;
    }

    this.isAllowedToEditNode = function ($el, textChecking) {
        if ($el.hasClass('ui-helper-hidden-accessible') || $el.hasClass('more-text')) return false;
        if (textChecking && getTextInsideElm($el) != '') return false;
        return true;
    }

    this.isDblClickEditingAllowed = function ($el) {
        var $li = $el.closest('li');
        if (isPlaceholderLi($li)) return false;

        var node = _this.getNodeFrom($li);

        if (node.type == "rule" && !options['auth']['can_edit_rule']) return false;

        if (Array.isArray(node.nodes)) return true;
        return false;
    }

    this.getSelectableObjForEditing = function ($el) {
        if ($el.hasClass('more-text')) {
            return $el.closest('.editable');
        }
        return null;
    }

    var updateNodesIds = function (nodes, newNodes) {
        for(var i = 0; i < nodes.length; i++) {
            if (newNodes[i] && newNodes[i]['pg-node-id']) {
                nodes[i]['pg-node-id'] = newNodes[i]['pg-node-id'];
            }

            if (nodes[i].nodes && nodes[i].nodes.length > 0 &&
                newNodes[i].nodes && newNodes[i].nodes.length > 0) {
                updateNodesIds(nodes[i].nodes, newNodes[i].nodes);
            }
        }
    }

    var updateNodeSelector = function (nodes) {
        // if (pgStylesheet.fileType == "css") {
        //     pinegrow.pgPostCSSHelper.addPgIdToSelector(nodes);
        // }
    }

    var updateDomIdsNodeAdded = function ($li) {
        var changingIdFinished = false;
        var lastNodeId = pgStylesheet.lastNodeId();
        var currentNodeId = _this.getNodeFrom($li)['pg-node-id'];
        var $oldLi = $li, $newLi;

        var pgRuleIdReg = new RegExp('pg-' + fileName + '-rule-id-([0-9]*)' + options['extra-rule-el-id']);

        var prevRemovedId = -1;

        while (currentNodeId <= lastNodeId) {
            $newLi = $('#pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);

            if ($newLi.length > 0 && $newLi.closest('.rule-list')[0] == $ruleList[0]) {
                while ($newLi.length > 0) {
                    currentNodeId++;

                    $oldLi = $('#pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);
                    $oldLi.removeAttr('id');

                    var currNode = _this.getNodeFrom($newLi);

                    var oldId = -1;
                    if ($newLi.attr('id')) {
                        var matched = $newLi.attr('id').match(pgRuleIdReg);
                        if (matched && matched.length > 1) {
                            oldId = parseInt(matched[1]);
                        }
                    }
                    else {
                        oldId = prevRemovedId;
                    }

                    if (options['on-rule-id-changed']) options['on-rule-id-changed'](currNode, pgStylesheet, currentNodeId, oldId, _this);

                    $newLi.attr('id', 'pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);
                    $newLi = $oldLi;

                    prevRemovedId = currentNodeId;
                }
                currentNodeId++;
            }
            else {
                if ($oldLi.length > 0) {
                    var currNode = _this.getNodeFrom($newLi);
                    if (options['on-rule-id-changed']) options['on-rule-id-changed'](currNode, pgStylesheet, currentNodeId, null, _this);
                    $oldLi.attr('id', 'pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);
                    $oldLi = [];
                }
                currentNodeId++;
            }
        }
    }

    var numOfRulesInsideNode = function (node) {
        var num = 0;
        var updateByVar = function (nodes) {
            for (var i = 0; i < nodes.length; i++) {
                var node = nodes[i];
                if (node.type == 'rule' && _this.pgStylesheet.canAddRuleIds(node)) {
                    num++;
                }

                if (node.nodes && node.nodes.length > 0) updateByVar(node.nodes);
            }
        }
        updateByVar(node.nodes || []);

        return num;
    }

    var updateDomIdsNodeRemoved = function ($li) {
        var node = _this.getNodeFrom($li);
        var changingIdFinished = false;
        var lastNodeId = pgStylesheet.lastNodeId();
        var currentNodeId;
        var pgRuleIdReg = new RegExp('pg-' + fileName + '-rule-id-([0-9]*)' + options['extra-rule-el-id']);
        var liId = $li.attr('id');
        if (!liId) return;

        var matchedRuleId = liId.match(pgRuleIdReg);
        if (matchedRuleId.length > 1) {
            currentNodeId = parseInt(matchedRuleId[1]);
        }
        var $oldLi = $li, $newLi;


        var by = 1 + numOfRulesInsideNode(node);
        while (currentNodeId <= lastNodeId + 1) {
            $newLi = $('#pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);

            if ($newLi.length > 0 && $newLi.closest('.rule-list')[0] == $ruleList[0]) {
                while ($newLi.length > 0) {
                    $oldLi = $('#pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);
                    $oldLi.removeAttr('id');


                    if ($newLi.closest('li[data-removed="true"]').length == 0) {
                        var newId = currentNodeId - by;
                        var currNode = _this.getNodeFrom($newLi);
                        if (currNode) {
                            if (options['on-rule-id-changed']) options['on-rule-id-changed'](currNode, pgStylesheet, newId, null, _this);
                            currNode['pg-node-id'] = newId;
                        }
                        $newLi.attr('id', 'pg-' + fileName + '-rule-id-' + newId + options['extra-rule-el-id']);
                    }

                    $newLi = $oldLi;
                }
                currentNodeId++;
            }
            else {
                if ($oldLi.length > 0) {
                    $oldLi.attr('id', 'pg-' + fileName + '-rule-id-' + currentNodeId + options['extra-rule-el-id']);
                    $oldLi = [];
                }
                currentNodeId++;
            }
        }
    }

    var updateDomIdsNode = function ($li) {
        var $next = $li;
        while (true) {
            var node = _this.getNodeFrom($next);
            if (!node['pg-node-id']) break;

            var $newLi = $('#pg-' + fileName + '-rule-id-' + node['pg-node-id'] + options['extra-rule-el-id']);
            if ($newLi[0] == $next[0])  {
                $next.attr('id', 'pg-' + fileName + '-rule-id-' + node['pg-node-id'] + options['extra-rule-el-id']);
                break;
            }
            else {
                $newLi.attr('id', '');
                $next.attr('id', 'pg-' + fileName + '-rule-id-' + node['pg-node-id'] + options['extra-rule-el-id']);
                $next = $newLi;
            }
        }
    }

    var updateNodesDomIds = function ($li, moved) {
        if (moved) {
            updateDomIdsNode($li);
        }
        else {
            if ($li.attr('data-removed') == "true") {
                updateDomIdsNodeRemoved($li);
            }
            else {
                updateDomIdsNodeAdded($li);
            }
        }

        var node = _this.getNodeFrom($li);
        if (node && node['pg-node-id']) {
            $li.attr('id', 'pg-' + fileName + '-rule-id-' + node['pg-node-id'] + options['extra-rule-el-id']);
        }
    }

    var availableWidth = 277;
    var nestedPadding = 15;
    var checkboxWidth = 15;
    var getAvailableWidth = function (node, level) {
        var takenWidth = level * nestedPadding;
        takenWidth += checkboxWidth;

        var currentNodeAvailableWidth;
        if (node.type == "decl") {
            currentNodeAvailableWidth = availableWidth;
        }
        else {
            currentNodeAvailableWidth = 170;
        }

        return currentNodeAvailableWidth - takenWidth;
    }

    var measureTextWidthCanvasContext = $('<canvas></canvas>').get(0).getContext("2d");
    measureTextWidthCanvasContext.font = '13px -apple-system,BlinkMacSystemFont,"Segoe UI",Roboto,Oxygen-Sans,Ubuntu,Cantarell,"Helvetica Neue",sans-serif';

    var measureTextWidth = function (text) {
        return measureTextWidthCanvasContext.measureText(text).width;
    }

    var lineHeight = 16;
    var measureHeight = function (blockWidth, textWidth, twoParts) {
        var wantedWidth = Math.ceil(textWidth / blockWidth);
        if (twoParts && wantedWidth > 1) wantedWidth += 1;
        return wantedWidth * lineHeight;
    }

    var getMoreAttrForNode = function (node) {
        var attrs = "";
        if (node['pg-node-id']) {
            attrs += ' id="pg-' + fileName + '-rule-id-' + node['pg-node-id'] + options['extra-rule-el-id'] + '"';
        }

        return attrs;
    }

    var getLiClassesForNode = function (node) {
        var classes = 'crsa-node crsa-' + node.type;
        if (node.type != 'rule' || (node.type == 'rule' && options['auth']['can_edit_rule'])) {
            classes += ' has-options';
        }
        if (node.type == 'rule' || node.type == 'atrule') {
            classes += ' can-has-children';
        }

        if (getUncollapsedNodes(node) == true) classes += ' uncollapse';
        return classes;
    }

    var filter_re = null;

    var updateFilterRegex = function () {
        //no need to create fuzzy search re for every containInFilter call. That's 1000s calls per paint.
        if (_this.filter) {
            //fuzzy gives very tsrange results
            //filter_re = crsaGetFuzzySearchRegxFrom(_this.filter.toLowerCase())
            filter_re = new RegExp(escapeRegExp(_this.filter), "i");
        }
    }

    var convertToTree = function (nodes, $ul, insertInside) {
        var buildSourceCode = [];
        var buildSourceIndex = 0;

        filter_re = null;
        updateFilterRegex();

        var forAll = function (nodes, level) {
            var length = nodes.length;

            for(var i = 0; i < length; i++) {
                var parserNode = nodes[i];

                if (insertInside && options['skip-showing-' + parserNode.type]) continue;

                //first check if matches filter, only then add to array. Adding + splicing very ineffecient
                if (filter_re) {
                    if (!pgcssh.searchInNodeAndSubnodes(parserNode, filter_re, pgStylesheet)) continue;
                }

                var liIndex = buildSourceIndex;

                buildSourceCode[buildSourceIndex++] = '<li draggable="true" data-node-indices="' + pinegrow.pgPostCSSHelper.getNodeIndices(parserNode) + '" class="' + getLiClassesForNode(parserNode) + '"';

                if (parserNode['pg-node-id']) {
                    buildSourceCode[buildSourceIndex++] = ' id="pg-' + fileName + '-rule-id-' + parserNode['pg-node-id'] + options['extra-rule-el-id'] + '"';
                }

                buildSourceCode[buildSourceIndex++] = '><div></div>';

                buildSourceCode[buildSourceIndex++] = '</li>';
            }
        }

        forAll(nodes, 0);

        $ul[0].innerHTML = buildSourceCode.join('');
    }

    var getStylePaneHtmlForNode = function (node, nestedShown) {
        if (node.type == "comment") {
            var ret = pgStylesheet.filter(node);
            if (ret.haveTo == "delete") return '';
            return '<span class="crsa-scss-prop comment-text editable" data-prop-name="text">' + node.text.replace(/</g, "&lt;", 'g').replace(/>/g, "&gt;", 'g') + '</span>';
        } else if (node.type == "atrule" && !node.variable) {
            var icon = '';
            if (node.nodes) {
                icon = '<i class="collapse-icon ' + ( nestedShown ? 'icon-down' : 'icon-right' ) + '"></i>';
            }

            var values = { name: node.name, params: node.params };
            var mixinLess = pgStylesheet.fileType == "less" && node.mixin;
            if (mixinLess) {
                values.name = `.${values.name}`;
                if (values.params.length == 0) {
                    values.params = "()";
                }
            }
            else {
                values.name = `@${values.name}`;
            }

            var ret = pgStylesheet.filter(node);
            if (ret.haveTo == "update") {
                values[ret.type] = ret.newValue;
            }


            var html = icon + '<span class="crsa-scss-prop atrule-name" data-prop-name="name">' + values.name + '</span>' +
                '&nbsp;<span class="crsa-scss-prop atrule-params" data-prop-name="params">' + values.params + '</span>';

            return html;
        } else if (node.type == "import") {
            return '<span class="crsa-scss-prop import-name" data-prop-name="name">@' + node.name + '</span>&nbsp;<span class="crsa-scss-prop import-importPath" data-prop-name="importPath">' + node.importPath + '</span>';
        } else if (node.type == "decl" || (node.type == 'atrule' && node.variable)) {
            var ret = pgStylesheet.filter(node);
            if (ret.haveTo == "delete") return '';

            var propName = 'prop';
            var value = node.prop;
            if (node.type == 'atrule') {
                propName = 'name';
                value = '@' + node.name;
            }

            return '<span class="crsa-scss-prop decl-prop editable" data-prop-name="' + propName + '">' + value + '</span>:&nbsp;<span class="crsa-scss-prop decl-value editable" data-prop-name="value">' + node.value + (node.important ? ' !important' : '') + '</span>';
        } else if (node.type == "rule") {
            var selector = node.selector;
            var ret = pgStylesheet.filter(node);
            if (ret.haveTo == "update") {
                selector = ret.newValue;
            }
            var collapseIcon = '<i class="collapse-icon ' + ( nestedShown ? 'icon-down' : 'icon-right' ) + '"></i>';
            var ruleSource = '<span class="crsa-scss-prop rule-selector" data-prop-name="selector">' + selector + '</span>'

            if (node.empty) return ruleSource;
            else            return collapseIcon + ruleSource;
        }
        return null;
    }

    // Show node inside li
    var showNodeFor = function ($li, uncollapseNode, childNode) {

        var forAll = function (list) {
            var length = list.length;
            for (var i = 0; i < length; i++) {
                var $li = $(list[i]);
                if ($li.hasClass('crsa-cm-placeholder') || $li.hasClass('placeholder') || $li.hasClass('placeholder-node')) continue;

                var node = _this.getNodeFrom($li);
                var $div = $li.find('> div');

                $div.html(getStylePaneHtmlForNode(node));

                if (node.type == "decl" || node.type == 'atrule') {
                    var $declValue = $div.find('> span.decl-value, > span.atrule-params');
                    updateColorPicker($declValue);
                }

                if ($li.hasClass('uncollapse') || uncollapseNode) {
                    var action = null;
                    if (options['uncollapse-all']) action = 'uncollapse-all';
                    toggleNodesApperanceFor($li, action, false, childNode);
                }
            }
        }

        forAll($li);
    }

    var $lastShown = null;
    var reachToTheEnd = false;
    var offsetThreshold = 500;

    this.updateShownNodes = function () {
        if (reachToTheEnd) return;
        if (!$lastShown) {
            $lastShown = $ul.find(' > li').first();
            showNodeFor($lastShown, options['uncollapse']);
        }

        if ($lastShown.parents('html').length == 0) {
            reachToTheEnd = false;
            return;
        }


        if (options['show-all']) {
            var $allNodes = $lastShown.nextAll();
            $allNodes.each(function (index, li) {
                var $li = $(li);
                showNodeFor($li, options['uncollapse']);
            });

            reachToTheEnd = true;
        }
        else {
            var $crsaCmRules = $container.closest('.lm_content');
            var height = $crsaCmRules.height();
            if (height == 0) return;

            var rulesTop = $crsaCmRules.offset().top;

            var $allNodes = $lastShown.nextAll();
            var length = $allNodes.length;
            var lastIndex = -1;
            $allNodes.each(function (index, li) {
                var $li = $(li);

                lastIndex = index;
                var offset = $li.offset().top - rulesTop;
                if (offset - offsetThreshold <= height) {
                    $lastShown = $li;
                    showNodeFor($li, options['uncollapse']);
                }
                else {
                    return false;
                }
            });

            if (lastIndex == length - 1) reachToTheEnd = true;
        }

        onRuleSelected(null, {rule: pinegrow.getSelectedRule()}, true /* skip scroll */);
    }

    this.onlyShowNode = function (node) {
        if (!$lastShown) {
            $lastShown = $ul.find(' > li').first();
            showNodeFor($lastShown, options['uncollapse'], node);
        }
    }

    this.addNewNodePlaceholderTo = function ($li) {
        var $lastLi = $li.find('> ul > li:last-child');
        addNewNodeLiClicked($lastLi, true, "rule");
    }

    this.addNewNodePlaceholderBefore = function ($li) {
        addNewNodeLiClicked($li, false, "rule");
    }

    this.addNewNodePlaceholderAfter = function ($li) {
        addNewNodeLiClicked($li, true, "rule");
    }

    this.uncollapseLi = function ($li) {
        toggleNodesApperanceFor($li, 'uncollapse');
    }

    this.collapseLi = function ($li) {
        toggleNodesApperanceFor($li, 'collapse');
    }

    this.addNewNodeIntoParent = function ($li, node) {
        insertNodeIntoParent($li, node);

        updatePgScssCodeUsingRules(function (resRootNode) {
            updateNodeSelector(pgStylesheet.rootNode.nodes);
            updateNodesDomIds($li);

            updateHighlightedRulesWithSelectedElm();
        });
    }

    this.moveNodeFromTo = function ($li, $placeholder, addedToNewCS, newCS) {
        var node = this.getNodeFrom($li);

        // $ul.data('properties-in-edit-mode', true);

        var $next_li = $placeholder.next();
        while ($next_li.hasClass('crsa-cm-placeholder') || $next_li.hasClass('placeholder')) $next_li = $next_li.next();

        var insertUnderNode = false;
        if ($next_li.length > 0) {
            if ($next_li.hasClass('placeholder-node')) {
                insertUnderNode = true;
            }

            if (!insertUnderNode) {
                var nextNode = _this.getNodeFrom($next_li);
                node.raws.before = nextNode.raws.before;

                pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: pgStylesheet});
                pinegrow.pgPostCSSHelper.removeNodeFrom(node);
                pinegrow.pgPostCSSHelper.insertNodeBefore(node, nextNode);
            }

            // $li.insertBefore($next_li);
        }
        else {
            var $prev_li = $placeholder.prev();
            while ($prev_li.hasClass('crsa-cm-placeholder') || $prev_li.hasClass('placeholder')) $prev_li = $prev_li.prev();

            if ($prev_li.length > 0) {
                if ($prev_li.hasClass('placeholder-node')) {
                    insertUnderNode = true;
                }

                if (!insertUnderNode) {
                    var prevNode = _this.getNodeFrom($prev_li);
                    node.raws.before = prevNode.raws.before;

                    pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: pgStylesheet});
                    pinegrow.pgPostCSSHelper.removeNodeFrom(node);
                    pinegrow.pgPostCSSHelper.insertNodeAfter(node, prevNode);
                }

                // $li.insertAfter($prev_li);
            }
        }

        if (insertUnderNode) {
            var $parentLi = $placeholder.closest('li');
            var parentNode = _this.getNodeFrom($parentLi);

            pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: pgStylesheet});
            pinegrow.pgPostCSSHelper.removeNodeFrom(node);
            pinegrow.pgPostCSSHelper.insertNodeAt(node, parentNode, 0);
        }

        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
        updatePgScssCodeUsingRules(function (resRootNode) {
            var newNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), nodeIndices);
            pinegrow.dispatchEvent('on_css_node_added', null, {node: newNode, stylesheet: pgStylesheet});
            // $ul.data('properties-in-edit-mode', false);
            // $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
            // .trigger('crsa-element-selected', { by: options['created-by'] });

            // selectRule(node, $li);
        });
    }

    // this.moveNodeFromTo = function ($li, $placeholder, addedToNewCS, newCS) {
    //     var node = this.getNodeFrom($li);

    //     var $next_li = $placeholder.next();
    //     while ($next_li.hasClass('crsa-cm-placeholder') || $next_li.hasClass('placeholder') || $next_li.hasClass('placeholder-node')) $next_li = $next_li.next();

    //     var insertUnderNode = false;
    //     if ($next_li.length > 0) {
    //         if ($next_li.hasClass('placeholder-node')) {
    //             insertUnderNode = true;
    //         }

    //         if (!insertUnderNode) {
    //             var nextNode = _this.getNodeFrom($next_li);
    //             node.raws.before = nextNode.raws.before;
    //             pinegrow.pgPostCSSHelper.removeNodeFrom(node);
    //             pinegrow.pgPostCSSHelper.insertNodeBefore(node, nextNode);
    //         }

    //         $li.insertBefore($next_li);
    //     }
    //     else {
    //         var $prev_li = $placeholder.prev();
    //         while ($prev_li.hasClass('crsa-cm-placeholder') || $prev_li.hasClass('placeholder') || $prev_li.hasClass('placeholder-node')) $prev_li = $prev_li.prev();

    //         if ($prev_li.length > 0) {
    //             if ($prev_li.hasClass('placeholder-node')) {
    //                 insertUnderNode = true;
    //             }

    //             if (!insertUnderNode) {
    //                 var prevNode = _this.getNodeFrom($prev_li);
    //                 node.raws.before = prevNode.raws.before;
    //                 pinegrow.pgPostCSSHelper.removeNodeFrom(node);
    //                 pinegrow.pgPostCSSHelper.insertNodeAfter(node, prevNode);
    //             }

    //             $li.insertAfter($prev_li);
    //         }
    //     }

    //     if (insertUnderNode) {
    //         var $parentLi = $placeholder.closest('li');
    //         var parentNode = _this.getNodeFrom($parentLi);
    //         pinegrow.pgPostCSSHelper.removeNodeFrom(node);
    //         pinegrow.pgPostCSSHelper.insertNodeAt(node, parentNode, 0);
    //     }

    //     updatePgScssCodeUsingRules(function (resRootNode) {
    //         $('body').trigger('crsa-stylesheets-changed', { by: options['created-by'] });
    //                  // .trigger('crsa-element-selected', { by: options['created-by'] });

    //         selectRule(node, $li);
    //     });
    // }

    var scrollTrigger = function () {
        if (scrollTrigger == null) return;

        _this.updateShownNodes();
    }

    $scrollParent.on('scroll', scrollTrigger);

    $container.get(0).addEventListener('contextmenu', function(event) {
        var $target = $(event.target);

        var varName = $target.text();

        if ($target.hasClass('more-text') && varName[0] == pgStylesheet.getDeclCharachter()) {
            var $menu = new CrsaContextMenu(null, $target);
            $menu.add("Go to definition...", null, function() {
                var $parent = $container.parent();
                var $li = $target.data('master-color').el;
                var newTop = $li.offset().top - $parent.offset().top;
                if (newTop != 0) {
                    $parent.scrollTop(newTop + $parent.scrollTop() - 30);
                }
            });

            $menu.add("Edit master color...", null, function() {
                pgStylesheet.varsPane.updateSearchValue(varName);
            });

            $menu.showAt(event.pageX, event.pageY, $container);
        }

        event.preventDefault();
        event.stopPropagation();
    }, true);

    $container.find('> .cs-head').click(function (e) {
        e.preventDefault();
        var $firstLi = $ul.find('> li:first-child');
        addNewNodeBeforeCurrentNode($firstLi);
    });

    this.getNodeFrom = function ($li) {
        if (!$li.attr('data-node-indices')) return null;
        return pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, $li.attr('data-node-indices').split(','));
    }

    function updateLiNodeData($ul) {
        var arr = $ul.find('li');
        var length = arr.length;
        for (var i = 0; i < length; i++) {
            var $li = $(arr[i])
            $li.data('node', _this.getNodeFrom($li));
        }
    }

    var updateNodeChildIndices = function (nodeIndices, level, $li) {
        var forAll = function (nodeIndices, $li) {
            var $next = $li;
            while ($next.length > 0) {
                if ($next.attr('data-node-indices')) {
                    var indices = $next.attr('data-node-indices').split(',');
                    var newIndices = nodeIndices + "," + indices[indices.length-1];

                    var node = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, newIndices.split(','));
                    // pinegrow.dispatchEvent('node-indices-changed', null, {
                    //     node: node,
                    //     stylesheet: pgStylesheet,
                    //     oldNodeIndices: $next.attr('data-node-indices').split(',')
                    // });

                    // Node indices changed
                    pinegrow.dispatchEvent('on_css_node_indices_changed', null, {
                        node: node,
                        stylesheet: pgStylesheet,
                        oldNodeIndices: $next.attr('data-node-indices').split(','),
                        currNodeIndices: newIndices.split(',')
                    });
                    if (options['on-node-indices-changed']) options['on-node-indices-changed'](node, pgStylesheet, $next.attr('data-node-indices').split(','), newIndices.split(','), _this);
                    $next.attr('data-node-indices', newIndices);

                    var $child = $next.find('> ul > li:first-child');
                    if ($child && $child.length > 0) {
                        forAll(newIndices, $child);
                    }
                }
                $next = $next.next();
            }
        }

        forAll(nodeIndices, $li);
    }

    var updateNodeIndicesAfter = function ($li, node, by, moveBackward, oldNodeIndices) {
        var nodeOldId = -1;

        var getNextItem = function ($el, backward) {
            if (backward) return $el.prev();
            else return $el.next();
        }

        var shouldStopChangingIndex = function (indexLevel) {
            if (nodeOldId < 0) return false;
            if (moveBackward) {
                if (indexLevel <= nodeOldId) return true;
            }
            else {
                if (indexLevel >= nodeOldId + 1) return true;
            }
            return false;
        }

        var forAll = function ($li, level, by, backward) {
            var $next = $li;

            while ($next.length > 0) {
                if ($next.attr('data-node-indices')) {
                    var indices = $next.attr('data-node-indices').split(',');
                    var indexLevel = parseInt(indices[level]);
                    if (shouldStopChangingIndex(indexLevel)) return;
                    indices[level] = indexLevel + by;

                    var node = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, indices);
                    // pinegrow.dispatchEvent('node-indices-changed', null, {
                    //     node: node,
                    //     stylesheet: pgStylesheet,
                    //     oldNodeIndices: $next.attr('data-node-indices').split(',')
                    // });

                    // Node indices changed
                    pinegrow.dispatchEvent('on_css_node_indices_changed', null, {
                        node: node,
                        stylesheet: pgStylesheet,
                        oldNodeIndices: $next.attr('data-node-indices').split(','),
                        currNodeIndices: indices
                    });
                    if (options['on-node-indices-changed']) options['on-node-indices-changed'](node, pgStylesheet, $next.attr('data-node-indices').split(','), indices, _this, $next.find('> ul > li').length > 0);
                    $next.attr('data-node-indices', indices.join(','));
                    var node = _this.getNodeFrom($next);
                    if (node.nodes && node.nodes.length > 0) {
                        var $new = $next.find('> ul > li:first-child');
                        updateNodeChildIndices(indices.join(','), level, $new);
                    }
                }

                $next = getNextItem($next, backward);
                while($next.hasClass('.ui-sortable-helper')) $next = getNextItem($next, backward);
            }
        }

        var level = pinegrow.pgPostCSSHelper.getNodeLevel(node);
        if (oldNodeIndices) nodeOldId = parseInt(oldNodeIndices.split(',')[level]);
        forAll(getNextItem($li, moveBackward), level, by, moveBackward);
    }

    var getIndexAndParentOfNewItem = function ($li, node) {
        var insertAt = 0,
            parentNode = null,
            siblingParent = null;

        var $next_li = $li.next();
        if ($next_li.hasClass('crsa-cm-placeholder') || $next_li.hasClass('placeholder') || $next_li.hasClass('placeholder-node')) $next_li = $next_li.next();

        var insertUnderNode = false;
        if ($next_li.length > 0) {
            if ($next_li.hasClass('placeholder-node')) {
                $next_li.insertBefore($li);
                insertUnderNode = true;
            }

            if (!insertUnderNode) {
                var nextNode = _this.getNodeFrom($next_li);
                siblingParent = nextNode.parent;
                insertAt = nextNode.parent.nodes.indexOf(nextNode);
            }
        }
        else {
            if (parentNode) {
                insertAt = parentNode.nodes ? parentNode.nodes.length : 0;
            }

            var $prev_li = $li.prev();
            if ($prev_li.hasClass('crsa-cm-placeholder') || $prev_li.hasClass('placeholder') || $prev_li.hasClass('placeholder-node')) $prev_li = $prev_li.prev();

            if ($prev_li.length > 0) {
                if ($prev_li.hasClass('placeholder-node')) {
                    insertUnderNode = true;
                }

                if (!insertUnderNode) {
                    var prevNode = _this.getNodeFrom($prev_li);
                    insertAt = prevNode.parent.nodes.indexOf(prevNode) + 1;
                    siblingParent = prevNode.parent;
                }
            }
        }

        if (!siblingParent || insertUnderNode) {
            var closest$li = $li.parent().closest('li');
            if (closest$li.length > 0) {
                insertAt = 0;
                siblingParent = _this.getNodeFrom(closest$li);
            }
            else {
                siblingParent = pgStylesheet.rootNode;
            }
        }

        return {
            index: insertAt,
            parent: siblingParent
        };
    }

    var insertNodeIntoParent = function ($li, node) {
        if (!node) {
            node = _this.getNodeFrom($li) || $li.data('node');
        }
        var siblingParent = false;

        var parentNode;
        var $parent_ul = $li.parent();
        if ($parent_ul.hasClass('sass-root-node')) {
            if (listOfNodes && listOfNodes.length > 0) {
                var $firstNode = $li.parent().find('> li:first-child');
                if ($li[0] == $firstNode[0]) {
                    $firstNode = $li.next();
                }
                parentNode = _this.getNodeFrom($firstNode).parent;
            }
            else {
                parentNode = pgStylesheet.rootNode;
            }
        }
        else {
            var $parent_li = $li.parent().closest('li');
            parentNode = _this.getNodeFrom($parent_li);
        }

        var insertAt = 0;
        var $next_li = $li.next();
        if ($next_li.hasClass('crsa-cm-placeholder') || $next_li.hasClass('placeholder') || $next_li.hasClass('placeholder-node')) $next_li = $next_li.next();

        var insertUnderNode = false;

        if ($next_li.length > 0) {
            if ($next_li.hasClass('placeholder-node')) {
                $next_li.insertBefore($li);
                insertUnderNode = true;
            }

            if (!insertUnderNode) {
                var nextNode = _this.getNodeFrom($next_li);
                siblingParent = nextNode.parent;
                insertAt = nextNode.parent.nodes.indexOf(nextNode);
                node.raws.before = nextNode.raws.before;

                if (options['inversed-insert'] && !$next_li.parent().hasClass('nested')) insertAt++;
            }
        }
        else {
            insertAt = parentNode.nodes ? parentNode.nodes.length : 0;

            var $prev_li = $li.prev();
            if ($prev_li.hasClass('crsa-cm-placeholder') || $prev_li.hasClass('placeholder') || $prev_li.hasClass('placeholder-node')) $prev_li = $prev_li.prev();

            if ($prev_li.length > 0) {
                if ($prev_li.hasClass('placeholder-node')) {
                    insertUnderNode = true;
                }

                if (!insertUnderNode) {
                    var prevNode = _this.getNodeFrom($prev_li);
                    siblingParent = prevNode.parent;
                    insertAt = prevNode.parent.nodes.indexOf(prevNode);
                    node.raws.before = prevNode.raws.before;

                    if (!options['inversed-insert'] || $prev_li.parent().hasClass('nested')) insertAt++;
                }
            }
        }

        if ($parent_ul.parent().hasClass('active-rule-container')) {
            parentNode = siblingParent;
        }

        if (insertUnderNode) {
            siblingParent = _this.getNodeFrom($li.closest('.nested').parent());
        }

        var nodeIndex = -1;
        if (node.parent) nodeIndex = node.parent.nodes.indexOf(node);

        var moveBy = 1, backward = false;
        if (node.parent == parentNode) {
            if (nodeIndex == insertAt - 1) return;
            if (nodeIndex < insertAt && nodeIndex != -1) {
                moveBy = -1;
                backward = true;
                insertAt--;
            }
        }

        var oldNodeIndices = $li.attr('data-node-indices');
        if (node.parent != parentNode && $OldNextli) {
            updateNodeIndicesAfter($OldNextli, node, -1);
            oldNodeIndices = null;
        }

        if (node.parent) {
            var parentChildrenNodes = node.parent.nodes;
            var index = parentChildrenNodes.indexOf(node);
            parentChildrenNodes.splice(index, 1);
        }

        if (node.parent != parentNode) {
            node.parent = parentNode;
        }

        node.parent.nodes.splice(insertAt, 0, node);

        var updateStylePane = false;

        var prevNode = node.prev();

        var isCurrentNodeVar = (node.type == "decl" && node.prop.startsWith(pgStylesheet.getDeclCharachter()));
        var isPrevNodeGroupName = (prevNode && prevNode.type == 'comment' && prevNode.text.startsWith(pgStylesheet.GROUP_TITLE_MARGIN));
        var isPrevNodeVar = (prevNode && prevNode.type == 'decl' && prevNode.prop.startsWith(pgStylesheet.getDeclCharachter()));

        if (isCurrentNodeVar && !isPrevNodeVar && !isPrevNodeGroupName) {
            updateStylePane = true;
        }

        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node).join(',');
        $li.attr('data-node-indices', nodeIndices);

        if (options['inversed-insert'] && !$li.parent().hasClass('nested')) backward = !backward;

        updateNodeChildIndices(nodeIndices, pinegrow.pgPostCSSHelper.getNodeLevel(node), $li.find('> ul > li:first-child'));
        updateNodeIndicesAfter($li, node, moveBy, backward, oldNodeIndices);
        return updateStylePane;
    }

    var elementSelectedAC = function ($elm, value, e, isDecl) {
        if (isDecl) value += ':';
        $elm.text(value);

        var code = e.keyCode || e.which;
        if (code != 9) {
            var e = jQuery.Event( 'keydown', { which: $.ui.keyCode.ENTER } );
            $elm.trigger(e);
        }
    }

    var searchForDeclProp = function ($el) {
        var moreProp = {};
        while (true) {
            var $li = $el.closest('li');
            var $prev = $li.prev();
            while ($prev.length > 0) {
                var node = _this.getNodeFrom($prev);
                if (node) {
                    if (node.type == "decl") {
                        if (node.prop[0] == pgStylesheet.getDeclCharachter()) {
                            if (!moreProp[node.prop]) {
                                moreProp[node.prop] = {
                                    node: node,
                                    el: $prev
                                };
                            }
                        }
                    }
                    else if (node.type == 'atrule' && node.variable) {
                        if (!moreProp[node.name]) {
                            moreProp[node.name] = {
                                node: node,
                                el: $prev
                            };
                        }
                    }
                }
                $prev = $prev.prev();
            }

            var $ul_parent = $li.parent();
            if ($ul_parent.length > 0) {
                if (!$ul_parent.hasClass('sass-root-node')) {
                    $el = $ul_parent;
                }
                else {
                    return moreProp;
                }
            }
            else {
                return moreProp;
            }
        }
    }

    var getVarNameList = function (varGroups) {
        var varNames = [];

        for (var i = 0; i < varGroups.length; i++) {
            var group = varGroups[i];
            for (var j = 0; j < group.length; j++) {
                var currVar = group[j];
                varNames.push(currVar.prop);
            }
        }

        return varNames;
    }

    var addAutoPropCompeleteTo = function ($el) {
        var simpleAutocomplete = new SimpleAutoComplete($el, {
            appendTo: $container,
            focus: function (e, value, $el) {
                $el.text(value);
                crsaSelectText($el[0]);
                return false;
            },
            select: function(e, value, $el) {
                if (value[0] == "@") {
                    elementSelectedAC($el, value, e, false);
                }
                else {
                    elementSelectedAC($el, value, e, true);
                }
                return false;
            },
            source: function (e, value, $el, response) {
                if (propNameArray.length == 0) {
                    propNameArray = Object.keys(autoCompSCSS);
                }

                var results = $.map(propNameArray, function(item) {
                    if (item.startsWith(value)) return item;
                    else return null;
                });

                response(results.slice(0, 6));
            }
        });
    }

    var getLastVarOfValue = function (value) {
        if (pgStylesheet.type == 'css') return null;

        var lastVar = null;
        var varReg = new RegExp('\\' + pgStylesheet.getDeclCharachter() + '[a-z0-9_\\-]*', 'ig');

        var matched = value.match(varReg);
        if (matched && matched.length > 0) lastVar = matched[matched.length - 1];

        return lastVar;
    }

    var addAutoValueCompeleteTo = function ($el) {
        var simpleAutocomplete = new SimpleAutoComplete($el, {
            appendTo: $container,
            focus: function (e, value, $el) {
                var newValue = value;
                if (pgStylesheet.type != 'css' && value.startsWith(pgStylesheet.getDeclCharachter())) {
                    var term = $el.text();

                    var lastVar = getLastVarOfValue(term);
                    if (lastVar && term.endsWith(lastVar)) {
                        newValue  = term.substr(0, term.length - lastVar.length);
                        newValue += value;
                    }
                }

                $el.text(newValue);
                crsaSetCaretPlacement($el[0], false);
                return false;
            },
            select: function(e, value, $el) {
                elementSelectedAC($el, value, e);
                return false;
            },
            source: function (e, value, $el, response) {
                var declPropName = $el.siblings('.decl-prop').text();
                var valueArr = [];

                if (autoCompSCSS[declPropName]) {
                    valueArr = autoCompSCSS[declPropName].values;
                }

                valueArr = valueArr.concat(pgStylesheet.getAutocompleteArr());
                valueArr = valueArr.concat(getVarNameList(pgStylesheet.varGroups));

                var term = value;

                var lastVar = getLastVarOfValue(term);
                if (lastVar && term.endsWith(lastVar)) term = lastVar;

                var results = $.map(valueArr, function(item) {
                    if (item.startsWith(term)) return item;
                    else return null;
                });

                response(results.slice(0, 6));
            }
        });
    }

    var addTriggers = function ($el) {
        $el
            .off('click', 'i.collapse-icon')
            .on('click', 'i.collapse-icon', function (e) {
                e.preventDefault();
                var $li = $(this).closest('li');

                if (pinegrow.keyState.alt) {
                    var collapseState = getLiCollapseState($li);

                    var action = '';
                    if (collapseState == 'uncollapsed') action = 'collapse';
                    else if (collapseState == 'collapsed')   action = 'uncollapse';

                    var $liChildren = $(this).closest('ul').find('> li.can-has-children');
                    $liChildren.each(function (i, li) {
                        var $li = $(li);
                        toggleNodesApperanceFor($li, action);
                    });
                }
                else {
                    toggleNodesApperanceFor($li);
                }
                return false;
            });

        $el
            .off('click', 'li', collapseUncollapseLi)
            .on('click', 'li', collapseUncollapseLi);

        $el
            .off('dblclick', 'li', editLiNode)
            .on('dblclick', 'li', editLiNode);
    }

    var onActiveRulesChanged = function(event) {
        if (onActiveRulesChanged == null) return;
        updateHighlightedRulesWithSelectedElm();
    }

    /*
     Not sure why this is here.

     var updateSourceInEditor = function () {
     // if (editorData) {
     //     var mirror = editorData.mirror;
     //     // getCrsaPageForIframe($.fn.crsa('getSelectedPage')).undoStack.add("Edit CSS rule", false, pgStylesheet);

     //     mirror.getDoc().setValue(pgStylesheet.filterSource(pgStylesheet.getSource(), true));
     //     mirror.getDoc().clearHistory();
     // }
     }
     */

    var addNewListItem = function (node, $li, insertInside, insertAfter, replaced) {
        var arrayIndex = 0;
        var sourceArray = [];

        var hasPgSelected = false;

        var classes = getLiClassesForNode(node);
        var hasUncollapse = classes.indexOf('uncollapse') > -1;
        if (replaced && $li.hasClass('pg-selected-rule')) {
            classes += ' pg-selected-rule';
        }

        var liIndex = arrayIndex;
        sourceArray[arrayIndex++] = '<li draggable="true" data-node-indices="' + pinegrow.pgPostCSSHelper.getNodeIndices(node) + '" class="' + classes + '"' + getMoreAttrForNode(node) + '>';

        sourceArray[arrayIndex++] = '<div>';
        var nodeHtmlText = getStylePaneHtmlForNode(node, hasUncollapse);
        if (nodeHtmlText == "") {
            sourceArray.splice(liIndex, sourceArray.length - liIndex);
            return;
        }
        else {
            sourceArray[arrayIndex++] = nodeHtmlText;
        }

        sourceArray[arrayIndex++] = '</div>';

        sourceArray[arrayIndex++] = '</li>';

        var $newLi = $(sourceArray.join(''));
        if (insertInside) {
            $newLi.appendTo($li);
            tryToRemovePlaceholderNode($li);

            // var $prevLi = $newLi.prev();
            // if ($prevLi.hasClass('placeholder-node')) $prevLi.remove();
        }
        else {
            if (insertAfter) {
                $newLi.insertAfter($li);
            }
            else {
                $newLi.insertBefore($li);
            }
        }

        if (hasUncollapse) {
            var $ul = $('<ul class="nested"></ul>').appendTo($newLi);
            placeNodesInside($ul, node.nodes);
        }

        return $newLi;
    }

    var updateNodeIndices = function (node, $ul, by, deleted) {
        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
        var level = pinegrow.pgPostCSSHelper.getNodeLevelFromIndices(nodeIndices);

        $ul.children().each(function (i, li) {
            var $li = $(li);

            if (!$li.attr('data-node-indices')) return;

            var currNodeIndices = $li.attr('data-node-indices').split(',');
            var currLevel = pinegrow.pgPostCSSHelper.getNodeLevelFromIndices(currNodeIndices);

            var currLevelIndex = parseInt(currNodeIndices[level]) || 0;
            var currParentLevelIndex = parseInt(currNodeIndices[level - 1]) || 0;

            var levelIndex = nodeIndices[level] || 0;
            var parentLevelIndex = nodeIndices[level - 1] || 0;

            if (level <= currLevel && levelIndex <= currLevelIndex && parentLevelIndex == currParentLevelIndex) {

                if (!deleted) currNodeIndices[level] = currLevelIndex + by;
                var rootNode = pinegrow.pgPostCSSHelper.getRootNodeFor(node);
                var currNode = pinegrow.pgPostCSSHelper.getNodeByIndices(rootNode, currNodeIndices);
                if (deleted) currNodeIndices[level] = currLevelIndex + by;

                // Node indices changed
                pinegrow.dispatchEvent('on_css_node_indices_changed', null, {
                    node: currNode,
                    stylesheet: pgStylesheet,
                    oldNodeIndices: $li.attr('data-node-indices').split(','),
                    currNodeIndices: currNodeIndices
                });
                if (options['on-node-indices-changed']) options['on-node-indices-changed'](currNode, pgStylesheet, $li.attr('data-node-indices').split(','), currNodeIndices, _this, $li.find('> ul > li').length > 0);

                $li.attr('data-node-indices', currNodeIndices.join(','));
            }

            var $nestedUl = $li.find('> ul');
            if ($nestedUl.length > 0) {
                updateNodeIndices(node, $nestedUl, by, deleted);
            }
        });
    }

    var updateRulePgId = function (node, $ul, by, deleted) {
        if (node.type != "rule") return;
        var ruleElIdReg = new RegExp('pg-' + fileName + '-rule-id-\([0-9]*\)' + options['extra-rule-el-id']);
        var rulePgId = parseInt(node['pg-node-id']);

        $ul.children().each(function (i, li) {
            var $li = $(li);

            if ($li.hasClass('crsa-rule')) {
                var currRuleId = -1;

                var matched = $li.attr('id').match(ruleElIdReg);
                if (matched && matched.length > 1) {
                    var currRuleId = parseInt(matched[1]);
                    if (rulePgId <= currRuleId) {

                        var currNode = _this.getNodeFrom($li);
                        var currNewNodeId = parseInt(currNode['pg-node-id']);

                        if (deleted) currNewNodeId += by;
                        $li.attr('id', 'pg-' + fileName + '-rule-id-' + currNewNodeId + options['extra-rule-el-id']);

                        if (options['on-rule-id-changed']) options['on-rule-id-changed'](currNode, pgStylesheet, currNewNodeId, currRuleId, _this);
                    }
                }
            }

            var $nestedUl = $li.find('> ul');
            if ($nestedUl.length > 0) {
                updateRulePgId(node, $nestedUl, by, deleted);
            }
        });
    }

    var onNodeAdded = function(page, obj) {
        if (onNodeAdded == null) return;
        if (!obj) return;
        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!$ul.data('properties-in-edit-mode')) {
                //debugger;
                //updateSourceInEditor();
                updateNodeIndices(updatedNode, $ul, 1, false);
                updateRulePgId(updatedNode, $ul, 1, false);

                var insertAt = updatedNode.parent.nodes.indexOf(updatedNode);

                var $parentEl = null;
                if (updatedNode.parent.type == "root") {
                    $parentEl = $ul;
                }
                else {
                    var parentNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode.parent);
                    var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);
                    var newNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, nodeIndices);

                    // TODO Check this
                    // var nodeSelector  = '.' + pgStylesheet.getUniqueName() + '-container' + options['extra-rule-el-id'];
                    //     nodeSelector += ' [data-node-indices="' + parentNodeIndices + '"]';
                    // $parentEl = $(nodeSelector).find('> ul');

                    var $nodeSelector = $ul.find('[data-node-indices="' + parentNodeIndices + '"]');
                    $parentEl = $nodeSelector.find('> ul');
                }

                if ($parentEl.length > 0) {
                    var $mainContainer = $parentEl.closest('.sass-root-node');
                    if ($mainContainer[0] == $ul[0]) {
                        var insertInside = false;
                        var insideAfter = true;

                        // var $nodeEl = $parentEl.find('> li:nth-child(' + insertAt + ')');
                        var $nodeEl = $($parentEl.find('> li:not(.placeholder-node)')[insertAt - 1]);
                        if ($nodeEl.length == 0) {
                            if (insertAt == 0) {
                                insideAfter = false;
                                $nodeEl = $parentEl.find('> li:first-child');
                                while ($nodeEl.hasClass('placeholder-node') || $nodeEl.hasClass('placeholder')) $nodeEl = $nodeEl.next();

                                if ($nodeEl.length == 0) {
                                    $nodeEl = $parentEl;
                                    insertInside = true;
                                }
                                else {
                                    var prevNode = _this.getNodeFrom($nodeEl);
                                    if (prevNode.prev() != updatedNode) return;
                                }
                            }
                            else {
                                $nodeEl = $parentEl.find('> li:last-child');
                                if ($nodeEl.length == 0) {
                                    $nodeEl = $parentEl;
                                    insertInside = true;
                                }
                            }
                        }
                        if ($nodeEl.length == 0) {
                            $nodeEl = $parentEl.find('> ul');
                            insertInside = true;
                        }

                        if (!insertInside && (!$nodeEl.parent().hasClass('nested') || isMediaNode(updatedNode.parent))) {
                            if (listOfNodes && listOfNodes.length > 0) return;
                        }

                        var $newLi = addNewListItem(updatedNode, $nodeEl, insertInside, insideAfter);
                        if (updatedNode.type == "decl" || updatedNode.type == 'atrule') {
                            updateColorPicker($newLi.find('.decl-value,.atrule-params'));
                        }

                        // stylePaneHelper.onListUpdated($newLi.parent());
                        var $prevLi = $newLi.prev();
                        // if ($prevLi.hasClass('add-new-node')) pg$Hide($prevLi);
                        if ($prevLi.hasClass('add-new-node')) $prevLi.detach();
                        if (options['on-node-added']) options['on-node-added'](updatedNode, pgStylesheet, _this, obj);
                    }
                }
            }
        }
    }

    var onNodeUpdated = function(page, obj) {
        if (onNodeUpdated == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;
        var oldValue = obj.oldValue || {};

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!$ul.data('properties-in-edit-mode')) {
                //updateSourceInEditor();
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);
                var currNode = updatedNode;

                // if (currNode) {
                //     var replacedNode = pinegrow.pgPostCSSHelper.replaceNodeWith(currNode, updatedNode);
                //     if (options['on-node-replaced']) options['on-node-replaced'](currNode, replacedNode, pgStylesheet, _this);
                // }

                // TODO Check this
                // var nodeSelector  = '.' + pgStylesheet.getUniqueName() + '-container' + options['extra-rule-el-id'];
                //     nodeSelector += ' [data-node-indices="' + nodeIndices + '"]';
                // var $nodeEl = $(nodeSelector);

                var $nodeSelector  = $ul.find('[data-node-indices="' + nodeIndices + '"]');
                var $nodeEl = $nodeSelector;

                if ($nodeEl.length > 0) {
                    var editedAttr = null;
                    var oldPropValue = null;
                    if (oldValue && oldValue.prop && currNode.type == 'decl') {
                        if (oldValue.prop != currNode.prop) {
                            editedAttr = 'prop';
                            oldPropValue = oldValue.prop;
                        }
                    }
                    if (options['on-node-changed']) options['on-node-changed'](currNode, pgStylesheet, oldPropValue, editedAttr, _this);

                    // if (currNode.type == 'decl') {
                    //     if (oldValue.prop == currNode.prop) {
                    //         if (options['on-node-changed']) options['on-node-changed'](currNode, pgStylesheet, null, null, _this);
                    //     }
                    // }
                    // if (options['on-node-changed']) options['on-node-changed'](currNode, pgStylesheet);
                    _this.updateNodeViewWith($nodeEl, updatedNode);
                }

                // if (updatedNode.type == "atrule") {
                //     if (options['on-node-changed']) options['on-node-changed'](currNode, pgStylesheet, null, null, _this);
                // }

                if (updatedNode.type == "decl" || updatedNode.type == "atrule") {
                    var $declValue = $nodeEl.find('.decl-value,.atrule-params');
                    updateContectedNode($declValue, updatedNode.value);
                }
            }
        }
    }

    var onNodeReplaced = function (page, obj) {
        if (onNodeReplaced == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var newNode = obj.node;
        var oldNode = obj.oldNode;

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!$ul.data('properties-in-edit-mode')) {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(newNode);

                var $nodeEl = $ul.find('[data-node-indices="' + nodeIndices + '"]');
                if ($nodeEl.length > 0) {
                    var $newLi = addNewListItem(newNode, $nodeEl, false, true, true);
                    $newLi.attr('class', $nodeEl.attr('class'));
                    $nodeEl.remove();

                    if (options['on-node-replaced']) options['on-node-replaced'](oldNode, newNode, pgStylesheet, _this);
                }
            }
        }
    }

    var updateNodeIndicesForUl = function (updatedNodeIndices, by) {
        $ul.find('[data-node-indices]').filter(function() {
            var indices = $(this).attr('data-node-indices').split(',');

            if (indices.length >= updatedNodeIndices.length) {
                var nodeIndex = updatedNodeIndices.length - 1;
                var parentIndex = nodeIndex - 1;

                var updatedNodeLastIndex = updatedNodeIndices[parentIndex] || -1;
                var currentNodeLastIndex = parseInt(indices[parentIndex]) || -1;

                if (updatedNodeLastIndex == currentNodeLastIndex) {
                    if (updatedNodeIndices[nodeIndex] < parseInt(indices[nodeIndex])) {
                        var node = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, indices);
                        indices[nodeIndex] = parseInt(indices[nodeIndex]) + by;

                        $(this).attr('data-node-indices', indices.join(','));
                    }
                }
            }
        });
    }

    var tryToAddPlaceholderNode = function ($ul) {
        if ($ul.length > 0) {
            if ($ul.find('> .crsa-node').length == 0) addPlaceholderNode($ul);
        }
    }

    var onNodeBeforeDeleted = function (page, obj) {
        if (onNodeBeforeDeleted == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!$ul.data('properties-in-edit-mode')) {
                //updateSourceInEditor();
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);
                var currNode = updatedNode;

                // if (options['on-before-node-deleted']) options['on-before-node-deleted'](currNode, pgStylesheet);

                var $nodeEl = $ul.find('[data-node-indices="' + nodeIndices + '"]');
                if ($nodeEl.length > 0) {
                    var $ulParent = $nodeEl.closest('.sass-root-node');
                    $nodeEl.attr('data-removed', true);
                    // updateNodesDomIds($nodeEl);

                    if (options['on-before-node-deleted']) {
                        if (currNode.nodes && currNode.nodes.length > 0) {
                            var children = currNode.nodes;
                            for (var i = 0; i < children.length; i++) {
                                options['on-before-node-deleted'](children[i], pgStylesheet, _this);
                            }
                        }

                        options['on-before-node-deleted'](currNode, pgStylesheet, _this);
                    }

                    removeNode$li($nodeEl);
                    var parentNode = updatedNode.parent;
                    if (parentNode.type == 'root' && parentNode.nodes.length <= 1) {
                        stylePaneHelper.onNodeRemoved();
                    }
                }

                var by = 1 + numOfRulesInsideNode(updatedNode);

                updateRulePgId(updatedNode, $ul, -1 * by, true);
                updateNodeIndices(updatedNode, $ul, -1, true);

                // updateNodeIndicesForUl(nodeIndices, -1);
            }
        }
    }

    var onNodeIndicesChanged = function (page, obj) {
        if (onNodeIndicesChanged == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var updatedIndicesNode = obj.node;
        var oldNodeIndices = obj.oldNodeIndices;

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!$ul.data('properties-in-edit-mode')) {
                var newNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedIndicesNode);

                // TODO Check this
                // var nodeSelector  = '.' + pgStylesheet.getUniqueName() + '-container' + options['extra-rule-el-id'];
                //     nodeSelector += ' [data-node-indices="' + oldNodeIndices + '"]';
                // var $nodeEl = $(nodeSelector);

                var nodeSelector  = $ul.find('[data-node-indices="' + oldNodeIndices + '"]');
                var $nodeEl = $nodeSelector;

                if ($nodeEl.length > 0) {
                    var $mainContainer = $nodeEl.closest('.sass-root-node');
                    if ($mainContainer[0] == $ul[0]) {
                        $nodeEl.attr('data-node-indices', newNodeIndices.join(','));
                    }
                }
            }
        }
    }


    this.updateNodeViewWith = function ($el, node) {
        if (!node || $el.length == 0) return;

        var $mainContainer = $el.closest('.sass-root-node');
        if ($mainContainer.data('properties-in-edit-mode')) return;

        var $div = $el.find('> div')
        if (node.type == "rule") {
            $div.find('span.rule-selector').html(this.pgStylesheet ? this.pgStylesheet.filterSource(node.selector) : node.selector);
        }
        else if (node.type == "decl" || (node.type == 'atrule' && node.variable)) {
            var prop = node.prop || (pgStylesheet.getDeclCharachter() + node.name);
            var value = node.value.replace('!important', '').trim();
            var $declProp = $div.find('span.decl-prop');
            var $declVal  = $div.find('span.decl-value');

            pgRemoveSpectrum($declVal, '.color-input');

            $declProp.html(prop);
            $declVal.html(value + (node.important ? ' !important' : ''));

            updateColorPicker($declVal);
        }
        else if (node.type == "atrule") {
            $div.find('span.atrule-name').html('@' + node.name);
            var $atruleParams = $div.find('span.atrule-params').html(node.params);

            updateColorPicker($atruleParams);
        }
        else if (node.type == "import") {
            $div.find('span.import-name').html('@' + node.name);
            $div.find('span.import-importPath').html(node.importPath);
        }
        else if (node.type == "comment") {
            $div.find('span.comment-text').html(node.text);
        }
    }

    this.getLengthWithoutPlaceholders = function (node) {
        var nodesLength = -1;
        if (node.nodes) {
            nodesLength = node.nodes.length;
        }

        return nodesLength;
    }

    this.showOnlySelectedRules = function () {
        updateHighlightedRulesWithSelectedElm();

        $ruleList.find('> ul > li').removeClass('show-rule');
        $.each($container.find('li.selected'), function (i, li) {
            var node = _this.getNodeFrom($(li));

            while (node && node.parent && node.parent.type != "root") {
                $(li).siblings().not('.selected').addClass('hide-rule');
                li = $(li).parent().closest('li');
                var node = node.parent;
            }

            if (node && node.parent && li) {
                $(li).removeClass('hide-rule').addClass('show-rule');
            }
        });

        $ruleList.find('> ul > li').not('.show-rule').addClass('hide-rule');
    }

    this.showAllRules = function () {
        $container.find('li.hide-rule').removeClass('hide-rule');
    }

    this.show = function () {
        if (pgStylesheet.getError()) return;

        var nodes;
        if (listOfNodes) nodes = listOfNodes;
        else             nodes = pgStylesheet.rootNode.nodes;

        $ul.data('node', pgStylesheet.rootNode);
        convertToTree(nodes, $ul);
        stylePaneHelper.setUl($ul);
        stylePaneHelper.onListUpdated();

        addTriggers($ul);
        stylePaneHelper.addTriggers($ul);

        if(!pgf.kids) {
            stylePaneHelper.supportInsertBorders();
        }

        stylePaneHelper.isAllowed = this.isAllowedToEditNode;
        stylePaneHelper.isDblClickAllowed = this.isDblClickEditingAllowed;
        stylePaneHelper.getSelectableObj = this.getSelectableObjForEditing;

        updateHighlightedRulesWithSelectedElm();
    }

    this.remove = function () {
        destroyed = true;
        pgRemoveSpectrum($ul, '.color-input');
        // var $selectedRule = $ul.find('.pg-selected-rule');
        // if ($selectedRule.length) selectRule(null);

        $ul.data('node', null);
        $ruleList.data('pg-sp', null);
        if (stylePaneHelper) {
            stylePaneHelper.remove();
            stylePaneHelper = null;
        }

        // event listeners
        pinegrow.removeEventHandler('on_css_node_added', onNodeAdded);
        pinegrow.removeEventHandler('on_css_node_updated', onNodeUpdated);
        pinegrow.removeEventHandler('on_css_node_replaced', onNodeReplaced);
        pinegrow.removeEventHandler('on_css_before_node_deleted', onNodeBeforeDeleted);
        pinegrow.removeEventHandler('node-indices-changed', onNodeIndicesChanged);
        pinegrow.removeEventHandler('on_css_active_rule_changed', onActiveRulesChanged);

        $('body')
            .off('crsa-rule-selected', onRuleSelected);

        $('#crsa-dummy-field')
            .off('keydown', dummyFieldKeydown)
            .off('keyup', dummyFieldKeyup);

        if ($scrollParent) $scrollParent.off('scroll', scrollTrigger);

        scrollTrigger = null;
        onActiveRulesChanged = null;
        onNodeUpdated = null;
        onNodeReplaced = null;
        onNodeBeforeDeleted = null;
        onNodeAdded = null;
        dummyFieldKeydown = null;
        dummyFieldKeyup = null;

        this.pgStylesheet.stylePane = pgStylesheet.stylePane = null;
        this.pgStylesheet = pgStylesheet = null;

        _this = null;
    }

    // $.getJSON( "scss-prop.json", function( data ) {
    //     autoCompSCSS = data;
    // });

    $('body').on('crsa-rule-selected', onRuleSelected);

    pinegrow.addEventHandler('on_css_node_added', onNodeAdded);
    pinegrow.addEventHandler('on_css_node_updated', onNodeUpdated);
    pinegrow.addEventHandler('on_css_node_replaced', onNodeReplaced);
    pinegrow.addEventHandler('on_css_before_node_deleted', onNodeBeforeDeleted);
    pinegrow.addEventHandler('node-indices-changed', onNodeIndicesChanged);
    pinegrow.addEventHandler('on_css_active_rule_changed', onActiveRulesChanged);
}
