window.PgStylesheet = function () {
    this.index = '';
    this.localFile = null;
    this.url = null;
    this.name = null;
    this.outFile = null;
    this.source = null;
    this.parsedSource = null;
    this.changed = false;
    this.changedOnce = false;
    this.appliedToPage = false;
    this.usedExistedCompiled = false;
    this.loaded = false;
    this.error = null;
    this.codeError = null;
    this.loadingFailed = false;
    this.ignore = false;
    this.compiledFile = false;
    this.nakedUrl = null;
    this.parentPgStylesheets = [];
    this.rootNode = null;
    this.importedFiles = [];
    this.recursiveImports = null;
    this.emptyNodes = [];
    this.varGroups = [];
    this.ignoreInSave = false;
    this.locked = false;
    this.lockedReason = null;
    this.needFormatting = false;
    this.formattingOptions = {};
    this.mappedRemote = false;

    this.isImported = function () {
        return this.parentPgStylesheets.length > 0
    }

    this.isImportedFor = function (page) {
        var csInfo = pinegrow.stylesheets.getStylesheetInfo(this, page);
        return csInfo && csInfo.parentPgStylesheet ? true : false;
    }

    this.setLoadingFailed = function(val) {
        this.loadingFailed = val;
        if(val) {
            pinegrow.showQuickError('Could not load ' + this.url);
        }
    }

    this.setLocked = function(v, reason, no_save) {
        this.locked = v;
        this.lockedReason = reason;
        if(!no_save) {
            this.setSetting({
                locked: v,
                locked_reason: reason
            });
            pinegrow.dispatchEvent('on_css_stylesheet_locked_changed', null, this, v);
        }
    }

    this.setNeedFormatting = function (v) {
        this.needFormatting = v;
    }

    this.isLocked = function() {
        return this.locked;
    }

    this.unlockWithAsk = function(done) {
        var _this = this;

        var msg = `<p>Are you sure you want to unlock this stylesheet?</p>`;
        if(!this.localFile && this.url) {
            msg += `<p>This is a remote stylesheet that can't be saved.`;
        }

        if(this.lockedReason) {
            msg += `<p>${this.lockedReason}</p>`;
        }
        pinegrow.showAlert(msg, 'Unlock the stylesheet?', 'Cancel', 'Yes, unlock it', null, function() {
            _this.setLocked(false, '');
            pinegrow.selectedElements.reselect();

            if(done) done();
        })
    }

    var locked_msg_displayed = false;

    this.canEdit = function(no_msg, rule) {
    //    return true;

        //disable for now
        if(this.locked && !no_msg) {
            pinegrow.selectedElements.reselect();
            if(!locked_msg_displayed) {
                locked_msg_displayed = true;
                var msg = `<p>${this.lockedReason || 'This stylesheet is locked.'}</p>`;

                if(rule) {
                    msg += `<p><b>Right-click on the CSS rule</b> in the Active Style tab to <b>override</b> or <b>duplicate</b> it in the editable stylesheet.</p>`;
                }
                pinegrow.showAlert(msg, `${this.name} is locked`, null, 'OK', function() {
                    locked_msg_displayed = false;
                }, function() {
                    locked_msg_displayed = false;
                });
            }
        }
        return !this.locked;
    }

    this.loadAndShowInActive = function(done) {
        this.ignore = false;
        var cStylesheet = null;
        var csInfo = pinegrow.stylesheets.getStylesheetInfo(this, pinegrow.getSelectedPage());
        if (csInfo && csInfo.compiledStylesheet) {
            cStylesheet = csInfo.compiledStylesheet;
            cStylesheet.ignore = this.ignore;
        }
        else {
            cStylesheet = this;
        }

        var doneLoading = function () {
            $('body').trigger('crsa-stylesheets-changed');
            crsaQuickMessage("CSS rules list updated.");
            done && done();
        }

        if(!this.loaded) {
            cStylesheet.load(cStylesheet.url, function() {
                doneLoading();
            });
        } else {
           this.regenerateAndSetCssSource(doneLoading);
        }
    }

    this.ignoreInActive = function() {
        var cs = this;

        var doCs = function(cs) {
            cs.ignore = true;
            if (cs.compiledStylesheets.length > 0) {
                cs.compiledStylesheets.forEach(function (compiledStylesheet) {
                    compiledStylesheet.ignore = cs.ignore;
                });
            }
            cs.parentPgStylesheets && cs.parentPgStylesheets.forEach(doCs);
        }

        var doIgnore = function() {
            doCs(cs);
            $('body').trigger('crsa-stylesheets-changed');
        }

        if(this.changed && this.localFile) {
            var $alert = showAlert(`The stylesheet ${cs.name} has unsaved changes. Hiding the stylesheet will cause these changes to be lost.`,  "The stylesheet has unsaved changes", "Cancel", "Save &amp; Hide", null, function() {
                cs.save(cs.localFile, function() {
                    pinegrow.refreshPageChangedStatusForAllPages(true);
                    doIgnore();
                });
            })
            $('<a href="#" class="btn pull-left">Don\'t save</a>').appendTo($alert.find('.modal-footer')).on('click', function(e) {
                e.preventDefault();
                $alert.modal('hide');
                doIgnore();
            });
        } else {
            doIgnore();
        }
    }

    this.addParentStylesheet = function (parentCs) {
        if (this.parentPgStylesheets.indexOf(parentCs) < 0) {
            this.parentPgStylesheets.push(parentCs);
        }
    }

    this.forEachStylesheet = function(func, skipFirst) {
        var result = null;
        if (!skipFirst) {
            result = func(this);
        }
        if (result != false) {
            var importedFiles = this.importedFiles;
            for (var i = 0; i < importedFiles.length; i++) {
                var impFile = importedFiles[i];
                var result = impFile.forEachStylesheet(func, false);
                if (result == false) break;
            }
        }
        return result;
    }

    this.forEachParentCS = function (func, skipFirst) {
        var cont = true;
        if (!skipFirst) {
            if (func(this) == false) cont = false;
        }
        if (cont) {
            this.parentPgStylesheets.forEach(function (cs) {
                cs.forEachParentCS(func, false);
            })
        }
    }

    this.isRemote = function() {
        return crsaIsRemoteUrl(this.url);
    }

    var getProjectInfo = function(cs) {
        var pi = null;
        var p = pinegrow.getCurrentProject();
        if (p) {
            pi = p.getProjectInfo();
            pi.parent_url = p.getUrl();
        } else {
            var pages = pinegrow.stylesheets.getPagesForStylesheet(cs);
            if(pages.length) {
                var page = pages[0];
                pi = page.getProjectInfo();
                pi.parent_url = page.url;
            }
        }
        if(pi && pi.parent_url) {
            pi.relative_url = crsaMakeLinkRelativeTo(cs.url, pi.parent_url);
        }
        return pi;
    }

    this.setSetting = function(key, value) {
        var pi = getProjectInfo(this);
        if(pi && this.url) {
            if(typeof key == 'string') {
                var d = {}
                d[key] = value;
                key = d;
            }
            var _this = this;
            $.each(key, function(k, v) {
                pi.setSettingForUrl(_this.url, k, v);
            })
            pi.save();
        }
    }

    this.getSetting = function(key) {
        var pi = getProjectInfo(this);
        if(pi && this.url) {
            return pi.getSettingForUrl(this.url, key);
        }
        return null;
    }

    this.fast = false; //when true, CS is handled without ids and rule parsing

    this.atruleWithNoChild = ['keyframes', '-webkit-keyframes', '-moz-keyframes', '-ms-keyframes', '-o-keyframes'];

    this.sourceStylesheet = null;
    this.compiledStylesheets = [];

    // Watch file vars
    this.isWatchingFileForChanges = false;
    this.watchReloadDialog = null;
    this.lastSavedAt = 0;
    this.GROUP_TITLE_MARGIN = "====";

    var _this = this;
    var breakpoints_units_re = '(px|em|rem|pt)';

    var uncollapsed_map = {};

    var delayed = new PgDelayedTasks();

    this.canAddRuleIdsSuper = function (rule) {
        var parent = rule.parent;
        if (parent && parent.type == 'atrule' && this.atruleWithNoChild.indexOf(parent.name) > -1) return false;

        return true;
    }

    this.canAddRuleIds = function (rule) {
        return this.canAddRuleIdsSuper(rule);
    }

    this.setCollapsedNode = function(node_id, value) {
        if(node_id) {
            uncollapsed_map[node_id] = value;
        }
    }

    this.setIgnore = function (value) {
        this.ignore = value;
        if (this.sourceStylesheet)   this.sourceStylesheet.ignore = value;
        if (this.compiledStylesheets.length > 0) {
            this.compiledStylesheets.forEach(function (compiledStylesheet) {
                compiledStylesheet.ignore = value;
            });
        }
    }

    this.getCollapsedNode = function(node_id) {
        return node_id ? (uncollapsed_map[node_id] || false) : false;
    }

    this.getUniqueName = function () {
        var indices = '-' + this.index;

        // var cs = this.parentPgStylesheet;
        // while (cs) {
        //     indices = '-' + cs.index + indices;
        //     cs = cs.parentPgStylesheet;
        // }

        return 'cs' + indices;
    }

    this.getUniqueId = function() {
        return this.index;
    }

    this.getFullUrlFor = function (relativeUrl, underscore) {
        var ext = `.${this.fileType}`;
        if (relativeUrl.endsWith(ext)) ext = '';

        var relUrlArr = relativeUrl.split("/");
        if (underscore) {
            var name = relUrlArr.pop();
            relUrlArr.push("_" + name);
        }

        if(crsaIsAbsoluteUrl(relativeUrl)) {
            return `${relUrlArr.join('/')}${ext}`
        } else {
            return crsaCombineUrlWith(this.url, `${relUrlArr.join('/')}${ext}`);
        }
    }

    this.setNakedUrl = function (url) {
        if (!url) return '';
        url = getUrlWithoutProxy(url);
        this.setUrl(url);
        var naked_url = crsaRemoveUrlParameters(url);
        if(pinegrow.httpServer) naked_url = pinegrow.httpServer.getOriginalUrl(naked_url);

        return naked_url;
    }

    this.unload = function() {
        pinegrow.undoManager.purgeCs(this);
        this.setLocalFile(null);
        if (this.watchReloadDialog) {
            try {
                this.watchReloadDialog.modal('hide');
            }
            catch(err) {}
        }
        pinegrow.dispatchEvent('on_stylesheet_unloaded', null, this);
    }

    this.setUrl = function(url) {
        var old_url = this.url;

        if (old_url) {
            if (url != old_url) {
                pinegrow.stylesheets.updateStylesheetByKey(old_url, url, this);
                pinegrow.stylesheets.updateUsedStylesheet(old_url, url, this);
            }
        }
        else {
            pinegrow.stylesheets.addStylesheetByKey(url, this);
        }

        this.url = url;
        this.name = url.split("/").pop();

        var localFile = crsaIsFileUrl(url) ? crsaMakeFileFromUrl(url) : null;

        if(localFile != this.localFile) {
            this.setLocalFile(localFile);
        }

        var locked = this.getSetting('locked');
        var lockedReason = this.getSetting('locked_reason');

        if(locked !== null) {
            this.locked = locked;
            this.lockedReason = lockedReason;
        } else {
            if (url && old_url !== url && !localFile && isApp()) {
                this.locked = true;
                this.lockedReason = `Remote stylesheet can't be saved.`;
            }
        }
    }

    this.setLocalFile = function(file) {
        if(this.isWatchingFileForChanges && this.localFile) {
            this.stopWatchingFileForChanges();
        }
        if (file) {
            file = file.replace(/\.css$/i, '.' + this.fileType);
        }

        this.localFile = file;

        if(this.localFile) {
            this.watchFileForChanges();
        }
    }

    this.getLocalFileName = function () {
        if (!this.localFile) return null;
        return this.localFile;
    }

    this.getDecodedLocalFileName = function () {
        if (!this.localFile) return null;
        var url = pgDecodeUrl(this.url);
        return crsaMakeFileFromUrl(url);
    }

    this.isImportedFilesChanged = function (currentImportList) {
        var importedFiles = this.importedFiles;
        if (importedFiles.length == currentImportList.length) {
            for (var i = 0; i < importedFiles.length; i++) {
                if (importedFiles[i].url != currentImportList[i].cs.url) {
                    return true;
                }
            }
        }
        else {
            return true;
        }

        return false;
    }

    this.getMainParnetStylesheetFor = function (page) {
        var stylesheet = this;
        while (true) {
            var csInfo = pinegrow.stylesheets.getStylesheetInfo(stylesheet, page);
            if (csInfo && csInfo.parentPgStylesheet) {
                stylesheet = csInfo.parentPgStylesheet;
            }
            else {
                break;
            }
        }

        return stylesheet;
    }

    this.getMainParnetStylesheets = function () {
        var mainParentList = [];

        this.forEachParentCS(function (cs) {
            if (cs.parentPgStylesheets.length == 0) {
                mainParentList.push(cs);
            }
        });

        return mainParentList;
    }

    this.setRecursiveImportsError = function (error) {
        var parentStylesheets = this.getMainParnetStylesheets();
        parentStylesheets.forEach(function (parentStylesheet) {
            parentStylesheet.recursiveImports = error;
        });
    }

    this.checkRecursiveImports = function (url, showErrorMsg) {
        var forAllCsParents = function (parentPgStylesheets) {
            for (var i = 0; i < parentPgStylesheets.length; i++) {
                var parentCS = parentPgStylesheets[i];

                if (url == parentCS.url) {
                    return true;
                }

                if (parentCS.parentPgStylesheets.length > 0) {
                    var value = forAllCsParents(parentCS.parentPgStylesheets);
                    if (value) return true;
                }
            }
            return false;
        }

        if (forAllCsParents([this])) {
            if (showErrorMsg) {
                pinegrow.showQuickError(`Recursive imports: one of <code>${crsaGetNameFromUrl(url)}</code> parents has the same URL.`);
            }
            this.recursiveImports = "Loop of imports";
            return true;
        }
        return false;
    }

    this.isAlreadyImported = function(url) {
        //prevent cyclic import
        //is cs is already imported
        var cs = this;
        do {
            /*
            for (var i = 0; i < cs.importedFiles.length; i++) {
                if (cs.importedFiles[i].url === url) {
                    //console.log('Already imported', url, cs)
                    return cs;
                }
            }*/
            if(cs.url === url) return cs;
            cs = cs.parentPgStylesheet;
        } while(cs);

        return false;
    }

    this.loadAllImportedFiles = function (done) {
        var list = this.importedFiles;
        var allImportedFiles = this.getImportedFiles();
        if (this.isImportedFilesChanged(allImportedFiles)) {
            list = allImportedFiles;
            this.importedFiles = [];
        }
        else {
            if (done) done();
            return;
        }

        //console.log('Loading all impprted files');

        var count = list.length;
        if (count == 0) {
            if (done) done();
            return;
        }

        var donesync = function () {
            count--;
            if (count == 0) {
                if (done) done();
            }
        }

        for (var i = 0; i < list.length; i++) {
            var pair = list[i];

            var already_cs = this.isAlreadyImported(pair.href);
            if(!already_cs) {
                this.importedFiles.push(pair.cs);

                if (pair.cs.loaded) {
                    donesync();
                }
                else {
                    pair.cs.loadFromUrl(pair.href, function () {
                        donesync()
                    });
                }
            } else {
                console.log(`${pair.href} is already imported in ${already_cs.url}!`);
                donesync();
            }
        }
    }

    this.setChanged = function(v) {
        this.changed = v;
        if (v) this.changedOnce = true;

        if (false && pgIsDev()) {
            console.log('------');
            console.log(`${this.name} changed: ${v}`);
            console.log("The following is not an error, it's there for debugging purposes");
            console.log(new Error().stack);
            console.log('------');
        }
    }

    var nodesNeedFormatting = 0; // for performance
    this.setNodeNeedFormatting = function (node, value, allNodesInside) {
        if (nodesNeedFormatting == 0 && !value) return;

        if (allNodesInside) {
            pgcssh.forAllNodes(node, function (childNode) {
                childNode.need_formatting = value;

                if (value) {
                    nodesNeedFormatting++;
                } else {
                    nodesNeedFormatting--;
                }
            });
        }
        else {
            node.need_formatting = value;

            if (value) {
                nodesNeedFormatting++;
            } else {
                nodesNeedFormatting--;
            }
        }
    }

    this.hasNodesNeedFormatting = function () {
        if (nodesNeedFormatting > 0) {
            return true;
        } else {
            return false;
        }
    }

    this.shouldFormatFile = function () {
        // needFormatting does not work with inline
        // Each change in the inline stylesheet using code editor create new loadStylesheet event
        return pinegrow.autoFormatCss != 'off' && (this.needFormatting || this.inline);
    }

    this.updateNodeFormatting = function (node, forceFormatting) {
        var shouldFormat = forceFormatting || this.shouldFormatFile();
        var nodesNeedFormatting = this.hasNodesNeedFormatting();

        if (shouldFormat || nodesNeedFormatting) {
            pgcssf.updateNodeRaws(node, this.inline, this.formattingOptions, shouldFormat, this);
            this.setNeedFormatting(false);
        }
    }

    this.formatCode = function (done) {
        var newSource = this.getSourceForRules(null, false, true);
        this.setSource(newSource, done, true, !this.inline);
    }

    this.filterSource = function (source, editor) {
        if (this.shouldFormatFile()) {
            return source.trim() + (this.inline ? '' : "\n");
        }
        return source;
    }

    this.getImportedStylesheetsByUniqueName = function (uniqueName) {
        var importedStylesheet = null;
        this.forEachStylesheet(function (cs) {
            if (cs.getUniqueName() == uniqueName) {
                importedStylesheet = cs;
                return false;
            }
        });

        return importedStylesheet;
    }

    this.getImportedStylesheets = function(list) {
        list = list || [];
        this.importedFiles.forEach(function(cs) {
            list.push(cs);
            cs.getImportedStylesheets(list);
        })
        return list;
    }

    this.needsSave = function(path) {
        path = path || this.localFile;
        if(!path && !this.url) return false;

        var changed = this.changed;

        var forAll = function (list) {
            for (var i = 0; i < list.length; i++) {
                var currStylesheet = list[i];
                changed = changed || currStylesheet.changed;

                var importedStylesheet = currStylesheet.importedFiles;
                if (importedStylesheet.length > 0) forAll(importedStylesheet);
            }
        }

        forAll(this.importedFiles);
        // this.importedFiles.forEach(function(cs) {
        //     changed = changed || cs.changed;
        // })

        if(!changed) {
            if(path) {
                if(crsaIsFileExist(path)) {
                    return false; //all ok
                } else {
                    if(this.mappedRemote) {
                        return false; //this is unchanged remote file that is mapped
                    } else {
                        return true; //not changed but not written yet
                    }
                }
            } else {
                return false; //not changed, remote, no need to save
            }
        }
        return true;
    }

    this.getNodeId = function (node) {
        return node['pg-node-id'];
    }

    this.getHtmlSourceForRule = function (node) {
        var sourceCode = "";
        var forAllNodes = function (nodes) {
            for (var i=0; i < nodes.length; i++) {
                var currentNode = nodes[i];

                if (currentNode.type == "comment") {
                    var ret = _this.filter(currentNode);
                    if (ret.haveTo != "delete") {
                        sourceCode += currentNode.raws.before;

                        sourceCode += '<span class="cm-cp-raw">';
                        if (node.raws.inline) sourceCode += '//';
                        else                  sourceCode += '/*';
                        sourceCode += currentNode.raws.left + currentNode.text + currentNode.raws.right;

                        if (!node.raws.inline) sourceCode += '*/';
                        sourceCode += '</span>';
                    }
                } else if (currentNode.type == "decl") {
                    var ret = _this.filter(currentNode);
                    if (ret.haveTo != "delete") {
                        sourceCode += currentNode.raws.before;
                        sourceCode += '<span class="cm-cp-prop">' + currentNode.prop + '</span>' + currentNode.raws.between + '<span class="cm-cp-val">' + currentNode.value + (currentNode.important ? ' !important' : '') + '</span>;';
                    }
                } else if (currentNode.type == "rule") {
                    sourceCode += currentNode.raws.before;
                    var selector = currentNode.selector;
                    var ret = _this.filter(currentNode);
                    if (ret.haveTo == "update") {
                        selector = ret.newValue;
                    }

                    sourceCode += '<span class="cm-cp-sel">' + selector + '</span>' + currentNode.raws.between
                    if (!currentNode.ruleWithoutBody) {
                        sourceCode += "{";
                    }
                } else if (currentNode.type == "atrule") {
                    sourceCode += currentNode.raws.before;
                    currentNode.name = currentNode.name.replace(/@/g, '');
                    sourceCode += '<span class="cm-cp-media">@' + currentNode.name + '</span>' + currentNode.raws.afterName + '<span class="cm-cp-params">' + currentNode.params + currentNode.raws.between + '</span>';
                    if (currentNode.nodes && !currentNode.ruleWithoutBody) {
                        sourceCode += "{";
                    }
                    else {
                        sourceCode += ";";
                    }
                }

                if (currentNode.nodes) {
                    if (currentNode.nodes.length) {
                        forAllNodes(currentNode.nodes);
                    }
                    sourceCode += currentNode.raws.after + "}";
                }
            }
        }

        forAllNodes([node]);
        return sourceCode.trim();
    }

    this.watchFileOnFileChanged = function(curr, prev) {
        var fs = require('fs');

        if (curr.mtime > prev.mtime && pinegrow.getSetting('auto-reloading', '1') == '1') {
            var ct = (new Date()).getTime();
            if (_this && ct - _this.lastSavedAt > 5000) {

                var file_source;

                try {
                    file_source = fs.readFileSync(_this.localFile, {encoding: "utf8"});
                } catch(err) {
                    return;
                }
                console.log(_this.url + ' changed');

                pinegrow.codeEditors.isCodeForUrlSameInExternalEditor(_this.url, /* _this.filterSource(_this.getSource())*/ file_source, function(same) {

                    if (_this.watchReloadDialog) {
                        try {
                            _this.watchReloadDialog.modal('hide');
                            _this.watchReloadDialog = null;
                        } catch (err) {
                        }
                    }

                    if(same) return;

                    if (_this.changed) {
                        _this.watchReloadDialog = showAlert('<p><b>' + _this.name + '</b> was modified outside of Pinegrow. The file in Pinegrow has unsaved changes. Do you want to reload it?</p><p><em>You can disable auto-reloading in Support -&gt; Settings.</em></p>', "Unsaved file modified outside of Pinegrow", "Don't reload", "Reload", function () {
                            _this.watchReloadDialog = null;
                        }, function () {
                            _this.watchReloadDialog = null;
                            _this.reload(function () {
                                $('body').trigger('crsa-stylesheets-changed', {cs: _this});
                                pinegrow.dispatchEvent('on_css_loaded_from_external', pinegrow.stylesheets.getPagesForStylesheet(_this), _this);
                            }, true);
                        });
                    } else {
                        _this.reload(function () {
                            $('body').trigger('crsa-stylesheets-changed', {cs: _this});
                            pinegrow.dispatchEvent('on_css_loaded_from_external', pinegrow.stylesheets.getPagesForStylesheet(_this), _this);
                        }, true);
                    }
                })
            }
        }
    }

    this.watchFileForChanges = function() {
        if(this.localFile && pinegrow.getSetting('auto-reloading', '1') == '1') {
            var fs = require('fs');
            //fs.watchFile(this.localFile, {persistent: true, interval: 1102}, this.watchFileOnFileChanged);
            if(this.nodeFileWatcher) {
                this.nodeFileWatcher.close();
                this.nodeFileWatcher = null;
            }
            try {
                this.nodeFileWatcher = fs.watch(this.localFile, {persistent: true}, function (event, filename) {
                    delayed.run('reloadFile', function () {
                        var curr = {
                            mtime: 1
                        }
                        var prev = {
                            mtime: 0
                        }
                        _this && _this.watchFileOnFileChanged(curr, prev);
                    }, 10);
                })
                this.isWatchingFileForChanges = true;
            } catch(err) {}
        }
    }

    this.stopWatchingFileForChanges = function() {
        if(this.localFile && this.isWatchingFileForChanges) {
            //var fs = require('fs');
            //fs.unwatchFile(this.localFile, this.watchFileOnFileChanged);
            if(this.nodeFileWatcher) {
                this.nodeFileWatcher.close();
                this.nodeFileWatcher = null;
            }
            this.isWatchingFileForChanges = false;
        }
    }

    this.insertRule = function (rule, parent, index) {
        parent = parent || this.getRootNode();
        index = index || parent.nodes.length;

        parent.nodes[index] = rule;
        rule.parent = parent;
    }

    this.deleteRule = function (rule) {
        var parent = rule.parent;
        var index = parent.nodes.indexOf(rule);

        if (index > -1) {
            parent.nodes.splice(index, 1);
        }
    }

    this.getAllRules = function () {
        if (!this.rootNode) return [];
        return this.rootNode.nodes;
    }

    this.getRules = function (done) {
        if (this.rootNode) {
            done(this.rootNode);
        }
        else {
            this.regenerateRules(function (rules) {
                done(rules);
            })
        }
    }

    this.getRootNode = function () {
        return this.rootNode;
    }

    this.getRuleByIndex = function(i, parent) {
        parent = parent || this.getRootNode();

        return parent.nodes.length > i ? parent.nodes[i] : null;
    }

    this.getCascadeError = function () {
        var error = null;
        this.forEachParentCS(function (cs) {
            var error = cs.getError();
            if (error) return false;
        });

        return error;
    }

    this.getAllImportedFiles = function (filter) {
        var allImportedFiles = [];
        var forAll = function (list) {
            for (let i = 0; i < list.length; i++) {
                var importedFile = list[i];
                var filterValue = true;
                if (filter) filterValue = filter(importedFile);
                if (filterValue === false) {
                    break;
                }
                else if (filterValue === undefined) {
                    filterValue = true;
                }

                if (filterValue) {
                    allImportedFiles.push(importedFile);

                    var newImportedFiles = importedFile.importedFiles;
                    if (newImportedFiles && newImportedFiles.length > 0) {
                        forAll(newImportedFiles);
                    }
                }
            }
        }

        forAll(this.importedFiles);
        return allImportedFiles;
    }

    this.attachAndFill = function (page, done, replacePgel, first_css) {//csspgdone
        var cs = this;
        if (this.inline) {
            PgStylesheetHelper().loadInlineStylesheet(page, function (pgStyle) {
                cs.inlineSheetPgel = pgStyle;
                cs.setAllCssText(cs.source);
                if (cs.loaded) {
                    cs.getCssSource(function(css) {
                        if (!cs.getError()) {
                            pinegrow.stylesheets.applyToPages(cs, done);
                        } else {
                            if (done) done();
                        }
                    });
                } else {
                    if (done) done();
                }
            }, replacePgel)
        }
        else {
            var csInfo = pinegrow.stylesheets.getStylesheetInfo(cs, page);
            if (csInfo && csInfo.compiledStylesheet) cs = csInfo.compiledStylesheet;

            var rel_url = crsaMakeLinkRelativeTo(cs.url, page.url);
            //rel_url = pinegrow.getProxyUrl(rel_url);
            var link = pgStylesheetHelper.loadStyleSheet(page, rel_url, function(pgLink) {
                if (cs.sourceStylesheet) {
                    cs.setSourceStylesheet(cs.sourceStylesheet);
                }

                if (cs.loaded) {
                    cs.getCssSource(function(css) {
                        if (!cs.getError()) {
                            pinegrow.stylesheets.applyToPages(cs, done);
                        } else {
                            if (done) done();
                        }
                    });
                } else {
                    if (done) done();
                }
            }, replacePgel, first_css);
        }
    }

    this.attachTo = function(page, pgLink, done, replacePgel, first_css) {//csspgdone
        if (pgLink) {
            if ((_this.changed || true) && _this.loaded && !_this.getError()) {
                _this.getCssSource(function(css) {
                    if (!_this.getError()) {
                        pinegrow.stylesheets.applyToPages(_this, function() {
                            for (var i = 0; i < _this.importedFiles.length; i++) {
                                pinegrow.stylesheets.applyToPages(_this.importedFiles[0]);
                            }
                            _this.appliedToPage = true;
                            done && done();
                        }, page);

                    } else {
                        if (done) done();
                    }
                });
            } else {
                if (done) done();
            }
        } else {
            this.attachAndFill(page, done, replacePgel, first_css)
            page.setPageChanged(true);
        }
    }

    this.reattachToAll = function(list, done) {
        var c = 0;

        _this.setChanged(true);
        if (list.length == 0) {
            if(done) done();
            return;
        }
        for (var i = 0; i < list.length; i++) {
            var $el = $(list[i].ownerNode);
            var page = getCrsaPageOfElement($el);
            // var iframe = getIframeOfElement($el).get(0);
            c++;
            this.attachTo(page, null, function() {
                c--;
                if (c == 0 && done) done();
            }, getElementPgNode($el));
        }
    }

    this.detachFrom = function(page) {
        var cs = this,
            found_ss = this.getPageStylesheetFrom(page);

        if (found_ss) {
            var pgLink = getElementPgNode($(found_ss.ownerNode));
            if(pgLink) pgLink.remove();
        }
        page.setPageChanged(true);
    }

    this.isEnabledForPage = function(page) {
        return pinegrow.stylesheets.isEnabledForPage(this, page);
    }

    this.setEnabledForPage = function(page, enabled) {
        pinegrow.stylesheets.setEnabledForPage(this, page, enabled);
    }

    this.getProjectInfo = function () {
        return new pgProjectInfo(this.localFile);
    }

    this.getPropValuePairFor = function (rule) {
        var values = {};
        var propWithDupObj = {
            "background-image": {
                startsWith: "-webkit"
            },
            "background": {
                startsWith: "-webkit"
            }
        };
        var propWithDupValues = Object.keys(propWithDupObj);

        var canHasDuplicate = function (prop) {
            if (!values[prop] || (values[prop] && propWithDupValues.indexOf(prop) > -1)) return true;
            else return false;
        }

        var setValue = function (prop, value) {
            if (values[prop]) {
                var hasPriority = false;
                if (propWithDupObj[prop]) hasPriority = true;

                if (hasPriority) {
                    var startsWith = propWithDupObj[prop].startsWith;
                    if (startsWith && value.startsWith(startsWith)) {
                        values[prop] = { value: value }
                    }
                }
            }
            else {
                values[prop] = { value: value }
            }
        }

        if (rule.type == "rule") {
            var nodes = rule.nodes;
            for (var i = 0; i < nodes.length; i++) {
                var node = nodes[i];
                if (node.type == "decl") {
                    if (canHasDuplicate(node.prop)) {
                        setValue(node.prop, node.value + (node.important ? ' !important' : ''));
                        // values[node.prop] = { value: node.value };
                    }
                    else {
                        throw new Error("Rule has multiple declarations of the same property. Please use CSS properties editor to edit it.");
                    }
                }
            }
        }
        return values;
    }

    //Changed to async, returns list in done callback
    this.getBreakpoints = function(done) {
        var doTask = function() {
            var unit_re = new RegExp(breakpoints_units_re, 'i');
            var re = new RegExp("([0-9\.]+)" + breakpoints_units_re,"g");
            var list = [];

            if(!_this.rootNode) {
                done(list);
                return;
            }

            var rules = _this.rootNode.nodes;
            if(rules) {
                for(var i = 0; i < rules.length; i++) {
                    if(rules[i].type == "atrule" && rules[i].name == "media") {
                        var m = rules[i].params.match(re);
                        if(m) {

                            for(var n = 0; n < m.length; n++) {
                                var cv = crsaSplitCSSValue(m[n]);
                                var b = cv.value;
                                var unit = cv.unit;

                                if(unit == 'px' && Math.floor(b) != b) {
                                    //filter out 575.98px and such
                                    continue;
                                }

                                var mm = rules[i].params.match(new RegExp("max\\-width\\:\\s*" + b + unit));
                                if(mm) {
                                    crsaOneUpCSSValue(cv);
                                }
                                b = cv.value + cv.unit;

                                if(list.indexOf(b) < 0) {
                                    list.push(b);
                                }
                            }
                        }
                    }
                }
            }




            if(done) done(list);
        }

        if(_this) {
            if (_this.rootNode) {
                doTask();
            } else if (_this.compiledFileSource) {
                //load first
                _this.setSource(_this.compiledFileSource, function () {
                    doTask();
                });
            } else {
                done([]);
            }
        } else {
            done([]);
        }
    }

    this.changeBreakpoint = function(old, b, allDone) {
        var updateSource = function (cs, done) {
            cs.setSourceForRules(cs.rootNode, function () {
                if (done) done(true);
            })
        }

        var doTask = function() {
            var re = new RegExp(":\\s*" + escapeRegExp(old),"g");

            var cv = crsaSplitCSSValue(old);
            var cvdown = crsaOneUpCSSValue(cv, -1);

            var cvb = crsaSplitCSSValue(b);
            var cvbdown = crsaOneUpCSSValue(cvb, -1);

            var re_small = new RegExp(":\\s*" + escapeRegExp(cvdown.value + cvdown.unit),"g");

            var changed = false;
            _this.forEachStylesheet(function (cs) {
                var rules = cs.rootNode.nodes;
                if (rules) {
                    for (var i = 0; i < rules.length; i++) {
                        if (rules[i].type == "atrule" && rules[i].name == "media") {

                            var nm = rules[i].params.replace(re, ": " + b);
                            nm = nm.replace(re_small, ": " + (cvbdown.value + cvbdown.unit));

                            if (nm != rules[i].params) {
                                rules[i].params = nm;
                                changed = true;
                            }
                        }
                    }
                }
                if (changed && cs.fileType == 'css' && cs != _this) {
                    updateSource(cs, allDone);
                }
            });

            if (changed) {
                updateSource(_this, allDone);
            } else {
                if (allDone) allDone(false);
            }
        }
        return doTask();
    }

    this.updateSource = function () {
        this.source = this.getSource();
    }

    this.ruleValueChanged = function (rule, declProp, declValue, done, skipDeleteNode, edited_by, dispatchEvent, forceAction, forceAddNew, changingType, ignoreUndo, skipApplyCode, insertAtIndex) {
        if (!declProp) return;
        var nodeChanged = false;
        var important = false;
        var _this = this;

        if(!this.canEdit(false, true)) {
            if(done) done();
            return;
        }

        // var oldRuleValue = pinegrow.pgPostCSSHelper.getSource(rule);

        if(declValue === '' || declValue === undefined || declValue === null) {
            declValue = null;
        } else {
            declValue = declValue + '';
        }
        //declValue = declValue || null; Allow 0
        if (declValue !== null) {
            if (declValue.indexOf('!important') > 0) {
                declValue = declValue.replace('!important', '');
                important = true;
            }

            declValue = declValue.trim();
        }

        this.forgetCachedSource();

        var declNodes;

        //accept declProp to be decl node or string
        if(typeof declProp === 'object') {
            declNodes = [declProp];
            forceAddNew = false;
            declProp = declProp.prop;
        } else {
            declNodes = pinegrow.pgPostCSSHelper.getChildrenDeclWithPropName(rule, declProp);
        }

        if (declNodes.length > 0 && !forceAddNew) {
            if (declValue === null && !skipDeleteNode) {


                var deleteNode = function (declNode) {
                    if (dispatchEvent) {
                        pinegrow.dispatchEvent('on_css_before_node_deleted', null, {
                            node: declNode,
                            stylesheet: _this,
                            edited_by: edited_by,
                            undoIgnore: ignoreUndo
                        });
                    }
                    pinegrow.pgPostCSSHelper.removeNodeFrom(declNode, declNode.parent);
                    pinegrow.dispatchEvent('on_css_node_deleted', null, {
                        node: declNode,
                        stylesheet: _this,
                        edited_by: edited_by,
                        undoIgnore: ignoreUndo
                    });
                }


                if (changingType === 'all') {
                    for (var i = 0; i < declNodes.length; i++) {
                        deleteNode(declNodes[i]);
                    }
                }
                else if (changingType === 'last') {
                    deleteNode(declNodes[declNodes.length - 1]);
                }
                else {
                    deleteNode(declNodes[0]);
                }
            }
            else {
                //var oldValue = declNode.value;
                //declNode.value = declValue;

                var declNode = declNodes[0];
                if (changingType === 'last') {
                    declNode = declNodes[declNodes.length - 1];
                }

                pgcssh.setNodeProperty(declNode, 'value', declValue);

                declNode.important = important;

                if (dispatchEvent) {
                    pinegrow.dispatchEvent('on_css_node_updated', null, {
                        node: declNode,
                        stylesheet: this,
                        edited_by: edited_by,
                        oldValue: pgcssh.getNodeOldValues(declNode),
                        undoCombineKey: 'ruleValueChanged_' + declProp,
                        //action: typeof dispatchEvent == 'string' ? dispatchEvent : null
                        undoRuleName: typeof dispatchEvent == 'string' ? dispatchEvent : null,
                        undoIgnore: ignoreUndo
                    });
                }
            }
            nodeChanged = true;
        }
        else if (declValue !== null) {
            declNode = pinegrow.pgPostCSSHelper.createNode("decl", declProp, declValue, rule, this);
            declNode.important = important;

            var lastDeclIndex = (insertAtIndex === undefined || insertAtIndex < 0) ? pinegrow.pgPostCSSHelper.getLastDeclIndex(rule) + 1 : insertAtIndex;

            pinegrow.pgPostCSSHelper.insertNodeAt(declNode, rule, lastDeclIndex);

            if (dispatchEvent) {
                pinegrow.dispatchEvent('on_css_node_added', null, {
                    node: declNode,
                    stylesheet: this,
                    edited_by: edited_by,
                    undoIgnore: ignoreUndo
                });
            }

            nodeChanged = true;
        }

        if (forceAction == "skip") nodeChanged = false;
        if (forceAction == "do") nodeChanged = true;

        if (skipApplyCode === false || (nodeChanged && !skipApplyCode)) {
            // this.setChanged(true); // Let's test the new changes

            if(this.updateSingleRule) {
                this.updateSingleRule(rule, done, declProp, declValue);
            } else {
                var rootNode = pinegrow.pgPostCSSHelper.getRootNodeFor(rule);
                var task = "setSourceForRules";

                if(done) {
                    delayed.cancel(task);
                    this.setSourceForRules(rootNode, done, false, true /* fastDelayed */);
                } else {
                    delayed.run(task, function() {
                        _this.setSourceForRules(rootNode, done, false, true /* fastDelayed */);
                    }, 250)
                }
            }
        }
        else {
            this.updateSource();
            if (done) done();
        }
        return declNode; //return the updated node
        // this.regenerateAndSetCssSource(done, false, rootNode);
    }

    /* Not used??
    this.ruleValueChangedSync = function (rule, declProp, declValue, skipDeleteNode, edited_by, dispatchEvent, forceChanging) {
        var sync = false;
        var deasync = require('deasync');

        this.ruleValueChanged(rule, declProp, declValue, function () {
            sync = true;
        }, skipDeleteNode, edited_by, dispatchEvent, forceChanging);

        while (!sync) { deasync.sleep(5); }
        return sync;
    }
    */

    this.ruleMediaChanged = function (rule, mediaParams, done, edited_by) {
        if(!this.canEdit(false, true)) {
            if (done) done(false);
            return;
        }

        if (!this.canAddChangeMediaQuery()) {
            if (done) done(false);
            return;
        }

        var nodeToAdd = [];
        var selectedRule = null;
        var mediaUpdated = false;
        var needRefresh = false; // TODO enhance this

        this.forgetCachedSource();

        var mediaRule = pinegrow.pgPostCSSHelper.getMediaNode(rule);
        if (mediaRule) {
            if (mediaRule.nodes.length == 1) {
                mediaRule.params = mediaParams;
                if (!mediaParams) {
                    pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: rule, stylesheet: this, edited_by: edited_by, will_be_added: true});
                    pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: mediaRule, stylesheet: this, edited_by: edited_by});
                    pinegrow.pgPostCSSHelper.replaceNodeWith(mediaRule, rule);

                    selectedRule = pinegrow.pgPostCSSHelper.getNodeIndices(rule);
                    nodeToAdd.push(selectedRule);
                }
                else {
                    mediaUpdated = true;
                }
            }
            else {
                pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: rule, stylesheet: this, edited_by: edited_by, will_be_added: true});

                if (mediaParams) {
                    var newMediaRule = pinegrow.pgPostCSSHelper.createNode("atrule", "media", mediaParams, null, this);
                    pinegrow.pgPostCSSHelper.insertNodeAfter(newMediaRule, mediaRule);
                    pinegrow.pgPostCSSHelper.insertNodeInside(rule, newMediaRule);
                    nodeToAdd.push(pinegrow.pgPostCSSHelper.getNodeIndices(newMediaRule));
                    nodeToAdd.push(pinegrow.pgPostCSSHelper.getNodeIndices(rule));
                    needRefresh = true;
                }
                else {
                    if (rule.next()) {
                        pinegrow.pgPostCSSHelper.insertNodeBefore(rule, mediaRule);
                    }
                    else {
                        pinegrow.pgPostCSSHelper.insertNodeAfter(rule, mediaRule);
                    }

                    nodeToAdd.push(pinegrow.pgPostCSSHelper.getNodeIndices(rule));
                }

                selectedRule = pinegrow.pgPostCSSHelper.getNodeIndices(rule);
                pinegrow.pgPostCSSHelper.removeNodeFrom(rule, mediaRule);
            }
        }
        else if (mediaParams) {
            var getFormattedMediaParams = function (params) {
                if (!params) return null;
                return (params || "").replace(/\s/g, '');
            }

            pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: rule, stylesheet: this, edited_by: edited_by, will_be_added: true});

            var found = false, next = false;
            var prevRule = rule.prev();
            if (prevRule && prevRule.type == 'atrule') {
                if (getFormattedMediaParams(prevRule.params) == mediaParams) {
                    mediaRule = prevRule;
                    found = true;
                }
            }
            else {
                var nextRule = rule.next();
                if (nextRule && getFormattedMediaParams(nextRule.params) == mediaParams) {
                    mediaRule = nextRule;
                    found = true;
                    next = true;
                }
            }

            if (found) {
                pinegrow.pgPostCSSHelper.removeNodeFrom(rule);

                if (next) {
                    pinegrow.pgPostCSSHelper.insertNodeAt(rule, mediaRule, 0);
                }
                else {
                    pinegrow.pgPostCSSHelper.insertNodeInside(rule, mediaRule);
                }

                nodeToAdd.push(pinegrow.pgPostCSSHelper.getNodeIndices(rule));
            }
            else {
                mediaRule = pinegrow.pgPostCSSHelper.createNode("atrule", "media", mediaParams, null, this);
                pinegrow.pgPostCSSHelper.replaceNodeWith(rule, mediaRule);
                pinegrow.pgPostCSSHelper.insertNodeInside(rule, mediaRule);
                nodeToAdd.push(pinegrow.pgPostCSSHelper.getNodeIndices(mediaRule));
            }

            selectedRule = pinegrow.pgPostCSSHelper.getNodeIndices(rule);
        }

        var rootNode = pinegrow.pgPostCSSHelper.getRootNodeFor(rule);
        this.setSourceForRules(rootNode, function () {
            var currRootNode = _this.getRootNode();
            for (var i = 0; i < nodeToAdd.length; i++) {
                var currNode = pinegrow.pgPostCSSHelper.getNodeByIndices(currRootNode, nodeToAdd[i]);
                pinegrow.dispatchEvent('on_css_node_added', null, {node: currNode, stylesheet: _this, edited_by: edited_by});
            }

            if (done) done(needRefresh);
            if (selectedRule) {
                var currSelectedRule = pinegrow.pgPostCSSHelper.getNodeByIndices(currRootNode, selectedRule);
                pinegrow.selectRule(currSelectedRule, { 'will_be_added': true });
            }

            if (mediaUpdated) {
                pinegrow.dispatchEvent('on_css_node_updated', null, {node: mediaRule, stylesheet: _this, edited_by: edited_by});
            }
        });
        // this.regenerateAndSetCssSource(null, false, rootNode);
    }

    this.ruleSupportsAtruleChanged = function (rule, params, done, edited_by) {
        if (!this.canAddChangeMediaQuery()) {
            if (done) done(false);
            return;
        }

        var nodeToAdded = [];
        var selectedRule = null;
        var needRefresh = false; // TODO enhance this

        this.forgetCachedSource();

        var supRule = pinegrow.pgPostCSSHelper.getSupportsAtruleNode(rule);
        if (supRule) {
            if (supRule.nodes.length == 1) {
                supRule.params = params;
                if (!params) {
                    pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: rule, stylesheet: this, edited_by: edited_by, will_be_added: true});
                    pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: supRule, stylesheet: this, edited_by: edited_by});
                    pinegrow.pgPostCSSHelper.replaceNodeWith(supRule, rule);

                    selectedRule = rule;
                    nodeToAdded.push(rule);
                }
                else {
                    pinegrow.dispatchEvent('on_css_node_updated', null, {node: supRule, stylesheet: this, edited_by: edited_by});
                }
            }
            else {
                pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: rule, stylesheet: this, edited_by: edited_by, will_be_added: true});

                if (params) {
                    var newMediaRule = pinegrow.pgPostCSSHelper.createNode("atrule", "supports", params, null, this);
                    pinegrow.pgPostCSSHelper.insertNodeBefore(newMediaRule, supRule);
                    pinegrow.pgPostCSSHelper.insertNodeInside(rule, newMediaRule);
                    nodeToAdded.push(newMediaRule);
                    nodeToAdded.push(rule);
                    needRefresh = true;
                }
                else {
                    pinegrow.pgPostCSSHelper.insertNodeBefore(rule, supRule);
                    nodeToAdded.push(rule);
                }

                selectedRule = rule;
                pinegrow.pgPostCSSHelper.removeNodeFrom(rule, supRule);
            }
        }
        else if (params) {
            pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: rule, stylesheet: this, edited_by: edited_by, will_be_added: true});

            supRule = pinegrow.pgPostCSSHelper.createNode("atrule", "supports", params, null, this);
            pinegrow.pgPostCSSHelper.replaceNodeWith(rule, supRule);
            pinegrow.pgPostCSSHelper.insertNodeInside(rule, supRule);

            selectedRule = rule;
            nodeToAdded.push(supRule);
        }

        var rootNode = pinegrow.pgPostCSSHelper.getRootNodeFor(rule);
        this.setSourceForRules(rootNode, function () {
            for (var i = 0; i < nodeToAdded.length; i++) {
                var currNode = nodeToAdded[i];
                pinegrow.dispatchEvent('on_css_node_added', null, {node: currNode, stylesheet: _this, edited_by: edited_by});
            }

            if (done) done(needRefresh);
            if (selectedRule) pinegrow.selectRule(selectedRule, { 'will_be_added': true });
        });
        // this.regenerateAndSetCssSource(null, false, rootNode);
    }


    this.genRuleValueChangedByIds = function (indices, declProp, declValue) {
        var rule = pinegrow.pgPostCSSHelper.getNodeByIndices(this.getRootNode(), indices);
        this.ruleValueChanged(rule, declProp, declValue);
    }

    this.getNodeWithPgId = function (id) {
        var forAll = function (nodes) {
            for (var i = 0; i < nodes.length; i++) {
                var currNode = nodes[i];
                var nodeId = currNode['pg-node-id'] || -1;
                if (currNode.selector) {
                    if (nodeId == -1) {
                        matchedId = currNode.selector.match(/\.pg-node-id-([0-9]+)/);
                        if (matchedId && matchedId.length > 0) nodeId = matchedId[1];
                    }

                    if (nodeId == id) return currNode;
                }

                if (currNode.nodes && currNode.nodes.length > 0) {
                    var node = forAll(currNode.nodes);
                    if (node) return node;
                }
            }
        }

        if (this.ignore) return null;
        var node = forAll(this.getRootNode().nodes);

        if (node) {
            // while (node.parent.type != "root") {
            //     node = node.parent;
            // }
            return node;
        }
        else {
            return null;
        }
    }

    this.fileSavedInEditor = function () {
        this.fileSaved();
    }

    this.fileSaved = function () {
        this.setChanged(false);
    }

    this.isVarValue = function (value) {
        return value.startsWith('$') || value.startsWith('@');
    }

    this.isDeclVarNode = function (node) {
        return node.type == 'decl' && node.prop[0] == this.getDeclCharachter();
    }

    this.filterRuleObjects = function (objects) {
        for(var obj in objects) {
            var match = obj.match(/\,\s\.pg-node-id-[0-9]*/);
            if (match && match.length > 0) {
                delete objects[obj];
            }
        }
        return objects;
    }

    this.removeSuper = function () {
        pinegrow.dispatchEvent('on_stylesheet_closed', null, this);
        pinegrow.stylesheets.removeStylesheetByKey(this.url);

        if(this.ignore && !this.elementStyle) {
            pinegrow.undoManager.purgeCs(this);
        }

        var importedFiles = this.importedFiles;
        for (var i = 0; i < importedFiles.length; i++) {
            if (importedFiles[i]) importedFiles[i].remove();
            importedFiles[i] = null;
        }
        this.importedFiles = [];

        var selectedRule = pinegrow.getSelectedRule();
        if (selectedRule) {
            var selectedRuleRoot = pinegrow.pgPostCSSHelper.getRootNodeFor(selectedRule);
            if(selectedRuleRoot.pgStylesheet && selectedRuleRoot.pgStylesheet.getUniqueName() === this.getUniqueName()) {
                if(!this.elementStyle) { //Matjaz: added this to prevent CSS Editor updates when selecting elements
                    $.fn.crsa('selectRule', null, null);
                }
            }
        }

        if (this.stylePane) {
            this.stylePane.remove();
            this.stylePane = null;
        }

        this.stopWatchingFileForChanges();
        var currRootNode = this.getRootNode();
        if (currRootNode) currRootNode.pgStylesheet = null;

        delayed.destroy();

        _this = null;
    }

    this.getFilteredSource = function () {
        return this.filterSource(this.getSource());
    }

    this.forgetSourceErrors = function () { }
}
