var PgStylesheetVars = function (pgStylesheet, $container, options) {
    options = options || {};
    options.customStylesheet = options.customStylesheet || null;
    options.useInsertBorder = options.useInsertBorder != undefined ? options.useInsertBorder : true;

    this.filter = '';
    this.showActiveOnly = false;
    this.active_vars_name = [];

    var _this = this;

    var numberOfComments = 3;
    var flags = ['default', 'global'];
    var items = [];
    for (var i = 0; i < flags.length; i++) {
        var flag = flags[i];
        items.push({ key: '!' + flags[i], 'name': flag.charAt(0).toUpperCase() + flag.slice(1) });
    }

    var $varMainList = $('<div/>', { class: 'crsa-vars-main-list' });
    var $groupMainList = null,
        $groupVarsContainer = null;

    var $stylesheetVars = $container.closest('.crsa-stylesheet-vars');
    var $scrollParent = $stylesheetVars.parent();

    var $nextPage = $('<div class="show-next-page"></div>');

    var $rulemenu = $('<div class="rule-menu"><i class="icon icon-bin crsa-remove-node"></i></div>');

    var page = 1;
    var numberOfItemPerPage = 10;
    var stopShowingVar = false;
    var offsetThreshold = 500;

    var varUpdating = false;

    var lastGroupIndex = 0, lastVarIndex = -1;
    var varsLength = 0;
    var $lastVarUl = null;

    var filter_re = null;

    var pgInsertBordersArr = [];

    var staticVarsArr = false;

    var allPgStylesheets = [];

    var grid = new PgUISectionGrid();
    $varMainList.append(grid.$element);

    var resetMainVars = function () {
        $groupMainList = null,
        $groupVarsContainer = null;

        page = 1;
        numberOfItemPerPage = 10;
        stopShowingVar = false;
        offsetThreshold = 500;

        varUpdating = false;

        lastGroupIndex = 0;
        lastVarIndex = -1;
        varsLength = 0;
        $lastVarUl = null;
    }

    var updateNodeOnStylesheet = function (node, prop, value, oldValue, currStylesheet) {
        var oldValueObj = { prop: '', value: '' };
        oldValueObj[prop] = oldValue;

        node[prop] = value;
        if (prop == 'value') oldValueObj.prop = node[prop];
        else                 oldValueObj.value = node.value;

        varUpdating = true;
        pinegrow.dispatchEvent('on_css_node_updated', null, {
            node: node,
            stylesheet: currStylesheet,
            oldValue: oldValueObj,
            undoCombineKey: 'varValueChanged_' + node[prop],
            action: 'Set ' + node[prop],
            undoRuleName: null
        });
        varUpdating = false;

        currStylesheet.setSourceForRules(currStylesheet.rootNode, function () { });
    }

    var updateColorPlaceholder = function ($field, color) {
        var $colorPlaceholder = $field.find('.color-value-placeholder');
        color = color || 'black';

        $colorPlaceholder.css("background-color", color);

        var $inputColor = $field.find('.crsa-input-color-picker');
        if ($inputColor.length > 0) {
            $inputColor.spectrum("set", color);
        }
    }

    var deleteNodeFromStylesheet = function (node, currStylesheet) {
        varUpdating = true;
        pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: currStylesheet});
        varUpdating = false;

        pinegrow.pgPostCSSHelper.removeNodeFrom(node);
        currStylesheet.setSourceForRules(currStylesheet.rootNode, function () {
            pinegrow.dispatchEvent('on_css_node_deleted', null, {node: node, stylesheet: currStylesheet});
        });
    }

    var getLastVarOfValue = function (value) {
        if (pgStylesheet.type == 'css') return null;

        var lastVar = null;
        var varReg = new RegExp('\\' + pgStylesheet.getDeclCharachter() + '[a-z0-9_\\-]*', 'ig');

        var matched = value.match(varReg);
        if (matched && matched.length > 0) lastVar = matched[matched.length - 1];

        return lastVar;
    }

    var getVarNameList = function (varGroups) {
        var varNames = [];

        for (var i = 0; i < varGroups.length; i++) {
            var group = varGroups[i];
            for (var j = 0; j < group.length; j++) {
                var currVar = group[j];
                var propVal = pgcssh.getVariableProp(currVar);
                varNames.push(propVal);
            }
        }

        return varNames;
    }

    var addNodeToStylesheet = function (type, frValue, sdValue, fun, mainNode, currStylesheet) {
        if (type == 'decl') {
            var node = pinegrow.pgPostCSSHelper.createNode(type, frValue, sdValue, null, currStylesheet);
            fun(node, mainNode);
            // pinegrow.pgPostCSSHelper.insertNodeInside(node, currStylesheet.getRootNode());


            currStylesheet.setSourceForRules(currStylesheet.rootNode, function () {
                varUpdating = true;
                pinegrow.dispatchEvent('on_css_node_added', null, {node: node, stylesheet: currStylesheet});
                varUpdating = false;
            });

            return node;
        }

        return null;
    }

    var addAutoCompeleteTo = function (inputObj) {
        var $input = inputObj.$input;
        $input.autocomplete({
            appendTo: $container,
            minLength: 0,
            search: function (e, ui) {
                var keyCode = e.keyCode || e.which;
                if (keyCode == 38 || keyCode == 40) {
                    return false;
                }
            },
            focus: function (e, ui) {
                return false;
            },
            select: function( event, ui ) {
                var value = ui.item.value;
                var newValue = value;
                if (pgStylesheet.type != 'css' && value.startsWith(pgStylesheet.getDeclCharachter())) {
                    var term = $(this).val();

                    var lastVar = getLastVarOfValue(term);
                    if (lastVar && term.endsWith(lastVar)) {
                        newValue  = term.substr(0, term.length - lastVar.length);
                        newValue += value;
                    }
                }

                $(this).val(newValue).trigger('input');
                this.focus();
                return false;
            },
            source: function(request, response) {
                var term = request.term;
                var allValues = getVarNameList(pgStylesheet.varGroups);
                    allValues = allValues.concat(pgStylesheet.getAutocompleteArr());

                var lastVar = getLastVarOfValue(term);
                if (lastVar && term.endsWith(lastVar)) term = lastVar;

                var results = $.ui.autocomplete.filter(allValues, term);
                response(results.slice(0, 6));
            }
        })
    }

    var getValueFlagPair = function (nodeValue) {
        var valArr = nodeValue.split(' ');
        var flag = valArr.pop();

        if (flag.startsWith('!')) {
            return {
                value: valArr.join(' '),
                flag: flag
            }
        }
        else {
            return {
                value: nodeValue,
                flag: ''
            }
        }
    }

    var getNodeFrom = function ($li, stylesheet) {
        if (!$li.attr('data-node-indices')) return null;

        stylesheet = stylesheet || pgStylesheet;
        return pinegrow.pgPostCSSHelper.getNodeByIndices(stylesheet.rootNode, $li.attr('data-node-indices').split(','));
    }

    // NOT USED
    var getCustomPrevNode = function (prop, mainNodeIndex) {
        var customStylesheet = options.customStylesheet;
        var rootNode = customStylesheet.getRootNode();
        var node = rootNode.last;

        if (!node) return null;

        while (true) {
            var propName = 'prop';
            var nodeType = 'decl';
            if (node.type == 'atrule') {
                propName = 'name';
                nodeType = 'atrule';
            }
            var nodeProp = pgcssh.getVariableProp(node);

            var mainNode = pinegrow.pgPostCSSHelper.findNodeInside(nodeType, propName, nodeProp, pgStylesheet.getRootNode());
            var mainNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(mainNode);

            var nodeIndex = mainNodeIndices[0];

            if (nodeIndex - mainNodeIndex < 0) {
                return node;
            }

            node = node.prev();
        }
    }

    var getMainNodeFor = function (currNode, nodeArr) {
        for (var i = 0; i < nodeArr.length; i++) {
            var node = nodeArr[i];

            var currNodeProp = pgcssh.getVariableProp(currNode);
            var nodeProp = pgcssh.getVariableProp(node);

            if (currNode.type == node.type && currNodeProp == nodeProp) {
                return node;
            }
        }
    }

    var getCustomNodeFor = function ($li) {
        var $customContainer = $li.find('.custom-container');

        if ($customContainer.length > 0) {
            return getNodeFrom($customContainer, options.customStylesheet);
        }

        return null;
    }

    var onVarValueChanged = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('li');

        var nodeVar = getNodeFrom($li);

        var inputObj = $li.data('input-field');
        var flagObj = $li.data('flag-field');

        var value = inputObj.$input.val() + ' ' + flagObj.$input.val();
        var oldNodeValue = $li.data('old-value');

        updateNodeOnStylesheet(nodeVar, 'value', value, oldNodeValue.value, pgStylesheet);

        if (e.type == 'change') {
            var propVal = pgcssh.getVariableProp(nodeVar);
            $li.data('old-value', {
                prop: propVal,
                value: nodeVar.value
            });
        }
    }

    var onVarCustomValueChanged = function (e) {
        var $target = $(e.target);
        var $li = $target.closest('li');
        var $customContainer = $target.closest('.custom-container');

        if ($customContainer.data('just-added')) return;

        var customStylesheet = options.customStylesheet;

        var nodeMainVar = getNodeFrom($li);
        var nodeCustomVar = getNodeFrom($customContainer, customStylesheet);
        var inputObj = $li.data('custom-field');
        var value = inputObj.$input.val();

        if (nodeCustomVar) {
            if (value) {
                var oldCustomValue = $customContainer.data('old-value');
                updateNodeOnStylesheet(nodeCustomVar, 'value', value, oldCustomValue.value, customStylesheet);
            }
            else {
                updateVarIndices(nodeCustomVar, -1, true);
                deleteNodeFromStylesheet(nodeCustomVar, customStylesheet);
                $customContainer.removeAttr('data-node-indices');
            }
        }
        else {
            if (value) {
                var api = new PgApi();

                api.overrideCSSVariable(nodeMainVar, customStylesheet, function(new_nodes, main_nodes) {
                    var newNodeLength = new_nodes.length;
                    if (newNodeLength) {
                        for (var i = 0; i < newNodeLength; i++) {
                            if (i != newNodeLength -1) {
                                var mainNode = getMainNodeFor(new_nodes[i], main_nodes);

                                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(mainNode);
                                var $nodeEl = $varMainList.find('[data-node-indices="' + nodeIndices.join(',') + '"]:not(.custom-container)');

                                if ($nodeEl.length) {
                                    var customField = $nodeEl.data('custom-field');

                                    customField.setValue(new_nodes[i].value);
                                    var $currCustomContainer = $nodeEl.find('.custom-container');

                                    var customNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(new_nodes[i]);
                                    $currCustomContainer.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(new_nodes[i]));
                                }
                            }
                            else {
                                $customContainer.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(new_nodes[i]));
                            }
                        }
                    }

                    $customContainer.data('just-added', false);
                }, value);

                $customContainer.data('just-added', true);

                // --- Add one node in the right order to the _custom.scss file ---
                // var mainNodeIndices = $li.attr('data-node-indices').split(',');
                // var mainNodeIndex = parseInt(mainNodeIndices[0]);

                // var prevNode = getCustomPrevNode(nodeMainVar.prop, mainNodeIndex);

                // var helperNode = null, fun = null;
                // if (prevNode) {
                //     helperNode = prevNode;
                //     fun = pinegrow.pgPostCSSHelper.insertNodeAfter;
                // }
                // else {
                //     helperNode = customStylesheet.getRootNode();
                //     fun = pinegrow.pgPostCSSHelper.insertNodeInside;
                // }

                // var customNode = addNodeToStylesheet('decl', nodeMainVar.prop, value, fun, helperNode, customStylesheet);
                // if (customNode) {
                //     updateVarIndices(customNode, 1, true);
                //     $customContainer.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(customNode));
                // }
            }
        }

        if (e.type == 'change' && nodeCustomVar) {
            var propVal = pgcssh.getVariableProp(nodeCustomVar);
            $customContainer.data('old-value', {
                prop: propVal,
                value: nodeCustomVar.value
            });
        }
    }

    var onVarNameClicked = function (e) {
        e.preventDefault();
        var $target = $(e.target);

        $target.attr('contenteditable', true);
        $target.focus();

        return false;
    }

    var onNewVarNameBlured = function (e) {
        e.preventDefault();
        var $target = $(e.target);
        var $li = $target.closest('li');

        var nodeVar = getNodeFrom($li);

        if (nodeVar) {
            var propVal = pgcssh.getVariableProp(nodeVar);
            $li.data('old-value', {
                prop: propVal,
                value: nodeVar.value
            });
            $target.removeAttr('contenteditable');
        }
        else {
            removeNode$li($li);
        }

        return false;
    }

    var onNewVarNameKeydown = function (e) {
        var code = e.keyCode || e.which;
        if (code == 13 || code == 9 || code == 27) {
            var $target = $(e.target);
            var $li = $target.closest('li');

            if (code == 27) {
                removeNode$li($li);
                return false;
            }
            else {
                var nodeAdded = false, insertAfter = false;
                var mainNode = null;
                var prop = $target.html();

                var $next = $li.next();

                if ($next.length > 0) {
                    var nextNode = getNodeFrom($next);
                    if (nextNode) {
                        mainNode = nextNode;
                        nodeAdded = true;
                        insertAfter = false;
                    }

                }
                else {
                    var $prev = $li.prev();
                    var prevNode = getNodeFrom($next);
                    if (prevNode) {
                        mainNode = prevNode;
                        nodeAdded = true;
                        insertAfter = true;
                    }
                }

                if (nodeAdded) {
                    var fun = null;
                    if (insertAfter) fun = pinegrow.pgPostCSSHelper.insertNodeAfter;
                    else             fun = pinegrow.pgPostCSSHelper.insertNodeBefore;

                    var newNode = addNodeToStylesheet('decl', prop, '', fun, mainNode, pgStylesheet);
                    if (newNode) {
                        $li.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(newNode));

                        var newNodeProp = pgcssh.getVariableProp(newNode);
                        $li.data('old-value', {
                            prop: newNodeProp,
                            value: newNode.value
                        });

                        $target
                            .off('click', onVarNameClicked)
                            .off('blur', onNewVarNameBlured)
                            .off('keydown', onNewVarNameKeydown);

                        $target
                            .on('click', onVarNameClicked)
                            .on('blur', onVarNameBlured)
                            .on('keydown', onVarNameKeydown)
                            .on('keyup', onVarNameKeyup);

                        var $customContainer = $li.find('.custom-container');
                        if ($customContainer.length > 0) {

                            var propName = 'prop';
                            var nodeType = 'decl';
                            if (newNode.type == 'atrule') {
                                propName = 'name';
                                nodeType = 'atrule'
                            }

                            var nodeCustomVar = pinegrow.pgPostCSSHelper.findNodeInside(nodeType, propName, newNodeProp, options.customStylesheet.getRootNode());
                            if (nodeCustomVar) {
                                $customContainer.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(nodeCustomVar));

                                var customVal = $li.data('custom-field');

                                var customValue = nodeCustomVar.value;

                                customVal.setValue(customValue);
                                updateColorPlaceholder(customVal.$field, customValue);

                                var customVarProp = pgcssh.getVariableProp(nodeCustomVar);
                                $customContainer.data('old-value', {
                                    prop: customVarProp,
                                    value: nodeCustomVar.value
                                });
                            }
                            else {
                                $customContainer.data('old-value', {
                                    prop: '',
                                    value: ''
                                });
                            }
                        }

                        $target.removeAttr('contenteditable');
                    }
                }
                else {
                    removeNode$li($li);
                }
            }

            return false;
        }
    }

    var onVarNameBlured = function (e) {
        e.preventDefault();
        var $target = $(e.target);
        var $li = $target.closest('li');

        var nodeVar = getNodeFrom($li);

        var propVal = pgcssh.getVariableProp(nodeVar);
        $li.data('old-value', {
            prop: propVal,
            value: nodeVar.value
        });
        $target.removeAttr('contenteditable');

        return false;
    }

    var onVarNameKeydown = function (e) {
        var code = e.keyCode || e.which;
        if (code == 13 || code == 9 || code == 27) {
            var $target = $(e.target);
            var $li = $target.closest('li');

            var nodeVar = getNodeFrom($li);
            if (code == 27) {
                var oldNodeValue = $li.data('old-value');
                $target.html(oldNodeValue.prop);

                var propName = 'prop';
                if (nodeVar.type == 'atrule') propName = 'name';
                updateNodeOnStylesheet(nodeVar, propName, oldNodeValue.prop, oldNodeValue.prop, pgStylesheet);

                // Update Custom stylesheet var name
                var customNode = getCustomNodeFor($li);
                if (customNode) {
                    var $customContainer = $li.find('.custom-container');
                    var oldCustomNodeValue = $customContainer.data('old-value');

                    updateNodeOnStylesheet(customNode, propName, oldNodeValue.prop, oldNodeValue.prop, options.customStylesheet);
                }
            }
            else {
                var propVal = pgcssh.getVariableProp(nodeCustomVar);
                $li.data('old-value', {
                    prop: propVal,
                    value: nodeVar.value
                });
            }

            $target.removeAttr('contenteditable');
            return false;
        }
    }

    var addNewNodeLiClicked = function ($li, after) {
        addVarAt(null, $li, false, after);
    }

    var isAllowedToAddNewNodeAt = function ($li, $target, mousePosition) {
        if ($target.hasClass('var-item')) return true;

        return false;
    }

    var keyupTimeout = null;
    var onVarNameKeyup = function (e) {
        if (keyupTimeout !== null) {
            clearTimeout(keyupTimeout);
        }
        keyupTimeout = setTimeout(function () {
            var $target = $(e.target);
            var $li = $target.closest('li');

            var nodeVar = getNodeFrom($li);
            var oldNodeValue = $li.data('old-value');

            var propName = 'prop';
            if (nodeVar.type == 'atrule') propName = 'name';
            updateNodeOnStylesheet(nodeVar, propName, $target.html(), oldNodeValue.prop, pgStylesheet);

            // Update Custom stylesheet var name
            var customNode = getCustomNodeFor($li);
            if (customNode) {
                var $customContainer = $li.find('.custom-container');
                var oldCustomNodeValue = $customContainer.data('old-value');

                updateNodeOnStylesheet(customNode, propName, $target.html(), oldCustomNodeValue.prop, options.customStylesheet);
            }
        }, 250);
    }

    // var expandVarsClicked = function (e) {
    //     var $target = $(e.target);
    //     var $expandVarsCont = $target.closest('.expand-vars-container');

    //     var $varsList = $expandVarsCont.siblings('.vars-list');

    //     slideToggle($varsList, function () {
    //         $expandVarsCont.toggleClass('closed');
    //         if ($expandVarsCont.hasClass('closed')) {
    //             pinegrow.dispatchEvent('on_css_var_collapsed', null, {});
    //             // getVarPage();
    //         }
    //     });
    // }

    var expandVarsClicked = function ($section, open) {
        if (!open) {
            pinegrow.dispatchEvent('on_css_var_collapsed', null, {});
        }
    }

    var addVarAt = function (currVar, $el, append, insertAfter) {
        var newNode = false;
        if (!currVar) newNode = true;

        var prop = '', value = '';
        if (!newNode) {
            prop = pgcssh.getVariableProp(currVar);
            value = currVar.value;
        }

        var valAndFlag = getValueFlagPair(value);

        var nodeIndicesAttr = '';
        if (!newNode) {
            var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(currVar);
            nodeIndicesAttr = ' data-node-indices="' + nodeIndices + '"';
        }
        var $li = $('<li'+ nodeIndicesAttr +' data-type="decl" class="var-item"></li>');

        if (append) {
            $li.appendTo($el);
        }
        else if (insertAfter) {
            $li.insertAfter($el);
        }
        else {
            $li.insertBefore($el);
        }

        var $varName = $('<span class="var-name">' + prop + '</span>').appendTo($li);
        if (newNode) {
            $varName
                .on('click', onVarNameClicked)
                .on('blur', onNewVarNameBlured)
                .on('keydown', onNewVarNameKeydown);
        }
        else {
            $varName
                .on('click', onVarNameClicked)
                .on('blur', onVarNameBlured)
                .on('keydown', onVarNameKeydown)
                .on('keyup', onVarNameKeyup);
        }

        if (newNode) $varName.click();

        var $row = $('<div class="row var-row"></div>').appendTo($li);

        var $valCont = $('<div class="col-sm-8 value-container"></div>').appendTo($row);
        var inputVal = new PgInputField('var-value', {
            type: 'color',
            use_type_features: ['slider'],
            name: 'Value',
            without_text: true,
            placeholder: 'Value'
        });

        addAutoCompeleteTo(inputVal);
        inputVal.appendTo($valCont);

        inputVal.setValue(valAndFlag.value);
        updateColorPlaceholder(inputVal.$field, valAndFlag.value)

        inputVal.on('input change', onVarValueChanged);
        $li.data('input-field', inputVal);

        var $flagCont = $('<div class="col-sm-4 flag-container"></div>').appendTo($row);
        var flagVal = new PgInputField('var-flag', {
            type: 'select',
            name: 'Flag',
            show_empty: true,
            icon: false,
            without_text: true,
            placeholder: 'Flag',
            append_new_to_list: true,
            default_value: valAndFlag.flag,
            options: items
        });

        flagVal.appendTo($flagCont);
        flagVal.setValue(valAndFlag.flag);

        flagVal.on('change', onVarValueChanged);
        $li.data('flag-field', flagVal);

        $li.data('old-value', {
            prop: prop,
            value: valAndFlag.value + ' ' + valAndFlag.flag
        });

        // Custom value
        if (options.customStylesheet) {
            var $customCont = $('<div class="col-sm-12 custom-container"></div>').appendTo($row);
            var customVal = new PgInputField('var-custom-value', {
                type: 'color',
                use_type_features: ['slider'],
                name: 'Custom value',
                without_text: true,
                placeholder: 'Custom value'
            });

            addAutoCompeleteTo(customVal);

            customVal.appendTo($customCont);
            $li.data('custom-field', customVal);

            var nodeCustomVar = null;
            if (!newNode) {
                nodeCustomVar = pinegrow.pgPostCSSHelper.findNodeInside('decl', 'prop', prop, options.customStylesheet.getRootNode());
            }

            if (nodeCustomVar) {
                $customCont.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(nodeCustomVar));


                var customValue = nodeCustomVar.value;

                customVal.setValue(customValue);
                updateColorPlaceholder(customVal.$field, customValue);

                var propVal = pgcssh.getVariableProp(nodeCustomVar);
                $customCont.data('old-value', {
                    prop: propVal,
                    value: nodeCustomVar.value
                });
            }
            else {
                $customCont.data('old-value', {
                    prop: '',
                    value: ''
                });
            }

            customVal.on('input change', onVarCustomValueChanged);
        }
    }

    var showVars = function (vars, $container) {
        var start = lastVarIndex + 1;
        var end   = lastVarIndex + numberOfItemPerPage;
        if (end >= vars.length) end = vars.length - 1;
        var varAdded = false;

        if (staticVarsArr) {
            start = 0;
            end = vars.length - 1;
        }

        if (start == 0) {
            $lastVarUl = $('<ul class="vars-list"></ul>').appendTo($container);
            if (options.useInsertBorder) {
                var pgInsertBorders = new PgInsertBorders($lastVarUl);
                pgInsertBorders.onLiClicked = addNewNodeLiClicked;
                pgInsertBorders.isAllowed = isAllowedToAddNewNodeAt;

                pgInsertBordersArr.push(pgInsertBorders);
            }
        }

        for (var i = start; i <= end; i++) {
            var currVar = vars[i];

            var skipCurrVar = false;

            var prop = pgcssh.getVariableProp(currVar);
            if (_this.filter && !prop.match(filter_re)) skipCurrVar = true;
            if (_this.showActiveOnly && (!_this.active_vars_name || _this.active_vars_name.indexOf(prop) == -1)) skipCurrVar = true;

            if (skipCurrVar && !staticVarsArr) {
                if (end < vars.length - 1) end++;
                continue;
            }

            allPgStylesheets.push(pgcssh.getStylesheetForNode(vars[i]));
            addVarAt(vars[i], $lastVarUl, true);

            if (!staticVarsArr) {
                lastVarIndex = i;
                varsLength++;

                varAdded = true;
                if (varsLength == page * numberOfItemPerPage) break;
            }
        }


        if (varAdded) {
            var $groupComment = $container.siblings('.group-comment-container');
            if ($groupComment.length == 0) {
                var prevNodes = getCommentsNodes(vars[0]);

                $groupComment = $('<div class="group-comment-container"></div>').insertBefore($container);
                showVarsComments(prevNodes, $groupComment);
            }
        }

        if (!staticVarsArr) {
            if (varsLength == page * numberOfItemPerPage) stopShowingVar = true;
        }
    }

    var showVarsComments = function (commentNodes, $container) {
        var $ul = $('<ul class="comments-list"></ul>').appendTo($container);

        for (var i = 0; i < commentNodes.length; i++) {
            commentNode = commentNodes[i];
            $('<li class="group-comment" data-node-indices="' + pinegrow.pgPostCSSHelper.getNodeIndices(commentNode) + '" data-type="comment">' + commentNode.text + '</li>').appendTo($ul);
        }
    }

    var getCommentsNodes = function (node) {
        var prevNodes = [],
            lastNode = node.prev(),
            comments = numberOfComments;

        while (lastNode && lastNode.type == 'comment' && comments != 0) {
            prevNodes.unshift(lastNode);
            lastNode = lastNode.prev();
            comments--;
        }

        return prevNodes;
    }

    var showVarsForAddedGroup = function (list, $group) {
        var prevNodes = getCommentsNodes(list[0]);

        var $newGroupComment = $('<div class="group-comment-container"></div>').appendTo($group);
        showVarsComments(prevNodes, $newGroupComment);

        var $newGroupVarsContainer = $('<div class="group-vars-container"></div>').appendTo($group);
        // var $groupExpand = $('<div class="expand-vars-container"><div class="expand-vars"><i class="icon icon-down crsa-open-icon"></i><i class="icon icon-up crsa-close-icon"></i></div></div>').appendTo($newGroupVarsContainer);
        // $groupExpand.click(expandVarsClicked);

        var $ul = $('<ul class="vars-list"></ul>').appendTo($newGroupVarsContainer);

        var pgInsertBorders = new PgInsertBorders($ul);
        pgInsertBorders.onLiClicked = addNewNodeLiClicked;
        pgInsertBorders.isAllowed = isAllowedToAddNewNodeAt;

        pgInsertBordersArr.push(pgInsertBorders);

        for (var i = 0; i < list.length; i++) {
            addVarAt(list[i], $ul, true);
        }
    }

    var showVarsForGroup = function (list, $group) {
        if (lastVarIndex == -1) {
            $groupVarsContainer = $('<div class="group-vars-container"></div>').appendTo($group);
            // var $groupExpand = $('<div class="expand-vars-container"><div class="expand-vars"><i class="icon icon-down crsa-open-icon"></i><i class="icon icon-up crsa-close-icon"></i></div></div>').appendTo($groupVarsContainer);
            // $groupExpand.click(expandVarsClicked);

        }

        showVars(list, $groupVarsContainer);
    }

    var showVarGroups = function () {
        var $mainList = $container.find('> .crsa-vars-main-list');
        if ($mainList.length == 0) {
            $varMainList.appendTo($container);
        }
        else {
            $varMainList = $mainList;
        }
        var list = pgStylesheet.varGroups;

        var start = lastGroupIndex;

        for (var i = start; i < list.length; i++) {
            if (stopShowingVar) break;

            if (!$groupMainList || lastVarIndex == -1) {
                var pgsec = new PgUISection(pgStylesheet.name, 'Group ' + (i + 1), null, null, null, {
                    onOpenClose: expandVarsClicked
                });
                grid.addSection(pgsec);
                grid.newSectionAdded(); // TODO find a better way
                // $groupMainList = $('<div class="vars-group"></div>').appendTo($varMainList);
                $groupMainList = pgsec.$content;
            }

            showVarsForGroup(list[i], $groupMainList);
            var $varItems = $groupMainList.find('.var-item');
            if ($varItems.length == 0) grid.popSection();

            lastGroupIndex = i;

            if (!stopShowingVar) lastVarIndex = -1;
        }

        if (stopShowingVar) $nextPage.appendTo($container);
        else                $nextPage.detach();
    }

    var removeCommentsBefore = function ($li) {
        while ($li.length > 0) {
            var $prev = $li.prev();
            $li.remove();
            $li = $prev;
        }
    }

    var getVarPage = function () {
        var height = $scrollParent.height();
        var varsTop = $scrollParent.offset().top;

        var $parent = $nextPage.parent();

        var $html = $container.closest('html');
        var isAdded = ($html.length != 0);
        var offset = $nextPage.offset().top - varsTop;

        while (isAdded && $parent.length != 0 && offset - offsetThreshold <= height) {
            stopShowingVar = false;
            showVarGroups();
            page++;

            $parent = $nextPage.parent();
            offset = $nextPage.offset().top - varsTop;
        }

        grid.updateLayout();
    }

    var showVarsArr = function (vars) {
        $varMainList.appendTo($container);

        var pgsec = new PgUISection(pgStylesheet.name, 'Group', null, null, null, {
            onOpenClose: expandVarsClicked
        });
        grid.addSection(pgsec);
        grid.newSectionAdded(); // TODO find a better way
        // $groupMainList = $('<div class="vars-group"></div>').appendTo($varMainList);
        var $content = pgsec.$content;
        var $groupVarsContainer = $('<div class="group-vars-container"></div>').appendTo($content);
        showVars(vars, $groupVarsContainer);
        grid.updateLayout();
    }

    var setUsedVarArr = function (nodes) {
        // var vars_name = [];

        var updateVarsName = function (rule) {
            var nodeStylesheet = pinegrow.pgPostCSSHelper.getStylesheetForNode(rule);
            var ruleNodes = rule.nodes;

            for (var i = 0; i < ruleNodes.length; i++) {
                var node = ruleNodes[i];
                if (node.type == "decl" && nodeStylesheet.isVarValue(node.value)) {
                    // vars_name.push(node.value);
                    resolveUsed(node.value);
                }
            }
        }


        for (var i = 0; i < nodes.length; i++) {
            if (nodes[i].type == 'rule') {
                updateVarsName(nodes[i]);
            }
        }

        // return vars_name;
    }

    var updateVarIndices = function (newNode, by, customVar) {
        var nodeLvl = pinegrow.pgPostCSSHelper.getNodeLevel(newNode);
        if (nodeLvl > 0) return;

        var newNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(newNode);
        var newNodeLevelIndex = newNodeIndices[0];

        var $varItems = $varMainList.find('.var-item,.group-comment');
        for (var i = $varItems.length - 1; i >= 0 ; i--) {
            var $nodeLi = $($varItems[i]);

            if (customVar) $nodeLi = $nodeLi.find('.custom-container');

            var nodeIndicesStr = $nodeLi.attr('data-node-indices');
            if (nodeIndicesStr) {
                var nodeIndices = nodeIndicesStr.split(',');
                var nodeLevelIndex = parseInt(nodeIndices[0]);

                if (nodeLevelIndex >= newNodeLevelIndex) {
                    nodeIndices[0] = nodeLevelIndex + by;
                    $nodeLi.attr('data-node-indices', nodeIndices.join(','));
                }
                else {
                    break;
                }
            }
        }
    }

    var onParentScrolled = function (e) {
        if (!staticVarsArr) getVarPage();
    }

    var onVarCollapsed = function () {
        if (!staticVarsArr) getVarPage();
    }

    var onNodeAdded = function(page, obj) {
        if (onNodeAdded == null || _this.showActiveOnly) return;
        if (!obj) return;
        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;

        var isVarNode = pgStylesheet.isDeclVarNode(updatedNode);

        var customStylesheet = options.customStylesheet;
        var hasCustomStylesheet = customStylesheet && customStylesheet.getUniqueName() == stylesheet.getUniqueName();

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!varUpdating && isVarNode) {
                var add = false;

                updateVarIndices(updatedNode, 1);

                var nextNode = updatedNode.next();
                if (nextNode) {
                    if (nextNode.type != 'comment') {
                        var nextNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(nextNode);
                        var $nextNodeEl = $varMainList.find('[data-node-indices="' + nextNodeIndices + '"]:not(.custom-container)');

                        if ($nextNodeEl.length > 0) {
                            addVarAt(updatedNode, $nextNodeEl, false, false);
                            add = true;
                        }
                    }
                }

                if (!add) {
                    var prevNode = updatedNode.prev();
                    if (prevNode) {
                        var prevNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(prevNode);
                        var $prevNodeEl = $varMainList.find('[data-node-indices="' + prevNodeIndices + '"]:not(.custom-container)');

                        if (prevNode.type == 'comment') {
                            removeCommentsBefore($prevNodeEl);
                        }
                        else {
                            if ($prevNodeEl.length > 0) {
                                addVarAt(updatedNode, $prevNodeEl, false, true);
                                add = true;
                            }
                        }
                    }
                }


                if (!add) {
                    var groupIndex = -1;

                    var varGroup = pgStylesheet.varGroups;
                    for (var i = 0 ; i < varGroup.length; i++) {
                        var currGroup = varGroup[i];
                        if (currGroup.indexOf(updatedNode) > -1) {
                            groupIndex = i;
                            break;
                        }
                    }

                    if (groupIndex > -1) {
                        var $varGroupSection = null;
                        var insertAfter = false;

                        var currGroup = varGroup[groupIndex];
                        var firstVar = currGroup[0];

                        var firstNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(firstVar);
                        var $firstNodeEl = $varMainList.find('[data-node-indices="' + firstNodeIndices + '"]:not(.custom-container)');

                        if ($firstNodeEl.length > 0) {
                            var $mainVarList = $firstNodeEl.closest('.crsa-vars-main-list');
                            var $showNextPage = $mainVarList.siblings('.show-next-page');
                            if ($showNextPage.length > 0) return;
                        }

                        if (groupIndex > 0) {
                            var prevGroup = varGroup[groupIndex - 1];
                            var firstVar = prevGroup[0];

                            var lastNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(firstVar);
                            var $lastNodeEl = $varMainList.find('[data-node-indices="' + lastNodeIndices + '"]:not(.custom-container)');
                            if ($lastNodeEl.length > 0) {
                                $varGroupSection = $lastNodeEl.closest('.pg-ui-section');
                                if (currGroup.length > 1) {
                                    var $mainVarList = $varGroupSection.closest('.crsa-vars-main-list');
                                    var $showNextPage = $mainVarList.siblings('.show-next-page');
                                    if ($showNextPage.length > 0) return;
                                }
                                insertAfter = true;
                            }
                        }
                        else {
                            var nextGroup = varGroup[groupIndex + 1];
                            var firstVar = nextGroup[0];

                            var lastNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(firstVar);
                            var $lastNodeEl = $varMainList.find('[data-node-indices="' + lastNodeIndices + '"]:not(.custom-container)');
                            if ($lastNodeEl.length > 0) {
                                $varGroupSection = $lastNodeEl.closest('.pg-ui-section');
                                insertAfter = false;
                            }
                        }

                        var insertAt = $varGroupSection.index();
                        if (insertAfter) insertAt++;

                        var pgsec = new PgUISection(pgStylesheet.name, 'New Group', null, null, null, {
                            onOpenClose: expandVarsClicked
                        });
                        grid.addSectionAt(pgsec, insertAt);
                        grid.newSectionAdded(); // TODO find a better way

                        var $newVarGroupSection = pgsec.$content;
                        showVarsForAddedGroup(varGroup[groupIndex], $newVarGroupSection);

                        grid.updateLayout();
                    }
                }
            }
        }
        else if (hasCustomStylesheet) {
            if (!varUpdating && isVarNode) {
                var groupIndex = -1, varIndex = -1;
                var mainVar = null;

                updateVarIndices(updatedNode, 1, true);

                var groupList = pgStylesheet.varGroups;
                var nodeFound = false;
                for (var i = 0; i < groupList.length; i++) {
                    var varList = groupList[i];

                    for (var j = 0; j < varList.length; j++) {
                        var varListProp = pgcssh.getVariableProp(varList[j]);
                        var updatedNodeProp = pgcssh.getVariableProp(updatedNode);

                        if (varListProp == updatedNodeProp) {
                            groupIndex = i; varIndex = j;
                            mainVar = varList[j];
                            nodeFound = true;
                            break;
                        }
                    }

                    if (nodeFound) break;
                }

                if (nodeFound) {
                    var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(mainVar);
                    var $nodeEl = $varMainList.find('[data-node-indices="' + nodeIndices + '"]');

                    if ($nodeEl.length > 0) {
                        var $customContainer = $nodeEl.find('.custom-container');

                        if ($customContainer.length > 0) {
                            $customContainer.attr('data-node-indices', pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode));

                            var customField = $nodeEl.data('custom-field');

                            customField.setValue(updatedNode.value);
                            updateColorPlaceholder(customField.$field, updatedNode.value);
                        }
                    }
                }
            }
        }
    }

    var onNodeUpdated = function(page, obj) {
        if (onNodeUpdated == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;
        var oldValue = obj.oldValue || {};

        var customStylesheet = options.customStylesheet;
        var hasCustomStylesheet = customStylesheet && customStylesheet.getUniqueName() == stylesheet.getUniqueName();

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!varUpdating) {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);

                var $nodeEl = $varMainList.find('[data-node-indices="' + nodeIndices + '"]:not(.custom-container)');

                if ($nodeEl.length > 0) {
                    if ($nodeEl.hasClass('custom-container')) {
                        var $li = $nodeEl.closest('li');
                        var customField = $li.data('custom-field');

                        customField.setValue(updatedNode.value);
                        updateColorPlaceholder(customField.$field, updatedNode.value);
                    }
                    else {
                        if ($nodeEl.data('type') == 'decl') {
                            _this.updateNodeViewWith($nodeEl, updatedNode);
                        }
                        else if ($nodeEl.data('type') == 'comment') {
                            $nodeEl.html(updatedNode.text);
                        }
                    }
                }
            }
        }
        else if (hasCustomStylesheet) {
            if (!varUpdating) {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);

                var $nodeEl = $varMainList.find('.custom-container[data-node-indices="' + nodeIndices + '"]');

                if ($nodeEl.length > 0) {
                    var $li = $nodeEl.closest('li');
                    var customField = $li.data('custom-field');

                    customField.setValue(updatedNode.value);
                    updateColorPlaceholder(customField.$field, updatedNode.value);
                }
            }
        }
    }

    var onNodeBeforeDeleted = function (page, obj) {
        if (onNodeBeforeDeleted == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;

        var customStylesheet = options.customStylesheet;
        var hasCustomStylesheet = customStylesheet && customStylesheet.getUniqueName() == stylesheet.getUniqueName();

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            if (!varUpdating) {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);

                var $nodeEl = $varMainList.find('[data-node-indices="' + nodeIndices + '"]:not(.custom-container)');

                if ($nodeEl.length > 0) {
                    $nodeEl.remove();
                }

                updateVarIndices(updatedNode, -1);
            }
        }
        else if (hasCustomStylesheet) {
            if (!varUpdating) {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(updatedNode);

                var $nodeEl = $varMainList.find('.custom-container[data-node-indices="' + nodeIndices + '"]');

                if ($nodeEl.length > 0) {
                    var $li = $nodeEl.closest('li');
                    var customField = $li.data('custom-field');

                    updateColorPlaceholder(customField.$field);
                    customField.setValue("");
                    $nodeEl.removeAttr('data-node-indices');
                    $nodeEl.closest('.custom-container').data('old-value', {
                        prop: '',
                        value: ''
                    });

                    updateVarIndices(updatedNode, -1, true);
                }
            }
        }
    }

    var onNodeIndicesChanged = function (page, obj) {
        if (onNodeIndicesChanged == null) return;
        if (!obj) return;

        var stylesheet = obj.stylesheet;
        var updatedNode = obj.node;
        var oldNodeIndices = obj.oldNodeIndices;
        var newNodeIndices = obj.currNodeIndices;

        var customStylesheet = options.customStylesheet;
        var hasCustomStylesheet = customStylesheet && customStylesheet.getUniqueName() == stylesheet.getUniqueName();

        if (pgStylesheet.getUniqueName() == stylesheet.getUniqueName()) {
            var $nodeEl = $varMainList.find('[data-node-indices="' + oldNodeIndices + '"]:not(.custom-container)');

            if ($nodeEl.length > 0) {
                $nodeEl = $nodeEl.last();

                var changeNodeIndices = false;
                if ($nodeEl.data('type') == 'decl') {
                    var $varName = $nodeEl.find('.var-name');
                    var propVal = pgcssh.getVariableProp(updatedNode);

                    if ($varName.html() == propVal) {
                        changeNodeIndices = true;
                    }
                }
                else if ($nodeEl.data('type') == 'comment') {
                    if ($nodeEl.html() == updatedNode.text) {
                        changeNodeIndices = true;
                    }
                }

                if (changeNodeIndices) $nodeEl.attr('data-node-indices', newNodeIndices);
            }
        }
        else if (hasCustomStylesheet) {
            var $nodeEl = $varMainList.find('.custom-container[data-node-indices="' + oldNodeIndices + '"]');

            if ($nodeEl.length > 0) {
                $nodeEl = $nodeEl.last();

                var changeNodeIndices = false;

                var $li = $nodeEl.closest('li');
                var $varName = $li.find('.var-name');
                var propVal = pgcssh.getVariableProp(updatedNode);

                if ($varName.html() == propVal) {
                    changeNodeIndices = true;
                }

                if (changeNodeIndices) $nodeEl.attr('data-node-indices', newNodeIndices);
            }
        }
    }

    var resolveUsed = function (value) {
        var varReg = new RegExp('\\' + pgStylesheet.getDeclCharachter() + '[a-z0-9_\\-]*', 'ig');
        var m = value.match(varReg);

        if (m) {
            m.forEach(function (used_var) {
                var newNode = pinegrow.pgPostCSSHelper.findNodeInside('decl', 'prop', used_var, pgStylesheet.getRootNode());
                if (newNode) {
                    resolveUsed(newNode.value);
                }

                if (_this.active_vars_name.indexOf(used_var) == -1) {
                    _this.active_vars_name.push(used_var);
                }
            })
        }
    }

    var updateActiveNodes = function () {
        var active_nodes = pinegrow.pgActiveRules.getActiveNodes();
        if (active_nodes) {
            setUsedVarArr(active_nodes.shown_nodes);
        }
        else {
            _this.active_vars_name = [];
        }
    }

    // var onSelectedElementsChanged = function () {
    //     updateActiveNodes();
    //     grid.clear();
    //     _this.show();
    // }

    var removeNode$li = function ($li) {
        pgRemoveSpectrum($li);
        pgRemoveMultiselects($li);

        $rulemenu.detach();
        $li.remove();
    }

    var removeNode = function ($li) {
        var $customContainer = $li.find('.custom-container');
        var customStylesheet = options.customStylesheet;

        var mainNode = getNodeFrom($li);
        var customNode = getNodeFrom($customContainer, customStylesheet);

        removeNode$li($li);

        if (mainNode)   {
            updateVarIndices(mainNode, -1);
            deleteNodeFromStylesheet(mainNode, pgStylesheet);
        }
        if (customNode) {
            updateVarIndices(customNode, -1, true);
            deleteNodeFromStylesheet(customNode, customStylesheet);
        }
    }

    var ruleMenuClicked = function (e) {
        e.preventDefault();
        var $target = $(e.target);
        var $li = $target.closest('li')

        if ($target.hasClass('crsa-remove-node')) {
            removeNode($li);
        }

        return false;
    }

    this.isPgStylesheetExists = function (currStylesheet) {
        for (var i = 0; i < allPgStylesheets.length; i++) {
            var stylesheet = allPgStylesheets[i];
            if (stylesheet == currStylesheet) return true;
        }
        if (options.customStylesheet == currStylesheet) return true;

        return false;
    }

    this.updateNodeViewWith = function ($li, node) {
        var propVal = pgcssh.getVariableProp(node);
        $li.find('> .var-name').html(propVal);

        var valAndFlag = getValueFlagPair(node.value);

        var inputField = $li.data('input-field');

        inputField.setValue(valAndFlag.value);
        updateColorPlaceholder(inputField.$field, valAndFlag.value)

        var flagField = $li.data('flag-field');
        flagField.setValue(valAndFlag.flag);
    }

    $scrollParent.on('scroll', onParentScrolled);

    this.show = function (vars) {
        resetMainVars();

        filter_re = null;
        if (_this.filter) {
            filter_re = new RegExp(escapeRegExp(_this.filter), "i");
        }

        if (this.showActiveOnly) updateActiveNodes();

        $container[0].innerHTML = '';
        // $noteMessage.appendTo($container);

        if (vars && vars.length > 0) {
            staticVarsArr = true;
            showVarsArr(vars);
        }
        else {
            $nextPage.appendTo($container);
            getVarPage();
        }

        // showHideNote();
    }

    this.updateLayout = function () {
        grid.updateLayout();
        getVarPage();
    }

    this.remove = function () {
        for (var i = 0; i < pgInsertBordersArr.length; i++) {
            var pgInsertBorders = pgInsertBordersArr[i];

            if (pgInsertBorders) {
                pgInsertBorders.remove();
                pgInsertBorders = null;
            }
        }
        pgInsertBordersArr = [];
        allPgStylesheets = [];

        $scrollParent.off('scroll', onParentScrolled);

        // event listeners
        pinegrow.removeEventHandler('on_css_node_added', onNodeAdded);
        pinegrow.removeEventHandler('on_css_node_updated', onNodeUpdated);
        pinegrow.removeEventHandler('on_css_before_node_deleted', onNodeBeforeDeleted);
        // pinegrow.removeEventHandler('on_css_node_indices_changed', onNodeIndicesChanged);

        pinegrow.removeEventHandler('on_css_var_collapsed', onVarCollapsed);

        $rulemenu.off('click', ruleMenuClicked);
    }

    $container.on('mouseenter', '.var-item', function (event) {
        var $li = $(event.target).closest('li');

        if ($li.hasClass('var-item')) {
            $li.addClass('hoverd');
            $rulemenu.appendTo($li);
        }
    });

    $container.on('mouseleave', '.var-item', function (event) {
        var $li = $(event.target).closest('li');

        if ($li.hasClass('var-item')) {
            $li.removeClass('hoverd');
            $rulemenu.detach();
        }
    });

    $rulemenu.on('click', ruleMenuClicked);

    pinegrow.addEventHandler('on_css_node_added', onNodeAdded);
    pinegrow.addEventHandler('on_css_node_updated', onNodeUpdated);
    pinegrow.addEventHandler('on_css_before_node_deleted', onNodeBeforeDeleted);
    // pinegrow.addEventHandler('on_css_node_indices_changed', onNodeIndicesChanged);

    // pinegrow.addEventHandler('on_selected_elements_changed', onSelectedElementsChanged);

    pinegrow.addEventHandler('on_css_var_collapsed', onVarCollapsed);
}
