/**
 * Created by Matjaz on 2/3/14.
 */

var Pinegrow = function() {

    var components = {};
    var frameworks = {};

    var delayedTasks = new PgDelayedTasks();

    this.defaults = {
        formatCss: 'default',
        codeTheme: 'Dark',
        font:"monospace",
        unsplashAppId: 'd3d0ad950808c9d241e097130ea0c4ca0d06ead44e05e8fa39c5054db8dd5bd7',
        autoprefixer: '> 1%, Last 4 versions, Firefox ESR, Opera 12.1'
    }

    var framework_urls = [];
    var _this = this;

    this.projectTemplates = [];

    this.sourceParser = true;
    this.httpServer = null;

    var ignore_clicks = false;

    var count = 0;

    var current_project = null;

    this.ready = false;

    this.cache_breaker = 0;

    this.autoFormatHtml = true;

    this.shouldHighlightElements = true;

    this.stats = new PgStats();

    this.code_editors = new PgCodeEditors();

    this.insight = new PgInsight();

    this.event_conductor = new PgEventConductor();

    this.lock_css_media = '(max-width:767px)'; //MT CSS TEST

    this.clipboard = new PgClipboard();

    this.getClipboard = function() {
        return this.clipboard;
    }

    this.showWorkspace = function(key, done) {
        window.pgDontSaveLayoutState = false;

        pgCloseAllUIWindows();

        var info_visible = null;
        if(pgUIViewManager.getView('info')) {
            info_visible = !pinegrow.isHiddenUIPanel('info');
        }

        pinegrow.showUIPanel('info');

        myLayout.restoreWorkspace(key);

        if(info_visible === false) {
            pinegrow.hideUIPanel('info');
        }

        var layout_state_key = 'pg_layout_state';
        layout_state_key += pgf.layout_prefix;

        localStorage[layout_state_key] = '';

        if(pgf.settings_on_backend) {
            pinegrow.setSetting(layout_state_key, '');
        }

        pinegrow.dispatchEvent('on_panel_moved_in_another_window', null);
        setTimeout(function() {
            window.pgDontSaveLayoutState = false;
            pinegrow.selectedElements.reselect();
            done && done();
        }, 1000)
    }

    this.isContributorMode = function() {
        var p = crsaStorage.getValue('activatedProduct');
        if(p && p.indexOf('CMSUSER') >= 0) return true;
        return pinegrow.getSetting('contributor-mode', 'false') == 'true';
    }

    var treeViews = []

    this.registerTreeView = function(view) {
        var idx = treeViews.indexOf(view);
        if(idx < 0) {
            treeViews.push(view)
        }
    }

    this.unregisterTreeView = function(view) {
        var idx = treeViews.indexOf(view);
        if(idx >= 0) {
            treeViews.splice(idx, 1)
        }
    }

    this.isOverTreeView = function(event, menu) {
        for(var i = 0; i < treeViews.length; i++) {
            var pos = treeViews[i].isOverTree(event, menu);
            if(pos) {
                pos.treeView = treeViews[i];
                return pos;
            }
        }
        return null;
    }

    this.getTreeViews = function() {
        return treeViews;
    }

    this.getTreeViewOfClass = function(cls) {
        for(var i = 0; i < treeViews.length; i++) {
            if(treeViews[i] instanceof cls) {
                return treeViews[i]
            }
        }
        return null;
    }

    this.updateStartScreen = function() {
        pinegrow.tutorialsManager &&  pinegrow.tutorialsManager.updateStartScreen(null, true);
    }

    this.openProject = function(dir, page, done, options) {
        var project = new CrsaProject();
        options = options || {}

        $.each(options, function(key, val) {
            project[key] = val;
        })
        if(!isApp()) {
            project.fromServer('project1', function(p) {
                pinegrow.setCurrentProject(p);
                crsaQuickMessage('Project was loaded.');
                //pinegrow.showTab('project');

                if (page && p.framework) {
                    page.addFramework(p.framework);
                }
                done && done(project);
            })
        } else {
            if(crsaIsFileUrl(dir)) {
                dir = crsaMakeFileFromUrl(dir);
            }
            project.fromFolder(dir, function (p) {
                pinegrow.setCurrentProject(p);
                if(!p.lesson) {
                    pinegrow.recentFiles.add(p.root.path, true);
                }
                crsaQuickMessage('Project was loaded.');
                pinegrow.openProjectPanelIfNoPageOpen();

                if (page && p.framework) {
                    page.addFramework(p.framework);
                }
                done && done(project);
            }, true);
        }
        return project;
    }

    this.getSetting = function(key, def, allow_empty) {
        if(allow_empty === undefined) allow_empty = true;
        if(!pgf.settings_on_backend) {
            key = 'settings-' + key;
        }

        var store = window.localStorage;

        if(pgf.settings_on_backend) {
            if(pgwpdata && pgwpdata.user_settings) {
                store = pgwpdata.user_settings;
            }
        }

        if(store && (key in store)) {
            var val = store[key];
            if(!allow_empty && val.length == 0) {
                val = def;
            }
            return val;
        }
        if(typeof def == 'undefined') def = null;
        return def;
    }

    this.setSetting = function(key, value) {
        if(value === null || value === undefined) value = '';
        if(!pgf.settings_on_backend) {
            key = 'settings-' + key;
        }

        var store = window.localStorage;

        if(pgf.settings_on_backend) {
            if(pgwpdata && pgwpdata.user_settings) {
                store = pgwpdata.user_settings;
            }
        }

        store[key] = value;

        if(pgf.settings_on_backend && pinegrow.currentUser) {
            pinegrow.currentUser.settingsChanged();
        }
    }

    this.openProjectBySelectingFolder = function(done) {
        if(canOpenProject()) {
            var project = new CrsaProject();
            project.openProjectBySelectingFolder(function(p) {
                pinegrow.setCurrentProject(p);
                pinegrow.recentFiles.add(p.root.path, true);
                crsaQuickMessage('Project was loaded.');
                pinegrow.openProjectPanelIfNoPageOpen();

                if(done) done(project);
            });
            pinegrow.stats.using('app.openproject');
        }
    }

    this.setCurrentProject = function(p) {
        if(current_project && current_project != p) {
            $('body').trigger('crsa-project-closed', current_project);
        }
        current_project = p;
        if(p) {
            p.showProjectExplorer($('#crsa-project'));
            _this.callGlobalFrameworkHandler('on_project_loaded', current_project);

            pinegrow.devTools.unregisterServiceWorkers();

            if(pgf.kids) {
                //restore display helpers
                pinegrow.setShowDisplayHelpers(true, false);
            }

            //$('#crsa-left-plane').show();
            //$('.empty-canvas').hide();
            $(window).trigger('resize');

             if(!p.framework && pgf.components) {
                //not yet loaded as framework
                p.framework = pinegrow.loadProjectAsLibrary(p.getDir());
            }
        }
        //updatePagesList();
        //$(window).trigger('resize');
    }

    this.getCurrentProject = function() {
        return current_project;
    }

    this.closeCurrentProject = function(pageOpen, force, done) {
        if(current_project) {
            current_project.is_closing_all_pages = true;

            pinegrow.closeAllPages(function(closed) {
                if(closed) {
                    var p = current_project;

                    current_project.closeProjectExplorer($('#crsa-project'));
                    pinegrow.setCurrentProject(null);

                    _this.callGlobalFrameworkHandler('on_project_closed', p);

                    if(p.framework) {
                        pinegrow.unloadFramework(p.framework);
                    }

                    p.destroy();

                    $(window).trigger('resize');

                    done && done();
                }
            }, true, force);
        }
        else {
            if (pageOpen) {
                this.showQuickMessage("There is no open project. Use <i>Close Page</i> instead.");
            }
            else {
                this.showQuickMessage("There is no open project.");
            }
        }
    }

    this.refreshCurrentProject = function(done, just_show, no_restore_tags) {
        if(current_project) {
            if(!no_restore_tags) {
                current_project.storeTags();
            }

            var update = function() {
                if(!no_restore_tags) {
                    current_project.restoreTags();
                }

                current_project.showProjectExplorer($('#crsa-project'));
                if(!just_show) {
                    _this.callGlobalFrameworkHandler('on_project_refreshed', current_project);
                }
                if(done) done();
            }

            //todo - get from server
            if(!isApp()) {
                update();
                return;
            }

            current_project.fromFolder(current_project.root.path, function(p) {
                update();
            }, true /* show pg files */)
        }
    }

    this.refreshCurrentProjectDelayed = function(fast) {
        delayedTasks.run('refreshCurrentProjectDelayed', function() {
            _this.refreshCurrentProject(null, fast);
        }, 1000);
    }

    this.getCurrentProjectInfo = function(cp) {
        if(current_project) {
            return current_project.getProjectInfo();
        }
        var page = cp || pinegrow.getSelectedPage();
        if(page) return page.getProjectInfo();
        return null;
    }


    this.refreshAllPages = function () {
        var pages = this.getAllPages();
        for (var i=0; i < pages.length; i++) {
            pages[i].refresh();
        }
    }

    this.getComponent = function(name, done) {
        if(name in components && false) {
            done(components[name]);
        } else {
            components[name] = new PgComponent(name, done);
        }
    }

    this.getProxyUrl = function(url, live) {
        if(this.sourceParser && this.httpServer) {
            return this.httpServer.makeProxyUrl(url, live);
        }
        return url;
    }

    this.hasInternalServer = function() {
        return this.httpServer && this.httpServer.url;
    }

    this.getOriginalUrl = function(url) {
        if(this.httpServer) {
            return this.httpServer.getOriginalUrl(url);
        }
        return url;
    }

    this.getApiUrl = function() {
        return this.httpServer.url;
    }

    // Set editable file types
    var editable_types = null;

    this.resetEditableFileTypes = function() {
        editable_types = null;
    }

    this.hideEditCode = function() {
        var sm = pgPageCodeEditorsStack;
        if(sm.isEditorShown()) {
            sm.hide();
            return true;
        }
        return false; //was not shown
    }

    this.editCode = function(page, dont_hide, read_only) {
        if(pinegrow.isContributorMode()) return null;

        var sm = pgPageCodeEditorsStack;
        if(page.codeEditor) {
            if(!sm.isEditorShown(page.codeEditor)) {
                sm.showEditor(page.codeEditor);
            } else {
                if(!dont_hide) {
                    sm.hide();
                }
            }
        } else {
            page.codeEditor = new PgInternalCodeEditorView(page, read_only);
            sm.showEditor(page.codeEditor);
        }
        return page.codeEditor;
    }

    this.editElementCode = function(pgel) {
        if(pgel) {
            var ce = new PgQuickCodeEdit(pgel);
        }
    }

    this.editCodeOfAllOpenPagesAndStylesheets = function() {
        if(pinegrow.isContributorMode()) return null;

        var first_editor = null;

        var sm = pgPageCodeEditorsStack;

        pinegrow.getAllPages().forEach(function(page) {
            if(!page.codeEditor) {
                page.codeEditor = new PgInternalCodeEditorView(page, false);
                if(!first_editor) {
                    first_editor = page.codeEditor;
                }
                //sm.showEditor(page.codeEditor);
            }
        })

        var project = pinegrow.getCurrentProject();

        pinegrow.getStylesheets().forEach(function(cs) {
            if(!cs.inline) {
                if(cs.url && project && project.isUrlInProject(cs.url)) {
                    pinegrow.stylesheets.editCode(cs);
                }
            }
        })

        if(!sm.isEditorShown()) {
            sm.show();
        }
        first_editor && pgPageCodeEditorsStack.showEditor(first_editor);
    }

    this.revealCSSRuleInCodeEditor = function(r) {

        if(pgf.kids) {
            this.editCodeOfAllOpenPagesAndStylesheets();
        }

        var cs = pgcssh.getStylesheetForNode(r);
        if(cs.elementStyle) return;

        var editor = pinegrow.codeEditors.isUrlOpen(cs.url);
        if(!editor) {
            editor = pinegrow.stylesheets.editCode(cs);
        } else {
            if(!pgPageCodeEditorsStack.isEditorShown(editor.view)) {
                pgPageCodeEditorsStack.showEditor(editor.view);
            }
        }
        if(editor) {
            editor.search(cs.getSourceForRules(r));
        }
    }

    this.openFileInCodeEditor = function(file, urlWithParams) {
        var url = isApp() ? crsaMakeUrlFromFile(file) : file;
        var page = pinegrow.getCrsaPageByUrl(url);
        var ext = crsaGetExtFromUrl(url);

        urlWithParams = urlWithParams || url;

        var editor = null;

        if(page) {
            editor = this.editCode(page);
        } else {
            //var cs = pinegrow.getStylesheetByUrl(url);
            if(['jpg', 'jpeg', 'gif', 'png'].indexOf(ext) >= 0) {
                pinegrow.showQuickMessage('Images can\'t be edited in the code editor.', 4000, false, 'error');
                return false;
            } else {
                var editor = pinegrow.codeEditors.isUrlOpen(url);
                if(editor) {
                    pinegrow.codeEditors.showEditor(editor);
                    return false;
                } else {
                    var watchForChanges = true;
                    var cs = pinegrow.stylesheets.getByUrl(urlWithParams, true);
                    if (cs) watchForChanges = false;

                    page = new CrsaPage();
                    page.fromFile(file, true /* only code */, watchForChanges /* watch for changes */ );

                    //check if the cs is open, then use that code
                    // var cs = pinegrow.stylesheets.getByUrl(url, true);
                    var readOnly = false;

                    if (cs && cs.loaded) {
                        // page.setCode(cs.getSource());
                        if (cs.compiledFile) {
                            cs.getSourceToSave(function (error, source) {
                                if (error) source = '';
                                page.setCode(cs.filterSource(source));
                            });
                        }
                        else {
                            page.setCode(cs.filterSource(cs.getSource()));
                        }
                        if(!cs.canEdit(true)) {
                            readOnly = true;
                        }
                        page.codeEditForStylesheet = cs;
                    }
                    ///

                    var project = this.getCurrentProject();
                    if (project) project.addBackgroundPage(page);
                    editor = this.editCode(page, false, readOnly);
                }
            }
        }
        return editor;
    }

    this.isFileEditable = function(file) {
        if(editable_types === null) {
            var editable = (pinegrow.getSetting('file-types', '') || '') + ',.html,.htm';
            editable_types = [];
            editable = editable.split(',');
            pinegrow.callGlobalFrameworkHandler('on_get_editable_file_types', editable);
            $.each(editable, function(i, ext) {
                try {
                    ext = $.trim(ext);
                    if(ext.length) {
                        if(ext.match(/^noext$/i)) {
                            editable_types.push(/^[^\.]+$/); //files without xtension
                            editable_types.push(/[\/\\][^\.]+$/); //files without xtension
                        } else {
                            editable_types.push(new RegExp(escapeRegExp(ext) + '$', 'i'));
                        }
                    }
                } catch(err) {
                    console.log('File type error: ' + err);
                }
            })
        }

        //exclude some common
        if(file.match(/functions\.php/)) return false;
        //

        for(var i = 0; i < editable_types.length; i++) {
            if(file.match(editable_types[i])) return true;
        }

        /*
        if(crsaIsRemoteUrl(file)) {
            return true; //assume remote is html
        }*/
        return false;
    }
    //End editable file types

    this.getMimeType = function(file) {
        file = crsaRemoveUrlParameters(file);

        if(pinegrow.isFileEditable(file)) {
            return "text/html";
        }

        var i = file.lastIndexOf("."),
            ext = (i === -1) ? "default" : file.substr(i),
            mimeTypes = {
                ".bmp": "image/bmp",
                ".css": "text/css",
                ".gif": "image/gif",
                ".ico": "image/x-icon",
                ".htm": "text/html",
                ".html": "text/html",
                ".php": "text/html",
                ".php5": "text/html",
                ".asp": "text/html",
                ".aspx": "text/html",
                ".cfm": "text/html",
                ".cfml": "text/html",
                ".cfc": "text/html",
                ".shtml": "text/html",
                ".jpg": "image/jpeg",
                ".jpeg": "image/jpeg",
                ".js": "application/javascript",
                ".json": "application/json",
                ".otf": "font/opentype",
                ".woff": "application/font-woff",
                ".woff2": "application/font-woff",
                ".ttf": "application/x-font-ttf",
                ".svg": "image/svg+xml",
                ".png": "image/png",
                ".text": "text/plain",
                ".txt": "text/plain",
                ".mp3": "audio/mpeg",
                ".mp4": "video/mp4",
                ".webm": "video/webm",
                ".ogv": "video/ogg",
                ".ogg": "audio/ogg",
                ".less": "text/x-less",
                ".scss": "text/x-scss",
                "default": null
                //"default": "application/octet-stream"
            };
        return mimeTypes[ext.toLowerCase()] || mimeTypes['default'];
        //return this.httpServer.getMimeType(file);
    }

    this.getSimpleMimeType = function(file) {
        var t = this.getMimeType(file);
        if(t) {
            if(t === 'image/x-icon') return 'icon';
            if(t.startsWith('image/')) return 'image';
            if(t === "application/javascript") return 'script';
            if(t === "text/css" || t === "text/x-less" || t === "text/x-scss") return "css";
            if(t === "text/html") return 'html';
        }
        return t;
    }

    this.getPlaceholderImage = function() {
        if(pgf.kids) {
            return `alien${getRand(1,15)}.png`;
        }
        var path;
        if(isApp() && false) { //map images to local in crsa.server
            //path = this.getProxyUrl(crsaMakeUrlFromFile(crsaGetAppDir() + 'placeholders/img'));
            path = crsaMakeUrlFromFile(crsaGetAppDir() + 'placeholders/img');
        } else {
            path = "https://pinegrow.com/placeholders/img";
        }
        var r = Math.round(Math.random() * 10) + 10;
        return path + r + '.jpg';
    }

    this.getThumbnailForPlaceholderImage = function(img) {
        return img.replace('.jpg', '_thumb.jpg');
    }

    this.createProxyLinksForPlaceholderImagesInCode = function(code) {
        return code.replace(/src="(https?:\/\/pinegrow\.com\/placeholders\/[^"]+)"/g, function(m, a, b, c) {
            return 'src="' + pinegrow.getProxyUrl(a) + '"';
        })
    }

    this.getPageForElement = function($el) {
        var $if = getIframeOfElement($el);
        if($if) {
            return getCrsaPageForIframe($if);
        }
        return null;
    }
    //Styles

    this.getStylesheets = function() {
        return pinegrow.stylesheets.getAllCrsaStylesheets();
    }

    this.getStylesheetByUrl = function(url) {
        return pinegrow.stylesheets.getStylesheetByKey(url);
        //this didnt return imported sass files
        var list = this.getStylesheets();
        for(var i = 0; i < list.length; i++) {
            if(url == list[i].url) return list[i];
        }
        return null;
    }

    this.getStylesheetByFileName = function(decodedName) {
        var list = this.getStylesheets();
        for(var i = 0; i < list.length; i++) {
            var cs = list[i];
            var parentCs = cs;
            if (decodedName == cs.getDecodedLocalFileName()) {
                return cs;
            }
            else {
                if (cs.sourceStylesheet) {
                    var sourceCs = cs.sourceStylesheet;
                    parentCs = sourceCs;
                    if (decodedName == sourceCs.getDecodedLocalFileName()) {
                        return sourceCs;
                    }
                }
            }

            var foundCS = null;
            parentCs.forEachStylesheet(function(cs) {
                if (decodedName == cs.getDecodedLocalFileName()) {
                    foundCS = cs;
                    return false;
                }
            }, true);

            if (foundCS) return foundCS;
        }
        return null;
    }

    this.addEditableAreaExtension = function(ext) {
        var pgc = pinegrow.findFrameworkByKey('pg.components').getComponentType('pgc.edit');
        pgc.pgcAddExtension(ext);
    }



    //Frameworks

    this.getCodeFromDefinition = function(def) {
        return getCodeFromDefinition(def);
    }

    this.addFramework = function(f, weight) {
        if(typeof weight == 'undefined') weight = 0;
        frameworks[f.key] = f;
        frameworks[f.key].order = weight;
        if(framework_urls.length > 0) {
            f.pluginUrl = framework_urls[framework_urls.length-1];
        }
        $('body').trigger('crsa-framework-loaded', f);
    }

    this.getFrameworks = function() {
        return frameworks;
    }

    this.getComponentType = function(type) {
       // for(var i = )
    }

    this.findFrameworkWithUrl = function(url) {
        var f = null;
        $.each(frameworks, function(i, fm) {
            if(fm.pluginUrl && fm.pluginUrl == url) {
                f = fm;
                return false;
            }
        });
        return f;
    }

    this.findFrameworkByKey = function(key) {
        var f = null;
        $.each(frameworks, function(i, fm) {
            if(fm.key == key) {
                f = fm;
                return false;
            }
        });
        return f;
    }

    this.findFrameworkByType = function(type) {
        var f = null;
        $.each(frameworks, function(i, fm) {
            if(fm.type == type) {
                f = fm;
                return false;
            }
        });
        return f;
    }

    this.loadAllFrameworksFromFile = function() {
        var flist = pinegrow.getFrameworksListFromStorage();
        //debugger;
        var elist = '';
        var idx = 0;

        var not_found = [];

        var loadOne = function() {
            if(idx >= flist.length) {
                if(elist.length > 0) {
                    showAlert(elist, "Not all frameworks could be loaded");
                }
                if(not_found.length) {
                    for(var i = 0; i < not_found.length; i++) {
                        var n = flist.indexOf(not_found[i]);
                        if(n >= 0) {
                            flist.splice(n, 1);
                        }
                    }
                    crsaStorage.setValue('frameworks', flist);
                }
            } else {
                var url = flist[idx];
                var file = crsaMakeFileFromUrl(url);
                var what = crsaIsFileOrDir(file);
                if(what == 'file') {
                    _this.loadFrameworkFromUrl(url, true, function(e) {
                        if(e) {
                            elist += '<p>Unable to load ' + url + ': ' + e + '</p>';
                            if(e == 'FILE_NOT_FOUND') {
                                not_found.push(url);
                            }
                        }
                        idx++;
                        loadOne();
                    });
                } else if(what == 'dir' && pgf.components) {
                    pinegrow.loadProjectAsLibrary(file, function() {
                        idx++;
                        loadOne();
                    });
                }
                else {
                    idx++;
                    loadOne();
                }
            }
        }
        loadOne();
    }

    this.fileLoaded = function() {
        if(framework_urls.length > 0) framework_urls.pop();
    }

    this.addScriptToPage = function(page, code, execute) {
        var pgBody = page.sourceNode.findOne('body');
        if(pgBody) {
            var n = pgCreateNodeFromHtml("<script>\n    " + code + "\n</script>");
            pgBody.append(n);

            if(execute) {
                this.executeScriptInPage(page, code);
            }
        }
    }

    this.hasScriptCodeOnPage = function(page, code) {
        var has = false;
        var re = /[\s\n]+/g;
        code = code.replace(re, '');
        page.sourceNode.find('script').forEach(function(s) {
            if(s.content && s.content.replace(re, '').indexOf(code) >= 0) {
                has = true;
            }
        })
        return has;
    }

    this.addScriptCodeToPageIfNeeded = function(page, code, name, execute, main_code) {
        name = name || 'Code';
        if(this.hasScriptCodeOnPage(page, main_code || code)) {
            pinegrow.showQuickMessage(name + ' is already on the page.');
        } else {
            this.addScriptToPage(page, code, execute);
            pinegrow.showQuickMessage(name + ' was added to the page.');
        }
    }

    var skipAskCode = {}

    this.askAboutAddingCodeToPage = function(page, code, title, msg, if_needed, main_code) {
        if(if_needed === undefined) if_needed = true;
        if(skipAskCode[code] || (if_needed && this.hasScriptCodeOnPage(page, main_code || code))) return;

        var html = '<p>' + msg + '</p><pre>' + escapeHtmlCode(code) + '</pre>';
        var $d = pinegrow.showAlert(html, title, 'Don\'t add', 'Add code to the page', null, function() {
            pinegrow.makeChanges(page, 'Add code to the page', function() {
                pinegrow.addScriptToPage(page, code, true);
            })
            pinegrow.showQuickMessage('Code was added to the page.');
        })
        this.addLeftButton($d, 'Don\'t ask again during this session', function() {
            skipAskCode[code] = true;
        })
    }

    this.executeScriptInPage = function(page, code) {
        if(!page.javascriptEnabled) return;
        page.sourceNode.findOne('body').executeScript(code);
    }

    var javascript_enabled = null;

    this.setJavascriptEnabled = function(val) {
        if(val === undefined) {
            val = pinegrow.getSetting('javascript-enabled', '1') == '1';
        }
        javascript_enabled = val;
        pinegrow.jsEnabledButton.setActive(val);
        pinegrow.setSetting('javascript-enabled', val ? '1' : '0');

        pinegrow.getAllPages().forEach(function(page) {
            page.setJavascriptEnabled(val);
        })
    }

    this.isJavascriptEnabled = function() {
        return javascript_enabled;
    }

    this.loadFrameworkFromUrl = function(url, ret, done) {
        var e = null;
        if(isApp()) {
            var fs = require('fs');

            if(crsaIsFileUrl(url)) {
                var file = crsaMakeFileFromUrl(url);
                if(crsaIsFileOrDir(file) != 'file') {
                    if(done) done('FILE_NOT_FOUND');
                    return false;
                }
            }
            try {
                var z = (new Date().getTime());
                var scr = '<script async="false" src="' + url + '?z=' + z + '"></script>';
                var $body = $('body');

                var scr = document.createElement('script');
                scr.async = false;
                $(scr).on('load', function(e) {
                    $(function() {
                        framework_urls.push(url);
                        try{ $('body').trigger('pinegrow-ready',pinegrow);
                            pinegrow.fileLoaded();
                            if(done) done();
                        } catch(err) {
                            pinegrow.fileLoaded();
                            if(done) done(err);
                        }});
                });

                scr.setAttribute('src',url);
                $body.get(0).appendChild(scr);

// var run = '<script async="false">$(function() { try{ $(\'body\').trigger(\'pinegrow-ready\',pinegrow); pinegrow.fileLoaded(); } catch(err) { pinegrow.fileLoaded(); throw err; }});</script>';
/*
                $.getScript( url )
                    .done(function( script, textStatus ) {
                        try{
                            $('body').trigger('pinegrow-ready',pinegrow);
                            pinegrow.fileLoaded();
                        } catch(err)
                        {
                            console.log(err);
                            pinegrow.fileLoaded();
                            //throw err;
                        }
                    })
                    .fail(function( jqxhr, settings, exception ) {
                        console.log(exception);
                    });
*/
                //$body.append(scr);



                var flist = pinegrow.getFrameworksListFromStorage();
                if(flist.indexOf(url) < 0) {
                    flist.push(url);
                    crsaStorage.setValue('frameworks', flist);
                }
            }
            catch(err) {
                e = err;
                if(!ret) {
                    showAlert('Unable to load ' + url + '. ' + err, "Unable to load framework");
                }
                console.log(err);
                if(done) done(err);
            }
        }
        return e;
    }

    this.saveFrameworksList = function() {
        var flist = [];
        $.each(frameworks, function(key, fm) {
            if(fm.pluginUrl) {
                flist.push(fm.pluginUrl);
            }
        });
        crsaStorage.setValue('frameworks', flist);
    }

    this.getFrameworksListFromStorage = function() {
        var flist = crsaStorage.hasValue('frameworks') ? crsaStorage.getValue('frameworks') : [];
        for(var i = 0; i < flist.length; i++) {
            if(!crsaIsAbsoluteUrl(flist[i])) {
                flist[i] = crsaMakeUrlFromFile(flist[i]);
            }
        }
        return flist;
    }

    this.unloadFramework = function(f) {
        var pages = $.fn.crsapages('getAllPages');
        $.each(pages, function(i, cp) {
            cp.removeFramework(f);
        });
        if(f.key in frameworks) {
            delete frameworks[f.key];
        }
        var idx = -1;
        var flist = pinegrow.getFrameworksListFromStorage();
        if(f.pluginUrl) {
            idx = flist.indexOf(f.pluginUrl);
        }
        if(idx >= 0) flist.splice(idx, 1);

        crsaStorage.setValue('frameworks', flist);
    }

    this.hasUnsavedFrameworks = function() {
        var b = false;
        $.each(frameworks, function(i, fm) {
            if(fm.changed && fm.user_lib) b = true;
        });
        return b;
    }

    this.saveChangedFrameworks = function(page, done) {
        var list = [];

        var source;

        if(page) {
            source = page.frameworks;
        } else {
            source = [];
            $.each(frameworks, function(key, val) {
                source.push(val);
            })
        }

        for(var i = 0; i < source.length; i++) {
            var f = source[i];
            if (f.user_lib && f.changed) {
                list.push(f);
            }
        }

        var doOne = function() {
            if(list.length) {
                var f = list.shift();

                f.save(null, function() {
                    doOne();
                });

            } else {
                pinegrow.frameworksChanged();
                if(done) done();
            }
        }

        doOne();
    }

    this.addFrameworkResourcesToPageOrProject = function(f, page_or_project, opts) {

        var options = $.extend({
            overwrite: false,
            same_project: false,
            done: null,
            resources: null,
            only_add_to_page: null
        }, opts || {})

        var onDone = function() {
            if(options.done) {
                options.done(f);
            }
        }

        var is_project = page_or_project instanceof CrsaProject;

        var add = function() {

            if (is_project) {

                f.addResourcesToProject(page_or_project, function () {
                    onDone(f);
                }, options.overwrite, options.same_project, options.only_add_to_page || false, options.resources);
            } else {
                if (!page_or_project.localFile && isApp()) {
                    pinegrow.showAlert('<p>Please save the page before adding resources to it.</p><p>After saving the page go to <b>Components -&gt; Add / update resources from libraries</b> to copy and add resources to the page.</p>', 'New page is not saved yet');
                } else {
                    f.addResourcesToPage(page_or_project, function () {
                        pinegrow.refreshCurrentProjectDelayed(true);
                        onDone(f);
                    }, options.overwrite, options.same_project, options.resources);
                }
            }
        }

        if(f.on_add_resources) {
            f.on_add_resources(is_project, function(go_ahead) {
                if(go_ahead) {
                    add();
                }
            })
        } else {
            add();
        }
    }

    this.activateFrameworkInCurrentContext = function(f, done, force_page) {
        var current_project = pinegrow.getCurrentProject();
        var page = pinegrow.getSelectedPage();

        if(!page) {
            if(isApp()) {
                pinegrow.showAlert('<p>The library is loaded. Now open a page and <b>Activate ' + f.name + '</b> in <b>Page -&gt; Manage libraries &amp; framewors</b>.</p>', 'Activate the library');
            }
            done && done();
        } else {
            if(current_project && current_project.isPageInProject(page) && !force_page) {
                //activate on all open pages in project
                pinegrow.getAllPages().forEach(function(cp) {
                    if(current_project.isPageInProject(cp)) {
                        cp.addFramework(f, true /* skip update */);
                    }
                });
                //update
                pinegrow.frameworksChanged();
                if(done) done(current_project);
                pinegrow.showQuickMessage('<b>' + f.name + '</b> activated for <b>project ' + current_project.getName() + '</b>');
            } else {
                //activate on the current page only
                page.addFramework(f, null, function() {
                    if(done) done(page);
                });
                pinegrow.showQuickMessage('<b>' + f.name + '</b> activated for <b>page ' + page.name + '</b>');
            }
        }
    }



    this.frameworksChanged = function() {
        $.each($.fn.crsapages('getAllPages'), function(i,cp) {
            cp.frameworksChanged();
        });

        pinegrow.dispatchEventDelayed('on_page_frameworks_changed', null);

        //$('body').trigger('crsa-frameworks-changed');
    }

    this.getDragMenu = function() {
        return $.fn.crsa('getDragMenu');
    }

    this.drag_drop_options = {
        force_parent: null
    }

    this.getChildOfForceParent = function(pgCurrent, on_change, error_str, allow_outside) {
        if(this.drag_drop_options.force_parent) {
            var p = this.drag_drop_options.force_parent;
            if(!pgCurrent.isDescendantOf(p)) {
                if(error_str !== false) {
                    pinegrow.showQuickError(error_str || `In this mode you can only move elements within the ${p.getName({short: true})} element.`);
                }
                return allow_outside ? pgCurrent : null;
            } else {
                var changed = false;
                while(pgCurrent.parent && pgCurrent.parent !== p) {
                    pgCurrent = pgCurrent.parent;
                    changed = true;
                }
                if(on_change && changed) {
                    on_change(pgCurrent);
                }
            }
        }
        return pgCurrent;
    }

    var isolated_element = null;
    var isolated_options = {}

    var auto_focus = pgf.auto_focus && this.getSetting('auto-focus', pgf.auto_focus ? "1" : "0") === '1';

    this.setAutoFocus = function(val) {
        auto_focus = val;
        this.setSetting('auto-focus', auto_focus ? "1" : "0");
        if(!auto_focus) {
            pinegrow.isolateElement(null);
        } else {
            var pgel = pinegrow.selectedElements.getLastSelected();
            if(pgel) {
                pinegrow.autoFocusOnElementIfNeeded(pgel);
            }
            pinegrow.selectedElements.reselect();
        }
    }

    this.getAutoFocus = function() {
        return auto_focus;
    }

    this.autoFocusOnElementIfNeeded = function(pgel) {
        if(pinegrow.getAutoFocus()) {
            if(pgel) {
                var section = pgel.closest('header,footer,section');

                if (section) {
                    if (section !== pinegrow.getIsolatedElement()) {
                        pinegrow.isolateElement(section, {}, true /* skip select */);
                    }
                } else {
                    var iso = pinegrow.getIsolatedElement();
                    if (iso && !pgel.isDescendantOf(iso)) {
                        pinegrow.isolateElement(null, {}, true /* skip select */);
                    }
                }
            }
        }
    }

    this.isolateElement = function(pgel, options, skip_select) {
        isolated_element = pgel;

        isolated_options = $.extend({
            tree_hide_scripts: true,
            drag_drop_only_top_level: pgf.focus_select_only_top_level
        }, options)

        this.drag_drop_options.force_parent = isolated_options.drag_drop_only_top_level ? pgel : null;

        pinegrow.dispatchEvent('on_element_isolated', pgel ? pgel.getPage() : null, pgel, isolated_options);

        if(skip_select) return;

        var select = [];
        pinegrow.selectedElements.forEachElement(function(node) {
            var changed = false;
            if(_this.getChildOfForceParent(node, function(changed_pgel) {
                changed = true;
                select.push(changed_pgel);
            }, false)) {
                if (!changed) {
                    select.push(node)
                }
            }
        })

        if(pgel && select.length === 0) {
            select.push(pgel);
        }

        pinegrow.selectedElements.selectIds(select.map(function(pgel) { return pgel.getId(); }));
    }

    this.getIsolatedElement = function(page) {
        if(isolated_element && isolated_element.isDeleted) {
            this.isolateElement(null);
            return null;
        }
        if(page && isolated_element && isolated_element.document !== page.sourceNode) return null;
        return isolated_element;
    }

    this.getIsolatedElementOptions = function() {
        return isolated_options;
    }

    this.pageHasIsolatedElement = function(page) {
        return isolated_element && isolated_element.document === page.sourceNode;
    }

    var event_handlers = {};

    this.addEventHandler = function(event, func, priority) {
        if(!event_handlers[event]) {
            event_handlers[event] = [];
        }
        if (priority) {
            var i = 0;
            var eventsArr = event_handlers[event];
            for (i = 0; i < eventsArr.length; i++) {
                var eventObj = eventsArr[i];

                if (eventObj.priority < priority) break;
            }

            event_handlers[event].splice(i, 0, {
                func: func,
                priority: priority
            });
        }
        else {
            event_handlers[event].push({func: func, priority: 0});
        }
    }

    this.removeEventHandler = function(event, func) {
        if(event_handlers[event]) {
            var i = 0;
            var eventsArr = event_handlers[event];
            for (i = 0; i < eventsArr.length; i++) {
                var eventObj = eventsArr[i];

                if (eventObj.func == func) break;
            }

            if (i < eventsArr.length) event_handlers[event].splice(i, 1);
        }
    }

    this.dispatchEvent = function(event, a, b, c, d, e, f) {

        var page = a instanceof CrsaPage ? a : null;

        var pages = (!page && Array.isArray(a) && a.length && a[0] instanceof CrsaPage) ? a : null;

        var getHandlers = function() {
            var list = [];
            if(page) {
                for(var i = 0; i < a.frameworks.length; i++) {
                    var f = a.frameworks[i];
                    if (f.hasOwnProperty(event)) {
                        list.push({
                            func: f[event],
                            priority: 1000
                        });
                    }
                }
            } else if(pages) {
                pages.forEach(function(page) {
                    for(var i = 0; i < page.frameworks.length; i++) {
                        var f = page.frameworks[i];
                        if(f.hasOwnProperty(event) && list.indexOf(f[event]) < 0) {
                            list.push({
                                func: f[event],
                                priority: 1000
                            });
                        }
                    }
                })
            }
            if (event_handlers[event]) {
                list = list.concat(event_handlers[event]);
            }
            return list;
        }

        var list = getHandlers();

        if(event.endsWith('_async')) {
            //a = page
            //b = info/result
            //c = done

            if (list.length) {
                var idx = 0;

                var doEvent = function() {
                    if(idx < list.length) {
                        list[idx].func(a, b, function(new_b) {
                            b = new_b;
                            idx++;
                            doEvent();
                        })
                    } else {
                        if(c) c(b);
                    }
                }

                doEvent();
            } else {
                if(c) c(b)
            }
        } else {
            if(page) page.currentFrameworkHandlerReturnValues[event] = null;
            var r = null;
            if (list.length) {
                for (var i = 0; i < list.length; i++) {
                    r = list[i].func(a, b, c, d, e, f);
                    if(page) page.currentFrameworkHandlerReturnValues[event] = r;
                }
            }
            if(page) page.currentFrameworkHandlerReturnValues[event] = null;
            return r;
        }
    }

    var delayed_event_triggers = {};

    this.dispatchEventDelayed = function(event, a, b, c, d, e, f) {
        this.dispatchEventDelayedForObject(a || this, event, a, b, c, d, e, f);
    }

    this.dispatchEventDelayedForObject = function(obj, event, a, b, c, d, e, f) {
        obj = obj || this;
        if(!delayed_event_triggers[event]) {
            delayed_event_triggers[event] = new WeakMap();
        }
        if(delayed_event_triggers[event].has(obj)) {
            clearTimeout(delayed_event_triggers[event].get(obj));
        }
        var delay = 750;

        delayed_event_triggers[event].set(obj, setTimeout(function() {
            delayed_event_triggers[event].delete(obj);
            _this.dispatchEvent(event, a, b, c, d, e, f);
        }, delay));
    }

    this.callGlobalFrameworkHandler = function(name, a, b, c, d, e, f, cp) {
        if(!cp) cp = this.getSelectedPage();
        var ret = null;
        $.each(this.getFrameworks(), function(key, f) {
            if(name in f && f[name]) {
                ret = f[name](cp, a, b, c, d, e, f);
            }
        });
        if (event_handlers[name]) {
            event_handlers[name].forEach(function(h) {
                ret = h.func(cp, a, b, c, d, e, f);
            })
        }
        return ret;
    }

    this.getCollection = function(collection_or_el) {
        if(collection_or_el) {
            if(collection_or_el instanceof PgCollection) {
                return collection_or_el;
            } else {
                return new PgCollection(collection_or_el);
            }
        } else {
            return pinegrow.selectedElements.getCollection().clone();
        }
    }

    this.getCollectionForElement = function(pgel) {
        var col;
        if(!pgel || pinegrow.selectedElements.isSelectedElement(pgel)) {
            col = pinegrow.getCollection();
        } else {
            col = new PgCollection();
            col.add(pgel);
        }
        return col;
    }

    this.showUpdateResourcesDialog = function(page, project, f_list) {
        var frameworks = [];
        var folder;
        var name;
        if(project) {
            //project
            frameworks = project.getFrameworks();
            name = project.getName();
            folder = project.getDir();
        }
        if(page) {
            //page
            for(var i = 0; i < page.frameworks.length; i++) {
                if(frameworks.indexOf(page.frameworks[i]) < 0) {
                    frameworks.push(page.frameworks[i]);
                }
            }
            name = page.name;
            folder = page.localFile ? require('path').dirname(page.localFile) : null;
        }
        var f_with_r = [];
        var current_project = pinegrow.getCurrentProject();
        var current_project_pi = null;
        if(current_project) {
            current_project_pi = current_project.getProjectInfo();
        }

        for(var i = 0; i < frameworks.length; i++) {
            if(f_list && f_list.indexOf(frameworks[i]) < 0) continue; //skip this one
            if(frameworks[i].resources.has()) {
                f_with_r.push(frameworks[i]);
            }
        }

        var isSameProject = function(f) {
            if(current_project && f.project && current_project.getDir() == f.project.getDir()) return true;
            if(current_project_pi && current_project_pi.getSetting('template_framework_id') == f.type) return true;
            return false;
        }

        var updateDisplay = function($d) {
            $d.html('');
            for(var i = 0; i < f_with_r.length; i++) {
                var f = f_with_r[i];
                var $li = $('<li style="margin-bottom:15px;"><b>' + (f.resources_name || f.name) + '</b></li>');
                if(f.pluginUrl && false) {
                    $li.append('<div><small class="text-muted">' + f.pluginUrl + '</small></div>');
                }
                if(f.resources.description) {
                    $li.append('<p class="text-muted" style="margin-bottom:0;">Resources: ' + f.resources.description + '</p>');
                }
                var same_project = isSameProject(f);

                if(folder && !same_project) {
                    $li.append('<p class="text-muted" style="margin-bottom:0;">Will be copied to: <b>./' + f.getResourceNamespace() + '</b></p>');
                }
                if(same_project) {
                    $li.append('<p style="margin-bottom:0;">These resources are already a part of the current project.</p>');
                }
                var $wrap = $li.append('<div class="files-list-wrapper"></div>');

                if(!same_project || true) {
                    if(page) {
                        $wrap.append('<a href="#" class="copy btn btn-sm btn-primary">Add to or update ' + page.name + '</a>');
                    }
                    if(project) {
                        $wrap.append('<a href="#" class="copy project btn btn-sm btn-primary" style="margin-left: 10px;">Add to or update all files in ' + project.getName() + '</a>');
                    }
                }
                $wrap.append('<a href="#" class="show-list btn btn-sm btn-link">Show files and folders</a>');

                var $list = $('<ul/>', {'class' : 'files-list'}).appendTo($li.find('.files-list-wrapper'));
                for(var j = 0; j < f.resources.list.length; j++) {
                    if(f.resources.list[j].code) continue;

                    $('<li>' + (f.resources.list[j].relative_url || f.resources.list[j].url) + '</li>').appendTo($list);
                }
                $list.hide();
                $li.data('framework', f);
                $d.append($li);

            }
            $d.find('a.copy').on('click', function(e) {
                e.preventDefault();
                var $li = $(e.delegateTarget).closest('li');
                var f = $li.data('framework');
                var is_project = $(e.delegateTarget).hasClass('project');
                var same_project = isSameProject(f);

                var overwrite = $over_input.is(':checked');

                pinegrow.addFrameworkResourcesToPageOrProject(f, is_project ? project : page, {
                    done: onDone,
                    overwrite: overwrite,
                    same_project: same_project
                })

            })
            $d.find('a.show-list')
                .on('click', function(e) {
                    e.preventDefault();
                    var $li = $(e.delegateTarget).closest('li');
                    var f = $li.data('framework');
                    var $list = $li.find('ul.files-list');
                    if($list.is(':visible')) {
                        $list.hide();
                        $(this).html('Show the list of files and folders')
                    } else {
                        $list.show();
                        $(this).html('Hide the list of files and folders')
                    }
                });
        }

        var onDone = function(f) {
            //pinegrow.refreshAllPages();
            if(f.on_after_resources_added) {
                f.on_after_resources_added();
            }
            pinegrow.showQuickMessage('<b>' + (f.resources_name || f.name) + '</b> resources copied to <b>' + name + '</b>');
            pinegrow.refreshCurrentProjectDelayed(true);
        }

        var html = '';
        if(f_with_r.length == 0) {
            html += '<div><p>No libraries or plugins that are activated on this page/project have any resources that can be copied to your page or project.</p></div>';
        } else {
            html += '<div><p>The following libraries and plugins have resources (CSS &amp; JS files, images...) that can be copied to your page or project. These resources are often necessary for components to display and behave properly. <b>Note: You don\'t have to update resources if your documents already include the necessary files.</b></p></div>';
        }
        var $b = $(html);
        var $ul = $('<ul/>').appendTo($b);

        var $overwrite = $(`<hr><div><label><input type="checkbox">&nbsp;&nbsp;Overwrite resources, if they already exist. <b>Check this if you want to update framework files to the latest version.</b></label></div>`).appendTo($b);
        var $over_input = $overwrite.find('input');

        pinegrow.showAlert($b, 'Update resources in page or project', null, 'Close');

        updateDisplay($ul);
    }

    this.goThroughAllForAttribute = function (page, attr, callback) {
        var list = page.sourceNode.findIncludingSelf('[' + attr + ']');
        for(var j = 0; j < list.length; j++) {
            var el = list[j];
            var url = el.getAttr(attr);
            if (callback) callback(el, attr, url);
        }
    }

    //Starting projects
    this.goThroughAllLinkedResources = function (page, callback) {
        var attrs = ['src', 'data-src', 'href', 'action'];

        for(var i = 0; i < attrs.length; i++) {
            this.goThroughAllForAttribute(page, attrs[i], callback);
        }
    }

    var getFixedUrl = function (currnet_url, target_url, file_url, callback) {
        crsaGetRelativeDistanceForUrls(target_url, file_url, function (distance, new_url) {
            if (distance < 0) {
                var backward = "";
                for (i = distance; i < 0; i++)
                    backward += "../";
                if (callback) callback(backward + currnet_url);
            }
            else if (distance > 0) {
                var backward = new_url;
                for (i = 0; i < distance; i++)
                    backward = "../" + backward;
                if (callback) callback(backward);
            }
            else if (distance == 0) {
                if (callback) callback(new_url);
            }
        });
    }

    this.copyMissingExternalFiles = function (page, sourceUrl, destUrl, projectBaseUrl) {

        page.sourceNode.walk(function(node) {
            if(node.isElement) {
                node.mapLinksToNewLocation(projectBaseUrl, destUrl, true /* only if dest file exists */);
            }
            return true;
        })

        var remote_template = crsaIsRemoteUrl(sourceUrl);

        this.goThroughAllLinkedResources(page, function (el, attr, url) {
            if (el.tagName == "a") return;

            if (url && !crsaIsAbsoluteUrl(url)) {
                var newUrl = crsaGetUrlToSmartCombine(destUrl, url);
                var new_file_url = crsaCombineUrlWith(destUrl, newUrl);

                var fs = require('fs');
                var source_file_url = crsaCombineUrlWith(sourceUrl, url);
                var new_file_path = crsaMakeFileFromUrl(new_file_url);

                if(crsaIsRemoteUrl(source_file_url)) {
                    //console.log(`REMOTE FILE ${source_file_url}`);
                    if(!crsaIsFileExist(new_file_path)) {
                        pgCopyRemoteUrlToLocalPath(source_file_url, new_file_path);

                        var cs = pinegrow.stylesheets.getByUrl(new_file_url);
                        if (cs) {
                            cs.reload();
                        }
                    }
                } else {

                    var source_file_path = crsaMakeFileFromUrl(source_file_url);

                    if (crsaIsFileExist(source_file_path) && !crsaIsFileExist(new_file_path)) {
                        crsaCreateDirs(crsaGetFileDir(new_file_path), fs);
                        crsaCopyFileSync(fs, source_file_path, new_file_path);

                        //Check if stylesheet is already loaded, but has empty content because the file didn't exist before
                        var cs = pinegrow.stylesheets.getByUrl(new_file_url);
                        if (cs) {
                            cs.reload();
                        }
                    }
                }
            }
        });
    }

    this.fixLinks = function (project, page, sourceUrl, destUrl, done, copy_external_files, skip_asking_user, skip_links_in_editable_areas) {

        var overwriteFiles = {};
        this.goThroughAllLinkedResources(page, function (el, attr, url) {
            url = crsaCleanUpUrl(url);

            if(url === null) {
                //no url
                //debugger;
            }
            if(url && !crsaIsAbsoluteUrl(url)) {

                var newUrl = crsaGetUrlToSmartCombine(destUrl, url);
                var new_file_url = crsaCombineUrlWith(destUrl, newUrl);
                /*
                var skip = false;
                if(skip_links_in_editable_areas) {
                    //is this in editable area?
                    var area = el.closest('[data-pgc-edit],[data-pgc-field]');
                    if()
                }*/
                if (!project.getFileForUrl(new_file_url) && !crsaIsFileOrDir(crsaMakeFileFromUrl(new_file_url))) {

                    var new_file_name = crsaGetNameFromUrl(new_file_url);
                    var currectPath = project.getFilesWithName(new_file_name);
                    if (currectPath.length > 0) {
                        if (currectPath.length == 1) {
                            getFixedUrl(url, currectPath[0].url, destUrl, function (fixedUrl) {
                                el.setAttr(attr, fixedUrl);
                            });
                        }
                        else {
                            var useFiles = [];
                            var breakLoop = false;
                            for (var i=0; i < currectPath.length; i++) {
                                crsaGetRelativeDistanceForUrls(currectPath[i].url, new_file_url, function (distance, new_url) {
                                    new_url = (distance < 0) ? url : new_url;

                                    if (!overwriteFiles[url]) {
                                        overwriteFiles[url] = {urls: []};
                                    }

                                    overwriteFiles[url]['name'] = new_file_name;
                                    overwriteFiles[url]['el'] = el;
                                    overwriteFiles[url]['attr'] = attr;
                                    overwriteFiles[url]['urls'].push(currectPath[i].path);
                                    useFiles.push({
                                        distance: distance,
                                        url: new_url
                                    });

                                    if (currectPath[i].url == crsaCombineUrlWith(sourceUrl, url)) {
                                        useFiles = [{
                                            distance: distance,
                                            url: new_url
                                        }];
                                        overwriteFiles[url]['urls'] = useFiles;
                                        breakLoop = true;
                                    }
                                });
                                if (breakLoop) break;
                            }

                            var currentFileNameArray = useFiles.sort(function (a, b) {
                                if (a.distance > b.distance) return 1;
                                if (a.distance < b.distance) return -1;
                                if (b.distance == undefined) return -1;
                                if (a.distance == undefined) return 1;
                            });

                            if (currentFileNameArray[0].distance < 0 || currentFileNameArray.length == 1) {
                                var backward = "";
                                for (var i = currentFileNameArray[0].distance; i < 0; i++)
                                    backward += "../";
                                var final_url = backward + currentFileNameArray[0].url;
                                el.setAttr(attr, final_url);
                                delete overwriteFiles[url];
                            }
                        }
                    }
                    else {
                        if (copy_external_files) {
                            if (el.tagName == "a" && attr == "href") return;

                            var fs = require('fs');
                            var source_file_url = crsaCombineUrlWith(sourceUrl, url);

                            var source_file_path = crsaMakeFileFromUrl(source_file_url);
                            var new_file_path = crsaMakeFileFromUrl(new_file_url);

                            if (crsaIsFileExist(source_file_path)) {
                                crsaCreateDirs(crsaGetFileDir(new_file_path), fs);
                                crsaCopyFileSync(fs, source_file_path, new_file_path);
                            }
                        }
                    }
                }
                else if (newUrl != url) {
                    el.setAttr(attr, newUrl);
                }
            } else {

            }
        });


        var hasBadLinks = !$.isEmptyObject(overwriteFiles);
        if (skip_asking_user) {
            if (done) done(hasBadLinks);
        }
        else {
            var cmc = new CrsaMultiChooser();
            for (file in overwriteFiles) {
                cmc.addItem(
                    "<b>"+file+"</b> missing. Use:",
                    file,
                    overwriteFiles[file].urls
                );
            }
            cmc.show(null, function (chosenFiles) {
                if (chosenFiles) {
                    for (var key in chosenFiles) {
                        var chosenFile = chosenFiles[key];
                        var fileObj = overwriteFiles[key];

                        getFixedUrl(null, crsaMakeUrlFromFile(chosenFile), page.url, function (fixedUrl) {
                            fileObj.el.setAttr(fileObj.attr, fixedUrl);
                        });
                    }
                }
                if (done) done(hasBadLinks);
            });
        }
    }

    this.addTemplateProject = function(project, index) {
        if(typeof index == 'undefined' || index === null) {
            this.projectTemplates.push(project);
        } else {
            this.projectTemplates.splice(index, 0, project);
        }
        $.fn.crsa('addTemplateProject', project, index);
    }

    this.addTemplateProjectFromTemplateDefinition = function(def) {
        var p = new CrsaProject();
        p.fromSimpleTemplate(def);
        this.addTemplateProject(p, 0);
        return p;
    }

    this.addTemplateProjectFromFolder = function(folder, done, index, skip_add_to_templates) {
        var p = new CrsaProject();
        p.templateBrowserIndex = index;
        p.fromFolder(folder, function(p) {
            p.sortItems();
            if(!skip_add_to_templates) _this.addTemplateProject(p, index);
            done(p);
        })
    }
    //Pages

    this.openPage = function(url, done, select_page, on_css_done, options) {

        var hash = crsaGetUrlHash(url);
        url = crsaRemoveUrlHash(url);

        //handle ' in url that causes pg to crash
        url = url.replace(/\'/g, '%27');

        options = options || {};

        if(/*!select_page*/ select_page === false) {
            options.skip_page_select = true;
        }

        if(pgf.use_page_wrappers) {
            var pi = pinegrow.getCurrentProjectInfo();

            if(options.open_with_wrapper) {
                var wrapper_url = pi.getSettingForUrl(url, 'wrapper_url');
                var wrapper_selector = pi.getSettingForUrl(url, 'wrapper_selector');

                if (wrapper_url) {
                    options.wrapper_url = wrapper_url;
                    options.wrapper_selector = wrapper_selector;
                }
            }
        }

        if(crsaIsFileUrl(url) && !this.isFileEditable(crsaMakeFileFromUrl(url))) {
            pinegrow.showQuickMessage('Files of this type can not be edited in Pinegrow.');
            return;
        }
        $.fn.crsa('openPage', url, /* null,*/ function(cp) {
            if(hash) cp.scrollToHash(hash);
            if(done) done(cp);
            if(/* !isApp() && */ (!options || !options.do_not_remember_page_open)) {
                var project = pinegrow.getCurrentProject();
                project && project.pageWasOpenedOrClosed(cp, true);
            }
        }, function(cp) {
            //on source loaded
            if(crsaIsFileUrl(url)) {
                cp.setLocalFile(crsaMakeFileFromUrl(url));
            } else {
                var mappedFile = pinegrow.getMappedRemoteFileForUrl(url, 'html');
                if(mappedFile) {
                    cp.setLocalFile(mappedFile);
                    cp.mappedRemote = true;
                    cp.mappedRemoteFirstSave = !crsaIsFileExist(mappedFile);
                }
            }
            if(options && options.on_source_loaded) {
                options.on_source_loaded(cp);
            }
        }, null, null, options);
    }

    this.openOrShowPage = function(url, done, select_page, on_css_done, options) {

        var hash = crsaGetUrlHash(url);
        url = crsaRemoveUrlHash(url);
        var cp = this.getCrsaPageByUrl(url);
        if(!cp) {
            $('body').trigger('crsa-page-changing');

            this.openPage(url, function(cp) {
                if(done) done(cp, true);
            }, select_page /* false */, on_css_done, options);
        } else {
            if(false && cp.loading && !cp.sourceNode) {
                pinegrow.showQuickError('The page is already in the process of being opened...');
            } else {
                cp.show(true);
                if (hash) cp.scrollToHash(hash);
            }
            if(done) done(cp, false);
        }
    }

    this.showOpenUrlDialogAndOpenPage = function(done, default_url) {
        pinegrow.showOpenUrlDialog(function(url) {
            pinegrow.openOrShowPage(url, function(page) {
                var project = pinegrow.getCurrentProject();
                project ? project.addRemoteUrl(url) : recentFiles.add(url);
                done && done(page);
            });
        }, default_url)
    }

    this.showPageWrapperDialog = function(page_url, error, open, page) {
        var pi = pinegrow.getCurrentProjectInfo();

        page = page || pinegrow.getCrsaPageByUrl(page_url);

        var dialog = pinegrow.showQuickDialog('Open in a wrapper page', 'Wrapper page', (error || '') + `<p>Select the URL of the page that will be used as a wrapper for opening this document.</p>`, {
            url: {
                type: 'text',
                name: 'URL of the page',
                value: pi.getSettingForUrl(page_url, 'wrapper_url'),
                helptext: isApp() ? `<p>Absolute URL of the wrapper page.</p>` : `<p>Absolute or relative URL of the wrapper page. The URL should be on the same domain or allow cross-domain access.</p>`,
                validate: function(pgel, fn, value) {
                    if(open && !value) {
                        return 'Wrapper URL is required.'
                    }
                }
            },
            selector: {
                type: 'text',
                name: 'CSS selector of the target element',
                value: pi.getSettingForUrl(page_url, 'wrapper_selector'),
                helptext: `<p>A CSS selector of the HTML element on the wrapper page where the document should be displayed.</p>`,
                validate: function(pgel, fn, value) {
                    if(open && !value) {
                        //return 'The selector is required.'
                    }
                },
                placeholder: 'default: article, main, body'
            }
        }, 'Cancel', 'Save' + (open ? ' &amp; Open' : ''), function() {}, function(values) {
            pi.setSettingForUrl(page_url, 'wrapper_url', values.url);
            pi.setSettingForUrl(page_url, 'wrapper_selector', values.selector);
            pi.save();

            if(open) {
                //setTimeout(function() {
                if(page) {
                    page.askClose(function() {
                        pinegrow.openOrShowPage(page_url, null, undefined, null, {
                            open_with_wrapper: true
                        })
                    });
                } else {
                    pinegrow.openOrShowPage(page_url, null, undefined, null, {
                        open_with_wrapper: true
                    });
                }
               // }, 100)
            }
            return true; //close
        }, true)
    }

    var example_count = 0;

    this.openExample = function(url) {
        var f = new PgFramework('pg.example.' + example_count++, 'Example');

        var page_url = pgMakeFileUrl(crsaGetBaseForUrl(window.location.href) + url);

        //var project_url = crsaGetBaseForUrl(url.replace('open:', ''));

        f.addTemplateProjectFromResourceFolder(crsaMakeFileFromUrl(crsaGetBaseForUrl(page_url)), function(p) {
            pinegrow.openOrShowPage(page_url, function(page) {
                page.setLocalFile(null);
                page.crsaProjectTemplate = p;
                p.framework = null;
            });
        }, 0, null, true, true)
    }

    this.duplicatePage = function(crsaPage) {

        var np = crsaPage.duplicate();
        np = null;
        //np.original_new_url = crsaPage.url;

        $.fn.crsa('openPage', crsaPage.url, function(newCrsaPage) {

        }, function(newCrsaPage) {

        }, np, null, null, function(newCrsaPage) {
            newCrsaPage.callFrameworkHandler('on_page_cloned', crsaPage);
        });
    }

    this.reloadPage = function(crsaPage) {
        if(crsaPage.changed) {
            showAlert("This page has unsaved changes. Are you sure you want to reload it?",  "Page has unsaved changes", "Cancel", "Reload", null, function() {
                $.fn.crsa('reloadPage', crsaPage, false);
            })
        } else {
            $.fn.crsa('reloadPage', crsaPage, false);
        }
    }

    this.showPage = function(cp) {
        cp.show(true);
        //this.pageTabs.showPage(cp, hide_others);
    }

    this.getAllPages = function() {
        return $.fn.crsapages('getAllPages');
    }

    this.getAllPageViews = function() {
        var r = [];
        this.getAllPages().forEach(function(page) {
            page.getPageViews().forEach(function(pv) {
                r.push(pv);
            })
        })
        return r;
    }

    this.getCrsaPageById = function(id) {
        var pages = $.fn.crsapages('getAllPages');
        for(var i = 0; i < pages.length; i++) {
            if(pages[i].uid == id) return pages[i];
        }
        return null;
    }

    this.getPageViewById = function(id) {
        var pages = $.fn.crsapages('getAllPages');
        for(var i = 0; i < pages.length; i++) {
            var pv = pages[i].getPageViewById(id);
            if(pv) return pv;
        }
        return null;
    }

    this.getCrsaPageByUrl = function(url) {
        var pages = $.fn.crsapages('getAllPages');
       /* var m = url.match(/pgid=([0-9]+)/);
        if(m) {
            var pageid = parseInt(m[1]);
            var cp = this.getCrsaPageById(pageid);
            if(cp) return cp;
        }
        for(var i = 0; i < pages.length; i++) {
            if(pages[i].url == url && !pages[i].live_update) return pages[i];
        }*/
        var decoded_url = pgDecodeUrl(url);

        for(var i = 0; i < pages.length; i++) {
            if(pages[i].url === url || pages[i].url === decoded_url) return pages[i];
            if(pages[i].mappedRemote) {
                //are we using mappedurl to lookup the page?
                var remote_url = pinegrow.getRemoteUrlFromMappedUrl(url);
                if(remote_url === pages[i].url) return pages[i];
            }
        }
        return null;
    }

    this.getSourceNodeOfUrl = function(url, clone, assign_ids, info) {
        info = info || {}

        var assignIdsIfNeeded = function(node) {
            if(assign_ids) {
                node.assignIdAndAddToCatalog(true);
            }
            return node;
        }
        //only for file at the moment
        var cp = this.getCrsaPageByUrl(url);
        if(cp && cp.sourceNode) {
            info.page = cp;
            return assignIdsIfNeeded(clone ? cp.sourceNode.clone(true) : cp.sourceNode);
        }

        var project = pinegrow.getCurrentProject();
        if(project) {
            var bck_page = project.getBackgroundPageForUrl(url);
            if(bck_page) {
                if(bck_page.onlyCode) {
                    var p = new pgParser();
                    p.assignIds = assign_ids;
                    p.parse(bck_page.getCode());
                    return p.rootNode;
                } else {
                    info.page = bck_page;
                    return assignIdsIfNeeded(clone ? bck_page.sourceNode.clone() : bck_page.sourceNode);
                }
            }
        }

        if(isApp()) {
            var fullPath = crsaMakeFileFromUrl(url);
            var fs = require('fs');
            try {
                if (fs.existsSync(fullPath)) {
                    var data = fs.readFileSync(fullPath);
                    var html = data.toString('utf8');
                    info.html = html;
                    if (info.skip_parse_if_html_equal && info.skip_parse_if_html_equal === html) {
                        info.parse_skipped = true;
                        return null;
                    }
                    var p = new pgParser();
                    p.assignIds = assign_ids;
                    p.parse(html);

                    pinegrow.dispatchEvent('on_got_source_node_of_url_from_file', null, p.rootNode, url);
                    return p.rootNode;
                }
            } catch (err) {
                return null;
            }
        } else {
            var html = project.getContentOfUrlWithLookup(url);
            if(html !== null) {
                info.html = html;
                if (info.skip_parse_if_html_equal && info.skip_parse_if_html_equal === html) {
                    info.parse_skipped = true;
                    return null;
                }
                var p = new pgParser();
                p.assignIds = assign_ids;
                p.parse(html);
                return p.rootNode;
            }
        }
        return null;
    }

    this.getCrsaPageForUrl = function(url, add_as_background) {
        //only for file at the moment
        var project = pinegrow.getCurrentProject();

        var cp = this.getCrsaPageByUrl(url);
        if(!cp) {
            cp = new CrsaPage(true /* not interactive */);
            cp.fromFile((!project || !project.loaded_from_backend) ? crsaMakeFileFromUrl(url) : url);
            if(add_as_background) {
                var p = pinegrow.getCurrentProject();
                if(p) {
                    p.addBackgroundPage(cp)
                }
            }
        }
        return cp;
    }


    this.getWorkingDir = function() {
        var project = pinegrow.getCurrentProject();
        if(project) return project.currentWorkingDir || project.getDir();
        var page = pinegrow.getSelectedPage();
        if(page && page.localFile) {
            return require('path').dirname(page.localFile);
        }
        return null;
    }

    this.setWorkingDir = function(dir) {
        var project = pinegrow.getCurrentProject();
        if(project) {
            project.currentWorkingDir = dir;
        }
    }

    this.getCrsaPageOfPgParserNode = function(pgel) {
        var pages = $.fn.crsapages('getAllPages');
        for(var i = 0; i < pages.length; i++) {
            if(pages[i].sourceNode == pgel.document) return pages[i];
        }
        return pgel.page || null;
    }

    this.isElementLocked = function(pgel, cp) {
        if(!cp) cp = this.getCrsaPageOfPgParserNode(pgel);
        if(!cp) return false;
        return cp.callFrameworkHandler('on_is_element_locked', pgel);
    }

    this.willChangeDom = function() {
        crsaWillChangeDom();
    }


    this.showUIPanel = function(name, flash, if_not_collapsed, reveal, next_to_panel) {
        var view = pgUIViewManager.getView(name);
        if(view) {
            view.showInUI(flash, if_not_collapsed, next_to_panel);
            if(reveal) {
                pgRevealElement(view.getTab().element);
            }
        }
        return view;
    }


    this.hideUIPanel = function(name) {
        var view = pgUIViewManager.getView(name);
        if(view) {
            view.hideInUI();
        }
    }

    this.isHiddenUIPanel = function(name) {
        var view = pgUIViewManager.getView(name);
        if(view) {
            return view.isHiddenInUI();
        }
        return true;
    }

    this.undoShowUIPanel = function(name) {
        var view = pgUIViewManager.getView(name);
        if(view) {
            view.undoShowInUI();
        }
        return view;
    }

    this.setNeedsUpdate = function($el, now) {
        $.fn.crsa('setNeedsUpdate', now, $el);
    }

    this.updateTree = function($el) {
        //this.setNeedsUpdate($el, true);

        this.updateTreeMultiple([$el]);
    }

    /*
    Update tree for multiple elements
    @list: array of $el elements for which to update the tree

    $el is a jQuery object with one element on the page that was changed

    Updating the tree is expensive. Should work in optimal way, update as few elements as possible
     */
    this.updateTreeMultiple = function(list) {
        return;//not used in 3.0
    }

    this.updateTreeLine = function($el) {
        return;//not used in 3.0
    }

    this.repaintTree = function() {
        return;//not used in 3.0
    }

    var selectedPageTab = null;

    this.setSelectedPageTab = function(tab) {
    //    if(selectedPageTab != tab) {
            if (selectedPageTab) {
                //selectedPageTab.setSelected(false);
            }
            selectedPageTab = tab;

            if (tab) {
                //selectedPageTab.setSelected(true);
                var page = tab.getPage();
                if (page.$iframe) {
                    pinegrow.setSelectedPage(page);
                } else {
                    //selectedPageTab.setSelected(true);
                }
            } else {
                pinegrow.setSelectedPage(null);
            }
    //    }
    }

    this.getSelectedPageTab = function() {
        return selectedPageTab;
    }

    this.getSelectedPage = function() {
        return $.fn.crsa('getSelectedCrsaPage');
    }

    this.setSelectedPage = function(crsaPage) {
        return $.fn.crsa('setSelectedPage', crsaPage);
    }

    this.isElementSelected = function(pgel) {
        if(!(pgel instanceof pgParserNode)) {
            console.error('pgParserNode expected');
            debugger;
            return false;
        }
        return this.selectedElements.isSelectedElement(pgel);
    }

    this.isDescendantSelected = function(pgel) {
        if(!(pgel instanceof pgParserNode)) {
            console.error('pgParserNode expected');
            debugger;
            return false;
        }
        return this.selectedElements.isDescendantSelected(pgel);
    }

    this.isElementOrDescendantSelected = function(pgel) {
        return this.isElementSelected(pgel) || this.isDescendantSelected(pgel);
    }

    this.getUniqueSelectorOfPgel = function(pgel) {
        return pgel.getAttr('data-pg-sel')
    }

    this.assignUniqueSelectorToPgel = function(pgel, change_event, force_new_sel) {
        var selector_attr = 'data-pg-sel';

        var cur_sel = pgel.getAttr(selector_attr);
        if(cur_sel && !force_new_sel) return cur_sel;

        var list = pgel.getUniqueSelectors(0, 9999, pgel.document, true);

        var sel = list.length ? list[0] : null;

        var func = function() {
            pgel.setAttribute(selector_attr, sel);
            pgel.setAttribute(selector_attr + '-auto', null);
        }

        if(sel) {
            if(change_event) {
                pinegrow.makeChanges(pgel, typeof change_event !== 'string' ? "Auto selector" : change_event, function () {
                    func();
                })
            } else {
                func();
            }
        }
        return sel;
    }

    this.assignNameToPgel = function(pgel, change_event) {
        var name_attr = 'data-pg-name';

        var cn = pgel.getAttr(name_attr);
        if(cn) return cn;

        var name = null;

        var page = pgel.getPage();

        if(page) {
            var names = []
            page.callFrameworkHandler('on_get_name_for_element', pgel, names);
            if(names.length) {
                name = names[0];
            }
        }

        if(!name) {
            if(pgel.tagName.match(/^h[1-6]$/i)) {
                name = `Heading ${pgel.tagName.replace('h', '')}`;
            }

            var parent_name = null;
            var i = 0;
            var p = pgel.parent;

            while(p && i < 4) {
                var n = p.getAttr(name_attr);
                if(n && ['body', 'html'].indexOf(p.tagName) < 0) {
                    parent_name = n;
                    break;
                }
                i++;
                p = pgel.parent;
            }

            switch(pgel.tagName) {
                case 'html':
                    name = 'HTML Element';
                    break;

                case 'body':
                    name = 'BODY Element';
                    break;
            }

            var map = {
                'img': 'Image',
                'p': 'Paragraph',
                'a': 'Link'
            }

            if(!name) {
                var tag_name = map[pgel.tagName] || (pgel.tagName.charAt(0).toUpperCase() + pgel.tagName.slice(1))

                var def = page.getMainType(pgel);
                if(def) {
                    var def_name = def.name_short || def.name;
                    if(def_name !== pgel.tagName) {
                        tag_name = def_name;
                    }
                    tag_name = map[tag_name] || tag_name;
                }

                var scope = parent_name ? p : pgel.document;

                var i = 1;
                do {
                    var n = i > 1 ? `${tag_name} ${i}` : tag_name;

                    if(!scope.findOne(`[data-pg-name="${n}"]`)) {
                        name = n;
                        break;
                    }
                    i++;
                }
                while(i < 100)
            }
            if(!name) {
                name = tag_name;
            }

            if(parent_name) {
                name = `${parent_name} - ${name}`;
            }

            if(change_event) {
                pinegrow.makeChanges(pgel, typeof change_event !== 'string' ? "Auto name" : change_event, function () {
                    pgel.setAttr(name_attr, name);
                })
            } else {
                pgel.setAttr(name_attr, name);
            }
        }
    }

    this.clearSelectedElements = function () {
        this.selectedElements.clear();
    }

    this.getSelectedElement = function() {
        return this.selectedElements.getLastSelected();
    }

    this.getSelectedElementName = function() {
        var el = this.getSelectedElement();
        if(el) {
            return el.getName();
        }
    }

    this.selectElement = function(pgel, user_action, multi, $el, event) {
        $.fn.crsa('selectElement', pgel, user_action, multi, $el, event);
    }

    this.reselectElement = function() {
        this.selectedElements.reselect();
    }

    this.getSelectedRule = function () {
        return $.fn.crsa('getSelectedRule');
    }

    this.isSelectedRule = function (rule) {
        return $.fn.crsa('isSelectedRule', rule);
    }

    this.getSelectedRuleIndices = function () {
        return $.fn.crsa('getSelectedRuleIndices');
    }

    this.getSelectedRuleStylesheet = function () {
        return $.fn.crsa('getSelectedRuleStylesheet');
    }

    this.selectRule = function (rule, options) {
        return $.fn.crsa('selectRule', rule, null, options);
    }

    this.autoSelectRule = function(pgel, shown_nodes, force_inline) {
        var select_rule = null;

        var selectRule = function (rule) {
            $.fn.crsa('selectRule', rule);
        }

        var pgel_selected_css_rule = null;
        if(pgel) {
            var nodeid = pgel.getData('selectedCssRule');

            if(nodeid === 'none') {
                selectRule(null);
                return true;
            }
            if(nodeid && !force_inline) {
                if(!nodeid.csurl) {
                    pgel_selected_css_rule = 'inline';
                } else {
                    pgel_selected_css_rule = pgcssh.getNodeFromGlobalId(nodeid);
                }
            }
        }



        if(!pgel_selected_css_rule) {
            var rule = this.getSelectedRule();
            if(rule && shown_nodes.indexOf(rule) >= 0) {
                selectRule(rule)
                return true;
            }
        }

        var selectFirst = function() {
            if(shown_nodes.length == 1) {
                selectRule(shown_nodes[0])
            } else if(shown_nodes[0].nodes.length) {
                selectRule(shown_nodes[0])
            } else if(pinegrow.activeElementStyleWasSelected) {
                selectRule(shown_nodes[0])
            } else if(shown_nodes[1].selector && shown_nodes[1].selector.startsWith('*')) {
                selectRule(shown_nodes[0])
            } else {
                selectRule(shown_nodes[1])
            }
        }

        if(shown_nodes.length) {
            if(pgel_selected_css_rule === 'inline') {
                selectRule(shown_nodes[0]);

            } else if(pgel_selected_css_rule && shown_nodes.indexOf(pgel_selected_css_rule) >= 0 && !pgel_selected_css_rule.selector.startsWith('*')) {
                //select the rule that was stored in pgel as last selected for that pgel
                selectRule(pgel_selected_css_rule);
            } else if (select_rule) {
                var r = select_rule.data || select_rule;
                if (shown_nodes.indexOf(r) >= 0) {
                    selectRule(r);
                } else {
                    selectFirst();
                    //selectRule(shown_nodes[0]);
                }
            } else {
                selectFirst();
                //selectRule(shown_nodes[0]);
            }
            return true; //rule selected
        }
        return false; //rule not selected
    }

    //return the css rule that is selected for pgel. If needed, get active rules
    this.getSelectedRuleForPgel = function(pgel, done) {
        var rule = this.getSelectedRule();
        if(rule) {
            done(rule)
        } else {
            var _this = this;
            pinegrow.pgActiveRules.clear();
            pinegrow.pgActiveRules.setPgel(pgel);

            pinegrow.pgActiveRules.updateMatchedRules(true, function (rules) {
                var an = pinegrow.pgActiveRules.getActiveNodes();
                var list = [];
                an.inline_rules.forEach(function(r) {
                    list.push(r)
                })
                an.nodes_to_show.forEach(function(ri) {
                    list.push(ri.node)
                })
                _this.autoSelectRule(pgel, list, false);
                done(_this.getSelectedRule());
            })
        }
    }

    this.highlightElement = function($el, color, force) {
        $.fn.crsa('highlightElement', $el, color, force);
    }

    this.highlightElementInTree = function(pgel) {
        pinegrow.getTreeViews().forEach(function(treeView) {
            if(treeView.markInsertPositionInTree) treeView.markInsertPositionInTree(pgel, 'self');
        })
    }

    this.refreshPageChangedStatusForAllPages = function(force) {
        this.getAllPages().forEach(function(page) {
            page.refreshPageChangedStatus(force);
        })
    }

    this.showContextMenuForElement = function(pgel, event, $clicked, $container) {

        event.preventDefault();
        event.stopPropagation();

        if(!pgel) return;

        $clicked = $clicked || $(event.delegateTarget);

        var for_all_selected = pinegrow.selectedElements.isSelectedElement(pgel);

        var selectedPage = pinegrow.getCrsaPageOfPgParserNode(pgel);
        if(!selectedPage) return;

        var def_actions;

        if(event.shiftKey) {
            //Only insert
            var im = new PgInsertLibElementMenu(pgel, for_all_selected ? null : pgel, true);
            def_actions = im.getActions();
        } else {
            def_actions = $.fn.crsa('getActionsMenuFor', pgel, for_all_selected);
            selectedPage.callFrameworkHandler('on_build_actions_menu', def_actions, pgel, for_all_selected);
        }

        var $target = pgel.get$DOMElement();

        if($clicked) {
            var app_win = pgGetAppWindowForDOMElement($clicked.get(0));
            if(app_win && app_win.window !== window) {
                //in external app window
                $target = $clicked;
                if(!$container) $container = $clicked.closest('body');
            }
        }

        var contextMenu = new CrsaContextMenu(null, $target);
        contextMenu.actions = def_actions;
        contextMenu.$target = $target;
        contextMenu.targetPgel = pgel;
        contextMenu.$clicked = $clicked;

        contextMenu.showAt(event.pageX, event.pageY, $container || $clicked.closest('body'));
        contextMenu.updatePosition();
    }

    this.showCSSRules = function($el, filter, active) {
        $.fn.crsa('showCSSRules', $el, filter, active);
    }

    this.scrollCanvasToElement = function($el) {
        debugger;
        $.fn.crsa('scrollCanvasToElement', $el);
    }

    this.scrollToElement = function($el) {
        debugger;
        $.fn.crsa('scrollCanvasToElement', $el);
    }

    this.addPageAction = function(name, func) {
        $.fn.crsapages('addCustomPageAction', name, func);
    }

    this.openProjectPanelIfNoPageOpen = function() {
        setTimeout(function() {
            var p = pinegrow.getCurrentProject();
            if(p && pinegrow.getAllPages().length === 0) {
                pinegrow.showTab('project');
            }
        }, 500)
    }

    this.showTab = function(tab) {
        var v = pgUIViewManager.getView(tab);
        if(v) {
            v.showInUI();
        }
        //$.fn.crsa('showTab', tab);
    }

    this.formatHtml = function(src) {
        var p = new pgParser();
        p.assignIds = false;
        p.parse(src);
        return p.rootNode.toStringOriginal(true, this.getFormatHtmlOptions());

    }

    this.getFormatHtmlOptions = function() {
        var indent = new Array(parseInt(pinegrow.getSetting('html-indent-size', '4')) + 1).join(' ');
        if(pinegrow.getSetting('html-indent-tabs', '0') == '1') indent = "\t";
        return {
            indent: indent
        }
    }

    var quickLibWin = null;

    this.getQuickLibWindow = function() {
        return quickLibWin;
    }

    this.showQuickInsert = function(pgel, $target, placement, static) {
        pgel = pgel || pinegrow.selectedElements.getLastSelected();

        if(!quickLibWin) {
            quickLibWin = new PgQuickPageLib(pgel, 'append', $target, placement, static);
        } else {
            quickLibWin.lib.setInsertInfo(pgel);
            quickLibWin.show(true, $target);
        }
        return quickLibWin;
    }

    this.populateCodeMirrorOptions = function(options) {
        options.theme = pinegrow.getSetting('code-theme-cm', pinegrow.defaults.codeTheme);
        options.indentUnit = parseInt(pinegrow.getSetting('html-indent-size', '4'));
        options.tabSize = options.indentUnit;
        options.indentWithTabs = pinegrow.getSetting('html-indent-tabs', '0') == '1';

        options.extraKeys = {
            "Ctrl-Q": function(cm){ cm.foldCode(cm.getCursor()); },
            "Ctrl-Space": "autocomplete",
            "Ctrl-U": function() {},
            "Cmd-U": function() {},
            "Ctrl-Z": function() {},
            "Cmd-Z": function() {}
        }
        options.undoDepth = 0;

        return options;
    }

    this.getCodeMirrorLintOptions = function() {
        return {
            options: {
                rules : {
                    "tagname-lowercase": true,
                    "attr-lowercase": true,
                    "attr-value-double-quotes": false,
                    "doctype-first": false,
                    "tag-pair": true,
                    "spec-char-escape": true,
                    "id-unique": true,
                    "src-not-empty": true,
                    "attr-no-duplication": true
                }
            }
        }
    }

    this.getHtmlIndentForLevel = function(level, options) {
        if(!options) options = this.getFormatHtmlOptions();
        var s = '';
        for(var i = 0; i < level; i++) {
            s += options.indent;
        }
        return s;
    }

    this.setIgnoreClicks = function(val) {
        ignore_clicks = val;
    }

    this.getIgnoreClicks = function() {
        return ignore_clicks;
    }

    this.getUniqueId = function(prefix) {
        return getUniqueId(prefix);
    }

    this.showNotice = function(msg, title, key, done, only_once, force, not_checked_by_default, with_cancel, extraButton, onExtra, options) {
        showNotice(msg, title, key, done, only_once, force, not_checked_by_default, with_cancel, extraButton, onExtra, options);
    }

    this.showAlert = function(msg, title, cancel, ok, onCancel, onOk, extraButton, onExtra) {
        return showAlert(msg, title, cancel, ok, onCancel, onOk, extraButton, onExtra);
    }



    this.addLeftButton = function($d, label, func) {
        $('<a href="#" class="btn pull-left">' + label + '</a>').appendTo($d.find('.modal-footer')).on('click', function(e) {
            e.preventDefault();
            $d.modal('hide');
            func();
        });
    }

    this.alert = function(msg, title) {
        return this.showAlert(msg, title || 'Alert', null, 'OK').css('z-index', 2000);
    }

    this.showEnterUrlPrompt = function(done, title, recent) {
        var $modal = showPrompt("Page url:", title || "Open page from url", null, "https://pinegrow.com", null, function(url) {
            if(url) {
                if(url.indexOf('://') < 0) {
                    url = 'http://' + url;
                }
                var a = document.createElement('a');
                a.href = url;
                url = a.href;
                a = null;

                done(url);
            }
        }, null, null, recent);
        return $modal;
    }

    this.showOpenUrlDialog = function(done, default_url) {
        var pro = pinegrow.isPRO() || pinegrow.isProTrialActive();

        var $modal = showPrompt("URL of the remote page or your local web server:", "Open URL", default_url || null, "https://pinegrow.com", null, function(url) {
            if(url) {
                if(url.indexOf('://') < 0) {
                    url = 'http://' + url;
                }
                var a = document.createElement('a');
                a.href = url;
                url = a.href;
                a = null;

                try {
                    var u = new URL(url);

                    if(!u.pathname.endsWith('/')) {
                        var paths = u.pathname.split('/');
                        if (paths[paths.length-1].indexOf('.') < 0) {
                            u.pathname += '/';
                        }
                    }
                    url = u.toString();

                    if(pro && !pinegrow.getCurrentProject()) {
                        pinegrow.showQuickError('Please select a folder for saving changed remote files.');
                        return false;
                    }

                    done(url);
                }
                catch (err) {
                }

            } else {
                pinegrow.showQuickError('Please enter a valid URL.');
                return false;
            }
        }, 'Cancel', 'Open URL');

        var $mb = $modal.find('.modal-body');
        var $info = $('<div></div>').appendTo($mb);

        //$mb.prepend(`<p></p>`);

        var updateInfo = function() {
            var project = pinegrow.getCurrentProject();
            if(pro) {
                var h = '';

                var info = `<p>That means that:</p><ul>
<li>Changed HTML documents and CSS stylesheets will be saved to the project folder.</li>
<li>Any such files that have been previously saved to this project will be loaded from the project.</li>
</ul>
<p><a href="${pgf.docs.remote_url_help}" class="external"><b>Check out this guide</b></a> to learn more about how you can edit remote projects and web applications with Pinegrow Web Editor.</p>`;

                if(!project) {
                    h += `<p>Select a folder/project that you wish to use for mapping changed remote files:</p><p><button class="pg-button pg-button-primary select-folder">Select a folder for saving changed remote files...</button></p>`;

                    h += `<p>The remote document is mapped to a local folder/project, so that you can easily save and reload edited HTML document and CSS stylesheets.</p>`;
                    h += info;


                } else {
                    h += `<p>This remote page <b>will be mapped to the current project</b> <code>${project.getDir()}</code>.</p>`;
                    h += info;
                }
                $info.html(h);

                $info.find('.select-folder').on('click', function(e) {
                    e.preventDefault();
                    pinegrow.openProjectBySelectingFolder(function() {
                        updateInfo();
                    })
                })

            } else {
                $info.html('<p>NOTE: <b>If you want to save changes</b> it is recommended to first save the page locally (for example, with &quot;Save as complete page&quot; in your browser) and then open it with <b>Open file</b>.</p>');
            }
        }

        updateInfo();

        return $modal;
    }

    this.showHelpImage = function(name) {
        return '<img class="help-image" src="images/help/' + name + '">';
    }

    this.showPrompt = function(notice, title, value, placeholder, onCancel, onOk, cancelLabel, okLabel, recent_values, validate) {
        showPrompt(notice, title, value, placeholder, onCancel, onOk, cancelLabel, okLabel, recent_values, validate);
    }

    this.showQuickDialog = function(title, section_title, msg, fields, cancelLabel, okLabel, onCancel, onOK, validate_on_submit) {
        var d = new PgDialog(title);

        d.values = {};
        d.okButtonLabel = okLabel || d.okButtonLabel;
        d.cancelButtonLabel = cancelLabel || d.cancelButtonLabel;
        d.validate_on_submit = validate_on_submit || false;

        d.before = msg;

        d.sections = {
            inputs: {
                name: section_title,
                is_closable: false,
                fields: {}
            }
        }

        var req = function(obj, prop, value, fieldDef, $field, values) {
            if(!value) {
                return pinegrow.getValidationError(fieldDef.name, 'req');
            }
            return null;
        }

        $.each(fields, function(key, def) {
            var i = $.extend({
                type: 'text'
            }, def)
            d.sections.inputs.fields[key] = i;

            if(i.required) {
                i.validate = req;
            }
            if(i.type === 'checkbox') {
                d.values[key] = def.default_value || null;
            } else {
                d.values[key] = def.value || null;
            }
        })

        d.onOK = function(async_close) {
            if(onOK) {
                return onOK(d.values, d, async_close);
            }
        }

        d.onCancel = function() {
            onCancel && onCancel();
        }

        d.show();

        return d;
    }

    var quickMessageRepeats = {};

    this.showQuickMessage = function(msg, duration, single, context, dont_repeat_delay) {
        if(dont_repeat_delay) {
            var t = (new Date()).getTime();
            if(quickMessageRepeats[msg] && t - quickMessageRepeats[msg] < dont_repeat_delay) {
                return;
            }
            quickMessageRepeats[msg] = t;
        }
        return crsaQuickMessage(msg, duration, single, context);
    }

    this.showQuickError = function(msg, single, duration) {
        return crsaQuickMessage(msg, duration || 4000, single, 'error');
    }

    this.showMode = function(mode, active) {
        pinegrow.showQuickMessage(mode + ': ' + (active ? 'ON' : 'OFF'))
    }

    this.showViewInDialog = function(title, msg, view) {
        var $d = pinegrow.showAlert(msg + '<div class="view-container"></div>', title, 'Close', null, function() {
            view.destroy();
        })

        $d.find('.view-container').append(view.get$Element());
        view.setAlwaysVisible(true);

        return $d;
    }

    this.showDialogMenu = function(title, intro, actions, no_close) {
        var html = ``;

        intro && (html += intro);

        html += `<div class="illustrated-menu"></div>`;

        var $d = this.showAlert(html, title, 'Close', null);

        $d.addClass('pg-illustrated-menu-dialog');

        this.showIllustratedMenu($d.find('.illustrated-menu'), actions, $d);

        if(no_close) {
            $d.find('.modal-footer,button.close').remove();
        }

        return $d;
    }

    this.showIllustratedMenu = function($d, actions, $dialog) {
        var html = ``;

        html += `<ul class="pg-dialog-menu-list">`;
        for(var i = 0; i < actions.length; i++) {
            html += `<li data-idx="${i}" class="${actions[i].class || ''}">`;

            html += pgSVGIcons.pg_logo;

            if(actions[i].url) {
                html += `<a href="${actions[i].url}" target="_blank" class="external">`;
            }
            html += `${actions[i].icon}<h4>${actions[i].label}</h4><p>${actions[i].desc}</p>`;
            if(actions[i].url) {
                html += `</a>`;
            }
            html += `</li>`;
        }
        html += `</ul>`;

        $d.html(html);

        $d.on('click', '.pg-dialog-menu-list li', function(e) {
            var a = actions[parseInt(this.getAttribute('data-idx'))];

            if(a.url) {
                if(isApp()) {
                    pinegrow.openExternalUrl(a.url);
                    e.preventDefault();
                    return;
                } else {
                    return;
                }
            };

            e.preventDefault();
            a.action(function(close) {
                if(close !== false) {
                    $dialog && $dialog.modal('hide');
                }
            });
        })
    }


    this.quitPinegrow = function() {
        var gui = require('nw.gui');
        var win = gui.Window.get();
        win.close();
    }

    this.restartPinegrow = function(ask) {

        if(ask) {

            var msg = `${typeof ask === 'string' ? ask : '<p>Please restart Pinegrow to continue.</p>'}`;
            var label = "Restart now";

            if(crsaHasChanges()) {
                msg += `<p>Please note that the editor <b>contains unsaved changes that will be lost</b> if you restart it now.</p>`;
                label = "Restart now and lose unsaved changes";
            }

            pinegrow.showAlert(msg, 'Restart is needed', "Later", label, function() {

            }, function() {
                pinegrow.restartPinegrow();
            })
        } else {
            //var nw = require('nw');
            var restart = function() {
                pinegrow.getAllPages().forEach(function (page) {
                    page.allowReload = true;
                })
                var win = nw.Window.get();
                win.reload();
            }

            if(pinegrow.devTools) {
                pinegrow.devTools.detach(restart);
            } else {
                restart();
            }
        }
    }

    this.showCreateCssRule = function(selector) {
        this.showUIPanel('style');
        pinegrow.activeCSSView.showCreateCssRule(selector);
    }

    var showOrHideLeftPanel = function(show) {
        var $leftPlane = $('#crsa-left-plane');
        var shown = !$leftPlane.hasClass('closed');
        var $leftHider = $leftPlane.find('i.hider');
        if(show != shown) $leftHider.trigger('click');
    }

    this.showLeftPanel = function() {
        showOrHideLeftPanel(true);
    }

    this.hideLeftPanel = function() {
        showOrHideLeftPanel(false);
    }


    this.evaluateUserVars = function(value, silent) {
        var invalid = [];
        if(!value) return value;
        var vars = null;
        var r = value.replace(/\$\{(.*)\}/g, function(match, varname) {
            if(vars === null) {
                vars = {};
                var str = pinegrow.getSetting('user-vars', '');
                str.split("\n").forEach(function(line) {
                    line = $.trim(line);
                    var a = line.split('=');
                    if(a.length == 2) {
                        vars[a[0].trim()] = a[1].trim();
                    }
                })
            }
            if(varname in vars) {
                return vars[varname];
            } else {
                invalid.push(varname);
            }
        })
        if(invalid.length && !silent) {
            pinegrow.showAlert(`<p>The following user variables are not defined: ${invalid.join(', ')}.</p><p><b>Use Support -&gt; Settings -&gt; Define user variables</b> to define these variables.</p>`, 'Missing user variables');
            return null;
        }
        return r;
    }

    this.openExternalUrl = function(url) {
        var gui = require('nw.gui');
        gui.Shell.openExternal(url);
    }

    this.addActionToElementForPreview = function(pgel, action_def, remove_other) {
        $.fn.crsa('addActionToElementForPreview', pgel, action_def, remove_other);
    }

    this.hasElementAction = function(pgel, action_def) {
        $.fn.crsa('hasElementAction', pgel, action_def);
    }

    this.showPreview = function($target, $content, cls, fixedX, code) {
        $.fn.crsa('showPreview', $target, $content, cls, fixedX, code);
    }

    this.hidePreview = function() {
        $.fn.crsa('hidePreview');
    }

    this.getValidationError = function(name, err) {
        switch(err) {
            case 'req':
                return name + ' is required.';
            default:
                return name + ' has invalid value';
        }
        return null;
    }

    this.validateField = function(pgel, fn, value, fdef, $field, values) {
        if(fdef && fdef.validate) {
            var err = fdef.validate(pgel, fn, value, fdef, $field, values);

            if($field) {
                var $msg = $field.data('error_msg');

                if (err) {
                    if(!$msg) {
                        $msg = $field.find('> p.error-message');
                        $field.data('error_msg', $msg);
                    }
                    if ($msg.get(0).innerHTML !== err || !$msg.pgIsVisible) {
                        fastdom.mutate(function() {
                            $msg.get(0).innerHTML = err;

                            if (!$msg.pgIsVisible) {
                                $field.addClass('error');
                                pg$Show($msg);
                                $msg.pgIsVisible = true;
                            }
                            if(!$field.hasClass('has-error')) $field.addClass('has-error');
                        })
                    }
                } else {
                    if($field.hasClass('has-error')) $field.removeClass('has-error');

                    if($msg && $msg.pgIsVisible) {
                        fastdom.mutate(function() {
                            $field.removeClass('error');
                            pg$Hide($msg);
                            $msg.pgIsVisible = false;
                            if($field.hasClass('has-error')) $field.removeClass('has-error');
                        })
                    }
                }
            }
            return err === null || err === undefined || err === '';
        } else {
            return true;
        }
    }

    this.validateAllFields = function($div, pgel, values) {
        $div.find('.crsa-field').each(function(i, f) {
            var $field = $(f);
            var fn = $field.data('crsa-field');
            pinegrow.validateField(pgel, fn, values.hasOwnProperty(fn) ? values[fn] : null, $field.data('crsa-field-def'), $field, values);
        });
    }

    this.getPgelFromSelector = function(pgel, sel, default_pgel, dom) {
        var r = this.getPgelFromSelectorMultiple(pgel, sel, true, dom);

        if(r.length) return r[0];

        /*
        if(sel) {
            if(sel === 'this') return pgel;
            var base = pgel;
            if(sel.startsWith('$')) {
                base = pgel.document;
                sel = sel.substr(1);
            }  else if(sel.startsWith('^')) {

                var sl = sel.substr(1).split('|');
                var p = pgel.closest(sl[0].trim());
                return (!p || sl.length === 1) ? p : this.getPgelFromSelector(p, sl[1].trim(), default_pgel);

            }
            return base.findOne(sel) || default_pgel;
        }
        */
        return default_pgel;
    }

    function processSelectorExtension(targets, sel_ext) {
        if(sel_ext) {
            var r = [];
            targets.forEach(function(target) {
                var el;

                switch(sel_ext) {
                    case 'previous':
                        el = target.prev() || target.parent.last();
                    case 'next':
                        el = target.next() || target.parent.first();
                        break;
                    case 'first':
                        el = target.parent.first();
                        break;
                    case 'last':
                        el = target.parent.last();
                        break;
                }

                if(r.indexOf(el) < 0) r.push(el);
            })
            return r;
        }
        return targets;
    }

    this.getPgelFromSelectorMultiple = function(pgel, sel, single, dom) {
        if(sel && pgel) {

            var sela = sel.split('->');
            var sel_ext = sela[1];
            sel = sela[0];

            if(sel === 'this') return processSelectorExtension([pgel]);

            var base = pgel;
            if(sel.startsWith('$')) {
                base = pgel.document;
                sel = sel.substr(1);
            } else if(sel.startsWith('^')) {

                var sl = sel.substr(1).split('|');
                var p = pgel.closest(sl[0].trim());
                if(!p) return [];
                return processSelectorExtension(sl.length === 1 ? [p] : this.getPgelFromSelectorMultiple(p, sl[1].trim(), single, dom));
            }

            if(dom) {
                var $base;

                var r = [];

                try {
                    if (base === pgel.document) {
                        $base = pgel.getPage().get$Html();
                    } else {
                        $base = base.get$DOMElement();
                    }

                    $base.find(sel).each(function(i, e) {
                        var n = getDomElementPgNode(e);
                        if(n) {
                            r.push(n);
                        }
                    })
                } catch(err) {

                }

            } else {
                var r = single ? base.findOne(sel) : base.find(sel);
                if (single) {
                    r = r ? [r] : [];
                }
            }
            return processSelectorExtension(r);
        }
        return [];
    }

    this.getSelectorFieldOptions = function(pgel, options, on_set, current_value) {

        options = $.extend({
            max_depth: 9999,
            multiple: false,
            parents: false
        }, options || {});

        var r = [
            {
                key: null,
                label: 'Not set',
                action: on_set,
                check: current_value === null || current_value === ''
            }
            //{
            //    type: 'divider'
            //}
        ]

        if(options.parents) {
            while (p.parent && p.parent.isElement) {
                p = p.parent;
                parents.push({
                    label: p.getName({short: true}),
                    pgel: p,
                    action: select,
                    on_mouseleave: on_mouseleave,
                    on_mouseenter: on_mouseenter
                })
            }

            if (parents.length) {
                parents.unshift({
                    type: 'header',
                    label: 'Parents'
                })
            }
        }

        var current_pgel = null;
        var current_pgels = [];

        if(current_value) {
            current_pgel = pinegrow.getPgelFromSelector(pgel, current_value);

            if(options.multiple) {
                current_pgels = pinegrow.getPgelFromSelectorMultiple(pgel, current_value);
            } else {
                current_pgels = [current_pgel];
            }
        }
        pgel.walkSelfAndChildren(function(node, depth) {
            if(depth > options.max_depth) return false;

            if(options.must_contain_or_be) {
                if(!node.isSelector(options.must_contain_or_be) && !node.findOne(options.must_contain_or_be)) return false;
            }

            if(!node.isElement) return true;

            if(options.filter && !options.filter(node)) return true;

            var name = node.getPlainTextName({short: false});
            var html_name = node.getName({short: false});

            var pg_name = node.getAttr('data-pg-name');
            if(pg_name) {
                name = pg_name + ' | ' + name;
                html_name = pg_name + ' | ' + html_name;
            }

            var sel = null;
            if(node === pgel) {
                sel = 'this'
            } else {
                var sels = node.getUniqueSelectors(0, 9999, pgel, true);
                if (sels.length) {
                    sel = sels[0];
                }
            }

            var submenu = null;

            if(options.multiple && sel !== 'this') {
                submenu = [];

                submenu.push({
                    type: 'header',
                    label: 'Select all elements with:'
                })
                /*
                submenu.push({
                    type: 'divider'
                })
                submenu.push({
                    type: 'header',
                    label: 'Select all elements with:'
                })*/
                submenu.push({
                    label: node.tagName,
                    key: node.tagName,
                    action: on_set,
                    check: node.tagName === current_value
                })
                node.getClasses().forEach(function(cls) {
                    if(cls !== 'pg-empty-placeholder') {
                        var sel = '.' + cls;

                        submenu.push({
                            label: sel,
                            key: sel,
                            action: on_set,
                            check: sel === current_value
                        })
                    }
                })
            }

            r.push({
                key: sel,
                //label: node === pgel ? 'current element' : name,
                label: node === pgel ? 'current element' : (`<span style="width:${(depth) * 15}px;display:inline-block;"></span>` + html_name),
                action: on_set,
                check: current_pgel === node  && current_pgels.length === 1,
                submenu: submenu
            })
            return true;
        })
        return r;
    }


    this.willMakeChange = function(page, $el, action) {
        willMakeChange(page.$iframe, $el ? (action + " / " + getElementName($el)) : action);
    }

    this.makeChanges = function(pgel_or_page, action, func) {
        func();
        if(pgel_or_page instanceof CrsaPage || pgel_or_page instanceof PgEmulatePageForUndo) {
            pinegrow.changeManager.didChangePage(pgel_or_page, action);
        } else {
            pinegrow.changeManager.didChange(pgel_or_page, action);
        }
    }

    //CS helpers
    this.getDefaultStylesheet = function(page, done, show_error, skip_load) {
        var url = pgcssh.getUrlOfStylesheetForOverrides(page);
        var detected = false;
        var cs;
        if(url) {
            cs = pinegrow.stylesheets.getByUrl(url);
            if(!cs) {
                url = null;
            }
        }
        if(!url) {
            //try to detect, take the last one
            page = page || pinegrow.getSelectedPage();
            if(page) {
                var list = page.getAttachedStylesheets(
                    {
                        loaded: true,
                        ignored: false,
                        imported: false,
                        inline: false,
                        remote: false
                    })
                if(list.length) {
                    url = list[list.length-1].url;
                    detected = true;
                }
            }
        }
        if(url) {
            cs = cs || pinegrow.stylesheets.getByUrl(url);
            if(cs) {
                if(cs.loaded || skip_load) {
                    done(cs, detected);
                } else {
                    cs.ignore = false;
                    cs.load(cs.url, function() {
                        done(cs, detected);
                    }, true /* force loading ignored cs */);
                }
            } else {
                if(show_error) {
                    pinegrow.showAlert('Default stylesheet with ' + csurl + ' not found or not loaded!', 'Error');
                }
                done(null);
            }
        } else {
            done(null);
        }
        return cs;
    }

    this.setDefaultStylesheet = function(page, cs) {
        if(cs) {
            pgcssh.setUrlOfStylesheetForOverrides(cs.url, page);
            pinegrow.dispatchEvent('on_default_stylesheet_selected', page, cs);
        }
    }

    //PLUGIN CONTROLS

    var plugin_controls = [];

    this.addPluginControlToTopbar = function(f, $control, show_always, func) {
        var $c = $('.main-menu .insert-custom-here');
        $control.insertBefore($c);
        plugin_controls.push({framework: f, control: $control, show_always: show_always, func: func});
        var func_show = func ? func() : false;
        if(!show_always && !func_show) $control.hide();
    }

    this.updatePluginControlsForSelectedPage = function (cp) {
        showPluginControlsForSelectedPage(cp);
    }

    var showPluginControlsForSelectedPage = function(cp) {
        var changed = false;
        for(var i = 0; i < plugin_controls.length; i++) {
            if(plugin_controls[i].show_always) continue;

            var func_show = plugin_controls[i].func ? plugin_controls[i].func() : false;

            if(plugin_controls[i].func) {
                if(plugin_controls[i].func()) {
                    plugin_controls[i].control.css('display', '');
                } else {
                    plugin_controls[i].control.css('display', 'none');
                }
            } else {

                if ((!cp || !cp.hasFramework(plugin_controls[i].framework))) {
                    plugin_controls[i].control.css('display', 'none');
                } else {
                    plugin_controls[i].control.css('display', '');
                }
            }
            changed = true;
        }
        if(changed) {
            arrangeTopBarSizes();
        }
    }

    $('body').on('crsa-page-selected crsa-frameworks-changed', function(e, crsaPage) {
        if(!crsaPage) crsaPage = pinegrow.getSelectedPage();
        showPluginControlsForSelectedPage(crsaPage);
    });

    //END PLUGIN CONTROLS





    this.showBuyProductScreen = function(p) {
        $.fn.crsa('showIntroScreen', p);
    }

    //Screenshots
    this.takePhotoOfUI = function (done, returnCanvasData, win) {
        var gui = require('nw.gui');
        win = win || gui.Window.get();

        //wait for menu to close
        setTimeout(function() {
            var width = $(window).width();
            var height = $(window).height();

            win.capturePage(function (datauri) {
                 try {
                    var img = new window.Image();
                    var wWidth = $(window).width();
                    var wHeight = $(window).height();

                    var canvas = document.createElement('canvas');
                    canvas.setAttribute('width', wWidth);
                    canvas.setAttribute('height', wHeight);

                    var ctx = canvas.getContext('2d');

                    img.addEventListener("load", function () {
                        ctx.drawImage(img, 0, 0, wWidth, wHeight);

                        var canvasData = null;
                        if (returnCanvasData) {
                            canvasData = ctx.getImageData(0, 0, width, height).data;
                        }

                        if (done) done(canvas, canvasData, datauri);
                    }, false);

                    img.src = datauri;

                } catch(err) {}
            }, { format : 'png', datatype : 'datauri'} );
        }, 100);
    }


    this.takePhotoOfPage = function (page, filename, done, img_width, returnCanvasData) {
        if (!page.$page) {
            if(done) done(null);
            return;
        }

        var gui = require('nw.gui');
        var win = gui.Window.get();

        var $page = page.$page;
        var $content = $page.find('>.content');
        var $iframe = $content.find('> iframe, > webview');

        var content_offset = $content.offset();

        $iframe.css('z-index', 10);

        var ui_zoom = window.innerWidth / document.body.clientWidth;

        page.removeCrsaStyles();
        var $body = $('body');

        $body.addClass('hide-quick-messages')

        //wait for menu to close
        setTimeout(function() {

            var left = content_offset.left * ui_zoom;
            var top = content_offset.top * ui_zoom;
            var width = $content.width() * ui_zoom;
            var height = $content.height() * ui_zoom;

            var scale = 1.0;

            if(img_width && img_width < width) {
                scale = 1.0 * img_width / width;
            }

            win.capturePage(function(datauri){

                try {
                    $body.removeClass('hide-quick-messages');

                    var getImageAfterResizing = function (img, done) {
                        var wWidth = $(window).width();
                        var wHeight = $(window).height();

                        if (wWidth < img.width) {
                            var resizeCanvas = document.createElement('canvas');
                            resizeCanvas.setAttribute('width', wWidth);
                            resizeCanvas.setAttribute('height', wHeight);

                            var resizeCtx = resizeCanvas.getContext('2d');
                            resizeCtx.drawImage(img, 0, 0, img.width, img.height, 0, 0, wWidth, wHeight);

                            var resizedImg = new window.Image();
                            resizedImg.addEventListener("load", function () {
                                if (done) done(resizedImg);
                            }, false);
                            resizedImg.src = resizeCanvas.toDataURL("image/png");
                        }
                        else {
                            if (done) done(img);
                        }
                    }

                    page.addCrsaStyles();

                    var canvas = document.createElement('canvas');
                    canvas.setAttribute('width', width);
                    canvas.setAttribute('height', height);
                    var ctx = canvas.getContext('2d');
                    var img = new window.Image();
                    img.addEventListener("load", function () {
                        getImageAfterResizing(img, function (newImg) {
                            ctx.drawImage(newImg, left, top, width, height, 0, 0, width, height);

                            var smallCanvas = scale < 1.0 ? downScaleCanvas(canvas, scale) : canvas;

                            if (filename) {
                                crsaCreateDirs(crsaGetDir(filename));
                                var image = smallCanvas.toDataURL('image/png');
                                image = image.replace('data:image/png;base64,', '');
                                var buffer = new Buffer(image, 'base64');
                                require('fs').writeFileSync(filename, buffer);
                            }

                            var canvasData = null;
                            if (returnCanvasData) {
                                canvasData = ctx.getImageData(0, 0, width, height).data;
                            }

                            if (done) done(smallCanvas, canvasData);
                        });
                        //resizeImage(big_file, parseInt(w/1), parseInt(h/1), big_file);
                        //var $d = $('<div style="position:fixed;top:0;left:0;z-index:10000;"></div>');
                        //$d.append(smallCanvas);
                        //$('body').append($d);

                    }, false);
                    img.src = datauri;

                    //var big_file = f.getResourceFile('images/' + list[current].type + '.png');
                    //fs.writeFileSync( big_file, buffer);
                } catch(err) {
                    if(done) done(null)
                }

                $iframe.css('z-index', '');

            }, { format : 'png', datatype : 'datauri'} );

        }, 100);
    }

    this.takePhotoOfElement = function($el, max_width, filename, done) {
        var gui = require('nw.gui');
        var win = gui.Window.get();

        var $iframe = getIframeOfElement($el);
        var $page = $iframe.closest('.page');

        $iframe.css('z-index', 10);

        crsaScrollToElementInIframe($iframe, $el);

        var pos = getElementPositionOnScreen($el, $iframe);

        var crsaPage = getCrsaPageForIframe($iframe);

        var ui_zoom = window.innerWidth / document.body.clientWidth;
        //var ui_zoom = 1;

        //console.log('ui zoom ' + ui_zoom, window)

        crsaPage.removeCrsaStyles();

        //wait for menu to close
        setTimeout(function() {

            var left = pos.left * ui_zoom;
            var top = pos.top * ui_zoom;
            var width = pos.width * ui_zoom;
            var height = pos.height * ui_zoom;

            var scale = 1.0;
            if(max_width && width > max_width) scale = 1.0 * max_width / width;

            win.capturePage(function(datauri){

                try {

                    var getImageAfterResizing = function (img, done) {
                        var wWidth = $(window).width();
                        var wHeight = $(window).height();

                        if (wWidth < img.width) {
                            var resizeCanvas = document.createElement('canvas');

                            resizeCanvas.setAttribute('width', wWidth);
                            resizeCanvas.setAttribute('height', wHeight);

                            var resizeCtx = resizeCanvas.getContext('2d');
                            resizeCtx.drawImage(img, 0, 0, img.width, img.height, 0, 0, wWidth, wHeight);

                            var resizedImg = new window.Image();
                            resizedImg.addEventListener("load", function () {
                                if (done) done(resizedImg);
                            }, false);
                            resizedImg.src = resizeCanvas.toDataURL("image/png");

                            //pgDebugCanvas(resizeCanvas);
                            //pgDebugPos(pos)
                        }
                        else {
                            if (done) done(img);
                        }
                    }

                    crsaPage.addCrsaStyles();

                    var canvas = document.createElement('canvas');
                    canvas.setAttribute('width', width);
                    canvas.setAttribute('height', height);
                    var ctx = canvas.getContext('2d');
                    var img = new window.Image();
                    img.addEventListener("load", function () {
                        getImageAfterResizing(img, function (newImg) {
                            ctx.drawImage(newImg, left, top, width, height, 0, 0, width, height);

                            //pgDebugCanvas(canvas);
                            console.log('bubi ' + left + ' ' + top);

                            var smallCanvas = scale < 1.0 ? downScaleCanvas(canvas, scale) : canvas;

                            if (filename) {
                                var image = smallCanvas.toDataURL('image/png');
                                image = image.replace('data:image/png;base64,', '');
                                var buffer = new Buffer(image, 'base64');
                                require('fs').writeFileSync(filename, buffer);
                            }

                            if (done) done(smallCanvas);
                        });
                        //resizeImage(big_file, parseInt(w/1), parseInt(h/1), big_file);
                        //var $d = $('<div style="position:fixed;top:0;left:0;z-index:10000;"></div>');
                        //$d.append(smallCanvas);
                        //$('body').append($d);

                    }, false);
                    img.src = datauri;

                    //var big_file = f.getResourceFile('images/' + list[current].type + '.png');
                    //fs.writeFileSync( big_file, buffer);
                } catch(err) {}

                $iframe.css('z-index', '');

            }, { format : 'png', datatype : 'datauri'} );

        }, 250);
    }

    var current_shown_workspace_theme = null;
    var quick_switch_workspace_theme_map = null;

    this.setWorkspaceTheme = function(key, just_show) {
        if(pgf.kids) return;

        if(pgf.product === 'WPPLUGIN') {

            if(pinegrow.startScreenManager && pinegrow.startScreenManager.isVisible()) {
                key = 'light';
                just_show = true;
            }

            if (just_show && current_shown_workspace_theme === key) return;

            var withSheet = function (link, func) {
                if (link.sheet) {
                    func(link.sheet)
                } else {
                    $(link).on('load', function () {
                        link.sheet && func(link.sheet);
                    })
                }
            }

            if (!quick_switch_workspace_theme_map) {

                var ver = '1';
                if(pgwpdata.version) {
                    ver = pgwpdata.version;
                }
                //we will handle this in a special way, load both and switch by enabling the stylesheet
                var $head = $('head');
                var $light = $('<link href="lib/crsa/edit.theme.light.css?v=' + ver + '" rel="stylesheet" class="pg-light-theme">');
                $light.appendTo($head);

                var $dark = $('<link href="lib/crsa/edit.theme.gray.css?v=' + ver + '" rel="stylesheet" class="pg-gray-theme">');
                $dark.appendTo($head);

                quick_switch_workspace_theme_map = {
                    light: $light.get(0),
                    gray: $dark.get(0)
                }
            }

            var map = quick_switch_workspace_theme_map;

            withSheet(map.light, function(sheet) {
                sheet.disabled = key !== 'light';
            })
            withSheet(map.gray, function(sheet) {
                sheet.disabled = key === 'light';
            })

            var $splash = $('.pg-splash-screen');
            if($splash.length) {
                var $sensor = $('.pg-splash-screen-sensor');
                if ($sensor.length) {
                    var splash_timer = setInterval(function () {
                        if ($sensor.is(':visible')) {
                            clearInterval(splash_timer);
                            $splash.css('opacity', 0);
                            setTimeout(function () {
                                arrangeTopBarSizes();
                                $splash.remove();
                            }, 300);
                        }
                    }, 200)
                } else {
                    $('.pg-splash-screen').remove();
                }
            }
            if(!just_show) {
                this.setSetting('workspace-theme', key);
            }
            return;
        }

        this.setSetting('workspace-theme', key);

        pgGetAppWindows().forEach(function(win) {
            _this.activateCurrentWorkspaceThemeInWindow(win);
        })

        if(pgf.product !== 'WPPLUGIN') {
            pinegrow.updateStartScreen();
        }
    }

    this.activateCurrentWorkspaceThemeInWindow = function(win) {
        var key = this.getWorkspaceTheme();
        var elid = 'workspacetheme';
        var url = 'lib/crsa/edit.theme.' + key + '.css';

        var $head = $(win.window.document).find('body');
        var $existing = $head.find('.' + elid);

        if($existing.attr('href') !== url) {

            var $new = $('<link href="' + url + '" rel="stylesheet" class="' + elid + '">');
            $new.appendTo($head);

            var $splash = $('.pg-splash-screen');
            var $sensor = $('.pg-splash-screen-sensor');
            if($sensor.length) {
                var splash_timer = setInterval(function () {
                    if($sensor.is(':visible')) {
                        clearInterval(splash_timer);
                        $splash.css('opacity', 0);
                        setTimeout(function() {
                            arrangeTopBarSizes();
                            $splash.remove();
                            }, 300);
                    }
                }, 200)
            } else {
                $('.pg-splash-screen').remove();
            }

            setTimeout(function() {
                $existing.remove();
                $(win.window).trigger('resize');
            }, 2000)
        } else {
            $('.pg-splash-screen').remove();
        }
    }

    this.getWorkspaceTheme = function() {
        var t = this.getSetting('workspace-theme', pgf.default_theme || 'gray');
        if(t === 'high.contrast') t = 'gray';
        return t;
    }


    var pgDisplayHelpersOn = false;

    this.getShowDisplayHelpers = function() {
        return pgDisplayHelpersOn;
    }

    this.setShowDisplayHelpers = function(v, msg) {
        pgDisplayHelpersOn = v;
        pinegrow.dispatchEvent('on_show_display_helpers_changed', null, v);
        pinegrow.helpersToolbarItem.toolbarItem.setActive(v);
        pinegrow.selectedElements.reselect();
        if(msg) {
            pinegrow.showMode('Show display helpers', v);
        }
    }


    //Photo browser
    this.showPhotoBrowser = function($target) {
        pgShowQuickImageBrowser($target);
    }

    this.getPhotoBrowserSearchTerm = function() {
        return pgQuickImageBrowserSearchTerm;
    }

    this.getMappedRemoteFileForUrl = function(url, type) {
        //Remove base
        if(!isApp() || !crsaIsRemoteUrl(url)) return null;

        var project = pinegrow.getCurrentProject();
        if(project) {
            //type = type || 'html';
            var def_name = 'data.bin';

            switch(type) {
                case 'html':
                    def_name = 'index_default.html';
                    break;
                case 'css':
                    def_name = 'style.css'
                    break;
            }

            try {
                var u = new URL(url);

                var pn = u.pathname;
                if(pn === '' || pn === '/') {
                    pn = def_name;
                } else if(pn.endsWith('/')) {
                    pn += def_name;
                }
                var fn = path.join(project.getDir(), 'mapped_urls', u.hostname, decodeURI(pn));
                return fn;
            }
            catch (err) {
            }
        }
        return null;
    }

    this.getRemoteUrlFromMappedUrl = function(url) {
        var project = pinegrow.getCurrentProject();
        if(project) {
            var base = project.getUrl() + '/mapped_urls/';

            if(url.startsWith(base)) {
                url = url.replace(base, 'http://');
                url = url.replace('index_default.html', '');
                return url;
            }
        }
        return null;
    }

    this.getMappedRemoteBaseDirForUrl = function(url) {
        var project = pinegrow.getCurrentProject();
        if(project) {
            try {
                var u = new URL(url);
                return path.join(project.getDir(), 'mapped_urls', u.hostname);
            }
            catch (err) {
            }
        }
        return null;
    }

    //Current device size
    var current_device_size = null;
    var sync_device_size_with_page_view = this.getSetting('prop-device-size-sync', '0') === '1';

    this.getSyncCurrentDeviceSizeWithPageView = function() {
        return sync_device_size_with_page_view;
    }

    this.setSyncCurrentDeviceSizeWithPageView = function(val) {
        sync_device_size_with_page_view = val;
        this.setSetting('prop-device-size-sync', val ? '1' : '0');
    }

    this.setCurrentDeviceSize = function(def, page, pv, force) {
        if(force || (page === this.getSelectedPage() && pv === page.activeView)) {
            current_device_size = def;
            this.dispatchEvent('on_current_device_size_changed', page, def);
        }
    }

    this.getCurrentDeviceSize = function() {
        return current_device_size;
    }
    //End current device size

    this.setPseudoStatesForCollection = function(col, states, done) {

        var cssStates = ['hover','active','focus','visited'];

        var list = [];

        states.forEach(function(s) {
            if(cssStates.indexOf(s) >= 0) {
                list.push(s)
            }
        })

        var updateStateClasses = function(pgel) {

            pgel.get$DOMElements().forEach(function($el) {
                cssStates.forEach(function(state) {
                    var cls = `pg-state-${state}`;
                    if(list.indexOf(state) >= 0) {
                        $el.addClass(cls);
                    } else {
                        $el.removeClass(cls);
                    }
                })
            })
        }

        col.forEachElementAsync(function (pgel, ondone) {
            pinegrow.devTools.forcePseudoStates(pgel, list, ondone);
            updateStateClasses(pgel);
        }, function (results) {
            pinegrow.dispatchEvent('on_current_pseudo_state_changed', pinegrow.getSelectedPage(), states);
            done && done();
        })
    }

    this.getCurrentClassStyle = function(for_pgel) {
        return this.classStyles ? this.classStyles.getCurrentStyle() : null;
    }

    this.requireProject = function(func) {

        if(!pinegrow.isPRO() && !pinegrow.isProTrialActive()) {
            pinegrow.showProductUpgradeMessage('PRO', 'Pinegrow PRO', 'Your edition of Pinegrow does not include Pinegrow PRO features. Please upgrade to the PRO edition to continue using projects, master pages and components.')
            return;
        }

        var p = pinegrow.getCurrentProject();
        if(p) {
            func(p);
        } else {
            pinegrow.showAlert('<p>This feature works with <b>Projects</b>. To use it, open the folder where your pages are located with <b>File -&gt; Open project</b>.</p>', 'Project is required');
        }
    }

    this.requireProjectDB = function(func) {
        var db = null;

        this.requireProject(function(p) {
            db = p.requireDB(function(db) {
                func(db, p)
            });
        })
        return db;
    }
}

var PgFramework = function(key, name) {
    this.key = key;
    this.name = name;
    this.show_in_action_tab = "ACT";
    this.detect = null;
    this.component_types = {};
    this.default = false;
    this.lib_sections = [];
    this.actions_sections = [];
    this.ignore_css_files = [];
    this.type = key; //default
    this.allow_single_type = false;
    this.pluginUrl = null;
    this.on_get_source = null; //function crsaPage
    this.user_lib = false;
    this.changed = false;
    this.not_main_types = false;
    this.product = null;
    this.trial = false;
    this.trial_start_message = '7 day trial was started.';
    this.trial_expired_message = 'The trial expired. Please purchase the product to continue using it.';
    this.common_sections = {};
    this.script_url = null;
    this.description = null;
    this.author = null;
    this.author_link = null;
    this.has_actions = false;
    this.info_badge = null;
    this.auto_updatable = false;
    this.order = 0;
    this.page_libs = [];
    this.quick_actions = [];

    this.show_in_manager = true;
    this.preview_images_base = 'images';

    this.resources = new PgResourcesList();
    this.resources_name = null;

    this.customResourceNamespace = null;

    var ordered_list = [];
    var ordered = false;
    var types_map = null;

    var defs_that_filter_tree = [];

    this.has_types_that_filter_tree = false;

    this.componentFiltersTree = function(def) {
        defs_that_filter_tree.push(def);
        this.has_types_that_filter_tree = true;
    }

    this.getTypeThatFiltersTree = function(pgel) {
        for(var i = 0; i < defs_that_filter_tree.length; i++) {
            if(isType(defs_that_filter_tree[i], null, pgel, true)) return defs_that_filter_tree[i];
        }
        return null;
    }

    var _this = this;

    this.addPageLib = function(item) {
        this.page_libs.push(item);
    }

    this.getPageLibs = function() {
        return this.page_libs;
    }

    this.addQuickAction = function(a) {
        this.quick_actions.push(a);
    }

    this.getQuickActions = function() {
        return this.quick_actions;
    }

    this.removeAllComponents = function() {
        this.component_types = {};
        this.lib_sections = [];
        this.actions_sections = [];
        ordered = false;
        ordered_list = [];
    }

    this.isTrialActive = function(no_msg) {
        if(pinegrow.isTrial()) return true;//just use global trial

        if(!no_msg) {
            pinegrow.showProductUpgradeMessage(this.product, this.name, this.trial_expired_message);
        }
        return false;
    }

    this.addComponentType = function(def) {
        if(def.selector && typeof def.selector == 'string') {
            if(def.selector.match(/^[a-z]+$/i)) {
                def.selector_tag = def.selector.toUpperCase();
            }
        }
        this.component_types[def.type] = def;
        if(!def.priority) def.priority = 1000;
        def.framework = this;

        if(this.common_sections) {
            if(!def.sections) {
                def.sections = {};
            }
            $.each(this.common_sections, function(key, sdef) {
                def.sections[key] = sdef;
            });
        }
        ordered_list.push(def);
        ordered = false;
    }

    this.removeComponentType = function(def) {
        if(def.type in this.component_types) {
            delete this.component_types[def.type];
            var idx = ordered_list.indexOf(def);
            if(idx >= 0) {
                ordered_list.splice(idx, 1);
            }
            ordered = false;

            for(var n = 0; n < this.lib_sections.length; n++) {
                var s = this.lib_sections[n];
                var idx = s.components.indexOf(def);
                if(idx >= 0) {
                    s.components.splice(idx, 1);
                }
            }
            for(var n = 0; n < this.actions_sections.length; n++) {
                var s = this.actions_sections[n];
                var idx = s.components.indexOf(def);
                if(idx >= 0) {
                    s.components.splice(idx, 1);
                }
            }

        }
    }

    this.replaceComponent = function(old_c, new_c) {
        this.component_types[old_c.type] = new_c;
        var idx = ordered_list.indexOf(old_c);
        if(idx >= 0) {
            ordered_list[idx] = new_c;
        }
        for(var n = 0; n < this.lib_sections.length; n++) {
            var s = this.lib_sections[n];
            var idx = s.components.indexOf(old_c);
            if(idx >= 0) {
                s.components[idx] = new_c;
            }
        }
        for(var n = 0; n < this.actions_sections.length; n++) {
            var s = this.actions_sections[n];
            var idx = s.components.indexOf(old_c);
            if(idx >= 0) {
                s.components[idx] = new_c;
            }
        }
        new_c.framework = this;
    }

    //todo30 - types
    var isType = function(def, $el, pgel, skip_actions) {
        if(!pgel) {
            debugger;
        }
        var isType = false;
        if(skip_actions && def.action) return false;
        if(def.selector_new) {
            isType = def.selector_new(pgel);
        } else if(typeof def.selector === 'function') {
            isType = def.selector(pgel);
        } else if(def.selector) {
            isType = pgel.isSelector(def.selector);
        } else if(def.selector_tag) {
            isType = pgel.tagName === def.selector_tag;
        }
        return isType;
    }

    this.isType = function(def, $el, pgel, skip_actions) {
        return isType(def, $el, pgel, skip_actions);
    }

    var orderTypesIfNeeded = function() {
        if(!ordered) {
            ordered_list.sort(function(a,b) {
                return a.priority - b.priority;
            });
            ordered = true;

            //return;

            var tags = 0;
            var selectors = 0;
            var funcs = 0;
            var match = 0;
            var no_sel = 0;

            var map = {
                tags: {},
                check_all: []
            };

            types_map = map;

            var re = /^([a-z0-9]*)?(\.[a-z0-9\-\_]*)?(\[[a-z\-]*\])?(\[[a-z\-]*="(.*)"\])?$/i;

            for(var i = 0; i < ordered_list.length; i++) {
                if(ordered_list[i].selector || ordered_list[i].selector_new) {
                    var info = {index: i, def: ordered_list[i]};

                    if(_this.no_type_maps || typeof ordered_list[i].selector == 'function' || ordered_list[i].selector_new ) {
                        map.check_all.push(info);
                        funcs++;
                    } else {

                        var m = ordered_list[i].selector.match(re);
                        if(m) {
                            var tag = m[1];
                            if(tag) {
                                if(!map.tags[tag]) map.tags[tag] = [];
                                map.tags[tag].push(info);
                            } else {
                                map.check_all.push(info);
                            }
                        } else {
                            map.check_all.push(info);
                        }
                        /*
                        var m = true;
                        var ss = ordered_list[i].selector.split(',');
                        var matches = [];
                        for(var n = 0; n < ss.length; n++) {
                            var match = ss[n].trim().match(re);
                            if(!match) {
                                m = false;
                                break;
                            } else {
                                matches.push(match);
                            }
                        }
                        if(m) {
                            match++;
                            var branch = map;
                            for(var n = 0; n < matches.length; n++) {
                                var parts = matches[n];
                                var tag = parts[1];
                                if(tag) {
                                    if(!map.tags[tag]) map.tags[tag] = [];
                                    branch = map.tags[tag];
                                } else {
                                    branch = map.no_tag;
                                }

                            }
                        } else {
                            console.log(ordered_list[i].selector);
                        }*/
                    }
                } else {
                    no_sel++;
                }
            }
            if(ordered_list.length) {
                //console.log(types_map);
                //console.log('Selector stats for ' + _this.key + ': match=' + match + ' no_sel=' + no_sel + ' tags=' + tags + ' sels=' + selectors + ' funcs=' + funcs + ' total=' + ordered_list.length);
            }
        }

    }

    this.getType = function($el, pgel, skip_actions) {
        orderTypesIfNeeded();

        var types = this.getTypes($el, pgel, skip_actions, true);
        for(var i = 0; i < types.length; i++) {
            if(!types[i].not_main_type) {
                return types[i];
            }
        }
        return null;
        /*
        return types.length ? types[0] : null;


        for(var i = 0; i < ordered_list.length; i++) {
            var def = ordered_list[i];
            if(def.not_main_type) continue;
            if(isType(def, $el, pgel, skip_actions)) return def;
        }
        return null;
        */
    }

    //todo30 - types
    this.getTypes = function($el, pgel, skip_actions, single) {
        orderTypesIfNeeded();

        if(single === undefined) single = false;
        var r = [];
        var info_list = [];
        var tag = pgel ? pgel.tagName : $el[0].tagName.toLowerCase();
        if(tag && types_map.tags[tag]) {
            for(var i = 0; i < types_map.tags[tag].length; i++) {
                if(isType(types_map.tags[tag][i].def, $el, pgel, skip_actions)) {
                    info_list.push(types_map.tags[tag][i]);
                    if(single) break;
                }
            }
        }
        for(var i = 0; i < types_map.check_all.length; i++) {
            if(isType(types_map.check_all[i].def, $el, pgel, skip_actions)) {
                info_list.push(types_map.check_all[i]);
                if(single) break;
            }
        }
        info_list.sort(function(a,b) {
            return a.index - b.index;
        });
        for(var i = 0; i < info_list.length; i++) {
            r.push(info_list[i].def);
            if(info_list[i].def.last_type) break;
        }
        return r;
/*

        var r = [];

        for(var i = 0; i < ordered_list.length; i++) {
            var def = ordered_list[i];
            if( isType(def, $el, pgel, skip_actions)) {
                r.push(def);
                if(def.last_type) break;
                //return def;
            }
        }
        return r;

 */
    }

    this.getComponentTypes = function() {
        return this.component_types;
    }

    this.getComponentType = function(type) {
        return this.component_types[type] || null;
    }

    this.addLibSection = function(section) {
        this.lib_sections.push(section);
        section.framework = this;
    }

    this.getLibSections = function() {
        return this.lib_sections;
    }

    this.getLibSection = function(key) {
        for(var s in this.lib_sections) {
            if(this.lib_sections[s].key == key) return this.lib_sections[s];
        }
        return null;
    }

    this.addActionsSection = function(section) {
        this.actions_sections.push(section);
        section.framework = this;
    }

    this.getActionsSections = function() {
        return this.actions_sections;
    }

    this.getAutoId = function(prefix) {
        var c = 0;
        var t;
        do {
            c++;
            t = prefix + (c > 0 ? c : '');
        }
        while((t in this.component_types));

        return {count: c, type: t};
    }

    this.getFileName = function() {
        if(this.pluginUrl) {
            return getPageName(this.pluginUrl);
        } else {
            return this.name.replace(/\s/ig,'') + '.js';
        }
    }

    this.getImagePreviewBaseUrl = function() {
        return this.getBaseUrl() + '/' + this.preview_images_base + '/';
    }

    this.getBaseUrl = function() {
        if(this.project) return crsaMakeUrlFromFile(this.project.getDir());
        return crsaGetBaseForUrl(this.pluginUrl ? this.pluginUrl : this.script_url);
    }

    this.getResourceUrl = function(relative_path) {
        return this.getBaseUrl() + '/' + relative_path;
    }

    this.getResourceFile = function(relpath) {
        return crsaMakeFileFromUrl(this.getResourceUrl(relpath)) ;
    }

    this.setScriptFileByScriptTagId = function(id, def_value) {
        var script = $('#' + id);
        if(script.length === 0 && !isApp() && def_value) {
            //search with src
            script = $(`script[src*="${def_value}"]`);
        }
        if(script.length) {
            if(isApp()) {
                this.script_url = pgMakeFileUrl(script.get(0).src); //get url if script is included directly into edit.html
            } else {
                this.script_url = script.get(0).src;
            }
        } else if(def_value && isApp()) {
            if(pgIsLive()) {
                //reference the loaded url not the packaged app
                this.script_url = def_value;
            } else {
                var path = require('path');
                this.script_url = crsaMakeUrlFromFile(path.join(process.cwd(), def_value));
            }
        }
    }

    /*
    this.addTemplateProjectFromResourceFolder = function(folder, done, index, filter_func, skip_add_to_templates, absolute_folder) {
        var path = require('path');
        var base_url = this.getBaseUrl();
        if(!base_url && !absolute_folder) return;
        var full_path = absolute_folder ? folder : path.join(crsaMakeFileFromUrl(base_url), folder);
        pinegrow.addTemplateProjectFromFolder(full_path, function(p) {
            p.name = _this.name;
            p.description = _this.description;
            p.author = _this.author;
            p.author_link = _this.author_link;
            p.info_badge = _this.info_badge || null;
            p.framework = _this;

            p.root.walkThroughFiles(function(node) {
                if(node.name == 'screenshots') {
                    return false;
                } else if(node.name.indexOf('.html') >= 0) {
                    node.tag = 'page';
                    node.image = p.root.url + '/screenshots' + '/' + node.name.replace('html', 'jpg');
                } else {
                    node.required = true;
                }
                if(filter_func) return filter_func(node, p);
            })

            if(done) {
                done(p);
            }
        }, index, skip_add_to_templates);
    }
*/
    this.addTemplateProjectFromResourceFolder = function(folder, done, index, filter_func, skip_add_to_templates, absolute_folder) {

        if(!isApp()) return;

        var path = require('path');
        var base_url = this.getBaseUrl();
        if(!base_url && !absolute_folder) return;
        var full_path = absolute_folder ? folder : path.join(crsaMakeFileFromUrl(base_url), folder);
        pinegrow.addTemplateProjectFromFolder(full_path, function(p) {
            p.name = _this.name;
            p.description = _this.description;
            p.author = _this.author;
            p.author_link = _this.author_link;
            p.info_badge = _this.info_badge || null;
            p.framework = _this;

            p.root.walkThroughFiles(function(node) {
                if(node.name == 'screenshots') {
                    return false;
                } else if(node.name.indexOf('.html') >= 0) {
                    node.tag = 'page';
                    node.image = p.root.url + '/screenshots' + '/' + node.name.replace('html', 'jpg');
                } else {
                    node.required = true;
                }
                if(filter_func) return filter_func(node, p);
            })

            if(done) {
                done(p);
            }
        }, index, skip_add_to_templates);
    }

    this.addTemplateProjectFromDefinition = function(def) {

        var p = pinegrow.addTemplateProjectFromTemplateDefinition(def);

        p.name = _this.name;
        p.description = _this.description;
        p.author = _this.author;
        p.author_link = _this.author_link;
        p.info_badge = _this.info_badge || null;
        p.framework = _this;
    }

    this.save = function(url, done) {

        url = url || this.pluginUrl;

        var _this = this;

        var saveAs = function() {
            crsaChooseFile(function(url, file) {
                if(file) {
                    _this.save(file, done);
                }
            }, _this.getFileName());
        }

        if(!url) {
            saveAs();
            return;
        }

        if(url != this.pluginUrl) {
            this.name = getPageName(url).replace(/\.js$/i,'');
            this.key = this.name;
        }
        var s = '';
        var comp_list = [];

        $.each(this.component_types, function(i, def) {
            var type = def.type;
            var vn = "comp_" + type.replace(/-/g,'_');
            s += def.toJSCode(vn);
            comp_list.push(vn);
        });

        var source = '\
$(function() {\n\
\n\
    //Wait for Pinegrow to wake-up\n\
    $("body").one("pinegrow-ready", function(e, pinegrow) {\n\
\n\
        //Create new Pinegrow framework object\n\
        var f = new PgFramework("' + this.key + '", "' + this.name + '");\n\
\n\
        //This will prevent activating multiple versions of this framework being loaded\n\
        f.type = "' + this.key + '";\n\
        f.allow_single_type = true;\n\
        f.user_lib = ' + (this.user_lib ? 'true' : 'false') + '\
\n\
' + s + '\n\
        //Tell Pinegrow about the framework\n\
        pinegrow.addFramework(f);\n\
            \n\
        var section = new PgFrameworkLibSection("' + this.key + '_lib", "Components");\n\
        //Pass components in array\n\
        section.setComponentTypes([' + comp_list.join(', ') + ']);\n\
\n\
        f.addLibSection(section);\n\
   });\n\
});';

        try {
            var file = crsaMakeFileFromUrl(url);
            var fs = require('fs');
            crsaWriteFileWithBackup(fs, file, source, "utf8");
            this.pluginUrl = url;
            pinegrow.saveFrameworksList();
            this.changed = false;
            pinegrow.showQuickMessage("Snippet library " + this.name + " saved.");
        }
        catch(err) {
            showAlert("File " + url + " could not be saved. " + err, "Error");
        }

        if(done) done();
    }

    this.getActionTypes = function(pgel) {
        return this.getTypes(null, pgel);
    }

    this.getActionTag = function(pgel) {
        var types = this.getActionTypes(pgel);
        if(types.length) {
            var at = '';
            for(var i = 0;i < types.length; i++) {
                var t;
                if(types[i].get_action_tag) {
                    t = types[i].get_action_tag(pgel, types[i]);
                } else {
                    t = (types[i].short_name || types[i].name);
                }
                at += (at.length ? ', ' : '') + t;
            }
            return at;
        }
        return null;
    }


    this.textForExample = function(what) {
        return 'for example, ' + what;
    }

    this.textDefaultValue = function(what) {
        return 'default, ' + what;
    }

    this.textUse = function(what) {
        return 'use ' + what;
    }

    this.fieldIsRequired = function(obj, prop, value, fieldDef, $field, values) {
        if(!value) {
            return pinegrow.getValidationError(fieldDef.name, 'req');
        }
        return null;
    }

    this.fieldIsValidFile = function(obj, prop, value) {
        if(value) {
            var p = pinegrow.getCurrentProject();
            var ok = true;
            if(isApp()) {
                ok = crsaIsFileExist(p.getAbsolutePath(value));
            } else if(p) {
                ok = p.getFileForUrl(value)
            }
            if(!ok) {
                return 'The file does not exist.'
            }
        }
    }

    //CSS helpers
    this.setCSSProperty = function(page, css_url, selector, prop, value) {

    }


    this.requireSelectedElement = function(callback) {
        var cp = pinegrow.getSelectedPage();
        if(!cp) {
            pinegrow.showQuickMessage('A page must be open for this to work.');
        } else {
            var pgel = pinegrow.getSelectedElement();
            if(!pgel) {
                pinegrow.showQuickMessage('An element must be selected.');
            } else {
                callback(cp, pgel);
            }
        }
    }

    this.requireSelectedPage = function(callback) {
        var cp = pinegrow.getSelectedPage();
        if(!cp) {
            pinegrow.showQuickMessage('A page must be open for this to work.');
        } else {
            callback(cp);
        }
    }


    this.on_element_inserted = function(page, pgel, $el, def) {

        if(def.framework && def.framework == this) {
            //that's us
            if(def.framework.resources.has()) {
                var res_namespace = page.getResourceNamespaceForFramework(def.framework);

                //skip this if framework is source framework for this project: TODO

                if(res_namespace) {
                    var fs = require('fs');

                    //var $html = page.get$Html();
                    //we should map urls
                    var attrs = ['src', 'data-src', 'href', 'action'];
                    for(var i = 0; i < attrs.length; i++) {
                        var attr = attrs[i];
                        var list = pgel.findIncludingSelf('[' + attr + ']');
                        for(var j = 0; j < list.length; j++) {
                            var el = list[j];
                            var url = el.getAttr(attr);
                            if(url && !crsaIsAbsoluteUrl(url)) {

                                url = res_namespace + url;
                                var file = crsaMakeFileFromUrl( page.makeAbsoluteUrl(url) );
                                if(crsaIsFileOrDir(file, fs) == 'file') {
                                    //resource file exists at new location, lets update url
                                    el.setAttr(attr, url);
                                    /*
                                    var $domel = el.get$DOMElement($html);
                                    if($domel) {
                                        $domel.attr(attr, url);
                                    }*/
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    this.addResourcesToPage = function(page, done, overwrite_existing, same_project, resources) {
        var path = require('path');

        var wasChanged = function() {
            page.setPageChanged(true);
            pinegrow.changeManager.didChangePage(page, 'Add resources', {
                source: 'add_resources'
            })
        }

        var project = pinegrow.getCurrentProject();

        var resource_folder = null;
        var changed = false;

        if(page.localFile) {
            if (project && project.isFileInProject(page.localFile)) {
                resource_folder = project.getDir();
            } else if (page.localFile) {
                resource_folder = path.dirname(page.localFile);
            }
        }

        resources = resources || this.resources;

        if(resource_folder) {

            var res_namespace = this.getResourceNamespacePath();
            if(same_project) {
                //
            } else {
                resource_folder = path.join(resource_folder, res_namespace);
            }
            //page.setResourceNamespaceForFramework(this, res_namespace);

            resources.copyFilesToFolder( resource_folder, function( errors ) {
                if(resources.addToPage(page, resource_folder)) {
                    //changed
                    wasChanged();
                    changed = true;
                    if(done) done();
                } else {
                    if(done) done();
                }
            }, overwrite_existing /* dont overwrite existing */)
        } else {
            if(resources.addToPage(page)) {
                //changed
                wasChanged();
                changed = true;
                if(done) done();
            } else {
                if(done) done();
            }
        }
        if (changed) {
            //pinegrow.refreshAllPages();
        }
    }

    this.addResourcesToProject = function(project, done, overwrite_existing, same_project, skip_add_to_pages, resources) {
        var path = require('path');

        var wasChanged = function() {
        }

        var resource_folder = project.getDir();

        var res_namespace = this.getResourceNamespacePath();

        if(same_project) {
            resource_folder += path.sep;
        } else {
            resource_folder = resource_folder + path.sep + res_namespace;
        }

        resources = resources || this.resources;

        /*var project_info = project.getProjectInfo();
        var resources = project_info.getSetting('resources') || {};
        resources[this.type] = res_namespace;
        project_info.setSetting('resources', resources);
        project_info.save();*/

        resources.copyFilesToFolder( resource_folder, function( errors ) {

            //add to pages
            if(skip_add_to_pages) {

                if(skip_add_to_pages instanceof CrsaPage) {
                    resources.addToPage(skip_add_to_pages, resource_folder);
                }
                if (done) done();
                pinegrow.refreshCurrentProjectDelayed(true);

            } else {

                project.forEachEditableFile(function (page, pageDone, status) {
                    //on page
                    if (resources.addToPage(page, resource_folder)) {
                        status.changed = true;
                        status.skip_refresh = true; //refresh was already done by addtopage
                    }
                    pageDone(page, status);

                }, function () {
                    //on done
                    if (done) done();
                    pinegrow.refreshCurrentProjectDelayed(true);

                }, 'Updating resources...')
            }

        }, overwrite_existing);
    }

    this.getResourceNamespace = function() {
        return this.customResourceNamespace !== null ? this.customResourceNamespace : 'components/' + (this.type || this.key) + '/';
    }

    this.getResourceNamespacePath = function() {
        return this.getResourceNamespace().replace(/\//g, require('path').sep);
    }

    //Helper method to easily add resource
    this.addResource = function(url, detect, always, options) {
        return this.resources.createAndAdd(url, this, detect, always, options);
        /*
        if(typeof detect === 'object') {
            options = detect;
            detect = null;
        }
        options = $.extend({
            resources_dir: 'template',
            detect: detect || false,
            always: always || false,
            replace: true,
            remote_source: null
        }, options || {})

        var relative_url = null;
        var source = null;

        if(crsaIsAbsoluteUrl(url)) {

        } else {
            relative_url = url;
            url = this.getResourceUrl(options.resources_dir + '/' + url);
            source = url;
        }
        res = new PgComponentTypeResource(options.remote_source || url);
        res.relative_url = relative_url;
        res.source = source;
        res.detect = options.detect;
        res.replace = options.replace;
        res.footer = res.type == 'application/javascript';
        if(!always) {
            //res.on_add_resource_filter = onAddResourceFilter;
        }
        this.resources.add(res);
        return res;
        */
    }

}

var PgFrameworkLibSection = function(key, name, small_name) {
    this.key = key;
    this.name = name;
    this.components = [];
    this.framework = null;
    this.small_name = small_name;

    this.addComponentType = function(pgComponentType) {
        this.components.push(pgComponentType);
    }

    this.setComponentTypes = function(list) {
        this.components = list;
    }

    this.getComponentTypes = function() {
        return this.components;
    }
}

var PgComponentTypeResource = function(url, code, force_type) {

    var _this = this;

    this.type = force_type || pinegrow.getMimeType(url);
    this.url = url;
    this.code = code;
    this.detect = null;
    this.footer = false;
    this.project = null;
    this.isFolder = false;
    this.source = null;
    this.relative_url = null;
    this.replace = false;
    this.place_before_user_style = false;

    this.detectAttrInPage = function (crsaPage) {
        if (!this.detect) return false;

        if(this.code) {
            var found = null;
            crsaPage.sourceNode.walk(function(node) {
                if(node.isElement && node.tagName === 'script') {
                    if(node.content && node.content.match(_this.detect)) {
                        found = node;
                        return false;
                    }
                }
                return true;
            })
            return found;
        }

        var tag = { name: '', attr: '' };
        if (this.type == 'text/css') {
            tag.name = 'link';
            tag.attr = 'href';
        }
        else if (this.type == 'application/javascript') {
            tag.name = 'script';
            tag.attr = 'src';
        }

        var found = false;
        var tags = crsaPage.sourceNode.findIncludingSelf(tag.name);
        for (var i = 0; i < tags.length; i++) {
            var el = tags[i];
            var url = el.getAttr(tag.attr);
            if (url) {
                var matched = url.match(this.detect);
                if (matched && matched.length > 0) {
                    found = el;
                    break;
                }
            }
        }

        return found;
    }

    this.isEqual = function(r) {
        return this.url == r.url || (this.code && this.code == r.code);
    }

    this.isCss = function() {
        return this.type == 'text/css';
    }

    this.addToPage = function(crsaPage, folder) {
        var url = this.code ? 'http://dummy' : this.url;

        if(this.relative_url && !crsaIsAbsoluteUrl(this.relative_url) && this.relative_url && folder) {
            if(isApp()) {
                url = crsaMakeUrlFromFile(require('path').join(folder, this.relative_url));
            } else {
                url = folder + this.relative_url;
            }
        }

        if(this.type == 'text/css') {
            var found_pgel = this.detectAttrInPage(crsaPage);
            var replace = false;
            if(found_pgel) {
                replace = typeof this.replace == 'function' ? this.replace(crsaPage, found_pgel) : this.replace;
            }
            if (!found_pgel || replace) {
                var insert_before = this.place_before_user_style ? 'style.css' : null;
                return crsaPage.addStylesheet(url, false, null, found_pgel, insert_before);
            }
            else {
                return null;
            }
        } else if(this.type == 'application/javascript') {
            var found_pgel = this.detectAttrInPage(crsaPage);
            var replace = false;
            if(found_pgel) {
                replace = typeof this.replace == 'function' ? this.replace(crsaPage, found_pgel) : this.replace;
            }
            if (!found_pgel || replace) {
                return crsaPage.addScript(url, this.footer, found_pgel, this.code);
            }
            else {
                return null;
            }
        } else if(this.isFolder) {

        }
        return false; //not changed
    }

    this.copyFilesToPage = function(crsaPage, done) {
        this.copyFilesToFolder( require('path').dirname( crsaPage.localFile ), done);
    }

    this.copyFilesToProject = function(project, done, overwrite_existing) {
        this.copyFilesToFolder(project.getDir(), done, overwrite_existing);
    }



    this.copyFilesToFolder = function(folder, done, overwrite_existing) {
        //copy to project
        var path = require('path');

        if(!this.relative_url) {
            if(done) done();
            return;
        }

        if (crsaIsAbsoluteUrl(this.relative_url || this.url)) {
            if (done) done();
            return;
        }

        if(!isApp()) {
            //if mapped url skip copy
            var project = pinegrow.getCurrentProject();
            if(project) {
                var oe = typeof overwrite_existing == 'function' ? overwrite_existing(this.relative_url) : overwrite_existing;

                oe = false; //in online version, force overwrite to false

                var dest_exists = project.hasContentForUrl(this.relative_url);
                if(dest_exists && !oe) {
                    if(done) done();
                    return;
                }

                var mapped = project.mapSourceUrlToDOMUrl(this.relative_url, false);
                if(mapped) {
                    done && done();
                    return;
                } else {
                    pgFetchRemoteContent(this.url).catch(function (error) {
                        console.log('Error getting file ' + _this.url, error);
                        if (done) done();
                    }).then(function (filedata) {
                        try {
                            if (filedata === undefined) throw 'Error getting remote file ' + _this.url;
                            project.setContentOfUrl(_this.relative_url, new TextDecoder().decode(filedata), done);
                            project.settingsFileIsChanged(_this.relative_url);//using this here to force the next save to save this file
                        } catch (err) {
                            console.log('Error getting remote file ' + _this.url, err);
                            done && done();
                        }
                    });
                }
            }
            return;
        }

        var relative_to_source_project = crsaMakeFileFromUrl(this.relative_url);
        var dest_path = path.join(folder, relative_to_source_project);


        var fs = require('fs');

        var dest_exists = false;
        try {
            var stat_dest = fs.statSync(dest_path);
            dest_exists = stat_dest.isFile() /* && stat_dest.mtime >= stat_source.mtime) */;
        } catch(err) {
            dest_exists = false;
        }

        var oe = typeof overwrite_existing == 'function' ? overwrite_existing(this.relative_url) : overwrite_existing;

        if(dest_exists && !oe) {
            if(done) done();
            return;
        }


        if(crsaIsRemoteUrl(this.url)) {

            var saveData = function(data) {
                try {
                    crsaCreateDirs(path.dirname(dest_path), fs);
                    fs.writeFileSync(dest_path, _this.remoteData);
                    console.log('Save remote file ' + _this.url + ' -> ' + dest_path);
                } catch(err) {
                    console.log('Could not save ' + dest_path, err);
                }
            }

            if(_this.remoteDataFetchedAt && (new Date()).getTime() - _this.remoteDataFetchedAt > 60*1000) {
                _this.remoteData = null;
            }
            if(_this.remoteData) {
                saveData();
                if(done) done();
            } else {
                //copy remote file to local
                pgFetchRemoteContent(this.url).catch(function (error) {
                    console.log('Error getting remote file ' + _this.url, error);
                    if (done) done();
                }).then(function (data) {
                    try {
                        if (data === undefined) throw 'Error getting remote file ' + _this.url;
                        _this.remoteData = new Buffer(data);
                        _this.remoteDataFetchedAt = (new Date()).getTime();
                        saveData();
                    } catch (err) {
                        console.log('Error getting remote file ' + _this.url, err);
                    }
                    if (done) done();
                });
            }

            return;
        }

        var source_path = crsaMakeFileFromUrl(this.url);

        if(source_path == dest_path) {
            if(done) done();
            return;
        }

        this.copy(dest_path, done, overwrite_existing);
    }

    /*
    this.existsInFolder = ===function(folder) {
        //copy to project
        var path = require('path');

        var relative_to_source_project = crsaMakeFileFromUrl(this.relative_url);
        if (crsaIsAbsoluteUrl(this.relative_url || this.url)) {
            return true;
        }

        var source_path = crsaMakeFileFromUrl(this.url);
        var dest_path = path.join(folder, relative_to_source_project);


        this.copy(dest_path, done, overwrite_existing);
    }

*/

    this.copy = function(dest, done, overwrite_existing, file_writter) {
        var path = require('path');
        var fs = require('fs');

        try {
            console.log('RESOURCES - COPY ' + this.source + ' -> ' + dest);

            crsaCreateDirs(path.dirname(dest), fs);

            var type = crsaIsFileOrDir(crsaMakeFileFromUrl(this.source), fs);

            if(type == 'dir') {
                crsaCopyFolder(crsaMakeFileFromUrl(this.source), dest, function(errors) {
                    if(done) done(errors);
                }, true /* quick, overwrite newer */ , overwrite_existing)
            } else if(type == 'file') {
                //console.log("RES " + this.source + ' -> ' + dest);

                var dest_exists = false;
                try {
                    var stat_dest = fs.statSync(dest);
                    var stat_source = fs.statSync(crsaMakeFileFromUrl(this.source));
                    //exists
                    dest_exists = stat_source.isFile() /* && stat_dest.mtime >= stat_source.mtime) */;
                } catch(err) {
                    dest_exists = false;
                }

                var oe = typeof overwrite_existing == 'function' ? overwrite_existing(this.relative_url) : overwrite_existing;

                if(!dest_exists || oe) {
                    if(file_writter) {
                        file_writter.copyFile(dest, crsaMakeFileFromUrl(this.source));
                    } else {
                        crsaCopyFileSync(fs, crsaMakeFileFromUrl(this.source), dest);
                    }
                } else {
                    //console.log('SKIP ' + dest);
                }

                if(done) done();
            } else {
                if(done) done(crsaMakeFileFromUrl(this.source) + ' does not exist.');
            }
        } catch(err) {
            console.log('COPY ERROR ' + err);
            if(done) done(err);
        }
    }
}

var PgResourcesList = function() {

    this.list = [];
    var _this = this;

    var description = null;

    this.clear = function () {
        this.list = [];
    }

    this.has = function() {
        return this.list.length > 0;
    }

    this.add = function(r) {
        for(var i = 0; i < this.list.length; i++) {
            if(this.list[i].isEqual(r)) return;
        }
        this.list.push(r);
    }

    this.createAndAdd = function(url, f, detect, always, options) {
        if(typeof detect === 'object') {
            options = detect;
            detect = null;
        }
        options = $.extend({
            resources_dir: 'template',
            detect: detect || false,
            always: always || false,
            replace: true,
            remote_source: null,
            place_before_user_style: false
        }, options || {})

        var relative_url = null;
        var source = null;

        if(crsaIsAbsoluteUrl(url)) {

        } else {
            relative_url = url;
            url = f ? f.getResourceUrl(options.resources_dir + '/' + url) : null;
            source = url;
        }
        res = new PgComponentTypeResource(options.remote_source || url);
        res.relative_url = relative_url;
        res.source = source;
        res.detect = options.detect;
        res.replace = options.replace;
        res.footer = res.type == 'application/javascript';
        res.place_before_user_style = options.place_before_user_style;

        if(!always) {
            //res.on_add_resource_filter = onAddResourceFilter;
        }
        this.add(res);
        return res;
    }

    this.detectRequiredResources = function(page, f) {
        var map = {};
        var remove = [];

        page.sourceNode.goThroughAllUrls(function(url, attr, pgel) {
            if(!url.startsWith('#') && !crsaIsAbsoluteUrl(url)) {
                //relative url
                var remote_source = (new URL(url, page.url)).toString();

                _this.createAndAdd(url, f, {
                    resources_dir: 'pagelib',
                    remote_source: remote_source
                })

                if(pgel.getAttribute('rel') === 'pg-require') {
                    remove.push(pgel)
                }

            }
            return url;
        })

        remove.forEach(function(pgel) {
            pgel.removeWithoutEvents();
        })

        return remove.length > 0;
    }

    /*
    @crsaPage: page to add to
    @folder: location of resource, either project dir or file dir
     */
    this.addToPage = function(crsaPage, folder, done) {
        var changed = false;
        var css_added = false;

        for(var i = 0; i < this.list.length; i++) {
            if(this.list[i].on_add_resource_filter && !this.list[i].on_add_resource_filter(this.list[i])) {
                continue;
            }
            if(this.list[i].addToPage(crsaPage, folder)) {
                changed = true;
                if(this.list[i].isCss()) {
                    css_added = true;
                }
            }
        }
        if(done) done(crsaPage, this);

        if(crsaPage.openedInEditor) {
            if(changed) {
                if(css_added) {
                    crsaPage.refreshWithUpdateStylesheets();
                } else {
                    crsaPage.refresh();
                }
            }
        }

        if(changed) {
            crsaPage.setPageChanged(true);
            pinegrow.changeManager.didChangePage(crsaPage, 'Add resources', {
                source: 'add_resources'
            })
        }
        return changed;
    }

    this.copyFilesToFolder = function(folder, done, overwrite_existing) {
        var idx = 0;

        var copyItem = function() {
            if(idx == _this.list.length) {
                if(done) done();
            } else if(idx < _this.list.length) {
                if(_this.list[idx].on_add_resource_filter && !_this.list[idx].on_add_resource_filter(_this.list[idx])) {
                    idx++;
                    copyItem();

                } else {
                    _this.list[idx].copyFilesToFolder(folder, function (errors) {
                        idx++;
                        copyItem();
                    }, overwrite_existing);
                }
            }
        }

        copyItem();
    }

}

var PgComponentType = function(type, name, options) {
    this.type = type;
    this.name = name;
    this.short_name = name;
    this.selector = null;
    this.code = null;
    this.preview = null;
    this.sections = null;
    this.priority = 1000;
    this.attribute = null; //for actions
    this.attribute_default = null;
    this.parameter_attributes = []; //array of {name, default}
    this.parent_selector = null;
    this.inherit_from = null;
    this.empty_placeholder = null;
    this.display_name = null;
    this.live_update = true;
    this.framework = null;
    this.set_value = null;//set_value(obj, value, values, oldValue, eventType);
    this.resources = null;

    if(options) {
        var _this = this;
        $.each(options, function(key, val) {
            _this[key] = val;
        })
    }
}

PgComponentType.prototype.addPrefix = function(prefix) {
    this.type = prefix + '.' + this.type;
    var r = {};
    $.each(this.sections, function (skey, sdef) {
        if (sdef.fields) {
            var fields = {};
            $.each(sdef.fields, function (fkey, fdef) {
                fields[prefix + '.' + fkey] = fdef;
            })
            sdef.fields = fields;
        }
        r[prefix + '.' + skey] = sdef;
    })
    this.sections = r;
}

PgComponentType.prototype.toJSCode = function(vn) {
    var escapeCode = function(code) {
        code = code.replace(/\r\n|\r|\n/g, '\\\n');
        return code.replace(/'/g,'\\\'');
    }

    var c = '\n\
        var ' + vn + ' = new PgComponentType(\'' + this.type + '\', \'' + escapeCode(this.name) + '\');\n\
        ' + vn + '.code = \'' + escapeCode(this.code) + '\';\n\
        ' + vn + '.parent_selector = ' + (this.parent_selector ? '\'' + this.parent_selector + '\'' : 'null') + ';\n\
        f.addComponentType(' + vn + ');\n\
        ';
    return c;
}

PgComponentType.prototype.addResource = function(res) {
    if(this.resources === null) {
        this.resources = new PgResourcesList();
    }
    this.resources.add(res);
}

PgComponentType.prototype.addResources = function(res_list) {
    for(var i = 0; i < res_list.length; i++) {
        this.addResource(res_list[i]);
    }
}

PgComponentType.prototype.addRequireCSS = function(res) {
    if(this.require_css.indexOf(res) < 0) this.require_css.push(res);
}

PgComponentType.prototype.getActionParameters = function() {
    if(this.action_parameter_list) return this.action_parameter_list;
    var r = [];
    $.each(this.sections, function(skey, sdef) {
        if(sdef.fields) {
            $.each(sdef.fields, function(fkey, fdef) {
                if(fdef.attribute) {
                    r.push({name: fdef.attribute, 'default': fdef.default ? fdef.default : null, def: fdef});
                }
            });
        }
    });
    this.action_parameter_list = r;
    return r;
}

PgComponentType.prototype.getActionValuesForPgel = function(pgel) {
    var r = {}
    if(this.attribute) {
        r[this.attribute] = pgel.getAttribute(this.attribute)
    }
    this.getActionParameters().forEach(function(a) {
        if(pgel.hasAttribute(a.name)) {
            r[a.name] = pgel.getAttribute(a.name)
        }
    })
    return r
}

PgComponentType.prototype.hasAction = function(pgel) {
    if (this.has_action) {
        return this.has_action(pgel);
    }
    return this.attribute && pgel.hasAttr(this.attribute);
}

PgComponentType.prototype.addEventHandler = function(event, func) {
    if(!this[event]) {
        this[event] = [func];
    } else {
        if(!Array.isArray(this[event])) {
            this[event] = [this[event]];
        }
        this[event].push(func);
    }
}





var PgPropertiesSection = function(key, name) {
    this.key = key;
    this.name = name;
    this.fields = {};

    this.addProperty = function(pgProp, key) {
        if(!key) key = pgProp.key;
        this.fields[pgProp.key] = pgProp;
    }
}

var PgProperty = function(key, name) {
    this.name = name;
    this.key = key;
    this.type = null;
    this.action = null;
    this.get_value = null;
    this.set_value = null;
    this.value = null;
    this.show_empty = null;
    this.attribute = null;
    this.options = null;
}

var PgComponent = function(name, done) {

    this.html = null;
    this.error = null;
    this.url = null;

    var _this = this;

    this.url = "components/" + name + '.html';

    if(isApp()) {
        var fs = require('fs');
        try {
            this.html = fs.readFileSync(crsaMakeFileFromUrl(this.url), {encoding: "utf8"});
            if(done) done(this);
        }
        catch(err) {
            this.error = err;
            if(done) done(this);
        }
    } else {
        $.ajax({
            url: this.url,
            data: null,
            dataType: 'text'
        })
            .done(function(data) {
                _this.html = data;
                if(done) done(_this);
            })
            .fail(function(a) {
                _this.error = a;
                if(done) done(_this);
            });
    }
}

var PgStats = function() {

    var events = [];
    var usage = {};
    var has_usage = false;

    var test = true;
    var _this = this;

    var install_id = crsaStorage.getValue('install_id');

    if(!install_id) {
        install_id = 'id' + new Date().getTime() + Math.round(Math.random() * 100000);
        crsaStorage.setValue('install_id', install_id);
    }

    setInterval(function() {
        _this.send();
    }, 1000*60*10);

    var version = crsaGetVersion();

    var app_run = parseInt(localStorage.app_run || "0");

    var fillUserData = function(e) {
        e.trialEmail = crsaStorage.getValue('userEmail');
        e.userEmail = crsaStorage.getValue('activatedUserEmail');
        e.serial = crsaStorage.getValue('activatedSerial');
        e.version = version;
        e.product = crsaStorage.getValue('activatedProduct');
        e.appRun = app_run;
        e.installId = install_id;
    }

    this.event = function(event, data, force) {
        var disable_stats = pinegrow && pinegrow.getSetting('disable-stats', '0') == '1';

        if(disable_stats && !force) return;

        var e = {};

        e.event = event;
        e.data = data || '';
        e.time = parseInt(new Date().getTime()/1000);
        fillUserData(e);

        events.push(e);

        if(test) this.send();
    }

    this.using = function(what) {
        if(!(what in usage)) {
            usage[what] = 0;
        }
        usage[what]++;
        has_usage = true;
    }

    var in_progress = false;

    this.send = function(done) {

        //disable
        if(done) done();
        return;
/*
        var disable_stats = pinegrow && pinegrow.getSetting('disable-stats', '0') == '1';

        //debugger;
        if(in_progress) {
            if(done) done();
            return;
        }

        if(events.length == 0 && (!has_usage || disable_stats)) {
            if(done) done();
            return;
        }

        var salt = 'jhuyuy575rhgty';

        var d = {};
        fillUserData(d);

        d.events = events;

        if(!disable_stats) {
            d.usage = usage;
        }

        var str = JSON.stringify(d);
        str = window.btoa(str);

        var url = 'http://shop.pinegrow.com/PinegrowEvents/event.php';

        //url = 'http://pinegrow/PinegrowEvents/event.php';

        in_progress = true;

        //console.log('sending stats...' + str);

        $.ajax({
            url: url,
            data: {str: str, id: md5(str + salt)},
            dataType: 'text',
            method: 'POST'
        }).done(function(data) {
            events = [];
            usage = {};
            has_usage = false;
            in_progress = false;

            //console.log('Got reply: ' + data);

            try {
                data = JSON.parse(data);
                if(data.msg_title) {
                    var delay = data.msg_delay || 10000;
                    setTimeout(function() {
                        pinegrow.showAlert(data.msg_text || '', data.msg_title);
                    }, delay);

                }
            } catch(err) {}

        }).fail(function(a, b, c) {
            in_progress = false;
        });

        setTimeout(function() {
            if(done) done();
        }, 1000);

 */
    }
}
