/**
 * Created by Matjaz on 06/11/15.
 */

var PgCodeValidator = function() {

    var _this = this;

    var known_tags = ['a','abbr','acronym','address','applet','area','article','aside','audio','b','base','basefont','bdi','bdo','bgsound','big','blink','blockquote','body','br','button','canvas','caption','center','cite','code','col','colgroup','command','content','data','datalist','dd','del','details','dfn','dialog','dir','div','dl','dt','element','em','embed','fieldset','figcaption','figure','font','footer','form','frame','frameset','h1','h2','h3','h4','h5','h6','head','header','hgroup','hr','html','i','iframe','image','img','input','ins','isindex','kbd','keygen','label','legend','li','link','listing','main','map','mark','marquee','menu','menuitem','meta','meter','multicol','nav','nobr','noembed','noframes','noscript','object','ol','optgroup','option','output','p','param','picture','plaintext','pre','progress','q','rp','rt','rtc','ruby','s','samp','script','section','select','shadow','slot','small','source','spacer','span','strike','strong','style','sub','summary','sup','table','tbody','td','template','textarea','tfoot','th','thead','time','title','tr','track','tt','u','ul','var','video','wbr','xmp']

    var known_tags_map = {}

    known_tags.forEach(function(tag) {
        known_tags_map[tag] = true;
    })

    var isKnown = function(tag) {
        return known_tags_map[tag] || false;
    }

    var headings = ['h1', 'h2', 'h3', 'h4', 'h5', 'h6'];
    var blocks = ['div', 'main', 'section'].concat(headings);

    var phrasing = ['abbr', 'audio', 'b', 'bdo', 'br', 'button', 'canvas', 'cite', 'code', 'command', 'data', 'datalist', 'dfn', 'em', 'embed', 'i', 'iframe', 'img', 'input', 'kbd', 'keygen', 'label', 'mark', 'math', 'meter', 'object', 'output', 'progress', 'q', 'ruby', 'samp', 'script', 'select', 'small', 'span', 'strong', 'sub', 'sup', 'svg', 'textarea', 'time', 'var', 'video', 'wbr']

    var self = 'self';

    //Construct rules

    var rules = {
        html: {
            only_children: ['head', 'body']
        },
        head: {
            only: ['title', 'base', 'link', 'style', 'meta', 'script', 'noscript', 'template', 'img'],
            parents: ['html']
        },
        title: {
            no: true,
            parents: ['head','svg']
        },
        base: {
            no: true,
            parents: ['head']
        },
        meta: {
            no: true
            //parents: ['head']
        },
        body: {
            parents: ['html']
        },
        p: {
            no: blocks,
            no_self: true
        },
        a: {
            no_self: true
        }
    }

    phrasing.forEach(function(h) {
        rules[h] = {
            no: blocks
        }
    })

    headings.forEach(function(h) {
        rules[h] = {
            no: headings
        }
    })

    //End construct rules

    var pgel_to_contained = {}

    var getContainedTags = function(node) {
        var map = {};
        var all = [];
        node.walk(function(pgel) {
            if(pgel.isElement && pgel.tagName) {
                if(isKnown(pgel.tagName)) {
                    if (!map[pgel.tagName]) {
                        all.push(pgel.tagName);
                        map[pgel.tagName] = true;
                    }
                }
            }
            return true;
        })
        pgel_to_contained[node.getId()] = all;
        return all;
    }

    var getChildrenTags = function(node) {
        var map = {};
        var all = [];
        if(node.children) {
            node.children.forEach(function (pgel) {
                if (pgel.isElement && pgel.tagName) {
                    if (isKnown(pgel.tagName)) {
                        if (!map[pgel.tagName]) {
                            all.push(pgel.tagName);
                            map[pgel.tagName] = true;
                        }
                    }
                }
            })
        }
        return all;
    }

    this.errors = [];

    var addError = function(pgel, msg) {
        _this.errors.push({pgel: pgel, error: msg});
    }

    var tagSelector = function(tag, sel) {
        return '<b data-tag="' + (sel || tag) + '">' + tag + '</b>';
    }

    var elementSelector = function(pgel) {
        return '<b data-pg-id="' + pgel.getId() + '">' + pgel.getName({short: true}) + '</b>';
    }

    this.validate = function(root) {

        var p = new CrsaProfile();

        root.walkSelfAndChildren(function(pgel) {
            var contained = null;
            var children = null;

            if(pgel.tagName) {
                var tag = pgel.tagName;
                if(rules[tag]) {
                    if(rules[tag].no) {
                        if(rules[tag].no === true) {
                            if(pgel.hasChildren()) {
                                addError(pgel, 'should not have children');
                            }
                        } else {
                            contained = getContainedTags(pgel);
                            for(var i = 0; i < rules[tag].no.length; i++) {
                                if(contained.indexOf(rules[tag].no[i]) >= 0) {
                                    addError(pgel, 'should not contain ' + tagSelector(rules[tag].no[i]) + ' elements');
                                }
                            }
                        }
                    }
                    if(rules[tag].no_self) {
                        contained = contained || getContainedTags(pgel);
                        if(contained.indexOf(pgel.tagName) >= 0) {
                            addError(pgel, 'should not contain ' + tagSelector(pgel.tagName) + ' elements.')
                        }
                    }
                    if(rules[tag].only) {
                        contained = contained || getContainedTags(pgel);
                        var badlist = [];
                        for(var i = 0; i < contained.length; i++) {
                            if(rules[tag].only.indexOf(contained[i]) < 0) {
                                badlist.push(tagSelector(contained[i]));
                            }
                        }
                        if(badlist.length) {
                            addError(pgel, 'should not contain ' + badlist.join(', ') + ' elements.')
                        }
                    }
                    if(rules[tag].only_children) {
                        children = children || getChildrenTags(pgel);
                        var badlist = [];
                        for(var i = 0; i < children.length; i++) {
                            if(rules[tag].only_children.indexOf(children[i]) < 0) {
                                badlist.push(tagSelector(children[i], '_' + children[i]));
                            }
                        }
                        if(badlist.length) {
                            addError(pgel, 'should not contain ' + badlist.join(', ') + ' elements.')
                        }
                    }

                    if(rules[tag].parents) {
                        if(pgel.parent && rules[tag].parents.indexOf(pgel.parent.tagName) < 0) {
                            addError(pgel, 'should not be a child of ' + elementSelector(pgel.parent) + '. Allowed parents are: ' + rules[tag].parents.join(', ') + '.');
                        }
                    }
                } else if(pgel.isSingleTag()) {
                    if(pgel.hasChildren()) {
                        addError(pgel, 'should not have children');
                    }
                } else if(!pgel.isProperlyClosed()) {
                    addError(pgel, 'is not properly closed' + (pgel.closingTag ? ' (closing tag is <b>' + pgel.closingTag + '</b>)' : '.'));
                }
            }
            return true;
        })

        p.show('Page Validation');

        return this.errors;
    }

    this.hasErrors = function() {
        return this.errors.length > 0;
    }

    this.describeErrors = function() {
        var html = '';
        if(this.errors.length === 0) {
            html += '<li>No errors found.</li>';
        } else {
            this.errors.forEach(function (e) {
                html += '<li data-pg-id="' + e.pgel.getId() + '"><span>' + e.pgel.getName({short: true}) + '</span> - ' + e.error + '</li>';
            })
        }
        return html;
    }

}



$(function () {

    if(!pgf.edit_html) return;

    $('body').one('pinegrow-ready', function (e, pinegrow) {//Wait for Pinegrow to start-up

        var auto_check = pinegrow.getSetting('auto-validate-code', '1') === '1';

        var f = new PgFramework('pg.code-validator', 'Code validator');
        pinegrow.addFramework(f);

        f.default = true;
        f.show_in_manager = false;

        f.validateHTML = function(node, auto_on_load) {

            var validator = new PgCodeValidator();

            validator.validate(node);

            if(!validator.hasErrors()) {
                if (auto_on_load) {
                    //nothing to say
                    return;
                } else {
                    //pinegrow.showQuickMessage('Everything looks fine!');
                }
            }

            var msg = '<p>The following HTML elements have problems:</p><ul>';
            msg += validator.describeErrors();
            msg += '</ul>';

            var page = node.getPage();

            var q = PgQuickDialog('HTML syntax errors in ' + page.getName(), function($element) {
                msg += '<p class="help">Click on the element on the above list to select it. If you can\'t see the element in the tree, switch to Source code <i class="icon icon-KODA"></i> tree mode.';
                if(auto_on_load) {
                    msg += ' Use "Page -&gt; Check for HTML errors" to validate the code at any time.';
                }
                msg += '</p>';

                $element.addClass('pg-code-validator-results');

                $element.on('click', '[data-tag]', function(e) {
                    e.preventDefault();
                    e.stopPropagation();

                    var tag = $(this).attr('data-tag');
                    var pgid = $(this).closest('li').attr('data-pg-id');

                    var pgel = getPgNodeByPgId(pgid);
                    if(pgel && tag) {
                        var only_kids = tag.startsWith('_');
                        tag = tag.replace('_', '');
                        pgel.scrollTo();

                        var els = pgel.find(tag, false, only_kids);
                        var more = false;
                        if(els.length > 10) {
                            els.splice(10, els.length - 10);
                            more = true;
                        }

                        var col = new PgCollection();
                        col.setList(els);

                        setTimeout(function() {
                            pinegrow.selectedElements.clear();
                            pinegrow.selectedElements.selectMany(col);
                            if(more) {
                                pinegrow.showQuickMessage('Only the first 10 <b>' + tag + '</b> elements in ' + pgel.getName({short:true}) + ' were selected.')
                            } else {
                                pinegrow.showQuickMessage('<b>' + tag + '</b> elements in ' + pgel.getName({short:true}) + ' were selected.')
                            }
                        }, 100)
                    }
                })

                $element.on('click', '[data-pg-id]', function(e) {
                    e.preventDefault();
                    var pgid = $(this).attr('data-pg-id');
                    var pgel = getPgNodeByPgId(pgid);
                    if(pgel) {
                        pgel.scrollTo();
                        setTimeout(function() {
                            pinegrow.selectElement(pgel);
                            pinegrow.showQuickMessage(pgel.getName({short:true}) + ' was selected.')
                        }, 100)
                    }
                })

                msg += '<button class="pg-button pg-button-primary refresh">Refresh</button>';

                $element.on('click', 'button.refresh', function(e) {
                    e.preventDefault();
                    validator = new PgCodeValidator();
                    validator.validate(node);

                    $element.find('ul').html(validator.hasErrors() ? validator.describeErrors() : '<li>No errors found.</li>');

                    page.refresh();
                })

                msg += '<label style="margin-left: 20px;position: relative;top: 2px;" class=" control-label auto-check"><input type="checkbox" ' + (auto_check ? 'checked="checked"' : '') + '> Check for errors on page open</label>';

                $element.on('change', 'input[type="checkbox"]', function(e) {
                    auto_check = $(this).is(':checked');
                    pinegrow.setSetting('auto-validate-code', auto_check ? '1' : '0');
                    pinegrow.showMode('Check for error on page open', auto_check);
                    !auto_check && pinegrow.showQuickMessage('Use Page -&gt; Check for HTML errors to check the current page.');
                })

                msg += `<p class="text-muted" style="border-top: 1px solid;
    margin-top: 20px;
    padding-top: 10px;">Please note that Pinegrow requires valid HTML code in order to work correctly. Therefore it is recommended to have the auto check enabled. If you encounter situations where correct code is flagged as invalid, let us know and we will improve the validator.</p>`

                /*
                setTimeout(function() {
                    addTooltip($element.find('.auto-check'), '');
                }, 10) */


                return msg;
            }, {width:480})

        }
/*
        f.on_project_item_get_context_menu = function(page, file, project) {

            var items = [];

            items.push({
                divider: true,
                header: 'Tools'
            })
            if(file.isEditable) {
                items.push({
                    label: "Validate HTML...",
                    func: function() {
                        var root = pinegrow.getSourceNodeOfUrl(file.url);
                        f.validateHTML(root);
                    }
                })
            }
            return items;
        }*/

        f.on_page_loaded = function(page, first_load) {
            if(first_load && auto_check) {
                f.validateHTML(page.sourceNode, true);
            }
        }

        f.on_page_menu = function(page, items) {

            items.push({
                label: 'Check for HTML errors...',
                kbd: null,
                action: function() {
                    f.validateHTML(page.sourceNode);
                }
            })

            return items;
        }

    })
});