class PgBackendProviderLocal {
    constructor() {
        var gui = require('nw.gui');
        this.path = require('path');
        this.data_dir = this.path.join(gui.App.dataPath, 'pg_tutorials');
        //crsaCreateDirs(this.data_dir);
    }

    callApi(name, data, done, error) {
        if(this[name]) {
            return this[name](data, done, error);
        } else {
            error && error({error: `Unknown local API call ${name}.`});
            return null;
        }
    }

    getProjectDir(pid) {
        return this.path.join(this.data_dir, pid);
    }

    apiGetProjects(data, done, error) {
        var data = {
            user: {
                name: 'Local user',
                level: 2,
                email: '',
                id: 1
            },
            projects: [],
            lessons: {}
        }

        var lessons = {}
        if(localStorage['lessons']) {
            lessons = JSON.parse(localStorage['lessons']);
        }
        $.each(lessons, function(key, val) {
            if(!(key in data.lessons)) {
                data.lessons[key] = val;
            }
        })

        if(localStorage.pgRecentFiles) {
            var list = JSON.parse(localStorage.pgRecentFiles);
            var use_thumbs = pinegrow.thumbnailMaker.enabled;

            var names = []

            data.projects = list.map(function(p) {
                var url = p.file;
                if(!url.startsWith('internal:')) {
                    url = crsaMakeUrlFromFile(p.file);
                }
                var thumb_url;
                if(use_thumbs) {
                    if (p.thumb) {
                        thumb_url = pgMakeFileUrl(crsaGetBaseForUrl(window.location.href) + p.thumb);
                    } else {
                        thumb_url = pinegrow.thumbnailMaker.getStoredThumbnailUrl(p.file);
                    }
                }
                var name = crsaGetNameFromUrl(url);
                names.push(url);

                return {
                    name: name,
                    url: url,
                    project_id: url,
                    thumb: thumb_url,
                    is_file: !p.project
                }
            })

            var short_names = pgGetUniqueShortNamesForUrls(names);

            for(var i = 0; i < short_names.length; i++) {
                var short = pgUrlToDisplayString(short_names[i]);
                if(short.length > 50) {
                    short = '&hellip;' + short.substr(short.length - 50);
                }
                data.projects[i].name = short;
            }

        }
        done(data);
    }

    apiLoadProject(data, done, error) {
        error({
            status: 404
        })
    }

    apiSaveProject(data, done, error) {
        var dir = this.getProjectDir(data.project_id);
        crsaCreateDirs(dir);
        $.each(data.urls, function(url, d) {

        })
    }

    apiDeleteProjects(data, done, error) {
        var _this = this;
        data.project_ids.forEach(function(pid) {
            var dir = _this.path.join(_this.data_dir, pid.replace('tutorial_', ''));
            if(crsaIsFileExist(dir)) {
                var f = new CrsaFile();
                f.path = dir;
                f.deleteFileAndFolder(dir);
            }
        })
        done && done();
    }

    apiSetLessonStatus(data, done, error) {
        var lessons = {}
        if(localStorage['lessons']) {
            lessons = JSON.parse(localStorage['lessons']);
        }
        lessons[data.lesson_id] = data.status;
        localStorage['lessons'] = JSON.stringify(lessons);
        done && done();
    }

    apiSetLessonStatusMultiple(data, done, error) {
        var lessons = {}
        if(localStorage['lessons']) {
            lessons = JSON.parse(localStorage['lessons']);
        }
        localStorage['lessons'] = JSON.stringify(lessons);
        done && done();
    }
}