$(function() {

    if(!pgf.edit_css) return;

    $('body').one('pinegrow-ready', function(e, pinegrow) {
        var f = new PgFramework('pg.css.grid', 'CSS Grid');
        f.type = "pg.css.grid";
        f.allow_single_type = true;

        f.description = 'CSS Grid';
        f.author = 'Pinegrow';
        f.author_link = 'https://pinegrow.com';

        f.has_actions = false;
        f.show_in_manager = false;
        f.default = true;

        f.setScriptFileByScriptTagId('plugin-pgcssgrid');
        pinegrow.addFramework(f);

        f.mode = 'css';
        if(pgIsDev()) {
            //f.mode = 'dm';
        }

        help_url = 'https://pinegrow.com/docs/css-grid/';

        pinegrow.addEventHandler('on_set_inline_style', function(page, o) {
            o.css += `
    .pinegrow-ui-helper-css-grid {
        position:absolute;
        left:0;
        top:0;
        right:0;
        bottom:0;
        pointer-events: none !important;
        display: none !important;
        grid-column: unset !important;
        grid-row: unset !important;
        grid-area: unset !important;
        background: transparent !important;
    }
    .crsa-show-css-grid .pinegrow-ui-helper-css-grid {
        display:block !important;
    }
    .pinegrow-ui-helper-relative {
        position: relative !important;
        }
        .pinegrow-ui-helper-css-grid-cell {
            position: absolute !important;
            pointer-events: none !important;
            outline:1px dashed var(--pg-grid-outline) !important; 
            outline-offset: 0 !important;
            z-index: 101 !important;
            filter: drop-shadow(0px 0px 0px rgba(255, 255, 255, 0.5)) !important;
            background: transparent !important;
        }
        
        .pinegrow-ui-helper-css-grid-hl > pgdiv > .pinegrow-ui-helper-css-grid-cell {
            background-color: #009fff1c;
        }
        .pinegrow-grid-cell-during-resize {
            z-index: 100 !important;
            width: 100% !important;
            height: 100% !important;
            box-sizing: border-box !important;
        }
        .pinegrow-grid-during-resize > pgdiv > .pinegrow-ui-helper-css-grid-cell {
            min-height: 20px !important;
            min-width: 20px !important;
        }
        .pinegrow-ui-helper-css-grid-cell-highlight {
            background-color: #009fff1c !important;
        }
`
        });

        var gridHelper = new PgPageViewHelper('CSS grid', 'css-grid-helper', 'crsa-show-css-grid', $('<div></div>'), `html { --pg-grid-outline: #COLOR#;}`, '1', null, true, 'gray');

        var css_grid_controls = [];
        var force_css_grid_control_for_pgel = null;

        var updateCSSGridControls = function() {
            css_grid_controls.forEach(function(gc) {
                gc.destroy();
            })
            css_grid_controls = [];

            var show = function(pgel) {
                var $el = pgel.get$DOMElement();
                $el = shouldGridBeShown($el);
                if($el) {
                    var gc = new PgCSSGridOverlayControl();
                    gc.show($el);
                    css_grid_controls.push(gc);
                }
            }

            var force_pgel = force_css_grid_control_for_pgel;

            pinegrow.selectedElements.forEachElement(function(pgel) {
                show(pgel);
                if(force_pgel == pgel) {
                    force_pgel = null;
                }
            })

            if(force_pgel) {
                show(force_pgel);
            }
        }

        var shouldGridBeShown = function($el) {
            if(!$el) return null;
            var display = $el.css('display');
            if(display == 'grid' || display == 'inline-grid') return $el;
            var $p = $el.parent();
            if($p.length == 0 || $p.get(0).nodeType != 1) return null;
            display = $p.css('display');
            if(display == 'grid' || display == 'inline-grid') return $p;
            return null;
        }

        pinegrow.addEventHandler('on_page_changes_done', function(page, info) {
            updateCSSGridControls();
        })

        /*
        //Changes for compiled stylesheets are not yet applied to the page when this event is triggered.
        //We have to use on_css_applied_to_pages instead.

        pinegrow.addEventHandler('on_css_stylesheet_changed', function(pagenull) {
            updateCSSGridControls();
        })
        */

        pinegrow.addEventHandler('on_css_applied_to_pages', function(pagenull) {
            updateCSSGridControls();
        })

        pinegrow.addEventHandler('on_selected_elements_changed', function(page) {
            updateCSSGridControls();
        });

        pinegrow.addEventHandler('on_page_view_resized', function() {
            updateCSSGridControls();
        });

        pinegrow.addEventHandler('on_get_menu_commands', function(page, pgel, commands, $el) {

            //if(!pgf.simple_element_menu) {
                commands.push({
                    icon: 'icon-grid',
                    tooltip: 'Edit CSS grid ' + pgShowKbd('CMD+G'),
                    action: function () {
                        editGrid(pgel);
                    }
                })
            //}
        })

        pinegrow.addEventHandler('on_build_actions_menu', function(page, actions, pgel, all_selected) {
            actions.push({
                label: 'Edit CSS Grid...',
                kbd: 'CMD G',
                action: function() {
                    editGrid(pgel);
                }
            })
        })

        var parseGapValue = function(val) {
            if(!val) val = 0;
            if(typeof val === 'string') val = val.replace('px', '');
            if(isNaN(val)) return 0;
            return parseFloat(val);
        }

        var force_active_css_rule = null;

        //Used by get points for menu and to return CSS Grid submenu
        var getMenuPointsWithCallback = f.gridGetMenuPointsWithCallback = function(page, pgel, menu, points_done, return_actions) {
            var points = [];

            if(!menu.pageView.isActive() && !return_actions) return;
            if(!gridHelper.getCurrentValue() && !return_actions) return;

            if(menu.$el.length == 0 || menu.$el.get(0).tagName == 'HTML') return;

            var grid_info = new PgCSSGridInfo(menu.$el);
            var parent_grid_info = new PgCSSGridInfo(menu.$el.parent());
            var cell_info = new PgCSSGridCellInfo(menu.$el, grid_info);

            force_active_css_rule = pgel.getData('cssGridActiveRule');

            if(grid_info.isGrid() && !(parent_grid_info.isGrid() && cell_info.isCell())) {
                var grid_area = new PgElementPageViewMenuPointButton(menu, '<i class="icon icon-grid"></i>');
                points.push(grid_area);

                if(return_actions) {
                    var actions = [
                        {
                            label: 'Edit grid...',
                            action: function () {
                                editGrid(pgel);
                            }
                        }
                    ]
                    points_done(actions)
                } else {
                    grid_area.onClick = function () {
                        editGrid(pgel);
                        return;
                    }
                }

            } else {
                var $p = menu.$el.parent();
                if ($p.length && $p.get(0).nodeType == 1) {

                    grid_info = parent_grid_info;//new PgCSSGridInfo($p);
                    cell_info.grid_info = grid_info;
                    cell_info.update();

                    //var rule = pinegrow.getSelectedRule();

                    //console.log(rule);

                    if (grid_info.isGrid() && cell_info.isCell()) {

                        var rule_to_force = force_active_css_rule;
                        if(!return_actions) {
                            force_active_css_rule = null;
                        }

                        var active_css_rule;
                        var rule_selection_done = false;

                        if(f.mode == 'css') {
                            var p = new CrsaProfile();
                            var active_rules = new PgActiveRules(menu.pgel);
                            active_rules.getMatchedRules(function(list) {
                                active_rules_list = list;

                                var prop_list = ['grid-area', 'grid-column-start', 'grid-column-end', 'grid-row-start', 'grid-row-end', 'grid-row', 'grid-column'];

                                if(rule_to_force) {
                                    if(rule_to_force == 'style_attribute') {
                                        active_css_rule = null;
                                        rule_selection_done = true;
                                    } else if(active_rules_list.indexOf(rule_to_force) >= 0) {
                                        active_css_rule = rule_to_force;
                                        rule_selection_done = true;
                                    }
                                }
                                if(!rule_selection_done) {
                                    if(hasStyleAttributeProps(menu.pgel, prop_list)) {
                                        active_css_rule = null;
                                        rule_selection_done = true;
                                    }
                                }
                                if(!rule_selection_done) {
                                    active_css_rule = pgFindCSSRuleWithProps(active_rules.getActiveNodes(), prop_list, false);
                                    if(!active_css_rule) {
                                        //lets use the selected rule in style panel
                                        if(pinegrow.selectedElements.isSelectedElement(menu.pgel)) {
                                            var sr = pinegrow.getSelectedRule();
                                            if(sr) {
                                                active_css_rule = sr;
                                                if(sr.selector === 'Style attribute') {
                                                    active_css_rule = null; //null for style attribute
                                                }
                                            }
                                        }
                                    }
                                }

                                var area_name = cell_info.getAreaName();

                                cell_info.colRowPosition = cell_info.getCellColRowPosition();

                                var placement, placement_str;

                                var active_points = false;

                                if (area_name && area_name !== 'auto') {
                                    placement = 'area';
                                    placement_str = ' <areaname style="position: relative;top: -2px;">' + area_name + '</areaname>';
                                } else if (cell_info.isAutoPlaced()) {
                                    placement = 'auto';
                                    placement_str = '';
                                } else {
                                    placement = 'fixed';
                                    placement_str = '';
                                    active_points = true;
                                }

                                if(!return_actions) {
                                    var grid_right = new PgMenuPointCSSGrid('right', menu, 'column-end', grid_info, cell_info);
                                    grid_right.active_css_rule = active_css_rule;
                                    grid_right.onDragStarted = updateCSSGridControls;
                                    points.push(grid_right);

                                    var grid_left = new PgMenuPointCSSGrid('left', menu, 'column-start', grid_info, cell_info);
                                    grid_left.active_css_rule = active_css_rule;
                                    grid_left.onDragStarted = updateCSSGridControls;
                                    points.push(grid_left);

                                    var grid_top = new PgMenuPointCSSGrid('top', menu, 'row-start', grid_info, cell_info);
                                    grid_top.active_css_rule = active_css_rule;
                                    grid_top.onDragStarted = updateCSSGridControls;
                                    points.push(grid_top);

                                    var grid_bottom = new PgMenuPointCSSGrid('bottom', menu, 'row-end', grid_info, cell_info);
                                    grid_bottom.active_css_rule = active_css_rule;
                                    grid_bottom.onDragStarted = updateCSSGridControls;
                                    points.push(grid_bottom);
                                }


                                if (area_name || true) {
                                    var label_str = '<i class="icon icon-template"></i>' + placement_str;

                                    if(f.mode == 'css') {
                                        if (active_css_rule) {
                                            label_str += '<rulename>' + pgcssh.getRuleDisplayName(active_css_rule) + '</rulename>';
                                        } else {
                                            label_str += '<rulename>Style attribute</rulename>';
                                        }
                                    }

                                    var grid_area = new PgElementPageViewMenuPointButton(menu, label_str);
                                    points.push(grid_area);

                                    var getActions = function() {
                                        var areas = grid_info.getAvailableAreas();

                                        var actions = [
                                            {
                                                label: 'Edit parent grid...',
                                                kbd: 'CMD G',
                                                action: function () {
                                                    editGrid(pgel.parent);
                                                }
                                            },
                                            {
                                                label: 'Edit item grid...',
                                                //kbd: 'CMD G',
                                                action: function () {
                                                    editGrid(pgel);
                                                }
                                            }
                                        ]

                                        if (areas.length || true) {
                                            actions.unshift({
                                                type: 'header',
                                                label: 'Grid'
                                            })

                                            actions.unshift({
                                                type: 'divider'
                                            })
                                        }
                                        actions.unshift({
                                            label: 'Auto positioned',
                                            check: placement == 'auto',
                                            action: function () {
                                                active_css_rule = pgGetCurrentRule(active_css_rule);

                                                if (placement != 'auto') {
                                                    //make auto
                                                    pgCSSGridSetGridAreaForPgel(pgel, active_css_rule, null);
                                                    pgCSSGridTriggerChangeEvent(pgel, active_css_rule, 'Auto position');
                                                } else {
                                                    pgCSSGridSetGridAreaForPgel(pgel, active_css_rule, cell_info.getGridAreaShorthand(cell_info.autoToFixedPlace()));
                                                    pgCSSGridTriggerChangeEvent(pgel, active_css_rule, 'Cols &amp; rows position');
                                                }
                                                pinegrow.selectedElements.reselect();
                                            }
                                        })

                                        actions.unshift({
                                            type: 'header',
                                            label: 'Grid item'
                                        })

                                        if (areas.length || true) {
                                            actions.unshift({
                                                type: 'divider'
                                            })
                                            actions.unshift({
                                                label: 'Set named area...',
                                                action: function() {
                                                    pinegrow.showPrompt(`<p>Enter the name of the new CSS grid area:</p>`, "Create &amp; set new CSS grid area", null, 'areaname', null, function(area_name) {

                                                        active_css_rule = pgGetCurrentRule(active_css_rule);

                                                        area_name = $.trim(area_name);
                                                        if(area_name.length && area_name.match(/^[a-z][a-z0-9\_\-]*$/i)) {
                                                            var grid_pgel = pgel.parent;
                                                            var ar = new PgActiveRules(grid_pgel);
                                                            ar.getMatchedRules(function() {
                                                                var grid_rules = ar.getActiveNodes();

                                                                var grid_rule = null;
                                                                var props = ['grid-template-areas', 'grid-template-columns', 'grid-template-rows', 'grid-auto-columns', 'grid-auto-rows'];

                                                                if(hasStyleAttributeProps(grid_pgel, props)) {

                                                                } else {
                                                                    grid_rule = pgFindCSSRuleWithProps(grid_rules, props);
                                                                }
                                                                var cell_pos = cell_info.getCellColRowPosition(true);
                                                                if (grid_info.createNamedArea(grid_pgel, grid_rule, area_name, cell_pos['grid-column-start'], cell_pos['grid-row-start'], cell_pos['grid-column-end'], cell_pos['grid-row-end'])) {
                                                                    //created
                                                                    pgSetCSSPropForActiveRule(pgel, active_css_rule, 'grid-area', area_name);
                                                                    pgCSSGridTriggerChangeEvent(pgel, active_css_rule,  'Set named area to ' + area_name);
                                                                    pinegrow.selectedElements.reselect();
                                                                } else {
                                                                    pinegrow.showQuickError('Sorry, couldn\'t create the new named area.');
                                                                }
                                                            })

                                                        } else {
                                                            pinegrow.showQuickError('Invalid area name.');
                                                        }
                                                    })

                                                }
                                            })
                                            if(area_name && false) {
                                                actions.unshift({
                                                    label: 'Update area ' + area_name,
                                                    action: function() {
                                                        var grid_pgel = pgel.parent;
                                                        var ar = new PgActiveRules(grid_pgel);
                                                        ar.getMatchedRules(function() {
                                                            var grid_rules = ar.getActiveNodes();

                                                            var props = ['grid-template-areas', 'grid-template-columns', 'grid-template-rows', 'grid-auto-columns', 'grid-auto-rows'];

                                                            if(hasStyleAttributeProps(grid_pgel, props)) {

                                                            } else {
                                                                grid_rule = pgFindCSSRuleWithProps(grid_rules, props);
                                                            }

                                                            var cell_pos = cell_info.getCellColRowPosition(true);
                                                            if (grid_info.createNamedArea(grid_pgel, grid_rule, area_name, cell_pos['grid-column-start'], cell_pos['grid-row-start'], cell_pos['grid-column-end'], cell_pos['grid-row-end'])) {
                                                            } else {
                                                                pinegrow.showQuickError('Sorry, couldn\'t create the new grid area.');
                                                            }
                                                        })

                                                    }
                                                })
                                            }
                                            areas.forEach(function (a) {
                                                actions.unshift({
                                                    label: a,
                                                    action: function () {
                                                        //set a
                                                        active_css_rule = pgGetCurrentRule(active_css_rule);

                                                        pgSetCSSPropForActiveRule(pgel, active_css_rule, 'grid-area', a);
                                                        pgCSSGridTriggerChangeEvent(pgel, active_css_rule, 'Set named area to ' + a);
                                                        pinegrow.selectedElements.reselect();
                                                    },
                                                    check: area_name == a
                                                })
                                            })
                                            /* actions.unshift({
                                             type: 'divider'
                                             })*/
                                            actions.unshift({
                                                type: 'header',
                                                label: 'Set named area:'
                                            })
                                        }

                                        if(active_rules_list) {
                                            actions.unshift({
                                                type: 'divider'
                                            })

                                            for(var i = active_rules_list.length - 1; i >= 0; i--) {
                                                (function(rule) {
                                                    actions.unshift({
                                                        label: pgcssh.getRuleDisplayName(rule),
                                                        check: rule === active_css_rule,
                                                        action: function() {
                                                            force_active_css_rule = rule;
                                                            pgel.setData('cssGridActiveRule', rule);
                                                            pinegrow.selectedElements.reselect();
                                                        }
                                                    })
                                                })(active_rules_list[i])
                                            }

                                            actions.unshift({
                                                label: 'Style attribute',
                                                check: null === active_css_rule,
                                                action: function() {
                                                    force_active_css_rule = 'style_attribute';
                                                    pgel.setData('cssGridActiveRule', force_active_css_rule);
                                                    pinegrow.selectedElements.reselect();
                                                }
                                            })

                                            actions.unshift({
                                                type: 'header',
                                                label: 'CSS Rule for changes:'
                                            })
                                        }
                                        return actions;
                                    }

                                    if(return_actions) {
                                        points_done( getActions() );
                                        return;
                                    } else {
                                        grid_area.onClick = function () {
                                            var actions = getActions();

                                            var menu = new CrsaContextMenu(actions, this.$element);

                                            menu.showForElement(this.$element, this.$element.closest('.page'));
                                        }

                                        grid_right.setIsNotSet(!active_points);
                                        grid_left.setIsNotSet(!active_points);
                                        grid_top.setIsNotSet(!active_points);
                                        grid_bottom.setIsNotSet(!active_points);

                                        grid_right.area_name_point = grid_area;
                                        grid_left.area_name_point = grid_area;
                                        grid_top.area_name_point = grid_area;
                                        grid_bottom.area_name_point = grid_area;
                                    }
                                }

                                points_done(points);

                            })
                            //var active_rules_list = active_rules.getActiveNodes();
                            //p.show('Get Active CSS Rules');
                        }
                    } else {
                        if(return_actions) {
                            var actions = [
                                {
                                    label: 'Edit grid...',
                                    action: function () {
                                        editGrid(pgel);
                                    }
                                }
                            ]
                            points_done(actions)
                        }
                    }
                }
            }
        }

        pinegrow.addEventHandler('on_get_menu_points_with_callback', function(page, pgel, menu, points_done) {
            getMenuPointsWithCallback(page, pgel, menu, points_done, false);
        })

        pinegrow.addEventHandler('on_inserted_with_dragdrop', function(pagenull, drag_menu) {
            if(drag_menu.gridPosition) {
                var area = `${drag_menu.gridPosition.row + 1} / ${drag_menu.gridPosition.col + 1} / ${drag_menu.gridPosition.row + 2} / ${drag_menu.gridPosition.col + 2}`;

                var func = function(pgel) {
                    if(f.mode == 'css') {
                        //var active_rules = new PgActiveRules(pgel);
                        var active_css_rule = null;//pgFindCSSRuleWithProps(active_rules, ['grid-area', 'grid-column-start', 'grid-column-end', 'grid-row-start', 'grid-row-end']);

                        pgCSSGridSetGridAreaForPgel(pgel, active_css_rule, area);
                        pgCSSGridTriggerChangeEvent(pgel, active_css_rule, "Set position in grid");
                    } else {
                        // 'dm'
                    }
                }

                drag_menu.inserted_elements.forEachElement(function(pgel) {
                    func(pgel);
                })
                drag_menu.moved_elements.forEachElement(function(pgel) {
                    func(pgel);
                })
            }
        })

        var editGrid = function(pgel, $target) {

            //bubiMagic();
            //return;

            force_css_grid_control_for_pgel = pgel;

            var q = new PgQuickCSSGridControlView(pgel, $target || $('body'));

            updateCSSGridControls();

            q.gridView.onChange = function() {
                var pgel = force_css_grid_control_for_pgel;
                updateCSSGridControls();
            }

            q.gridView.onDeviceSizeChanged = function(page, def) {
                updateCSSGridControls();
            }

            q.gridView.onHighlightCell = function(col, row) {
                var $el = pgel.get$DOMElement();
                if($el) {
                    var $c = $el.find('> .pinegrow-ui-helper-css-grid');
                    $c.find('> .pinegrow-ui-helper-css-grid-cell-highlight').removeClass('pinegrow-ui-helper-css-grid-cell-highlight');
                    if (col !== undefined && row !== undefined) {
                        $c.find('[data-col-row="' + col + '-' + row + '"]').addClass('pinegrow-ui-helper-css-grid-cell-highlight');
                    }
                }
            }

            q.win.onDestroy = function() {
                force_css_grid_control_for_pgel = null;
                updateCSSGridControls();
            }
        }

        pinegrow.showCSSGridControl = editGrid;


        pinegrow.addEventHandler('on_key_pressed', function(page, e, status) {
            if(status.key == 'g' && status.ctrl && !status.edit) {

                status.done = true;
                var pgel = pinegrow.selectedElements.getLastSelected();
                if(pgel) {
                    var $el = pgel.get$DOMElement();
                    var grid_info = new PgCSSGridInfo($el);
                    if(grid_info.isGrid()) {
                        editGrid(pgel);
                        return;
                    }

                    $el = pgel.parent.get$DOMElement();
                    grid_info = new PgCSSGridInfo($el);
                    if(grid_info.isGrid()) {
                        editGrid(pgel.parent);
                        return;
                    }

                    editGrid(pgel);
                }
            }
        })


        //Helpers and controls
        var pgDMGetCurrentSizeSlug = function(pv) {
            if(pv && pv.currentStandardSizeDef && ('slug' in pv.currentStandardSizeDef)) return pv.currentStandardSizeDef.slug;
            return '-md';
        }

        var pgDMGetCurrentSizeDef = function(pv) {
            if(pv && pv.currentStandardSizeDef) return pv.currentStandardSizeDef;
            return null;
        }

        var parseColsRowsTemplateValue = function(str) {
            var open_re = /[\(\[]/;
            var close_re = /[\)\]]/;

            var r = [];

            var with_next = null;

            var setName = function(item, with_next) {
                item.name = with_next.replace(/[\[\]]/g, '');
                //item.name_parsed = item.name.replace(/(\-start|\-end)$/g, '');

                item.names = item.name.split(/\s+/).map(function(name) {
                    var d = {
                        name: name,
                        name_parsed: name.replace(/(\-start|\-end)$/g, '')
                    }
                    d.start = null;
                    if(name.endsWith('-start')) {
                        d.start = true;
                    } else if(name.endsWith('-end')) {
                        d.start = false;
                    }
                    return d;
                })

                item.getName = function(start) {
                    for(var i = 0; i < item.names.length; i++) {
                        if(item.names[i].start === start) return item.names[i].name_parsed;
                    }
                    return null;
                }
            }

            splitCssValue(str, false, open_re, close_re, true).forEach(function(size) {
                if(size.startsWith('[')) {
                    with_next = size;
                } else {
                    var item = {
                        size: size
                    }
                    if(with_next) {
                        setName(item, with_next);
                        with_next = null;
                    }
                    r.push(item)
                }
            })
            if(with_next) {
                var item = {
                    size: '',
                    last: true
                }
                setName(item, with_next);
                r.push(item);
            }
            return r;
        }

        var pgCSSGridParseTemplateAreas = function(areas_str) {
            if(areas_str == 'none') return null;
            var cells = [];
            (areas_str || '').split(/["']/).forEach(function(row) {
                row = $.trim(row);
                if(row.length) {
                    var row_cells = [];
                    cells.push(row_cells);
                    row.split(/\s+/).forEach(function (area) {
                        if(area == '.') area = '';
                        row_cells.push(area);
                    })
                }
            })
            return cells;
        }

        var pgGetAreasFromCells = function(cells, only_if_has) {
            var has_areas = false;
            var areas = '';
            cells.forEach(function(row) {
                if(areas.length) {
                    areas += ' ';
                }
                areas += `"`;
                areas += row.map(function(a) {
                    if($.trim(a).length == 0) {
                        a = '.';
                    }
                    if(a != '.') has_areas = true;
                    return a;
                }).join(' ');
                areas += `"`;
            })
            if(!has_areas && only_if_has) return null;
            return areas;
        }

        var hasStyleAttributeProps = function(pgel, props) {
            for(var i = 0; i < props.length; i++) {
                if(pgel.getInlineStylePropertyValue(props[i])) return true;
            }
            return false;
        }

        var pgFindCSSRuleWithProps = function(active_rules, props, force_sel_rule) {
            active_rules = active_rules || pinegrow.pgActiveRules.getActiveNodes();

            if(!active_rules) return null;

            var rule = null;

            if(force_sel_rule) {
                rule = pinegrow.getSelectedRule();
                if(rule && active_rules.nodes_to_show.indexOf(rule) >= 0) {
                    var cs = pgcssh.getStylesheetForNode(rule);
                    if(!cs.elementStyle) {
                        return rule;
                    }
                }
            }

            if(active_rules.nodes_to_show) {
                for (var i = 0; i < active_rules.nodes_to_show.length; i++) {
                    //has grid props?
                    for (var j = 0; j < active_rules.nodes_to_show[i].node.nodes.length; j++) {
                        var n = active_rules.nodes_to_show[i].node.nodes[j];
                        if (n.type == 'decl' && props.indexOf(n.prop) >= 0) {
                            rule = active_rules.nodes_to_show[i].node;
                            return rule;
                        }
                    }
                }
            }
            return null;
        }

        var pgGetCSSRulePropValue = function(rule, prop) {
            for(var j = 0; j < rule.nodes.length; j++) {
                var n = rule.nodes[j];
                if(n.type === 'decl' && n.prop === prop) {
                    return n.value;
                }
            }
        }

        var pgGetCurrentRule = function(rule) {
            if(rule) {
                var cs = pgcssh.getStylesheetForNode(rule);
                if (cs && pgcssh.getRootNodeFor(rule) !== cs.getRootNode()) {
                    var i = pgcssh.getNodeIndices(rule);
                    var new_rule = pgcssh.getNodeByIndices(cs.getRootNode(), i);
                    if (new_rule) {
                        rule = new_rule;
                    }
                }
            }
            return rule;
        }

        var pgSetCSSPropForActiveRule = function(pgel, active_rule, prop, val, skip_apply_code) {
            var map = {
                'gap': 'grid-gap',
                'column-gap': 'grid-column-gap',
                'row-gap': 'grid-row-gap'
            }
            if(active_rule) {
                //var pgStylesheet = obj.pgStylesheet;
                //var rule = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.getRootNode(), obj.indices);
                var cs = pgcssh.getStylesheetForNode(active_rule);

                if(map[prop] && pgGetCSSRulePropValue(active_rule, map[prop])) {
                    //delete depreceated prop
                    cs.ruleValueChanged(active_rule, map[prop], null, null, false, 'cssgridcontrol', false, null, null, null, null, true);
                }
                cs.ruleValueChanged(active_rule, prop, val, function() {
                    //on done
                    if(cs.inline) {
                        //updating inline css rebuilds the css rules.
                    }
                }, false, 'cssgridcontrol', false, null, null, null, null, skip_apply_code);
            } else {
                if(map[prop]) {
                    pgel.setInlineStylePropertyValue(map[prop], null);
                }
                pgel.setInlineStylePropertyValue(prop, val);
                /*if(changed_event) {
                    pinegrow.changeManager.didChange(pgel, typeof changed_event == 'string' ? changed_event : 'Set inline style', info);
                }*/
            }
        }

        var pgCSSGridSetGridAreaForPgel = function(pgel, css_rule, val) {

            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-column', null, false);
            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-row', null, false);

            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-column-start', null, false);
            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-column-end', null, false);
            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-row-start', null, false);
            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-row-end', null, false);

            pgSetCSSPropForActiveRule(pgel, css_rule, 'grid-area', val);
        }

        var pgCSSGridTriggerChangeEvent = function(pgel, active_rule, name, edited_by, info) {
            if(active_rule) {
                var cs = pgcssh.getStylesheetForNode(active_rule);
                pinegrow.dispatchEvent('on_css_node_updated', null, {
                    node: active_rule,
                    stylesheet: cs,
                    edited_by: edited_by,
                    //undoCombineKey: 'ruleValueChanged_' + declProp,
                    undoRuleName: name
                });
            } else {
                pinegrow.changeManager.didChange(pgel, name, info);
                //A special event, caught in the Style panel to update the panel
                pinegrow.dispatchEvent('on_css_attribute_style_changed', pgel.getPage(), pgel);
            }
        }

        var dmSetGridItemProp = function(pgel, prop, value) {
            var grid_skill = pgDmStore.getSkillByKey('grid.item');
            pgDmStore.setSkillParam(pgel, grid_skill, prop, value);
        }

        var dmGetGridItemProp = function(pgel, prop) {
            var grid_skill = pgDmStore.getSkillByKey('grid.item');
            return pgDmStore.getSkillParam(pgel, grid_skill, prop);
        }

        var PgDmCSSGridControlView = function(pgel, pv) {

            pv = pv || pinegrow.getSelectedPage().activeView;
            var _this = this;

            this.mode = f.mode; //or css

            var $element = $('<div class="pg-dm-grid-view"></div>');
            var view = new PgUIView($element, 'pgdmgrid');

            this.view = view;
            this.pageView = pv;

            $element.addClass(this.mode);

            var page = pgel.getPage();

            var delayedTasks = new PgDelayedTasks();

            this.setDeviceDef = function(def) {
                this.dmSizeDef = def;
                this.dmSizeSlug = def ? def.slug : '-md';
            }

            this.setDeviceDef(pv ? pgDMGetCurrentSizeDef(pv) : null);

            var all_device_sizes = [];
            page.callFrameworkHandler('on_get_device_sizes', all_device_sizes);

            //change, only return current size
            var findDMCssVarThatAffectsCurrentSize = function(var_template) {
                var current_idx = -1;

                for(var i = 0; i < all_device_sizes.length; i++) {
                    if(all_device_sizes[i].size == _this.dmSizeDef.size) {
                        current_idx = i;
                    }
                }

                for(var i = current_idx; i == current_idx; i--) {
                    var current_slug = all_device_sizes[i].slug;
                    var prop = var_template.replace('#SLUG', current_slug);
                    var val = pgel.getInlineStylePropertyValue(prop);
                    if(val) {
                        return {
                            val: val,
                            prop: prop,
                            slug: current_slug,
                            size: all_device_sizes[i].size
                        }
                    }
                }
                return null;
            }

            var getAllDMGridDeviceSizes = function() {
                var r = [];
                var vars = ['c', 'r', 'a'];
                var t = '--g#SLUG-';
                for(var i = 0; i < all_device_sizes.length; i++) {
                    var has = false;
                    var prop = t.replace('#SLUG', all_device_sizes[i].slug);
                    for(var n = 0; n < vars.length; n++) {
                        if(pgel.getInlineStylePropertyValue(prop + vars[n])) {
                            has = true;
                            break;
                        }
                    }
                    if(has) {
                        r.push(all_device_sizes[i].size);
                    }
                }
                return r;
            }

            var $selector = $('<div class="pg-dm-grid-view-selector"></div>').appendTo($element);
            var $grid_c = $('<div class="pg-dm-grid-view-container"></div>').appendTo($element);
            var $options = $('<div class="pg-dm-grid-view-options"></div>').appendTo($grid_c);
            var $grid = $('<div class="pg-dm-grid-view-grid"></div>').appendTo($grid_c);

            crsaHandleExternalLinks($element);

            var cols = [];
            var rows = [];
            var cells = []; //rows of cols

            var last_col_line = null;
            var last_row_line = null;

            var auto_col = null;
            var auto_row = null;
            var gap = null;
            var gap_row = null;
            var gap_eq = false;
            var display = null;

            var has_grid = false;

            var area_colors = {}
            var all_areas = [];

            var getAllAreas = function() {
                all_areas = [];

                cells.forEach(function(row) {
                    row.forEach(function(area) {
                        if(area && all_areas.indexOf(area) < 0) {
                            all_areas.push(area);
                        }
                    })
                })
            }

            var active_rules = null;

            var active_rule = null;

            var findCSSRuleForThisGrid = function(ar, pgel) {
                var grid_props = ['grid-template-rows', 'grid-template-columns', 'grid', 'grid-template', 'grid-template-areas', 'grid-auto-columns', 'grid-auto-rows'];
                if(hasStyleAttributeProps(pgel, grid_props)) return null;
                return pgFindCSSRuleWithProps(ar, grid_props, /*pinegrow.selectedElements.isSelectedElement(pgel) */ false);
            }



            //console.log(active_rule);

            var css = {}

            var getPropFromRule = function(rule, prop) {
                return pgGetCSSRulePropValue(rule, prop);
            }

            var getCSSPropsFromActiveRule = function() {
                var map = {
                    'gap':  'grid-gap',
                    'column-gap': 'grid-column-gap',
                    'row-gap': 'grid-row-gap'
                }
                $.each(css, function(prop, val) {
                    if(active_rule) {
                        css[prop] = getPropFromRule(active_rule, prop);
                        if(!css[prop] && map[prop]) {
                            css[prop] = getPropFromRule(active_rule, map[prop]);
                        }
                    } else {
                        css[prop] = pgel.getInlineStylePropertyValue(prop);
                        if(!css[prop] && map[prop]) {
                            css[prop] = pgel.getInlineStylePropertyValue(map[prop]);
                        }
                    }
                })
            }



            var getDMVarForProp = function(prop, slug) {
                if(slug === undefined) slug = _this.dmSizeSlug;
                switch(prop) {
                    case 'grid-template-columns':
                        return `--g${slug}-c`;
                    case 'grid-template-rows':
                        return `--g${slug}-r`;
                    case 'grid-template-areas':
                        return `--g${slug}-a`;
                    case 'grid-gap':
                        return `--g${slug}-gap`;
                }
                return null;
            }

            var dmSetGridProp = function(prop, value) {
                var grid_skill = pgDmStore.getSkillByKey('grid');
                pgDmStore.setSkillParam(pgel, grid_skill, prop, value);
            }

            var dmGetGridProp = function(prop) {
                var grid_skill = pgDmStore.getSkillByKey('grid');
                return pgDmStore.getSkillParam(pgel, grid_skill, prop);
            }



            //this.

            var makeAreaColors = function() {
                getAllAreas();

                area_colors = {};
                /*
                var c = 1.0;
                var d = -(0.8 / (1.0 * all_areas.length));

                all_areas.forEach(function(a) {
                    area_colors[a] = `rgba(0, 152, 204, ${c})`;
                    c += d;
                    //if(c < 0.4) c = 1.0;
                })*/

                var i = 0;
                chroma.scale(['#0497cc','#b2d12e' /*, '#ffc800'*/])
                    .mode('lch').colors(all_areas.length).forEach(function(c) {
                    area_colors[all_areas[i]] = c;
                    i++;
                })
            }

            var getCell = function(r, c) {
                if(r < cells.length && c < cells[r].length) {
                    return cells[r][c];
                }
                return '';
            }

            var addCol = function(c, idx) {
                cols.splice(idx, 0, c)

                var row_i = 0;
                cells.forEach(function(row) {
                    row.splice(idx, 0, getCell(row_i, idx));
                    row_i++;
                })
            }

            var addRow = function(c, idx) {
                rows.splice(idx, 0, c);

                var row = [];
                for(var i = 0; i < cols.length; i++) {
                    row.push(getCell(idx, i));
                }
                cells.splice(idx, 0, row);
            }

            var removeCol = function(idx) {
                cols.splice(idx, 1);

                cells.forEach(function(row) {
                    row.splice(idx, 1);
                })
            }

            var removeRow = function(idx) {
                rows.splice(idx, 1);
                cells.splice(idx, 1);
            }

            var def_new_size_col = '1fr';
            var def_new_size_row = '100px';

            var md = function(cls, $d) {
                return $('<div class="pg-dm-grid-view-grid-' + cls + '"></div>').appendTo($d || $grid);
            }

            var tester = new PgCSSStatementTester();

            var validate = function(input, prop, value) {
                var ok = tester.test(prop, value);
                input.setError(ok ? null : 'Invalid value');
                return ok;
            }

            var updateGrid = function() {
                if(!makeSurePgelIsOk()) return;

                $options.empty();
                $grid.empty();
                $selector.empty();

                $grid.css('min-height', '0');

                var size_options = [
                    {key: '100px', name: '100px', html: '100px <small>Fixed size</small>'},
                    {key: '1fr', name: '1fr', html: '1fr <small>Divide available space</small>'},
                    {key: 'auto', name: 'auto', html: 'auto <small>Content size</small>'},
                    {key: 'minmax(100px, 1fr)', name: 'minmax(100px, 1fr)', html: 'minmax(100px, 1fr) <small>Min 100px, max 1fr</small>'},
                    {key: 'minmax(100px, auto)', name: 'minmax(100px, auto)', html: 'minmax(100px, auto) <small>Min 100px, max the content size</small>'},
                    {key: 'fit-content(40%)', name: 'fit-content(40%)', html: 'fit-content(40%) <small>Content size, but not larger than 40%</small>'},
                    {key: 'repeat(3, 200px)', name: 'repeat(3, 200px)', html: 'repeat(3, 200px) <small>Repeat 3 times</small>'},
                    {key: 'repeat(auto-fill, minmax(100px, 1fr))', name: 'repeat(auto-fill, minmax(100px, 1fr))', html: 'repeat(auto-fill, minmax(100px, 1fr)) <small>Fill with as many cols/rows as possible</small>'},
                    {key: 'repeat(auto-fit, minmax(100px, 1fr))', name: 'repeat(auto-fit, minmax(100px, 1fr))', html: 'repeat(auto-fit, minmax(100px, 1fr)) <small>Fit the available cols/rows into the space</small>'},
                    {key: 'none', name: 'none', html: 'None <small>Track template is not defined. Only valid with a single row or column.</small>'}
                ]

                var auto_size_options = [
                    {key: '100px', name: '100px', html: '100px <small>Fixed size</small>'},
                    {key: '1fr', name: '1fr', html: '1fr <small>Divide available space</small>'},
                    {key: 'auto', name: 'auto', html: 'auto <small>Content size</small>'},
                    {key: 'minmax(100px, 1fr)', name: 'minmax(100px, 1fr)', html: 'minmax(100px, 1fr) <small>Min 100px, max 1fr</small>'},
                    {key: 'minmax(100px, auto)', name: 'minmax(100px, auto)', html: 'minmax(100px, auto) <small>Min 100px, max the content size</small>'},
                    {key: 'fit-content(40%)', name: 'fit-content(40%)', html: 'fit-content(40%) <small>Content size, but not larger than 40%</small>'}
                ]

                var addInput = function(size, $c, list, idx) {
                    var input = new PgInputField('size', {
                        type: 'slider',
                        name: 'W',
                        slider_def_unit: 'px',
                        slider_min: 0,
                        without_text: true,
                        options: size_options
                    });

                    input.setValue(size);
                    validate(input, 'grid-template-columns', size)

                    input.on('input change', function(e) {
                        var size = input.getValue();

                        if(validate(input, 'grid-template-columns', size)) {
                            list[idx].size = size;
                            onChange(e.type === 'change');
                        }
                    })

                    input.appendTo($c);
                }

                has_grid = cols.length > 0 && rows.length > 0;

                //Options
                if(_this.mode == 'dm') {
                    var $d = md('devices', $selector);
                    var html = `<label>Grid definitions for device sizes:</label><ul>`;

                    var grid_sizes = getAllDMGridDeviceSizes();

                    var tabsToolbar = new PgButtonToolbar();
                    $d.prepend(tabsToolbar.$element.addClass('pg-tabs-toolbar'));

                    var deviceTabs = new PgToolbarTabButtons(tabsToolbar, 'tabs');

                    for(var i = 0; i < all_device_sizes.length; i++) {
                        var def = all_device_sizes[i];
                        var has = grid_sizes.indexOf(def.size) >= 0;
                        var current = _this.dmSizeDef && (def.size == _this.dmSizeDef.size);

                        var tab = deviceTabs.addTab(def.name_short, null, def, current).setClosable(false);

                        if(has) {
                            tab.$element.append('<i class="icon icon-check"></i>')
                        }
                        html += `<li><button class="btn pg-button ${current ? 'pg-button-primary' : ''}" data-idx="${i}">${def.name_short}${has ? '<i class="icon icon-check"></i>' : ''}</button></li>`;
                    }

                    deviceTabs.onTabSelected = function(tab) {
                        var def = tab.data;
                        _this.setDeviceDef(def);
                        _this.showGrid();

                        var page = pinegrow.getSelectedPage();
                        if(page) {
                            page.view.setActiveViewSizeWithDef(def);
                        }
                        if(_this.onDeviceSizeChanged) {
                            _this.onDeviceSizeChanged(page, def);
                        }
                    }

                    html += '</ul>';

                    //$d.html(html);
                    /*
                     $d.on('click', 'button', function(e) {
                     var $a = $(this);
                     var idx = parseInt($a.attr('data-idx'));
                     _this.setDeviceDef(all_device_sizes[idx]);
                     _this.showGrid();
                     return false;
                     })*/
                }


                    if (_this.mode == 'css') {
                        /*
                        var input_rule = new PgInputField('rule', {
                            type: 'displayhtmlwithlabel',
                            name: 'CSS Rule',
                            html: `<span class="pgBold">${active_rule ? pgcssh.getRuleDisplayName(active_rule) : 'Style attribute'}</span> (automatic)`,
                            without_text: false
                        });

                        input_rule.appendTo($options);
                        */

                        var shown_nodes = active_rules.getShownNodes();

                        var input_sel_rule = new PgInputField('rule', {
                            type: 'select',
                            name: 'CSS Rule',
                            without_text: false,
                            show_empty: false,
                            options: function() {
                                var i = 0;
                                var r = [
                                    {key: 'item-1', name: 'Style attribute'}
                                ]
                                r = r.concat( shown_nodes.map(function(rule) {
                                    return {
                                        key: `item${i++}`,
                                        name: pgcssh.getRuleDisplayName(rule)
                                    }
                                }))
                                return r;
                            }
                        });

                        var sel_idx = -1;
                        for(var i = 0; i < shown_nodes.length; i++) {
                            if(shown_nodes[i] === active_rule) {
                                sel_idx = i;
                                break;
                            }
                        }
                        input_sel_rule.setValue(`item${sel_idx}`);

                        input_sel_rule.on('change', function (e) {
                            var idx = parseInt(input_sel_rule.getValue().replace('item', ''));
                            idx = Math.min(idx, shown_nodes.length -1);
                            var rule = idx >= 0 ? shown_nodes[idx] : 'style_attribute';
                            _this.showGrid(rule);
                        })

                        input_sel_rule.appendTo($selector);

                        //$options.append('<hr/>');
                    }

                if(has_grid) {

                    //$grid.show();
                    var $gaps = $(`<div class="pg-css-grid-field-gaps crsa-field"><label>Gap col/row</label><div></div></div>`).appendTo($options);

                    $gaps = $gaps.find('div');

                    var input_gap = new PgInputField('gap', {
                        type: 'slider',
                        name: 'Gap',
                        slider_def_unit: 'px',
                        slider_min: 0,
                        without_text: true
                    });

                    input_gap.setValue(gap);
                    validate(input_gap, 'gap', gap)

                    input_gap.on('change input', function (e) {
                        gap = input_gap.getValue();
                        if(gap_eq) {
                            input_gap_row.setValue(gap);
                            gap_row = gap;
                        }
                        if(validate(input_gap, 'gap', gap)) {
                            onChange(e.type === 'change');
                        }
                    })

                    input_gap.appendTo($gaps);

                    var bm = new PgToggleButtonMaker();

                    var input_gap_eq = new PgInputField('gap-eq', {
                        type: 'select',
                        name: 'Gap eq',
                        show_empty: true,
                        toggle_buttons: true,
                        without_text: true,
                        options: [
                            {
                                key: 'true',
                                name: 'Equal gaps',
                                html: bm.makeIcon('same', {title: 'Make gaps the same.'})
                            }
                        ]
                    });

                    input_gap_eq.setValue(gap_eq ? 'true' : null);
                    input_gap_eq.on('change', function (e) {
                        gap_eq = input_gap_eq.getValue() === 'true';
                        if(gap_eq) {
                            var val = gap || gap_row;
                            input_gap.setValue(val);
                            input_gap_row.setValue(val);
                            gap = gap_row = val;
                        }
                        onChange(true);
                    })
                    input_gap_eq.appendTo($gaps);


                    var input_gap_row = new PgInputField('gap-row', {
                        type: 'slider',
                        name: 'Gap row',
                        slider_def_unit: 'px',
                        slider_min: 0,
                        without_text: true
                    });

                    input_gap_row.setValue(gap_eq ? gap : gap_row);
                    validate(input_gap_row, 'gap', gap_row)

                    input_gap_row.on('change input', function (e) {
                        gap_row = input_gap_row.getValue();
                        if(gap_eq) {
                            input_gap.setValue(gap_row);
                            gap = gap_row;
                        }
                        if(validate(input_gap_row, 'gap', gap_row)) {
                            onChange(e.type === 'change');
                        }
                    })

                    input_gap_row.appendTo($gaps);


                    var input_auto_col = new PgInputField('auto_col', {
                        type: 'slider',
                        name: 'Auto Columns',
                        slider_def_unit: 'px',
                        slider_min: 0,
                        without_text: false,
                        options: auto_size_options
                    });

                    input_auto_col.setValue(auto_col);
                    validate(input_auto_col, 'grid-auto-columns', auto_col)

                    input_auto_col.on('change input', function (e) {
                        auto_col = input_auto_col.getValue();

                        if(validate(input_auto_col, 'grid-auto-columns', auto_col)) {
                            onChange(e.type == 'change');
                        }
                    })

                    input_auto_col.appendTo($options);


                    var input_auto_row = new PgInputField('auto_row', {
                        type: 'slider',
                        name: 'Auto Rows',
                        slider_def_unit: 'px',
                        slider_min: 0,
                        without_text: false,
                        options: auto_size_options
                    });

                    input_auto_row.setValue(auto_row);
                    validate(input_auto_row, 'grid-auto-rows', auto_row)

                    input_auto_row.on('change input', function (e) {
                        auto_row = input_auto_row.getValue();
                        if(validate(input_auto_row, 'grid-auto-rows', auto_row)) {
                            onChange(e.type == 'change');
                        }
                    })

                    input_auto_row.appendTo($options);

                    if(!auto_row && !auto_col) {
                        input_auto_col.$field.hide();
                        input_auto_row.$field.hide();

                        var $adv = $('<p style="text-align:right;margin-bottom:0;"><a href="#" class="help-link show-more">Auto columns and rows...</a></p>').appendTo($options);
                        $adv.on('click', 'a', function(e) {
                            e.preventDefault();
                            input_auto_col.$field.show();
                            input_auto_row.$field.show();
                            $adv.hide();
                        })
                    }


                }

                if(!has_grid) {

                    //$grid.hide();

                    var copyGrid = function(from_slug, to_slug) {
                        var css = $.extend({}, grid_css_fields);
                        $.each(css, function(prop, val) {
                            var cssvar = getDMVarForProp(prop, from_slug);
                            if(cssvar) {
                                var val = pgel.getInlineStylePropertyValue(cssvar);
                                var cssvar_to = getDMVarForProp(prop, to_slug);
                                if(val) {
                                    pgel.setInlineStylePropertyValue(cssvar_to, val);
                                }
                            }
                        })
                    }

                    var closest_size = null;

                    if(_this.mode == 'dm') {
                        var idx = 0;
                        for (var i = 0; i < all_device_sizes.length; i++) {
                            if (_this.dmSizeDef.size == all_device_sizes[i].size) {
                                idx = i;
                                break;
                            }
                        }

                        for (var i = idx - 1; idx >= 0; idx--) {
                            if (grid_sizes.indexOf(all_device_sizes[idx].size) >= 0) {
                                closest_size = all_device_sizes[idx];
                                break;
                            }
                        }

                        if (!closest_size) {
                            for (var i = idx + 1; idx < all_device_sizes.length; idx++) {
                                if (grid_sizes.indexOf(all_device_sizes[i].size) >= 0) {
                                    closest_size = all_device_sizes[i];
                                    break;
                                }
                            }
                        }
                    }

                    var create_str, button_str;
                    if(_this.mode == 'dm') {
                        create_str = `<p>CSS Grid for device size ${_this.dmSizeDef.name_short} is not defined. Would you like to define it?</p>`;
                        button_str = `Create grid for ${_this.dmSizeDef.name_short}`;
                    } else {
                        create_str = `<p>CSS Grid is not defined for the selected CSS rule. Would you like to define it?</p>`;
                        button_str = `Create CSS Grid`;
                    }
                    $(create_str).appendTo($options);
                    $element.addClass('no-grid');

                    var $new = $(`<button class="btn pg-button create-grid">${button_str}</button>`).appendTo($options);

                    $new.on('click', function(e) {
                        e.preventDefault();
                        if(closest_size) {
                            copyGrid(closest_size.slug, _this.dmSizeDef.slug);
                            pinegrow.showQuickMessage('Grid copied from ' + closest_size.name_short + '.');
                            _this.showGrid();
                            onChange(true);
                        } else {
                            addCol({size: def_new_size_col});
                            addCol({size: def_new_size_col});
                            addCol({size: def_new_size_col});
                            addRow({size: def_new_size_row});
                            updateGrid();
                            onChange(true);
                            _this.showGrid(/*active_rule ? active_rule : 'style_attribute'*/);
                        }

                    })

                    var $help = $('<p style="margin-top:40px;"><a href="' +  help_url + '" class="external help-link">Learn about CSS Grid...</a></p>').appendTo($options);

                    /*
                     if(grid_sizes.length) {
                     var $clone = $(`<button class="btn pg-button pg-button-primary">Clone grid from</button>`).appendTo($options);

                     var clone_from = new PgInputField('clonefrom', {
                     type: 'select',
                     name: 'Gap',
                     without_text: true,
                     show_empty: false,
                     options: grid_sizes.map(function(s) {
                     return {key: s, name: s.toUpperCase()}
                     })
                     });



                     if(closest_size) {
                     clone_from.setValue(closest_size.size);
                     }

                     clone_from.appendTo($options);
                     }*/

                } else {
                    $element.removeClass('no-grid');

                    var $rc = null;//md('rowc', $grid);

                    var $corner = md('corner');
                    $corner.html(`<div><span>Columns</span><span>Rows</span></div>`);


                    var col_i = 0;
                    cols.forEach(function (c) {
                        var size = c.size;
                        var $c = md('col');
                        if (col_i == 0 || true) {
                            $(`<i class="add-col add-col-first icon icon-plus" data-idx="${col_i}">`).appendTo($c);
                        }

                        addInput(size, $c, cols, col_i);

                        $(`<i class="del-col icon icon-bin" data-idx="${col_i}">`).appendTo($c);
                        $(`<i class="add-col icon icon-plus" data-idx="${col_i + 1}">`).appendTo($c);
                        col_i++;
                    })

                    makeAreaColors();

                    var row_i = 0;
                    rows.forEach(function (r) {
                        var size = r.size;
                        //var $rc = md('rowc', $grid);
                        var $r = md('row');

                        if (row_i == 0 || true) {
                            $(`<i class="add-row add-row-first icon icon-plus" data-idx="${row_i}">`).appendTo($r);
                        }
                        addInput(size, $r, rows, row_i);

                        $(`<i class="del-row icon icon-bin" data-idx="${row_i}">`).appendTo($r);
                        $(`<i class="add-row icon icon-plus" data-idx="${row_i + 1}">`).appendTo($r);

                        for (var i = 0; i < cols.length; i++) {
                            (function (row_i, i) {
                                var $e = md('cell', $rc);
                                var a = cells[row_i][i];

                                $e.attr('data-area', a);
                                $e.attr('data-col', i);
                                $e.attr('data-row', row_i);

                                if(a) {
                                    $e.addClass('has-area');
                                }

                                var input = new PgInputField('size', {
                                    type: 'text',
                                    name: 'W',
                                    without_text: true,
                                    placeholder: 'area name',
                                    fill_list_on_open: true,
                                    options: function () {
                                        getAllAreas();
                                        var r = [];

                                        all_areas.forEach(function (a) {
                                            if (a.length) {
                                                r.push({key: a, name: a});
                                            }
                                        })
                                        return r;
                                    }
                                });

                                input.setValue(a);

                                input.on('change', function (e) {
                                    var a = input.getValue();
                                    cells[row_i][i] = a;
                                    $e.attr('data-area', a);
                                    if(a) {
                                        $e.addClass('has-area');
                                    } else {
                                        $e.removeClass('has-area')
                                    }

                                    if (!(a in area_colors)) {
                                        makeAreaColors();
                                        $grid.find('.pg-dm-grid-view-grid-cell').each(function (i, e) {
                                            var $e = $(e);
                                            var a = $e.attr('data-area');
                                            $e.css({
                                                'background-color': area_colors[a]
                                            })
                                        })
                                    } else {
                                        $e.css({
                                            'background-color': area_colors[a]
                                        })
                                    }
                                    onChange(e.type == 'change');
                                })

                                input.appendTo($e);

                                //$e.html(a);
                                $e.css({
                                    'background-color': area_colors[a]
                                })
                            })(row_i, i)
                        }
                        row_i++;
                    })

                    $grid.css({
                        'grid-template-columns': `minmax(120px, 1fr) repeat(${cols.length}, minmax(80px, 1fr))`,
                        'grid-template-rows': `50px repeat(${rows.length}, minmax(70px, 70px))`
                    })

                    updateLines();
                    updateLines(); //maybe lines caused the scrollbars to appear
                }


            }

            var updateLines = function() {
                //update line names
                if(!$grid.is(':visible')) return;

                $grid.css('min-height', '0');

                var gs = window.getComputedStyle($grid.get(0));

                var gcolgap = parseGapValue(gs.gridColumnGap);
                var growgap = parseGapValue(gs.gridRowGap);
                var gw = parseFloat(gs.width.replace('px', ''));
                var gh = parseFloat(gs.height.replace('px', ''));

                var gcols = gs.gridTemplateColumns.split(' ').map(function(v) {
                    return parseFloat(v.replace('px', ''))
                });
                var grows = gs.gridTemplateRows.split(' ').map(function(v) { return parseFloat(v.replace('px', ''))});

                var max_col_line_w = 0;

                var showLine = function(col, i) {
                    var $cline = $grid.find(`.pg-dm-grid-view-grid-col-line[data-idx="${i-1}"]`);
                    if($cline.length === 0) {
                        $cline = md('col-line', $grid);
                        var str = `<hr><span>${i}</span>`;
                        if (col.name) {
                            str += `<linename>${col.name}</linename>`;
                            $cline.addClass('has-name');
                        }
                        $cline.html(str);
                        $cline.attr('data-idx', i - 1);
                        addTooltip($cline, 'Click to edit the track name.');
                    }
                    $cline.attr('style', `left:${x}px;min-width:${gh - 86}px;`);
                    x += gcols[i] + gcolgap;

                    var w = $cline.width();
                    if(max_col_line_w < w) max_col_line_w = w;
                }

                var x = gcols[0] + gcolgap/2;
                for(var i = 1; i < gcols.length; i++) {
                    showLine(cols[i-1], i);
                }
                showLine(last_col_line, gcols.length);

                $grid.css('min-height', (max_col_line_w + 86) + 'px');

                var showRow = function(row, i) {
                    var $cline = $grid.find(`.pg-dm-grid-view-grid-row-line[data-idx="${i-1}"]`);
                    if($cline.length === 0) {

                        $cline = md('row-line', $grid);

                        var str = `<hr><span>${i}</span>`;
                        if (row.name) {
                            str += `<linename>${row.name}</linename>`;
                            $cline.addClass('has-name');
                        }
                        $cline.attr('data-idx', i - 1);
                        $cline.html(str);
                        addTooltip($cline, 'Click to edit the track name.');
                    }
                    $cline.attr('style', `left:${row_x}px;top:${y}px;min-width:${gw - row_x}px;`);
                    y += grows[i] + growgap;
                }

                var row_x = gcols[0] + gcolgap + 2;
                var y = grows[0] + growgap/2;
                for(var i = 1; i < grows.length; i++) {
                    showRow(rows[i-1], i);
                }
                showRow(last_row_line, grows.length);
            }

            view.onResize = function() {
                if(has_grid && grid_initialized) {
                    updateLines();
                    updateLines(); //maybe lines caused the scrollbars to appear
                }
            }

            $grid.on('click', '.add-col', function(e) {
                var $a = $(this);
                var idx = parseInt($a.attr('data-idx'));
                var item = {size: idx < cols.length ? cols[idx].size : cols[idx-1].size};
                if(idx == cols.length) {
                    item.name = last_col_line.name;
                    last_col_line.name = null;
                }
                addCol(item, idx);
                updateGrid();
                onChange(true);
            })

            $grid.on('click', '.del-col', function(e) {
                var $a = $(this);
                var idx = parseInt($a.attr('data-idx'));
                removeCol(idx);
                updateGrid();
                onChange(true);
            })

            $grid.on('click', '.add-row', function(e) {
                var $a = $(this);
                var idx = parseInt($a.attr('data-idx'));
                var item = {size: idx < rows.length ? rows[idx].size : rows[idx-1].size};
                if(idx == rows.length) {
                    item.name = last_row_line.name;
                    last_row_line.name = null;
                }
                addRow(item, idx);
                updateGrid();
                onChange(true);
            })

            $grid.on('click', '.del-row', function(e) {
                var $a = $(this);
                var idx = parseInt($a.attr('data-idx'));
                removeRow(idx);
                updateGrid();
                onChange(true);
            })

            var editTrackName = function(cols, $a, last_col) {

                var validate = function(value) {
                    var a = value.split(' ');
                    for(var i = 0; i < a.length; i++) {
                        if(a[i].length) {
                            /*
                            if(!a[i].endsWith('-start') && !a[i].endsWith('-end')) {
                                return false;
                            }*/
                        }
                    }
                    return true;
                }

                var idx = parseInt($a.attr('data-idx'));
                var col = idx < cols.length ? cols[idx] : last_col;
                pinegrow.showPrompt(`<p>Edit the track name. Track names must end with <b>-start</b> or <b>-end</b> if you wish to use grid handles for element placement along those tracks, for example <code>main-start</code>.</p><p>Separate multiple names with spaces, for example <code>main-start content-start</code>.</p><p>Leave empty to clear the track name.</p>`, 'Edit track name', col.name || '', 'for example, content-end', null, function(value) {
                    if(value && (!validate(value))) {
                        pinegrow.showQuickError('Invalid track name(s).');
                    } else {
                        col.name = value;
                        updateGrid();
                        onChange(true);
                    }
                })
                onChange(true);
            }

            $grid.on('click', '.pg-dm-grid-view-grid-col-line', function(e) {
                var $a = $(this);
                editTrackName(cols, $a, last_col_line);
            })

            $grid.on('click', '.pg-dm-grid-view-grid-row-line', function(e) {
                var $a = $(this);
                editTrackName(rows, $a, last_row_line);
            })

            $grid.on('mouseover', '.pg-dm-grid-view-grid-cell', function() {
                var $cell = $(this);
                var col = $cell.attr('data-col');
                var row = $cell.attr('data-row');
                if(_this.onHighlightCell) {
                    _this.onHighlightCell(col, row);
                }
            })

            $grid.on('mouseout', '.pg-dm-grid-view-grid-cell', function() {
                if(_this.onHighlightCell) {
                    _this.onHighlightCell();
                }
            })

            var onChange = function(change_event) {

                if(!makeSurePgelIsOk()) return;

                var apply = function() {
                    var list = [];
                    var css_vals = _this.getCSS();
                    $.each(css_vals, function (prop, val) {
                        if(val === 'none') val = null;
                        if(val) {
                            val = val.replace(/"/g, "'");
                        }
                        list.push({prop: prop, val: val});
                    });

                    /*
                    if(css_vals['gap'] !== null) {
                        //remove other gaps
                        pgSetCSSPropForActiveRule(pgel, active_rule, 'column-gap', null, true);
                        pgSetCSSPropForActiveRule(pgel, active_rule, 'row-gap', null, true);
                    } else if(css_vals['row-gap'] !== null || css_vals['column-gap'] !== null) {
                        //remove gap
                        pgSetCSSPropForActiveRule(pgel, active_rule, 'gap', null, true);
                    }

                     */

                    for(var i = 0; i < list.length; i++) {
                        if(_this.mode === 'css') {
                            pgSetCSSPropForActiveRule(pgel, active_rule, list[i].prop, list[i].val, i !== list.length - 1 /* only apply code on the last one */);
                        } else {
                            // 'dm'
                            dmSetGridProp(list[i].prop, list[i].val);
                            /*
                            var cssvar = getDMVarForProp(list[i].prop);
                            if(cssvar) {
                                pgel.setInlineStylePropertyValue(cssvar, list[i].val);
                            }*/
                        }
                    }
                }

                active_rule = pgGetCurrentRule(active_rule);

                if(!change_event) {
                    apply();
                } else {
                    if (!active_rule) {
                        //pinegrow.makeChanges(pgel, 'Set CSS grid', function () {
                            apply();
                        //})
                        //pinegrow.selectedElements.reselectDelayed();
                        //if(pinegrow.pgActiveRules)
                        //pinegrow.dispatchEvent('on_css_node_updated', null, {node: rule, stylesheet: this, edited_by: edited_by});
                    } else {
                        apply();
                    }
                    pgCSSGridTriggerChangeEvent(pgel, active_rule, 'Set CSS grid', null, {source: _this})

                    //hack, events not working properly, trigger CSS panel refresh
                    delayedTasks.run('refreshCssPanel', function() {
                        var cs = null;
                        if(active_rule) {
                            cs = pgcssh.getStylesheetForNode(active_rule);
                        }
                        pinegrow.activeCSSView.refreshCurrentDisplay(cs);
                    }, 1000);

                }


                if(_this.onChange) {
                    _this.onChange();
                }
            }

            var grid_css_fields = {
                'grid-template-columns': null,
                'grid-template-rows': null,
                'grid-template-areas': null,
                'grid-auto-columns': null,
                'grid-auto-rows': null,
                'gap': null,
                'column-gap': null,
                'row-gap': null,
                'display': null
            }

            var makeSurePgelIsOk = function() {
                var o_pgel = pgel;

                if(pgel) {
                    pgel = pgel.replacedWith || getPgNodeByPgId(pgel.getId());
                }
                if(pgel && pgel !== o_pgel) {
                    active_rules && active_rules.destroy();
                    active_rules = new PgActiveRules(pgel);
                }
                if(!pgel) {
                    pinegrow.showQuickError('The target element doesn\'t exist any more. Please close the CSS Grid editor.');
                    return false;
                }
                return true;
            }

            var grid_initialized = false;

            active_rules = new PgActiveRules(pgel);

            this.showGrid = function(force_rule) {

                if(!makeSurePgelIsOk()) return;

                grid_initialized = false;

                var withActiveRules = function() {

                    cols = [];
                    rows = [];
                    cells = []; //rows of cols

                    auto_col = null;
                    auto_row = null;
                    gap = null;
                    gap_row = null;

                    css = $.extend({}, grid_css_fields);

                    grid_initialized = true;

                    var ar = active_rules.getActiveNodes();
                    //console.log(ar);

                    active_rule = null;

                    if(this.mode === 'css') {
                        var done = false;
                        if(force_rule) {
                            if (force_rule === 'style_attribute') {
                                active_rule = null;
                                done = true;
                            } else {
                                if (active_rules.getShownNodes().indexOf(force_rule) >= 0) {
                                    active_rule = force_rule;
                                    done = true;
                                }
                            }
                        }
                        if(!done) {
                            active_rule = findCSSRuleForThisGrid(ar, pgel);
                        }
                    }

                    if(this.mode === 'css') {
                        getCSSPropsFromActiveRule();
                    } else {
                        // 'dm'
                        $.each(css, function(prop, val) {
                            css[prop] = dmGetGridProp(prop);
                            /*
                             var cssvar = getDMVarForProp(prop, '#SLUG');
                             if(cssvar) {
                             var info = findDMCssVarThatAffectsCurrentSize(cssvar);
                             if(info) {
                             css[prop] = info.val;
                             } else {
                             css[prop] = null;
                             }
                             //css[prop] = pgel.getInlineStylePropertyValue(cssvar);
                             }*/
                        })
                    }

                    var cols_str, rows_str, areas_str, auto_col_str, auto_row_str, gap_str, gap_column_str, gap_row_str;
                    cols_str = css['grid-template-columns'];
                    rows_str = css['grid-template-rows'];
                    areas_str = css['grid-template-areas'];
                    auto_col_str = css['grid-auto-columns'] || null;
                    auto_row_str = css['grid-auto-rows'] || null;
                    gap_str = css['gap'] || css['column-gap'];
                    gap_column_str = css['column-gap'];
                    gap_row_str = css['row-gap'];

                    display = null;
                    if(css['display'] !== 'inline-grid') {
                        display = 'grid';
                    } else {
                        display = 'inline-grid';
                    }

                    if(!cols_str) cols_str = '';
                    if(!rows_str) rows_str = '';

                    last_col_line = {size: '', name: null};
                    last_row_line = {size: '', name: null};

                    auto_col = auto_col_str;
                    auto_row = auto_row_str;
                    gap = gap_str;
                    gap_row = gap_row_str;
                    gap_eq = !!css['gap'] || (gap === gap_row);


                    var i = 0;

                    parseColsRowsTemplateValue(cols_str).forEach(function(item) {
                        var size = item.size;
                        if(item.last) {
                            last_col_line = item;
                        } else {
                            addCol(item, i++);
                        }
                    })

                    i = 0;
                    parseColsRowsTemplateValue(rows_str).forEach(function(item) {
                        var size = item.size;
                        if(item.last) {
                            last_row_line = item;
                        } else {
                            addRow(item, i++);
                        }
                    })


                    if(cols.length == 0 && rows.length) {
                        addCol({size: 'none'});
                    } else if(rows.length == 0 && cols.length) {
                        addRow({size: 'none'});
                    }
                    var row_i = 0;
                    (areas_str || '').split(/["']/).forEach(function(row) {
                        row = $.trim(row);
                        if(row.length) {
                            var col_i = 0;
                            row.split(/\s+/).forEach(function (area) {
                                if(area == '.') area = '';
                                if (row_i < rows.length && col_i < cols.length) {
                                    cells[row_i][col_i] = area;
                                }
                                col_i++;
                            })
                            row_i++;
                        }
                    })
                    updateGrid();
                }

                active_rules.clearCallbackList();

                active_rules.updateMatchedRules(true, function() {
                    withActiveRules.call(_this);
                })


            }

            this.getCSS = function() {
                var combine = function(item) {
                    var str = item.size;
                    if(item.name) str = '[' + item.name + '] ' + str;
                    return str;
                }

                var cols_css = cols.map(combine).join(' ');
                var rows_css = rows.map(combine).join(' ');
                var areas = pgGetAreasFromCells(cells, true);
                var display_css = display;
                var gap_css = gap;
                var row_gap_css = gap_eq ? gap_css : gap_row;

                if(cols_css == 'none' || rows_css == 'none') {
                    areas = null;
                }
                if(cols_css == 'none') {
                    cols_css = null;
                }
                if(rows_css == 'none') {
                    rows_css = null;
                }

                if(cols_css && last_col_line.name) {
                    cols_css += ` [${last_col_line.name}]`;
                }
                if(rows_css && last_row_line.name) {
                    rows_css += ` [${last_row_line.name}]`;
                }
                if(cols.length == 0 || rows.length == 0) {
                    display_css = null;
                    cols_css = null;
                    rows_css = null;
                    areas = null;
                    gap_css = null;
                    row_gap_css = null;
                }
                var r = {
                    'display': display_css,
                    'grid-template-columns': cols_css,
                    'grid-template-rows': rows_css,
                    'grid-template-areas': areas,
                    'grid-auto-columns': auto_col,
                    'grid-auto-rows': auto_row,
                    'gap': null,
                    'column-gap': null,
                    'row-gap': null
                }
                if(gap_css === row_gap_css) {
                    r.gap = gap_css;
                } else {
                    r['column-gap'] = gap_css;
                    r['row-gap'] = row_gap_css;
                }
                return r;
            }

            this.showGrid();

            var onPageViewResized = function() {
                _this.showGrid(/* active_rule */);
            }

            pinegrow.addEventHandler('on_page_view_resized', onPageViewResized);
            pinegrow.addEventHandler('on_active_page_view_changed', onPageViewResized);

            var onCSSNodeChanged = function(page, obj) {
                if(_this.mode == 'css' && (!obj.edited_by || obj.edited_by != 'cssgridcontrol')) {
                    if(obj.node && obj.node.parent == active_rule) {
                        _this.showGrid( active_rule );
                    } else if(obj.stylesheet.elementStyle && !active_rule && obj.stylesheet.element === pgel) {
                        _this.showGrid( active_rule );
                    }
                }
            }

            pinegrow.addEventHandler('on_css_node_deleted', onCSSNodeChanged);
            pinegrow.addEventHandler('on_css_node_updated', onCSSNodeChanged);
            pinegrow.addEventHandler('on_css_node_added', onCSSNodeChanged);

            var onPageChanges = function(page, info, col) {
                if(info && info.source && info.source == _this) return;
                if(col && col.contains(pgel)) {
                    _this.showGrid( active_rule );
                }
            }

            pinegrow.addEventHandler('on_page_changes_done', onPageChanges);


            var onUndo = function() {
                _this.showGrid( active_rule );
            }

            pinegrow.addEventHandler('on_undo_redo', onUndo);


            view.onDestroy = function() {
                pinegrow.removeEventHandler('on_page_view_resized', onPageViewResized);
                pinegrow.removeEventHandler('on_active_page_view_changed', onPageViewResized);
                pinegrow.removeEventHandler('on_css_node_deleted', onCSSNodeChanged);
                pinegrow.removeEventHandler('on_css_node_updated', onCSSNodeChanged);
                pinegrow.removeEventHandler('on_css_node_added', onCSSNodeChanged);
                pinegrow.removeEventHandler('on_page_changes_done', onPageChanges);
                pinegrow.removeEventHandler('on_undo_redo', onUndo);
                delayedTasks.destroy();
            }
        }

        var PgQuickCSSGridControlView = function(pgel, $target) {


            var uv = this.gridView = new PgDmCSSGridControlView(pgel);

            var win = this.win = new PgQuickWindow('CSS Grid for ' + pgel.getName({short: true}), uv, 'quickCSSGridControl', 600, 600, null, true);
            win.noRightEdge();

            win.onFloatingChanged = function() {
                uv.showGrid();
            }

            win.showFor$El($target || pgel.get$DOMElement(), true);

            win.addIconToHeader('icon-271', 'Learn more about CSS grid.', function() {
                pinegrow.openExternalUrl(help_url);
            })
        }

        var PgCSSGridInfo = function($el, closest) {

            var el = $el.get(0);

            this.$el = $el;
            this.el = el;

            this.s = null;

            if(el.nodeType == 1) {
                var s = this.s = window.getComputedStyle(el);
            }

            if(closest) {
                var $gel = $el;
                var gs = s;
                while(!this.isGrid(gs) && $gel.length) {
                    $gel = $gel.parent();
                    if($gel.length) {
                        var gel = $gel.get(0);
                        if(gel.nodeType == 1) {
                            gs = window.getComputedStyle(gel);
                        }
                    }
                }
                if($gel.length) {
                    s = this.s = gs;
                    this.$el = $el = $gel;
                    this.el = el = $el.get(0);
                }
            }

            //console.log(s);

            this.update();

            this.position = s.position;
        }


        PgCSSGridInfo.prototype.update = function() {
            if(!this.s) return;

            this.cols = parseColsRowsTemplateValue(this.s.gridTemplateColumns).map(function(i) {
                i.value = parseFloat(i.size.replace('px', ''));
                return i;
            })

            this.num_of_cols = this.cols.length;
            if(this.num_of_cols > 0 && this.cols[ this.num_of_cols - 1].last) {
                this.num_of_cols--;
            }

            this.rows = parseColsRowsTemplateValue(this.s.gridTemplateRows).map(function(i) {
                i.value = parseFloat(i.size.replace('px', ''));
                return i;
            })
            this.num_of_rows = this.rows.length;
            if(this.num_of_rows > 0 && this.rows[ this.num_of_rows - 1].last) {
                this.num_of_rows--;
            }

            this.gap_col = parseGapValue(this.s.gridColumnGap);
            this.gap_row = parseGapValue(this.s.gridRowGap);

            this.rect = this.el.getBoundingClientRect();
        }

        PgCSSGridInfo.prototype.createNamedArea = function(pgel, active_rule, area, col_start, row_start, col_end, row_end) {
            col_start = parseInt(col_start);
            row_start = parseInt(row_start);
            col_end = parseInt(col_end);
            row_end = parseInt(row_end);

            var area_str = null;
            if(active_rule) {
                area_str = pgGetCSSRulePropValue(active_rule, 'grid-template-areas');
            } else {
                area_str = pgel.getInlineStylePropertyValue('grid-template-areas');
            }
            var cells;

            if(area_str) {
                cells = pgCSSGridParseTemplateAreas(area_str);
            }
            if(!cells) {
                cells = [];
                for(var i = 0; i < this.num_of_rows /* this.rows.length */; i++) {
                    var row = [];
                    for(var j = 0; j < this.num_of_cols /* this.cols.length */; j++) {
                        row.push('.');
                    }
                    cells.push(row);
                }
            }

            for(var i = 0; i < cells.length; i++) {
                for(var j = 0; j < cells[i].length; j++) {
                    if(cells[i][j] == area) {
                        cells[i][j] = '.';
                    }
                }
            }

            var max_rows = cells.length;
            var max_cols = max_rows > 0 ? cells[0].length : 0;

            if(max_rows == 0 || max_cols == 0) return false;

            if(col_start < 1 || col_start > max_cols || col_end - 1 < 1 || col_end - 1 > max_cols || row_start < 1 || row_start > max_rows || row_end - 1 < 1 || row_end - 1> max_rows || col_start > col_end - 1|| row_start > row_end - 1) {
                return false;
            }

            for(var y = row_start; y <= row_end - 1; y++) {
                for(var x = col_start; x <= col_end - 1; x++) {
                    cells[y-1][x-1] = area;
                }
            }

            active_rule = pgGetCurrentRule(active_rule);

            var areas = pgGetAreasFromCells(cells);
            pgSetCSSPropForActiveRule(pgel, active_rule, 'grid-template-areas', areas);
            pgCSSGridTriggerChangeEvent(pgel, active_rule, 'Create grid area');
            return true;
        }

        PgCSSGridInfo.prototype.getNumber = function(num) {
            return parseFloat(num.replace('px', ''));
        }

        PgCSSGridInfo.prototype.isGrid = function(s) {
            s = s || this.s;
            return s && (s.display == 'grid' || s.display == 'inline-grid');
        }

        PgCSSGridInfo.prototype.makeFixedSizeInDOM = function() {
            var _this = this;

            var withMinSize = function(str, min) {
                return parseColsRowsTemplateValue(str).map(function(item) {
                    val = _this.getNumber(item.size);
                    if(val < min) val = min;
                    if(item.name) val = '[' + item.name + '] ' + val;
                    return val + 'px';
                }).join(' ');
            }

            pgSaveOriginalStyle(this.$el);
            this.$el.css({
                'grid-template-columns': withMinSize(this.s.gridTemplateColumns, 0),
                'grid-template-rows': withMinSize(this.s.gridTemplateRows, 40)
            })
        }

        PgCSSGridInfo.prototype.restoreOriginalSizeInDOM = function() {
            pgRestoreOriginalStyle(this.$el);
        }

        PgCSSGridInfo.prototype.getCellRect = function(col, row) {
            var r = {
                left: this.getXForColStart(col),
                right: this.getXForColEnd(col),
                top: this.getYForRowStart(row),
                bottom: this.getYForRowEnd(row)
            }
            r.width = r.right - r.left;
            r.height = r.bottom - r.top;
            r.name = `${row + 1} / ${col + 1} / ${row + 2} / ${col + 2}`;
            return r;
        }

        PgCSSGridInfo.prototype.getAvailableAreas = function() {
            var r = [];
            var has = {};
            var cells = pgCSSGridParseTemplateAreas(this.s.gridTemplateAreas);
            if(cells) {
                for(var row_i = 0; row_i < cells.length; row_i++) {
                    for(var col_i = 0; col_i < cells[row_i].length; col_i++) {
                        if(cells[row_i][col_i] != '' && cells[row_i][col_i] != '.' && !has[cells[row_i][col_i]]) {
                            has[cells[row_i][col_i]] = true;
                            r.push(cells[row_i][col_i]);
                        }
                    }
                }
            }
            return r;
        }

        PgCSSGridInfo.prototype.getAreaPosition = function(area) {
            var cells = pgCSSGridParseTemplateAreas(this.s.gridTemplateAreas);
            if(cells) {
                var r = {
                    'grid-column-start': null,
                    'grid-column-end': null,
                    'grid-row-start': null,
                    'grid-row-end': null
                }
                var row_i, col_i;

                for(row_i = 0; row_i < cells.length; row_i++) {
                    for(col_i = 0; col_i < cells[row_i].length; col_i++) {
                        if(cells[row_i][col_i] == area) {
                            if(r['grid-row-start'] === null) {
                                r['grid-row-start'] = row_i + 1;
                            }
                            if(r['grid-column-start'] === null) {
                                r['grid-column-start'] = col_i + 1;
                            }
                            r['grid-column-end'] = col_i + 2;
                            r['grid-row-end'] = row_i + 2;
                        }
                    }
                }
                return r;
            } else {
                return null;
            }
        }


        PgCSSGridInfo.prototype.getColForX = function(x, def, end) {
            return this.getTrackForX(x, def, end);
        }

        PgCSSGridInfo.prototype.getTrackForX = function(x, def, end) {
            return this.getTrackForValue('col', x, def, end)

            if(def === undefined || def === null) def = this.num_of_cols;
            var sx = this.getNumber(this.s.paddingLeft);
            var tolerance = end ? -1.0 : 1.0;
            for(var i = 0; i < this.num_of_cols; i++) {
                var track_start = sx;
                var track_end = sx + this.gap_col;
                var track_mid = sx + this.gap_col/2;

                var col_half = this.cols[i].value / 2;

                var area_start = track_start;
                if(i > 0) {
                    area_start -= this.cols[i-1].value / 2;
                }
                var area_end = track_end + this.cols[i].value / 2;

                if(area_start <= x && x <= area_end) {
                    return i;
                }
                if(x < area_start) {
                    //the match was not picked up earlier or due to decimal precision
                    return i;
                }
                sx += this.gap_col + this.cols[i].value;
            }
            return def;
        }

        PgCSSGridInfo.prototype.getXForColStart = function(col) {
            var x = this.getNumber(this.s.paddingLeft);
            for(var i = 0; i < this.num_of_cols; i++) {
                if(i == col) return x;
                x += this.cols[i].value;
                x += this.gap_col;
            }
            return x;
        }


        PgCSSGridInfo.prototype.getXForColEnd = function(col) {
            var x = this.getNumber(this.s.paddingLeft);
            var tolerance = 1.0;
            for(var i = 0; i < this.num_of_cols; i++) {
                x += this.cols[i].value;
                if(i == col) return x;
                x += this.gap_col;
            }
            return x;
        }

        PgCSSGridInfo.prototype.getRowForY = function(x, def, end) {
            return this.getTrackForY(x, def, end);
        }

        PgCSSGridInfo.prototype.getRowForYOld = function(x, def, end) {
            if(def === undefined || def === null) def = this.num_of_rows;
            var sx = this.getNumber(this.s.paddingTop);
            var tolerance = end ? -1.0 : 1.0;
            for(var i = 0; i < this.num_of_rows; i++) {
                if(sx <= x && sx + this.rows[i].value + tolerance >= x) return i;
                if(sx + this.rows[i].value <= x && x < sx + this.rows[i].value + this.gap_row) {
                    //in gap
                    return i;
                }
                sx += this.gap_row + this.rows[i].value;
            }
            return def;
        }

        PgCSSGridInfo.prototype.getColForXOld = function(x, def, end) {
            if(def === undefined || def === null) def = this.num_of_cols;
            var sx = this.getNumber(this.s.paddingLeft);
            var tolerance = end ? -1.0 : 1.0;
            for(var i = 0; i < this.num_of_cols; i++) {
                if(sx <= x && sx + this.cols[i].value + tolerance >= x) return i;
                if(sx + this.cols[i].value <= x && x < sx + this.cols[i].value + this.gap_col) {
                    //in gap
                    return i;
                }
                sx += this.gap_col + this.cols[i].value;
            }
            return def;
        }

        PgCSSGridInfo.prototype.getTrackForValue = function(dim, x, def, end) {
            var num, items, gap, sx;

            if(dim === 'row') {
                num = this.num_of_rows;
                items = this.rows;
                gap = this.gap_row;
                sx = this.getNumber(this.s.paddingTop) + this.getNumber(this.s.borderTopWidth);
            } else {
                num = this.num_of_cols;
                items = this.cols;
                gap = this.gap_col;
                sx = this.getNumber(this.s.paddingLeft) + this.getNumber(this.s.borderLeftWidth);
            }
            if(def === undefined || def === null) def = num;

            //console.log(dim, items);
            //console.log(`x = ${x}`);

            var prev_dist = null;

            var getDistFromTrack = function(i) {
                var track_start = sx;
                var track_end = sx + ((i > 0) ? gap : 0);
                var measure_from = end ? track_start : track_start;

                var dist = end ? measure_from - x : x - measure_from;

                if(Math.abs(dist) < 1.0) dist = 0;

                //dist = Math.abs(dist)

                //console.log(`track ${i} = ${track_start} - ${track_end}, measure from ${measure_from}, dist = ${dist}`)
                return dist;
            }

            for(var i = 0; i < num + 1; i++) {
                var dist = getDistFromTrack(i);

                if(prev_dist !== null && prev_dist >= 0 && (prev_dist < dist || dist < 0)) {
                    return i-1;
                }

                prev_dist = dist;

                /*

                 var area_start = track_start;

                 console.log(`start = ${area_start}`);

                 if(i > 0) {
                 area_start -= this.rows[i-1].value / 2;
                 console.log(`adjusted start = ${area_start}`);
                 }
                 var area_end = track_end + this.rows[i].value;// / 2;

                 console.log(`end = ${area_end}`);

                 if(area_start <= x && x <= area_end) {
                 return i;
                 }
                 if(x < area_start) {
                 //the match was not picked up earlier or due to decimal precision
                 return i;
                 }
                 */
                if(i < items.length) {
                    sx += gap + items[i].value;
                }
            }

            return def;
        }


        PgCSSGridInfo.prototype.getTrackForY = function(x, def, end) {
            return this.getTrackForValue('row', x, def, end);
        }

        PgCSSGridInfo.prototype.getYForRowStart = function(row) {
            var x = this.getNumber(this.s.paddingTop);
            for(var i = 0; i < this.num_of_rows; i++) {
                if(i == row) return x;
                x += this.rows[i].value;
                x += this.gap_row;
            }
            return x;
        }

        PgCSSGridInfo.prototype.getYForRowEnd = function(row) {
            var x = this.getNumber(this.s.paddingTop);
            for(var i = 0; i < this.num_of_rows; i++) {
                x += this.rows[i].value;
                if(i == row) return x;
                x += this.gap_row;
            }
            return x;
        }

        PgCSSGridInfo.prototype.placeAllAutoCellsInDOM = function() {
            var _this = this;
            this.$el.children().each(function(i, ch) {
                var $ch = $(ch);
                var cell = new PgCSSGridCellInfo($ch, _this);
                if(cell.isCell() && cell.isAutoPlaced()) {
                    pgSaveOriginalStyle($ch);
                    var css = cell.autoToFixedPlace();
//                    console.log(css);
                    $ch.css(css);
                }
            })
        }

        PgCSSGridInfo.prototype.restorePlacedCellsInDOM = function() {
            var _this = this;
            this.$el.children().each(function(i, ch) {
                var $ch = $(ch);
                pgRestoreOriginalStyle($ch);
            })
        }

        window.PgCSSGridInfo = PgCSSGridInfo;

        var PgCSSGridCellInfo = function($el, grid_info) {

            var el = $el.get(0);

            this.s = null;
            this.el = el;

            if(el.nodeType == 1) {
                var s = this.s = window.getComputedStyle(el);
            }


            this.grid_info = grid_info;

            this.update();


            //console.log(s);

            //this.cols = s.gridTemplateColumns.split(' ').map(function(v) { return parseFloat(v.replace('px', '')); });
        }

        PgCSSGridCellInfo.prototype.update = function() {
            if(this.grid_info) {
                this.rect = this.el.getBoundingClientRect();

                this.x = this.rect.left - this.grid_info.rect.left;
                this.end_x = this.x + this.rect.width;
                this.y = this.rect.top - this.grid_info.rect.top;
                this.end_y = this.y + this.rect.height;
            }
        }

        PgCSSGridCellInfo.prototype.getGridCSSValue = function(prop) {
            return this.s['grid' + prop];
        }

        PgCSSGridCellInfo.prototype.isCell = function() {
            return this.s && (['absolute', 'fixed'].indexOf(this.s.position) < 0 && this.s.display != 'none');
        }

        PgCSSGridCellInfo.prototype.isAutoPlaced = function() {
            return this.s.gridArea == "auto / auto / auto / auto" && (this.s.gridColumnStart == 'auto' || this.s.gridColumnEnd == 'auto' || this.s.gridRowStart == 'auto' || this.s.gridRowEnd == 'auto');
        }

        PgCSSGridCellInfo.prototype.getAreaName = function() {
            if(this.s.gridColumnStart.match(/^[^0-9\-\_][a-zA-Z\_\-0-9]*$/) && this.s.gridColumnStart == this.s.gridColumnEnd && this.s.gridColumnStart == this.s.gridRowStart && this.s.gridColumnStart == this.s.gridRowEnd) {
                if(this.s.gridColumnStart == 'auto') return null;
                //are these named areas or tracks?

                var findName = function(name, list) {
                    for(var i = 0; i < list.length; i++) {
                        if(list[i].names) {
                            for(var j = 0; j < list[i].names.length; j++) {
                                if(list[i].names[j].name_parsed == name && list[i].names[j].start !== null) {
                                    return true;
                                }
                            }
                        }
                    }
                    return false;
                }

                if(findName(this.s.gridColumnStart, this.grid_info.cols) &&  findName(this.s.gridColumnStart, this.grid_info.rows)) {
                    return null;
                }
                return this.s.gridColumnStart;
            }
            return null;
        }

        PgCSSGridCellInfo.prototype.autoToFixedPlace = function() {
            var col_start = isNumber(this.s.gridColumnStart) ? this.s.gridColumnStart : this.grid_info.getColForX(this.x) + 1;
            var col_end = isNumber(this.s.gridColumnEnd) ? this.s.gridColumnEnd : this.grid_info.getColForX(this.end_x, null, true) + 1;//2;
            var row_start = isNumber(this.s.gridRowStart) ? this.s.gridRowStart : this.grid_info.getRowForY(this.y) + 1;
            var row_end = isNumber(this.s.gridRowEnd) ? this.s.gridRowEnd : this.grid_info.getRowForY(this.end_y, null, true) + 1;//2;

            var d = {
                'grid-column-start': col_start.toString(),
                'grid-column-end': col_end.toString(),
                'grid-row-start': row_start.toString(),
                'grid-row-end': row_end.toString()
            }
            //console.log('autoToFixedPlace', d)
            return d;
        }

        var isIdent = function(str) {
            return typeof str == 'string' && str.match(/^[^0-9\-\_][a-zA-Z\_\-0-9]*$/);
        }

        var isNumber = function(str) {
            return typeof str == 'number' || (typeof str == 'string' && str.match(/^[0-9]+$/));
        }

        PgCSSGridCellInfo.prototype.getCellColRowPosition = function(force_numbers) {
            var area = this.getAreaName();
            if(area) {
                var r = this.grid_info.getAreaPosition(area);
                if(r) return r;
            }
            if(force_numbers || this.isAutoPlaced()) {
                return this.autoToFixedPlace();
            }

            var r = {};
            var pos;

            //col start
            if(isIdent(this.s.gridColumnStart)) {
                r['grid-column-start'] = this.s.gridColumnStart;
                /*
                pos = this.grid_info.getAreaPosition(this.s.gridColumnStart);
                if(pos) {
                    r['grid-column-start'] = pos['grid-column-start'];
                }*/
            } else if(isNumber(this.s.gridColumnStart)) {
                r['grid-column-start'] = this.s.gridColumnStart;
            }

            if(!('grid-column-start' in r)) {
                r['grid-column-start'] = this.grid_info.getColForX(this.x) + 1;
            }

            //col end
            if(isIdent(this.s.gridColumnEnd)) {
                r['grid-column-end'] = this.s.gridColumnEnd;
                /*
                pos = this.grid_info.getAreaPosition(this.s.gridColumnEnd);
                if(pos) {
                    r['grid-column-end'] = pos['grid-column-end'];
                }*/
            } else if(isNumber(this.s.gridColumnEnd)) {
                r['grid-column-end'] = this.s.gridColumnEnd;
            }

            if(!('grid-column-end' in r)) {
                r['grid-column-end'] = this.grid_info.getColForX(this.end_x, null, true) + 1;//2;
            }

            //row start
            if(isIdent(this.s.gridRowStart)) {
                r['grid-row-start'] = this.s.gridRowStart;
                /*
                pos = this.grid_info.getAreaPosition(this.s.gridRowStart);
                if(pos) {
                    r['grid-row-start'] = pos['grid-row-start'];
                }*/
            } else if(isNumber(this.s.gridRowStart)) {
                r['grid-row-start'] = this.s.gridRowStart;
            }

            if(!('grid-row-start' in r)) {
                r['grid-row-start'] = this.grid_info.getRowForY(this.y) + 1;
            }

            //row end
            if(isIdent(this.s.gridRowEnd)) {
                r['grid-row-end'] = this.s.gridRowEnd;
                /*
                pos = this.grid_info.getAreaPosition(this.s.gridRowEnd);
                if(pos) {
                    r['grid-row-end'] = pos['grid-row-end'];
                }*/
            } else if(isNumber(this.s.gridRowEnd)) {
                r['grid-row-end'] = this.s.gridRowEnd;
            }

            if(!('grid-row-end' in r)) {
                r['grid-row-end'] = this.grid_info.getRowForY(this.end_y, null, true) + 1;//2;
            }

            return r;
        }

        PgCSSGridCellInfo.prototype.getColName = function(row_start, start) {
            if(start === undefined) start = false;
            if(isIdent(row_start)) return row_start;
            var row_idx = parseInt(row_start) - 1;
            if(row_idx >= 0 && row_idx < this.grid_info.cols.length && this.grid_info.cols[row_idx].name) {
                return this.grid_info.cols[row_idx].getName(start) || row_start;
            }
            return row_start;
        }

        PgCSSGridCellInfo.prototype.getRowName = function(row_start, start) {
            if(start === undefined) start = false;
            if(isIdent(row_start)) return row_start;
            var row_idx = parseInt(row_start) - 1;
            if(row_idx >= 0 && row_idx < this.grid_info.rows.length && this.grid_info.rows[row_idx].name) {
                return this.grid_info.rows[row_idx].getName(start) || row_start;
            }
            return row_start;
        }

        PgCSSGridCellInfo.prototype.getGridAreaShorthand = function(r) {
            /*var r = {
                'grid-row': null,
                'grid-column': null,
                'grid-column-start': null,
                'grid-column-end': null
            }
            if(!r) return null*/
            var rs = this.getRowName(r['grid-row-start'], true);
            var cs = this.getColName(r['grid-column-start'], true);
            var re = this.getRowName(r['grid-row-end']);
            var ce = this.getColName(r['grid-column-end']);

            if(rs && rs === cs && cs === re && re === ce) {
                return rs;
            }
            return `${rs} / ${cs} / ${re} / ${ce}`
        }

        window.PgCSSGridCellInfo = PgCSSGridCellInfo;

        var PgCSSGridOverlayControl = function() {

            var $grid = null;
            var $current_el = null;

            this.show = function($el) {
                $current_el = $el;

                var info = new PgCSSGridInfo($el);

                var cols = info.cols;
                var rows = info.rows;
                var gap_col = info.gap_col;
                var gap_row = info.gap_row;

                var x = info.getNumber(info.s.paddingLeft);
                var y = info.getNumber(info.s.paddingTop);

                var md = function(cls, st, attrs) {
                    return `<pgdiv class="pinegrow-ui-helper pinegrow-ui-helper-${cls}" style="${st || ''}" ${attrs ? attrs : ''}>`;
                }

                var html = md('css-grid');

                var total_w = 0;
                var total_h = 0;
                var total_col_w = 0;
                var total_row_h = 0;

                for(var yi = 0; yi < rows.length; yi++) {
                    if(rows[yi].last) break;
                    total_h += gap_row + rows[yi].value;
                    total_row_h += rows[yi].value
                }

                for(var xi = 0; xi < cols.length; xi++) {
                    if(cols[xi].last) break;
                    total_w += gap_col + cols[xi].value;
                    total_col_w += cols[xi].value;
                }

                var space_w = info.rect.width - x - info.getNumber(info.s.paddingRight);
                var space_h = info.rect.height - y - info.getNumber(info.s.paddingBottom);

                var sx = info.getNumber(info.s.paddingLeft);

                switch(info.s.justifyContent) {
                    case 'center':
                        sx = Math.max(0, sx + (space_w - total_w) / 2);
                        break;
                    case 'end':
                    case 'right':
                    case 'flex-end':
                        sx = Math.max(0, sx + (space_w - total_w));
                        break;
                    case 'space-between':
                        if(total_w < space_w && info.num_of_cols > 1) {
                            gap_col = (space_w - total_col_w) / (info.num_of_cols - 1);
                        }
                        break;
                    case 'space-around':
                        if(total_w < space_w && info.num_of_cols > 0) {
                            gap_col = (space_w - total_col_w) / (info.num_of_cols);
                            sx = sx + gap_col / 2;
                        }
                        break;
                    case 'space-evenly':
                        if(total_w < space_w && info.num_of_cols > 0) {
                            gap_col = (space_w - total_col_w) / (info.num_of_cols + 1);
                            sx = sx + gap_col;
                        }
                        break;
                }

                var sy = y;

                switch(info.s.alignContent) {
                    case 'center':
                        sy = Math.max(0, sy + (space_h - total_h) / 2);
                        break;
                    case 'end':
                    case 'flex-end':
                        sy = Math.max(0, sy + (space_h - total_h));
                        break;
                    case 'space-between':
                        if(total_h < space_h && info.num_of_rows > 1) {
                            gap_row = (space_h - total_row_h) / (info.num_of_rows - 1);
                        }
                        break;
                    case 'space-around':
                        if(total_h < space_h && info.num_of_rows > 0) {
                            gap_row = (space_h - total_row_h) / (info.num_of_rows);
                            sy = sy + gap_row/2;
                        }
                        break;
                    case 'space-evenly':
                        if(total_h < space_h && info.num_of_rows > 0) {
                            gap_row = (space_h - total_row_h) / (info.num_of_rows + 1);
                            sy = sy + gap_row;
                        }
                        break;
                }

                y = sy;

                for(var yi = 0; yi < rows.length; yi++) {
                    if(rows[yi].last) break;

                    x = sx;
                    for(var xi = 0; xi < cols.length; xi++) {
                        if(cols[xi].last) break;

                        html += md('css-grid-cell', `top:${y}px;left:${x}px;width:${cols[xi].value}px;height:${rows[yi].value}px;`, `data-col-row="${xi}-${yi}"`) + '</pgdiv>';
                        x += gap_col + cols[xi].value;
                    }
                    y += gap_row + rows[yi].value;
                }
                html += '</pgdiv>';

                $grid = $(html);

                if(info.position == 'static') {
                    $el.addClass('pinegrow-ui-helper-relative');
                }

                $el.append($grid);

            }

            this.update = function() {
                if($grid) $grid.remove();
                this.show($current_el);
            }

            var current_hl = false;

            this.setHighlight = function(v) {
                if(current_hl != v) {
                    if (v) {
                        $grid.addClass('pinegrow-ui-helper-css-grid-hl');
                    } else {
                        $grid.removeClass('pinegrow-ui-helper-css-grid-hl');
                    }
                    current_hl = v;
                }
            }

            this.destroy = function() {
                if($grid) $grid.remove();
            }
        }

        window.PgCSSGridOverlayControl = PgCSSGridOverlayControl;



//Menu points
        var PgMenuPointCSSGrid = function(placement, menu, type, grid_info, cell_info) {

            PgElementPageViewMenuPoint.call(this, placement, menu);

            var _this = this;

            this.mode = f.mode; //or css

            this.placement = placement;

            this.type = type;

            this.menu = menu;

            this.tooltip = '';
            this.active_css_rule = null;

            this.$grid = menu.$el.parent();

            this.info = grid_info || new PgCSSGridInfo(this.$grid);
            this.cell_info = cell_info || new PgCSSGridCellInfo(menu.$el, this.info);

            this.onPageViewsUpdated = function(page, update_id) {
                if(!_this.info.update_id || _this.info.update_id != update_id) {
                    _this.info.update();
                    _this.info.update_id = update_id;
                }
                if(!_this.cell_info.update_id || _this.cell_info.update_id != update_id) {
                    _this.cell_info.update();
                    _this.cell_info.update_id = update_id;
                }
            }

            pinegrow.addEventHandler('on_page_views_updated', this.onPageViewsUpdated);

            this.update();
        }

        extendObject(PgMenuPointCSSGrid, PgElementPageViewMenuPoint)


        PgMenuPointCSSGrid.prototype.setIsNotSet = function(val) {
            PgElementPageViewMenuPoint.prototype.setIsNotSet.call(this, val);

            if(val) {
                this.tooltip = 'Positioned automatically or with a named area. Move to position with columns &amp; rows.';
            } else {
                this.tooltip = null;
            }
        }


        PgMenuPointCSSGrid.prototype.update = function() {

            switch(this.type) {
                case 'column-start':
                    if(this.cell_info.getGridCSSValue('ColumnStart') == 'auto') {
                        this.setText('A');
                        this.setIsNotSet(true);
                    } else {
                        this.setText(this.cell_info.colRowPosition['grid-' + this.type]);
                    }
                    break;
                case 'column-end':
                    if(this.cell_info.getGridCSSValue('ColumnEnd') == 'auto') {
                        this.setText('A');
                        this.setIsNotSet(true);
                    } else {
                        this.setText(this.cell_info.colRowPosition['grid-' + this.type]);
                    }
                    break;
                case 'row-start':
                    if(this.cell_info.getGridCSSValue('RowStart') == 'auto') {
                        this.setText('A');
                        this.setIsNotSet(true);
                    } else {
                        this.setText(this.cell_info.colRowPosition['grid-' + this.type]);
                    }
                    break;
                case 'row-end':
                    if(this.cell_info.getGridCSSValue('RowEnd') == 'auto') {
                        this.setText('A');
                        this.setIsNotSet(true);
                    } else {
                        this.setText(this.cell_info.colRowPosition['grid-' + this.type]);
                    }
                    break;
            }
            //var sizes = this.options.sizes;

            //var parent_w = this.menu.$el.parent().innerWidth();
            //this.unitSize = this.snapX = Math.max(10, (parent_w / this.options.num_columns) * this.menu.pageView.zoom);
        }


        PgMenuPointCSSGrid.prototype.add = function(menu, $container) {
            PgElementPageViewMenuPoint.prototype.add.call( this, menu, $container );
            this.update();
        }

        PgMenuPointCSSGrid.prototype.onReposition = function() {
            //update text in box
            if(!this.in_drag) {
                this.update();
            }
        }

        PgMenuPointCSSGrid.prototype.onDragStart = function(x, y) {
            this.original_z = this.menu.$el.css('z-index');

            this.info.makeFixedSizeInDOM();

            this.info.update();
            this.cell_info.update();

            pgSaveOriginalStyle(this.menu.$el);

            this.info.placeAllAutoCellsInDOM();
            this.menu.$el.addClass('pinegrow-grid-cell-during-resize');
            this.info.$el.addClass('pinegrow-grid-during-resize');

            if(this.mode == 'css') {
                //this.active_css_rule = pgFindCSSRuleWithProps(null /* def active rules */, ['grid-area', 'grid-column-start', 'grid-column-end', 'grid-row-start', 'grid-row-end']);
            }

            this.current_value = null;

            if(this.onDragStarted) this.onDragStarted.call(this);
            /*
             this.menu.$el.css({
             'z-index': '1000',
             'width': '100%',
             'height': '100%'
             });*/

            //this.setLabel(x);
            //this.setText(x);
        }

        PgMenuPointCSSGrid.prototype.onDrag = function(x, y) {
            var _this = this;

            var nx = x / this.menu.pageView.zoom;
            var ny = y / this.menu.pageView.zoom;

            var col_start_x, col_end_x, row_start_y, row_end_y;

            col_start_x = col_end_x = row_start_y = row_end_y = 0;

            if(pinegrow.keyState.shift && pgIsDev()) {
                pinegrow.keyState.shift = false;
                debugger;
            }

            switch(this.type) {
                case 'column-start':
                    col_start_x = nx;
                    var col_for_x = this.info.getColForX(this.cell_info.x + col_start_x, 0);
                    this.cell_info.colRowPosition['grid-' + this.type] = Math.min(col_for_x + 1, this.info.num_of_cols);
                    //console.log(col_start_x, col_for_x, this.cell_info.x, x, nx);
                    break;

                case 'column-end':
                    col_end_x = nx;
                    this.cell_info.colRowPosition['grid-' + this.type] = Math.min(this.info.getColForX(this.cell_info.end_x + col_end_x, null, true) + 1/*2*/, this.info.num_of_cols + 1 );
                    break;

                case 'row-start':
                    row_start_y = ny;
                    var v = this.info.getRowForY(this.cell_info.y + row_start_y, 0) + 1;
                    this.cell_info.colRowPosition['grid-' + this.type] = Math.min(v, this.info.num_of_rows);
                    break;

                case 'row-end':
                    row_end_y = ny;
                    this.cell_info.colRowPosition['grid-' + this.type] = Math.min(this.info.getRowForY(this.cell_info.end_y + row_end_y, null, true) + 1/*2*/, this.info.num_of_rows + 1);
                    break;
            }

            var gridArea = this.current_value = this.cell_info.getGridAreaShorthand(this.cell_info.colRowPosition);

            if(this.mode === 'css') {

                this.active_css_rule = pgGetCurrentRule(this.active_css_rule);
                console.log('GRID DRAG UPDATE')
                pgCSSGridSetGridAreaForPgel(this.menu.pgel, this.active_css_rule, gridArea);
            } else {
                // 'dm'
                dmSetGridItemProp(this.menu.pgel, 'grid-area', gridArea);
                /*
                var dms = pgDMGetCurrentSizeSlug(this.menu.pageView);
                var dims = ['cs', 'ce', 'rs', 're'];
                dims.forEach(function(dim) {
                    _this.menu.pgel.removeClassesStartingWith(`g${dms}-${dim}-`);
                })
                this.menu.$el.css('grid-area', '');

                this.menu.pgel.addClass(`g${dms}-cs-${this.cell_info.colRowPosition['grid-column-start']}`);
                this.menu.pgel.addClass(`g${dms}-ce-${this.cell_info.colRowPosition['grid-column-end']}`);
                this.menu.pgel.addClass(`g${dms}-rs-${this.cell_info.colRowPosition['grid-row-start']}`);
                this.menu.pgel.addClass(`g${dms}-re-${this.cell_info.colRowPosition['grid-row-end']}`);
                */
            }

            this.update();

            this.setIsNotSet(false);

            if(this.area_name_point) {
                this.area_name_point.destroy();
                this.area_name_point = null;
            }

            this.menu.reposition();
            /*
             this.menu.$el.css({
             'z-index': '1000',
             'width': '100%',
             'height': '100%'
             });
             */

            return {x: x /* x_for_col - this.cell_info.end_x */, y:y};
        }

        PgMenuPointCSSGrid.prototype.onClick = function() {
        }

        PgMenuPointCSSGrid.prototype.onDragStop = function(x, y) {
            pinegrow.removeEventHandler('on_page_views_updated', this.onPageViewsUpdated);

            if(this.mode == 'css') {
                pgForgetOriginalStyle(this.menu.$el);
            } else {
                // 'dm'
                //pgForgetOriginalStyle(this.menu.$el);
                pgRestoreOriginalStyle(this.menu.$el);
            }
            this.menu.$el.removeClass('pinegrow-grid-cell-during-resize');
            this.info.$el.removeClass('pinegrow-grid-during-resize');
            //this.menu.$el.css('z-index', this.original_z || '');
            this.setLabel();
            this.info.restoreOriginalSizeInDOM(); //this will also restore track names

            //this.info.update();
            //this.cell_info.update();

            //get track names
            //this.current_value = this.cell_info.getGridAreaShorthand(this.cell_info.colRowPosition);

            this.info.restorePlacedCellsInDOM();

            if(this.mode == 'css') {
                if (this.current_value) {
                    this.active_css_rule = pgGetCurrentRule(this.active_css_rule);
                    pgCSSGridSetGridAreaForPgel(this.menu.pgel, this.active_css_rule, this.current_value /* trigger changed event */);
                    pgCSSGridTriggerChangeEvent(this.menu.pgel, this.active_css_rule, 'Set position in grid')
                }
            } else {
                // 'dm'
                pinegrow.changeManager.didChange(this.menu.pgel, 'Change position in grid');
            }
            pinegrow.selectedElements.reselect();
        }

    })

    var PgCSSStatementTester = function() {

        var $div = $('<div/>');

        this.test = function(prop, value) {
            if(typeof value !== 'string' || value === '') return true;

            var ok = true;

            try {
                if(!value.match(/[\$\@]/)) { //skip SASS/LESS
                    $div.css(prop, value);
                    var testval = $div.css(prop);
                    ok = testval !== '';
                    $div.css(prop, '');
                }
                if(ok) {
                    //detect missing ( and )
                    var ps = 0;
                    for(var i = 0; i < value.length; i++) {
                        var ch = value.charAt(i);
                        if(ch === '(') {
                            ps++;
                        } else if(ch === ')') {
                            ps--;
                        }
                    }
                    if(ps !== 0) ok = false;
                }
            }
            catch(err) {}
            return ok;
        }
    }

});
