/*
Copyright Humane tehnologije d.o.o.
 */

var PgInsight = function() {

    var _this = this;

    var forEachSource = function(func) {
        //current page for now
        var project = pinegrow.getCurrentProject();
        var page = pinegrow.getSelectedPage();

        if(page) {
            if(project && project.isPageInProject(page)) {
                project.forEachOpenPage(function(p) {
                    func(p.sourceNode);
                });
            } else {
                func(page.sourceNode);
            }
        }
    }

    var cache = null;
    var stylesheets_done = null;
    let openedProject;

    var contents_tags = ['title'];

    var handlers_on_begin_inspecting_project = [];
    var handlers_on_inspect_page = [];
    var handlers_on_end_inspecting_project = [];

    this.addHandler = function(on_begin, on_inspect_page, on_end) {
        if(on_begin) {
            handlers_on_begin_inspecting_project.push(on_begin);
        }
        if(on_inspect_page) {
            handlers_on_inspect_page.push(on_inspect_page);
        }
        if(on_end) {
            handlers_on_end_inspecting_project.push(on_end);
        }
    }

    this.ext = {}; //for extensions

    var beginInspectingProject = function() {
        cache = {
            classes_map : {},
            classes_list : [],
            classes_changed : false,
            classes_id: 0,
            css_vars_map: {},
            css_vars_list: [],
            css_vars_id: 0,
            css_vars_changed: false,
            ids_map : {},
            ids_list : [],
            ids_changed : false,
            attribute_values : {},
            tags_attribute_values : {},
            keywords_map : {},
            keywords_list : [],
            keywords_changed : false,
            contents : {},
            meta: {},
            path_map : {},
            path_list : [],
            path_changed : false,
            path_id: 0
        }
        stylesheets_done = {};

        handlers_on_begin_inspecting_project.forEach(function(func) {
            func(_this);
        })
    }

    beginInspectingProject(); //init cache

    var attributes_by_tags = ['img', 'iframe', 'style', 'script'];

    var addClass = function(cls, source, sourceid) {
        if(cls === null || cls.length === 0) return;

        if(cls.startsWith('pg-main-cs-') || cls === 'pg-empty-placeholder') return;

        if(!(cls in cache.classes_map)) {
            cache.classes_id++;
            var item = {key: cls, name: cls, count: 1, sources: [], sourceids: [], id: cache.classes_id};
            cache.classes_map[cls] = item;
            cache.classes_list.push(item);
            cache.classes_changed = true;
            if(source && item.sources.indexOf(source) < 0) {
                item.sources.push(source);
            }
            if(sourceid && item.sourceids.indexOf(sourceid) < 0) {
                item.sourceids.push(sourceid);
            }
        } else {
            cache.classes_map[cls].count++;
            var item = cache.classes_map[cls];
            if(source && item.sources.indexOf(source) < 0) {
                item.sources.push(source);
            }
            if(sourceid && item.sourceids.indexOf(sourceid) < 0) {
                item.sourceids.push(sourceid);
            }
        }
    }

    var addCSSVar = function(variable, source, sourceid) {
        if(variable === null || variable.length === 0) return;
        variable = variable.replace('--', '');
        if(variable.length === 0) return;

        if(!(variable in cache.css_vars_map)) {
            cache.css_vars_id++;
            var item = {key: variable, name: variable, count: 1, sources: [], sourceids: [], id: cache.css_vars_id};
            cache.css_vars_map[variable] = item;
            cache.css_vars_list.push(item);
            cache.css_vars_changed = true;
            if(source && item.sources.indexOf(source) < 0) {
                item.sources.push(source);
            }
            if(sourceid && item.sourceids.indexOf(sourceid) < 0) {
                item.sourceids.push(sourceid);
            }
        } else {
            cache.css_vars_map[variable].count++;
            var item = cache.css_vars_map[variable];
            if(source && item.sources.indexOf(source) < 0) {
                item.sources.push(source);
            }
            if(sourceid && item.sourceids.indexOf(sourceid) < 0) {
                item.sourceids.push(sourceid);
            }
        }
    }

    var removeClassesForSourceId = function(sourceid) {
        if(sourceid) {
            for(var i = 0; i < cache.classes_list.length; i++) {
                var item = cache.classes_list[i];
                var idx = item.sourceids.indexOf(sourceid);
                if(idx >= 0) {
                    item.sourceids.splice(idx, 1);
                    if(item.sourceids.length === 0) {
                        delete cache.classes_map[item.key];
                        cache.classes_list.splice(i, 1);
                        i--;
                    }
                    cache.classes_changed = true;
                }
            }
        }
    }

    var addId = function(id) {
        if(id === null || id.length === 0) return;
        if(!(id in cache.ids_map)) {
            var item = {key: id, name: id, count: 1}
            cache.ids_map[id] = item;
            cache.ids_list.push(item);
            cache.ids_changed = true;
        } else {
            cache.ids_map[id].count++;
        }
    }

    var addPath = function(path) {
        if(path === null || path.length === 0) return;
        if(!(path in cache.path_map)) {
            var item = {key: path, name: path, count: 1}
            cache.path_map[path] = item;
            cache.path_list.push(item);
            cache.path_changed = true;
        } else {
            cache.path_map[path].count++;
        }
    }

    var doCSSRules = function(p, source, sourceid) {
        try {
            if (p.cssRules) {
                for (var i = 0; i < p.cssRules.length; i++) {
                    if (p.cssRules[i].type == 1) {
                        var sel = p.cssRules[i].selectorText;
                        var classes = sel.match(/\.([a-zA-Z][a-zA-Z0-9\-\_]*)/g);
                        var ids = sel.match(/\#([a-z0-9\-\_]*)/g);
                        if (classes) {
                            for (let n = 0; n < classes.length; n++) {
                                var cls = classes[n].substr(1, classes[n].length - 1);
                                if (!cls.startsWith('pg-node-id') && !cls.startsWith('pg-main-')) {
                                    addClass(cls, source, sourceid);
                                }
                            }
                        }
                        if (ids) {
                            for (let n = 0; n < ids.length; n++) {
                                var id = ids[n].substr(1, ids[n].length - 1);
                                addId(id, source);
                            }
                        }
                        //look for css vars
                        if(p.cssRules[i].style) {
                            for(let n = 0; n < p.cssRules[i].style.length; n++) {
                                const prop = p.cssRules[i].style.item(n);
                                if(prop.startsWith('--')) {
                                    addCSSVar(prop, source, sourceid);
                                }
                            }
                        }
                    }
                    doCSSRules(p.cssRules[i], source, sourceid);
                }
            }
        } catch(err) {} //this will catch not allowed cssRules error
    }

    var doStylesheet = function(sheet) {
        //console.log('doStylesheet', sheet)
        var cs = pinegrow.stylesheets.getStylesheetForHTMLElement(sheet.ownerNode);
        //doCSSRules(sheet, sheet.href ? crsaGetNameFromUrl(sheet.href.toString()) : 'inline', cs ? ('c' + cs.getUniqueId()) : null);
        doCSSRules(sheet, sheet.href ? crsaGetNameFromUrl(sheet.href.toString()) : 'inline', cs ? ('c' + cs.getUniqueId()) : null);

        //if(pgf.kids) {
            pinegrow.dispatchEventDelayed('on_insight_css_reindexed', null, cs);
        //}
    }

    var inspectAProjectItem = function(page, changedPgel, do_html, do_css, do_path) {
        if(do_html === undefined) do_html = true;
        if(do_css === undefined) do_css = true;
        if(do_path === undefined) do_path = true;

        var handlers_on_inspect_node = [];

        handlers_on_inspect_page.forEach(function(func) {
            var inspect_node_func = func(_this, page);
            if(inspect_node_func) {
                handlers_on_inspect_node.push(inspect_node_func);
            }
        })



        if(do_html) {
            if(pgf.kids) {
                changedPgel = null;
            }
            var sourceNode = changedPgel ? changedPgel : page.sourceNode;
            var sourceid = 'p' + page.uid;

            if(sourceNode === page.sourceNode) {
                removeClassesForSourceId(sourceid);
            }

            sourceNode.walkSelfAndChildren(function(node) {
                if(node.isElement) {
                    var list = node.getClasses();
                    for(var i = 0; i < list.length; i++) {
                        addClass(list[i], null, sourceid);
                    }

                    var id = node.getAttr('id');
                    if(id) addId(id);

                    var attrs = node.getAttrList();
                    var tag_attrs = null;
                    if(attributes_by_tags.indexOf(node.tagName) >= 0) {
                        if(!(node.tagName in cache.tags_attribute_values)) {
                            cache.tags_attribute_values[node.tagName] = {};
                        }
                        tag_attrs = cache.tags_attribute_values[node.tagName];
                    }
                    for(var i = 0; i < attrs.length; i++) {
                        var attr_value = attrs[i].value;
                        if(attr_value) {
                            if(attr_value.startsWith('http')) {
                                attr_value = pinegrow.getOriginalUrl(attr_value);
                            }
                            if(!(attrs[i].name in cache.attribute_values)) {
                                cache.attribute_values[attrs[i].name] = {};
                            }
                            cache.attribute_values[attrs[i].name][attr_value] = true;

                            if(tag_attrs) {
                                if(!(attrs[i].name in tag_attrs)) {
                                    tag_attrs[attrs[i].name] = {};
                                }
                                tag_attrs[attrs[i].name][attr_value] = true;
                            }
                        }
                    }
                    if(node.tagName == 'meta') {
                        var metaName = node.getAttr('name');
                        var metaContent = (node.getAttr('content') || '').trim();
                        switch(metaName) {
                            case 'keywords':
                                var list = metaContent.split(',');
                                for(var i = 0; i < list.length; i++) {
                                    var kw = list[i].trim();
                                    if(kw.length && !(kw in cache.keywords_map)) {
                                        cache.keywords_map[kw] = true;
                                        cache.keywords_list.push({key: kw, name: kw});
                                        cache.keywords_changed = true;
                                    }
                                }
                                break;
                            default:
                                if(metaContent.length) {
                                    if(!(metaName in cache.meta)) {
                                        cache.meta[metaName] = {};
                                    }
                                    cache.meta[metaName][metaContent] = true;
                                }
                                break;
                        }
                    }
                    if(contents_tags.indexOf(node.tagName) >= 0) {
                        if(!(node.tagName in cache.contents)) {
                            cache.contents[node.tagName] = {};
                        }
                        cache.contents[node.tagName][node.html()] = true;
                    }
                    handlers_on_inspect_node.forEach(function(func) {
                        func(_this, page, node);
                    })
                }
                return true;
            })
            // if(changedPgel) return; //quick, just do the changed element
        }

        //do csss
        if(do_css) {
            if(page.openedInEditor) {
                var $html = page.get$Html();
                if($html) {
                    $html.find('style, link[rel="stylesheet"]').each(function(i, e) {
                        if(e.getAttribute('id') != 'crsa-inline-styles') {
                            if(e.href) {
                                if(stylesheets_done[e.href.toString()]) return true;
                                stylesheets_done[e.href.toString()] = true;
                            }
                            if(e.sheet) {
                                doStylesheet(e.sheet);
                            }
                        }
                    })
                }
            }
        }

        if(do_path){
            //todo create logic to handle path fetching
        }
    }

    this.inspectPageOrPgel = inspectAProjectItem;

    var onChangedCss = function() {
        stylesheets_done = {};
        pinegrow.getAllPages().forEach(function(page) {
            inspectAProjectItem(page, null, false, true, false);
        })
    }

    var sortOptionsList = function(list) {
        var c = new Intl.Collator();
        list.sort(function(a,b) {
            return c.compare(a.key, b.key);
        })
    }

    this.sortOptionsList = sortOptionsList;

    var rebuildCacheLists = function() {
        if(cache.classes_changed) {
            sortOptionsList(cache.classes_list);
            cache.classes_changed = false;

            cache.classes_list.forEach(function(item) {
                if(item.sources.length) {
                    var html = item.sources[0];
                    if(item.sources.length > 1) {
                        html += ' + ' + (item.sources.length - 1) + ' more';
                    }
                    item.html = '<span>' + item.name + '</span><small>' + html + '</small>';

                    //item.html = item.sourceids.join(',') + item.html;
                }
            })
        }
        if(cache.ids_changed) {
            sortOptionsList(cache.ids_list);
            cache.ids_changed = false;
        }

        if(cache.path_changed){
            cache.path_changed = false;

        }
    }

    // this was added to help the editor intellisense
    this.rebuild = function(section){

        switch (section){
            case "id":
                cache.ids_changed = true;
                break;
            case "class":
                cache.classes_changed = true;
                break;

            case "path":
                cache.path_changed=true;
                break;

            default:
                cache.ids_changed = true;
                cache.classes_changed = true;
                cache.path_changed=true;
                break;
        }
        rebuildCacheLists();
    }

    var endInspectingProject = function() {
        rebuildCacheLists();
        /*
        for(var attr in cache.attribute_values) {
            if (cache.attribute_values.hasOwnProperty(attr)) {
                cache.attribute_values[attr].sort();
            }
        }
        for(var tag in cache.tags_attribute_values) {
            if (cache.tags_attribute_values.hasOwnProperty(tag)) {
                for(var attr in cache.tags_attribute_values[tag]) {
                    if (cache.tags_attribute_values[tag].hasOwnProperty(attr)) {
                        cache.tags_attribute_values[tag][attr].sort();
                    }
                }
            }
        }
        */
        //console.log(cache);

        handlers_on_end_inspecting_project.forEach(function(func) {
            func(_this);
        })
    }

    var profile_total = null;

    $('body').one('pinegrow-ready', function (e, pinegrow) {

        var f = new PgFramework('pg.insight.events', 'PG Insight');
        pinegrow.addFramework(f);

        f.default = true;
        f.show_in_manager = false;

        f.on_project_scan_begin = function(nothing, project, ref) {
            profile_total = new CrsaProfile();
            beginInspectingProject();
            profile_total.pause();
        }

        f.on_project_scan_page = function(nothing, project, page, ref) {
            profile_total.resume();
            inspectAProjectItem(page);
            profile_total.pause();
        }

        f.on_project_scan_end = function(nothing, project, ref) {
            profile_total.show('PgInsight.on_project_scan_end begin');
            profile_total.resume();
            endInspectingProject();
            profile_total.show('PgInsight.on_project_scan_end end');
        }

        var rescan_timeout = null;

        var rescan_delay = 1000;

        var scanPage = function(page, changedPgel, do_html, do_css, do_path) {
	        if(rescan_timeout) {
                clearTimeout(rescan_timeout);
            }
            rescan_timeout = setTimeout(function() {
                var profile = new CrsaProfile();
                var start = new Date().getTime();

                stylesheets_done = {};

                try {
                    inspectAProjectItem(page, changedPgel, do_html, do_css , do_path);
                } catch(err) {}
                endInspectingProject();
                profile.show('PgInsight.rescanChangedPage');
                var elapsed = new Date().getTime() - start;

                rescan_delay = elapsed * 50;
                if(rescan_delay < 1000) rescan_delay = 1000;
                if(rescan_delay > 60000) rescan_delay = 60000;

                //rescan_delay = 500;

                //console.log('PgInsight.rescan_delay = ' + rescan_delay);

                rescan_timeout = null;
            }, /* changedPgel ? 0 : */ rescan_delay)

        }

        f.on_page_changed = function(page, pgel, info) {
            var event_type = null;

            if(info) event_type = info.event_type || null;

            //Disable scanning CSS on HTML changes
            if(event_type && event_type == 'change') {
                scanPage(page, null, true, /* true */ false, false);
                return;
            } else if(document.activeElement && document.activeElement.classList.contains('crsa-input')) {
                return;
            }
	        scanPage(page, pgel, true, false /* true */);
        }

        f.on_page_loaded = f.on_plugin_activated = function(page) {
	        scanPage(page, null, true, true, true);
        }

        let getFilePath =  (children) => {
            let exclude = [".git", "_pgbackup", "_pginfo", "dockerfile", ".gitignore"];

             Array.from(children).forEach(file => {
                if (!exclude.includes(file.name)) {
                    if (file.isFile) {
                        // let fileEXT = crsaGetExtFromUrl(file.url);
                        //
                        // if (["html"].includes(fileEXT)) {
                        //     let page = pinegrow.getCrsaPageForUrl(file.url);
                        //     scanPage(page, null, true, true, true);
                        // }

                        // todo (PEP) add file path to intellisense
                    } else {
                        // todo (PEP) add folder path to intellisense
                        getFilePath(file.children);
                    }
                }

            })
        }

        // added this to cache all css, id and path on project loaded
        f.on_project_loaded = (selPage, project)=>{
            openedProject = project;
           getFilePath(openedProject.root.children);
        }

        /*
        pinegrow.addEventHandler('on_css_source_changed', function () {
            var page = pinegrow.getSelectedPage();
            if(page) {
                scanPage(page, null, false, true);
            }
        })
        */

        var stylesheet_change_timers = new WeakMap();

        pinegrow.addEventHandler('on_css_applied_to_pages', function(page, pgStylesheet, ss_list, first_apply) {
            //first_apply is true when the ss is applied for the first time when page view is loaded. we can ignore those.
            if(ss_list.length && (!first_apply || !isApp())) {
                var ss = ss_list[0];
                if(ss) {
                    if (stylesheet_change_timers.has(ss)) {
                        clearTimeout(stylesheet_change_timers.get(ss));
                    }
                    stylesheet_change_timers.set(ss, setTimeout(function () {
                        if(!ss.ownerNode && ss._pg_get_sheet) {
                            //in browser the sheet is recreated when innerhtml of the style is set. a helper function that return sthe updated ss is passed from there
                            ss = ss._pg_get_sheet();
                        }
                        if (ss && ss.ownerNode) {
                            doStylesheet(ss);
                            rebuildCacheLists();
                        }
                        //console.log('scan done for ' + pgStylesheet.url)
                        stylesheet_change_timers.delete(ss);
                    }, pgf.kids ? 1000 : 5000));
                }
                //console.log('schedule scan after ' + 5000 + ' for ' + pgStylesheet.url)
            }
        });

    });


    this.find = function(selector) {
        var r = [];
        forEachSource( function(sourceNode) {
            r = r.concat(sourceNode.find(selector));
        })
        return r;
    }

    var objectValuesToOptionList = function(values) {
        var r = Object.keys(values).sort();
        var list = [];
        for(var i = 0; i < r.length; i++) {
            list.push({key: r[i], name: r[i]});
        }
        return list;
    }

    this.getClasses = function() {
        return cache.classes_list;
    }

    this.getClassesForPage = function(page) {
        var sources = [];
        page.stylesheets.stylesheets.forEach(function(cs) {
            if(cs.sourceStylesheet) cs = cs.sourceStylesheet;
            sources.push('c' + cs.getUniqueId());
        })
        sources.push('p' + page.uid);

        var r = [];
        cache.classes_list.forEach(function(item) {
            for(var i = 0; i < item.sourceids.length; i++) {
                if(sources.indexOf(item.sourceids[i]) >= 0) {
                    r.push(item);
                    break;
                }
            }
        })
        return r;
    }

    this.getCSSVarsForPage = function(page) {
        var sources = [];
        page.stylesheets.stylesheets.forEach(function(cs) {
            if(cs.sourceStylesheet) cs = cs.sourceStylesheet;
            sources.push('c' + cs.getUniqueId());
        })
        sources.push('p' + page.uid);

        var r = [];
        cache.css_vars_list.forEach(function(item) {
            for(var i = 0; i < item.sourceids.length; i++) {
                if(sources.indexOf(item.sourceids[i]) >= 0) {
                    r.push(item);
                    break;
                }
            }
        })
        return r;
    }

    this.getValuesForAttribute = function(attribute, tag) {
        var map = null;
        if(tag) {
            map = cache.tags_attribute_values[tag] || {};
        } else {
            map = cache.attribute_values;
        }
        return objectValuesToOptionList(map[attribute] || {});
    }

    this.getContentsForTag = function(tag) {
        return objectValuesToOptionList(cache.contents[tag] || {});
    }

    this.getMetaValuesForName = function(name) {
        return objectValuesToOptionList(cache.meta[name] || {});
    }

    this.getIds = function() {
        return cache.ids_list;
    }

    this.getKeywords = function() {
        if(cache.keywords_changed) {
            sortOptionsList(cache.keywords_list);
            cache.keywords_changed = false;
        }
        return cache.keywords_list;
    }
}