
//will move this to PG core later
var PgPluginBuilder = function (f) {

    var count = 0;

    this.debug = false;

    var previewIframe;

    this.usePreviewIframe = function(options) {
        var url = f.getResourceUrl('preview/index.html');
        previewIframe = $('<iframe class="pg-preview-frame" src="' + url + '" frameborder="0" scrolling="false"></iframe>').appendTo($('body')).get(0);
        pgHide(previewIframe);

        var scale = 1.0;
        var empty_body_h = null;

        var delayed_tasks = new PgDelayedTasks();

        var isStuck = function() {
            if(previewIframe && $(previewIframe).is(':visible') && !pgPreview.isVisible()) {
                pgHide(previewIframe);
            } else {
                delayed_tasks.run('isStuck', isStuck, 5000);
            }
        }

        f.on_lib_show_preview = function(page, c, $cel, preview) {
            if(!previewIframe.contentDocument || !previewIframe.contentDocument.documentElement) {
                return false;
            }

            var stage = previewIframe.contentDocument.documentElement.querySelector(c.preview_stage_selector || '#preview-stage');

            var current_h = 0;

            var updateHeight = function(h) {
                content_h = h || stage.clientHeight;
                if(current_h != content_h) {
                    previewIframe.style.height = (content_h + empty_body_h) + 'px';
                    current_h = content_h;
                    preview.setContentSize(content_w * scale, content_h * scale + empty_body_h);
                }
            }

            preview.onShowFirstMutate = function(win) {

                if(win !== previewIframe.ownerDocument.defaultView) {
                    $(previewIframe).appendTo(win.document.body);
                }

                while(stage.hasChildNodes()) {
                    stage.removeChild(stage.lastChild);
                }

                if(empty_body_h === null) {
                    empty_body_h = previewIframe.contentDocument.body.clientHeight;
                }

                stage.insertBefore($cel.get(0), null);

                stage.querySelectorAll('img').forEach(function(img) {
                    img.addEventListener('load', function() {
                        updateHeight();
                    })
                })

                if(c.big_preview) {
                    previewIframe.classList.add('pg-preview-frame-big');
                    scale = 0.5;
                } else {
                    previewIframe.classList.remove('pg-preview-frame-big');
                    scale = 1.0;
                }

            }

            var content_h = 0;
            var content_w = 400;

            preview.onShowSecondMeasure = function() {

            }

            preview.onShowThirdMutate = function(x, y, info, fixed) {
                pgFastdomTasks.stopAll('hidePreview' + f.key);

                pgShow(previewIframe);

                delayed_tasks.run('isStuck', isStuck, 5000);

                content_h = stage.clientHeight;
                content_w = previewIframe.contentDocument.body.clientWidth;

                updateHeight(content_h);

                var cw = (content_w || 300) * scale;
                var ch = content_h * scale;

                var pw = cw + empty_body_h;

                preview.setContentSize(cw, ch + empty_body_h);

                if(!fixed) {

                    if (info.pos == 'left') {
                        info.x = Math.max(0, info.target_pos.left - pw - 0);
                    } else {
                        info.x = info.target_pos.left + info.container_w;
                        if (info.x + pw > info.win_w) {
                            info.x = info.win_w - pw;
                        }
                    }

                    var add_h = 80;

                    if (info.y + ch + add_h > info.win_h) {
                        info.y = info.win_h - ch - add_h;
                    }
                    if (info.y < 0) {
                        info.y = 0;
                    }
                }

                previewIframe.style.left = (info.x + 8) + 'px';
                previewIframe.style.top = (info.y + 8) + 'px';

                return info;
            }

            pgPreview.onHide = function() {
                var task = pgFastdomTasks.create('hidePreview' + f.key);
                task.mutate(function() {
                    pgHide(previewIframe);
                    delayed_tasks.cancel('isStuck');
                })
            }
            return true;
        }
    }

    var getType = function (what) {
        what = what || 'comp';
        count++;
        return f.type + '.' + what + '.' + count;
    }

    this.addSection = function (name) {
        var section = new PgFrameworkLibSection(getType('section'), name);
        f.addLibSection(section);
        return section;
    }

    this.readFile = function (name) {
        var url = crsaIsAbsoluteUrl(name) ? name : f.getResourceUrl(name);
        return pinegrow.getSourceNodeOfUrl(url, false /* no clone */, false /* no assign_ids */);
    }

    this.getSourceNode = function(url, done) {
        var url = crsaIsAbsoluteUrl(url) ? url : f.getResourceUrl(url);
        if(crsaIsFileUrl(url)) {
            done(pinegrow.getSourceNodeOfUrl(url, false /* no clone */, false /* no assign_ids */));
        } else {
            $.ajax({
                url: url,
                data: null,
                dataType: 'text'
            })
                .done(function(data) {
                    var p = new pgParser();
                    p.assignIds = false;
                    p.parse(data);
                    done(p.rootNode);
                })
                .fail(function(a) {
                   done(null);
                });
        }
    }

    this.replaceHolderImages = function(pgel) {
        pgel.walkSelfAndChildren(function(node) {
            var src = node.getAttribute('data-src');
            if(src && src.indexOf('holder.js') >= 0) {
                src = pinegrow.getPlaceholderImage();
                node.setAttribute('src', src);
                node.removeAttribute('data-src');
            } else {
                if(node.tagName === 'svg' && node.hasClass('bd-placeholder-img')) {
                    node.replaceWith( pgCreate(`<img src="${pinegrow.getPlaceholderImage()}" />`) );
                }
            }
            return true;
        })
    }

    this.add = function (name, code, options, section) {
        var type = getType();

        var c = new PgComponentType(type, name);

        c.code = code;

        c.preview_image = c.type + '.png';
        c.button_image = c.preview_image;
        c.no_lib_label = true;

        //c.tags = 'block';

        if (options) {
            $.each(options, function (key, val) {
                c[key] = val;
            });
        }

        if (this.debug) {
            try {
                var $el = createElementFromDefinition(c);
            } catch (err) {
                console.log('Invalid definition for ' + c.name + ' (' + c.type + ') ' + err);
            }
        }

        f.addComponentType(c);

        if (section) {
            section.addComponentType(c);
        }
        return c;
    }

    this.addSourceComponent = function (node, selector, name, options, section) {
        var els = node.find(selector);
        var c;
        els.sort(function(a, b) {
            if(a.startStringIndex > b.startStringIndex) {
                return 1;
            } else if(a.startStringIndex < b.startStringIndex) {
                return -1;
            } else {
                return 0;
            }
        });
        for (var i = 0; i < els.length; i++) {
            c = this.add(name + (els.length > 1 ? (' ' + (i + 1)) : ''), els[i].toStringOriginal(true), options, section);
        }
        return c;
    }

    this.addScriptToComponent = function (c, script) {

        c.add_script_on_insert = script;
        /*
         if(c.addEventHandler) {
         c.addEventHandler('on_inserted', function($el, page) {
         //replace ids
         var id = $el.attr('id');
         if(id) {
         script = script.replace(/\$ELEMENT_ID/g, id);
         }
         })
         }
         */
    }

    this.takePhotos = function (stage_url, done) {
        var url = f.getResourceUrl(stage_url);

        var gui = require('nw.gui');

        var current = 0;

        var preview_win = gui.Window.open(url, {
            width: 1024,
            height: 600,
            transparent: false
        });

        var $body;
        var $photo;

        var loaded = false;

        preview_win.on('loaded', function () {
            if (loaded) return;
            loaded = true;

            var first_item = true;

            setTimeout(function () {

                $body = $(preview_win.window.document.body);
                var $stage = $body.find('#photo-stage');

                var photo_list = [];

                for (var t in f.component_types) {
                    photo_list.push(f.component_types[t]);
                }

                var doOne = function () {
                    if (photo_list.length === 0) {
                        if (done) done();
                    } else {
                        var c = first_item ? photo_list[0] : photo_list.shift(); //repeat the first one
                        first_item = false;

                        takePhoto(c, preview_win, $stage, function () {
                            doOne();
                        })
                    }
                }

                doOne();

            }, 500);
        })
    }

    var takePhoto = function (c, win, $stage, done) {
        var gui = require('nw.gui');

        var body = $stage.closest('body').get(0);

        //pinegrow.scrollToElement($el);

        var $el = createElementFromDefinition(c);

        if (!$el) {
            done();
            return;
        }

        var filename = f.getResourceFile('images/' + c.type + '.png');

        var delay = 100;

        var $stage_inner = $stage;
        $stage.empty();

        var script = null;

        if (c.take_photo) {
            if (c.take_photo.delay) {
                delay = c.take_photo.delay;
            }
            if (c.take_photo.stage_inner) {
                //like <div><div class="take_photo_inner"></div></div>
                var $wrap = $(c.take_photo.stage_inner).appendTo($stage);
                $stage_inner = $stage.find('.take_photo_inner');
            }
            if (c.take_photo.script) {
                script = c.take_photo.script;
            } else if (c.add_script_on_insert) {
                script = c.add_script_on_insert;
                if (script.indexOf('$ELEMENT_ID') >= 0) {
                    var id = $el.attr('id') || getUniqueId('element');
                    $el.attr('id', id);
                    script = script.replace(/\$ELEMENT\_ID/g, id);
                }
            }

        }


        $stage_inner.append($el);

        if (script) {
            var scr = body.ownerDocument.createElement('script');
            scr.async = false;
            scr.text = script;
            body.appendChild(scr);
        }

        if (c.take_photo && c.take_photo.func) {
            c.take_photo.func($el);
        }

        var ui_zoom = 1;

        //wait for menu to close
        setTimeout(function () {

            if (c.take_photo && c.take_photo.selector) {
                $el = $el.find(c.take_photo.selector);
            }

            if (c.take_photo && c.take_photo.get_subject) {
                $el = c.take_photo.get_subject($el);
            }

            var zoom = 1;

            var pos = $el.offset();
            pos.width = $el.outerWidth();
            pos.height = $el.outerHeight();

            //debugger;
            if (pos.height > win.window.innerHeight) {

                zoom = win.window.innerHeight / pos.height;
                $el.css('transform', 'scale(' + zoom + ', ' + zoom + ')').css('transform-origin', 'top left');
                pos.width *= zoom;
                pos.height *= zoom;
            }

            setTimeout(function () {

                var left = pos.left * ui_zoom;
                var top = pos.top * ui_zoom;
                var width = pos.width * ui_zoom;
                var height = pos.height * ui_zoom;

                if (width == 0 || height == 0) {
                    console.log('Can not takePhoto, dimensions are 0', $el);
                    if (done) done();
                }

                var scale = 1.0;

                var max_width = 400;

                if (max_width && width > max_width) scale = 1.0 * max_width / width;

                win.capturePage(function (datauri) {

                    var canvas = document.createElement('canvas');
                    canvas.setAttribute('width', width);
                    canvas.setAttribute('height', height);
                    var ctx = canvas.getContext('2d');
                    var img = new window.Image();
                    img.addEventListener("load", function () {
                        ctx.drawImage(img, left, top, width, height, 0, 0, width, height);

                        var smallCanvas = scale < 1.0 ? downScaleCanvas(canvas, scale) : canvas;

                        if (filename) {
                            var image = smallCanvas.toDataURL('image/png');
                            image = image.replace('data:image/png;base64,', '');
                            var buffer = new Buffer(image, 'base64');
                            require('fs').writeFileSync(filename, buffer);
                        }

                        if (done) done(smallCanvas);

                    }, false);
                    img.src = datauri;

                    $el.remove();

                }, {format: 'png', datatype: 'datauri'});

            }, 100);

        }, delay);
    }

    this.addResource = function(url, detect) {
        var relative_url = null;
        var source = null;

        if(crsaIsAbsoluteUrl(url)) {

        } else {
            relative_url = url;
            if(isApp()) {
                url = f.getResourceUrl('resources/' + url);
                source = url;
            }
        }
        res = new PgComponentTypeResource(url);
        res.relative_url = relative_url;
        res.source = source;
        res.detect = detect;
        res.replace = true;
        res.footer = res.type == 'application/javascript';
        f.resources.add(res);
        return res;
    }

    this.addResourceCode = function(code, detect, footer, resources) {
        res = new PgComponentTypeResource('http://dummy/file.js');
        res.code = code;
        res.detect = detect;
        res.replace = true;
        res.footer = footer;
        (resources || f.resources).add(res);
        return res;
    }
}