/**
 * Created by Matjaz on 06/11/15.
 */

class PgSearch {

    constructor() {
        var _this = this;
        this.results = [];

        pinegrow.addEventHandler('on_page_menu', function(page, items) {
            items.push({
                label: 'Remove attributes...',
                kbd: null,
                action: function() {
                    pinegrow.showPrompt(`Enter the name of the attribute you want to remove from all elements on the page. Use <code>*</code> to match a pattern, for example <code>data-pg-*</code>.`, 'Remove attributes', '', 'data-pg-*', null, function(value) {
                        var count = 0;
                        value = value.trim()
                        if(value === '') {
                            pinegrow.showQuickError('Please enter a value.')
                            return false;
                        }
                        var restr = '^' + pgParserEscapeRegExp(value) + '$';
                        restr = restr.replace('\\*', '.*');

                        var re = new RegExp(restr, 'i');
                        pinegrow.makeChanges(page, 'Remove attribute ' + value, function() {
                            page.sourceNode.walk(function(pgel) {
                                if(pgel.isElement) {
                                    pgel.getAttrList().slice(0).forEach(function(a) {
                                        if(a.name.match(re)) {
                                            pgel.removeAttribute(a.name)
                                            count++;
                                        }
                                    })
                                }
                                return true;
                            })
                        })
                        if(count === 0) {
                            pinegrow.showQuickError('No such attributes found!')
                        } else {
                            pinegrow.showQuickMessage(count + ' attributes removed!')
                        }
                        return false;
                    }, 'Close', 'Remove');

                    //_this.search(page.sourceNode);
                }
            })

            return items;
        });
    }

    addError(pgel, attr, val) {
        this.results.push({pgel: pgel, attr: attr, value: val});
    }

    elementSelector(pgel) {
        return '<b data-pg-id="' + pgel.getId() + '">' + pgel.getName({short: true}) + '</b>';
    }


    describeResults() {
        var html = '';
        if(this.results.length === 0) {
            html += '<li>No elements found.</li>';
        } else {
            this.results.forEach(function (e) {
                var text = crsaGetSummaryStr(e.pgel.text(null, true) || '', 10, true);
                html += '<li data-pg-id="' + e.pgel.getId() + '"><span>' + e.pgel.getName({short: true}) + '</span></li>';
            })
        }
        return html;
    }

    search(node) {
        var _this = this;

        var msg = '<p class="status"></p><p>Results:</p><ul style="user-select: text;">';
        msg += '</ul>';

        var page = node.getPage();

        var q = new PgQuickDialog('Remove attributes in ' + page.getName(), function($element) {
            msg += '<p class="help">Click on the element on the above list to select it.';
            msg += '</p>';

            var update = function() {
                var $ul = $element.find('ul');
                $ul.html( _this.describeResults());
            }

            $element.addClass('pg-code-validator-results');

            $element.on('click', '[data-pg-id]', function(e) {
                e.preventDefault();
                var pgid = $(this).attr('data-pg-id');
                var pgel = getPgNodeByPgId(pgid);
                if(pgel) {
                    pgel.scrollTo();
                    setTimeout(function() {
                        pinegrow.selectElement(pgel);
                        pinegrow.showQuickMessage(pgel.getName({short:true}) + ' was selected.')
                    }, 100)
                }
            })

            msg += '<button class="pg-button pg-button-primary refresh">Refresh</button>';

            $element.on('click', 'button.refresh', function(e) {
                e.preventDefault();
                validate();
            })

            return msg;
        }, {width:480})

    }

}



$(function () {

    if(!pgf.edit_html) return;

    var pgSearch = null;

    $('body').one('pinegrow-ready', function (e, pinegrow) {//Wait for Pinegrow to start-up
        pgSearch = new PgSearch();
    })
});