class PgSimplePugNode {
    constructor(tag, id, classes, text, spaces) {
        this.tag = tag;
        this.id = id;
        this.classes = classes || [];
        this.children = [];
        this.parent = null;
        if(text) {
            this.children.push(text);
        }
        this.spaces = spaces;
    }

    addChild(child) {
        this.children.push(child);
        if(typeof child !== 'string') child.parent = this;
    }

    toString(indent) {
        if(indent === undefined) indent = '';
        var html = indent;
        var root = this.tag === '__root__';

        if(this.tag) {
            if(!root) {
                html += `<${this.tag}`;
                if(this.id) {
                    html += ` id="${this.id}"`;
                }
                if (this.classes.length) {
                    html += ` class="${this.classes.join(' ')}"`;
                }
            }
            switch(this.tag.toLowerCase()) {
                case 'a':
                    html += ' href="#"';
                    break
                case 'img':
                    html += ' src="$IMAGE_URL"';
                    break;
            }
            if(this.children.length) {
                !root && (html += '>');
                var prev_is_text = false;
                var add_nl = false;
                this.children.forEach(function(child) {
                    if(typeof child === 'string') {
                        if(prev_is_text) html += ' ';
                        html += child;
                        prev_is_text = true;
                    } else {
                        html += (html.length ? "\n" : '') + child.toString(indent + (root ? '' : '    '));
                        prev_is_text = false;
                        add_nl = true;
                    }
                })
                if(add_nl && !root) html += "\n" + indent;
            }
            if(!root) {
                var self_closed = pgSingleTags.indexOf(this.tag.toLowerCase()) >= 0;
                if (this.children.length || !self_closed) {
                    (this.children.length === 0) && (html += '>');
                    html += `</${this.tag}>`;
                } else {
                    html += ' />';
                }
            }
        }
        return html;
    }
}

class PgSimplePug {
    constructor() {
    }

    render(code) {
        var lines = code.split(/[\n\r]+/);
        var prev_spaces = null;
        var root = new PgSimplePugNode('__root__');
        var prev_node = null;
        var line_count = 0;

        lines.forEach(function(line) {
            line_count++;
            var m = line.match(/^(\s*)(.*)$/);
            if(m) {
                var spaces = m[1].replace(/\t/,'    ').length;

                if(prev_spaces === null) {
                    prev_spaces = spaces;
                    root.spaces = spaces;
                    prev_node = root;
                }

                var line_m = m[2].match(/^([^\s\.\#]+)(#[^\s\.]+)?(\.\S+)?\s*(.*)/);
                if(!line_m) {
                    throw `Syntax error in line ${line_count} - ${line}. Only tags, ids, classes and text are supported.`
                }

                var tag = line_m[1];
                var id = line_m[2];
                if(id) {
                    id = id.replace('#', '');
                }
                var classes = line_m[3];
                var text = line_m[4];

                var node = null;

                if(tag === '|') {
                    node = text || '';
                } else {
                    if(classes) {
                        classes = classes.replace('.', '').split('.');
                    }
                    node = new PgSimplePugNode(tag, id,classes || [], text, spaces);
                }

                if(prev_node.spaces < spaces || prev_node === root) {
                    //this will be a child
                    prev_node.addChild(node);
                } else if(prev_node.spaces === spaces) {
                    //sibling
                    prev_node.parent.addChild(node);
                } else {
                    //find closest parent with this level
                    var p = prev_node.parent;
                    while(p && p.spaces > spaces) {
                        p = p.parent;
                    }
                    if(!p || !p.parent) {
                        throw `Bad indent in line ${line_count} - ${line}`;
                    }
                    p.parent.addChild(node);
                }
                if(typeof node !== 'string') {
                    prev_node = node;
                }
            } else {

            }
        })

        //console.log(`---${root.toString()}---`);

        return root.toString();
    }
}