$(function () {
    $('body').one('pinegrow-ready', function (e, pinegrow) {//Wait for Pinegrow to start-up

        // Create new Pinegrow framework object
        var f = new PgFramework('pg.project.items', 'Project items');
        pinegrow.addFramework(f);

        f.default = true;
        f.show_in_manager = false;

        var requireSelectedElement = function(callback) {
            f.requireSelectedElement(callback);
        }

        var openTemplateBrowser = function (projectTemplates, selectedProjcetTemplate, callback) {
            if(projectTemplates) {
                crsaTemplateBrowser = new CrsaProjectBrowser();
                crsaTemplateBrowser.title = "Add a page to your project";
                crsaTemplateBrowser.intro = '<p>Choose a page to add to your project:</p>';
                crsaTemplateBrowser.setProjects(projectTemplates);

                crsaTemplateBrowser.selectedProject = selectedProjcetTemplate;
                crsaTemplateBrowser.onFileSelected = function(cf, selectedProject) {
                    callback(cf, selectedProject);
                };
                crsaTemplateBrowser.show();
            }
        }

        var addNewFolderTo = function (destPath, folderName) {
            folderName = folderName || "";

            if(!isAppForProject()) {
                pinegrow.showAlert(`<p>At the moment, empty folders can not be created directly. To create a folder include it in the name of the file when creating it with New page or New file command.</p><p>For example, use the new page name <code>subfolder/mypage.html</code> to create the new page and the subfolder <code>subfolder</code>.</p>`, 'About creating folders', null, 'Got it');
                return;
            }

            pinegrow.showPrompt('Name of the new folder', 'New folder', folderName, 'Folder name', null, function(name) {
                if (name.length) {
                    var path = require('path');
                    var folder_path = path.join(destPath, name);
                    if (!crsaIsFileExist(folder_path)) {
                        crsaCreateDirs(folder_path);
                        pinegrow.refreshCurrentProject();
                    } else {
                        pinegrow.showAlert("Folder or file with the same name (<b>" + name + "</b>) exists. Please choose a different name.", "Folder or file already exists", null, "Ok", null, function () {
                            addNewFolderTo(destPath, name);
                        });
                    }
                } else {
                    pinegrow.showAlert('Folder name should not be blank', 'Error', null, "Ok", null, function () {
                        addNewFolderTo(destPath, '');
                    });
                }
            })
        }



        var addNewFileTo = function (destUrl, name) {
            var project = pinegrow.getCurrentProject();

            if(!isAppForProject()) {
                var h = new PgProjectMenuHelper(project);
                h.createNewPage(null, name, destUrl ? project.getRelativeUrl(destUrl) : null);
            } else {
                var s = project.getSettingsForNewFileDialog();

                openTemplateBrowser(s.projects, s.selected, function (cf, selectedProject) {
                    project.createFromMasterPage(destUrl, cf, selectedProject)
                });
            }
        }

        var createNewFile = function(destPath, name) {
            var $m = pinegrow.showPrompt('<p>Enter the name of the new file.</p>' + (!isAppForProject() ? '<p>Include a folder name (for example, <code>folder/myfile.js</code>) if you wish to create the page in a new folder.</p>' : ''), 'Create new file', name || '', 'File name', null, function(name) {
                if (name.length) {
                    var project = pinegrow.getCurrentProject();
                    var exists = false;
                    if(isAppForProject()) {
                        var path = require('path');
                        var file_path = path.join(destPath, name);
                        exists = crsaIsFileExist(file_path);
                    } else {
                        file_path = (destPath ? (destPath + '/') : '') + name;
                        file_path = project.getRelativeUrl(file_path);
                        exists = project.getFileForUrl(file_path);
                    }
                    if (!exists) {

                        var createEmptyFile = function() {
                            if(isAppForProject()) {
                                var fs = require('fs');
                                fs.writeFileSync(file_path, '', {encoding: 'utf8'});
                                pinegrow.refreshCurrentProjectDelayed(true);
                            } else {
                                project.addFile(file_path, '');
                                project.settingsFileIsChanged(file_path);
                                pinegrow.backend.saveDelayed();
                                pinegrow.refreshCurrentProjectDelayed(true);
                            }
                        }

                        if(pinegrow.isFileEditable(file_path)) {
                            pinegrow.showAlert(isisAppForProject() ? `<p>Would you like to create a new HTML page based on one of the ready-made templates or master pages?</p>` : `<p>Would you like to add the boilerplate HTML code to the new page?</p>`, 'Are you creating a new HTML page?', 'No, just create the file', isAppForProject() ? 'Yes, show me the templates' : 'Yes, add the HTML code', function() {
                                createEmptyFile();

                            }, function() {
                                addNewFileTo(isAppForProject() ? crsaMakeUrlFromFile(destPath) : destPath, name);
                            })
                        } else {
                            createEmptyFile();
                        }
                    } else {
                        pinegrow.showAlert("Folder or file with the same name (<b>" + name + "</b>) exists. Please choose a different name.", "File already exists", null, "Ok", null, function () {
                            createNewFile(destPath, name);
                        });
                    }
                } else {
                    pinegrow.showAlert('File name should not be blank', 'Error', null, "Ok", null, function () {
                        createNewFile(destPath, '');
                    });
                }
            }, null, null, null, function(val) {
                var info = {};
                if(!pinegrow.backend.isFileNameAllowed(val, info)) {
                    return info.reason;
                }
            })
            //$m.find('.modal-body').prepend(`<p></p>`);
        }

        f.on_project_menu = function(page, project) {
            var items = [];

            if(isAppForProject()) {
                items.push({
                    label: 'Reload project',
                    func: function () {
                        pinegrow.refreshCurrentProject(null, false, true /* no restore tags */);
                    }
                })
            }

            if(pgf.edit_html) {

                items.push({
                    label: 'Add new page...',
                    func: function () {
                        addNewFileTo(project.getUrl());
                    }
                })
                items.push({
                    label: 'Add new folder...',
                    func: function () {
                        addNewFolderTo(project.getDir());
                    }
                })

                items.push({
                    divider: true
                })

                items.push({
                    label: 'Create new file...',
                    func: function () {
                        createNewFile(project.getDir());
                    },
                    helptext: 'Use this command to create a blank non-HTML file. Use Add new page for HTML files.'
                })
            }
            if(isAppForProject()) {
                items.push({
                    label: 'Delete backups...',
                    func: function() {
                        pinegrow.showAlert("<p>Are you sure you want to delete all <b>_pgbackup</b> folders in the project?</p><p>_pgbackup folders contain backup copies of changed files, if backup is enabled in Settings.</p>", "Delete all backups?", "No", "Yes", null, function() {

                            project.findFilesByName('_pgbackup').forEach(function(file) {
                                file.delete();
                            })
                            pinegrow.showQuickMessage('Backups deleted!');
                            pinegrow.refreshCurrentProject();
                        });
                        pinegrow.stats.using('prj.deletebackups');
                    }
                })


                var remote_submenu = []

                items.push({
                    divider: true
                })

                items.push({
                    label: 'Remote URLs',
                    submenu: remote_submenu
                })

                remote_submenu.push({
                    label: 'Open URL...',
                    action: function () {
                        pinegrow.showOpenUrlDialogAndOpenPage();
                    }
                })

                var project_urls = project.getRemoteUrls();

                var submenu = [];

                project_urls.forEach(function (url) {
                    var urlstr = pgUrlToDisplayString(url);
                    if (urlstr.length > 50) {
                        urlstr = '&hellip;' + urlstr.substr(urlstr.length - 50);
                    }

                    submenu.push({
                        label: `<span class="" title="${url}">${urlstr}</span><span class="extracmd pg-project-urls-list-remove" style="opacity:0.7;margin-left:20px;position:absolute;right:8px;" title="Remove from the list."><i class="icon icon-close"></i></span>`,
                        extra_action: function () {
                            project.removeRemoteUrl(url);
                            pinegrow.showQuickMessage('URL was removed from the list.');
                        },
                        action: function () {
                            pinegrow.openOrShowPage(url, function (page, opened) {
                                if (opened) {
                                    project.addRemoteUrl(url);
                                }
                            })
                        }
                    })
                })

                if (submenu.length === 0) {
                    submenu.push({
                        type: 'header',
                        label: 'The list is empty.'
                    })
                }

                remote_submenu.push({
                    label: 'Recent URLs',
                    submenu: submenu
                })

                remote_submenu.push({
                    label: 'Learn more...',
                    action: function () {
                        pinegrow.openExternalUrl(pgf.docs.remote_url_help);
                    }
                })


                items.push({
                    divider: true
                })

                items.push({
                    label: 'Close project',
                    func: function () {
                        pinegrow.closeCurrentProject();
                    }
                })
            }

            return items;
        }

        f.on_project_item_preview = function(page, file) {
            if(file && file.isImage()) {
                var fs = require('fs');
                var fileSizeInBytes = null;

                if(isAppForProject()) {
                    try {
                        var stats = fs.statSync(file.path)
                        fileSizeInBytes = stats["size"];
                    } catch (err) {
                    }
                }

                if(!isAppForProject() && file.url.endsWith('.svg')) {
                    if(file.content) {
                        return {
                            code: file.content
                        }
                    }
                } else {
                    return {
                        code: '<img src="' + file.getMappedUrlOrUrl() + '" />', info: '', onLoaded: function (el) {
                            return `${el.naturalWidth} x ${el.naturalHeight}` + (fileSizeInBytes !== null ? `, ${Math.round(fileSizeInBytes / 1024)} KB` : '');
                        }
                    };
                }
            }
        }


        f.on_project_item_get_context_menu = function(page, file, project) {
            var closeFilePage = function (fileURL) {
                var filePage = pinegrow.getCrsaPageByUrl(fileURL);
                if (filePage && filePage.view){
                    filePage.view.closeTab();
                    return true;
                }
                return false;
            }

            var pi = pinegrow.getCurrentProjectInfo();

            var items = [];

            if(pgf.edit_html || file.isEditable) {
                items.push({
                    divider: true,
                    header: 'Page'
                })
            }

            if(pgf.edit_html) {
                if (!file.isFile) {
                    items.push({
                        label: "Add new folder...",
                        func: function () {
                            addNewFolderTo(file.path);
                            pinegrow.stats.using('prj.newfolder');
                        }
                    })
                    items.push({
                        label: "Add new page...",
                        func: function () {
                            addNewFileTo(file.url);
                            pinegrow.stats.using('prj.newpage');
                        }
                    })
                    items.push({
                        label: 'Create new file...',
                        func: function () {
                            createNewFile(file.path);
                        }
                    })
                }
            }


            if(file.isEditable) {
                //check if it is partial
                var pgc = pinegrow.findFrameworkByKey('pg.components');
                var submenu = null;

                if(pgc) {
                    submenu = pgc.pgcGetPartialOpeningOptions(file.url, pi, file.name);
                }
                if(submenu) {

                    items.push({
                        label: 'Open',
                        submenu: submenu
                    })

                } else {

                    items.push({
                        label: 'Open',
                        func: function () {
                            var page = pinegrow.getCrsaPageByUrl(file.url);
                            if(pgf.use_page_wrappers && page && page.wrapper_node) {
                                page.askClose(function() {
                                    pinegrow.openOrShowPage(file.url);
                                })
                            } else {
                                pinegrow.openOrShowPage(file.url);
                            }
                            pinegrow.stats.using('prj.openfile');
                        }
                    })
                }
            }



            if(pgf.edit_html) {

                /*
                if (isAppForProject() && file.isEditable) {
                    items.push({
                        label: 'Open as partial...',
                        func: function () {
                            pinegrow.showAlert('<p>"Open as partial" is now implemented through the <b>Partial</b> action in the Actions panel.</p><p><a href="https://pinegrow.com/docs/pinegrow-pro/partials/" class="external">Learn more about working with partials.</a></p>', 'Partials are back!', null, 'OK');
                            //var openPartial = new pgOpenPartialInContainer(file.url, null, true);
                            pinegrow.stats.using('prj.openpartial');
                        }
                    })
                }

                 */


                if (file.isFile) {
                    items.push({
                        label: 'Open in code editor',
                        func: function () {
                            pinegrow.openFileInCodeEditor(file.path);
                            pinegrow.stats.using('prj.editcode');
                        }
                    })
                }
            }

            if(pgf.use_page_wrappers) {
                items.push({
                    divider: true,
                    header: 'Wrapper'
                })

                items.push({
                    label: 'Open in wrapper',
                    func: function() {
                        var pi = pinegrow.getCurrentProjectInfo();

                        var wrapper_url = pi.getSettingForUrl(file.url, 'wrapper_url');
                        var wrapper_selector = pi.getSettingForUrl(file.url, 'wrapper_selector');

                        if (!wrapper_url) {
                            pinegrow.showPageWrapperDialog(file.url, null, true)
                        } else {
                            var open = function() {
                                pinegrow.openPage(file.url, null, true, null, {
                                    open_with_wrapper: true
                                })
                            }

                            var page = pinegrow.getCrsaPageByUrl(file.url);
                            if(page) {
                                page.askClose(function() {
                                    open();
                                })
                            } else {
                                open();
                            }
                        }
                    }
                })

                items.push({
                    label: 'Wrapper settings...',
                    func: function() {
                        pinegrow.showPageWrapperDialog(file.url)
                    }
                })
            }

            if(isAppForProject()) {
                var file_page = pinegrow.getCrsaPageByUrl(file.url) || project.getBackgroundPageForUrl(file.url);
                if (file_page) {
                    items.push({
                        divider: true,
                    })

                    items.push({
                        label: 'Save',
                        func: function () {
                            if (canSavePage(file_page)) {
                                file_page.save(null, true, true, false, true);
                            }
                        }
                    })
                }
            }

            if(pgf.edit_html) {
                items.push({
                    divider: true,
                })

                if(file.isFile) {
                    if (isAppForProject()) {
                        items.push({
                            label: 'Duplicate...',
                            func: function () {
                                pinegrow.stats.using('prj.duplicate');
                                pinegrow.showPrompt('Name of the new file', 'Duplicate file', file.name, null, null, function (name) {
                                    if (name.length && name != file.name) {
                                        try {
                                            var project_info = project.getProjectInfo();
                                            var page_info = project_info.getSettingsForFile(file.path);

                                            var new_file = file.duplicate(name);
                                            for (var key in page_info) {
                                                project_info.setSettingForFile(new_file.path, key, page_info[key]);
                                            }
                                            project_info.save();

                                            var onDone = function (notHtml) {
                                                if (new_file && !notHtml) {
                                                    pinegrow.openOrShowPage(new_file.url, function (page) {
                                                        pinegrow.refreshCurrentProject();
                                                        page.callFrameworkHandler('on_page_cloned', pinegrow.getCrsaPageForUrl(file.url));
                                                    });
                                                } else {
                                                    pinegrow.refreshCurrentProject();
                                                }
                                            }

                                            if (pinegrow.isFileEditable(crsaMakeFileFromUrl(new_file.url))) {
                                                //this would fit better with comps
                                                var new_cp = pinegrow.getCrsaPageForUrl(new_file.url);
                                                if (new_cp) {
                                                    var defs = new_cp.sourceNode.find('[data-pgc-define]');
                                                    if (defs.length > 0) {
                                                        var names = [];
                                                        for (var i = 0; i < defs.length; i++) {
                                                            names.push(defs[i].getAttr('data-pgc-define'));
                                                        }
                                                        pinegrow.showAlert('<p>The page contains <b>component definitions</b> for components: ' + names.join(', ') + '.</p><p>Do you want to:<ul><li><b>Keep these definitions in the new page</b> (and have duplicate definitions, which you would need to fix later)<br />- or -</li><li><b>Replace component definitions with component instances in the new page</b>?</li></ul></p>', 'What to do with component definitions?', 'Keep definitions', 'Replace with instances', function () {
                                                            //keep
                                                            onDone();
                                                        }, function () {
                                                            //replace
                                                            pinegrow.findFrameworkByKey('pg.components').changeComponentDefinitionsToComponentInstances(new_cp.sourceNode);
                                                            new_cp.save();
                                                            onDone();
                                                        })
                                                    } else {
                                                        onDone();
                                                    }
                                                    new_cp
                                                } else {
                                                    onDone(true);
                                                }
                                            } else {
                                                onDone(true);
                                            }
                                        } catch (err) {
                                            pinegrow.showAlert(err, 'Error');
                                        }
                                    } else {
                                        pinegrow.showQuickMessage('Nothing to do...');
                                    }

                                })
                            }
                        })
                    } else {
                        items.push({
                            label: 'Duplicate...',
                            func: function() {
                                var h = new PgProjectMenuHelper(project);
                                h.createNewPage(project.getContentOfUrlWithLookup(file.url), null, crsaGetBaseForUrl(file.url), !pinegrow.isFileEditable(file.name));
                            }
                        })
                    }
                }

                if(!isAppForProject()) {
                    items.push({
                        label: 'Rename...',
                        func: function() {
                            var h = new PgProjectMenuHelper(project);
                            h.renameFile(file.url);
                        }
                    })
                }

                if(true) {
                    items.push({
                        label: 'Delete...',
                        func: function () {
                            pinegrow.showAlert("Are you sure you want to delete <b>" + file.name + "</b>?", "Delete file/folder", "No", "Yes, delete it!", null, function () {

                                var onDelete = function () {
                                    closeFilePage(file.url);
                                    pinegrow.showQuickMessage('<b>' + file.name + '</b> deleted successfully.');
                                    pinegrow.refreshCurrentProject();
                                }

                                if (isAppForProject()) {
                                    if (file.delete()) {
                                        onDelete();
                                    } else {
                                        pinegrow.showQuickMessage('Failed to delete <b>' + file.name + '</b>.');
                                    }
                                } else {
                                    var list = []
                                    var list_urls = [];

                                    if(file.isFile) {
                                        list.push(file);
                                        list_urls.push(file.url);
                                    } else {
                                        //folder
                                        file.walkThroughFiles(function(ch) {
                                            if(ch.isFile) {
                                                list.push(ch);
                                                list_urls.push(ch.url);
                                            }
                                        })

                                    }

                                    pinegrow.backend.deletePages(project, list_urls, function () {
                                        list.forEach(function(f) {
                                            closeFilePage(f.url);
                                            project.removeFileNodeFromList(f);
                                        })
                                        if(!file.isFile) {
                                            project.removeFileNodeFromList(file);
                                        }
                                        pinegrow.showQuickMessage(`${list.length > 1 ? (list.length + ' files ') : list[0].name} deleted successfully.`);
                                        pinegrow.refreshCurrentProject();
                                    });
                                }
                            });
                            pinegrow.stats.using('prj.delete');
                        }
                    })
                }

                if(isAppForProject()) {
                    items.push({
                        label: 'Rename...',
                        func: function () {
                            pinegrow.stats.using('prj.rename');
                            pinegrow.showPrompt('Enter the new file name', 'Rename file/folder', file.name, "New file name", null, function (name) {
                                if (name.length && name != file.name) {
                                    var filePage = pinegrow.getCrsaPageByUrl(file.url);
                                    if (filePage) filePage.save(null, true);

                                    var renameFile = function () {
                                        try {
                                            var project_info = project.getProjectInfo();
                                            var page_info = project_info.getSettingsForFile(file.path);

                                            var new_file = file.rename(name);
                                            for (var key in page_info) {
                                                project_info.setSettingForFile(new_file.path, key, page_info[key]);
                                            }
                                            project_info.save();
                                            pinegrow.showQuickMessage('<b>' + file.name + '</b> renamed successfully.');
                                            if (closeFilePage(file.url)) {
                                                if (new_file) {
                                                    pinegrow.openOrShowPage(new_file.url, function (cp) {
                                                        cp.detectAndAddFrameworks();
                                                        pinegrow.updatePluginControlsForSelectedPage(cp);
                                                    });
                                                }
                                            }
                                            pinegrow.refreshCurrentProject();
                                        } catch (err) {
                                            pinegrow.showQuickMessage('Failed to rename <b>' + file.name + '</b>.');
                                        }
                                    }

                                    if (file.doesFileExistInSameDir(name)) {
                                        if (!file.isDirectoryInTheSameDir(name)) {
                                            pinegrow.showQuickMessage("You can't overwrite directory");
                                        } else {
                                            pinegrow.showAlert("Are you sure you want to overwrite <b>" + name + "</b>?", "overwrite file", "No", "Yes", null, function () {
                                                renameFile();
                                            });
                                        }
                                    } else {
                                        renameFile();
                                    }

                                } else {
                                    pinegrow.showQuickMessage('Nothing to do...');
                                }
                            });
                        }
                    })

                    if (file.isFile) {
                        items.push({
                            label: 'Fix links...',
                            func: function () {
                                pinegrow.stats.using('prj.fixlinks');

                                if (!page) {
                                    pinegrow.showAlert('Please open the page first.', 'The page must be open');
                                } else {
                                    pinegrow.fixLinks(project, page, null, page.url, function () {
                                        page.updateStylesheetsList(function () {
                                            page.detectAndAddFrameworks(function () {
                                                page.refresh();
                                                pinegrow.changeManager.didChangePage(page, 'Fix links');
                                            })
                                        });
                                    });
                                }
                            }
                        })
                    }
                }

                var selectedElementName = pinegrow.getSelectedElementName() || 'selected element';
                var selected_pgel = pinegrow.selectedElements.getLastSelected();

                //var mimeType = pinegrow.getMimeType(file.name);

                var simpleType = file.getSimpleType();

                if (selected_pgel) {
                    if (simpleType == 'image') {
                        items.push({
                            divider: true,
                            header: 'Images'
                        })
                        items.push({
                            label: 'Insert image after <b>' + selectedElementName + '</b>',
                            func: function () {
                                requireSelectedElement(function (cp, pgel) {
                                    var url = cp.makeRelativeUrl(file.url);

                                    var code = '<img src="' + url + '" />';
                                    var imgel = pgCreate(code);

                                    pinegrow.makeChanges(pgel, 'Insert image', function () {
                                        imgel.insertAfter(pgel);
                                    })
                                    pinegrow.showQuickMessage('Image inserted.');
                                });
                            }
                        })

                        if (selected_pgel.canHaveChildren()) {
                            items.push({
                                label: 'Append image to <b>' + selectedElementName + '</b>',
                                func: function () {
                                    requireSelectedElement(function (cp, pgel) {
                                        var url = cp.makeRelativeUrl(file.url);

                                        var code = '<img src="' + url + '" />';
                                        var imgel = pgCreate(code);

                                        pinegrow.makeChanges(pgel, 'Append image', function () {
                                            pgel.append(imgel);
                                        })
                                        pinegrow.showQuickMessage('Image appended.');
                                    });
                                }
                            })
                        }

                        if (selected_pgel.tagName === 'img') {
                            items.push({
                                label: 'Replace image',
                                func: function () {
                                    requireSelectedElement(function (cp, pgel) {
                                        var url = cp.makeRelativeUrl(file.url);

                                        pinegrow.makeChanges(pgel, 'Replace image', function () {
                                            pinegrow.selectedElements.forEachElement(function (pgel) {
                                                var url = pgel.getPage().makeRelativeUrl(file.url);
                                                pgel.setAttribute('src', url)
                                            })
                                        })
                                        pinegrow.showQuickMessage('Image replaced.');
                                    });
                                }
                            })
                        }
                    }
                }

                if (simpleType == 'script') {
                    items.push({
                        divider: true,
                        header: 'Javascript'
                    })

                    var addScript = function (footer) {
                        f.requireSelectedPage(function (page) {
                            var used = file.hasScriptOnPage(page);

                            if (!used) {
                                pinegrow.makeChanges(page, 'Include script', function () {
                                    file.addScriptToPage(page, footer);
                                })
                                pinegrow.showQuickMessage('Script included. Refresh the page view to run it (if Javascript is enabled in Pinegrow).');
                            } else {
                                pinegrow.showQuickError('The script is already included in the ' + used + ' of the page.');
                            }
                        });
                    }

                    items.push({
                        label: 'Include in page Head',
                        func: function () {
                            addScript();
                        }
                    })

                    items.push({
                        label: 'Include in page Footer',
                        func: function () {
                            addScript(true);
                        }
                    })
                }

                if (simpleType == 'css') {
                    if (pinegrow.getMimeType(file.url) === "text/css") {
                        items.push({
                            divider: true,
                            header: 'Stylesheet'
                        })

                        var addStylesheet = function () {
                            f.requireSelectedPage(function (page) {
                                var used = file.hasStylesheetOnPage(page);

                                if (!used) {
                                    pinegrow.makeChanges(page, 'Include stylesheet', function () {
                                        file.addStylesheetToPage(page);
                                    })
                                    pinegrow.showQuickMessage('Stylesheet included.');
                                } else {
                                    pinegrow.showQuickError('The stylesheet is already included in the page.');
                                }
                            });
                        }

                        items.push({
                            label: 'Link the stylesheet to the page',
                            func: function () {
                                addStylesheet();
                            }
                        })
                    }
                }

                //Link to
                items.push({
                    divider: true,
                    header: 'Links'
                })

                var pgel = pinegrow.selectedElements.getLastSelected();

                var label = (pgel && pgel.tagName == 'a') ? 'Link to' : 'Set href/src of';
                items.push({
                    label: label + ' <b>' + selectedElementName + '</b>',
                    func: function () {
                        requireSelectedElement(function (cp, pgel) {
                            var url = cp.makeRelativeUrl(file.url);
                            var src_tags = 'img,script,iframe';
                            var attr;
                            if (pgel.is(src_tags)) {
                                attr = 'src';
                            } else {
                                attr = 'href';
                            }
                            pinegrow.makeChanges(pgel, 'Set ' + attr, function () {
                                pgel.attr(attr, url);
                            })
                            pinegrow.showQuickMessage('<b>' + attr + '</b> set to ' + url);
                        });
                    }
                })

                items.push({
                    label: 'Append link to <b>' + selectedElementName + '</b>',
                    func: function () {
                        requireSelectedElement(function (cp, pgel) {
                            var url = cp.makeRelativeUrl(file.url);

                            var code = '<a href="' + url + '">' + file.name + '</a>';
                            var linkel = pgCreate(code);

                            pinegrow.makeChanges(pgel, 'Insert link', function () {
                                pgel.append(linkel);
                            })
                            pinegrow.showQuickMessage('Link appended.');
                        });
                    }
                })

                items.push({
                    label: 'Insert link after <b>' + selectedElementName + '</b>',
                    func: function () {
                        requireSelectedElement(function (cp, pgel) {
                            var url = cp.makeRelativeUrl(file.url);

                            var code = '<a href="' + url + '">' + file.name + '</a>';
                            var linkel = pgCreate(code);

                            pinegrow.makeChanges(pgel.parent, 'Insert link', function () {
                                linkel.insertAfter(pgel);
                            })
                            pinegrow.showQuickMessage('Link inserted.');
                        });
                    }
                })
            }

            return items;
        }

        f.aaaon_project_loaded = function(selectedPage, project) {
            var links_count = 0;

            project.forEachEditableFile( function( page, pageDone, status ) {
                //on page
                var links = page.sourceNode.find('[href]');
                links_count += links.length;
                status.errors.push('Oh no!');
                pageDone(page, status);

            }, function() {
                //on done
                pinegrow.showQuickMessage('Found ' + links_count + ' links.');
            }, 'Just fooling around...')
        }

    });

});