
class PgApi {

    constructor() {
        this.preview_elements = {};
    }


    previewCodeInPage(ref, code, placement, pgid) {
        this.clearPreviewCodeInPage(ref);
        var pgel = pgid ? getPgNodeByPgId(pgid) : pinegrow.getSelectedElement();
        if(pgel) {
            var $el = pgel.get$DOMElement();
            if($el) {
                var $pel = $(code).attr('data-pg-preview-ref', ref);

                switch(placement) {
                    case 'append':
                        $el.append($pel);
                        break;
                    case 'prepend':
                        $el.prepend($pel);
                        break;
                    case 'insertAfter':
                        $pel.insertAfter($el);
                        break;
                    case 'insertBefore':
                        $pel.insertBefore($pel);
                        break;
                }
                this.preview_elements[ref] = $pel;
            }
        }
    }

    clearPreviewCodeInPage(ref) {
        if(this.preview_elements[ref]) {
            this.preview_elements[ref].remove();
            delete this.preview_elements[ref];
        }
    }

    duplicateElements(collection_or_el, done, group) {

        var collection = pinegrow.getCollection(collection_or_el);

        var changed_list = [];

        var inserted = new PgCollection();

        var last_pgel = null;
        var is_selected = false;

        if(group) {
            collection.orderByDocumentPosition();
        }

        var ok = true;

        collection.forEachElement(function(pgel, i, total) {
            if(!pgel.canDuplicate()) {
                pinegrow.showQuickError(`Can't duplicate ${pgel.tagName} element.`);
                ok = false;
            }
        })

        if(!ok) return inserted;

        pinegrow.repeater.repeat(function(repeat_i, repeat_count) {

            collection.forEachElement(function(pgCurrent, i, total) {

                is_selected = pinegrow.selectedElements.isSelectedElement(pgCurrent);

                var problems = new pgParserSourceProblem(pgCurrent);

                if (!pgCurrent) {
                    problems.add('element', pgCurrent.getName(), 'duplicate');
                }
                if (!problems.ok()) {
                    showAlert(problems.toString(), "Can't duplicate this element");
                    return;
                }

                if (!canMakeChange(pgCurrent, 'duplicate_element')) return;

                var cp = pinegrow.getCrsaPageOfPgParserNode(pgCurrent);

                willMakeChange(pgCurrent.getPage(), "Duplicate element / " + pgCurrent.getName());

                var last = (i == total - 1) && (repeat_i == repeat_count - 1);

                var pgNew = pgCurrent.clone();

                pinegrow.dispatchEvent('on_duplicate_element_async', cp, pgNew, function (pgNew) {

                    if(!last_pgel) last_pgel = pgCurrent;

                    if(group && i == 0) last_pgel = collection.getLast()

                    pgNew.insertAfter(last_pgel);
                    inserted.add(pgNew);

                    last_pgel = pgNew;

                    /*

                    pinegrow.repeater.repeatOthers(function () {
                        var pgNewOther = pgNew.clone();
                        pgNewOther.insertAfter(last_pgel);
                        inserted.add(pgNewOther);
                        last_pgel = pgNewOther;
                    })

                    */

                    if(last) {
                        if(done) {
                            done(inserted);
                        }
                        pinegrow.repeater.done();
                        pinegrow.changeManager.didInsert(inserted);

                        if(is_selected) {
                            if (inserted.getLength() <= 12) {
                                pinegrow.selectedElements.selectMany(inserted, true);
                            } else {
                                pinegrow.selectElement(last_pgel);
                            }
                        }
                        if(pgf.kids) {
                            //var $el = pgNew.get$DOMElement();
                            //if($el && $el.css('position') === 'absolute') {
                                pinegrow.showQuickMessage('Element duplicated and selected.')
                            //}
                        }
                    }
                })
            })

        })
    }

    removeElements(collection_or_el) {
        var collection = pinegrow.getCollection(collection_or_el);

        var changed_list = [];

        collection.forEachElement(function(pgCurrent, i, total) {

            var is_selected = pinegrow.selectedElements.isSelectedElement(pgCurrent);

            var problems = new pgParserSourceProblem(pgCurrent);

            if(!pgCurrent) {
                problems.add('element', 'Element', 'remove');
            }
            if(!problems.ok()) {
                showAlert(problems.toString(), "Can't remove this element");
                return;
            }

            if(!canMakeChange(pgCurrent, 'delete_element')) return;

            //var cp = pinegrow.getCrsaPageOfPgParserNode(pgCurrent);

            var page = pgCurrent.getPage();

            pinegrow.dispatchEvent('on_before_delete_element_async', page, pgCurrent, function(delete_element) {

                if(delete_element) {

                    willMakeChange(page, "Delete element / " + pgCurrent.getName(), pgCurrent, 'delete');

                    var last = i == total - 1;

                    var select_next = null;

                    if(last && is_selected) {
                        select_next = pgCurrent.next();
                        if (!select_next) select_next = pgCurrent.prev();
                        if (!select_next) select_next = pgCurrent.parent;
                    }

                   // pgCurrent.removedFromPage = page;
                    pgCurrent.remove();

                    if(last) {
                        if(select_next) {
                            pinegrow.selectElement(select_next);
                        }

                        pinegrow.changeManager.didDelete(collection);

                        pinegrow.highlightElement(null);
                    }
                }
            });

        })
    }

    mergeElements(collection_or_el, dest_pgel) {
        var collection = pinegrow.getCollection(collection_or_el);

        collection.normalize();

        collection.orderByDocumentPosition();

        var deleted_list = new PgCollection();

        var dest_pos = dest_pgel.getPositionInDocument();

        if(!canMakeChange(dest_pgel, 'insert_element')) return;

        collection.forEachElement(function(pgCurrent, i, total) {

            var is_dest = pgCurrent == dest_pgel;

            if(!is_dest) {
                if(!canMakeChange(pgCurrent, 'delete_element')) return;

                if(pgCurrent.positionInDocument < dest_pos) {
                    pgCurrent.moveChildrenToElement(dest_pgel, true /* prepend */);
                } else {
                    pgCurrent.moveChildrenToElement(dest_pgel, false /* append */);
                }
                pgCurrent.remove();
                deleted_list.add(pgCurrent);
            }
        })

        pinegrow.changeManager.batchChanges(function() {
            pinegrow.changeManager.didDelete(deleted_list);
            pinegrow.changeManager.didChange(dest_pgel, 'Merge elements');
        }, 'Merge elements');

        pinegrow.selectElement(dest_pgel);
    }

    splitParent(collection_or_el) {
        var collection = pinegrow.getCollection(collection_or_el);

        var inserted = new PgCollection();
        var changed = new PgCollection();

        collection.normalize();

        collection.forEachElement(function(pgCurrent, i, total) {

            if(!pgCurrent.parent || !pgCurrent.parent.parent) return;

            if(!canMakeChange(pgCurrent.parent, 'edit_content')) return;
            if(!canMakeChange(pgCurrent.parent.parent, 'insert_element')) return;

            var new_parent = pgCurrent.parent.clone(false, false, true /* just top element */);

            new_parent.insertAfter(pgCurrent.parent);

            //Move pgCurrent and all later to new parent
            var idx = pgCurrent.parent.children.indexOf(pgCurrent);

            var list = pgCurrent.parent.children;

            while(idx >= 0 && idx < list.length) {
                var node = list[idx];
                new_parent.append(node);
            }

            inserted.add(new_parent);
            changed.add(pgCurrent.parent.parent);
        })

        pinegrow.changeManager.batchChanges(function() {
            pinegrow.changeManager.didInsert(inserted);
            pinegrow.changeManager.didChange(changed, 'Split element');
        }, 'Split element')
    }

    insertElements(code_or_col_or_el, dest_collection_or_el, placement, options) {
        options = options || {};
        options = $.extend({select: false, scroll: false, highlight: false, repeater: false, format_html: false, fix_placement: false, before_body_scripts: false, append_before_selector: null}, options);

        placement = placement || 'append';

        var count = options.repeater ? pinegrow.repeater.getCount() : 1;

        var collection = pinegrow.getCollection(dest_collection_or_el);

        var inserted_elements = new PgCollection();

        var change_inside = placement == 'append' || placement == 'prepend';

        var add_elements;
        var is_text = false;

        if(typeof code_or_col_or_el === 'string') {
            add_elements = [];
            var p = new pgParser();
            p.parse(code_or_col_or_el);

            for(var j = 0; j < count; j++) {
                for (var i = 0; i < p.rootNode.children.length; i++) {
                    add_elements.push(j == 0 ? p.rootNode.children[i] :  p.rootNode.children[i].clone());
                }
            }

        } else {
            add_elements = pinegrow.getCollection(code_or_col_or_el).getList().slice();
            for(var j = 1; j < count; j++) {
                pinegrow.getCollection(code_or_col_or_el).getList().forEach(function(pgel) {
                    add_elements.push(pgel.clone());
                })
            }
        }

        for(var i = 0; i < add_elements.length; i++) {
            if(!add_elements[i].isElementOrScript) {
                is_text = true;
                break;
            }
        }

        var first_el = null;

        var updated_parents = new PgCollection();
        collection.forEachElement(function(pgel, i, total) {
            if(placement == 'insertAfter' || placement == 'insertBefore') {
                updated_parents.add(pgel.parent);
            } else {
                updated_parents.add(pgel);
            }
        })
        updated_parents.normalize();

        var html_map;
        var first_parent = null;

        if(is_text) {
            html_map = new Map();
            updated_parents.forEachElement(function(pgel) {
                html_map.set(pgel, pgel.html(null, true));
                if(!first_parent) {
                    first_parent = pgel;
                }
            })
        }

        var doInsertion = function() {
            collection.forEachElement(function(pgel, i, total) {

                var cur_placement = placement;

                for(var i = 0; i < add_elements.length; i++) {
                    if (add_elements[i].isElementOrScript && !canMakeChange(change_inside ? pgel : pgel.parent, 'insert_element', {inserted: add_elements[i]})) return inserted_elements;
                }



                if(cur_placement == 'append' || cur_placement == 'prepend') {
                    if(!pgel.canHaveChildren()) {
                        if(options.fix_placement) {
                            cur_placement = 'insertAfter';
                        } else {
                            pinegrow.showQuickError(pgel.getName({short: true}) + ` can't have sub-elements.`);
                            return;
                        }
                    }
                }

                if(options.fix_placement && ['body', 'head', 'html'].indexOf(pgel.tagName) >= 0 && (cur_placement == 'insertAfter' || cur_placement == 'insertBefore')) {
                    cur_placement = 'append';
                }

                if(pgel.tagName == 'body' && cur_placement == 'append' && options.before_body_scripts) {
                    var has_scripts = false;
                    add_elements.forEach(function(pgel) {
                        if(pgel.tagName == 'script') {
                            has_scripts = true;
                        }
                    })

                    if(!has_scripts) {
                        var last_non_script = pgel.findLastNonScriptChild();
                        if (last_non_script) {
                            pgel = last_non_script;
                            cur_placement = 'insertAfter';
                        } else {
                            cur_placement = 'prepend';
                        }
                    }
                }

                if(options.append_before_selector) {
                    var ap_pgel = pgel.findOne(options.append_before_selector);
                    if(ap_pgel) {
                        pgel = ap_pgel;
                        cur_placement = 'insertBefore';
                    }
                }

                var cloned_add_elements = add_elements.slice();

                while(cloned_add_elements.length) {
                    var idx = (cur_placement == 'insertAfter' || cur_placement == 'prepend') ? cloned_add_elements.length - 1 : 0;
                    var addel = total > 1 ? cloned_add_elements[idx].clone() : cloned_add_elements[idx];

                    inserted_elements.add(addel);

                    if(!first_el && addel.isElementOrScript) {
                        first_el = addel;
                    }

                    switch(cur_placement) {
                        case 'append':
                            pgel.append(addel);
                            break;
                        case 'prepend':
                            pgel.prepend(addel);
                            break;
                        case 'insertAfter':
                            addel.insertAfter(pgel);
                            break;
                        case 'insertBefore':
                            addel.insertBefore(pgel);
                            break;
                    }

                    cloned_add_elements.splice(idx, 1);
                }
            })
        }

        if(is_text) {
            first_parent.withoutEvents(function() {
                doInsertion();
            })
            updated_parents.forEachElement(function(pgel) {
                pgel.updateInner(html_map.get(pgel))
            })
            pinegrow.changeManager.didChange(updated_parents, 'Insert elements');
        } else {
            doInsertion();
            pinegrow.changeManager.didInsert(inserted_elements, null, options.format_html);
        }

        if(options.repeater) {
            pinegrow.repeater.done();
        }

        if(first_el) {//todo30 - some helpers, send event to dom player?
            var $first_el = first_el.get$DOMElement();
            if($first_el) {
                if (options.select) {
                    pinegrow.selectElement($first_el);
                } else if (options.scroll) {
                    first_el.scrollTo();
                }
                if(options.highlight) {
                    setTimeout(function(){
                        pinegrow.highlightElement($first_el);
                    }, 250);
                }
            }
        }

        return inserted_elements;
    }

    pasteAttributes(collection_or_el, code) {
        var collection = pinegrow.getCollection(collection_or_el)

        var list = [];

        var $dummy_span = $('<span ' + code + '></span>');
        var el = $dummy_span[0];
        for (var i = 0, atts = el.attributes, n = atts.length; i < n; i++){
            list.push({name: atts[i].nodeName, value: $dummy_span.attr(atts[i].nodeName)});
        }

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            for(var i = 0; i < list.length; i++) {
                if(!canMakeChange(pgel, 'attr', list[i].name)) {
                    allowed = false;
                    break;
                }
            }
        })

        if(!allowed) return;

        collection.forEachElement(function(pgel) {
            for(var i = 0; i < list.length; i++) {
                pgel.setAttr(list[i].name, list[i].value);
            }
        })

        pinegrow.changeManager.didChange(collection, 'Paste attributes');
    }

    setAttribute(collection_or_el, name, value, change_info) {
        var collection = pinegrow.getCollection(collection_or_el)

        if(collection.getLength() == 0) return;

        var list;
        if(typeof name === 'string') {
            list = [{name: name, value: value}]
        } else {
            list = name;
        }

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            list.forEach(function(attr) {
                if(!canMakeChange(pgel, 'attr', attr.name)) {
                    allowed = false;
                }
            })
        })

        if(!allowed) return;

        collection.forEachElement(function(pgel) {
            list.forEach(function(attr) {
                var quote = null;
                if(typeof attr.value === 'string') {
                    if(attr.value.indexOf(`"`) >= 0) {
                        quote = `'`;
                    }
                    if(attr.value.indexOf(`'`) >= 0) {
                        quote = `"`;
                    }
                }
                pgel.setAttr(attr.name, attr.value, false, quote, (change_info && change_info.event_info) ? change_info.event_info : null);
            })

        })

        if(change_info && change_info.no_change_event) return;

        var event_name = (change_info && change_info.event_name) ? change_info.event_name : ((list.length == 1 ? 'Set attribute ' : 'Set attributes ') + list.map(function(attr) { return attr.name }).join(', '));

        pinegrow.changeManager.didChange(collection, event_name, change_info);
    }

    removeAttribute(collection_or_el, name, change_info) {
        var collection = pinegrow.getCollection(collection_or_el)

        if(collection.getLength() == 0) return;

        var list;
        if(typeof name === 'string') {
            list = [name]
        } else {
            list = name;
        }

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            list.forEach(function(attr) {
                if(!canMakeChange(pgel, 'attr', attr)) {
                    allowed = false;
                }
            })
        })

        if(!allowed) return;

        collection.forEachElement(function(pgel) {
            list.forEach(function(attr) {
                pgel.removeAttr(attr);
            })
        })

        pinegrow.changeManager.didChange(collection, (list.length == 1 ? 'Remove attribute ' : 'Remove attributes ') + list.join(', '), change_info);
    }

    getClassSorter(pgel) {
        var page = pgel.getPage();
        return page ? (page.classSorter || null) : null;
    }

    addClass(collection_or_el, cls_list, change_info, remove_classes) {
        var _this = this;

        var collection = pinegrow.getCollection(collection_or_el)

        if(typeof cls_list == 'string') cls_list = [cls_list];

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            for(var i = 0; i < cls_list.length; i++) {
                if (!canMakeChange(pgel, 'add_class', cls_list[i])) {
                    allowed = false;
                    break;
                }
            }
        })

        if(!allowed) return;

        collection.forEachElement(function(pgel) {
            var page = pgel.getPage();

            if(remove_classes) {
                var remove_list = remove_classes;

                if(typeof remove_classes === 'function') {
                    remove_list = remove_classes(pgel);
                }
                remove_list.forEach(function(cls) {
                    if(page) {
                        page.classSetter.removeClass(pgel, cls, _this.getClassSorter(pgel));
                    } else {
                        pgel.removeClass(cls, _this.getClassSorter(pgel));
                    }
                })
            }

            cls_list.forEach(function(cls) {
                if(page) {
                    page.classSetter.addClass(pgel, cls, _this.getClassSorter(pgel));
                } else {
                    pgel.addClass(cls, _this.getClassSorter(pgel));
                }
                pinegrow.recentClasses.add(cls, true);
            })
        })

        var name = cls_list.length == 1 ? cls_list[0] : (cls_list[0] + ' + ' + (cls_list.length - 1) + ' more');

        pinegrow.changeManager.didChange(collection, 'Add class ' + name, change_info);

        pinegrow.dispatchEventDelayed('on_element_classes_changed', null, collection, change_info);

    }

    getFlexSettings(pgel) {
        var r = {
            flex: false,
            inline: false,
            direction: false,
            grid: false
        }
        var $el = pgel.get$DOMElement();

        if($el) {
            var display = $el.css('display');
            if(display === 'flex') {
                r.flex = true;
            } else if(display === 'inline-flex') {
                r.flex = true;
                r.inline = true;
            } else {
                r.inline = display.indexOf('inline') >= 0;
                r.grid = display.indexOf('grid') >= 0;
            }
            var dir = $el.css('flex-direction');
            if(dir) {
                if(dir.indexOf('row') >= 0) {
                    r.direction = 'row';
                } else {
                    r.direction = 'column';
                }
            }
        }
        return r;
    }

    hasClass(pgel, cls) {
        var page = pgel.getPage();
        if(page) {
            return page.classSetter.hasClass(pgel, cls);
        } else {
            return pgel.hasClass(cls);
        }
    }

    getClasses(pgel) {
        var page = pgel.getPage();
        if(page) {
            return page.classSetter.getClasses(pgel);
        } else {
            return pgel.getClasses();
        }
    }

    removeClass(collection_or_el, cls_list, change_info) {
        var _this = this;
        var collection = pinegrow.getCollection(collection_or_el)

        if(typeof cls_list == 'string') cls_list = [cls_list];

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            for(var i = 0; i < cls_list.length; i++) {
                if (!canMakeChange(pgel, 'remove_class', cls_list[i])) {
                    allowed = false;
                    break;
                }
            }
        })

        if(!allowed) return;

        collection.forEachElement(function(pgel) {
            var page = pgel.getPage();

            cls_list.forEach(function(cls) {
                pinegrow.classStyles && pinegrow.classStyles.ensureStyleWithClassIsSelected(pgel, cls);

                if(page) {
                    page.classSetter.removeClass(pgel, cls, _this.getClassSorter(pgel));
                } else {
                    pgel.removeClass(cls, _this.getClassSorter(pgel));
                }
            })
        })

        var name = cls_list.length == 1 ? cls_list[0] : (cls_list[0] + ' + ' + (cls_list.length - 1) + ' more');

        pinegrow.changeManager.didChange(collection, 'Remove class ' + name, change_info);

        pinegrow.dispatchEventDelayed('on_element_classes_changed', null, collection, change_info);
    }

    //Class properties - set class-based framework property, supporting class styles
    setClassProperty(collection_or_el, cls, values, class_style) {
        var col = pinegrow.getCollection(collection_or_el);
        var pgel = col.getFirst();
        if(!pgel) return false;

        if(class_style === 'auto') {
            class_style = pinegrow.classStyles.getCurrentStyleForPgel(pgel)
        }

        var style = class_style;

        var current_cls = null;

        var getCurrentClass = function(list) {
            var map = {}
            for(var i = 0; i < list.length; i++) {
                map[list[i]] = true;
            }
            for(var i = 0; i < values.length; i++) {
                if(map[values[i].key]) {
                    return values[i].key;
                }
            }
            return null;
        }

        //is op allowed
        var allowed = true;
        col.forEachElement(function(pgel) {
            if (!canMakeChange(pgel, 'add_class', cls)) {
                allowed = false;
            }
        })

        if(!allowed) return;

        if(style) {
            if(style.locked) {
                pinegrow.showQuickError('The style is locked.')
                return false;
            } else {
                current_cls = getCurrentClass(style.getClasses());
                pinegrow.makeChanges(pgel, 'Set ' + cls, function () {
                    pinegrow.classStyles.addClassToStyle(pgel, style, cls, current_cls);
                })
            }
        } else {
            var styles = pinegrow.classStyles.getStylesForPgel(pgel);
            if(styles.length) {
                //this is inline class
                pinegrow.makeChanges(col, 'Set ' + cls, function () {
                    col.forEachElement(function(pgel) {
                        current_cls = getCurrentClass(pinegrow.classStyles.getInlineClassesForPgel(pgel));
                        pinegrow.classStyles.addClassToStyle(pgel, null, cls, current_cls);
                    })

                })
            } else {
                this.addClass(col, [cls], null, function(pgel) {
                    var c = getCurrentClass(pgel.getClasses());
                    return c ? [c] : [];
                });
            }
        }
        return true;
    }

    addAction(collection_or_el, def, change_info, optional_values) {
        var collection = pinegrow.getCollection(collection_or_el)

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            if(!canMakeChange(pgel, 'add_action', def)) {
                allowed = false;
            }
        })

        if(!allowed) return;

        var opname = "Add action / " + def.name;

        collection.forEachElement(function(pgel) {
            var can_add = true;
            if (def.on_can_add_action) {
                if (!def.on_can_add_action(pgel, pgel.getPage(), def)) {
                    can_add = false;
                }
            }
            if (can_add) {
                $.fn.crsa('addActionToElement', pgel, def, pgel.getPage(), optional_values);
            }
        })

        if(change_info && change_info.no_change_event) return;

        pinegrow.changeManager.didChange(collection, opname, change_info);
    }

    removeAction(collection_or_el, def, change_info) {
        var collection = pinegrow.getCollection(collection_or_el)

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            if(!canMakeChange(pgel, 'add_action', def)) {
                allowed = false;
            }
        })

        if(!allowed) return;

        var opname = "Remove action / " + def.name;

        collection.forEachElement(function(pgel) {
            $.fn.crsa('removeActionFromElement', pgel, def, pgel.getPage());
        })

        if(change_info && change_info.no_change_event) return;

        pinegrow.changeManager.didChange(collection, opname, change_info);
    }

    changeTag(collection_or_el, tag) {
        var collection = pinegrow.getCollection(collection_or_el)

        var t = typeof tag == 'string' ? {tag: tag} : tag;

        t.change_parent = t.change_parent || [];

        //is op allowed
        var allowed = true;
        collection.forEachElement(function(pgel) {
            if (!canMakeChange(pgel, 'delete_element')) {
                allowed = false;
            }
            if(pgel.parent && t.parent_tag && pgel.parent.tagName != t.parent_tag) {
                if (!canMakeChange(pgel.parent, 'insert_element')) {
                    allowed = false;
                }
            }
        })

        if(!allowed) return;

        var change_parents = new PgCollection();

        collection.orderByDocumentPosition();

        collection.forEachElement(function(pgel) {
            pgel = pgel.replaceTag(t.tag);
            if(t.parent_tag && pgel.parent) {
                if(pgel.parent.tagName != t.parent_tag) {
                    //is just after parent tag?
                    var prev = pgel.prev();
                    if(prev && prev.tagName == t.parent_tag) {
                        pgel.appendTo(prev);
                    } else if(t.change_parent.indexOf(pgel.parent.tagName) >= 0) {
                        //change parent
                        change_parents.add(pgel.parent);
                    } else {
                        //insert new parent
                        var newp = pgCreate('<' + t.parent_tag + '></' + t.parent_tag + '>');
                        newp.insertBefore(pgel);
                        pgel.appendTo(newp);
                    }
                }
            }
        })

        change_parents.forEachElement(function(pgel) {
            pgel.replaceTag(t.parent_tag);
        })

        pinegrow.changeManager.didChange(collection, 'Change to ' + t.tag);
    }

    wrapWithElement(collection_or_el, code, check, on_wrapped, skip_select) {

        var collection = pinegrow.getCollection(collection_or_el).clone();

        var inserted = new PgCollection();

        collection.forEachElement(function(pgCurrent, i, total) {

            var problems = new pgParserSourceProblem(pgCurrent);

            if(!pgCurrent) {
                problems.add('element', pgCurrent, 'wrap with element');
            }
            if(!problems.ok()) {
                showAlert(problems.toString(), "Can't create the element");
                return;
            }

            if(!canMakeChange(pgCurrent.parent, 'insert_element')) return;

            var error = check ? check(pgCurrent) : false;

            if(error) {
                pinegrow.showQuickMessage(error, 4000, false, 'error');
                return;
            }

            var pgNew = pgCreateNodeFromHtml(code);

            pgNew.insertBefore(pgCurrent);

            pgNew.append(pgCurrent);

            on_wrapped && on_wrapped(pgNew, pgCurrent);

            if(!skip_select) {
                pinegrow.selectedElements.replace(pgCurrent, pgNew);
            }

            inserted.add(pgNew);

            if(i == total - 1) {
                pinegrow.changeManager.didInsert(inserted);

                pinegrow.selectedElements.reselect();
            }
        })
    }


    wrapContentWithElement(collection_or_el, code, check, on_wrapped) {

        var collection = pinegrow.getCollection(collection_or_el).clone();

        var inserted = new PgCollection();

        collection.forEachElement(function(pgCurrent, i, total) {

            var problems = new pgParserSourceProblem(pgCurrent);

            if(!pgCurrent) {
                problems.add('element', pgCurrent, 'wrap with element');
            }
            if(!problems.ok()) {
                showAlert(problems.toString(), "Can't create the element");
                return;
            }

            if(!canMakeChange(pgCurrent, 'insert_element')) return;

            var error = check ? check(pgCurrent) : false;

            if(error) {
                pinegrow.showQuickMessage(error, 4000, false, 'error');
                return;
            }

            var pgNew = pgCreateNodeFromHtml(code);

            pgCurrent.moveChildrenToElement(pgNew, false, true);

            pgCurrent.append(pgNew);

            on_wrapped && on_wrapped(pgNew, pgCurrent);

            inserted.add(pgNew);

            if(i == total - 1) {
                pinegrow.changeManager.didInsert(inserted);

                pinegrow.selectedElements.reselect();
            }
        })
    }


    ///CSS related easy-to-use APIs
    setInlineStyle(collection_or_el, rule, change_info) {
        var collection = pinegrow.getCollection(collection_or_el)

        if(collection.getLength() === 0) return;

        var html = '';
        rule.nodes.forEach(function(node) {
            if(node.type === 'decl' && node.prop.length && node.value.length) {
                html += `${node.prop}: ${node.value + (node.important ? ' !important' : '')};`
            } else if(node.type === 'comment') {
                html += `/* ${node.text} */`;
            }
        })
        this.setAttribute(collection, 'style', html, change_info)
        var pgel = collection.getFirst()
        if(!change_info || !change_info.skip_change_event) {
            pinegrow.dispatchEvent('on_css_attribute_style_changed', pgel.getPage(), pgel, change_info);
        }
    }

    addCSSRuleToStylesheet(rule_code, cs, position, action_info, done /* (error, nodes, main_nodes) */, main_nodes, before) {
        if(typeof rule_code === 'object') {
            rule_code = cs.getSourceForRules(rule_code);
            rule_code = cs.filterSource(rule_code, true);
        }
        cs.parseSource(rule_code, function (error, postcssSource, newRootNode) {
            if (error) {
                if(done) done(error);
                pinegrow.showAlert('<p>Could not insert the CSS rule.</p><p>' + error + '</p>', 'Error');
            } else {
                var idx = 0;

                if(typeof position == 'object') {
                    idx = -1;
                    var last_node = position;
                    newRootNode.nodes.forEach(function(node) {
                        if(before) {
                            pgcssh.insertNodeBefore(node, last_node);
                        } else {
                            pgcssh.insertNodeAfter(node, last_node);
                            last_node = node;
                        }
                        if(idx < 0) {
                            idx = node.parent.nodes.indexOf(node);
                        }
                    })
                } else {
                    switch (position) {
                        case 'append':
                            idx = cs.getRootNode().nodes.length;
                            break;
                        case 'prepend':
                            idx = 0;
                            break;
                        default:
                            idx = position;
                            break;
                    }

                    pgcssh.insertNodesAt(newRootNode.nodes, cs.getRootNode(), idx);
                }

                cs.regenerateAndSetCssSource(function() {
                    var root = cs.getRootNode();
                    var new_nodes = [];
                    for(var i = 0; i < newRootNode.nodes.length; i++) {
                        new_nodes.push(newRootNode.nodes[i]);
                    }
                    /*
                    for(var i = idx; i < idx + newRootNode.nodes.length; i++) {
                        if(i < root.nodes.length) {
                            new_nodes.push(root.nodes[i]);
                        }
                    }*/
                    var info = typeof action_info == 'object' ? action_info : {action: (action_info || 'Add CSS rule')};
                    info.list = [cs];

                    pinegrow.dispatchEvent('on_css_source_changed', null, info);

                    if(done) done(null, new_nodes, main_nodes);
                })
            }
        }, false);
    }

    overrideCSSVariableOLD(node, cs, done, custom_value) {

        var findVarsBefore = function(cs, node, used_vars, map) {
            var idx = node.parent.nodes.indexOf(node);
            var list = [];
            map = map || {};
            for(var i = idx - 1; i >= 0; i--) {
                if(isVariable(node.parent.nodes[i])) {
                    list.push(node.parent.nodes[i].prop);
                    // map[node.parent.nodes[i].prop] = node.parent.nodes[i].value;
                    map[node.parent.nodes[i].prop] = node.parent.nodes[i];
                    // main_nodes.push(node.parent.nodes[i]);
                }
            }

            var resolveUsed = function(value) {
                var m = value.match(/\$[a-z0-9_\-]+/ig);
                if (m) {
                    m.forEach(function (used_var) {
                        var more = [], used_var_added = false;
                        if (used_vars.indexOf(used_var) < 0 && used_var in map) {
                            more.push(used_var);
                            used_var_added = true;
                        }
                        more.forEach(function (mv) {
                            // resolveUsed(mv, map[mv]);
                            resolveUsed(map[mv].value);
                        });
                        if (used_var_added) {
                            used_vars.push(used_var);
                        }
                    })
                }
            }

            resolveUsed(custom_value || node.value);

            return list;
        }

        var autoDetectPositionInCs = function(origcs, cs, node, info) {
            var used_vars = [];
            var map = {};
            info = info || {};
            var list = findVarsBefore(origcs, node, used_vars, map);
            var root = cs.getRootNode();
            var last_var_before_import = null;
            var position = null;

            for(var i = root.nodes.length - 1; i >= 0; i--) {
                if(isVariable(root.nodes[i])) {
                    if(!position) {
                        last_var_before_import = root.nodes[i];
                        if (list.indexOf(root.nodes[i].prop) >= 0) {
                            position = root.nodes[i];
                        }
                    }
                    var idx = used_vars.indexOf(root.nodes[i].value);
                    if(idx >= 0) {
                        used_vars.splice(idx, 1);
                    }
                } else if(root.nodes[i].type == 'atrule' && root.nodes[i].name == 'import') {
                    last_var_before_import = null;
                }
            }

            info.used_vars = used_vars;
            info.used_vars_code = '';

            for(var i = used_vars.length - 1; i >= 0; i--) {
                main_nodes.unshift(map[used_vars[i]]);
                info.used_vars_code = `${used_vars[i]}: ${map[used_vars[i]].value.replace(/ !global/g, '')};` + info.used_vars_code;
            }

            if(position) return position;

            if(!last_var_before_import) {
                for(var i = 0; i < root.nodes.length; i++) {
                    if(root.nodes[i].type != 'comment') {
                        break;
                    }
                    last_var_before_import = root.nodes[i];
                }
            }
            return last_var_before_import || 'prepend';
        }

        var isVariable = pgcssh.isVariable;
        var sourcecs = pgcssh.getStylesheetForNode(node);

        var position = null;

        var main_nodes = [];

        var act;

        if(sourcecs == cs) {
            position = 'append';
            act = 'Duplicate';
        } else {
            act = 'Override';
            var info = {};
            position = autoDetectPositionInCs(sourcecs, cs, node, info);
        }

        var rule_code = sourcecs.getSourceForRules(node);
        rule_code = sourcecs.filterSource(rule_code, true);

        if(custom_value) {
            rule_code = `${node.prop}: ${custom_value};`;
        }

        if(info && info.used_vars_code) {
            rule_code = info.used_vars_code + rule_code;
        }

        if(sourcecs !== cs && isVariable(node)) {
            rule_code = rule_code.replace(/ !default/g, '');
        }

        main_nodes.push(node);
        this.addCSSRuleToStylesheet(rule_code, cs, position, {
            action: act + ' ' + pgcssh.getNodeDisplayName(node)
        }, done, main_nodes);
    }

    async overrideCSSVariable(node, cs, done, custom_value) {

        var findVarsBefore = function(cs, node, used_vars, map) {
            var idx = node.parent.nodes.indexOf(node);
            var list = [];
            map = map || {};
            for(var i = idx - 1; i >= 0; i--) {
                if(isVariable(node.parent.nodes[i])) {
                    list.push(node.parent.nodes[i].prop);
                    map[node.parent.nodes[i].prop] = node.parent.nodes[i];
                }
            }

            var resolveUsed = function(value) {
                var m = value.match(varReg);
                if (m) {
                    m.forEach(function (used_var) {
                        var more = [], used_var_added = false;
                        if (used_vars.indexOf(used_var) < 0 && used_var in map) {
                            more.push(used_var);
                            used_var_added = true;
                        }
                        more.forEach(function (mv) {
                            resolveUsed(map[mv].value);
                        });
                        if (used_var_added) {
                            used_vars.push(used_var);
                        }
                    })
                }
            }

            resolveUsed(custom_value || node.value);

            return list;
        }

        var findMainNodeFor = function (node) {
            for (var i = 0; i < main_nodes.length; i++) {
                if (main_nodes[i].prop == node.prop) return main_nodes[i];
            }
        }

        var getCustomFunNode = function (main_node, mainNodeIndex, level) {
            var rootNode = cs.getRootNode();
            var node = rootNode.last;

            if (!node) {
                return {
                    node: rootNode,
                    fun: pcssh.insertNodeInside
                };
            }

            var fun = null;
            var lastImpNode = null;

            while (node) {
                if (node.type == 'decl') {
                    var mainNode = pcssh.findNodeInside('decl', 'prop', node.prop, main_node.parent);
                    if (mainNode) {
                        if (main_node.prop == node.prop) return null;

                        var mainNodeIndices = pcssh.getNodeIndices(mainNode);

                        var nodeIndex = mainNodeIndices[level];

                        if (nodeIndex - mainNodeIndex < 0) {
                            return {
                                node: node,
                                fun: pcssh.insertNodeAfter
                            };
                        }

                        lastImpNode = node;
                        node = node.prev();
                    }
                    else {
                        return {
                            node: node,
                            fun: pcssh.insertNodeAfter
                        };
                    }
                }
                else {
                    if (pcssh.getNodeType(node) == 'import') {
                        lastImpNode = node;
                    }
                    node = node.prev();
                }
            }

            if (lastImpNode) {
                return {
                    node: lastImpNode,
                    fun: pcssh.insertNodeBefore
                };
            }
            else {
                return {
                    fun: pcssh.insertNodeAt,
                    node: rootNode,
                    index: 0
                }
            }
        }

        var filterNodeValue = function (value) {
            return value.replace(/ !global/g, '').replace(/ !default/g, '');
        }

        var autoDetectPositionInCs = function(origcs, cs, node) {
            var used_vars = [];
            var map = {};
            var list = findVarsBefore(origcs, node, used_vars, map);


            for(var i = used_vars.length - 1; i >= 0; i--) {
                var currNode = map[used_vars[i]];
                main_nodes.unshift(currNode);
                custom_nodes.unshift(pcssh.createNode('decl', used_vars[i], filterNodeValue(currNode.value), null, cs));
            }
        }

        var updateCustomValue = async function (cs, nodes, callback) {
            var isFuntion = function (value) {
                if (value.indexOf('(') > -1) {
                    var { name } = getCSSFunctionNameAndParams(value);
                    if (!cs.isBuitInFunction(name)) return true;
                }

                return false;
            }

            var askForValue = async function (node) {
                return new Promise(function (resolve, reject) {
                    var $alert = pinegrow.showAlert(
                        "<p><b>" + node.prop + "</b> is using a custom function. Using this function in the value of the custom variable will most likely cause compilation errors.</p><p>Please input the custom value here:</p><p><div class='node-value'></div></p>",
                        "Set the variable value",
                        "Cancel", "Ok",
                        function() {
                            reject('delete');
                        }, function() {
                            resolve(inputVal.getValue());
                        }
                    );

                    var $valueCont = $alert.find('.node-value');
                    var inputVal = new PgInputField('node-value', {
                        type: 'text',
                        name: 'Value',
                        without_text: false
                    });
                    inputVal.setValue(node.value);
                    inputVal.appendTo($valueCont);
                });
            }

            for (var i = 0; i < nodes.length; i++) {
                var node = nodes[i];
                var value = node.value;
                if (isFuntion(value)) {
                    try {
                        value = await askForValue(node);
                    } catch (e) {
                        if (e == 'delete') {
                            return false;
                        }
                    }

                    node.value = value;
                }
                callback(node);
            }

            return true;
        }

        var varReg = new RegExp('\\' + cs.getDeclCharachter() + '[a-z0-9_\\-]*', 'ig');
        var pcssh = pinegrow.pgPostCSSHelper;

        var isVariable = pgcssh.isVariable;
        var sourcecs = pgcssh.getStylesheetForNode(node);

        var main_nodes = [];
        var custom_nodes = [];

        var custom_nodes_indices = [];

        var node_value = null;

        if (sourcecs == cs) {
            node_value = node.value;
        }
        if (sourcecs != cs) {
            node_value = filterNodeValue(node.value);
            autoDetectPositionInCs(sourcecs, cs, node);
        }


        var main_custom_node = pcssh.createNode('decl', node.prop, node_value, null, cs);

        if (custom_value) {
            main_custom_node.value = custom_value;
        }

        custom_nodes.push(main_custom_node);
        main_nodes.push(node);

        var filter_custom_nodes = [];
        var add_custom_nodes = true;
        if (sourcecs == cs) {
            add_custom_nodes = await updateCustomValue(cs, custom_nodes, function (custom_node, done) {
                var main_node = findMainNodeFor(custom_node);
                pcssh.insertNodeInside(custom_node, main_node.parent);

                filter_custom_nodes.push(custom_node);
                if (done) done();
            });
        }
        else {
            add_custom_nodes = await updateCustomValue(cs, custom_nodes, function (custom_node, done) {
                var main_node = findMainNodeFor(custom_node);

                var nodeLevel = pcssh.getNodeLevel(main_node);

                var mainNodeIndices = pcssh.getNodeIndices(main_node);
                var mainNodeIndex = mainNodeIndices[nodeLevel];

                var funNodeObj = getCustomFunNode(main_node, mainNodeIndex, nodeLevel);
                if (funNodeObj != null) {
                    filter_custom_nodes.push(custom_node);
                    funNodeObj.fun(custom_node, funNodeObj.node, funNodeObj.index);
                }
                if (done) done();
            });
        }

        if (!add_custom_nodes) return;

        for (var i = 0; i < filter_custom_nodes.length; i++) {
            custom_nodes_indices.push(pcssh.getNodeIndices(filter_custom_nodes[i]));
        }


        cs.setSourceForRules(cs.rootNode, function () {
            var new_custom_nodes = [];
            for (var i = 0; i < custom_nodes_indices.length; i++) {
                var custom_node = pcssh.getNodeByIndices(cs.rootNode, custom_nodes_indices[i]);
                new_custom_nodes.push(custom_node);
                pinegrow.dispatchEvent('on_css_node_added', null, {node: custom_node, stylesheet: cs});
            }

            if (done) done(new_custom_nodes, main_nodes);
        });
    }

    editCSSRule(rule) {
        pinegrow.selectRule(rule);
    }

    extractCSSForElementAndSelector(pgel, sel, done) {
        var page = pgel.getPage();

        var code = '';
        var re = new RegExp(`^${escapeRegExp(sel)}($|[\\s>])`, 'gm');

        var current_media = null;
        var indent = '';

        var atrules = [];
        var current_atrules_level = 0;
        var current_atrules_str = '';
        var current_atrules = [];

        var getAtrulesCode = function(start_level) {
            var c = '';
            for(var i = start_level; i < atrules.length; i++) {
                c += atrules[i] + ' {\n';
            }
            return c;
        }

        var checkNode = function(cs, node, sel) {
            if(node.type === 'rule') {
                if(node.selector === sel || node.selector.match(re)) {

                    var start_at = 0;

                    for(var start_at = 0; start_at < atrules.length; start_at++) {
                        if(current_atrules[start_at] === atrules[start_at]) {

                        } else {
                            break;
                        }
                    }

                    for(var i = start_at; i < current_atrules.length; i++) {
                        code += '}\n';
                    }
                    current_atrules = atrules.slice(0);

                    var atrule_code = getAtrulesCode(start_at);

                    //var opt_atrule_code = atrule_code.replace(current_atrules_str, '');

                    current_atrules_str = atrule_code;

                    var c = cs.getSourceForRules(node);

                    code += atrule_code + '\n' + c;

                    return;

                    /*
                     var media = null;
                     if(node.parent.type == 'atrule') {
                     media = `@${node.parent.name} ${node.parent.params}`;
                     }
                     if(current_media !== media) {
                     if(current_media) {
                     code += '\n}\n';
                     }
                     if(media) {
                     code += '\n' + media + ' {\n';
                     indent = '    ';
                     } else {
                     indent = '';
                     }
                     current_media = media;
                     }

                     code += '\n' + indent + c;
                     */
                }
            }
            if(node.nodes) {
                if(node.type === 'atrule') {
                    atrules.push(`@${node.name} ${node.params}`);
                }
                node.nodes.forEach(function(n) {
                    checkNode(cs, n, sel);
                })
                if(node.type === 'atrule') {
                    atrules.pop();
                }
            }
        }

        var doCS = function(cs, sel) {
            if(cs.rootNode) {
                checkNode(cs, cs.rootNode, sel);
            }
            if(cs.importedFiles) {
                cs.importedFiles.forEach(doCS);
            }
        }

        page.stylesheets.stylesheets.forEach(function(cs) {
            if(cs.sourceStylesheet) cs = cs.sourceStylesheet;
            if(cs.loaded) {
                console.log(cs.rootNode);
                doCS(cs, sel);
            }
        })

        for(var i = 0; i < current_atrules.length; i++) {
            code += '}\n';
        }

        //Lets format the code
        var allPostcssLib = require('postcss');
        var postcss = allPostcssLib.postcss;

        var syntaxRes = postcss().process(code, { syntax: allPostcssLib.postcssSCSS, parser: allPostcssLib.postcssSCSS, stringifier: allPostcssLib.postcssSCSS.stringify });
        syntaxRes.then(function (postcssRes) {

            var rootNode = postcssRes.root;

            pgcssf.updateNodeRaws(rootNode, false, {}, true, null);

            var source = pinegrow.pgPostCSSHelper.getSource(rootNode, allPostcssLib.postcssSCSS.stringify);

            done(source);
        });
    }

    extractComponent(pgel, done) {
        var id = pgel.getAttr('data-pgc-define');
        var sel = pgel.getAttr('data-pgc-define-extract-css');

        var page = pgel.getPage();

        var d = {
            html: pgel.toStringOriginal(true)
        }

        var script = page.sourceNode.findOne(`script[data-pgc-define-script="${id}"]`);

        if(script) {
            d.js = script.html();
        }

        var onDone = function() {
            done && done(d);
        }

        if(sel) {
            this.extractCSSForElementAndSelector(pgel, sel, function(css) {
                d.css = css;
                onDone();
            })
        } else {
            onDone();
        }
    }

    transformElement(pgel, to_pgel) {
        var col = pinegrow.getCollectionForElement(pgel);

        pinegrow.makeChanges(col, "Transform element", function() {

            col.forEachElement(function(pgel) {
                if(pgel) {

                    var content_elements = ['p','span','h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'img', 'svg', 'button', 'a'];
/*
                    var extras = {
                        img: {
                            getValue: function(pgel) {
                                return {
                                    src: pgel.getAttr('src'),
                                    srcset: pgel.getAttr('srcset'),
                                }
                            },
                            setValue: function(pgel, val) {
                                val.src && pgel.setAttr('src', val.src);
                                val.srcset ? pgel.setAttr('srcset', val.srcset) : pgel.removeAttribute('srcset');
                            }
                        },
                        svg: {
                            getValue: function(pgel) {
                                return pgel;
                            },
                            setValue: function(pgel, val) {
                                var attrs = ['id', 'class', 'style'];
                                val = val.clone();
                                attrs.forEach(function(attr) {
                                    if(pgel.hasAttribute(attr)) {
                                        val.setAttribute(attr, pgel.getAttribute(attr));
                                    }
                                })
                                pgel.replaceWith(val);
                            }
                        },
                        a: {
                            getValue: function(pgel) {
                                return {
                                    href: pgel.getAttr('href'),
                                    target: pgel.getAttr('target'),
                                    html: pgel.html()
                                }
                            },
                            setValue: function(pgel, val) {
                                val.href && pgel.setAttr('href', val.href);
                                val.target ? pgel.setAttr('target', val.target) : pgel.removeAttribute('target');
                                pgel.html(val.html);
                            }
                        },
                    }

                    var values = {};

                    pgel.walkSelfAndChildren(function(node) {
                        if(node.isElement && content_elements.indexOf(node.tagName) >= 0) {
                            if(extras[node.tagName]) {
                                var v = extras[node.tagName].getValue(node);
                            } else {
                                var v = node.html();
                            }

                            if(!values[node.tagName]) {
                                values[node.tagName] = [];
                            }
                            values[node.tagName].push(v);
                            return 'skip_children';
                        }
                        return true;
                    })

                    //Any unused values from before
                    var unused = pgel.getData('transform_values');
                    if(unused) {
                        $.each(unused, function(key, val) {
                            if(!values[key]) {
                                values[key] = val;
                            } else {
                                values[key] = values[key].concat(val);
                            }
                        })
                    }
*/

                    var new_pgel = to_pgel.clone();

                    if(new_pgel.tagName === 'svg') {
                        var attrs = ['id', 'class', 'style', 'width', 'height'];

                        attrs.forEach(function(attr) {
                            if(pgel.hasAttribute(attr)) {
                                new_pgel.setAttribute(attr, pgel.getAttribute(attr));
                            }
                        })
/*
                        pgel.getAttrList().forEach(function(ad) {
                            if(ad.name.startsWith('data-')) {
                                new_pgel.setAttr(ad.name, ad.value);
                            }
                        })*/
                    }

/*
                    new_pgel.walkSelfAndChildren(function(node) {
                        if(node.isElement && content_elements.indexOf(node.tagName) >= 0 && values[node.tagName] && values[node.tagName].length) {
                            var v = values[node.tagName].shift();

                            if(extras[node.tagName]) {
                                extras[node.tagName].setValue(node, v);
                            } else {
                                node.html(v);
                            }
                            return 'skip_children';
                        }
                        return true;
                    })
                    */

                    pgel.replaceWith(new_pgel);

                    pinegrow.selectedElements.replace(pgel, new_pgel);

                    //new_pgel.setData('transform_values', values);
                }
            })
        })

        col.getLength() > 0 && pinegrow.changeManager.didChange(col, 'Replace');
        //pinegrow.selectedElements.selectMany(select, true)
    }

    moveUpOrDown(collection_or_el, up) {
        var collection = pinegrow.getCollection(collection_or_el);

        var changed = new PgCollection();

        var name = up ? 'Move up' : 'Move down';

        collection.normalize();

        if(!up) {
            //reverse for down
            collection.reverse();
        }

        collection.forEachElement(function(pgCurrent, i, total) {

            if(!pgCurrent.parent || !pgCurrent.parent.parent) return;

            if(!canMakeChange(pgCurrent.parent, 'edit_content')) return;

            if(up) {
                var prev = pgCurrent.prev();

                if(!prev) {
                    pinegrow.showQuickError('The element is already at the top.');
                } else {
                    if(collection.contains(prev)) {
                        //prev was also moved
                        pgCurrent.insertAfter(prev);
                    } else {
                        pgCurrent.insertBefore(prev);
                    }
                    changed.add(pgCurrent.parent);
                }
            } else {
                var next = pgCurrent.next();
                if(!next) {
                    pinegrow.showQuickError('The element is already at the bottom.');
                } else {
                    if(collection.contains(next)) {
                        //next was also moved
                        pgCurrent.insertBefore(next);
                    } else {
                        pgCurrent.insertAfter(next);
                    }
                    changed.add(pgCurrent.parent);
                }
            }

            if(i === total - 1) {
                pgCurrent.scrollTo();
            }
        })

        changed.getLength() > 0 && pinegrow.changeManager.didChange(changed, name);
    }

    moveUp(collection_or_el) {
        return this.moveUpOrDown(collection_or_el, true)
    }

    moveDown(collection_or_el) {
        return this.moveUpOrDown(collection_or_el, false)
    }

    /*
    this.addCSSCodeToProject = function(code, key, relative_url) {

        var project = pinegrow.getCurrentProject();
        var url = project.getAbsoluteUrl(relative_url);
        var cs = pinegrow.stylesheets.getByUrl(url);

        if(!cs) {
            var file = crsaMakeFileFromUrl(url);
            if(!crsaIsFileExist(file)) {

            }
        } else {

        }

    }*/
}