class PgCSSApi {
    constructor() {
        this.api = new PgApi();
    }

    getStylesheetFromNode(node) {
        return pgcssh.getStylesheetForNode(node);
    }

    isNodeImport(node) {
        return (node && (node.type === "atrule" || node.type === "import") && node.name === "import");
    }

    updateStylesheetCodeFromRules(cs, singleChangedRule, done) {
        if (!cs) return;
        if(singleChangedRule) {
            if(cs.updateSingleRule) {
                cs.updateSingleRule(singleChangedRule, function() {
                    if (done) done(cs.getRootNode());
                })
                return;
            }
        }
        cs.setSourceForRules(cs.rootNode, function () {
            if (done) done(cs.getRootNode());
        }, false, true /* fastDelayed */);
    }

    commentNode(node, done) {
        const rule = pgcssh.getRuleForNode(node);
        let cs = null;
        if(rule.inline) {

        } else {
            cs = this.getStylesheetFromNode(node);
        }
        var source = cs ? cs.filterSource(pinegrow.pgPostCSSHelper.getSource(node)) : pinegrow.pgPostCSSHelper.getSource(node);
        var commentSource = source.replace(/\/\*/, '||*|').replace(/\*\//, '|*||');
        var commentNode = pinegrow.pgPostCSSHelper.createNode('comment', commentSource, null, node, cs);

        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
        cs && pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: cs});

        pinegrow.pgPostCSSHelper.replaceNodeWith(node, commentNode);
        commentNode['pg-node-id'] = node['pg-node-id'];

        if (cs && this.isNodeImport(node)) {
            cs.ignoreImportedFile(node, true);
        }

        if(rule.inline) {
            this.api.setInlineStyle(pinegrow.getCollectionForElement(rule.inline_for_pgel), rule, {
                event_name: 'Comment style'
            })
            done && done();
        } else {
            this.updateStylesheetCodeFromRules(cs, null, function () {
                delete commentNode['pg-node-id']; //not sure why
                var addedNode = pinegrow.pgPostCSSHelper.getNodeByIndices(cs.getRootNode(), nodeIndices);
                pinegrow.dispatchEvent('on_css_node_added', null, {node: addedNode, stylesheet: cs});
                done && done();
            })
        }
    }

    uncommentNode(node, done) {
        const _this = this;
        const rule = pgcssh.getRuleForNode(node);
        let cs = null;
        if(!rule.inline) {
            cs = this.getStylesheetFromNode(node);
        }
        var source = cs ? cs.filterSource(pinegrow.pgPostCSSHelper.getSource(node)) : pinegrow.pgPostCSSHelper.getSource(node);
        source = source.trim();
        if (source.startsWith('/*') && source.endsWith('*/')) {
            source = source.substr(2, source.length - 4).trim();
        }
        var commentSource = source.replace(/\|\|\*\|/, '/*').replace(/\|\*\|\|/, '*/');
        pinegrow.pgPostCSSHelper.parsedCSSSourceFromText(commentSource, function (err, css, root) {
            if (err || root.nodes.length === 0) {
                pinegrow.showQuickError("The content is not a valid CSS expression.");
                if (done) done(false);
            }
            else {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
                var callback = function (done) {
                    if(rule.inline) {
                        _this.api.setInlineStyle(pinegrow.getCollectionForElement(rule.inline_for_pgel), rule, {
                            event_name: 'Uncomment style'
                        })
                        done && done(true);
                    } else {
                        _this.updateStylesheetCodeFromRules(cs, null, function (resRootNode) {
                            var addedNode = pinegrow.pgPostCSSHelper.getNodeByIndices(cs.getRootNode(), nodeIndices);
                            //setUncollapsedNodes(addedNode, true);

                            pinegrow.dispatchEvent('on_css_node_added', null, {
                                node: addedNode,
                                stylesheet: cs,
                                operation: 'ignore'
                            });
                            if (done) done(true);
                        });
                    }
                }

                var newNode = root.nodes[0];
                if (cs && cs.fileType === "css") {
                    if (_this.isNodeImport(newNode) && cs.isImported()) {
                        pinegrow.showQuickError('Imports are not allowed here.');
                        if (done) done(false);
                        return;
                    }
                }

                node['ignore'] = false;

                pinegrow.pgPostCSSHelper.copyNodeRawsTo(newNode, node);
                cs && pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: cs, operation: 'ignore'});

                pinegrow.pgPostCSSHelper.replaceNodeWith(node, newNode);
                callback(done);
            }
        })
    }

    deleteNode(node, done) {
        const rule = node.type !== 'rule' && node.parent && node.parent.type === 'rule' ? node.parent : null;//pgcssh.getRuleForNode(node);
        let cs = this.getStylesheetFromNode(node);

        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
        cs && pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: cs});

        pinegrow.pgPostCSSHelper.removeNodeFrom(node);

        if(!cs && rule && rule.inline) {
            this.api.setInlineStyle(pinegrow.getCollectionForElement(rule.inline_for_pgel), rule, {
                event_name: 'Delete style ' + node.prop
            })
            done && done();
        } else {
            this.updateStylesheetCodeFromRules(cs, null, function () {
                pinegrow.dispatchEvent('on_css_node_deleted', null, {node: node, stylesheet: cs});
                done && done();
            })
        }
    }

    deleteNodeAndKeepChildren(node, done) {
        if(!node.nodes || node.nodes.length === 0) {
            this.deleteNode(node, done)
        } else {
            let cs = this.getStylesheetFromNode(node);
            var after = node;
            while(node.nodes.length) {
                var n = node.nodes[0];
                pinegrow.pgPostCSSHelper.insertNodeAfter(n, after);
                pinegrow.pgPostCSSHelper.removeNodeFrom(n, node);
                after = n;
            }

            this.updateStylesheetCodeFromRules(cs, null, function () {
                var info = {}
                info.list = [cs];
                info.action = 'Remove CSS node';
                pinegrow.dispatchEvent('on_css_source_changed', null, info);

                done && done();
            })
        }
    }

    changeMediaQueryForRule(rule, params, done, info) {
        info = info || {}
        var mediaRule = pinegrow.pgPostCSSHelper.getMediaNode(rule);
        if(!mediaRule) {
            const cs = this.getStylesheetFromNode(rule);
            cs.ruleMediaChanged(rule, params, done, info.edited_by || null)
        } else {
            this.changeMediaQuery(mediaRule, params, done, info)
        }
    }

    changeMediaQuery(node, params, done, info) {
        info = $.extend({}, {
            dispatchEvent: true,
            undoCombineKey: 'atRuleChanged',
            undoRuleName: typeof dispatchEvent == 'string' ? dispatchEvent : null,
        }, info || {})

        const cs = this.getStylesheetFromNode(node);
        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);

        node.params = params;
        this.updateStylesheetCodeFromRules(cs, null, function () {
            info.dispatchEvent && pinegrow.dispatchEvent('on_css_node_updated', null, {node: node, stylesheet: cs, ...info});
            var new_node = pinegrow.pgPostCSSHelper.getNodeByIndices(cs.getRootNode(), nodeIndices);
            done && done(new_node);
        })
    }

    //returns the new or updated prop decl node
    addRuleProperty(rule, prop, value, done, info) {
        info = $.extend({}, {
            dispatchEvent: true,
            source: 'api',
            insertAtIndex: -1
        }, info || {})
        const _this = this;
        if(rule.inline) {
            //add prop to rule object
            var important = false;
            if (value) {
                if (value.indexOf('!important') > 0) {
                    value = value.replace('!important', '');
                    important = true;
                }
                value = value.trim();
            }
            var declNode = pinegrow.pgPostCSSHelper.createNode("decl", prop, value, rule, null);
            declNode.important = important;
            var lastDeclIndex = info.insertAtIndex < 0 ? pinegrow.pgPostCSSHelper.getLastDeclIndex(rule) + 1 : info.insertAtIndex;
            pinegrow.pgPostCSSHelper.insertNodeAt(declNode, rule, lastDeclIndex);

            info.event_name = 'Set ' + prop
            if(!info.dispatchEvent) {
                info.skip_change_event = true;
            }
            this.api.setInlineStyle(pinegrow.getCollectionForElement(rule.inline_for_pgel), rule, info)

            return declNode;

        } else {
            const cs = this.getStylesheetFromNode(rule);
            return cs.ruleValueChanged(rule, prop, value, done, false, info.source, info.dispatchEvent, null, true, null, false, false, info.insertAtIndex);
        }
    }

    setRuleProperty(prop_node, prop, value, done, info) {
        info = $.extend({}, {
            dispatchEvent: true
        }, info || {})
        const _this = this;
        const rule = pgcssh.getRuleForNode(prop_node);

        if (prop_node.prop !== prop) {
            pgcssh.setNodeProperty(prop_node, 'prop', prop);
        }

        if(rule.inline) {
            var important = false;
            if (value) {
                if (value.indexOf('!important') > 0) {
                    value = value.replace('!important', '');
                    important = true;
                }
                value = value.trim();
            }
            prop_node.important = important;
            prop_node.value = value;

            info.event_name = 'Set ' + prop;
            if(!info.dispatchEvent) {
                info.skip_change_event = true;
            }
            this.api.setInlineStyle(pinegrow.getCollectionForElement(rule.inline_for_pgel), rule, info)

            return prop_node;

        } else {
            const cs = this.getStylesheetFromNode(prop_node);
            return cs.ruleValueChanged(rule, prop_node, value, done, false, info.source, info.dispatchEvent, null, true);
        }
    }

    addOrSetRuleProperty(rule, prop, value, done, info) {
        var props = pinegrow.pgPostCSSHelper.getChildrenDeclWithPropName(rule, prop);
        if(props.length) {
            this.setRuleProperty(props[props.length - 1], prop, value, done, info)
            return props[props.length - 1];
        } else {
            return this.addRuleProperty(rule, prop, value, done, info)
        }
    }

    getInsertCSSPropMenu(on_selected, stylePreview) {
        var getSubmenu = function(list, add_list) {
            var show_desc = pinegrow.getSetting('css-menu-show-desc', '1') === '1';

            return function() {
                var r = add_list ? add_list.slice(0) : [];
                list.forEach(function(item) {
                    if(item.name === '-') {
                        var d = {
                            type: 'divider'
                        }
                    } else if(item.name === 'pgmore') {
                        var d = {
                            label: 'More...',
                            action: function() {}
                        }
                        if (item.sub && item.sub.length) {
                            d.submenu = getSubmenu(item.sub);
                        }
                    } else {
                        var d = {
                            class: 'pg-css-menu-prop',
                            label: '<i class="icon icon-minus"></i>' + '<code>' + item.name + '</code>',
                            action: function () {
                                on_selected && on_selected(item.name, '');
                            }
                        }
                        var info = pinegrow.devTools.getCSSPropertyInfo(item.name);
                        var add = []

                        if(info) {
                            if(show_desc) {
                                if (info.desc) {
                                    add.push({
                                        type: 'html',
                                        html: info.desc_html,
                                        on_click: function ($li, event) {

                                        }
                                    })
                                    add.push({
                                        type: 'divider'
                                    })
                                }
                            }
                            if(info.examples && info.examples.length) {
                                add.push({
                                    label: 'Examples',
                                    submenu: info.examples.map(function(e) {
                                        return {
                                            label: '<code>' + e + '</code>',
                                            action: function() {
                                                var idx = e.indexOf(':');
                                                on_selected && on_selected(e.substr(0, idx).trim(), e.substr(idx + 1).trim().replace(';', ''))
                                            },
                                            on_mouseenter: function() {
                                                var idx = e.indexOf(':');
                                                stylePreview && stylePreview.setStyle(e.substr(0, idx).trim(), e.substr(idx + 1).trim().replace(';', ''));
                                            },
                                            on_mouseleave: function() {
                                                stylePreview && stylePreview.restore();
                                            },
                                            check: false
                                        }
                                    })
                                })
                            }
                            if(info.values) {
                                var buttons = []

                                /*
                                add.push({
                                    type: 'header',
                                    label: 'Values'
                                })*/
                                add.push({
                                    buttons: buttons
                                })
                                info.values.forEach(function(val) {
                                    buttons.push({
                                        label: val,
                                        class: 'pg-css-menu-value',
                                        action: function() {
                                            on_selected && on_selected(item.name, val);
                                        },
                                        on_mouseenter: function() {
                                            stylePreview && stylePreview.setStyle(item.name, val);
                                        },
                                        on_mouseleave: function() {
                                            stylePreview && stylePreview.restore();
                                        },
                                        check: false
                                    })
                                })
                            }
                            if(add.length) {
                                add.push({
                                    type: 'divider'
                                })
                            }
                        }
                        if((item.sub && item.sub.length) || add.length) {
                            //console.log(item.name, item.sub, add)
                            d.submenu = getSubmenu(item.sub || [], add);
                        }
                    }
                    r.push(d)
                })
                return r;
            }
        }

        var a = [];

        var si = new PgSmartInputForCSS(stylePreview)

        si.onPropSelected = function(prop, value) {
            stylePreview.restore();
            return on_selected(prop, value);
            //si.destroy();
            //si = null;
        }

        a.push({
            type: 'smartinput',
            input: si,
            on_menu_shown: function() {
                si.focus()
            }
        })

        a.push({
            type: 'divider'
        })

        var info = pinegrow.devTools.getCSSPropsInfo();
        $.each(info, function(section, def) {

            var d = {
                label: `<i class="icon ${def.icon}" style="margin-right:8px;"></i>${def.name}`,
                submenu: getSubmenu(def.sub),
                class: 'pg-css-menu-cat'
            }
            a.push(d);
        })

        a.push({
            type: 'divider'
        })
        a.push({
            label: 'Show descriptions',
            check: function() {
                return pinegrow.getSetting('css-menu-show-desc', '1') === '1';
            },
            action: function() {
                var val = !(pinegrow.getSetting('css-menu-show-desc', '1') === '1');
                pinegrow.setSetting('css-menu-show-desc', val ? '1' : '0')
            }
        })
        return a;
    }

    walkRules(cs, func, node, level) {
        const _this = this;
        node = node || cs.getRootNode();
        level = level || 0;
        if(node) {
            if(node.type === 'rule') {
                if(func(node, level) === false) return false;
                level++;
            }
            if(node.nodes) {
                for(var i = 0; i < node.nodes.length; i++) {
                    if(_this.walkRules(cs, func, node.nodes[i], level) === false) {
                        return false;
                    }
                }
            }
        }
    }

    getMediaMatchRe(size) {
        return new RegExp(`\\(\\s*min\\-width:\\s*${pgParserEscapeRegExp(size + 'px')}\\s*\\)`);
    }
    matchesMedia(media, size_re) {
        return !!media.params.match(size_re);
    }

    getRuleForMediaSize(cs, selector, size, done, create) {
        const _this = this;
        var found = null;
        var size_re = size ? this.getMediaMatchRe(size) : null;
        var media_to_rules = {}
        var first = null;
        this.walkRules(cs, function(rule, level) {
            if(level === 0 && rule.selector === selector) {
                var media = pgcssh.getMediaNode(rule);
                var rule_size = media ? _this.getMediaSizeForRule(rule) : 0;

                media_to_rules[rule_size] = media || rule;
                if(!first) first = media || rule;

                if(!found) {
                    if (!media && !size) {
                        found = rule;
                        return false;
                    } else if (media && _this.matchesMedia(media, size)) {
                        found = rule;
                        return false;
                    }
                }
            }
        })
        if(!found && create) {
            var rule_code = `
${selector} {
}`;
            if(size) {
                rule_code = `
@media (min-width: ${size}px) { ${rule_code} 
}`;
            }
            //where to insert
            var pos_after = null;
            $.each(media_to_rules, function(rule_size, rule_or_media) {
                if(rule_size < size) {
                    pos_after = rule_or_media;
                }
            })

            var pos;
            var before = false;
            if(pos_after === null && first) {
                pos = first;
                before = true;
            } else {
                pos = pos_after;
            }

            this.api.addCSSRuleToStylesheet(rule_code, cs, pos, {}, function(error, nodes, main_nodes) {
                done && done(size ? nodes[0].nodes[0] : nodes[0]);
            }, null, before)
        } else {
            done && done(found);
        }
    }

    getMediaSizeForRule(rule) {
        var media = pgcssh.getMediaNode(rule);
        if(!media) return 0;

        var m = media.params.match(/min\-width:\s*([0-9\.]+)px\s*\)/)
        if(m) {
            return parseInt(m[1]);
        }
        return 0;
    }

    showInsertCSSPropMenuForRule(rule, $target, pgel) {
        const _this = this;

        var name = pgcssh.getRuleDisplayName(rule);

        var actions = [
            {
                type: 'html',
                html: `<p style="margin-bottom:0;padding: 5px 10px 0px;"><span style="color:rgba(255,255,255,0.70);">Set property for</span> <span style="color:rgba(255,255,255,1)">${name}</span></p>`
            },
            {
                type: 'divider'
            }
        ]

        var stylePreview = new PgFastDOMStylePreviewForPgel(pgel);

        const m = new CrsaContextMenu(actions.concat(this.getInsertCSSPropMenu(function(prop, value) {
            if(prop && value !== '') {
                stylePreview.restore();
                _this.addOrSetRuleProperty(rule, prop, value, function() {

                })
            }
            if(value === null || value === '') {
                //if no value, we close the menu and let the user enter the value
                pinegrow.showQuickError('Please enter the property value.')
                //m.close();
                return false;
            }
        }, stylePreview)), $target);

        $target ? m.showForElement($target, null, 4) : m.showAtPgel(pgel);

        m.onClosed = function() {
            stylePreview.destroy();
        }

        m.keepOpenOnClick = true;

        return m;
    }

    highlightNodeInCode(node, force) {
        var done = false;

        if(pgPageCodeEditorsStack.highlight_css_rule && pgPageCodeEditorsStack.isEditorShown()) {
            var cs = pgStylesheetHelper.getStylesheetFromNode(node);
            if(cs && cs.url) {
                var editor = pinegrow.codeEditors.isUrlOpen(cs.url);
                if(editor) {
                    if(!pgPageCodeEditorsStack.isEditorShown(editor.view)) {
                        pgPageCodeEditorsStack.showEditor(editor.view);
                    }
                    var pos = pgcssh.getNodePositionInSource(node);
                    editor.highlightRange(pos);
                    done = true;
                }
            }
        }
        if(!done && force) {
            var cs = pgStylesheetHelper.getStylesheetFromNode(node);
            if(cs && cs.url) {
                pinegrow.stylesheets.editCode(cs);
                var editor = pinegrow.codeEditors.isUrlOpen(cs.url);
                setTimeout(function() {
                    var pos = pgcssh.getNodePositionInSource(node);
                    editor.highlightRange(pos);
                }, 500)
            } else {
                pinegrow.showQuickError('Sorry, unable to show the source of this stylesheet.')
            }
        }
    }

    getContextMenuActionsForRule(node, $li) {
        var _this = this;
        var only_classes = false;
        var cs = pgcssh.getStylesheetForNode(node);
        var locked = !cs.canEdit(true);
        var main_cs = cs;
        var r = pgcssh.getRuleForNode(node);

        var $menu = new CrsaContextMenu(null, $li);

        var createCSSelector = function($li, suffix) {
            suffix = suffix || '';
            var options = pinegrow.stylesheets.getAllAsSelectOptions(null, true);
            //options.unshift({name: '[current stylesheet]', key:'current'})

            var cslist = pinegrow.getSelectedPage().getAttachedStylesheets(
                {
                    loaded: false,
                    ignored: true,
                    imported: true,
                    inline: false,
                    locked: false
                })

            var options = cslist.map(function (cs) {
                var short = crsaGetNameFromUrl(cs.url)
                return {name: short + (cs.locked ? ' &#128274;' : ''), key: cs.url};
            })

            var in_input = new PgInputField('stylesheet', {
                type: 'select',
                show_empty: true,
                name: 'in',
                without_text: false,
                system_field: true,
                default_value: null,
                options: options
            });

            var def_value = pgcssh.getUrlOfStylesheetForOverrides(null, suffix);
            def_value && in_input.setValue(def_value);

            $li.addClass('has-field');
            in_input.$field.addClass('field-in-menu');

            in_input.appendTo($li);

            //if(options.stylesheet_value) {
            //  in_input.setValue(options.stylesheet_value)
            //}

            in_input.on('change', function(e) {
                var v = in_input.getValue();
                if(v) {
                    var cs = pinegrow.stylesheets.getByUrl(v);
                    if(cs && !cs.canEdit(true)) {
                        pinegrow.showQuickError('The selected stylesheet is locked.');
                        in_input.setValue('');
                        return;
                    }
                }
                pgcssh.setUrlOfStylesheetForOverrides(v, null, suffix);
            })
        }

        var withSelectedCS = function(func, suffix) {
            suffix = suffix || '';
            var cs;
            var csurl = pgcssh.getUrlOfStylesheetForOverrides(null, suffix);
            if(!csurl) {
                pinegrow.showQuickMessage('Please select the destination stylesheet.', 4000, false, 'error');
                return;
            }
            if (csurl == 'current') {
                cs = main_cs;
            } else {
                cs = pinegrow.stylesheets.getByUrl(csurl);
            }
            if(cs) {
                if(!cs.canEdit()) return;

                cs.ignore = false;
                if(cs.loaded) {
                    func(cs);
                } else {
                    cs.load(cs.url, function() {
                        func(cs);
                    }, true /* force loading ignored cs */);
                }
            } else {
                pinegrow.showAlert('Stylesheet with ' + csurl + ' not found or not loaded!', 'Error');
            }
        }

        var se = pinegrow.getSelectedElement();

        var getSelectableRule = function(node) {
            if(node.type === 'rule') return node;

            for(var i = 0; i < node.nodes.length; i++) {
                var rule = getSelectableRule(node.nodes[i]);
                if(rule) return rule;
            }
            return null;
        }

        var duplicate = function(cs, r, node, pos) {
            if(!cs.canEdit()) return;
            node = node || r;

            if (pgcssh.isMedia(node.parent)) node = node.parent;
            var position = pos || (cs === main_cs ? node : 'append');
            var api = new PgApi();
            api.addCSSRuleToStylesheet(r, cs, position, {
                action: 'Duplicate ' + pgcssh.getNodeDisplayName(node)
            }, function(error, new_nodes) {
                if(!error) {
                    if(new_nodes.length) {
                        var rule = getSelectableRule(new_nodes[0]);

                        var show_cs = true;
                        if(show_cs &&  cs !== main_cs) {
                            //show this cs
                            pinegrow.activeCSSView.setActiveCs(cs);
                        }
                        if(rule) {
                            pinegrow.selectRule(rule);
                        }
                    }
                }
            })
        }

        var override = function(cs, r) {
            if(!cs.canEdit()) return;
            var node = r;
            if (pgcssh.isMedia(node.parent)) node = node.parent;

            var api = new PgApi();
            var code = pgcssh.getSelectorWithoutPgId(r.selector) + ' {}';

            var position = cs === main_cs ? node : 'append';
            api.addCSSRuleToStylesheet(code, cs, position, {
                action: 'Override ' + pgcssh.getNodeDisplayName(r)
            }, function (error, new_nodes) {
                if (!error) {
                    if (new_nodes.length) {
                        var rule = getSelectableRule(new_nodes[0]);

                        var show_cs = true;
                        if(show_cs &&  cs !== main_cs) {
                            //show this cs
                            pinegrow.activeCSSView.setActiveCs(cs);
                        }
                        if(rule) {
                            pinegrow.selectRule(rule);
                        }
                    }
                }
            })
        }

        var duplicateNonRule = function (cs, node, position) {
            var rule_code = cs.getSourceForRules(node);
            rule_code = cs.filterSource(rule_code, true);

            duplicate(cs, rule_code, node, position);
        }

        if(locked) {
            $menu.add({
                label: 'This stylesheet is locked.',
                action: function() {}
            })
        }
        if(r) {

            if (!only_classes) {
                if (r.selector && !locked) {
                    $menu.add("Edit...", null, function () {
                        pinegrow.selectRule(r);
                    });
                }
                $menu.add("Edit code...", null, function () {
                    _this.editRuleSource(r, $li);
                });

                if(cs && cs.url) {
                    $menu.add("Reveal in code", null, function () {
                        _this.highlightNodeInCode(r, true)
                    });
                }

            }
        }

        var isVariable = pgcssh.isVariable;

        if (pgcssh.getNodeType(node) === 'decl' && r === node.parent) {
            $menu.add("", null, null, 'divider');
            $menu.add("!important", null, function () {
                if (cs.canEdit(false, false)) {
                    var oldValue = node.value;
                    var value = oldValue;
                    if (value.endsWith('!important') || node.important) {
                        node.value = value.replace(/\s*!important/, '');
                        node.important = false;
                    } else {
                        node.important = true;
                    }

                    var nodeIndices = pgcssh.getNodeIndices(node);
                    var rootNode = pgcssh.getRootNodeFor(node);

                    cs.setSourceForRules(rootNode, function () {
                        var updatedNode = pgcssh.getNodeByIndices(cs.getRootNode(), nodeIndices);
                        pinegrow.dispatchEvent('on_css_node_updated', null, {
                            node: updatedNode,
                            stylesheet: cs,
                            edited_by: 'context_menu',
                            oldValue: pgcssh.getNodeOldValues(updatedNode),
                            undoCombineKey: 'ruleValueChanged_' + updatedNode.prop
                        });
                    }, false, true /* fastDelayed */ );
                }
            }, null, null, node.value.endsWith('!important') || node.important);
        }

        if (r) {
            $menu.add("", null, null, 'divider');
            $menu.add("Duplicate the whole rule", null, null, 'header');

            if(!locked) {
                $menu.add("Duplicate", null, function () {
                    duplicate(main_cs, r);
                });
            }

            $menu.add("<span>Duplicate</span>", null, function () {
                withSelectedCS(function (cs) {
                    duplicate(cs, r);
                })
            }, null, null, null, function ($li) {
                //onShow
                createCSSelector($li);
            });
        } else if (!isVariable(node)) {
            $menu.add("", null, null, 'divider');

            var cs_suffix = '-media';
            $menu.add("<span>Duplicate</span>", null, function () {
                withSelectedCS(function (cs) {
                    duplicateNonRule(cs, node);
                }, cs_suffix)
            }, null, null, null, function ($li) {
                //onShow
                createCSSelector($li, cs_suffix);
            });
        }
        else {
            $menu.add("", null, null, 'divider');
            //$menu.add("Duplicate the whole rule", null, null, 'header');

            if(!locked) {
                $menu.add("Duplicate", null, function () {
                    duplicateNonRule(main_cs, node);
                });
            }

            var isVarNode = isVariable(node);

            var label = isVarNode ? 'Customize' : 'Duplicate';
            var cs_suffix = isVarNode ? '-vars' : '';

            $menu.add("<span>" + label + "</span>", null, function () {
                withSelectedCS(function (cs) {
                    var api = new PgApi();

                    api.overrideCSSVariable(node, cs, function(new_nodes) {
                        if(new_nodes.length) {
                            pinegrow.activeCSSView.setActiveCs(cs);
                            pinegrow.selectRule(new_nodes.pop());
                        }
                    })
                }, cs_suffix)
            }, null, null, null, function ($li) {
                //onShow
                createCSSelector($li,  cs_suffix);
            });
        }

        if (pgcssh.getNodeType(node) === 'decl' || pgcssh.getNodeType(node) === 'atrule') {

            if (cs.isCompilableStylesheet() && !locked) {
                // var fr_lvl_nodes_obj = crsaGetAllUsedNodes(cs, node, true);
                var fr_lvl_nodes_obj = cs.smartGetAllUsedNodes(node, true);
                var fr_lvl_nodes = fr_lvl_nodes_obj.nodes;

                if (fr_lvl_nodes.length > 0) {
                    $menu.add("", null, null, 'divider');
                    $menu.add("Edit variables", null, null, 'header');

                    for (var i = 0; i < fr_lvl_nodes.length; i++) {
                        var curr_node = fr_lvl_nodes[i];

                        var prop = pgcssh.getVariableProp(curr_node)
                        $menu.add("Edit " + prop, null, function (pgel, e, $li, data) {
                            if (!data || !data.node) return;

                            var element_node = data.node;
                            var $element = $('<div class="var-dialog-content"></div>');
                            var view = new PgVarDialogView($element);

                            var nodeStylesheet = pgcssh.getStylesheetForNode(element_node);

                            var parents = nodeStylesheet.getMainParnetStylesheets();
                            var nodes = [];

                            parents.forEach(function(pcs) {
                                var list = pcs.smartGetAllUsedNodes(element_node);
                                if(list) {
                                    nodes = nodes.concat(list.nodes);
                                }
                            })

                            //console.log(nodes);

                            // var used_nodes = crsaGetAllUsedNodes(nodeStylesheet, curr_node);
                            //var used_nodes = nodeStylesheet.smartGetAllUsedNodes(element_node);
                            //var nodes = used_nodes.nodes;

                            var varsPane = new PgStylesheetVars(nodeStylesheet, $element, {
                                useInsertBorder: false
                            });
                            varsPane.show(nodes);

                            view.setStylesheetVar(varsPane);
                            var win = new PgQuickWindow('Edit ' + pgcssh.getVariableProp(element_node), view, 'vars-editor', 380, 400);
                            // win.showAtRandomPosition();
                            win.showFor$El($li);

                            var onCssSourceChanged = function (p, info) {
                                if (info) {
                                    var list = info.list;
                                    for (var i = 0; i < list.length; i++) {
                                        var cslist = list[i];
                                        if (varsPane.isPgStylesheetExists(cslist)) {
                                            win.close();
                                            break;
                                        }
                                    }
                                }
                            }

                            pinegrow.addEventHandler('on_css_source_changed', onCssSourceChanged);

                            win.onDestroy = function() {
                                pinegrow.removeEventHandler('on_css_source_changed', onCssSourceChanged);
                            }

                        }, null, {node: curr_node});
                    }
                }
            }

        }

        if(r && r.selector) {

            $menu.add("", null, null, 'divider');
            $menu.add("Create empty rule with this selector", null, null, 'header');

            if(!locked) {
                $menu.add("<span>Override</span>", null, function () {
                    override(main_cs, r);
                }, null, null, null, function ($li) {
                });
            }

            $menu.add("<span>Override</span>", null, function () {
                withSelectedCS(function(cs) {
                    override(cs, r);
                })
            }, null, null, null, function($li) {
                //onShow
                createCSSelector($li);
            });
        }

        if(!locked) {
            $menu.add("", null, null, 'divider');
            $menu.add("Delete", null, function () {
                if (!main_cs.canEdit(false, true)) return;
                _this.deleteNode(r)
            });
        }

        if(r) {
            var classes = r.selector ? pgcssh.getClassesFromSelector(r.selector) : [];

            if (classes.length) {
                if (se) {
                    var remove = [];
                    var add = [];

                    $.each(classes, function (i, cls) {
                        if (se.hasClass(cls)) {
                            remove.push(cls);
                        } else {
                            add.push(cls);
                        }
                    })

                    if (add.length) {
                        $menu.add("", null, null, 'divider');
                        $menu.add("Add class to&nbsp;&nbsp;<b>" + pinegrow.selectedElements.getName() + "</b>", null, null, 'header');

                        add.forEach(function (cls) {
                            $menu.add('+ ' + cls, null, function () {
                                var api = new PgApi();
                                api.addClass(null, cls)
                            });
                        })
                    }

                    var all = se.getClasses();
                    var other = [];
                    /*
                     if (all) {
                     $.each(all, function(i, cls) {
                     if(remove.indexOf(cls) < 0) {
                     other.push(cls);
                     }
                     });
                     }*/

                    if (remove.length > 0 || other.length > 0) {
                        $menu.add("", null, null, 'divider');
                        $menu.add("Remove class from&nbsp;&nbsp;<b>" + pinegrow.selectedElements.getName() + "</b>", null, null, 'header');
                        $.each(remove, function (i, cls) {
                            $menu.add('- ' + cls, null, function () {
                                var api = new PgApi();
                                api.removeClass(null, cls)
                            });
                        })
                        $.each(other, function (i, cls) {
                            if (cls.indexOf('crsa-') !== 0) {
                                $menu.add('- ' + cls, null, function () {
                                    var api = new PgApi();
                                    api.removeClass(null, cls)
                                });
                            }
                        })
                    }
                } else {
                    $menu.add("Select element first, then assign class to it.", null, null, 'header');
                }
            }
        }

        /*
        if (!locked && r && r.nodes !== undefined) {
            $menu.add("", null, null, 'divider');
            $menu.add("Add rule inside...", null, function() {
                //TODO
            });
            $menu.add("Add rule after...", null, function() {
                //TODO
            });
            $menu.add("Add rule before...", null, function() {
                //TODO
            });
        }

         */

        return $menu.actions;
    }

    editRuleSource(rule, $target) {
        rule = pgcssh.getRuleForNode(rule);
        var cs = pgcssh.getStylesheetForNode(rule);

        var q = new PgQuickCSSRuleCodeEdit(rule, $target);

        if(!cs.canEdit(true)) {
            pinegrow.showQuickError('Opened in read-only mode because the stylesheet is locked.');
        }
    }
}