var PgDialogFieldsView = function(def, values, validate_on_submit) {

    var $element = $('<div class="pg-dialog-fields"></div>');
    var view = new PgUIView($element);
    this.view = view;

    var obj = {}; //dummy obj

    this.onValueChanged = function(fn, value) {};

    var _this = this;
    var fields = []
    var $fields_map = {}

    var grid = new PgUISectionGrid();
    $element.append(grid.$element);
    grid.onGetNumberOfColumns = function() { return 1; }



    var getValue = function(obj, fn) {
        return (fn in values) ? values[fn] : null;
    }

    var setValue = function(obj, fn, value, a, b, def) {
        if(def.neg_value && !value) {
            value = def.neg_value;
        }
        values[fn] = value;
        _this.onValueChanged(fn, value);
    }

    this.isValid = function() {
        if(!validate_on_submit) {
            return grid.$element.find('p.error-message:visible').length === 0;
        }

        var valid = true;
        $.each(def.sections, function(key, s) {
            $.each(s.fields, function (fn, fdef) {
                if(fdef.validate) {
                    var ok = pinegrow.validateField(obj, fn, values.hasOwnProperty(fn) ? values[fn] : null, fdef, $fields_map[fn], values);
                    valid = valid && ok;
                }
            })
        })
        return valid;
    }

    this.populateDefaultValues = function() {
        $.each(def.sections, function(key, s) {
            $.each(s.fields, function (fn, fdef) {
                if(fdef.default_value && (values[fn] === undefined || values[fn] === null || values[fn] === '')) {
                    values[fn] = fdef.default_value;
                }
            })
        })
    }

    this.show = function() {
        var $dest = grid.$element;
        var $scrollParent = $dest.parent();

        var current_scroll_top = $scrollParent.scrollTop();
        grid.clear();

        var sections_array = [];
        var on_fields_created = []

        //Pre-create custom cotrols so that definition is done before we get values
        $.each(def.sections, function(key, s) {
            sections_array.push(s);
            s.section_key = key;

            $.each(s.fields, function (fn, fdef) {
                if (fdef.type == 'custom') {
                    if (fdef.control && !fdef.control_defined) {
                        var control = fdef.control();
                        var section = control.define(s.component_definition);
                        if(section) sections_array.push(section);
                        fdef.control_defined = true;
                    }
                }
            })
        })

        $.each(sections_array, function(i, s) {
            var key = s.section_key;
            var pgsec = new PgUISection(s.name, null, s, null, s.force_open || false);
            pgsec.is_closable = s.is_closable !== undefined ? s.is_closable : true;

            pgsec._propData = {
                controls: [],
                fields: []
            }

            grid.addSection(pgsec);

            var $c = pgsec.$content;

            $.each(s.fields, function(fn, fdef) {

                var $field;

                var $append_to = fdef.append_to ? $c.parent().find(fdef.append_to) : $c

                if(fdef.hasOwnProperty("show") && !fdef.show) return true;

                fdef.action = 'custom';

                fdef.get_value = getValue;

                fdef.set_value = function(obj, value, a, b, c, def) {
                    setValue(obj, fn, value, b, c, def);
                    return value;
                }

                if(fdef.type === 'custom') {
                    if(fdef.control) {
                        var control;

                        control = fdef.control();
                        control.setObject(obj);
                        control.setValues(values);
                        control.show($scrollParent);
                        $append_to.append(control.$element);

                        $field = control.$element;
                        pgsec._propData.controls.push(control);
                    }
                } else {
                    var field = new PgPropertyField($c, obj, fn, fdef, values, $scrollParent);

                    $field = field.$field;

                    pgsec._propData.fields.push(field)
                }
                $fields_map[fn] = $field;

                if (!fdef.without_icon) {
                    if (s.icons) {
                        $field.addClass('with-icons');
                        $('<i class="field-icon icon-' + fn + '"></i>').insertAfter($field.find('>label'));
                    }
                }
                if(fdef.on_fields_created) {
                    on_fields_created.push({func: fdef.on_fields_created, obj: obj, field: $field, def: fdef, name: fn, default_value: fdef.default_value || null});
                }
                if(fdef.validate && !validate_on_submit) {
                    pinegrow.validateField(obj, fn, values.hasOwnProperty(fn) ? values[fn] : null, fdef, $field, values);
                }

                if(fdef.show_if) {
                    fdef.show_if_selector = '.fields';
                    var show = typeof fdef.show_if == 'function' ? fdef.show_if( values, obj) : crsaEvaluateShowIfString(fdef.show_if, values, fdef, obj);
                    if(show) {
                        pg$Show($field, fdef.control ? 'block' : 'flex');
                    } else {
                        pg$Hide($field);
                    }
                }
                if(fdef.on_field_updated) {
                    fdef.on_field_updated(obj, $field, fdef, control, fn, null);

                    if(control && control.shown_fields) {
                        $.each(control.shown_fields, function(key, field) {
                            fdef.on_field_updated(obj, field.$field, field.def, key, field);
                        })
                    }
                }

            });
        })

        for(var i = 0; i < on_fields_created.length; i++) {
            var val = values.hasOwnProperty(on_fields_created[i].name) ? values[on_fields_created[i].name] : null;
            if(val === null && on_fields_created[i].default_value) val = on_fields_created[i].default_value;
            on_fields_created[i].func(
                on_fields_created[i].obj,
                on_fields_created[i].name,
                val,
                on_fields_created[i].def,
                on_fields_created[i].field,
                values);

        }
        on_fields_created = null;

        grid.updateLayout();
    }

    view.onDestroy = function() {
        var $dest = grid.$element;
        $dest.find('.crsa-input-color-picker').spectrum('destroy');
        pgRemoveMultiselects($dest);
        $dest.trigger('crsa-will-clear-area');
        grid.clear();
    }
}

var PgDialog = function(title) {

    var _this = this;

    this.before = '';
    this.after = '';
    this.sections = {}
    this.values = {}
    this.cancelButtonLabel = 'Cancel';
    this.okButtonLabel = 'OK';
    this.onCancel = function() {}
    this.onOK = function() {}
    this.validate_on_submit = false;
    this.$dialog = null;

    var fv = null;
    var $d;

    this.setFieldValue = function(field, val) {
        if(fv) {
            var $i = fv.view.get$Element().find('.crsa-field-' + field + ' input.crsa-input');
            $i.val(val);
            $i.trigger('change')
        }
    }

    this.get$OKButton = function() {
        return $d.find('.modal-footer button.ok');
    }

    this.get$Dialog = function() {
        return $d;
    }

    var closed = function() {
        fv && fv.view.destroy();
    }

    this.hide = function() {
        closed();
        $d.modal('hide');
    }

    this.show = function() {

        var def = new PgComponentType('pg.dialog', 'Dialog dummy def', {
            selector: null,
            not_main_type: true,
            sections: this.sections
        });

        fv = new PgDialogFieldsView(def, this.values, this.validate_on_submit);
        $d = pinegrow.showAlert(this.before + '<div class="fields"></div>' + this.after, title, this.cancelButtonLabel, this.okButtonLabel, function() {
            closed();
            _this.onCancel();

        }, function() {
            if(fv && !fv.isValid()) {
                pinegrow.showQuickError('The form contains errors.')
                return false;
            }
            if(fv) {
                fv.populateDefaultValues();
            }
            if(_this.onOK(function() {
                $d.modal('hide');
                closed();
            }) === false) {
                //will be closed async
                return false;
            } else {
                closed();
            }
        })
        this.$dialog = $d;
        $d.find('.fields').append(fv.view.get$Element());

        fv.show();

        var first_field = fv.view.get$Element().get(0).querySelector('input,textarea,select');
        first_field && first_field.focus();

        //crsaHandleExternalLinks($d);
        return $d;
    }
}



