var PgPaintToolManager = function() {

    var _this = this;
    var tool = null;

    const $icon = $('<div class="pg-paint-tool-icon"></div>')
    var $cursor;

    var fastdomTasks = pgFastdomTasks;
    var fastdomKey = 'paintToolCursor';

    var mode = pinegrow.getSetting('paint-tools-mode', 'set');

    this.getMode = function() {
        return mode;
    }

    this.setMode = function(m) {
        mode = m;
        pinegrow.setSetting('paint-tools-mode',m);
        pinegrow.showMode(m === 'set' ? 'Set mode' : 'Paint mode', true);
    }

    var onMouseMove = function(e) {
        if(mode === 'paint') {
            var x = e.pageX;
            var y = e.pageY;

            var task = fastdomTasks.create(fastdomKey);

            task.mutate(function () {
                $cursor.css({
                    top: y + 'px',
                    left: x + 'px'
                })
            })
        }
    }

    this.onMouseMoveOverPage = function(e) {
        if(!tool || mode !== 'paint') return;

        var $e = $(e.target);
        var $iframe = getIframeOfElement($e);

        if($iframe) {
            var page = getCrsaPageForIframe($iframe);
            var pv = page.getPageViewOf$DOMElement($e, $iframe);
            if(pv) {
                var task = fastdomTasks.create(fastdomKey);

                task.measure(function() {
                    var pos = pv.pageRectToGlobalRect({
                        left: e.pageX,
                        top: e.pageY,
                        width: 0,
                        height: 0
                    })

                    task.mutate(function() {
                        $cursor && $cursor.css({
                            top: pos.top + 'px',
                            left: pos.left + 'px'
                        })
                    })
                })
            }
        }
    }

    this.setTool = function(t, event) {
        const $body = $('body');
        this.clear();

        if(mode === 'paint') {
            tool = t;

            $icon.append(tool.onGetIcon());
            $body.addClass('pg-dm-paint-text');

            $body.on('mousemove', onMouseMove);

            $cursor = t.getCursor();
            $body.append($cursor);

            pinegrow.dispatchEvent('on_paint_tool_changed', null, tool);

            if(!event) {
                var pos = pinegrow.mousePosition.getPosition();
                event = {
                    pageX: pos.left,
                    pageY: pos.top
                }
            }
            onMouseMove(event);
        } else {
            //set mode
            var col = pinegrow.getCollection();
            if(col.getLength() === 0) {
                pinegrow.showQuickMessage('Select an element first.')
            } else {
                col.forEachElement(function (pgel) {
                    t.onClick(pgel);
                })
            }
        }

        //var $page_area = $body.find('.pg-page-area');
        //$bin.appendTo($page_area);

    }

    this.clear = function() {
        $icon.empty();
        $('body').removeClass('pg-dm-paint-text');
        if(tool) {
            tool.onDestroy();
            tool = null;
        }
        $('body').off('mousemove', onMouseMove);
        $cursor && $cursor.remove();

        pinegrow.dispatchEvent('on_paint_tool_changed', null, null);
        //$bin.detach();
    }

    this.getTool = function() {
        return tool;
    }

    pinegrow.addEventHandler('on_element_clicked', function(page, pgel, $el, status) {
        if(tool) {
            status.stop = true;
            tool.onClick.call(tool, pgel, $el);
        }
    })
}

var PgPaintTool = function() {

    var _this = this;

    this.action = 'class';
    this.action_name = null;
    this.class = null;
    this.all_classes = [];
    this.svg_support = false;
    this.icon = null;
    this.icon_color;
    this.background_color;
    this.cursor_html;
    this.on_before_set;
    this.toggle_class = false;

    this.prev_pgel = null;
    this.prev_class = null;

    this.getCursor = function() {
        return $(`<div class="pg-paint-tool-cursor">${this.cursor_html}</div>`);

        return $(`<div class="pg-paint-tool-cursor" style="background-color: ${this.background_color}"><i style="color: ${this.icon_color}" class="icon ${this.icon}"></i></div>`);
    }

    this.onGetIcon = function() {};

    this.setClasses = function(pgel, classes) {
        var set_svg = false;
        var svg_parent = pgel.closest('svg');
        var in_svg = svg_parent && pgel !== svg_parent;

        if(_this.svg_support && in_svg && classes.length) {
            if(pgel.getAttribute('stroke') !== 'currentColor' && pgel.getAttribute('fill') !== 'currentColor') {
                set_svg = true;
            }
        }

        var all_classes = this.all_classes;

        if(in_svg && this.on_filter_svg) {
            this.on_filter_svg(classes, all_classes);
        }

        var api = new PgApi();

        var name = _this.action_name || (classes.length ? ('Set class ' + classes[0]) : 'Clear');

        pinegrow.changeManager.batchChanges(function() {

            pinegrow.makeChanges(pgel, name, function() {
                var skip_default = false;
                if(_this.on_before_set) {
                    if(_this.on_before_set(pgel, classes) === false) {
                        //if false is returned we skip default
                        skip_default = true;
                    }
                }
                if(!skip_default) {
                    if(_this.toggle_class) {
                        if(classes.length === 0) {
                            //keep clear behaviour consistent
                            all_classes.forEach(function (cls) {
                                api.hasClass(pgel, cls) && api.removeClass(pgel, cls)
                            })
                        } else {
                            api.hasClass(pgel, classes[0]) ? api.removeClass(pgel, classes[0]) : api.addClass(pgel, classes[0]);
                        }
                    } else {
                        all_classes.forEach(function (cls) {
                            api.hasClass(pgel, cls) && api.removeClass(pgel, cls)
                        })
                        classes.forEach(function (cls) {
                            if (cls.length) {
                                api.addClass(pgel, cls);
                            }
                        })
                        set_svg && pgel.setAttribute(_this.svg_support, 'currentColor');
                    }
                }
            })

        }, name)

    }

    this.onClick = function(pgel) {
       // if(this.class) {

            var classes = this.class ? this.class.split(' ') : [];
/*
            if(this.prev_pgel === pgel) {
                //cycle through classes
                var idx = this.all_classes.indexOf(this.prev_class);
                if(idx === this.all_classes.length - 1) {
                    idx = -1;
                }
                idx++;
                this.prev_class = this.all_classes[idx];
                this.setClasses(pgel, [this.prev_class]);
                return;
            }

            this.prev_pgel = pgel;
            this.prev_class = classes[0];

 */

            this.setClasses(pgel, classes)
      //  }

        this.on_after_set && this.on_after_set(pgel);
    };

    this.onDestroy = function() {};
}


class PgFloatingToolsManager {
    constructor() {
        this.quick_win = null;

        var _this = this;

        this.timer = null;

        pgQuickWindowManager.setRememberInSettings('quickFloatingTools');

        pinegrow.panelManager.registerPanel({
            panel_category: 'Tools',
            panel_key: 'floatingTools',
            title: 'Floating tools',
            panel_description: 'Floating palette of tools that let you "paint" elements on the page.',
            panel_on_show: function() {
                _this.show();
                _this.showAtMousePosition();
            },
            panel_on_hide: function() {
                _this.hide();
            },
            panel_on_is_visible: function() {
                return _this.quick_win && _this.quick_win.visible;
            },
            panel_kbd: '&#8679; &#8679;',
            icon: ''
        })

        pinegrow.addEventHandler('on_page_loaded', function(page) {
            if(pinegrow.getAllPages().length === 1) {
                //first page opened
                if(pinegrow.getSetting('floating-tools', 'visible') === 'visible') {
                    setTimeout(function() {
                        if(!_this.quick_win || !_this.quick_win.visible) {
                            var sections = []
                            pinegrow.dispatchEvent('on_get_floating_tools_sections', pinegrow.getSelectedPage(), sections);
                            if(sections.length) {

                                _this.show();

                                var pos = pgQuickWindowManager.getPosition('quickFloatingTools');
                                if (pos) {
                                    _this.quick_win.showAtPosition(pos.x, pos.y, null, true);
                                } else {
                                    _this.showAtMousePosition();
                                }
                            }
                        }
                    }, 1500)
                }
            }
        })
    }

    show() {
        var show_tooltip = false;

        if(!this.quick_win) {
            show_tooltip = true;
            this.quick_win = PgQuickFloatingToolsWindow();

            var _this = this;
            this.quick_win.$element.on("transitionend", function(e) {
                //just react to events on the element itself, for focus top bar
                if(e.target === _this.quick_win.$element.get(0)) {
                    _this.quick_win.$element.removeClass('animate-move');
                }
            });
        }
        this.timer && clearTimeout(this.timer);
        this.quick_win.show();

        var _this = this;
        if(show_tooltip) {
            setTimeout(function() {
                addTooltip(_this.quick_win.$element, `Press ${pgShowKbd('SHIFT')} ${pgShowKbd('SHIFT')} (the SHIFT key twice) to display the toole palette at the current mouse position.`, {trigger: 'manual'})
                _this.quick_win.$element.tooltip('show');
            }, 2000)
        }
    }

    hide() {
        if(this.quick_win) {
            this.quick_win.hide();
        }
    }

    showAtMousePosition() {
        this.show();

        var pos = pinegrow.mousePosition.getPosition();

        this.quick_win.$element.addClass('animate-move');
        this.quick_win.$element.css({
            top: pos.top + 'px',
            left: pos.left + 'px'
        })
    }
}


var PgQuickFloatingToolsWindow = function($target) {

    var v = new PgFloatingToolsView();

    var win = new PgQuickWindow('Paint mode', v, 'quickFloatingTools', 400, 80, null, false);
    win.$element.addClass('pg-floating-tools-win');

    v.quick_win = win;

    win.hideOnClose = true;
    win.isClosable = false;

    win.onHide = function(close_all) {
        pinegrow.paintToolManager.clear();
        if(!close_all) {
            pinegrow.setSetting('floating-tools', 'hidden');

            pinegrow.showNotice(`<p>To show the floating tools:</p>
<ul>
<li>Press ${pgShowKbd('SHIFT')} ${pgShowKbd('SHIFT')} (the SHIFT key twice) to show or move the tools to the current mouse position.</li>
<li>Use Window -&gt; Show / Hide panels -&gt; Floating tools.</li>
</ul>`, 'How to show the floating tools?', 'floating-tools-show', null, true);
        }
    }

    win.onShow = function() {
        pinegrow.setSetting('floating-tools', 'visible');
        v.updateSize();
    }

    win.onMove = function() {

    }

    win.noRightEdge();
    //win.hideOnClose = true;

    $target = $target || $('body');

    win.showAtRandomPosition();

    if($target) {
        //win.showFor$El($target);
    }

    v.update();

    win.onDestroy = function() {
    }


    return win;
}

class PgFloatingToolsView {
    constructor() {
        this.$element = $('<div class="pg-floating-tools"><i class="icon icon-Pin6 mover"></i></div>');
        this.$sections = $('<div class="pg-floating-tools-sections"></div>').appendTo(this.$element);
        this.$bin = $('<div class="pg-floating-tools-bin"><i class="icon icon-bin"></i></div>').appendTo(this.$element);
        addTooltip(this.$bin, `Drop the paint tool in the bin to stop using it or press the ${pgShowKbd('ESC')} key.`)

        this.$no_tools = $('<div class="pg-floating-tools-no-sections">No tools are defined.</div>').appendTo(this.$element);

        this.view = new PgUIView(this.$element);

        var _this = this;
        this.displayed_sections = [];
        this.sections = [];

        var first_time = true;

        this.more_section = new PgFloatingToolsSection('more', 'More tools');

        var more_control = new PgFloatingToolsControl({
            type: 'button',
            name: '<i class="icon icon-settings"></i>',
            helptext: 'Click to open the Properties panel where you can edit properties of the selected element.',
            func: function() {
                var $button = more_control.$element;

                var a = [
                    {
                        type: 'header',
                        label: 'Mode'
                    },
                    {
                        label: 'Paint mode',
                        action: function() {
                            pinegrow.paintToolManager.setMode('paint');
                        },
                        check: pinegrow.paintToolManager.getMode() === 'paint',
                        submenu_icon: 'icon-none',
                        submenu: [
                            {
                                type: 'html',
                                html: `<p style="color:rgba(255,255,255,0.75);margin-bottom:0;padding:10px 15px;">In Paint mode you click on the tool to pick it up and then click on elements on the page to apply the tool.</p>`
                            }
                        ],
                    },
                    {
                        label: 'Set mode',
                        action: function() {
                            pinegrow.paintToolManager.setMode('set');
                        },
                        check: pinegrow.paintToolManager.getMode() === 'set',
                        submenu_icon: 'icon-none',
                        submenu: [
                            {
                                type: 'html',
                                html: `<p style="color:rgba(255,255,255,0.75);margin-bottom:0;padding:10px 15px;">In Set mode you click on the tool to apply it to one or more selected elements.</p>`
                            }
                        ],
                    },
                    {
                        type: 'divider'
                    },
                    {
                        label: 'More tools &amp; controls...',
                        action: function() {
                            var pgel = pinegrow.selectedElements.getLastSelected();
                            PgQuickProperties(null, pgel, $button);
                            //pinegrow.showUIPanel('prop', false, false, true);
                        }
                    }
                ]

                if(!$button.data('pg-menu')) {

                    var m = new CrsaContextMenu(a, $button);
                    m.showForElement($button, null, 4);

                    setTimeout(function () {
                        m.closeOnMouseOut = true;
                    }, 300)
                }

                return false;
            }
        })

        this.more_section.addControl(more_control);

        pinegrow.addEventHandler('on_page_frameworks_changed', function(page) {
            _this.update();
        });

        pinegrow.addEventHandler('on_page_selected', function(page) {
            _this.update();
        });

        var api = new PgApi();

        var updateSelectedElement = function(page) {
            if(pinegrow.paintToolManager.getMode() === 'set') {
                var pgel = pinegrow.selectedElements.getLastSelected();
                var info = {
                    classes: pgel ? api.getClasses(pgel) : []
                }
                _this.sections.forEach(function(section) {
                    section.onElementSelected(pgel, info);
                })
            }
        }

        pinegrow.addEventHandler('on_selected_elements_changed', updateSelectedElement);
        pinegrow.addEventHandler('on_current_pseudo_state_changed', updateSelectedElement);
        pinegrow.addEventHandler('on_current_device_size_changed', updateSelectedElement);


        pinegrow.addEventHandler('on_paint_tool_changed', function(pagenull, tool) {
            if(tool) {
                _this.$bin.addClass('active');
                if(first_time) {
                    _this.$bin.tooltip('show');
                    first_time = false;
                }
            } else {
                _this.$bin.removeClass('active');
            }
        });

        this.$bin.on('click', function(e) {
            e.preventDefault();
            pinegrow.paintToolManager.clear();
        })
    }

    update() {
        var _this = this;
        this.view.whenVisible(function() {
            _this.getSections();
            if(!pgAreArraysEqual(_this.displayed_sections, _this.sections)) {
                _this.displayed_sections.forEach(function (s) {
                    s.detach();
                })
                _this.displayed_sections = _this.sections;

                _this.displayed_sections.forEach(function (s) {
                    s.attach(_this.$sections);
                })
                if(_this.displayed_sections.length === 0) {
                    pg$Show(_this.$no_tools, 'flex');
                } else {
                    pg$Hide(_this.$no_tools);
                }
                //update size
                _this.updateSize()
            }
        }, 'update')
    }

    updateSize() {
        if(this.quick_win) {
            this.quick_win.$element.css('width', '1000px');
            var w = this.quick_win.$element.find('.pg-quick-win-content').width();
            this.quick_win.$element.css('width', (w + 60) + 'px');
        }
    }

    onPageSelected(page) {

    }

    getSections() {
        var sections = []
        pinegrow.dispatchEvent('on_get_floating_tools_sections', pinegrow.getSelectedPage(), sections);
        if(sections.length) {
            sections.push(this.more_section)
        }
        this.sections = sections;
    }
}


class PgFloatingToolsSection {
    constructor(key, name) {
        this.$element = $(`<div class="pg-floating-tools-section"><sectionname>${name}</sectionname></div>`);
        this.$controls = $(`<div class="controls"></div>`).prependTo(this.$element);

        this.controls = []
    }

    addControl(c) {
        this.controls.push(c)
        c.attach(this.$controls);
    }

    attach($container) {
        this.$element.appendTo($container);
    }

    detach() {
        this.$element.detach();
    }

    onElementSelected(pgel, info) {
        this.controls.forEach(function(c) {
            c.onElementSelected(pgel, info);
        })
    }
}

class PgFloatingToolsControl {
    constructor(def, options) {
        var _this = this;

        this.options = $.extend({}, {
            buttons: true,
            toggle_class: false
        }, options || {});

        def.compact_select = true;
        def.without_text = true;

        if(this.options.compact_current_html) {
            def.compact_current_html = this.options.compact_current_html;
        }

        this.classPreview = null;
        this.classPreviewValues = null;

        var bm = new PgToggleButtonMaker();

        this.def = def;
        this.field = null;

        if(def.type === 'checkbox') {
            //convert to select
            def.type = 'select';
            def.toggle_buttons = true;
            def.options = [
                {
                    key: def.value,
                    name: def.name,
                    html: bm.makeText(def.name)
                }
            ]
        }



        var label = this.options.label || def.name;
        this.$element = $(`<div class="pg-floating-tools-control pg-floating-tools-control-${def.type}">${label}</div>`)

        /*
        if(def.type === 'select' && !def.toggle_buttons) {
            def.toggle_buttons = true;
            def.options = def.options.map(function(o) {
                o.html = bm.makeText(o.name);
                return o;
            })
        }

         */

        if(def.colors) {
            //color picker
            this.$color_input = $('<input type="hidden" />').appendTo(this.$element);
            this.color_picker = new PgColorPicker(this.$color_input, def.colors, def.name, function() {
                return {};
            }, {
                //settings
                on_color_highlighted: function(val) {
                    if(!_this.classPreview) _this.getClassPreview(def);
                    if(_this.classPreview) {
                        if (val) {
                            _this.classPreview.setClass(val, _this.classPreviewValues);
                        } else {
                            _this.classPreview.restore();
                        }
                    }
                },
                on_show: function() {
                    _this.getClassPreview(def);
                },
                on_value_set: function() {
                    _this.destroyClassPreview();
                },
                on_hide: function() {
                    _this.destroyClassPreview();
                }
            })
            this.$color_input.on('change', function(e) {
                var list = typeof _this.def.options === 'function' ? _this.def.options() : _this.def.options;
                var val = _this.$color_input.val();

                if(!val) {
                    _this.setTool({
                        key: null,
                        all_classes: _this.makeResponsiveClasses(list.map(function (i) {
                            return i.key;
                        })),
                        cursor: 'Clear ' + _this.options.label || _this.def.name
                    })
                } else {
                    var cursor = _this.options.cursor || val;
                    cursor = cursor.replace('$COLOR', _this.color_picker.getColorValue());

                    _this.setTool({
                        key: _this.makeResponsiveClasses(val),
                        all_classes: _this.makeResponsiveClasses(list.map(function (i) {
                            return i.key;
                        })),
                        cursor: cursor
                    })
                }
            })
        }

        if(def.action === 'apply_class_multiple') {
            def.action = 'apply_class';
            def.multiple = false;
        }

        //if(this.def.type !== 'button') {
            this.$element.on('mouseenter', this.onMouseEnter.bind(this));
            this.$element.on('mouseleave', this.onMouseLeave.bind(this));
            this.$element.on('click', this.onClick.bind(this));
        //}

        if(this.def.helptext) {
            addTooltip(this.$element, this.def.helptext)
        }
    }

    getClassPreviewObj() {
        return pinegrow.selectedElements.getLastSelected();
    }

    getClassPreviewResponsiveClass(cls, fdef, pgel) {
        return fdef.responsive_class ? fdef.responsive_class(cls, pgel) : cls;
    }

    getClassPreview(def) {
        if(pinegrow.paintToolManager.getMode() !== 'set') return;

        if(!this.classPreview || this.classPreview.destroyed) {
            var obj = this.getClassPreviewObj();
            if (obj) {
                var items = typeof def.options === 'function' ? def.options(def) : def.options;

                this.classPreviewValues = items.filter(function (i) {
                    return !!i.key;
                })
                var dummy_cls = '___dummy___';
                this.classPreview = new PgFastDOMClassPreviewForPgel(obj, this.getClassPreviewResponsiveClass(dummy_cls, def, obj).replace(dummy_cls, ''));
            }
        }
        return this.classPreview;
    }

    destroyClassPreview() {
        if(this.classPreview) {
            this.classPreview.destroy();
            this.classPreview = null;
        }
    }

    onClick(event) {
        if(event) {
            event.preventDefault();
            event.stopImmediatePropagation();
            event.stopPropagation();
        }

        pinegrow.paintToolManager.clear();

        if(this.def.type === 'button') {
            this.def.func();
        }
    }

    getClassesMap() {
        var map = {}
        var _this = this;
        if(this.def.options) {
            var list = typeof this.def.options === 'function' ? this.def.options() : this.def.options;
            list.forEach(function (i) {
                var res_key = _this.makeResponsiveClasses(i.key);

                map[res_key] = i;

                if(_this.options.submenu) {
                    var options_submenu = typeof _this.options.submenu === 'function' ? _this.options.submenu(i.key) : _this.options.submenu;
                    options_submenu.items.forEach(function(si) {
                        map[_this.makeResponsiveClasses(si.key)] = si;
                    })
                }

            })
        }

        return map;
    }

    onToolApplied(pgel) {
        var api = new PgApi();
        var info = {
            classes: api.getClasses(pgel)
        }
        this.onElementSelected(pgel, info);
    }

    onElementSelected(pgel, info) {
        let has = false;
        const map = this.getClassesMap();
        for(var i = 0; i < info.classes.length; i++) {

            if(map[info.classes[i]]) {
                //has value
                has = map[info.classes[i]];
                break;
            }
        }
        var label = this.options.label || this.def.name;

        if(this.options.on_get_current_value_label) {
            label = this.options.on_get_current_value_label.call(this, info, map);
        }

        if(has !== false) {
            this.$element.addClass('has-value')
            if(!this.options.on_get_current_value_label) {
                if (this.color_picker) {
                    label = this.options.cursor.replace(/\$COLOR/, has.color);
                    this.color_picker.get$Input().val(has.key);
                } else if (this.def.toggle_buttons) {
                    label = this.options.label || has.html;
                }
            }
        } else {
            this.$element.removeClass('has-value')
        }
        if(typeof label === 'function') label = label();

        if(label !== null && this.$element.get(0).innerHTML !== label) {
            this.$element.get(0).innerHTML = label;
        }
    }

    setTool(d) {
        var tool = new PgPaintTool();
        tool.class = d.key || '';
        tool.all_classes = d.all_classes;
        tool.toggle_class = this.options.toggle_class;


        //tool.svg_support = 'fill';
        //tool.icon = 'icon-ikone1-100-copy-63';
        //tool.icon_color = _this.colors.getColor(key);
        tool.cursor_html = d.cursor;

        if(this.options.svg_support) {
            tool.svg_support = this.options.svg_support;
        }
        if(this.options.on_filter_svg) {
            tool.on_filter_svg = this.options.on_filter_svg;
        }

        if(d.on_before_set) {
            tool.on_before_set = d.on_before_set;
        }

        var _this = this;
        if(pinegrow.paintToolManager.getMode() === 'set') {
            tool.on_after_set = function(pgel) {
                _this.onToolApplied(pgel)
            }
        }

        //tool.background_color = pgGetColorWithEnoughContrast(tool.icon_color, 'white', 'black');
        pinegrow.paintToolManager.setTool(tool);

    }

    makeResponsiveClasses(list) {
        var _this = this;
        if(list && this.def.responsive_class) {
            if(typeof list === 'string') {
                return this.def.responsive_class(list);
            } else {
                return list.map(function (cls) {
                    return _this.def.responsive_class(cls);
                })
            }
        } else {
            return list;
        }
    }

    onMouseEnter(event) {

        //tools is moving under the mouse
        if(this.$element.closest('.pg-quick-win').hasClass('animate-move')) return;

        //if(this.def.type === 'button') return;

        this.mouse_is_over = true;
        this.show_timer && clearTimeout(this.show_timer);

        this.show_timer = null;

        var menu = this.$element.data('pg-menu');

        if(event && !menu && pgCurrentContextMenu) {
            //a menu is already open, perhaps the mouse is just travelling over this control on the way to the menu
            var velocity = pinegrow.mousePosition.getYVelocity();
            //console.log('Yvelocity', velocity)
            if(velocity > 10) {
                this.show_timer = setTimeout(this.onMouseEnter.bind(this), velocity > 10 ? 250 : 50);
                return;
            }
        }


        var _this = this;

        var pgel = pinegrow.selectedElements.getLastSelected();
        var is_set = pinegrow.paintToolManager.getMode() === 'set';
        var api = is_set ? new PgApi() : null;

        var isActive = function(cls) {
            if(is_set && pgel) {
                return api.hasClass(pgel, cls);
            } else {
                return false;
            }
        }

        if(this.def.type === 'button') {
            this.onClick(event)
            return;
        }

        if(this.color_picker) {
            pgCloseContextMenus();
            if(!this.color_picker.isOpen()) {
                this.color_picker.close_on_mouseleave = false;
                this.color_picker.showWithPosition(this.$element);
                this.color_picker.keepOpenOnClick = pinegrow.paintToolManager.getMode() === 'set';

                setTimeout(function () {
                    _this.color_picker.close_on_mouseleave = true;
                }, 300)
            }
            return;
        }

        var addClear = function(buttons, all_classes, name) {
            if(!_this.options.buttons) {
                actions.push({
                    type: 'divider'
                })
            }

            buttons.push({
                label: _this.options.buttons ? '<i class="pg-tb-button pg-tb-button-icon icon icon-close" style="font-size:13px !important;top:3px;"></i>' : '<i class="icon icon-close" style="color:white;margin-right:8px;"></i> Clear',
                class: 'pg-floating-tool-menu-item',
                data: {
                    key: null,
                    all_classes: all_classes,
                    on_before_set: _this.options.on_before_set || null,
                    cursor: 'Clear ' +  name
                },
                action: setTool
            })
        }

        var actions = []

        //pgCloseContextMenus();

        var setTool = function() {
            _this.destroyClassPreview();
            if(this.require_submenu_selection && this.submenu) return;
            _this.setTool(this.data);
        }

        var makeResponsiveClasses = this.makeResponsiveClasses.bind(this);

        if(this.def.options) {
            var list = typeof this.def.options === 'function' ? this.def.options() : this.def.options;
            var buttons = [];
            var all_classes = []

            list.forEach(function(i) {
                all_classes.push(i.key);
            })

            all_classes = makeResponsiveClasses(all_classes);

            list.forEach(function(i) {
                var html;
                if(i.html) {
                    html = typeof i.html === 'function' ? i.html() : i.html;
                } else {
                    html = i.name;
                }
                var submenu = null;

                var has_checked_submenu = false;

                if(_this.options.submenu) {
                    var submenu = [];

                    var options_submenu = typeof _this.options.submenu === 'function' ? _this.options.submenu(i.key) : _this.options.submenu;

                    var req_sub = options_submenu.require_submenu_selection || false;

                    var si_all_classes = options_submenu.all_classes;
                    si_all_classes = makeResponsiveClasses(si_all_classes);

                    if(options_submenu.combine_with_parent) {
                        si_all_classes = si_all_classes.concat(all_classes);
                    }

                    var submenu_def = {
                        options: si_all_classes.map(function(cls) {
                            return {
                                key: cls
                            }
                        })
                    }

                    options_submenu.items.forEach(function(si) {
                        if(!options_submenu.filter || options_submenu.filter(i.key)) {
                            var si_key = makeResponsiveClasses( si.key );
                            if (options_submenu.combine_with_parent) {
                                si_key = makeResponsiveClasses(i.key) + ' ' + si_key;
                            }
                            var si_html;
                            if(si.html) {
                                si_html = si.html;
                            } else {
                                si_html = `<span>${si.name}</span>`;
                            }
                            var d = {
                                label: si_html,
                                class: 'pg-floating-tool-menu-item',
                                data: {
                                    key: si_key,
                                    all_classes: si_all_classes,
                                    cursor: options_submenu.combine_with_parent ? html + ' ' + si_html : si_html,
                                    on_before_set: _this.options.on_before_set || null
                                },
                                action: setTool
                            }

                            d.on_mouseenter = function() {
                                if(!_this.classPreview) _this.getClassPreview(_this.def);
                                _this.classPreview && _this.classPreview.setClass(this.data.key, _this.classPreviewValues);
                            }
                            d.on_mouseleave = function(bd) {
                                _this.classPreview && _this.classPreview.restore();
                            }

                            if(is_set) {
                                d.check = isActive(si_key) && (has_checked_submenu = true);
                            }
                            submenu.push(d)
                        }
                    })

                    submenu.push({
                        type: 'divider'
                    })

                    addClear(submenu, si_all_classes, options_submenu.name)

                    if(submenu.length) {
                        submenu.unshift({
                            type: 'header',
                            label: options_submenu.name
                        })
                    }
                }

                var res_key = makeResponsiveClasses(i.key);

                var bd = {
                    label: html,
                    class: 'pg-floating-tool-menu-item',
                    data: {
                        key: res_key,
                        all_classes: all_classes,
                        on_before_set: _this.options.on_before_set || null,
                        cursor: html
                    },
                    action: setTool,
                    submenu: (submenu && submenu.length) ? submenu : null,
                    require_submenu_selection: req_sub
                }

                if(is_set) {
                    if(_this.options.buttons) {
                        if(isActive(res_key)) {
                            bd.class += ' active';
                        }
                    } else {
                        bd.check = isActive(res_key) || has_checked_submenu;
                    }
                }

                if(!bd.submenu) {
                    bd.on_mouseenter = function() {
                        if(!_this.classPreview) _this.getClassPreview(_this.def);
                        _this.classPreview && _this.classPreview.setClass(this.data.key, _this.classPreviewValues);
                    }
                    bd.on_mouseleave = function(bd) {
                        _this.classPreview && _this.classPreview.restore();
                    }
                } else {

                    bd.on_submenu_open = function() {
                        _this.getClassPreview(submenu_def)
                    }
                    bd.on_submenu_close = function() {
                        _this.destroyClassPreview();
                    }
                }
                buttons.push(bd)
            })
        }

        if(this.options.buttons) {
            actions.push({
                buttons: buttons
            })
        } else {
            actions = buttons;
        }

        actions.unshift({
            type: 'header',
            label: this.options.name || this.def.name
        })

        if(!this.options.no_clear_on_first_level) {
            addClear(buttons, all_classes, this.options.name || this.def.name)
        }

        if(!menu) {

            _this.$element.addClass('active');
            var m = new CrsaContextMenu(actions, this.$element);
            m.showForElement(this.$element, null, 4, -60);
            m.onClosed = function() {
                _this.$element.removeClass('active');
                _this.destroyClassPreview();
            }
            m.keepOpenOnClick = pinegrow.paintToolManager.getMode() === 'set';

            //wait a bit, mouse is moving over the menu
            setTimeout(function () {
                m.closeOnMouseOut = true;
            }, 300)

            if(this.options.buttons) {
                this.getClassPreview(this.def);
            }
        }
    }

    onMouseLeave(event) {
//return;
        var _this = this;

        this.mouse_is_over = false;

        this.show_timer && clearTimeout(this.show_timer);
        this.show_timer = null;

        this.mouseout_timer && clearTimeout(this.mouseout_timer);

        var _this = this;

        this.mouseout_timer = setTimeout(function() {
            if(_this.color_picker) {
                if(_this.color_picker.isOpen() && !_this.color_picker.isMouseOver()) {
                    _this.color_picker.close();
                }
            } else {
                var menu = _this.$element.data('pg-menu');
                if (menu && !_this.mouse_is_over) {
                    if (!menu.isMouseOverOrSubmenuOpen()) {
                        menu.close();
                    }
                    _this.mouseout_timer = null;
                }
            }
        }, 200)
    }

    onGetObj() {
        return {}
    }

    onGetValues() {
        return {}
    }

    onValueChanged(value, html, options) {
        if(this.def.action === 'apply_class') {
            var classes = options.map(function(i) { return i.key; })
            classes = this.makeResponsiveClasses(classes);

            var tool = new PgPaintTool();
            tool.class = this.makeResponsiveClasses(value);
            tool.all_classes = classes;

            //tool.svg_support = 'fill';
            //tool.icon = 'icon-ikone1-100-copy-63';
            //tool.icon_color = _this.colors.getColor(key);
            tool.cursor_html = html;

            if(this.options.svg_support) {
                tool.svg_support = this.options.svg_support;
            }
            if(this.options.on_filter_svg) {
                tool.on_filter_svg = this.options.on_filter_svg;
            }

            if(this.options.on_before_set) {
                tool.on_before_set = this.options.on_before_set;
            }
            //tool.background_color = pgGetColorWithEnoughContrast(tool.icon_color, 'white', 'black');
            pinegrow.paintToolManager.setTool(tool);
        }
    }

    getField() {
        if(!this.field) {
            var _this = this;
            this.field = new PgPropertyField(this.$element, this.onGetObj, this.def.key, this.def, this.onGetValues);
            this.field.on('change', function(e) {
                var val = _this.field.getCurrentValueQuick();
                var html = _this.field.$field.find('.pg-tb-compact-current').html() || val;

                if(_this.def.options) {
                    var list = typeof _this.def.options === 'function' ? _this.def.options() : _this.def.options;
                }

                _this.onValueChanged(val, html, list);
            })
        }
        return this.field;
    }

    attach($container) {
        //this.getField();
        $container.append(this.$element);
    }
}