class PgProjectMenuHelper {
    constructor(p) {
        this.project = p || pinegrow.getCurrentProject();
    }

    askToRename(done, text, title) {
        var _this = this;
        pinegrow.backend.askToRenameProject(this.project, function() {
            done && done(_this.project)
        }, text, title);

    }

    close(done, text, title) {
        if(this.project.ask_to_rename_on_close) {
            this.project.ask_to_rename_on_close = false;

            this.askToRename(function() {
                pinegrow.backend.saveAndCloseCurrentProject(done);
            }, text, title)
        } else {
            pinegrow.backend.saveAndCloseCurrentProject(done);
        }
    }

    download(done) {
        if(!pinegrow.currentUser.isPro()) {
            pinegrow.currentUser.onlyForProNotice('Project download')
        } else {
            var dl = new PgProjectDownloader();
            dl.createAndDownloadZIP(done)
        }
    }

    exploreFiles() {
        pinegrow.showUIPanel('project', false, false, true);
    }

    duplicate() {
        var p = this.project;
        if(pinegrow.backend.canCreateProject()) {
            pinegrow.showPrompt(`<p>Enter the name of the new project:</p>`, 'Duplicate project', p.name, '', null, function (value) {
                var url = 'project_' + pgMakeSlug(value);
                p.name = value;
                pinegrow.backend.saveAsNewProject(p, url, function () {
                    pinegrow.showQuickMessage('The project was duplicated. Open it from the Start screen.')
                })
            })
        }
    }

    delete(done) {
        var p = this.project;
        pinegrow.showAlert(`<p>Are you sure you want to delete the project? You won't be able to get it back!</p>`, 'Delete project?', 'Cancel', 'Yes, delete it!', null, function() {
            pinegrow.backend.deleteProject(p, function() {
                if(!pinegrow.getCurrentProject()) {
                    done && done();
                } else {
                    pinegrow.closeCurrentProject(null, true, function () {
                        done && done();
                    });
                }
            })
        })
    }

    getShareMenuActions() {
        var p = this.project;

        var access = p.access_view;
        var clone = p.access_clone;

        var setAccess = function(a) {
            p.access_view = a;
            access = a;
            pinegrow.backend.saveDelayed();
        }

        var setClone = function(c) {
            p.access_clone = clone = c;
            pinegrow.backend.saveDelayed();
        }

        var a = []

        a.push({
            type: 'header',
            label: 'Get the link'
        })

        var preview_url = pinegrow.browserPreview.getPublicUrlForProject();

        a.push({
            link_url: preview_url,
            label: `Open in browser`,
            dont_handle_click: true
        })

        a.push({
            label: 'Copy the link to clipboard',
            action: function() {
                var $input = $(`<input type="text" value="${preview_url}" style="position:absolute;top:-100px;left:0;">`).appendTo($('body'));
                $input.get(0).select();
                document.execCommand("copy");
                pinegrow.showQuickMessage('Link was copied to the clipboard.');
            }
        })

        a.push({
            type: 'divider'
        })

        a.push({
            type: 'header',
            label: pgf.featured_projects === 'global' ? 'Who can edit this project?' : 'Who can view the project?'
        })

        a.push({
            label: 'Just me',
            action: function() {
                setAccess('private');
            },
            check: access === 'private'
        })

        if(pgf.kids) {
            a.push({
                label: 'Everybody in my group',
                action: function () {
                    setAccess('group');
                },
                check: access === 'group'
            })
        }

        if(pinegrow.currentUser.canPublishPublicProjects()) {

            a.push({
                label: pgf.featured_projects === 'global' ? 'Other users' : 'Everybody',
                action: function () {
                    setAccess('public');
                },
                check: access === 'public'
            })
        }

        if(pgf.use_nice_project_menu) {

            a.push({
                type: 'divider'
            })

            a.push({
                type: 'header',
                label: 'Can people copy the project?'
            })

            a.push({
                label: 'Allow copying',
                action: function () {
                    setClone(clone === 'clone' ? '' : 'clone')
                },
                check: clone === 'clone'
            })
        }
        return a;
    }

    getRevisionsMenuActions() {
        var _this = this;
        var p = this.project;

        var a = [];

        if(!p.revisions || p.revisions.length === 0) {
            a.push({
                type: 'header',
                label: 'The project has no revisions.'
            })
            return a;
        }

        a.push({
            type: 'header',
            label: 'Project revisions<br>Click on the revision to preview it.'
        })

        var previewRevision = function() {
            var revision_id = this.key;
            var p = _this.project;

            if((revision_id === 0 && !p.loaded_revision_info) || (p.loaded_revision_info && p.loaded_revision_info.id == revision_id)) {
                pinegrow.showQuickError('This revision is already loaded.')
                return;
            }
            _this.close(function() {
                pinegrow.tutorialsManager.openProject(p.url, null, p.isLoadedFromAnotherUser, null, revision_id)
            })
        }

        a.push({
            type: 'divider'
        })

        a.push({
            label: 'Current',
            key: 0,
            action: previewRevision,
            check: !p.loaded_revision_info,
            submenu: [
                {
                    label: 'Open',
                    key: 0,
                    action: previewRevision
                },
                {
                    label: 'Open in new window',
                    key: 0,
                    dont_handle_click: true,
                    link_url: pinegrow.backend.getProjectEditUrl(p)
                },
            ]
        })

        a.push({
            type: 'divider'
        })

        p.revisions.forEach(function(rev) {
            a.push({
                label: `${rev.user} at ${new Date(rev.time * 1000).toLocaleString()}`,
                key: rev.id,
                action: previewRevision,
                check: p.loaded_revision_info && p.loaded_revision_info.id == rev.id,
                submenu: [
                    {
                        label: 'Open',
                        key: rev.id,
                        action: previewRevision
                    },
                    {
                        label: 'Open in new window',
                        key: rev.id,
                        dont_handle_click: true,
                        link_url: pinegrow.backend.getProjectEditUrl(p, rev.id)
                    },
                    {
                        label: 'Delete...',
                        action: function() {
                            pinegrow.showAlert(`<p>Are you sure you want to delete the revision created by <code>${rev.user}</code>on <code>${new Date(rev.time * 1000).toLocaleString()}</code>?</p>`, 'Delete the revision?', 'Cancel', 'Yes, delete the revision', null, function() {
                                pinegrow.backend.deleteProjectRevision(p, rev.id);
                            })
                        }
                    }
                ]
            })
        })

        a.push({
            type: 'divider'
        })

        a.push({
            label: 'Delete all revisions...',
            action: function() {
                pinegrow.showAlert(`<p>Are you sure you want to delete all saved project revisions?</p>`, 'Delete the revisions?', 'Cancel', 'Yes, delete all revisions', null, function() {
                    pinegrow.backend.deleteAllProjectRevisions(p);
                })
            }
        })

        return a;
    }

    createNewPage(html, suggested_name, dir, just_file) {
        var project = this.project;
        dir = dir || '';

        if(dir && !dir.endsWith('/')) dir += '/';

        var base_rel_url = html ? dir + 'index.html' : 'index.html';

        var getName = function(v) {
            if(!just_file && !v.endsWith('.html')) v += '.html';
            return v;
        }

        var what = just_file ? 'file' : 'page';

        pinegrow.showPrompt(`<p>Enter the name of the new ${what}.</p> <p>Include a folder name (for example, <code>folder/mypage.html</code>) if you wish to create the ${what} in a new folder.</p>`, 'Give a name to the new ' + what, suggested_name, 'mypage.html', null, function(value) {
            var rel_url = dir + getName(value);

            var new_url = project.getAbsoluteUrl(rel_url);

            if(!html && !just_file) {
                html = project.getContentOfUrl('_new.html', function(code) {
                    html = code;
                })

                if(!html) {
                    var pi = project.getProjectInfo();
                    var source_project_url = project.source_project_id || pi.getSetting('source-project-url');

                    if(source_project_url) {
                        var start_project = pinegrow.tutorialsManager.getLessonByKey(source_project_url.replace('https://tutorial_', ''));
                        if(start_project) {
                            html = start_project.getHiddenFileContent('_new.html');
                        }
                    }
                }
            }

            if(!just_file) {
                html = html || `<!doctype html><html><head><link href="style.css" rel="stylesheet"></head><body></body></html>`;

                var p = new pgParser();
                p.assignIds = false;
                p.parse(html);
                p.rootNode.mapLinksToNewLocation(project.getAbsoluteUrl(base_rel_url), project.getAbsoluteUrl(rel_url), false, true);
                html = p.rootNode.toStringOriginal(true);
            }

            project.setContentOfUrl(rel_url, html, function() {
                if(!just_file) {
                    pinegrow.openOrShowPage(new_url, function (page) {
                        page.setPageChanged(true);
                        project.pageWasOpenedOrClosed(page, true);
                        pinegrow.backend.saveDelayed();
                        pinegrow.refreshCurrentProjectDelayed(true);
                    }, true);
                } else {
                    pinegrow.backend.saveDelayed();
                    pinegrow.refreshCurrentProjectDelayed(true);
                }
            })
        }, null, null, null, function(v) {
            if(!v.match(/^[a-z0-9\.\-\_\/]+(\.html|\.php)?$/i)) {
                return `The page name should contain only letters, numbers and characters <code>-</code>, <code>_</code> and <code>.</code>.${just_file ? '' : ' The name should end with <code>.html</code> or <code>.php</code>.'}`;
            } else if(v.match(/^[0-9]+(\.html|\.php)?$/)) {
                return `The page name should not be a number. Use <code>page${v}</code> instead.`;
            }
            v = getName(v);
            if(project.getFileForUrl(project.getAbsoluteUrl(v))) {
                return `A page with the name <code>${v}</code> already exists.`;
            }
            return null;
        })
    }

    renameFile(url) {
        var p = this.project;
        var file = p.getFileForUrl(url);

        var type = file.isFile ? 'file' : 'folder';

        pinegrow.showPrompt(`Enter the new ${type} name:`, 'Rename the ' + type + ' ' + file.name, file.name, `New ${type} name`, null, function (name) {
            p.renameProjectFileUrl(url, name);
            pinegrow.refreshCurrentProjectDelayed(true);

            pinegrow.backend.saveCurrentProject(function() {
                pinegrow.showQuickMessage('Done.')
            });

            pinegrow.showQuickMessage('Renaming...')
        })


    }

    restartLesson() {
        var project = this.project;
        var lesson = this.project.lesson;
        var tutorial = lesson.tutorial;
        //get all lesson parts
        var parts = tutorial.getAllLessonParts(project.lesson);
        var lesson_idx = parts.indexOf(project.lesson);

        var restartParts = function() {
            pinegrow.backend.cancelSaveDelayed();
            pinegrow.showQuickMessage('Restarting... Please wait.');
            pinegrow.backend.deleteProjects(parts.map(function (lesson) {
                return lesson.getProjectId();
            }), function () {

                pinegrow.backend.setLessonStatusMultiple(parts, 'none', function () {
                    parts.forEach(function (lesson) {
                        lesson.setStatus('none');
                    })
                    tutorial.openLesson(parts[0].key);
                })
            })
        }

        if(parts.length > 1) {
            var main_lesson = parts[0];
            var main_name = main_lesson.main_lesson_project_name || main_lesson.name.split(' - ')[0];
            //this is a part of the multipart lesson
            //previos lessons are saved before
            //two choices: restart just this part, or all

            if(lesson_idx === 0) {
                pinegrow.showAlert(`<p>This ${pgf.lesson_name} is the first part of the project <code>${main_name}</code>.</p>
<p>Restarting the first ${pgf.lesson_name} will restart all ${pgf.lesson_name}s in the project and erase what you did so far.</p>
<p>What would you like to do?</p>`, `Do you want to restart?`, `No, go back`, null, null, null, 'Restart the whole project', function () {

                    restartParts();
                })

            } else {
                pinegrow.showAlert(`<p>This ${pgf.lesson_name} is a part of the project <code>${main_name}</code>.</p>
<p>You can:</p>
<ul>
<li>Restart the project from this ${pgf.lesson_name} onward.</li>
<li>Restart the whole project. You will go back to the first ${pgf.lesson_name} in the project.</li>
</ul>
<p>Restarting the ${pgf.lesson_name} will erase what you did so far. What would you like to do?</p>`, `Do you want to restart?`, `No, go back`, `Restart from this ${pgf.lesson_name} onward`, null, function () {
                    parts.splice(0, lesson_idx);
                    restartParts();

                }, 'Restart the whole project', function () {
                    restartParts();
                })
            }

        } else {

            pinegrow.showAlert(`<p>Restarting the ${pgf.lesson_name} will erase what you did so far. Do you want to restart?</p>`, `Do you want to restart?`, `No, let's go back`, `Yes, restart the ${pgf.lesson_name}`, null, function () {

                restartParts();
                /*
                pinegrow.backend.cancelSaveDelayed();
                pinegrow.showQuickMessage('Restarting... Please wait.');
                pinegrow.backend.deleteProject(project, function () {
                    pinegrow.backend.setLessonStatus(project.lesson, 'none', function() {
                        project.lesson.setStatus('none');
                        tutorial.openLesson(project.lesson.key);
                    })
                });

                 */
            })
        }
    }
}

var PgQuickNiceProjectView = function($target) {

    if(!pgf.use_nice_project_menu) {
        var projectMenuHelper = new PgProjectMenuHelper();

        var actions = []

        actions.push({
            label: 'Close project',
            action: function() {
                projectMenuHelper.close();
            }
        })

        actions.push({
            type: 'divider'
        })

        actions.push({
            label: 'Explore files',
            action: function() {
                projectMenuHelper.exploreFiles();
            }
        })

        actions.push({
            type: 'divider'
        })

        actions.push({
            label: 'Revisions',
            submenu: projectMenuHelper.getRevisionsMenuActions()
        })

        actions.push({
            type: 'divider'
        })

        if(projectMenuHelper.project.isTutorialLesson()) {
            actions.push({
                type: 'header',
                label: 'Tutorial lesson'
            })

            actions.push({
                label: 'Save as a normal project...',
                action: function() {
                    projectMenuHelper.duplicate();
                }
            })

            actions.push({
                label: 'Restart the lesson...',
                action: function() {
                    projectMenuHelper.restartLesson();
                }
            })
        } else {
            actions.push({
                label: 'Rename...',
                action: function() {
                    projectMenuHelper.askToRename();
                }
            })

            actions.push({
                label: 'Duplicate...',
                action: function() {
                    projectMenuHelper.duplicate();
                }
            })
        }



        actions.push({
            type: 'divider'
        })

        actions.push({
            label: 'Download',
            action: function() {
                projectMenuHelper.download();
            }
        })

        actions.push({
            label: 'Share',
            submenu: projectMenuHelper.getShareMenuActions()
        })

        actions.push({
            type: 'divider',
            header: 'Danger zone'
        })

        actions.push({
            type: 'header',
            label: 'Danger zone'
        })

        actions.push({
            label: 'Delete...',
            action: function() {
                projectMenuHelper.delete();
            }
        })

        /*
        if(pgIsDev()) {
            actions.push({
                type: 'divider',
                header: 'Starter templates'
            })

        }

         */

        var m = new CrsaContextMenu(actions, $target);
        m.showForElement($target, null, 8);
        return;
    }

    var pv = new PgNiceProjectView();

    var win = new PgQuickWindow('Project', pv, 'quickProject', 500, 500);

    this.win = win;
    pv.win = win;

    win.showFor$El($target, false, 0, 'down');

    pv.show(pinegrow.getCurrentProject());
    pinegrow.focusRect.coverAll();

    var onFocusRectClosed = function() {
        win.close();
    }

    pinegrow.addEventHandler('on_focus_rect_closed', onFocusRectClosed);

    win.onDestroy = function() {
        pinegrow.focusRect.hide();
        pinegrow.removeEventHandler('on_focus_rect_closed', onFocusRectClosed);
    }

}

var PgNiceProjectView = function() {

    var projectMenuHelper = new PgProjectMenuHelper();

    var $element = $(`<div class="pg-nice-project-view">
<div class="pg-nice-project-commands"></div>
<div class="pg-nice-project-name rename"><span></span><i class="icon icon-Editable" style="margin-left:5px;"></i></div>
<div class="pg-nice-project-pages"></div>
</div>`);

    var $name = $element.find('.pg-nice-project-name span');
    var $pages = $element.find('.pg-nice-project-pages');
    var $commands = $element.find('.pg-nice-project-commands');

    var askToRename = function(done, text, title) {
        projectMenuHelper.askToRename(function() {
            _this.show(project);
        }, text, title)
    }
    $element.find('.rename').on('click', function(e) {
        e.preventDefault();
        askToRename();
    })

    addTooltip($name.parent(), 'Click to rename the project.')

    var view = new PgUIView($element, 'nice_project', 'Project');

    this.view = view;

    var _this = this;

    var project = null;

    var done = function() {
        _this.win.close();
    }

    this.show = function(p) {

        var options = {
            title: 'The Project menu',
            pages: true,
            buttons: {
                close: {
                    label: 'Close project',
                    classes: '',
                    tooltip: 'Close the project and go back to the start screen. Your project is saved automatically.'
                },
                extras: [
                    {
                        label: 'Share',
                        tooltip: 'Decide who can view and clone the project.',
                        action: function($b, close) {
                            var a = projectMenuHelper.getShareMenuActions();

                            var m = new CrsaContextMenu(a, $b);
                            m.showForElement($b);
                            return true;
                        }
                    },
                    {
                        label: 'More',
                        tooltip: 'Download, duplicate, delete and other project commands.',
                        action: function($b, close) {
                            var a = []

                            a.push({
                                label: 'Download...',
                                action: function() {
                                    projectMenuHelper.download();
                                }
                            })

                            a.push({
                                type: 'divider'
                            })

                            a.push({
                                label: 'Explore files...',
                                action: function() {
                                    done();
                                    projectMenuHelper.exploreFiles()
                                }
                            })

                            var supported_tags = [
                                'bootstrap 5',
                                'tailwind css',
                                'plain html',
                                'interactions',
                                'tutorial'
                            ]

                            a.push({
                                label: 'Tags',
                                submenu: supported_tags.map(function(tag) {
                                    return {
                                        label: tag,
                                        check: p.project_tags.indexOf(tag) >= 0,
                                        action: function() {
                                            var idx = p.project_tags.indexOf(tag);
                                            if(idx < 0) {
                                                p.project_tags.push(tag);
                                            } else {
                                                p.project_tags.splice(idx, 1);
                                            }
                                            p.project_tags_changed = true;
                                            pinegrow.backend.saveDelayed();
                                        }
                                    }
                                })
                            })

                            a.push({
                                type: 'divider'
                            })


                            a.push({
                                label: 'Duplicate project...',
                                action: function() {
                                    projectMenuHelper.duplicate();
                                }
                            })

                            a.push({
                                type: 'divider'
                            })

                            a.push({
                                type: 'header',
                                label: 'Danger zone'
                            })

                            a.push({
                                label: 'Delete project...',
                                action: function() {
                                    projectMenuHelper.delete()
                                }
                            })

                            var m = new CrsaContextMenu(a, $b);
                            m.showForElement($b);
                            return true;
                        }
                    }
                ]
            }
        }
        project = p;

        //Commands
        var html = `<`;

        $commands.html('');

        pinegrow.dispatchEvent('on_show_nice_project_view', null, p, options, done);

        //Home
        var $home = $(`<button class="pg-button ${options.buttons.close.classes || ''}">${options.buttons.close.label}</button>`).appendTo($commands);
        options.buttons.close.tooltip && addTooltip($home, options.buttons.close.tooltip);

        $home.on('click', function(e) {
            e.preventDefault();
            projectMenuHelper.close(done, `<p>Why not give your project a proper name?</p>`, `Name the project`)
        })

        options.buttons.extras.forEach(function(b) {
            var $b = $(`<button class="pg-button ${b.classes || ''}">${b.label}</button>`).appendTo($commands);
            b.tooltip && addTooltip($b, b.tooltip);

            $b.on('click', function(e) {
                e.preventDefault();
                if(b.action($b, done) === true) {
                    //handler will call done
                } else {
                    done();
                }
                return false;
            })
        })
/*
        var $share = $(`<button class="pg-button">Share</button>`).appendTo($commands);
        addTooltip($share, 'Show your project to the Universe.');

        var $download = $(`<button class="pg-button">Download</button>`).appendTo($commands);
        addTooltip($download, 'Download the project files to your computer.');
*/

        if(options.pages) {
            //Name
            $name.html(p.name || p.url);
            pg$Show($name.parent());

            //Add page
            var html = `<div class="pg-nice-project-new-page" data-page=""><i class="icon icon-plus"></i>Add new page</div>`;

            //Pages
            var files = p.getEditableFiles({}, p);

            files = files.filter(function(f) {
                var name = crsaGetNameFromUrl(f.url);
                return !name.startsWith('_') || pinegrow.currentUser.isAdmin();
            })

            files.forEach(function (page) {
                //var thumb = p.getThumbnailForUrl(page.url);
                html += `<div data-page="${page.url}">`;
                if(page.thumb) {
                    html += `<div class="image-wrapper-4x3"><img src="${page.thumb}"></div>`;
                }
                html += `<span>${page.name}</span><div class="commands">
<i class="icon icon-duplicate duplicate" title="Duplicate the page."></i>
<i class="icon icon-bin delete" title="Delete the page."></i>
</div></div>`;
            })

            $pages.html(html);

            addTooltip($pages.find('.commands i'), '');
            pg$Show($pages, 'grid');
        } else {
            pg$Hide($pages);
            pg$Hide($name.parent());
        }

        this.win.setTitle(options.title);
    }

    $pages.on('click', 'div.commands i', function(e) {
        e.preventDefault();
        e.stopPropagation();
        e.stopImmediatePropagation();

        var $i = $(this);

        var url = $i.closest('[data-page]').attr('data-page');

        var page_name = crsaGetNameFromUrl(url);
        var page = pinegrow.getCrsaPageByUrl(url);
        var file = project.getFileForUrl(url);

        if($i.hasClass('delete')) {
            pinegrow.showAlert(`<p>Do you want to delete the page <code>${page_name}</code>? You can't undo this operation.`, 'Are you sure?', 'Cancel', 'Yes, delete it!', null, function() {
                var name;

                if(page) {
                    page.close();
                    page.view.closeTab();
                }

                project.removeFileNodeFromList(file);
                //project.page_set_changed = true;

                _this.show(project);

                pinegrow.backend.deletePage(project, url, function() {
                    pinegrow.showQuickMessage(`The page ${page_name} was deleted!`);
                });
            })
        } else if($i.hasClass('duplicate')) {
            var html;
            if(page) {
                html = page.getCodeForSave();
                createNewPage(html);
            } else {
                project.getContentOfUrl(url, function(html) {
                    createNewPage(html);
                })
            }
            done();
        }
    })

    $pages.on('click', 'div[data-page]', function(e) {
        var url = this.getAttribute('data-page');
        if(url) {
            pinegrow.openOrShowPage(url, function(page, open) {
                if(open) {
                    project.pageWasOpenedOrClosed(page, true);
                }
            }, true);
        } else {
            //New page
            if(project.source_project_id) {
                //get templates from the source project
                pinegrow.backend.withSourceProject(project, function(sp) {
                    if(sp) {
                        var templates = sp.getEditableFiles();

                        var html = `<div class="pg-nice-project-templates">`;
                        templates.forEach(function(f) {
                            html += `<div data-name="${f.name}"><div class="image-wrapper-4x3">`;
                            if(f.thumb) {
                                html += `<img src="${f.thumb}">`;
                            }
                            html += `</div><span>${f.name}</span></div>`;
                        })
                        html += `</div>`;

                        var $d = pinegrow.showAlert(html, 'Select a template for the new page', 'Cancel');

                        $d.find('[data-name]').on('click', function(e) {
                            var $t = $(this);
                            var name = $t.attr('data-name');
                            var html = sp.getContentOfUrl(name);
                            $d.modal('hide');
                            createNewPage(html, name);
                        })
                    } else {
                        pinegrow.showQuickError('Could not load the source project ' + project.source_project_id + '.')
                    }
                })
            } else {
                createNewPage();
            }
        }
        done();
    })

    var createNewPage = function(html, suggested_name) {
        var h = new PgProjectMenuHelper(project);
        return h.createNewPage(html, suggested_name);
    }
}

var PgProjectView = function() {

    pinegrow.projectView = this;

    var $element = $('<div class="pg-project-view crsa-panel"></div>');

    var view = new PgUIView($element, 'project', 'Project');

    this.view = view;

    var _this = this;

    view.tabIcon = 'icon-project';

    var expanded_folders = {};

    var search = new PgSearchBox({
        placeholder: 'Search project',
        placeholder_active: null,
        toolbar_on: true
    });

    var toolbar = search.getToolbar();

    if(isApp()) {
        toolbar.addSection('settings');

        var settingsButton = new PgButtonToolbarItem('icon-tools', 'Manage libraries and plugins.');
        toolbar.addItem('settings', settingsButton);
        settingsButton.onClick = function () {
            $.fn.crsacss('showFrameworkManager', pinegrow.getSelectedPage());
        }
    }

    search.$element.appendTo($element);

    search.onSearch = function(f) {
        updateFilter(f);
    }

    var spacer = new PgSearchBoxSpacer();
    spacer.$element.appendTo($element);

    var notice = new PgPanelNotice();
    notice.$element.appendTo($element);

    notice.$element.on('click', '.open-folder', function(e) {
        e.preventDefault();
        if(canOpenProject()) {
            pinegrow.openProjectBySelectingFolder();
        }
    })

    notice.$element.on('click', '.open-current', function(e) {
        e.preventDefault();
        var page = pinegrow.getSelectedPage();
        if(page && canOpenProject()) {
            if(page.localFile) {
                var path = require('path');
                var dir = path.dirname(page.localFile)
                pinegrow.openProject(dir, page);
            } else {
                pinegrow.showQuickError('Please, save the page first.');
            }
        }
    })

    notice.$element.on('click', '.save-current', function(e) {
        e.preventDefault();
        var page = pinegrow.getSelectedPage();
        if(page) {
            $('.menu-file-save').trigger('click');
        }
    })

    var $content = $('<div class="pg-project-view-content crsa-project-list"></div>').appendTo($element);

    var project = null;

    var getFileForPath = function(path) {
        var a = path.split(',');
        var n = project.root;
        for(var i = 1; i < a.length; i++) {
            var idx = parseInt(a[i]);
            if(n.children.length <= idx) return null;
            n = n.children[idx];
        }
        return n;
    }

    var updateFilter = function(filter) {
        var re = filter.length > 0 ? new RegExp(escapeRegExp(filter), 'i') : null;
        var $all = $content.find('li.project-item');
        if(re) {
            $all.removeClass('has-found');
            $all.each(function(i, a) {
                var $li = $(a);
                var name = $li.attr('data-file-name');
                var found = re && name.match(re);
                if(found) {
                    $li.addClass('found').removeClass('not-found');

                    var $pli = $li.parent().closest('li.project-item');
                    while($pli.length) {
                        //if closed, open
                        if($pli.hasClass('folder-closed')) {
                            $pli.removeClass('folder-closed');
                            $pli.find('> .folder-icon').removeClass('icon-right').addClass('icon-down');
                            $pli.addClass('search-expanded');
                            //expanded_folders[file.url] = true;
                        }

                        if($pli.hasClass('has-found') && !$pli.hasClass('folder-closed')) {
                            break;
                        }
                        $pli.addClass('has-found');
                        $pli = $pli.parent().closest('li.project-item');
                    }

                } else {
                    $li.removeClass('found').addClass('not-found');
                }
            });
        }

        if(!re) {
            $all.removeClass('found not-found has-found');
        /*    $all.each(function(li) {
                var $li = $(li);
                var n = getFileForPath($li.attr('data-path'));
                if(!expanded_folders[n.url]) {
                    $li.addClass('folder-closed');
                    $li.find('> .folder-icon').removeClass('icon-down').addClass('icon-right');
                }
            })*/
        }

        $content.find('li.search-expanded:not(.found,.has-found)').each(function(i, e) {
            var $li = $(e);
            $li.addClass('folder-closed');
            $li.removeClass('search-expanded');
            $li.find('> .folder-icon').removeClass('icon-down').addClass('icon-right');
        })
    }

    var updateNotice = function() {
        var page = pinegrow.getSelectedPage();
        var html = '<p>Open a folder as Project to show it here.</p>';
        if(page) {
            if(page.localFile) {
                html += '<p class="buttons"><button class="pg-button open-current">Open ' + page.name + ' as project</button></p>';
            } else {
                html = '<p>Please save the page first, then open its folder as project.</p><p class="buttons"><button class="pg-button save-current">Save the page</button></p>';
            }
        } else {
            html += '<p class="buttons"><button class="pg-button open-folder">Open a folder</button></p>';
        }
        notice.show(html)
    }

    var $urls = null;

    var updateRemoteUrls = function() {

        return;

        if(!$urls) return;

        var url_list = project.getRemoteUrls();

        if(url_list.length === 0) {
            $urls.hide();
        } else {
            $urls.show();
        }

        var url_html = '';

        url_list.forEach(function(url) {
            url_html += `<div data-url="${url}" title="${url}"><i class="icon icon-minus"></i>${url}</div>`;
        })

        $urls.find('.pg-project-urls-list').html(url_html);

        $urls.on('click', '[data-url]', function(e) {
            e.preventDefault();
            var $e = $(this);
            var url = $e.attr('data-url');
            pinegrow.openOrShowPage(url, function(page, opened) {
                if(opened) {
                    project.addRemoteUrl(url);
                }
            });
        })

        $urls.on('contextmenu', '[data-url]', function(e) {
            var $e = $(this);

            var url = $e.attr('data-url');

            var a = [
                {
                    label: 'Open URL',
                    action: function() {
                        pinegrow.openOrShowPage(url, function(page, opened) {
                            if(opened) {
                                project.addRemoteUrl(url);
                            }
                        });
                    }
                },
                {
                    type: 'divider'
                },
                {
                    label: 'Delete from the list',
                    action: function() {
                        project.removeRemoteUrl(url);
                    }
                }
            ]

            var m = new CrsaContextMenu(a, $e);
            m.showForElement($e);

            e.preventDefault();

            return false;
        })

        $urls.find('.help').on('click', function(e) {
            e.preventDefault();
            pinegrow.openExternalUrl(pgf.docs.remote_url_help);
        })
    }

    this.show = function(p) {
        view.whenVisible(function() {

            project = p;

            if(!p) {
                updateNotice();
                $content.hide();
                return;
            } else {
                notice.show();
                $content.show();
            }

            //expanded_folders = {};

            project.sortItems();

            var $lib = $content;
            var $container = $lib;

            $currentContainer = $lib;

            var filter = null;
            var $header = $lib.find('>.header').html('');
            var $input = $('<input/>', {class: 'form-control filter-form', placeholder: 'search'}).appendTo($header);
            crsaAddCancelSearch($input, 'top: 16px;right: 46px;');
            /* var $cancel = $('<a/>', {class: 'btn btn-link', href: '#'}).html('&times;').appendTo($lib).on('click', function(e) {
             e.preventDefault();
             $input.val('').trigger('input');
             })*/
            var selectedCrsaPage;

            $content.html('');

            var $list = $('<ul/>').appendTo($content);

            var currentPage = null;

            var pageChanged = function(crsaPage) {
                if(currentPage) {
                    var $li = _this.find$liForUrl(currentPage.url);
                    if($li) {
                        $li.removeClass('selected-page');
                    }
                }
                currentPage = crsaPage;
                if(crsaPage) {
                    var $li = _this.find$liForUrl(crsaPage.url);
                    if($li) {
                        $li.addClass('selected-page');
                    }
                }
            }

            view.on('crsa-page-selected.project', function(e, crsaPage) {
                pageChanged(crsaPage);
            });

            var updateList = function(sections) {
                var handlers = {
                    on_project_menu : [],
                    on_project_item_get_tags : [],
                    on_project_item_get_context_menu : [],
                    on_project_item_preview : []
                };

                $.each(pinegrow.getFrameworks(), function(key, f) {
                    $.each(handlers, function(handler, list) {
                        if(handler in f && f[handler]) {
                            list.push(f[handler]);
                        }
                    });
                });

                var callHandler = function(handler, a, b, c) {
                    var cp = pinegrow.getSelectedPage();
                    var r = [];
                    for(var i = 0; i < handlers[handler].length; i++) {
                        var ret = handlers[handler][i](cp, a, b, c);
                        if(ret) r.push(ret);
                    }
                    return r;
                }

                var $scroll = $content;

                var showMenu = function(menu_items_sets, e, $target) {
                    var $menu = new CrsaContextMenu(null, $target);

                    for(var i = 0; i < menu_items_sets.length; i++) {
                        for(var m = 0; m < menu_items_sets[i].length; m++) {
                            if(menu_items_sets[i][m].divider) {
                                $menu.add("", null, null, 'divider');
                                if(menu_items_sets[i][m].header) {
                                    $menu.add(menu_items_sets[i][m].header, null, null, 'header');
                                }
                            } else {
                                var item = $menu.add(menu_items_sets[i][m].label, menu_items_sets[i][m].key || null, menu_items_sets[i][m].func);
                                item.submenu = menu_items_sets[i][m].submenu || null;
                                item.check = menu_items_sets[i][m].check || null;
                            }
                        }
                    }
                    $menu.showAt(e.pageX, e.pageY, $scroll);
                    $menu.updatePosition();
                }

                var selectedCrsaPage = currentPage;
                $list.html('');

                $content.find('>h2').remove();
                var $title = $('<h2 class="project-name" style="display:inline-block;"></h2>').prependTo($content);

                $title.html(project.name + '&nbsp;<i class="icon icon-down" style="position:relative;top:3px;"></i>');

                $title.on('click contextmenu', function(e) {
                    e.preventDefault();
                    e.stopPropagation();
                    var menu_items_sets = callHandler( 'on_project_menu', project);
                    if(menu_items_sets.length) {
                        showMenu(menu_items_sets, e, $(this));
                    }
                })

                if(isApp()) {
                    addTooltip($title, `<code>${project.getDir()}</code><br>Click for project options.`);
                } else {
                    addTooltip($title, `Click for project options.`);
                }

                $urls = $(`<div class="pg-project-urls"><h4>Remote URLs <i class="help icon icon-271" title="Learn about editing remote projects..."></i></h4><div class="pg-project-urls-list"></div><h4 style="margin-top:20px;">Local files</h4></div>`).appendTo($content);

                updateRemoteUrls();


                filter = $input.val();
                var filterRegEx = filter && filter.length > 0 ? new RegExp(escapeRegExp(filter),'i') : null;

                var html = '';

                var draggable = ['image', 'css', 'script'];

                var dont_show = {}

                if(!isApp()) {
                    dont_show['_pginfo'] = true;
                    dont_show['pinegrow.json'] = true;
                    dont_show['projectdb.pgml'] = true;
                }
                var showLevel = function(node, path) {
                    for(var i = 0; i < node.children.length; i++) {
                        var n = node.children[i];
                        if(dont_show[n.name]) continue;

                        var item_path = path + ',' + i;
                        var icon = '';
                        var icon_after = '';
                        var classes = '';
                        if(!n.isFile) {
                            if (expanded_folders[n.url]) {
                                icon = '<i class="folder-icon icon icon-down"></i>';
                            }
                            else {
                                icon = '<i class="folder-icon icon icon-right"></i>';
                                classes = "folder-closed";
                            }

                        }
                        var name_class = (n.isEditable) ? 'editable' : '';

                        var type = n.getSimpleType();

                        if(draggable.indexOf(type) >= 0 && pgf.edit_html) {
                            if(type !== 'css' || (pinegrow.getMimeType(n.url) === "text/css")) {
                                classes += ' draggable';
                                //icon_after = '<i class="after icon icon-add" title="Drag the item to the page or to the tree."></i>'
                                n.addTag({
                                    icon: 'Pin6',
                                    desc: 'You can drag the item to the page or to the tree.',
                                    text: 'draggable'
                                })
                            }
                        }

                        html += '<li draggable class="project-item ' + (n.isFile ? "file" : "folder ") + classes + '" data-path="' + item_path + '" data-file-name="' + n.name + '" data-url="' + n.url + '">' + icon + '<name class="' + name_class + '">' + n.name + icon_after + '<tags>' + makeTagsHtml(n) + '</tags></name>';

                        if(n.children.length) {
                            html += '<ul>';
                            showLevel(n, item_path);
                            html += '</ul>';
                        }
                        html += '</li>';
                    }
                }

                showLevel(project.root, '');

                $list.html(html);

                $list.on('contextmenu.project', 'li.project-item name, li.project-item i.folder-icon', function(e) {
                        e.stopPropagation();
                        e.preventDefault();

                        var $el = $(this);
                        var $li = $el.closest('li');
                        var file = getFileForPath($li.attr('data-path'));

                        var menu_items_sets = callHandler('on_project_item_get_context_menu', file, project);

                        pinegrow.dispatchEvent('on_project_item_get_context_menu_regular', null, file, project, menu_items_sets);

                        if(menu_items_sets.length) {
                            showMenu(menu_items_sets, e, $(this));
                        }
                        return false;
                    })
                $list.on('click.project', 'li.project-item name, li.project-item i.folder-icon', function(e) {
                        var $el = $(this);
                        e.preventDefault();
                        //$el.closest('li').toggleClass('crsa-action-closed');
                        var $li = $el.closest('li');
                        var file = getFileForPath($li.attr('data-path'));
                        //console.log(file);
                        if(file) {
                            if(file.isFile) {
                                if(file.isEditable) {

                                    var pgc = pinegrow.findFrameworkByKey('pg.components');
                                    var submenu = null;

                                    if(pgc) {
                                        submenu = pgc.pgcGetPartialOpeningOptions(file.url, null, file.name);
                                    }
                                    if(submenu) {

                                        submenu.unshift({
                                            type: 'header',
                                            label: 'Open'
                                        })

                                        var menu = new CrsaContextMenu(submenu, $li);
                                        menu.showForElement($li);
                                        return false;

                                    } else {
                                        var partial_info = pgGetOpenPartialInContainerInfo(file.path);
                                        if (false && partial_info && partial_info.auto_open) {
                                            new pgOpenPartialInContainer(file.url);
                                            pinegrow.stats.using('prj.openpartial');
                                        } else {
                                            var show_one = !e.ctrlKey && !e.metaKey;
                                            if (show_one) pinegrow.pageTabs.hideAll();

                                            var mapped_url = pinegrow.getRemoteUrlFromMappedUrl(file.url);
                                            if (mapped_url) {
                                                pinegrow.openOrShowPage(mapped_url);
                                                pinegrow.showQuickMessage('Opening the saved remote file as URL...');
                                            } else {

                                                pinegrow.openOrShowPage(file.url, function (cp) {
                                                }, show_one, true /* select openned page */);
                                            }
                                            pinegrow.stats.using('prj.openfile');
                                        }
                                    }
                                } else {

                                    if(pinegrow.openFileInCodeEditor(file.path)) {
                                        if(file.getSimpleType() != 'css') {
                                            pinegrow.showQuickMessage('This file can be only opened in code editor.');
                                        }
                                    }
                                    //pinegrow.stats.using('prj.editcode');
                                }
                            } else if(!file.isFile) {
                                var $li = $el.closest('li');
                                var $icon = $li.find('> .folder-icon');
                                if($li.hasClass('folder-closed')) {
                                    $li.removeClass('folder-closed');
                                    $icon.removeClass('icon-right').addClass('icon-down');
                                    expanded_folders[file.url] = true;
                                } else {
                                    $li.addClass('folder-closed');
                                    $icon.removeClass('icon-down').addClass('icon-right');
                                    expanded_folders[file.url] = false;
                                }
                            }
                        }
                    })
                $list.on('mouseenter.project', 'li.project-item name, li.project-item i.folder-icon', function(e) {
                        var $el = $(this);
                        e.preventDefault();
                        var $li = $el.closest('li');
                        var file = getFileForPath($li.attr('data-path'));

                        var p = callHandler('on_project_item_preview', file);

                        if(p.length) {
                            var $pel = $(p[0].code);
                            if($pel.length && $li.get(0).ownerDocument.defaultView === window) {
                                pgPreview.show($li, $pel, $pel.is('img') ? 'with-image' : null, null);
                                if(p[0].onLoaded) {
                                    $pel.one('load', function(e) {
                                        var html = p[0].onLoaded(this);
                                        pgPreview.setDescription(html);
                                    })
                                }
                            }
                        }
                    })
                $list.on('mouseleave.project', 'li.project-item name, li.project-item i.folder-icon', function(e) {
                    pgPreview.hide();
                    });

            }

            updateList();
        })
    }

    this.show();

    this.close = function($lib) {
        this.show(null);
    }

    this.find$liForUrl = function(url) {
        //var $li = $content.find('li[data-url="' + url + '"]');
        var li = $content[0].querySelector('li[data-url="' + url + '"]');
        return li ? $(li) : null;
        //return $li.length ? $li : null;
    }

    var makeTagsHtml = function(file) {
        var tags = '';
        for(var i = 0; i < file.tags.length; i++) {
            if('show' in file.tags[i] && !file.tags[i].show) continue;

            var c = 'tag-type-' + (file.tags[i].type || 'info');
            if(file.tags[i].class) {
                c += ' ' + file.tags[i].class;
            }
            var text = file.tags[i].icon ? '<i class="icon-' + file.tags[i].icon + '"></i>' : file.tags[i].text;
            tags += '<tag class="' + c + '"' + (file.tags[i].desc ? ' title="' + file.tags[i].desc + '"' : '') + '>' + text + '</tag>';
        }
        return tags;
    }

    this.updateTagsForFile = function(file, $li) {
        //var html = makeTagsHtml(file);
        //if(html === file.tags_html) return;
        if(!$li) $li = this.find$liForUrl(file.url);
        if($li) {
            $li.find('>name tags').html(makeTagsHtml(file));
        }
    }

    var dragMenu = null;
    var dragHelper = new PgDragHelper($content, '.project-item.file', null, true);

    dragHelper.onBefore = function($el) {

        if(!pgf.edit_html) return false;

        var $li = $el.closest('li');
        var file = getFileForPath($li.attr('data-path'));

        var simpleType = file.getSimpleType();
        return ['image', 'css', 'script'].indexOf(simpleType) >= 0;
    }

    dragHelper.on( 'dragStart', function(x, y, event) {
        $.fn.crsapages('showOverlays');

        var $el = dragHelper.dragTarget;
        var $li = $el.closest('li');
        var file = getFileForPath($li.attr('data-path'));

        var simpleType = file.getSimpleType();
        var def = {name: 'Element', parent_selector: null};

        var pgel = null;
        var onInsert = null;
        var force_position_label = null;
        var force_position = null;
        var options = null;
        var onInsertedElements = null;

        switch(simpleType) {
            case 'image':
                var url = file.url;//cp.makeRelativeUrl(file.url);
                var code = '<img src="' + url + '" />';
                pgel = pgCreate(code);
                def.name = 'Image ' + file.name;

                onInsert = function() {
                    if(this.positionPgel) {
                        var page = this.positionPgel.getPage();
                        pgel.setAttribute('src', page.makeRelativeUrl(url));
                    }
                    return false;
                }
                break;
            case 'script':
            case 'css':
                var code = simpleType === 'css' ? `<link rel="stylesheet" href="${file.url}">` : `<script src="${file.url}"></script>`;
                pgel = pgCreate(code);
                def.name = file.name;

                def.parent_selector = 'html';
                force_position_label = 'Add to the page';
                force_position = 'top';

                options = {
                    disableClone: true,
                    disableGrid: true,
                    disableRepeater: true,
                    disableDisableParentSelector: true,
                    disableText: true
                }

                onInsert = function() {
                    if(this.getOver() !== 'page') return false; //let dd handle the tree

                    if(this.positionPgel) {
                        var page = this.positionPgel.getPage();

                        pinegrow.makeChanges(page, 'Add ' + file.name, function () {
                            if (simpleType === 'css') {
                                if (page.addStylesheet(file.url, true, function () {

                                    })) {
                                    pinegrow.showQuickMessage(`<b>${file.name}</b> was added to <b>${page.name}</b>.`)
                                } else {
                                    pinegrow.showQuickError(`<b>${file.name}</b> is already included on <b>${page.name}</b>.`)
                                }
                            } else if (simpleType === 'script') {
                                if (page.addScript(file.url, true)) {
                                    pinegrow.showQuickMessage(`<b>${file.name}</b> was added to the footer of <b>${page.name}</b>. Refresh the page view to re-run the Javascript code.`)
                                } else {
                                    pinegrow.showQuickError(`<b>${file.name}</b> is already included on <b>${page.name}</b>.`)
                                }
                            }
                        })
                    }
                    return true;
                }

                onInsertedElements = function(col) {
                    var page = false;
                    col.forEachElement(function(pgel) {
                        var attr = simpleType === 'css' ? 'href' : 'src';
                        page = pgel.getPage();
                        pgel.setAttribute(attr, page.makeRelativeUrl(file.url));
                    })
                    if(page && simpleType === 'css') {
                        //page.updateStylesheetsList();
                    }
                }
                break;

        }

        if (pgel) {
            dragMenu = new PgDragMenu(def, false, true, def.parent_selector, null, options);

            dragMenu.formatHtml = true;
            dragMenu.onInsert = onInsert;
            dragMenu.onInsertedElements = onInsertedElements;
            dragMenu.parentSelector = def.parent_selector;

            dragMenu.forcePositionLabel = force_position_label;
            dragMenu.forcePosition = force_position;

            dragMenu.pgel = pgel;
            dragMenu.handleMoveEvent(event, dragMenu);
        }

    }, this );

    dragHelper.on( 'drag', function(x, y, event) {
        if(dragMenu) {
            dragMenu.handleMoveEvent(event, dragMenu);
        }
    }, this );

    dragHelper.on( 'dragStop', function(x, y) {
        $.fn.crsapages('showOverlays', true);

        if (!dragMenu) return; //was already handled by the tree
        if (dragMenu.insert()) {
        } else {
        }

        dragMenu.destroy();
        dragMenu = null;

        crsaFuncs.setDraggedMenu(null);

        crsaFuncs.showInsertionLine(null);
        pinegrow.highlightElement(null);

    }, this );

    pinegrow.addEventHandler('on_page_selected', function(page) {
        if(!project) {
            updateNotice();
        }
    });

    pinegrow.addEventHandler('on_page_saved', function(page, saved_csss, first_save) {
        if(project && first_save) {
            pinegrow.refreshCurrentProject();
        }
    });

    pinegrow.addEventHandler('on_project_remote_urls_updated', function(pagenull) {
        updateRemoteUrls();
    });

}

var CrsaProject = function() {
    this.name = '';
    this.root = null;
    this.localPath = null;
    this.url = null;
    this.entityMap = {};
    this.frameworks = [];
    this.db = null;

    var pages_cache = {};

    this.path_separator = require('path').sep;

    var _this = this;
    var url_to_file_map = {};

    var projectFiles = {};

    var revision = 0;
    var saved_revision = 0;

    this.wasChanged = function() {
        revision++;
    }

    this.hasUnsavedChanges = function() {
        return revision > saved_revision;
    }

    this.getRevision = function() {
        return revision;
    }

    this.revisionWasSaved = function(r) {
        saved_revision = r;
    }

    this.isTutorialLesson = function() {
        if(this.lesson && this.lesson.tutorial && this.lesson.tutorial.startProjects) {
            return false;
        }
        return !!this.lesson;
    }

    this.getSettingsForNewFileDialog = function() {
        this.master_page = true;
        this.show_only_master_pages = true;

        var pi = this.getProjectInfo();
        var template_id = pi.getSetting('template_framework_id');
        var selectedProjectTemplate = pinegrow.projectTemplates.filter(function(temp){ return temp.framework.type == template_id; });
        if (selectedProjectTemplate.length > 0)
            selectedProjectTemplate = selectedProjectTemplate[0];
        else
            selectedProjectTemplate = pinegrow.projectTemplates[0];

        var projectTemplates = [this];

        if(!pinegrow.isContributorMode()) {
            projectTemplates = projectTemplates.concat(pinegrow.projectTemplates);
            for(var i = 0; i < projectTemplates.length; i++) {
                if(projectTemplates[i].show_only_master_pages && !$.isEmptyObject(this.framework.master_pages)) {
                    selectedProjectTemplate = projectTemplates[i];
                    break;
                }
            }
        }
        return {
            projects: projectTemplates,
            selected: selectedProjectTemplate
        }
    }

    this.createFromMasterPage = function(destUrl, cf, selectedProject, fileName) {
        var project = this;

        if (!fileName) {
            if (!selectedProject.master_page) fileName = cf.name;
        }

        pinegrow.showPrompt('Name of the new page', 'New page', fileName, 'Page name', null, function(name) {
            if (name.length) {

                var ext = crsaGetExtFromUrl(name);
                if(!ext) {
                    name = name + '.html';
                }

                var doTask = function() {
                    var file_url = crsaCombineCurrentUrlWith(destUrl, name);
                    if(!project.getFileForUrl(file_url)) {
                        try {
                            var new_file_path = crsaMakeFileFromUrl(file_url);
                            //var new_file = cf.copyFileToNewPath(new_file_path);

                            var doWithNewFile = function(new_file) {
                                var project_info = project.getProjectInfo();
                                var info = project_info.getSettingsForFile(cf.path);
                                if(info && info.frameworks) {
                                    project_info.setSettingForFile(new_file.path, 'frameworks', info.frameworks);
                                    project_info.save();
                                }

                                pinegrow.openPage(new_file.url, function (cp) {

                                    //pinegrow.fixLinks(project, cp, cf.url, new_file.url, function () {
                                        if (selectedProject == project) {
                                            cp.crsaProjectTemplate = project;//selectedProjectTemplate;
                                            var comp = pinegrow.findFrameworkByKey('pg.components');
                                            comp.useMasterPage(cp, project, cf, true /* skip refresh */);
                                            comp.updateEditableAreas(cp);
                                        }
                                        else {
                                            cp.crsaProjectTemplate = selectedProject;
                                        }

                                        cp.detectAndAddFrameworks(function() {
                                            cp.save(function () {
                                                pinegrow.refreshCurrentProject(function () {
                                                    //wait a bit, don't refresh from pageOpened done handler
                                                    setTimeout(function() {
                                                        cp.refresh();
                                                    }, 50)
                                                });
                                            }, true, false, false, true, true /* first save */);
                                        });
                                    //}, true);
                                }, true, null, {
                                    on_source_parsed: function(cp) {
                                        if(cf.remove_components) {
                                            cp.sourceNode.withoutEvents(function() {
                                                cp.sourceNode.removeAttrIfStartsWith('data-pgc', true);
                                            })
                                        }
                                        if(cf.remove_local_links) {
                                            cp.sourceNode.withoutEvents(function() {
                                                cp.sourceNode.goThroughAllUrls(function(url, attr) {
                                                    if(attr === 'href' && url.endsWith('.html')) {
                                                        if(!crsaIsAbsoluteUrl(url)) {
                                                            return '#';
                                                        }
                                                    }
                                                    return url;
                                                })
                                            })
                                        }
                                        pinegrow.copyMissingExternalFiles(cp, cf.url, new_file.url, (selectedProject.master_page && selectedProject === project) ? cf.url : project.getAbsoluteUrl('index.html') /* dummy url to indicate where the file would go if created at the top of the project */);
                                    }
                                });
                            }

                            cf.copyFileToNewPath(new_file_path, doWithNewFile);
                        }
                        catch(err) {
                            pinegrow.showAlert(err, 'Error');
                        }
                    } else if (name) {
                        pinegrow.showAlert("The page <b>" + name + "</b> already exists. Please choose a different name.", "File with Same Name Found", null, "Ok", null, function () {
                            _this.createFromMasterPage(destUrl, cf, selectedProject, name);
                        });
                    }
                }

                //if(!pinegrow.isFileEditable(name)) {

                //    doTask();
               // } else {
                    doTask();
                //}

            } else {
                pinegrow.showAlert('The page name should not be blank.', 'The page needs a name', null, "Ok", null, function () {
                    _this.createFromMasterPage(destUrl, cf, selectedProject, '');
                });
            }
        });
    }


    this.addBackgroundPage = function(page) {
        pages_cache[page.url] = page;
        page.isBackgroundPage = true;
    }

    this.getBackgroundPageForUrl = function(url) {
        return pages_cache[url] || null;
    }

    this.removeBackgroundPageForUrl = function(url) {
        if(url in pages_cache) {
            pages_cache[url].isBackgroundPage = false;
            delete pages_cache[url];
        }
    }

    this.getAllBackgroundPages = function() {
        return Object.keys(pages_cache).map(function (key) {
            return pages_cache[key];
        });
    }

    this.closeAllBackgroundPages = function() {
        var list = this.getAllBackgroundPages();
        list.forEach(function(page) {
            if(page.codeEditor) {
                page.codeEditor.closeTab();
                page.codeEditor = null;
            }
            _this.removeBackgroundPageForUrl(page.url);
        })
    }

    this.destroy = function() {
        this.closeAllBackgroundPages();
        if(this.db) {
            this.db.destroy();
        }
    }

    var updateFileNamesIndex = function (force) {
        force = force || false;
        if (force || Object.keys(projectFiles).length === 0) {
            projectFiles = {};
            _this.root.walkThroughFiles(function(node) {
                if (!projectFiles[node.name]) {
                    projectFiles[node.name] = [];
                }
                projectFiles[node.name].push(node);
            });
        }
    }

    this.getFilesWithName = function (name) {
        var files = []
        this.root.walkThroughFiles(function(node) {
            if(node.name === name) {
                files.push(node);
            }
        });
        return files;
        //better to just lookup files every time
        //return (name in projectFiles) ? projectFiles[name] : [];
    }

    this.getFile = function(path) {
        var paths = path.split('/');
        var current_node = this.root;
        for(var i = 0; i < paths.length; i++) {
            var p = paths[i];
            if(p.length === 0) continue;

            var found = false;

            for(var n = 0; n < current_node.children.length; n++) {
                var child = current_node.children[n];
                if(child.name === p) {
                    if(/* child.isFile || */ i === paths.length - 1) {
                        return child;
                    } else {
                        current_node = child;
                        found = true;
                        break;
                    }
                }
            }

            if(!found) {
                return null;
            }
        }
        return null;
    }

    this.toJSON = function() {
        var d = {};
        d.name = this.name;
        d.root = this.root;
        d.localPath = this.localPath;
        d.url = this.url;

        return JSON.stringify(d);
    }

    this.getDocuments = function() {
        var list = [];
        $.each(this.root.children, function(i,f) {
            if(f.name.match(/\.html$/i)) list.push(f);
        });
        return list;
    }

    this.fromSimpleTemplate = function(def) {
        this.url = def.url;
        this.name = def.name;

        this.description = def.description || null;
        this.author = def.author || null;
        this.author_link = def.author_link || null;
        this.info_badge = def.info_badge || null;
        this.templateBrowserIndex = def.templateBrowserIndex || null;

        if(def.frameworks) {
            this.frameworks = def.frameworks;
        }

        this.root = new CrsaFile();
        this.root.isFile = false;
        this.root.name = def.name;
        this.root.url = def.url;

        var path = crsaGetAppDir();
        this.root.path = path + crsaMakeFileFromUrl(def.url);

        var addPages = function(pages, node, url) {
            var sep = isApp() ? require('path').sep : '/';
            $.each(pages, function(i,p) {
                if(!p.children) {
                    var f = new CrsaFile();
                    f.name = p.name || p.src;
                    f.desc_name = p.desc_name || null,
                    f.isFile = true;
                    f.url = p.url || (url + f.name);
                    if(path) f.path = node.path + f.name;
                    f.tag = p.tag || 'page';
                    if(p.required) f.required = true;
                    f.image = p.image ? url + p.image : url + 'screenshots' + sep + f.name.replace('html', 'jpg');
                    f.image_direct = p.image_direct || null;
                    f.template_data = p;
                    f.video = p.video || null;
                    f.description = p.description || null;
                    f.subpages = p.subpages || null;
                    f.remove_components = p.remove_components || false;
                    f.remove_local_links = p.remove_local_links || false;
                    //node.children.push(f);
                    node.addChild(f);
                } else {
                    var d = new CrsaFile();
                    d.name = p.src;
                    d.isFile = false;
                    d.url = url + d.name;
                    if(p.tag) d.tag = p.tag;
                    if(p.required) d.required = true;
                    if(path) d.path = node.path + d.name + sep;
                    //node.children.push(d);
                    node.addChild(d);
                    addPages(p.children, d, d.url + sep);
                }
            });
        }
        addPages(def.pages, _this.root, def.url);
        updateFileNamesIndex(true);
    }

    this.openProjectBySelectingFolder = function(done) {
        crsaChooseFile(function(url, file) {
            _this.fromFolder(file, done, true);
        }, false, false, null, true);
    }

    this.reload = function(done) {
        this.fromFolder(this.getDir(), done, this.show_pg_files || false);
    }

    this.fromFolder = function(dir, done, show_pg_files, limit_files) {
        this.url = crsaMakeUrlFromFile(dir);
        this.name = this.name || decodeURIComponent(crsaGetNameFromUrl(this.url));
        this.show_pg_files = show_pg_files;

        this.root = new CrsaFile();
        this.root.isFile = false;
        this.root.name = this.name;
        this.root.url = this.url;

        this.file_count = 0;
        this.file_size_count = 0;
        this.too_many_files = false;

        var path = dir;
        this.root.path = path;

        var fs = require('fs');
        var pathmodule = require('path');
        var sep = pathmodule.sep;

        url_to_file_map = {};

        var ignore_folders = [];
        var folders = pinegrow.getSetting('ignore-folders', 'node_modules');
        try {
            ignore_folders = folders.split(',');
            for(var i = 0; i < ignore_folders.length; i++) {
                ignore_folders[i] = $.trim(ignore_folders[i]);
            }
        } catch(err) {}

        var shouldIgnoreFolder = function(folder) {
            var r = ignore_folders.indexOf(folder) >= 0;
            return r;
        }

        var walk = function(node, dir, done) {
            var results = [];

            //console.log('Reading folder ' + dir);

            fs.readdir(dir, function(err, list) {
                if (err) {
                    console.log('Error reading folder ' + dir, err);
                    return done(err);
                }
                var pending = list.length;
                if (!pending) return done(null, results);

                list.forEach(function(filename) {

                    if(filename.charAt(0) == '.' || shouldIgnoreFolder(filename) || (!show_pg_files && (filename == '_pgbackup' || filename == 'pinegrow.json' || filename == '_pgexport'))) {
                        if (!--pending) done(null, results);
                        return true;
                    }
                    var file = pathmodule.resolve(dir, filename);

                    fs.stat(file, function(err, stat) {
                        _this.file_count++;

                        if(stat && stat.size) {
                            _this.file_size_count += stat.size;
                        }
                        if(limit_files && (_this.file_count > 1000 || _this.file_size_count > 1024 * 1024 * 100)) {
                            _this.too_many_files = true;
                            if (!--pending) done(null, results);
                            return;
                        }

                        if (stat && stat.isDirectory()) {
                            var d = new CrsaFile();
                            d.name = filename;
                            d.isFile = false;
                            d.url = node.url + '/' + encodeURIComponent(d.name);
                            d.path = file + sep;
                            //node.children.push(d);
                            node.addChild(d);

                            var d_url = d.url
                            if(is_windows) {
                                d_url = d_url.toLocaleLowerCase();
                            }

                            url_to_file_map[d_url] = d;

                            walk(d, file, function(err, res) {
                                results = results.concat(res);
                                if (!--pending) done(null, results);
                            });
                        } else {
                            var f = new CrsaFile();
                            f.name = filename;
                            f.isFile = true;
                            f.url = node.url + '/' + encodeURIComponent(f.name);
                            f.path = file;
                            f.isEditable = pinegrow.isFileEditable(file);
                            //node.children.push(f);
                            node.addChild(f);

                            var d_url = f.url
                            if(is_windows) {
                                d_url = d_url.toLocaleLowerCase();
                            }

                            url_to_file_map[d_url] = f;

                            results.push(file);
                            if (!--pending) done(null, results);
                        }
                    });
                });
            });
        };
        walk(_this.root, dir, function(err) {
            updateFileNamesIndex(true);

            if(pgf.project_db) {
                _this.requireDB();
            }
            done(_this);
        });
    }

    this.addFile = function(name, content) {
        var sep = '/';

        var paths = name.split(sep);
        var file_name = paths.pop();
        var folders = paths;

        var f = new CrsaFile();
        f.name = file_name;
        f.isFile = true;
        f.url = _this.root.url + sep + name;
        f.path = f.url;
        f.isEditable = pinegrow.isFileEditable(file_name);
        f.content = content;

        var p = _this.root;
        folders.forEach(function(folder) {
            var folder_url = p.url + sep + folder;
            if(!url_to_file_map[folder_url]) {
                var ff = new CrsaFile();
                ff.name = folder;
                ff.isFile = false;
                ff.url = folder_url;
                ff.path = folder_url;
                //p.children.push(ff);
                p.addChild(ff);
                url_to_file_map[folder_url] = ff;
            }
            p = url_to_file_map[folder_url];
        })

        p.addChild(f);
        //p.children.push(f);

        url_to_file_map[f.url] = f;
    }

    var updateUrlsOfPagesAndStylesheetsAfterRename = function(old_url, url) {

        var replace = function(str, old, new_url) {
            var re = new RegExp(`^(${pgParserEscapeRegExp(old)})($|\\/)`);
            return str.replace(re, new_url + '$2');
        }

        url_to_file_map = {}

        _this.walkThroughFiles(function(node, parent) {
            if(parent) {
                //delete url_to_file_map[node.url]
                var rel_old = _this.getRelativeUrl(node.url);
                node.url = parent.url + '/' + node.name;
                node.path = parent.path + '/' + node.name;
                url_to_file_map[node.url] = node;

                var rel_new = _this.getRelativeUrl(node.url);
                if(rel_new !== rel_old) {
                    node.renamedFromOriginalUrl = rel_old;
                }
            }
        })

        if(old_url) {
            var changed_pages = [];
            var project_info = _this.getProjectInfo();

            pinegrow.getAllPages().forEach(function (page) {
                var page_url = replace(page.url, old_url, url);
                if(page_url !== page.url) {
                    var rel_new = _this.getRelativeUrl(page_url);
                    var rel_old = _this.getRelativeUrl(page.url);
                    if(rel_new !== rel_old) {
                        project_info.renameSettingsForUrl(rel_old, rel_new);
                        project_info.renameSettingsForFile(rel_old, rel_new);
                        changed_pages.push(page);
                        page.renamedFromOriginalUrl = rel_old;
                    }
                    page.rename(page_url);
                }
            })

            if(changed_pages.length) {
                $.fn.crsapages('updateFilesName');
                changed_pages.forEach(function(page) {
                    page.view.updateTitle(page.tab_name);
                })
                project_info.save();
            }

            pinegrow.getStylesheets().forEach(function (cs) {
                if (cs.url) {
                    var cs_url = replace(cs.url, old_url, url);
                    var old_cs_url = cs.url;

                    if(cs_url !== old_cs_url) {
                        var rel_new = _this.getRelativeUrl(cs_url);
                        var rel_old = _this.getRelativeUrl(cs.url);

                        cs.setUrl(cs_url);

                        if(rel_new !== rel_old) {
                            cs.renamedFromOriginalUrl = rel_old;
                        }

                        pinegrow.getAllPages().forEach(function (page) {
                            page.getPageViews().forEach(function (pv) {
                                var $style = pv.get$Html().find(`style[data-pg-style-url="${old_cs_url}"]`);
                                if ($style.length) {
                                    $style.attr('data-pg-style-url', cs_url);
                                }
                            })
                        })
                    }
                }
            })
        }

        var old_rel_url = _this.getRelativeUrl(old_url);
        var new_rel_url = _this.getRelativeUrl(url);

        if(old_rel_url !== new_rel_url) {
            var map = {}
            $.each(map_url_source_to_dom, function (source_url, dom_url) {
                map[replace(source_url, old_rel_url, new_rel_url)] = dom_url;
            })
            map_url_source_to_dom = map;

            map = {}
            $.each(map_url_dom_to_source, function (dom_url, source_url) {
                map[dom_url] = replace(source_url, old_rel_url, new_rel_url);
            })
            map_url_dom_to_source = map;
        }
    }

    this.setUrl = function(url) {
        var old_url = this.url;
        this.url = url;
        this.root.url = this.url;
        this.root.path = !this.loaded_from_backend ? crsaMakeFileFromUrl(url) : url;

        updateUrlsOfPagesAndStylesheetsAfterRename(old_url, url);
    }

    //rename folders and any project file
    this.renameProjectFileUrl = function(old_url, new_name) {
        if(!this.loaded_from_backend) console.error('Only for projects loaded from backend.');

        var f = this.getFileForUrl(old_url);
        if(f) {
            var old_name = f.name;
            f.name = new_name;
            f.path = f.parent.path + '/' + f.name;
            f.url = f.parent.url + '/' + f.name;

            updateUrlsOfPagesAndStylesheetsAfterRename(old_url, f.url);
        }
    }

    this.fromServer = function(project_url, done, project_user_slug, revision_id) {
        this.url = project_url;
        this.name = decodeURIComponent(crsaGetNameFromUrl(project_url));

        this.root = new CrsaFile();
        this.root.isFile = false;
        this.root.name = this.name;
        this.root.url = this.url;

        var path = this.url;
        this.root.path = path;

        this.loaded_from_backend = true;

        url_to_file_map = {};

        var addFile = this.addFile;

        var withProject = function() {
            if(!_this.lesson) _this.sortItems();

            if(_this.db) {
                _this.db.load(_this);
            }

            addUrlMapsFromMappedFileNodes();

            done(_this);
        }

        var onDone = function() {
            if(_this.lesson && _this.lesson.tutorial.startProjects) {
                //this is a start project, no need to load
                withProject()
            } else {
                if (pinegrow.backend) {
                    pinegrow.backend.loadProject(_this, function () {
                        withProject();
                    }, project_user_slug, null, revision_id)
                } else {
                    done(_this);
                }
            }
        }

        pinegrow.dispatchEvent('on_before_load_project_from_server', null, this, function() {
            updateFileNamesIndex(true);

            if (pgf.project_db) {
                _this.requireDB();
            }
            onDone();
        })

        /*

        if(list.length) {
            list[0](this, function() {
                updateFileNamesIndex(true);

                if (pgf.project_db) {
                    _this.requireDB();
                }
                onDone();
            })
        } else {

            updateFileNamesIndex(true);

            if (pgf.project_db) {
                _this.requireDB();
            }
            onDone();
        }*/
    }


    this.requireDB = function(done, skip_create) {
        if(!this.db) {
            this.db = new PgProjectDB(this);
            this.db.init(skip_create);
        } else {
            if(!this.db.isInitialized()) {
                this.db.create();
            }
        }
        done && done(this.db);
        return this.db;
    }

    this.walkThroughFiles = function(func, node, parent) {
        node = node || this.root;
        var r = func(node, parent);
        if(r === false) return;
        for(var i = 0; i < node.children.length; i++) {
            this.walkThroughFiles(func, node.children[i], node);
        }
    }

    var stored_tags = {};

    this.storeTags = function() {
        stored_tags = {};
        this.walkThroughFiles(function(node) {
            stored_tags[node.url] = {
                tag: node.tag,
                tags: node.tags
            }
        })
    }

    this.restoreTags = function() {
        this.walkThroughFiles(function(node) {
            var d = stored_tags[node.url] || null;
            if(d) {
                node.tag = d.tag;
                node.tags = d.tags.slice(0);
            }
        })
    }


    this.setVersionForUrl = function(rel_url, version) {
        var f = this.getFileForUrl(rel_url);
        if(f) {
            f.version = version;
        }
    }

    this.getVersionForUrl = function(rel_url) {
        var f = this.getFileForUrl(rel_url);
        return (f && f.version) ? f.version : 0;
    }

    this.getContentOfUrlWithLookup = function(rel_url) {
        rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(rel_url));
        var url = this.url + '/' + rel_url;

        var type = pinegrow.getSimpleMimeType(rel_url);
        if(type === 'css') {
            var cslist = pinegrow.stylesheets.findCrsaStylesheetForUrl(url, false, true);
            var cs = cslist.length > 0 ? cslist[0] : null;

            if (cs && cs.loaded) {
                var content = null;

                if(cs.sourceBeingCurrentlyParsed) {
                    //source of the cs was set but the processing is not finished yet. we are using this instead of getSource that would still get us the old version
                    content = cs.sourceBeingCurrentlyParsed;
                } else {
                    cs.getCssSource(function (css) {
                        css = css || '';
                        content = css;
                    }, false);
                }
                if(content !== null) {
                    return content;
                }
            }
        }

        var page = pinegrow.getCrsaPageByUrl(url);

        if(!page) {
            page = this.getBackgroundPageForUrl(url);
        }
        if(page) {
            return page.getSource();
        }

        return this.getContentOfUrl(rel_url);
    }

    this.getContentOfUrl = function(rel_url, done) {
        rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(rel_url));

        if(crsaIsAbsoluteUrl(rel_url)) {
            done && done('Remote file ' + rel_url);
        } else {
            if(!this.loaded_from_backend) {
                var fs = require('fs');
                var content = null;
                try {
                    content = fs.readFileSync(this.getAbsolutePath(rel_url), {encoding: 'utf8'});

                } catch(err) {}
                done && done(content);
                return content;

            } else {
                var f = this.getFileForUrl(this.url + '/' + rel_url);
                done && done(f ? f.content : '');
                return f ? f.content : null;
            }
        }
    }

    this.hasContentForUrl = function(rel_url) {
        rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(rel_url));
        if(!this.loaded_from_backend) {
            return require('fs').existsSync(this.getAbsolutePath(rel_url));
        } else {
            var f = this.getFileForUrl(this.url + '/' + rel_url);
            if (f && f.content !== undefined && f.content !== null) return true;
            return false;
        }
    }



    this.getViewCodeForSave = function(url, filter_url) {
        url = this.getAbsoluteUrl(url);
        var page = pinegrow.getCrsaPageByUrl(url);
        var node;

        var mapUrls = function(node) {
            node.withoutEvents(function() {
                node.goThroughAllUrls(function(url, attr, node) {
                    if(filter_url) {
                        return filter_url(url, attr, node);
                    } else {
                        return _this.mapSourceUrlToDOMUrl(url);
                    }
                })

                node.walkSelfAndChildren(function(node) {
                    if(node.isElement) {
                        var s = node.getAttribute('style');
                        if(s) {
                            var ns = pgAddProxyLinksToStyleAttribute(s);
                            if(ns !== s) {
                                node.setAttribute('style', ns);
                            }
                        }
                    }
                    return true;
                })
            })
        }

        var keep_collapsed = pinegrow.getSetting('keep-collapsed', '0') == '1';

        if(page) {
            node = page.sourceNode.clone(true);
            mapUrls(node);
            return node.toStringForCodeEditorOrSave(page, true, keep_collapsed, true);
        } else {
            var cslist = pinegrow.stylesheets.findCrsaStylesheetForUrl(url, false, true);
            var cs = cslist.length > 0 ? cslist[0] : null;

            if (cs && cs.loaded) {
                var code;

                cs.getCssSource(function (css) {
                    css = css || '';
                    code = pgAddProxyLinksToStyleAttribute(css);
                }, false);

                return code;

            } else {
                var f = this.getFileForUrl(url);
                if (f.content) {
                    switch(f.getSimpleType()) {
                        case 'html':
                            var p = new pgParser();
                            p.assignIds = false;
                            p.parse(f.content);
                            if (p.rootNode) {
                                mapUrls(p.rootNode);
                                return p.rootNode.toStringForCodeEditorOrSave(null, true, keep_collapsed, true);
                            } else {
                                return null;
                            }
                            break;

                        case 'css':
                            return pgAddProxyLinksToStyleAttribute(f.content);
                            break;

                        default:
                            return null;//f.content;
                    }

                }
            }
        }
        return null;
    }

    //remember and restore open pages
    var open_pages = [];

    this.pageWasOpenedOrClosed = function(page, opened) {
        var rel_url = this.getRelativeUrl(page.url);
        var idx = open_pages.indexOf(rel_url);
        if(opened) {
            if(idx < 0) open_pages.push(rel_url);
        } else {
            if(idx >= 0) open_pages.splice(idx, 1);
        }
        var pi = pinegrow.getCurrentProjectInfo();
        pi.setSetting('open-pages', open_pages);
        pi.save();
    }

    this.getListOfOpenPages = function() {
        var pi = pinegrow.getCurrentProjectInfo();
        return pi.getSetting('open-pages') || [];
    }


    this.changed_setting_files = [];

    this.settingsFileIsChanged = function(file) {
        if(this.changed_setting_files.indexOf(file) < 0) {
            this.changed_setting_files.push(file);
        }
        pinegrow.backend.saveDelayed();
    }

    this.setContentOfUrlWithLookup = function(rel_url, content, force_save) {
        rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(rel_url));
        var url = this.url + '/' + rel_url;

        var page = pinegrow.getCrsaPageByUrl(url);

        if(!page) {
            page = this.getBackgroundPageForUrl(url);
        }

        if(page) {
            page.setCode(content);
            pinegrow.dispatchEvent('on_page_changes_done', page, {});
            if(!force_save) return true;
        }

        var type = pinegrow.getSimpleMimeType(rel_url);
        if(type === 'css') {
            var cslist = pinegrow.stylesheets.findCrsaStylesheetForUrl(url, false, true);
            var cs = cslist.length > 0 ? cslist[0] : null;

            if (cs && cs.loaded) {
                cs.setSource(content, function() {
                    pinegrow.dispatchEvent('on_css_source_changed', null, {list: [cs]});
                })
                if(!force_save) return true;
            }
        }

        return this.setContentOfUrl(rel_url, content);
    }

    this.setContentOfUrl = function(rel_url, content, done) {
        var changed = true;
        rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(rel_url));

        if(crsaIsAbsoluteUrl(rel_url)) {
            done && done('Remote file ' + rel_url);
        } else {
            if(!this.loaded_from_backend) {
                var fs = require('fs');
                try {
                    crsaWriteFileWithBackup(fs, this.getAbsolutePath(rel_url), content, 'utf8');
                } catch(err) {}
                done && done(true);
                return true;
            } else {
                var f = this.getFileForUrl(this.url + '/' + rel_url);
                if (f) {
                    if (f.content === content) changed = false;
                    f.content = content;
                } else {
                    this.addFile(rel_url, content);
                }
                done && done(changed);
            }
        }
        return changed;
    }

    this.setThumbnailOfUrl = function(rel_url, thumb) {
        rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(rel_url));
        var f = this.getFileForUrl(this.url + '/' + rel_url);
        if(f) {
            f.thumb = thumb;
        }
    }


    this.getInternalUrl = function(rel_url) {
        return (new URL(rel_url, this.url + '/')).toString();
    }

    this.getRelativeUrlFromInternalUrl = function(url) {
        return url.replace(this.url + '/', '');
    }

    this.findFilesByName = function(name) {
        var r = [];
        this.walkThroughFiles(function(node) {
            if(node.name === name) r.push(node);
        })
        return r;
    }

    var sortNodeChildren = function(node) {
        node.children.sort( function(a, b) {
            return ((a.isFile ? 'b' : 'a') + a.name).localeCompare((b.isFile ? 'b' : 'a') + b.name);
        })
        for(var i = 0; i < node.children.length; i++) {
            sortNodeChildren( node.children[i]);
        }
    }

    this.sortItems = function() {
        sortNodeChildren(this.root);
    }

    this.removeFileNodeFromList = function(file_node) {
        delete url_to_file_map[file_node.url];
        //if folder
        file_node.walkThroughFiles(function(f) {
            delete url_to_file_map[f.url];
        })

        var idx = file_node.parent.children.indexOf(file_node);
        if(idx >= 0) {
            file_node.parent.children.splice(idx, 1);
        }
        /*
        this.walkThroughFiles(function(node, parent) {
            if(node === file_node) {
                var idx = parent.children.indexOf(node);
                if(idx >= 0) {
                    parent.children.splice(idx, 1);
                }
            }
        })*/
    }

    this.getFileForUrl = function(url) {
        if(!crsaIsAbsoluteUrl(url)) {
            url = this.url + '/' + url;
        }
        if(is_windows) {
            url = url.toLocaleLowerCase();
        }
        return url_to_file_map[url] || null;
    }

    this.closeProjectExplorer = function($lib) {
        if(pinegrow.projectView) {
            pinegrow.projectView.close();
        }
    }

    this.showProjectExplorer = function() {
        if(pinegrow.projectView) {
            pinegrow.projectView.show(this);
        }
    }

    this.copyRequiredFilesTo = function(dir, filter_func, fileWriter, only_dirs) {

        var is_app = isApp();

        if(is_app) {
            var fs = require('fs');
            var path = require('path');
            var path_sep = path.sep;
        } else {
            var path_sep = '/';
        }

        if(dir === null) dir = '';

        var copyNodes = function(nodes, dest) {
            $.each(nodes, function(i, cf) {
                var req = cf.required || (filter_func && filter_func(cf));
                if(!req) return true;
                var dest_path = dest + cf.name;
                var exists = is_app ? crsaIsFileOrDir(dest_path, fs) : (cf.isFile && fileWriter.fileExists(dest_path));

                if(exists && cf.isFile && !only_dirs) {
                    //var sourcestat = fs.statSync(cf.path);
                    //var deststat = fs.statSync(dest_path);
                    if(/* deststat.mtime < sourcestat.mtime || */ fileWriter) { //this was dangerous
                        exists = false; //overwrite
                    }
                }
                if(!exists) {
                    if(cf.isFile) {
                        if(!only_dirs) {
                            //console.log('copy ' + cf.path + ' to ' + dest_path);
                            if(!fileWriter) {
                                crsaCopyFileSync(fs, cf.path, dest_path);
                            } else {
                                fileWriter.copyFile(dest_path, cf.path);
                            }
                        }
                    } else {
                        is_app && fs.mkdirSync(dest_path);
                    }
                }
                if(cf.children) {
                    copyNodes(cf.children, dest_path + path_sep);
                }
            });
        }
        copyNodes(this.root.children, dir);
    }

    this.copyRequiredFilesToZip = function(zip, done) {

        var count = 0;

        var copyNodes = function(nodes, zip, folder) {

            $.each(nodes, function(i, cf) {
                if(!cf.required) return true;

                if(cf.isFile) {
                    console.log('copy ' + cf.url + ' to ' + folder);
                    count++;
                    $.ajax({
                        url: cf.url,
                        data: null,
                        dataType: 'text'
                    }).done(function(data) {
                        setTimeout(function() {
                            zip.file(cf.name, data);
                            count--;
                            if(count == 0) done();
                        }, 0);
                    }).fail(function() {
                        setTimeout(function() {
                            count--;
                            console.log('can not read ' + cf.url);
                            if(count == 0) done();
                        }, 0);
                    });
                } else {
                    var subzip = zip.folder(cf.name);
                    if(cf.children) {
                        copyNodes(cf.children, subzip, folder + '/' + cf.name);
                    }
                }
            });
        }
        copyNodes(this.root.children, zip, '/');
        if(count == 0) done();
    }

    //project API
    this.getDir = function() {
        return this.root.path;
    }

    this.getUrl = function() {
        return this.root.url;
    }

    this.getName = function() {
        return this.root.name;
    }

    this.isFileInProject = function(file) {
        return file.startsWith( this.getDir());
    }

    this.isUrlInProject = function(url) {
        return url.startsWith( this.getUrl());
    }

    this.isPageInProject = function(page) {
        if(this.loaded_from_backend) return true;
        return page.localFile && this.isFileInProject(page.localFile);
    }

    this.getProjectInfo = function() {
        return new pgProjectInfo(this.getDir());
    }

    this.getRelativeUrl = function (url) {
        var relativeUrl = url.replace(this.getUrl(), '');
        if (relativeUrl[0] == "/")
            relativeUrl = relativeUrl.substr(1, relativeUrl.length);
        return relativeUrl;
    }

    this.getRelativePath = function(file) {
        return file.replace( this.getDir() + this.path_separator, '');
    }

    this.getAbsolutePath = function(relative_file) {
        return path.join( this.getDir(), relative_file);
    }

    this.getAbsoluteUrl = function(relative_file) {
        if(crsaIsAbsoluteUrl(relative_file)) return relative_file;
        if(relative_file.startsWith('/')) relative_file = relative_file.substr(1);
        var purl = this.getUrl();
        if(!purl.endsWith('/')) purl += '/';
        return new URL(relative_file, purl).toString();
    }

    this.forEachOpenPage = function(func, ref_page) {
        if(ref_page && !this.isPageInProject(ref_page)) return;
        var pages = pinegrow.getAllPages();
        for (var i=0; i < pages.length; i++) {
            if(this.isPageInProject(pages[i])) {
                if(ref_page && ref_page == pages[i]) continue;
                func(pages[i]);
            }
        }
    }

    this.getEditableFiles = function(ignore_map, project) {
        project = project || this;
        var files = [];

        var findNodes = function(node) {
            if(node.isEditable) {
                files.push(node);
            }
            if(!node.name.startsWith('_pg')) {
                var folders = []

                for(var i = 0; i < node.children.length; i++) {

                    if(ignore_map) {
                        var rel_url = project.getRelativeUrl(node.children[i].url);
                        if(ignore_map[rel_url]) continue;
                    }

                    if(!node.children[i].isFile) {
                        folders.push(node.children[i])
                    } else {
                        findNodes(node.children[i]);
                    }
                }

                folders.forEach(function(folder) {
                    findNodes(folder)
                })
            }
        }

        findNodes(this.root);
        return files;
    }

    this.getFilesWithContent = function() {
        var files = [];

        var findNodes = function(node) {
            if('content' in node && node.content !== null) {
                files.push(node);
            }
            if(!node.name.startsWith('_pg')) {
                var folders = []

                for(var i = 0; i < node.children.length; i++) {

                    if(!node.children[i].isFile) {
                        folders.push(node.children[i])
                    } else {
                        findNodes(node.children[i]);
                    }
                }

                folders.forEach(function(folder) {
                    findNodes(folder)
                })
            }
        }

        findNodes(this.root);
        return files;
    }

    this.getThumbnailForUrl = function(url) {
        var rel_url = this.getRelativeUrl(this.getRelativeUrlFromInternalUrl(url));
        var f = this.getFileForUrl(this.url + '/' + rel_url);
        if(f) {
            return f.thumb || null;
        }
        return null;
    }

    this.forEachEditableFile = function(func, done, title, reference, ignore_map, silent) {

        if(!title) title = 'Processing files...';

        var errors = [];
        var summary = [];

        var files = this.getEditableFiles(ignore_map, this);

        var progressUI = function() {

            var $ui = $('<div><p>index.html(90 of 100)</p>\
            <div class="progress">\
                <div class="progress-bar" style="width: 0%;">\
                </div>\
            </div></div>');

            var $title = $ui.find('h3');
            var $txt = $ui.find('p');
            var $bar = $ui.find('.progress-bar');

            var $dialog = makeAndShowDialog(title, null, "Stop", $ui, function() {
                $dialog = null;
            }, function() {
                $dialog = null;
                stopProcessing();
            });
            $dialog.css('z-index', 2000);

            this.hide = function() {
                if($dialog) {
                    $dialog.remove();
                    $dialog = null;
                }
            }

            this.update = function(file, num, total) {
                var percent = parseInt(num * 100 / total);
                $bar.attr("style", "width:" + percent + "%;");
                $txt.html(file + ' (' + num + ' of ' + total + ')');
            }
        }

        var onDone = function() {
            pinegrow.callGlobalFrameworkHandler('on_project_scan_end', _this, reference);

            if(nextFileTimer) {
                clearTimeout(nextFileTimer);
                nextFileTimer = null;
            }

            if(silent) {
                done && done(errors);
                return;
            }

            if(files) {
                ui.update('Done!', files.length, files.length);
            } else {
                ui.update('Done!', 100, 100);
            }
            files = null;
            if(summary.length) {
                ui.hide();
                var str = '<table class="table table-striped table-condensed table-hover"><thead><tr><td>File</td><td>Messages</td></tr></thead><tbody>';
                for(var i = 0; i < summary.length; i++) {
                    str += '<tr><td><b>' + _this.getRelativePath(summary[i].file.path) + '</b></td><td>';
                    for(var j = 0; j < summary[i].status.errors.length; j++) {
                        str += '<p class="error">' + summary[i].status.errors[j] + '</p>';
                    }
                    for(var j = 0; j < summary[i].status.warnings.length; j++) {
                        str += '<p class="warning">' + summary[i].status.warnings[j] + '</p>';
                    }
                    for(var j = 0; j < summary[i].status.messages.length; j++) {
                        str += '<p class="message">' + summary[i].status.messages[j] + '</p>';
                    }
                    str += '</td></tr>';
                }
                str += '</tbody></table>';

                pinegrow.showAlert('<p>The following happened during processing:</p>' + str, 'Status report', null, "OK", function() {
                    if(done) done(errors);
                }, function() {
                    if(done) done(errors);
                });
            } else {
                setTimeout(function() {
                    ui.hide();
                }, 500);
                if(done) done(errors);
            }
        }

        var idx = 0;
        var nextFileTimer = null;

        var stopProcessing = function() {
            if(nextFileTimer) clearTimeout(nextFileTimer);
            nextFileTimer = null;
            idx = files.length;
            onDone();
        }

        pinegrow.callGlobalFrameworkHandler('on_project_scan_begin', this, reference);

        var doNode = function() {

            nextFileTimer = null;
            if(files && idx < files.length) {
                var node = files[idx];

                ui && ui.update(node.name, idx + 1, files.length);

                try {
                    var page = pinegrow.getCrsaPageForUrl( node.url );

                    if(page.sourceNode) {
                        //idx++;
                        //nextFileTimer = setTimeout(doNode, 1);
                        var status = new pgProjectItemIteratorStatus();

                        func(page, function (page, status) {
                            var changed = status && status.changed;
                            if (changed) {
                                if (page.openedInEditor) {
                                    if (!status.skip_refresh) {
                                        page.refresh();
                                    }
                                } else {
                                    pages_cache[page.url] = page; //store page
                                }
                                page.setPageChanged(true, true /* force update */);
                            }
                            if (status.save) {
                                page.save(null, true);
                            }
                            if (status) {
                                if (status.errors.length || status.warnings.length || status.messages.length) {
                                    summary.push({file: node, status: status});
                                }
                            }
                            idx++;
                            var time = (new Date().getTime());
                            if (time - node_batch_start_time < 50) {
                                doNode();
                            } else {
                                node_batch_start_time = time;
                                nextFileTimer = setTimeout(doNode, 1);
                            }
                        }, status, node);

                        pinegrow.callGlobalFrameworkHandler('on_project_scan_page', _this, page, reference, node);
                    } else {
                        //no sourceNode
                        idx++;
                        nextFileTimer = setTimeout(doNode, 1);
                    }

                    page = null;

                } catch(err) {
                    var status = new pgProjectItemIteratorStatus();
                    status.errors.push(err.toString());
                    errors.push({file: node, error: err.toString()});

                    if(status.errors.length || status.warnings.length || status.messages.length) {
                        summary.push({file: node, status: status });
                    }
                    idx++;
                    //doNode();
                    nextFileTimer = setTimeout(doNode, 1);

                    console.log('FOR EACH ERROR ' + err);
                }

            } else {
                //done
                onDone();
            }
        }

        if(!silent) {
            var ui = new progressUI();
        }

        var node_batch_start_time = (new Date().getTime());
        doNode();
    }

    this.getFrameworks = function() {
        var list = [];
        var pi = new pgProjectInfo(this.getDir());
        var keys = pi.getSetting('frameworks');
        if(keys) {
            for(var i = 0; i < keys.length; i++) {
                var f = pinegrow.findFrameworkByKey(keys[i])
                if(f) {
                    list.push(f);
                }
            }
        }
        return list;
    }

    this.updateTagsForFile = function(file, $li) {
        if(!pinegrow.projectView) return null;
        return pinegrow.projectView.updateTagsForFile(file, $li);
    }

    this.getRemoteUrls = function() {
        var pi = this.getProjectInfo();
        return pi.getSetting('remote-urls') || [];
    }

    this.addRemoteUrl = function(url) {
        var pi = this.getProjectInfo();
        var list = pi.getSetting('remote-urls') || [];
        var idx = list.indexOf(url);
        if(idx >= 0) {
            list.splice(idx, 1);
        }
        list.unshift(url);
        pi.setSetting('remote-urls', list);
        pi.save();
        pinegrow.dispatchEvent('on_project_remote_urls_updated', null, this);
    }

    this.removeRemoteUrl = function(url) {
        var pi = this.getProjectInfo();
        var list = pi.getSetting('remote-urls') || [];
        var idx = list.indexOf(url);
        if(idx >= 0) {
            list.splice(idx, 1);
        }
        pi.setSetting('remote-urls', list);
        pi.save();
        pinegrow.dispatchEvent('on_project_remote_urls_updated', null, this);
    }

    var map_url_source_to_dom = {};
    var map_url_dom_to_source = {};

    var getMappedUrlFromFile = function(f) {
        return f.getMappedUrl();
    }

    var addUrlMapsFromMappedFileNodes = function() {
        _this.walkThroughFiles(function(f) {
            var mu = getMappedUrlFromFile(f);
            if(mu) {
                _this.addSourceUrlToDOMUrlMap(_this.getRelativeUrl( f.url), mu);
            }
        })
    }

    this.mapSourceUrlToDOMUrl = function(url, def) {
        if(url === null) return url;
        if(def === undefined) def = url;
        return map_url_source_to_dom[url] || def;
    }

    this.mapDOMUrlToSourceUrl = function(url) {
        if(url === null) return url;
        return map_url_dom_to_source[url] || url;
    }

    this.addSourceUrlToDOMUrlMap = function(source_url, dom_url) {
        map_url_dom_to_source[dom_url] = source_url;
        map_url_source_to_dom[source_url] = dom_url;
    }

    this.getMapForSourceToDOMUrls = function() {
        return map_url_source_to_dom;
    }
}

var pgProjectItemIteratorStatus = function() {
    this.changed = false;
    this.skip_refresh = false;
    this.viewUpdated = false;
    this.errors = [];
    this.warnings = [];
    this.messages = [];
}

var pgProjectInfo = function(file_or_dir, dont_search) {

     var browser = !isApp();
     var project = pinegrow.getCurrentProject();

     if(project && project.loaded_from_backend) {
         browser = true;
     }

     if(browser) {
         file_or_dir = project.getUrl();
     }

    //if(!browser) {
        var path = require('path');
        var fs = require('fs');
    //}

    var getFileNameRelativeToProject = function(file, project_dir) {
        return file.replace( project_dir + path.sep, '');
    }

    var getRelativeUrl = function(url, purl) {
        url = url.replace(purl, '');
        if(url.startsWith('/')) url = url.replace('/', '');
        return url;
    }

    var getAbsoluteUrl = function(url, purl) {
        if(crsaIsAbsoluteUrl(url)) return url;
        if(url.startsWith('/')) url = url.replace('/', '');
        if(!purl.endsWith('/')) purl += '/';
        return purl + url;
    }

    var findAndReadProjectFile = function(file, dont_search) {
        if(browser) {
            var json = project.getContentOfUrl('pinegrow.json');
            if(json) {
                var info = JSON.parse(json);
                return {data: info, file: file, project_dir: project.getUrl(), file_key: ''};
            }
            return null;
        }
        if(!file || !isApp()) return null;

        var dir = crsaGetDir(file, path, fs);

        while(dir) {
            var pf = path.join(dir, 'pinegrow.json');
            try {
                var json = fs.readFileSync(pf, {encoding: "utf8"});
                var info = JSON.parse(json);
                //console.log('Project file found: ' + pf);
                return {data: info, file: pf, project_dir: dir, file_key: getFileNameRelativeToProject(file, dir)};
            } catch(err) {

            }
            if(dont_search) break;

            try {
                var parent_dir = path.join(dir, '..');
                if(parent_dir == dir) break;
                dir = parent_dir;
            }
            catch(err) {
                break;
            }
        }
        return null;
    }

    var getNewProjectInfo = function(file) {
        if(!file || browser) return {file: file, project_dir: project ? project.getUrl() : null, data: { files: {}}, file_key: null};

        //var existing = findAndReadProjectFile();
        //if(existing) return existing.file;

        //no file yet

        var currentProject = pinegrow.getCurrentProject();
        var pf;

        if(currentProject && currentProject.isFileInProject(file)) {
            pf = path.join(currentProject.getDir(), 'pinegrow.json');
        }
        pf = path.join( crsaGetDir(file, path, fs), 'pinegrow.json');
        var project_dir = path.dirname(pf);
        return {file: pf, project_dir: project_dir, data: { files: {}}, file_key: getFileNameRelativeToProject( file, project_dir)};
    }

    var saveProjectInfo = function(info) {
        if(browser) {
            if(project.setContentOfUrl('pinegrow.json', JSON.stringify(info.data))) {
                //changed
                project.settingsFileIsChanged('pinegrow.json');
            }
            return;
        }
        if(!info.file) return; //ignore pi for remote documents

        var fs = require('fs');
        var json = JSON.stringify(info.data);
        crsaWriteFileWithBackup(fs, info.file, json, "utf8");
    }

    this.was_just_created = false;
    var info = findAndReadProjectFile( file_or_dir, crsaIsFileOrDir(file_or_dir) == 'dir' || dont_search );
    if(!info) {
        info = getNewProjectInfo( file_or_dir );
        this.was_just_created = true;
    }

    this.save = function() {
        saveProjectInfo(info);
    }

    this.saveDelayed = function() {
        delayed.run('save', function() {
            saveProjectInfo(info);
        }, 1000)
    }

    this.getSetting = function(key, def) {
        if(key in info.data) return info.data[key];
        return def || null;
    }

    this.setSetting = function(key, value) {
        info.data[key] = value;
    }

    this.getSettingsForFile = function(file) {
        if(!file) return null;
        var r = getFileNameRelativeToProject( file, info.project_dir );
        if(info.data.files && info.data.files[r]) {
            return info.data.files[r];
        }
        return null;
    }

    this.getSettingForFile = function(file, key) {
        if(!file) return null;
        var r = getFileNameRelativeToProject( file, info.project_dir );
        if(info.data.files && info.data.files[r]) {
            return info.data.files[r][key] || null;
        }
        return null;
    }

    this.setSettingForFile = function(file, key, value) {
        if(!info) return null;
        var r = getFileNameRelativeToProject( file, info.project_dir );
        if(!info.data.files) {
            info.data.files = {};
        }
        if(!info.data.files[r]) {
            info.data.files[r] = {};
        }
        info.data.files[r][key] = value;
    }

    this.getSettingForUrl = function(url, key) {
        if(!url || !info || !info.project_dir) return null;
        var r = getRelativeUrl( url, browser ? info.project_dir : crsaMakeUrlFromFile(info.project_dir ));
        if(info.data.urls && info.data.urls[r]) {
            return (key in info.data.urls[r]) ? info.data.urls[r][key] : null;
        }
        return null;
    }

    this.renameSettingsForUrl = function(old_url, new_url) {
        if(!info || !info.project_dir) return null;
        if(info.data.urls && info.data.urls[old_url]) {
            info.data.urls[new_url] = info.data.urls[old_url];
            delete info.data.urls[old_url];
        }
    }

    this.renameSettingsForFile = function(old_url, new_url) {
        if(!info || !info.project_dir) return null;
        if(info.data.urls && info.data.files[old_url]) {
            info.data.files[new_url] = info.data.files[old_url];
            delete info.data.files[old_url];
        }
    }

    this.setSettingForUrl = function(url, key, value) {
        if(!url || !info || !info.project_dir) return null;
        var r = getRelativeUrl( url, browser ? info.project_dir : crsaMakeUrlFromFile(info.project_dir ));
        if(!info.data.urls) {
            info.data.urls = {};
        }
        if(!info.data.urls[r]) {
            info.data.urls[r] = {};
        }
        info.data.urls[r][key] = value;
    }


    this.getDir = function() {
        return info.project_dir;
    }

    this.makeFileRelative = function(file) {
        if(!info || !info.project_dir) return file;
        return getFileNameRelativeToProject( file, info.project_dir );
    }

    this.makeFileAbsolute = function(file) {

    }

    this.getRelativeUrlFromFile = function(file) {
        if(!info || !info.project_dir) return file;
        if(browser) {
            return getRelativeUrl(file, info.project_dir);
        }
        var url = crsaMakeUrlFromFile(file);
        var purl = crsaMakeUrlFromFile(info.project_dir);
        return getRelativeUrl(url, purl);
    }

    this.getFileFromRelativeUrl = function(url) {
        if(!info || !info.project_dir) return url;
        var purl = crsaMakeUrlFromFile(info.project_dir);
        return getAbsoluteUrl(url, purl);
    }

    this.makeUrlRelative = function(url) {
        if(!info || !info.project_dir) return url;
        var purl = browser ? info.project_dir : crsaMakeUrlFromFile(info.project_dir);
        return getRelativeUrl(url, purl);
    }

    this.makeUrlAbsolute = function(url) {
        if(!info || !info.project_dir) return url;
        var purl = browser ? info.project_dir : crsaMakeUrlFromFile(info.project_dir);
        return getAbsoluteUrl(url, purl);
    }

    this.getSourceStylesheetForCSSStylesheet = function(relative_url) {
        if(info.data.compiledStylesheet && info.data.compiledStylesheet[relative_url]) {
            return info.data.compiledStylesheet[relative_url];
        }
        return null;
    }

    this.getDir = function() {
        return info.project_dir;
    }

    this.saveFile = function(rel_name, data) {
        if(browser) {
            if(project.setContentOfUrl('_pginfo/' + rel_name, data)) {
                //changed
                project.settingsFileIsChanged('_pginfo/' + rel_name);
            }
            return;
        }
        if(!info.file) return null;
        var file = path.join(this.getDir(), '_pginfo', rel_name);
        crsaCreateDirs(path.dirname(file), fs);
        fs.writeFileSync(file, data, typeof data === 'string' ? {encoding: 'utf8'} : {});
    }

    this.loadFile = function(rel_name) {
        if(browser) {
            return project.getContentOfUrl('_pginfo/' + rel_name);
        }
        if(!info.file) return null;
        var file = path.join(this.getDir(), '_pginfo', rel_name);
        try {
            return fs.readFileSync(file, {encoding: 'utf8'});
        } catch(err) {
            return null;
        }
    }

    this.deleteFile = function(rel_name) {
        if(!info.file) return;
        var file = path.join(this.getDir(), '_pginfo', rel_name);
        try {
            return fs.unlinkSync(file);
        } catch(err) {
            return null;
        }
    }

}