var PgUISection = function(name, cat, sec_def, default_closed, force_open, options) {

    var _this = this;

    this.name = name;
    this.category = cat;
    this.parent = null;

    if(default_closed === undefined || default_closed === null) default_closed = false;

    this.key = 'sec_open_' + escapeHtmlCode(name).replace(/"/g,'').replace(/'/g, '') + '_' + (cat ? escapeHtmlCode(cat) : '');
    this.sec_def = sec_def || {};

    this.options = options || {};

    if(!this.sec_def.hasOwnProperty('closed')) {
        var k = this.key;
        if(k in localStorage) {
            this.sec_def.closed = localStorage[k] != '1'; //not open
        } else if('default_closed' in this.sec_def) {
            this.sec_def.closed = this.sec_def.default_closed;
        } else {
            this.sec_def.closed = default_closed;
        }
    }

    this.is_closable = this.sec_def.is_closable !== undefined ? this.sec_def.is_closable : true;

    this.open = !this.sec_def.closed || !this.is_closable;

    if(force_open === 'closed') {
        this.open = false;
    } else if(force_open) this.open = true;

    this.originalHeight = null;

    this.$element = $(`<div class="pg-ui-section ${this.sec_def.section_class || ''}" data-key="${this.key}"><h2 class="section-title"><span>` + this.name + '</span>' + (this.category ? ('<small>' + this.category + '</small>') : '') + '<i class="icon icon-right closed"></i><i class="icon icon-down opened"></i></h2></div>');

    if(!this.is_closable) {
        this.open = true;
        this.$element.find('.icon.closed, .icon.opened').remove();
    }

    this.$content = $('<div class="pg-ui-section-content"></div>').appendTo(this.$element);

    this.$title = this.$element.find('h2');

    crsaHandleExternalLinks(this.$title);

    if(!this.open) {
        this.$content.hide().css('height', 0);
        this.$element.addClass('section-closed');
    }

    this.onOpenClosed = null;

    if(sec_def && sec_def.pinnable) {
        this.$element.find('.icon').remove();
        var $pin = $('<i class="icon pin-close icon-Pin pin"></i>').appendTo(this.$title);

        $pin.on('click', function(e) {
            e.preventDefault();
            var $i = $(e.target);
            if($i.hasClass('pin')) {
                if(_this.onPin) {
                    _this.onPin(_this);
                }
                $i.removeClass('pin icon-Pin').addClass('icon-close');
            } else {
                if(_this.onClose) {
                    _this.onClose(_this);
                }
                $i.addClass('pin icon-Pin').removeClass('icon-close');
            }
        })

        if(sec_def.pinned) {
            $pin.removeClass('pin icon-Pin').addClass('icon-close');
        }

    } else {
        if(this.is_closable) {
            var $h = this.$element.find('>h2')
                .on('click', function (e) {
                    e.preventDefault();
                    if (e.altKey && _this.parent) {
                        _this.parent.openCloseAll(!_this.open)
                    } else {
                        _this.openClose(!_this.open);
                    }
                })
            var tt = 'Click to open / close. ' + pgShowKbd('ALT') + ' + click to open / close all sections.';
            if(sec_def && sec_def.collapse_tooltip) {
                tt = sec_def.collapse_tooltip;
            }
            if(options && options.tooltip) {
                tt = options.tooltip + '<br>' + tt;
            }
            addTooltip($h.find('span,.icon'), tt);
        }
    }
}

PgUISection.prototype.detach = function(what) {
    this.$element.detach();
}

PgUISection.prototype.appendTo = function(what) {
    this.$content.append(what);
}

PgUISection.prototype.setTitle = function(title) {
    this.$title.find('>span').html(title);
}

PgUISection.prototype.toggle = function(callback, no_animate) {
    this.open = !this.open;

    var $el = this.$content;
    if (!this.originalHeight && this.open) {
        this.originalHeight = $el.css('height', 'auto').show().outerHeight();
        $el.hide().css({height: 0});
    }

    if (this.open) {
        if(no_animate) {
            $el.show().css({height: this.originalHeight});
            if (callback) callback();
        } else {
            $el.show().animate({height: this.originalHeight}, {
                duration: 250, complete: function () {
                    $el.height('auto');
                    if (callback) callback();
                }
            });
        }
    } else {
        if(no_animate) {
            $el.hide().css({height: 0});
            if (callback) callback();
        } else {
            $el.animate({height: 0}, {
                duration: 250, complete: function () {
                    $el.hide();
                    if (callback) callback();
                }
            });
        }
    }
    if(this.onOpenClosed) {
        this.onOpenClosed(this.open);
    }
}

PgUISection.prototype.forceOpen = function() {
    if(!this.open) {
        this.toggle(null, true);
    }
}

PgUISection.prototype.forceClose = function() {
    if(this.open) {
        this.toggle(null, true);
    }
}

PgUISection.prototype.show = function() {
    this.$element.show();
}

PgUISection.prototype.hide = function() {
    this.$element.hide();
}

PgUISection.prototype.openClose = function(open, no_animate) {
    var _this = this;
    if(this.open != open) {
        _this.toggle(function () {
            if (_this.open) {
                _this.$element.removeClass('section-closed');
                if (_this.sec_def) {
                    _this.sec_def.closed = false;
                    localStorage[_this.key] = '1'; //open
                }
            } else {
                _this.$element.addClass('section-closed');
                if (_this.sec_def) {
                    _this.sec_def.closed = true;
                    localStorage[_this.key] = '0'; //open
                }

            }
            if (_this.options.onOpenClose) {
                _this.options.onOpenClose(_this.$element, _this.open);
            }
        }, no_animate)
    }
}

PgUISection.prototype.setHasData = function(d) {
    if(d) {
        this.$element.addClass('has-data');
    } else {
        this.$element.removeClass('has-data');
    }
}

PgUISection.prototype.scrollTo = function(animate) {
    if(animate === undefined) animate = true;
    var _this = this;
    pgScrollToItemInContainer(this.$element, this.parent.$element, animate, true, function() {
        pg$Flash(_this.$element, 'bck_flash');
    });
}

var PgUISectionGrid = function() {

    this.sections = [];

    this.$element = $('<div class="pg-grid"></div>');
    this.$columns;

    this.minWidth = 280;
    this.maxWidth = 40000;

    this.prevW = 0;
    this.prevH = 0;

    this.num_columns = 0;

    this.new_section_id = 0;

    this.fit = false;

    this.fastdom_key = pgGetFastdomTasks(this.$element.get(0)).getUniqueKey();

    this._sectionChromeH = 14 + 30 + 4 + 12 + 3 - 14 - 3;
}

PgUISectionGrid.prototype.openCloseAll = function(open) {
    this.sections.forEach(function(s) {
        s.openClose(open);
    })
}

PgUISectionGrid.prototype.popSection = function () {
    var lastSection = this.sections.pop();
    lastSection.parent = null;
    return lastSection;
}

PgUISectionGrid.prototype.addSection = function(s) {
    this.sections.push(s);
    s.parent = this;
}

PgUISectionGrid.prototype.addSectionAt = function(s, index) {
    this.sections.splice(index, 0, s);
    s.parent = this;
}

PgUISectionGrid.prototype.clear = function() {
    this.sections.forEach(function(s) {
        s.detach();
    })
    if(this.$columns) this.$columns.remove();
    this.$columns = null;
    this.sections = [];
    this.num_columns = 0;

    pgGetFastdomTasks(this.$element.get(0)).stopAll(this.fastdom_key)
}

PgUISectionGrid.prototype.newSectionAdded = function() {
    var sections = this.sections;
    this.clear();
    this.sections = sections;
}

PgUISectionGrid.prototype.updateLayout = function(width, height) {

    var fastdom = false;

    var task = fastdom ? pgGetFastdomTasks(this.$element.get(0)).create(this.fastdom_key) : pgGetFastdomTasks(this.$element.get(0)).createDummy();

    var _this = this;

    task.measure(function() {

        var w = width || _this.$element.width() - 12; //available width
        var h = height || _this.$element.height();

        //how many columns should we have?
        //var num_c = Math.floor(w / ((this.minWidth + this.maxWidth)/2));
        var num_c = Math.floor(w / _this.minWidth);
        //if(w / num_c > this.maxWidth) num_c++;

        if(_this.sections.length < num_c) num_c = _this.sections.length;

        if(num_c < 1) num_c = 1;

        if(_this.onGetNumberOfColumns) {
            num_c = _this.onGetNumberOfColumns(w, h, _this.sections.length, num_c);
        }

        var col_w = w / num_c;

        if(w === _this.prevW && h === _this.prevH && _this.num_columns === num_c) {
            return;
        }

        _this.prevW = w;
        _this.prevH = h;

        task.mutate(function() {
            var new_columns = [];

            if(_this.num_columns != num_c) {
                //create columns and distribute items
                for(var i = 0; i < num_c; i++) {
                    var $c = $('<div class="pg-grid-col"></div>');
                    new_columns.push($c.get(0));

                    for(var s_idx = i; s_idx < _this.sections.length; s_idx += num_c) {
                        $c.append(_this.sections[s_idx].$element);
                    }
                }
                if(_this.$columns) _this.$columns.remove();
                _this.$columns = $(new_columns);
                _this.$columns.appendTo(_this.$element);

                _this.num_columns = num_c;
            }

            var wp = 100.0 / (num_c * 1.0);
            //set columns width
            var cw = Math.round(w / num_c);
            _this.$columns.each(function(i, c) {
                //$(c).css('width', (i < num_c - 1 ? cw : (w - cw * (num_c-1))) + 'px');
                $(c).css({
                    'max-width': wp + '%',
                    'flex': '0 0 ' + wp + '%'
                });
            })

            if(col_w < 360) {
                _this.$element.get(0).classList.add('pg-grid-narrow');
            } else {
                _this.$element.get(0).classList.remove('pg-grid-narrow');
            }

            if(_this.fit) {
                var sec_per_col = Math.ceil(_this.sections.length / num_c);
                var h_for_content = Math.floor((h - sec_per_col * _this._sectionChromeH) / sec_per_col);
                var min_h = 80;
                if(h_for_content < min_h) {
                    h_for_content = min_h;
                }
                _this.sections.forEach(function(s) {
                    s.$content.css('height', h_for_content);
                })
            }

            _this.sections.forEach(function(s) {
                if(s.onLayoutChanged) s.onLayoutChanged();
            })
        })
    })
}

PgUISectionGrid.prototype.destroy = function() {
    this.clear();
    this.$element.remove();
}