
var PgElementSelectorHelper = function() {

    var _this = this;

    var getDB = function() {
        return pinegrow.getCurrentProject().db;
    }

    this.verifySelectorsOnPage = function(page) {
        var problems = [];

        page.sourceNode.walk(function(node) {
            if(node.isElement) {
                var sel = node.getAttr('data-pg-sel');
                if(sel) {
                    if(page.sourceNode.findOne(sel) !== node) {
                        problems.push(node)
                    }
                }
            }
            return true;
        })
        return problems;
    }

    this.getSelectorBadnessScore = function(sel) {
        var s = 0;
        s += (sel.match(/\>/g) || []).length * 5;
        s += (sel.match(/nth-of-type/g) || []).length * 10;
        return s;
    }

    this.isThereABetterAutoSelector = function(pgel, sel) {

        var list = pgel.getUniqueSelectors(0, 9999, pgel.document, true);
        var new_sel = list.length ? list[0] : null;

        if(new_sel && new_sel !== sel) {
            if(this.getSelectorBadnessScore(new_sel) < this.getSelectorBadnessScore(sel)) {
                return new_sel;
            }
        }
        return null;
    }

    this.getSelectorHtml = function(sel) {
        sel = sel.replace(/\> /g, '>');
        var a = sel.split(' ');
        for(var i = 0; i < a.length; i++) {
            a[i] = a[i].replace(/(>)?([^#.]+)?(#[^.]+)?(\..+)?/g, function(m, p, tag, id, cls) {
                var html = '';
                if(p) {
                    html += `<child>&gt; </child>`;
                }
                if(tag) {
                    html += `<nametag>${tag}</nametag>`;
                }
                if(id) {
                    html += `<nameid>${id}</nameid>`;
                }
                if(cls) {
                    html += `<nameclass>${cls}</nameclass>`;
                }
                return html;
            })
        }
        return `<span class="pg-selector">${a.join(' ')}</span>`;
    }

    this.getActionSetName = function(element) {
        var n = element.getAttr('name') || '&lt;no name&gt;';
        var sel = element.getAttr('selector');
        if(sel) {
            n += ' - ' + sel;
        }
        var actions = element.find('action');
        if(actions.length) {
            n += ' - ' + actions[0].getAttr('name');
            if(actions.length > 1) {
                n += ` +${actions.length - 1}`
            }
        }
        return n;
    }

    this.getActionsForMenu = function(pgel, func) {
        var r = []

        var list = pgel.getUniqueSelectors(0, 100);

        if(list.length) {
            list.forEach(function(p) {
                var orig_sel = p;
                p = p.replace(/\> /g, '>');

                var parts = p.split(' ');
                var html = '';

                var parent_sel = '';
                for(var i = 0; i < parts.length; i++) {
                    parent_sel += (parent_sel ? ' ' : '') + parts[0];
                    html += `<span class="pg-sel-menu-path-part" data-sel="${parent_sel}">${_this.getSelectorHtml(parts[i])}</span>`;
                }

                r.push({
                    selector: orig_sel,
                    label: html,
                    action: function() {
                        func(p)
                    }
                })
            })
        }

        return r;
    }

    this.displayIssuesNotice = function(num) {
        pinegrow.showAlert(`<p>The project has ${num} unresolved issue${num > 1 ? 's' : ''} with selectors mapping WordPress actions to their respective HTML elements.</p><p><a href="https://pinegrow.com/docs/wordpress/selectors/" class="external">Learn how to resolve mapping issues.</a></p>`, 'Action sets mapping issues', `I'll deal with it later`, 'Go to Action mappings panel', null, function() {
            pinegrow.showUIPanel('db_tree');
            pinegrow.projectDBTreeView.setFilterOption('issues');
        })
    }

    var withoutHistory = function(pgel_page, func) {
        if(pgel_page instanceof pgParserNode) {
            pgel_page = pgel_page.getPage();
        }
        pinegrow.undoManager.getHistoryForPage(pgel_page).ignore(func)
    }

    this.mapDBElementToPgel = function(element, pgel, sel, force_new_sel) {

        withoutHistory(pgel, function() {
            if(!sel || force_new_sel) {
                sel = pinegrow.assignUniqueSelectorToPgel(pgel, false /* no change event */, force_new_sel);
            }

            _this.replaceElementSelector(element, sel, true, true /* add actions */);
        })
        pgel.getPage().setPageChanged(true);
        pinegrow.selectedElements.reselect();
    }

    var unmapDBElementFromPgelWithoutEvents = function(element, pgel) {
        var col = new PgCollection();

        withoutHistory(pgel, function() {

            if(pgel) {
                _this.removeElementActionsFromPgel(element, pgel);
            }

            element.removeAttribute('selector');
            element.removeAttribute('desc');
            element.setAttr('status', 'unused');

            col.add(element);
        })

        return col;
    }



    this.changeActionSetName = function(element, name) {
        var page = this.getOpenPageForElement(element);

        var col = new PgCollection();

        withoutHistory(page, function() {
            var mapped_pgel = _this.getMappedPgel(element);
            if (mapped_pgel) {
                mapped_pgel.removeAttr('data-pg-name')
                if(name) {
                    mapped_pgel.setAttr('data-pg-name', name)
                }
            }
            element.removeAttr('name');
            if(name) {
                element.setAttr('name', name);
            }
        })
        page.setPageChanged(true);
        getDB().changed(col);
        pinegrow.selectedElements.reselect();

    }

    this.unmapDBElementFromPgel = function(element, pgel) {
        var col = unmapDBElementFromPgelWithoutEvents(element, pgel);

        var page = (pgel ? pgel.getPage() : this.getOpenPageForElement(element)) || pinegrow.getSelectedPage();

        withoutHistory(page, function () {
            if(pgel) {
                pinegrow.changeManager.didChange(pgel, 'Remove action');
            } else {
                pinegrow.changeManager.didChangePage(page, 'Remove action');
            }
        })

        page.setPageChanged(true);
        getDB().changed(col);
        pinegrow.selectedElements.reselect();
    }

    this.unmapOneDBElementFromPgelAndMapAnother = function(unmap_element, map_element, pgel, with_sel) {
        var col = new PgCollection();
        col.add(unmap_element);
        col.add(map_element);

        unmapDBElementFromPgelWithoutEvents(unmap_element, pgel)

        this.mapDBElementToPgel(map_element, pgel, with_sel)

        getDB().changed(col);
    }

    this.makeElementUnused = function(element) {
        var mapped_pgel = this.getMappedPgel(element);
        this.unmapDBElementFromPgel(element, mapped_pgel);
    }

    this.deleteElement = function(element) {
        var page = this.getOpenPageForElement(element);
        withoutHistory(page, function() {
            var mapped_pgel = _this.getMappedPgel(element);
            if (mapped_pgel) {
                unmapDBElementFromPgelWithoutEvents(element, mapped_pgel)
            }
            element.remove();
        });
        page.setPageChanged(true);
        getDB().changed();
    }

    var updateStatus = function(element, page) {
        if(page) {
            var list = page.sourceNode.find(element.getAttr('selector'), false, false, function(node) {
                if(node.hasAttr('wp-preview') && !node.hasAttr('data-pg-sel')) {
                    return true; //just preview
                }
                return false;
            }, null)
            var status = 'ok';
            if(list.length === 0) {
                status = 'none'
            } else if(list.length > 1) {
                status = 'multiple';
            }
            element.setAttr('status', status)
        } else {
            element.setAttr('status', 'ok')
        }
    }

    this.addElementActionsToPgel = function(element, pgel) {
        //remove all actions first
        //a quick hack
        pgel.removeAttrIfStartsWith('wp-');
        pgel.removeAttrIfStartsWith('cms-');

        element.find('action').forEach(function(a) {
            a.getAttrList().forEach(function(attr) {
                if(attr.name !== 'id' && attr.name !== 'name') {
                    pgel.setAttribute(attr.name, attr.value, true, attr.quote)
                }
            })
        })
    }

    this.removeElementActionsFromPgel = function(element, pgel) {
        //this will do
        //a quick hack
        //pgel.removeAttrIfStartsWith('wp-');
        //pgel.removeAttrIfStartsWith('cms-');


        element.find('action').forEach(function(a) {
            a.getAttrList().forEach(function(attr) {
                if(pgel.hasAttr(attr.name)) {
                    pgel.removeAttribute(attr.name)
                }
            })
        })
    }

    this.replaceElementSelector = function(element, sel, replace_instances, add_actions) {
        var current_sel = element.getAttr('selector');
        var doc = element.closest('document');
        var name = element.getAttr('name')

        var changed_pgels = new PgCollection();

        //pgParserNode.prototype.withoutEvents(function() {

        var page = pinegrow.getCrsaPageByUrl(pinegrow.getCurrentProject().getAbsoluteUrl(doc.getAttr('url')));

        var col = new PgCollection();

        withoutHistory(page, function() {

            if(page) {
                if(current_sel) {
                    var current_pgel = _this.getMappedPgel(element);
                    //var current_pgel = page.sourceNode.findOne(current_sel);
                    if (current_pgel) {
                        current_pgel.removeAttr('data-pg-sel');
                        current_pgel.removeAttr('data-pg-name');
                        _this.removeElementActionsFromPgel(element, current_pgel);
                        changed_pgels.add(current_pgel)
                    }
                }

                var pgel = page.sourceNode.findOne(sel)
            }

            element.setAttr('selector', sel);
            updateStatus(element, page);

            if(pgel) {
                pgel.setAttr('data-pg-sel', sel);
                if(name) {
                    pgel.setAttr('data-pg-name', name);
                }

                if(add_actions) {
                    _this.addElementActionsToPgel(element, pgel);
                }
                changed_pgels.add(pgel)
            }

            col.add(element);

            if(current_sel) {
                doc.find('element').forEach(function (node) {
                    var s = node.getAttr('selector');
                    if (s && s.startsWith(current_sel + ' ')) {
                        var new_s = s.replace(current_sel + ' ', sel + ' ');
                        node.setAttr('selector', new_s);
                        node.setAttr('status', 'ok');

                        col.add(node);

                        if(page) {
                            var current_pgel = page.sourceNode.findOne(s);
                            if(current_pgel) {
                                //is this element already mapped to another action set?
                                var another_as = getDB().getElementNodeForPgel(current_pgel, page);
                                if(another_as) {
                                    //have to make it unused, otherwise will be lost
                                    unmapDBElementFromPgelWithoutEvents(another_as, current_pgel);
                                    col.add(another_as);
                                }

                                current_pgel.removeAttr('data-pg-sel');
                                _this.removeElementActionsFromPgel(node, current_pgel);
                                changed_pgels.add(current_pgel)
                            }

                            var pgel = page.sourceNode.findOne(new_s);
                            if(pgel) {
                                pgel.setAttr('data-pg-sel', new_s)
                                var name = node.getAttr('name');

                                if(name) {
                                    pgel.setAttr('data-pg-name', name);
                                }

                                changed_pgels.add(pgel)

                                if(add_actions) {
                                    _this.addElementActionsToPgel(node, pgel);
                                }
                            }
                        }
                        updateStatus(element, page);
                    }
                })
            }
          //  })
            if(changed_pgels.getLength() > 0) {
                pinegrow.changeManager.didChange(changed_pgels, 'Replace selector')
            }
        })
        getDB().changed(col);
    }

    this.openDocument = function(doc, done) {
        var url = pinegrow.getCurrentProject().getAbsoluteUrl(doc.getAttr('url'));

        pinegrow.openOrShowPage(url, function(page) {
            if(done) done(page)
        }, true)
    }

    this.getOpenPageForElement = function(element) {
        var doc = element.closest('document');
        var url = pinegrow.getCurrentProject().getAbsoluteUrl(doc.getAttr('url'));
        return pinegrow.getCrsaPageByUrl(url)
    }

    var normalizeSelector = function(sel) {
        return sel.replace(/\s/g, '');
    }

    this.getMappedPgel = function(element) {
        var page = this.getOpenPageForElement(element);
        var sel = element.getAttr('selector');

        var pgel = null;

        if(sel) {
            sel = normalizeSelector(sel);
            page.sourceNode.walk(function (node) {
                if (node.isElement) {
                    var node_sel = node.getAttr('data-pg-sel');
                    if(node_sel) {
                        if(sel === normalizeSelector(node_sel)) {
                            pgel = node;
                            return false;
                        }
                    }
                }
                return true;
            })
        }
        return pgel;
    }

    this.getPgelTargetedBySelector = function(element) {
        var page = this.getOpenPageForElement(element);
        var sel = element.getAttr('selector');

        return sel ? page.sourceNode.findOne(sel) : null;
    }

    this.selectMappedPgel = function(element, done) {
        var doc = element.closest('document');

        var select = function(page) {
            var pgel = null;

            var sel = element.getAttr('selector');

            if(sel) {
                page.sourceNode.walk(function (node) {
                    if (node.isElement && node.getAttr('data-pg-sel') === sel) {
                        pgel = node;
                        return false;
                    }
                    return true;
                })
            }

            if(pgel) {
                pinegrow.selectElement(pgel);
                pgel.scrollTo();
                return pgel;
            } else {
                pinegrow.showQuickError('The action set is not used on the page.')
            }
            return null;
        }

        this.openDocument(doc, function(page) {
            var r = select(page);
            if(done) done(r)
        })
    }

    this.selectMatchedPgels = function(element, done) {
        var doc = element.closest('document');

        var select = function(page) {
            var sel = element.getAttr('selector');

            var pgels = sel ? page.sourceNode.find(sel) : [];
            if(pgels.length) {
                var col = new PgCollection();
                for(var i = 0; i < pgels.length; i++) {
                    if(i === 5) {
                        pinegrow.showQuickMessage(`...and ${pgels.length - i} more elements.`);
                        break;
                    } else {
                        col.add(pgels[i]);
                    }
                }
                pinegrow.selectedElements.selectMany(col, true);
                col.getFirst().scrollTo();
                return pgels;
            } else {
                pinegrow.showQuickError('The action set is not used on the page.')
            }
            return null;
        }

        this.openDocument(doc, function(page) {
            var r = select(page);
            if(done) done(r)
        })
    }

    this.selectAction = function(action, done) {
        var element = action.closest('element');
        this.selectMappedPgel(element, function(pgel) {
            if(pgel) pinegrow.showUIPanel('cms');
            if(done) done(pgel);
        })
    }

    this.isAutoSelectorOn = function() {
        return pinegrow.getSetting('actions-auto-selector', '1') === '1';
    }

}

var PgElementSelectorBox = function(options) {

    var _this = this;

    this.selector_value = null;

    this.auto_selector = pinegrow.getSetting('actions-auto-selector', '1') === '1';
    this.show_hints = pinegrow.getSetting('actions-selector-show-hints', '1') === '1';

    var sel_helper = new PgElementSelectorHelper();

    this.onChanged = function(term) {};

    this.$element = $('<div></div>', {class: 'pg-element-selector-box'});

    var $selector = $(`<div class="pg-element-selector"></div>`).appendTo(this.$element);

    var $selectorHint = $(`<div class="pg-element-selector-hint"></div>`).appendTo(this.$element);

    crsaHandleExternalLinks($selectorHint)

    pg$Hide($selectorHint);

    var setSelectorHint = function(val) {
        if(val) {
            $selectorHint.html(val);
            pg$Show($selectorHint);
        } else {
            pg$Hide($selectorHint);
        }
    }

    $selector.on('click contextmenu', function(e) {
        e.preventDefault();
        var actions = sel_helper.getActionsForMenu(pinegrow.selectedElements.getLastSelected(), function(sel) {
            _this.setSelectorValue(sel)
        })

        actions.forEach(function(a) {
            a.check = a.selector === _this.selector_value;
        })

        actions.push({
            type: 'divider'
        })

        actions.push({
            type: 'header',
            label: 'Options'
        })

        actions.push({
            label: 'Automatically add selector',
            check: _this.auto_selector,
            action: function() {
                _this.auto_selector = !_this.auto_selector;
                pinegrow.setSetting('actions-auto-selector', _this.auto_selector ? '1' : '0')
                pinegrow.showMode('Auto add selector', _this.auto_selector)
                pinegrow.selectedElements.reselect();
            }
        })

        actions.push({
            label: 'Show hints',
            check: _this.show_hints,
            action: function() {
                _this.show_hints = !_this.show_hints;
                pinegrow.setSetting('actions-selector-show-hints', _this.show_hints ? '1' : '0')
                pinegrow.showMode('Show hints', _this.show_hints)
                pinegrow.selectedElements.reselect();
            }
        })

        var menu = new CrsaContextMenu(actions, $selector);
        menu.onBeforeMenuShown = function() {
            menu.$menu_ul.addClass('pg-element-selector-menu')
        }
        menu.showForElement($selector, null, 8);

        return false;
    })

    var name_input = new PgInputField('name', {
        type: 'text',
        show_empty: false,
        name: 'Name',
        without_text: true,
        slider_def_unit: '',
        placeholder: 'Element name',
        validate: function(obj, fn, value) {
            if(!value && _this.show_hints) {
                return 'HINT: name the element.'
            }
        }
    });
    name_input.$field.addClass('blue-error')
    name_input.appendTo(this.$element);

    name_input.$input.on('contextmenu', function(e) {
        var actions = [
            {
                label: 'Guess a name',
                action: function() {
                    var pgel = pinegrow.selectedElements.getLastSelected();
                    if(pgel) {
                        pinegrow.assignNameToPgel(pgel, true);
                        pinegrow.selectedElements.reselect();
                    }
                }
            }
        ]
        var menu = new CrsaContextMenu(actions, name_input.$input);
        menu.showForElement(name_input.$input)

        return false;
    })

    var notice = new PgPanelNotice(true, true);
    notice.$element.appendTo(this.$element).css('margin-right', '0px');

    var onInput = function() {
        term = input.getValue();
        if(_this.onChanged) _this.onChanged(term);
        if(toolbar && !options.toolbar_on) {
            if(term.length) {
                toolbar.show();
            } else {
                toolbar.hide();
            }
        }
        if(term.length) {
            pg$Show($close_icon);
            pg$Hide($icon);
        } else {
            pg$Hide($close_icon);
        }
        delayTimer = null;
    }

    name_input.on('change', function(e) {
        _this.onNameChanged(name_input.getValue())
    })

    name_input.on('keydown', function(e) {
        if(e.which == 13) {
            _this.onNameChanged(name_input.getValue())
        }
    })

    /*
    input.on('focus', function () {
        input.$input.attr('placeholder', options.placeholder_active || options.placeholder);
        pg$Hide($icon);
    });

    input.on('blur', function () {
        input.$input.attr('placeholder', options.placeholder);
        if(term.length === 0) pg$Show($icon);
    });*/

    this.focus = function() {
        name_input.$input.get(0).focus();
    }

    this.setSelectorValue = function(val, no_event) {
        $selector.html(val ? sel_helper.getSelectorHtml(val) : 'Choose a selector&hellip;');

        this.selector_value = val;

        if(!no_event) {
            _this.onSelectorChanged(val);
        }

        if(!val && !this.auto_selector) {
            notice.show(`Element needs a selector before you can add WordPress actions. Giving it a name is also a good idea.`)
        } else {
            notice.show(null)
        }

        if(val) {
            if(sel_helper.getSelectorBadnessScore(val) > 15 && _this.show_hints) {
                setSelectorHint('HINT: Use element ids for stronger selectors. <a href="https://pinegrow.com/docs/wordpress/selectors/" class="external"><i class="icon icon-271"></i> More info&hellip;</a>')
            } else {
                setSelectorHint(null);
            }
        } else {
            setSelectorHint(null);
        }
        return;
        term = val;
        input.setValue(val);
        if(!no_event) input.$input.trigger('input');
    }

    this.setNameValue = function(val, no_event) {
        term = val;
        name_input.setValue(val);
        if(!no_event) name_input.$input.trigger('input');
    }

    this.getValue = function() {
        return name_input.getValue();
    }

    this.getInput = function() {
        return input;
    }

    var shown = true;

    this.show = function() {
        if(!shown) {
            this.$element.show();
            shown = true;
        }
    }

    this.hide = function() {
        if(shown) {
            this.$element.hide();
            shown = false;
        }
    }

    this.destroy = function() {
        if(delayTimer) {
            clearTimeout(delayTimer);
            delayTimer = null;
        }
    }
}

var PgActionsToPgelMap = function() {

    var pages = new WeakMap();

    var build = function(page) {
        var p = new CrsaProfile();
        actions_to_pgels = new WeakMap();

        if(page) {
            page.sourceNode.walk(function(node) {
                if(node.isElement) {
                    page.getAllActionTypes(node).forEach(function(a) {
                        var list = actions_to_pgels.get(a);
                        if(!list) {
                            list = [];
                            actions_to_pgels.set(a, list);
                        }
                        if(list.indexOf(node) < 0) {
                            list.push(node);
                        }
                    });
                }
                return true;
            })
        }
        pages.set(page, actions_to_pgels);
        p.show('buildActionsToPgelsMap')
        return actions_to_pgels;
    }

    this.getPgelsForAction = function(a, page) {
        var map = pages.get(page);
        if(!map) {
            map = build(page);
        }
        return map.get(a) || [];
    }

    this.getListForAction = function(a, page) {
        var map = pages.get(page);
        if(map) {
            return map.get(a);
        }
        return null;
    }

    this.actionAdded = function(a, pgel) {
        var page = pgel.getPage();
        var list = this.getListForAction(a, page);
        if(!list) {
            if(!pages.get(page)) {
                build(page);
            }
            list = [];
            pages.get(page).set(a, list);
        }
        if(list.indexOf(pgel) < 0) {
            list.push(pgel);
        }
    }

    this.actionRemoved = function(a, pgel) {
        var page = pgel.getPage();
        var list = this.getListForAction(a, page);
        if(list) {
            var idx = list.indexOf(pgel);
            if(idx >= 0) {
                list.splice(idx, 1);
            }
        }
    }

    var deletePage = function(page) {
        if(pages.has(page)) {
            pages.delete(page);
        }
    }

    var onPageChanged = function(page, info) {
        if(info && (info.action_added || info.action_removed)) return;
        deletePage(page);
    }

    /*
    pinegrow.addEventHandler('on_page_changes_done', onPageChanged);
    pinegrow.addEventHandler('on_page_closed', function(page) {
        deletePage(page);
    });*/
}

var PgQuickAddActions = function(type) {

    var h;

    if(h === undefined) {
        h = Math.floor($('body').height() * 0.75);
    }
    var allActions = new PgActionsView(type, false);

    var win = this.win = new PgQuickWindow('Quick Add Actions', allActions, 'quickAddActions', 400, h, false);

    //win.noRightEdge();
    var pgel = pinegrow.selectedElements.getLastSelected();

    if(pgel) {
        win.showForPgel(pgel);
    } else {
        win.showAtRandomPosition();
    }

    //view.quickWin = win;
    //view.setQuickWinTitle = !title;

    allActions.show();

    win.onDestroy = function() {
    }
}

var PgCombinedActionsView = function(type, icon) {

    var $element = $('<div class="crsa-panel"></div>');
    var view = new PgUIView($element, type.toLowerCase(), type == 'CMS' ? 'WP' : 'Actions');

    this.view = view;
    view.combinedActionsView = this;

    view.tabIcon = icon;

    var assignedActions = this.assignedActionsView = new PgActionsView(type, true);
    var allActions = new PgActionsView(type, false);

    allActions.assignedActionsView = assignedActions;

    var all_name = 'All actions';

    switch(type) {
        case 'CMS':
            all_name = 'All WordPress Actions';
            break;
        case 'ANIMATIONS':
            all_name = 'All Interactions';
            break;
    }

    var combo = new PgCombinedPanelView('actionsCombo' + type, assignedActions.view, allActions.view, all_name, 50);

    assignedActions.combo = combo;

    $element.append(combo.view.get$Element());

    view.onAddedToLayout = function(c) {
        combo.onAddedToLayout();
    }

    view.onResize = function() {
        combo.view.onResize();
        assignedActions.updateLayout();
        allActions.updateLayout();
    }
}

var PgActionsView = function(type, show_assigned, icon) {

    var $element = $('<div class="crsa-panel pg-lib pg-actions"></div>');

    var view = new PgUIView($element, type.toLowerCase() + (show_assigned ? '_assigned' : '_list'));

    this.view = view;

    this.panel_type = type;

    view.tabIcon = icon;

    this.pluginActivated = true;

    var itemName = 'Action';

    var use_selectors = pgf.use_selectors;
    var element_selector = null;
    var selector_attr = 'data-pg-sel';

    var show_html_when_no_el_selected = false;

    if(type == 'CMS') {
        //show_html_when_no_el_selected = true;
        itemName = 'WordPress Action'
    } else if(type === 'ANIMATIONS') {
        itemName = 'Interaction';
    }

    var assignedTitle = itemName + 's for';

    this.showAssignedActions = show_assigned;

    var open_sections_map = {};

    var active_controls = [];

    var destroyActiveControls = function() {
        active_controls.forEach(function(c) {
            c.destroy();
        })
        active_controls = [];
    }

    if(show_assigned) {
        $element.addClass('pg-actions-assigned');

        var $name = $('<div class="pg-show-properties-name"></div>').prependTo($element);
        pg$Hide($name);

        $name.on('click', 'i.copy-actions', function(e) {
            e.preventDefault();

            var r = []
            $element.find('li.crsa-action.crsa-action-on').each(function(i, li) {
                var def = getActionDef($(li));
                r.push({
                    def: def,
                    values: def.getActionValuesForPgel(selectedPgel)
                })
            })
            if(r.length) {
                pinegrow.clipboard.copySpecial('actions', r)
            }
        })
    } else {
        $element.addClass('pg-actions-list');
    }

    var showName = function() {
        view.whenVisible(function() {
            if($name) {
                if(pinegrow.selectedElements.hasSelection()) {
                    $name.get(0).innerHTML = '<span>' + assignedTitle + '</span><name>' + pinegrow.selectedElements.getName({
                        short: true,
                        def: null,
                        num_classes: -1
                    }) + '</name><i class="cmd icon icon-273 copy-actions" title="Copy all actions to clipboard"></i>';
                    pg$Show($name);
                } else {
                    pg$Hide($name);
                }
            }
        }, 'showName');
    }

    var updateComboPanels = function() {
        if(use_selectors && !element_selector.auto_selector) {
            if (selectedPgel) {
                var sel = selectedPgel.getAttr(selector_attr)
                if(sel) {
                    _this.combo.showSecondPanel();
                } else {
                    _this.combo.hideSecondPanel();
                }
            }
        }
    }

    var selectedPgel = null;

    this.getSelectedPgel = function() {
        return selectedPgel;
    }

    var currentPage = null;
    var currentFactoryLibSections = null;
    var filter = null;
    var filterRegEx = null;

    var sections_map = [];
    var actions_map = [];

    var currentTabName = type;

    var _this = this;

    var showHelp = pinegrow.getSetting('actions-show-help-text', '0') == '1';
    if (showHelp) $element.addClass('show-help-text');

    this.setShowHelp = function(v) {
        if (v) {
            $element.addClass('show-help-text');
        } else {
            $element.removeClass('show-help-text');
        }
    }

    if(this.showAssignedActions) {

        if(use_selectors) {
            $element.addClass('pg-actions-with-selectors');

            element_selector = new PgElementSelectorBox();

            element_selector.onSelectorChanged = function(sel) {

                //var sel_helper = new PgElementSelectorHelper();
                //sel_helper.replaceElementSelector()

                pinegrow.makeChanges(selectedPgel, 'Set selector', function() {
                    selectedPgel.setAttribute(selector_attr, sel);
                    selectedPgel.removeAttribute(selector_attr + '-auto');
                })

                onElementSelected();
                //_this.show();
                //updateComboPanels()
            }

            element_selector.onNameChanged = function(name) {

                pinegrow.makeChanges(selectedPgel, 'Set name', function() {
                    if(name) {
                        selectedPgel.setAttribute('data-pg-name', name);
                    } else {
                        selectedPgel.removeAttribute('data-pg-name');
                    }
                })
            }

            element_selector.$element.appendTo($element);

            //var spacer = new PgSearchBoxSpacer();
            //spacer.$element.appendTo($element);
        }

    } else {
        var search = new PgSearchBox({
            placeholder: 'search',
            placeholder_active: null,
            toolbar_on: true
        });

        var helpToggle = search.addIcon({
            icon: 'icon-tools',
            label: 'Settings &amp; tools',
            func: function (item) {

                var actions = [
                    {
                        label: 'Show help texts',
                        action: function() {
                            var v = !(pinegrow.getSetting('actions-show-help-text', '0') === '1');
                            if (v) {
                                $element.addClass('show-help-text');
                                pinegrow.setSetting('actions-show-help-text', '1');
                            } else {
                                $element.removeClass('show-help-text');
                                pinegrow.setSetting('actions-show-help-text', '0');
                            }
                            if(_this.assignedActionsView) {
                                _this.assignedActionsView.setShowHelp(v);
                            }
                            pinegrow.showMode('Show help text', v);
                        },
                        check: pinegrow.getSetting('actions-show-help-text', '0') === '1'
                    }
                ]

                pinegrow.dispatchEvent('on_actions_panel_help_menu', currentPage, _this, actions);

                var menu = new CrsaContextMenu(actions, item.$element);
                menu.showForElement(item.$element);


                /*

                var v = !item.getActive();
                item.setActive(v);
                if (v) {
                    $element.addClass('show-help-text');
                    pinegrow.setSetting('actions-show-help-text', '1');
                } else {
                    $element.removeClass('show-help-text');
                    pinegrow.setSetting('actions-show-help-text', '0');
                }
                if(_this.assignedActionsView) {
                    _this.assignedActionsView.setShowHelp(v);
                }
                pinegrow.showMode('Show help text', v);
                */
            }
        })

        //helpToggle.setActive(showHelp);

        search.$element.appendTo($element);

        search.onSearch = function (f) {
            filter = f;
            _this.show();
            _this.updateUsage();
        }

        var spacer = new PgSearchBoxSpacer();
        spacer.$element.appendTo($element);

        this.focusSearch = function () {
            search.focus();
        }
    }

    var $msg = $('<div class="alert alert-info"></div>').appendTo($element).hide();

    var notice = new PgPanelNotice();
    notice.$element.appendTo($element);

    notice.$element.on('click', '.activate-wp', function(e) {
        e.preventDefault();
        var wpf = pinegrow.findFrameworkByKey('wordpress.pinegrow');
        if(wpf) {
            pinegrow.activateFrameworkInCurrentContext(wpf, function() {

                wpf.showWordPressSiteSetupDialog();
                pinegrow.selectedElements.reselect();

            });
        }
    })

    var showMessage = function(msg) {
        if(msg) {
            notice.show(msg);
            grid.$element.hide();
            if(_this.combo && (selectedPgel || _this.combo.isStatic())) {
                var pos = notice.$element.position();

                _this.combo.enterStaticMode(pos.top + notice.$element.height() + 60);
            }
        } else {
            notice.show();
            grid.$element.show();
            if(_this.combo) {
                _this.combo.exitStaticMode();
            }
        }
    }


    var updatePgelsWithActions = function() {
        if(currentPage) {

        }
    }

    var grid = new PgUISectionGrid();
    $element.append(grid.$element);

    crsaHandleExternalLinks(grid.$element);

    var grid_sections = [];

    view.onResize = function() {
        grid.updateLayout();
    }

    this.updateLayout = function() {
        grid.updateLayout();
    }

    //var actions = new CrsaActionsPanel(type);
   // actions.showAssignedActions = show_assigned;

    var assigned_section = null;
    var assigned_section_def = {
    };

    var quick_section = null;
    var quick_section_def = {
    };

    if(this.showAssignedActions) {
        assigned_section = new PgUISection('All assigned actions', null, assigned_section_def, null, true);
    }

    this.show = function(sections, done) {

        destroyActiveControls();

        view.whenVisible(function () {

            //actions.show($element);
            currentPage = pinegrow.getSelectedPage();
            selectedPgel = pinegrow.selectedElements.getLastSelected();

            if(show_html_when_no_el_selected && currentPage && !selectedPgel) {
                selectedPgel = currentPage.sourceNode.findOne('html');
            }

            showName();

            keepScrollPosition(function() {

                pgRemoveMultiselects(grid.$element);
                grid.clear();

                _this.pluginActivated = true;

                if (type === 'CMS') {
                    var wpf = pinegrow.findFrameworkByKey('wordpress.pinegrow');
                    if (wpf && !wpf.default && currentPage && !currentPage.hasFramework(wpf)) {
                        _this.pluginActivated = false;
                        if(_this.showAssignedActions) {
                            var html = '<p>WordPress Theme Builder is not activated on this page.</p><p class="buttons"><button class="pg-button activate-wp">Activate WordPress</button></p>';
                            showMessage(html);
                        }
                        if(done) done(false)
                        return;
                    }
                }

                if(_this.showAssignedActions) {
                    var panel_info = {
                        panel: _this,
                        show_actions: true,
                        show_selector: true,
                        show_assigned_with_message: false,
                        message: null,
                        on_after: null,
                        pgel: selectedPgel,
                        page: currentPage
                    }

                    if (!selectedPgel || !currentPage) {
                        panel_info.message = 'Click on any element on the page to edit its actions or to add new HTML elements to the page.'
                        panel_info.show_actions = false;
                        panel_info.show_selector = false;
                    }

                    if(currentPage) {
                        currentPage.callFrameworkHandler('on_assigned_actions_shown', panel_info)
                    } else {
                        pinegrow.dispatchEvent('on_assigned_actions_shown', null, panel_info)
                    }

                    if(panel_info.show_actions) {
                        _this.combo.showSecondPanel();
                    } else {
                        _this.combo.hideSecondPanel();
                    }

                    if(panel_info.message) {
                        showMessage(panel_info.message);
                    } else {
                        showMessage()
                    }

                    if(element_selector) {
                        if (panel_info.show_selector) {
                            element_selector.show();
                        } else {
                            element_selector.hide();
                        }
                    }
                    if(panel_info.on_after) {
                        panel_info.on_after();
                    }
                    if(panel_info.message && !panel_info.show_assigned_with_message) {
                        if(done) done(false)
                        return;
                    }
                }

                filterRegEx = filter && filter.length > 0 ? new RegExp(escapeRegExp(filter), 'i') : null;

                if(!currentPage) return;

                if (!sections) sections = currentPage.getActionsSections();
                currentFactoryLibSections = sections;

                var displayed_sections = sections;

                grid_sections = [];

                if (_this.showAssignedActions) {
                    grid.addSection(assigned_section);
                    grid_sections.push(assigned_section);

                    if(quick_section) {
                        grid.addSection(quick_section);
                        grid_sections.push(quick_section);
                    }

                } else {
                    $.each(displayed_sections, function (i, sec_def) {
                        if (sec_def.framework.show_in_action_tab != currentTabName) return;
                        if(sec_def.show_if && !sec_def.show_if(currentPage)) return;
                        var cat_match = true;
                        if (filterRegEx) cat_match = sec_def.name.match(filterRegEx);

                        sections_map.push(sec_def);

                        var inner_html = '';

                        $.each(sec_def.getComponentTypes(), function (i, eltype) {
                            var def = eltype;
                            if (def) {
                                inner_html += getHTMLForAction(def, cat_match);
                            }

                        });

                        if (inner_html.length) {

                            var pgsec = new PgUISection(sec_def.name, sec_def.small_name || sec_def.framework.name, sec_def, null, filterRegEx, {
                                tooltip: sec_def.tooltip || null
                            });

                            pgsec.$content.get(0).innerHTML = '<ul>' + inner_html + '</ul>';

                            grid.addSection(pgsec);

                            grid_sections.push(pgsec);

                            addTooltip(pgsec.$content.find('name span[title]'), 'dummy', {
                                use_title_attr: true
                            })
                        }

                    });
                }
                grid.updateLayout();
                if(done) done(true)
            })
        })
    }

    this.updateUsage = function() {
        destroyActiveControls();

        view.whenVisible(function() {
            updateUsage();
        }, 'updateUsage', 2)
    }

    var updateUsage = function () {

        destroyActiveControls();
        var $div = grid.$element;

        value_sections = [];
        values = {};

        grid_sections.forEach(function(pgsec) {
            var has_data = false;

            pgsec.$content.find('li.crsa-action').each(function (i, a) {
                var $li = $(a);

                var def = getActionDef($li);

                if(_this.showAssignedActions) {
                    pgRemoveMultiselects($li);
                    $li.find('.action-fields').remove();
                } else {
                    /*
                    var list = pinegrow.actionsToPgelMap.getPgelsForAction(def, currentPage);
                    var $list = $li.find('> .element-list');
                    var html = '';
                    list.forEach(function(pgel) {
                        html += `<li>${pgel.getName({short: false})}</li>`;
                    })
                    $list.get(0).innerHTML = html;
                    */
                }


                if (selectedPgel) {
                    if (hasAction(selectedPgel, def) || def.action_is_default_action) {
                        $li.addClass('crsa-action-on');
                        if(_this.showAssignedActions) {
                            showFields(selectedPgel, def, $li, grid.$element);
                        }
                        has_data = true;
                    } else {
                        $li.removeClass('crsa-action-on');
                        if(_this.showAssignedActions) {
                            removeFields($li);
                        }
                    }
                } else {
                    $li.removeClass('crsa-action-on');
                }
            });

            pgsec.setHasData(has_data);
        })
    }

    var updateAssignedActions = function() {
        view.whenVisible(function() {
            if (!_this.showAssignedActions || !selectedPgel) return;
            if(!_this.pluginActivated) return;

            var selectedCrsaPage = selectedPgel.getPage();
            if (!selectedCrsaPage) return;

            if(element_selector) {
                var sel = selectedPgel.getAttribute(selector_attr);

                var auto_sel = element_selector.auto_selector;

                /*
                if(!sel) {
                    if(auto_sel) {
                        sel = pinegrow.assignUniqueSelectorToPgel(selectedPgel, true);
                    }
                }
                */
                element_selector.setSelectorValue(sel, true /* no change event */);
                element_selector.setNameValue(selectedPgel.getAttribute('data-pg-name'), true /* no change event */);

                updateComboPanels();
            }

            var used_actions = selectedCrsaPage.getAllActionTypes(selectedPgel);
            var html = '<ul>';
            var has = false;

            if(used_actions.length === 0) {
                pinegrow.dispatchEvent('on_get_default_actions', selectedCrsaPage, selectedPgel, _this, used_actions);
            }

            if (used_actions.length) {
                for (var i = 0; i < used_actions.length; i++) {
                    if (used_actions[i].framework.show_in_action_tab == currentTabName) {
                        html += getHTMLForAction(used_actions[i], false);
                        has = true;
                    }
                }
            }

            if(has) {
                showMessage();
            } else {
                if(element_selector) {
                    if(sel) {
                        showMessage(`<p>Click on one or more ${itemName}s in the list below to add them to the selected element.</p>`);
                    } else {
                        showMessage();
                    }
                } else {
                    showMessage(`<p>Click on one or more ${itemName}s in the list below to add them to the selected element.</p>`);
                }
            }
            pgRemoveMultiselects(assigned_section.$content);
            keepScrollPosition(function() {
                assigned_section.$content.get(0).innerHTML = html + '</ul>';
            })


            addTooltip(assigned_section.$content.find('name span[title]'), 'dummy', {
                use_title_attr: true
            })
        }, 'updateAssigned', 1)
    }

    var onPageSelected = function(crsaPage, force) {
        var sections = null;
        var changed = false;

        var prevSelectedPgel = selectedPgel;

        selectedPgel = crsaPage ? pinegrow.selectedElements.getLastSelectedOnPage(crsaPage) : null;

        if(selectedPgel != prevSelectedPgel) changed = true;

        currentPage = crsaPage;

        if (!crsaPage) {
            changed = true;
        } else {
            if(!changed) {
                sections = crsaPage.getActionsSections();
                if (currentFactoryLibSections == null || sections == null) {
                    changed = true;
                } else {
                    if (currentFactoryLibSections.length != sections.length) {
                        changed = true;
                    } else {
                        for (var i = 0; i < sections.length; i++) {
                            if (sections[i] != currentFactoryLibSections[i]) {
                                changed = true;
                                break;
                            }
                        }
                    }
                }
            }
        }
        if(notice && notice.isShown()) changed = true;

        if (changed || force) {
            //updateList(sections);
            _this.show(sections, function(shown) {
                if(shown) {
                    updateAssignedActions();
                    if(selectedPgel) {
                        _this.updateUsage();
                    }
                }
            });
        } else {
            updateAssignedActions();
            _this.updateUsage();
        }
    }

    var onElementSelected = function(page, info) {
        var pgel = pinegrow.selectedElements.getLastSelected();
        var selectedCrsaPage = pinegrow.getSelectedPage();

        if(show_html_when_no_el_selected && selectedCrsaPage && !pgel) {
            pgel = selectedCrsaPage.sourceNode.findOne('html');
        }

        var was_empty = selectedPgel == null;
        selectedPgel = pgel;

        if (currentPage != selectedCrsaPage) {
            onPageSelected(selectedCrsaPage);
        } else {
            if (!_this.pluginActivated || _this.showAssignedActions || was_empty || currentFactoryLibSections == null || currentFactoryLibSections.length === 0 || !selectedPgel || !grid_sections || grid_sections.length === 0) {
                _this.show(null, function(shown) {
                    if(shown) {
                        updateAssignedActions();
                        _this.updateUsage();
                    }
                });
            } else {
                updateAssignedActions();
                _this.updateUsage();
            }
        }
        showName();

        //PgQuickAddActions(type);
/*
        if(pgel) {
            console.log('Unique selectors')
            pgel.getUniqueSelectors().forEach(function(sel) {
                console.error(sel)
            })
        }*/
    }

    var onActionAdded = function(page, pgel, action) {
        if(_this.showAssignedActions) {
            updateAssignedActions();
        }
        _this.updateUsage();
        if(_this.showAssignedActions) {
            highlightAction(action);
        }
    }

    var onActionRemoved = function(page, pgel, action) {
        if(_this.showAssignedActions) {
            updateAssignedActions();
        }
        _this.updateUsage();
    }

    var onActionClicked = function(page, action, info) {
        if(info.source != _this) {
            highlightAction(action);
        }
    }

    var onProjectLoaded = function() {
        currentPage = null; //force redraw
        onElementSelected(currentPage);
    }

    pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);
    pinegrow.addEventHandler('on_page_selected', function(page) { onPageSelected(page) });
    pinegrow.addEventHandler('on_project_loaded', onProjectLoaded);
    pinegrow.addEventHandler('on_action_added', onActionAdded);
    pinegrow.addEventHandler('on_action_removed', onActionRemoved);
    pinegrow.addEventHandler('on_action_clicked', onActionClicked);

    this.refreshPanel = function() {
        currentPage = null; //force redraw
        var page = pinegrow.getSelectedPage();
        if(!_this.showAssignedActions) {
            onPageSelected(page, true)
        }
        onElementSelected(currentPage);
    }

    pinegrow.addEventHandler('on_action_panel_should_be_refreshed', this.refreshPanel);

    view.on('crsa-frameworks-changed', function() {
        _this.refreshPanel();
    })

    var hasAction = function (pgel, def) {
        if (def.has_action) {
            return def.has_action(pgel);
        }
        return def.attribute && pgel.hasAttr(def.attribute);
    }

    var getHTMLForAction = function (def, cat_match) {
        var html = '';
        var extraname = '';
        var helptext = '';
        var tags = '';
        var tags_array = null;
        var icons = '';

        icons = '';

        if (def.framework && def.framework.get_preview_for_action && _this.showAssignedActions) {
            // TODO: @add-icon magic icon (fa-magic)
            if(isApp()) {
                icons += '<i class="icon icon-WP action-magic only-active" title="Import WordPress content for this element."></i>';
            }
            icons += '<i class="icon icon-KODA show-code" title="Preview the PHP code."></i>';
        }
        if(def.on_get_action_icons && _this.showAssignedActions) {
            def.on_get_action_icons(def).forEach(function(i) {
                icons = `<i class="icon ${i.icon} action-icon" title="${i.title}"></i>` + icons;
            })
        }

        icons += '<i title="Remove action from the element" class="icon icon-bin remove-action only-active" title="Remove the action from the element."></i>';

        if(show_assigned) {
            icons = '<i title="Copy action to clipboard" class="icon icon-273 copy-action"></i>' + icons;

            if(def.meta && def.meta.helplink) {
                var help_icon = '<i title="Learn how to use this action..." class="icon icon-271 help-action"></i>';
                if(!isApp()) {
                    help_icon = `<a href="${def.meta.helplink}" target="_blank" style="text-decoration: none;margin-right: 8px;">${help_icon}</a>`;
                }
                icons = help_icon + icons;
            }
        }

        icons = '<icons>' + icons + '</icons>';

        if (def.meta) {
            extraname = def.meta.extra_name ? '<small> / ' + def.meta.extra_name + '</small>' : '';
            helptext = def.meta.helptext ? def.meta.helptext : '';
            if (def.meta.helplink) {
                helptext += '<a href="' + def.meta.helplink + '" class="external" target="_blank"><i class="icon icon-271"></i></a>';
                //helptext += ' <a href="' + def.meta.helplink + '" class="external">Learn more...</a>';
            }
            if (def.meta.tags && def.meta.tags.length) {
                tags = def.meta.tags.join(' ');
                tags_array = tags;
            }
        }

        if (filterRegEx) {
            if (!cat_match && !def.name.match(filterRegEx) && !extraname.match(filterRegEx) /*&& !tags.match(filterRegEx) */) {
                return '';
            }
        }
        //if (!matchTags(tags_array, selected_tags)) return '';

        //var $item = $('<li/>', { 'class' : 'crsa-action crsa-action-on'}).html('<name>' + def.name + extraname + icons + '</name>');

        //$item.data('crsa-factory-def', def); //!!!!!

        actions_map.push(def);

        var action_name = def.name;

        if(!_this.showAssignedActions && def.short_name) {
            action_name = def.short_name;
        }

        var tooltip = '';

        if(/*!show_assigned &&*/ def.meta && def.meta.helptext) {
            tooltip = ` title="${def.meta.helptext.replace(/"/g, "'")}`;
            tooltip += show_assigned ? '' : '<br><br>Click to add to the selected element.';
            tooltip += '"';
        }

        html += '<li data-action-index="' + (actions_map.length - 1) + '" data-action="' + def.type + '" class="crsa-action"><name>' + '<span' + tooltip + '>' + action_name + '</span>' + extraname + icons + '</name>';

        var has_this_action = false;

        if (selectedPgel) {
            if (hasAction(selectedPgel, def)) {
                has_this_action = true;
            }
        }

        /*
        if(!_this.showAssignedActions) {
            html += '<ul class="element-list"></ul>';
        }*/

        if (tags.length) {
            //html += '<div class="action-tags">' + tags + '</div>';
        }
        if (helptext.length) {
            html += '<p class="help-text">' + helptext + '</p>';
        }

        html += '</li>';
        //$item.appendTo($dest);
        return html;
    }

    var onInputValueUpdated = function ($field, $input, value, fdef) {
        var $parent = $input.parent();
        if (value) {
            //$parent.addClass('has-value');
            $field.addClass('has-value');
        }
        else  {
            //$parent.removeClass('has-value');
            $field.removeClass('has-value');
        }

        if ($field.hasClass('crsa-field-color')) {
            var $colorPlaceholder = $field.find('.color-value-placeholder');
            var spectrumVal = getStylePropWithoutImportant(value);

            var $colorInput = $input.data('color-input');
            if ($colorInput && $colorInput.length > 0) {
                $colorInput.spectrum("set", spectrumVal, false);
            }

            $colorPlaceholder.css('background-color', 'black')
                .css('background-color', spectrumVal);
        }

        if (fdef.type == 'image' || fdef.file_picker) {
            $input.trigger('pg-on-value-updated');
        }

        if (fdef.rich) {
            var $crsaSelectVal = $input.siblings('.crsa-select-val');
            var crsaSelect = $crsaSelectVal.data('crsa-select');
            crsaSelect.paintVal();
        }
    }

    var value_sections = []; //keep this global
    var values = {};

    function showFields(selectedElement, def, $item, $scrollParent, show_repeat_item_idx) {
        show_repeat_item_idx = show_repeat_item_idx || 0;

        var def_sections = def.sections ? Object.values(def.sections) : [];

        if (def.on_before_show_fields) {
            def.on_before_show_fields(currentPage, def_sections /* value_sections */, selectedElement, def);
        }

        def_sections.forEach(function(s) {
            if(value_sections.indexOf(s) < 0) {
                value_sections.push(s);
            }
        })

        //var values = {};

        //Pre-create custom cotrols so that definition is done before we get values
        $.each(value_sections, function(i, s) {
            $.each(s.fields, function (fn, fdef) {
                if (fdef.type === 'custom') {
                    if (fdef.control && !fdef.control_defined) {
                        var control = fdef.control();
                        if(!s.component_definition) {
                            s.component_definition = {};
                        }
                        var section = control.define(s.component_definition);
                        if(section) value_sections.push(section);
                        fdef.control_defined = true;
                    }
                }
            })
        })


        //try {
        var def_values = selectedElement ? getValuesForObject(selectedElement, def_sections) : {};
        $.each(def_values, function(key, val) {
            values[key] = val;
        })
        //}
        //catch (err) {
            //console.log(err);
       // }

        var getValues = function(pgel) {
            return pgel === selectedElement ? values : getValuesForObject(pgel, value_sections);
        }

        _this.refreshValuesForSelectedElement = function() {
            if(selectedElement) {
                var v = getValuesForObject(selectedElement, value_sections);
                $.each(v, function(key, val) {
                    values[key] = val;
                })
            }
        }

        if(def.on_action_after_get_values) {
            def.on_action_after_get_values(selectedElement, def, values);
        }

        var $props = $('<div/>', {class: 'action-fields'});

        var count = def_sections.length;

        var on_fields_created = [];

        $.each(def_sections, function (i, s) {
            if ((s.hasOwnProperty("show") && !s.show) || !s.name) return true;

            if (i > 0 && !s.no_section_label) {
                $('<div class="crsa-field crsa-field-label"><label>' + s.name + '</label></div>').appendTo($props);
                //$('<h3/>').html(s.name).appendTo($props);
            }

            var $c = $('<div/>').appendTo($props);
            var has_repeat_items = false;

            var add_repeat_item_label = 'Add field';

            var createFields = function (fields) {
                var $addto = $c;
                var $prev_addto = null;
                var in_section = false;
                var section_has_value = false;
                var $section = null;
                var $repeat_item = null;

                var ri = 0;
                var field_list;

                $.each(fields, function (fn, fdef) {
                    var $field;
                    var field_in_section = false;
                    var $s = null;

                    if(fdef.start_repeat_item) {
                        if(def.max_repeat_items && ri > show_repeat_item_idx && $repeat_item && !$repeat_item.hasClass('has-value')) {
                            //the previous section was empty, we can stop painting sections
                            return false;
                        }
                        if(fdef.add_repeat_item_label) {
                            add_repeat_item_label = fdef.add_repeat_item_label;
                        }
                        ri++;
                        $addto = $('<div class="pg-action-repeat-item" data-repeat-item="' + ri + '"><h5>' + fdef.start_repeat_item + '</h5></div>').appendTo($c);
                        has_repeat_items = true;

                        if(fdef.reorder_repeat_items) {
                            $(`<div class="reorder-repeat-item">
<i class="icon icon-up repeat-item-move up"></i><i class="icon icon-down repeat-item-move down"></i>
</div>`).appendTo($addto.find('h5'));
                        }

                        if(ri > 1 || fdef.delete_first_repeat_item) {
                            $('<i class="icon icon-bin remove-repeat-item"></i>').appendTo($addto);
                        }
                        field_list = [];
                        $addto.data('repeat-item-fields', field_list);
                        $addto.data('repeat-item-fdef', fdef);

                        if(fdef.repeat_item_class) {
                            $addto.addClass(fdef.repeat_item_class);
                        }

                        if(ri <= show_repeat_item_idx) {
                            $addto.addClass('has-value');
                        }

                        $repeat_item = $addto;
                        in_section = false;
                    }

                    if(fdef.end_section) {
                        $addto = $prev_addto;
                        in_section = false;
                    }

                    if(fdef.start_section) {
                        $prev_addto = $addto;

                        var section_key = `${fn}.${fdef.start_section}`;

                        $s = $(`<div class="pg-action-section ${open_sections_map[section_key] ? 'open' : ''}"><h6><i class="icon icon-right"></i><i class="icon icon-down"></i> ${fdef.start_section}</h6><div class="fields"></div></div>`).appendTo($addto);
                        $addto = $s.find('.fields');

                        in_section = true;
                        section_has_value = false;

                        $s.on('click', 'h6', function(e) {
                            e.preventDefault();

                            if($s.hasClass('open')) {
                                $s.removeClass('open');
                                open_sections_map[section_key] = false;
                            } else {
                                $s.addClass('open');
                                open_sections_map[section_key] = true;
                            }
                        })

                        $section = $s;
                    }

                    field_in_section = in_section;


                    if (fdef.type === 'custom') {
                        if(fdef.control) {
                            var control = fdef.control();
                            control.actionsView = _this;
                            control.setObject(selectedElement);
                            control.setValues(values, {});
                            control.show($scrollParent);
                            $addto.append(control.$element);
                            $field = control.$element;

                            active_controls.push(control);
                        } else {
                            $field = fdef.show($addto, selectedElement, fn, fdef, values);
                        }
                    } else {
                        $field = $.fn.crsa('addInputField', $addto, selectedElement, fn, fdef, getValues /* values */, false, $scrollParent);
                    }
                    var $input = $field.find('.crsa-input');

                    if (fdef.on_fields_created) {
                        on_fields_created.push({
                            func: fdef.on_fields_created,
                            obj: selectedElement,
                            field: $field,
                            def: fdef,
                            name: fn,
                            default_value: fdef.default_value || null
                        });
                    }
                    if(field_list) field_list.push(fn);

                    var value = values.hasOwnProperty(fn) ? values[fn] : null;
                    pinegrow.validateField(selectedElement, fn, value, fdef, $field, values);

                    if(value !== null && value !== '') {
                        $addto.addClass('has-value');

                        if(field_in_section && !section_has_value && $section) {
                            $section.addClass('has-value');
                            section_has_value = true;
                        }
                    }

                    if(fdef.show_if) {
                        var show = typeof fdef.show_if == 'function' ? fdef.show_if(values, selectedElement, $field) : crsaEvaluateShowIfString(fdef.show_if, values, fdef, selectedElement);
                        if(show) {
                            pg$Show($field, 'flex');
                        } else {
                            pg$Hide($field);
                        }
                    }

                    onInputValueUpdated($field, $input, value, fdef);

                    $input.on('input change', function (e) {
                        if('value' in e.target) {
                            var val = e.target.value;
                            var $input = $(e.target);

                            if(e.target.type === 'checkbox') {
                                val = $input.is(':checked');
                            }

                            onInputValueUpdated($field, $input, val /* $input.val() */, fdef);

                            if(field_in_section) {
                                var has_value = false;
                                var $section = $field.closest('.pg-action-section');
                                $section.find('.crsa-field').each(function(i, e) {
                                    if(e.classList.contains('has-value')) {
                                        has_value = true;
                                    }
                                })
                                if(has_value) {
                                    $section.addClass('has-value');
                                } else {
                                    $section.removeClass('has-value');
                                }
                            }
                        }
                    });

                });
            }

            createFields(s.fields);

            var dynamic_fields = {};
            currentPage.callFrameworkHandler("on_show_action_fields", dynamic_fields, def, selectedElement);//todo30 sel is pgel

            createFields(dynamic_fields);

            if(has_repeat_items) {
                $c.on('click', '.remove-repeat-item', function(e) {
                    var $item = $(this).closest('.pg-action-repeat-item');
                    var idx = $item.attr('data-repeat-item');

                    var fdef = $item.data('repeat-item-fdef');
                    if(fdef.on_repeat_item_removed) {
                        fdef.on_repeat_item_removed(_this, $item, fdef, idx, values);
                    } else {
                        removeRepeatItem($item);
                    }
                })

                var $add = $(`<div class="pg-action-repeat-item-add"><button class="pg-button"><i class="icon icon-plus"></i> ${add_repeat_item_label}</button></div>`).appendTo($c);
                $add.on('click', 'button', function(e) {
                    e.preventDefault();
                    var $unused = $c.find('.pg-action-repeat-item').not('.has-value,[data-repeat-item="1"]');
                    if($unused.length) {
                        $unused.first().addClass('has-value');
                    } else {

                        var num = $c.find('.pg-action-repeat-item').length;

                        if(!def.max_repeat_items || def.max_repeat_items <= num) {
                            pinegrow.showQuickError('Sorry, no more items are available.');
                        } else {
                            keepScrollPosition(function () {
                                removeFields($item);
                                showFields(selectedElement, def, $item, $scrollParent, num + 1);
                            })
                        }
                    }
                })
            }

            $c.on('click', '.repeat-item-move', function(e) {
                var $item = $(this).closest('.pg-action-repeat-item');
                var idx = parseInt($item.attr('data-repeat-item'));
                var to_idx;
                if(this.classList.contains('up')) {
                    to_idx = idx - 1;
                } else {
                    to_idx = idx + 1;
                }
                swapRepeatItems($item, idx, to_idx);
            })

        });

        for (var i = 0; i < on_fields_created.length; i++) {
            var val = values.hasOwnProperty(on_fields_created[i].name) ? values[on_fields_created[i].name] : null;
            if (val === null && on_fields_created[i].default_value) val = on_fields_created[i].default_value;
            on_fields_created[i].func(
                on_fields_created[i].obj,
                on_fields_created[i].name,
                val,
                on_fields_created[i].def,
                on_fields_created[i].field,
                values);
        }
        on_fields_created = null;

        $props.insertAfter($item.find('>name'));
        //$item.append($props);
        return $props;
    }

    var removeFields = function ($item) {
        $item.find('>.action-fields').remove();
    }

    var getSectionDef = function ($li) {
        return sections_map[parseInt($li.attr('data-section-index'))];
    }

    var getActionDef = function ($li) {
        return actions_map[parseInt($li.attr('data-action-index'))];
    }

    var highlightAction = function(a) {
        $element.find('li.crsa-action').each(function(i, e) {
            var $a = $(e);
            if(a == getActionDef($a)) {
                pgScrollToItemInContainer($a, grid.$element, false, false, null, -10);

                pg$Flash($a, 'orange_border_flash');
                var $input = $a.find('input');
                if($input.length) $input.get(0).focus();
                return false;
            }
        })
    }

    this.highlightAction = highlightAction;

    var addEventsToList = function () {

        /*
        grid.$element.on('mouseenter', 'i.show-code', function (event) {
            if (selectedPgel) {
                var $li = $(this).closest('li');
                var def = getActionDef($li);
                var code = def.framework.get_preview_for_action(selectedPgel, def);

                var $el = crsaFuncs.createPreviewElementFromDefinition(def);
                if (!$el) {
                    pgPreview.show($li, $('<pre>' + escapeHtmlCode(code) + '</pre>'), 'cm-preview actions-preview', 300);
                } else {
                    pgPreview.show($li, $el, 'actions-preview', 300, code);
                }
            }
        });

        grid.$element.on('mouseleave', 'i.show-code', function (event) {
            pgPreview.hide();
        });
*/
        grid.$element.on('contextmenu', 'name', function (e) {

            e.stopPropagation();
            e.preventDefault();

            var $li = $(this).closest('li');
            var def = getActionDef($li);

            var $menu = new CrsaContextMenu(null, $li);

            var name = selectedPgel.getName();
            var action_name = '<b>' + def.name + '</b>';

            if (selectedPgel) {

                if (hasAction(selectedPgel, def)) {
                    $menu.add("Remove action from selected element", null, null, 'header');
                    $menu.add("Remove " + action_name + " from <b>" + name + "</b>", null, function () {
                        addAndRemoveAction($(this));
                    });
                } else {
                    $menu.add("Add action to selected element", null, null, 'header');
                    $menu.add("Add " + action_name + " to <b>" + name + "</b>", null, function () {
                        $(this).closest('name').click();
                    });
                    if (def.not_placeable && pgf.edit_html) {

                        $menu.add("", null, null, 'divider');
                        $menu.add("Can\'t create new HTML element: " + def.not_placeable, null, null, 'header');

                    }
                }
                if (def.not_placeable || !def.code || !pgf.edit_html) {

                } else {
                    var am = crsaFuncs.findActionMenuForInsertingDefIntoEl(def, selectedPgel);

                    $menu.add("", null, null, 'divider');
                    $menu.add("Create new HTML element with this action", null, null, 'header');

                    $menu.add("Prepend to <b>" + name + "</b>", null, function () {
                        crsaFuncs.insertThroughActionMenu(am, null, def, false, true);
                    });
                    $menu.add("Append to <b>" + name + "</b>", null, function () {
                        crsaFuncs.insertThroughActionMenu(am, null, def, false, false);
                    });
                    $menu.add("Insert before <b>" + name + "</b>", null, function () {
                        crsaFuncs.insertBeforeOrAfter(null, def, false, false);
                    });
                    $menu.add("Insert after <b>" + name + "</b>", null, function () {
                        crsaFuncs.insertBeforeOrAfter(null, def, false, true);
                    });
                    $menu.add("Replace <b>" + name + "</b>", null, function () {
                        crsaFuncs.replaceElement(null, def);
                    });
                }

            } else {
                if(pgf.edit_html) {
                    $menu.add('First select an existing element on the page and then use the right click to place the new element.', null, function () {
                    }, 'header');
                }
            }

            $menu.showAt(e.pageX, e.pageY, grid.$element);
            return false;
        });

        grid.$element.on('click', 'i.show-code', function (event) {
            if (selectedPgel) {
                var $li = $(this).closest('li');
                var def = getActionDef($li);
                pinegrow.hidePreview();
                event.preventDefault();
                event.stopImmediatePropagation();
                selectedPgel.getPage().callFrameworkHandler('on_action_show_code_clicked', selectedPgel, def, $li);
            }
        });

        grid.$element.on('click', 'i.remove-action', function (event) {
            addAndRemoveAction($(this), null, null, true /* only remove */);
            event.preventDefault();
            event.stopImmediatePropagation();
        });

        grid.$element.on('click', 'i.copy-action', function (event) {
            event.preventDefault();
            event.stopImmediatePropagation();

            var $li = $(this).closest('li');
            var def = getActionDef($li);

            pinegrow.clipboard.copySpecial('actions', [{
                def: def,
                values: def.getActionValuesForPgel(selectedPgel)
            }])
        });

        if(isApp()) {
            grid.$element.on('click', 'i.help-action', function (event) {
                event.preventDefault();
                event.stopImmediatePropagation();

                var $li = $(this).closest('li');
                var def = getActionDef($li);

                pinegrow.openExternalUrl(def.meta.helplink);
            });
        }

        grid.$element.on('click', 'i.action-magic', function (event) {
            if (selectedPgel) {
                var selectedCrsaPage = selectedPgel.getPage();
                var $li = $(this).closest('li');
                var def = getActionDef($li);
                event.preventDefault();
                event.stopImmediatePropagation();
                if (selectedPgel && hasAction(selectedPgel, def)) {
                    selectedCrsaPage.callFrameworkHandler('on_action_magic_clicked', selectedPgel, def, $li);
                } else {
                    var code = selectedCrsaPage.callFrameworkHandler('on_action_magic_preview_clicked', selectedPgel, def, function (code) {
                        if (code) {
                            pgPreview.show($li, code, 'actions-preview', 300);
                        }
                    });
                }
            }
        });

        grid.$element.on('click', 'i.action-icon', function (event) {
            if (selectedPgel) {
                var $i = $(this);
                var $li = $i.closest('li');
                var def = getActionDef($li);
                if (def.on_get_action_icons) {
                    def.on_get_action_icons(def).forEach(function (i) {
                        if ($i.hasClass(i.icon)) {
                            i.func(selectedPgel, def, $li);
                        }
                    })
                }
            }
            return false;
        })

        grid.$element.on('click', 'name', function (event) {
            if(event.target.tagName === 'I' || event.target.tagName === 'A') return;

            event.preventDefault();
            event.stopImmediatePropagation();
            if(selectedPgel) {
                var $target = $(this);
                var $li = $target.closest('li');
                var def = getActionDef($li);
                if(def.on_before_add_action) {
                    def.on_before_add_action(selectedPgel, def);
                }
                addAndRemoveAction($target, true, def);
            } else {
                pinegrow.showQuickError('Please select an element on the page first, then assign action to it.')
            }
        });

    }

    addEventsToList();

    var addAndRemoveAction = function ($target, only_add, def_arg, only_remove) {
        var $li = $target.closest('li');
        var def = def_arg || getActionDef($li);
        var pos_top = $li.position().top;

        //event.preventDefault();
        //event.stopImmediatePropagation();
        var collection = pinegrow.getCollection();

        var changed_collection = new PgCollection();

        var msg;
        var opname;

        var info = {};

        collection.forEachElement(function (selectedPgel, i, total) {

            if (selectedPgel) {
                var elname = selectedPgel.getName();
                var problems = new pgParserSourceProblem(selectedPgel, null, def.ignore_lock ? true : false);
                var selectedCrsaPage = selectedPgel.getPage();

                if (!selectedPgel) {
                    problems.add('element', elname, 'change');
                }
                if (!problems.ok()) {
                    showAlert(problems.toString(), "Can't edit this element");
                    return;
                }

                if (!hasAction(selectedPgel, def)) {

                    if(only_remove) return;

                    var can_add = true;
                    if (def.on_can_add_action) {
                        if (!def.on_can_add_action(selectedPgel, selectedCrsaPage, def)) {//todo30 - check handlers, last $el arg removed
                            can_add = false;
                        }
                    }
                    if(can_add) {
                        var can_add_status = {
                            can_add: true
                        }
                        selectedCrsaPage.callFrameworkHandler('on_can_add_action', selectedPgel, def, can_add_status);
                        can_add = can_add_status.can_add;
                    }
                    if (can_add) {
                        info.action_added = def;

                        if (!canMakeChange(selectedPgel, 'add_action', def)) return;

                        willMakeChange(selectedCrsaPage, "Add action / " + elname);

                        opname = "Add action / " + def.name;

                        if (!msg) {
                            msg = 'Action <b>' + def.name + '</b> was added to <b>' + collection.getName() + '</b>';
                        }

                        var modified_pgel = $.fn.crsa('addActionToElement', selectedPgel, def, selectedCrsaPage);

                        if (modified_pgel == selectedPgel || true) {
                            if(_this.showAssignedActions) {
                                showFields(selectedPgel, def, $li, $li.closest('.content'));
                            }
                            $li.addClass('crsa-action-on');
                        }

                        changed_collection.add(selectedPgel);

                        pinegrow.stats.using('action.' + (def.type) || 'unknown');
                    }
                } else {
                    if (only_add) {
                        if (i == 0) {
                            if(_this.showAssignedActions) {
                                pinegrow.showQuickMessage('Use <i class="icon icon-bin"></i> icon to remove the action from the element.', 4000, false, 'error');
                            } else {
                                selectedCrsaPage.callFrameworkHandler('on_action_clicked', def, {
                                    source: _this
                                });
                            }
                        }
                        return;
                    }

                    if (!canMakeChange(selectedPgel, 'remove_action', def)) return;

                    willMakeChange(selectedCrsaPage, "Remove action / " + def.name);

                    opname = "Remove action / " + elname;

                    info.action_removed = def;

                    var modified_pgel = $.fn.crsa('removeActionFromElement', selectedPgel, def, selectedCrsaPage);
                    if (!msg) {
                        msg = 'Action <b>' + def.name + '</b> was removed from <b>' + collection.getName() + '</b>';
                    }

                    if (modified_pgel == selectedPgel || true) {
                        if(_this.showAssignedActions) {
                            removeFields($li);
                        }
                        $li.removeClass('crsa-action-on');
                    }
                    changed_collection.add(selectedPgel);
                }
            }
        })

        /*
        if(_this.showAssignedActions) {
            if ($li.closest('.assigned-actions').length) {
                updateAssignedActions();
            } else {
                updateUsage();
            }
        }
        var new_top = $li.position().top;
        var $content = grid.$element;
        var st = $content.scrollTop();
        st = st - (pos_top - new_top);
        $content.scrollTop(st);
        */

        if(changed_collection.getLength() > 0) {
            pinegrow.changeManager.didChange(collection, opname, info);
        }
        if(msg) {
            pinegrow.showQuickMessage(msg);
        }
    }

    var keepScrollPosition = function(func) {
        var $content = grid.$element;
        var st = $content.scrollTop();
        func();
        $content.scrollTop(st);
    }

    this.scrollTo = function(el_or_sel) {
        if(typeof el_or_sel === 'string') {
            el_or_sel = grid.$element.get(0).querySelector(el_or_sel);
        }
        el_or_sel && pgScrollToItemInContainer($(el_or_sel), grid.$element, false, true);
    }

    var removeRepeatItem = function($item) {
        var $li = $item.closest('li');
        var def = getActionDef($li);
        var pos_top = $li.position().top;
        var idx = $item.attr('data-repeat-item');
        var ri_fields = $item.data('repeat-item-fields');
        var max_items = 10;
        if(def.max_repeat_items) {
            max_items = def.max_repeat_items;
        }

        var collection = pinegrow.getCollection();

        var changed_collection = new PgCollection();

        var msg;
        var opname;

        collection.forEachElement(function (selectedPgel, i, total) {

            var elname = selectedPgel.getName();
            var problems = new pgParserSourceProblem(selectedPgel, null, def.ignore_lock ? true : false);
            var selectedCrsaPage = selectedPgel.getPage();

            var can_add = true;
            if (def.on_can_add_action) {
                if (!def.on_can_add_action(selectedPgel, selectedCrsaPage, def)) {//todo30 - check handlers, last $el arg removed
                    can_add = false;
                }
            }
            if (can_add) {
                if (!canMakeChange(selectedPgel, 'add_action', def)) return;

                willMakeChange(selectedCrsaPage, "Edit action / " + elname);

                opname = "Edit action / " + def.name;

                ri_fields.forEach(function(rif) {
                    selectedPgel.removeAttribute(rif);
                })

                //compact list, dummy implementation, can be optimized
                var did_move = false;

                var start = def.delete_first_repeat_item ? 1 : 2;

                var getAttrName = function(rif, ri) {
                    if(ri === 1) {
                        return rif;
                    } else {
                        return rif + '-' + ri;
                    }
                }

                while(true) {
                    did_move = false;
                    for (var ri = 1; ri < max_items; ri++) {
                        var has = false;

                        ri_fields.forEach(function (rif) {
                            rif = rif.replace(/-[0-9]+$/, '');
                            if (selectedPgel.hasAttribute(getAttrName(rif, ri))) {
                                has = true;
                            }
                        })

                        if (!has) {
                            ri_fields.forEach(function (rif) {
                                rif = rif.replace(/-[0-9]+$/, '');
                                if (selectedPgel.hasAttribute(getAttrName(rif, ri + 1))) {
                                    selectedPgel.setAttribute(getAttrName(rif, ri), selectedPgel.getAttribute(getAttrName(rif, ri + 1)));
                                    selectedPgel.removeAttribute(getAttrName(rif, ri + 1));
                                    did_move = true;
                                }
                            })
                        }
                    }
                    if(!did_move) break;
                }

                showFields(selectedPgel, def, $li, $li.closest('.content'));

                changed_collection.add(selectedPgel);
            }
        })

        if(_this.showAssignedActions) {
            if ($li.closest('.assigned-actions').length) {
                updateAssignedActions();
            } else {
                updateUsage();
            }
        }
        var new_top = $li.position().top;
        var $content = grid.$element;
        var st = $content.scrollTop();
        st = st - (pos_top - new_top);
        $content.scrollTop(st);

        if(changed_collection.getLength() > 0) {
            pinegrow.changeManager.didChange(collection, opname);
        }
    }

    var swapRepeatItems = function($item, idx_a, idx_b) {
        var $li = $item.closest('li');
        var def = getActionDef($li);
        var pos_top = $li.position().top;
        var ri_fields = $item.data('repeat-item-fields');

        var collection = pinegrow.getCollection();

        var changed_collection = new PgCollection();

        var msg;
        var opname;

        if(idx_b < 0) {
            pinegrow.showQuickError('The item is already at the top.');
            return;
        }

        var getAttrName = function(rif, ri) {
            if(ri === 1) {
                return rif;
            } else {
                return rif + '-' + ri;
            }
        }


        collection.forEachElement(function (selectedPgel, i, total) {

            var elname = selectedPgel.getName();
            var problems = new pgParserSourceProblem(selectedPgel, null, def.ignore_lock ? true : false);
            var selectedCrsaPage = selectedPgel.getPage();

            var can_add = true;
            if (def.on_can_add_action) {
                if (!def.on_can_add_action(selectedPgel, selectedCrsaPage, def)) {//todo30 - check handlers, last $el arg removed
                    can_add = false;
                }
            }
            if (can_add) {
                if (!canMakeChange(selectedPgel, 'add_action', def)) return;

                willMakeChange(selectedCrsaPage, "Reorder action / " + elname);

                opname = "Reorder action / " + def.name;

                //check if idx_b is empty
                var has = false;

                ri_fields.forEach(function (rif) {
                    rif = rif.replace(/-[0-9]+$/, '');
                    if (selectedPgel.hasAttribute(getAttrName(rif, idx_b))) {
                        has = true;
                    }
                })

                if(!has) {
                    pinegrow.showQuickError('The item is already at the bottom.');
                    return;
                }


                var rename = function(from, to) {
                    ri_fields.forEach(function (rif) {
                        rif = rif.replace(/-[0-9]+$/, '');
                        if (selectedPgel.hasAttribute(getAttrName(rif, from))) {
                            selectedPgel.setAttribute(getAttrName(rif, to), selectedPgel.getAttribute(getAttrName(rif, from)));
                            selectedPgel.removeAttribute(getAttrName(rif, from));
                        }
                    })
                }

                rename(idx_a, 999);
                rename(idx_b, idx_a);
                rename(999, idx_b);

                showFields(selectedPgel, def, $li, $li.closest('.content'));

                changed_collection.add(selectedPgel);
            }
        })


        if(_this.showAssignedActions) {
            if ($li.closest('.assigned-actions').length) {
                updateAssignedActions();
            } else {
                updateUsage();
            }
        }
        var new_top = $li.position().top;
        var $content = grid.$element;
        var st = $content.scrollTop();
        st = st - (pos_top - new_top);
        $content.scrollTop(st);

        if(changed_collection.getLength() > 0) {
            pinegrow.changeManager.didChange(collection, opname);
        }
    }

    view.onDestroy = function() {
        _this.show(null);
        pinegrow.removeEventHandler('on_action_added', onActionAdded);
        pinegrow.removeEventHandler('on_action_removed', onActionRemoved);
        pinegrow.removeEventHandler('on_action_clicked', onActionClicked);
        pinegrow.removeEventHandler('on_project_loaded', onProjectLoaded);
    }
}