class PgClassProperties {
    constructor() {
    }

    init() {
        this.cats = []
        this.cats_map = {}
    }

    getCat(key) {
        return this.cats_map[key];
    }

    addProperty(cat_key, key, name, values) {
        var p = new PgAddClassContextMenuProperty(key, name, values);
        this.getCat(cat_key).add(p);
    }
}
class PgAddClassContextMenuItem {
    constructor(key) {
        this.key = key;
        this.menu = null;
    }

    setMenu(menu) {
        this.menu = menu;
    }
}

class PgAddClassContextMenuCategory extends PgAddClassContextMenuItem {
    constructor(key, name) {
        super(key);
        this.name = name;
        this.items = []
    }

    add(item) {
        this.items.push(item)
    }

    getAction(menu, classPreview) {
        this.setMenu(menu)
        var submenu = this.items.map(function(item) {
            return item.getAction(menu, classPreview);
        })
        return {
            label: this.name,
            submenu: submenu
        }
    }
}

class PgAddClassContextMenuProperty extends PgAddClassContextMenuItem {
    constructor(key, name, values) {
        super(key);
        this.name = name;
        this.values = values;
    }

    setValue(value) {
        this.menu.setClass(this, value)
    }

    getAction(menu, classPreview) {
        var _this = this;

        this.setMenu(menu);

        var onSet = function() {
            _this.setValue(this.key);
        }

        var onMouseEnter = function() {
            classPreview.setClass(this.key, _this.values);
        }

        var onMouseLeave = function() {
            classPreview.restore();
        }

        var submenu = this.values.map(function(value) {
            return {
                label: value.name,
                key: value.key,
                action: onSet,
                on_mouseenter: onMouseEnter,
                on_mouseleave: onMouseLeave
            }
        })
        return {
            label: this.name,
            submenu: submenu
        }
    }
}



class PgAddClassContextMenu {

    /*
    Spacing
       Margin
       Margin x
       Margin y

    Text
       Font size
     */
    constructor(pgel, class_style, prefix) {
        this.pgel = pgel;
        this.page = pgel.getPage();
        this.class_style = class_style;
        this.prefix = prefix || '';

        this.onClassSet = function() {}

        this.list = []

        var test_cat = new PgAddClassContextMenuCategory('test', 'Cool test');
        var test1 = new PgAddClassContextMenuProperty('text.size', 'Text size', [
            {key: 'text-2xl', name: '2xl'},
            {key: 'text-3xl', name: '3xl'},
            {key: 'text-4xl', name: '4xl'},
            {key: 'text-5xl', name: '5xl'},
        ])
        test_cat.add(test1)
        var test2 = new PgAddClassContextMenuProperty('koli', 'Koli', [
            {key: 'koli-1', name: 'Koli 1'},
            {key: 'koli-2', name: 'Koli 2'},
            {key: 'koli-3', name: 'Koli 3'},
        ])
        test_cat.add(test2)
        var test3 = new PgAddClassContextMenuProperty('misko', 'Misko', [
            {key: 'misko-1', name: 'Misko 1'},
            {key: 'misko-2', name: 'Misko 2'},
            {key: 'misko-3', name: 'Misko 3'},
        ])
        test_cat.add(test3)

        this.list.push(test_cat);

        pinegrow.dispatchEvent('on_get_classes_menu', this.page, this.pgel, this.list);
    }

    setClass(class_obj, value) {
        var _this = this;

        var cls = value;
        var style = this.class_style;

        //destroy here so that restore is called before the class is set
        this.classPreview.restore();

        var api = new PgApi();
        var r = api.setClassProperty(this.pgel, this.prefix + cls, class_obj.values.map(function(d) {
            return {key: _this.prefix + d.key}
        }), style);
        r && this.onClassSet(cls);

        this.classPreview.remember();

        return r;
    }

    getMenu($el) {
        var _this = this;
        var actions = []

        this.classPreview = new PgFastDOMClassPreviewForPgel(this.pgel, this.prefix);

        this.list.forEach(function(item) {
            var a = item.getAction(_this, _this.classPreview);
            actions.push(a)
        })

        var m = new CrsaContextMenu(actions, $el);

        m.keepOpenOnClick = true;

        m.onClosed = function() {
            _this.classPreview.destroy();
        }

        return m;
    }

    showFor$el($el) {
        var m = this.getMenu($el)
        m.showForElement($el);
    }

    showForEvent(event) {
        var m = this.getMenu($(event.target))
        m.showAt(event.pageX, event.pageY);
    }
}

var PgQuickAssignClass = function(pgel, $target) {

    var target_pgel = null;
    if(pgel) target_pgel = pgel;

    pgel = pgel || pinegrow.selectedElements.getLastSelected();

    if(!pgel) {
        pinegrow.showQuickMessage('Please, first select an element.');
        return;
    }

    var ac = new PgAssignClassView();

    var name = pgel ? pgel.getName({short:true}) : pinegrow.selectedElements.getName();

    var win = new PgQuickWindow('Add classes to ' + name, ac, 'quickAssignClasses', 640, 320, null, true);

    win.noRightEdge();

  /*  win.addIconToHeader('icon-duplicate', 'Pin to CODE panel', function() {

    })*/

    ac.quickWin = win;
    ac.target_pgel = target_pgel;

    if($target) {
        win.showFor$El($target);
    } else {
        win.showForPgel(pgel);
    }

    ac.onShown();

    ac.show(pgel);

    win.onDestroy = function() {
    }
}

var pgAssignClassLastSearchTerm = null;

var PgAssignClassView = function() {

    var $element = $('<div class="pg-assign-classes"></div>');

    var view = new PgUIView($element);

    this.view = view;

    this.target_pgel = null;

    var _this = this;

    var api = new PgApi();

    var current_pgel = null;

    var show_max = 100;

    var search = new PgSearchBox({
        placeholder: 'Create or search... Use ↑, ↓ and ENTER keys.',
        placeholder_active: null,
        toolbar_on: false
    });

    var toolbar = search.getToolbar();

    search.$element.appendTo($element);

    var search_term_re = null;
    var search_term = null;

    search.onSearch = function(f) {
        //_this.search(f);
        try {
            pgAssignClassLastSearchTerm = search_term = f;
            search_term_re = f.length ? new RegExp(splitTerm(f).join('|'), 'i') : null;
            updateAllList();
            //updateAssigned();
            updateRecent();
        }
        catch(err) {}
    }

    this.onShown = function() {
        search.focus();
    }



    var splitTerm = function(t, escape) {
        var r = t.split(/[\s,]+/);
        var l = [];
        r.forEach(function(cls) {
            cls = $.trim(cls);
            if(cls.length) {
                if(escape) cls = escapeRegExp(cls);
                l.push(cls);
            }
        })
        return l;
    }

    var createClass = function(term, toggle) {
        if(toggle === undefined) toggle = false;

        var list = splitTerm(term || search_term);

        var remove = false;

        if(!list.length) return;

        remove = toggle && current_pgel.hasClass(list[0]);

        list.forEach(function(cls) {
            //set event type, otherwise pg.inight will think we are still typing in input field
            if(remove) {
                api.removeClass(_this.target_pgel, cls, {event_type: 'change'});
            } else {
                api.addClass(_this.target_pgel, cls, {event_type: 'change'});
                if (addToRecent(cls)) {

                }
            }
        })
        updateRecent();

    }


    var spacer = new PgSearchBoxSpacer();
    spacer.$element.appendTo($element);

    var $createButton = $('<button class="btn btn-primary">Add class</button>').appendTo(search.$element);

    $createButton.on('click', function(e) {
        e.preventDefault();
        createClass(null, false);
    })

    var $container = $('<div class="pg-assign-classes-c"></div>').appendTo($element);

    var $all = $('<div class="pg-assign-classes-all pg-list"><p>All</p></div>').appendTo($container);

    var $assigned = $('<div class="pg-assign-classes-assigned"><p>Assigned</p></div>').appendTo($container);

    var $recent = $('<div class="pg-assign-classes-recent pg-list"><p>Recent</p></div>').appendTo($container);

    var $ul_all = $('<ul class="pg-assign-classes-list"></ul>').appendTo($all);

    var $ul_assigned = $('<ul class=""></ul>').appendTo($assigned);

    var $ul_recent = $('<ul class="pg-assign-classes-list"></ul>').appendTo($recent);

    var mover = new PgArrowsMover(search.getInput(), $ul_all, '> li');

    mover.onEnter = function($item) {
        if($item) {
            if($item.hasClass('show-more')) {
                showAll();
                return;
            }
            createClass($item.attr('data-class'), true /* toggle */);
        } else {
            createClass();
        }
    }

    var showAll = function() {
        show_max = 9999999;
        updateAllList();
    }

    var addContextMenu = function($ul) {
        new PgClassButtonHelper($ul);
    }

    var onLiClicked = function(e, is_recent) {
        e.preventDefault();
        var $li = $(e.currentTarget);

        if($li.hasClass('show-more')) {
            showAll();
            return;
        }

        var cls = $li.attr('data-class');

        if(!current_pgel) return;

        var add = true;
        if(current_pgel.hasClass(cls)) {
            api.removeClass(_this.target_pgel, cls, {source: _this});
            add = false;
        } else {
            api.addClass(_this.target_pgel, cls, {source: _this});
            addToRecent(cls);
            if(!is_recent) updateRecent();
        }
        updateHasClass(cls, add, $li);
        updateAssigned();
    }

    $ul_all.on('click', 'li', function(e) {
        onLiClicked(e, false);
    })

    $ul_recent.on('click', 'li', function(e) {
        onLiClicked(e, true);
    })

    var updateHasClass = function(cls, add, $li) {
        if(!$li) {
            $li = $ul_all.add($ul_recent).find('[data-class="' + cls + '"]');
        }
        if(add) {
            $li.addClass('has-class');
        } else {
            $li.removeClass('has-class');
        }
    }

    var updateHasClassForAll = function() {
        $ul_all.add($ul_recent).find('.has-class').removeClass('has-class');

        if(!current_pgel) return;

        var q = [];
        current_pgel.getClasses().forEach(function(cls) {
            q.push('[data-class="' + cls + '"]');
        })
        if(q.length) {
            $ul_all.add($ul_recent).find(q.join(',')).addClass('has-class');
        }
    }

    $ul_assigned.on('click', 'li', function(event) {
        var $li = $(this);
        var cls = $li.attr('data-class');

        if(!current_pgel) return;

        if(current_pgel.hasClass(cls)) {
            api.removeClass(_this.target_pgel, cls, {source: _this});
            $li.addClass('class-off');
            updateHasClass(cls, false);
        } else {
            api.addClass(_this.target_pgel, cls, {source: _this});
            $li.removeClass('class-off');
            if(addToRecent(cls, true)) {
                updateRecent();
            }
            updateHasClass(cls, true);
        }

        event.preventDefault();
        event.stopPropagation();
    })

    $ul_assigned.on('click', '.crsa-input-rule-remove', function (event) {
        var $li = $(this).parent();
        var cls = $li.attr('data-class');

        if(!current_pgel) return;

        api.removeClass(_this.target_pgel, cls);

        event.preventDefault();
        event.stopPropagation();
    });

    addContextMenu($ul_all);
    addContextMenu($ul_assigned);
    addContextMenu($ul_recent);

    var encodeClassName = function(cls) {
        if(cls.startsWith('<span')) return cls;
        return escapeHtmlCode(cls);
    }


    var updateAllList = function() {
        var all = current_pgel ? pinegrow.insight.getClassesForPage(current_pgel.getPage()) : pinegrow.insight.getClasses();

        //var alltxt = '';

        var html = '';
        var c = 0;

        for(var i = 0; i < all.length; i++) {
            if(search_term_re && !all[i].key.match(search_term_re)) {
                continue;
            }
            c++;
            var has = current_pgel && current_pgel.hasClass(all[i].key);
            html += '<li class="' + (has ? 'has-class' : '') + '" data-class="' + all[i].key + '" data-id="' + all[i].id + '">' + encodeClassName(all[i].html || all[i].name) + '</li>';

            if(show_max == c) {
                html += '<li class="show-more">' + show_max + ' results shown. Show all.</li>';
                break;
            }
        }
        //console.log(alltxt);

        mover.update('data-id', function() {
            $ul_all.get(0).innerHTML = html;
        })
    }

    var updateAssigned = function() {

        var html = '';

        if(current_pgel) {
            var all = current_pgel.getClasses();

            for (var i = 0; i < all.length; i++) {
               /* if (search_term_re && !all[i].match(search_term_re)) {
                    continue;
                }*/
                html += '<li class="pg-button pg-class-button" data-class="' + all[i] + '">' + encodeClassName(all[i]) + '<i class="icon icon-close crsa-input-rule-remove"></i></li>';
            }
        }
        $ul_assigned.get(0).innerHTML = html;
    }

    var recent_shown = false;

    var addToRecent = function(cls, skip_top) {
        return pinegrow.recentClasses.add(cls, skip_top);
    }

    var updateRecent = function() {
        var html = '';

        var recent = pinegrow.recentClasses.getRecent();

        for(var i = 0; i < recent.length; i++) {
            /* no search for recent
            if(search_term_re && !recent[i].match(search_term_re)) {
                continue;
            }*/
            if(recent[i]) {
                var has = current_pgel && current_pgel.hasClass(recent[i]);
                html += '<li class="' + (has ? 'has-class' : '') + '" data-class="' + recent[i] + '">' + encodeClassName(recent[i]) + '</li>';
            }
        }
        $ul_recent.get(0).innerHTML = html;
        recent_shown = true;
    }

    this.show = function(pgel) {
        current_pgel = pgel;

        if(false && pgAssignClassLastSearchTerm && pgAssignClassLastSearchTerm != search_term) {
            search.setValue(pgAssignClassLastSearchTerm);
        } else {
            updateAllList();
            updateRecent();
        }
        updateAssigned();
    }



    var onElementSelected = function(page, pages, info) {
        var pgel = pinegrow.selectedElements.getLastSelected();
        _this.show(pgel);
        _this.quickWin.setTitle('Add classes to ' + pinegrow.selectedElements.getName());
    }

    var onPageChanged = function(page, pgel, info) {
        if(pgel == current_pgel && (!info.source || info.source != _this)) {
            updateAssigned();
            updateHasClassForAll();
        }
    }

    view.onResize = function() {
    }

    pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);
    pinegrow.addEventHandler('on_page_changed', onPageChanged);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_selected_elements_changed', onElementSelected);
        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
    }
}

var PgRecentClasses = function() {
    ///Recent
    var recent = [];

    if(localStorage.pgRecentClasses) {
        try {
            recent = JSON.parse(localStorage.pgRecentClasses);
        } catch(err) {}
    }

    var max_recent = 100;

    this.add = function(cls, skip_top) {
        if(!cls) return false;

        var changed = false;
        var idx = recent.indexOf(cls);
        if(idx >= 0) {
            if(!skip_top) {
                recent.splice(idx, 1);
                changed = true;
            }
        } else {
            skip_top = false;
        }
        if(!skip_top) {
            recent.unshift(cls);
            changed = true;
        }
        if(recent.length > max_recent) {
            recent.splice(max_recent, recent.length - max_recent);
            changed = true;
        }
        if(changed) {
            localStorage.pgRecentClasses = JSON.stringify(recent);
        }
        return changed;
    }

    this.getRecent = function() {
        return recent;
    }

}

var PgClassButtonHelper = function($ul, pgel, source) {

    $ul.on('contextmenu', 'li', function(e) {
        var $b = $(this);
        var cls = $b.attr('data-class');
        if(cls) {
            var menu = new CrsaContextMenu(null, $(this));

            if(pgf.edit_html && pgel) {
                menu.add("Toggle", null, function () {
                    var api = new PgApi();
                    if (pgel.hasClass(cls)) {
                        api.removeClass(null, cls, {source: source});
                        $b.addClass('class-off');
                    } else {
                        api.addClass(null, cls, {source: source});
                        $b.removeClass('class-off');
                    }
                });

                if(pgel.hasClass(cls)) {
                    menu.add("Remove", null, function () {
                        var api = new PgApi();
                        api.removeClass(null, cls, {source: source});
                        $b.addClass('class-off');

                    });
                }
            }

            if(pgf.edit_css) {

                menu.add("Create CSS rule...", null, function () {
                    createClassCSSRule(cls);
                });

                menu.add("Explore CSS rules...", null, function () {
                    showClassCSS(cls);
                });
            }

            menu.showAt(e.pageX, e.pageY);
        }
        return false;
    })

    var createClassCSSRule = function(cls) {
        pinegrow.showUIPanel('style');
        pinegrow.activeCSSView.openCreateNewRuleDialog('.' + cls, null, null, cls);
    }

    var showClassCSS = function(cls) {
        pinegrow.showUIPanel('style');
        pinegrow.activeCSSView.show(null, '.' + cls);
    }


}