var PgPageLib = function() {
    var _this = this;

    this.url = null;
    this.sourceNode = null;
    this.originalSourceNode = null;
    this.pageLib = true;
    this.javascriptEnabled = true;

    this.parsedStatus = null;
    this.parsedError = null;

    this.destroy = function() {
        if(this.sourceNode) {
            this.sourceNode.withoutEvents(function () {
                _this.sourceNode.remove();
            })
        }

        if(this.originalSourceNode) {
            this.originalSourceNode.withoutEvents(function () {
                _this.originalSourceNode.remove();
            })
        }
    }

    this.handleWrapperOnPageLoad = function(statusFunc) {
        statusFunc('done');
    }

    this.getHTMLNodeForBrowserWithSupportForWrapper = function(html_full_node) {
        return html_full_node;
    }

    var whenParsedQueue = [];

    this.whenParsed = function(loadFunc, doFunc, errorFunc) {
        if(this.parsedStatus == 'done') {
            doFunc(this);

        } else if(!this.parsedStatus == 'error') {
            errorFunc(this, _this.parsedError);

        } else if(this.parsedStatus === null) {
            _this.parsedStatus = 'progress';

            whenParsedQueue.push({doFunc: doFunc, errorFunc: errorFunc});

            loadFunc(this, function(status, error) {
                _this.parsedStatus = status;
                _this.parsedError = error;

                if(_this.sourceNode) {
                    var html = _this.sourceNode.findOne('html');
                    html && html.addClass('pg-page-lib-view');
                }

                /*
                if(status == 'done') {
                    _this.originalSourceNode = _this.sourceNode;
                    _this.sourceNode = _this.originalSourceNode.clone(true, true);
                    _this.sourceNode.walk(function(node) {
                        if(node.isElement) {
                            //node.setAttr('draggable', 'true');
                        }
                        return true;
                    })
                }*/

                whenParsedQueue.forEach(function(funcs) {
                    (status == 'done' ? funcs.doFunc : funcs.errorFunc)(_this, _this.parsedError);
                })
                whenParsedQueue = [];
            })

        } else if(this.parsedStatus = 'progress') {
            whenParsedQueue.push({doFunc: doFunc, errorFunc: errorFunc});
        }
    }

    this.callGlobalFrameworkHandler = function() {}
}

var PgPageLibManager = function() {
    var map = {};

    this.add = function(url, page) {
        map[url] = page;
    }

    this.get = function(url) {
        return map[url] || null;
    }

    this.remove = function(url) {
        var p = this.get(url);
        if(p) {
            p.destroy();
            delete map[url];
        }
    }
}


var PgPageLibsTabbedView = function(quickWin) {

    var _this = this;

    var idx = -1;

    var list_def = []

    var list = [];

    var page = pinegrow.getSelectedPage();

    if(page) {
        list = page.getPageLibs();
        //page.callFrameworkHandler('on_get_page_libs', list);
    }

    if(!list.length) {
        list = list_def;
    }

    list.unshift({lib: true})

    var lib = null;
    var views = [];
    var views_map = {};

    var custom = [];

    if(localStorage.pg_custom_page_libs) {
        custom = JSON.parse(localStorage.pg_custom_page_libs);
    }

    var findCustom = function(url) {
        for(var i = 0; i < custom.length; i++) {
            if(custom[i].url == url) return i;
        }
        return -1;
    }

    var addCustom = this.addCustom = function(item) {
        var idx = findCustom(item.url);
        if(idx < 0) {
            custom.push(item);
        } else {
            custom[idx] = item;
        }
        saveCustom();
    }

    var removeCustom = function(url) {
        var idx = findCustom(url);
        if(idx >= 0) {
            custom.splice(idx, 1);
        }
        saveCustom();
    }

    var saveCustom = function() {
        localStorage.pg_custom_page_libs = JSON.stringify(custom);
    }

    var current_view = null;

    var setCurrentView = function(v) {
        if(v != current_view) {
            if(current_view) {
                pg$Hide(current_view.view.get$Element());
            }
            pg$Show(v.view.get$Element(), v == lib ? 'block' : 'block');
            current_view = v;

            pgUIViewManager.updateVisibilityInArea($content);
        }
    }


    var $element = $('<div class="pg-page-lib-tabs"><div class="pg-page-lib-tabs-content"></div></div>');

    var view = this.view = new PgUIView($element, quickWin ? null : 'library', 'Library');

    view.tabIcon = 'icon-import';

    view.onResize = function() {
        if(lib) {
            lib.view.whenVisible(function() {
                lib.view.onResize();
            }, 'resize')
        }
        views.forEach(function(pv) {
            pv.updateSize();
        })
    }

    var $content = $element.find('.pg-page-lib-tabs-content');

    var onPagelibsChanged = function() {
        var list = [];

        var page = pinegrow.getSelectedPage();

        var libs_for_page = {};

        if(page) {
            //page.callFrameworkHandler('on_get_page_libs', list);
            list = page.getPageLibs();

            list.forEach(function(i) {
                var auto_open = 'auto_open_tab' in i ? i.auto_open_tab : true;

                if(auto_open) {
                    if (!views_map[i.url]) {
                        open(i)
                    }
                    libs_for_page[i.url] = true;
                }
            })
        }

        //hide tabs that have frameworks and are not on the list
        pageTabs.getTabs().forEach(function(tab) {
            if(tab.data && tab.data instanceof PgPageLibView) {
                if(tab.data.item.framework && !libs_for_page[tab.data.item.url]) {
                    pageTabs.hideTab(tab);
                } else {
                    pageTabs.showTab(tab);
                }
            }
        })

        //Is active tab hidden?
        var activeTab = pageTabs.getActiveTab();
        if(activeTab && !pageTabs.isTabVisible(activeTab)) {
            var done = false;
            if(page && page.currentPageLibUrl) {
                var tab = pageTabs.findTabWithFilter(function(t) {
                    if(t.data && t.data.item && t.data.item.url == page.currentPageLibUrl) {
                        return true;
                    } else {
                        return false;
                    }
                })
                if(tab) {
                    pageTabs.selectTab(tab);
                    done = true;
                }
            }
            if(!done) {
                pageTabs.selectTab(pageTabs.getTabs()[0]);
            }
        }

        if(list.length) {
            $element.addClass('with-page-libs');
            $element.removeClass('without-page-libs');
        } else {
            $element.removeClass('with-page-libs');
            $element.addClass('without-page-libs');
        }
    }

    view.on('crsa-page-selected', onPagelibsChanged);
    view.on('crsa-frameworks-changed', onPagelibsChanged);

    var tabsToolbar = new PgButtonToolbar();
    $element.prepend(tabsToolbar.$element.addClass('pg-tabs-toolbar'));

    var pageTabs = new PgToolbarTabButtons(tabsToolbar, 'tabs');

    //var activeRulesTab = sheetTabs.addTab('Active', 'Active rules for the selected element', null, true);

    pageTabs.onTabSelected = function(tab) {
        var v = tab.data;
        setCurrentView(v);
        if(v instanceof PgPageLibView) {
            var page = pinegrow.getSelectedPage();
            if(page) {
                page.currentPageLibUrl = v.item.url;
            }
        } else if(v instanceof PgImageBrowserView) {
            if(!v.images_shown) {
                v.onShown();
            }
        }
    }

    pageTabs.onTabClosed = function(tab) {
        var v = tab.data;
        v.view.destroy();
        delete views_map[v.item.url];
    }

    var open = function(item, show) {
        if(views_map[item.url]) {
            pageTabs.selectTab(pageTabs.findTabWithData(views_map[item.url]));
        } else {
            var v = new PgPageLibView(item, _this);
            v.quickWin = quickWin;
            $content.append(v.view.get$Element());
            v.view.whenVisible(function () {
                v.open(item);
            }, 'openLib')
            views_map[item.url] = v;

            pg$Hide(v.view.get$Element());

            if (show) {
                setCurrentView(v);
            }

            var tab = pageTabs.addTab(item.tab_name || item.name || ('Tab'), item.url, v, show).setClosable(true);

            v.pageLibTab = tab;

            views.push(v);
        }
    }

    var close = function(url) {
        if(views_map[url]) {
            pageTabs.removeTab(pageTabs.findTabWithData(views_map[url]));
        }
    }

    tabsToolbar.addSection('icons');

    var addTabButton = new PgButtonToolbarItem('icon-plus', 'Open a page as library...');

    !pgf.kids && tabsToolbar.addItem('icons', addTabButton);

    addTabButton.onClick = function() {
        var actions = [];

        var list = [];

        var page = pinegrow.getSelectedPage();

        if(page) {
            //page.callFrameworkHandler('on_get_page_libs', list);
            list = page.getPageLibs();

            var header = null;

            list.forEach(function(i) {
                if(i.framework && header != i.framework.name) {
                    actions.push({
                        type: 'header',
                        label: i.framework.name
                    })
                    header = i.framework.name;
                }
                actions.push({
                    label: i.name,
                    action: function() {
                        open(i, true);
                    }
                })
            })
        }

        if(custom.length) {
            actions.push({
                type: 'divider'
            })

            actions.push({
                type: 'header',
                label: 'Custom'
            })
            custom.forEach(function(i) {
                actions.push({
                    label: i.name || crsaGetSummaryStr(i.url, 40, false),
                    action: function() {
                        open(i, true);
                    },
                    check: '<i class="icon icon-close"></i>',
                    check_action: function(e) {
                        removeCustom(i.url);
                        close(i.url);
                        e.preventDefault();
                        e.stopPropagation();
                        menu.close();
                    },
                    check_tooltip: 'Remove from the list'
                })
            })

            actions.push({
                type: 'divider'
            })
        }

        if(isApp()) {
            actions.push({
                type: 'header',
                label: 'Load custom page library'
            })
            /*
             actions.push({
             label: 'Open file...',
             action: function() {
             crsaChooseFile(function(url, file) {
             var item = makeItemFromUrl(url);
             open(item, true);
             addCustom(item);
             })
             }
             })*/

            actions.push({
                label: 'Open...',
                action: function () {
                    addOrEditPageLibItem(null, function (item) {
                        open(item, true);
                        addCustom(item);
                    })
                    /*
                     pinegrow.showEnterUrlPrompt(function(url) {
                     var item = makeItemFromUrl(url);
                     open(item, true);
                     addCustom(item);
                     }, 'Load library from URL')
                     */
                }
            })
        }

        if(actions.length === 0) {
            if(page) {
                pinegrow.showQuickError('No page libraries are available.')
            } else {
                pinegrow.showQuickError('Please open a page first.')
            }
        }

        var menu = new CrsaContextMenu(actions, addTabButton.$element);
        menu.showForElement(addTabButton.$element, null, 8);
    }

    var addOrEditPageLibItem = this.addOrEditPageLibItem = function(item_param, done) {
        var edit = item_param;

        var values = item_param ? cloneObject(item_param) : {};

        var d = new PgDialog(edit ? `Edit page library ${crsaGetNameFromUrl(item_param.url)}` : 'Add a new page library');

        d.values = values;
        d.okButtonLabel = edit ? 'Save changes' : 'Add';

        d.before = `<p>Page libraries let you open a local file or URL and use it as a component library from where you can drag components to the page. <a href="https://pinegrow.com/docs/master-pinegrow/page-libraries/" class="external">Learn more about Page libraries.</a></p>`;
        //d.cancelButtonLabel = `I'll do it later`;

        d.sections = {
            url: {
                name: 'Page Library information',
                is_closable: false,
                fields: {
                    url: {
                        type: 'text',
                        name: 'URL or local file',
                        file_picker: true,
                        file_picker_no_url: false,
                        validate: function (pgel, prop, value, fieldDef, $field, values) {
                            if (!value) {
                                return pinegrow.getValidationError(fieldDef.name, 'req');
                            }
                        },
                        disabled: edit || false,
                        placeholder: 'for example, https://pinegrow.com/buttons.html',
                        helptext: `<p>Select the local file or enter a URL you want to use as a page library.</p>`
                    },
                    name: {
                        type: 'text',
                        name: 'Library name',
                        placeholder: 'default, URL name',
                        helptext: `<p>Give the page library a name, for example &quot;Cool Buttons&quot; File or URL name is used by default.</p>`
                    },
                    selector: {
                        type: 'text',
                        name: 'Components selector',
                        placeholder: 'default, *',
                        helptext: `<p>A CSS selector for selecting HTML elements that can be dragged to the page, for example <code>section, header, .component</code>. By default, all elements (<code>*</code>) are draggable.</p>`
                        //validate : f.requireSlug
                    },
                    navigation_selector: {
                        type: 'text',
                        name: 'Navigation selector',
                        placeholder: 'default, none',
                        helptext: `<p>A CSS selector for navigation areas where Pinegrow should let the page handle the clicks even if test clicks is off.</p>`
                        //validate : f.requireSlug
                    }/*,
                     direct: {
                     type: 'checkbox',
                     name: 'Load the URL directly',
                     value: 'true',
                     helptext: `<p>By default, URLs are loaded through the internal web server. This can interfere with page navigation and login methods. Check this to load the URL directly. This option has no effect on local files.</p>`
                     }*/
                }
            }
        }

        d.onOK = function () {
            var item = d.values;

            item.name = item.name || crsaGetNameFromUrl(item.url);
            item.selector = item.selector || '*';
            item.direct = !crsaIsFileUrl(item.url);
            done(item);
        }

        d.show();
    }

    var addResource = function(url) {

        var r = new PgComponentTypeResource(url);

    }

    var v = lib = this.lib = new PgLibraryView(true, quickWin);
    v.quickWin = quickWin;
    $content.append(v.view.get$Element());

    v.view.whenVisible(function() {
        v.show();
    })

    if(pgQuickInsertLastFilter) {
        v.setFilter(pgQuickInsertLastFilter);
    }

    setCurrentView(v);

    pageTabs.addTab(pgf.kids ? 'Elements' : 'List', pgf.kids ? 'HTML elements that you can insert on the page.' : 'Elements and components from all active frameworks.', v, true).setClosable(false);

    if(pgf.insert_in_top_menu) {
        var images_view = new PgImageBrowserView();
        pageTabs.addTab('Images', 'Insert images to the page.', images_view, false).setClosable(false);
        $content.append(images_view.view.get$Element());
        pg$Hide(images_view.view.get$Element());
    }

    onPagelibsChanged();

    var makeItemFromUrl = function(url) {
        return {
            url: url,
            selector: '*',
            name: crsaGetNameFromUrl(url)
        }
    }
}


var PgPageLibView = function(item, page_lib_tabbed_view) {

    var _this = this;
    this.item = item;

    var $element = $('<div class="pg-page-lib"></div>');

    var view = this.view = new PgUIView($element);

    var css_spread = 'translate3d(20px, 20px, 0px) rotate3d(0,0,0, 30deg) !important;';

    var css = '\
    .pg-lib-item, .pg-lib-item-temp-draggable {\
        cursor: grab !important;\
        position: relative;\
    }\
    html:not(.pg-lib-no-transforms) .pg-lib-item, html:not(.pg-lib-no-transforms) .pg-lib-item-temp-draggable {\
        transition:all 0.1s !important;\
    }\
    html:not(.pg-lib-no-transforms) .pg-lib-item:hover, html:not(.pg-lib-no-transforms) .pg-lib-item-temp-draggable:hover {\
        transform:translate(2px, 2px) !important;\
        box-shadow:-2px -2px 8px 0px rgba(0,0,0,0.5) !important;\
    }\
    html.pg-lib-no-transforms .pg-lib-item:hover, html.pg-lib-no-transforms .pg-lib-item-temp-draggable:hover {\
        outline:4px solid #0098cc;\
    }\
    .pg-lib-item-spread {\
        //perspective:600px;\
        box-shadow:0 0 2px #039dff;\
        z-index:1;\
        margin-bottom:80px;\
    }\
    .pg-lib-item-spread * {\
        transform:' + css_spread + '\
        box-shadow:0 0 2px #039dff;\
    }\
    .pg-lib-item-spread *:hover {\
        transform:' + css_spread + '\
        box-shadow:0 0 2px #039dff;\
    }\
    [pg-draggable] *{\
        cursor: grab !important;\
    }\
    html body .pg-lib-item-locked {\
        //outline:4px solid #33add6 !important;\
        //outline-offset: -4px;\
        transform: translate(5px, 5px) !important;\
        box-shadow: -6px -6px 0px 0px #33add6 !important;\
    }\
    html body .pg-lib-item-locked:hover {\
        transform: translate(10px, 10px) !important;\
        box-shadow: -6px -6px 0px 0px #33add6, -2px -2px 8px 0px rgba(0,0,0,0.5) !important;\
    }\
    \
    ';

    css += `
.pg-lib-hide {
    display: none;
}

.pg-lib-hover-menu {
    position: absolute;
    right:-5px;
    top:-20px;
    height:20px;
    background-color: #33add6;
    z-index:999;
    white-space: nowrap;
    font-size: 11px;
    transform: scale(var(--pg-scale-to-100));
    transform-origin: 100% 100%;
    cursor: pointer !important;
    text-transform: uppercase;
    padding-top: 3px;
    padding: 3px 4px;
}

.pg-lib-hover-menu a {
    color: #ffffff;
    text-decoration: none;
}

.pg-empty-placeholder {
    min-height: 100px;
    box-shadow: 0 0 0 1px rgba(0,0,0,0.15);
    background-image:url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAYAAACp8Z5+AAAADklEQVQIW2NgQAXGZHAAGioAza6+Hk0AAAAASUVORK5CYII=);
}
`

    var $iframe;
    var $crumbs;

    var topics = []
    var topics_dropdown = null;
    var zoom_dropdown = null;
    var tocbar;
    var toolbar;

    var navBackButton;
    var navForwardButton;
    var history = []
    var history_idx = -1;

    var current_url = null;

    var test_clicks = false;
    var pick_all = item.selector === '*';

    var addHistory = function() {
        var url = $iframe.get(0).contentWindow.location.href;
        url = pinegrow.getOriginalUrl(url);

        if(history.length && history[history_idx].url === url) {
            return;
        }

        if(history.length) {
            history.splice(history_idx + 1, history.length - history_idx - 1);
        }

        history.push({
            url: url
        });

        history_idx = history.length - 1;
    }

    var updateNavigation = function() {
        if(history_idx > 0) {
            navBackButton.$element.removeClass('disabled');
        } else {
            navBackButton.$element.addClass('disabled');
        }
        if(history_idx < history.length - 1) {
            navForwardButton.$element.removeClass('disabled');
        } else {
            navForwardButton.$element.addClass('disabled');
        }
    }

    var makeUrl = function(url) {
        if(!_this.item.direct) {
            url = pinegrow.getProxyUrl(url);
            //url = crsaAppendQueryToUrl(url, ['pgpagelib=1'])
        }
        return url;
    }

    var reload = function(purl) {
        var url = purl || $iframe.get(0).contentWindow.location.href;
        url = pinegrow.getOriginalUrl(url);

        url = pgRemoveUrlParameter(url, 'pgz');
        var page = pinegrow.pageLibManager.get(url);

        if(page) {
            pinegrow.pageLibManager.remove(url);
        }

        navigateTo(url, true);
    }

    this.reload = reload;

    var navigateTo = function(url, force_reload, done) {

        var src_html = '';

        var whenReady = function() {
            current_url = url;
            var orig_url = url;
            url = makeUrl(url);

            if(force_reload && _this.item.direct) {
                url = crsaAppendQueryToUrl(url, [`pgz=${(new Date()).getTime()}`]);
            }

            if(isApp()) {
                $iframe.attr('src', url);
            } else {
                $iframe.get(0).srcdoc = src_html;
            }

            if(item.pages_toc && topics_dropdown) {
                var name = null;
                for(var i = 0; i < item.pages_toc.length; i++) {
                    if(item.pages_toc[i].url === orig_url) {
                        name = item.pages_toc[i].name;
                        break;
                    }
                }
                topics_dropdown.toolbarItem.$element.find('topic').html(name || 'Index');
            }

            done && done();
        }

        $loading.appendTo($page);

        if(!_this.item.direct) {
            var page = pinegrow.pageLibManager.get(url);

            if(!page) {
                page = new PgPageLib();
                page.name = 'PageLib';
                page.url = url;
                page.forceReload = force_reload;
                page.forcePgIds = true;

                pinegrow.pageLibManager.add(url, page);

                if(!isApp()) {
                    $.ajax({
                        url: url,
                        data: null,
                        dataType: 'text'
                    })
                        .done(function(data) {
                            var p = new pgParser();
                            p.assignIds = true;
                            p.parse(data);
                            page.sourceNode = p.rootNode;
                            var base = pgCreate(`<base href="${url}"/>`);
                            page.sourceNode.findOne('head').prepend(base);
                            src_html = page.sourceNode.toStringWithIds(true);
                            whenReady();
                        })
                        .fail(function(a) {
                            //done(null);
                        });
                    return;
                }
            } else {
                if(!isApp()) {
                    src_html = page.sourceNode.toStringWithIds(true);
                }
                whenReady();
                return;
            }
        }

        whenReady()
    }
    //addHistory(item.url);

    this.initToc = function() {
        $toc = $('<div class="pg-page-lib-tabs-toc"></div>').prependTo($element);

        //tocbar = new PgButtonToolbar();
        //tocbar.addSection('toc');

        toolbar = new PgButtonToolbar();

        $toc.append(toolbar.$element);

        //toolbar.addSection('topics');
        toolbar.addSection('nav');
        toolbar.addSection('cmd');
        toolbar.addSection('zoom');

        topics_dropdown = new PgButtonToolbarItemWithDropdown('<span><topic>' + (item.pages_toc ? item.name : '') + '</topic><i class="icon icon-down"></i></span>', 'Jump to content section');

        toolbar.addItem('nav', topics_dropdown.toolbarItem);


        //Navigation
        navBackButton = new PgButtonToolbarItem('icon-left', 'Navigate back.');

        navBackButton.onClick = function() {
            //$iframe.get(0).contentWindow.history.back();
            //return;

            if(history_idx > 0) {
                history_idx--;
                navigateTo(history[history_idx].url);
            } else {
                pinegrow.showQuickMessage('You are at the beginning of the history.')
            }
        }


        navForwardButton = new PgButtonToolbarItem('icon-right', 'Navigate forward.');

        navForwardButton.onClick = function() {
            //$iframe.get(0).contentWindow.history.forward();
            //return;

            if(history_idx < history.length - 1) {
                history_idx++;
                navigateTo(history[history_idx].url);
            } else {
                pinegrow.showQuickMessage('You are at the end of the history.')
            }
        }

        //Back goes back in page iframe also????
        if(!item.no_history) {
            toolbar.addItem('nav', navBackButton);
            toolbar.addItem('nav', navForwardButton);
        }

        updateNavigation();
        ///End nav



        //topics_dropdown.toolbarItem.$element.appendTo($toc);


        zoom_dropdown = new PgButtonToolbarItemWithDropdown('<span><i class="icon icon-zoom"></i><topic></topic></span>', 'Zoom view');

        [100, 75, 50, 25].forEach(function(z) {
            zoom_dropdown.addAction({
                label: `${z}%`,
                action: function() {
                    _this.zoomView(z);
                },
                check: function() {return z == current_zoom; }
            })
        })

        //zoom_dropdown.toolbarItem.$element.appendTo($toc);
        toolbar.addItem('zoom', zoom_dropdown.toolbarItem);

        this.zoomView(parseInt(pinegrow.getSetting('page-lib-zoom', 100)))

        var testClicksButton = new PgButtonToolbarItem('icon-touch', 'Let the page handle the clicks. Useful for navigation and interaction with the page.');

        testClicksButton.$element.addClass('page-lib-test-clicks-control');

        toolbar.addItem('cmd', testClicksButton);

        testClicksButton.onClick = function() {
            test_clicks = !this.getActive();

            this.setActive(test_clicks);

            if(test_clicks) {
                //selectElement(null);
                //highlightElement(null);
                //crsaQuickMessage('');
            } else {
                //pinegrow.showQuickMessage('Use ' + pgShowKbd('ALT') + ' + CLICK to do a test click when test clicks are disabled.', 3000);
            }
        }


        toolbar.addSection('options');

        var optionsButton = new PgButtonToolbarItem('icon-down', 'Page library options');

        toolbar.addItem('icons', optionsButton);

        /*
         Options:
         Direct url
         All
         Help
         */

        optionsButton.onClick = function() {
            var actions = [
                {
                    type: 'header',
                    label: 'Load'
                },
                {
                    label: 'Through internal web server',
                    check: function() {
                        return !item.direct;
                    },
                    action: function() {
                        item.direct = false;
                        reload();
                        pinegrow.showQuickMessage('Loading the page through the internal web server...')
                    }
                },
                {
                    label: 'Directly',
                    check: function() {
                        return item.direct;
                    },
                    action: function() {
                        if(crsaIsFileUrl(item.url)) {
                            pinegrow.showQuickError('Local page libraries can\'t be loaded directly.');
                        } else {
                            item.direct = true;
                            reload();
                            pinegrow.showQuickMessage('Loading the page directly...')
                        }
                    }
                },
                {
                    type: 'divider',
                },
                {
                    type: 'header',
                    label: 'Pick elements'
                },
                {
                    label: 'Any',
                    check: function() {
                        return pick_all;
                    },
                    action: function() {
                        pick_all = true;
                        pickAllUpdated();
                        pinegrow.showQuickMessage('Drag any element from the page. Click to select, double-click to break apart.');
                    }
                },
            ]

            actions = [];

            /*

             if(item.selector !== '*') {
             actions.push({
             label: item.selector_description || 'Only components',
             check: function() {
             return !pick_all;
             },
             action: function() {
             pick_all = false;
             pickAllUpdated();
             pinegrow.showQuickMessage('Drag components from the page.')
             }
             })
             }


             actions.push({
             type: 'divider',
             })

             */

            actions.push({
                type: 'header',
                label: 'Page library'
            })

            actions.push({
                label: 'Settings...',
                action: function() {
                    if(item.dynamic) {
                        pinegrow.showQuickError(`This page library is defined by ${item.framework ? item.framework.name : 'a plugin'} and can't be edited.`);
                    } else {
                        page_lib_tabbed_view.addOrEditPageLibItem(item, function (changed_item) {
                            page_lib_tabbed_view.addCustom(changed_item);
                            _this.item = item = changed_item;
                            _this.pageLibTab.setLabel(item.tab_name || item.name);
                            reload(item.url);
                        })
                    }
                }
            });

            /*
            actions.push({
                label: 'Create a new page...',
                action: function() {
                    pinegrow.openOrShowPage(item.url, function(page) {
                        page.localFile = null;
                        page.pageIsRemoteTemplate = crsaIsAbsoluteUrl(item.url);
                        page.source_framework = item.framework;
                    });
                }
            })*/

            actions.push({
                type: 'divider',
            })

            actions.push({
                label: 'Reload',
                action: function() {
                    reload();
                }
            });

            if(item.on_page_lib_menu) {
                item.on_page_lib_menu(actions, _this);
            }

            var menu = new CrsaContextMenu(actions, this.$element);
            menu.showForElement(this.$element);
            return false;
        }


        /*
         $toc.on('click', 'li', function (e) {
         var $li = $(this);
         var target = $li.attr('data-toc-target');
         _this.scrollToTopic(target);
         })*/
        $crumbs = $('<ul class="pg-page-lib-crumbs"></ul>');

        $crumbs.appendTo($element);

        $crumbs.on('click', 'li', function(e) {
            $crumbs.find('.locked').removeClass('locked');

            var $li = $(this);
            var pgid = $li.attr('data-pg-id');
            var $node;

            var doc = selected_element_doc || _this.getDocument();
            if(pgid) {
                $node = $(doc.body).find('[data-pg-id="' + pgid + '"]');
            } else {
                var did = $li.attr('data-pg-dyn-id');
                if(did) {
                    $node = $(doc.body).find('[data-pg-dyn-id="' + did + '"]');
                }
            }
            if($node && $node.length) {
                if(_this.elementLocked == $node.get(0)) {
                    _this.selectElement(null);
                    updateCrumbs();
                } else {
                    _this.selectElement($node, selected_element_doc);
                    $li.addClass('locked');
                }
            }
        })

        updateCrumbs();
    }

    var count_id = 0;

    var pickAllUpdated = function() {
        updateCrumbs();
        updateDraggableElements();
    }

    var selected_element_doc = null;

    var updateCrumbs = function($node) {
        var html = '';
        if($node) {
            while ($node && $node.length && !$node.is('body')) {
                var pgid = $node.attr('data-pg-id');
                if (pgid) {
                    var pgel = getPgNodeByPgId(pgid);
                    if (pgel) {
                        var sel = $node.get(0) == _this.elementLocked ? 'locked' : '';
                        html = '<li class="divider"><i class="icon icon-right"></i></li><li data-pg-id="' + pgid + '" class="' + sel + '">' + pgel.getName({short: true}) + '</li>' + html;
                    }
                } else {
                    //plain html stuff
                    var sel = $node.get(0) == _this.elementLocked ? 'locked' : '';
                    var did = 'did_' + count_id++;
                    $node.attr('data-pg-dyn-id', did);

                    html = '<li class="divider"><i class="icon icon-right"></i></li><li data-pg-dyn-id="' + did + '" class="' + sel + '">' + getElementName($node, {display_name: 'tag'}, true, true) + '</li>' + html;
                }
                $node = $node.parent();
            }
        } else {
            if(pick_all) {
                html = '<li class="instructions"><b>Drag</b> elements to the page. <b>Click</b> to show element selection path. <b>Double-click</b> to break an element apart.</li>';
            } else {
                html = '<li class="instructions"><b>Drag</b> components to the page. <b>Click</b> to break components apart.</li>';
            }
        }
        $crumbs.html(html);
        _this.updateSize();
    }

    this.elementLocked = null;

    this.selectElement = function($node, doc) {
        doc = doc || _this.getDocument();
        $(doc.body).find('.pg-lib-item').removeAttr('pg-draggable').removeAttr('draggable').removeClass('pg-lib-item');
        $(doc.body).find('.pg-lib-item-locked').removeClass('pg-lib-item-locked');

        if($node) {
            $node.attr('pg-draggable', 'true').attr('draggable', 'true').addClass('pg-lib-item');
            this.elementLocked = $node.get(0);
            $node.addClass('pg-lib-item-locked');
            selected_element_doc = doc;
        } else {
            this.elementLocked = null;
            selected_element_doc = null;
        }
    }



    var updateToc = function() {
        //pg$Hide($toc);

        if(topics_dropdown) {
            topics_dropdown.toolbarItem.hide();
        }

        _this.whenLoaded(function() {
            var $toc_ul = $toc.find('ul');
            var html = _this.getTocHtml();
            if(html) {
                //$toc_ul.get(0).innerHTML = html;
                topics_dropdown.clearActions();
                topics.forEach(function(t) {
                    topics_dropdown.addAction(t.name, function() {
                        if(item.pages_toc) {
                            navigateTo(t.id);
                        } else {
                            _this.scrollToTopic(t.id);
                        }

                    })
                })

                //pg$Show($toc, 'flex');
                if(topics_dropdown) {
                    topics_dropdown.toolbarItem.show();
                }
            }
            _this.updateSize()
        })
    }

    var current_zoom = 100;

    this.zoomView = function(z) {
        current_zoom = z;
        zoom_dropdown.toolbarItem.$element.find('topic').html(`${z}%`);
        $iframe.css({
            transformOrigin: '0 0',
            transform: `scale(${z / 100.0})`,
            width: `${100 * 100 / z}%`,
            height: `${100 * 100 / z}%`
        })
        pinegrow.setSetting('page-lib-zoom', z);

        try {
            _this.getDocument().body.style.setProperty("--pg-scale-to-100", 1.0 / (z / 100.0));
        } catch(err) {}
    }




    this.getDocument = function() {
        var iframe = $iframe.get(0);
        return iframe.contentDocument;
    }

    view.onShow = function() {
        var $body = $(_this.getDocument().scrollingElement);
        var st = $body.scrollTop();
        $body.scrollTop(st - 1);
    }

    view.onHide = function() {
        if(scrollTimer) {
            clearTimeout(scrollTimer);
            scrollTimer = null;
        }
    }

    var scrollTimer = null;

    var loaded = false;
    var when_loaded = [];

    this.whenLoaded = function(func) {
        if(loaded) {
            func()
        } else {
            when_loaded.push(func);
        }
    }


    var onScroll = function() {
        if(item.pages_toc) return;

        var $content = $(_this.getDocument().scrollingElement);
        var st = $content.scrollTop();
        var h = $content.height();
        var idx = -1;
        $content.find('.pg-lib-section').each(function(i, s) {
            var $s = $(s);
            var pos = $s.position();
            var sh = $s.height();
            if(pos.top + sh < st || pos.top > h + st) {
                //not visible
            } else {
                topics_dropdown.toolbarItem.$element.find('topic').html($s.find('.pg-lib-section-title').html() || 'No title');
                //$toc.find('.active-toc').removeClass('active-toc');
                //$toc.find('[data-toc-target="' + $s.attr('id') + '"]').addClass('active-toc');
                return false;
            }
        })
    }

    var getTopNodeOnLocation = function(node) {
        try {
            var rect = node.getBoundingClientRect();

            var p = node;
            var take = node;

            var spacer = 4;

            while (true) {
                var prect = p.getBoundingClientRect();
                if ((Math.abs(prect.left - rect.left) >= spacer || Math.abs(prect.top - rect.top) >= spacer) || Math.abs(prect.right - rect.right) >= spacer || prect.height > 2000) {
                    break;
                }
                if (p.tagName == 'BODY') break;
                take = p;
                p = p.parentNode;
            }

            return take;
        } catch(err) {
            return node;
        }
    }

    var $page;

    var updateDraggableElements = function(doc) {

        doc = doc || _this.getDocument();
        var selector = item.selector || '*';

        pick_all = item.selector === '*';

        $(doc.body).find('*').removeAttr('draggable');
        $(doc.body).find('[pg-draggable]').removeAttr('pg-draggable');
        $(doc.body).find('[pg-lib-item-temp-draggable]').removeAttr('pg-lib-item-temp-draggable');

        if(!pick_all) {
            $(doc.body).find(selector).attr('pg-draggable', 'true').attr('draggable', 'true').addClass('pg-lib-item');
        }
        if(item.navigation_selector) {
            $(doc.body).find(`${item.navigation_selector}, ${item.navigation_selector} *`).removeAttr('pg-draggable').removeAttr('draggable').removeClass('pg-lib-item');
        }
    }

    var $loading = null;

    this.open = function() {
        item = this.item;

        var url = item.url;
        var selector = item.selector;

        /*
         Moved to navigateTo()

         var page = new PgPageLib();
         page.name = 'PageLib';
         page.url = url;

         pinegrow.pageLibManager.add(url, page);
         */

        var ua = navigator.userAgent + ' PinegrowInternal PinegrowPageLib';

        $page = $('<div/>', { class : 'page-lib'}).html('<div class="content"><iframe class="content-iframe" frameborder="0" nwfaketop nwdisable nwUserAgent="' + ua + '"></iframe></div>');

        $loading = $('<div class="iframe-loading-overlay"><div><i class="icon-REFRESH icon-spin"></i>Loading, please wait...</div></div>').appendTo($page);

        $iframe = $page.find('.content-iframe');

        $page.appendTo($element);

        var focus_handler_added = false;

        $iframe.on('load', function() {
            $loading.detach();

            addHistory();
            updateNavigation();

            _this.zoomView(current_zoom);

            if(!focus_handler_added) {
                $($iframe.get(0).contentWindow)
                    .on('focus', function () {
                        //console.log('iframe focus');
                        pgAppWindowHasFocus(pgGetAppWindowForDOMElement($iframe.get(0)));
                    })
                    .on('blur', function () {
                        //console.log('iframe blur');
                        pgAppWindowLostFocus(pgGetAppWindowForDOMElement($iframe.get(0)));
                    })

                //focus_handler_added = true;
            }

            var dragMenu = null;

            var parent = null;

            var prevX = -1;
            var prevY = -1;

            var hideWinTimer = null;

            var direct = item.direct || false;

            var doc = _this.getDocument();

            crsaHandleExternalLinks($(doc.body));

            if(!item.direct) {
                //patch absolute links and svg icons
                $(doc.body).find('a').each(function(i,e) {
                    if(e.getAttribute('aahref')) {
                        //disabled svg icons
                        e.removeAttribute('onclick');
                        e.setAttribute('href', e.getAttribute('aahref'))
                        e.removeAttribute('aahref')
                    }
                    var href = e.getAttribute('href');
                    if(href && crsaIsAbsoluteUrl(href) && !e.classList.contains('external')) {
                        e.setAttribute('href', pinegrow.getProxyUrl(pinegrow.getOriginalUrl(href)));
                    }
                })
            }

            var spreadNode = function($node) {
                $node = $node.closest(selector);

                if($node.length) {
                    $node.addClass('pg-lib-item-spread');

                    $node.find('*').each(function(i,e) {
                        var $e = $(e);
                        $e.attr('pg-draggable', 'true');
                        $e.attr('draggable', 'true');
                        $e.addClass('pg-lib-item-temp-draggable');
                    })
                    //pinegrow.showQuickMessage('')
                }
            }

            var unspread = function(doc) {
                doc = doc || _this.getDocument();
                doc.querySelectorAll('.pg-lib-item-spread').forEach(function(node) {
                    node.classList.remove('pg-lib-item-spread');
                })

                doc.querySelectorAll('.pg-lib-item-temp-draggable').forEach(function(node) {
                    node.classList.remove('pg-lib-item-temp-draggable');
                    node.removeAttribute('pg-draggable');
                    node.removeAttribute('draggable');
                })
            }



            var wireHandlers = function(doc) {

                $(doc.body).find('[target]').attr('target', '');

                if(item.no_spreading) {
                    doc.documentElement.classList.add('pg-lib-no-transforms');
                }

                updateDraggableElements(doc);

                $(doc.head).append('<style>' + css + '</style>');

                doc.addEventListener("contextmenu", function (event) {

                    event.preventDefault();
                    event.stopPropagation();
                    /*
                     var $el = $(event.target);

                     var menu = new CrsaContextMenu(null, $el);

                     if(pinegrow.selectedElements.hasSelection()) {

                     helper.addInsertActions(menu, def);

                     }*/

                    return false;
                }, true, false)

                var pb = new PgPluginBuilder();

                var show_hover_menu = item.show_hover_menu || false;
                var $hover_menu = $(`<div class="pg-lib-hover-menu"><a href="#" class="replace">Replace it</a></div>`);
                var hover_menu_$c = null;

                $hover_menu.find('.replace').on('click', function(e) {
                    e.preventDefault();

                    if(!hover_menu_$c) return;

                    var d = getPgelToInsert(hover_menu_$c);

                    var sel_pgel = pinegrow.selectedElements.getLastSelected();

                    if(sel_pgel) {
                        var api = new PgApi();
                        api.transformElement(sel_pgel, d.pgel);
                    } else {
                        pinegrow.showQuickError('Please select an element first.')
                    }
                })

                doc.addEventListener("mouseover", function (event) {

                    if(test_clicks) {
                        return;
                    }

                    var node = event.target;
                    var $node = $(node);

                    if(item.navigation_selector) {
                        if($node.closest(item.navigation_selector).length) return;
                    }

                    if(pick_all) {

                        event.preventDefault();

                        if (_this.elementLocked) return;

                        if ($node.closest('.pg-lib-item-spread').length) return;

                        $(doc.body).find('.pg-lib-item').removeAttr('pg-draggable').removeAttr('draggable').removeClass('pg-lib-item');


                        $node = $(getTopNodeOnLocation(node));

                        $node.attr('pg-draggable', 'true').attr('draggable', 'true').addClass('pg-lib-item');
                    }
                })

                if(show_hover_menu) {
                    $(doc).on('mouseenter', '.pg-lib-item', function() {

                        var $c = $(this);

                        var pgel = pinegrow.selectedElements.getLastSelected();

                        if(show_hover_menu && pgel && pgel.canDelete()) {
                            //var $c = $node.closest('.pg-lib-item');
                            if($c.length) {
                                if(!$hover_menu.parent().is($c)) {
                                    $hover_menu.prependTo($c);
                                    hover_menu_$c = $c;
                                }
                            }
                        }

                    })

                    $(doc).on('mouseleave', '.pg-lib-item', function() {

                        var $c = $(this);

                        if(show_hover_menu) {
                            if($c.length) {
                                if($hover_menu.parent().is($c)) {
                                    $hover_menu.detach();
                                    hover_menu_$c = null;
                                }
                            }
                        }

                    })
                }


                doc.addEventListener("click", function (event) {

                    if(test_clicks) {
                        highlightUIElement($('.page-lib-test-clicks-control'), 'show-click');
                        return;
                    }

                    var node = event.target;
                    var $node = $(node);

                    if($node.closest('.pg-lib-hover-menu').length) return;

                    if((item.no_spreading || (item.navigation_selector && $node.closest(item.navigation_selector).length > 0))) {
                        var anode = node.closest('a');

                        if(anode && anode.hasAttribute('data-pg-lib-template')) {
                            var template_url = (new URL(anode.getAttribute('data-pg-lib-template'), pinegrow.getOriginalUrl( $iframe.get(0).contentWindow.location.href))).toString();

                            event.preventDefault();

                            pinegrow.openOrShowPage(template_url, function(page) {
                                page.localFile = null;
                                page.pageIsRemoteTemplate = crsaIsRemoteUrl(template_url);
                                page.source_framework = item.framework;
                            });

                            return false;
                        }

                        if(!item.direct && anode /* && anode.classList.contains('pg-lib-link') */) {
                            var href = anode.getAttribute('href');
                            if(href && !href.startsWith('#')) {
                                if(isApp()) {
                                    href = (new URL(href, $iframe.get(0).contentWindow.location.href)).toString();
                                } else {
                                    href = (new URL(href, current_url)).toString();
                                }
                                navigateTo(pinegrow.getOriginalUrl(href));

                                //without this bs icons dont work and lots of dmp files are generated
                                event.preventDefault();
                                event.stopPropagation();
                                event.stopImmediatePropagation();
                                return false;
                            }
                            return;
                        }
                        return;
                    }

                    if(!pick_all && item.selector && $node.closest(item.selector).length === 0) {
                        return;
                    }

                    event.preventDefault();
                    //event.stopImmediatePropagation();
                    //event.stopPropagation();


                    if (event.altKey) {
                        //return;

                        if($node.is('a')) {
                            var url = node.href;

                            //url = pinegrow.getProxyUrl(url);
                            //url = crsaAppendQueryToUrl(url, ['pgpagelib=1'])
                            $iframe.attr('src', url);
                        }
                        return;
                    }

                    var just_off = $node.is('.pg-lib-item-spread') || $node.closest('.pg-lib-item-spread').length > 0;

                    if(just_off) {
                        unspread(doc);
                        return false;
                    }

                    if(pick_all) {
                        $node = $(getTopNodeOnLocation(node));

                        if (!_this.elementLocked || $node.get(0) != _this.elementLocked) {
                            _this.selectElement($node, doc);
                            updateCrumbs($node);
                        } else {
                            _this.selectElement(null, doc);
                            updateCrumbs(null);
                        }
                    } else {
                        spreadNode($node);
                    }
                })

                if(!item.no_spread) {
                    doc.addEventListener("dblclick", function (event) {

                        if (test_clicks) {
                            highlightUIElement($('.page-lib-test-clicks-control'), 'show-click');
                            return;
                        }

                        event.preventDefault();

                        var node = event.target;

                        var $node = $(node);

                        var just_off = $node.is('.pg-lib-item-spread') || $node.closest('.pg-lib-item-spread').length > 0;

                        unspread(doc);

                        if (just_off) return;

                        spreadNode($node);
                    })
                }

                var getPgelToInsert = function($el) {
                    if($el.hasClass('pg-lib-svg-wrapper')) {
                        $el = $el.find('svg');
                    }
                    var pgid = $el.attr('data-pg-id');
                    //console.log(pgid);

                    var rest = [];

                    if(item.get_code) {
                        //pgel = pgCreate(item.get_code($el, pick_all));
                        var pgel_list = pgCreateNodeFromHtml(item.get_code($el, pick_all), true, true);
                        if(pgel_list) {
                            pgel_list = pgel_list.filter(function(pgel) {
                                return pgel.isElementOrScript;
                            })
                            if(pgel_list.length) {
                                pgel = pgel_list.shift();
                                rest = pgel_list;
                            }
                        }
                    } else {
                        if (pgid) {
                            var pgel = getPgNodeByPgId(pgid);
                            if (pgel) {
                                parent = pgel.parent;
                                pgel = pgel.clone();
                                //clear data in case it got set through a previously cloned element
                                pgel.walkSelfAndChildren(function(n) {
                                    n.clearData();
                                })
                            }
                        } else {
                            pgel = pgCreate($el.get(0).outerHTML);
                        }
                    }

                    var processPgel = function(pgel) {
                        collapseNewPgelIfNeeded(pgel);

                        if(item.on_pgel_created) {
                            item.on_pgel_created(pgel);
                        }

                        var skip_replace_images = pgel.hasClass('pg-lib-item-skip-replace-images');

                        var remove_classes = [];

                        if(item.removeClasses) {
                            remove_classes = remove_classes.concat(item.removeClasses);
                        }

                        pgel.walkSelfAndChildren(function(node) {
                            var classes = node.getClasses().slice(0);
                            classes.forEach(function(cls) {
                                if(cls.startsWith('pg-lib-item') || remove_classes.indexOf(cls) >= 0) {
                                    node.removeClass(cls);
                                }
                            })
                            node.removeAttribute('draggable');
                            node.removeAttribute('pg-draggable');
                            node.removeAttribute('data-pg-require');
                            node.removeAttribute('data-pg-require-plugins');

                            if(item.remove_class_styles) {
                                node.removeAttribute('data-pg-class-style');
                                node.removeAttribute('data-pg-class-style-inline');
                            }
                            return true;
                        })

                        if(!skip_replace_images) {
                            pb.replaceHolderImages(pgel);
                        }
                    }

                    pgel && processPgel(pgel);

                    if(rest) {
                        rest.forEach(processPgel);
                    }

                    return {pgel: pgel, rest: rest};
                }

                var dragHelper = new PgDragHelper($(doc.body), '[pg-draggable]', null, true);
                dragHelper._oAnotherDocument = $(document);//$(doc);

                dragHelper.on( 'dragStart', function(x, y, event) {
                    $.fn.crsapages('showOverlays');

                    if(hideWinTimer) {
                        clearTimeout(hideWinTimer);
                        hideWinTimer = null;
                    }

                    prevX = event.pageX;
                    prevY = event.pageY;

                    var $el = dragHelper.dragTarget;
                    var parent = null;

                    var required_resources = [];
                    var required_resources_code = [];
                    var required_plugins = [];

                    var $rel = $el;

                    var doRequire = function($rel) {
                        ($rel.attr('data-pg-require') || '').split(',').forEach(function(r) {
                            r = r.trim();
                            if(r.length) {
                                if(required_resources.indexOf(r) < 0) {
                                    required_resources.push(r);
                                }
                            }
                        });
                        ($rel.attr('data-pg-require-plugins') || '').split(',').forEach(function(r) {
                            r = r.trim();
                            if(r.length) {
                                if(required_plugins.indexOf(r) < 0) {
                                    required_plugins.push(r);
                                }
                            }
                        });
                    }

                    while($rel.length) {
                        doRequire($rel);
                        if($rel.is('html')) break;
                        $rel = $rel.parent();
                    }

                    $el.find('[data-pg-require],[data-pg-require-plugins]').each(function(i, e) {
                        doRequire($(e));
                    })

                    var doRequireCode = function($rel) {
                        required_resources_code.push({
                            code: $rel.html(),
                            detect: $rel.attr('data-pg-require-code'),
                            footer: $rel.closest('head').length === 0
                        })
                    }

                    $el.closest('html').find('[data-pg-require-code]').each(function(i,e) {
                        doRequireCode($(e));
                    })

                    var d = getPgelToInsert($el);
                    var pgel = d.pgel;
                    var rest = d.rest;


                    if(pgel) {



                        //processPgel(pgel);

                        //if(rest) {
                        //    rest.forEach(processPgel);
                        //}

                        var def = null;
                        var page = pinegrow.getSelectedPage();
                        if(page) {
                            def = page.getMainType(null, pgel);
                        }

                        def = def || {name: 'Element'};

                        if(required_resources) {
                            var new_def = {
                                name: def.name,
                                framework: item.framework,
                                resources: new PgResourcesList(),
                                on_inserted: item.on_inserted || null,
                                required_plugins: required_plugins
                            }

                            required_resources.forEach(function(r) {
                                var remote_source = (new URL(r, item.url)).toString();

                                new_def.resources.add(item.framework.addResource(r, {
                                    resources_dir: item.resources_dir || 'pagelib',
                                    remote_source: remote_source,
                                    place_before_user_style: true
                                }))
                            })

                            var pb = new PgPluginBuilder(item.framework);

                            required_resources_code.forEach(function(rc) {
                                pb.addResourceCode(rc.code, rc.detect ? new RegExp(pgParserEscapeRegExp(rc.detect)) : null, rc.footer, new_def.resources);
                            })

                            def = new_def;
                        }

                        var parent_selector = null;

                        if(parent && selector == '*') {
                            var pid = parent.getAttribute('id');
                            if(pid) {
                                parent_selector = '#' + pid;
                            } else if(parent.getClasses().length) {
                                parent_selector = parent.getClasses().map(function(cls) {
                                    return '.' + cls;
                                }).join(',');
                            } else if(['body','main','header','footer','nav','section'].indexOf(parent.tagName) >= 0) {
                                parent_selector = parent.tagName;
                            }
                        }

                        dragMenu = new PgDragMenu(def || {name: 'Element'}, false, true, /* parent_selector */ null, pgel);

                        dragMenu.pgel = pgel;
                        dragMenu.selectNew = true;
                        dragMenu.formatHtml = true;
                        //dragMenu.parentSelector = parent_selector;

                        if(rest && rest.length) {
                            dragMenu.morePgels = rest;
                        }

                        dragMenu.handleMoveEvent(event, dragMenu);

                        dragMenu.onInsert = function(menu) {
                            if(item.require_saved_page) {
                                var page = this.positionPgel.getPage();
                                if (!page || page.isNew()) {
                                    pinegrow.showQuickError('Please save the page first.');
                                    return true; //cancel insert
                                }
                            }
                        }





                        /*
                         hideWinTimer = setTimeout(function () {
                         if (_this.quickWin) {
                         _this.quickWin.hide();
                         }
                         }, 500)
                         */
                    }
                    $element.trigger('pg-drag-start');

                }, this );

                dragHelper.on( 'drag', function(x, y, event) {

                    dragMenu.handleMoveEvent(event, dragMenu);
                }, this );

                dragHelper.on( 'dragStop', function(x, y) {
                    $.fn.crsapages('showOverlays', true);

                    if(dragMenu.insert()) {
                    }
                    dragMenu.destroy();
                    dragMenu = null;

                    crsaFuncs.showInsertionLine(null);
                    pinegrow.highlightElement(null);

                    /*
                     if(hideWinTimer) {
                     clearTimeout(hideWinTimer);
                     hideWinTimer = null;
                     }

                     if(_this.quickWin) {
                     setTimeout(function() {
                     _this.quickWin.show();

                     //top scroll hack
                     var $body = $(_this.getDocument().body);
                     var st = $body.scrollTop();
                     $body.scrollTop(st - 1);
                     }, 500)
                     }*/

                    $element.trigger('pg-drag-stop');

                }, this );

                if(item.no_drag_image) {
                    dragHelper.onGetDragImage = function () {
                        return 'none';
                    }
                }

                var ghost = null;

                doc.addEventListener("keydown", function( event ) {
                    $.fn.crsa('processKeydownEvent', event);
                }, true);

                doc.addEventListener("keyup", function( event ) {
                    $.fn.crsa('processKeyupEvent', event);
                }, true);


            }

            wireHandlers(_this.getDocument());

            $(doc.body).find('iframe').each(function(i, iframe) {
                if(iframe.contentDocument) {
                    wireHandlers(iframe.contentDocument);
                } else {
                    $(iframe).on('load', function (e) {
                        wireHandlers(e.target.contentDocument);
                    })
                }
            })

            _this.getDocument().addEventListener("scroll", function() {
                if(scrollTimer) {
                    clearTimeout(scrollTimer);
                }
                scrollTimer = setTimeout(function() {
                    onScroll();
                }, 300)
            })

            loaded = true;

            when_loaded.forEach(function(f) {
                f();
            })
            when_loaded = [];

            onScroll();

            if(item.on_page_lib_loaded) {
                item.on_page_lib_loaded(doc, _this)
            }
        })

        navigateTo(url);

        this.initToc();

        updateToc();
    }

    this.getTocHtml = function() {
        var html = '';
        var id_prefix = 'pglibtoc_';

        topics = []

        if(item.pages_toc) {
            item.pages_toc.forEach(function(d) {
                html += '<li data-toc-target="' + d.url + '">' + d.name + '</li>';
                topics.push({name: d.name, id: d.url})
            })

        } else {

            $(_this.getDocument()).find('.pg-lib-section').each(function (i, s) {
                var $s = $(s);
                var title = $s.find('.pg-lib-section-title').html();
                if (title) {
                    var target = id_prefix + i;
                    $s.attr('id', target);
                    html += '<li data-toc-target="' + target + '">' + title + '</li>';
                    topics.push({name: title, id: target})
                }
            })
        }
        return html;
    }

    this.scrollToTopic = function(target) {
        var $doc = $(_this.getDocument());
        var $s = $doc.find('#' + target);
        if($s.length == 0) {
            this.getTocHtml();
            $s = $doc.find('#' + target);
        }
        var pos = $s.position();
        $(_this.getDocument().scrollingElement).scrollTop(pos.top);
    }

    var parentView = {
        updateSize: function() {}
    }

    var current_ch = current_th = 0;

    this.updateSize = function() {
        view.whenVisible(function() {
            if(!$crumbs) return;

            var task = view.getFastdomTasks().create('resizePageLib' + view.getId());

            $crumbs.css('height', 'auto');

            task.measure(function() {
                var ch = $crumbs.outerHeight();
                var th = $toc.is(':visible') ? ($toc.height() + 10 + 'px') : '10px';

                if(current_ch != ch || current_th != th) {
                    task.mutate(function () {
                        $crumbs.css('height', ch + 'px');
                        $page.css({
                            bottom: ch + 'px'
                            //top: th
                        })
                    })
                    current_ch = ch;
                    current_th = th;
                }
            })
        }, 'resizeView')
    }

    view.onResize = function() {
    }

    view.onDestroy = function() {
        view.getFastdomTasks().stopAll('resizePageLib' + view.getId());

        if($iframe) {
            $iframe.remove();
        }
        pinegrow.pageLibManager.remove(_this.item.url);
        if(scrollTimer) {
            clearTimeout(scrollTimer);
            scrollTimer = null;
        }
    }
}


var PgCombinedLibView = function() {

    var lib = new PgLibraryView(true);
    var page_lib = new PgPageLibsTabbedView(this);

    var combined = new PgCombinedPanelView('combined-lib', lib.view, page_lib.view, 'lib', 'icon-import');

    var view = this.view = combined.view;

    page_lib.initToc();
    lib.show();
}



var PgQuickPageLib = function(pgel, position, $target, placement, static) {

    var _this = this;

    var pageLib = new PgPageLibsTabbedView(this);

    var lib = this.lib = pageLib.lib;

    var win = this.win = new PgQuickWindow('Insert element', pageLib, 'quickPageLib', 400, 600, false, null, static);

    if(static) {
        lib.keep_open_on_drag = pinegrow.getSetting('lib-keep-open-on-drag', '0') === '1';
        var $keep_icon = win.addIconToHeader('icon-Pin', 'Keep the insert window open', function ($i) {
            lib.keep_open_on_drag = !lib.keep_open_on_drag;
            pinegrow.setSetting('lib-keep-open-on-drag', lib.keep_open_on_drag ? '1' : '0');
            $keep_icon.css('opacity', lib.keep_open_on_drag ? 1 : 0.5);
            pinegrow.showMode('Keep insert window open', lib.keep_open_on_drag);
        });
        $keep_icon.css('opacity', lib.keep_open_on_drag ? 1 : 0.5);
    }

    win.$element.addClass('pg-lib-quick-win');

    win.hideOnClose = true;
    win.noRightEdge();

    var insertIntoSelected = pinegrow.selectedElements.isSelectedElement(pgel);

    lib.setInsertInfo(pgel, position);
    lib.showPreviews = true;

    lib.onInsertPositionChanged = function(pos) {
        var title;
        switch(pos) {
            case 'append':
                title = 'Append to';
                break;
            case 'prepend':
                title = 'Prepend to';
                break;
            case 'insertBefore':
                title = 'Insert before';
                break;
            case 'insertAfter':
                title = 'Insert after';
                break;
        }
        var pgel = lib.getInsertPgel();

        if(pgel) {
            title += ' ' + pgel.getName({short: true});
        }
        //win.setTitle(title);
    }

    lib.onInsertPositionChanged(position);

    //pageLib.initToc();

    if(placement === 'cover_left') {
        win.showCover(placement);
    } else if($target) {
        win.showFor$El($target, false, pgf.insert_in_top_menu ? 0 : 80, placement);
    } else if(pgel) {
        win.showForPgel(pgel);
    } else {
        win.showFor$El($('body'), false, 0);
        //win.showCover();
    }

    var hideWinTimer = null;

    win.$element.on('pg-drag-start', function() {
        hideWinTimer = setTimeout(function () {
            _this.hide();
        }, 100)
    })

    win.$element.on('pg-drag-stop', function() {
        if(hideWinTimer) {
            clearTimeout(hideWinTimer);
            hideWinTimer = null;
        }
        if(!pgf.insert_in_top_menu) {
            setTimeout(function () {
                _this.show();
            }, 100)
        } else {
            if(lib.keep_open_on_drag) {
                _this.show();
            } else {
                if (_this.win.visible) {
                    _this.hide();
                }
            }
        }
    })

    // pageLib.open('http://getbootstrap.com/docs/4.0/components/card/');

    win.onDestroy = function() {
    }

    win.onHide = function() {
        pinegrow.dispatchEvent('on_quick_lib_window_hidden', null, _this);
    }

    this.hide = function() {
        pgPreview.hide();
        win.hide();
    }

    this.show = function(focus) {
        win.show(true);
        pgQuickWindowManager.focus(win);
    }
}


var PgLibraryView = function(handleSelected, quick) {

    if(handleSelected === undefined) handleSelected = true;
    //View implementation is responsible for creating DOM and adding it to document.

    var insertOnClick = handleSelected && quick;

    if(pgf.kids) {
        insertOnClick = false;
    }

    var $element = $('<div class="aacrsa-panel pg-lib"></div>');

    var view = new PgUIView($element, quick ? null : 'librarylist');

    this.view = view;

    view.tabIcon = 'icon-import';

    this.showPreviews = true;
    this.handleSelectedElements = handleSelected;

    var _this = this;
    var current_pgel = null;
    var currentPage = pinegrow.getSelectedPage();
    var currentFactoryLibSections = null;

    var current_prop_obj;
    var filter = null;

    this.onInserted = function() {};
    this.onBeforeInserted = function() {};

    var insertPgel = null;
    var insertPosition = 'append';

    this.setInsertInfo = function(pgel, position) {
        if(pgel) {
            insertPgel = pgel;
            currentPage = pgel.getPage();
        }
        insertPosition = position || insertPosition;

        if(this.onInsertPositionChanged) this.onInsertPositionChanged(insertPosition);

        for(var i = 0; i < icons.length; i++) {
            if(icons[i].data.position == insertPosition) {
                toggleIcons(icons[i]);
                break;
            }
        }
    }

    this.getInsertPgel = function() {
        return insertPgel;
    }

    this.onInsertPositionChanged = null;

    this.getFilter = function() {
        return filter;
    }

    this.setFilter = function(f) {
        search.setValue(f);
    }

    var search = new PgSearchBox({
        placeholder: 'search',
        placeholder_active: null,
        toolbar_on: true
    });

    var icons = [];

    var toggleIcons = function(icon) {
        icons.forEach(function(i) {
            i.setActive(i === icon);
            if(i === icon) {
                insertPosition = icon.data.position;
                if(_this.onInsertPositionChanged) _this.onInsertPositionChanged(insertPosition);
            }
        })
    }

    if(insertOnClick) {

        icons.push(search.addIcon({
            icon: '<i class="icon icon-Prepend"><span class="path1"></span><span class="path2"></span></i>',
            label: 'Append to the selected element(s).',
            position: 'append',
            func: toggleIcons
        }))

        icons.push(search.addIcon({
            icon: '<i class="icon icon-Append"><span class="path1"></span><span class="path2"></span></i>',
            label: 'Prepend to the selected element(s).',
            position: 'prepend',
            func: toggleIcons
        }))

        icons.push(search.addIcon({
            icon: '<i class="icon icon-Ins_before"><span class="path1"></span><span class="path2"></span></i>',
            label: 'Insert before the selected element(s).',
            position: 'insertBefore',
            func: toggleIcons
        }))

        icons.push(search.addIcon({
            icon: '<i class="icon icon-Ins_after"><span class="path1"></span><span class="path2"></span></i>',
            label: 'Insert after the selected element(s).',
            position: 'insertAfter',
            func: toggleIcons
        }))
    }

    if(!quick) {
        var toolbar = search.getToolbar();

        toolbar.addSection('settings');

        var settingsButton = new PgButtonToolbarItem('icon-tools', 'Manage frameworks and plugins.');
        toolbar.addItem('settings', settingsButton);
        settingsButton.onClick = function () {
            $.fn.crsacss('showFrameworkManager', pinegrow.getSelectedPage());
        }
    }

    var grid = new PgUISectionGrid();

    if(!pgf.kids) {
        search.$element.appendTo($element);
    }

    search.onSearch = function(f) {
        filter = f;
        _this.show(currentFactoryLibSections, f);
    }

    /*
     var mover = new PgArrowsMover(search.getInput(), grid.$element, '.crsa-factory-element');

     mover.onEnter = function($item) {
     if($item) {

     } else {

     }
     }*/


    var spacer = new PgSearchBoxSpacer();

    grid.$element.addClass('pg-search-box-results');

    if(!pgf.kids) {
        spacer.$element.appendTo($element);
    } else {
        grid.$element.css({
            top: 0
        })
    }

    this.focusSearch = function() {
        search.focus();
    }

    $element.append(grid.$element);

    var $scrollParent = grid.$element;

    var selectedInsertSection = new PgUISection('Insert into selected', '');
    var hasSelectedInsert = false;

    var lastUsedSection = new PgUISection('Recently used', '');

    var quickInsertSection = new PgUISection(`Insert code <small>HTML or <a href="#" class="pug-info">${isApp() ? 'PUG' : 'SIMPLE'}</a></small>`, '', null, pgf.kids /* closed by default */);

    var codeInsert = new PgLibraryCodeInsert(quick && !pgf.kids);

    quickInsertSection.onOpenClosed = function() {
        codeInsert && codeInsert.resized();
    }
    quickInsertSection.$content.append(codeInsert.$element);

    quickInsertSection.$title.find('.pug-info').on('click', function(e) {
        e.preventDefault();
        e.stopPropagation();

        var msg = `<p>Use the format <code>tag#id.class1.class2 Text</code> to describe the HTML structure. Each element should be in a separate line. Use indents to nest the elements.</p><p>Example:</p>
<pre>
div#hero.text-left.text-primary.bg-white
  h1 Hello
  p.lead How are you?
  ul
    li Item 1
    li Item 2
  div.content
  | Just text
</pre>`;

        if(isApp()) {
            msg += `<p>Pinegrow uses <a href="https://pugjs.org/language/tags.html" class="external">Pug</a> to translate the description to HTML.</p>`;
        } else {
            msg += `<p>Although this is similar to Pug syntax, Pinegrow in the browser uses a simplified converter that only supports tags, ids, classes and text.</p>`
        }
        pinegrow.showAlert(msg, 'About the simplified HTML syntax');
        return false;
    })

    this.updateLayout = function() {
        grid.updateLayout();
    }

    var grid_sections_map = {};

    var clearGridSectionsMap = function() {
        grid_sections_map = {};
    }

    var createButton = function(def, repeater, on_click) {
        repeater = false;
        on_click = false;
        var $item;
        var r = repeater ? pinegrow.repeater.showLiveRepeatCount() : '';
        var oc = on_click ? ' <i class="icon icon-right-arrow"></i>' : '';
        if(def.button_image) {
            var src = def.framework.getImagePreviewBaseUrl() + def.button_image + '?z=' + pinegrow.cache_breaker;
            var name_html = def.no_lib_label ? '' : ('<name>' + def.name + oc + '</name>');
            $item = $('<div/>', { 'class' : 'crsa-factory-element-image crsa-factory-element crsa-factory-element-type-' + def.type }).html('<figure><img alt="' + def.name + '" src="' + src + '"/></figure>' + name_html + r).data('crsa-factory-def', def);
        } else {
            $item = $('<div/>', { 'class' : 'crsa-factory-element crsa-factory-element-text crsa-factory-element-type-' + def.type}).html('<name>' + def.name + r + oc + '</name>').data('crsa-factory-def', def);
        }
        $item.attr('draggable', 'true');
        return $item;
    }

    var getGridSection = function(sec_def) {
        if(!(sec_def.key in grid_sections_map)) {

            var cat = null;
            var section_class = null;

            if(sec_def.framework && !pgf.kids) {
                cat = sec_def.framework.name;
                if(sec_def.framework.user_lib) {
                    cat = '<libname>' + cat + '</libname> <changed>*</changed>';
                    section_class = 'user-lib';
                }
            }

            if(pgf.kids) {
                sec_def.is_closable = false;
            }

            var pgsec = new PgUISection(sec_def.name, cat, sec_def);

            if(section_class) {
                pgsec.$element.addClass(section_class);
            }

            if(sec_def.framework && sec_def.framework.user_lib) {
                pgsec.$title.find('small').on('click contextmenu', function(e) {
                    pgPreview.hide();

                    var $menu = new CrsaContextMenu(null, $(this));

                    var saveAs = function() {
                        crsaChooseFile(function(url, file) {
                            sec_def.framework.save(file, function() {
                                pinegrow.frameworksChanged();
                            });
                        }, sec_def.framework.getFileName());
                    }

                    $menu.add("Add as HTML snippet", null, function() {
                        if(pinegrow.selectedElements.hasSelection()) {
                            pinegrow.getCollection().forEachElement(function(pgel) {
                                addAsComponent(pgel, sec_def.framework);
                            })
                        } else {
                            showAlert("First select the element on the page than add it to the snippet library.");
                        }
                    });

                    $menu.add("Save", null, function() {
                        sec_def.framework.save(sec_def.framework.pluginUrl, function() {
                            pinegrow.frameworksChanged();
                        });
                    });

                    $menu.add("Save as...", null, function() {
                        if(sec_def.framework.pluginUrl) {
                            sec_def.framework.save(sec_def.framework.pluginUrl, function() {
                                pinegrow.frameworksChanged();
                            });
                        } else {
                            saveAs();
                        }
                    });

                    $menu.showAt(e.pageX, e.pageY);
                    return false;
                })
            }

            var $c = pgsec.$content;
            var $dest = $('<div/>', {class: 'pg-lib-section-content'}).appendTo($c);

            $.each(sec_def.getComponentTypes(), function(i, def) {

                var $item = createButton(def);

                $item.appendTo($dest);
            });
            grid_sections_map[sec_def.key] = pgsec;
        }
        return grid_sections_map[sec_def.key];
    }

    var filterGridSection = function(pgsec, sec_def, filterRegEx, deflist) {
        deflist = deflist || sec_def.getComponentTypes();
        var list = pg$FindAll(pgsec.$content, '.crsa-factory-element');

        var cat_match = (sec_def && filterRegEx) ? sec_def.name.match(filterRegEx) : (sec_def || false);

        var empty = true;

        $.each(deflist, function(i, def) {
            if(filterRegEx && !cat_match && !def.name.match(filterRegEx)) {
                pgHide(list[i]);
            } else {
                pgShow(list[i]);
                empty = false;
            }
        })
        return !empty;
    }

    this.show = function(sections, filter) {
        var profile = new CrsaProfile();

        var filterRegEx = filter && filter.length > 0 ? new RegExp(escapeRegExp(filter),'i') : null;

        var current_scroll_top = $scrollParent.scrollTop();

        grid.clear();

        if(!currentPage) {
            currentFactoryLibSections = null;
            //work around the issue where code editor disappears after closing and reopening project
            //we destroy the editor when the lib panel is empty
            //looks like the editor is removed when the element is detached
            codeInsert && codeInsert.destroy();
            codeInsert = null;
            return;
        }

        var $dest = grid.$element;

        if(!sections) sections = currentPage.getLibSections();
        currentFactoryLibSections = sections;

        if(hasSelectedInsert && this.handleSelectedElements && !pgf.kids) {
            grid.addSection(selectedInsertSection);
        }

        if(!pgf.kids && pgLibraryLastUsed.getList().length && filterGridSection(lastUsedSection, null, filterRegEx, pgLibraryLastUsed.getList())) {
            grid.addSection(lastUsedSection);
        }

        if(!pgf.kids) {
            //recreate the code editor if it was destroyed after the lib was empty before
            if (!codeInsert) {
                codeInsert = new PgLibraryCodeInsert(quick);
                quickInsertSection.$content.append(codeInsert.$element);
            }
            grid.addSection(quickInsertSection);
        }

        //todo30 - user lib

        $.each(sections, function(i, sec_def) {
            //var key = sec_def.section_key;

            var pgsec = getGridSection(sec_def);

            if(sec_def.framework.user_lib) {
                pgsec.$title.find('libname').text(sec_def.framework.name);
                var $changed = pgsec.$title.find('changed');
                if(sec_def.framework.changed) {
                    pg$Show($changed, 'inline');
                } else {
                    pg$Hide($changed);
                }
            }

            if(filterGridSection(pgsec, sec_def, filterRegEx)) {
                grid.addSection(pgsec);
            }

            /*
             if(empty && (!sec_def.framework.user_lib || filterRegEx)) {
             //$tit.remove();
             }*/

        })

        profile.show('showLib');

        grid.updateLayout();
        codeInsert && codeInsert.resized();

        $scrollParent.scrollTop(current_scroll_top);
    }

    var updateList = function(sections) {
        view.whenVisible(function() {
            _this.show(sections, filter);
        })
    }

    var onPageChanged = function(event, crsaPage) {
        var sections = null;
        var changed = false;

        currentPage = crsaPage;

        if(!crsaPage) {
            changed = true;
            //$manage.hide();
        } else {
            sections = crsaPage.getLibSections();
            if(currentFactoryLibSections == null || sections == null) {
                changed = true;
            } else {
                if(currentFactoryLibSections.length != sections.length) {
                    changed = true;
                } else {
                    for(var i = 0; i < sections.length; i++) {
                        if(sections[i] != currentFactoryLibSections[i]) {
                            changed = true;
                            break;
                        }
                    }
                }
            }
            //$manage.show();
        }
        if(changed) {
            updateList(sections);
        }
    }

    var onElementSelected = function(page, pgel) {
        view.whenVisible(function() {
            _this.show(currentFactoryLibSections);
        })
        current_pgel = pgel;
    }

    var onLastUsedChanged = function() {
        var list = pgLibraryLastUsed.getList();

        lastUsedSection.$content.empty();

        var $dest = $('<div/>', {class: 'pg-lib-section-content'}).appendTo(lastUsedSection.$content);

        list.forEach(function(def) {
            createButton(def).appendTo($dest).addClass('pg-lib-recent-item');
        })

        if(!lastUsedSection.parent) {
            updateList(currentFactoryLibSections);
        }
    }

    onLastUsedChanged();

    var onUpdateInsertIntoSelected = function() {
        var has = false;

        if(!_this.handleSelectedElements) return;

        var pgel = pinegrow.selectedElements.getLastSelected();

        if(pgel) {
            currentPage = pgel.getPage();
            var defs = currentPage.getAllTypes(null, pgel);
            var add = [];
            $.each(defs, function(i, def) {
                if(def.action_menu && def.action_menu.add) {
                    add.push(def.action_menu);
                }
            });

            if(!add.length) {
                //return;
            } else {
                selectedInsertSection.$content.empty();

                var $dest = $('<div/>', {class: 'pg-lib-section-content'}).appendTo(selectedInsertSection.$content);

                $.each(add, function (i, action_menu) {
                    $.each(action_menu.add, function (j, type) {
                        var def = currentPage.getTypeDefinition(type);
                        if (def) {
                            var $item = createButton(def, true, true);
                            $item.data('action-menu', action_menu);
                            $item.addClass('crsa-factory-element-link')
                            $item.appendTo($dest);
                        }
                    });
                });
                has = true;
            }
            if(insertOnClick) {
                _this.setInsertInfo(pgel);
                search.getToolbar().getSection('icons').$element.show();
            }

        } else {
            if(insertOnClick) {
                search.getToolbar().getSection('icons').$element.hide();
            }

        }
        if(has != hasSelectedInsert) {
            hasSelectedInsert = has;
            updateList(currentFactoryLibSections);
        }
    }

    onUpdateInsertIntoSelected();

    var onFrameworksChanged = function() {
        clearGridSectionsMap();
        updateList();
    }

    view.onResize = function() {
        codeInsert && codeInsert.resized();
        grid.updateLayout();
    }

    view.on('crsa-frameworks-changed', onFrameworksChanged);// todo30 - repopulate the tree
    view.on('crsa-page-selected', onPageChanged);
    view.on('crsa-update-lib-display', onFrameworksChanged);

    pinegrow.addEventHandler('on_lib_last_used_changed', onLastUsedChanged);

    pinegrow.addEventHandler('on_selected_elements_changed', onUpdateInsertIntoSelected);

    var delayedTasks = new PgDelayedTasks();

    view.onDestroy = function() {
        _this.show(null);
        codeInsert && codeInsert.destroy();
        pinegrow.removeEventHandler('on_lib_last_used_changed', onLastUsedChanged);

        pinegrow.removeEventHandler('on_selected_elements_changed', onUpdateInsertIntoSelected);

        delayedTasks.destroy();
        pgPreview.hide();
    }

    var tooltip_active = false;

    var wireEvents = function() {
        $element.on('click', '.crsa-factory-element', function(e) {
            e.preventDefault();

            pgPreview.hide();

            var $el = $(this);
            var am = $el.data('action-menu');
            var def = $el.data('crsa-factory-def');

            if(insertPgel && !pgf.kids) {
                //one click insert
                var api = new PgApi();
                var created = createPgElFromDefinitionMultiple(def);
                if (created) {
                    var dest = pinegrow.selectedElements.isSelectedElement(insertPgel) ? pinegrow.selectedElements.getCollection() : insertPgel;

                    _this.onBeforeInserted(created.collection, dest, insertPosition);

                    api.insertElements(created.collection, dest, insertPosition, {highlight: true, scroll: true, repeater: true, format_html: true});

                    pgLibraryLastUsed.add(def);

                    pinegrow.showQuickMessage(def.name + ' inserted.');

                    _this.onInserted(created.collection, dest, insertPosition);
                }
            } else {
                if (am && !pgf.kids) {
                    //insert into selected
                    if (!pinegrow.selectedElements.hasSelection()) return;
                    crsaFuncs.insertThroughActionMenu(am, null, def, false);
                } else {
                    if (!tooltip_active) {
                        var tree_name = pgf.kids ? 'Structure panel' : 'Tree panel';
                        $el.tooltip({
                            container: 'body',
                            trigger: 'manual',
                            title: `Drag & Drop me on the page or to the ${tree_name}. Right click to insert into the selected element.`
                        });
                        var $cont = $el.closest('body');
                        $el.data('bs.tooltip').options.container = $cont.length ? $cont : 'body';

                        $el.tooltip('show');
                        setTimeout(function () {
                            $el.tooltip('destroy');
                            tooltip_active = false;
                        }, 4000);
                        //dbl_timer = null;
                        tooltip_active = true;
                    }
                }
            }
        })
        $element.on('contextmenu', '.crsa-factory-element', function(e) {
            e.preventDefault();

            var $el = $(this);
            var def = $el.data('crsa-factory-def');
            var fm = def.framework || null;

            pgPreview.hide();

            var helper = new PgContextMenuHelpers(null, null);

            var $menu = new CrsaContextMenu(null, $el);

            if(pinegrow.selectedElements.hasSelection()) {

                helper.addInsertActions($menu, def);

            }

            if(fm.user_lib && !$el.hasClass('pg-lib-recent-item')) {
                $menu.add("", null, null, 'divider');
                $menu.add("HTML Snippet", null, null, 'header');

                $menu.add("Rename...", null, function() {
                    showPrompt("Enter the new name for the component:", "Rename component", def.name, null, null, function(name) {
                        if(name && def.name != name) {
                            def.name = name;
                            fm.changed = true;
                            pinegrow.frameworksChanged();
                        }
                    });
                });
                $menu.add("Update", null, function() {
                    var pgel = pinegrow.selectedElements.getLastSelected();
                    if(pgel) {

                        showAlert("Do you want to update component <b>" + def.name + "</b> with the content of element <b>" + pgel.getName() + "</b>?", "Confirm update", "Cancel", "Update it", null, function() {
                            copyElementToDef(pgel, def, pgel.getPage(), false);
                            fm.changed = true;
                            pinegrow.frameworksChanged();
                            crsaQuickMessage("Component updated");
                        });
                    } else {
                        showAlert("First select the element on the page that you'll use to update the component with.");
                    }
                });

                $menu.add("Delete", null, function() {
                    showAlert("Are you sure? Deleting component <b>" + def.name + "</b> can't be undone.", "Confirm delete", "Cancel", "Delete it", null, function() {
                        fm.removeComponentType(def);
                        fm.changed = true;
                        pinegrow.frameworksChanged();
                    });
                });
            }


            $menu.add("", null, null, 'divider');
            //$menu.add("Insert", null, null, 'header');

            $menu.add("Copy code", null, function() {
                var code = getCodeFromDefinition(def);
                copyCodeToClipboard(code);
                //setLastPreviewedDef(def);
            });

            var actions = [];
            currentPage.callFrameworkHandler('on_build_lib_actions_menu', actions, def);

            for(var i = 0; i < actions.length; i++) {
                $menu.add(actions[i].label, null, actions[i].action);
            }
            $menu.showAt(e.pageX, e.pageY, grid.$element);
            return false;
        })

        $element.on('mouseenter', '.crsa-factory-element', function(e) {
            if(pgGetDragMenu() || !_this.showPreviews) return; //todo30

            var $el = $(this);

            delayedTasks.run('showPreview', function() {
                //if mouse moved out out the window we didn't get the mouseleave event, probably
                //detect that case and abort
                if(pgGetAppWindowUnderMouse() !== pgGetAppWindowForDOMElement($element.get(0))) {
                    return;
                }
                var def = $el.data('crsa-factory-def');

                var $pel = crsaFuncs.createPreviewElementFromDefinition(def);

                if($pel && def.framework && !def.preview_image && def.framework.on_lib_show_preview && def.framework.on_lib_show_preview(null, def, $pel, pgPreview)) {

                }
                if($el.get(0).ownerDocument === document) {
                    pgPreview.show($el, $pel, def.preview_image ? 'with-image' : null, removeCrsaClassesFromHtml(getCodeFromDefinition(def)));
                }
            }, 300)

        });

        $element.on('mouseleave', '.crsa-factory-element', function(e) {
            delayedTasks.cancel('showPreview');
            pgPreview.hide();
        });

        var dragHelper = new PgDragHelper($element, '.crsa-factory-element', null, true);

        var dragMenu;
        var activeDef;

        dragHelper.on( 'dragStart', function(x, y, event) {
            $.fn.crsapages('showOverlays');

            var $li = dragHelper.dragTarget;
            var def = $li.data('crsa-factory-def');

            pgPreview.hide();

            activeDef = def;

            dragMenu = new PgDragMenu(def, false, true);

            dragMenu.formatHtml = true;

            var created = createPgElFromDefinitionMultiple(def);

            if (created) {
                dragMenu.pgel = created.pgel;
                dragMenu.morePgels = created.rest;

                dragMenu.handleMoveEvent(event, dragMenu);
            }

            $element.trigger('pg-drag-start');

        }, this );

        dragHelper.onSetDragData = function(dt) {
            //var code = getCodeFromDefinition(activeDef);
            //dt.setData("text/plain", code);
        }

        dragHelper.on( 'drag', function(x, y, event) {

            dragMenu.handleMoveEvent(event, dragMenu);
        }, this );

        dragHelper.on( 'dragStop', function(x, y) {
            $.fn.crsapages('showOverlays', true);

            if(dragMenu.insert()) {
                pgLibraryLastUsed.add(activeDef);
            }
            dragMenu.destroy();
            dragMenu = null;
            activeDef = null;

            crsaFuncs.showInsertionLine(null);
            pinegrow.highlightElement(null);

            $element.trigger('pg-drag-stop');

        }, this );
    }

    wireEvents();
    this.setInsertInfo();
}

var PgLibraryCodeInsert = function(click_to_insert) {

    var pug = null;

    if(isApp()) {
        pug = require('pug');
    } else {
        pug = new PgSimplePug();
    }

    var current_code = '';

    var _this = this;

    this.$element = $('<div class="pg-code-insert"><div class="pg-code-insert-editor"></div></div>');
    var view = this.view = new PgUIView(this.$element, "code_insert");

    this.$editor = this.$element.find('> .pg-code-insert-editor');

    this.$item = $('<div/>', { 'class' : 'crsa-factory-element crsa-factory-element-text'}).html('<name>' + 'Code bibo' + '</name><i class="clear icon icon-close"></i>').data('crsa-factory-def', null);
    this.$item.attr('draggable', 'true');

    //var $helptext = $('<p class="pg-code-insert-editor-help">Enter HTML or <a href="http://jade-lang.com/" class="external">JADE</a>:</p>').appendTo(this.$element);

    //crsaHandleExternalLinks($helptext);

    var $clear = this.$item.find('.clear');

    $clear.on('click', function(e) {
        if(CodeEditorVsc) {
            CodeEditorVsc.setValue('');
        }
        return false;
    })

    addTooltip($clear, 'Clear code in the text editor.')

    var $errortext = $('<p class="pg-code-insert-editor-error"></p>').appendTo(this.$element);

    var current_error = '';

    var showError = function(err) {
        if(err !== current_error) {
            if(err) {
                $errortext.html(err);
                if(!current_error) {
                    pg$Show($errortext);
                    pg$Hide(_this.$item);
                }
            } else {
                if(current_error) {
                    pg$Hide($errortext);
                }
            }
            current_error = err;
        }
        if(!err && current_code) pg$Show(_this.$item);
        if(!err && current_code.length === 0) {
            pg$Hide(_this.$item);
        }
    }

    showError(null);

    pg$Hide(this.$item);

    addTooltip(this.$item, click_to_insert ? 'Click to insert to the page' : 'Drag me to the page!', {
        placement: 'bottom',
        trigger: 'hover'
    });

    var $itemname = this.$item.find('name');

    this.$item.appendTo(this.$element);

    let oldMode = "pug";

    let updateItem = async function (code) {

        code = $.trim(code);

        if (code.length === 0) {
            //pg$Show($helptext);
            showError(null);
            pg$Hide(_this.$item);
            current_code = code;
            return;
        } else {
            //pg$Show(_this.$item);
            if (current_code.length === 0) {
                //pg$Hide($helptext);
            }
        }
        var old_code = current_code;

        current_code = code;

        var is_pug;
        var html;

        /**@var {Position}**/
        let editorPosition = CodeEditorVsc.getPosition();

        if(pug) {
            if (editorPosition.lineNumber > 1) {
                is_pug = code.indexOf('<') < 0
            } else {
                is_pug = code.match(/(\w+(>|{))|((^<\w+)|(^\s+<\w+))/si) ? false : true;
            }
        } else {
            is_pug = false;
        }

        if (is_pug && oldMode != "pug") {
            try {
                await setMode('pug');
                html = pug.render(code, {});
                //editorPosition.column++;
                CodeEditorVsc.setPosition(editorPosition);
                oldMode = "pug"
            } catch (err) {
                showError('PUG Syntax error: ' + err.msg);
                return;
            }
        } else if(!is_pug && oldMode != "html"){
            await setMode('html');
            html = code;
            //editorPosition.column++;
            CodeEditorVsc.setPosition(editorPosition);
            oldMode = "html";
        } else{
            try {
                html = is_pug ? pug.render(code, {}) : code;
            } catch(err) {
                showError('PUG Syntax error: ' + err.msg);
                return;
            }
        }

        var def = {
            name: 'Bibo',
            code: html,
            original_code: code
        }
        _this.$item.data('crsa-factory-def', def);

        var created = createPgElFromDefinitionMultiple(def, false, false);

        if (created) {
            var err = null;

            if (created.text_nodes) {
                err = 'Inserting text outside of tags is not supported. Use inline editing or edit code.';
            }
            if (!err && created.syntax_errors) {
                err = 'Syntax error. Maybe tags or attributes are not closed?';
            }
            if (!err) {
                def.name = created.pgel.getName({max_text_len: 10});
                if (created.rest.length) {
                    def.name += ' + ' + created.rest.length;
                }
                $itemname.html(def.name);

                showError(null);

                if (old_code.length == 0) {
                    _this.$item.tooltip('show');
                    setTimeout(function () {
                        _this.$item.tooltip('hide');
                    }, 2000);
                }
            } else {
                showError(err);
            }
        } else {
            // !/(^\w+\s)/.test(code)
            //get current line content and ignore error
            let currentLine = CodeEditorVsc.getModel().getLineContent(editorPosition.lineNumber);
            if(!currentLine.match(/\w+(>|{)/si))
            showError('Syntax error');
        }


    }

    var CodeEditorVsc;
    var current_mode = null;

    var setMode = function(mode) {
        if(mode !== current_mode) {
            var new_model = null;
            switch(mode) {
                case 'jade':
                case 'pug':
                    new_model = monaco.editor.createModel(CodeEditorVsc.getValue(), "pug");
                    break;
                default:
                    new_model = monaco.editor.createModel(CodeEditorVsc.getValue(), "html");
            }
            current_mode = mode;
            if(new_model) {
                CodeEditorVsc.setModel(new_model);
            }
        }
    }

    var makeEditor = function() {

        let win = _this.$editor.get(0).ownerDocument.defaultView;
        CodeEditorVsc = monacoInstance.openEditor(null,null, win, _this.$element.children(), true)

        pinegrow.code_editors.register(CodeEditorVsc);

        // $editor_el.attr('data-has-codemirror','').data('codeMirrorInstance', CodeEditorVsc);

        CodeEditorVsc.updateOptions({wordWrap:"on"});

        CodeEditorVsc.onDidChangeModelContent( async function () {
            await updateItem(CodeEditorVsc.getValue({preserveBOM: true}));
        });

        CodeEditorVsc.onDidFocusEditorText( function() {
            pgEditorAttributeIntellisense.focusedUrl = pinegrow.getSelectedPage().url;
        });

        setMode('pug');
    }

    makeEditor();

    this.destroy = function() {
        pinegrow.code_editors.unregister(CodeEditorVsc);
        CodeEditorVsc.dispose();
        this.$element.remove();
    }

    view.onBeforeMovedToAnotherWindow = function (){
        _this.destroy();
    }

    view.onMovedToAnotherWindow = function (win) {
        makeEditor();
        CodeEditorVsc.layout();
    }

    this.resized = function() {
        if(CodeEditorVsc) {
            // if(this.$editor.get(0).ownerDocument) {
            //     CodeEditorVsc.pgSetWindow(this.$editor.get(0).ownerDocument.defaultView);
            // }
            // CodeEditorVsc.refresh();
            CodeEditorVsc.layout()
        }
    }

}

var pgQuickInsertLastFilter = null;

var PgQuickInsert = function(pgel, position, handleSelected, $target) {

    if(handleSelected === undefined) handleSelected = true;

    /*
     var win = pgQuickWindowManager.getWindow('quickWindow');
     if(win) {
     win.flash();
     return;
     }*/

    var lib = new PgLibraryView(handleSelected, true);

    var insertIntoSelected = pinegrow.selectedElements.isSelectedElement(pgel);

    lib.setInsertInfo(pgel, position);
    lib.showPreviews = true;

    lib.onInsertPositionChanged = function(pos) {
        var title;
        switch(pos) {
            case 'append':
                title = 'Append to';
                break;
            case 'prepend':
                title = 'Prepend to';
                break;
            case 'insertBefore':
                title = 'Insert before';
                break;
            case 'insertAfter':
                title = 'Insert after';
                break;
        }
        var pgel = lib.getInsertPgel();

        if(pgel) {
            title += ' ' + pgel.getName({short: true});
        }
        win.setTitle(title);
    }

    var win = new PgQuickWindow('Quick insert', lib, 'quickInsert', 570, 360);
    lib.onInsertPositionChanged(position);

    win.noRightEdge();

    if(!handleSelected) {
        win.$element.addClass('pg-quick-insert-no-selected');
    }

    if($target) {
        win.showFor$El($target, false, 80);
    } else {
        win.showForPgel(pgel);
    }

    lib.show();
    if(pgQuickInsertLastFilter) {
        lib.setFilter(pgQuickInsertLastFilter);
    }
    lib.focusSearch();

    //pgOpenViewInFloatingPanel('Insert into selected', 'lib_quick', lib.view);

    var onElementSelected = function() {
        if(insertIntoSelected) {
            var pgel = pinegrow.selectedElements.getLastSelected();
            if(pgel) {
                lib.setInsertInfo(pgel);
            } else {
                //todo30 - sel el notice
            }
        }
    }

    pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);


    win.onDestroy = function() {
        pgQuickInsertLastFilter = lib.getFilter();
        pinegrow.removeEventHandler('on_selected_elements_changed', onElementSelected);
    }

    this.getWindow = function() {
        return win;
    }

    this.getLib = function() {
        return lib;
    }
}


var PgLibraryLastUsed = function() {

    var list = [];

    var _this = this;

    var notify = function() {
        pinegrow.dispatchEvent('on_lib_last_used_changed', null, this);
    }

    this.add = function(def) {
        var idx = -1;
        for(var i = 0; i < list.length; i++) {
            if(list[i] === def || list[i].code === def.code) {
                idx = i;
                break;
            }
        }
        if(idx >= 0) {
            list.splice(idx, 1);
        }
        list.unshift(def);

        var max = 10;
        if(list.length > max) {
            list.splice(max, list.length - max);
        }
        notify();
    }

    this.clear = function() {
        list = [];
        notify();
    }

    this.getList = function() {
        return list;
    }
}

window.pgLibraryLastUsed = new PgLibraryLastUsed();

var pgInsertLibMenuCurrentPosition = 'insertAfter';

var PgContextMenuHelpers = function(pgel, def_place) {

    var current_icons = null;

    this.getPlacementIcons = function() {
        var fertile = pgel ? pgel.canHaveChildren() : true;
        var icons = `<span class="pg-lib-menu-place-icons">`;

        if(pgel.canHaveSiblings()) {
            icons += `<i data-pos="insertBefore" class="icon icon-Ins_before ${def_place === 'insertBefore' ? 'active' : ''}" title="Insert before"><span class="path1"></span><span class="path2"></span></i><i data-pos="insertAfter" class="icon icon-Ins_after ${def_place === 'insertAfter' ? 'active' : ''}" title="Insert after"><span class="path1"></span><span class="path2"></span></i>`
        }

        if(fertile) {
            icons += `<i data-pos="prepend" class="icon icon-Append ${def_place === 'prepend' ? 'active' : ''}" title="Insert inside on the top"><span class="path1"></span><span class="path2"></span></i><i data-pos="append" class="icon icon-Prepend ${def_place === 'append' ? 'active' : ''}" title="Insert inside at the end"><span class="path1"></span><span class="path2"></span></i>`
        }
        icons += `</span>`;

        current_icons = icons;

        return icons;
    }

    this.getPlacementFromEvent = function(event) {
        return $(event.target).closest('[data-pos]').attr('data-pos');
    }

    this.getMenuItemWithPlacement = function(name) {
        current_icons = current_icons || this.getPlacementIcons();
        return {
            label: '<span class="name">' + name + '</span>' + current_icons,
            class: 'pg-lib-menu-item'
        }
    }

    this.getMenuItemWithoutPlacement = function(name) {
        return {
            label: '<span class="name">' + name + '</span>',
            class: 'pg-lib-menu-item'
        }
    }

    this.getMenuFixedSizeClass = function(size) {
        size = size || 'md';
        return 'pg-menu-fixed-size-' + size;
    }

    this.addInsertActions = function(menu, def) {
        var am = crsaFuncs.findActionMenuForInsertingDefIntoEl(def, pinegrow.selectedElements.getLastSelected());

        var name = pinegrow.selectedElements.getName();

        menu.add("", null, null, 'divider');
        menu.add("Insert", null, null, 'header');

        menu.add("Prepend to <b>" + name + "</b>" + pinegrow.repeater.showLiveRepeatCount(), null, function() {
            crsaFuncs.insertThroughActionMenu(am, /*selectedElement.data*/ null, def, false, true);
            //setLastPreviewedDef(def);
        });
        menu.add("Append to <b>" + name + "</b>" + pinegrow.repeater.showLiveRepeatCount(), null, function() {
            crsaFuncs.insertThroughActionMenu(am, /*selectedElement.data*/ null, def, false, false);
            //setLastPreviewedDef(def);
        });
        menu.add("Insert before <b>" + name + "</b>" + pinegrow.repeater.showLiveRepeatCount(), null, function() {
            crsaFuncs.insertBeforeOrAfter(/*selectedElement.data*/ null, def, false, false);
            //setLastPreviewedDef(def);
        });
        menu.add("Insert after <b>" + name + "</b>" + pinegrow.repeater.showLiveRepeatCount(), null, function() {
            crsaFuncs.insertBeforeOrAfter(/*selectedElement.data*/ null, def, false, true);
//                    setLastPreviewedDef(def);
        });
        menu.add("Replace <b>" + name + "</b>" + pinegrow.repeater.showLiveRepeatCount(), null, function() {
            crsaFuncs.replaceElement(null, def);
        });
    }
}

var PgInsertLibElementMenu = function(pgel, target_pgel, no_hint, placement, recommended) {

    if(target_pgel === undefined) target_pgel = pgel;

    var currentPage = pgel.getPage();
    var def = currentPage.getMainType(null, pgel);

    var $dest = pgel.get$DOMElement();

    var actions = []

    var sections = currentPage.getLibSections();

    var current_f = null;

    var add_fm_header = false;

    var show_previews = pinegrow.getSetting('lib-insert-menu-show-previews', "0") === "1";

    var def_place = placement || pgInsertLibMenuCurrentPosition;
    var fertile = pgel.canHaveChildren();

    if(!fertile && ['prepend', 'append'].indexOf(def_place) >= 0) {
        def_place = 'insertAfter';
    }

    if(!pgel.canHaveSiblings() && ['insertAfter', 'insertBefore'].indexOf(def_place) >= 0) {
        def_place = 'append';
        if(!fertile) {
            def_place = null;
        }
    }

    var helper = new PgContextMenuHelpers(pgel, def_place);

    var delayedTasks = new PgDelayedTasks();

    var showPreview = function(def, $el) {
        delayedTasks.run('showPreview', function() {

            var $pel = crsaFuncs.createPreviewElementFromDefinition(def);

            if($pel && def.framework && !def.preview_image && def.framework.on_lib_show_preview && def.framework.on_lib_show_preview(null, def, $pel, pgPreview)) {

            }
            if($el.get(0).ownerDocument === document) {
                pgPreview.show($el, $pel, def.preview_image ? 'with-image' : null, removeCrsaClassesFromHtml(getCodeFromDefinition(def)));
            }
        }, 300)
    }

    var getInsertionLinePos = function(pos) {
        switch(pos) {
            case 'insertBefore':
                return 'top';
            case 'insertAfter':
                return 'bottom';
            case 'prepend':
                return 'inside-top';
            case 'append':
                return 'inside-bottom';
        }
    }

    var addPreview = function(d) {

        var handlePos = function(e) {
            //console.log(e)
            crsaFuncs.showInsertionLine($dest, getInsertionLinePos($(e.target).closest('[data-pos]').attr('data-pos')))
        }

        var handlePosOut = function(e) {
            crsaFuncs.showInsertionLine($dest, getInsertionLinePos(def_place))
        }

        d.on_mouseenter = function ($a, event) {

            if(show_previews) {
                showPreview(this.def, $a)
            }
            delayedTasks.cancel('cancelTransparent');
            crsaFuncs.showInsertionLine($dest, getInsertionLinePos(def_place))
            pgMakeParentContextMenusTransparent(true)

            $a.on('mouseenter', '[data-pos]', handlePos)
            $a.on('mouseleave', '[data-pos]', handlePosOut)
        }

        d.on_mouseleave = function ($a) {
            if(show_previews) {
                pgPreview.hide();
                delayedTasks.cancel('showPreview');
            }
            crsaFuncs.showInsertionLine(null)

            $a.off('mouseenter', '[data-pos]', handlePos)
            $a.off('mouseleave', '[data-pos]', handlePosOut)

            delayedTasks.run('cancelTransparent', function() {
                pgMakeParentContextMenusTransparent(false)
            }, 500)
        }

    }

    var insert = function(dummy, event) {
        var def = this.def;

        var pos = helper.getPlacementFromEvent(event);

        if(pos) {
            pgInsertLibMenuCurrentPosition = def_place = pos;
        }

        if(!def_place) {
            pinegrow.showQuickError(`Can't insert the element here.`);
            return;
        }
        var api = new PgApi();
        var created = createPgElFromDefinitionMultiple(def);
        if (created) {
            var dest = pinegrow.getCollectionForElement(pgel);

            api.insertElements(created.collection, dest, def_place, {highlight: true, scroll: true, repeater: true, format_html: true});

            pgLibraryLastUsed.add(def);

            pinegrow.showQuickMessage(def.name + ' inserted.');
        }
    }

    var getPlacementOptions = function(def) {

        var m = new PgPlacementOptionsMenu(function(pos) {
            pgInsertLibMenuCurrentPosition = def_place = pos;
            insert.call(this);
        }, function(b) {
            b.def = def;
        }, def_place, pgel)

        return m.getActions();
    }

    var getPosName = function() {
        switch(def_place) {
            case 'insertBefore':
                return 'insert before';
            case 'insertAfter':
                return 'insert after';
            case 'prepend':
                return 'prepend';
            case 'append':
                return 'append';
        }
    }

    var icons = helper.getPlacementIcons();

    actions.unshift({
        label: `Insert into ${def.name}:`,
        type: "header"
    });

    //Quick insert
    if (def.action_menu && def.action_menu.add) {

        var action_menu = {};
        if (def.action_menu.on_add) {
            action_menu.on_add = def.action_menu.on_add;
        }

        $.each(def.action_menu.add, function (i, type) {
            var tdef = currentPage.getTypeDefinition(type);
            if (tdef) {
                actions.push({
                    label: tdef.name + pinegrow.repeater.showLiveRepeatCount(),
                    action: function () {
                        crsaFuncs.insertThroughActionMenu(action_menu, target_pgel, tdef, false);
                    }
                });
            }
        });
    }


    //Lorem ipusm
    if (pgel.canHaveChildren()) {
        actions.push({
            label: "Lorem Ipsum text", action: function () {

                var collection = pinegrow.getCollection(target_pgel);

                collection.forEachElement(function (pgCurrent, i, total) {

                    var cp = pgCurrent.getPage();
                    var def = cp.getMainType(null, pgCurrent);

                    if (!canMakeChange(pgCurrent, 'edit_content')) return;

                    willMakeChange(cp, "Insert Lorem Ipsum to / " + pgCurrent);
                    var text = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasellus pulvinar faucibus neque, nec rhoncus nunc ultrices sit amet. Curabitur ac sagittis neque, vel egestas est. Aenean elementum, erat at aliquet hendrerit, elit nisl posuere tortor, id suscipit diam dui sed nisi. Morbi sollicitudin massa vel tortor consequat, eget semper nisl fringilla. Maecenas at hendrerit odio. Sed in mi eu quam suscipit bibendum quis at orci. Pellentesque fermentum nisl purus, et iaculis lectus pharetra sit amet.';
                    var html = pgCurrent.html();
                    if (['h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'a', 'button', 'li', 'dd', 'dt'].indexOf(pgCurrent.tagName) >= 0) {
                        text = 'Lorem ipsum dolor sit amet';
                    }
                    var $newel = createElementFromDefinition(def);
                    if ($newel && $newel.html() == html) {
                    } else if (html.length > 0) {
                        text = ' ' + text;
                    }

                    pgCurrent.append(pgCreateNodeFromHtml(text));
                })
                pinegrow.changeManager.didChange(collection, 'Add Lorem Ipsum');
            }
        });
    }
    /////

    if(pgLibraryLastUsed.getList().length) {
        var submenu = []
        actions.push({
            type: 'divider'
        })
        actions.push({
            label: 'Recently used',
            submenu: submenu,
            submenu_class: 'pg-lib-menu-items'
        })

        pgLibraryLastUsed.getList().forEach(function(def) {
            var m = helper.getMenuItemWithPlacement(def.name);
            m.def = def;
            m.action = insert;
            addPreview(m);
            submenu.push(m)
        })

        submenu.unshift({
            type: 'header',
            label: `Click to insert:`
        })
    }

    var rec_list = []

    $.each(sections, function(i, sec_def) {

        if(current_f !== sec_def.framework) {
            add_fm_header = true;
            current_f = sec_def.framework;
        }

        var submenu = [];

        var sec_action = {
            label: sec_def.name,
            submenu: submenu,
            submenu_class: 'pg-lib-menu-items'
        }

        var check_pgel = def_place.startsWith('insert') ? pgel.parent : pgel;

        $.each(sec_def.getComponentTypes(), function(i, def) {

            var m = placement ? helper.getMenuItemWithoutPlacement(def.name) : helper.getMenuItemWithPlacement(def.name);
            m.def = def;
            m.action = insert;
            addPreview(m);
            submenu.push(m)

            if(recommended) {
                if(def.on_can_insert_into) {
                    if(def.on_can_insert_into(check_pgel, def)) {
                        var m = placement ? helper.getMenuItemWithoutPlacement(def.name) : helper.getMenuItemWithPlacement(def.name);
                        m.def = def;
                        m.action = insert;
                        addPreview(m);
                        rec_list.push(m)
                    }
                }
            }
        });

        if(submenu.length) {
            submenu.unshift({
                type: 'header',
                label: `Click to insert:`
            })
            if(add_fm_header) {
                actions.push({
                    type: 'divider'
                })
                actions.push({
                    type: 'header',
                    label: current_f.name
                })
                add_fm_header = false;
            }
            actions.push(sec_action)
        }
    })

    if(rec_list.length) {
        //actions = rec_list.concat(actions);

        rec_list.push({
            type: 'divider'
        })

        rec_list.push({
            label: 'More',
            submenu: actions
        })

        actions = rec_list;
    }

    if(placement) {
        actions.unshift({
            type: 'header',
            label: getPosName(placement) + ' ' + pgel.getName({short: true})
        })
    }
    actions.push({
        type: 'divider'
    })
    actions.push({
        type: 'header',
        label: 'Options'
    })
    actions.push({
        label: 'Show previews',
        check: show_previews,
        action: function() {
            show_previews = !show_previews;
            pinegrow.setSetting('lib-insert-menu-show-previews', show_previews ? "1" : "0");
        }
    })
    if(!no_hint) {
        actions.push({
            type: 'html',
            html: `<p class="pg-menu-hint">Use ${pgShowKbdCombo('SHIFT+Right-click')} to<br>open this menu directly.</p>`
        })
    }

    this.getActions = function() {
        return actions;
    }
}

var PgPlacementOptionsMenu = function(action, on_button, def_place, pgel) {

    def_place = def_place || 'insertAfter'

    var fertile = pgel.canHaveChildren();

    var r = []
    r.push({
        type: 'header',
        label: `Choose position:`
    })

    var buttons = [];
    [
        {pos:'insertBefore', name: 'Before'},
        {pos:'prepend', name: 'Prepend', fertile: true},
        {pos:'append', name: 'Append', fertile: true},
        {pos:'insertAfter', name: 'After'}
    ].forEach(function(place) {

        if(!place.fertile || place.fertile === fertile) {

            var b = {
                action: function () {
                    action.call(this, place.pos);
                },
                label: place.name,
                class: `pg-lib-menu-place-button pg-lib-menu-place-${place.pos} ${place.pos === def_place ? 'active' : ''}`
            }

            if(on_button) {
                on_button(b)
            }

            buttons.push(b)
        }
    })
    r.push({
        buttons: buttons,
        html: `<div class="pg-lib-menu-place"><div class="pg-lib-menu-place-dest">${pgel.getName({short: true})}</div></div>`
    })

    this.getActions = function() {
        return r;
    }
}

var PgSelectMenu = function(pgel, only_kids) {

    var list;
    var parents = []
    var p = pgel;

    var select = function() {
        pinegrow.selectElement(this.pgel)
        this.pgel.scrollTo();
    }

    var on_mouseenter = function() {
        pinegrow.highlightElement(this.pgel.get$DOMElement())
    }

    var on_mouseleave = function() {
        pinegrow.highlightElement(null)
    }

    if(!only_kids) {
        while (p.parent && p.parent.isElement) {
            p = p.parent;
            parents.push({
                label: p.getName({short: true}),
                pgel: p,
                action: select,
                on_mouseleave: on_mouseleave,
                on_mouseenter: on_mouseenter
            })
        }

        if (parents.length) {
            parents.unshift({
                type: 'header',
                label: 'Parents'
            })
        }
    }

    var children = [];

    pgel.getChildren().forEach(function(p) {

        var d = {
            label: p.getName({short: true}),
            pgel: p,
            action: select,
            on_mouseleave: on_mouseleave,
            on_mouseenter: on_mouseenter
        }

        if(children.length < 100) {

            children.push(d)

            if (p.getChildren().length) {
                d.submenu = function () {
                    var pgm = new PgSelectMenu(p, true);
                    return pgm.getActions();
                }
            }
        }
    })

    if(children.length) {

        children.unshift({
            type: 'header',
            label: 'Children'
        })
        if(parents.length) {
            children.unshift({
                type: 'divider'
            })
        }
    }

    list = parents.concat(children);

    this.getActions = function() {
        return list;
    }
}


$('body').one('pinegrow-ready', function(e, pinegrow) {

    var quickLibWin = null;

    if(!pgf.kids) {
        pinegrow.addEventHandler('on_key_pressed', function (page, e, status) {
            if ((e.key === '+') && !status.input && !status.edit && !status.ctrl) {
                status.done = true;
                if(pgf.insert_in_top_menu) {
                    showTopMenuInsertLibrary(true);
                } else {
                    pinegrow.showQuickInsert();
                }
                return;


                if (pgel) {
                    var qi = new PgQuickInsert(pgel, 'append');
                } else {
                    pinegrow.showQuickMessage('Please select an element first.')
                }
            }
        })
    }
})
