class PgSmartInput {

    constructor(placeholder) {
        var _this = this;

        this.$element = $('<div class="pg-smart-input"></div>');
        this.$results = $('<div class="pg-smart-input-results"></div>');
        this.results = []
        this.menu = null;

        this.field = new PgInputField('value', {
            type: 'text',
            show_empty: false,
            name: 'Search',
            without_text: true,
            auto_resize: false,
            placeholder: placeholder,
            with_clear_icon: true
        });

        this.field.appendTo(this.$element);

        this.$results.appendTo(this.$element)

        this.field.on('input', this.onValueInput.bind(this))
        this.field.on('focus', this.onFocus.bind(this))

        this.field.on('keydown', this.onKeyDown.bind(this))

        this.$results.on('click', '.result', function(e) {
            var idx = parseInt(this.getAttribute('data-idx'))
            _this.onResultSelected(_this.results[idx])
        })

        var mover = this.mover = new PgArrowsMover(this.field.$input, this.$results, '> .result');
        mover.select_first = true;

        mover.onEnter = function($item) {
            if($item) {
                var idx = parseInt($item.attr('data-idx'))
                _this.onResultSelected(_this.results[idx])
            } else {
                if(_this.results.length) {
                    _this.onResultSelected(_this.results[0])
                } else {
                    pinegrow.showQuickError('Please enter a valid CSS property value.')
                }
            }
        }

        mover.onSelected = function($item) {
            if($item) {
                var idx = parseInt($item.attr('data-idx'))
                _this.onResultPreview(_this.results[idx])
            } else {
                _this.onResultPreview(null)
            }
        }
    }

    onKeyDown(event) {
        if(event.key === 'Tab') {
            event.preventDefault();
            event.stopPropagation();

            if(this.results.length) {
                var $item = this.mover.get$Item();
                if($item) {
                    var idx = $item.attr('data-idx');
                    this.field.setValue(this.results[idx].full_value);
                    this.focus(true);
                }
            }
        }
    }

    addToMenu(menu, $el) {
        this.menu = menu;
        $el.append(this.$element);
    }

    focus(at_end) {
        this.field.focus(at_end)
    }

    showResults(results) {
        var max = 10;
        this.results = results;
        var html = '';

        if(results.length) {
            html += `<div class="info">${pgShowKbd('DOWN')} ${pgShowKbd('UP')} ${pgShowKbd('ENTER')} ${pgShowKbd('TAB')}</div>`;
        }
        var idx = 0;
        for(var idx = 0; idx < results.length; idx++) {
            html += `<div class="result" data-idx="${idx}">${results[idx].html}</div>`
            if(idx === max) break;
        }

        this.$results.get(0).innerHTML = html;
        this.mover.update('data-idx');
    }

    onValueInput(event, value) {
        var val = event ? event.target.value : value;
        var results = this.getResultsForShorthand(val);
        this.showResults(results)

        if(this.menu) {
            this.menu.closeSubmenus();
        }
    }

    onFocus(event) {
        if(this.menu) {
            this.menu.closeSubmenus();
        }
    }

    onResultSelected(result) {
        if(this.menu) {
            //this.menu.close();
        }
    }

    onResultPreview(result) {}

    getResultsForShorthand(short) {
        return []
    }

    destroy() {

    }
}

class PgSmartInputForCSS extends PgSmartInput {

    constructor(stylePreview) {
        super('Example: mt10');
        this.matcher = new PgSmartCSSPropsMatcher();

        this.currentMatchedProps = []
        this.props = {
            'border-top': ''
        }

        this.stylePreview = stylePreview;
    }

    getResult(prop, val) {
        return {
            html: `<span class="prop">${prop}</span>: <span class="value">${val}</span>`,
            prop: prop,
            value: val,
            full_value: prop + (val.length ? (' ' + val.trim()) : '')
        }
    }

    onResultPreview(result) {
        if(this.stylePreview) {
            if (result) {
                this.stylePreview.setStyle(result.prop, result.value);
            } else {
                this.stylePreview.restore()
            }
        }
    }

    getResultsForShorthand(short) {
        var _this = this;

        pinegrow.highlightElement(null);

        return this.matcher.matchProps(short, pinegrow.devTools.getCSSPropertiesOptions()).map(function(pi) {
            return _this.getResult(pi.key, pi.value || '');
        })

        var _this = this;
        var a = short.split(/[:\s]+/);
        var r = []
        if(a.length > 1) {
            r.push(this.getResult(a[0], short.replace(a[0], '').replace(/^[\s:]+/,'')));
        } else {
            var list;
            if(short.length === 1 || true) {
                list = pinegrow.devTools.getCSSPropertiesOptions()
            } else {
                list = this.currentMatchedProps;
            }
            this.currentMatchedProps = this.matchProps(short, list);

            this.currentMatchedProps.forEach(function(p) {
                r.push(_this.getResult(p.key, 'inherit'))
            })
        }


        return r;
    }

    onResultSelected(result) {
        super.onResultSelected(result);
        if(this.onPropSelected(result.prop, result.value) !== false) {
            this.field.setValue('');
            this.onValueInput(null, '')
        }
    }

    onPropSelected(prop, value) {

    }
}


class PgSmartCSSPropsMatcher {

    constructor() {
        this.force = {
            m: 'margin',
            mb: 'margin-bottom',
            mt: 'margin-top',
            ml: 'margin-left',
            mr: 'margin-right',
            p: 'padding',
            pb: 'padding-bottom',
            pt: 'padding-top',
            pl: 'padding-left',
            pr: 'padding-right',
            c: 'color',
            bgc: 'background-color'
        }
    }

    test(short) {
        console.log(`Match ${short}:`)
        var testList = [
            {key: 'border'},
            {key: 'border-top'},
            {key: 'border-top-style', values: [
                'solid', 'dashed', 'dotted'
                ]},
            {key: 'display', values: [
                'block', 'inline', 'flex', 'grid', 'inline-block', 'inline-flex'
                ]
            }
        ]
        testList = pinegrow.devTools.getCSSPropertiesOptions()
        var r = this.matchProps(short, testList).slice(0,10)
        r.forEach(function(r) {
            console.log(`${r.key} ${r.value} ${r.distance}`)
        })
    }

    matchProps(short, list, match_values) {
        var _this = this;
        //first letter should start the prop
        var first = short.charAt(0);
        var chars = short.split('');

        var short_prop = short;

        var vip_re = /^(margin|padding)/;
        var ignore_units = ['background', 'background-image', 'background-repeat', 'line-height', 'flex-grow', 'flex-shrink', 'flex', 'z-index', 'opacity', 'font-weight', 'flex-size', 'order', 'grid-row', 'grid-column', 'grid-row-start', 'grid-row-end', 'grid-column-start', 'grid-column-end', 'columns', 'column-rule']

        var r = [];
        list.forEach(function(prop) {
            var name = match_values ? prop : prop.key;
            if(name.charAt(0) === first) {
                var value = '';
                var in_value = false;
                var prev_match_idx = 0;
                var dist = 1;
                for(var i = 1; i < chars.length; i++) {
                    var has = false;
                    if(!in_value) {
                        for (var j = prev_match_idx + 1; j < name.length; j++) {
                            if (chars[i] === name.charAt(j)) {
                                if (chars[i] === '-' && i + 1 < chars.length && chars[i + 1] !== name.charAt(j + 1)) {
                                    //if char is - the next char must match (-x)
                                    return;
                                }
                                prev_match_idx = j;
                                has = true;
                                if (name.charAt(j - 1) === '-') {
                                    //promote matches right after -
                                    dist += j / 2;
                                } else {
                                    dist += j;
                                }
                                short_prop = short;
                                break;
                            }
                        }
                    }
                    if(!has && !match_values) {
                        //is this the value already?
                        if(!in_value && (chars[i] === ':' || chars[i] === ' ')) {
                            in_value = true;
                            short_prop = short.substr(0, i);
                            continue;
                        }
                        if(chars[i].match(/[0-9]/)) {
                            in_value = true;
                            short_prop = short.substr(0, i);
                            /*
                            r.push({
                                key: name,
                                prop: name,
                                value: value,
                                distance: dist,
                                //html: html
                            })
                            return;*/
                            //break;
                        } else {
                            //check values
                            if(prop.values) {
                                var value_short = short.substr(i);
                                var matched_values = _this.matchProps(value_short, prop.values, true);
                                if(matched_values.length) {

                                    if(value !== '') {
                                        var spaces = '';
                                        //handle 'border 10 solid'
                                        if(value.endsWith(' ')) {
                                            value = value.trimEnd();
                                            spaces = ' ';
                                        }
                                        //add units if needed
                                        var val_info = pgParseCSSValue(value);
                                        if(val_info.type === 'value' && !val_info.unit) {
                                            if(ignore_units.indexOf(name) < 0) {
                                                value += 'px';
                                            }
                                        }
                                        value += spaces;
                                    }

                                    if(value.length && !value.endsWith(' ')) {
                                        value += ' '
                                    }
                                    matched_values.forEach(function (vr) {
                                        r.push({
                                            key: name,
                                            prop: name,
                                            value: value + vr.key,
                                            distance: dist + 100,
                                            //html: html
                                        })
                                    })
                                    return;
                                }
                            }
                        }
                        if(in_value) {
                            value += chars[i];
                            continue;
                        }
                        return;
                    } else if(!has && !in_value) {
                        return;
                    }
                }

                if(value !== '') {
                    dist += 100; //matching props should have priority

                    //add units if needed
                    value = value.trimEnd();
                    var val_info = pgParseCSSValue(value);
                    if(val_info.type === 'value' && !val_info.unit) {
                        if(ignore_units.indexOf(name) < 0) {
                            value += 'px';
                        }
                    }
                }
                if(_this.force[short_prop] && name === _this.force[short_prop]) {
                    dist = 0;
                }

                if(name.match(vip_re)) {
                    dist = dist / 2;
                }
                r.push({
                    key: name,
                    prop: name,
                    value: value,
                    distance: dist,
                    //html: html
                })
            }
        })

        r.sort(function(a, b) {
            if(a.distance < b.distance) {
                return -1;
            } else if(a.distance === b.distance) {
                if(a.prop.length < b.prop.length) {
                    return -1;
                } else if(a.prop.length === b.prop.length) {
                    return 0;
                }
                return 1;
            } else {
                return 1;
            }
        })

        return r;
    }

    runTests() {
        this.test('dii')
        //this.test('bo10')
        //this.test('bo-t');
    }
}
