


var PgPageTreeManager = function() {
    var event_handler = new PgEventHandler();

    //event_handler.handle_dom_events = false;//not interested in dom-only events like executeScript

    event_handler.on_event = function(event) {
        event.page.tree && event.page.tree.handleEvent(event);
    }

    event_handler.start();

    this.destroy = function() {
        event_handler.destroy();
    }
}

var PgTreeView = function() {

    pinegrow.registerTreeView(this);

    var $element = $('<div/>', {id: 'pg-tree', class: 'pg-tree'});

    var $tree = $('<div/>', {class: 'pg-tree-container'}).appendTo($element);

    var search = new PgSearchBox({
        toolbar_on: pgf.kids ? false : true
    });

    if(!pgf.kids) {

        search.addIcon({
            icon: 'icon-130',
            label: 'Select search results',
            func: function () {
                searchSelect('select');
            }
        })

        search.addIcon({
            icon: 'icon-137',
            label: 'Add search results to selection',
            func: function () {
                searchSelect('add');
            }
        })

        search.addIcon({
            icon: 'icon-138',
            label: 'Deselect search results',
            func: function () {
                searchSelect('deselect');
            }
        })

    }

    var toolbar = this.toolbar = search.getToolbar();

    if(!pgf.kids) {
        var $toolbar_section = toolbar.addSection('settings').$element;
    } else {
        search.$element.css({
            'margin-right': '10px'
        })
    }

    var stylesButton = new PgButtonToolbarItem('icon-filter', 'Select what kind of elements to show in the tree.');

    if(!pgf.kids) {
        toolbar.addItem('settings', stylesButton);
    }

    var filters = {
        styles: {label: 'Inline styles', desc: 'Show only the elements with inline styles'},
        images: {label: 'Images', desc: 'Show only images'},
        links: {label: 'Links', desc: 'Show only links'},
        texts: {label: 'Texts', desc: 'Show only the elements with textual content'},
        components: {label: 'Components &amp; Editable', desc: 'Show only components and editable areas'}
    }

    if(pinegrow.hasActivatedProduct('WP') || pinegrow.isTrial()) {
        filters.wp = {label: 'WordPress actions', desc: 'Show only the elements with WordPress actions'};
    }

    if(pinegrow.hasActivatedProduct('WC') || pinegrow.isTrial()) {
        filters.wc = {label: 'WooCommerce actions', desc: 'Show only the elements with WooCommerce actions'};
    }

    var show_interactions = false;

    if(pinegrow.hasActivatedProduct('IA') || pinegrow.isTrial()) {
        filters.ia = {label: 'Interactions', desc: 'Show only the elements with Interactions'};
        show_interactions = pinegrow.getSetting('pgia-show-in-tree', '1') === '1';
    }

    if(!pgf.components) {
        delete filters.components;
    }

    var default_filter_options = {};

    $.each(filters, function(key, name) {
        default_filter_options[key] = false;
    })

    var filter_options = default_filter_options;

    stylesButton.onClick = function () {

        var menu = new CrsaContextMenu(null, this.$element);

        menu.add({
            type: 'header',
            label: 'Filter (show el. with):'
        });

        $.each(filters, function(key, info) {
            menu.add({
                label: info.label,
                action: function() {
                    var on = !(filter_options[key] === true);
                    _this.setFilterOption(on ? key : null);
                },
                check: filter_options[key] === true
            })
        })

        menu.add({
            type: 'divider'
        });

        menu.add({
            label: 'Clear filter',
            action: function() {
                _this.setFilterOption(null);
            }
        })

        menu.add({
            type: 'divider'
        });

        menu.add({
            type: 'header',
            label: 'Settings'
        });

        if(pinegrow.hasActivatedProduct('IA') || pinegrow.isTrial()) {
            menu.add({
                label: 'Show interactions',
                action: function() {
                    show_interactions = !show_interactions;
                    pinegrow.setSetting('pgia-show-in-tree', show_interactions ? '1' : '0');
                    page.tree.setShowInteractions(show_interactions);
                    page.tree.create();
                    showTreeForPage(page);
                },
                check: show_interactions === true
            })
        }

        menu.showForElement(this.$element, null, 8);
    }

    this.setFilterOption = function(key, skip_update) {
        filter_options = $.extend({}, default_filter_options);
        if(key) {
            var info = filters[key];
            filter_options[key] = true;
            stylesButton.setActive(true);
            pinegrow.showMode(info.desc, true);
            setFilterMessage(info.label);
        } else {
            stylesButton.setActive(false);
            setFilterMessage();
        }
        page.tree.setFilterOptions(filter_options);
        if(!skip_update) {
            if (currentPage) currentPage.tree.research(true);
        }
    }


    var modeButton = new PgButtonToolbarItem('icon-down', 'Tree options (Source or DOM, auto collapse...)');

    if(!pgf.kids) {
        toolbar.addItem('settings', modeButton);
    }

    var dom_mode = pinegrow.getSetting('tree-dom-mode', '0') === '1';
    var auto_collapse_level = parseInt(pinegrow.getSetting('tree-auto-collapse', isApp() ? 0 : 0));

    modeButton.onClick = function () {

        var actions = [
            {
                type: 'header',
                label: 'Show'
            },
            {
                label: 'Source code structure',
                help: 'Show the page structure as defined in the source code.',
                check: !dom_mode,
                action: function() {
                    dom_mode = false;
                    pinegrow.showMode('Show source code structure', true)

                    pinegrow.setSetting('tree-dom-mode', dom_mode ? '1' : '0');
                    page.tree.setDomMode(dom_mode);
                    page.tree.create();
                    showTreeForPage(page);
                }
            },
            {
                label: 'DOM structure',
                help: 'Show the current DOM browser structure.',
                check: dom_mode,
                action: function() {
                    dom_mode = true;
                    pinegrow.showMode('Show DOM structure', true)

                    pinegrow.setSetting('tree-dom-mode', '1');
                    page.tree.setDomMode(dom_mode);
                    page.tree.create();
                    showTreeForPage(page);
                }
            },
            {
                type: 'divider'
            },
            {
                type: 'header',
                label: 'Options'
            },
            {
                label: 'Auto collapse',
                check: auto_collapse_level > 0,
                submenu: [0, 'Off', 1,2,3,4,5].map(function(level) {
                    if(level === 0) {
                        return {
                            type: 'header',
                            label: 'Collapse from level'
                        }
                    } else {
                        return {
                            label: level,
                            action: function() {
                                auto_collapse_level = level === 'Off' ? 0 : level;
                                pinegrow.setSetting('tree-auto-collapse', auto_collapse_level);
                                page.tree.setAutoCollapse(auto_collapse_level);
                                page.tree.create();
                                showTreeForPage(page);
                                pinegrow.showMode('Auto collapse tree', level !== 'Off');
                            },
                            check: auto_collapse_level === (level === 'Off' ? 0 : level)
                        }
                    }
                })
            }
        ]
        var m = new CrsaContextMenu(actions, modeButton.$element);
        m.showForElement(modeButton.$element);
        return false;
    }

    //modeButton.setActive(!dom_mode);

    var $selects = search.$element.find('.pg-button-toolbar-section.icons');

    if(!pgf.kids) {
        search.$element.appendTo($element);

        var spacer = new PgSearchBoxSpacer();
        spacer.$element.appendTo($element);
    }

    search.onSearch = function(filter) {
        var p = new CrsaProfile();
        if(currentPage) currentPage.tree.search(filter);

        if(filter.length) {
            pg$Show($selects, 'inline-block');
        } else {
            pg$Hide($selects);
        }

        p.show('SEARCH');
    }

    pg$Hide($selects);

    var searchSelect = function(mode) {
        var col = new PgCollection();
        tableGrid.data.forEach(function(data) {
            if(data.search === 1) {
                var pgel = _this.getPgel(data);
                if(pgel) {
                    col.add(pgel);
                }
            }
        })
        var multi = mode == 'add';

        if(mode == 'select') {
            pinegrow.selectedElements.clear();
        }
        if(mode == 'deselect') {
            pinegrow.selectedElements.deselectMany(col);
        } else {
            pinegrow.selectedElements.selectMany(col);
        }
    }


    var $filterMsg = $('<div class="pg-tree-filter-msg"><div>Show only: <span></span><i class="close icon icon-close"></i></div></div>').appendTo($element);

    pg$Hide($filterMsg);

    $filterMsg.on('click', function(e) {
        e.preventDefault();
        _this.setFilterOption(null);
    })

    var setFilterMessage = function(msg) {
        if(msg) {
            $filterMsg.find('span').html(msg);
            pg$Show($filterMsg);
            $element.addClass('with-filter-msg');
        } else {
            pg$Hide($filterMsg);
            $element.removeClass('with-filter-msg');
        }
    }

    if(pgf.auto_focus) {

        var $focusMsg = $('<div class="pg-tree-filter-msg"><div>Focus on: <span></span><i class="icon icon-down"></i></div><label><input type="checkbox" name="checkbox" value="1"' + (pinegrow.getAutoFocus() ? ' checked' : '') + '>Auto-focus on sections</label></div>').appendTo($element);

        //pg$Hide($focusMsg);

        $focusMsg.find('> div').on('click', function (e) {
            e.preventDefault();
            //pinegrow.isolateElement(null);

            var $el = $(this)

            var page = pinegrow.getSelectedPage();

            if (page) {
                var actions = [];

                if (pinegrow.getIsolatedElement(page)) {
                    actions.push({
                        label: 'Show whole page',
                        action: function () {
                            pinegrow.isolateElement(null)
                        }
                    })

                    actions.push({
                        type: 'divider'
                    })
                }

                actions.push({
                    type: 'header',
                    label: 'Focus on:'
                })

                actions.push({
                    label: 'Page structure',
                    action: function () {
                        pinegrow.isolateElement(page.sourceNode.findOne('body'), {
                            tree_hide_scripts: true,
                            drag_drop_only_top_level: true
                        })
                    }
                })

                var imp_sel = 'header,footer,section';

                page.sourceNode.find(imp_sel).forEach(function (pgel) {
                    actions.push({
                        label: pgel.getName({short: true}),
                        action: function () {
                            pinegrow.isolateElement(pgel, {}, true);
                            pinegrow.selectElement(pgel);
                            pgel.scrollTo();
                        }
                    })
                })

                var m = new CrsaContextMenu(actions, $el);
                m.showForElement($el);
            } else {
                pinegrow.showQuickError('Please open a page first')
            }

            return false;
        })

        $focusMsg.find('input').on('change', function (e) {
            var val = $(this).is(':checked');
            pinegrow.setAutoFocus(val);
        })

        var setFocusMessage = function (msg) {
            if (msg) {
                $focusMsg.find('> div').html(`Focus on <span>${msg}</span><i class="icon icon-down"></i>`);
                $focusMsg.removeClass('inactive');
                //pg$Show($focusMsg);
                //$element.addClass('with-filter-msg');
            } else {
                //pg$Hide($focusMsg);
                $focusMsg.find('> div').html(`Whole page is displayed<i class="icon icon-down"></i>`);
                $focusMsg.addClass('inactive');
                //$element.removeClass('with-filter-msg');
            }
        }
    }
    /*
    OFF Auto focus on sections
    HAS Focus on Section  - Disable auto-focus
    NONE Focus on whole Page - Disable auto-focus

     */


    var $content = $element.find('.editor');

    var view = this.view = new PgUIView($element, 'tree', 'Tree');

    view.classForContainer = 'pg-tree-panel-container';

    var _this = this;

    view.tabIcon = 'icon-Pin11';

    var currentPage;
    var page;

    var manager = new PgPageTreeManager();

    view.onShow = function() {
        //scroll to selected element if not visible
        if(currentPage && currentPage.tree.scrollToSelected) {
            tableGrid.scrollToDelayed(currentPage.tree.scrollToSelected);
            currentPage.tree.scrollToSelected = null;
        }
    }


    var withPgel = function(data, func, single) {
        var pgel = _this.getPgelOfDataItem(data);
        if(pgel) {
            if(pinegrow.selectedElements.isSelectedElement(pgel) && !single) {
                func(pinegrow.selectedElements.getCollection());
            } else {
                func(pgel);
            }
        }
    }

    var showItem = function(data) {
        withPgel(data, function(pgel) {
            var first = pgel instanceof pgParserNode ? pgel : pgel.getFirst();
            if(!first.isHidden()) {
                first.scrollTo();
                setTimeout(function() {
                    pinegrow.highlightElement(first.get$DOMElement());
                }, 50);
            } else {
                var c = pinegrow.getCollection(pgel);
                c.forEachElement(function (pgel) {
                    pgel.show();
                })
            }
        })
    }

    var indentOutdent = function(data, what) {
        var all_done = true;
        var some_done = false;
        var count = 0;
        var c = null;
        withPgel(data, function(pgel) {
            c = pinegrow.getCollection(pgel);
            c.forEachElement(function(pgel) {
                count++;
                var d = pgel[what]();
                all_done = all_done && d;
                some_done = some_done || d;
            })
        })
        if(!all_done) {
            if(count == 1) {
                pinegrow.showQuickMessage('The element could not be moved.')
            } else {
                pinegrow.showQuickMessage('Some elements could not be moved.')
            }
        }
        if(some_done) {
            pinegrow.changeManager.didChange(c, 'Level ' + what);
        }
    }

    var collapseExpandAll = function(data, collapse) {
        tableGrid.forEachDataAndDescendants(data, function(ch) {
            ch.collapsed = collapse;
            var pgel = _this.getPgel(ch);
            if(pgel) {
                setCollapsedForPgel(pgel, ch.collapsed);

                /*
                if (ch.collapsed) {
                    pgel.setAttr('data-pg-collapsed');
                } else {
                    pgel.removeAttr('data-pg-collapsed');
                }*/
            }
        })
        tableGrid.refresh();
    }

    var setCollapsedForPgel = function(pgel, val) {
        if(pgel.tagName === 'svg') {
            if(val) {
                pgel.removeAttr('data-pg-collapsed');
            } else {
                pgel.setAttr('data-pg-collapsed', 'false');
            }
        } else {
            val ? pgel.setAttr('data-pg-collapsed') : pgel.removeAttr('data-pg-collapsed');
        }
        pinegrow.dispatchEventDelayed('on_elements_collapsed_expanded', pgel.getPage());
    }

    var menuItems = []

    var is_editable = true;//['html','head','body'].indexOf(pgel.tagName) < 0;

    if(pgf.edit_html && is_editable) {

        menuItems.push(
            {
                icon: 'icon-bin',
                name: 'Delete',
                func: function (data) {
                    withPgel(data, function (pgel) {
                        var api = new PgApi();
                        api.removeElements(pgel);
                    })
                },
                show: function(data, ctx) {
                    const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                    ctx.pgel = pgel;
                    return pgel && pgel.canDelete();
                }
            })

        menuItems.push({
            icon: 'icon-duplicate',
            name: 'Duplicate',
            func: function (data) {
                withPgel(data, function (pgel) {
                    var api = new PgApi();
                    api.duplicateElements(pgel);
                })
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canDuplicate();
            }
        })
    }

    if(is_editable && !pgf.kids) {
        menuItems.push({
            icon: 'icon-hide',
            name: 'Hide / show in Pinegrow',
            key: 'hide',
            func: function (data, $item) {
                withPgel(data, function (pgel) {
                    var c = pinegrow.getCollection(pgel);
                    var hide = $item.find('i').hasClass('icon-hide');
                    c.forEachElement(function (pgel) {
                        if (hide) {
                            pgel.hide();
                        } else {
                            pgel.show();
                        }
                    })

                    $item.find('i').removeClass('icon-hide icon-see').addClass(hide ? 'icon-see' : 'icon-hide');
                })
            }
        })
    }

    if(pgf.edit_html && !pgf.kids) {
        menuItems.push({
            icon: 'icon-up',
            name: 'Move up',
            func: function (data) {
                var api = new PgApi();
                withPgel(data, function(pgel) {
                    api.moveUp(pgel)
                })
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canMove();
            }
        })

        menuItems.push({
            icon: 'icon-down',
            name: 'Move down',
            func: function (data) {
                var api = new PgApi();
                withPgel(data, function(pgel) {
                    api.moveDown(pgel)
                })
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canMove();
            }
        })

        menuItems.push({
            icon: 'icon-131',
            name: 'Indent one level',
            func: function (data) {
                indentOutdent(data, 'indent');
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canMove();
            }
        })

        menuItems.push({
            icon: 'icon-132',
            name: 'Outdent one level',
            func: function (data) {
                indentOutdent(data, 'outdent');
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canMove();
            }
        })

        menuItems.push({
            icon: '<i class="icon icon-Ins_before"><span class="path1"></span><span class="path2"></span></i>',
            name: 'Insert before',
            func: function (data) {
                showQuickInsertIntoTree(data, 'insertBefore');
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canHaveSiblings();
            }
        })

        menuItems.push({
            icon: '<i class="icon icon-Ins_after"><span class="path1"></span><span class="path2"></span></i>',
            name: 'Insert after',
            func: function (data) {
                showQuickInsertIntoTree(data, 'insertAfter');
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canHaveSiblings();
            }
        })

        menuItems.push({
            icon: '<i class="icon icon-Append"><span class="path1"></span><span class="path2"></span></i>',
            name: 'Insert into at beginning (prepend)',
            func: function (data) {
                showQuickInsertIntoTree(data, 'prepend');
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canHaveChildren();
            }
        })

        menuItems.push({
            icon: '<i class="icon icon-Prepend"><span class="path1"></span><span class="path2"></span></i>',
            name: 'Insert into at end (append)',
            func: function (data) {
                showQuickInsertIntoTree(data, 'append');
            },
            show: function(data, ctx) {
                const pgel = ctx.pgel || _this.getPgelOfDataItem(data);
                ctx.pgel = pgel;
                return pgel && pgel.canHaveChildren();
            }
        })
    }

    if(!pgf.kids) {
        menuItems.push({
            icon: 'icon-right',
            name: 'Collapse / Expand all',
            func: function (data) {
                var pgel = _this.getPgelOfDataItem(data);
                var collapse = !pgel.isCollapsed();
                collapseExpandAll(data, collapse)
            }
        })

        /*
        menuItems.push({
            icon: 'icon-down',
            name: 'Expand all',
            func: function (data) {
                collapseExpandAll(data, false)
            }
        })

         */
    }


    var menu = new PgTreeNodeMenu(menuItems);

    menu.hide();

    this.setPage = function(p) {
        page = p;
    }

    var dimension;

    var hasHScrollBar = true;
    var hasVScrollBar = true;

    var tableGrid = new PgTableGrid();

    if(pgf.kids) {
        tableGrid.$element.css('top', 0);
    }

    tableGrid.draggable = pgf.edit_html;

    var quickInsertConnector = null;

    this.getPgel = function(data) {
        return this.getPgelOfDataItem(data);
        this.populateDataItem(data);
        return data.pgel || null;
    }

    tableGrid.onScroll = function() {
        if(quickInsertConnector) {
            quickInsertConnector.updatePosition();
        }
        updateMenuPosition();
    }

    $tree.append(tableGrid.$element);

    menu.$element.appendTo($element);

    tableGrid.setTemplate('<div><i data-element="icon" class="collapse-icon icon-down"></i><name data-element="name"></name><i data-element="hidden" class="icon icon-hide show-hide"></i></div>');

    tableGrid.setTemplate('<div><i data-element="icon" class="collapse-icon icon-113"></i><name data-element="name"></name><div data-element="progress" class="pgia-progress"></div></div>', 'interaction');

    tableGrid.setTemplate('<div><i data-element="icon" class="collapse-icon icon-Scene_Scroll"></i><name data-element="name"></name><div data-element="progress" class="pgia-progress"></div></div>', 'scrollscene');

    tableGrid.onSetDataToCell = function(data, cell) {

        _this.populateDataItem(data);

        if(cell.type === 'interaction' || cell.type === 'scrollscene') {
            cell.setHtml('name', data.name);// + ' ' + data.pgia.id);
            var p = cell.getElement('progress').style.width = '0';

            pgInteractionsTreeCellManager.register(data.pgia.id, cell);

            var name_el = cell.getElement('name');

            if(data.pgia.desc) {
                name_el.setAttribute('title', data.pgia.desc);
                name_el.classList.add('has-desc');
            } else {
                name_el.removeAttribute('title');
                name_el.classList.remove('has-desc');
            }
        } else {

            cell.setHtml('name', data.name);

            if ((data.selected || false) !== (cell.isSelected || false)) {
                if (data.selected) {
                    cell.$element.addClass('selected');
                } else {
                    cell.$element.removeClass('selected');
                }
                cell.isSelected = data.selected;
            }

            var icon_cls = null;

            if (tableGrid.hasChildren(data)) {
                //cell.removeClass('icon', 'noicon');
                if (data.collapsed) {
                    icon_cls = 'icon-right';

                    //cell.addClass('icon', 'icon-right');
                    //cell.removeClass('icon', 'icon-minus', 'icon-down');
                    cell.$element.addClass('collapsed');
                } else {
                    icon_cls = 'icon-down';

                    //cell.removeClass('icon', 'icon-minus', 'icon-right');
                    //cell.addClass('icon', 'icon-down');
                    cell.$element.removeClass('collapsed');
                }
            } else {
                icon_cls = data.icon || 'icon-minus';

                //cell.removeClass('icon', 'icon-down', 'icon-right');
                //cell.addClass('icon', );
                cell.$element.removeClass('collapsed');
            }
            if (icon_cls !== (cell.icon_cls || 'icon-down')) {
                cell.removeClass('icon', cell.icon_cls || 'icon-down');
                cell.addClass('icon', icon_cls);
                cell.icon_cls = icon_cls;
            }
            if (data.hide_in_pg) {
                cell.getElement('hidden').style.display = 'block';
                cell.$element.addClass('hide-in-pg');
            } else if (data.hide_in_pg_child) {
                cell.getElement('hidden').style.display = 'none';
                cell.$element.addClass('hide-in-pg');
            } else {
                cell.getElement('hidden').style.display = 'none';
                cell.$element.removeClass('hide-in-pg');
            }

            if (data.pgel) {
                cell.$element.removeClass('dyn');
            } else {
                cell.$element.addClass('dyn');
            }

            if (!pgf.focus_in_tree) {
                if (data.focused) {
                    cell.$element.addClass('focused');
                } else {
                    cell.$element.removeClass('focused');
                }
            }
        }

        var search = data.search || 0;
        cell.$element.removeClass('search-1 search-2 search-0').addClass('search-' + search);
    }

    tableGrid.onRepaint = function() {
        menu.setBottom(tableGrid.getHScrollBarSize());
        menu.setRight(tableGrid.getVScrollBarSize());
        updateMenuPosition();
    }

    this.getPgelOfDataItem = function (data) {
        if (data.pgid) {
            return getPgNodeByPgId(data.pgid);
        } else if(data.dyn_element) {
            //dynamic element
            var pgel = getElementPgNode($(data.dyn_element));
            if(pgel) {
                data.pgid = pgel.getId();
                if(currentPage && currentPage.tree) {
                    currentPage.tree.treeData_PgIdMap[data.pgid] = data;
                }
                return pgel;
            }
        }
        return null;
    }

    var classes_filter = null;
    var num_classes = -1;

    this.populateDataItem = function (data, html, pgel) {

        if(data.populated) return;

        data.populated = true;

        pgel = pgel || data.pgel || _this.getPgelOfDataItem(data);
        if(pgel) {

            var def = currentPage.getMainType(null, pgel);

            if (def) {

                var action_tag = currentPage.getActionTag(pgel);

                var info = {
                    def: def,
                    num_classes: num_classes >= 0 ? num_classes : (action_tag ? 1 : 2),
                    action_tag: action_tag,
                    classes_filter: classes_filter
                };

                info.name = pgel.getAttr('data-pg-name');

                if(pgel.parent && pgel.parent.rootNode && !pgel.dynamic) {
                    //page top
                    var page = pgel.getPage();
                    if(page) {
                        info.name = info.name ? ('<pagename>' + page.getName() + '</pagename> / ' + info.name) : ('<pagename>' + page.getName() + '</pagename>');
                    }
                }

                data.name = pgel.getName(info); //getElementName($el, def, true, true /* cls */, true, false, action_tag, _this.currentCrsaPage);

                data.search_string = info.search_string || '';

                if (def.tags) {
                    //div.className += ' tag-' + def.tags;
                }

                var action_tag_search = currentPage.getActionTag(pgel, true);
                if(action_tag_search) {
                    data.search_string += ':' + action_tag_search;
                    //data.name += ' <action>' + action_tag + '</action>';
                }

                //data.selected = pinegrow.selectedElements.isSelectedElement(pgel);
            }

            if(pgel.getAttribute('style') && !pgf.kids) {
                data.name += ' <icons><i class="inline-styles icon icon-Pin15" title="Inline style is set. Click to edit it in the Style panel."></i></icons>';
            }
            data.pgel = pgel;

            data.hide_in_pg = pgel.hasAttr('data-pg-hidden');

        } else {
            var node = getElementOfTreeNodeFast(data);
            if(node) {
                var info = {};
                data.name = pgGetDOMElementName(node, true, true, info);
                data.search_string = info.search_string;
            }
        }
    }


    var populateAll = function() {
        var html = currentPage.get$Html().get(0);
        tableGrid.data.forEach(function(item) {
            _this.populateDataItem(item, html);
        })
    }

    var activeQuickInsertPosition = null;
    var activeQuickInsertData = null;
    var activeQuickInsertWin = null;

    var infoPositionToQuickInsertPosition = function(p) {
        switch(p) {
            case 'left':
            case 'top':
                return 'insertBefore';
            case 'right':
            case 'bottom':
                return 'insertAfter';
            case 'inside-top':
            case 'inside-left':
                return 'prepend';
            case 'inside-bottom':
            case 'inside-right':
                return 'append';
        }
        return p;
    }

    this.markInsertPositionInTree = function(data_or_pgel, quickInsertPosition, skip_scroll) {
        var data;
        if(data_or_pgel instanceof pgParserNode) {
            if(currentPage == data_or_pgel.getPage()) {
                data = currentPage.tree.getTreeDataItemForPgId(data_or_pgel.getId());
                quickInsertPosition = infoPositionToQuickInsertPosition(quickInsertPosition);
            }
        } else {
            data = data_or_pgel;
        }
        if(data && !data.hidden) {
            var idx, level, add_rows;
            add_rows = 0;
            if (quickInsertPosition == 'insertBefore') {
                idx = data.visible_idx;
                level = data.level;
            } else if (quickInsertPosition == 'insertAfter') {
                idx = data.visible_idx + 1 + tableGrid.getNumberOfDescendants(data.visible_idx, tableGrid.visibleData);
                level = data.level;
            } else if (quickInsertPosition == 'prepend') {
                idx = data.visible_idx + 1;
                level = data.level + 1;
                add_rows = 1;
            } else if (quickInsertPosition == 'append') {
                idx = data.visible_idx + 1 + tableGrid.getNumberOfDescendants(data.visible_idx, tableGrid.visibleData);
                level = data.level + 1;
                add_rows = 1;
            }
            if(quickInsertPosition === 'self') {
                tableGrid.showInsertionLineAt(-1);
                tableGrid.setEmptyLineAt(-1);
                tableGrid.borderItem(data, add_rows);
            } else {
                tableGrid.showInsertionLineAt(idx, level);
                tableGrid.setEmptyLineAt(idx);
                tableGrid.borderItem(data, add_rows);
            }

            if(!skip_scroll) {
                tableGrid.scrollTo(data);
            }
        } else {
            tableGrid.showInsertionLineAt(-1);
            tableGrid.setEmptyLineAt(-1);
            tableGrid.borderItem(null);
        }
    }

    this.markRangeInTree = function(pgel, pgel_end) {
        var data, data_end;
        if(pgel && pgel_end && currentPage == pgel.getPage()) {
            data = currentPage.tree.getTreeDataItemForPgId(pgel.getId());
            data_end = currentPage.tree.getTreeDataItemForPgId(pgel_end.getId());
        }

        if(data && !data.hidden && data_end && !data_end.hidden) {
            var idx, level, add_rows;
            add_rows = 1 + tableGrid.getNumberOfDescendants(data.visible_idx, tableGrid.visibleData);
            if(pgel != pgel_end) {
                add_rows = data_end.visible_idx - data.visible_idx + tableGrid.getNumberOfDescendants(data_end.visible_idx, tableGrid.visibleData) + 1;
            }
            tableGrid.scrollTo(data);
            return tableGrid.borderItem(data, false, add_rows);

        } else {
            tableGrid.borderItem(null);
            return false;
        }
    }

    var showQuickInsertIntoTree = function(data, quickInsertPosition) {
        if(!data) return false;

        var pgel = _this.getPgel(data);
        if(!pgel) return false;

        activeQuickInsertPosition = quickInsertPosition;

        var markTree = function() {
            _this.markInsertPositionInTree(data, activeQuickInsertPosition);
        }

        pgQuickWindowManager.close('quickInsert');

        activeQuickInsertData = data;
        markTree();

        tableGrid.quickInsertLine.hide();

        var win = activeQuickInsertWin = new PgQuickInsert(pgel, quickInsertPosition, false, tableGrid.$insertionLine);

        quickInsertConnector = new PgConnector();
        quickInsertConnector.$from = tableGrid.$insertionLine;
        quickInsertConnector.toOffset = 0;

        quickInsertConnector.$to = win.getWindow().$element;

        pinegrow.connectors.add(quickInsertConnector);

        win.getWindow().onMoveDrag = win.getWindow().onMove = win.getWindow().onResize = function() {
            quickInsertConnector.updatePosition();
        }

        var onPageChangedDone = function(page) {
            if(page == currentPage) {
                tableGrid.updateDataIndexes();
                markTree();
                quickInsertConnector.updatePosition();
            }
        }

        pinegrow.addEventHandler('on_page_changes_done', onPageChangedDone);

        win.getLib().onBeforeInserted = function() {
            if((activeQuickInsertPosition == 'prepend' || activeQuickInsertPosition == 'append') && data.collapsed) {
                data.collapsed = false;
                pgel.removeAttr('data-pg-collapsed');
            }
        }

        win.getWindow().onDestroy = function() {

            quickInsertConnector.destroy();
            quickInsertConnector = null;
            tableGrid.setEmptyLineAt(-1);
            tableGrid.showInsertionLineAt(-1);
            tableGrid.borderItem(null);
            activeQuickInsertData = null;
            activeQuickInsertWin = null;
            pinegrow.removeEventHandler('on_page_changes_done', onPageChangedDone);
        }

    }

    var wireTreeElements = function() {

        tableGrid.$container
            .on('contextmenu', '.pg-table-grid-cell', function (event) {
                event.preventDefault();
                event.stopPropagation();

                if(pgf.kids) {
                    return false;
                }

                var data = getDataForEvent(event);
                var pgel = getPgElForEvent(event);

                if(data && data.template_type === 'interaction') {

                } else if(pgel) {
                    pinegrow.showContextMenuForElement(pgel, event, $(this), tableGrid.$container.closest('.pg-tree-container'));
                } else {
                    pgDynElementMessage('Can\'t edit dynamic elements.');
                }
                return false;
            })
            .on('mouseover', '.pg-table-grid-cell', function (event) {
                if(!tableGrid.inScroll) {
                    var data = getDataForEvent(event);

                    if (!pinegrow.getDragMenu()) {
                        var pgel = _this.getPgelOfDataItem(data);
                        if (pgel) {
                            pinegrow.highlightElement(pgel.get$DOMElement());
                        }
                    }

                }
                event.stopImmediatePropagation();
            })
            .on('mouseout', '.pg-table-grid-cell', function (event) {
                if(!tableGrid.inScroll) {
                    if (!pinegrow.getDragMenu()) {
                        pinegrow.highlightElement(null);
                    }

                }
                event.stopImmediatePropagation();
            })
            .on('click', 'i.collapse-icon', function (event) {

                var data = getDataForEvent(event);
                var pgel = _this.getPgelOfDataItem(data);

                data.collapsed = !(data.collapsed || false);

                if(pgel) {
                    setCollapsedForPgel(pgel, data.collapsed);
                    /*
                    if(data.collapsed) {
                        pgel.setAttr('data-pg-collapsed');
                    } else {
                        pgel.removeAttr('data-pg-collapsed');
                    }*/
                    if(pgel.tagName == 'head') {
                        pgel.getPage().show_head_in_tree = !data.collapsed;
                    }
                }
                // = false;

                if(event.altKey) {
                    tableGrid.forEachVisibleSibling(data, function(ch) {
                        ch.collapsed = data.collapsed;
                        var pgel = _this.getPgel(ch);
                        if(pgel) {
                            setCollapsedForPgel(pgel, ch.collapsed);
                            /*
                            if (ch.collapsed) {
                                pgel.setAttr('data-pg-collapsed');
                            } else {
                                pgel.removeAttr('data-pg-collapsed');
                            }*/
                        }
                        //ch.collapse_on_deselect = false;
                    })
                }

                tableGrid.refresh();
                event.preventDefault();
                event.stopPropagation();
            })
            .on('click', 'i.icon-hide', function (event) {

                var data = getDataForEvent(event);
                showItem(data);
                event.preventDefault();
                event.stopPropagation();
            })
            .on('click', '.pg-table-grid-cell', function (event) {

                var pgel = getPgElForEvent(event);
                var data = getDataForEvent(event);

                $('#crsa-dummy-field').get(0).focus();

                if(pgel) {
                    if(quickInsertPosition) {
                        showQuickInsertIntoTree(data, quickInsertPosition);
                    } else {
                        var sel_done = false;

                        var status = {
                            stop: false,
                            multi: crsaIsCmdPressed(event),
                            source: 'tree',
                            event: event,
                            tree_data: data
                        };

                        pinegrow.dispatchEvent('on_element_clicked', currentPage, pgel, null, status);

                        if(status.stop) {
                            sel_done = true;

                        } else if (event.shiftKey) {
                            var selected = false;
                            var last_pgel = pinegrow.selectedElements.getLastSelectedOnPage(currentPage);
                            if (last_pgel) {
                                var last_data = currentPage.tree.getTreeDataItemForPgId(last_pgel.getId());
                                var inc = last_data.visible_idx > data.visible_idx ? -1 : 1;
                                var col = new PgCollection();
                                for (var i = last_data.visible_idx + inc; i !== data.visible_idx + inc; i = i + inc) {
                                    if(tableGrid.visibleData[i].level == last_data.level) {
                                        var cpgel = _this.getPgelOfDataItem(tableGrid.visibleData[i]);
                                        if (cpgel) {
                                            //pinegrow.selectedElements.add(cpgel);
                                            col.add(cpgel);
                                            selected = true;
                                        }
                                    }
                                }
                                pgel.scrollTo();
                                if(selected) {
                                    pinegrow.selectedElements.selectMany(col);
                                }
                                sel_done = true;

                                if(!selected) {
                                    pinegrow.showQuickMessage('SHIFT + Click can only select elements on the same level.');
                                }
                            }
                        }
                        if (!sel_done) {
                            pinegrow.selectElement(pgel, 'tree', crsaIsCmdPressed(event));
                            pgel.scrollTo();

                            if($(event.target).is('.inline-styles')) {
                                pinegrow.showUIPanel('style', false);
                                //console.log(pinegrow.pgActiveRules.getActiveNodes());
                                pinegrow.activeCSSView.updateRulesDisplayOnElementSelect(true);
                            }
                        }
                    }
                } else {
                    //todo30 - handle dyn el select
                    pgDynElementMessage(quickInsertPosition ? 'Can\'t edit dynamic elements.' : 'Can\'t select a dynamic element.');
                }

                event.stopImmediatePropagation();
                event.preventDefault();

                $('body').trigger('click');
                // return false;
            });

    }


    var getDataForEvent = function(event) {
        var div = $(event.currentTarget).closest('div');
        return tableGrid.visibleData[parseInt(div.attr('data-visible-idx'))];
    }

    var getPgElForEvent = function(event) {
        var data = getDataForEvent(event);
        return _this.getPgelOfDataItem(data);
    }


    var getElementOfTreeNode = function ($li, $html) {
        if (!$html) $html = currentPage.get$Html();
        var el = getElementOfTreeNodeFast($li.get(0), $html.get(0));
        return el ? $(el) : null;
    }

    var getElementOfTreeNodeFast = function (data, html) {
        html = html || currentPage.get$Html().get(0);
        var el;
        var liid = data.treeId;
        if (liid && html) {
            if (html.getAttribute('data-pg-tree-id') == liid) return html;
            el = html.querySelector('[data-pg-tree-id="' + liid + '"]');
        }
        if(data.pgid) {
            if (html.getAttribute('data-pg-id') == data.pgid) return html;
            el = html.querySelector('[data-pg-id="' + data.pgid + '"]');
        }
        return el;
    }

    var getPgelOfTreeNode = function ($li) {
        return getPgelOfTreeNodeFast($li.get(0));
    }

    var getPgelOfTreeNodeFast = function (li) {
        var pgid = li.getAttribute('data-pg-id');
        if (pgid) {
            return getPgNodeByPgId(pgid);
        } else {
            return null;
        }
    }

    var getTreeNodeForPgel = function(pgel) {
        var pgid = pgel instanceof pgParserNode ? pgel.getId() : pgel;
        var node = treeTop.find('[data-pg-id="' + pgid + '"]');
        if(node.length) return node;
        return null;
    }

    /*
    var getOrCreateTreeNodeForPgel = function(pgid) {
        return getTreeNodeForPgel(pgid) || createTreeNodeForPgel(pgid);
    }*/

    this.isOverTree = function(event, drag_menu) {

        if(drag_menu && drag_menu.dragType !== 'html') return null;
        if(view.getWindow() !== event.view) return null;

        if(view.isVisible() && event.pageX >= dimension.left && event.pageX <= dimension.right - (menu.isHidden() ? 0 : (menu.width + (tableGrid.getVScrollBarSize()))) && event.pageY >= dimension.top && event.pageY <= dimension.bottom) {
            return {left: event.pageX - dimension.left, top: event.pageY - dimension.top /* - 38 */};
        }
        return null;
    }

    this.getDropPositionInfo = function(event, menu, tree_pos) {

        var info = tableGrid.dragOver(tree_pos);

        if(info) {
            if(info.data.pgid) {
                info.target_pgel = getPgNodeByPgId(info.data.pgid);
                info.$e = info.target_pgel.get$DOMElement();
                //info.$e = getPgNodeByPgId(info.data.pgid).get$DOMElement();
            } else {
                info.invalid = "Dyn element"; //todo30
                //console.log(info)

                if(info.position == 'bottom') {
                    //DYN
                    //-----

                    //Check if next is not dyn
                    var idx = info.data.visible_idx;
                    var next = (tableGrid.visibleData.length > idx) ? tableGrid.visibleData[idx] : null;

                    if(next && next.level == info.data.level && next.pgid) {
                        info.data = next;
                        info.position = 'top';
                        info.invalid = false;
                    } else if(next && next.level < info.data.level) {
                        //append it to parent
                        var parent = tableGrid.getParentWithLevel(idx, info.data.level - 1);
                        if(parent && parent.pgid) {
                            info.data = parent;
                            info.position = 'inside-bottom';
                            info.invalid = false;
                        }
                    }
                } else if(info.position == 'top') {
                    //Check if prev is not dyn
                    var idx = info.data.visible_idx;
                    var prev = (idx - 1 >= 0) ? tableGrid.visibleData[idx - 1] : null;

                    if(prev && prev.level == info.data.level && prev.pgid) {
                        info.data = prev;
                        info.position = 'bottom';
                        info.invalid = false;
                    } else if(prev && prev.level < info.data.level && prev.pgid) {
                        info.data = prev;
                        info.position = 'inside-top';
                        info.invalid = false;
                    }
                }
                if(info.data.pgid) {
                    info.target_pgel = getPgNodeByPgId(info.data.pgid);
                    info.$e = info.target_pgel.get$DOMElement();
                }
            }
        }
        tableGrid.showInsertionPoint(info);

        return info;
    }


    wireTreeElements();

    var dragHelper = new PgDragHelper($element, '.pg-table-grid-cell', null, true);

    var dragMenu;

    dragHelper.onGetDragImage = function() {
        return 'none';
    }

    dragHelper.onBefore = function($el) {
        if(!pgf.edit_html) return false;

        var data = currentPage.tree.getDataFor$Element($el);
        if(data.no_sort) return false;
        return true;
    }

    dragHelper.on( 'dragStart', function(x, y, event) {
       $.fn.crsapages('showOverlays');

        var data = currentPage.tree.getDataFor$Element(dragHelper.movedElement);
        var pgid = data.pgid;

        var pgel = getPgNodeByPgId(pgid);
        var def = currentPage.getMainType(null, pgel);

        dragMenu = new PgDragMenu(def, true, true, null, pgel);

        var cloned_cont = $('<div class="pg-tree-drag-container with-tree"></div>');

        var num_desc = tableGrid.getNumberOfDescendants(data.idx, tableGrid.data) + 1;

        //if(num_desc > 3) num_desc = 3;

        var first_level = data.level;

        for(var i = 0; i < num_desc; i++) {
            if(i == 4) break;
            var cell = tableGrid.createCellForData(tableGrid.data[data.idx + i]);// new PgTableGridCell(tableGrid.template);
            cell.setData(tableGrid.data[data.idx + i]);
            tableGrid.onSetDataToCell(tableGrid.data[data.idx + i], cell);
            cell.showAt((cell.data.level - first_level) * tableGrid.levelPadding, i * tableGrid.cellHeight, cloned_cont);
        }
        if(num_desc > 4) {
            $('<div class="pg-table-grid-cell"><name>+ ' + (num_desc - 4) + ' more</name></div>').appendTo(cloned_cont).css({
                top: (4 * tableGrid.cellHeight) + 'px',
                left: ((tableGrid.data[data.idx + 4].level + 1 - first_level) * tableGrid.levelPadding) + 'px'
            });
        }
        cloned_cont.css('height', Math.min(num_desc, 5) * tableGrid.cellHeight);

        dragMenu.addElement(cloned_cont);
        dragMenu.movedPgel = pgel;
        dragMenu.pgel = pgel;

        tableGrid.dragStarted(data);

        dragMenu.handleMoveEvent(event, dragMenu);

        menu.hide();

    }, this );

    dragHelper.on( 'drag', function(x, y, event) {

        dragMenu.handleMoveEvent(event, dragMenu);
    }, this );

    dragHelper.on( 'dragStop', function(x, y) {
        $.fn.crsapages('showOverlays', true);

        tableGrid.dragStopped();

        dragMenu.insert();
        dragMenu.destroy();
        dragMenu = null;

        crsaFuncs.showInsertionLine(null);
        pinegrow.highlightElement(null);

        tableGrid.showInsertionPoint(null);


    }, this );


    var updateDimensions = function() {
        //dimension = $element.get(0).getBoundingClientRect();
        dimension = tableGrid.$element.get(0).getBoundingClientRect();
    }

    view.onResize = function() {
        updateDimensions();
        tableGrid.onRepaint();
        tableGrid.show();
        menu.updateLayout();
    }

    /*
    view.on('crsa-page-selected', function (event, selectedPage) {
        view.whenVisible(function () {
            showTreeForPage(selectedPage);
        })
    })
    */

    view.onDestroy = function() {
        pinegrow.unregisterTreeView(_this);

        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        pinegrow.removeEventHandler('on_page_loaded', onPageLoaded);
        pinegrow.removeEventHandler('on_page_refreshed', onPageLoaded);
        pinegrow.removeEventHandler('on_page_selected', onPageSelected);
        pinegrow.removeEventHandler('on_drag_drop_operation_done', onDragDropDone);
        pinegrow.removeEventHandler('on_page_frameworks_changed', onPageFrameworksChanged);
        pinegrow.removeEventHandler('on_element_isolated', onElementIsolated);
        manager.destroy();
        $element.off('mousemove', onMouseMoved);
        $element.off('mouseleave', onMouseOut);
        $element.off('mouseenter', onMouseIn);
        //$element.off('mouseenter mouseleave', noop);
    }

    var onPageChanged = function(page, pgel, info) {
    }

    var showTreeForPage = function(page) {
        currentPage = page;

        if(page) {

            classes_filter = page.tree.options.classes_filter;
            num_classes = page.tree.options.num_classes;

            var has_isolated = pinegrow.pageHasIsolatedElement(page);

            if(has_isolated) {
                modeButton.hide();

                if(pgf.focus_in_tree) {
                    var focus_pgel = pinegrow.getIsolatedElement();
                    setFocusMessage(focus_pgel.getName({short: true}));
                }
            } else {
                if(pgf.focus_in_tree) {
                    setFocusMessage(null)
                }
                modeButton.show();
            }

            page.tree.setShowInteractions(show_interactions);
            page.tree.setFilterOptions(filter_options);
            page.tree.setTreeView(_this);
            page.tree.setTableGrid(tableGrid);

            if(dom_mode !== page.tree.getDomMode()) {
                page.tree.setDomMode(dom_mode);
                page.tree.setAutoCollapse(auto_collapse_level);
                page.tree.create();

            } else {
                var pgel = pinegrow.selectedElements.getLastSelectedOnPage(page);
                if (!pgel) {
                    currentMenuData = null;
                    menu.hide();
                } else {
                    currentMenuData = page.tree.getTreeDataItemForPgId(pgel.getId());
                    menu.show();
                    updateMenuPosition();
                }
            }
        } else {
            currentMenuData = null;
            tableGrid.setData([]);
            menu.hide();
        }

        _this.setPage(page);

        menu.updateLayout();
    }

    var onPageLoaded = function(page) {
        page.tree.setShowInteractions(show_interactions);
        page.tree.setDomMode(dom_mode);
        page.tree.setAutoCollapse(auto_collapse_level);
        page.tree.create();
        if(page === currentPage) {
            showTreeForPage(page);
        }
    }

    var onPageFrameworksChanged = function(page) {
        if(page) {
            page.tree.create();
        }
        if((page && page === currentPage) || (page === null && currentPage)) {
            if(currentPage && currentPage.tree) {
                showTreeForPage(currentPage);
            }
        }
    }

    var onElementIsolated = function(page, pgel) {
        page = page || currentPage;
        if(page && page.tree) {
            page.tree.create();
            showTreeForPage(page);
        }
    }

    var onPageSelected = function(page) {
        view.whenVisible(function () {
            updateDimensions();
            showTreeForPage(page);
        })
    }

    var onDragDropDone = function(page, drag_menu) {
        if(tableGrid) {
            tableGrid.dragStopped();
        }
    }

    var noop = function() {}

    var currentMenuData = null;

    var hoveredCell = null;

    var mouseSpeed = new PgMouseSpeed();

    var quickInsertPosition = null;
    var quickInsertPositionTimer = null;

    var onMouseMoved = function(event) {
        var pos = _this.isOverTree(event);

        mouseSpeed.track(event);

        if(currentMenuData && menu.isHidden()) menu.show();

        quickInsertPosition = null;
        tableGrid.quickInsertLine.show(null);

        if(quickInsertPositionTimer) {
            clearTimeout(quickInsertPositionTimer);
            quickInsertPositionTimer = null;
        }

        if(pos && currentPage) {
            var grid = tableGrid;

            var so = grid.getScrollOffset();
            pos.top += so.top;
            pos.left += so.left;

            var idx = grid.getVisibleDataIndexForY(pos.top);
            if(idx == -2) idx = grid.getVisibleDataIndexForY(pos.top - grid.cellHeight);
            var data = grid.visibleData[idx];
            var level = Math.max(0, Math.floor(pos.left / grid.levelPadding) - 1);

            var dragPos = tableGrid.getDragPos();
           // console.log(dragPos);

            if(dragMenu && dragPos) {
                //grid.highlightLevel(idx, dragPos.level);
            } else {
                grid.highlightLevel(idx, level);

                //show quick insert line
                if(pgf.edit_html && !pgf.kids) {
                    if (data && !pinegrow.getDragMenu() && !data.no_insert) {
                        var edge = 4;
                        var delay = 450;
                        var insertPosition = null;
                        var level = data.level;

                        var y = grid.getYForVisibleDataIndex(idx);

                        //console.log(`y = ${y}, pos.top = ${pos.top}, edge = ${edge}`)
                        if (pos.top <= y + edge) {
                            insertPosition = 'insertBefore';
                        } else if (pos.top >= y + grid.cellHeight - edge) {
                            var next = grid.visibleData.length > (data.visible_idx + 1) ? grid.visibleData[data.visible_idx + 1] : null;

                            if (!next || next.level <= data.level) {
                                insertPosition = 'insertAfter';
                            } else {
                                insertPosition = 'prepend';
                                level = data.level + 1;
                            }
                        }
                        //console.log(insertPosition);

                        if (insertPosition) {
                            if (activeQuickInsertData && ((activeQuickInsertData == data && activeQuickInsertPosition == insertPosition) || (activeQuickInsertData.visible_idx == data.visible_idx + 1 && activeQuickInsertPosition == 'insertBefore' && (insertPosition == 'insertAfter' || insertPosition == 'prepend')) || (activeQuickInsertData.visible_idx == data.visible_idx - 1) && (activeQuickInsertPosition == 'insertAfter' || activeQuickInsertPosition == 'prepend') && insertPosition == 'insertBefore')) {
                                //skip
                            } else {
                                quickInsertPositionTimer = setTimeout(function () {
                                    quickInsertPosition = insertPosition;
                                    grid.quickInsertLine.show(data, level, insertPosition == 'insertBefore' ? 'top' : 'bottom');
                                }, delay)
                            }
                        }
                    }
                }
            }

            if(!dragMenu) {
                var data = grid.visibleData[idx];
                var top = grid.getYForVisibleDataIndex(idx) - so.top;

                var cell = grid.getCellForData(data);

                if(hoveredCell != cell) {

                    var movingToMenu = menu.isOverMenu(pos);

                    //{left: pos.left, top: pos.top}
                    /*
                    var distX = dimension.width - 32 - (pos.left - so.left);
                    if(mouseSpeed.speedX > 0) {
                        var timeToX = distX / mouseSpeed.speedX;
                        var yAtTime = pos.top - so.top + mouseSpeed.speedY * timeToX;
                        if(menu.isYAlignedWithMenu(yAtTime, 20)) {
                            movingToMenu = true;
                        }
                        //console.log(distX, timeToX, yAtTime);
                    }
*/
/*
                    if(data) {
                        if (!movingToMenu) {

                            currentMenuData = data;

                            menu.showFor(top + so.top, currentMenuData.selected || false, currentMenuData, so.top);
                            //menu.showInCell(cell);
                            setClassesForMenu(currentMenuData);


                            if (hoveredCell) {
                                hoveredCell.$element.removeClass('hovered');
                            }
                            if (cell) cell.$element.addClass('hovered');
                            hoveredCell = cell;
                        }
                    } else {
                        menu.hide();
                    }
                    */
                }

            }
        }
    }

    var setClassesForMenu = function(data) {
        var all = 'for-selected for-hidden for-collapsed for-search';

        var c = '';

        if(data.selected) c += ' for-selected';
        if(data.hidden) c += ' for-hidden';
        if(data.collapsed) c += ' for-collapsed';
        if((data.search || 0) === 1) c += ' for-search';

        menu.$menu.removeClass(all).addClass(c);
    }

    var onMouseOut = function(event) {
        tableGrid.highlightLevel(-1);
        tableGrid.quickInsertLine.hide();
        if(quickInsertPositionTimer) {
            clearTimeout(quickInsertPositionTimer);
            quickInsertPositionTimer = null;
        }
        menu.hide();
        if (hoveredCell) {
            hoveredCell.$element.removeClass('hovered');
        }
        hoveredCell = null;
    }

    var onMouseIn = function(event) {
        if(currentMenuData) menu.show();
    }

    $element.on('mousemove', onMouseMoved);
    $element.on('mouseleave', onMouseOut);
    $element.on('mouseenter', onMouseIn);

    pinegrow.addEventHandler('on_page_changed', onPageChanged);
    pinegrow.addEventHandler('on_page_loaded', onPageLoaded);
    pinegrow.addEventHandler('on_page_refreshed', onPageLoaded);
    pinegrow.addEventHandler('on_page_selected', onPageSelected);
    pinegrow.addEventHandler('on_drag_drop_operation_done', onDragDropDone);
    pinegrow.addEventHandler('on_element_isolated', onElementIsolated);

    pinegrow.addEventHandler('on_page_frameworks_changed', onPageFrameworksChanged);

    var updateMenuPosition = function() {
        if(currentMenuData) {
            var so = tableGrid.getScrollOffset();
            var top = tableGrid.getYForVisibleDataIndex(currentMenuData.visible_idx) - so.top;

            menu.showFor(top, top + tableGrid.cellHeight, currentMenuData.selected || false, currentMenuData);
        }
    }

    this.onSelected = function(data) {

        currentMenuData = data;

        var pgel = this.getPgel(data);

        if(pgel) {
            var $item = menu.find('hide');
            var i = $item.data('icon') || $item.get(0).querySelector('i');
            $item.data('icon', i);

            var add, remove;
            if(pgel.isHidden()) {
                add = 'icon-see';
                remove = 'icon-hide';
            } else {
                add = 'icon-hide';
                remove = 'icon-see';
            }
            if(i.classList.contains(remove) || !i.classList.contains(add)) {
                fastdom.mutate(function () {
                    i.classList.remove(remove);
                    i.classList.add(add);
                })
            }

            //$item.find('i').removeClass('icon-hide icon-see').addClass(pgel.isHidden() ? 'icon-see' : 'icon-hide');
        }
        updateMenuPosition();
    }

    this.onDeselected = function(data) {
        //if(currentMenuData === data) setClassesForMenu(data);
        if(!pinegrow.selectedElements.hasSelection()) {
            currentMenuData = null;
            menu.hide();
        }
    }
}






var PgPageTree = function(page) {

    var _this = this;

    var max_tree_level = 0;
    var li_id_count = 0;
    var currentTreeRoot;
    var show_interactions;

    var dom_mode = false;
    var auto_collapse_level = 0;

    var filter_options = {
        styles: false,
        images: false
    }

    var has_filter_options = false;

    var treeData = [];
    var treeData_PgIdMap = {};
    var helper;

    this.getDomMode = function() {
        return dom_mode;
    }

    this.setDomMode = function(v) {
        dom_mode = v;
    }

    this.setAutoCollapse = function(v) {
        auto_collapse_level = v;
    }

    this.setShowInteractions = function(v) {
        show_interactions = v;
    }

    this.setFilterOptions = function(opt) {
        var o = has_filter_options;
        has_filter_options = false;
        $.each(opt, function(key, val) {
            if(val) {
                has_filter_options = true;
                return true;
            }
        })
        filter_options = opt;
        research_on_show = o !== has_filter_options;
    }

    var treeView;

    this.setTreeView = function(tv) {
        treeView = tv;
    }

    var tableGrid = null;

    this.setTableGrid = function(tg) {
        tableGrid = tg;
        this.$element = tableGrid.$element;
        tg.setData(treeData);

        if(research_on_show) {
            this.research();
        }
    }

    var treeTop = $('<div/>').addClass('crsa-tree-branch pg-tree-ul');

    this.$element = null;

    this.highlightOnMouseOver = true;

    this.destroy = function() {
        nextFrame.cancelAll();
        tableGrid = null;
        treeView = null;
        if(helper) helper.destroy();
        helper = null;
        treeData = null;
        treeData_PgIdMap = null;
        _this = null;
    }

    this.getDataFor$Element = function($el) {
        return tableGrid.getDataFor$Element($el);
    }

    this.updatePgIdMap = function() {
        var p = new CrsaProfile();
        treeData_PgIdMap = {};
        treeData.forEach(function(data) {
            if(data.pgid) {
                treeData_PgIdMap[data.pgid] = data;
            }
        })
        p.show('updatePgIdMap');
    }

    var removeItemPgidMap = function(data) {
        if(data && data.pgid) delete treeData_PgIdMap[data.pgid];
    }

    var removeItemAndChildrenFromPgidMap = function(data) {
        helper.withDataAndChildren(data, removeItemPgidMap);
    }

    var getTreeDataItemForPgId = function(pgid) {
        return treeData_PgIdMap[pgid] || null;
    }

    this.getTreeDataItemForPgId = function(pgid) {
        return getTreeDataItemForPgId(pgid);
    }

    var getDataItemsForPgIdAndDescendants = function(pgid) {
        var data = getTreeDataItemForPgId(pgid);
        if(data) {
            var list = [data];
            var num = data.idx + helper.getNumberOfDescendants(data.idx, treeData);
            for(var i = data.idx + 1; i <= num; i++) {
                list.push(treeData[i]);
            }
            return list;
        }
        return null;
    }

    var createDataItemsForPgEl = function(pgel, level) {
        var list = [];
        if(dom_mode) {
            var $el = pgel.get$DOMElement();
            if ($el) {
                createDataItem(list, $el ? $el.get(0) : null, false, level);
            }
        } else {
            createDataItemForPgEl(list, pgel, false, level);
        }
        return list;
    }

    var refreshGrid = function() {
        nextFrame.do('refreshGrid', function() {
            _this.research();
            if(tableGrid) {
                tableGrid.refresh();

                if(treeView && !treeView.view.isVisible()) {
                    treeView.view.whenVisible(function() {
                        tableGrid.show();
                    })
                }
            }
        })
    }

    var previous_selector_was_invalid = false;

    const SEARCH_IN = 1;
    const SEARCH_OUT = 3;
    const SEARCH_HAS_CHILD = 2;

    var currentFilter = null;

    this.research = function(force) {
        var obj = {
            selected_ids: {},
            changed: false
        }
        page.callFrameworkHandler('on_custom_tree_filter', obj);

        if(currentFilter || obj.changed || has_filter_options || force) this.search(currentFilter, true, obj);
    }

    var research_on_show = false;

    this.search = function(filter, research, custom_obj) {
        if(!treeView) {
            research_on_show = true;
            return;
        }

        research_on_show = false;

        if(filter === null) filter = '';
        var re = (filter.length > 0) ? new RegExp(escapeRegExp(filter), 'i') : null;

        var force_sel = false;
        var new_filter = filter;

        if (filter.startsWith('$')) {
            filter = filter.substr(1);
            force_sel = true;
        }

        //try searching for selector
        var selected_ids = {};

        currentFilter = filter;

        ////
        var obj;

        if(custom_obj) {
            obj = custom_obj;
        } else {
            var obj = {
                selected_ids: {},
                changed: false
            }
            page.callFrameworkHandler('on_custom_tree_filter', obj);
        }

        var custom_ids = obj.selected_ids;
        var custom_filter = obj.changed;

        if(!filter && !custom_filter && !has_filter_options) {
            treeData.forEach(function(data) {
                data.search = 0;
                data.force_search = false;
                if(data.peek_in) {
                    data.collapsed = true;
                    data.peek_in = false;
                }
            })
            tableGrid.refresh();
            return;
        }

        //find selector
        if (filter) {
            try {
                page.get$Html().find(filter).each(function (i, e) {
                    var pgid = e.getAttribute('data-pg-id');
                    if (pgid) {
                        selected_ids[pgid] = true;
                    }
                })
                previous_selector_was_invalid = false;
            } catch (err) {
                previous_selector_was_invalid = true;
            }
        }
/*
        if (custom_filter) {
            if (filter) {
                var matched_ids = {};
                $.each(custom_ids, function (key, val) {
                    if (key in selected_ids) {
                        matched_ids[key] = true;
                    }
                })
                selected_ids = matched_ids;
            } else {
                selected_ids = custom_ids;
            }
        }
        */

if(!helper) return;

        var html = page.get$Html().get(0);

        treeData.forEach(function(data) {
            data.search = 0;
            if(data.peek_in) {
                data.collapsed = true;
                data.peek_in = false;
            }
        })

        helper.forEachWithParents(function(data, parents) {
            treeView.populateDataItem(data, html);

            var inc = false;

            if(!force_sel && (!re || data.search_string.match(re))) {
                inc = true;
            } else if(selected_ids[data.pgid]) {
                inc = true;
            }
            if (inc && custom_filter) {
                inc = custom_ids[data.pgid] || false;
            }
            if(has_filter_options) {
                if(data.pgel) {
                    if(filter_options.styles) {
                        if (data.pgel.getAttribute('style')) {
                            inc = inc && true;
                        } else {
                            inc = false;
                        }
                    } else if(filter_options.images) {
                        inc = inc && data.pgel.tagName === 'img';
                    } else if(filter_options.links) {
                        inc = inc && data.pgel.tagName === 'a';
                    } else if(filter_options.components) {
                        inc = inc && (data.pgel.findAttributesStartingWith('data-pgc').length > 0);
                    } else if(filter_options.wp) {
                        inc = inc && ((data.pgel.findAttributesStartingWith('wp-').length > 0 || data.pgel.findAttributesStartingWith('cms-').length > 0));

                    } else if(filter_options.wc) {
                        inc = inc && ((data.pgel.findAttributesStartingWith('wc-').length > 0));

                    } else if(filter_options.ia) {
                        inc = inc && (data.pgel.findAttributesStartingWith('data-pg-ia').length > 0);

                    } else if(filter_options.texts) {
                        //console.log(data.pgel.text(null, true));
                        inc = inc && data.pgel.text(null, true).trim().length > 0;
                    }
                }
            }
            if(inc) {
                //selected_ids[data.pgid] = true;
                data.search = SEARCH_IN;

                for(var i = parents.length - 1; i >= 0; i--) {
                    if(parents[i].search === SEARCH_OUT) {
                        parents[i].search = SEARCH_HAS_CHILD;
                        if(parents[i].collapsed) {
                            parents[i].peek_in = true;
                            //data.peeked = true;
                        }
                        parents[i].collapsed = false;
                    } else {
                        break;
                    }
                    //parents[i].collapsed = false;
                }
            } else {
                if(false && data.force_search && research) {
                    for(var i = parents.length - 1; i >= 0; i--) {
                        if(parents[i].search === SEARCH_OUT) {
                            parents[i].search = SEARCH_HAS_CHILD;
                        } else {
                            break;
                        }
                        //parents[i].collapsed = false;
                    }
                } else {
                    data.search = SEARCH_OUT;
                }
            }
        })

        tableGrid.refresh();
    }

    var unpopulateDataItem = function(data) {
        if(data) {
            data.populated = false;
            data.def = null;
        }
    }

    var updateBranchForPgel = function(pgel, data) {
        if(data === undefined) {
            data = getTreeDataItemForPgId(pgel.getId())
        }
        var list = createDataItemsForPgEl(pgel);
        helper.replaceWith(data, list);
    }

    var updateSelectedItems = function() {
        pinegrow.selectedElements.forEachElement(function(pgel) {
            var data = getTreeDataItemForPgId(pgel.getId());
            if(data) data.selected = true;
        })
    }

    var forceDisplayInSearch = function(list_or_data, val) {
        if(Array.isArray(list_or_data)) {
            list_or_data.forEach(function(data) {
                data.force_search = val;
            })
        } else if(list_or_data) {
            list_or_data.force_search = val;
        }
    }

    this.scrollToSelected = null;

    this.handleEvent = function(event) {

        var pgel = event.getPgel();
        var data = getTreeDataItemForPgId(event.pgId);
        var $oli;

        try {

            switch (event.operation) {
                case 'selected':
                    if(data) {
                        this.scrollToSelected = null;
                        data.selected = true;

                        if (helper.peek(data)) {
                            helper.reindexData(0, true /* update hidden status */);
                            refreshGrid(); //was uncollapsed
                        } else {
                            if (tableGrid) tableGrid.updateItem(data);
                        }
                        if (tableGrid) {
                            tableGrid.scrollToDelayed(data);
                        }
                        this.scrollToSelected = data;

                        if (treeView) treeView.onSelected(data);
                    }
                    break;

                case 'deselected':
                    if(data) {
                        data.selected = false;

                        if(this.scrollToSelected == data) {
                            this.scrollToSelected = null;
                        }

                        if(helper.unpeek(data)) {
                            helper.reindexData(0, true /* update hidden status */);
                            refreshGrid();
                        } else {
                            if(tableGrid) tableGrid.updateItem(data);
                        }
                        //tableGrid.scrollTo(data);
                        if(treeView) treeView.onDeselected(data);

                    }
                    break;
                case 'show':
                    data.hide_in_pg = false;
                    data.hide_in_pg_child = false;
                    helper.withChildren(data, function(data) {
                        if(!data.hide_in_pg) {
                            data.hide_in_pg_child = false;
                        } else return false;
                        //data.collapsed = !(data.uncollapse_on_show || false);
                    })
                    helper.reindexData();
                    refreshGrid();
                    break;

                case 'hide':
                    data.hide_in_pg = true;
                    helper.withChildren(data, function(data) {
                        if(!data.hide_in_pg) {
                            data.hide_in_pg_child = true;
                        } else return false;
                        //data.uncollapse_on_show = !data.collapsed;
                        //data.collapsed = true;
                    })
                    helper.reindexData();
                    refreshGrid();
                    break;

                case 'remove':
                case 'detach':
                    if(data) {
                        helper.removeItem(data, removeItemPgidMap);
                        //this.updatePgIdMap();
                        helper.reindexData();
                        refreshGrid();
                    }
                    break;
                case 'insertAfter':
                case 'insertBefore':
                    var dest_data = getTreeDataItemForPgId(event.objectId);
                    var list = getDataItemsForPgIdAndDescendants(event.pgId) || createDataItemsForPgEl(pgel);
                    forceDisplayInSearch(list, true);
                    helper[event.operation](dest_data, list);
                    helper.reindexData();
                    refreshGrid();
                    break;
                case 'insertAtIndex':
                    var dest_data = getTreeDataItemForPgId(event.objectId);
                    var list = getDataItemsForPgIdAndDescendants(event.pgId) || createDataItemsForPgEl(pgel);
                    var destPgel = getPgNodeByPgId(event.objectId);
                    var idx = destPgel.getChildPos(pgel, true);
                    helper.insertAtIndex(dest_data, list, idx);
                    forceDisplayInSearch(list, true);
                    helper.reindexData();
                    refreshGrid();
                    break;

                case 'appendPrepend':
                    var dest_data = getTreeDataItemForPgId(event.objectId);
                    if(event.pgId) {
                        var list = getDataItemsForPgIdAndDescendants(event.pgId) || createDataItemsForPgEl(pgel);
                        event.data.prepend ? helper.prepend(dest_data, list) : helper.append(dest_data, list);
                        forceDisplayInSearch(list, true);
                        helper.reindexData();
                        refreshGrid();
                    } else {
                        unpopulateDataItem(dest_data);
                        if(tableGrid) tableGrid.updateItem(dest_data);
                    }
                    break;
                case 'removeAllChildren':
                    helper.withChildren(data, removeItemPgidMap);
                    helper.removeDescendants(data);
                    helper.reindexData();
                    //this.updatePgIdMap();
                    refreshGrid();
                    break;
                case 'detag':
                    removeItemPgidMap(data);
                    helper.delevel(data);
                    forceDisplayInSearch(data, true);
                    helper.reindexData();
                    //this.updatePgIdMap();
                    refreshGrid();
                    break;
                case 'replaceContentWithElement':
                    helper.withChildren(data, removeItemPgidMap);
                    var list = getDataItemsForPgIdAndDescendants(event.objectId) || createDataItemsForPgEl(getPgNodeByPgId(event.objectId));
                    helper.replaceContentWith(data, list);
                    forceDisplayInSearch(list, true);
                    helper.reindexData();
                    //this.updatePgIdMap();
                    refreshGrid();
                    break;
                case 'replaceContentWithContentOf':
                    helper.withChildren(data, removeItemPgidMap);
                    var list = getDataItemsForPgIdAndDescendants(event.objectId) || createDataItemsForPgEl(getPgNodeByPgId(event.objectId));
                    if(list.length > 1) {
                        list.shift();
                        helper.replaceContentWith(data, list);
                        forceDisplayInSearch(list, true);
                        helper.reindexData();
                        //this.updatePgIdMap();
                        refreshGrid();
                    }
                    break;
                case 'replaceWith':
                    removeItemAndChildrenFromPgidMap(data);
                    var list = getDataItemsForPgIdAndDescendants(event.objectId) || createDataItemsForPgEl(getPgNodeByPgId(event.objectId));
                    helper.replaceWith(data, list);
                    list[0].selected = data.selected || false;
                    forceDisplayInSearch(list, true);
                    helper.reindexData();
                    //this.updatePgIdMap();
                    refreshGrid();
                    break;
                case 'removeAttr':
                case 'setAttr':
                case 'removeClass':
                case 'addClass':
                case 'replaceTag':
                //case 'removeAttrIfStartsWith':
                    var update_all = false;

                    //console.log(event)
                    if(event.data.attr && event.data.attr.startsWith('data-pg-ia')) {
                        if(event.data.info && event.data.info.pgia_animation) {
                            //Animation definition changed, we don't need to update the deep tree for that
                            break;
                        } else {
                            update_all = true;
                        }
                    }
                    if(!update_all) {
                        unpopulateDataItem(data);
                        forceDisplayInSearch(data, true);
                        //updateSelectedItems();
                        if (tableGrid) tableGrid.updateItem(data);
                        break;
                    }
                case 'update':
                case 'updateInner':
                case 'html':
                    if(!data) {
                        //perhaps pgid changed, new id not found. try with old id
                        if(event.data && event.data.domPgId) {
                            data = getTreeDataItemForPgId(event.data.domPgId);
                        }
                    }
                    if(data) {
                        removeItemAndChildrenFromPgidMap(data);
                        updateBranchForPgel(pgel, data);
                        forceDisplayInSearch(data, true);
                        helper.reindexData();
                        //this.updatePgIdMap();
                        updateSelectedItems();
                        refreshGrid();
                    }
                    break;

            }
        }
       catch(err) {
            console.log('treeError', err);
       }
    }

    var ignore_scripts = false;
    var force_parent = null;
    var classes_filter = null;

    this.options = {
        num_classes: -1,
        classes_filter: null
    }

    this.create = function() {
        //generate tree snapshot of the page

        treeData = [];
        treeData_PgIdMap = this.treeData_PgIdMap = {};

        this.options = {
            num_classes: -1,
            classes_filter: null
        }

        page.callFrameworkHandler('on_get_tree_options', this.options);


        var $ul = treeTop;
        var only_kids = false;

        var current_max_tree_level = max_tree_level;

        max_tree_level = 0;

        var test = false;

        var treeRoot = currentTreeRoot = page.get$Html();
        var isolated_element = pinegrow.getIsolatedElement(page);

        var dom_root = treeRoot.get(0);
        var pgel_root = page.sourceNode.findOne('html') || page.sourceNode.children[0];

        ignore_scripts = false;
        force_parent = null;

        if(isolated_element) {
            pgel_root = isolated_element;
            dom_root = isolated_element.get$DOMElement().get(0);
            var opts = pinegrow.getIsolatedElementOptions();
            ignore_scripts = opts.tree_hide_scripts;
            force_parent = opts.drag_drop_only_top_level ? isolated_element : null;
            dom_mode = false;
            //treeView.setFilterOption(null, true);
        }

        var profile = new CrsaProfile(true);

        var start_level = 0;

        if(dom_mode) {
            var html = createDataItem(treeData, dom_root, only_kids);
        } else {
            var html = pgel_root && createDataItemForPgEl(treeData, pgel_root, only_kids);
        }

        profile.show('create tree data items');

        helper = new PgTreeDataHelper(treeData);

        helper.reindexData();

        this.research();
    }

    //helper methods



    var createDataItem = function (treeData, e, only_kids, level) {
        if (!level) level = 0;

        var cls = 'sort';

        if ((e.tagName == 'BODY' && $(e).children().length > 0) || e.tagName == 'HEAD' || e.tagName == 'HTML') {
            cls = 'no-sort';
        }

        var tags = '';//def && def.tags ? ' tag-' + def.tags : '';

        var eye = 'icon-see';
        var i_class = '';
        var name = e.tagName;
        var dyn_element = null;

        if (name == 'STYLE') {
            if (e.id == 'crsa-inline-styles') return '';
        }

        if (isElementHidden(e)) {
            cls += ' crsa-tree-node-hidden'
            eye = 'icon-hide';
        }

        var lid = ++li_id_count;
        e.setAttribute('data-pg-tree-id', lid);

        var attr = 'data-pg-tree-id="' + lid + '"';

        var pgid = e.getAttribute('data-pg-id');
        if (!pgid) {
            cls += ' dyn';
            dyn_element = e;
        } else {
            attr += ' data-pg-id="' + pgid + '" ';
        }

        if (name == 'HTML') {
            name = page.name;
            tags += ' tag-block';
            i_class = 'icon-Pin10';
            eye += ' hide';
            attr += ' aadata-li-populated="true"';
            cls += ' doc-root';
        }

        var fata = {level: level, tags: tags, classes: cls, name: name, pgid: pgid, treeId: lid, shown: true, collapsed: isCollapsed(e, level), dyn_element: dyn_element};
        treeData.push(fata);

        if(pgid) {
            treeData_PgIdMap[pgid] = fata;
        }

        if(pgid) {
            addInteractions(treeData, e, pgid, level);
        }

        var children = e.childNodes;
        if (children.length > 0 && e.tagName !== 'aaaSVG') {
            for (var i = 0; i < children.length; i++) {
                var node = children.item(i);
                if (node.nodeType == 1 /* && node.tagName != 'BR' */) {

                    //Dont show page ui elements in tree
                    if(pgIsNodePageUI(node)) continue;

                    if(ignore_scripts && node.tagName === 'SCRIPT') continue;

                    createDataItem(treeData, node, false, level + 1);
                } else if(node.nodeType == 3) {
                    //fata = {level: level + 1, tags: null, classes: null, name: i + ' [' + node.nodeValue + ']', pgid: null, treeId: null, shown: true};
                    //treeData.push(fata);
                }
            }
        }
    }

    var addInteractions = function(treeData, pgel, pgid, level) {
        if(!show_interactions) return;

        var json = pgel.getAttribute('data-pg-ia');
        if(json) {
            try {
                var ad = JSON.parse(json);
                if(ad.l) {
                    /*
                     treeData.push({
                     level: level + 1,
                     name: 'Interactions',
                     pgid: pgid,
                     pgel: pgel,
                     populated: true,
                     show: true,
                     collapsed: false,
                     no_sort: true,
                     interaction: ic++
                     })*/

                    var ic = 0;
                    ad.l.forEach(function(a) {
                        var name = (a.name || 'Interaction ' + (ic + 1));
                        if(a.on) {
                            switch(a.on) {
                                case 'mobile':
                                    name += ' [SM]';
                                    break;
                                case 'desktop':
                                    name += ' [LG]';
                                    break;
                                case 'off':
                                    name += ' [OFF]';
                                    break;
                            }
                        }
                        if(a.nopg) {
                            name += ' [NOPG]';
                        }
                        treeData.push({
                            level: level + 1,
                            name: name,
                            pgid: pgid,
                            //pgel: pgel,
                            populated: true,
                            shown: true,
                            collapsed: false,
                            no_sort: true,
                            no_insert: true,
                            icon: 'icon-113',
                            template_type: 'interaction',
                            search_string: name,
                            pgia: {
                                type: 'a',
                                action: 'i',
                                id: pgid + ':animation_' + ic,
                                idx_zero: ic++,
                                desc: (a.a && a.a.desc) ? a.a.desc : null
                            }
                        })
                    })
                }
            } catch(err) {}
        }

        var json = pgel.getAttribute('data-pg-ia-scene');
        if(json) {
            try {
                var ad = JSON.parse(json);
                if(ad.l) {
                    var ic = 0;
                    ad.l.forEach(function(a) {
                        var name = a.name || ('Scroll scene' + (ad.length > 1 ? (' ' + (ic + 1)) : ''));

                        if(a.on) {
                            switch(a.on) {
                                case 'mobile':
                                    name += ' [SM]';
                                    break;
                                case 'desktop':
                                    name += ' [LG]';
                                    break;
                                case 'off':
                                    name += ' [OFF]';
                                    break;
                            }
                        }
                        if(a.nopg) {
                            name += ' [NOPG]';
                        }

                        treeData.push({
                            level: level + 1,
                            name: name,
                            pgid: pgid,
                            //pgel: pgel,
                            populated: true,
                            shown: true,
                            collapsed: false,
                            no_sort: true,
                            no_insert: true,
                            template_type: 'scrollscene',
                            search_string: name,
                            pgia: {
                                type: 'a',
                                action: 's',
                                id: pgid + ':scene_item_' + ic,
                                idx_zero: ic++,
                                desc: (a.a && a.a.desc) ? a.a.desc : null
                            }
                        })
                    })
                }
            } catch(err) {}
        }


    }

    var createDataItemForPgEl = function (treeData, pgel, only_kids, level) {
        if (!level) level = 0;

        var cls = 'sort';
        var no_sort = false;

        var def = page.getTypeThatFiltersTree(pgel);

        if ((pgel.tagName == 'body' && pgel.children.length > 0) || pgel.tagName == 'head' || pgel.tagName == 'html') {
            cls = 'no-sort';
            no_sort = true;
        }

        var force_collapsed = false;
        var focused = false;

        if(force_parent) {
            if(pgel === force_parent) {
                cls = 'no-sort';
                no_sort = true;
                focused = true;
            } else if(/* pgf.focus_select_only_top_level && */ pgel.isDescendantOf(force_parent)) {
                force_collapsed = true;
            }
        }

        var tags = '';
        var attr = '';

        var eye = 'icon-seen';
        var i_class = '';
        var name = pgel.tagName;

        if (isPgElHidden(pgel)) {
            cls += ' crsa-tree-node-hidden'
            eye = 'icon-hide';
        }

        var pgid = pgel.getId();
        if (!pgid) {
            cls += ' dyn';
        } else {
            attr += ' data-pg-id="' + pgid + '" ';
        }

        if (name == 'html') {
            name = page.name;
            tags += ' tag-block';
            i_class = 'icon-Pin10';
            eye += ' hide';
            attr += ' aadata-li-populated="true"';
            cls += ' doc-root';
        }



        var fata = {level: level, tags: tags, classes: cls, name: name, pgid: pgid, pgel: pgel, treeId: 0, shown: true, collapsed: isPgElCollapsed(pgel, level) || force_collapsed, no_sort: no_sort, focused: focused};
        treeData.push(fata);

        if(pgid) {
            treeData_PgIdMap[pgid] = fata;
        }

        //if(force_parent && pgel !== force_parent) return;


        ///extension begin
        //List Interactions
        if(true) {
            addInteractions(treeData, pgel, pgid, level);
        }
        //ext end

        var children = pgel.children;

        if(def && def.on_filter_tree_children) {
            children = def.on_filter_tree_children(pgel);
        }
        if (children.length > 0 && pgel.tagName !== 'aaasvg') {
            for (var i = 0; i < children.length; i++) {
                var node = children[i];
                if (node.isElementOrScript) {

                    //Dont show page ui elements in tree
                    //if(pgIsNodePageUI(node)) continue;
                    if(ignore_scripts && node.tagName === 'script') continue;

                    createDataItemForPgEl(treeData, node, false, level + 1);
                }
            }
        }
    }


    var isElementHidden = function(e) {
        return e.getAttribute('data-pg-hidden') == '';
    }

    var isPgElHidden = function(pgel) {
        return pgel.hasAttribute('data-pg-hidden');
    }

    var isCollapsed = function (e, level) {
        if(pgf.kids) return false;
        if(auto_collapse_level > 0 && level >= auto_collapse_level) return true;
        if (e.tagName == 'HEAD') return !page.show_head_in_tree;
        if(e.hasAttribute('data-pg-collapsed')) {
            return e.getAttribute('data-pg-collapsed') !== 'false';
        }
        return e.tagName === 'SVG' || e.tagName === 'svg'; //collapsed by default
    }

    var isPgElCollapsed = function (pgel, level) {
        if(pgf.kids) return false;
        if(auto_collapse_level > 0 && level >= auto_collapse_level) return true;
        return pgel.isCollapsed();
    }

    var selectElement = function(pgid) {
        var $li = getTreeNodeForPgel(pgid);
        if($li) {
            $li.addClass('crsa-tree-node-selected');
            scrollToTreeNode($li);
        }
    }

    var deselectElement = function(pgid) {
        var $li = getTreeNodeForPgel(pgid);
        if($li) {
            $li.removeClass('crsa-tree-node-selected');
        }
    }

    var removeElement = function(pgid) {
        var $li = getTreeNodeForPgel(pgid);
        if($li) {
            $li.remove();
        }
    }

    var scrollToTreeNode = function($li) {
        var test = $li.get(0).getBoundingClientRect();
        var test2 = treeTop.get(0).getBoundingClientRect();

        var lip = {left: test.left - test2.left, top: test.top - test2.top};
        //var lip = $li.position();
        var $tree = treeTop.closest('.tree-container');

        var tree_h = $tree.height();
        var li_h = 20;//$li.height();
        var tree_st = $tree.scrollTop();

        if (lip.top >= 0 && lip.top < tree_h - li_h) {

        } else {
            var y = lip.top + tree_st - 100;
            if (y < 0) y = 0;
            $tree.stop().animate({scrollTop: y}, 150);
        }

    }

}
