var PgQuickUndoView = function() {

    var uv = new PgUndoView();

    var win = new PgQuickWindow('Undo stack', uv, 'quickUndo', 300, 520);

    uv.quickWin = win;

    win.showFor$El($('body'));

    var p = pinegrow.getSelectedPage();

    if(p) {
        uv.show(p);
    }

    win.onDestroy = function() {
    }
}

var PgUndoView = function() {

    var css = false;

    var $element = $('<div class="pg-undo-view"></div>');


    var view = new PgUIView($element, 'history');

    view.tabIcon = 'icon-undo';

    this.view = view;
    var _this = this;

    var page = null;
    var states = [];

    var history = null;


    var tabsToolbar = new PgButtonToolbar();
    $element.append(tabsToolbar.$element.addClass('pg-tabs-toolbar'));
    var tabs = new PgToolbarTabButtons(tabsToolbar, 'tabs');

    var combinedTab = tabs.addTab( pgf.edit_html ? 'Combined history' : 'History', pgf.edit_html ? 'Combined history for all pages and stylesheets.' : 'Undo history for all pages', null, pinegrow.undoManager.getUseCombined());

    combinedTab.setClosable(false);

    var htmlTab = tabs.addTab('HTML', 'Selected page HTML history', null, !pinegrow.undoManager.getCssActive());
    htmlTab.setClosable(false);

    var cssTab = tabs.addTab('CSS', 'CSS Stylesheets history', null, pinegrow.undoManager.getCssActive());
    cssTab.setClosable(false);

    tabs.onTabSelected = function(tab) {
        pinegrow.undoManager.setCssActive(tab == cssTab);
        _this.show();
    }

    tabsToolbar.addSection('icons');
    var modeIcon = new PgButtonToolbarItem('icon-down', 'History options.');

    if(pgf.edit_html && false /* depreceated */) {
        //tabsToolbar.addItem('icons', modeIcon);

        modeIcon.onClick = function () {

            var tab = this;

            var actions = [
                {
                    label: 'Separate histories for pages and CSS',
                    check: function () {
                        return !pinegrow.undoManager.getUseCombined();
                    },
                    action: function () {
                        var combined = !pinegrow.undoManager.getUseCombined();
                        pinegrow.showAlert('<p>Changing the history mode will erase undo histories for all pages and stylesheets.</p><p>Are you ok with that?</p>', 'This action will purge all undo histories', 'Cancel', 'Yes', null, function () {
                            updateUseCombinedTabs(combined);
                            pinegrow.undoManager.setUseCombined(combined);
                            pinegrow.showMode('Separate histories', !combined);
                        })
                    }
                }
            ]
            var menu = new CrsaContextMenu(actions);
            menu.showForElement(tab.$element);
        }
    }

    var updateUseCombinedTabs = function(val) {
        if(val) {
            htmlTab.hide();
            cssTab.hide();
            combinedTab.show();
            combinedTab.setActive(true);
        } else {
            htmlTab.show();
            htmlTab.setActive(!pinegrow.undoManager.getCssActive());

            cssTab.show();
            cssTab.setActive(pinegrow.undoManager.getCssActive());

            combinedTab.hide();
        }
    }

    updateUseCombinedTabs(pinegrow.undoManager.getUseCombined());

    var $content = $('<div class="pg-panel-tab-content"></div>').appendTo($element);
    var $ul = $('<ul class="pg-assign-classes-list"></ul>').appendTo($content);

    this.show = function(p) {
        view.whenVisible(function() {

            history = pinegrow.undoManager.getActiveHistory();

            var combined = pinegrow.undoManager.getUseCombined();

            page = pinegrow.getSelectedPage();

            css = history.css;

            $ul.empty();

            var info = history.getInfo();

            var html = '';
            var before = true;

            var last_page_name = null;
            var many_page_names = false;
            var last_cs_name = null;
            var many_cs_names = false;

            states = info.states;

            var start_name = '<span>Start<span>';

            if(info.start_state) {
                start_name = '<span>' + info.start_state.name + '<span>';
            }

            if(info.current_state === null) {
                before = false;
                html += '<li class="active">' + start_name + '</li>';
            } else {
                html += '<li class="start">' + start_name + '</li>';
            }

            var i = 0;
            states.forEach(function(state) {
                //console.log(state);

                var cls = '';
                if(state == info.current_state) {
                    cls = 'active';
                    before = false;
                } else if(!before) {
                    cls = 'after';
                }
                var name = '';
                if(state.element_names && (state.page_name || '') !== 'Project') {
                    name = state.element_names;
                    cls += ' with-name';
                }
                html += '<li class="' + cls + '" data-state-id="' + i + '">';
                if(combined) {
                    if(state.page_name) {
                        html += '<span class="undo-page-name">' + state.page_name + '</span>';
                        if(last_page_name !== null && last_page_name != state.page_name) {
                            many_page_names = true;
                        } else {
                            last_page_name = state.page_name;
                        }
                    }

                    if(state.cs_name) {
                        html += '<span class="undo-cs-name">' + state.cs_name + '</span>';
                        if(last_cs_name !== null && last_cs_name != state.cs_name) {
                            many_cs_names = true;
                        } else {
                            last_cs_name = state.cs_name;
                        }
                    }
                }
                html += '<span>' + state.name + '</span>' + name + '</li>';
                //states.push(state);
                i++;
            })
            $ul.html(html);

            if(many_page_names) {
                $ul.addClass('show-page-names');
            } else {
                $ul.removeClass('show-page-names');
            }

            if(many_cs_names) {
                $ul.addClass('show-cs-names');
            } else {
                $ul.removeClass('show-cs-names');
            }

            var active_y = $ul.find('> .active').position().top;
            var h = $element.height();
            var st = $element.scrollTop();
            if(active_y - 30 < st || active_y > st + h - 30) {
                $element.scrollTop(active_y - Math.floor(h/2));
            }
        })
    }

    var getState = function($li) {
        var idx = parseInt($li.attr('data-state-id'));
        if(idx < states.length) {
            return states[idx];
        }
        return null;
    }

    $ul.on('click', 'li', function(e) {
        e.preventDefault();
        var $li = $(this);
        var state = getState($li);
        history.restoreState(state);
    })

    view.onResize = function() {
    }

    var onPageSelected = function(page) {
        if(!css) {
            _this.show(page);
        }
    }

    var onStateAdded = function(p, undo) {
        tabs.selectTab(p == null ? cssTab : htmlTab);
        //_this.show();
    }

    pinegrow.addEventHandler('on_undo_state_added', onStateAdded);
    pinegrow.addEventHandler('on_undo_redo', onStateAdded);
    pinegrow.addEventHandler('on_page_selected', onPageSelected);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_page_selected', onPageSelected);
        pinegrow.removeEventHandler('on_undo_state_added', onStateAdded);
        pinegrow.removeEventHandler('on_undo_redo', onStateAdded);
    }
}

/*
$('body').one('pinegrow-layout-ready', function() {
    var q = new PgQuickUndoView();
})
*/