$(function() {

    $('body').one('pinegrow-ready', function(e, pinegrow) {
        var f = new PgFramework('bs5', 'Bootstrap 5');
        f.type = "bootstrap";
        f.allow_single_type = true;

        f.description = '<a href="http://getbootstrap.com/">Bootstrap 5</a> starting pages and components. Templates are based on Bootstrap 5.2.2.';
        f.author = 'Pinegrow';
        f.author_link = 'http://pinegrow.com';
        // f.info_badge = 'Beta';

        var simple = false;

        if(pgf.bootstrap) {
            f.default = pgf.bootstrap.activated_by_default || false;
            f.show_in_manager = !f.default;
            simple = pgf.bootstrap.simple || false;
        }

        var addLibSection = function(sec) {
            if(!simple) {
                f.addLibSection(sec);
            }
        }

        f.ignore_css_files = [/(^|\/)bootstrap\.(css|less)/i, /(^|\/)bootstrap\.min\.(css|less)/i, /(^|\/)font(\-|)awesome(\.min|)\.(css|less)/i];

        if(!isApp()) {
            f.detect = function (page) {
                return page.hasStylesheet(/bootstrap/i);
            }
        } else {
            f.detect_async = function (crsaPage, done) {
                crsaPage.stylesheetsThatContain(/bootstrap\s+v5/im, /bootstrap(\.min|)\.(css|less)/i, function (list) {
                    done(list.length > 0)
                });
            }
        }

        var setCurrentBsVersion = function(ver) {
            current_bs_version = ver;
            f.name_short = 'Bootstrap ' + ver;
        }

        var getBootstrapVersionForPage = function(page, force) {
            if(force_bs_version != 'auto') {
                setCurrentBsVersion( parseFloat(force_bs_version));
                return;
            }
            if(page.data.bootstrapVersion && !force) {
                setCurrentBsVersion(page.data.bootstrapVersion);
                //console.log('bs version = ' + current_bs_version);
            } else {
                page.stylesheetsThatContain(/bootstrap\sv(5[0-9\.]*)/i, /bootstrap(\.min|)\.(css|less)/i, function (list, matches, async) {
                    if(list.length > 0) {
                        var a = matches[0][1].split('.');
                        var version = parseFloat(a[0]);
                        if (a.length > 1) {
                            version += parseFloat(a[1]) / 10.0;
                        }
                        setCurrentBsVersion(version);
                        page.data.bootstrapVersion = version;
                        if (async) {
                            pinegrow.selectedElements.reselect(); //trigger update
                        }
                        console.log('bs version = ' + current_bs_version + ' ' + async);
                    }
                }, force);
            }
        }

        var reqVersion = function(min_ver) {
            return current_bs_version >= min_ver;
        }

        var skipVersion = function(min_ver) {
            return current_bs_version < min_ver;
        }

        var getPlaceholderImage = function () {
            return pinegrow.getPlaceholderImage();
        }

        var getGridPreview = function(t) {
            var colcont = '<div style="height:60px;background:#eee;"></div>';

            var sel = 'border:2px solid #0098cc;';
            var conts, rows, cols;
            conts = rows = cols = '';

            switch(t) {
                case 'container':
                    conts = sel;
                    break;
                case 'row':
                    rows = sel;
                    break;
                case 'column':
                    cols = sel;
                    break;
            }

            return '<div class="container-fluid" style="' + conts + '">\
                <div class="row" style="' + rows + '">\
                    <div class="col-4" style="' + cols + '">' + colcont + '</div>\
                    <div class="col-4">' + colcont + '</div>\
                    <div class="col-4">' + colcont + '</div>\
                </div>\
                <div class="row" style="margin-top:12px;">\
                    <div class="col-4">' + colcont + '</div>\
                    <div class="col-4">' + colcont + '</div>\
                    <div class="col-4">' + colcont + '</div>\
                </div>\
            </div>';
        }

        f.setScriptFileByScriptTagId('plugin-bootstrap-5-mt', 'frameworks/bootstrap5/bootstrap5.js'); //get url if script is included directly into edit.html

        var bs_version = 5.2;
        var bs_minor_version = "5.2.0";

        var current_bs_version;

        setCurrentBsVersion(bs_version);

        var force_bs_version = pinegrow.getSetting('bs5-force-version', 'auto');
        if (force_bs_version != 'auto') {
            setCurrentBsVersion(parseFloat(force_bs_version));
        }

        pinegrow.addFramework(f, 3);

        var bm = new PgToggleButtonMaker();
        var prefix = 'bs5.';

        var isColumn = function(pgel) {
            if(pgel.tagName == 'div') {
                if(pgel.parent && pgel.parent.isSelector('.row') && !pgel.hasClass('row')) return true;
                if(pgel.getData(prefix + 'was-column')) return true;
            }

            var cls = pgel.getAttr('class');
            if(cls) {
                if(cls.match(/(\s|^)col($|[\-\s])/i)) return true; // handle .col as well
            }

            return false;
        }

        var num_columns = parseInt(pinegrow.getSetting('bootstrap-col-num', '12')) || 12;
        var sizes = ["xs", "sm", "md", "lg", "xl", "xxl"];
        var sizes_names = ["Extra Small", "Small", "Medium", "Large", "Extra Large", "Extra X Large"];
        var sizes_breakpoints = [0, 576, 768, 992, 1200, 1400];
        var sizes_view_sizes = [320, 640, 768, 1024, 1280, 1440];
        var sizes_icons = ['icon-Xs', 'icon-IPAD', 'icon-tableta-lezeca', 'icon-ikone100--copy_laptop', 'icon-Lg', 'icon-Lg'];

        var size_for_all = sizes[0];

        var sizes_breakpoints_map = {};
        for(var i = 0; i < sizes.length; i++) {
            sizes_breakpoints_map[sizes[i]] = sizes_breakpoints[i];
        }


        // Bootstrap helper
        var bs_options = {
            num_columns: num_columns,
            sizes: sizes,
            sizes_breakpoints: sizes_breakpoints,
            sizes_breakpoints_map: sizes_breakpoints_map,
            sizes_view_sizes: sizes_view_sizes,
            sizes_icons: sizes_icons,
            size_for_all: size_for_all,
            sizes_names: sizes_names,
            prefix: prefix,
            col_prefix_class: 'col',
            isColumn: isColumn,
            onClickOnColumnPoint: function() {
                var q = PgQuickProperties([def_all.sections[this.options.prefix + 'columns']], this.menu.pgel, this.$element, 'Column size &amp; order', 300);
            },
            reqVersion: reqVersion,
            skipVersion: skipVersion,
            version_name: `Bootstrap ${bs_minor_version}`
        }

        var gh = PgBootstrapHelpers(bs_options, f, 5);

        // Return a function that will create the control
        var getDevicesControlFactory = gh.getDevicesControlFactory;

        // Add device selector on prop sections with multi-device controls
        var onSectionWithDevicesControlsShown = gh.onSectionWithDevicesControlsShown;

        // Make field label for provided size
        var makeFieldLabelForSize = gh.makeFieldLabelForSize;


        //all section and field keys should be prefixed by prefix bs5.
        var addPrefixToSectionsAndFields = function (sections) {
            return gh.addPrefixToSectionsAndFields(sections);
        }

        //* Function helpers
        var getBootstrapColor = f.bsGetBootstrapColor = function (type) {
            switch (type) {
                case 'white':
                    return '#fff';
                case 'light':
                    return '#f8f9fa';
                case 'dark':
                    return '#343a40';
                case 'primary':
                    return '#0d6efd';
                case 'secondary':
                    return '#6c757d';
                case 'info':
                    return '#0dcaf0';
                case 'success':
                    return '#198754';
                case 'warning':
                    return '#ffc107';
                case 'danger':
                    return '#dc3545';
                case 'muted':
                    return '#777';
                case 'dark':
                    return '#212529';
                case 'black':
                    return '#000000';
                case 'body':
                    return '#fff';
                default:
                    return '#fff';
            }
        }

        var getButtonTypeToggleButton = f.bsGetButtonTypeToggleButton = function(type, title, h) {
            var attr = {}
            if (title) attr.title = title;

            if(type == 'link') {
                return bm.makeColor(null, 'T', {
                    'text-decoration': 'underline',
                    'attributes': attr
                });
            } else {
                var text_color = type == 'light' ? '#333' : '#fff';
                /*
                var opts = {
                    'color': text_color,
                    'attributes': attr
                }

                 */
                var opts = {title: type, default_color: getBootstrapColor(type)}
                if(h) {
                    opts.height = h;
                }
                return bm.makeColorFromCSSVariable('--bs-' + type, opts);
                //return bm.makeColor(getBootstrapColor(type), '', opts);
            }
        }

        f.bsGetTextColorOptions = function(suff, all) {
            var list = [
                { 'key' : 'text-white' + suff,     'name' : 'White',     html: bm.makeColorTextFromCSSVariable('--bs-white', {title: 'White', default_color: getBootstrapColor('white')}), 'ahtml' : bm.makeColorText( getBootstrapColor('white') )     },
                { 'key' : 'text-light' + suff,     'name' : 'Light',     html: bm.makeColorTextFromCSSVariable('--bs-light', {title: 'Light', default_color: getBootstrapColor('light')}), 'ahtml' : bm.makeColorText( getBootstrapColor('light') )     },
                { 'key' : 'text-dark' + suff,      'name' : 'Dark',      html: bm.makeColorTextFromCSSVariable('--bs-dark', {title: 'Dark', default_color: getBootstrapColor('dark')}), 'ahtml' : bm.makeColorText( getBootstrapColor('dark') )      },
                { 'key' : 'text-primary' + suff,   'name' : 'Primary',   html: bm.makeColorTextFromCSSVariable('--bs-primary', {title: 'Primary', default_color: getBootstrapColor('primary')}), 'ahtml' : bm.makeColorText( getBootstrapColor('primary') )   },
                { 'key' : 'text-secondary' + suff, 'name' : 'Secondary', html: bm.makeColorTextFromCSSVariable('--bs-secondary', {title: 'Secondary', default_color: getBootstrapColor('secondary')}), 'ahtml' : bm.makeColorText( getBootstrapColor('secondary') ) },
                { 'key' : 'text-success' + suff,   'name' : 'Success',   html: bm.makeColorTextFromCSSVariable('--bs-success', {title: 'Success', default_color: getBootstrapColor('success')}), 'ahtml' : bm.makeColorText( getBootstrapColor('success') )   },
                { 'key' : 'text-info' + suff,      'name' : 'Info',      html: bm.makeColorTextFromCSSVariable('--bs-info', {title: 'Info', default_color: getBootstrapColor('info')}), 'ahtml' : bm.makeColorText( getBootstrapColor('info') )      },
                { 'key' : 'text-warning' + suff,   'name' : 'Warning',   html: bm.makeColorTextFromCSSVariable('--bs-warning', {title: 'Warning', default_color: getBootstrapColor('warning')}), 'ahtml' : bm.makeColorText( getBootstrapColor('warning') )   },
                { 'key' : 'text-danger' + suff,    'name' : 'Danger',    html: bm.makeColorTextFromCSSVariable('--bs-danger', {title: 'Danger', default_color: getBootstrapColor('danger')}), 'ahtml' : bm.makeColorText( getBootstrapColor('danger') )    }
            ]

            if (reqVersion(5.1)) {
                list.push(
                    { 'key' : 'text-black' + suff, 'name' : 'Black',     html: bm.makeColorTextFromCSSVariable('--bs-black-rgb', {title: 'Black', default_color: getBootstrapColor('black')}), 'ahtml' : bm.makeColorText( getBootstrapColor('black') )    }
                )
            }

            if(all) {
                list.unshift({
                    'key': 'text-muted',
                    'name': 'Muted',
                    //html: bm.makeColorTextFromCSSVariable('--white', {title: 'White'}),
                    'html': bm.makeColorText(getBootstrapColor('muted'))
                });
                list.push({
                    'key': 'text-body',
                    'name': 'Body',
                    'html': bm.makeColorText('#212529', {attributes: {title: 'Body text color'}})
                });
                list.push({
                    'key': 'text-black-50',
                    'name': 'Black 50%',
                    'html': bm.makeColorText('rgba(0,0,0,0.5)', {attributes: {title: 'Black transparent 50%'}})
                });
                list.push({
                    'key': 'text-white-50',
                    'name': 'White 50%',
                    'html': bm.makeColorText('rgba(255, 255, 255, 0.5)', {attributes: {title: 'White transparent 50%'}})
                });
            }
            return list;
        }


        var textColorOptions = f.bsGetTextColorOptions('');

        f.bsGetLinkColorOptions = function(suff) {
            return [
                {
                    'key': 'link-light' + suff,
                    'name': 'Light',
                    html: bm.makeColorTextFromCSSVariable('--bs-white', {
                        title: 'White',
                        default_color: getBootstrapColor('white')
                    }),
                    'html': bm.makeColorText(getBootstrapColor('light'))
                },
                {'key': 'link-dark' + suff, 'name': 'Dark', 'html': bm.makeColorText(getBootstrapColor('dark'))},
                {
                    'key': 'link-primary' + suff,
                    'name': 'Primary',
                    html: bm.makeColorTextFromCSSVariable('--bs-primary', {
                        title: 'Primary',
                        default_color: getBootstrapColor('primary')
                    }),
                    //'html': bm.makeColorText(getBootstrapColor('primary'))
                },
                {
                    'key': 'link-secondary' + suff,
                    'name': 'Secondary',
                    html: bm.makeColorTextFromCSSVariable('--bs-secondary', {
                        title: 'Secondary',
                        default_color: getBootstrapColor('secondary')
                    }),
                    //'html': bm.makeColorText(getBootstrapColor('secondary'))
                },
                {
                    'key': 'link-success' + suff,
                    'name': 'Success',
                    html: bm.makeColorTextFromCSSVariable('--bs-success', {
                        title: 'Success',
                        default_color: getBootstrapColor('success')
                    }),
                    //'html': bm.makeColorText(getBootstrapColor('success'))
                },
                {
                    'key': 'link-info' + suff, 'name': 'Info',
                    html: bm.makeColorTextFromCSSVariable('--bs-info', {
                        title: 'Info',
                        default_color: getBootstrapColor('info')
                    }),
                    //'html': bm.makeColorText(getBootstrapColor('info'))
                },
                {
                    'key': 'link-warning' + suff,
                    'name': 'Warning',
                    html: bm.makeColorTextFromCSSVariable('--bs-warning', {
                        title: 'Warning',
                        default_color: getBootstrapColor('warning')
                    }),
                    //'html': bm.makeColorText(getBootstrapColor('warning'))
                },
                {
                    'key': 'link-danger' + suff,
                    'name': 'Danger',
                    html: bm.makeColorTextFromCSSVariable('--bs-danger', {
                        title: 'Danger',
                        default_color: getBootstrapColor('danger')
                    }),
                   // 'html': bm.makeColorText(getBootstrapColor('danger'))
                },
            ];
        }

        var linkColorOptions = f.bsGetLinkColorOptions('');

        f.bsGetBackgroundColorOptions = function(suff) {
            var list = [
                {'key' : 'bg-white' + suff,       'name' : 'White',       html: getButtonTypeToggleButton('white',       'White')},
                {'key' : 'bg-light' + suff,       'name' : 'Light',       html: getButtonTypeToggleButton('light',       'Light')},
                {'key' : 'bg-dark' + suff,        'name' : 'Dark',        html: getButtonTypeToggleButton('dark',        'Dark')},
                {'key' : 'bg-primary' + suff,     'name' : 'Primary',     html: getButtonTypeToggleButton('primary',     'Primary')},
                {'key' : 'bg-secondary' + suff,   'name' : 'Secondary',   html: getButtonTypeToggleButton('secondary',   'Secondary')},
                {'key' : 'bg-success' + suff,     'name' : 'Success',     html: getButtonTypeToggleButton('success',     'Success')},
                {'key' : 'bg-info' + suff,        'name' : 'Info',        html: getButtonTypeToggleButton('info',        'Info')},
                {'key' : 'bg-warning' + suff,     'name' : 'Warning',     html: getButtonTypeToggleButton('warning',     'Warning')},
                {'key' : 'bg-danger' + suff,      'name' : 'Danger',      html: getButtonTypeToggleButton('danger',      'Danger')},
                {'key' : 'bg-transparent' + suff, 'name' : 'Transparent', html: getButtonTypeToggleButton('transparent', 'Transparent')}
            ];

            if (reqVersion(5.1)) {
                list.push(
                    {'key' : 'bg-body' + suff,    'name' : 'Body',        html: getButtonTypeToggleButton('body',        'Body')}
                )
            }
            return list;
        }

        var bgColorOptions = f.bsGetBackgroundColorOptions('');

        f.bsGetBorderColorOptions = function(suff) {
            return [
                {'key' : 'border-white' + suff,     'name' : 'White',     html: getButtonTypeToggleButton('white', null, '4px')},
                {'key' : 'border-light' + suff,     'name' : 'Light',     html: getButtonTypeToggleButton('light', null, '4px')},
                {'key' : 'border-dark' + suff,      'name' : 'Dark',      html: getButtonTypeToggleButton('dark', null, '4px')},
                {'key' : 'border-primary' + suff,   'name' : 'Primary',   html: getButtonTypeToggleButton('primary', null, '4px')},
                {'key' : 'border-secondary' + suff, 'name' : 'Secondary', html: getButtonTypeToggleButton('secondary', null, '4px')},
                {'key' : 'border-success' + suff,   'name' : 'Success',   html: getButtonTypeToggleButton('success', null, '4px')},
                {'key' : 'border-info' + suff,      'name' : 'Info',      html: getButtonTypeToggleButton('info', null, '4px')},
                {'key' : 'border-warning' + suff,   'name' : 'Warning',   html: getButtonTypeToggleButton('warning', null, '4px')},
                {'key' : 'border-danger' + suff,    'name' : 'Danger',    html: getButtonTypeToggleButton('danger', null, '4px')}
            ]
        }

        var borderColorOptions = f.bsGetBorderColorOptions('');

        // column def
        var column_span_values = [
            {key: '', name: 'Distribute'},
            {key: '-auto', name: 'Auto'}
        ];

        var order_values = [];

        var offset_values = [];

        for (var i = 0; i <= num_columns; i++) {
            if (i == 0) {
                // order first
                order_values.push({key: '-first', name: 'First'});
                // order 0
                order_values.push({key: '-' + i, name: i});
            }
            else {
                column_span_values.push({key: '-' + i, name: i});

                if (i < num_columns / 2) {
                    // order from 1-5
                    order_values.push({key: '-' + i, name: i});
                } else if (i == num_columns / 2) {
                    // order last
                    order_values.push({key: '-last', name: "Last"});
                }
            }

            if (i != num_columns) {
                offset_values.push({key: '-' + i, name: i});
            }
        }

        // Return a function that will create the control
        var getGridControlFactory = function(control_id, rows) {
            //Returns function that creates the control
            return function() {
                //will keep helper functions and vars private

                //Create control, needs a unique id
                var c = new PgCustomPropertyControl(prefix + control_id);

                //Register subfields.
                c.onDefine = function () {

                    //These fields will not be shown when just registered. This is neccessary to get values.
                    for (var n = 0; n < rows.length; n++) {
                        for (var m = 0; m < sizes.length; m++) {
                            var field = prefix + rows[n].field_prefix + '.' + sizes[m];
                            this.registerInputField(field, createFieldDef(rows[n], sizes[m]))
                        }
                    }
                }

                //Show control. Return the control $el.
                c.onShow = function () {
                    var $table = $("<table/>", {class: 'grid-control columns-control seven-col-grid'});
                    var html = '<td></td>';
                    sizes.forEach(function(size) {
                        html += '<td><label class="grid-control-size">' + size + '</label></td>';
                    })
                    var $row = $("<tr/>").html(html).appendTo($table);

                    for (var n = 0; n < rows.length; n++) {
                        $row = $("<tr/>", {class: ''}).appendTo($table);
                        var $td = $("<td/>").html('<label>' + rows[n].name + '</label>').appendTo($row);

                        for (var m = 0; m < sizes.length; m++) {
                            $td = $("<td/>").appendTo($row);
                            var field = prefix + rows[n].field_prefix + '.' + sizes[m];
                            this.showInputField($td, field, createFieldDef(rows[n], sizes[m]));
                        }
                    }
                    return $table;
                }

                //Called when control is recycled and new values are set
                c.onSetValues = function () {
                    //Nothing to do in this case. PG will take care of updating sub fields with new values
                }

                //Helper functions and vars

                var createFieldDef = function(row, size) {
                    var size_part = (size == size_for_all) ? '' : ('-' + size);
                    var base = row.class_prefix + size_part;
                    var span_select = {
                        'type': 'select',
                        'name': null,
                        'action': 'apply_class',
                        draggable_list : true,
                        'show_empty': true,
                        'options': [],
                        'on_changed' : function(pgel, prop, value) {
                            pgel.setData(prefix + 'was-column', true);
                            if(value) {
                                if(false && !pgel.parent.hasClass('row') && !pgel.parent.hasClass('form-row')) {
                                    pinegrow.showAlert('<p>Column must have a Row parent element.</p><p>Would you like to add a Row?</p>', 'Every Column needs a Row', 'No', 'Yes, add a Row', null, function() {
                                        if(!pgel.isDeleted) {
                                            var row = pgCreate('<div class="row"></div>');
                                            row.insertBefore(pgel);
                                            row.append(pgel);
                                            pinegrow.showQuickMessage('Parent Row added.');
                                        }
                                    })
                                }
                            }
                        }
                    }
                    row.values.forEach(function(d) {
                        span_select.options.push({key: base + d.key, name: d.name});
                    })
                    return span_select;
                }

                //Return the control
                return c;
            };
        }

        var columns_section_def = {
            name : "Columns",
            fields : {
                layout_control: {
                    type: 'custom',
                    name: 'layout_control',
                    action: 'none',
                    control: getGridControlFactory('column-span', [
                        {
                            field_prefix: 'col',
                            class_prefix: 'col',
                            values: column_span_values,
                            name: 'Size'
                        },
                        {
                            field_prefix: 'order',
                            class_prefix: 'order',
                            values: order_values,
                            name: 'Order'
                        },
                        {
                            field_prefix: 'offset',
                            class_prefix: 'offset',
                            values: offset_values,
                            name: 'Offset'
                        }
                    ])
                }
            }
        };

        //* Utilities
        var onJSChangedRefresh = function(pgel) {
            pinegrow.showQuickMessage('Refresh the page to see the changes.', 2000, false, null, 5000);
        }

        var onTooltipsChangedRefresh = function(pgel) {
            pinegrow.showQuickMessage('Refresh the page to see the changes.', 2000, false, null, 5000);
        }

        var onTooltipsChanged = function(pgel) {
            gh.executeScriptInPageDelayed(pgel.getPage(), 'var tooltipTriggerList = [].slice.call(document.querySelectorAll(\'[data-bs-toggle="tooltip"]\'));\nvar tooltipList = tooltipTriggerList.forEach(function (tooltipTriggerEl) {\n    (new bootstrap.Tooltip(tooltipTriggerEl)).dispose();\n})' + bsTooltipsSection._tooltipsCode);
        }

        var bsTooltipsSection = {
            name : 'Tooltip',
            _tooltipsCode: 'var tooltipTriggerList = [].slice.call(document.querySelectorAll(\'[data-bs-toggle="tooltip"]\'))\nvar tooltipList = tooltipTriggerList.map(function (tooltipTriggerEl) {\n    return new bootstrap.Tooltip(tooltipTriggerEl)\n})',
            _tooltipsMainCode: 'var tooltipTriggerList = [].slice.call(document.querySelectorAll(\'[data-bs-toggle="tooltip"]\'))\nvar tooltipList = tooltipTriggerList.map(function (tooltipTriggerEl) {\n    return new bootstrap.Tooltip(tooltipTriggerEl)\n})',
            fields : {
                tooltip : {
                    'type' : 'checkbox',
                    'name' : 'Tooltip',
                    'value' : "1",
                    'action' : 'custom',
                    get_value: function(pgel) {
                        return pgel.attr('data-bs-toggle') == 'tooltip';
                    },
                    set_value: function(pgel, value, values, oldValue, eventType) {
                        if(value) {
                            pgel.attr('data-bs-toggle', 'tooltip');
                            pinegrow.askAboutAddingCodeToPage(pgel.getPage(), bsTooltipsSection._tooltipsCode, 'Enable Bootstrap tooltips', 'Bootstrap tooltips need to be enabled with the call to the .tooltip() function. The following code activates all tooltips on the page:', true, bsTooltipsSection._tooltipsMainCode);

                        } else {
                            if(pgel.attr('data-bs-toggle') == 'tooltip') {
                                pgel.removeAttr('data-bs-toggle');
                            }
                        }
                        onTooltipsChangedRefresh(pgel);
                        return value;
                    }
                },
                tooltptext : {
                    type: 'text',
                    name: 'Text',
                    live_update : true,
                    'action' : 'custom',
                    get_value: function(pgel) {
                        var t = pgel.attr('title');
                        if(!t || t.length == 0) {
                            t = pgel.attr('data-bs-original-title');
                        }
                        return t;
                    },
                    set_value: function(pgel, value, values, oldValue, eventType) {
                        if(value) {
                            pgel.attr('title', value);
                            if(pgel.attr('data-bs-original-title')) {
                                pgel.attr('data-bs-original-title', value);
                            }
                        } else {
                            pgel.removeAttr('title');
                            pgel.removeAttr('data-bs-original-title');
                        }
                        return value;
                    },
                    on_changed : function(pgel, prop, value, oldValue) {
                        if(!oldValue) {
                            onTooltipsChangedRefresh(pgel)
                        } else {
                            onTooltipsChanged(pgel)
                        }
                    }
                },
                tooltipos : {
                    type: 'select',
                    name: 'Placement',
                    action: 'element_attribute',
                    attribute : 'data-bs-placement',
                    'show_empty' : true,
                    'options' : [
                        {'key' : 'top', 'name' : 'Top'},
                        {'key' : 'right', 'name' : 'Right'},
                        {'key' : 'bottom', 'name' : 'Bottom'},
                        {'key' : 'left', 'name' : 'Left'},
                        {'key' : 'auto top', 'name' : 'Auto top'},
                        {'key' : 'auto right', 'name' : 'Auto right'},
                        {'key' : 'auto bottom', 'name' : 'Auto bottom'},
                        {'key' : 'auto left', 'name' : 'Auto left'}
                    ],
                    on_changed : onTooltipsChangedRefresh
                },
                tooltiphtml : {
                    type: 'checkbox',
                    name: 'Html',
                    action: 'element_attribute',
                    attribute: 'data-bs-html',
                    value: 'true',
                    negvalue: 'false',
                    live_update : false,
                    on_changed : onTooltipsChangedRefresh
                },
                tooltiptrigger : {
                    type: 'select',
                    name: 'Trigger',
                    action: 'element_attribute',
                    attribute : 'data-bs-trigger',
                    'show_empty' : true,
                    'options' : [
                        {'key' : 'click', 'name' : 'Click'},
                        {'key' : 'hover', 'name' : 'Hover'},
                        {'key' : 'focus', 'name' : 'Focus'},
                        {'key' : 'manual', 'name' : 'Manual'}
                    ],
                    on_changed : onTooltipsChangedRefresh
                },
                tooltpdelay : {
                    type: 'text',
                    name: 'Delay in ms',
                    action: 'element_attribute',
                    attribute: 'data-bs-delay',
                    live_update : true,
                    on_changed : onTooltipsChangedRefresh
                },
                tooltpcontainer : {
                    type: 'text',
                    name: 'Container',
                    action: 'element_attribute',
                    attribute: 'data-bs-container',
                    live_update : false,
                    on_changed : onTooltipsChangedRefresh
                },
                tooltipcode : {
                    'type' : 'button',
                    'action': 'button',
                    'name' : 'Enable tooltips on the page',
                    'func' : function(pgel, $button) {
                        pinegrow.makeChanges(pgel, 'Add tooltips code', function() {
                            pinegrow.addScriptCodeToPageIfNeeded(pgel.getPage(), bsTooltipsSection._tooltipsCode, 'Tooltips code', true /* execute */, bsTooltipsSection._tooltipsMainCode);
                        })
                    }
                }
            }
        }

        var onPopoversChanged = function(pgel) {
            gh.executeScriptInPageDelayed(pgel.getPage(), 'var popoverTriggerList = [].slice.call(document.querySelectorAll(\'[data-bs-toggle="popover"]\'));\nvar popoverList = popoverTriggerList.forEach(function (popoverTriggerEl) {\n    bootstrap.Popover(popoverTriggerEl).dispose();\n})' + bsPopoverSection._popoversCode);
        }

        var bsPopoverSection = {
            name : 'Popover',
            _popoversCode: 'var popoverTriggerList = [].slice.call(document.querySelectorAll(\'[data-bs-toggle="popover"]\'));\nvar popoverList = popoverTriggerList.map(function (popoverTriggerEl) {\n    return new bootstrap.Popover(popoverTriggerEl);\n})',
            _popoversMainCode: 'var popoverTriggerList = [].slice.call(document.querySelectorAll(\'[data-bs-toggle="popover"]\'));\nvar popoverList = popoverTriggerList.map(function (popoverTriggerEl) {\n    return new bootstrap.Popover(popoverTriggerEl);\n})',
            fields : {
                popover : {
                    'type' : 'checkbox',
                    'name' : 'Popover',
                    'value' : "1",
                    'action' : 'custom',
                    get_value: function(pgel) {
                        return pgel.attr('data-bs-toggle') == 'popover';
                    },
                    set_value: function(pgel, value, values, oldValue, eventType) {
                        if(value) {
                            pgel.attr('data-bs-toggle', 'popover');
                            pinegrow.askAboutAddingCodeToPage(pgel.getPage(), bsPopoverSection._popoversCode, 'Enable Bootstrap popovers', 'Bootstrap popovers need to be enabled with the call to the .popover() function. The following code activates all popovers on the page:', true, bsPopoverSection._popoversMainCode);
                        } else {
                            if(pgel.attr('data-bs-toggle') == 'popover') {
                                pgel.removeAttr('data-bs-toggle');
                            }
                        }
                        onTooltipsChangedRefresh(pgel);
                        return value;
                    }
                },
                popovertext : {
                    type: 'text',
                    name: 'Title',
                    'action' : 'custom',
                    get_value: function(pgel) {
                        var t = pgel.attr('title');
                        if(!t || t.length == 0) {
                            t = pgel.attr('data-bs-original-title');
                        }
                        return t;
                    },
                    set_value: function(pgel, value, values, oldValue, eventType) {
                        if(value) {
                            pgel.attr('title', value);
                            if(pgel.attr('data-bs-original-title')) {
                                pgel.attr('data-bs-original-title', value);
                            }
                        } else {
                            pgel.removeAttr('title');
                            pgel.removeAttr('data-bs-original-title');
                        }
                        return value;
                    },
                    on_changed : function(pgel, prop, value, oldValue) {
                        if(!oldValue) {
                            onTooltipsChangedRefresh(pgel)
                        } else {
                            onPopoversChanged(pgel)
                        }
                    }
                },
                popovercontent : {
                    type: 'text',
                    name: 'Content',
                    action: 'element_attribute',
                    attribute: 'data-bs-content',
                    on_changed : onPopoversChanged
                },
                popoverpos : {
                    type: 'select',
                    name: 'Placement',
                    action: 'element_attribute',
                    attribute : 'data-bs-placement',
                    'show_empty' : true,
                    'options' : [
                        {'key' : 'top', 'name' : 'Top'},
                        {'key' : 'right', 'name' : 'Right'},
                        {'key' : 'bottom', 'name' : 'Bottom'},
                        {'key' : 'left', 'name' : 'Left'},
                        {'key' : 'auto top', 'name' : 'Auto top'},
                        {'key' : 'auto right', 'name' : 'Auto right'},
                        {'key' : 'auto bottom', 'name' : 'Auto bottom'},
                        {'key' : 'auto left', 'name' : 'Auto left'}
                    ],
                    on_changed : onTooltipsChangedRefresh
                },
                popoverhtml : {
                    type: 'checkbox',
                    name: 'Html',
                    action: 'element_attribute',
                    attribute: 'data-bs-html',
                    value: 'true',
                    negvalue: 'false',
                    live_update : false,
                    on_changed : onTooltipsChangedRefresh
                },
                popovertrigger : {
                    type: 'select',
                    name: 'Trigger',
                    action: 'element_attribute',
                    attribute : 'data-bs-trigger',
                    'show_empty' : true,
                    'options' : [
                        {'key' : 'click', 'name' : 'Click'},
                        {'key' : 'hover', 'name' : 'Hover'},
                        {'key' : 'focus', 'name' : 'Focus'},
                        {'key' : 'manual', 'name' : 'Manual'}
                    ],
                    on_changed : onTooltipsChangedRefresh
                },
                popoverdelay : {
                    type: 'text',
                    name: 'Delay in ms',
                    action: 'element_attribute',
                    attribute: 'data-bs-delay',
                    live_update : false,
                    on_changed : onTooltipsChangedRefresh
                },
                popovercontainer : {
                    type: 'text',
                    name: 'Container',
                    action: 'element_attribute',
                    attribute: 'data-bs-container',
                    live_update : false,
                    on_changed : onTooltipsChangedRefresh
                },
                popovercode : {
                    'type' : 'button',
                    'action': 'button',
                    'name' : 'Enable popovers on the page',
                    'func' : function(pgel, $button) {
                        pinegrow.makeChanges(pgel, 'Add popovers code', function() {
                            pinegrow.addScriptCodeToPageIfNeeded(pgel.getPage(), bsPopoverSection._popoversCode, 'Popovers code', true /* execute */, bsPopoverSection._popoversMainCode);
                        })
                    }
                }
            }
        }

        var bsScrollSpySection = {
            name: 'Scroll Spy',
            fields: {
                scrollspy: {
                    'type' : 'checkbox',
                    'name' : 'Scroll Spy',
                    'value' : "scroll",
                    'action': 'element_attribute',
                    'attribute': 'data-bs-spy'
                },
                scrollspytarget: {
                    'type' : 'text',
                    'name' : 'Target',
                    'action': 'element_attribute',
                    'attribute': 'data-bs-target'
                },
                scrollspyoffset: {
                    'type' : 'text',
                    'name' : 'Offset',
                    'action': 'element_attribute',
                    'attribute': 'data-bs-offset'
                }
            }
        }

        var bsDataTarget = {
            type: 'text',
            name: 'Data target',
            type: 'text',
            live_update: true,
            action : 'custom',
            get_value: function(pgel) {
                return pgel.attr('data-bs-target');
            },
            set_value: function(pgel, value, values, oldValue, eventType) {
                if(value) {
                    pgel.attr('data-bs-target', value);
                } else {
                    pgel.removeAttr('data-bs-target');
                }
                return value;
            }
        }

        var bsDataToggle = {
            type: 'select',
            name: 'Data toggle',
            live_update: false,
            show_empty: true,
            options : [
                {key : 'dropdown', name : 'Dropdown' },
                {key : 'offcanvas', name : 'Offcanvas' },
                {key : 'modal', name : 'Modal' },
                {key : 'tab', name : 'Tab' },
                {key : 'pill', name : 'Pill' },
                {key : 'tooltip', name : 'Tooltip' },
                {key : 'popover', name : 'Popover' },
                // {key : 'button', name : 'Button' },
                // {key : 'buttons', name : 'Buttons' },
                {key : 'collapse', name : 'Collapse' },
                {key : 'list', name : 'List' },
            ],
            action : 'custom',
            get_value: function(pgel) {
                return pgel.attr('data-bs-toggle');
            },
            set_value: function(pgel, value, values, oldValue, eventType) {
                if(value) {
                    pgel.attr('data-bs-toggle', value);
                } else {
                    pgel.removeAttr('data-bs-toggle');
                }
                return value;
            }
        }

        var bsDataDisplay = {
            type: 'checkbox',
            name: 'Static Display',
            action: 'element_attribute',
            value: 'static',
            attribute: 'data-bs-display',
            attribute_without_value : false
        }

        var class_disabled = {
            type: 'checkbox',
            action : 'apply_class',
            name: 'Disabled',
            value: 'disabled'
        };

        var class_active = {
            type: 'checkbox',
            action : 'apply_class',
            name: 'Active',
            value: 'active'
        };

        //Add BS common sections to all elements
        var def_all = new PgComponentType(prefix + 'all', 'All elements', {
            selector: function (pgel) {
                return true
            },
            name: 'Div',
            display_name: 'tag',
            priority: 2001,
            sections: addPrefixToSectionsAndFields({
                bstextcontext: {
                    name: 'Text',
                    compact: true,
                    default_closed: false,
                    fields: {
                        bsemphasis: {
                            'type': 'select',
                            'name': 'Text color',
                            'action': 'apply_class',
                            'show_empty': true,
                            'toggle_buttons': true,
                            'options': function() { return f.bsGetTextColorOptions('', true); }
                        },
                        bslinkcolor: {
                            'type': 'select',
                            'name': 'Link color',
                            'action': 'apply_class',
                            'show_empty': true,
                            'toggle_buttons': true,
                            'options': linkColorOptions,
                            show_if: function (values, pgel) {
                                return pgel.is('a');
                            }
                        },
                        bstabindex: {
                            'type': 'text',
                            'name': 'Tab index',
                            'action' : 'element_attribute',
                            'attribute' : 'tabindex',
                            show_if: function (values, pgel) {
                                return pgel.is('a');
                            }
                        },
                        bstexttransform: {
                            'type': 'select',
                            'name': 'Transform',
                            'action': 'apply_class',
                            'show_empty': true,
                            'toggle_buttons': true,
                            'options': [
                                {'key': 'text-lowercase',  'name': 'Lowercase', 'html': bm.makeIcon('lowercase')},
                                {'key': 'text-uppercase',  'name': 'Uppercase', 'html': bm.makeIcon('uppercase')},
                                {'key': 'text-capitalize', 'name': 'Capitalize', 'html': bm.makeIcon('sentence-case')}
                            ]
                        },
                        bsfontstyle: {
                            'type': 'select',
                            'name': 'Font style',
                            'action': 'apply_class',
                            'show_empty': true,
                            'toggle_buttons' : true,
                            'options' : [
                                {
                                    'key' : 'fst-normal',
                                    'name' : 'Normal',
                                    'html' : bm.makeIcon('style_normal', { attributes: { 'title': 'Normal' } })
                                },
                                {
                                    'key' : 'fst-italic',
                                    'name' : 'Italic',
                                    'html' : bm.makeIcon('style_italic', { attributes: { 'title': 'Italic' } })
                                }
                            ]
                        },
                        bsfontdecoration: {
                            'type': 'select',
                            'name': 'Font decoration',
                            'action': 'apply_class',
                            'show_empty': true,
                            'toggle_buttons' : true,
                            'options' : [
                                {
                                    'key' : 'text-decoration-none',
                                    'name' : 'None',
                                    'html' : bm.makeIcon('decoration-none', { attributes: { 'title': 'None' } })
                                },
                                {
                                    'key' : 'text-decoration-underline',
                                    'name' : 'Underline',
                                    'html' : bm.makeIcon('decoration-under', { attributes: { 'title': 'Underline' } })
                                },
                                {
                                    'key' : 'text-decoration-line-through',
                                    'name' : 'Line through',
                                    'html' : bm.makeIcon('decoration-through', { attributes: { 'title': 'Line through' } })
                                }
                            ]
                        },
                        textopacity: {
                            'type': 'select',
                            'name': 'Text Opacity',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'text-opacity-25', 'name': '25'},
                                {'key': 'text-opacity-50', 'name': '50'},
                                {'key': 'text-opacity-75', 'name': '75'},
                                {'key': 'text-opacity-100', 'name': '100'},
                            ],
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                        displayheadings: {
                            'type' : 'select',
                            'name' : 'Heading Size',
                            'action' : 'apply_class',
                            'show_empty' : true,
                            'options' : [
                                { 'key' : 'display-1',  'name' : '1 (big)' },
                                { 'key' : 'display-2',  'name' : '2' },
                                { 'key' : 'display-3',  'name' : '3' },
                                { 'key' : 'display-4',  'name' : '4' },
                                { 'key' : 'display-5',  'name' : '5' },
                                { 'key' : 'display-6',  'name' : '6 (small)' },
                                { 'key' : 'h1',  'name' : 'h1' },
                                { 'key' : 'h2',  'name' : 'h2' },
                                { 'key' : 'h3',  'name' : 'h3' },
                                { 'key' : 'h4',  'name' : 'h4' },
                                { 'key' : 'h5',  'name' : 'h5' },
                                { 'key' : 'h6',  'name' : 'h6' },
                            ]
                        },
                        bsfontsize: {
                            'type': 'select',
                            'name': 'Font size',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'fs-1', 'name': '1'},
                                {'key': 'fs-2', 'name': '2'},
                                {'key': 'fs-3', 'name': '3'},
                                {'key': 'fs-4', 'name': '4'},
                                {'key': 'fs-5', 'name': '5'},
                                {'key': 'fs-6', 'name': '6'},
                            ]
                        },
                        bsfontweight: {
                            'type': 'select',
                            'name': 'Font weight',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'fw-bold', 'name': 'Bold'},
                                {'key': 'fw-semibold', 'name': 'Semibold'},
                                {'key': 'fw-bolder', 'name': 'Bolder'},
                                {'key': 'fw-normal', 'name': 'Normal'},
                                {'key': 'fw-light', 'name': 'Light'},
                                {'key': 'fw-lighter', 'name': 'Lighter'},
                            ]
                        },
                        bsineheight: {
                            'type': 'select',
                            'name': 'Line height',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'lh-1', 'name': '1'},
                                {'key': 'lh-sm', 'name': 'Small'},
                                {'key': 'lh-base', 'name': 'Base'},
                                {'key': 'lh-lg', 'name': 'Large'},
                            ]
                        },
                        bspointevent: {
                            'type': 'select',
                            'name': 'Point event',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {key: 'pe-none', name: 'None'},
                                {key: 'pe-auto', name: 'Auto'},
                            ]
                        },
                        bstextwrap: {
                            'type': 'select',
                            'name': 'Text wrap',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {key: 'text-wrap', name: 'Wrap'},
                                {key: 'text-nowrap', name: 'No wrap'},
                            ]
                        },
                        bslead: {
                            'type': 'checkbox',
                            'name': 'Lead',
                            'action': 'apply_class',
                            'value': 'lead'
                        },
                        bstextbreak: {
                            'type': 'checkbox',
                            'name': 'Text break',
                            'action': 'apply_class',
                            'value': 'text-break',
                        },
                        bstextrest: {
                            'type': 'checkbox',
                            'name': 'Text reset',
                            'action': 'apply_class',
                            'value': 'text-reset',
                        },
                        texttruncate: {
                            'type': 'checkbox',
                            'name': 'Truncate',
                            'action': 'apply_class',
                            'value': 'text-truncate'
                        },
                        bstextmonospace: {
                            'type': 'checkbox',
                            'name': 'Monospace',
                            'action': 'apply_class',
                            'value': 'font-monospace',
                        },
                        bsstretchedlink: {
                            'type': 'checkbox',
                            'name': 'Stretched link',
                            'action': 'apply_class',
                            'value': 'stretched-link',
                            show_if: function (values, pgel) {
                                return pgel.is('a');
                            }
                        }
                    }
                },
                bsbackground: {
                    name: 'Background',
                    compact: true,
                    default_closed: false,
                    fields: {
                        bsbckcontext: {
                            'type': 'select',
                            'name': 'Background',
                            'action': 'apply_class',
                            'show_empty': true,
                            'toggle_buttons': true,
                            'options': bgColorOptions
                        },
                        bsbckgradient: {
                            'type': 'checkbox',
                            'name': 'Gradient',
                            'action': 'apply_class',
                            'value': 'bg-gradient'
                        },
                        bsbgopacity: {
                            'type': 'select',
                            'name': 'Opacity',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'bg-opacity-10', 'name': '10'},
                                {'key': 'bg-opacity-25', 'name': '25'},
                                {'key': 'bg-opacity-50', 'name': '50'},
                                {'key': 'bg-opacity-75', 'name': '75'},
                                {'key': 'bg-opacity-100', 'name': '100'},
                            ],
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                        bsboxshadow: {
                            'type': 'select',
                            'name': 'Box shadow',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {key: 'shadow-none', name: 'None'},
                                {key: 'shadow-sm', name: 'Small'},
                                {key: 'shadow', name: 'Regular'},
                                {key: 'shadow-lg', name: 'Large'}
                            ]
                        },
                        bsplaceholder: {
                            'type': 'checkbox',
                            'name': 'Placeholder',
                            'action': 'apply_class',
                            'value': 'placeholder',
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                        bsplaceholderanimation: {
                            'type': 'select',
                            'name': 'Placeholder Animation',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                { key: 'placeholder-glow', name: 'Glow' },
                                { key: 'placeholder-wave', name: 'Wave' }
                            ],
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        }
                    }
                },
                bsspacing: {
                    name: "Spacing",
                    on_section_shown: onSectionWithDevicesControlsShown,
                    fields: {
                        margin_control: {
                            type: 'custom',
                            name: 'layout_control',
                            action: 'none',
                            control: gh.getSpacingControl()
                        },
                        rowgap: {
                            type: 'custom',
                            name: 'layout_control',
                            action: 'none',
                            control: gh.getGappingControl()
                        },
                        cssgridgap: {
                            type: 'custom',
                            name: 'css_grid_gap',
                            action: 'none',
                            control: getDevicesControlFactory('cssgrid.gap', 'CSS Grid Gap', function (size, size_token) {
                                var base = 'gap' + size_token + '-';

                                var field = {
                                    'type': 'select',
                                    'name': makeFieldLabelForSize('CSS Grid Gap', size),
                                    'action': 'apply_class',
                                    'show_empty': true,
                                    //'toggle_buttons': true,
                                    'options': [
                                        {key: base + '0', name: '0'},
                                        {key: base + '1', name: '1'},
                                        {key: base + '2', name: '2'},
                                        {key: base + '3', name: '3'},
                                        {key: base + '4', name: '4'},
                                        {key: base + '5', name: '5'}
                                    ]
                                }
                                return field;
                            })
                        },
                    }
                },
                bstextalign: {
                    name: "Layout",
                    on_section_shown: onSectionWithDevicesControlsShown,
                    fields: {
                        bstextalign: {
                            type: 'custom',
                            name: 'text_align',
                            action: 'none',
                            control: getDevicesControlFactory('text.align', 'Text Align', function (size, size_token) {
                                var base = 'text' + size_token + '-';

                                var field = {
                                    'type': 'select',
                                    'name': makeFieldLabelForSize('Text align', size),
                                    'action': 'apply_class',
                                    'show_empty': true,
                                    'toggle_buttons': true,
                                    'options': [
                                        {
                                            'key': base + 'start',
                                            'name': 'Start',
                                            'html': bm.makeIcon('align-left', {attributes: {'title': 'Start'}})
                                        },
                                        {
                                            'key': base + 'center',
                                            'name': 'Center',
                                            'html': bm.makeIcon('align-center', {attributes: {'title': 'Center'}})
                                        },
                                        {
                                            'key': base + 'end',
                                            'name': 'End',
                                            'html': bm.makeIcon('align-right', {attributes: {'title': 'End'}})
                                        }
                                    ]
                                }

                                return field;
                            })
                        },
                        divfloat: {
                            type: 'custom',
                            name: 'Float',
                            action: 'none',
                            control: getDevicesControlFactory('float', 'Float', function (size, size_token) {
                                var base = 'float' + size_token + '-';

                                var field = {
                                    'type': 'select',
                                    'name': makeFieldLabelForSize('Float', size),
                                    'action': 'apply_class',
                                    'show_empty': true,
                                    'toggle_buttons': true,
                                    'options': [
                                        {
                                            'key': base + 'none',
                                            'name': 'None',
                                            'html': bm.makeIcon('float', {attributes: {'title': 'None'}})
                                        },
                                        {
                                            'key': base + 'start',
                                            'name': 'Start',
                                            'html': bm.makeIcon('float-left"', {attributes: {'title': 'Start'}})
                                        },
                                        {
                                            'key': base + 'end',
                                            'name': 'End',
                                            'html': bm.makeIcon('float-right', {attributes: {'title': 'End'}})
                                        }
                                    ]
                                }

                                return field;
                            })
                        },
                        divdisplay: {
                            type: 'custom',
                            name: 'Display',
                            action: 'none',
                            control: getDevicesControlFactory('display', 'Display', function (size, size_token) {
                                var base = 'd' + size_token + '-';

                                var field = {
                                    'type': 'select',
                                    'name': makeFieldLabelForSize('Display', size),
                                    'action': 'apply_class',
                                    'show_empty': true,
                                    'options': [
                                        {'key': base + 'none',         'name': 'None'},
                                        {'key': base + 'inline',       'name': 'Inline'},
                                        {'key': base + 'inline-block', 'name': 'Inline block'},
                                        {'key': base + 'block',        'name': 'Block'},
                                        {'key': base + 'grid',         'name': 'Grid'},
                                        {'key': base + 'table',        'name': 'Table'},
                                        {'key': base + 'table-row',    'name': 'Table row'},
                                        {'key': base + 'table-cell',   'name': 'Table cell'},
                                        {'key': base + 'flex',         'name': 'Flex'},
                                        {'key': base + 'inline-flex',  'name': 'Inline flex'},
                                    ]
                                }

                                return field;
                            })
                        }
                    }
                },
                bslayout: {
                    name: 'Display',
                    fields: {
                        bsverticalalign: {
                            type: 'select',
                            name: 'Vertical align',
                            action: 'apply_class',
                            show_empty: true,
                            options: [
                                {key: 'align-baseline', name: 'Baseline'},
                                {key: 'align-top', name: 'Top'},
                                {key: 'align-middle', name: 'Middle'},
                                {key: 'align-bottom', name: 'Bottom'},
                                {key: 'align-text-top', name: 'Text top'},
                                {key: 'align-text-bottom', name: 'Text bottom'},
                            ]
                        },
                        bsoverflow: {
                            type: 'select',
                            name: 'Overflow',
                            action: 'apply_class',
                            show_empty: true,
                            options: [
                                {key: 'overflow-auto', name: 'Auto'},
                                {key: 'overflow-hidden', name: 'Hidden'},
                                {key: 'overflow-visible', name: 'Visible'},
                                {key: 'overflow-scroll', name: 'Scroll'},
                            ]
                        },
                        bstranslate: {
                            'type': 'select',
                            'name': 'Translate',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'translate-middle', 'name': 'Middle'},
                                {'key': 'translate-middle-x', 'name': 'Middle X'},
                                {'key': 'translate-middle-y', 'name': 'Middle Y'},
                            ]
                        },
                        clearfix: {
                            'type': 'checkbox',
                            'name': 'Clear floats',
                            'action': 'apply_class',
                            'value': 'clearfix'
                        },
                        bsvirticalrule: {
                            'type': 'checkbox',
                            'name': 'Vertical rule',
                            'action': 'apply_class',
                            'value': 'vr',
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                        bsstack: {
                            'type': 'select',
                            'name': 'Stack type',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'vstack', 'name': 'Vertical'},
                                {'key': 'hstack', 'name': 'Horizontal'},
                            ],
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                        bsopacity: {
                            'type': 'select',
                            'name': 'Opacity',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': '0', 'name': '0'},
                                {'key': '25', 'name': '25'},
                                {'key': '50', 'name': '50'},
                                {'key': '75', 'name': '75'},
                                {'key': '100', 'name': '100'},
                            ],
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                        bsposition: {
                            'type': 'select',
                            'name': 'Position',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'position-static', 'name': 'Static'},
                                {'key': 'position-relative', 'name': 'Relative'},
                                {'key': 'position-absolute', 'name': 'Absolute'},
                                {'key': 'position-fixed', 'name': 'Fixed'},
                                {'key': 'position-sticky', 'name': 'Sticky'}
                            ]
                        },
                        bsposfixed: {
                            'type': 'select',
                            'name': 'Position fixed',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'fixed-top', 'name': 'Top'},
                                {'key': 'fixed-bottom', 'name': 'Bottom'},
                            ]
                        },
                        bspossticky: {
                            'type': 'select',
                            'name': 'Position sticky',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'sticky-top', 'name': 'All sizes'},
                                {'key': 'sticky-sm-top', 'name': 'Small'},
                                {'key': 'sticky-md-top', 'name': 'Medium'},
                                {'key': 'sticky-xl-top', 'name': 'X Large'},
                                {'key': 'sticky-xxl-top', 'name': 'XX Large'},
                            ]
                        },
                        bsarrange: {
                            type: 'custom',
                            name: 'element_position',
                            action: 'none',
                            control: gh.getArrangeControl(),
                        },
                        bswidth: {
                            'type': 'select',
                            'name': 'Width',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'w-auto', 'name': 'Auto'},
                                {'key': 'w-25', 'name': '25%'},
                                {'key': 'w-50', 'name': '50%'},
                                {'key': 'w-75', 'name': '75%'},
                                {'key': 'w-100', 'name': '100%'},
                                {'key': 'vw-100', 'name': '100vw'}
                            ]
                        },
                        bsmxwidth: {
                            'type': 'checkbox',
                            'name': 'Max width',
                            'action': 'apply_class',
                            'value': 'mw-100'
                        },
                        bsmnwidth: {
                            'type': 'checkbox',
                            'name': 'Min VW width',
                            'action': 'apply_class',
                            'value': 'min-vw-100',
                        },
                        bsheight: {
                            'type': 'select',
                            'name': 'Height',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'h-auto', 'name': 'Auto'},
                                {'key': 'h-25', 'name': '25%'},
                                {'key': 'h-50', 'name': '50%'},
                                {'key': 'h-75', 'name': '75%'},
                                {'key': 'h-100', 'name': '100%'},
                                {'key': 'vh-100', 'name': '100vh'}
                            ]
                        },
                        bsmxheight: {
                            'type': 'checkbox',
                            'name': 'Max height',
                            'action': 'apply_class',
                            'value': 'mh-100'
                        },
                        bsmnheight: {
                            'type': 'checkbox',
                            'name': 'Min VH height',
                            'action': 'apply_class',
                            'value': 'min-vh-100',
                        },
                        bsuserselect: {
                            'type': 'select',
                            'name': 'User select',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {key: 'user-select-all', name: 'All'},
                                {key: 'user-select-auto', name: 'Auto'},
                                {key: 'user-select-none', name: 'None'},
                            ]
                        },
                        bscaptiontop: {
                            type: 'checkbox',
                            name: 'Caption top',
                            action: 'apply_class',
                            value: 'caption-top'
                        }
                    }
                },
                bsborders: {
                    name: 'Border',
                    fields: {
                        hasborder: {
                            type: 'select',
                            name: 'Border',
                            action: 'apply_class_multiple',
                            show_empty: true,
                            multiple: true,
                            options: [
                                {key: 'border', name: 'All sides'},
                                {key: 'border-top', name: 'Top'},
                                {key: 'border-end', name: 'End'},
                                {key: 'border-bottom', name: 'Bottom'},
                                {key: 'border-start', name: 'Start'},
                            ]
                        },
                        hideborder: {
                            type: 'select',
                            name: 'Hide',
                            action: 'apply_class_multiple',
                            show_empty: true,
                            multiple: true,
                            options: [
                                {key: 'border-0', name: 'All sides'},
                                {key: 'border-top-0', name: 'Top'},
                                {key: 'border-end-0', name: 'End'},
                                {key: 'border-bottom-0', name: 'Bottom'},
                                {key: 'border-start-0', name: 'Start'},
                            ]
                        },
                        widthbotder: {
                            type: 'select',
                            name: 'Width',
                            action: 'apply_class',
                            show_empty: true,
                            options: [
                                {key: 'border-1', name: '1'},
                                {key: 'border-2', name: '2'},
                                {key: 'border-3', name: '3'},
                                {key: 'border-4', name: '4'},
                                {key: 'border-5', name: '5'},
                            ]
                        },
                        raduisborder: {
                            type: 'select',
                            name: 'Rounded',
                            action: 'apply_class_multiple',
                            show_empty: true,
                            multiple: true,
                            options: [
                                {key: 'rounded', name: 'All sides'},
                                {key: 'rounded-top', name: 'Top'},
                                {key: 'rounded-end', name: 'End'},
                                {key: 'rounded-bottom', name: 'Bottom'},
                                {key: 'rounded-start', name: 'Start'},
                                {key: 'rounded-circle', name: 'Circle'},
                                {key: 'rounded-pill', name: 'Pill'},
                            ]
                        },
                        raduisdegreeborder: {
                            type: 'select',
                            name: 'Rounded degree',
                            action: 'apply_class',
                            show_empty: true,
                            options: [
                                {key: 'rounded-0', name: '0'},
                                {key: 'rounded-1', name: '1'},
                                {key: 'rounded-2', name: '2'},
                                {key: 'rounded-3', name: '3'},
                                {key: 'rounded-4', name: '4'},
                                {key: 'rounded-5', name: '5'},
                            ]
                        },
                        bordercolor: {
                            type: 'select',
                            name: 'Border',
                            action: 'apply_class',
                            show_empty: true,
                            toggle_buttons: true,
                            options: borderColorOptions
                        }
                    }
                },
                columns: columns_section_def,
                bsvisible: {
                    name: 'Visibility',
                    compact: true,
                    fields: {
                        bsvisibility: {
                            'type': 'select',
                            'name': 'Visibility',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'visible', 'name': 'Visible'},
                                {'key': 'invisible', 'name': 'Invisible'}
                            ]
                        },
                        bsvisually: {
                            'type': 'select',
                            'name': 'Visually hidden',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'visually-hidden', 'name': 'Always'},
                                {'key': 'visually-hidden-focusable', 'name': 'Not focused'}
                            ]
                        },
                        bsprint: {
                            'type': 'select',
                            'name': 'Print',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                {'key': 'd-print-none',         'name': 'None'},
                                {'key': 'd-print-block',        'name': 'Block'},
                                {'key': 'd-print-inline',       'name': 'Inline'},
                                {'key': 'd-print-inline-block', 'name': 'Inline block'},
                                {'key': 'd-print-grid',         'name': 'Grid'},
                                {'key': 'd-print-table',        'name': 'Table'},
                                {'key': 'd-print-table-row',    'name': 'Table row'},
                                {'key': 'd-print-table-cell',   'name': 'Table cell'},
                                {'key': 'd-print-flex',         'name': 'Flex'},
                                {'key': 'd-print-inline-flex',  'name': 'Inline flex'}
                            ]
                        },
                    }
                },
                bsflex: {
                    name: "Flex container",
                    on_section_shown: onSectionWithDevicesControlsShown,
                    fields: {
                        flex_display: {
                            type: 'custom',
                            name: 'flex_display',
                            action: 'none',
                            control: getDevicesControlFactory('flex.display', 'Display', function (size, size_token) {
                                var base = 'd' + size_token + '-';

                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Display', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'flex',
                                            name: 'Flex',
                                            html: bm.makeText('Flex', {attributes: {title: 'Flex'}})
                                        },
                                        {
                                            key: base + 'inline-flex',
                                            name: 'Inline flex',
                                            html: bm.makeText('Inline flex', {attributes: {title: 'Inline flex'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_direction: {
                            type: 'custom',
                            name: 'flex_direction',
                            action: 'none',
                            control: getDevicesControlFactory('flex.direction', 'Direction', function (size, size_token) {
                                var base = 'flex' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Direction', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'row',
                                            name: 'Row',
                                            html: bm.makeIcon('225', {attributes: {'title': 'Row'}})
                                        },
                                        {
                                            key: base + 'row-reverse',
                                            name: 'Row rev.',
                                            html: bm.makeIcon('224', {attributes: {'title': 'Row reverse'}})
                                        },
                                        {
                                            key: base + 'column',
                                            name: 'Column',
                                            html: bm.makeIcon('227', {attributes: {'title': 'Column'}})
                                        },
                                        {
                                            key: base + 'column-reverse',
                                            name: 'Col. rev.',
                                            html: bm.makeIcon('226', {attributes: {'title': 'Column reverse'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_justify_content: {
                            type: 'custom',
                            name: 'flex_justify_content',
                            action: 'none',
                            control: getDevicesControlFactory('flex.justify.content', 'Justify content', function (size, size_token) {
                                var base = 'justify-content' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Justify content', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'start',
                                            name: '',
                                            html: bm.makeIcon('223', {attributes: {'title': 'Start'}})
                                        },
                                        {
                                            key: base + 'end',
                                            name: '',
                                            html: bm.makeIcon('218', {attributes: {'title': 'End'}})
                                        },
                                        {
                                            key: base + 'center',
                                            name: '',
                                            html: bm.makeIcon('219', {attributes: {'title': 'Center'}})
                                        },
                                        {
                                            key: base + 'between',
                                            name: '',
                                            html: bm.makeIcon('206', {attributes: {'title': 'Between'}})
                                        },
                                        {
                                            key: base + 'around',
                                            name: '',
                                            html: bm.makeIcon('205', {attributes: {'title': 'Around'}})
                                        },
                                        {
                                            key: base + 'evenly',
                                            name: '',
                                            html: bm.makeIcon('205', {attributes: {'title': 'Evenly'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_align_items: {
                            type: 'custom',
                            name: 'flex_align_items',
                            action: 'none',
                            control: getDevicesControlFactory('flex.align.items', 'Align items', function (size, size_token) {
                                var base = 'align-items' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Align items', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'start',
                                            name: '',
                                            html: bm.makeIcon('223', {attributes: {'title': 'Start'}})
                                        },
                                        {
                                            key: base + 'end',
                                            name: '',
                                            html: bm.makeIcon('220', {attributes: {'title': 'End'}})
                                        },
                                        {
                                            key: base + 'center',
                                            name: '',
                                            html: bm.makeIcon('221', {attributes: {'title': 'Center'}})
                                        },
                                        {
                                            key: base + 'baseline',
                                            name: '',
                                            html: bm.makeIcon('212', {attributes: {'title': 'Baseline'}})
                                        },
                                        {
                                            key: base + 'stretch',
                                            name: '',
                                            html: bm.makeIcon('214', {attributes: {'title': 'Stretch'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_align_content: {
                            type: 'custom',
                            name: 'flex_align_content',
                            action: 'none',
                            control: getDevicesControlFactory('flex.align.content', 'Align content', function (size, size_token) {
                                var base = 'align-content' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Align content', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'start',
                                            name: '',
                                            html: bm.makeIcon('217', {attributes: {'title': 'Start'}})
                                        },
                                        {
                                            key: base + 'end',
                                            name: '',
                                            html: bm.makeIcon('216', {attributes: {'title': 'End'}})
                                        },
                                        {
                                            key: base + 'center',
                                            name: '',
                                            html: bm.makeIcon('215', {attributes: {'title': 'Center'}})
                                        },
                                        {
                                            key: base + 'between',
                                            name: '',
                                            html: bm.makeIcon('210', {attributes: {'title': 'Between'}})
                                        },
                                        {
                                            key: base + 'around',
                                            name: '',
                                            html: bm.makeIcon('209', {attributes: {'title': 'Around'}})
                                        },
                                        {
                                            key: base + 'stretch',
                                            name: '',
                                            html: bm.makeIcon('214', {attributes: {'title': 'Stretch'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_wrap: {
                            type: 'custom',
                            name: 'flex_wrap',
                            action: 'none',
                            control: getDevicesControlFactory('flex.wrap', 'Wrap', function (size, size_token) {
                                var base = 'flex' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Wrap', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'nowrap',
                                            name: '',
                                            html: bm.makeIcon('203', {attributes: {'title': 'No wrap'}})
                                        },
                                        {
                                            key: base + 'wrap',
                                            name: '',
                                            html: bm.makeIcon('204', {attributes: {'title': 'Wrap'}})
                                        },
                                        {
                                            key: base + 'wrap-reverse',
                                            name: '',
                                            html: bm.makeIcon('202', {attributes: {'title': 'Wrap reverse'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        }
                    }
                },
                bsflexChild: {
                    name: "Flex child",
                    on_section_shown: onSectionWithDevicesControlsShown,
                    fields: {
                        flex_grow: {
                            type: 'custom',
                            name: 'flex_grow',
                            action: 'none',
                            control: getDevicesControlFactory('flex.grow', 'Grow', function (size, size_token) {
                                var base = 'flex' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Grow', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {key: base + 'grow-0', name: '0'},
                                        {key: base + 'grow-1', name: '1'}
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_shrink: {
                            type: 'custom',
                            name: 'flex_shrink',
                            action: 'none',
                            control: getDevicesControlFactory('flex.shrink', 'Shrink', function (size, size_token) {
                                var base = 'flex' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Shrink', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {key: base + 'shrink-0', name: '0'},
                                        {key: base + 'shrink-1', name: '1'}
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_align_self: {
                            type: 'custom',
                            name: 'flex_align_self',
                            action: 'none',
                            control: getDevicesControlFactory('flex.align.self', 'Align self', function (size, size_token) {
                                var base = 'align-self' + size_token + '-';
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Align self', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: true,
                                    options: [
                                        {
                                            key: base + 'start',
                                            name: '',
                                            html: bm.makeIcon('211', {attributes: {'title': 'Start'}})
                                        },
                                        {
                                            key: base + 'end',
                                            name: '',
                                            html: bm.makeIcon('208', {attributes: {'title': 'End'}})
                                        },
                                        {
                                            key: base + 'center',
                                            name: '',
                                            html: bm.makeIcon('213', {attributes: {'title': 'Center'}})
                                        },
                                        {
                                            key: base + 'baseline',
                                            name: '',
                                            html: bm.makeIcon('212', {attributes: {'title': 'Baseline'}})
                                        },
                                        {
                                            key: base + 'stretch',
                                            name: '',
                                            html: bm.makeIcon('207', {attributes: {'title': 'Stretch'}})
                                        }
                                    ]
                                }
                                return field;
                            })
                        },
                        flex_order: {
                            type: 'custom',
                            name: 'flex_order',
                            action: 'none',
                            control: getDevicesControlFactory('flex.order', 'Order', function (size, size_token) {
                                var base = 'order' + size_token + '-';

                                var options = [];
                                options.push({key: base + 'first', name: 'First'});

                                for (var i = 0; i <= 5; i++) {
                                    options.push({key: base + i, name: i});
                                }

                                options.push({key: base + 'last', name: 'Last'});
                                var field = {
                                    type: 'select',
                                    name: makeFieldLabelForSize('Order', size),
                                    action: 'apply_class',
                                    show_empty: true,
                                    toggle_buttons: false,
                                    options: options
                                }
                                return field;
                            })
                        },
                        flex_fill: {
                            type: 'custom',
                            name: 'flex_fill',
                            action: 'none',
                            control: getDevicesControlFactory('flex.fill', 'Fill', function (size, size_token) {
                                var base = 'flex' + size_token + '-';
                                var field = {
                                    type: 'checkbox',
                                    name: makeFieldLabelForSize('Fill', size),
                                    action: 'apply_class',
                                    value: base + 'fill'
                                }
                                return field;
                            })
                        }
                    }
                },
                bstooltip: bsTooltipsSection,
                bspopover: bsPopoverSection,
                bsScrollSpy: bsScrollSpySection,
                bsjs: {
                    name: 'Toggle',
                    fields: {
                        bsdatatarget: bsDataTarget,
                        bsdatatoggle: bsDataToggle,
                        bsdatadisplay: bsDataDisplay
                    }
                },
            })
        });
        f.addComponentType(def_all);

        var def_placeholder = new PgComponentType(prefix + 'container', 'Container', {
            'selector' : '.placeholder',
            'sections' : addPrefixToSectionsAndFields({
                bsplaceholder : {
                    name : "Placeholder options",
                    fields : {
                        bsplaceholdersize : {
                            'type': 'select',
                            'name': 'Placeholder size',
                            'action': 'apply_class',
                            'show_empty': true,
                            options: [
                                { key: 'placeholder-xs', name: 'Wave' },
                                { key: 'placeholder-lg', name: 'Wave' },
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_placeholder);

        //* Grid section

        // Container
        var def_container = new PgComponentType(prefix + 'container', 'Container', {
            'selector' : '.container,.container-fluid,.container-sm,.container-md,.container-lg,.container-xl,.container-xxl',
            tags: 'major',
            preview: getGridPreview('container'),
            'code' : '<div class="container"></div>',
            empty_placeholder : true,
            action_menu: {
                add: [prefix + 'row']
            },
            'inline_edit' : true,
            'sections' : addPrefixToSectionsAndFields({
                containeropt : {
                    name : "Container",
                    fields : {
                        cosize : {
                            name: 'Size',
                            type: 'select',
                            action: "apply_class",
                            options: [
                                { 'key': 'container', 'name': 'Normal' },
                                { 'key': 'container-sm', 'name': 'Fluid up to SM' },
                                { 'key': 'container-md', 'name': 'Fluid up to MD' },
                                { 'key': 'container-lg', 'name': 'Fluid up to LG' },
                                { 'key': 'container-xl', 'name': 'Fluid up to XL' },
                                { 'key': 'container-xxl', 'name': 'Fluid up to XXL' },
                                { 'key': 'container-fluid',  'name': 'Fluid' }
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_container);

        // Row
        var def_row = new PgComponentType(prefix + 'row', 'Row', {
            'selector' : 'div.row',
            tags: 'major',
            preview: getGridPreview('row'),
            invalid_drop_msg : "Row can only be dropped into a Container. Drop it to the tree if you want to put it somewhere else.",
            'code' : '<div class="row"></div>',
            empty_placeholder : true,
            action_menu : {
                add: [prefix + 'column']
            },
            'sections' : addPrefixToSectionsAndFields({
                rowcols: {
                    name : "Row layout",
                    on_section_shown : onSectionWithDevicesControlsShown,
                    fields: {
                        rowcolsoptions : {
                            type: 'custom',
                            name: 'Number of columns',
                            action: 'none',
                            control: getDevicesControlFactory('row-cols', 'Row Columns', function (size, size_token) {
                                var base = 'row-cols' + size_token + '-';

                                var field = {
                                    'type' : 'select',
                                    'name' : makeFieldLabelForSize('Row Columns', size),
                                    'action' : 'apply_class',
                                    'show_empty' : true,
                                    'options' : [
                                        { 'key' : base + 'auto', 'name' : 'Auto' },
                                        { 'key' : base + '1', 'name' : '1' },
                                        { 'key' : base + '2', 'name' : '2' },
                                        { 'key' : base + '3', 'name' : '3' },
                                        { 'key' : base + '4', 'name' : '4' },
                                        { 'key' : base + '5', 'name' : '5' },
                                        { 'key' : base + '6', 'name' : '6' },
                                    ]
                                }

                                return field;
                            })
                        },
                    }
                },
            })
        });
        f.addComponentType(def_row);

        var def_column = new PgComponentType(prefix + 'column', 'Column', {
            tags: 'major',
            selector : isColumn,
            parent_selector: '.row',
            code : '<div class="col-md-4"><h3>Column title</h3>\
                <p>Cras justo odio, dapibus ac facilisis in, egestas eget quam. Donec id elit non mi porta gravida at eget metus. Nullam id dolor id nibh ultricies vehicula ut id elit.</p>\
            </div>',
            empty_placeholder : true,
            name : 'Column',
            sections : addPrefixToSectionsAndFields({
                columns : columns_section_def
            })
        });
        f.addComponentType(def_column);


        // Column clear
        var def_clear_columns = new PgComponentType(prefix + 'clear-columns', 'Clear columns', {
            'tags' : 'major',
            'priority' : 100,
            'selector' : '',
            'selector_new' : function(pgel) {
                return pgel.hasClass('w-100') && pgel.parent && pgel.parent.hasClass('row');
            },
            'code' : '<div class="w-100"></div>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_clear_columns);


        // Grid section
        var grid_section = new PgFrameworkLibSection(prefix + 'grid', 'Grid');
        grid_section.setComponentTypes([def_container, def_row, def_column, def_clear_columns]);
        addLibSection(grid_section);

        //* Text & Images
        var textAndImageSection = [];

        var hoptions = [
            {key: 'h1', name: 'Heading 1'},
            {key: 'h2', name: 'Heading 2'},
            {key: 'h3', name: 'Heading 3'},
            {key: 'h4', name: 'Heading 4'},
            {key: 'h5', name: 'Heading 5'},
            {key: 'h6', name: 'Heading 6'}
        ];

        var headings = ['h1', 'h2', 'h3', 'h4', 'h5', 'h6'];
        $.each(headings, function(i,h) {
            var def_heading = new PgComponentType(prefix + h, h, {
                'selector' : h,
                'code' : '<' + h + '>Heading ' + (i+1) + '</' + h + '>',
                action_menu: {
                    add: [prefix + 'small']
                },
                'sections' : addPrefixToSectionsAndFields({
                    'headstyle' : {
                        name : 'Heading',
                        fields : {
                            headlevel : {
                                'type' : 'select',
                                'name' : 'Level',
                                'action' : 'custom',
                                'show_empty' : false,
                                options: hoptions,
                                get_value: function(pgel) {
                                    for(var i = 0; i < hoptions.length; i++) {
                                        if(pgel.tagName == hoptions[i].key) {
                                            return hoptions[i].key;
                                        }
                                    }
                                    return null;
                                },
                                set_value: function(pgel, value, values, oldValue, eventType) {
                                    var newpgel = pgel.replaceTag(value);
                                    //$.fn.crsa('setSelectElementOnUpdate', newpgel);todo30
                                    return value;
                                }
                            },
                        }
                    }
                })
            });
            f.addComponentType(def_heading);

            textAndImageSection.push(def_heading);
        });

        var def_secondary_text = new PgComponentType(prefix + 'small', 'Small', {
            'selector' : 'small',
            'code' : '<small>Secondary text</small>',
            'preview' : '<h1><small>Secondary text</small></h1>',
            'sections' : addPrefixToSectionsAndFields({})
        })
        f.addComponentType(def_secondary_text);
        textAndImageSection.push(def_secondary_text);

        var def_paragraph = new PgComponentType(prefix + 'p', 'p', {
            'selector' : 'p',
            'code' : '<p>Paragraph</p>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_paragraph);
        textAndImageSection.push(def_paragraph);

        var def_abbreviation = new PgComponentType(prefix + 'abbr', 'Abbreviation', {
            'selector' : 'abbr',
            'code' : '<abbr title="Description">term</abbr>',
            'sections' : addPrefixToSectionsAndFields({
                'abbreviationstyle' : {
                    name : 'Abbreviation',
                    fields : {
                        abbreviationtext : {
                            'type' : 'text',
                            'name' : 'Text',
                            'action' : 'element_html'
                        },
                        abbreviationdescription : {
                            'type' : 'text',
                            'name' : 'Description',
                            'action' : 'element_attribute',
                            'attribute' : 'title'
                        },
                        abbreviationinitialism : {
                            'type' : 'checkbox',
                            'name' : 'All caps',
                            'action' : 'apply_class',
                            'value' : 'initialism'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_abbreviation);
        textAndImageSection.push(def_abbreviation);

        // Blockquote
        var def_blockquote = new PgComponentType(prefix + 'blockquote', 'Blockquote', {
            'selector_new' : function(pgel) {
                if (pgel.tagName == 'blockquote' || pgel.hasClass('blockquote')) return true;
                return false;
            },
            tags: 'major',
            'code' : '<figure>\
                <blockquote class="blockquote">\
                    <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Integer posuere erat a ante.</p>\
                </blockquote>\
                <figcaption class="blockquote-footer">\
                    Someone famous in <cite title="Source Title">Source Title</cite>\
                </figcaption>\
            </figure>',
            'sections' : addPrefixToSectionsAndFields({
                'blockquotedata' : {
                    name: "Blockquote Info",
                    fields: {
                        blockquotetext: {
                            type: 'text',
                            name: 'Quote',
                            action: 'custom',
                            get_value: function (pgel) {
                                var p = pgel.findOne('p');
                                if (!p) return null;
                                return p.html();
                            },
                            set_value: function (pgel, value, values, oldValue) {
                                var p = pgel.findOne('p');
                                if(!p) {
                                    p = pgCreate('<p></p>');
                                    pgel.prepend(p);
                                }
                                p.html(value);
                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_blockquote);
        textAndImageSection.push(def_blockquote);

        var def_blockquote_footer = new PgComponentType(prefix + 'blockquote-footer', 'Blockquote footer', {
            'selector': '.blockquote-footer',
            tags: 'major',
            code: '<figcaption class="blockquote-footer">\
                Someone famous in <cite title="Source Title">Source Title</cite>\
            </figcaption>',
            'sections' : addPrefixToSectionsAndFields({
                'blockquotedatafooter' : {
                    name: "Blockquote footer Info",
                    fields: {
                        blockquotecitation: {
                            type: 'text',
                            name: 'Citation',
                            action: 'custom',
                            get_value: function (pgel) {
                                var s = pgel.html();
                                s = s.trim().replace(/\n+\s*/g, ' ');
                                s = s.replace(/<cite.*<\/cite>/i, '@SOURCE');
                                return s;
                            },
                            set_value: function (pgel, value, values, oldValue) {
                                var source = values[prefix + 'blockquotesource'];
                                if (source && source.length > 0) {
                                    source = '<cite title="' + source + '">' + source + '</cite>';
                                } else {
                                    source = '';
                                }
                                if ((!oldValue || oldValue.length == 0) && (source && source.length > 0)) value += " @SOURCE";
                                var text = value.replace(/@source/i, source);

                                pgel.html(text);
                                return value;
                            }
                        },
                        blockquotesource: {
                            type: 'text',
                            name: 'Source',
                            action: 'custom',
                            get_value: function (pgel) {
                                var pgcite = pgel.findOne('cite');
                                if (!pgcite) return null;
                                return pgcite.text();
                            },
                            set_value: function (pgel, value, values) {
                                var citation = values.citation;
                                if (value && value.length > 0) {
                                    if (!citation) {
                                        citation = '@SOURCE';
                                    } else {
                                        if (!citation.match(/@source/i)) {
                                            citation += ' @SOURCE';
                                        }
                                    }
                                }

                                var pgcite = pgel.findOne('cite');
                                if (!pgcite) {
                                    pgcite = pgCreateNodeFromHtml("<cite title='Source Title'></cite");
                                    var pgbqfooter = pgel.findOne('.blockquote-footer');
                                    if (!pgbqfooter) {
                                        pgbqfooter = pgCreateNodeFromHtml("<div class='blockquote-footer'>" + value + "</div>");
                                        pgel.append(pgbqfooter);
                                    }

                                    pgbqfooter.append(pgcite);
                                }

                                pgcite.html(value);

                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_blockquote_footer);
        textAndImageSection.push(def_blockquote_footer);

        var def_img = new PgComponentType(prefix + 'img', 'Image', {
            'selector' : 'img,svg',
            'code' : '<img src="' + getPlaceholderImage() + '">',
            big_preview: true,
            'sections' : addPrefixToSectionsAndFields({
                image : {
                    name : 'Image options',
                    fields : {
                        imgfluid : {
                            type: 'checkbox',
                            name: 'Fluid',
                            action: 'apply_class',
                            value: 'img-fluid'
                        },
                        imgthumbnail : {
                            type: 'checkbox',
                            name: 'Thumbnail',
                            action: 'apply_class',
                            value: 'img-thumbnail'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_img);
        textAndImageSection.push(def_img);

        var def_a_link = new PgComponentType(prefix + 'a', 'Link', {
            'selector' : function() { return false; },
            'code' : '<a href="#">Link</a>',
        });
        f.addComponentType(def_a_link);
        textAndImageSection.push(def_a_link);

        var def_div = new PgComponentType(prefix + 'div', 'div', {
            'selector' : 'div',
            'code' : '<div></div>',
            empty_placeholder: true,
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_div);
        textAndImageSection.push(def_div);

        //* Text & Images
        var text_image_section = new PgFrameworkLibSection(prefix + 'textAndImage', 'Text &amp; Images');
        text_image_section.setComponentTypes(textAndImageSection);
        addLibSection(text_image_section);

        //* Buttons
        // Button
        var getButtonSizeToggleButton = function(size) {
            return bm.makeText(size);
        }
        // Button
        var def_button = new PgComponentType(prefix + 'button', 'Button', {
            'selector' : '.btn',
            'code' : '<button type="button" class="btn btn-light">Label</button>',
            'content_selector' : null,
            'drag_code' : null, //same as panel code
            'allowed_in' : null, //everywhere
            'sections' : addPrefixToSectionsAndFields({
                'button' : {
                    name : 'Button',
                    default_closed: false,
                    fields : {
                        'buttonlabel' : {
                            type: 'text',
                            name: 'Label',
                            action: 'element_html',
                            update_tree: true
                        },
                        'buttontag' : {
                            'type' : 'select',
                            'name' : 'Element',
                            'action' : 'custom',
                            toggle_buttons: true,
                            'options' : [
                                {
                                    'key' : 'a',
                                    'name' : 'Link - a tag',
                                    html: bm.makeText('&lt;a&gt;')
                                },
                                {
                                    'key' : 'button',
                                    'name' : 'Button - button tag',
                                    html: bm.makeText('&lt;button&gt;')
                                }
                            ],
                            show_empty: false,
                            get_value: function(pgel) {
                                return pgel.tagName == 'button' ? 'button' : 'a';
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                var tag = value;
                                pgel = pgel.replaceTag(tag);

                                if(value == 'a') {
                                    pgel.removeAttr('type');
                                    pgel.attr('href', pgel.getData('remember-href') || '#');
                                } else {
                                    pgel.attr('type', 'button');
                                    pgel.setData('remember-href', pgel.attr('href'));
                                    pgel.removeAttr('href');
                                }
                                return value;
                            }
                        },
                        'buttontype' : {
                            'type' : 'select',
                            'name' : 'Type',
                            'action' : 'apply_class',
                            show_empty: true,
                            toggle_buttons: true,
                            'options' : [
                                { 'key' : 'btn-light',     'name' : 'Light',     html: getButtonTypeToggleButton('light') },
                                { 'key' : 'btn-dark',      'name' : 'Dark',      html: getButtonTypeToggleButton('dark') },
                                { 'key' : 'btn-primary',   'name' : 'Primary',   html: getButtonTypeToggleButton('primary') },
                                { 'key' : 'btn-secondary', 'name' : 'Secondary', html: getButtonTypeToggleButton('secondary') },
                                { 'key' : 'btn-info',      'name' : 'Info',      html: getButtonTypeToggleButton('info') },
                                { 'key' : 'btn-success',   'name' : 'Success',   html: getButtonTypeToggleButton('success') },
                                { 'key' : 'btn-warning',   'name' : 'Warning',   html: getButtonTypeToggleButton('warning') },
                                { 'key' : 'btn-danger',    'name' : 'Danger',    html: getButtonTypeToggleButton('danger') },
                                { 'key' : 'btn-link',      'name' : 'Link',      html: getButtonTypeToggleButton('link') }
                            ]
                        },
                        'buttonoutlinetype' : {
                            'type' : 'select',
                            'name' : 'Outline',
                            'action' : 'apply_class',
                            show_empty: true,
                            toggle_buttons: true,
                            'options' : [
                                {
                                    'key' : 'btn-outline-light',
                                    'name' : 'Light',
                                    html: getButtonTypeToggleButton('light')
                                },
                                {
                                    'key' : 'btn-outline-dark',
                                    'name' : 'Dark',
                                    html: getButtonTypeToggleButton('dark')
                                },
                                {
                                    'key' : 'btn-outline-primary',
                                    'name' : 'Primary',
                                    html: getButtonTypeToggleButton('primary')
                                },
                                {
                                    'key' : 'btn-outline-secondary',
                                    'name' : 'Secondary',
                                    html: getButtonTypeToggleButton('secondary')
                                },
                                {
                                    'key' : 'btn-outline-info',
                                    'name' : 'Info',
                                    html: getButtonTypeToggleButton('info')
                                },
                                {
                                    'key' : 'btn-outline-success',
                                    'name' : 'Success',
                                    html: getButtonTypeToggleButton('success')
                                },
                                {
                                    'key' : 'btn-outline-warning',
                                    'name' : 'Warning',
                                    html: getButtonTypeToggleButton('warning')
                                },
                                {
                                    'key' : 'btn-outline-danger',
                                    'name' : 'Danger',
                                    html: getButtonTypeToggleButton('danger')
                                }
                            ]
                        },
                        'buttonsize' : {
                            'type' : 'select',
                            'name' : 'Size',
                            'action' : 'apply_class',
                            'show_empty' : true,
                            toggle_buttons: true,
                            'options' : [
                                {
                                    'key' : 'btn-lg',
                                    'name' : 'Large',
                                    html: getButtonSizeToggleButton('lg')
                                },
                                {
                                    'key' : 'btn-sm',
                                    'name' : 'Small',
                                    html: getButtonSizeToggleButton('sm')
                                }
                            ]
                        },
                        'buttonactive' : {
                            type: 'checkbox',
                            name: 'Active',
                            action: 'apply_class',
                            value: 'active'
                        },
                        'buttondisabled': {
                            type: 'checkbox',
                            name: 'Disabled',
                            action: 'custom',
                            value: '1',
                            get_value: function(pgel) {
                                return (pgel.attr('disabled') != null || pgel.hasClass('disabled')) ? '1' : null;
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                if(pgel.tagName == 'button') {
                                    if(value) {
                                        pgel.attr('disabled', 'disabled');
                                    } else {
                                        pgel.removeAttr('disabled');
                                    }
                                } else {
                                    if(value) {
                                        pgel.addClass('disabled');
                                    } else {
                                        pgel.removeClass('disabled');
                                    }
                                }
                                return value;
                            }
                        }
                    }
                },
                css : {
                    //inherit this from html framework
                    inherit: true
                }
            })
        });
        f.addComponentType(def_button);

        // group button
        var def_button_group = new PgComponentType(prefix + 'button-group', 'Button group', {
            'selector' : 'div.btn-group, div.btn-group-vertical',
            'code' : '<div class="btn-group" role="group" aria-label="Basic example">\
                <button type="button" class="btn btn-secondary">Left</button>\
                <button type="button" class="btn btn-secondary">Middle</button>\
                <button type="button" class="btn btn-secondary">Right</button>\
            </div>',
            tags: 'major',
            action_menu: {
                'add' : [prefix + 'button', prefix + 'button-dropdown', prefix + 'button-group'],
            },
            'sections' : addPrefixToSectionsAndFields({
                'buttongroup' : {
                    name : 'Button group',
                    fields : {
                        buttongroupsize : {
                            type: 'select',
                            action: 'apply_class',
                            show_empty: true,
                            name: 'Size',
                            options: [
                                {key: 'btn-group-lg', name: "Large"},
                                {key: 'btn-group-sm', name: "Small"}
                            ]
                        },
                        buttongrouporientation : {
                            name: "Orientation",
                            show_empty: false,
                            type: 'select',
                            action: 'apply_class',
                            options: [
                                {key: 'btn-group', name: "Horizontal"},
                                {key: 'btn-group-vertical', name: "Vertical"}
                            ]
                        },
                    }
                }
            })
        });
        f.addComponentType(def_button_group);

        // group checkbox button
        var def_button_group_checkbox = new PgComponentType(prefix + 'button-group-checkbox', 'Checkboxes Button group', {
            'selector' : null,
            'code' : function() {
                var id1 = getUniqueId('groupcheckbox', null, 0);
                var id2 = getUniqueId('groupcheckbox', null, id1);
                var id3 = getUniqueId('groupcheckbox', null, id2);

                return '<div class="btn-group" role="group" aria-label="Basic checkbox toggle button group">\
                    <input type="checkbox" class="btn-check" id="' + id1 + '" autocomplete="off">\
                    <label class="btn btn-outline-primary" for="' + id1 + '">Checkbox 1</label>\
                    <input type="checkbox" class="btn-check" id="' + id2 + '" autocomplete="off">\
                    <label class="btn btn-outline-primary" for="' + id2 + '">Checkbox 2</label>\
                    <input type="checkbox" class="btn-check" id="' + id3 + '" autocomplete="off">\
                    <label class="btn btn-outline-primary" for="' + id3 + '">Checkbox 3</label>\
                </div>';
            }
        });
        f.addComponentType(def_button_group_checkbox);

        // group radio button
        var def_button_group_radio = new PgComponentType(prefix + 'button-group-radio', 'Radio Button group', {
            'selector' : null,
            'code' : function() {
                var id1 = getUniqueId('groupradio', null, 0);
                var id2 = getUniqueId('groupradio', null, id1);
                var id3 = getUniqueId('groupradio', null, id2);

                return '<div class="btn-group" role="group" aria-label="Basic radio toggle button group">\
                    <input type="radio" class="btn-check" name="btnradio" id="' + id1 + '" autocomplete="off" checked>\
                    <label class="btn btn-outline-primary" for="' + id1 + '">Radio 1</label>\
                    <input type="radio" class="btn-check" name="btnradio" id="' + id2 + '" autocomplete="off">\
                    <label class="btn btn-outline-primary" for="' + id2 + '">Radio 2</label>\
                    <input type="radio" class="btn-check" name="btnradio" id="' + id3 + '" autocomplete="off">\
                    <label class="btn btn-outline-primary" for="' + id3 + '">Radio 3</label>\
                </div>';
            }
        });
        f.addComponentType(def_button_group_radio);

        // button toolbar
        var def_button_toolbar = new PgComponentType(prefix + 'button-toolbar', 'Button toolbar', {
            'selector' : 'div.btn-toolbar',
            'code' : '<div class="btn-toolbar" role="toolbar" aria-label="Toolbar with button groups">\
                <div class="btn-group me-2" role="group" aria-label="group">\
                    <button type="button" class="btn btn-primary">1</button>\
                    <button type="button" class="btn btn-primary">2</button>\
                    <button type="button" class="btn btn-primary">3</button>\
                    <button type="button" class="btn btn-primary">4</button>\
                </div>\
            </div>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_button_toolbar);

        // close button
        var def_close_button = new PgComponentType(prefix + 'close-button', 'Close button', {
            'selector' : 'button.btn-close',
            'code' : '<button type="button" class="btn-close" aria-label="Close"></button>',
            'sections' : addPrefixToSectionsAndFields({
                'closebutton' : {
                    name : 'Close button',
                    default_closed: false,
                    fields : {
                        closebtndisabled: {
                            type: 'checkbox',
                            name: 'Disabled',
                            action: 'element_attribute',
                            attribute: 'disabled'
                        },
                        closebtnwhite: {
                            type: 'checkbox',
                            name: 'White',
                            action: 'apply_class',
                            value: 'btn-close-white'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_close_button);

        var def_button_dropdown = new PgComponentType(prefix + 'button-dropdown', 'Button dropdown', {
            'selector_new' : function(pgel) {
                if ((pgel.hasClass('dropdown') || pgel.hasClass('btn-group')) && pgel.find('> .dropdown-menu').length > 0) return true;
                return false;
            },
            'code' : function() {
                var dpId = getUniqueId('dropdown', null, 0);

                return '<div class="dropdown">\
                    <button class="btn btn-secondary dropdown-toggle" type="button" id="' + dpId + '" data-bs-toggle="dropdown" aria-expanded="false">\
                        Dropdown button\
                    </button>\
                    <ul class="dropdown-menu" aria-labelledby="' + dpId + '">\
                        <li><a class="dropdown-item" href="#">Action</a></li>\
                        <li><a class="dropdown-item" href="#">Another action</a></li>\
                        <li><a class="dropdown-item" href="#">Something else here</a></li>\
                    </ul>\
                </div>'
            },
            tags: 'major',
            'action_menu' : {
                'add' : [prefix + 'dropdown-item', prefix + 'dropdown-toggle'],
                'on_add' : function (pgel, pgnew, newdef, prepend) {
                    if (newdef.type == prefix + 'dropdown-item') {
                        var pgul = pgel.findOne('> .dropdown-menu');
                        if (!pgul) {
                            pgul = pgCreateNodeFromHtml('<div class="dropdown-menu" role="menu"></div>');
                            pgel.append(pgul);
                        }

                        if (prepend) {
                            pgul.prepend(pgnew);
                        } else {
                            pgul.append(pgnew);
                        }
                    } else {
                        if (prepend) {
                            pgel.prepend(pgnew);
                        } else {
                            pgel.append(pgnew);
                        }
                    }
                }
            },
            'sections' : addPrefixToSectionsAndFields({
                'buttondropdown' : {
                    name : 'Variations',
                    fields : {
                        buttondropdowndropup : {
                            name : "Position",
                            type : "select",
                            action: 'apply_class',
                            show_empty: true,
                            options: [
                                {key: 'dropup', name: 'Up'},
                                {key: 'dropend', name: 'End'},
                                {key: 'dropstart', name: 'Start'},
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_button_dropdown);

        var dropdown_item_classes = [
            {key: 'dropdown-header', name: 'Header'},
            {key: 'dropdown-divider', name: 'Divider'},
            {key: 'dropdown-item-text', name: 'Text'},
            {key: 'disabled', name: 'Disabled'}
        ];

        var def_dropdown_menu = new PgComponentType(prefix + 'dropdown-menu', 'Dropdown menu', {
            'selector': '.dropdown-menu',
            'code': '<ul class="dropdown-menu"></ul>',
            'tags': 'major',
            'sections' : addPrefixToSectionsAndFields({
                'dropdownmenu' : {
                    name : 'Dropdown menu',
                    fields : {
                        'ddmenudark': {
                            type: 'checkbox',
                            name: 'Dark menu',
                            action: 'apply_class',
                            value: 'dropdown-menu-dark'
                        },
                    }
                },
                'dropdownmenualign' : {
                    name : "Layout",
                    on_section_shown : onSectionWithDevicesControlsShown,
                    fields: {
                        dropdownmenuPos : {
                            type: 'custom',
                            name: 'menu_align',
                            action: 'none',
                            control: getDevicesControlFactory('dropdownmenu.align', 'Align', function (size, size_token) {
                                var base = 'dropdown-menu' + size_token + '-';

                                var field = {
                                    'type' : 'select',
                                    'name' : makeFieldLabelForSize('Align', size),
                                    'action' : 'apply_class',
                                    'show_empty' : true,
                                    'toggle_buttons' : true,
                                    'options' : [
                                        { 'key' : base + 'start', 'name' : 'Start', 'html' : bm.makeIcon('align-left')   },
                                        { 'key' : base + 'end',   'name' : 'End',   'html' : bm.makeIcon('align-right')  }
                                    ]
                                }

                                return field;
                            })
                        }
                    }
                },
            })
        });
        f.addComponentType(def_dropdown_menu);

        var def_dropdown_toggle = new PgComponentType(prefix + 'dropdown-toggle', 'Dropdown toggle', {
            'selector': '.dropdown-toggle',
            'code': '<button class="btn btn-secondary dropdown-toggle" type="button" data-bs-toggle="dropdown" aria-expanded="false">\
                Dropdown button\
            </button>',
            tags: 'major',
        });
        f.addComponentType(def_dropdown_toggle);

        var def_dropdown_item = new PgComponentType(prefix + 'dropdown-item', 'Dropdown item', {
            'selector_new' : function(pgel) {
                return (pgel.hasClass('dropdown-item') || pgel.hasClass('dropdown-header') || pgel.hasClass('dropdown-divider') || pgel.hasClass('dropdown-item-text')) && pgel.parent && pgel.parent.hasClass('dropdown-menu');
            },
            parent_selector: 'ul.dropdown-menu',
            'code' : '<a class="dropdown-item" href="#">Dropdown item</a>',
            'sections' : addPrefixToSectionsAndFields({
                'dropdownitem' : {
                    name : 'Dropdown item',
                    fields : {
                        dropdownitemtype : {
                            name : "Type",
                            type : "select",
                            options : dropdown_item_classes,
                            show_empty: true,
                            action: 'custom',
                            get_value: function (pgel) {
                                var val = null;
                                $.each(dropdown_item_classes, function(i, c) {
                                    if (pgel.hasClass(c.key)) {
                                        val = c.key;
                                        return false;
                                    }
                                });
                                return val;
                            },
                            set_value: function (pgel, value, values) {
                                var removeClasses = [], addClasses = [];

                                if (value == 'dropdown-header') {
                                    pgel = pgel.replaceTag('h6');
                                    pgel.removeAttr('href');
                                    removeClasses.push('dropdown-item', 'disabled', 'dropdown-divider', 'dropdown-item-text');
                                    addClasses.push(value);
                                }
                                else if (value == 'dropdown-divider') {
                                    pgel = pgel.replaceTag('div');
                                    pgel.html('');
                                    pgel.removeAttr('href');
                                    removeClasses.push('dropdown-item', 'disabled', 'dropdown-header', 'dropdown-item-text');
                                    addClasses.push(value);
                                }
                                else if (value == 'dropdown-item-text') {
                                    pgel = pgel.replaceTag('span');
                                    pgel.removeAttr('href');
                                    removeClasses.push('dropdown-item', 'disabled', 'dropdown-divider', 'dropdown-header');
                                    addClasses.push(value);
                                }
                                else if (value == 'disabled') {
                                    pgel = pgel.replaceTag('a');
                                    var hrefAttr = pgel.getAttr('href');
                                    if (!hrefAttr) pgel.setAttr('href', '#');
                                    removeClasses.push('dropdown-header', 'dropdown-divider', 'dropdown-item-text');
                                    addClasses.push(value, 'dropdown-item');
                                }
                                else {
                                    pgel = pgel.replaceTag('a');
                                    var hrefAttr = pgel.getAttr('href');
                                    if (!hrefAttr) pgel.setAttr('href', '#');
                                    removeClasses.push('dropdown-header', 'dropdown-divider', 'disabled', 'dropdown-item-text');
                                    addClasses.push('dropdown-item');
                                }

                                for (var i = 0; i < removeClasses.length; i++) {
                                    pgel.removeClass(removeClasses[i]);
                                }

                                for (var i = 0; i < addClasses.length; i++) {
                                    pgel.addClass(addClasses[i]);
                                }

                                return value;
                            }
                        },
                        dropdownitemlabel : {
                            name : "Label",
                            type : "text",
                            action: 'custom',
                            get_value: function (pgel) {
                                return pgel.html();
                            },
                            set_value: function (pgel, value, values) {
                                pgel.html(value);
                                return value;
                            }
                        },
                        dropdownitemactive : {
                            type: 'checkbox',
                            name: 'Active',
                            action: 'apply_class',
                            value: 'active'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_dropdown_item);


        // button section
        var button_section = new PgFrameworkLibSection(prefix + 'button', 'Buttons');
        button_section.setComponentTypes([def_button, def_button_group, def_button_group_checkbox, def_button_group_radio, def_button_toolbar, def_close_button, def_button_dropdown]);
        addLibSection(button_section);


        //* UI Elements
        // Badge
        var def_badge = new PgComponentType(prefix + 'badge', 'Badge', {
            'selector' : 'span.badge',
            'code' : '<span class="badge bg-primary">42</span>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_badge);

        // Alert
        var def_alert = new PgComponentType(prefix + 'alert', 'Alert', {
            'selector' : 'div.alert',
            'code' : '<div class="alert alert-success">\
                <strong>Well done!</strong> You successfully read this important alert message.\
            </div>',
            action_menu: {
                'add' : [prefix + 'alert-title', prefix + 'alert-p', prefix + 'hr'],
                actions : [
                    {label: "Style links", action: function (pgel) {
                        var pga = pgel.find('a');
                        if (pga.length > 0) {
                            for (var i = 0; i < pga.length; i++) {
                                pga[i].addClass('alert-link');
                            }
                        }
                    }}
                ]
            },
            'sections' : addPrefixToSectionsAndFields({
                'alertstyle' : {
                    name : 'Alert',
                    fields : {
                        'alertclose' : {
                            'type' : 'checkbox',
                            'name' : 'Dismissable',
                            'value' : "1",
                            'action' : 'custom',
                            get_value: function (pgel) {
                                return pgel.find("button.btn-close").length > 0 ? "1" : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var pgb = pgel.findOne('button.btn-close');

                                if (value) {
                                    pgel.addClass('alert-dismissible');
                                    if (!pgb) {
                                        pgb = pgCreateNodeFromHtml('<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>');
                                        pgel.prepend(pgb);
                                    }
                                } else {
                                    pgel.removeClass('alert-dismissible');
                                    pgb.remove();
                                }

                                return value;
                            }
                        },
                        'alertvariation' : {
                            type: 'select',
                            name: 'Type',
                            action: 'apply_class',
                            toggle_buttons: true,
                            'options' : [
                                {'key' : 'alert-light',     'name' : 'Light',     html: getButtonTypeToggleButton('light')},
                                {'key' : 'alert-dark',      'name' : 'Dark',      html: getButtonTypeToggleButton('dark')},
                                {'key' : 'alert-primary',   'name' : 'Primary',   html: getButtonTypeToggleButton('primary')},
                                {'key' : 'alert-secondary', 'name' : 'Secondary', html: getButtonTypeToggleButton('secondary')},
                                {'key' : 'alert-success',   'name' : 'Success',   html: getButtonTypeToggleButton('success')},
                                {'key' : 'alert-info',      'name' : 'Info',      html: getButtonTypeToggleButton('info')},
                                {'key' : 'alert-warning',   'name' : 'Warning',   html: getButtonTypeToggleButton('warning')},
                                {'key' : 'alert-danger',    'name' : 'Danger',    html: getButtonTypeToggleButton('danger')}
                            ]
                        },
                    }
                }
            })
        });
        f.addComponentType(def_alert);

        var def_alert_title = new PgComponentType(prefix + 'alert-title', 'Alert title', {
            'selector' : 'div.alert-heading',
            'code' : '<h4 class="alert-heading">Well done!</h4>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_alert_title);

        var def_alert_p = new PgComponentType(prefix + 'alert-p', 'Alert text', {
            'selector' : null,
            'code' : '<p>Aww yeah, you successfully read this important alert message. This example text is going to run a bit longer so that you can see how spacing within an alert works with this kind of content.</p>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_alert_p);

        var getTablePreviewCode = function(t) {
            return '<table class="table preview-' + t + '">\
                <thead>\
                    <tr>\
                        <th>#</th>\
                        <th>First Name</th>\
                        <th>Last Name</th>\
                        <th>Username</th>\
                    </tr>\
                </thead>\
                <tbody>\
                    <tr class="sel">\
                        <td>1</td>\
                        <td class="sel">Mark</td>\
                        <td>Otto</td>\
                        <td>@mdo</td>\
                    </tr>\
                    <tr>\
                        <td>2</td>\
                        <td>Jacob</td>\
                        <td>Thornton</td>\
                        <td>@fat</td>\
                    </tr>\
                    <tr>\
                        <td>3</td>\
                        <td>Larry</td>\
                        <td>the Bird</td>\
                        <td>@twitter</td>\
                    </tr>\
                </tbody>\
            </table>';
        }

        var isTableRowEmpty = function (pgtr) {
            var empty = true;
            var pgtd = pgtr.find(">td");

            for (var i = 0; i < pgtd.length; i++) {
                if ($.trim(pgtd[i].html()).length > 0) {
                    empty = false;
                    return false;
                }
            }

            return empty;
        }

        var countNonEmptyTableRows = function(pgtbody) {
            var c = 0;
            var pgrows = pgtbody.find(">tr");

            for (var i = 0; i < pgrows.length; i++) {
                if (isTableRowEmpty(pgrows[i])) c++;
            }
            return c;
        }

        var responsiveTableOptions = [
            { key: 'table-responsive',     name: 'Xs' },
            { key: 'table-responsive-sm',  name: 'Sm' },
            { key: 'table-responsive-md',  name: 'Md' },
            { key: 'table-responsive-lg',  name: 'Lg' },
            { key: 'table-responsive-xl',  name: 'Xl' },
            { key: 'table-responsive-xxl', name: 'XXl' },
        ];

        var tableContext = {
            'type' : 'select',
            'name' : 'Context',
            'action' : 'apply_class',
            'show_empty' : true,
            'toggle_buttons' : true,
            'options' : [
                {'key' : 'table-light',     'name' : 'Light',     html: getButtonTypeToggleButton('light')},
                {'key' : 'table-dark',      'name' : 'Dark',      html: getButtonTypeToggleButton('dark')},
                {'key' : 'table-primary',   'name' : 'Primary',   html: getButtonTypeToggleButton('primary')},
                {'key' : 'table-secondary', 'name' : 'Secondary', html: getButtonTypeToggleButton('secondary')},
                {'key' : 'table-success',   'name' : 'Success',   html: getButtonTypeToggleButton('success')},
                {'key' : 'table-info',      'name' : 'Info',      html: getButtonTypeToggleButton('info')},
                {'key' : 'table-warning',   'name' : 'Warning',   html: getButtonTypeToggleButton('warning')},
                {'key' : 'table-danger',    'name' : 'Danger',    html: getButtonTypeToggleButton('danger')}
            ]
        };

        var tableActive = {
            'type': 'checkbox',
            'name': 'Active',
            'action': 'apply_class',
            'value': 'table-active'
        }

        var def_table = new PgComponentType(prefix + 'table', 'Table', {
            'selector' : 'table',
            tags: 'major',
            'code' : '<table class="table">\
                <thead>\
                    <tr>\
                        <th scope="col">#</th>\
                        <th scope="col">First Name</th>\
                        <th scope="col">Last Name</th>\
                        <th scope="col">Username</th>\
                    </tr>\
                </thead>\
                <tbody>\
                    <tr>\
                        <th scope="row">1</th>\
                        <td>Mark</td>\
                        <td>Otto</td>\
                        <td>@mdo</td>\
                    </tr>\
                    <tr>\
                        <th scope="row">2</th>\
                        <td>Jacob</td>\
                        <td>Thornton</td>\
                        <td>@fat</td>\
                    </tr>\
                    <tr>\
                        <th scope="row">3</th>\
                        <td>Larry</td>\
                        <td>the Bird</td>\
                        <td>@twitter</td>\
                    </tr>\
                </tbody>\
            </table>',
            preview: getTablePreviewCode('table'),
            action_menu: {
                add: [prefix + 'table-head', prefix + 'table-body']
            },
            'sections' : addPrefixToSectionsAndFields({
                'tabledata' : {
                    name: "Table options",
                    fields: {
                        'tablecontext': tableContext,
                        'tablerows': {
                            type: 'slider',
                            name: 'Rows',
                            action: 'custom',
                            live_update: false,
                            slider_def_unit: '',
                            get_value: function (pgel) {
                                var pgtbody = pgel.findOne('>tbody');
                                var pgrows = pgtbody.find('>tr');
                                return pgrows.length;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                value = parseInt(value) || 0;
                                var pgtbody = pgel.findOne('>tbody');
                                if (!pgtbody) {
                                    pgtbody = pgCreateNodeFromHtml('<tbody></tbody>');
                                    pgel.append(pgtbody);
                                }
                                var pgrows = pgtbody.find('>tr').reverse();

                                var cols = parseInt(values[prefix + 'tablecolumns']) || 0;
                                var change = false;
                                var empty_count = countNonEmptyTableRows(pgtbody);
                                var new_value = value;

                                if (pgrows.length < value) {
                                    for (var n = 0; n < value - pgrows.length; n++) {
                                        var pgtr = pgCreateNodeFromHtml('<tr></tr>');
                                        pgtbody.append(pgtr);
                                        for (var c = 0; c < cols; c++) {
                                            pgtr.append(pgCreateNodeFromHtml('<td></td>'));
                                        }
                                    }
                                    change = true;
                                } else if (pgrows.length > value) {
                                    var remove = pgrows.length - value;

                                    for (var i = 0; i < pgrows.length; i++) {
                                        var pgtr = pgrows[i];
                                        if (true || isTableRowEmpty(pgtr)) {
                                            pgtr.remove();
                                            remove--;
                                            if(remove == 0) break;
                                        }
                                    }

                                    new_value = value + remove;
                                }

                                if (eventType == "change") {
                                    value = new_value;
                                }
                                if (value == 0) return null;
                                else            return value;
                            }
                        },
                        'tablecolumns': {
                            type: 'slider',
                            name: 'Columns',
                            action: 'custom',
                            slider_def_unit: '',
                            live_update: false,
                            get_value: function (pgel) {
                                var pgtbody = pgel.findOne('>tbody');
                                var pgrows = pgtbody.find('>tr');
                                var max_c = 0;

                                for (var i = 0; i < pgrows.length; i++) {
                                    var c = pgrows[i].find('>td').length;
                                    if (c > max_c) max_c = c;
                                }

                                var pgheadrows = pgel.find('>thead >tr');

                                for (var i = 0; i < pgheadrows.length; i++) {
                                    var c = pgheadrows[i].find('>th').length;
                                    if (c > max_c) max_c = c;
                                }

                                return max_c;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {

                                crsaWillChangeDom();

                                value = parseInt(value) || 0;

                                var pgtbody = pgel.findOne('>tbody');
                                var pgthead = pgel.findOne('>thead');

                                var pgrows = [];
                                if (pgtbody) pgrows = pgrows.concat(pgtbody.find('>tr'));
                                if (pgthead) pgrows = pgrows.concat(pgthead.find('>tr'));

                                var new_value = value;
                                if (value) {
                                    var max_c = 0;

                                    for (var i = 0; i < pgrows.length; i++) {
                                        var pgcells = pgrows[i].find('>td,>th').reverse();

                                        if (pgcells.length > value) {
                                            var remove = pgcells.length - value;

                                            for (var n = 0; n < pgcells.length; n++) {
                                                var pgtd = pgcells[n];
                                                if (true || $.trim(pgtd.html()) == '') {
                                                    //$td.remove();
                                                    pgtd.remove();
                                                    remove--;
                                                    if (remove == 0) break;
                                                }
                                            }
                                        } else if (pgcells.length < value) {
                                            for(var n = pgcells.length; n < value; n++) {
                                                var tag = pgrows[i].parent.tagName == 'thead' ? 'th' : 'td';
                                                pgrows[i].append(pgCreateNodeFromHtml('<' + tag + '></' + tag + '>'));
                                                //$('<' + tag +'/>').appendTo($(tr));
                                            }
                                        }
                                        var c = pgrows[i].find('>td,>th').length;
                                        if (c > max_c) max_c = c;
                                    }

                                    new_value = max_c;
                                }
                                if (eventType == "change") {
                                    // $.fn.crsa("setNeedsUpdate", false, $el);
                                    value = new_value;
                                }

                                if (value == 0) return null;
                                else            return value;
                            }
                        },
                        'tableresponsive' : {
                            'type' : 'select',
                            'name' : 'Responsive',
                            // 'action' : 'apply_class_multiple',
                            'action' : 'custom',
                            'show_empty' : true,
                            'multiple' : true,
                            'options' : responsiveTableOptions,
                            get_value: function (pgel) {
                                var value = [];
                                var pgp = pgel.parent;

                                for (var i = 0; i < responsiveTableOptions.length; i++) {
                                    var resOpt = responsiveTableOptions[i];
                                    if (pgp.hasClass(resOpt.key)) value.push(resOpt.key);
                                }

                                return value.join(',');
                            },
                            set_value: function (pgel, value, values, oldValue) {
                                var pgp = pgel.parent;
                                var parentClasses = pgp.getClasses();

                                var currValues = [];
                                if (value) currValues = value.split(',');

                                var hasResClass = false;
                                if (parentClasses.length != 0) {
                                    for (var i = 0; i < responsiveTableOptions.length; i++) {
                                        var resOpt = responsiveTableOptions[i];
                                        if (pgp.hasClass(resOpt.key)) {
                                            hasResClass = true;
                                            break;
                                        }
                                    }
                                }

                                if (parentClasses.length != 0 && !hasResClass) {
                                    var pgdiv = pgCreateNodeFromHtml('<div></div>');
                                    pgdiv.insertBefore(pgel);
                                    pgdiv.append(pgel);
                                }

                                pgp = pgel.parent;

                                for (var i = 0; i < responsiveTableOptions.length; i++) {
                                    var resOpt = responsiveTableOptions[i];
                                    if (currValues.indexOf(resOpt.key) > -1) {
                                        pgp.addClass(resOpt.key);
                                    }
                                    else {
                                        pgp.removeClass(resOpt.key);
                                    }
                                }

                                return value;
                            }
                        },
                        'tableborder': {
                            'type' : 'select',
                            'name' : 'Border',
                            'action' : 'apply_class',
                            'show_empty' : true,
                            'toggle_buttons' : false,
                            'options' : [
                                { key: 'table-bordered',   name: 'Bordered' },
                                { key: 'table-borderless', name: 'Borderless' }
                            ],
                        },
                        'tablestriped' : {
                            'type' : 'checkbox',
                            'name' : 'Striped',
                            'action' : 'apply_class',
                            'value' : 'table-striped'
                        },
                        'tablehoverrows' : {
                            'type' : 'checkbox',
                            'name' : 'Hover rows',
                            'action' : 'apply_class',
                            'value' : 'table-hover'
                        },
                        'tablecondensed' : {
                            'type' : 'checkbox',
                            'name' : 'Small',
                            'action' : 'apply_class',
                            'value' : 'table-sm'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_table);

        var def_table_head = new PgComponentType(prefix + 'table-head', 'Table heading', {
            'selector' : 'thead',
            'parent_selector' : 'table',
            'code' : '<thead><tr><th></th></tr></thead>',
            'preview' : getTablePreviewCode('thead'),
            'sections' : addPrefixToSectionsAndFields({
                'tablehead': {
                    name: 'Table head',
                    fields: {
                        theadcontext: tableContext,
                    }
                }
            }),
            action_menu: {
                add: [prefix + 'table-tr']
            },
        });
        f.addComponentType(def_table_head);

        var def_table_body = new PgComponentType(prefix + 'table-body', 'Table body', {
            'selector' : 'tbody',
            'parent_selector' : 'table',
            'code' : '<tbody><tr><td></td></tr></tbody>',
            'preview' : getTablePreviewCode('tbody'),
            action_menu: {
                add: [prefix + 'table-tr']
            },
            'sections' : addPrefixToSectionsAndFields({
                'tablebody': {
                    name: 'Table body',
                    fields: {
                        tbodycontext: tableContext,
                    }
                }
            })
        });
        f.addComponentType(def_table_body);

        var def_table_row = new PgComponentType(prefix + 'table-tr', 'Table row', {
            'selector' : 'tr',
            'parent_selector' : 'table,tbody,thead,tfooter',
            'code' : '<tr><td></td></tr>',
            'preview' : getTablePreviewCode('tr'),
            action_menu: {
                add: [prefix + 'table-td', prefix + 'table-th']
            },
            'sections' : addPrefixToSectionsAndFields({
                'tablerowstyle' : {
                    name : 'Table row',
                    fields : {
                        'tablerowcontext' : tableContext,
                        'tablerowactive' : tableActive
                    }
                }
            })
        });
        f.addComponentType(def_table_row);

        var def_table_head_cell = new PgComponentType(prefix + 'table-th', 'Table cell', {
            'selector' : 'th',
            'parent_selector' : 'tr',
            'code' : '<th></th>',
            'preview' : getTablePreviewCode('td'),
            'sections' : addPrefixToSectionsAndFields({
                'tableheadcellstyle' : {
                    name : 'Table cell',
                    fields : {
                        'tableheadcellcontext' : tableContext,
                        'tableheadcellactive' : tableActive
                    }
                }
            })
        });
        f.addComponentType(def_table_head_cell);

        var def_table_cell = new PgComponentType(prefix + 'table-td', 'Table cell', {
            'selector' : 'td',
            'parent_selector' : 'tr',
            'code' : '<td></td>',
            'preview' : getTablePreviewCode('td'),
            'sections' : addPrefixToSectionsAndFields({
                'tablecellstyle' : {
                    name : 'Table cell',
                    fields : {
                        'tablecellcontext' : tableContext,
                        'tablecellactive' : tableActive
                    }
                }
            })
        });
        f.addComponentType(def_table_cell);

        var def_figure = new PgComponentType(prefix + 'figure', 'Figure', {
            'tags' : 'major',
            'selector' : 'figure, .figure',
            'code' : '<figure class="figure">\
                <img src="' + getPlaceholderImage() + '" class="figure-img img-fluid rounded" alt="A generic square placeholder image with rounded corners in a figure.">\
                <figcaption class="figure-caption">A caption for the above image.</figcaption>\
            </figure>',
            action_menu: {
                add: [prefix + 'figure-img', prefix + 'figure-caption']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_figure);

        var def_figure_img = new PgComponentType(prefix + 'figure-img', 'Figure image', {
            'selector' : '.figure-img',
            'code' : '<img src="' + getPlaceholderImage() + '" class="figure-img img-fluid rounded" alt="A generic square placeholder image with rounded corners in a figure.">',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_figure_img);

        var def_figure_caption = new PgComponentType(prefix + 'figure-caption', 'Figure caption', {
            'selector' : '.figure-caption',
            'code' : '<figcaption class="figure-caption">A caption for the above image.</figcaption>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_figure_caption);

        var def_ratio = new PgComponentType(prefix + 'ratio', 'Ratio', {
            'tags' : 'major',
            'priority' : 100,
            'selector' : '.ratio',
            'code' : '<div class="ratio ratio-16x9"></div>',
            'sections' : addPrefixToSectionsAndFields({
                ratio: {
                    name : "Ratio",
                    fields : {
                        ratiodimension : {
                            name: 'Dimension',
                            type: 'select',
                            action: "apply_class",
                            options: [
                                { 'key': 'ratio-1x1', 'name': '1x1' },
                                { 'key': 'ratio-4x3', 'name': '4x3' },
                                { 'key': 'ratio-16x9', 'name': '16x9' },
                                { 'key': 'ratio-21x9', 'name': '21x9' },
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_clear_columns);

        var def_card_group = new PgComponentType(prefix + 'card-group', 'Card group', {
            'tags' : 'major',
            'selector' : 'div.card-group',
            'priority' : 100,
            big_preview: true,
            'code' : '<div class="card-group">\
                <div class="card">\
                    <img class="card-img-top" src="' + getPlaceholderImage() + '" alt="Card image cap">\
                    <div class="card-body">\
                        <h4 class="card-title">Card title</h4>\
                        <p class="card-text">This is a wider card with supporting text below as a natural lead-in to additional content. This content is a little bit longer.</p>\
                    </div>\
                    <div class="card-footer">\
                        <small class="text-muted">Last updated 3 mins ago</small>\
                    </div>\
                </div>\
                <div class="card">\
                    <img class="card-img-top" src="' + getPlaceholderImage() + '" alt="Card image cap">\
                    <div class="card-body">\
                        <h4 class="card-title">Card title</h4>\
                        <p class="card-text">This card has supporting text below as a natural lead-in to additional content.</p>\
                    </div>\
                    <div class="card-footer">\
                        <small class="text-muted">Last updated 3 mins ago</small>\
                    </div>\
                </div>\
                <div class="card">\
                    <img class="card-img-top" src="' + getPlaceholderImage() + '" alt="Card image cap">\
                    <div class="card-body">\
                        <h4 class="card-title">Card title</h4>\
                        <p class="card-text">This is a wider card with supporting text below as a natural lead-in to additional content. This card has even longer content than the first to show that equal height action.</p>\
                    </div>\
                    <div class="card-footer">\
                        <small class="text-muted">Last updated 3 mins ago</small>\
                    </div>\
                </div>\
            </div>',
            action_menu: {
                add: [prefix + 'card']
            },
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_card_group);

        var def_card = new PgComponentType(prefix + 'card', 'Card', {
            'tags' : 'major',
            'selector' : 'div.card',
            'priority' : 100,
            'code' : '<div class="card">\
                <img class="card-img-top" src="' + getPlaceholderImage() + '" alt="Card image cap">\
                <div class="card-body">\
                    <h4 class="card-title">Card title</h4>\
                    <h6 class="card-subtitle mb-2 text-muted">Card subtitle</h6>\
                    <p class="card-text">Some quick example text to build on the card title and make up the bulk of the card\'s content.</p>\
                    <a href="#" class="btn btn-primary">Go somewhere</a>\
                </div>\
            </div>',
            action_menu: {
                add: [prefix + 'card-img', prefix + 'card-header', prefix + 'card-body', prefix + 'card-img-overlay', prefix + 'card-footer', prefix + 'list-group']
            },
            'sections' : addPrefixToSectionsAndFields({
                'card' : {
                    name : 'Card',
                    fields : {
                        cardtype : {
                            type: 'select',
                            name: 'Type',
                            action: 'apply_class',
                            show_empty: true,
                            toggle_buttons: true,
                            options: bgColorOptions
                        },
                        cardborder : {
                            type: 'select',
                            name: 'Border',
                            action: 'apply_class',
                            show_empty: true,
                            toggle_buttons: true,
                            options: borderColorOptions
                        }
                    }
                }
            })
        });
        f.addComponentType(def_card);

        var cardPosOptions = [
            { key: 'card-img', name: 'Not positioned' },
            { key: 'card-img-top', name: 'Top' },
            { key: 'card-img-bottom', name: 'Bottom' },
        ]
        var def_card_img = new PgComponentType(prefix + 'card-img', 'Card image', {
            'tags' : 'major',
            'selector': 'img.card-img,img.card-img-top,img.card-img-bottom',
            'priority' : 100,
            'code' : '<img class="card-img" src="' + getPlaceholderImage() + '" alt="Card image cap">',
            'sections' : addPrefixToSectionsAndFields({
                ciopt: {
                    name: 'Card image',
                    fields: {
                        cipos: {
                            type: 'select',
                            name: 'Position',
                            action: 'custom',
                            options: cardPosOptions,
                            get_value: function (pgel) {
                                var classes = pgel.getClasses();
                                for (var i = 0; i < cardPosOptions.length; i++) {

                                    //This looks wrong
                                    var className = classes[i];
                                    //Should be
                                    var className = cardPosOptions[i].key;

                                    if (classes.indexOf(className) > -1) return className;
                                }

                                return null;
                            },
                            set_value: function (pgel, value, values, oldValue) {
                                pgel.removeClass(oldValue);
                                var pgparent = pgel.parent;

                                if (oldValue != value) {
                                    if (value == 'card-img-top') {
                                        pgparent.prepend(pgel);
                                    }
                                    else if (value == 'card-img-bottom') {
                                        pgparent.append(pgel);
                                    }
                                }

                                pgel.addClass(value);
                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_card_img);

        var def_card_header = new PgComponentType(prefix + 'card-header', 'Card header', {
            'tags' : 'major',
            'selector' : 'div.card-header',
            'priority' : 100,
            'code' : '<div class="card-header">Header</div>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_card_header);

        var def_card_body = new PgComponentType(prefix + 'card-body', 'Card body', {
            'selector' : 'div.card-body',
            'priority' : 100,
            'code' : '<div class="card-body">\
                <h4 class="card-title">Card title</h4>\
                <p class="card-text">Some quick example text to build on the card title and make up the bulk of the card\'s content.</p>\
                <a href="#" class="btn btn-primary">Go somewhere</a>\
            </div>',
            action_menu: {
                add: [prefix + 'card-title', prefix + 'card-subtitle', prefix + 'card-text', prefix + 'card-link']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_card_body);

        var def_card_img_overlay = new PgComponentType(prefix + 'card-img-overlay', 'Card image overlay', {
            'selector' : 'div.card-img-overlay',
            'priority' : 100,
            'code' : '<div class="card-img-overlay">\
                <h4 class="card-title">Card title</h4>\
                <p class="card-text">This is a wider card with supporting text below as a natural lead-in to additional content. This content is a little bit longer.</p>\
                <p class="card-text">Last updated 3 mins ago</p>\
            </div>',
            action_menu: {
                add: [prefix + 'card-title', prefix + 'card-subtitle', prefix + 'card-text', prefix + 'card-link']
            },
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_card_img_overlay);

        var def_card_body_title = new PgComponentType(prefix + 'card-title', 'Card title', {
            'selector' : '.card-title',
            'priority' : 100,
            'code' : '<h4 class="card-title">Card title</h4>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_card_body_title);

        var def_card_body_subtitle = new PgComponentType(prefix + 'card-subtitle', 'Card subtitle', {
            'selector' : '.card-subtitle',
            'priority' : 100,
            'code' : '<h6 class="card-subtitle mb-2 text-muted">Card subtitle</h6>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_card_body_subtitle);

        var def_card_body_text = new PgComponentType(prefix + 'card-text', 'Card text', {
            'selector' : 'p.card-text',
            'priority' : 100,
            'code' : '<p class="card-text">Some quick example text to build on the card title and make up the bulk of the card\'s content.</p>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_card_body_text);

        var def_card_body_link = new PgComponentType(prefix + 'card-link', 'Card link', {
            'selector' : 'a.card-link',
            'priority' : 100,
            'code' : '<a href="#" class="card-link">Card link</a>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_card_body_link);

        var def_card_footer = new PgComponentType(prefix + 'card-footer', 'Card footer', {
            'tags' : 'major',
            'selector' : 'div.card-footer',
            'priority' : 100,
            'code' : '<div class="card-footer">Footer</div>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_card_footer);

        // List group
        var def_list_group = new PgComponentType(prefix + 'list-group', 'List group', {
            'selector' : 'ul.list-group,div.list-group',
            'priority': 100,
            'code' : '<ul class="list-group">\
                <li class="list-group-item">Cras justo odio</li>\
                <li class="list-group-item">Dapibus ac facilisis in</li>\
                <li class="list-group-item">Morbi leo risus</li>\
                <li class="list-group-item">Porta ac consectetur ac</li>\
                <li class="list-group-item">Vestibulum at eros</li>\
            </ul>',
            'action_menu' : {
                'add' : [prefix + 'list-group-item']
            },
            tags: 'major',
            'sections' : addPrefixToSectionsAndFields({
                'listgroupstyle' : {
                    name : 'List group',
                    fields : {
                        'listgrouphorizontal' : {
                            'type' : 'select',
                            'name' : 'Horizontal',
                            'action' : 'apply_class',
                            'show_empty': true,
                            'options' : [
                                { key: 'list-group-horizontal',    name: 'All' },
                                { key: 'list-group-horizontal-sm', name: 'SM and up'  },
                                { key: 'list-group-horizontal-md', name: 'MD and up'  },
                                { key: 'list-group-horizontal-lg', name: 'LG and up'  },
                                { key: 'list-group-horizontal-xl', name: 'XL and up'  },
                                { key: 'list-group-horizontal-xxl', name: 'XXL and up'  },
                            ]
                        },
                        'listgrouppercent' : {
                            'type' : 'checkbox',
                            'name' : 'Linked',
                            'value' : '1',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                return pgel.tagName == 'div' ? '1' : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var dmain = value == '1' ? 'div' : 'ul';
                                var ditem = value == '1' ? 'a' : 'li';
                                if (pgel.tagName != dmain) {
                                    pgel = pgel.replaceTag(dmain);

                                    var items = pgel.children;
                                    for (var i = 0; i < items.length; i++) {
                                        var item = items[i];
                                        if (item.tagName != 'text') {
                                            item = item.replaceTag(ditem);
                                            if (ditem == 'a') {
                                                item.setAttr('href', '#');
                                            }
                                            else {
                                                item.removeAttr('href');
                                            }
                                        }
                                    }
                                }

                                return value;
                            }
                        },
                        'listgroupflush' : {
                            'type' : 'checkbox',
                            'name' : 'Flush',
                            'value' : 'list-group-flush',
                            'action' : 'apply_class'
                        },
                        'listgroupnumbered' : {
                            'type' : 'checkbox',
                            'name' : 'Numbered',
                            'value' : 'list-group-numbered',
                            'action' : 'apply_class'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_list_group);

        var list_group_fields = {
            listitem : {
                name : 'List item',
                fields : {
                    'listgroupitemactive' : {
                        'type' : 'checkbox',
                        'name' : 'Active',
                        'value' : 'active',
                        'action' : 'apply_class'
                    },
                    'listgroupitemdisabled' : {
                        'type' : 'checkbox',
                        'name' : 'Disabled',
                        'value' : 'disabled',
                        'action' : 'apply_class'
                    },
                    'listgroupitemaction' : {
                        'type' : 'checkbox',
                        'name' : 'Action',
                        'value' : 'list-group-item-action',
                        'action' : 'apply_class'
                    },
                    'listgroupitemtype' : {
                        type: 'select',
                        name: 'Type',
                        action: 'apply_class',
                        toggle_buttons: true,
                        'options' : [
                            {'key' : 'list-group-item-light',     'name' : 'Light',     html: getButtonTypeToggleButton('light')},
                            {'key' : 'list-group-item-dark',      'name' : 'Dark',      html: getButtonTypeToggleButton('dark')},
                            {'key' : 'list-group-item-primary',   'name' : 'Primary',   html: getButtonTypeToggleButton('primary')},
                            {'key' : 'list-group-item-secondary', 'name' : 'Secondary', html: getButtonTypeToggleButton('secondary')},
                            {'key' : 'list-group-item-success',   'name' : 'Success',   html: getButtonTypeToggleButton('success')},
                            {'key' : 'list-group-item-info',      'name' : 'Info',      html: getButtonTypeToggleButton('info')},
                            {'key' : 'list-group-item-warning',   'name' : 'Warning',   html: getButtonTypeToggleButton('warning')},
                            {'key' : 'list-group-item-danger',    'name' : 'Danger',    html: getButtonTypeToggleButton('danger')}
                        ]
                    }
                }
            }
        }

        var def_list_group_item = new PgComponentType(prefix + 'list-group-item', 'List group item', {
            'selector' : 'li.list-group-item',
            priority: 100,
            parent_selector: 'ul,ol',
            'code' : '<li class="list-group-item">Cras justo odio</li>',
            'action_menu' : {
                'add' : ['h4', 'p', 'img', prefix + 'badge']
            },
            'sections' : addPrefixToSectionsAndFields(list_group_fields)
        });
        f.addComponentType(def_list_group_item);

        // Progress
        var def_progress = new PgComponentType(prefix + 'progress', 'Progress', {
            'selector' : 'div.progress',
            'code' : '<div class="progress">\
                <div class="progress-bar" role="progressbar" style="width: 25%" aria-valuenow="25" aria-valuemin="0" aria-valuemax="100"></div>\
            </div>',
            'preview' : '<div class="progress">\
                <div class="progress-bar" role="progressbar" style="width: 15%" aria-valuenow="15" aria-valuemin="0" aria-valuemax="100"></div>\
                <div class="progress-bar bg-success" role="progressbar" style="width: 30%" aria-valuenow="30" aria-valuemin="0" aria-valuemax="100"></div>\
            </div>',
            tags: 'major',
            action_menu : {
                add: [prefix + 'progress-bar']
            },
            'sections' : addPrefixToSectionsAndFields({ })
        });
        f.addComponentType(def_progress);

        var def_progress_bar = new PgComponentType(prefix + 'progress-bar', 'Progress bar', {
            'selector' : 'div.progress-bar',
            parent_selector: 'div.progress',
            'code' : '<div class="progress-bar" role="progressbar" style="width: 25%" aria-valuenow="25" aria-valuemin="0" aria-valuemax="100"></div>',
            'preview' : '<div class="progress">\
                <div class="progress-bar" role="progressbar" style="width: 30%" aria-valuenow="30" aria-valuemin="0" aria-valuemax="100"></div>\
            </div>',
            'sections' : addPrefixToSectionsAndFields({
                'progressbarstyle' : {
                    name : 'Progress bar',
                    fields : {
                        'progressbarstriped' : {
                            'type' : 'checkbox',
                            'name' : 'Striped',
                            'value' : "progress-bar-striped",
                            'action' : 'apply_class'
                        },
                        'progressbaranimated' : {
                            'type' : 'checkbox',
                            'name' : 'Animated',
                            'value' : "progress-bar-animated",
                            'action' : 'apply_class'
                        },
                        'progressbarpercent' : {
                            'type' : 'text',
                            'name' : 'Completed %',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                var v = pgel.getAttr('aria-valuenow');
                                if (v !== null) return v;

                                var styles = styleToObject(pgel.getAttr('style'));

                                var w = styles.width;
                                w = w ? w.replace('%', '') : null;
                                return w;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var v = value == null ? 50 : parseInt(value);
                                if (v > 100) v = 100;
                                value = v.toString();
                                pgel.setAttr('aria-valuenow', value);
                                pgel.setAttr('style', 'width:' + value + '%');
                                var t = value + '% Complete';
                                for (var n = 0; n < bgColorOptions.length; n++) {
                                    if (pgel.hasClass(bgColorOptions[n].key)) {
                                        t += ' (' + bgColorOptions[n].name + ')';
                                        break;
                                    }
                                }

                                var pgspan = pgel.findOne('span');
                                if (pgspan) pgspan.html(t);
                                return value;
                            }
                        },
                        'progressbarlabel': {
                            'type' : 'text',
                            'name' : 'Label',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                return pgel.text();
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                pgel.html(value);
                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_progress_bar);

        var def_spinner = new PgComponentType(prefix + 'spinner', 'Spinner', {
            selector: '.spinner-border,.spinner-grow',
            code: '<div class="spinner-border" role="status">\
                <span class="visually-hidden">Loading...</span>\
            </div>',
            'sections': addPrefixToSectionsAndFields({
                'spinner': {
                    name: "Spinner",
                    fields: {
                        spinnerType: {
                            'type': 'select',
                            'name': 'Type',
                            'action': 'apply_class',
                            'options': [
                                { key: 'spinner-border', name: 'Border' },
                                { key: 'spinner-grow',   name: 'Grow' }
                            ]
                        },
                        spinnerSize: {
                            type: 'checkbox',
                            name: 'Small',
                            action: 'custom',
                            value: '1',
                            get_value: function (pgel) {
                                return pgel.hasClass('spinner-border-sm') || pgel.hasClass('spinner-grow-sm') ? '1' : false;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                if (value) {
                                    if (pgel.hasClass('spinner-border')) pgel.addClass('spinner-border-sm');
                                    if (pgel.hasClass('spinner-grow')) pgel.addClass('spinner-grow-sm');
                                }
                                else {
                                    pgel.removeClass('spinner-border-sm');
                                    pgel.removeClass('spinner-grow-sm');
                                }

                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_spinner);

        var def_toast = new PgComponentType(prefix + 'toast', 'Toast', {
            selector: '.toast',
            code: '<div class="toast" role="alert" aria-live="assertive" aria-atomic="true">\
                <div class="toast-header">\
                    <img src="https://via.placeholder.com/15/000/fff" class="rounded me-2" alt="Placeholder image">\
                    <strong class="me-auto">Bootstrap</strong>\
                    <small>11 mins ago</small>\
                    <button type="button" class="btn-close" data-bs-dismiss="toast" aria-label="Close"></button>\
                </div>\
                <div class="toast-body">\
                    Hello, world! This is a toast message.\
                </div>\
            </div>',
            'action_menu': {
                'add': [prefix + 'toast-header', prefix + 'toast-body']
            },
            'sections' : addPrefixToSectionsAndFields({
                'toast': {
                    name: "Toast",
                    fields: {
                        toastShowInPG: {
                            type: 'checkbox',
                            name: 'Show in PG',
                            action: 'custom',
                            value: '1',
                            get_value: function (pgel) {
                                var $el = pgel.get$DOMElement();
                                return $el.hasClass('show') ? '1' : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var $el = pgel.get$DOMElement();
                                if (value) {
                                    $el.addClass('show');
                                }
                                else {
                                    $el.removeClass('show');
                                }

                                return value;
                            }
                        },
                        toastautohide: {
                            type: 'checkbox',
                            name: 'Auto hide',
                            value: 'true',
                            default_value: 'true',
                            negvalue: 'false',
                            action: 'element_attribute',
                            attribute: 'data-bs-autohide'
                        },
                        toastdelay: {
                            type: 'text',
                            name: 'Delay in ms',
                            action: 'element_attribute',
                            attribute: 'data-bs-delay',
                            live_update : false,
                        }
                    }
                }
            })
        });
        f.addComponentType(def_toast);

        var def_toast_header = new PgComponentType(prefix + 'toast-header', 'Toast header', {
            selector: '.toast-header',
            code: '<div class="toast-header">\
                <img src="https://via.placeholder.com/15/000/fff" class="rounded me-2" alt="Placeholder image">\
                <strong class="me-auto">Bootstrap</strong>\
                <small>11 mins ago</small>\
                <button type="button" class="btn-close" data-bs-dismiss="toast" aria-label="Close"></button>\
            </div>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_toast_header);

        var def_toast_body = new PgComponentType(prefix + 'toast-body', 'Toast body', {
            selector: '.toast-body',
            code: '<div class="toast-body">\
                Hello, world! This is a toast message.\
            </div>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_toast_body);

        // UI Element section
        var ui_element = new PgFrameworkLibSection(prefix + 'uiElement', 'UI elements');
        ui_element.setComponentTypes([def_badge, def_alert, def_table, def_figure, def_ratio, def_card_group, def_card, def_list_group, def_progress, def_spinner, def_toast]);
        addLibSection(ui_element);

        //* Nav
        var def_nav = new PgComponentType(prefix + 'nav', 'Nav', {
            'selector' : '.nav',
            'code' : '<ul class="nav">\
                <li class="nav-item">\
                    <a class="nav-link active" href="#">Active</a>\
                </li>\
                <li class="nav-item">\
                    <a class="nav-link" href="#">Longer nav link</a>\
                </li>\
                <li class="nav-item">\
                    <a class="nav-link" href="#">Link</a>\
                </li>\
                <li class="nav-item">\
                    <a class="nav-link disabled" href="#">Disabled</a>\
                </li>\
            </ul>',
            'priority' : 100,
            'action_menu' : {
                'add' : [prefix + 'navbar-nav-item']
            },
            'sections' : addPrefixToSectionsAndFields({
                nav : {
                    name : 'Nav',
                    fields : {
                        'navtype' : {
                            'type' : 'select',
                            'name' : 'Type',
                            'action' : 'apply_class',
                            'show_empty' : true,
                            'options' : [
                                {'key' : 'nav-tabs', 'name' : 'Tabs'},
                                {'key' : 'nav-pills', 'name' : 'Pills'}
                            ]
                        },
                        'navfill' : {
                            name : 'Fill',
                            type : 'checkbox',
                            action : 'apply_class',
                            value : 'nav-fill'
                        },
                        'navjustified' : {
                            type: 'checkbox',
                            name: 'Justified',
                            action: 'apply_class',
                            value: 'nav-justified'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_nav);

        var def_nav_item = new PgComponentType(prefix + 'navbar-nav-item', 'Navbar nav item', {
            'selector' : '.nav-item',
            'code' : '<li class="nav-item">\
                <a class="nav-link" href="#">Link</a>\
            </li>',
            'priority' : 100,
            'action_menu' : {
                'add' : [prefix + 'navbar-nav-link']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_nav_item);

        var def_nav_link = new PgComponentType(prefix + 'navbar-nav-link', 'Nav link', {
            'selector' : '.nav-link',
            'parent_selector' : '.nav-item',
            'code' : '<a class="nav-link" href="#">Link</a>',
            'sections' : addPrefixToSectionsAndFields({
                navlink : {
                    name : 'Nav link Options',
                    fields : {
                        active: class_active,
                        disabled: class_disabled,
                    }
                }
            })
        });
        f.addComponentType(def_nav_link);


        //* List
        var def_list = new PgComponentType(prefix + 'list', 'List', {
            tags: 'major',
            'selector' : function(pgel) {
                return (pgel.tagName == 'ol' || pgel.tagName == 'ul') && !pgel.hasClass('list-group');
            },
            'code' : '<ul>\
                <li>Lorem ipsum dolor sit amet</li>\
                <li>Consectetur adipiscing elit</li>\
                <li>Integer molestie lorem at massa</li>\
                <li>Nulla volutpat aliquam velit\
                    <ul>\
                        <li>Phasellus iaculis neque</li>\
                        <li>Purus sodales ultricies</li>\
                    </ul>\
                </li>\
            </ul>',
            'action_menu' : {
                'add' : [prefix + 'list-item']
            },
            'sections' : addPrefixToSectionsAndFields({
                'liststyle' : {
                    name : 'List',
                    fields : {
                        'listordered' : {
                            'type' : 'checkbox',
                            'name' : 'Ordered',
                            'action' : 'custom',
                            'value' : '1',
                            get_value: function (pgel) {
                                return pgel.tagName == "ol" ? "1" : null;
                            },
                            set_value: function (pgel, value, values) {
                                var tag = value == "1" ? 'ol' : 'ul';
                                pgel = pgel.replaceTag(tag);

                                return value;
                            }
                        },
                        'liststyle' : {
                            'type' : 'select',
                            'name' : 'Style',
                            'action' : 'apply_class',
                            'show_empty': true,
                            'options': [
                                { key: "list-unstyled", name: 'Unstyled' },
                                { key: "list-inline", name: 'Inline' },
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_list);

        var def_list_item_def = new PgComponentType(prefix + 'list-item', 'List item', {
            'selector' : 'li',
            parent_selector: 'ol,ul',
            'code' : '<li>List item</li>',
            'name' : 'List Item',
            'sections' : addPrefixToSectionsAndFields({
                'liststyle' : {
                    name : 'List',
                    fields : {
                        'listiteminline' : {
                            type: 'checkbox',
                            name: 'Inline',
                            action: 'apply_class',
                            value: 'list-inline-item'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_list_item_def);

        // Description
        var def_description = new PgComponentType(prefix + 'description', 'Description list', {
            'selector' : 'dl',
            'code' : '<dl class="row">\
                <dt class="col-sm-3">Description lists</dt>\
                <dd class="col-sm-9">A description list is perfect for defining terms.</dd>\
                <dt class="col-sm-3">Euismod</dt>\
                <dd class="col-sm-9">\
                    <p>Vestibulum id ligula porta felis euismod semper eget lacinia odio sem nec elit.</p>\
                    <p>Donec id elit non mi porta gravida at eget metus.</p>\
                </dd>\
                <dt class="col-sm-3">Malesuada porta</dt>\
                <dd class="col-sm-9">Etiam porta sem malesuada magna mollis euismod.</dd>\
                <dt class="col-sm-3 text-truncate">Truncated term is truncated</dt>\
                <dd class="col-sm-9">Fusce dapibus, tellus ac cursus commodo, tortor mauris condimentum nibh, ut fermentum massa justo sit amet risus.</dd>\
                <dt class="col-sm-3">Nesting</dt>\
                <dd class="col-sm-9">\
                    <dl class="row">\
                        <dt class="col-sm-4">Nested definition list</dt>\
                        <dd class="col-sm-8">Aenean posuere, tortor sed cursus feugiat, nunc augue blandit nunc.</dd>\
                    </dl>\
                </dd>\
            </dl>',
            tags: 'major',
            action_menu: {
                add: [prefix + 'description-term', prefix + 'description-def']
            },
            'sections' : addPrefixToSectionsAndFields({ })
        });
        f.addComponentType(def_description);

        var def_description_term = new PgComponentType(prefix + 'description-term', 'Description term', {
            'selector' : 'dt',
            'parent_selector': 'dl',
            'code' : '<dt class="col-sm-3">Term</dt>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_description_term);

        var def_description_def = new PgComponentType(prefix + 'description-def', 'Description definition', {
            'selector' : 'dd',
            'parent_selector': 'dl',
            'code' : '<dd class="col-sm-9">Term definition.</dd>',
            'sections' : addPrefixToSectionsAndFields({}),
            action_menu: {
                add: [prefix + 'description']
            },
        });
        f.addComponentType(def_description_def);

        var list = new PgFrameworkLibSection(prefix + 'list', 'List');
        list.setComponentTypes([def_list, def_list_item_def, def_description]);
        addLibSection(list);

        //* Tabs
        var findTabsContentPanes = function(pgtabs) {
            var nextPg = pgtabs.next();
            if (nextPg && nextPg.isSelector('.tab-content')) {
                return nextPg;
            }
            var pga = pgtabs.findOne('a');
            if (pga) {
                if (pga.getAttr('data-bs-toggle') == 'tab') {
                    var pgp = findTabPane(pga);
                    if (pgp) {
                        var pgdiv = pgp.parent;
                        if (pgdiv.isSelector('.tab-content')) return pgdiv;
                    }
                }
            }
            return null;
        }

        var findTabSel = function(pga) {
            var sel = pga.getAttr('data-bs-target');
            if (!sel) sel = pga.getAttr('href');
            if (sel) sel = sel.replace('#','');
            if (sel && sel.length == 0) sel = null;
            return sel;
        }

        var findTabPane = function(pga) {
            var sel = findTabSel(pga);
            if (!sel) return null;
            try {
                var pgp = pga.closest('body').findOne(sel);
                return pgp;
            }
            catch(err) {}
            return null;
        }

        //* Navbar
        var def_navbar = new PgComponentType(prefix + 'navbar', 'Navbar', {
            'selector' : '.navbar',
            'code' : function(env) {
                var id = getUniqueId('navbarToggler');
                var ddId = getUniqueId('navbarDropdownMenuLink');

                return '<nav class="navbar navbar-expand-lg navbar-light bg-light">\
                    <div class="container-fluid">\
                        <a class="navbar-brand" href="#">Navbar</a>\
                        <button class="navbar-toggler" type="button" data-bs-toggle="collapse" data-bs-target="#' + id + '" aria-controls="" aria-expanded="false" aria-label="Toggle navigation">\
                            <span class="navbar-toggler-icon"></span>\
                        </button>\
                        <div class="collapse navbar-collapse" id="' + id + '">\
                            <ul class="navbar-nav me-auto mb-2 mb-lg-0">\
                                <li class="nav-item">\
                                    <a class="nav-link active" href="#">Home</a>\
                                </li>\
                                <li class="nav-item">\
                                    <a class="nav-link" href="#">Link</a>\
                                </li>\
                                <li class="nav-item">\
                                    <a class="nav-link disabled" href="#">Disabled</a>\
                                </li>\
                                <li class="nav-item dropdown">\
                                    <a class="nav-link dropdown-toggle" href="#" id="' + ddId + '" data-bs-toggle="dropdown" aria-haspopup="true" aria-expanded="false">\
                                        Dropdown link\
                                    </a>\
                                    <div class="dropdown-menu" aria-labelledby="' + ddId + '">\
                                        <a class="dropdown-item" href="#">Action</a>\
                                        <a class="dropdown-item" href="#">Another action</a>\
                                        <a class="dropdown-item" href="#">Something else here</a>\
                                    </div>\
                                </li>\
                            </ul>\
                            <form class="d-flex">\
                                <input class="form-control me-2" type="search" placeholder="Search" aria-label="Search">\
                                <button class="btn btn-outline-success" type="submit">Search</button>\
                            </form>\
                        </div>\
                    </div>\
                </nav>';
            },
            tags: 'major',
            'action_menu' : {
                'add' : [prefix + 'navbar-brand', 'link', prefix + 'navbar-collapse', prefix + 'navbar-toggler', prefix + 'form-navbar', prefix + 'navbar-list', prefix + 'navbar-nav-item', prefix + 'navbar-dropdown', prefix + 'navbar-text'],
                'on_add' : function(pgel, pgnew, newdef, prepend) {
                    if (['link', prefix + 'navbar-nav-item', prefix + 'navbar-dropdown'].indexOf(newdef.type) >= 0) {
                        var pgdiv = pgel.findOne('.navbar-collapse') || pgel;
                        var pgul = pgdiv.findOne('ul');
                        if (!pgul) {
                            var pgul = pgCreateNodeFromHtml('<ul class="navbar-nav"></ul>');
                            pgdiv.append(pgul);
                        }

                        if(prepend) {
                            pgul.prepend(pgnew);
                        } else {
                            pgul.append(pgnew);
                        }
                    } else if([prefix + 'navbar-brand', prefix + 'navbar-collapse', prefix + 'navbar-toggler'].indexOf(newdef.type) >= 0) {
                        var pgContainer = pgel.findOne('.container-fluid') || pgel;
                        if (prepend) {
                            pgContainer.prepend(pgnew);
                        } else {
                            pgContainer.append(pgnew);
                        }
                    } else if([prefix + 'form-navbar', prefix + 'navbar-list'].indexOf(newdef.type) >= 0) {
                        var pgdiv = pgel.findOne('.navbar-collapse') || pgel;
                        if (prepend) {
                            pgdiv.prepend(pgnew);
                        } else {
                            pgdiv.append(pgnew);
                        }
                    }
                }
            },
            'sections' : addPrefixToSectionsAndFields({
                navbar : {
                    name : 'Navbar style',
                    fields : {
                        navbarstyle : {
                            type: 'select',
                            name: 'Style',
                            action: "apply_class",
                            options: [
                                { key: 'navbar-light', name: 'Light' },
                                { key: 'navbar-dark',  name: 'Dark' }
                            ]
                        },
                        navbarcollapse : {
                            name : 'Collapsable',
                            'type': 'checkbox' ,
                            value: '1',
                            action: "custom",
                            get_value: function (pgel) {
                                return pgel.findOne('.navbar-collapse') ? '1' : null;
                            },
                            set_value: function (pgel, value, values, oldValue) {
                                if (value == oldValue) {
                                    return value;
                                }

                                if (value) {
                                    var id = getUniqueId('navbarToggler');
                                    var pgbutton = pgCreateNodeFromHtml('<button class="navbar-toggler" type="button" data-bs-toggle="collapse" data-bs-target="#' + id + '" aria-controls="" aria-expanded="false" aria-label="Toggle navigation">\
                                        <span class="navbar-toggler-icon"></span>\
                                    </button>');

                                    var pgnav = pgel.find('.navbar-nav');
                                    var pgform = pgel.find('.d-flex');

                                    var pgcollapse = pgCreateNodeFromHtml('<div class="collapse navbar-collapse" id="' + id + '"></div>');

                                    var pgContainer = pgel.findOne('.container-fluid');
                                    if (!pgContainer) {
                                        pgContainer = pgel.findOne('.container');
                                        if (!pgContainer) {
                                            pgContainer = pgel;
                                        }
                                    }
                                    pgContainer.append(pgbutton);
                                    pgContainer.append(pgcollapse);

                                    for (var i = 0; i < pgnav.length; i++) {
                                        pgcollapse.append(pgnav[i]);
                                    }
                                    for (var i = 0; i < pgform.length; i++) {
                                        pgcollapse.append(pgform[i]);
                                    }
                                }
                                else {
                                    var pgnavtoggle = pgel.findOne('.navbar-toggler');
                                    if (pgnavtoggle) pgnavtoggle.remove();

                                    var pgcollapse = pgel.findOne('.navbar-collapse');
                                    var pgchildren = pgcollapse.children;
                                    var pgContainer = pgel.findOne('.container-fluid');
                                    if (!pgContainer) {
                                        pgContainer = pgel.findOne('.container');
                                        if (!pgContainer) {
                                            pgContainer = pgel;
                                        }
                                    }

                                    while (pgchildren.length > 0) {
                                        pgContainer.append(pgchildren[0]);
                                    }

                                    pgcollapse.remove();
                                }

                                return value;
                            }
                        },
                        navbartype : {
                            name : 'Type',
                            type : 'select' ,
                            show_empty: true,
                            action: "apply_class",
                            options: [
                                {key: "fixed-top", name: "Fixed - top", tip: "Add padding-top to body to prevent navbar overlapping the content."},
                                {key: "fixed-bottom", name: "Fixed - bottom", tip: "Add padding-bottom to body to prevent navbar overlapping the content."},
                                {key: "sticky-top", name: "Sticky - top"}
                            ]
                        },
                        navbarexpand : {
                            name : 'Expand',
                            type : 'select',
                            show_empty: true,
                            action: "apply_class",
                            options: [
                                {key: "navbar-expand",     name: "Xs"},
                                {key: "navbar-expand-sm",  name: "Sm"},
                                {key: "navbar-expand-md",  name: "Md"},
                                {key: "navbar-expand-lg",  name: "Lg"},
                                {key: "navbar-expand-xl",  name: "Xl"},
                                {key: "navbar-expand-xxl", name: "XXl"},
                            ]
                        }
                    }
                }
            })
        });

        if(pgIsMagic()) {
            def_navbar.on_filter_tree_children = function (pgel) {
                var r = [];
                r = r.concat(pgel.find('.navbar-brand'));
                r = r.concat(pgel.find('.navbar-nav'));
                return r;
            }

            f.componentFiltersTree(def_navbar);
        }
        f.addComponentType(def_navbar);

        var def_nav_toggler = new PgComponentType(prefix + 'navbar-toggler', 'Navbar toggler', {
            'selector' : '.navbar-toggler',
            'code' : '<button class="navbar-toggler" type="button" data-bs-toggle="collapse" data-bs-target="" aria-controls="" aria-expanded="false" aria-label="Toggle navigation">\
                <span class="navbar-toggler-icon"></span>\
            </button>',
            'sections' : addPrefixToSectionsAndFields({
                navtoggler : {
                    name : 'Toggler',
                    fields : {
                        'navtogglertarget' : {
                            'type' : 'text',
                            'name' : 'Target',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                var target = pgel.getAttr('data-bs-target')
                                return target.length > 0 ? target.substr(1, target.length) : '';
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                pgel.setAttr('data-bs-target', '#' + value);
                                pgel.setAttr('aria-controls', value);

                                var pgcollapse = pgel.siblingOne('.collapse');
                                pgcollapse.setAttr('id', value);

                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_nav_toggler);

        var def_nav_collapse = new PgComponentType(prefix + 'navbar-collapse', 'Navbar collapse', {
            'selector' : '.navbar-collapse',
            'code' : '<div class="collapse navbar-collapse"></div>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_nav_collapse);

        var def_brand = new PgComponentType(prefix + 'navbar-brand', 'Navbar brand', {
            'selector' : '.navbar-brand',
            'code' : '<a class="navbar-brand" href="#">Brand</a>',
            'sections' : addPrefixToSectionsAndFields({
                brand : {
                    name : 'Link',
                    fields : {
                        'brandhref' : {
                            name : 'Url',
                            type : 'text',
                            action: 'element_attribute',
                            attribute: 'href'
                        },
                        'brandtarget' : {
                            name : 'Target',
                            type : 'text',
                            action: 'element_attribute',
                            attribute: 'target'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_brand);

        var def_navbar_form = new PgComponentType(prefix + 'form-navbar', 'Form navbar', {
            'selector' : null,
            'code' : '<form class="d-flex">\
                <input class="form-control me-2" type="search" placeholder="Search" aria-label="Search">\
                <button class="btn btn-outline-success" type="submit">Search</button>\
            </form>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_navbar_form);

        var def_navbar_text = new PgComponentType(prefix + 'navbar-text', 'Navbar text', {
            'selector' : '.navbar-text',
            'code' : '<span class="navbar-text">\
                Navbar text with an inline element\
            </span>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_navbar_text);

        var def_navbar_list = new PgComponentType(prefix + 'navbar-list', 'Navbar list', {
            'selector' : 'ul.navbar-nav',
            'code' : '<ul class="navbar-nav"></ul>',
            tags: 'major',
            properity: 101,
            'action_menu' : {
                'add' : ['link', prefix + 'navbar-dropdown'],
                'on_add' : function (pgel, pgnew, newdef, prepend) {
                    var pgli = pgCreateNodeFromHtml('<li></li>');
                    pgel.append(pgli);

                    if (prepend) {
                        pgli.prepend(pgnew);
                    } else {
                        pgli.append(pgnew);
                    }
                }
            },
            'sections' : addPrefixToSectionsAndFields({
                'navbarnav' : {
                    name : 'Navbar',
                    fields : {
                        'navbarnavscroll' : {
                            'type' : 'checkbox',
                            'name' : 'Scroll',
                            'action' : 'apply_class',
                            'value' : 'navbar-nav-scroll'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_navbar_list);

        var def_navbar_dropdown = new PgComponentType(prefix + 'navbar-dropdown', 'Navbar dropdown', {
            'selector' : null,
            'code': function (env) {
                var ddId = getUniqueId('navbarDropdownMenuLink');

                return '<li class="nav-item dropdown">\
                    <a class="nav-link dropdown-toggle" href="#" id="' + ddId + '" data-bs-toggle="dropdown" aria-haspopup="true" aria-expanded="false">\
                        Dropdown link\
                    </a>\
                    <div class="dropdown-menu" aria-labelledby="' + ddId + '">\
                        <a class="dropdown-item" href="#">Action</a>\
                        <a class="dropdown-item" href="#">Another action</a>\
                        <a class="dropdown-item" href="#">Something else here</a>\
                    </div>\
                </li>'
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_navbar_dropdown);

        //* Tabs
        var def_tabs = new PgComponentType(prefix + 'tabs', 'Tabs', {
            'selector' : 'ul.nav-tabs,ul.nav-pills',
            'code' : '<ul class="nav nav-tabs" role="tablist">\
                <li class="nav-item">\
                    <a class="nav-link active" href="#" data-bs-toggle="tab" role="tab" aria-controls="" aria-expanded="true">Home</a>\
                </li>\
                <li class="nav-item">\
                    <a class="nav-link" href="#" data-bs-toggle="tab" role="tab" aria-controls="" aria-expanded="true">Profile</a>\
                </li>\
                <li class="nav-item">\
                    <a class="nav-link" href="#" data-bs-toggle="tab" role="tab" aria-controls="" aria-expanded="true">Messages</a>\
                </li>\
                <li class="nav-item">\
                    <a class="nav-link" href="#" data-bs-toggle="tab" role="tab" aria-controls="" aria-expanded="true">Settings</a>\
                </li>\
            </ul>',
            'action_menu' : {
                'add': [prefix + 'navbar-nav-item', prefix + 'navbar-dropdown', prefix + 'tab-pane'],
                'on_add' : function(pgel, pgnew, newdef, prepend) {
                    var pgpanes = findTabsContentPanes(pgel);

                    if (pgnew.isSelector('.tab-pane')) {
                        if (!pgpanes) {
                            pgpanes = pgCreateNodeFromHtml('<div class="tab-content"></div>').insertAfter(pgel);
                        }

                        if (prepend) {
                            pgpanes.prepend(pgnew);
                        } else {
                            pgpanes.append(pgnew);
                        }
                    } else {
                        if (pgnew.tagName != 'li') {
                            pgnew = pgCreateNodeFromHtml('<li></li>').append(pgnew).addClass('dropdown');
                        }

                        if (prepend) {
                            pgel.prepend(pgnew);
                        } else {
                            pgel.append(pgnew);
                        }
                    }

                    if (pgpanes && pgnew.tagName == 'li') {
                        var pgaparent = pgnew;
                        if (newdef.type == prefix + 'navbar-dropdown') {
                            pgaparent = pgnew.findOne('.dropdown-menu');
                        }

                        var pgas = pgaparent.find('> a');
                        for (var i = 0; i < pgas.length; i++) {
                            var pga = pgas[i];
                            var sel = findTabSel(pga);
                            var pgp = findTabPane(pga);

                            if (!pgp || pgp.length == 0) {
                                if (!sel) {
                                    sel = getUniqueId('tab', null, 0);
                                    pga.setAttr('href', '#' + sel);
                                }
                                pga.setAttr('data-bs-toggle', 'tab');

                                pgnew = pgCreateNodeFromHtml('<div class="tab-pane" id="' + sel + '"><p>Tab ' + sel + ' content goes here...</p></div>');
                                if (prepend) {
                                    pgpanes.prepend(pgnew);
                                } else {
                                    pgpanes.append(pgnew);
                                }
                            }
                        }
                    }
                    else {
                        var pgli = pgCreateNodeFromHtml('<li class="nav-item"><a class="nav-link" href="#" data-bs-toggle="tab" role="tab" aria-controls="" aria-expanded="true">Link</a></li>');

                        var pga = pgli.findOne('> a');
                        var sel = getUniqueId('tab', null, 0);
                        pga.setAttr('href', '#' + sel);
                        pgnew.setAttr('id', sel);

                        if (prepend) {
                            pgel.prepend(pgli);
                        } else {
                            pgel.append(pgli);
                        }
                    }
                }
            },
            tags : 'major',
            'sections' : addPrefixToSectionsAndFields({
                'tabs' : {
                    name : 'Tabs and Pills',
                    fields : {
                        'tabswithpanes' : {
                            'type' : 'checkbox',
                            'name' : 'With panes',
                            'action' : 'custom',
                            'value' : '1',
                            get_value: function (pgel) {
                                return findTabsContentPanes(pgel) ? '1' : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var pgpanes = findTabsContentPanes(pgel);

                                if (value) {
                                    var active = pgel.find('> li').indexOf(pgel.find('> li > a.active')[0]);
                                    if (active < 0) active = 0;

                                    if (!pgpanes) {
                                        var html = pgel.data['crsa-tab-panes'];
                                        if (html) {
                                            pgpanes = pgCreateNodeFromHtml(html);
                                        } else {
                                            pgpanes = pgCreateNodeFromHtml('<div class="tab-content"></div>');
                                            var pglilink = pgel.find('>li a');

                                            for (var i = 0; i < pglilink.length; i++) {
                                                var pgtab = pglilink[i];

                                                var sel = findTabSel(pgtab);
                                                if (!sel) {
                                                    sel = getUniqueId('tab', null, i);
                                                    pgtab.setAttr('data-bs-target', '#' + sel);
                                                }
                                                pgtab.setAttr('data-bs-toggle', 'tab');

                                                var pgp = pgCreateNodeFromHtml('<div class="tab-pane" id="' + sel + '">Tab ' + sel + ' content goes here...</div>');
                                                pgpanes.append(pgp);
                                                if (i == active) {
                                                    pgp.addClass('active');
                                                }
                                            }
                                        }
                                        pgpanes.insertAfter(pgel);
                                    }
                                } else {
                                    if (pgpanes) {
                                        pgel.data['crsa-tab-panes'] = pgpanes.get$DOMElement();
                                        pgpanes.remove();
                                    }
                                }
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_tabs);

        var def_tab_content = new PgComponentType(prefix + 'tab-content', 'Tab content', {
            'selector' : '.tab-content',
            'code' : '<div class="tab-content"></div>',
            'action_menu' : {
                'add' : [prefix + 'tab-pane']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_tab_content);

        var def_tab_pane = new PgComponentType(prefix + 'tab-pane', 'Tab pane', {
            'selector' : '.tab-pane',
            parent_selector: '.tab-content',
            'code' : '<div class="tab-pane"><p>Tab content goes here...</p></div>',
            'sections' : addPrefixToSectionsAndFields({
                name: 'Tab Pane',
                fields: {
                    'bstabpanefade': {
                        type: 'checkbox',
                        name: 'Fade',
                        action: 'apply_class',
                        value: 'fade'
                    },
                    'bstabpaneactive': {
                        type: 'checkbox',
                        name: 'Active',
                        action: 'apply_class',
                        value: 'active'
                    },
                    'bstabpaneshow': {
                        type: 'checkbox',
                        name: 'Show',
                        action: 'apply_class',
                        value: 'show'
                    }
                }
            })
        });
        f.addComponentType(def_tab_pane);

        // Breadcrumb
        var def_breadcrumb = new PgComponentType(prefix + 'breadcrumb', 'Breadcrumbs', {
            'tags' : 'major',
            'selector' : 'ol.breadcrumb',
            'code' : '<ol class="breadcrumb">\
                <li class="breadcrumb-item"><a href="#">Home</a></li>\
                <li class="breadcrumb-item"><a href="#">Library</a></li>\
                <li class="breadcrumb-item active">Data</li>\
            </ol>',
            'priority' : 100,
            'action_menu' : {
                'add' : [prefix + 'breadcrumb-item']
            },
            'sections' : addPrefixToSectionsAndFields({
                'bsbreadcrumb': {
                    'name': 'Breadcrumb',
                    'fields': {
                        'bsbkdivider': {
                            'type': 'text',
                            'name': 'Divider',
                            'action': 'custom',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                pgp = pgel.parent;
                                if (pgp.is('nav')) {
                                    var style = pgp.attr('style');
                                    var styleObj = styleToObject(style);
                                    return styleObj['--bs-breadcrumb-divider'];
                                }

                                return null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                pgp = pgel.parent;
                                if (!pgp.is('nav')) {
                                    var pgp = pgCreate('<nav></nav>');
                                    pgp.insertBefore(pgel);
                                    pgp.append(pgel);
                                }

                                var style = pgp.attr('style');
                                var styleObj = styleToObject(style);
                                styleObj['--bs-breadcrumb-divider'] = value;

                                pgp.attr('style', objectToStyle(styleObj));

                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_breadcrumb);

        var def_breadcrumb_item = new PgComponentType(prefix + 'breadcrumb-item', 'Breadcrumb item', {
            'selector' : '.breadcrumb-item',
            'code' : '<li class="breadcrumb-item"><a href="#">Link</a></li>',
            'priority' : 100,
            'sections' : addPrefixToSectionsAndFields({
                'breadcrumb-item' : {
                    name : 'Breadcrumb item',
                    fields : {
                        'active' : {
                            'type' : 'checkbox',
                            'name' : 'Active',
                            'value' : '1',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                return pgel.hasClass('active') ? '1' : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                if (value) {
                                    pgel.addClass('active');
                                    pgel.html(pgel.text());
                                } else {
                                    var pga = pgCreate('<a href="#">' + pgel.text() + '</a>');
                                    pgel.text('')
                                    pgel.append(pga);
                                    pgel.removeClass('active');
                                }

                                return value;
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_breadcrumb_item);

        // Pagination
        var def_pagination = new PgComponentType(prefix + 'pagination', 'Pagination', {
            tags: 'major',
            'selector' : 'ul.pagination',
            'code' : '<ul class="pagination">\
                <li class="page-item"><a class="page-link" href="#">Previous</a></li>\
                <li class="page-item"><a class="page-link" href="#">1</a></li>\
                <li class="page-item"><a class="page-link" href="#">2</a></li>\
                <li class="page-item"><a class="page-link" href="#">3</a></li>\
                <li class="page-item"><a class="page-link" href="#">Next</a></li>\
            </ul>',
            'priority' : 100,
            'action_menu' : {
                'add' : [prefix + 'pagination-item', prefix + 'pagination-previous', prefix + 'pagination-next'],
                'on_add' : function(pgel, pgnew, newdef, prepend) {
                    if (newdef.type == prefix + 'pagination-previous') {
                        pgel.prepend(pgnew);
                    } else if (newdef.type == prefix + 'pagination-next') {
                        pgel.append(pgnew);
                    } else {
                        var max = 0;
                        var pglist = pgel.find('>li');

                        for (var i = 0; i < pglist.length; i++) {
                            var pgli = pglist[i]
                            var val = parseInt(pgli.text());
                            if (val && val > max) max = val;
                        }

                        max++;
                        pgnew.html(pgnew.html().replace('_NUM_', max));

                        if (pglist.length > 0) {
                            var pglast = pglist[pglist.length - 1];
                            var t = pglast.text();
                            if (t.match(/(&raquo;|»)/i)) {
                                pgnew.insertBefore(pglast);
                            } else {
                                pgel.append(pgnew);
                            }
                        } else {
                            pgel.append(pgnew);
                        }
                    }
                }
            },
            'sections' : addPrefixToSectionsAndFields({
                'paginationlist' : {
                    name : 'Pagination',
                    fields : {
                        paginationstacked : {
                            type: 'select',
                            name: 'Size',
                            action: 'apply_class',
                            'show_empty' : true,
                            'options' : [
                                {'key' : 'pagination-lg', 'name' : 'Large'},
                                {'key' : 'pagination-sm', 'name' : 'Small'}
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_pagination);

        var def_pagination_item = new PgComponentType(prefix + 'pagination-item', 'Pagination item', {
            'selector' : '.page-item',
            'parent_selector' : 'ul.pagination',
            'priority' : 100,
            'code' : '<li class="page-item"><a class="page-link" href="#">_NUM_</a></li>',
            'sections' : addPrefixToSectionsAndFields({
                'paginationitem' : {
                    name : 'Pagination item',
                    fields : {
                        paginationitemactive: class_active,
                        paginationitemdisabled: class_disabled
                    }
                }
            })
        });
        f.addComponentType(def_pagination_item);

        var def_pagination_link = new PgComponentType(prefix + 'pagination-link', 'Pagination link', {
            'selector' : '.page-link',
            'parent_selector' : 'li.page-item',
            'code' : '<a class="page-link" href="#">_NUM_</a>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_pagination_link);

        var def_pagination_previous = new PgComponentType(prefix + 'pagination-previous', 'Pagination previous', {
            'selector' : null,
            'parent_selector' : 'ul.pagination',
            'code' : '<li class="page-item"><a class="page-link" href="#">&laquo;</a></li>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_pagination_previous);

        var def_pagination_next = new PgComponentType(prefix + 'pagination-next', 'Pagination next', {
            'selector' : null,
            'parent_selector' : 'ul.pagination',
            'code' : '<li class="page-item"><a class="page-link" href="#">&raquo;</a></li>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_pagination_next);

        var navigation_section = new PgFrameworkLibSection(prefix + 'navigation', 'Navigation');
        navigation_section.setComponentTypes([def_navbar, def_breadcrumb, def_tabs, def_pagination]);
        addLibSection(navigation_section);

        // Form
        var attr_disabled = {
            type: 'checkbox',
            action: 'element_attribute',
            attribute: 'disabled',
            name: 'Disabled',
            value: 'disabled'
        };

        var formlabelfor = {
            'type' : 'text',
            'name' : 'For field id',
            'action' : 'element_attribute',
            'attribute' : 'for'
        };

        var control_sizing = {
            'type' : 'select',
            'name' : 'Size',
            'action' : 'apply_class',
            'show_empty' : true,
            'options' : [
                {key: 'form-control-sm', name: 'Small'},
                {key: '',                name: 'Default'},
                {key: 'form-control-lg', name: 'Large'}
            ]
        };

        var validation = {
            'type' : 'select',
            'name' : 'Validation',
            'action' : 'apply_class',
            'show_empty' : true,
            'options' : [
                {key: 'is-valid', name: 'Valid'},
                {key: 'is-invalid', name: 'Not valid'}
            ]
        };

        var readonly = {
            type: 'checkbox',
            action: 'element_attribute',
            attribute: 'readonly',
            name: 'Read Only',
            value: 'readonly'
        }

        var required = {
            type: 'checkbox',
            action: 'element_attribute',
            attribute: 'required',
            name: 'Required',
            value: 'required',
            empty_attribute: true
        };

        var def_label = new PgComponentType(prefix + 'label', 'Label', {
            'selector' : 'label',
            'code' : '<label>Label</label>',
            'sections' : addPrefixToSectionsAndFields({
                'labelstyle' : {
                    name : 'Label options',
                    fields : {
                        'labelforid' : formlabelfor
                    }
                }
            })
        });
        f.addComponentType(def_label);

        var form_check_inline = {
            type: 'checkbox',
            name: 'Inline',
            action: 'apply_class',
            value: 'form-check-inline'
        };

        var def_form_help = new PgComponentType(prefix + 'form-text', 'Help text', {
            'selector' : '.form-text',
            'code' : '<span class="form-text">A block of help text.</span>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_help);

        var def_form = new PgComponentType(prefix + 'form', 'Form', {
            tags: 'major',
            'selector' : 'form',
            'code' : function() {
                var emialInput = getUniqueId('exampleInputEmail', null, 0);
                var passowrdInput = getUniqueId('exampleInputPassword', null, 0);
                var fileInput = getUniqueId('exampleInputFile', null, 0);
                var checkInput = getUniqueId('exampleInputCheck', null, 0);

                return '<form role="form">\
                    <div class="mb-3">\
                        <label for="' + emialInput + '">Email address</label>\
                        <input type="email" class="form-control" id="' + emialInput + '" placeholder="Enter email" aria-describedby="emailHelp">\
                    </div>\
                    <div class="mb-3">\
                        <label for="' + passowrdInput + '">Password</label>\
                        <input type="password" class="form-control" id="' + passowrdInput + '" placeholder="Password">\
                        <div class="form-text">\
                            Your password must be 8-20 characters long, contain letters and numbers, and must not contain spaces, special characters, or emoji.\
                        </div>\
                    </div>\
                    <div class="mb-3">\
                        <label for="' + fileInput + '">File input</label>\
                        <input type="file" id="' + fileInput + '">\
                    </div>\
                    <div class="mb-3 form-check">\
                        <input type="checkbox" class="form-check-input" id="' + checkInput + '">\
                        <label class="form-check-label" for="' + checkInput + '">Check me out</label>\
                    </div>\
                    <button type="submit" class="btn btn-primary">Submit</button>\
                </form>'
            },
            action_menu: {
                add: [prefix + 'form-textarea-block', prefix + 'form-checkbox-block', prefix + 'form-radio-block', prefix + 'form-select-block', prefix + 'form-input-group', prefix + 'form-range-block', prefix + 'form-file-block', prefix + 'form-color-block', prefix + 'form-datalists-block', prefix + 'form-floating', prefix + 'form-fieldset']
            },
            'sections' : addPrefixToSectionsAndFields({
                'layout' : {
                    name : 'Form style',
                    fields : {
                        'formvalidated' : {
                            'type' : 'checkbox',
                            'name' : 'Validated',
                            'action' : 'apply_class',
                            'value' : 'was-validated'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_form);

        var def_form_fieldset = new PgComponentType(prefix + 'form-fieldset', 'Fieldset', {
            'selector' : 'fieldset',
            parent_selector: 'form',
            'code' : '<fieldset></fieldset>',
            action_menu: {
                add: [prefix + 'form-textarea-block', prefix + 'form-checkbox-block', prefix + 'form-radio-block', prefix + 'form-select-block', prefix + 'form-file-block', prefix + 'form-color-block']
            },
            'sections' : addPrefixToSectionsAndFields({
                'formfieldsetstyle' : {
                    name : 'Fieldset style',
                    fields : {
                        formfieldsetdisabled: attr_disabled
                    }
                }
            })
        });
        f.addComponentType(def_form_fieldset);

        var def_form_label = new PgComponentType(prefix + 'form-label', 'Label', {
            'selector' : '.form-label,.col-form-label',
            'code' : '<label class="form-label">Label</label>',
            'sections' : addPrefixToSectionsAndFields({
                'formlabelstyle' : {
                    name : 'Label options',
                    fields : {
                        'formcolsize' : {
                            'type' : 'select',
                            'name' : 'Column Size',
                            'action' : 'apply_class',
                            'show_empty': true,
                            'options' : [
                                { key: 'col-form-label-lg', name: 'Large' },
                                { key: 'col-form-label-sm', name: 'Small' }
                            ]
                        }
                    }
                }
            }),
        });
        f.addComponentType(def_form_label);

        var def_form_input = new PgComponentType(prefix + 'form-input', 'Input', {
            'selector_new' : function(pgel) {
                if(pgel.tagName == 'input' && pgel.getAttribute('type') != 'checkbox' && pgel.getAttribute('type') != 'radio') return true;
                return false;
            },
            'code' : '<input type="text" class="form-control" placeholder="Placeholder text">',
            'sections' : addPrefixToSectionsAndFields({
                'forminputdata' : {
                    name : 'Input options',
                    fields : {
                        'forminputtype' : {
                            'type' : 'select',
                            'name' : 'Type',
                            'action' : 'custom',
                            'options' : [
                                {key: 'text', name: 'Text'},
                                {key: 'password', name: 'Password'},
                                {key: 'number', name: 'Number'},
                                {key: 'email', name: 'Email'},
                                {key: 'file', name: 'File'},
                                {key: 'url', name: 'Url'},
                                {key: 'search', name: 'Search'},
                                {key: 'tel', name: 'Tel'},
                                {key: 'color', name: 'Color'},
                                {key: 'datetime', name: 'Datetime'},
                                {key: 'datetime-local', name: 'Datetime local'},
                                {key: 'date', name: 'Date'},
                                {key: 'month', name: 'Month'},
                                {key: 'time', name: 'Time'},
                                {key: 'week', name: 'Week'}
                            ],
                            get_value: function (pgel) {
                                return pgel.getAttr('type');
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                pgel.setAttr('type', value);
                                return value;
                            }
                        },
                        'forminputvalue': {
                            type: 'text',
                            name: 'Value',
                            action: 'custom',
                            attribute: 'value',
                            get_value: function (pgel) {
                                return pgel.getAttr('value');
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                pgel.setAttr('value', value);
                                return value;
                            }
                        },
                        'forminputplaceholder': {
                            type: 'text',
                            name: 'Placeholder',
                            action: 'element_attribute',
                            attribute: 'placeholder'
                        },
                        'forminputcontrolsize' : control_sizing,
                        'forminputvalidation': validation,
                        'forminputdisabled': attr_disabled,
                        'forminputreadonly': readonly,
                        'forminputrequired': required
                    }
                }
            })
        });
        f.addComponentType(def_form_input);

        var def_form_ta_block = new PgComponentType(prefix + 'form-textarea-block', 'Textarea block', {
            'selector' : null,
            parent_selector: 'form,fieldset',
            'code' : function(env) {
                var id = getUniqueId('formInput');

                return '<div class="mb-3">\
                    <label for="' + id + '" class="form-label">Example textarea</label>\
                    <textarea class="form-control" id="' + id + '" rows="3"></textarea>\
                </div>';
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_ta_block);

        var def_form_checkbox = new PgComponentType(prefix + 'form-checkbox', 'Checkbox', {
            'selector' : 'input.form-check-input[type="checkbox"]',
            'code' : '<input class="form-check-input" type="checkbox" value="option1">',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_checkbox);

        var def_form_radio_button = new PgComponentType(prefix + 'form-radio-button', 'Radio button', {
            'selector' : 'input.btn-check[type="radio"]',
            'code' : '<input type="radio" class="btn-check" autocomplete="off">',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_radio_button);

        var def_form_checkbox_button = new PgComponentType(prefix + 'form-checkbox-button', 'Checkbox button', {
            'selector' : 'input.btn-check[type="checkbox"]',
            'code' : '<input type="checkbox" class="btn-check" autocomplete="off">',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_checkbox_button);

        var def_form_check_label_button = new PgComponentType(prefix + 'form-check-label-button', 'Checkbox label button', {
            'selector' : 'label.btn',
            'code' : '<label class="btn btn-primary">Single toggle</label>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_check_label_button);

        var def_form_check_label = new PgComponentType(prefix + 'form-check-label', 'Check label', {
            'selector' : '.form-check-label',
            'code' : '<label class="form-check-label">Check me out</label>',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_check_label);

        var def_form_checkbox_block = new PgComponentType(prefix + 'form-checkbox-block', 'Checkbox block', {
            'selector_new' : function(pgel) {
                if (pgel.hasClass('form-check')) {
                    var pgRadio = pgel.findOne('input[type="checkbox"]');
                    if (pgRadio) return true;
                }

                return false;
            },
            'code' : function(env) {
                var id = getUniqueId('formInput');
                return '<div class="mb-3 form-check">\
                    <input type="checkbox" class="form-check-input" id="' + id + '">\
                    <label class="form-check-label" for="' + id + '">Check me out</label>\
                </div>'
            },
            action_menu: {
                add: [prefix + 'form-checkbox', prefix + 'form-check-label', prefix + 'form-checkbox-button', prefix + 'form-check-label-button', prefix + 'form-text']
            },
            'sections' : addPrefixToSectionsAndFields({
                'checkboxgroupstyle' : {
                    name : 'Style',
                    fields : {
                        'checkboxgroupinline': form_check_inline,
                        'checkboxgroupswitch': {
                            type: 'checkbox',
                            name: 'Switch',
                            action: 'apply_class',
                            value: 'form-switch'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_form_checkbox_block);

        var def_form_checkbox_button_block = new PgComponentType(prefix + 'form-checkbox-button-block', 'Checkbox button block', {
            'selector' : null,
            'code' : function(env) {
                var id = getUniqueId('checkboxButton');
                return '<div class="form-check mb-3">\
                    <input type="checkbox" class="btn-check" id="' + id + '" autocomplete="off">\
                    <label class="btn btn-primary" for="' + id + '">Single toggle</label>\
                </div>'
            },
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_checkbox_button_block);

        var def_form_radio = new PgComponentType(prefix + 'form-radio', 'Radio', {
            'selector' : 'input.form-check-input[type="radio"]',
            'code' : '<input class="form-check-input" type="radio" name="inlineRadioOptions" value="option1">',
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_radio);

        var def_form_radio_block = new PgComponentType(prefix + 'form-radio-block', 'Radio block', {
            'selector_new' : function(pgel) {
                if (pgel.hasClass('form-check')) {
                    var pgRadio = pgel.findOne('input[type="radio"]');
                    if (pgRadio) return true;
                }

                return false;
            },
            'code' : function(env) {
                var id = getUniqueId('formInput');
                return '<div class="form-check">\
                    <input class="form-check-input" type="radio" name="inlineRadioOptions" id="' + id + '" value="option1">\
                    <label class="form-check-label" for="' + id + '">\
                        Option one\
                    </label>\
                </div>'
            },
            action_menu: {
                add: [prefix + 'form-radio', prefix + 'form-check-label', prefix + 'form-radio-button', prefix + 'form-check-label-button', prefix + 'form-text']
            },
            'sections' : addPrefixToSectionsAndFields({
                'radiogroupstyle' : {
                    name : 'Style',
                    fields : {
                        'radiogroupgroup' : {
                            'type' : 'text',
                            'name' : 'name',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                var pginput = pgel.findOne('input[type="radio"]');
                                return pginput.getAttr('name');
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var pginput = pgel.findOne('input[type="radio"]');
                                pginput.setAttr('name', value);
                                return value;
                            }
                        },
                        'radiogroupinline' : form_check_inline
                    }
                }
            })
        });
        f.addComponentType(def_form_radio_block);

        var def_form_radio_button_block = new PgComponentType(prefix + 'form-radio-button-block', 'Radio button block', {
            'selector' : null,
            'code' : function(env) {
                var id = getUniqueId('radioButton');
                return '<div class="form-check mb-3">\
                    <input type="radio" class="btn-check" name="options" id="' + id + '" autocomplete="off">\
                    <label class="btn btn-secondary" for="' + id + '">Radio</label>\
                </div>'
            },
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_radio_button_block);

        var def_form_select = new PgComponentType(prefix + 'form-select', 'Select', {
            'selector' : '.form-select',
            'code' : '<select class="form-select" aria-label="Default select example">\
                <option selected>Open this select menu</option>\
                <option value="1">One</option>\
                <option value="2">Two</option>\
                <option value="3">Three</option>\
            </select>',
            action_menu: {
                add: [prefix + 'form-select-option']
            },
            'sections' : addPrefixToSectionsAndFields({
                'formselectstyle' : {
                    name : 'Select style',
                    fields : {
                        'formselectcontrolsize' : {
                            'type': 'select',
                            'name': 'Size',
                            'action': 'apply_class',
                            'show_empty': true,
                            'options': [
                                { key: 'form-select-lg', name: 'Large' },
                                { key: 'form-select-sm', name: 'Small' }
                            ]
                        },
                        'formselectmultiple' : {
                            'type' : 'checkbox',
                            'name' : 'Multiple',
                            'action' : 'element_attribute',
                            'attribute' : 'multiple',
                            'value' : 'multiple',
                            'attribute_without_value' : true
                        },
                    }
                }
            })
        });
        f.addComponentType(def_form_select);

        var def_form_select_block = new PgComponentType(prefix + 'form-select-block', 'Select block', {
            selector : null,
            parent_selector: 'form,fieldset',
            'code' : function(env) {
                var id = getUniqueId('formInput');
                return '<div class="mb-3">\
                    <label for="' + id + '">Field label</label>\
                    <select id="' + id + '" class="form-select" aria-label="Default select example">\
                        <option selected>Open this select menu</option>\
                        <option value="1">One</option>\
                        <option value="2">Two</option>\
                        <option value="3">Three</option>\
                    </select>\
                </div>'
            },
            action_menu: {
                add: [prefix + 'form-select', prefix + 'form-label', prefix + 'form-text']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_select_block);

        var def_form_input_group = new PgComponentType(prefix + 'form-input-group', 'Input group', {
            'selector' : 'div.input-group',
            'priority' : 100,
            'code' : '<div class="input-group mb-3">\
                <span class="input-group-text" id="basic-addon1">@</span>\
                <input type="text" class="form-control" placeholder="Username" aria-label="Username" aria-describedby="basic-addon1">\
            </div>',
            action_menu : {
                'add' : [prefix + 'input-group-text', prefix + 'form-input', prefix + 'button', prefix + 'button-dropdown']
            },
            'sections' : addPrefixToSectionsAndFields({
                'inputgroupstyle' : {
                    name : 'Input group',
                    fields : {
                        inputgroupsize: {
                            name: 'Size',
                            type: 'select',
                            action: 'apply_class',
                            show_empty : true,
                            options: [
                                {key: 'input-group-lg', name: 'Large'},
                                {key: 'input-group-sm', name: 'Small'}
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_form_input_group);

        var def_form_range = new PgComponentType(prefix + 'form-range', 'Range', {
            'selector' : '.form-range',
            'code' : '<input type="range" class="form-range">',
            'sections' : addPrefixToSectionsAndFields({
                name: "Options",
                fields: {
                    rangeblockmin: {
                        'type' : 'text',
                        'name' : 'Min',
                        'action' : 'element_attribute',
                        'attribute' : 'min'
                    },
                    rangeblockmax: {
                        'type' : 'text',
                        'name' : 'Max',
                        'action' : 'element_attribute',
                        'attribute' : 'max'
                    },
                    rangeblockstep: {
                        'type' : 'text',
                        'name' : 'Step',
                        'action' : 'element_attribute',
                        'attribute' : 'step'
                    }
                }
            })
        });
        f.addComponentType(def_form_range);

        var def_form_range_block = new PgComponentType(prefix + 'form-range-block', 'Range block', {
            'priority' : 100,
            'code' : function() {
                var id = getUniqueId('customRange', null, 0);

                return '<div class="form-group">\
                    <label for="' + id + '" class="form-label">Example range</label>\
                    <input type="range" class="form-range" id="' + id + '">\
                </div>'
            },
            action_menu : {
                'add' : [prefix + 'form-range']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_range_block);

        var def_form_file = new PgComponentType(prefix + 'form-file', 'File input', {
            'selector': null,
            'code': '<input type="file" class="form-control">',
            'sections': addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_file);

        var def_form_file_block = new PgComponentType(prefix + 'form-file-block', 'File block', {
            'selector': null,
            parent_selector: 'form,fieldset',
            'code': function (env) {
                var id = getUniqueId('formFile');
                return '<div class="mb-3">\
                    <label for="' + id + '" class="form-label">Default file input example</label>\
                    <input class="form-control" type="file" id="' + id + '">\
                </div>';
            },
            action_menu : {
                'add' : [prefix + 'form-file']
            },
            'sections': addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_file_block);

        var def_form_color = new PgComponentType(prefix + 'form-color', 'Color input', {
            'selector': '.form-control-color',
            'code': '<input type="color" class="form-control form-control-color" value="#563d7c" title="Choose your color">',
            'sections': addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_color);

        var def_form_color_block = new PgComponentType(prefix + 'form-color-block', 'Color block', {
            'selector': null,
            parent_selector: 'form,fieldset',
            'code': function (env) {
                var id = getUniqueId('formColor');
                return '<div class="mb-3">\
                    <label for="' + id + '" class="form-label">Color picker</label>\
                    <input type="color" class="form-control form-control-color" id="' + id + '" value="#563d7c" title="Choose your color">\
                </div>';
            },
            action_menu : {
                'add' : [prefix + 'form-color']
            },
            'sections': addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_color_block);

        var def_form_datalists_block = new PgComponentType(prefix + 'form-datalists-block', 'Datalists block', {
            'selector': null,
            parent_selector: 'form,fieldset',
            'code': function (env) {
                var id = getUniqueId('formDatalists');
                return '<div class="mb-3">\
                    <label for="' + id + '" class="form-label">Datalist example</label>\
                    <input class="form-control" list="datalistOptions" id="' + id + '" placeholder="Type to search...">\
                    <datalist id="datalistOptions">\
                        <option value="San Francisco">\
                        <option value="New York">\
                        <option value="Seattle">\
                        <option value="Los Angeles">\
                        <option value="Chicago">\
                    </datalist>\
                </div>';
            },
            'sections': addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_datalists_block);

        var def_form_floating = new PgComponentType(prefix + 'form-floating', 'Form floating', {
            'selector': '',
            'code': function() {
                var id = getUniqueId('floatingInput', null, 0);
                return '<div class="form-floating mb-3">\
                    <input type="email" class="form-control" id="' + id + '" placeholder="name@example.com">\
                    <label for="' + id + '">Email address</label>\
                </div>'
            },
            'sections': addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_form_floating);

        var def_form_feedback_invalid = new PgComponentType(prefix + 'form-feedback-invalid', 'Invalid feedback', {
            'selector' : '.invalid-feedback',
            'code' : '<div class="invalid-feedback">A block of invalid feedback text.</div>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_feedback_invalid);

        var def_form_feedback_valid = new PgComponentType(prefix + 'form-feedback-valid', 'Valid feedback', {
            'selector' : '.valid-feedback',
            'code' : '<div class="valid-feedback">Looks good!</div>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_feedback_valid);

        var def_form_tooltip_invalid = new PgComponentType(prefix + 'form-tooltip-invalid', 'Invalid tooltip', {
            'selector' : '.invalid-tooltip',
            'code' : '<div class="valid-tooltip">A block of invalid tooltip text.</div>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_tooltip_invalid);

        var def_form_tooltip_valid = new PgComponentType(prefix + 'form-tooltip-valid', 'Valid tooltip', {
            'selector' : '.valid-tooltip',
            'code' : '<div class="valid-tooltip">Looks good!</div>',
            'sections' : addPrefixToSectionsAndFields({})
        });
        f.addComponentType(def_form_tooltip_valid);

        var form = new PgFrameworkLibSection(prefix + 'forms', 'Forms');
        form.setComponentTypes([def_form, def_form_fieldset, def_form_help, def_form_ta_block, def_form_checkbox_block, def_form_checkbox_button_block, def_form_radio_block, def_form_radio_button_block, def_form_select_block, def_form_input_group, def_form_range_block, def_form_color_block, def_form_datalists_block, def_form_feedback_invalid, def_form_feedback_valid, def_form_tooltip_invalid, def_form_tooltip_valid]);
        addLibSection(form);

        //* Javascript
        // Accordion
        var def_accordion = new PgComponentType(prefix + 'accordion', 'Accordion', {
            'selector' : '.accordion',
            big_preview: true,
            'code' : function() {
                var pid = getUniqueId('panels', null, 0);
                var id1 = getUniqueId('collapse', null, 0);
                var id2 = getUniqueId('collapse', null, id1);
                var id3 = getUniqueId('collapse', null, id2);

                return '<div class="accordion" id="' + pid + '">\
                    <div class="accordion-item">\
                        <h2 class="accordion-header" id="' + id1 + '">\
                            <button class="accordion-button" type="button" data-bs-toggle="collapse" data-bs-target="#collapseOne" aria-expanded="true" aria-controls="collapseOne">\
                            Accordion Item #1\
                            </button>\
                        </h2>\
                        <div id="collapseOne" class="accordion-collapse collapse show" aria-labelledby="' + id1 + '" data-bs-parent="#' + pid + '">\
                            <div class="accordion-body">\
                                <strong>This is the first item\'s accordion body.</strong> It is hidden by default, until the collapse plugin adds the appropriate classes that we use to style each element. These classes control the overall appearance, as well as the showing and hiding via CSS transitions. You can modify any of this with custom CSS or overriding our default variables. It\'s also worth noting that just about any HTML can go within the <code>.accordion-body</code>, though the transition does limit overflow.\
                            </div>\
                        </div>\
                    </div>\
                    <div class="accordion-item">\
                        <h2 class="accordion-header" id="' + id2 + '">\
                            <button class="accordion-button collapsed" type="button" data-bs-toggle="collapse" data-bs-target="#collapseTwo" aria-expanded="false" aria-controls="collapseTwo">\
                            Accordion Item #2\
                            </button>\
                        </h2>\
                        <div id="collapseTwo" class="accordion-collapse collapse" aria-labelledby="' + id2 + '" data-bs-parent="#' + pid + '">\
                            <div class="accordion-body">\
                                <strong>This is the second item\'s accordion body.</strong> It is hidden by default, until the collapse plugin adds the appropriate classes that we use to style each element. These classes control the overall appearance, as well as the showing and hiding via CSS transitions. You can modify any of this with custom CSS or overriding our default variables. It\'s also worth noting that just about any HTML can go within the <code>.accordion-body</code>, though the transition does limit overflow.\
                            </div>\
                        </div>\
                    </div>\
                    <div class="accordion-item">\
                        <h2 class="accordion-header" id="' + id3 + '">\
                            <button class="accordion-button collapsed" type="button" data-bs-toggle="collapse" data-bs-target="#collapseThree" aria-expanded="false" aria-controls="collapseThree">\
                            Accordion Item #3\
                            </button>\
                        </h2>\
                        <div id="collapseThree" class="accordion-collapse collapse" aria-labelledby="' + id3 + '" data-bs-parent="#' + pid + '">\
                            <div class="accordion-body">\
                                <strong>This is the third item\'s accordion body.</strong> It is hidden by default, until the collapse plugin adds the appropriate classes that we use to style each element. These classes control the overall appearance, as well as the showing and hiding via CSS transitions. You can modify any of this with custom CSS or overriding our default variables. It\'s also worth noting that just about any HTML can go within the <code>.accordion-body</code>, though the transition does limit overflow.\
                            </div>\
                        </div>\
                    </div>\
                </div>';
            },
            tags: 'major',
            priority: 100,
            action_menu: {
                add: [prefix + 'accordion-item']
            },
            on_inserted : function() {
            },
            on_changed : function() {
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_accordion);

        var def_accordion_item = new PgComponentType(prefix + 'accordion-item', 'Accordion item', {
            'selector' : null,
            'code' : function() {
                var id1 = getUniqueId('collapse', null, 0);

                return '<div class="accordion-item">\
                    <h2 class="accordion-header" id="' + id1 + '">\
                        <button class="accordion-button" type="button" data-bs-toggle="collapse" data-bs-target="#collapseOne" aria-expanded="true" aria-controls="collapseOne">\
                        Accordion Item\
                        </button>\
                    </h2>\
                    <div id="collapseOne" class="accordion-collapse collapse show" aria-labelledby="' + id1 + '" data-bs-parent="#' + pid + '">\
                        <div class="accordion-body">\
                            <strong>This is the first item\'s accordion body.</strong> It is hidden by default, until the collapse plugin adds the appropriate classes that we use to style each element. These classes control the overall appearance, as well as the showing and hiding via CSS transitions. You can modify any of this with custom CSS or overriding our default variables. It\'s also worth noting that just about any HTML can go within the <code>.accordion-body</code>, though the transition does limit overflow.\
                        </div>\
                    </div>\
                </div>';
            },
            tags: 'major',
            priority: 100,
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_accordion_item);

        var def_collapse = new PgComponentType(prefix + 'collapse', 'Collapse', {
            'selector' : '.collapse',
            'code' : `<div class="collapse">
                <div class="card card-body">
                    Some placeholder content for the collapse component. This panel is hidden by default but revealed when the user activates the relevant trigger.
                </div>
            </div>`,
            tags: 'major',
            priority: 100,
            'sections' : addPrefixToSectionsAndFields({
                collapse: {
                    name : 'Collapse options',
                    fields : {
                        collapsemulti : {
                            'type' : 'checkbox',
                            'name' : 'Mlti collapse',
                            'action' : 'apply_class',
                            'value' : 'multi-collapse',
                        },
                        collapseHorizontal : {
                            'type' : 'checkbox',
                            'name' : 'Horizontal',
                            'action' : 'apply_class',
                            'value' : 'collapse-horizontal',
                            show_if: function () {
                                return reqVersion(5.1);
                            }
                        },
                    }
                }
            })
        });
        f.addComponentType(def_collapse);

        var callSlider = function(pgel, func, msg) {
            var id = pgel.getId();

            code = `
                var myCarousel = document.querySelector('[data-pg-id="` + id + `"]');
                var carousel = new bootstrap.Carousel(myCarousel);
                carousel.` + func + `;
            `;

            pinegrow.setIgnoreClicks(true);
            pgel.executeScript(code);
            pinegrow.setIgnoreClicks(false);
            if(msg) {
                pinegrow.showQuickMessage(msg);
            }
        }

        var gotoSlide = function(pgel) {
            var slider = pgel.closest('.carousel');
            callSlider(slider, "pause()");
            var num = pgel.index();
            callSlider(slider, "to(" + num + ")", "Going to slide " + (num + 1) + "...");
        }

        var def_carousel = new PgComponentType(prefix + 'carousel', 'Carousel', {
            'selector' : '.carousel',
            'code' : function() {
                var pid = getUniqueId('carousel', null, 0);

                return '<div id="' + pid + '" class="carousel slide" data-bs-ride="carousel">\
                    <ol class="carousel-indicators">\
                        <li data-bs-target="#' + pid + '" data-bs-slide-to="0" class="active"></li>\
                        <li data-bs-target="#' + pid + '" data-bs-slide-to="1"></li>\
                        <li data-bs-target="#' + pid + '" data-bs-slide-to="2"></li>\
                    </ol>\
                    <div class="carousel-inner">\
                        <div class="carousel-item active" data-bs-interval="10000">\
                            <img src="' + getPlaceholderImage() + '" class="d-block w-100">\
                            <div class="carousel-caption d-none d-md-block">\
                                <h5>First slide label</h5>\
                                <p>Nulla vitae elit libero, a pharetra augue mollis interdum.</p>\
                            </div>\
                        </div>\
                        <div class="carousel-item" data-bs-interval="2000">\
                            <img src="' + getPlaceholderImage() + '" class="d-block w-100">\
                            <div class="carousel-caption d-none d-md-block">\
                                <h5>Second slide label</h5>\
                                <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit.</p>\
                            </div>\
                        </div>\
                        <div class="carousel-item">\
                            <img src="' + getPlaceholderImage() + '" class="d-block w-100">\
                            <div class="carousel-caption d-none d-md-block">\
                                <h5>Third slide label</h5>\
                                <p>Praesent commodo cursus magna, vel scelerisque nisl consectetur.</p>\
                            </div>\
                        </div>\
                    </div>\
                    <a class="carousel-control-prev" href="#' + pid + '" role="button" data-bs-slide="prev">\
                        <span class="carousel-control-prev-icon" aria-hidden="true"></span>\
                        <span class="visually-hidden">Previous</span>\
                    </a>\
                    <a class="carousel-control-next" href="#' + pid + '" role="button" data-bs-slide="next">\
                        <span class="carousel-control-next-icon" aria-hidden="true"></span>\
                        <span class="visually-hidden">Next</span>\
                    </a>\
                </div>';
            },
            tags: 'major',
            priority: 100,
            action_menu: {
                actions : [
                    {label: "Pause Carousel", action: function(pgel) {
                        callSlider(pgel, "pause()", "Carousel paused.");
                    }},
                    {label: "Resume Carousel", action: function(pgel) {
                        callSlider(pgel, "cycle()", "Carousel resumed.");
                    }},
                    {label: "Next Slide", action: function(pgel) {
                        callSlider(pgel, "next()");
                    }},
                    {label: "Previous Slide", action: function(pgel) {
                        callSlider(pgel, "prev()");
                    }}
                ],
                add: [prefix + 'carousel-item', prefix + 'carousel-prev', prefix + 'carousel-next'],
                on_add : function (pgel, pgnew, newdef, prepend) {
                    if (newdef.type == prefix + 'carousel-item') {
                        var pgol = pgel.find('.carousel-indicators');
                        var pginner = pgel.find('.carousel-inner');
                        pginner = pginner[0];

                        var i = 0;
                        if (pginner) {
                            i = pginner.children.length;
                        }

                        var id = pgel.getAttr('id');
                        if (!id) id = '';
                        var pgli = pgCreateNodeFromHtml('<li data-bs-target="#' + id + '" data-bs-slide-to="' + i + '"></li>');
                        if (prepend) {
                            if (pginner) pginner.prepend(pgnew);
                            if (pgol.length > 0) pgol[0].prepend(pgli);
                        } else {
                            if (pgol.length > 0) pgol[0].append(pgli);
                            if (pginner) pginner.append(pgnew);
                        }
                    }
                    else if (newdef.type == prefix + 'carousel-prev' || newdef.type == prefix + 'carousel-next') {
                        pgel.append(pgnew);
                    }
                }
            },
            on_inserted : function() {
            },
            on_changed : function() {
            },
            'sections' : addPrefixToSectionsAndFields({
                carousel : {
                    name : 'Carousel options',
                    fields : {
                        carouselfade : {
                            'type' : 'checkbox',
                            'name' : 'Fade',
                            'action' : 'apply_class',
                            'value' : 'carousel-fade',
                        },
                        carouseldark : {
                            'type' : 'checkbox',
                            'name' : 'Dark',
                            'action' : 'apply_class',
                            'value' : 'carousel-dark',
                        },
                        carouselinterval : {
                            'type' : 'text',
                            'name' : 'Interval',
                            'action' : 'element_attribute',
                            'attribute' : 'data-bs-interval'
                        },
                        carouselkeyboard : {
                            'type' : 'select',
                            'name' : 'Keyboard',
                            'action' : 'element_attribute',
                            'attribute' : 'data-bs-keyboard',
                            'show_empty' : true,
                            'options' : [
                                {key: 'true', name: 'Yes'},
                                {key: 'false', name: 'No'}
                            ]
                        },
                        carouselpause : {
                            'type' : 'select',
                            'name' : 'Pause on',
                            'action' : 'element_attribute',
                            'attribute' : 'data-bs-pause',
                            'show_empty' : true,
                            'options' : [
                                {key: 'hover', name: 'Hover'}
                            ]
                        },
                        carouselwrap : {
                            'type' : 'select',
                            'name' : 'Wrap',
                            'action' : 'element_attribute',
                            'attribute' : 'data-bs-wrap',
                            'show_empty' : true,
                            'options' : [
                                {key: 'true', name: 'Yes'},
                                {key: 'false', name: 'No'}
                            ]
                        },
                        carouseltouch : {
                            'type' : 'select',
                            'name' : 'Touch',
                            'action' : 'element_attribute',
                            'attribute' : 'data-bs-touch',
                            'show_empty' : true,
                            'options' : [
                                {key: 'true', name: 'Yes'},
                                {key: 'false', name: 'No'}
                            ]
                        }
                    }
                }
            })
        });
        f.addComponentType(def_carousel);

        if(pgIsMagic()) {
            def_carousel.on_filter_tree_children = function (pgel) {
                var r = [];
                r = r.concat(pgel.find('.carousel-item'));
                return r;
            }

            f.componentFiltersTree(def_carousel);
        }

        var def_carousel_item = new PgComponentType(prefix + 'carousel-item', 'Carousel slide', {
            'selector' : '.carousel-item',
            'parent_selector' : '.carousel-inner',
            'code' : function() {
                return '<div class="carousel-item" data-bs-interval="10000">\
                    <img src="' + getPlaceholderImage() + '" class="d-block w-100">\
                    <div class="carousel-caption d-none d-md-block">\
                        <h5>First slide label</h5>\
                        <p>Nulla vitae elit libero, a pharetra augue mollis interdum.</p>\
                    </div>\
                </div>';
            },
            'action_menu': {
                add: [prefix + 'img', prefix + 'carousel-caption'],
            },
            tags: 'major',
            priority: 100,
            on_selected: function(pgel) {
                gotoSlide(pgel);
            },
            'sections' : addPrefixToSectionsAndFields({
                'carouselitemslide' : {
                    name : 'Slide',
                    fields : {
                        carouselitemactive : {
                            'type' : 'checkbox',
                            'name' : 'Active slide',
                            'action' : 'apply_class',
                            'value' : 'active'
                        },
                        carouseliteminterval: {
                            'type' : 'text',
                            'name' : 'Interval',
                            'action' : 'element_attribute',
                            'attribute' : 'data-bs-interval'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_carousel_item);

        var def_carousel_caption = new PgComponentType(prefix + 'carousel-caption', 'Carousel caption', {
            'selector' : '.carousel-caption',
            'parent_selector' : '.carousel-inner',
            'code' : '<div class="carousel-caption d-none d-md-block">\
                <h5>First slide label</h5>\
                <p>Nulla vitae elit libero, a pharetra augue mollis interdum.</p>\
            </div>',
            tags: 'major',
            priority: 100,
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_carousel_caption);

        var def_carousel_inner = new PgComponentType(prefix + 'carousel-inner', 'Carousel inner', {
            'selector' : '.carousel-inner',
            parent_selector: '.carousel',
            'code' : '<div class="carousel-inner"></div>',
            action_menu: {
                add: [prefix + 'carousel-item']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_carousel_inner);

        var def_carousel_indicators = new PgComponentType(prefix + 'carousel-indicators', 'Carousel indicators', {
            'selector' : '.carousel-indicators',
            parent_selector: '.carousel',
            'code' : '<ol class="carousel-indicators"></ol>',
            action_menu: {
                add: [prefix + 'carousel-indicator']
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_carousel_indicators);

        var def_carousel_indicator = new PgComponentType(prefix + 'carousel-indicator', 'Carousel indicator', {
            'selector_new': function (pgel) {
                return pgel.parent && pgel.parent.hasClass('carousel-indicators');
            },
            'code' : '<li data-bs-target="" data-bs-slide-to=""></li>',
            action_menu: {
                add: [prefix + 'carousel-indicator']
            },
            'sections' : addPrefixToSectionsAndFields({
                'carouselindicator' : {
                    name : 'Carousel',
                    fields : {
                        carouselindicatordatatarget : {
                            type: 'text',
                            name: 'Data target',
                            type: 'text',
                            live_update: true,
                            action : 'custom',
                            get_value: function(pgel) {
                                return pgel.attr('data-bs-target');
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                if(value) {
                                    pgp = pgel.parent;
                                    if (pgp) {
                                        var allPgChildren = pgp.children;
                                        for (var i = 0; i < allPgChildren.length; i++) {
                                            if (allPgChildren[i].closingTag != 'text') {
                                                allPgChildren[i].setAttr('data-bs-target', value);
                                            }
                                        }

                                        var pgcarousel = pgp.parent;
                                        if (pgcarousel) {
                                            pgcarousel.setAttr('id', value.replace('#', ''));

                                            var pgprev = pgcarousel.findOne('.carousel-control-prev');
                                            if (pgprev) pgprev.setAttr('href', value);

                                            var pgnext = pgcarousel.findOne('.carousel-control-next');
                                            if (pgnext) pgnext.setAttr('href', value);
                                        }
                                    }
                                    else {
                                         pgel.setAttr('data-bs-target', value);
                                    }
                                } else {
                                    pgel.removeAttr('data-bs-target');
                                }
                                return value;
                            }
                        },
                        carouselindicatorslideto : {
                            type: 'text',
                            name: 'Slide to',
                            action: 'element_attribute',
                            attribute: 'data-bs-slide-to'
                        }
                    }
                }
            })
        });
        f.addComponentType(def_carousel_indicator);

        var def_carousel_prev = new PgComponentType(prefix + 'carousel-prev', 'Carousel prev', {
            'selector' : '.carousel-control-prev',
            parent_selector: '.carousel',
            'code' : '<a class="carousel-control-prev" href="#" role="button" data-bs-slide="prev">\
                <span class="carousel-control-prev-icon" aria-hidden="true"></span>\
                <span class="visually-hidden">Previous</span>\
            </a>',
            on_inserted : function(pgel, page) {
                var pgp = pgel.parent;
                if (pgp) {
                    var id = pgp.getAttr('id');
                    pgel.setAttr('href', '#' + id);
                }
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_carousel_prev);

        var def_carousel_next = new PgComponentType(prefix + 'carousel-next', 'Carousel next', {
            'selector' : '.carousel-control-next',
            parent_selector: '.carousel',
            'code' : '<a class="carousel-control-next" href="#" role="button" data-bs-slide="next">\
                <span class="carousel-control-next-icon" aria-hidden="true"></span>\
                <span class="visually-hidden">Next</span>\
            </a>',
            on_inserted : function(pgel, page) {
                var pgp = pgel.parent;
                if (pgp) {
                    var id = pgp.getAttr('id');
                    pgel.setAttr('href', '#' + id);
                }
            },
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_carousel_next);

        //* Offcanvas
        var def_offcanvas = new PgComponentType(prefix + 'offcanvas', 'Offcanvas', {
            'selector' : '.offcanvas',
            big_preview: true,
            'code' : function() {
                return '<div class="offcanvas offcanvas-start" tabindex="-1" aria-labelledby="offcanvasLabel">\
                    <div class="offcanvas-header">\
                        <h5 class="offcanvas-title" id="offcanvasLabel">Offcanvas</h5>\
                        <button type="button" class="btn-close text-reset" data-bs-dismiss="offcanvas" aria-label="Close"></button>\
                    </div>\
                    <div class="offcanvas-body">\
                        Content for the offcanvas goes here. You can place just about any Bootstrap component or custom elements here.\
                    </div>\
                </div>';
            },
            tags: 'major',
            priority: 100,
            action_menu: {
                add: [prefix + 'offcanvas-header', prefix + 'offcanvas-body']
            },
            on_inserted : function() {
            },
            on_changed : function() {
            },
            'sections' : addPrefixToSectionsAndFields({
                bsoffcanvas : {
                    name : 'Modal options',
                    fields : {
                        bsoffcanvasplacement : {
                            type: 'select',
                            name: 'Placement',
                            show_empty: false,
                            action: 'apply_class',
                            options: [
                                { key: 'offcanvas-start', name: 'Start' },
                                { key: 'offcanvas-end', name: 'End' },
                                { key: 'offcanvas-top', name: 'Top' },
                                { key: 'offcanvas-bottom', name: 'Bottom' },
                            ]
                        },
                        bsoffcanvasshow: {
                            type: 'checkbox',
                            name: 'Show',
                            action: 'apply_class',
                            value: 'show',
                            on_changed: function() {
                                onJSChangedRefresh();
                            }
                        },
                        bsoffcanvasbackdrop : {
                            'name' : 'Backdrop',
                            'type' : 'checkbox',
                            'value' : 'true',
                            default_value: 'true',
                            negvalue: 'false',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-backdrop',
                            on_changed: function() {
                                onJSChangedRefresh();
                            }
                        },
                        bsoffcanvaskeyboard : {
                            'name' : 'Keyboard',
                            'type' : 'checkbox',
                            'value' : 'true',
                            default_value: 'true',
                            negvalue: 'false',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-keyboard',
                            on_changed: function() {
                                onJSChangedRefresh();
                            }
                        },
                        bsoffcanvaskeyboard : {
                            'name' : 'Keyboard',
                            'type' : 'checkbox',
                            'value' : 'true',
                            default_value: 'false',
                            negvalue: 'false',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-keyboard',
                            on_changed: function() {
                                onJSChangedRefresh();
                            }
                        },
                        bsoffcanvasscroll : {
                            'name' : 'Scroll',
                            'type' : 'checkbox',
                            'value' : 'true',
                            default_value: 'false',
                            negvalue: 'false',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-scroll',
                            on_changed: function() {
                                onJSChangedRefresh();
                            }
                        },
                    }
                }
            })
        });
        f.addComponentType(def_offcanvas);

        var def_offcanvas_header = new PgComponentType(prefix + 'offcanvas-header', 'Offcanvas header', {
            'selector' : '.offcanvas-header',
            big_preview: true,
            'code' : '<div class="offcanvas-header">\
                <h5 class="offcanvas-title" id="offcanvasLabel">Offcanvas</h5>\
                <button type="button" class="btn-close text-reset" data-bs-dismiss="offcanvas" aria-label="Close"></button>\
            </div>',
            tags: 'major',
            priority: 100,
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_offcanvas_header);

        var def_offcanvas_body = new PgComponentType(prefix + 'offcanvas-body', 'Offcanvas body', {
            'selector' : '.offcanvas-body',
            big_preview: true,
            'code' : '<div class="offcanvas-body">\
                Content for the offcanvas goes here. You can place just about any Bootstrap component or custom elements here.\
            </div>',
            tags: 'major',
            priority: 100,
            'sections' : addPrefixToSectionsAndFields({
            })
        });
        f.addComponentType(def_offcanvas_body);

        //* Modals

        // modal size options
        var modalSizeOptions = [
            {key: '', name: 'Normal', ahtml: bm.makeText('Normal')},
            {key: 'modal-sm', name: 'Small', ahtml: bm.makeText('Small')},
            {key: 'modal-lg', name: 'Large', ahtml: bm.makeText('Large')},
            {key: 'modal-xl', name: 'X Large', ahtml: bm.makeText('X Large')}
        ];

        var modalFullscreen = [
            {key: 'modal-fullscreen', name: 'Always'},
            {key: 'modal-fullscreen-sm-down', name: 'Small and less'},
            {key: 'modal-fullscreen-md-down', name: 'Medium and less'},
            {key: 'modal-fullscreen-lg-down', name: 'Large and less'},
            {key: 'modal-fullscreen-xl-down', name: 'X Large and less'},
            {key: 'modal-fullscreen-xxl-down', name: 'XX Large and less'},
        ];


        var def_modal = new PgComponentType(prefix + 'modal', 'Modal', {
            'selector' : '.modal',
            'code' : function() {
                var id = getUniqueId('modal', null, 0);
                return '<div class="modal fade pg-show-modal" id="' + id + '" tabindex="-1" role="dialog" aria-hidden="true">\
                    <div class="modal-dialog">\
                        <div class="modal-content">\
                            <div class="modal-header">\
                                <h4 class="modal-title">Modal title</h4>\
                                <button type="button" class="btn-close" data-bs-dismiss="modal" aria-label="Close"></button>\
                            </div>\
                            <div class="modal-body">\
                                <p>Modal body text goes here.</p>\
                            </div>\
                            <div class="modal-footer">\
                                <button type="button" class="btn btn-default" data-bs-dismiss="modal">Close</button>\
                                <button type="button" class="btn btn-primary">Save changes</button>\
                            </div>\
                        </div>\
                    </div>\
                </div>'
            },
            preview: '<div class="modal open fade in" style="display:block;opacity:1;height:300px;width:500px;position:relative;">\
                <div class="modal-dialog modal-sm">\
                    <div class="modal-content">\
                        <div class="modal-header">\
                            <h4 class="modal-title">Modal title</h4>\
                            <button type="button" class="close" data-dismiss="modal" aria-hidden="true">&times;</button>\
                        </div>\
                        <div class="modal-body">\
                            <p>One fine body&hellip;</p>\
                        </div>\
                        <div class="modal-footer">\
                            <button type="button" class="btn btn-default" data-dismiss="modal">Close</button>\
                            <button type="button" class="btn btn-primary">Save changes</button>\
                        </div>\
                    </div>\
                </div>\
            </div>\
            ',
            _editModalTriggers: function(pgel, $clicked) {
                var q = new PgQuickSelectElement({
                    title: 'Edit triggers for showing the modal',
                    description: '<p>Select an element (usually a button or a link) that should open the modal.</p>',
                    reselect: true,
                    target: $clicked,
                    current_col: function() {
                        var mid = pgel.getAttribute('id');
                        if(mid) {
                            var col = new PgCollection();
                            col.setList(pgel.document.findWithMultipleAttrValues([
                                {
                                    name: 'data-bs-toggle',
                                    value: 'modal'
                                },
                                {
                                    name: 'data-bs-target',
                                    value: '#' + mid
                                }
                            ]))
                            return col;
                        } else {
                            return null;
                        }
                    },
                    remove: true,
                    func: function(col, remove_col) {
                        var mid = pgel.getAttribute('id');
                        if(!mid) {
                            mid = pgel.getUniqueIdAttribute('modal');
                            pgel.setAttribute('id', mid);
                        }
                        var api = new PgApi();
                        api.setAttribute(col, [
                            {
                                name: 'data-bs-toggle',
                                value: 'modal'
                            },
                            {
                                name: 'data-bs-target',
                                value: '#' + mid
                            }
                        ])
                        api.removeAttribute(remove_col, ['data-bs-toggle', 'data-bs-target']);
                        pinegrow.showQuickMessage('The modal triggers are set.')
                    }
                })
            },
            on_get_menu_commands: function(pgel, list, $el) {
                var _this = this;
                list.push({
                    icon: 'icon-267',
                    label: 'Edit modal triggers...',
                    header: 'Modal',
                    tooltip: 'Edit triggers for showing the modal.',
                    action: function(pgel, $clicked) {
                        _this._editModalTriggers(pgel, $clicked || $el)
                    }
                })
            },
            tags: 'major',
            on_inserted : function(pgel, page) {
            },
            'sections' : addPrefixToSectionsAndFields({
                bsmodal : {
                    name : 'Modal options',
                    fields : {
                        bsmodalsize : {
                            'type' : 'select',
                            'name' : 'Size',
                            'action' : 'custom',
                            'show_empty' : false,
                            'options' : modalSizeOptions,
                            toggle_buttons: false,
                            get_value: function (pgel) {
                                var pgm = pgel.findOne('.modal-dialog');

                                for (const option of modalSizeOptions) {
                                    if (pgm && pgm.hasClass(option.key)) return option.key;
                                }

                                return null
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var pgm = pgel.findOne('.modal-dialog');

                                if (pgm) {
                                    for (const option of modalSizeOptions) {
                                        pgm.removeClass(option.key);
                                    }

                                    if (value) pgm.addClass(value);
                                }
                                return value;
                            }
                        },
                        bsmodalfullscreen : {
                            'type' : 'select',
                            'name' : 'Full screen',
                            'action' : 'custom',
                            'show_empty' : true,
                            'options' : modalFullscreen,
                            get_value: function (pgel) {
                                var pgm = pgel.findOne('.modal-dialog');

                                for (const option of modalFullscreen) {
                                    if (pgm && pgm.hasClass(option.key)) return option.key;
                                }

                                return null
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var pgm = pgel.findOne('.modal-dialog');

                                if(pgm) {
                                    for (const option of modalFullscreen) {
                                        pgm.removeClass(option.key);
                                    }

                                    if (value) pgm.addClass(value);
                                }
                                return value;
                            }
                        },
                        bsmodalbackdrop : {
                            'type' : 'select',
                            'name' : 'Backdrop',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-backdrop',
                            'show_empty' : true,
                            toggle_buttons: false,
                            'options' : [
                                {key: 'true', name: 'Yes (default)', ahtml: bm.makeText('Yes')},
                                {key: 'false', name: 'No', ahtml: bm.makeText('No')},
                                {key: 'static', name: 'Static', ahtml: bm.makeText('Static')}
                            ]
                        },
                        bsmodalcentered : {
                            type: 'checkbox',
                            name : 'Centered',
                            value: '1',
                            action: "custom",
                            get_value: function (pgel) {
                                var pgModalDialog = pgel.findOne('.modal-dialog');
                                if (pgModalDialog) {
                                    return pgModalDialog.hasClass('modal-dialog-centered') ? '1' : null;
                                }

                                return null;
                            },
                            set_value: function (pgel, value, values) {
                                var pgModalDialog = pgel.findOne('.modal-dialog');
                                if (value) {
                                    if (pgModalDialog) {
                                        pgModalDialog.addClass('modal-dialog-centered');
                                        return '1';
                                    }
                                    else {
                                        return null
                                    }
                                }
                                else {
                                    if (pgModalDialog) {
                                        pgModalDialog.removeClass('modal-dialog-centered');
                                    }
                                }

                                return value;
                            }
                        },
                        bsmodalanimate : {
                            'type' : 'checkbox',
                            'name' : 'Animate',
                            'value' : 'fade',
                            'action' : 'apply_class'
                        },
                        bsmodalstatic : {
                            'type': 'checkbox',
                            'name': 'Static',
                            'value': 'modal-static',
                            'action': 'apply_class'
                        },
                        bsmodalescape : {
                            'type' : 'select',
                            'name' : 'Close on ESC',
                            'type' : 'checkbox',
                            'value' : 'true',
                            default_value: 'true',
                            negvalue: 'false',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-keyboard'
                        },
                        bsmodalfocus : {
                            'type' : 'select',
                            'name' : 'Focus when open',
                            'type' : 'checkbox',
                            'value' : 'true',
                            default_value: 'true',
                            negvalue: 'false',
                            'action' : 'element_attribute',
                            attribute : 'data-bs-focus'
                        },
                        bsmodaleditshow : {
                            'type' : 'checkbox',
                            'name' : 'Show during editing',
                            'value' : 'pg-show-modal',
                            'action' : 'custom',
                            get_value: function (pgel) {
                                return pgel.hasClass('pg-show-modal') ? 'pg-show-modal' : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                if (value) {
                                    pgel.addClass(value);
                                } else {
                                    pgel.removeClass('pg-show-modal');
                                }
                                removeDisplayStyle(pgel);
                                return value;
                            }
                        },
                        bsmodalscrollable : {
                            'type' : 'checkbox',
                            'name' : 'Scrollable',
                            'action' : 'custom',
                            'value' : '1',
                            get_value: function (pgel) {
                                var pgdialog = pgel.findOne('.modal-dialog');
                                return pgdialog && pgdialog.hasClass('modal-dialog-scrollable') ? '1' : null;
                            },
                            set_value: function (pgel, value, values, oldValue, eventType) {
                                var pgdialog = pgel.findOne('.modal-dialog');
                                if (pgdialog) {
                                    if (value) {
                                        pgdialog.addClass('modal-dialog-scrollable');
                                    }
                                    else {
                                        pgdialog.removeClass('modal-dialog-scrollable');
                                    }
                                }

                                return value;
                            }
                        },
                        bsmodaltriggers : {
                            'type' : 'button',
                            'action': 'button',
                            'name' : 'Modal triggers...',
                            'func' : function(pgel, $button) {
                                def_modal._editModalTriggers(pgel, $button);
                            }
                        }
                    }
                }
            })
        });
        f.addComponentType(def_modal);

        var javascript_section = new PgFrameworkLibSection(prefix + 'javascript', 'Javascript');
        javascript_section.setComponentTypes([def_accordion, def_carousel, def_offcanvas, def_modal]);
        addLibSection(javascript_section);

        f.on_get_menu_points = function(page, pgel, menu, points) {
            gh.getMenuPoints(page, pgel, menu, points);
        }

        f.on_page_view_resized = function(page, pv) {
            gh.onPageViewResized(page, pv);
        }

        //GRID
        f.on_get_device_sizes = function(page, list) {
            gh.getDeviceSizesForPageViewMenu(page, list);
        }

        f.on_page_shown_in_live_view = function(page, pv) {
            gh.show(pv);
        }

        var floating_sections = null;

        f.on_get_floating_tools_sections = function(page, sections) {

            if(!floating_sections) {
                var getControl = function (sec, fn, opts) {
                    var def = cloneObject(def_all.sections[prefix + sec].fields[prefix + fn]);
                    def.key = fn;
                    return new PgFloatingToolsControl(def, opts);
                }


                var getMoreControl = function(sec, def_fns, opts) {
                    var def = {
                        name: 'More',
                        key: sec,
                        type: 'select',
                        options: []
                    }
                    var defs_map = {}

                    def_fns.forEach(function(fn) {
                        defs_map[fn] = cloneObject(def_all.sections[prefix + sec].fields[prefix + fn]);
                        if(defs_map[fn].type === 'checkbox') {
                            defs_map[fn].type = 'select';
                            defs_map[fn].options = [
                                {
                                    key: defs_map[fn].value,
                                    name: defs_map[fn].name
                                }
                            ]
                        }
                        def.options.push({
                            key: fn,
                            name: defs_map[fn].name
                        })
                    })

                    opts = opts || {}
                    opts.buttons = false;
                    opts.no_clear_on_first_level = true;
                    opts.submenu = function(fn) {
                        var options = typeof defs_map[fn].options === 'function' ? defs_map[fn].options() : defs_map[fn].options;
                        return {
                            items: options,
                            combine_with_parent: false,
                            name: defs_map[fn].name,
                            require_submenu_selection: true,
                            all_classes: options.map(function(i) { return i.key })
                        }
                    }
                    return new PgFloatingToolsControl(def, opts);
                }


                floating_sections = [];

                var text = new PgFloatingToolsSection(prefix + 'text', 'Text');

                text.addControl(getControl('bstextcontext', 'bsemphasis', {
                    label: `<i class="icon icon-ikone1-100-copy-63"></i>`,
                    on_before_set: function (pgel, classes) {
                        var svg = pgel.closest('svg');
                        if(svg && pgel !== svg) {
                            if(classes.length === 0) {
                                pgel.removeAttribute('stroke');
                            } else {
                                //pgel.removeAttribute('stroke');
                                pgel.setAttribute('stroke', `var(--bs-${classes[0].replace('text-', '')})`);
                                pgel.setInlineStylePropertyValue('stroke', ``);
                            }
                            return false;
                        }
                    }
                })); //text color
                //text.addControl( getControl( 'bstextcontext', 'bslinkcolor')); //link color
                text.addControl(getControl('bstextcontext', 'bstexttransform', {
                    label: `<i class="icon icon-uppercase"></i>`
                })); //text transform
                text.addControl(getControl('bstextcontext', 'bsfontstyle', {
                    label: `<i class="icon icon-style_italic"></i>`
                })); //font style
                text.addControl(getControl('bstextcontext', 'bsfontdecoration', {
                    label: `<i class="icon icon-decoration-under"></i>`
                })); //font deco
                //colors.addControl( getFieldDef( bstextcontext, bsfontstyle)); //font style
                // colors.addControl( getFieldDef( bstextcontext, bsfontstyle)); //font style

                text.addControl(getMoreControl('bstextcontext', ['displayheadings', 'bsfontsize', 'bsfontweight', 'bsineheight', 'bspointevent', 'bstextwrap', 'bslead', 'bstextbreak', 'texttruncate', 'bstextmonospace'], {
                    label: `<i class="icon icon-down"></i>`,
                    name: 'More options'
                }));
                floating_sections.push(text);

                floating_sections.push(text);

                var background = new PgFloatingToolsSection(prefix + 'background', 'Background');
                background.addControl(getControl('bsbackground', 'bsbckcontext', {
                    label: `<i class="icon icon-Pin21"></i>`,
                    name: 'Background context',
                    on_before_set: function (pgel, classes) {

                        var colors = ['light', 'dark', 'primary', 'secondary', 'success', 'info', 'warning', 'danger'];
                        var specials = ['btn', 'alert', 'table', 'btn-outline', 'badge', 'list-group-item'];

                        var svg = pgel.closest('svg');
                        if(svg && pgel !== svg) {
                            if(classes.length === 0) {
                                pgel.removeAttribute('fill');
                            } else {
                                //pgel.removeAttribute('fill');
                                pgel.setAttribute('fill', `var(--bs-${classes[0].replace('bg-', '')})`);
                                pgel.setInlineStylePropertyValue('fill', ``);
                                //pgel.setInlineStylePropertyValue('fill', `var(--bs-${classes[0].replace('bg-', '')})`);
                            }
                            return false;
                        } else {
                            for (var i = 0; i < specials.length; i++) {
                                if (pgel.hasClass(specials[i])) {
                                    var all_classes_re = new RegExp(`^(${pgParserEscapeRegExp(specials[i])})\-(${colors.join('|')})$`, '');
                                    pgel.getClasses().forEach(function (cls) {
                                        if (cls.match(all_classes_re) || cls.startsWith('bg-')) {
                                            pgel.removeClass(cls);
                                        }
                                    })

                                    classes.forEach(function (cls) {
                                        pgel.addClass(cls.replace('bg-', specials[i] + '-'));
                                    })
                                    //dont do default action
                                    return false;
                                }
                            }
                        }
                    }
                }));
                background.addControl(getControl('bsbackground', 'bsbckgradient', {
                    label: '<i class="icon icon-blur"></i>',
                    toggle_class: true,
                    name: 'Toggle gradient'
                }));
                background.addControl(getControl('bsbackground', 'bsboxshadow', {
                    label: '<i class="icon icon-Pin23"></i>'
                }));
                floating_sections.push(background);

                var addBorderIfNeeded = function (pgel) {
                    var border_classes = ['border', 'border-top', 'border-end', 'border-bottom', 'border-start'];
                    var has = false;
                    for (var i = 0; i < border_classes.length; i++) {
                        if (pgel.hasClass(border_classes[i])) {
                            has = true;
                            break;
                        }
                    }
                    if (!has) pgel.addClass('border');
                }

                var border = new PgFloatingToolsSection(prefix + 'border', 'Border');
                border.addControl(getControl('bsborders', 'bordercolor', {
                    on_before_set: addBorderIfNeeded,
                    label: '<i class="icon icon-Pin20"></i>',
                    buttons: false,
                    submenu: {
                        name: 'Border sides',
                        all_classes: ['border', 'border-top', 'border-end', 'border-bottom', 'border-start', 'border-0', 'border-top-0', 'border-end-0', 'border-bottom-0', 'border-start-0'],
                        combine_with_parent: true,
                        items: [
                            {key: 'border', name: 'All sides'},
                            {key: 'border-top', name: 'Top'},
                            {key: 'border-end', name: 'End'},
                            {key: 'border-bottom', name: 'Bottom'},
                            {key: 'border-start', name: 'Start'},
                            {key: 'border-bottom border-top', name: 'Top &amp; Bottom'},
                            {key: 'border-start border-end', name: 'Start &amp; End'},
                        ]
                    },
                    on_get_current_value_label: function(info, map) {
                        var all = f.bsGetBorderColorOptions('');
                        for(var i = 0; i < all.length; i++) {
                            if(info.classes.indexOf(all[i].key) >= 0) {
                                return all[i].html;
                            }
                        }
                        return this.options.label;
                    }
                }));
                border.addControl(getControl('bsborders', 'widthbotder', {
                    on_before_set: addBorderIfNeeded,
                    name: 'Border width',
                    label: `<svg xmlns="http://www.w3.org/2000/svg" width="1em" height="1em" fill="currentColor" class="bi bi-border-width" viewBox="0 0 16 16">
    <path d="M0 3.5A.5.5 0 0 1 .5 3h15a.5.5 0 0 1 .5.5v2a.5.5 0 0 1-.5.5H.5a.5.5 0 0 1-.5-.5v-2zm0 5A.5.5 0 0 1 .5 8h15a.5.5 0 0 1 .5.5v1a.5.5 0 0 1-.5.5H.5a.5.5 0 0 1-.5-.5v-1zm0 4a.5.5 0 0 1 .5-.5h15a.5.5 0 0 1 0 1H.5a.5.5 0 0 1-.5-.5z"/>
</svg>`
                }));
                var radius = getControl('bsborders', 'raduisborder', {
                    label: '<i class="icon icon-round_lu"></i>',
                    buttons: false
                })
                radius.def.options = [
                    {key: 'rounded', name: 'All sides'},
                    {key: 'rounded-0', name: 'All sides none'},
                    {key: 'rounded-1', name: 'All sides small'},
                    {key: 'rounded-2', name: 'All sides medium'},
                    {key: 'rounded-3', name: 'All sides large'},
                    {key: 'rounded-top', name: 'Top'},
                    {key: 'rounded-bottom', name: 'Bottom'},
                    {key: 'rounded-start', name: 'Left'},
                    {key: 'rounded-end', name: 'Right'},
                    {key: 'rounded-circle', name: 'Circle'},
                    {key: 'rounded-pill', name: 'Pill'}
                ]
                border.addControl(radius);
                floating_sections.push(border);

                var flexJustifyContentOptions = [
                    {
                        key: 'justify-content-' + 'start',
                        name: 'Start',
                        html: bm.makeIcon('223', {attributes: {'title': 'Start'}})
                    },
                    {
                        key: 'justify-content-' + 'end',
                        name: 'End',
                        html: bm.makeIcon('218', {attributes: {'title': 'End'}})
                    },
                    {
                        key: 'justify-content-' + 'center',
                        name: 'Center',
                        html: bm.makeIcon('219', {attributes: {'title': 'Center'}})
                    },
                    {
                        key: 'justify-content-' + 'between',
                        name: 'Between',
                        html: bm.makeIcon('206', {attributes: {'title': 'Between'}})
                    },
                    {
                        key: 'justify-content-' + 'around',
                        name: 'Around',
                        html: bm.makeIcon('205', {attributes: {'title': 'Around'}})
                    },
                    {
                        key: 'justify-content-' + 'evenly',
                        name: 'Evenly',
                        html: bm.makeIcon('205', {attributes: {'title': 'Evenly'}})
                    }
                ]

                var flexAlignItemsOptions = [
                    {
                        key: 'align-items-' + 'start',
                        name: 'Start',
                        html: bm.makeIcon('223', {attributes: {'title': 'Start'}})
                    },
                    {
                        key: 'align-items-' + 'end',
                        name: 'End',
                        html: bm.makeIcon('220', {attributes: {'title': 'End'}})
                    },
                    {
                        key: 'align-items-' + 'center',
                        name: 'Center',
                        html: bm.makeIcon('221', {attributes: {'title': 'Center'}})
                    },
                    {
                        key: 'align-items-' + 'baseline',
                        name: 'Baseline',
                        html: bm.makeIcon('212', {attributes: {'title': 'Baseline'}})
                    },
                    {
                        key: 'align-items-' + 'stretch',
                        name: 'Stretch',
                        html: bm.makeIcon('214', {attributes: {'title': 'Stretch'}})
                    }
                ]


                var onSetAlign = function(pgel, classes, row_prefix, col_prefix) {
                    var api = new PgApi();
                    //check if has flex
                    var flex = api.getFlexSettings(pgel);
                    if(flex.grid) {
                        pinegrow.showQuickError('The element uses CSS Grid to position its children. Use the grid properties for the layout.');
                        return false;
                    }
                    var add_classes = [];
                    var remove_classes = [];

                    if(!flex.flex) {
                        add_classes.push( flex.inline ? 'd-inline-flex' : 'd-flex');
                    }

                    var prefix = row_prefix;
                    var cls = classes[0];
                    if(flex.direction === 'column') {
                        //map
                        cls = cls.replace(prefix, col_prefix);
                        prefix = col_prefix;
                    }
                    if(prefix === 'align-items-') {
                        remove_classes = flexAlignItemsOptions.map(function (i) {
                            return i.key
                        });
                    } else {
                        remove_classes = flexJustifyContentOptions.map(function (i) {
                            return i.key
                        });
                    }

                    remove_classes.filter(function(cls) {
                        return api.hasClass(pgel, cls);
                    })

                    add_classes.push(cls);

                    api.removeClass(pgel, remove_classes);
                    api.addClass(pgel, add_classes);

                    return false;
                }

                var align = new PgFloatingToolsSection(prefix + 'align', 'Align content');

                align.addControl(new PgFloatingToolsControl({
                    key: prefix + 'align.h',
                    type: 'select',
                    name: 'Align children horizontally w. Flex',
                    action : 'apply_class',
                    show_empty : true,
                    toggle_buttons: true,
                    options: flexJustifyContentOptions
                }, {
                    label: `<i class="icon icon-219"></i>`,
                    on_before_set: function(pgel, classes) {
                        return onSetAlign(pgel, classes, 'justify-content-', 'align-items-');
                    }
                }));

                align.addControl(new PgFloatingToolsControl({
                    key: prefix + 'align.v',
                    type: 'select',
                    name: 'Align children vertically w. Flex',
                    action : 'apply_class',
                    show_empty : true,
                    toggle_buttons: true,
                    options: flexAlignItemsOptions
                }, {
                    label: `<i class="icon icon-221"></i>`,
                    on_before_set: function(pgel, classes) {
                        return onSetAlign(pgel, classes, 'align-items-', 'justify-content-');
                    }
                }));

                //floating_sections.push(align);


            }
            floating_sections.forEach(function(s) {
                sections.push(s);
            })
        }

        var extension_added = false;

        var plugin_builder = null;

        f.on_plugin_activated = function(page, done) {
            gh.onFrameworkActivatedOnPage(page);

            if(!plugin_builder) {
                plugin_builder = new PgPluginBuilder(f);
                plugin_builder.usePreviewIframe();
            }


            if(extension_added) {
                if(done) done(f);
                return;
            }

            if(typeof PgEditableAreaExtension !== 'undefined') {
                var ext = new PgEditableAreaExtension('bs5', 'Bootstrap', f);

                ext.addFieldFromProp('button-type', 'Button type', def_button.sections[prefix + 'button'].fields[prefix + 'buttontype'], '.btn');

                ext.addFieldFromProp('button-size', 'Button size', def_button.sections[prefix + 'button'].fields[prefix + 'buttonsize'], '.btn');

                pinegrow.addEditableAreaExtension(ext);
            }
            extension_added = true;
            //end add extension*/

            if(done) done(f);
        }

        f.on_set_inline_style = function(page, o) {
            o.css += gh.getViewCss();

            //modals
            o.css += '\
            .pg-show-modal:not(.show) {\
            position: relative;\
            top: auto;\
            right: auto;\
            left: auto;\
            bottom: auto;\
            display: block !important;\
            opacity: 1 !important;\
            }\
            .pg-show-modal:not(.show) .modal-dialog {\
            transform: translate(0, 1px) !important;\
            }';
        }


        f.on_page_changes_done = function(page, info, page_col) {
            gh.updateHelpersAfterChange(page, info, page_col);
        }

        f.on_inline_editing_done = function(page, pgel) {
            gh.updateHelpersAfterChange(page, null, new PgCollection(pgel));
        }

        f.on_page_selected = function(page) {
            if(page) {
                getBootstrapVersionForPage(page);
            }
        }

        f.on_before_properties_shown = function(page, panel, options) {
            if(page && !page.data.bootstrapVersion) {
                getBootstrapVersionForPage(page);
            }
            if(panel) {
                //options.show_section_nav_control = true; disabled for now
            }
        }

        /*
        $.each(f.component_types, function(k, def) {
            $.each(def.sections, function(sk, sec_def) {
                $.each(sec_def.fields, function(fn, fdef) {
                    console.log(fdef);
                })
            });
        })

        f.on_get_class_prop_options = function(page, info) {
            var ci = getClassInfo(info.class);

            info.done = true;
        }

         */

        var sections = [
            {name: 'Text', sections: ['tw.text'], icon: 'icon-Pin18'},
            {name: 'Margin, padding and dimension', sections: ['tw.spacing','tw.dimension'], icon: 'icon-Pin16'},
            {name: 'Display and position', sections: ['tw.display', 'tw.position'], icon: 'icon-Pin17'},
            {name: 'Background', sections: ['tw.background'], icon: 'icon-Pin21'},
            {name: 'Border', sections: ['tw.border'], icon: 'icon-Pin20'},
            {name: 'Flexbox', sections: ['tw.flexcontainer', 'tw.flexchild'], icon: 'icon-Pin22'},
            {name: 'Shadows &amp; Rings', sections: ['tw.shadow'], icon: 'icon-Pin23'},
            {name: 'Transform and transition', sections: ['tw.transform', 'tw.transition'], icon: 'icon-Pin24'},
            {name: 'Grid', sections: ['tw.grid'], icon: 'icon-grid'},
            {name: 'List', sections: ['tw.list'], icon: 'icon-Pin25'},
            {name: 'More (interactivity, accessibility, SVG, table)', icon: '<i class="icon icon-repeat-X" style="position: relative;top: 12px;"></i>', sub_sections: [
                    {name: 'Interactivity', sections: ['tw.interactivity']},
                    {name: 'Accessibility', sections: ['tw.accessibility']},
                    {name: 'SVG', sections: ['tw.svg']},
                    {name: 'Table', sections: ['tw.table']}
                ]},
        ]

        f.on_get_sections_for_prop_section_nav = function(page, list) {

            sections.forEach(function(s) {
                list.push(s);
            })
        }


        var design_provider = new PgDmDesignProviderForBootstrap5(f);

        pgDmStore.registerDesignProvider('bs5', design_provider);

        f.on_dm_detect_design_providers = function(page, list) {
            list.push('bs5');
        }

        if (!simple) {
            f.on_get_page_libs = function (page, list) {
                gh.getPageLibs([
                    {name: 'Components', url: 'index.html'}
                ]).forEach(function (item) {
                    list.push(item);
                })
            }

            f.on_page_menu = function (page, items) {
                items.push({
                    type: 'divider'
                    //header: 'Components'
                })

                items.push({
                    label: 'Customize &amp; update Bootstrap Theme',
                    action: function () {
                        pinegrow.getComponent('bootstrap5-theme-before', function (pc) {
                            var html = pc.html;
                            pinegrow.showAlert(html, 'Want to customize Bootstrap with SASS variables?', 'No, maybe later', 'Sure, let\'s do it!', null, function () {
                                customizeBootstrap();
                            });
                        });
                    }
                })
            }

            // css cdn replacer
            var onReplaceBootstrapCSSCDN = function (page, pgel) {
                var current_url = pgel.getAttribute('href');
                var info = page.getProjectInfo();
                if (info.getSourceStylesheetForCSSStylesheet(current_url)) {
                    //this stylesheet is compiled from sass/less source, don't replace it
                    return false;
                }
                return true;
            }

            // css replacer
            var onReplaceBootstrapCSS = function (page, pgel) {
                var current_url = pgel.getAttribute('href');
                var info = page.getProjectInfo();
                if (info.getSourceStylesheetForCSSStylesheet(current_url)) {
                    //this stylesheet is compiled from sass/less source, don't replace it
                    return false;
                }
                if (current_url === 'bootstrap/css/bootstrap.min.css' || current_url === 'bootstrap/css/bootstrap.css') {
                    return false;
                }
                return true;
            }

            var onReplaceBootstrapJS = function (page, pgel) {
                var current_url = pgel.getAttribute('src');
                if (current_url === 'bootstrap/js/bootstrap.min.js' || current_url === 'bootstrap/js/bootstrap.js') {
                    return false;
                }
                return true;
            }

            // Resources
            var res;

            res = gh.addResource('https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css', /bootstrap(\.min|)\.css/i);
            //determines if the current LINK href that matches detect re should be replaced
            res.replace = onReplaceBootstrapCSSCDN;

            gh.addResource('https://cdn.jsdelivr.net/npm/@popperjs/core@2.10.2/dist/umd/popper.min.js', /popper(\.min|)\.js/i);
            gh.addResource('https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.min.js', /bootstrap(\.min|)\.js/i);


            //gh.addResource('assets/js/jquery.min.js', /jquery[\-\.0-9]*(\.min|)\.js/i);
            gh.addResource('assets/js/popper.min.js', /popper(\.min|)\.js/i);
            res = gh.addResource('bootstrap/js/bootstrap.min.js', /bootstrap(\.min|)\.js/i);
            res.replace = onReplaceBootstrapJS;

            //res = gh.addResource('bootstrap/js/bootstrap.js', /bootstrap(\.min|)\.js/i);
            //res.replace = onReplaceBootstrapJS;

            res = gh.addResource('bootstrap/css/bootstrap.min.css', /bootstrap(\.min|)\.css/i);
            res.replace = onReplaceBootstrapCSS;

            //res = gh.addResource('bootstrap/css/bootstrap.css', /bootstrap(\.min|)\.css/i);
            //res.replace = onReplaceBootstrapCSS;

            f.on_after_resources_added = function () {
                pinegrow.getAllPages().forEach(function (page) {
                    getBootstrapVersionForPage(page, true /* force */);
                })
            }

            //Register starting page template
            var notRequiredFiles = ["pricing.css", "starter-template.css", "dashboard.css", "chartjs.css", "cheatsheet.css", "cheatsheet-rtl.css", "sign-in.css", "product.css", "blog.css", "blog-rtl.css", "grid.css", "navbar.css", "navbar-static.css", "form-validation.css", "carousel.css", "sticky-footer.css", "sticky-footer-navbar.css", "offcanvas.css", "cover.css"];
            var templatesOrder = ["index.html", "starter-template.html", "dashboard.html", "cheatsheet.html", "cheatsheet-rtl.html", "sign-in.html", "product.html", "album.html", "blog.html", "blog-rtl.html", "grid.html", "navbar.html", "pricing.html", "checkout.html", "checkout-rtl.html", "sticky-footer.html", "sticky-footer-navbar.html", "offcanvas.html", "carousel.html", "cover.html"];

            f.addTemplateProjectFromResourceFolder('template', null, 0, function (node) {
                var currentFilesName = notRequiredFiles.filter(function (fileName) {
                    return node.name == fileName;
                });
                if (currentFilesName && currentFilesName.length > 0) {
                    node.required = false;
                }

                var nodeIndex = templatesOrder.indexOf(node.name);
                if (nodeIndex >= 0) node.order = nodeIndex;
            });

            pinegrow.addEventHandler('on_project_loaded', function(pagenull, project) {
                if (!isApp()) {
                    //add resources to url map
                    project.addSourceUrlToDOMUrlMap('assets/js/popper.min.js', f.getResourceUrl('template/assets/js/popper.min.js'));
                    project.addSourceUrlToDOMUrlMap('bootstrap/js/bootstrap.min.js', f.getResourceUrl('template/bootstrap/js/bootstrap.min.js'));
                    project.addSourceUrlToDOMUrlMap('bootstrap/css/bootstrap.min.css', f.getResourceUrl('template/bootstrap/css/bootstrap.min.css'));
                    return;
                }
            });

            //Bootstrap theme
            var customizeBootstrap = function () {
                var project = pinegrow.getCurrentProject();
                var page = pinegrow.getSelectedPage();

                if (project) {
                    var res = new PgComponentTypeResource(f.getResourceUrl('bootstrap_theme'));
                    res.isFolder = true;
                    res.relative_url = 'bootstrap_theme';
                    res.source = res.url;

                    var custom_file = crsaMakeFileFromUrl(project.getUrl() + '/bootstrap_theme/custom.scss');

                    var doTask = function (update) {

                        var has_local_bs = true;

                        if(!update) {

                            page.stylesheetsThatContain(/v5/i, /bootstrap/i, function (list) {
                                var cs = null;
                                for (var i = 0; i < list.length; i++) {
                                    if (list[i].localFile && !list[i].mappedRemote) {
                                        cs = list[i];
                                        break;
                                    }
                                }
                                if (!cs) {
                                    pinegrow.showAlert('<p>Couldn\'t detect a local Bootstrap 5 stylesheet on the page.</p><p>Customizing the Bootstrap theme doesn\'t work if bootstrap.css is included remotly, from a CDN.</p><p>To proceed, include a local copy of bootstrap.css to your page and repeat the Customize Bootstrap Theme command.</p>', 'Local bootstrap.css was not found on the page');

                                    has_local_bs = false;
                                }
                            })
                        }

                        if (!has_local_bs) return;

                        res.copyFilesToProject(project, function () {

                            pinegrow.refreshCurrentProjectDelayed(true);

                            if (update) {
                                pinegrow.showAlert('<p>Bootstrap theme was updated to the latest Bootstrap version.</p><p>Please note that this did not update any Bootstrap JS files. Download them from the official Bootstrap website and copy JS files into the project.</p>', 'Bootstrap theme updated');
                                return;
                            }
                            //find bs cs
                            page.stylesheetsThatContain(/v5/i, /bootstrap/i, function (list) {
                                var cs = null;
                                for (var i = 0; i < list.length; i++) {
                                    if (list[i].localFile && !list[i].mappedRemote) {
                                        cs = list[i];
                                        break;
                                    }
                                }
                                if (!cs) {
                                    //should not happen
                                } else {
                                    //set scss source for bs cs
                                    var url = project.getUrl() + '/bootstrap_theme/custom.scss';

                                    var newStylesheet = pinegrow.stylesheets.getStylesheetByKey(url);
                                    if (!newStylesheet) {
                                        cs.ignore = false;

                                        newStylesheet = PgSCSSStylesheet();
                                        newStylesheet.load(url, function () {
                                            pinegrow.dispatchEvent('on_stylesheet_closed', null, cs);
                                            cs.setSourceStylesheet(newStylesheet);
                                            $('body').trigger('crsa-stylesheets-changed');

                                            pgcssh.setUrlOfStylesheetForOverrides(url, null, '-vars')
                                            pinegrow.refreshCurrentProject();

                                            pinegrow.showQuickMessage('Done');

                                            pinegrow.getComponent('bootstrap5-theme', function (pc) {
                                                var html = pc.html;
                                                pinegrow.showAlert(html, 'Customize your Bootstrap theme');
                                            });

                                            page.refresh();
                                        }, false, false, cs);
                                    }

                                    pinegrow.stylesheets.stylsheetUsed({
                                        url: url,
                                        pgStylesheet: newStylesheet
                                    }, 'source', cs);
                                }
                            });

                        }, function (rel_url) {
                            //console.log('oe ' + rel_url);
                            if (rel_url.indexOf('custom.scss') >= 0) return false;
                            return true;
                        } /*  overwrite_existing callback */)
                    }

                    if (crsaIsFileExist(custom_file)) {
                        pinegrow.showAlert(`<p>Custom Bootstrap theme already exists in the <code>bootstrap_theme</code> folder.</p><p>Do you want to update it to the latest Bootstrap version? The file <code>custom.scss</code> that contains your custom settings will not be overwritten.</p>`, 'Customized theme already exists - Want to update?', 'Cancel', 'Update', null, function () {

                            pinegrow.showAlert('<p>We need to close all open pages first. You will be asked to save changes.</p>', 'Close open pages?', 'Cancel', 'Close pages &amp; update', null, function () {

                                pinegrow.closeAllPages(function (closed) {
                                    if (closed) {
                                        doTask(true);
                                    }
                                })
                            });
                        });
                        return;
                    } else {
                        doTask();
                    }

                } else {
                    pinegrow.showAlert('<p>Customizing Bootstrap theme works with projects.</p><p>Open your page as a project first.</p>', 'Open the page as a project');
                }
            }
        }
    });
});
